/*! \file dbintd.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: dbintd.c,v 1.5 2001/02/23 19:02:01 stevek Exp $
 *
 * \brief   Database inteface daemon
 * \author  Steve Kiernan
 * \version $Revision: 1.5 $
 * \date    $Date: 2001/02/23 19:02:01 $
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/syslog.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include <wss_db.h>
#include <wss_vm.h>

#include "dbintd.h"


struct sock_node
{
   int                socket;
   struct sock_node  *next;
   struct sock_node **prev;
};

/* Private globals */
static int               version   = 0;
static int               notLoaded = FALSE;

static int               sock;
static int               dbsize;

static struct sock_node *socketList;

/* Public globals */
int                      call;

/* Externals */
extern wss_char_t       *wss_db_filename;

#ifdef DBINTD_THREADED
static void *WorkThread      __P(( void * ));
#endif /* DBINTD_THREADED */
static void  CheckSigHandler __P(( int ));
static void  ExitHandler     __P(( void ));
extern int   DoCommand       __P(( int, int ));

int
main( int argc, char *argv[] )
{
   struct sigaction   sa;
   struct sigaction   osa;
   struct sockaddr_un un;
   mode_t             um;
   int                nextsock;
   int                retc;
   fd_set             readset;

#ifndef DEBUG_DBINT
   if( ( getuid() != 0 ) && ( geteuid() != 0 )  && ( getppid() != 1 ) )
   {
      fprintf( stderr, "dbintd: Must be run by the kernel module\n" );
      return (EPERM);
   }
#endif /* DEBUG_DBINT */

   /* Make sure we have 3 arguments to the binary (plus the binary path) */
   if( argc != 4 )
   {
      fprintf( stderr, "dbintd: Invalid number of arguments\n" );
      return (EINVAL);
   }

   /* Set variables from arguments */
   wss_db_filename = argv[1];
   dbsize = atoi(argv[2]);
   call = atoi(argv[3]);

   /* Try calling the syscall for a test */
   sa.sa_handler = CheckSigHandler;
   sigemptyset( &sa.sa_mask );
   sa.sa_flags = 0;
   sigaction( SIGSYS, &sa, &osa );
   syscall( call, DBINT_GET_VERSION, &version );
   sigaction( SIGSYS, &osa, NULL );

   /* Exit if module is not loaded */
   if( notLoaded )
   {
      fprintf( stderr, "dbintd: Module not loaded\n" );
      return (EINVAL);
   }

   /* Exit if interface versions mismatched */
   if( version != DBINT_VERSION )
   {
      fprintf( stderr, "dbintd: Mismatched versions (%d != %d)\n",
               version, DBINT_VERSION );
      return (EINVAL);
   }

   /* Trap SIGPIPIE and ignore */
   sa.sa_handler = SIG_IGN;
   sigemptyset( &sa.sa_mask );
   sa.sa_flags = 0;
   sigaction( SIGPIPE, &sa, NULL );

   /* Create UNIX domain socket */
   un.sun_family = AF_UNIX;
   if( sizeof(_PATH_DBINTSOCK) >= sizeof(un.sun_path) )
   {
      fprintf( stderr, "dbintd: Socket name too long\n" );
      return (EINVAL);
   }
   strcpy( un.sun_path, _PATH_DBINTSOCK );
   un.sun_len = strlen( un.sun_path );

   if( ( sock = socket( AF_UNIX, SOCK_STREAM, 0 ) ) < 0 )
   {
      perror( "socket" );
      return (errno);
   }

   /* Set atexit to clean up on exit */
   atexit( ExitHandler );

   /* Bind socket into filesystem */
   um = umask(000);
   unlink( un.sun_path );
   if( bind( sock, (struct sockaddr *) &un, sizeof(un) ) < 0 )
   {
      perror( "bind" );
      return (errno);
   }
   (void) umask( um );

   /* Listen to socket for connections (5 lookahead) */
   (void) listen( sock, 5 );

   /* Map database into address space */
   if( wss_vm_init( dbsize ) )
      wss_db_init( FALSE );

   /* Start up logging (and change name) */
   openlog( "dbintd", LOG_CONS|LOG_PID, LOG_DAEMON );

   FD_ZERO( &readset );
   FD_SET( sock, &readset );

   nextsock = sock + 1;
   for( ;; )
   {
      struct sockaddr_un  un2;
      int                 len2;
      fd_set              fdset;
      int                 so2;
      struct sock_node   *node;
      struct sock_node   *next;

      len2 = sizeof( un2 );

      /* Copy read set */
      memcpy( &fdset, &readset, sizeof(fdset) );

      /* Select for read on socket */
      retc = select( nextsock, &fdset, (fd_set *) NULL, (fd_set *) NULL,
                     (struct timeval *) NULL );
      if( retc < 0 )
      {
         if( errno == EINTR )
            continue;

         syslog( LOG_ERR, "select: %s", strerror( errno ) );
         return (errno);
      }

      if( FD_ISSET( sock, &fdset ) )
      {
         /* Accept a new connection */
         so2 = accept( sock, (struct sockaddr *) &un2, &len2 );
         if( so2 < 0 )
         {
            if( errno != EINTR )
            {
               syslog( LOG_ERR, "accept: %s", strerror( errno ) );
               return (errno);
            }
         }
         else
         {
            /* Save connection information */
            syslog( LOG_INFO, "accepted connection (socket %d)", so2 );

            node = (struct sock_node *)malloc( sizeof( *node ) );
            node->next = socketList;
            node->prev = &socketList;
            socketList = node;
            node->socket = so2;
            if( so2 >= nextsock )
               nextsock = so2 + 1;

            FD_SET( so2, &readset );
         }
      }

      for( node = socketList; node; node = next )
      {
         next = node->next;

         if( FD_ISSET( node->socket, &fdset ) )
         {
            db_command_t cmd;

            syslog( LOG_INFO, "reading from socket %d", node->socket );
            retc = read( node->socket, &cmd, sizeof(cmd) );
            if( retc == 0 )
            {
               /* Hit end-of-file */
               retc = -1;
               errno = EPIPE;
            }
            else if( retc > 0 )
               retc = DoCommand( cmd.cmd, node->socket );

            if( ( retc < 0 ) && ( errno != EINTR ) )
            {
               syslog( LOG_INFO, "closing socket %d", node->socket );
               *(node->prev) = node->next;
               FD_CLR( node->socket, &readset );
               close( node->socket );
               free( node );
            }
         }
      }
   }

   return (0);
}

/* SIGSYS signal handler */
static void
CheckSigHandler( int sig )
{
   notLoaded = TRUE;
}

/* Exit handler */
static void
ExitHandler()
{
   close( sock );
   wss_vm_shutdown();
   unlink( _PATH_DBINTSOCK );
}

