/*! \file dbintd_cmds.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: dbintd_cmds.c,v 1.7 2001/02/23 19:02:01 stevek Exp $
 *
 * \brief   Database inteface daemon command processing routines
 * \author  Steve Kiernan
 * \version $Revision: 1.7 $
 * \date    $Date: 2001/02/23 19:02:01 $
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include <sys/syslog.h>

#include <wrerrno.h>
#include <wss_db.h>
#include <wss_vm.h>

#include <libwrv.h>

#include "dbintd.h"
#include "dbinterface.h"

int
DoCommand( int cmd, int sock )
{
   int retc  = 0;
   int error = 0;

   syslog( LOG_INFO, "processing command %d from socket %d", cmd, sock );

   switch( cmd )
   {
      case DBOP_VERSION:
      {
         /* Get interface version */
         int pktsize;
         int vers;

         pktsize = sizeof(vers);
         vers = DBINT_VERSION;

         if( ( retc = write( sock, &pktsize, sizeof(pktsize) ) ) < 0 )
            break;

         retc = write( sock, &vers, sizeof(vers) );
         break;
      }

      case DBOP_SELECT:
         break;

      case DBOP_INSERT:
         break;

      case DBOP_UPDATE:
         break;

      case DBOP_DELETE:
         break;

      case DBOP_HANDLE:
      {
         /* Get table handle by name */
         TCHAR   name[MAXNAMELEN];
         int     pktsize;
         wqtab_t tab;

         if( ( retc = read( sock, &pktsize, sizeof(pktsize) ) ) < 0 )
            return (retc);

         if( pktsize >= sizeof(name) )
         {
            error = ENAMETOOLONG;
            break;
         }

         if( ( retc = read( sock, name, pktsize ) ) < 0 )
            return (retc);

         name[pktsize] = WSS_MAKESTR('\0');

         tab = wss_db_find_table_by_name( name );
         if( tab == NULL )
            error = ENOENT;
         else
         {
            u_long id = (u_long) abs2rel(tab);

            wss_db_unlock_table(tab);

#ifdef DEBUG_DBINT
            syslog( LOG_INFO, "DBOP_HANDLE: returning 0x%lx for '%s'",
                    id, name );
#endif /* DEBUG_DBINT */

            pktsize = sizeof(id);

            /* Write out packet size */
            if( ( retc = write( sock, &pktsize, sizeof(pktsize) ) ) < 0 )
               return (retc);

            /* Write out table id */
            if( ( retc = write( sock, &id, sizeof(id) ) ) < 0 )
               return (retc);
         }

         break;
      }

      case DBOP_COLDEFS:
      {
         /* Get column definitions */
         wrv_h_t data;
         wqtab_t tab;
         wqcd_t  col_defs;
         int     pktsize;

         if( ( retc = read( sock, &pktsize, sizeof(pktsize) ) ) < 0 )
            return (retc);

         if( pktsize != sizeof(data) )
            error = EINVAL;
         else
         {
            if( ( retc = read( sock, &data, sizeof(data) ) ) < 0 )
            return (retc);

            if( ( tab = wss_db_find_table_by_handle( data.handle ) ) == NULL )
               error = ENOENT;
            else
            {
               wqcd col;
               int  i;

               pktsize = sizeof(tab->col_count);
               col_defs = (wqcd_t)( ( tab->col_defs )?rel2abs( tab->col_defs )
                                                     :NULL );

               for( i = 0; i < tab->col_count; i++ )
               {
                  wqcd_t cd = col_defs + i;
                  wss_char_t *name = rel2abs( cd->name );
                  size_t namelen = ( _tcslen( name ) + 1 ) * sizeof(wss_char_t);

                  pktsize += sizeof(*cd) + namelen;
               }

               /* Write out packet size */
               if( ( retc = write( sock, &pktsize, sizeof(pktsize) ) ) < 0 )
               {
                  wss_db_unlock_table( tab );
                  return (retc);
               }

               /* Write out number of column definitions */
               if( ( retc = write( sock, &(tab->col_count),
                                   sizeof(tab->col_count) ) ) < 0 )
               {
                  wss_db_unlock_table( tab );
                  return (retc);
               }

               for( i = 0; i < tab->col_count; i++ )
               {
                  wqcd_t cd = col_defs + i;
                  wss_char_t *name = rel2abs( cd->name );
                  size_t namelen = ( _tcslen( name ) + 1 ) * sizeof(wss_char_t);

                  /* Copy column definition structure */
                  memcpy( &col, cd, sizeof(*cd) );

                  /* Set name length in bytes in place of name pointer */
                  col.name = (wss_char_t *) namelen;

                  /* Write out structure */
                  if( ( retc = write( sock, &col, sizeof(col) ) ) < 0 )
                  {
                     wss_db_unlock_table( tab );
                     return (retc);
                  }

                  /* Write out name */
                  if( ( retc = write( sock, name, namelen ) ) < 0 )
                  {
                     wss_db_unlock_table( tab );
                     return (retc);
                  }
               }

               wss_db_unlock_table( tab );
            }
         }

         break;
      }

      case DBOP_NEWROW:
      {
         /* Insert empty row */
         int pktsize;
         int error;
         wrv_h_t data;
 
         if( ( retc = read( sock, &data, sizeof(data) ) ) < 0 )
            break;

         error = wss_db_insert_nulls( data.handle );
         switch( error )
         {
            case WRE_NOMEM:
               error = ENOMEM;
               break;

            case WRE_NOTFOUND:
               error = ENOENT;
               break;

            case WRE_INVALID_PARAMETER:
               error = EINVAL;
               break;

            default:
               break;
         }

         if( error )
            pktsize = -1;
         else
            pktsize = 0;

         if( ( retc = write( sock, &pktsize, sizeof(pktsize) ) ) < 0 )
            break;

         if( error )
            retc = write( sock, &error, sizeof(error) );
         break;
      }

      default:
         break;
   }

   /* If error, send -1 and error code */
   if( error )
   {
      int code = -1;

      if( ( retc = write( sock, &code, sizeof(code) ) ) < 0 )
         return (retc);

      if( ( retc = write( sock, &error, sizeof(error) ) ) < 0 )
         return (retc);
   }

   return (retc);
}

