/*! \file dbintd_subr.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: dbintd_subr.c,v 1.6 2001/02/23 19:02:01 stevek Exp $
 *
 * \brief   Subroutines for supporting WSS/WSI database functionality
 * \author  Steve Kiernan
 * \version $Revision: 1.6 $
 * \date    $Date: 2001/02/23 19:02:01 $
 */

#include <wr_conf.h>
#include <stdarg.h>
#include <wsi_stdlib.h>

#include <stdio.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WR_WINNT
#include <sys/syslog.h>
#include <sys/mman.h>
#include <unistd.h>
#else
#include <io.h>
#endif

#include <wss_db.h>
#include <wss_vm.h>
#include <wss_wrvevent.h>

#include <dbintd.h>


u_long wss_config_db_handle;

vm_offset_t wsi_vm_init(size_t size);
void panic( const char *, ... );

extern wss_char_t *wss_db_filename;
extern int         call;

#ifdef WR_WINNT
#define stat  _stat

wss_char_t *
wsi_strdup( const wss_char_t *str )
{
   return _tcsdup( str );
}
#else
__inline__ wss_char_t *
wsi_strdup( const wss_char_t *str )
{
   return strdup( str );
}
#endif /* WR_WINNT */

wss_char_t *
wss_strdup( const wss_char_t *str )
{
   return wsi_strdup( str );
}

void
wsi_vm_shutdown()
{
   _tprintf( MAKESTR("VM Shutting down...\n") );
#ifdef WR_WINNT
   if( !UnmapViewOfFile( wss_vm_data ) )
#else
   if( munmap( (caddr_t)wss_vm_data, wss_vm_size ) < 0 )
#endif
      perror( "munmap" );
}

vm_offset_t
wsi_vm_init( size_t size )
{
#ifdef WR_WINNT
   HANDLE       hFile;
   DWORD        nBytes;
#else
   int          fd;
#endif /* WR_WINNT */
   struct stat  sb;
   off_t        offset;
   size_t       size_in_bytes = size * 1048576;

   if( stat( wss_db_filename, &sb ) )
   {
#ifdef WR_WINNT
      if( ( hFile = CreateFile( wss_db_filename, GENERIC_READ | GENERIC_WRITE,
                                0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL,
                                NULL ) ) == INVALID_HANDLE_VALUE )
#else
      if( ( fd = open( wss_db_filename, O_RDWR | O_CREAT, 0644 ) ) < 0 )
#endif /* WR_WINNT */
      {
         perror( "open" );
         return 0;
      }

#ifdef WR_WINNT
      offset = SetFilePointer( hFile, size_in_bytes - 1, NULL, FILE_BEGIN );
#else
      offset = lseek( fd, size_in_bytes - 1, SEEK_SET );
#endif /* WR_WINNT */
      wss_vm_size = size_in_bytes;

#ifdef WR_WINNT
      WriteFile( hFile, "x", 1, &nBytes, NULL );
#else
      write( fd, "x", 1 );
#endif /* WR_WINNT */
   }
   else
   {
#ifdef WR_WINNT
      if( ( hFile = CreateFile( wss_db_filename, GENERIC_READ | GENERIC_WRITE,
                                0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
                                NULL ) ) == INVALID_HANDLE_VALUE )
#else
      if( ( fd = open( wss_db_filename, O_RDWR ) ) < 0 )
#endif /* WR_WINNT */
      {
         perror( "open" );
         return 0;
      }
      wss_vm_size = sb.st_size;
      if( wss_vm_size != size_in_bytes )
         return 0;
   }

#ifdef WR_WINNT
   wss_vm_data = MapViewOfFile( CreateFileMapping( hFile, NULL, PAGE_READWRITE,
                                                   0, wss_vm_size, NULL ),
                                FILE_MAP_READ | FILE_MAP_WRITE, 0, 0,
                                wss_vm_size );
#else
   wss_vm_data = (void *)mmap( (caddr_t) 0, wss_vm_size,
                               (PROT_READ|PROT_WRITE), MAP_SHARED, fd, 0 );
#endif /* WR_WINNT */

#ifdef WR_WINNT
   if( wss_vm_data == NULL )
#else
   if( (int)wss_vm_data == (int)MAP_FAILED )
#endif /* WR_WINNT */
   {
      _tprintf( MAKESTR("map failed!\n") );
      return 0;
   }

   return( (vm_offset_t)wss_vm_data );
}

void
panic( const char *msg, ... )
{
   va_list ap;
   va_start( ap, msg );

   fprintf( stderr, "panic: " );
   vfprintf( stderr, msg, ap );
   fprintf( stderr, "\n" );

   abort();
}

void *
wsi_malloc( size_t size )
{
   return( malloc( size ) );
}

void
wsi_free( void *ptr )
{
   free( ptr );
}

void
wss_vm_mark( void *blk )
{
   /* noop */
}

void
wss_vm_sweep()
{
   /* noop */
}

void
wss_db_lock()
{
   int flag = DBF_LOCK;
#ifdef DEBUG_DBINT
   syslog( LOG_INFO, "Calling wss_db_lock via syscall\n" );
#endif /* DEBUG_DBINT */
   syscall( call, DBINT_LOCK, &flag );
}

void
wss_db_unlock()
{
   int flag = DBF_UNLOCK;
#ifdef DEBUG_DBINT
   syslog( LOG_INFO, "Calling wss_db_unlock via syscall\n" );
#endif /* DEBUG_DBINT */
   syscall( call, DBINT_LOCK, &flag );
}

void
wss_db_lock_table( wqtab_t tab )
{
   struct table_lock_args
   {
      int flag;
      u_long id;
   }
   args;

   args.flag = DBF_LOCK;
   args.id = (u_long)abs2rel(tab);

#ifdef DEBUG_DBINT
   syslog( LOG_INFO, "Calling wss_db_lock_table(0x%lx) via syscall\n",
           args.id );
#endif /* DEBUG_DBINT */
   syscall( call, DBINT_LOCK_TABLE, &args );
}

void
wss_db_unlock_table( wqtab_t tab )
{
   struct table_lock_args
   {
      int flag;
      u_long id;
   }
   args;

   args.flag = DBF_UNLOCK;
   args.id = (u_long)abs2rel(tab);

#ifdef DEBUG_DBINT
   syslog( LOG_INFO, "Calling wss_db_unlock_table(0x%lx) via syscall\n",
           args.id );
#endif /* DEBUG_DBINT */
   syscall( call, DBINT_LOCK_TABLE, &args );
}

void
wss_wrvevent_post( wrv_event_sig_t signature, u_long key, u_long extra,
                   void *buffer, u_long length )
{
   struct event_args
   {
      wrv_event_sig_t  signature;
      u_long           key;
      u_long           extra;
      void            *buffer;
      u_long           length;
   }
   args;

   args.signature = signature;
   args.key = key;
   args.extra = extra;
   args.buffer = buffer;
   args.length = length;

   syscall( call, DBINT_POST_EVENT, &args );
}

