/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: dbintd_vm.c,v 1.5 2001/02/23 19:02:01 stevek Exp $
 *
 * Authors: Lee Badger, Chris Vance
 *
 * This file contains functions for managing VM used by the WSS.
 * Currently, it creates a wss.db file in the current directory
 * and maps it to kernel space for use by the DBMS functions.
 *      Process-specific allocation routines are provided to allow
 *      parameter replacement (memory needs to be in user-space).
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>

#include "wr_conf.h"
#include "wsi_stdlib.h"
#include "wss_db.h"
#include "wss_vm.h"
#include "wss_malloc.h"


/***************************************************************************\
 * Private structures and types                                            *
\***************************************************************************/

/*
 * The following mem malloc/free structures adapted from "The C
 * Programming Language" by Kernighan and Ritchie, chapter 8.
 */
typedef double ALIGN;
typedef union header
{
   struct
   {
      union header *next;  /* next free block */
      unsigned      size;  /* size of this free block, in units */
   }
   s;
   ALIGN            x;     /* force alignment of blocks */
}
HEADER;

/*
 * The first word of the vm region points to the root table.  The next
 * structure is for wss_vm_malloc/wss_vm_free.
 */
struct wss_vm_header
{
   struct wql_table *root_table; /* relative address of first table  */
   u_long            free_bytes;
   HEADER           *trace;
   HEADER           *base;       /* base is now the first aligned
                                    block after the vm header */
};


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

struct wql_table **wss_vm_root_table;

void   *wss_vm_data;
size_t  wss_vm_size;

volatile HEADER   *save = NULL;

/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static short                 mark_and_sweep = 1;          
static struct wss_vm_header *wss_vm_head    = NULL;
static HEADER               *allocp         = NULL; /* last allocated block,
                                                       an absolute ptr */
static int                   header_size    = 0;


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

int psize __P((void *mem));

#ifdef DEBUG_VM
static int _wss_vm_aligned_ptr   __P(( void *rel, char *file, int line ));
static int _wss_vm_valid_ptr     __P(( void *rel, char *file, int line ));
static int _wss_vm_valid_abs_ptr __P(( void *abs, char *file, int line ));
#endif /* DEBUG_VM */


/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/

extern vm_offset_t wsi_vm_init     __P(( size_t size ));
extern void        wsi_vm_shutdown __P(( void ));
extern void        dbck_panic      __P(( const char *msg, ... ));


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#ifdef DEBUG_VM
#   define WR_VM_ALIGNED_PTR(rel)   _wss_vm_aligned_ptr(rel,__FILE__,__LINE__)
#   define WR_VM_VALID_PTR(rel)     _wss_vm_valid_ptr(rel,__FILE__,__LINE__)
#   define WR_VM_VALID_ABS_PTR(abs) _wss_vm_valid_abs_ptr(abs,__FILE__,__LINE__)
#else
#   define WR_VM_ALIGNED_PTR(rel)   ((((unsigned)rel)%sizeof(HEADER))==0)
#   define WR_VM_VALID_PTR(rel)     ((rel==0) || ((((vm_size_t)rel)<wss_vm_size ) && WR_VM_ALIGNED_PTR(rel)))
#   define WR_VM_VALID_ABS_PTR(abs) (((char*)abs>(char*)wss_vm_head) && WR_VM_VALID_PTR(abs2rel(abs)))
#endif /* !DEBUG_VM */

#define WR_VM_MARKED    (1<<(sizeof(unsigned)*8-1))
#define WR_VM_SIZE_MASK ~WR_VM_MARKED
#define WR_VM_SIZE(blk) (mark_and_sweep?(blk)->s.size&WR_VM_SIZE_MASK:(blk)->s.size)


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

#ifdef DEBUG_VM
static int __inline__
_wss_vm_aligned_ptr( void *rel, char *file, int line )
{
   if( ( ((unsigned)rel) % sizeof(HEADER) ) != 0 )
   {
      printf( "_wss_vm_aligned_ptr: unaligned pointer ==> %p (%s:%d)\n", rel,
              file, line );
      return 0;
   }
   else
      return 1;
}

static int __inline__
_wss_vm_valid_ptr( void *rel, char *file, int line )
{
   if( _wss_vm_aligned_ptr(rel, file, line) )
   {
      if( (rel == NULL) || (((vm_size_t)rel) < wss_vm_size) )
         return 1;
      else
      {
         printf( "_wss_vm_valid_ptr: relative pointer out of range ==> %p "
                 "(%s:%d)\n", rel, file, line );
         return 0;
      }
   }
   else
      return 0;
}

static int __inline__
_wss_vm_valid_abs_ptr( void *abs, char *file, int line )
{
   if( (char *)abs > (char *)wss_vm_head )
      return( _wss_vm_valid_ptr( abs2rel(abs), file, line ) );
   else
   {
      printf( "_wss_vm_valid_abs_ptr: absolute pointer out of range ==> "
              "%p <= %p (%s:%d)\n", abs, wss_vm_head, file, line );
      return 0;
   }
}
#endif /* DEBUG_VM */



/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*
 * VM management initialization function
 */
int
wss_vm_init( size_t size )
{
   vm_offset_t offset;
   size_t      size_in_bytes = size * 1024 * 1024;
    
   if( ( size_in_bytes / sizeof(HEADER) ) > WR_VM_MARKED )
   {
      printf("WSS database too large to allow mark-and-sweep operations.\n");
      mark_and_sweep = 0;
   }
    
   /*
    * Request the specified size of mapped memory
    */
   if( ( offset = wsi_vm_init(size) ) )
   {
      int vmh_size = sizeof(struct wss_vm_header);
      int vmb_size = sizeof(HEADER);

      wss_vm_head = (struct wss_vm_header*) offset;
      wss_vm_root_table = &wss_vm_head->root_table;

      /* Determine size of header aligned to block size */
      header_size = (vmh_size + vmb_size) & ~(vmb_size - 1);

#ifdef DEBUG_VM
      printf( "wss_vm_init: header size = %d bytes\n", header_size );
      printf( "wss_vm_init: %d blocks @ %d bytes = %d bytes\n",
              size_in_bytes / vmb_size, vmb_size, size_in_bytes );
#endif /* DEBUG_VM */
   }

   return offset;
}

/*
 * VM management shutdown function
 */
void
wss_vm_shutdown()
{
   wsi_vm_shutdown();
}

/*
 * For debugging; prints the size of (absolute) memory gotten from
 * wss_vm_malloc()
 */
int
psize( void *mem )
{
   HEADER *m;

   m = (HEADER*)mem - 1;
   return(WR_VM_SIZE(m) * sizeof(HEADER));
}


/*
 * Walk trace list and return 1 if we find the specified block, otherwise return 0
 */
int
wss_vm_valid_alloced( void *rel )
{
   register HEADER *mem = (HEADER *)rel - 1;
   
   if( WR_VM_VALID_PTR(rel) && ( wss_vm_head->trace != NULL ) )
   {
      register HEADER *trace = wss_vm_head->trace;
      register HEADER *q;
      for( q = (HEADER *)rel2abs(trace); q;
           q = (HEADER*)((q->s.next)?rel2abs(q->s.next):NULL) )
      {
#ifdef DEBUG_VM
         if( !WR_VM_VALID_ABS_PTR(q) )
         {
            printf( "wr_wm_valid_alloced: ptr out of range in trace ==> %p\n", q );
            dbck_panic( "wss_vm_valid_alloced" );
         }
#endif /* DEBUG_VM */
         if( q->s.next == mem )
            break;
      }

      if( q == NULL )
      {
         if( mem == trace )
            return 1;
         else
            return 0;
      }

      return 1;
   }
   else
      return 0;
}

/*
 * Check for valid null-terminated string
 * String length must be less than size of block minus the header
 */
int
wss_vm_valid_str( const char *rel )
{
   HEADER *blk = (HEADER *)rel2abs(rel) - 1;
   int i;
   char *ptr;

   if( !wss_vm_valid_alloced((void *)rel) ) return 0;

   for( i = 0, ptr = (char *)(blk + 1);
        *ptr && ( i < (WR_VM_SIZE(blk) - 1) * sizeof(HEADER) );
        i++, ptr++ );

   if( *ptr )
      return 0;
   else
      return 1;
}


/*
 * The following mem malloc/free code adapted from "The C Programming
 * Language" by Kernighan and Ritchie, chapter 8.
 */


/*
 * Like malloc(3), but allocates (absolute ptrs) from the VM region.
 * This routine maintains all internal pointers as relative addresses
 * within the VM region.  This means that memory allocated by DB
 * structures in the VM region will be preserved across runs of the
 * WSS.
 *
 * Returns NULL on failure (i.e., if no memory was available).
 */
#ifdef DEBUG_VM
void *
wss_vm_malloc_debug( unsigned nbytes, const char *file, int line )
#else
void *
wss_vm_malloc( unsigned nbytes )
#endif /* DEBUG_VM */
{
   register HEADER *p, *q;
   register int nunits;
   register HEADER *base = NULL;

#ifdef DEBUG_VM
   printf( "wss_vm_malloc: Allocating %d bytes (%s:%d)\n", nbytes, file, line );
#endif /* DEBUG_VM */

   /* compute the number of needed units */
#ifdef DEBUG_VM
   nunits = 2 + ( (nbytes + sizeof(HEADER) - 1) / sizeof(HEADER) );
#else
   nunits = 1 + ( (nbytes + sizeof(HEADER) - 1) / sizeof(HEADER) );
#endif

   /* 
    * If this is the first malloc/free for this run of the wss,
    * initialize the header.
    */
   if( ( q = allocp ) == NULL )
   {
      base = (HEADER *)rel2abs(wss_vm_head->base);
#ifdef DEBUG_VM
      printf( "wss_vm_malloc: Setting allocp to %p\n", base );
#endif /* DEBUG_VM */
      if( wss_vm_head->base->s.next == NULL )
      {
         /* new wss.db file */
         base->s.next = (HEADER*)abs2rel(base);
         base->s.size = ( wss_vm_size - header_size ) / sizeof(HEADER);
         wss_vm_head->trace = NULL;
         wss_vm_head->free_bytes = base->s.size * sizeof(HEADER);
      }
      allocp = q = base;
#ifdef DEBUG_VM
      if( allocp == NULL ) printf( "wss_vm_malloc: allocp set null\n" );
#endif /* DEBUG_VM */
   }

#ifdef DEBUG_VM
   if( !WR_VM_VALID_ABS_PTR(allocp) )
   {
      printf( "wss_vm_malloc: allocp out of range ==> %p\n", allocp );
      dbck_panic( "wss_vm_malloc" );
   }
   printf( "wss_vm_malloc: q = %p(%p), allocp = %p(%p), base = %p(%p)\n", q,
           abs2rel(q), allocp, abs2rel(allocp), base,
           (void *)((base!=NULL)?(abs2rel(base)):(void *)base) );
#endif /* DEBUG_VM */

   /* Find a free block: use first one big enough. */
   for( p=(HEADER*)rel2abs(q->s.next); ; q=p, p=(HEADER*)rel2abs(p->s.next) )
   {
#ifdef DEBUG_VM
      printf( "wss_vm_malloc: Checking %p(%p)\n", p, abs2rel(p) );
      if( !WR_VM_VALID_ABS_PTR(p) )
      {
         printf( "wss_vm_malloc: free block out of range ==> %p\n", p );
         dbck_panic( "wss_vm_malloc" );
      }
#endif /* DEBUG_VM */
      if( WR_VM_SIZE(p) >= nunits )
      {
         /* big enough */
         if( WR_VM_SIZE(p) == nunits )
         {
            /* exactly */
            q->s.next = p->s.next;  /* both addrs rel */
         }
         else
         {
            p->s.size = WR_VM_SIZE(p) - nunits;
            p += WR_VM_SIZE(p);
#ifdef DEBUG_VM
            if( !WR_VM_VALID_ABS_PTR(p) )
            {
               printf( "wss_vm_malloc: new block out of range ==> %p\n", p );
               dbck_panic( "wss_vm_malloc" );
            }
#endif /* DEBUG_VM */
            p->s.size = nunits;
         }

         wss_vm_head->free_bytes -= nunits * sizeof(HEADER);

         /* Add block to trace list */
         p->s.next = wss_vm_head->trace;
         wss_vm_head->trace = (HEADER *)abs2rel(p);

#ifdef DEBUG_VM
         *((int *)(p + nunits - 1)) = 0xdeadbeef;
#endif /* DEBUG_VM */

         allocp = q;
#ifdef DEBUG_VM
         if( allocp == NULL ) printf( "wss_vm_malloc: allocp set null\n" );
#endif /* DEBUG_VM */
         return( (void*) (p+1) );
      }
      if( p == allocp )
      {
         /* wrapped around free list */
         return NULL;  /* out of memory */
      }
   }
}

/*
 * Like free(3), but returns (absolute addressed) memory to the VM region.
 */
#ifdef DEBUG_VM
void
wss_vm_free_debug( void *addr, const char *file, int line )
#else
void
wss_vm_free( void *addr )
#endif /* DEBUG_VM */
{
   register HEADER *mem, *p;

#ifdef DEBUG_VM
   printf( "wss_vm_free: Freeing block %p (%s:%d)\n", addr, file, line );
#endif /* DEBUG_VM */

   save = NULL;

   mem = (HEADER *)addr - 1;     /* point to header */

   /* 
    * If this is the first malloc/free for this run of the wss,
    * initialize the header.
    */
   if( allocp == NULL )
   {
      allocp = (HEADER *)rel2abs(wss_vm_head->base);
      if( allocp->s.next == NULL )
      {
         printf( "BAD Error: trying to free from VM before any allocations\n" );
         /* new wss.db file */
         allocp->s.next = wss_vm_head->base;
         allocp->s.size = ( wss_vm_size - header_size ) / sizeof(HEADER);
         wss_vm_head->trace = NULL;
         wss_vm_head->free_bytes = allocp->s.size * sizeof(HEADER);
      }
#ifdef DEBUG_VM
      if( allocp == NULL ) printf( "wss_vm_free: allocp set null\n" );
#endif /* DEBUG_VM */
   }

#ifdef DEBUG_VM
   if( !WR_VM_VALID_ABS_PTR(allocp) )
   {
      printf( "wss_wm_free: allocp out of range ==> %p\n", allocp );
      dbck_panic( "wss_vm_free" );
   }
   if( !WR_VM_VALID_ABS_PTR(mem) )
   {
      printf( "wss_wm_free: free block out of range ==> %p\n", mem );
      dbck_panic( "wss_vm_free" );
   }
#endif /* DEBUG_VM */

   wss_vm_head->free_bytes += WR_VM_SIZE(mem) * sizeof(HEADER);

#ifdef DEBUG_VM
   if( *((int *)mem) == 0xdeadbeef )
      dbck_panic( "wss_vm_free: allocation stomping detected at block %p(%p)\n", mem,
             abs2rel(mem) );
   if( *((int *)(mem + mem->s.size - 1)) != 0xdeadbeef )
      dbck_panic( "wss_vm_free: bounds stomping detected at block %p(%p)\n", mem,
             abs2rel(mem) );
#endif /* DEBUG_VM */
   if( wss_vm_head->trace != NULL )
   {
      register HEADER  *trace   = wss_vm_head->trace;
      register HEADER **q       = &wss_vm_head->trace;
#ifdef DEBUG_VM
      HEADER           *q_debug = trace;
#ifdef DEBUG_VM_TRACK
      printf( "wss_vm_free: tracking %p:", mem );
#endif /* DEBUG_VM_TRACK */
#endif /* DEBUG_VM */
      for( p = (HEADER *)rel2abs(trace); p;
           q = &p->s.next,
#ifdef DEBUG_VM
           q_debug = (HEADER *)abs2rel(p),
#endif /* DEBUG_VM */
           p = (HEADER*)((p->s.next)?rel2abs(p->s.next):NULL) )
      {
#ifdef DEBUG_VM
#ifdef DEBUG_VM_TRACK
         if( (unsigned)p <= (unsigned)mem )
            printf( " %p==%p", p, mem );
#endif /* DEBUG_VM_TRACK */
         if( !WR_VM_VALID_ABS_PTR(p) )
         {
#ifdef DEBUG_VM_TRACK
            printf( "\n" );
#endif
            printf( "wss_wm_free: ptr out of range in trace ==> %p\n", p );
            dbck_panic( "wss_vm_free" );
         }
#endif /* DEBUG_VM */
         if( p == mem )
         {
#ifdef DEBUG_VM
#ifdef DEBUG_VM_TRACK
            printf( "\n" );
#endif
            printf( "wss_vm_free: Setting *q (&%p->s.next) = %p (next = %p)\n",
                    q_debug, mem->s.next,
                    (mem->s.next)?((HEADER *)rel2abs(mem->s.next))->s.next:NULL );
#endif /* DEBUG_VM */
            *q = mem->s.next;
            break;
         }
      }
#ifdef DEBUG_VM
#ifdef DEBUG_VM_TRACK
      if( p == NULL )
         printf( "\n" );
#endif /* DEBUG_VM_TRACK */
#endif /* DEBUG_VM */

      if( p == NULL )
      {
         if( (HEADER *)abs2rel(mem) == trace )
            wss_vm_head->trace = mem->s.next;
#ifdef DEBUG_VM
         else
         {
            HEADER *head_base = (HEADER *)rel2abs(wss_vm_head->base);
            for( p = (HEADER *)rel2abs(head_base->s.next); p;
                 p = (HEADER*)rel2abs(p->s.next) )
            {
               if( p == mem )
                  dbck_panic( "wss_vm_free: already freed block ==> %p\n", mem );
               if( p == head_base ) break;
            }
            dbck_panic( "wss_vm_free: unable to find block in trace ==> %p\n", mem);
         }
#endif /* DEBUG_VM */
      }
   }

#ifdef DEBUG_VM
/*   mem->s.next = allocp->s.next;
   allocp->s.next = (HEADER *)abs2rel(mem);
   allocp = mem; */
#else
   /* find free blocks adjacent to the free'd memory*/
   for( p = allocp; !( ( mem > p ) && ( mem < (HEADER*)rel2abs(p->s.next) ) );
        p = (HEADER*)rel2abs(p->s.next) )
   {
      if( ( p >= (HEADER*)rel2abs(p->s.next) ) &&
          ( ( mem > p ) || ( mem < (HEADER*)rel2abs(p->s.next) ) ) )
      {
         break;      /* at one end or the other */
      }
   }
   if( mem + WR_VM_SIZE(mem) == (HEADER*)rel2abs(p->s.next) )
   {
      /* join to upper free block */
      mem->s.size = WR_VM_SIZE(mem) + WR_VM_SIZE((HEADER*)rel2abs(p->s.next));
      mem->s.next = ((HEADER*)rel2abs(p->s.next))->s.next;
   }
   else
   {
      mem->s.next = p->s.next;
   }

   if( p + WR_VM_SIZE(p) == mem )
   {
      /* join to lower block */
      p->s.size = WR_VM_SIZE(p) + WR_VM_SIZE(mem);
      p->s.next = mem->s.next;
   }
   else
   {
      p->s.next = (HEADER*)abs2rel(mem);
   }
   allocp = p;
#endif /* DEBUG_VM */

#ifdef DEBUG_VM
   if( allocp == NULL ) printf( "wss_vm_free: allocp set null\n" );
#endif /* DEBUG_VM */
}

