/*! \file test_dbintd.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: test_dbintd.c,v 1.6 2001/02/23 19:02:01 stevek Exp $
 *
 * \brief   Database inteface daemon test program
 * \author  Steve Kiernan
 * \version $Revision: 1.6 $
 * \date    $Date: 2001/02/23 19:02:01 $
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/syslog.h>
#include <sys/stat.h>

#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include <wss_db.h>

#include <libwrv.h>

#include "dbintd.h"

int DbCommand __P(( int, int, void *, size_t, void *, size_t * ));

static TCHAR  ConfigTableName[] = ".CONFIG";
static u_long ConfigTableId     = 0;

int
main( int argc, char *argv[] )
{
   struct sockaddr_un  un;
   int                 sock;
   void               *pdata;
   size_t              pktsize;

   /* Create UNIX domain socket */
   un.sun_family = AF_UNIX;
   if( sizeof(_PATH_DBINTSOCK) >= sizeof(un.sun_path) )
   {
      fprintf( stderr, "test_dbintd: Socket name too long\n" );
      return (EINVAL);
   }
   strcpy( un.sun_path, _PATH_DBINTSOCK );
   un.sun_len = strlen( un.sun_path );

   if( ( sock = socket( AF_UNIX, SOCK_STREAM, 0 ) ) < 0 )
   {
      perror( "socket" );
      return (errno);
   }

   /* Bind socket into filesystem */
   if( connect( sock, (struct sockaddr *) &un, sizeof(un) ) < 0 )
   {
      perror( "connect" );
      return (errno);
   }

   if( DbCommand( sock, DBOP_VERSION, (void *) NULL, 0, (void *) &pdata,
                  &pktsize ) < 0 )
   {
      perror( "DbCommand ==> DBOP_VERSION" );
      return (errno);
   }

   if( *(int *)pdata != DBINT_VERSION )
   {
      free( pdata );
      fprintf( stderr, "test_dbintd: Mismatched versions\n" );
      return (EINVAL);
   }

   free( pdata );

   if( DbCommand( sock, DBOP_HANDLE, ConfigTableName, sizeof(ConfigTableName),
                  (void *) &pdata, &pktsize ) < 0 )
   {
      perror( "DbCommand ==> DBOP_HANDLE" );
      return (errno);
   }

   if( pktsize != sizeof(ConfigTableId) )
   {
      free( pdata );
      fprintf( stderr, "test_dbintd: Packet size mismatch\n" );
      return (EINVAL);
   }

   ConfigTableId = *(u_long *)pdata;

   printf( "%s = 0x%x\n", ConfigTableName, (u_int) ConfigTableId );

   free( pdata );

   {
      wrv_h_t     data;
      int         ncols;
      wss_byte_t *ptr;
      wqcd_t      cd;
      int         i;

      data.handle = ConfigTableId;

      if( DbCommand( sock, DBOP_COLDEFS, &data, sizeof(data), (void *) &pdata,
                     &pktsize ) < 0 )
      {
         perror( "DbCommand ==> DBOP_COLDEFS" );
         return (errno);
      }

      ncols = *( (int *) pdata );
      ptr = (wss_byte_t *) pdata + sizeof(ncols);
      cd = (wqcd_t) ptr;

      if( ncols )
         printf( "Type  Max Length   Name\n" );
      for( i = 0; i < ncols; i++ )
      {
         wss_char_t *name = (wss_char_t *) (cd + 1);
         wqcd_t      ncd  = cd;

         cd = (wqcd_t) ((wss_byte_t *) name  + (size_t) cd->name);

         printf( "%4d  %10d   %s\n", ncd->dtype, ncd->max_len, name );
      }
   }

   close( sock );

   return (0);
}


int
DbCommand( int sock, int cmd, void *out, size_t outsize, void *in,
           size_t *insize )
{
   db_command_t *pcmd;
   wss_byte_t   *pRead;
   size_t        pktsize;
   size_t        totalRead;
   int           nread;

   pcmd = (void *)malloc( sizeof(db_command_t) );
   if( pcmd == NULL )
      return (-1);

   pcmd->cmd = cmd;

   if( write( sock, pcmd, sizeof(db_command_t) ) < 0 )
   {
      free( pcmd );
      return (-1);
   }

   if( out && outsize )
   {
      if( write( sock, &outsize, sizeof(outsize) ) < 0 )
      {
         free( pcmd );
         return (-1);
      }

      if( write( sock, out, outsize ) < 0 )
      {
         free( pcmd );
         return (-1);
      }
   }

   if( read( sock, &pktsize, sizeof(pktsize) ) < 0 )
   {
      free( pcmd );
      return (-1);
   }

   if( pktsize < 0 )
   {
      int error;
      free( pcmd );
      if( read( sock, &error, sizeof(error) ) >= 0 )
         errno = error;
      return (-1);
   }

   pcmd = realloc( pcmd, pktsize );
   if( pcmd == NULL )
      return (-1);

   pRead = (wss_byte_t *)pcmd;
   totalRead = 0;

   /* Keep reading until we have all the data */
   while( pktsize )
   {
      if( ( nread = read( sock, pRead, pktsize ) ) < 0 )
      {
         free( pcmd );
         return (-1);
      }
      pktsize -= nread;
      totalRead += nread;
      pRead += nread;
   }

   if( in )
   {
      *(char **)in = (char *) pcmd;
      if( insize )
         *insize = totalRead;
   }
   else
      free( pcmd );

   return (0);
}

