/*! \file dbinterface.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: dbinterface.c,v 1.12 2001/02/23 19:02:02 stevek Exp $
 *
 * \brief   Database inteface module
 * \author  Steve Kiernan
 * \version $Revision: 1.12 $
 * \date    $Date: 2001/02/23 19:02:02 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/eventhandler.h>
#include <sys/kernel.h>
#include <sys/kthread.h>
#include <sys/linker.h> 
#include <sys/lock.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/proc.h>
#include <sys/ptrace.h>
#include <sys/sysctl.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/unistd.h>
   
#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/vm_extern.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>
#include <sys/shm.h>

#include <wsi_mutex.h>
#include <wss_vm.h>
#include <wss_db.h>
#include <wss_wrvevent.h>
#include <wss_malloc.h>

#include <dbintd/dbintd.h>


static int dbinterface_mod_event __P(( module_t, int, void * ));

static moduledata_t dbinterface_mod =
{
   "WSS Database Interface",
   dbinterface_mod_event,
   NULL,
};
DECLARE_MODULE( dbinterface, dbinterface_mod, SI_SUB_KTHREAD_IDLE,
                SI_ORDER_MIDDLE );

static char *dbintdpaths[] =
{
   "/sbin/dbintd",
   NULL
};

static void dbintd_atexit __P(( struct proc *p ));
static void dbintd_start  __P(( void ));

typedef struct lock_held lock_held_t;

struct lock_held
{
   struct proc        *p;
   WSI_MUTEX_POINTER  (lock);
   int                 cmd;
   lock_held_t        *next;
   lock_held_t       **prev;
};

struct dbintd_args
{
   int   code;
   void *args;
};
int dbintd_syscall __P(( struct proc *, struct dbintd_args * ));

static struct proc      *dbintdproc    = NULL;
static int               dbintdsyscall = NO_SYSCALL;
static struct lock_held *lock_list     = NULL;
WSI_MUTEX_PROTECT       (lock_list);

extern wss_char_t       *wss_db_filename;
extern int               wss_db_size;
WSI_MUTEX_EXPORT        (db);

static struct kproc_desc dbint_kp =
{
   "dbintd",
   dbintd_start,
   &dbintdproc
};
SYSINIT( dbinterface, SI_SUB_KTHREAD_IDLE, SI_ORDER_FIRST, kproc_start,
         &dbint_kp )

static struct sysent newent = { 2, (sy_call_t *) dbintd_syscall };
static struct sysent oldent;

#ifdef DEBUG_DBINT
static struct sysent oldent;
#endif /* DEBUG_DBINT */


static void
dbintd_start()
{
   vm_offset_t addr;
   struct execve_args args;
   struct proc *p;
   char **pathp, *path, *ucp, **uap, *arg0, *arg1, *arg2, *arg3;
   char synum[4];
   char dbsize[32];
   int i;
   int error = 0;

   /* Process needs to be suspended prior to shutdown sync */
   EVENTHANDLER_REGISTER( shutdown_pre_sync, shutdown_kproc, dbintdproc,
                          SHUTDOWN_PRI_LAST );

   p = curproc;

   /* Reparent process to init */
   proc_reparent( p, initproc );

   /* Make up a name */
   strcpy( p->p_comm, "dbintd" );

   /* Grab a page to hold parameters */
   addr = trunc_page( USRSTACK - PAGE_SIZE );
   if( vm_map_find( &p->p_vmspace->vm_map, NULL, 0, &addr, PAGE_SIZE, FALSE,
                    VM_PROT_ALL, VM_PROT_ALL, 0 ) != 0 )
   {
      printf( "dbinterface: couldn't allocate argument space\n" );
      goto doexit;
   }
   p->p_vmspace->vm_maxsaddr = (caddr_t)addr;
   p->p_vmspace->vm_ssize = 1;

   if( syscall_register( &dbintdsyscall, &newent, &oldent ) )
   {
      printf( "dbinterface: couldn't find an open syscall entry\n" );
      goto doexit;
   }
   printf( "dbinterface: added syscall #%d\n", dbintdsyscall );

   sprintf( synum, "%d", dbintdsyscall );
   sprintf( dbsize, "%d", wss_db_size );

   /* Try to execute dbintd from known paths */
   for( pathp = &dbintdpaths[0]; (path = *pathp) != NULL; pathp++ )
   {
      ucp = (char *)USRSTACK;

      /* Move out the syscall number (arg 3) */
      for( i = strlen(synum) + 1; i >= 0; i-- )
         (void)subyte( --ucp, synum[i] );
      arg3 = ucp;

      /* Move out the database size (arg 2) */
      for( i = strlen(dbsize) + 1; i >= 0; i-- )
         (void)subyte( --ucp, dbsize[i] );
      arg2 = ucp;

      /* Move out the database path (arg 1) */
      for( i = strlen(wss_db_filename) + 1; i >= 0; i-- )
         (void)subyte( --ucp, wss_db_filename[i] );
      arg1 = ucp;

      /* Move out the file name (arg 0) */
      for( i = strlen(path) + 1; i >= 0; i-- )
         (void)subyte( --ucp, path[i] );
      arg0 = ucp;

      /* Move out the arg pointers */
      uap = (char **)((intptr_t)ucp & ~(sizeof(intptr_t)-1));
      (void)suword( (caddr_t)--uap, (long)0 ); /* terminator */
      (void)suword( (caddr_t)--uap, (long)(intptr_t)arg3 );
      (void)suword( (caddr_t)--uap, (long)(intptr_t)arg2 );
      (void)suword( (caddr_t)--uap, (long)(intptr_t)arg1 );
      (void)suword( (caddr_t)--uap, (long)(intptr_t)arg0 );

      /* Point at the arguments */
      args.fname = arg0;
      args.argv = uap;
      args.envv = NULL;

      /* Now try to exec the program */
      if( ( error = execve(p, &args) ) == 0 )
         return;

      if( error != ENOENT )
         printf( "exec %s: error %d\n", path, error );
   }

   printf( "dbinterface: dbintd not found\n" );

   dbintdproc = NULL;

doexit:
   exit1( p, error );
}

static int
dbinterface_mod_event( module_t mod, int type, void *data )
{
   int error = 0;

   switch( type ) 
   {
      case MOD_UNLOAD:
         /* Kill the dbintd */
         if( dbintdproc )
         {
            /* Signal the dbintd process to die */
            psignal( dbintdproc, SIGKILL );

            /* Wait for the dbintd to exit */
            error = tsleep( (caddr_t)dbintdproc->p_pptr, PWAIT | PCATCH,
                            "wait", 0 );
            if( error )
               break;
         }
         error = syscall_deregister( &dbintdsyscall, &oldent );
         rm_at_exit( dbintd_atexit );
         WSI_MUTEX_DESTROY(lock_list);
         break;

      case MOD_LOAD:
         /* Initialize the mutex */
         WSI_MUTEX_INIT(lock_list);

         /* Hook exit call */
         at_exit( dbintd_atexit );
         break;

      default:
         break;
   }
   return (error);
}

static void
dbintd_atexit( struct proc *p )
{
   if( p == dbintdproc )
   {
      lock_held_t *plock;

      dbintdproc = NULL;

      /* Release any reserved locks */
      WSI_MUTEX_ENTER(lock_list);
      for( plock = lock_list; plock; plock = plock->next )
      {
         if( plock->p == p )
         {
            WSI_MUTEX_EXIT_POINTER( plock->lock_lock );

            *(plock->prev) = plock->next;
            wss_free( plock );
            return;
         }
      }
      WSI_MUTEX_EXIT(lock_list);
   }
}

static int dbintd_version = DBINT_VERSION;

static void
add_lock_held( struct proc *p, WSI_MUTEX_POINTER(lock), int cmd )
{
   lock_held_t *plock;

   plock = (lock_held_t *) wss_malloc( sizeof(*plock) );
   plock->p = p;
   plock->lock_lock = lock_lock;
   plock->cmd = cmd;
   plock->next = lock_list;
   plock->prev = &lock_list;
   if( lock_list )
      lock_list->prev = &(plock->next);
   lock_list = plock;
#ifdef DEBUG_DBINT
   printf( "dbinterface: added lock 0x%lx held by process #%d\n",
           (u_long) lock_lock, (int) p->p_pid );
#endif /* DEBUG_DBINT */
}

static void
remove_lock_held( struct proc *p, WSI_MUTEX_POINTER(lock) )
{
   lock_held_t *plock;

   for( plock = lock_list; plock; plock = plock->next )
      if( ( plock->p == p ) && ( plock->lock_lock == lock_lock ) )
      {
         *(plock->prev) = plock->next;
         wss_free( plock );
#ifdef DEBUG_DBINT
         printf( "dbinterface: removed lock 0x%lx held by process #%d\n",
                 (u_long) lock_lock, (int) p->p_pid );
#endif /* DEBUG_DBINT */
         return;
      }

#ifdef DEBUG_DBINT
   printf( "dbinterface: remove_lock_held called for non-held lock 0x%lx by "
           "process #%d\n", (u_long) lock_lock, (int) p->p_pid );
#endif /* DEBUG_DBINT */
}

static int
is_lock_held( struct proc *p, WSI_MUTEX_POINTER(lock) )
{
   lock_held_t *plock;

   for( plock = lock_list; plock; plock = plock->next )
      if( ( plock->p == p ) && ( plock->lock_lock == lock_lock ) )
         return (TRUE);
   return (FALSE);
}

int
dbintd_syscall( struct proc *p, struct dbintd_args *uap )
{
   int          error = 0;

#ifndef DEBUG_DBINT
   if( p != dbintdproc )
      return (ENOSYS);
#endif /* !DEBUG_DBINT */

   switch( uap->code )
   {
      case DBINT_GET_VERSION:
         error = copyout( &dbintd_version, uap->args, sizeof(int) );
         break;

      case DBINT_LOCK:
      {
         int flag;

         if( ( error = copyin( uap->args, &flag, sizeof(flag) ) ) )
            break;

         switch( flag )
         {
            case DBF_UNLOCK:
               WSI_MUTEX_ENTER(lock_list);
               if( is_lock_held( p, WSI_MUTEX_ADDR(db) ) )
               {
                  WSI_MUTEX_EXIT(db);
                  remove_lock_held( p, WSI_MUTEX_ADDR(db) );
               }
               else
               {
#ifdef DEBUG_DBINT
                  printf( "DB lock not held for unlock (%d)\n",
                          (int) curproc->p_pid );
#endif /* DEBUG_DBINT */
                  remove_lock_held( p, WSI_MUTEX_ADDR(db) );
                  psignal( p, SIGSEGV );
               }
               WSI_MUTEX_EXIT(lock_list);
               break;

            case DBF_LOCK:
               WSI_MUTEX_ENTER(lock_list);
               if( is_lock_held( p, WSI_MUTEX_ADDR(db) ) )
               {
#ifdef DEBUG_DBINT
                  printf( "DB lock already held for unlock (%d)\n",
                          (int) curproc->p_pid );
#endif /* DEBUG_DBINT */
                  psignal( p, SIGSEGV );
               }
               else
               {
                  WSI_MUTEX_ENTER(db);
                  add_lock_held( p, WSI_MUTEX_ADDR(db), uap->code );
               }
               WSI_MUTEX_EXIT(lock_list);
               break;
         }
         break;
      }

      case DBINT_LOCK_TABLE:
      {
         struct table_lock_args
         {
            int flag;
            u_long id;
         }
         args;
         wqtab_t tab;

         if( ( error = copyin( uap->args, &args, sizeof(args) ) ) )
            break;

         if( wss_vm_root_table == NULL )
         {
            error = ENOENT;
            break;
         }

         for( tab = (wqtab_t)rel2abs(*wss_vm_root_table);
              tab && ( (u_long)abs2rel(tab) != args.id );
              tab = (wqtab_t)((tab->next)?rel2abs(tab->next):NULL) )
            continue;

         if( tab == NULL )
         {
            error = ENOENT;
            break;
         }

         switch( args.flag )
         {
            case DBF_UNLOCK:
               WSI_MUTEX_ENTER(lock_list);
               if( is_lock_held( p,
                                 WSI_MUTEX_ADDR(WQL_TABLE_OPAQUE(tab)->table) ) )
               {
                  wss_db_unlock_table( tab );
                  remove_lock_held( p,
                                    WSI_MUTEX_ADDR(WQL_TABLE_OPAQUE(tab)->table) );
               }
               else
               {
#ifdef DEBUG_DBINT
                  printf( "DB table (0x%lx) lock not held for unlock (%d)\n",
                          (u_long) abs2rel(tab), (int) curproc->p_pid );
#endif /* DEBUG_DBINT */
                  remove_lock_held( p,
                                    WSI_MUTEX_ADDR(WQL_TABLE_OPAQUE(tab)->table) );
                  psignal( p, SIGSEGV );
               }
               WSI_MUTEX_EXIT(lock_list);
               break;

            case DBF_LOCK:
               WSI_MUTEX_ENTER(lock_list);
               if( is_lock_held( p,
                                 WSI_MUTEX_ADDR(WQL_TABLE_OPAQUE(tab)->table) ) )
               {
#ifdef DEBUG_DBINT
                  printf( "DB table (0x%lx) lock already held for lock (%d)\n",
                          (u_long) abs2rel(tab), (int) curproc->p_pid );
#endif /* DEBUG_DBINT */
                  psignal( p, SIGSEGV );
               }
               else
               {
                  wss_db_lock_table( tab );
                  add_lock_held( p,
                                 WSI_MUTEX_ADDR(WQL_TABLE_OPAQUE(tab)->table),
                                 uap->code );
               }
               WSI_MUTEX_EXIT(lock_list);
               break;
         }
         break;
      }

      case DBINT_POST_EVENT:
      {
         struct event_args
         {
            wrv_event_sig_t  signature;
            u_long           key;
            u_long           extra;
            void            *buffer;
            u_long           length;
         }
         args;

         if( ( error = copyin( uap->args, &args, sizeof(args) ) ) )
            break;

         if( args.buffer && args.length )
         {
            wss_byte_t *buffer = wss_malloc( args.length );

            if( ( error = copyin( args.buffer, buffer, args.length ) ) )
            {
               wss_free( buffer );
               break;
            }

            args.buffer = buffer;
         }

         wss_wrvevent_post( args.signature, args.key, args.extra, args.buffer,
                            args.length );

         break;
      }

      default:
         error = ENOSYS;
   }

   return (error);
}

