/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: gwm.cxx,v 1.68 2001/04/05 18:40:22 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: The Global Wrapper Manager service
 *
 *          This Windows NT service manages the global state of wrappers.
 *          It communicates with both wrapped processes' local wrapper
 *          manager instances and WSS management clients via a set of
 *          named pipes.
 */

#include <wr_conf.h>

#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>

#include <wss_init.h>
#include <wss_module.h>
#include <wss_mutex.h>
#include <wss_process.h>
#include <wss_wrapper.h>
#include <wss_wrv.h>
#include <wss_wrvevent.h>
#include <wss_criteria.h>
#include <wss_audit.h>

#include <linker_pe.h>

#include <wsi_time.h>

#include <eventfs.h>
#include <pipesrv.h>

#include "service.h"
#include "gwm.h"
#include "gwm_comm.h"


// external functions
//
extern "C" {
int  IMPORT wsi_export_init( void );
void IMPORT linker_crit_init( void );
void IMPORT eventfs_interface_init( void );
void IMPORT eventfs_interface_uninit( void );
int  IMPORT eventfs_init( void );

BOOL IMPORT eventfs_interface_register( eventfs_interface_t * );
}


// prototypes
//
int GWMLoadWrapper( wss_mod_t *wrm, HINSTANCE wrapperInstance );
int GWMLoadActivation( wss_mod_t *wrm, HINSTANCE criteriaInstance );
int GWMLoadModule( TCHAR *moduleName );

// this event is signalled when the service should end
//
HANDLE g_hServerStopEvent = NULL;

BOOL g_bDebugging = FALSE;
BOOL g_bConsole   = FALSE;

DB_HANDLERS g_dbHandlers =
{
   wss_create,
   wss_drop,
   wss_insert,
   wss_select,
   wss_delete,
   wss_update,
   wss_db_find_table_by_names,
   
   wss_audit_write
};

eventfs_interface_t g_efsIFace =
{
   EFSPURENAME(eventfs_construct_stream),
   EFSPURENAME(eventfs_destroy_group),
   EFSPURENAME(eventfs_destroy_stream),
   EFSPURENAME(eventfs_flush_stream),
   EFSPURENAME(eventfs_get_event_mask),
   EFSPURENAME(eventfs_init),
   EFSPURENAME(eventfs_post_event),
   EFSPURENAME(eventfs_post_event_by_types),
   EFSPURENAME(eventfs_register_event_type),
   EFSPURENAME(eventfs_register_group),
   EFSPURENAME(eventfs_register_stream)
};


////////////////////////////////////////////////////////////////////////////
// Public functions
//
DWORD inline WaitMutex( HANDLE hMutex, DWORD time )
{
   DWORD dwWait;
   while( ( dwWait = WaitForSingleObject( hMutex, time ) ) != WAIT_OBJECT_0 );
   return dwWait;
}

VOID ServiceStart (DWORD dwArgc, LPTSTR *lpszArgv)
{
   DWORD                dwWait;
   LPPIPESRV_QUEUE_INFO gwmpqiClient = new PIPESRV_QUEUE_INFO();
   LPPIPESRV_QUEUE_INFO gwmpqiLWM = new PIPESRV_QUEUE_INFO();

   if( dwArgc > 0 )
   {
      g_bConsole = TRUE;
      g_bDebugging = TRUE;
   }
   g_bPipeSrvDebugging = g_bDebugging;


   ///////////////////////////////////////////////////
   //
   // Service initialization
   //

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   // create the event object. The control handler function signals
   // this event when it receives the "stop" control code.
   //
   g_hServerStopEvent = CreateEvent( NULL, TRUE, FALSE, NULL );

   if( g_hServerStopEvent == NULL )
      goto cleanup;

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   // initialize the EventFS interface
   //
   eventfs_interface_init();

   // register eventfs interface handlers
   //
   if( !eventfs_interface_register( &g_efsIFace ) )
   {
      if( g_bDebugging )
         _tprintf( _TEXT("eventfs_interface_register failed.\n") );

      goto cleanup;
   }

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   // initialize EventFS
   eventfs_init();

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   GWMLWMCommSetup();

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   GWMClientCommSetup();

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   // initialize the WSS
   //
   if( wss_init( WSS_INIT_DB | WSS_INIT_CONF ) )
   {
      if( g_bDebugging )
         _tprintf( _TEXT("wss_init failed.\n" ) );

      goto cleanup;
   }

   // initialize the PE linker module
   //
   linker_pe_init();

   // initialize the criteria linker module
   //
   linker_crit_init();

   // initialize the locally exported symbols
   //
   wsi_export_init();

   // register database command handlers
   //
   if( !WRImpRegisterDBHandlers( &g_dbHandlers ) )
   {
      if( g_bDebugging )
         _tprintf( _TEXT("WRImpRegisterDBHandlers failed.\n") );

      goto cleanup;
   }

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   // initialize pipe queues
   //
   gwmpqiLWM->thread = 0;
   gwmpqiLWM->bReadOnly = FALSE;
   gwmpqiLWM->lpszPipeName = GWM_LWM_PIPE;
   gwmpqiLWM->nInstances = 0;
   gwmpqiLWM->nThreads = 0;
   gwmpqiLWM->nActive = 0;
   gwmpqiLWM->pData = NULL;
   gwmpqiLWM->fOpen = NULL;
   gwmpqiLWM->fClose = GWMLWMCommCloseCallback;
   gwmpqiLWM->fProcessing = GWMLWMCommProcessPacket;
   gwmpqiLWM->hEvents[0] = g_hServerStopEvent;
   gwmpqiLWM->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] = NULL;
   gwmpqiLWM->hEvents[PIPESRV_PIPE_ACTIVE_EVENT] = NULL;

   gwmpqiClient->thread = 0;
   gwmpqiClient->bReadOnly = FALSE;
   gwmpqiClient->lpszPipeName = GWM_CLIENT_PIPE;
   gwmpqiClient->nInstances = 0;
   gwmpqiClient->nThreads = 0;
   gwmpqiClient->nActive = 0;
   gwmpqiClient->pData = NULL;
   gwmpqiClient->fOpen = GWMClientCommOpenCallback;
   gwmpqiClient->fClose = GWMClientCommCloseCallback;
   gwmpqiClient->fProcessing = GWMClientCommProcessPacket;
   gwmpqiClient->hEvents[0] = g_hServerStopEvent;
   gwmpqiClient->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] = NULL;
   gwmpqiClient->hEvents[PIPESRV_PIPE_ACTIVE_EVENT] = NULL;

   InitializeCriticalSection( &gwmpqiClient->critSec );
   InitializeCriticalSection( &gwmpqiLWM->critSec );

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;
   
   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   gwmpqiLWM->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] =
      CreateEvent( NULL, TRUE, FALSE, NULL );

   if( gwmpqiLWM->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] == NULL )
      goto cleanup;

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   gwmpqiLWM->hEvents[PIPESRV_PIPE_ACTIVE_EVENT] =
      CreateSemaphore( NULL, 0, 10000, NULL );

   if( gwmpqiLWM->hEvents[PIPESRV_PIPE_ACTIVE_EVENT] == NULL )
      goto cleanup;

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   gwmpqiClient->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] =
      CreateEvent( NULL, TRUE, FALSE, NULL );

   if( gwmpqiClient->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] == NULL )
      goto cleanup;

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   gwmpqiClient->hEvents[PIPESRV_PIPE_ACTIVE_EVENT] =
      CreateSemaphore( NULL, 0, 10000, NULL );

   if( gwmpqiClient->hEvents[PIPESRV_PIPE_ACTIVE_EVENT] == NULL )
      goto cleanup;

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   wss_wrvevent_register_hook( GWMClientCommPostEvent );

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   gwmpqiLWM->thread = _beginthread( PipeSrvAcceptThread, 0, gwmpqiLWM );

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_START_PENDING, NO_ERROR, 3000 ) )
      goto cleanup;

   gwmpqiClient->thread = _beginthread( PipeSrvAcceptThread, 0, gwmpqiClient );

   // report the status to the service control manager.
   //
   if( !ReportStatusToSCMgr( SERVICE_RUNNING, NO_ERROR, 0 ) )
      goto cleanup;

   //
   // End of initialization
   //
   ////////////////////////////////////////////////////////

   ////////////////////////////////////////////////////////
   //
   // Service is now running, wait for shutdown
   // TBD: Should this thread be doing some work instead,
   //      since it's here?
   //

   while ( 1 )
   {
      dwWait = WaitForSingleObject( g_hServerStopEvent, INFINITE );

      if( dwWait == WAIT_OBJECT_0 ) // server stop signaled
         break;
   }

cleanup:

   // TBD: cleanup events allocated for pipe queues

   if( g_hServerStopEvent )
      CloseHandle( g_hServerStopEvent );
}


VOID ServiceStop()
{
   if( g_hServerStopEvent )
      SetEvent( g_hServerStopEvent );
}



