/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: layout.cxx,v 1.6 2001/04/05 18:40:22 stevek Exp $
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include "layout.h"

LayoutManager::LayoutManager( HWND window, LPCTSTR resname )
{
   hWnd = window;
   bit = TRUE;
   initialized = FALSE;

   // Grab window client rect
   GetClientRect( hWnd, &rect );

   // Save the current dialog procedure
   dlgProc =
      (BOOL(*)(HWND,UINT,WPARAM,LPARAM)) GetWindowLong( hWnd, DWL_DLGPROC );

   // Save old user data
   StoreUserData( (LPVOID) GetWindowLong( hWnd, GWL_USERDATA ) );

   // Enumerate child windows and construct layout information
   EnumChildWindows( hWnd, _EnumProc, (LONG) this );

   // Resolve layout information from resource
   if( resname )
   {
      HMODULE hMod = GetModuleHandle( NULL );
      HRSRC hRes = FindResourceEx( hMod, MAKEINTRESOURCE(LAYOUTINFO), resname,
                                   MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL) );

      if( hRes )
      {
         HGLOBAL hGlobal = LoadResource( hMod, hRes );
         if( hGlobal )
         {
#define RES_SHORT(res,indx)  (*((PSHORT)((pd_byte_t *)res + dwProcessed + indx)))
#define RES_USHORT(res,indx) (*((PUSHORT)((pd_byte_t *)res + dwProcessed + indx)))
#define RES_TYPE(res)      RES_USHORT(res,0)
#define RES_NAME(res)      ((RES_TYPE(res)==LRC_NAMED) \
                            ? (pd_char_t *)((pd_byte_t *)res + dwProcessed + \
                                             sizeof(USHORT)) \
                            : NULL)
#define RES_ID(res)        ((RES_TYPE(res)==LRC_ID) \
                            ? *((PSHORT)(res + dwProcessed) + 1) \
                            : -1)

            DWORD       dwResSize = SizeofResource( hMod, hRes );
            pd_byte_t *pRes      = (pd_byte_t *) LockResource( hGlobal );

            // Process resource data
            DWORD dwProcessed = 0;
            DWORD idnameIndex;
            for( ; ( idnameIndex = sizeof(USHORT) ); dwProcessed += idnameIndex )
            {
               if( ( dwProcessed + idnameIndex ) >= dwResSize ) break;

               HWND hControl = NULL;
               switch( RES_TYPE(pRes) )
               {
                  // Named component (usually an IDC_STATIC)
                  case LRC_NAMED:
                  {
                     pd_char_t *resName = RES_NAME(pRes);
                     idnameIndex += ( wsi_strlen(resName) + 1 ) *
                                    sizeof(pd_char_t);
                     LAYOUT_NAME_MAP::iterator it = compNames.find( resName );
                     if( it != compNames.end() )
                        hControl = (*it).second->hControl;
                     break;
                  }

                  // Component with numeric id
                  case LRC_ID:
                  {
                     int resId = RES_ID(pRes);
                     idnameIndex += sizeof(SHORT);
                     hControl = GetDlgItem( hWnd, resId );
                     break;
                  }
               }

               if( ( dwProcessed + idnameIndex ) >= dwResSize ) break;
               USHORT flags = RES_USHORT(pRes, idnameIndex);
               idnameIndex += sizeof(USHORT);

               if( ( dwProcessed + idnameIndex ) >= dwResSize ) break;
               int resId = RES_SHORT(pRes, idnameIndex);
               idnameIndex += sizeof(USHORT);

               if( hControl )
               {
                  LAYOUT_MAP::iterator it = components.find( hControl );
                  if( it != components.end() )
                  {
                     Layout *pLayout = (*it).second;
                     pLayout->style = flags;
                     pLayout->hRelative = ((resId>=0)
                                           ? GetDlgItem( hWnd, resId )
                                           : NULL);
                  }
               }
            }
         }
      }
   }

   // Set user data on window to point to us
   SetWindowLong( hWnd, GWL_USERDATA, (LONG) this );

   // Replace the dialog proc with ours (we'll sit between the app's
   // dialog proc and Win32 and handle any messages not handled by the
   // app).
   SetWindowLong( hWnd, DWL_DLGPROC, (LONG) _DlgProc );
}

LayoutManager::~LayoutManager()
{
   SetWindowLong( hWnd, DWL_DLGPROC, (LONG) dlgProc );
   SetWindowLong( hWnd, GWL_USERDATA, (LONG) FetchUserData() );
}

BOOL CALLBACK
LayoutManager::_EnumProc( HWND hChild, LPARAM lParam )
{
   LayoutManager *pLMan = (LayoutManager *) lParam;

   Layout *pLayout = new Layout();
   pLayout->hControl = hChild;
   pLayout->rx = 0;
   pLayout->ry = 0;

   pLMan->AddComponent( hChild, pLayout );

   return TRUE;
}

void
LayoutManager::AddComponent( HWND hComp, Layout *pLayout )
{
   pd_char_t buf[MAXPATHLEN];
   memset( buf, 0, sizeof(buf) );

   if( SendMessage( hComp, WM_GETTEXT, (WPARAM) MAXPATHLEN, (LPARAM) buf ) )
      compNames.insert( LAYOUT_NAME_MAP::value_type( buf, pLayout ) );
   components.insert( LAYOUT_MAP::value_type( hComp, pLayout ) );
}

BOOL CALLBACK
LayoutManager::_DlgProc( HWND hwndDialog, UINT uMsg, WPARAM wParam,
                         LPARAM lParam )
{
   LayoutManager *pLMan;
   BOOL           bRet = FALSE;
   
   pLMan = (LayoutManager *) GetWindowLong( hwndDialog, GWL_USERDATA );

   switch( uMsg )
   {
     case WM_WINDOWPOSCHANGING:
     {
        if( ((LPWINDOWPOS)lParam)->flags & SWP_SHOWWINDOW )
           pLMan->DoLayoutInit();
        break;
     }

     case WM_SIZE:
     {
        pLMan->DoLayout( LOWORD(lParam), HIWORD(lParam) );
        break;
     }

     default:
        break;
   }

   bRet |= (pLMan->dlgProc)( hwndDialog, uMsg, wParam, lParam );

   // WARNING! Do not remove the following code line!
   // It works around a serious bug in the VC++ compiler
   BOOL foo = TRUE;

   return bRet;
}

void
LayoutManager::DoLayoutInit()
{
   if( initialized ) return;

   initialized = TRUE;

   for( LAYOUT_MAP::iterator it = components.begin();
        it != components.end();
        it++ )
   {
      HWND    hChild  = (*it).first;
      Layout *pLayout = (*it).second;

      WINDOWPLACEMENT wp;
      wp.length = sizeof(wp);
      GetWindowPlacement( hChild, &wp );
      CopyRect( &(pLayout->rect), &wp.rcNormalPosition );
   }

   for( it = components.begin(); it != components.end(); it++ )
   {
      Layout *pLayout = (*it).second;
      if( pLayout->hRelative )
      {
         LAYOUT_MAP::iterator relit = components.find( pLayout->hRelative );

         if( relit != components.end() )
         {
            Layout *pRelLayout = (*relit).second;

            if( pLayout->style & LSF_RELATIVE_X )
               pLayout->rx = pLayout->rect.right - pRelLayout->rect.left;

            if( pLayout->style & LSF_RELATIVE_Y )
               pLayout->ry = pLayout->rect.top - pRelLayout->rect.bottom;
         }
      }

      if( pLayout->style & LSF_SNAP_X )
         pLayout->rx = ( rect.right - rect.left ) - pLayout->rect.right;

      if( pLayout->style & LSF_SNAP_Y )
         pLayout->ry = ( rect.bottom - rect.top ) - pLayout->rect.bottom;
   }
}

BOOL
LayoutManager::StoreUserData( LPVOID data )
{
   map.insert( USERDATA_MAP::value_type( 0, data ) );
   return TRUE;
}

BOOL
LayoutManager::StoreUserData( DWORD id, LPVOID data )
{
   if( id == 0 ) return FALSE;

   map.insert( USERDATA_MAP::value_type( id, data ) );
   return TRUE;
}

LPVOID
LayoutManager::FetchUserData( DWORD id )
{
   USERDATA_MAP::iterator it = map.find( id );
   return( ( it != map.end() )?(*it).second:NULL );
}

LPVOID
LayoutManager::FetchUserData()
{
   return FetchUserData( 0 );
}

void
LayoutManager::DoLayout( Layout *pLayout )
{
   if( !pLayout || ( pLayout->bit == bit ) ) return;

   pLayout->bit = bit;

   LPRECT crect = &(pLayout->rect);

   if( pLayout->hRelative )
   {
      LAYOUT_MAP::iterator findit = components.find( pLayout->hRelative );
      Layout *pRelLayout = ( findit != components.end() )
                           ? (*findit).second
                           : NULL;

      // Layout relative component
      RECT   originRect = { 0, 0, 0, 0 };
      LPRECT rrect;
      if( pRelLayout )
      {
         DoLayout( pRelLayout );
         rrect = &(pRelLayout->rect);
      }
      else
         rrect = &originRect;

      if( pLayout->style & LSF_RELATIVE_X )
      {
         int adjust = ( rrect->right + pLayout->rx ) - crect->left;
         crect->left += adjust;
         crect->right += adjust;
      }

      if( pLayout->style & LSF_RELATIVE_Y )
      {
         int adjust = ( rrect->bottom + pLayout->ry ) - crect->top;
         crect->top += adjust;
         crect->bottom += adjust;
      }

      if( pLayout->style & LSF_CENTER_X )
      {
         int cwidth = crect->right - crect->left;
         int rwidth = rrect->right - rrect->left;
         int adjust = ( rwidth - cwidth ) / 2;

         crect->left = rrect->left + adjust;
         crect->right = rrect->left + adjust + cwidth;
      }
      
      if( pLayout->style & LSF_CENTER_Y )
      {
         int cheight = crect->bottom - crect->top;
         int rheight = rrect->bottom - rrect->top;
         int adjust = ( rheight - cheight ) / 2;

         crect->top = rrect->top + adjust;
         crect->bottom = rrect->top + adjust + cheight;
      }
   }

   if( pLayout->style & LSF_SNAP_X )
   {
      int cwidth = crect->right - crect->left;
      int width = rect.right - rect.left;

      crect->right = width - pLayout->rx;
      crect->left = crect->right - cwidth;
   }

   if( pLayout->style & LSF_SNAP_Y )
   {
      int cheight = crect->bottom - crect->top;
      int height = rect.bottom - rect.top;

      crect->bottom = height - pLayout->ry;
      crect->top = crect->bottom - cheight;
   }

   if( pLayout->style & LSF_STRETCH_X )
   {
      int adjust = ( rect.right - rect.left ) -
                   ( orect.right - orect.left );
      crect->right += adjust;
   }

   if( pLayout->style & LSF_STRETCH_Y )
   {
      int adjust = ( rect.bottom - rect.top ) -
                   ( orect.bottom - orect.top );
      crect->bottom += adjust;
   }

   MoveWindow( pLayout->hControl, crect->left, crect->top,
               crect->right - crect->left, crect->bottom - crect->top,
               TRUE );
}

void
LayoutManager::DoLayout( WORD width, WORD height )
{
   orect.left = rect.left;
   orect.right = rect.right;
   orect.top = rect.top;
   orect.bottom = rect.bottom;

   rect.right = rect.left + width;
   rect.bottom = rect.top + height;

   bit = !bit;

   for( LAYOUT_MAP::iterator it = components.begin();
        it != components.end();
        it++ )
      DoLayout( (*it).second );
}

