/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: lwmcomm.cxx,v 1.80 2001/04/05 18:40:22 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Routines to communicate with wrapped processes' local wrapper
 *          manager instances
 */

#include <wr_conf.h>

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <process.h>
#include <tchar.h>
#include <Lmcons.h>

#include "service.h"
#include "gwm.h"
#include "gwm_comm.h"
#include "debugwin.h"
#include "resource.h"

#include <wss_malloc.h>
#include <wss_module.h>
#include <wss_process.h>
#include <wss_wrapper.h>
#include <wss_wrvevent.h>
#include <wss_audit.h>
#include <wss_vm.h>
#include <wss_db.h>
#include <wsi_exec.h>
#include <wrerrno.h>

#include <stack>


typedef struct _GWM_LWM_PACKET_DISPATCH GWM_LWM_PACKET_DISPATCH,
                                        *LPGWM_LWM_PACKET_DISPATCH;

struct _GWM_LWM_PACKET_DISPATCH
{
   PipeSrvCommDispatchFunc fDispatch;
   LPTSTR                  tcDescription;
};


GWM_LWM_PACKET_DISPATCH g_packetDispatch[] =
{
   { LWMComm_RequestWrappers, _TEXT("Request wrappers") },
   { LWMComm_Message,         _TEXT("Message") },
   { LWMComm_CreateTable,     _TEXT("Create DB table") },
   { LWMComm_DropTable,       _TEXT("Drop DB table") },
   { LWMComm_InsertTableRow,  _TEXT("Insert row into DB table") },
   { LWMComm_SelectTableRows, _TEXT("Select row(s) from DB table") },
   { LWMComm_DeleteTableRows, _TEXT("Delete row(s) from DB table") },
   { LWMComm_UpdateTableRows, _TEXT("Update row(s) in DB table") },
   { LWMComm_RequestTable,    _TEXT("Request DB table header info") },
   { LWMComm_CreatingChild,   _TEXT("Process spawning a child") },
   { LWMComm_CreatedChild,    _TEXT("Process completed child creation") },
   { LWMComm_FindTable,       _TEXT("Find table identifier by name") },
   { LWMComm_SyncResend,      _TEXT("Resend last response packet") },
   { LWMComm_AuditPost,       _TEXT("Post an Audit event") }
};

#define LWMCOMM_DISPATCH_TABLE_SIZE ( sizeof(g_packetDispatch) / \
                                      sizeof(GWM_LWM_PACKET_DISPATCH) )


// Local process flag to indicate that a process has local DB tables
//    associated with it
#define WRPF_LOCAL_DBACCESS WRPF_LOCAL

extern BOOL g_bDebugging;

extern "C" {
WSS_MUTEX_EXPORT(db);
}

class DebugWin    g_lwmcommDebugWin;
DWORD             g_dwPendingChildren = 0;
WSS_MUTEX_PROTECT(g_dwPendingChildren);
GWM_PROC_PENDING  g_procsPendingMap;
WSS_MUTEX_PROTECT(g_procsPendingMap);


/////////////////////////////////////////////////////////////////////////////
// private functions
//
static void
_GWMLWMPostDropTableEvent( ULONG tableId )
{
   wss_proc_t               *wrp;
   PIPESRV_COMM_RESP_HEADER  commHdr;
   DWORD                     dwId;
   
   dwId = GetTickCount() << GWM_COMM_ID_SHIFT;
   commHdr.commandId = LWM_COMM_EVENT_DROP_TABLE | ( dwId & GWM_COMM_ID_MASK );
   commHdr.pktlen    = sizeof(commHdr) + sizeof(ULONG);
   commHdr.errorCode = 0;

   WSS_MUTEX_ENTER( wss_proc_list );
   for( wrp = wss_proc_list.lh_first; wrp; wrp = wrp->wrp_p_list.le_next )
   {
      commHdr.pid = wrp->wrp_p->p_pid;
      if( (LPPIPESRV_PIPE_INFO)WSI_PROC_OPAQUE(wrp)->hLWMPipe != NULL )
         GWMLWMCommWritePacket( (LPPIPESRV_PIPE_INFO)WSI_PROC_OPAQUE(wrp)->hLWMPipe,
                                &commHdr, (PVOID)&tableId );
   }
   WSS_MUTEX_EXIT( wss_proc_list );
}

static void
_GWMLWMEventHook( const pd_char_t *group, u_long types, size_t length,
                  const void *buffer)
{
    if (types & WREV_DROPTAB) {
        u_long key = *((u_long*)buffer);
        _GWMLWMPostDropTableEvent( key );
    }
}

static void
_GWMLWMCommConvertRelExprValToAbs( pd_byte_t *base, sxval_t val )
{
   if( val )
   {
      switch( val->which )
      {
         case literal:
            if( ( val->dtype == pd_char_t_t ) && val->v.l.cval )
               val->v.l.cval = (pd_char_t *)( base + (DWORD)val->v.l.cval );
            break;

         case field:
            break;

         case variable:
            // Not implemented yet.
            break;
      }
   }
}

static void
_GWMLWMCommConvertRelScritToAbs( pd_byte_t *base, wqsx_t scrit )
{
   if( scrit )
   {
      switch( scrit->which )
      {
         case term:
            scrit->un.exp = (tx_t)( base + (DWORD)scrit->un.exp );
            if( scrit->un.exp->left )
            {
               scrit->un.exp->left = (sxval_t)( base +
                                                (DWORD)scrit->un.exp->left );
               _GWMLWMCommConvertRelExprValToAbs( base, scrit->un.exp->left );
            }
            if( scrit->un.exp->right )
            {
               scrit->un.exp->right = (sxval_t)( base +
                                                 (DWORD)scrit->un.exp->right );
               _GWMLWMCommConvertRelExprValToAbs( base, scrit->un.exp->right );
            }

            break;

         case nonterm:
            scrit->un.children = (ntx_t)( base + (DWORD)scrit->un.children );
            if( scrit->un.children->left )
            {
               scrit->un.children->left =
                  (wqsx_t)( base + (DWORD)scrit->un.children->left );
               _GWMLWMCommConvertRelScritToAbs( base,
                                                scrit->un.children->left );
            }
            if( scrit->un.children->right )
            {
               scrit->un.children->right =
                  (wqsx_t)( base + (DWORD)scrit->un.children->right );
               _GWMLWMCommConvertRelScritToAbs( base,
                                                scrit->un.children->right );
            }

            break;
      }
   }
}

static void
_GWMLWMCommConvertRelUpdateExprToAbs( pd_byte_t *base, wquexp_t expr )
{
   if( expr )
   {
      switch( expr->which )
      {
         case term:
            if( expr->w.terminal )
            {
               expr->w.terminal = (sxval_t)( base + (DWORD)expr->w.terminal );
               _GWMLWMCommConvertRelExprValToAbs( base, expr->w.terminal );
            }
            break;

         case nonterm:
            if( expr->w.e.left )
            {
               expr->w.e.left = (wquexp_t)( base + (DWORD)expr->w.e.left );
               _GWMLWMCommConvertRelUpdateExprToAbs( base, expr->w.e.left );
            }
            if( expr->w.e.right )
            {
               expr->w.e.right = (wquexp_t)( base + (DWORD)expr->w.e.right );
               _GWMLWMCommConvertRelUpdateExprToAbs( base, expr->w.e.right );
            }

            break;
      }
   }
}

static void
_GWMLWMCommConvertRelSetScriptToAbs( pd_byte_t *base, wqusl_t script )
{
   while( script )
   {
      if( script->left )
      {
         script->left = (sxval_t)( base + (DWORD)script->left );
         _GWMLWMCommConvertRelExprValToAbs( base, script->left );
      }
      if( script->right )
      {
         script->right = (wquexp_t)( base + (DWORD)script->right );
         _GWMLWMCommConvertRelUpdateExprToAbs( base, script->right );
      }

      script = (wqusl_t)( script->next ? ( base + (DWORD)script->next )
                                       : NULL );
   }
}

static BOOL
_GWMLWMCommCreateProcess( LPPIPESRV_PIPE_INFO pInfo, proc_t *p, pd_char_t *cwd,
                          pd_char_t *path )
{
   LPPIPESRV_COMM_RESP_HEADER  pCommHdr   = &(pInfo->commHdr);
   wss_proc_t                 *wrp        = NULL;
   wss_proc_t                 *wrpParent  = NULL;
   proc_t                     *pParent    = NULL;
   BOOL                        bNewParent = FALSE;
   PVOID                       data       = NULL;

   if( p == NULL )
      return FALSE;

   if( ( wrpParent = wss_pfind( p->p_ppid ) ) == NULL )
   {
      pParent = (proc_t *)malloc( sizeof(proc_t) );
      pParent->p_pid = 0; // TBD: find out the parent pid
      pParent->p_ppid = 0;
      bNewParent = TRUE;
   }
   else
   {
      pParent = wrpParent->wrp_p;
      wss_proc_rele( wrpParent );
   }

   wss_proc_fork( pParent, p, 0 );

   if( bNewParent )
      free( pParent );

   (void) wsi_exec( p, (pd_char_t *)cwd, (pd_char_t *)path );

   if( cwd )
      wss_free( cwd );

   if( path )
      wss_free( path );

   // determine wrappers needed by this process
   wrp = wss_pfind( p->p_pid );
   if( wrp )
   {
      g_lwmcommDebugWin.NewProcess( p->p_pid );

      int wrapperSize = sizeof(DWORD);

      WSI_PROC_OPAQUE(wrp)->hLWMPipe = (HANDLE)pInfo;

      WSS_MUTEX_ENTER( wrp->wrp_wi_list );
      for( wss_wrapper_instance_t *wi =
              (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
           wi;
           wi = wi->wi_next )
         wrapperSize += ( wsi_strlen( wi->wi_wrapper->w_module->wrm_linked->linker_file_path ) + 1 ) *
                        sizeof(pd_char_t);
      WSS_MUTEX_EXIT( wrp->wrp_wi_list );

      if( wrapperSize > sizeof(DWORD) )
         wrapperSize += sizeof(pd_char_t);
      wrapperSize += ((UNLEN + 1) * sizeof( pd_char_t ));
      if( ( data = malloc( wrapperSize ) ) != NULL )
      {
         pd_byte_t *ptr = (pd_byte_t *)data;
         memset( data, 0, wrapperSize );
         wsi_strcpy( (pd_char_t *)ptr, p->p_user_name);
         ptr += ((UNLEN + 1) * sizeof( pd_char_t ));
         ProcDebugNode *pNode = g_lwmcommDebugWin.FindProc( p->p_pid );
         if( pNode )
            *((DWORD *)ptr) = pNode->GetFlags();
         ptr += sizeof(DWORD);

         if( wrapperSize > sizeof(DWORD) )
         {
            WSS_MUTEX_ENTER( wrp->wrp_wi_list );
            for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
                 wi;
                 wi = wi->wi_next )
            {
               int len = wsi_strlen( wi->wi_wrapper->w_module->wrm_linked->linker_file_path ) + 1;

               if( pInfo->commHdr.flags & GWM_COMM_UNICODE )
                  memcpy( ptr, wi->wi_wrapper->w_module->wrm_linked->linker_file_path,
                          len * sizeof(pd_char_t) );
               else
                  WideCharToMultiByte( CP_ACP, NULL,
                                       wi->wi_wrapper->w_module->wrm_linked->linker_file_path, 
                                       len, (char *)ptr, 
                                       len * sizeof(pd_char_t),
                                       NULL, NULL );

               ptr += ( len * sizeof(pd_char_t) );
            }
            WSS_MUTEX_EXIT( wrp->wrp_wi_list );
         }

         pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER) + wrapperSize;
      }

      wss_proc_rele( wrp );
   }

   if( data == NULL )
   {
      pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);
      pCommHdr->errorCode = -WRE_ERROR;
   }
   else
      pCommHdr->errorCode = 0;

   pCommHdr->flags = pInfo->commHdr.flags & GWM_COMM_UNICODE;

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_RequestWrappers: sending response\n") );

   // send response to waiting process
   BOOL bRet = GWMLWMCommWritePacket( pInfo, pCommHdr, data );

   if( data )
      free( data );

//   if( wrp )
//      g_lwmcommDebugWin.PostMessage( p->p_pid, 0, PD_MAKESTR("Process Requested Wrappers\n") );

   return bRet;
}


/////////////////////////////////////////////////////////////////////////////
// public functions
//

void
GWMLWMCommSetup()
{
   wss_wrvevent_init();

   wss_wrvevent_register_hook( _GWMLWMEventHook );

   WSS_MUTEX_INIT(g_dwPendingChildren);
   WSS_MUTEX_INIT(g_procsPendingMap);


   HICON hIcon = (HICON) LoadImage( GetModuleHandle( NULL ),
                                    MAKEINTRESOURCE(IDI_WRAPPERS),
                                    IMAGE_ICON, 16, 16, 0 );

   g_lwmcommDebugWin.TrayIcon( hIcon,
                               PD_MAKESTR("Global Wrapper Manager") );
}

//
// GWMProcessLWMComm
//
// Process LWM communication packets from passed pipe
//
// Peeks the pipe to see if any data is available, if so, reads a packet
// If there is no data available, this pipe will be returned to the
// active queue.
//
BOOL
GWMLWMCommProcessPacket( LPPIPESRV_PIPE_INFO pInfo )
{
   BOOL bRet;
          
   bRet = GWMLWMCommReadPacket( pInfo );
   if( !bRet )
      return TRUE;

#ifdef DEBUG_LWM_PACKET
   if( g_bDebugging )
      _tprintf( _TEXT("ProcessPacket: ID = 0x%x, Command = 0x%x\n"),
                GWM_COMM_ID(pInfo->commHdr),
                GWM_COMM_CMD(pInfo->commHdr) );
#endif /* DEBUG_LWM_PACKET */

   // process packet
   //
   if( GWM_COMM_CMD(pInfo->commHdr) >= LWMCOMM_DISPATCH_TABLE_SIZE )
   {
      if( g_bDebugging )
         _tprintf( _TEXT("GWMProcessLWMComm: invalid packet encountered\n") );
   }
   else
   {
      if( g_bDebugging &&
          ( GWM_COMM_CMD(pInfo->commHdr) != LWM_COMM_MESSAGE ) )
      {
         _tprintf( _TEXT("Processing: %s\n"),
                   g_packetDispatch[GWM_COMM_CMD(pInfo->commHdr)].tcDescription );
      }

      bRet = (*g_packetDispatch[GWM_COMM_CMD(pInfo->commHdr)].fDispatch)( pInfo );
      if( !bRet )
         return TRUE;
   }

   return FALSE;
}

BOOL
GWMLWMCommReadPacket( LPPIPESRV_PIPE_INFO pInfo )
{
   OVERLAPPED os;
   DWORD      dwWait;
   HANDLE     hEvents[2];
   DWORD      cbRead;
   BOOL       bRet;
   
   hEvents[0] = pInfo->hEvents[0];
   hEvents[1] = pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT];

   WSS_MUTEX_ENTER( pInfo->readPipe );

   // init the overlapped structure
   //
   memset( &os, 0, sizeof(OVERLAPPED) );
   os.hEvent = hEvents[1];
   ResetEvent( hEvents[1] );

   // clear packet header
   //
   memset( &(pInfo->commHdr), 0, sizeof(PIPESRV_COMM_HEADER) );

   // grab packet header
   //
   bRet = ReadFile( pInfo->hPipe, &(pInfo->commHdr), sizeof(PIPESRV_COMM_HEADER),
                    &cbRead, &os );

   if( !bRet )
   {
      switch( GetLastError() )
      {
         case ERROR_IO_PENDING:
            dwWait = WaitForMultipleObjects( 2, hEvents, FALSE, INFINITE );
            if( ( dwWait != WAIT_OBJECT_0 + 1 ) ||
                ( pInfo->commHdr.pktlen == 0 ) )
            {
               // not overlapped i/o event
               // - error occured or server stop signaled
               WSS_MUTEX_EXIT( pInfo->readPipe );
               return FALSE;
            }
            break;

         case ERROR_HANDLE_EOF:
#ifdef DEBUG_IO
            if( g_bDebugging )
               _tprintf( _TEXT("EOF reached\n") );
#endif /* DEBUG_IO */

            WSS_MUTEX_EXIT( pInfo->readPipe );
            return FALSE;

         default:
#ifdef DEBUG_IO
            if( g_bDebugging )
               _tprintf( _TEXT("Error: %ld\n"), GetLastError() );
#endif /* DEBUG_IO */

            WSS_MUTEX_EXIT( pInfo->readPipe );
            return FALSE;
      }
   }

   // grab data
   //
   if( pInfo->commHdr.pktlen > sizeof(PIPESRV_COMM_HEADER) )
   {
      DWORD buflen = pInfo->commHdr.pktlen - sizeof(PIPESRV_COMM_HEADER);

      if( pInfo->pBuf )
         pInfo->pBuf = realloc( pInfo->pBuf, buflen );
      else
         pInfo->pBuf = malloc( buflen );
      
      // init the overlapped structure
      //
      memset( &os, 0, sizeof(OVERLAPPED) );
      os.hEvent = hEvents[1];
      ResetEvent( hEvents[1] );

      // clear packet data buffer
      //
      memset( pInfo->pBuf, 0, buflen );

      bRet = ReadFile( pInfo->hPipe, pInfo->pBuf, buflen, &cbRead, &os );

      if( !bRet )
      {
         switch( GetLastError() )
         {
            case ERROR_IO_PENDING:
               dwWait = WaitForMultipleObjects( 2, hEvents, FALSE, INFINITE );
               if( dwWait != WAIT_OBJECT_0 + 1 )
               {
                  // not overlapped i/o event
                  // - error occured or server stop signaled
                  WSS_MUTEX_EXIT( pInfo->readPipe );
                  return FALSE;
               }
               break;

            case ERROR_HANDLE_EOF:
#ifdef DEBUG_IO
               if( g_bDebugging )
                  _tprintf( _TEXT("EOF reached\n") );
#endif /* DEBUG_IO */

               WSS_MUTEX_EXIT( pInfo->readPipe );
               return FALSE;

            default:
#ifdef DEBUG_IO
               if( g_bDebugging )
                  _tprintf( _TEXT("Error: %ld\n"), GetLastError() );
#endif /* DEBUG_IO */

               WSS_MUTEX_EXIT( pInfo->readPipe );
               return FALSE;
         }
      }
   }

   WSS_MUTEX_EXIT( pInfo->readPipe );
   return TRUE;
}

BOOL
GWMLWMCommWritePacket( LPPIPESRV_PIPE_INFO pInfo,
                       LPPIPESRV_COMM_RESP_HEADER pCommHdr,
                       PVOID pData )
{
   OVERLAPPED os;
   DWORD      dwWait;
   HANDLE     hEvents[2];
   DWORD      dwDatalen;
   DWORD      cbWritten;
   BOOL       bRet;
   
   dwDatalen = pCommHdr->pktlen - sizeof(*pCommHdr);

   hEvents[0] = pInfo->hEvents[0];
   hEvents[1] = pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT];

   WSS_MUTEX_ENTER( pInfo->writePipe );

   // init the overlapped structure
   //
   memset( &os, 0, sizeof(OVERLAPPED) );
   os.hEvent = hEvents[1];
   ResetEvent( hEvents[1] );

   // send out the header...
   //
   bRet = WriteFile( pInfo->hPipe, pCommHdr, sizeof(*pCommHdr), &cbWritten,
                     &os );

   if( !bRet )
   {
      switch( GetLastError() )
      {
         case ERROR_IO_PENDING:
#ifdef DEBUG_IO
            if( g_bDebugging )
               _tprintf( _TEXT("Waiting on WriteFile(3)...\n") );
#endif /* DEBUG_IO */

            dwWait = WaitForMultipleObjects( 2, hEvents, FALSE, INFINITE );
            if( dwWait != WAIT_OBJECT_0 + 1 )
            {
               // not overlapped i/o event
               // - error occured or server stop signaled
               WSS_MUTEX_EXIT( pInfo->writePipe );
               return FALSE;
            }
            break;

         default:
#ifdef DEBUG_IO
            if( g_bDebugging )
               _tprintf( _TEXT("WriteFile(3) Error: %ld\n"), GetLastError() );
#endif /* DEBUG_IO */

            WSS_MUTEX_EXIT( pInfo->writePipe );
            return FALSE;
      }
   }

#ifdef DEBUG_IO
   if( g_bDebugging )
      _tprintf( _TEXT("WriteFile(3): wrote %ld bytes\n"), sizeof(*pCommHdr) );
#endif /* DEBUG_IO */

   // send out the data...
   //
   if( dwDatalen > 0 )
   {
      bRet = WriteFile( pInfo->hPipe, pData, dwDatalen, &cbWritten, &os );

      if( !bRet )
      {
         switch( GetLastError() )
         {
            case ERROR_IO_PENDING:
#ifdef DEBUG_IO
               if( g_bDebugging )
                  _tprintf( _TEXT("Waiting on WriteFile(4)...\n") );
#endif /* DEBUG_IO */
               dwWait = WaitForMultipleObjects( 2, hEvents, FALSE, INFINITE );
               if( dwWait != WAIT_OBJECT_0 + 1 )
               {
                  // not overlapped i/o event
                  // - error occured or server stop signaled
                  WSS_MUTEX_EXIT( pInfo->writePipe );
                  return FALSE;
               }
               break;

            default:
#ifdef DEBUG_IO
               if( g_bDebugging )
                  _tprintf( _TEXT("WriteFile(4) Error: %ld\n"), GetLastError() );
#endif /* DEBUG_IO */

               WSS_MUTEX_EXIT( pInfo->writePipe );
               return FALSE;
         }
      }
#ifdef DEBUG_IO
      if( g_bDebugging )
         _tprintf( _TEXT("WriteFile(4): wrote %ld bytes\n"), dwDatalen );
#endif /* DEBUG_IO */
   }
   else
   {
#ifdef DEBUG_IO
      if( g_bDebugging )
         _tprintf( _TEXT("WriteFile(4) no data, no write.\n") );
#endif /* DEBUG_IO */
   }

   // process cached response
   if( ( pCommHdr->flags & GWM_COMM_RESENT ) == 0 )
   {
      LPPIPESRV_LAST_PACKET_DATA pPacket;
      WSS_MUTEX_ENTER( pInfo->pLastPackets );
      PIPESRV_LAST_PACKETS::iterator it = pInfo->pLastPackets->find( pCommHdr->tid );
      if( it == pInfo->pLastPackets->end() )
         pPacket = new PIPESRV_LAST_PACKET_DATA;
      else
      {
         // Use old packet cache data structure
         pPacket = (*it).second;
         pInfo->pLastPackets->erase( it );

         if( pPacket->pCommHdr )
         {
            free( pPacket->pCommHdr );
            pPacket->pCommHdr = NULL;
         }
         if( pPacket->pData )
         {
            free( pPacket->pData );
            pPacket->pData = NULL;
         }
      }

      if( pPacket )
      {
         pPacket->pCommHdr =
            (LPPIPESRV_COMM_RESP_HEADER) malloc( sizeof(*pCommHdr) );
         if( pPacket->pCommHdr )
         {
            memcpy( pPacket->pCommHdr, pCommHdr, sizeof(pCommHdr) );
            if( pData )
            {
               pPacket->pData = malloc( dwDatalen );
               if( pPacket->pData )
               {
                  memcpy( pPacket->pData, pData, dwDatalen );
                  pInfo->pLastPackets->insert( PIPESRV_LAST_PACKETS::value_type(pCommHdr->tid, pPacket) );
               }
               else
               {
                  free( pPacket->pCommHdr );
                  free( pPacket );
               }
            }
            else
               pPacket->pData = NULL;
         }
         else
            free( pPacket );
      }
      WSS_MUTEX_EXIT( pInfo->pLastPackets );
   }

   WSS_MUTEX_EXIT( pInfo->writePipe );
   return TRUE;
}

void
GWMLWMCommCloseCallback( LPPIPESRV_PIPE_INFO pInfo )
{
   wss_proc_t *wrp;

   WSS_MUTEX_ENTER( wss_proc_list );
   for( wrp = wss_proc_list.lh_first; wrp; wrp = wrp->wrp_p_list.le_next )
   {
      if( WSI_PROC_OPAQUE(wrp)->hLWMPipe == (HANDLE)pInfo )
      {
         WSS_MUTEX_EXIT( wss_proc_list );
#ifndef GWM_POST_MESSAGE_ALWAYS
         if( g_bDebugging )
#endif /* GWM_POST_MESSAGE_ALWAYS */
            g_lwmcommDebugWin.Flush( wrp->wrp_p->p_pid );

         WSS_MUTEX_ENTER(g_dwPendingChildren);
         DWORD dwPendingChildren = WSI_PROC_OPAQUE(wrp)->dwPendingChildren;
         WSI_PROC_OPAQUE(wrp)->dwPendingChildren = 0;
         g_dwPendingChildren -= dwPendingChildren;

         if( !g_dwPendingChildren )
         {
            WSS_MUTEX_ENTER(g_procsPendingMap);
            for( GWM_PROC_PENDING::iterator it = g_procsPendingMap.begin();
                 it != g_procsPendingMap.end();
                 it++ )
            {
               LPGWM_PROC_PENDING_DATA pPendData = (*it).second;

               // continue child creation
               _GWMLWMCommCreateProcess( pPendData->pPipe,
                                         pPendData->p,
                                         pPendData->cwd,
                                         pPendData->path );

               wss_free( pPendData );                  
            }

            // remove child pending from map
            g_procsPendingMap.clear();

            WSS_MUTEX_EXIT(g_procsPendingMap);
         }
         WSS_MUTEX_EXIT(g_dwPendingChildren);

         int pid = wrp->wrp_p->p_pid;
         BOOL bHasLocals = (wrp->wrp_flags & WRPF_LOCAL_DBACCESS) ? TRUE
                                                                  : FALSE;
         wss_proc_exit( wrp->wrp_p );

         // clean up cached responses
         WSS_MUTEX_ENTER( pInfo->pLastPackets );
         PIPESRV_LAST_PACKETS::iterator it = pInfo->pLastPackets->begin();
         while( it != pInfo->pLastPackets->end() )
         {
            LPPIPESRV_LAST_PACKET_DATA pPacket = (*it).second;

            pInfo->pLastPackets->erase( it );
            if( pPacket->pCommHdr )
            {
               free( pPacket->pCommHdr );
               pPacket->pCommHdr = NULL;
            }
            if( pPacket->pData )
            {
               free( pPacket->pData );
               pPacket->pData = NULL;
            }
            delete pPacket;
            it = pInfo->pLastPackets->begin();
         }
         WSS_MUTEX_EXIT( pInfo->pLastPackets );

         if( bHasLocals )
         {
            // walk through the database and drop all local tables
            wqtab_t tab;

            std::stack < u_long > tableStack;

            WSS_MUTEX_ENTER(db);
            tab = (wqtab_t)DB_REL2ABS(DBHEAD->db_tables);

            while( tab )
            {
               pd_char_t *name;

               wqtab_t atab = (wqtab_t)DB_REL2ABS(tab);

               for( name = DB_REL2ABS(atab->name);
                    *name && (*name != '.');
                    name++ )
                  continue;

               if( *name )
               {
                  pd_char_t *pidstr = ++name;
                  while( *name && (*name != '.') &&
                         (*name >= '0' || *name <= '9') )
                     name++;
                  if( *name == '.' )
                  {
                     pd_char_t pidval[MAX_PATH];

                     DWORD len = (DWORD)name - (DWORD)pidstr;

                     if( len < ( MAX_PATH * sizeof(pd_char_t) ) )
                     {
                        memcpy( pidval, pidstr, len );
                        len /= sizeof(pd_char_t);
                        pidval[len] = PD_MAKESTR('\0');

                        int tabpid = _ttoi(pidval);
                        if( tabpid == pid )
                           tableStack.push( (u_long)tab );
                     }
                  }
               }

               tab = atab->next;
            }
            WSS_MUTEX_EXIT(db);

            while( !tableStack.empty() )
            {
               wss_drop( tableStack.top() );
               tableStack.pop();
            }
         }

         return;
      }
   }
   WSS_MUTEX_EXIT( wss_proc_list );
}

BOOL
LWMCommStringIsValid( PVOID buffer, DWORD length, BOOL unicode )
{
   if( unicode )
   {
      pd_char_t *last = (pd_char_t *)buffer +
                         ( length / sizeof(pd_char_t) ) - 1;

      // Check for misalignment and non-null-terminated string
      if( ( length % sizeof(pd_char_t) ) || *last )
         return FALSE;
   }
   else if( *((PCHAR)buffer + length - 1) )
      return FALSE;

   return TRUE;
}

void
GWMLWMCommSendTraceCallsEvent( int pid, BOOL state )
{
   wss_proc_t *wrp = wss_pfind( pid );

   if( wrp )
   {
      PIPESRV_COMM_RESP_HEADER  commHdr;
   
      DWORD dwId = GetTickCount() << GWM_COMM_ID_SHIFT;
      commHdr.commandId = LWM_COMM_DEBUG_TRACE_CALLS |
                          ( dwId & GWM_COMM_ID_MASK );
      commHdr.pid       = wrp->wrp_p->p_pid;
      commHdr.pktlen    = sizeof(commHdr) + sizeof(BOOL);
      commHdr.errorCode = 0;

      LPPIPESRV_PIPE_INFO pPipe = (LPPIPESRV_PIPE_INFO)WSI_PROC_OPAQUE(wrp)->hLWMPipe;
      wss_proc_rele( wrp );

      if( pPipe )
         GWMLWMCommWritePacket( pPipe, &commHdr, (PVOID)&state );
   }
}

void
GWMLWMCommSendCountCallsEvent( int pid, BOOL state )
{
   wss_proc_t *wrp = wss_pfind( pid );

   if( wrp )
   {
      PIPESRV_COMM_RESP_HEADER  commHdr;
   
      DWORD dwId = GetTickCount() << GWM_COMM_ID_SHIFT;
      commHdr.commandId = LWM_COMM_DEBUG_COUNT_CALLS |
                          ( dwId & GWM_COMM_ID_MASK );
      commHdr.pid       = wrp->wrp_p->p_pid;
      commHdr.pktlen    = sizeof(commHdr) + sizeof(BOOL);
      commHdr.errorCode = 0;

      LPPIPESRV_PIPE_INFO pPipe = (LPPIPESRV_PIPE_INFO)WSI_PROC_OPAQUE(wrp)->hLWMPipe;
      wss_proc_rele( wrp );

      if( pPipe )
         GWMLWMCommWritePacket( pPipe, &commHdr, (PVOID)&state );
   }
}

void
GWMLWMCommSendThreadMessagesEvent( int pid, BOOL state )
{
   wss_proc_t *wrp = wss_pfind( pid );

   if( wrp )
   {
      PIPESRV_COMM_RESP_HEADER  commHdr;
   
      DWORD dwId = GetTickCount() << GWM_COMM_ID_SHIFT;
      commHdr.commandId = LWM_COMM_DEBUG_THR_MESSAGES |
                          ( dwId & GWM_COMM_ID_MASK );
      commHdr.pid       = wrp->wrp_p->p_pid;
      commHdr.pktlen    = sizeof(commHdr) + sizeof(BOOL);
      commHdr.errorCode = 0;

      LPPIPESRV_PIPE_INFO pPipe = (LPPIPESRV_PIPE_INFO)WSI_PROC_OPAQUE(wrp)->hLWMPipe;
      wss_proc_rele( wrp );

      if( pPipe )
         GWMLWMCommWritePacket( pPipe, &commHdr, (PVOID)&state );
   }
}

void
GWMLWMCommSendProcessMessagesEvent( int pid, BOOL state )
{
   wss_proc_t *wrp = wss_pfind( pid );

   if( wrp )
   {
      PIPESRV_COMM_RESP_HEADER  commHdr;
   
      DWORD dwId = GetTickCount() << GWM_COMM_ID_SHIFT;
      commHdr.commandId = LWM_COMM_DEBUG_PROC_MESSAGES |
                          ( dwId & GWM_COMM_ID_MASK );
      commHdr.pid       = wrp->wrp_p->p_pid;
      commHdr.pktlen    = sizeof(commHdr) + sizeof(BOOL);
      commHdr.errorCode = 0;

      LPPIPESRV_PIPE_INFO pPipe = (LPPIPESRV_PIPE_INFO)WSI_PROC_OPAQUE(wrp)->hLWMPipe;
      wss_proc_rele( wrp );

      if( pPipe )
         GWMLWMCommWritePacket( pPipe, &commHdr, (PVOID)&state );
   }
}


/////////////////////////////////////////////////////////////////////////////
// dispatch functions
//
BOOL
LWMComm_RequestWrappers( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER  pCommHdr   = &(pInfo->commHdr);
   proc_t                     *p          = (proc_t *)malloc( sizeof(proc_t) );
   int                         pathSize   = 0;
   LPTSTR                      cwd        = NULL;
   LPTSTR                      path       = NULL;

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_RequestWrappers\n") );

   p->p_pid = pCommHdr->pid;
   p->p_ppid = 0; // TBD: find out the parent pid

   // Impersonate the user on the other end of the pipe
   // and get their ID
   TCHAR *lpszUserName;   
   DWORD  dwUserNameLength = 250;

   lpszUserName = (TCHAR *)malloc(dwUserNameLength);
   ImpersonateNamedPipeClient( pInfo->hPipe );
   GetUserName(lpszUserName,&dwUserNameLength);
   RevertToSelf();
   
   p->p_user_name=lpszUserName;

   // process packet if there's data
   if( pCommHdr->pktlen > sizeof(PIPESRV_COMM_HEADER) )
   {
      int dataSize = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);

      pd_char_t *cwdAndPath = (pd_char_t *)pInfo->pBuf;

      if( LWMCommStringIsValid( pInfo->pBuf, dataSize,
                                   (pInfo->commHdr.flags &
                                      GWM_COMM_UNICODE) ? TRUE : FALSE ) )
      {

         if( !(pInfo->commHdr.flags & GWM_COMM_UNICODE) )
         {
            // Convert ASCII string to Unicode
            pInfo->pBuf = malloc( dataSize * sizeof(pd_char_t) );
            MultiByteToWideChar( CP_ACP, 0,
                                 (LPCSTR)cwdAndPath, dataSize,
                                 (pd_char_t *)pInfo->pBuf,
                                 dataSize * sizeof(pd_char_t) );
            free( cwdAndPath );
            cwdAndPath = (pd_char_t *)pInfo->pBuf;
            dataSize *= sizeof(pd_char_t);
         }

         // determine path size
         int pathSize = ( wsi_strlen( cwdAndPath ) + 1 ) *
                        sizeof(pd_char_t);

         if( pathSize < dataSize )
         {
            // copy path from data buffer
            path = (LPTSTR)wss_malloc( pathSize );
            memcpy( path, cwdAndPath, pathSize );

            // copy cwd from data buffer
            cwd = (LPTSTR)wss_malloc( dataSize - pathSize );
            memcpy( cwd, (BYTE *)cwdAndPath + pathSize,
                    dataSize - pathSize );


         }
      }
   }

   // if we have no cwd, make cwd null string
   if( cwd == NULL )
   {
      cwd = (LPTSTR)wss_malloc( sizeof(TCHAR) );
      memset( cwd, 0, sizeof(TCHAR) );
   }

   // if we have no path, make path null string
   if( path == NULL )
   {
      path = (LPTSTR)wss_malloc( sizeof(TCHAR) );
      memset( path, 0, sizeof(TCHAR) );
   }

   WSS_MUTEX_ENTER(g_dwPendingChildren);
   DWORD                   dwPending = g_dwPendingChildren;
   LPGWM_PROC_PENDING_DATA pPendData = NULL;
   if( dwPending )
   {
      WSS_MUTEX_ENTER(g_procsPendingMap);

      GWM_PROC_PENDING::iterator it = g_procsPendingMap.find( p->p_pid );
      if( it == g_procsPendingMap.end() )
      {
         pPendData = (LPGWM_PROC_PENDING_DATA) wss_malloc( sizeof(GWM_PROC_PENDING_DATA) );

         pPendData->pPipe = pInfo;
         pPendData->p = p;
         pPendData->cwd = cwd;
         pPendData->path = path;
         pPendData->sleep = TRUE;

         g_procsPendingMap.insert( GWM_PROC_PENDING::value_type( p->p_pid,
                                                                 pPendData ) );
      }
      else
      {
         dwPending = 0;
         pPendData = (*it).second;

         g_procsPendingMap.erase( it );

         wss_proc_t *wrpParent = wss_pfind( pPendData->ppid );
         if( wrpParent )
         {
            WSI_PROC_OPAQUE(wrpParent)->dwPendingChildren--;
            g_dwPendingChildren--;

            wss_proc_rele( wrpParent );
         }
      }
      WSS_MUTEX_EXIT(g_procsPendingMap);
   }
   WSS_MUTEX_EXIT(g_dwPendingChildren);

   BOOL bRet = TRUE;
   if( !dwPending )
   {
      if( pPendData )
      {
         p->p_ppid = pPendData->ppid;
         wss_free( pPendData );
      }

      bRet = _GWMLWMCommCreateProcess( pInfo, p, cwd, path );
   }
   return bRet;
}

BOOL
LWMComm_Message( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER  pCommHdr = &(pInfo->commHdr);
   int                         dataLen  = pCommHdr->pktlen -
                                          sizeof(PIPESRV_COMM_HEADER);
   DWORD                      *pTid     = (DWORD *)pInfo->pBuf;
   pd_char_t                  *msg      = (pd_char_t *)( pTid + 1 );
   int                         retval   = -1;

   if( !(pInfo->commHdr.flags & GWM_COMM_UNICODE) )
   {
      // Convert ASCII string to Unicode
      pInfo->pBuf = malloc( dataLen * sizeof(pd_char_t) );
      MultiByteToWideChar( CP_ACP, 0,
                           (LPCSTR)msg, dataLen,
                           (pd_char_t *)pInfo->pBuf,
                           dataLen * sizeof(pd_char_t) );
      free( msg );
      msg = (pd_char_t *)pInfo->pBuf;
   }

#ifndef GWM_POST_MESSAGE_ALWAYS
   if( g_bDebugging )
#endif /* GWM_POST_MESSAGE_ALWAYS */
      g_lwmcommDebugWin.PostMessage( pCommHdr->pid, *pTid, msg );

   return TRUE;
}

BOOL
LWMComm_CreateTable( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_CreateTable\n") );

   int dataLen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);
   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);

   // Make sure the data is large enough for a struct create_block
   if( dataLen < sizeof(struct create_block) )
   {
      int error = -WRE_INVALID_PARAMETER;
      return GWMLWMCommWritePacket( pInfo, pCommHdr, (PVOID)&error );
   }

   struct create_block *recvData = (struct create_block *)pInfo->pBuf;

   // Make sure the data is large enough for the names and column def
   //   structures
   if( dataLen < ( (UINT)recvData->cd + recvData->count * sizeof(wqcd) ) )
   {
      int error = -WRE_INVALID_PARAMETER;
      return GWMLWMCommWritePacket( pInfo, pCommHdr, (PVOID)&error );
   }

   // Check for null termination of last name string (if we have any
   // column defs -- we should, but we're not _that_ trusting)
   if( recvData->count &&
       ( *( (pd_char_t *)( (pd_byte_t *)recvData + dataLen -
                            sizeof(pd_char_t) ) ) != PD_MAKESTR('\0') ) )
   {
      int error = -WRE_INVALID_PARAMETER;
      return GWMLWMCommWritePacket( pInfo, pCommHdr, (PVOID)&error );
   }

   // Translate the relative pointers to absolute
   recvData->wname = (pd_char_t *)( (pd_byte_t *)recvData +
                                     (UINT)recvData->wname );
   recvData->tname = (pd_char_t *)( (pd_byte_t *)recvData +
                                     (UINT)recvData->tname );
   recvData->cd    = (wqcd_t)( (pd_byte_t *)recvData +
                               (UINT)recvData->cd );

   // Now check through all the column defs and make sure the data is
   //   large enough to contain all the names as it states
   for( int i = 0; i < recvData->count; i++ )
   {
      wqcd_t cd = recvData->cd + i;

      if( dataLen < (UINT)cd->name )
      {
         int error = -WRE_INVALID_PARAMETER;
         return GWMLWMCommWritePacket( pInfo, pCommHdr, (PVOID)&error );
      }

      // Looks okay for this one, translate relative to absolute
      cd->name = (pd_char_t *)( (pd_byte_t *)recvData +
                                 (UINT)cd->name );

   }

   // Well, we're relatively confident that we have a good data block, or
   // at least we shouldn't crash accessing the data members.  Let's get
   // to the character set conversion....
   if( !(pInfo->commHdr.flags & GWM_COMM_UNICODE) )
   {
      char *name     = (char *)recvData->wname;
      int   nameSize = strlen(name);

      // Convert ASCII string to Unicode
      recvData->wname = (pd_char_t *)wss_malloc( nameSize *
                                                  sizeof(pd_char_t) );
      MultiByteToWideChar( CP_ACP, 0, (LPCSTR)name, nameSize,
                           recvData->wname, nameSize * sizeof(pd_char_t) );

      name     = (char *)recvData->tname;
      nameSize = strlen(name);

      // Convert ASCII string to Unicode
      recvData->tname = (pd_char_t *)wss_malloc( nameSize *
                                                  sizeof(pd_char_t) );
      MultiByteToWideChar( CP_ACP, 0, (LPCSTR)name, nameSize,
                           recvData->tname, nameSize * sizeof(pd_char_t) );

      for( i = 0; i < recvData->count; i++ )
      {
         wqcd_t cd = recvData->cd + i;

         name     = (char *)cd->name;
         nameSize = strlen(name);

         // Convert ASCII string to Unicode
         cd->name = (pd_char_t *)wss_malloc( nameSize * sizeof(pd_char_t) );
         MultiByteToWideChar( CP_ACP, 0, (LPCSTR)name, nameSize,
                              cd->name, nameSize * sizeof(pd_char_t) );
      }
   }

   // Whew! That's all done, on with the table creation
   u_long id;
   pCommHdr->errorCode = wss_create( recvData->wname, recvData->instance,
                                     recvData->tname, FALSE, recvData->cd,
                                     recvData->count,
                                     &id );

   if( pCommHdr->errorCode == 0 )
   {
      BOOL local = FALSE;
      wqtab_t tab = wss_db_find_table_by_handle( id );
      if( tab->flags & DB_FLAG_LOCAL )
         local = TRUE;
      wss_db_unlock_table( tab );

      wss_proc_t *wrp = wss_pfind( pCommHdr->pid );
      if( wrp )
      {
         WSS_MUTEX_ENTER( wrp->wrp );
         wrp->wrp_flags |= WRPF_LOCAL_DBACCESS;
         WSS_MUTEX_EXIT( wrp->wrp );
         wss_proc_rele( wrp );
      }
   }

   // Free any allocated data if we converted from ASCII to Unicode
   if( !(pInfo->commHdr.flags & GWM_COMM_UNICODE) )
   {
      wss_free( recvData->wname );
      wss_free( recvData->tname );

      for( i = 0; i < recvData->count; i++ )
         wss_free( (recvData->cd + i)->name );
   }

   // Rend response packet
   pCommHdr->pktlen += sizeof(u_long);
   BOOL retcode = GWMLWMCommWritePacket( pInfo, pCommHdr, &id );

   return retcode;
}

BOOL
LWMComm_DropTable( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_DropTable\n") );

   int dataLen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);
   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);

   if( dataLen < ( sizeof(PIPESRV_COMM_HEADER) + sizeof(u_long) ) )
   {
      pCommHdr->errorCode = -WRE_ERROR;
   }
   else
   {
      u_long tableId = *((u_long *)pInfo->pBuf);

      pCommHdr->errorCode = wss_drop( tableId );
   }
   return GWMLWMCommWritePacket( pInfo, pCommHdr, NULL );
}

BOOL LWMComm_InsertTableRow( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_InsertTableRow\n") );

   struct insert_block *iBlock = (struct insert_block *)pInfo->pBuf;

   int dataLen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);
   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);

   // Find the specified table
   wqtab_t tab = wss_db_find_table_by_handle( iBlock->id );
   if( tab == NULL )
   {
      pCommHdr->errorCode = -WRE_ERROR;
      return GWMLWMCommWritePacket( pInfo, pCommHdr, NULL );
   }

   wqrow_t rows = (wqrow_t)(iBlock + 1);
   wqrow_t row;
   int     nrows;
   for( row = rows, nrows = 0;
        row && ( nrows < iBlock->rowCount );
        row = row->next, nrows++ )
   {
      row->next =
         (wqrow_t)( row->next ? ( (pd_byte_t *)iBlock + (DWORD)row->next )
                              : NULL );

      if( row->vals == NULL )
      {
         wss_db_unlock_table( tab );

         pCommHdr->errorCode = -WRE_ERROR;
         return GWMLWMCommWritePacket( pInfo, pCommHdr, NULL );
      }

      row->vals = (wqval_t)( (pd_byte_t *)iBlock + (DWORD)row->vals );
      for( int i = 0; i < tab->col_count; i++ )
      {
         wqcd_t  cd  = (wqcd_t)DB_REL2ABS(tab->col_defs) + i;
         wqval_t val = row->vals + i;

         if( ( cd->dtype == pd_char_t_t ) && val->v.cval )
            val->v.cval = (pd_char_t *)( (pd_byte_t *)iBlock +
                                          (DWORD)val->v.cval );
      }
   }

   wss_db_unlock_table( tab );

   pCommHdr->errorCode = wss_insert( iBlock->id, FALSE, rows );

   BOOL bRet = GWMLWMCommWritePacket( pInfo, pCommHdr, NULL );

   return bRet;
}

BOOL
LWMComm_SelectTableRows( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_SelectTableRows\n") );

   struct select_block *sBlock = (struct select_block *)pInfo->pBuf;

   int dataLen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);
   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);

   wqsx_t scrit = NULL;
   if (dataLen - sizeof(struct select_block)>0) {
     scrit = (wqsx_t)( sBlock + 1 );
     _GWMLWMCommConvertRelScritToAbs( (pd_byte_t *)sBlock, scrit );
   }

   wqtab new_table;

   int error = wss_select( sBlock->id, FALSE, scrit, sBlock->takeFirst,
                           sBlock->orderBy, sBlock->desc, &new_table );
   pCommHdr->errorCode = error;

   PVOID sendData = NULL;
   if( error > 0 )
   {
      // Determine memory necessary to hold packet data
      int reqMem = sizeof(wqtab) + new_table.col_count * sizeof(wqcd) +
                   new_table.row_count * sizeof(wqrow) +
                   new_table.row_count * new_table.col_count * sizeof(wqval);

      for( int i = 0; i < new_table.col_count; i++ )
      {
         wqcd_t cd = new_table.col_defs + i;
         reqMem += ( wsi_strlen( cd->name ) + 1 ) * sizeof(pd_char_t);
      }

      for( wqrow_t row = new_table.rows; row; row = row->next )
      {
         for( i = 0; i < new_table.col_count; i++ )
         {
            wqcd_t  cd  = new_table.col_defs + i;
            wqval_t val = row->vals + i;

            if( ( cd->dtype == pd_char_t_t ) && val->v.cval )
               reqMem += ( wsi_strlen( val->v.cval ) + 1 ) * sizeof(pd_char_t);
         }
      }

      // Convert returned table to packet data
      DWORD index = 0;
      sendData = wss_malloc( reqMem );

      wqtab_t tab = (wqtab_t)( (pd_byte_t *)sendData + index );
      memcpy( tab, &new_table, sizeof(wqtab) );
      index += sizeof(wqtab);
      tab->col_defs = (wqcd_t)index;

      memcpy( (pd_byte_t *)sendData + index, new_table.col_defs,
              new_table.col_count * sizeof(wqcd) );
      index += new_table.col_count * sizeof(wqcd);

      for( i = 0; i < new_table.col_count; i++ )
      {
         wqcd_t cd  = new_table.col_defs + i;
         wqcd_t ncd = (wqcd_t)( (pd_byte_t *)sendData +
                                (DWORD)tab->col_defs ) + i;

         ncd->name = (pd_char_t *)index;
         int len = ( wsi_strlen( cd->name ) + 1 ) * sizeof(pd_char_t);
         memcpy( (pd_byte_t *)sendData + index, cd->name, len );
         index += len;
      }

      wqrow_t *last = &(tab->rows);
      for( row = new_table.rows; row; row = row->next )
      {
         wqrow_t nrow = (wqrow_t)( (pd_byte_t *)sendData + index );
         *last = (wqrow_t)index;
         last = &nrow->next;
         index += sizeof(wqrow);

         memcpy( nrow, row, sizeof(wqrow) );
         nrow->vals = (wqval_t)index;
         index += new_table.col_count * sizeof(wqval);
         memcpy( (pd_byte_t *)sendData + (int)nrow->vals, row->vals,
                 new_table.col_count * sizeof(wqval) );

         for( i = 0; i < new_table.col_count; i++ )
         {
            wqcd_t  cd   = new_table.col_defs + i;
            wqval_t val  = row->vals + i;
            wqval_t nval = (wqval_t)((pd_byte_t *)sendData + 
                                     (DWORD)nrow->vals) + i;

            if( ( cd->dtype == pd_char_t_t ) && val->v.cval )
            {
               nval->v.cval = (pd_char_t *)index;
               int len = ( wsi_strlen( val->v.cval ) + 1 ) * sizeof(pd_char_t);
               memcpy( (pd_byte_t *)sendData + index, val->v.cval, len );
               index += len;
            }
         }
      }

      if( index != reqMem )
         panic( "LWMComm_SelectTableRows: computed and actual memory usage do "
                "not match!" );

      pCommHdr->pktlen += reqMem;
   }

   return GWMLWMCommWritePacket( pInfo, pCommHdr, sendData );
}

BOOL
LWMComm_DeleteTableRows( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_DeleteTableRows\n") );

   struct delete_block *dBlock = (struct delete_block *)pInfo->pBuf;

   int dataLen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);
   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);

   wqsx_t scrit = (dataLen > sizeof(*scrit) ) ? (wqsx_t)( dBlock + 1 ) : NULL;
   _GWMLWMCommConvertRelScritToAbs( (pd_byte_t *)dBlock, scrit );

   pCommHdr->errorCode = wss_delete( dBlock->id, FALSE, scrit,
                                     dBlock->takeFirst );
   return GWMLWMCommWritePacket( pInfo, pCommHdr, NULL );
}

BOOL
LWMComm_UpdateTableRows( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_UpdateTableRows\n") );

   struct update_block *uBlock = (struct update_block *)pInfo->pBuf;

   int dataLen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);
   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);

   wqsx_t scrit = (wqsx_t)( uBlock + 1 );
   _GWMLWMCommConvertRelScritToAbs( (pd_byte_t *)uBlock, scrit );

   wqusl_t set_script = (wqusl_t)( (pd_byte_t *)uBlock +
                                   (DWORD)uBlock->setScript );
   _GWMLWMCommConvertRelSetScriptToAbs( (pd_byte_t *)uBlock, set_script );

   pCommHdr->errorCode = wss_update( uBlock->id, FALSE, scrit,
                                     uBlock->takeFirst, set_script );

   return GWMLWMCommWritePacket( pInfo, pCommHdr, NULL );
}

BOOL
LWMComm_RequestTable( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_RequestTable\n") );

   u_long id = *((u_long *)pInfo->pBuf);

   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);
   pCommHdr->errorCode = 0;

   LPDB_CACHE sendData = NULL;
   wqtab_t    tab      = wss_db_find_table_by_handle( id );
   if( tab )
   {
      pd_char_t *tabName = DB_REL2ABS(tab->name);
      int nameLen = ( wsi_strlen( tabName ) + 1 ) * sizeof(pd_char_t);
      int reqMem = sizeof(DB_CACHE) + nameLen;

      for( int i = 0; i < tab->col_count; i++ )
      {
         wqcd_t      cd     = (wqcd_t)DB_REL2ABS(tab->col_defs) + i;
         pd_char_t *cdName = DB_REL2ABS(cd->name);

         reqMem += sizeof(wqcd) +
                   ( ( wsi_strlen( cdName ) + 1 ) * sizeof(pd_char_t) );
      }

      sendData = (LPDB_CACHE)wss_malloc( reqMem );
      memset( sendData, 0, sizeof(DB_CACHE) );

      int index = sizeof(DB_CACHE);

      sendData->id = id;
      sendData->colCount = tab->col_count;

      sendData->name = (pd_char_t *)index;
      memcpy( (pd_byte_t *)sendData + index, tabName, nameLen );
      index += nameLen;

      sendData->colDefs = (wqcd_t)index;
      index += tab->col_count * sizeof(wqcd);
      for( i = 0; i < tab->col_count; i++ )
      {
         wqcd_t cd  = (wqcd_t)DB_REL2ABS(tab->col_defs) + i;
         wqcd_t ncd = (wqcd_t)( (pd_byte_t *)sendData +
                                (DWORD)sendData->colDefs ) + i;
         memcpy( ncd, cd, sizeof(wqcd) );

         pd_char_t *cdName  = DB_REL2ABS(cd->name);
         int         nameLen = ( wsi_strlen( cdName ) + 1 ) *
                               sizeof(pd_char_t);
         ncd->name = (pd_char_t *)index;
         memcpy( (pd_byte_t *)sendData + (DWORD)ncd->name, cdName, nameLen );
         index += nameLen;
      }

      if( index != reqMem )
         panic( "LWMComm_RequestTable: computed and actual memory usage do "
                "not match!" );

      pCommHdr->pktlen += reqMem;

      wss_db_unlock_table( tab );
   }
   else
      pCommHdr->errorCode = -WRE_NOTFOUND;

   BOOL bRet = GWMLWMCommWritePacket( pInfo, pCommHdr, (PVOID)sendData );

   if( sendData )
      wss_free( sendData );

   return bRet;
}

BOOL
LWMComm_CreatingChild( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_CreatingChild\n") );

   wss_proc_t *wrp = wss_pfind( pCommHdr->pid );
   if( wrp )
   {
      WSS_MUTEX_ENTER(g_dwPendingChildren);
      DWORD dwPending = WSI_PROC_OPAQUE(wrp)->dwPendingChildren++;
      g_dwPendingChildren++;
      WSS_MUTEX_EXIT(g_dwPendingChildren);

      wss_proc_rele( wrp );
   }

   return TRUE;
}

BOOL
LWMComm_CreatedChild( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_CreatedChild\n") );

   BOOL bRet = TRUE;

   if( pCommHdr->pktlen >= ( sizeof(PIPESRV_COMM_HEADER) + sizeof(DWORD) ) )
   {
      DWORD child = *((DWORD *)pInfo->pBuf);

      if( child )
      {
         wss_proc_t *wrp = wss_pfind( pCommHdr->pid );
         if( wrp )
         {
            WSS_MUTEX_ENTER(g_dwPendingChildren);
            WSS_MUTEX_ENTER(g_procsPendingMap);
            GWM_PROC_PENDING::iterator it = g_procsPendingMap.find( child );
            if( it != g_procsPendingMap.end() )
            {
               DWORD dwPending = WSI_PROC_OPAQUE(wrp)->dwPendingChildren--;
               g_dwPendingChildren--;

               LPGWM_PROC_PENDING_DATA pPendData = (*it).second;

               // remove child pending from map
               g_procsPendingMap.erase( it );

               // continue child creation
               pPendData->p->p_ppid = pCommHdr->pid;
               bRet = _GWMLWMCommCreateProcess( pPendData->pPipe,
                                                pPendData->p,
                                                pPendData->cwd,
                                                pPendData->path );

               wss_free( pPendData );
            }
            else
            {
               LPGWM_PROC_PENDING_DATA pPendData = (LPGWM_PROC_PENDING_DATA) wss_malloc( sizeof(GWM_PROC_PENDING_DATA) );
               pPendData->p = NULL;
               pPendData->ppid = pCommHdr->pid;
               pPendData->cwd = NULL;
               pPendData->path = NULL;
               pPendData->sleep = FALSE;

               g_procsPendingMap.insert( GWM_PROC_PENDING::value_type( child,
                                                                       pPendData ) );
            }

            if( !g_dwPendingChildren )
            {
               for( GWM_PROC_PENDING::iterator it = g_procsPendingMap.begin();
                    it != g_procsPendingMap.end();
                    it++ )
               {
                  LPGWM_PROC_PENDING_DATA pPendData = (*it).second;

                  // remove child pending from map
                  g_procsPendingMap.erase( it );

                  // continue child creation
                  bRet = _GWMLWMCommCreateProcess( pPendData->pPipe,
                                                   pPendData->p,
                                                   pPendData->cwd,
                                                   pPendData->path );

                  wss_free( pPendData );                  
               }
            }
            WSS_MUTEX_EXIT(g_procsPendingMap);
            WSS_MUTEX_EXIT(g_dwPendingChildren);
         }

         wss_proc_rele( wrp );
      }
   }

   return bRet;
}

BOOL
LWMComm_FindTable( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_FindTable\n") );

   pd_char_t *fullName = NULL;

   // process packet if there's data
   if( pCommHdr->pktlen > sizeof(PIPESRV_COMM_HEADER) )
   {
      int dataSize = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);

      fullName = (pd_char_t *)pInfo->pBuf;

      if( LWMCommStringIsValid( pInfo->pBuf, dataSize,
                                   (pInfo->commHdr.flags &
                                      GWM_COMM_UNICODE) ? TRUE : FALSE ) )
      {

         if( !(pInfo->commHdr.flags & GWM_COMM_UNICODE) )
         {
            // Convert ASCII string to Unicode
            pInfo->pBuf = malloc( dataSize * sizeof(pd_char_t) );
            MultiByteToWideChar( CP_ACP, 0,
                                 (LPCSTR)fullName, dataSize,
                                 (pd_char_t *)pInfo->pBuf,
                                 dataSize * sizeof(pd_char_t) );
            free( fullName );
            fullName = (pd_char_t *)pInfo->pBuf;
            dataSize *= sizeof(pd_char_t);
         }
      }
   }

   LPVOID data = NULL;
   u_long id;

   pCommHdr->pktlen = sizeof(PIPESRV_COMM_RESP_HEADER);
   if( fullName == NULL )
      pCommHdr->errorCode = -WRE_ERROR;
   else
   {
      wqtab_t tab = wss_db_find_table_by_name( fullName );
      if( tab == NULL )
      {
         pCommHdr->errorCode = -WRE_NOTFOUND;
      }
      else
      {
         id = (u_long) DB_ABS2REL(tab);
         wss_db_unlock_table( tab );

         pCommHdr->errorCode = 0;
         pCommHdr->pktlen += sizeof(id);
         data = (LPVOID)&id;
      }
   }

   pCommHdr->flags = pInfo->commHdr.flags & GWM_COMM_UNICODE;

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_FindTable: sending response\n") );

   // send response to waiting process
   BOOL bRet = GWMLWMCommWritePacket( pInfo, pCommHdr, data );

   return bRet;
}

BOOL
LWMComm_AuditPost( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   // process packet if there's data
   if( pCommHdr->pktlen > sizeof(PIPESRV_COMM_HEADER) )
   {
      struct event_block *packet = (struct event_block*)pInfo->pBuf;
      wss_audit_write( packet->types, packet->buffer, packet->length );
   }

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_AuditPost\n") );

   return TRUE;
}

BOOL
LWMComm_SyncResend( LPPIPESRV_PIPE_INFO pInfo )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &(pInfo->commHdr);

   if( g_bDebugging )
      _tprintf( _TEXT("LWMComm_SyncResend: !!! RESYNC !!!\n") );

   pCommHdr->flags |= GWM_COMM_RESENT;

   WSS_MUTEX_ENTER( pInfo->pLastPackets );
   PVOID pData = NULL;
   PIPESRV_LAST_PACKETS::iterator it = pInfo->pLastPackets->find( pCommHdr->tid );
   if( it == pInfo->pLastPackets->end() )
   {
      pCommHdr->commandId = *(DWORD *)pInfo->pBuf;
      pCommHdr->errorCode = -ENOENT;
   }
   else
   {
      LPPIPESRV_LAST_PACKET_DATA pPacket = (*it).second;

      memcpy( pCommHdr, pPacket->pCommHdr, sizeof(*pCommHdr) );
      if( pPacket->pData )
         pData = pPacket->pData;
   }
   WSS_MUTEX_EXIT( pInfo->pLastPackets );

   return GWMLWMCommWritePacket( pInfo, pCommHdr, pData );
}

