/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 *  Author: Calvin Ko   
 *  Email : calvin_ko@nai.com
 *
 *  $Id: expr.h,v 1.11 2001/02/23 19:02:03 stevek Exp $
 */

#ifndef _EXPR_H_0699
#define _EXPR_H_0699

#include <stdio.h>
#include <string.h>
#include <string>
#include <iostream.h>
#include "source.h"
#include "smodule.h"
#include "type.h"
#include <list>

typedef set<string> StrSet;

struct WrEvent {
     int i;
};

class SpecIdentExpr;

class Expr {
public:
     enum IdType { Para, Ret, InVar};
     enum eType { Ident, SpecIdent, RecElm, ArrayElm, PtrToElm, PtrRef, 
                  Const, Binary, Unary, PostFix, Trinary, Assmt, List,
                  FuncCall, Decl, EvSel, TermDef };
     Type 	type;
     int  	lvalue;
     source *	code;
     Expr *	next;

     Expr () {   
         code = new source;   
         next = NULL;
     };

     virtual ~Expr() {
         delete code;
     }

     /** Return Expression Type */
     virtual eType etype() const = 0;

	/** equal operator */
     virtual int operator ==(const Expr &e) = 0;

     /** not equal operator */
     int operator !=(const Expr &e) {
          return !(*this == e);
     }

     /** print the expression to a stream */
     virtual ostream& print(ostream& out=cerr) const {
          out << "Expr";
          return out;
     };

	friend ostream& operator<<(ostream&, const Expr&);

     virtual const char *str() 	{	return ""; 	};

     /** return the first ident in a var ref, e.g. return a for a.b.c */
     virtual const char *head() 	{ 	return NULL; 	};
     virtual source *tail() 		{	return NULL; 	};

     virtual Type compute_type(WrEvent *)   {
          return Type(TC_ERR);    
     };
};

inline ostream& operator<< (ostream& out, const Expr& e) {
     return e.print(out); 
}

class UnaryExpr: public Expr {
public:
     Expr *e1;
     string op;
     UnaryExpr(const string& ops, Expr *expr1) : Expr() {
          op = ops;
          e1 = expr1;  
          lvalue = expr1->lvalue;
     }
     eType etype() const {   return (Unary);         }
     int operator ==(const Expr& e) {
          if (e.etype() == Unary) {
               UnaryExpr& be = (UnaryExpr&) e;
               return op == be.op && *e1 == *be.e1;
          } else
               return 0;
     }
     ostream& print(ostream& out=cerr) const {
          out << op << *e1;
          return out;
     }
};

class PostFixExpr : public Expr {
public:
     Expr *e1;
     string op;
     PostFixExpr(const string& ops, Expr *expr1) : Expr() {
          op = ops;
          e1 = expr1;  
          lvalue = expr1->lvalue;
     }
     eType etype() const {   return (PostFix);         }

     int operator ==(const Expr& e) {
          if (e.etype() == PostFix) {
               PostFixExpr& be = (PostFixExpr&) e;
               return op == be.op && *e1 == *be.e1;
          } else
               return 0;
     }

     ostream& print(ostream& out=cerr) const {
          out << *e1 << op;
          return out;
     }
};

class AssmtExpr: public Expr {
public:
     Expr *e1, *e2;
     string op;
     AssmtExpr(const string& ops, Expr *expr1, Expr *expr2) : Expr() {
          op = ops;
          e1 = expr1;  
          e2 = expr2;
     }
     eType etype() const {   return Assmt;         }
     int operator ==(const Expr& e) {
          if (e.etype() == Assmt) {
               AssmtExpr& ae = (AssmtExpr& ) e;
               return op == ae.op && *e1 == *ae.e1 && *e2 == *ae.e2;
          } else
               return 0;
     }

     ostream& print(ostream& out=cerr) const {
          out << *e1 << op << *e2;
          return out;
     }
};

class BinaryExpr: public Expr {
public:
     Expr *e1, *e2;
     string op;
     BinaryExpr(const string& ops, Expr *expr1, Expr *expr2) : Expr() {
          op = ops;
          e1 = expr1;  
          e2 = expr2;
          lvalue = 0;
     }

     eType etype() const {   return (Binary);         }

     ostream& print(ostream& out=cerr) const {
          out << *e1 << " " << op << " " << *e2;
          return out;
     }

     int operator ==(const Expr& e) {
          if (e.etype() == Binary) {
               BinaryExpr& be = (BinaryExpr& ) e;
               return op == be.op && *e1 == *be.e1 && *e2 == *be.e2;
          } else
               return 0;
     }

};

class EvSelExpr: public Expr {
public:
     AEvent 	aevent;
     Expr 	*cond;
	VarRefTbl	*svartbl;

     EvSelExpr(AEvent a, VarRefTbl *t, Expr *e) {
     	aevent = a;
		svartbl = t;	
		cond = e;
     };

     eType etype() const {   return (EvSel);         }

     int operator ==(const Expr& e) {
          if (e.etype() == EvSel) {
               EvSelExpr& ese = (EvSelExpr&) e;
               return aevent == ese.aevent && *cond == *ese.cond;
          } else
               return 0;
     }

     ostream& print(ostream& out=cerr) const {
          out << "(" << aevent << ", " << *cond << " )";
          return out;
     }
};

class TrinaryExpr: public Expr {
public:
     Expr *e1, *e2, *e3;
     string op;
     TrinaryExpr(const string& ops, Expr *expr1, Expr *expr2, Expr *expr3) 
     : Expr() 
     {
          op = ops; 
          e1 = expr1;
          e2 = expr2;
          e3 = expr3;
          lvalue = 0;
     }
     eType etype() const {   return (Trinary);         }
      
     int operator ==(const Expr& e) {
          if (e.etype() == Trinary) {
               TrinaryExpr& te = (TrinaryExpr& ) e;
               return (op == te.op) && 
                       (*e1 == *te.e1) && *e2 == *te.e2 && *e3 == *te.e3;
          } else
               return 0;
     }
     ostream& print(ostream& out=cerr) const {
          out << *e1 << " ? " << *e2 << " : " << *e3;
          return out;
     }
};

class DeclExpr: public Expr {
public:
     char *vname;
     source *d;
     Expr *init;
     DeclExpr(source *dsrc, Expr *expr, char *vn) : Expr() {
          vname = vn;
          d = dsrc;
          init = expr;  
     }
     eType etype() const {   return (Decl);         }
       
     int operator ==(const Expr& e) {
          if (e.etype() == Decl) {
               return 1;
          } else
               return 0;
     }

     ostream& print(ostream& out=cerr) const {
          out << "DeclExpr";
          return out;
     }
};

class ListExpr: public Expr {
public:
     Expr *elist;
     ListExpr(Expr *l=NULL) : Expr() {
          elist = l; 
     }

     eType etype() const {   return List;          }

     ostream& print(ostream& out=cerr) const {
          Expr *e;
          for (e = elist; e != NULL; e = e->next)
               if (e != elist)
                    out << ", " << *e;
               else
                    out << *e;
          return out;
     }

     Expr *append(Expr *e) {
          if (elist == NULL) elist = e;
          else {
               Expr *e1 = elist;
               while (e1->next != NULL) e1 = e1->next;
               e1->next = e;
          }
     }

     int operator ==(const Expr& e) {
           if (e.etype() == List)
                return 1;
           else
                return 0;
     }

	Expr *dumpList() {
		Expr *e = elist;
		elist = NULL;
		return e;	
	}
};

class TermDefExpr : public ListExpr {
public:
	int		type;		/* 1 - positive, 0 - negative */

	TermDefExpr(Expr *l, int t=1) : ListExpr(l) {
		type = t;
	};
	eType 	etype() const		{	return TermDef;	}

	/* for generation of a termdef name */
	string 	getName() {
		EvSelExpr *e = (EvSelExpr *) elist;
		return e->aevent.sevent->name;
	}
};


class FuncCallExpr: public Expr {
public:
     Expr *func, *plist; 
     FuncCallExpr(Expr *f, Expr *elist) : Expr() {
         func = f;
         plist = elist; 
     }
     eType etype() const {   return FuncCall;       }
      
     int operator ==(const Expr& e) {
           if (e.etype() == FuncCall)
                return 1;
           else
                return 0;
     }
     ostream& print(ostream& out=cerr) const {
          out << *func << "(";
          if (plist != NULL) out << *plist;
          out << ")";
          return out;
     }
};

class RecElmExpr: public Expr {
public:
     Expr *expr;
     char *name;

     RecElmExpr(Expr *e, const char *n) : Expr() {
         expr = e;
         name = strdup(n);
	 lvalue = 1;
     }

     eType etype() const {   return (RecElm);         }

     virtual ostream& print(ostream& out=cerr) const {
          out << *expr << "." << name;
          return out;
     }

       
     int operator ==(const Expr& e) {
          if (e.etype() == RecElm) {
               const RecElmExpr& e1 = (const RecElmExpr&) e;
               return *expr == *e1.expr && !strcmp(name, e1.name);
          } else
               return 0;
     }
     const char *head() { 
          return expr->head();
     }
     source *tail() { 
          source *sp = expr->tail();
          SourceAppend2(sp, ".", name);
          return sp;
     }

     Type compute_type(WrEvent *);
};

class ArrayElmExpr: public Expr {
public:
     Expr *expr, *iexpr;
     ArrayElmExpr(Expr *e, Expr *e1) : Expr() {
         expr  = e;
         iexpr = e1;
         lvalue = 1;
     }

     eType etype() const {   return (ArrayElm);         }

     virtual ostream& print(ostream& out=cerr) const {
          out << *expr << "[" << *iexpr << "]";
          return out;
     };

     int operator ==(const Expr& e) {
          if (e.etype() == ArrayElm) {
               const ArrayElmExpr& e1 = (const ArrayElmExpr&) e;
               return *expr == *e1.expr && *iexpr == *e1.iexpr;
          } else
               return 0;
     }

     Type compute_type(WrEvent *);
     
};

class IdentExpr: public Expr {
public:
     char *name;
     IdentExpr(const char *n) : Expr() {
         name = strdup(n);
         lvalue = 1;
     }
     eType etype() const {   return (Ident);         }

     virtual ostream& print(ostream& out=cerr) const {
          out << name;
          return out;
     }

     const char *str()   {   return name;            }
     int operator ==(const Expr& e) {
          if (e.etype() == Ident) {
                 return !strcmp(name, ((const IdentExpr&) e).name);
          } else
          return 0;
     }
     const char *head() {
          return name;
     };
     source *tail() { 
          source *sp = new source;
          return sp;
     }
     Type compute_type(WrEvent *);

};

class SpecIdentExpr: public Expr {
public:
     char *name;
     char *offset;
     mlabel *label;
     IdType idtype;
     SpecIdentExpr(const char *n) : Expr() {
         name = strdup(n);
         lvalue = 1;
     }
     eType etype() const {   return (SpecIdent);         }

     virtual ostream& print(ostream& out=cerr) const {
          out << "$" << name;
          return out;
     }

     int operator ==(const Expr& e) {
          if (e.etype() == SpecIdent) {
                 return !strcmp(name, ((const SpecIdentExpr&) e).name);
          } else
                 return 0;
     }
};

class PtrRefExpr: public Expr {
public:
     Expr *expr;

     PtrRefExpr(Expr *e) : Expr() {
         expr = e;
         lvalue = 1;
     };

     eType etype() const {   return (PtrRef);         }

     virtual ostream& print(ostream& out=cerr) const {
          out << "*" << *expr;
          return out;
     }

     int operator ==(const Expr& e) {
          if (e.etype() == PtrRef) {
               const PtrRefExpr& e1 = (const PtrRefExpr&) e;
               return *expr == *e1.expr;
          } else
               return 0;
     }
     const char *head() {
          return expr->head();
     }
     source *tail() {
          return expr->tail();
     }
     Type compute_type(WrEvent *);
};

class PtrToElmExpr: public Expr {
public:
     Expr *expr;
     char *name;
     PtrToElmExpr(Expr *e, const char *n) : Expr() {
         expr = e;
         name = strdup(n);
         lvalue = 1;
     }
     eType etype() const {   return (PtrToElm);         }

     virtual ostream& print(ostream& out=cerr) const {
          out << *expr << "->" << name;
          return out;
     }

     int operator ==(const Expr& e) {
          if (e.etype() == PtrToElm) {
               const PtrToElmExpr& e1 = (const PtrToElmExpr&) e;
               if (*expr == *e1.expr) {
                    return (!strcmp(name, e1.name) || !strcmp(e1.name, "*"));
               } else
                    return 0;
          } else
               return 0;
     }
     const char *head() { 
          return expr->head();
     }
     source *tail() { 
          source *sp = expr->tail();
          if (!sp->is_empty()) {
              SourceAppend2(sp, "->", name);
          } else
              sp->append(name);
          return sp;
     }

     Type compute_type(WrEvent *);
};

class ConstExpr : public Expr {
public:     
     char *val;
     ConstExpr(const char *v, Type t = cstrType) : Expr() {   
         val = strdup(v);    
         type = t;
         lvalue = 0;
     }
     ConstExpr(int i) : Expr() {   
         char buf[64];
         sprintf(buf, "%d", i);
         val = strdup(buf); 
         type = intType;
         lvalue = 0;

     }

     eType etype() const      {    return Const;      }

     virtual ostream& print(ostream& out=cerr) const {
          out << val;
          return out;
     }

     int operator ==(const Expr& e) {
          return e.etype() == Const && 
                 !strcmp( val, ((const ConstExpr&) e).val);
     }
     const char *head() { 
          return val;
     }
     source *tail() { 
          source *sp = new source;
          return sp;
     }
     Type compute_type(WrEvent *) {   return type;    }
};

#endif /* _EXPR_H_0699 */
