/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 *  $Id: gram.cc,v 1.28 2001/04/05 18:40:24 stevek Exp $
 */
#include <stdio.h>
#include <string.h>
#include <string>

#include "conf.h"
#include "gram.h"
#include "pegc.h"
#include "source.h"

extern void pegerror(const char *s, const char *s1="");
extern int pegwarning(const char *s, const char *s1="");
extern SModule *curr_module;
extern char *outfilename;

static char *def_comments =
"#include \"../../../wr.include/local.ch\"\n"
"#include \"../../../wr.include/libwr.h\"\n"
"#define  TRACK_FD  1\n"
"#define  STACKSIZE 64\n"
"#define  P_SUCC                   0\n"
"#define  P_FAIL                   1\n"
"#define  P_WRONGTK                2\n"
"#define  P_EXECFAIL               3\n"
"#define  P_STACKVALUE_OVERFLOW    4\n"
"\n"
"\n"
"/*\n"
" * NTERMS = number of terminals, terminal id from 1 to NTERMS\n"
" * NVARS  = number of nonterm (vars), var id from 1 to NVARS\n"
" * NACTS  = number of actions, action id from 0 to NACT - 1\n"
" * NSYMS  = number of terms + nonterms\n"
" * NITEMS = no of terms + nonterms + actions\n"
" */\n";



/** constructor */
PEGram::PEGram(const char *n) 
{
     name = n;
	nerr = 0;			/* number of error initialized to 0 */
     nrule = nterm = nvars = nactions = 0;

     nil = new Term("nil");
     nil->id = 0;
     nilfirsts = new GSymSet;
     nilfirsts->insert(nil);
     nil->firsts = nilfirsts;
     eof = new Term("eof");
     eof->id = 0;
     symtbl.insert(make_pair(string("nil"), nil));
     nvt = 2;

     err = new Term("err");
     err->id = 1;
     err->firsts = new GSymSet;
     err->firsts->insert(err);
     symtbl.insert(make_pair(string("err"), err));
}

/** perform some sanity checks on the PE-grammar */
int PEGram::check()

{

    /* Check whether all used terminal are defined and all defined
     * terminal are being used
     */
    int i;
    for (i = nvt; i < nterm+nvt; i++) {
         if (tdtbl.find(idtbl[i]->name.c_str()) == tdtbl.end()) {
              	cerr << "Error: Terminal '" << idtbl[i]->name 
				<< "' used but not defined\n";
			nerr++;
         }
    };

    RuleList::iterator iter;
    for (iter = rlist.begin(); iter != rlist.end(); iter++) {
         Rule *r = *iter;
         idtbl[r->lhs->id]->defined = 1;
         GSymList::iterator giter;
         for (giter = r->rhs->begin(); giter != r->rhs->end(); giter++) {
              if ((*giter)->objtype() == GSym::NTERM)
                   idtbl[(*giter)->id]->used = 1;
         }
    }
    for (i=0; i<nvars; i++) {
         NTerm *n = (NTerm *) idtbl[i+nterm+nvt];
         if (n->used == 1 && n->defined == 0) {
         		cerr << "Error: " << *n << " used but not defined in any rules\n";
			nerr++;
         }
    };

    /* 
     * Check all the hypernotions/nonterms
     * defined and used, have problems.
    RuleList::iterator iter;
    for (iter = rlist.begin(); iter != rlist.end(); iter++) {
         Rule *r = *iter;
         r->lhs->defined = 1;
         GSymList::iterator giter;
         for (giter = r->rhs->begin(); giter != r->rhs->end(); giter++) {
              (*giter)->used = 1;
         }
    }
    for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
         if ((*giter).second->objtype() == GSym::NTERM) {
              if ((*giter).second->used == 0 && (*giter).second->defined == 1)
                    pegwarning((*giter).second->name.c_str(), 
                               " defined but not used"); 
              else 
              if ((*giter).second->used == 1 && (*giter).second->defined == 0)
                    pegwarning((*giter).second->name.c_str(), 
                               " is not defined"); 
         }
    } 
    */

}

/** Process a PEGram.  This function will do
 *  1) some sanity checking on the grammar (e.g., terms defined but no used)
 *  2) compute the FIRSTS, FOLLOWS, and the parse table
 *  3) compute the attributes that will be referenced for each relevant System event, which
 *     will be used in code generation or auditdata-request-specification.
 *
 */
int PEGram::process()

{
    RuleList::const_iterator iter;

    /* allocate idtbl */
    idtbl = new GSym *[nterm+nvars+2];

    /* Adjust the id of all nonterm and initialize the idtbl
	* -- each GSym has an id and idtbl[id] should give the associated GSym.	
     * We have two special terms, eof/nil(id==0) and error (id==1)
	*/
    GSymTbl::iterator giter;
    for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
         if ((*giter).second->objtype() == GSym::NTERM) {
              (*giter).second->id += (nterm + nvt - 1);
         }
         idtbl[(*giter).second->id] = (*giter).second;
    }
    idtbl[0] = eof; 		/* set specially */
    GSymList::iterator giter1;
    for (giter1 = alist.begin(); giter1 != alist.end(); giter1++) {
         (*giter1)->id += nterm + nvars + nvt;
    };
    if (dflag & DF_GRAM_BASE) {
         cerr << "no of terms = " << nterm << "\n";
         cerr << "no of vars = " << nvars << "\n";
    };
    if (dflag & DF_GRAM_SYM) {
       for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
           cerr << (*giter).second->id << ":\t" 
                << *(*giter).second << "\n";
       } 
    };
    check();

    set_firsts();
    set_follows();
    set_lasts();

    if (dflag & DF_GRAM_PTABLE) {
		print_firsts();
		print_lasts();
		print_follows();
	}
	
    set_parsetable();
    set_optimized_parsetable();

	if (nerr == 0) {
		propagate_attr_ref();
		check_attr_ref();	
	}

	if (nerr == 0) {
		source *src = gencode();
		FILE *fp = fopen(outfilename, "w");
		fprintf(fp, "%s", src->all());
		fclose(fp);                                              
	};

}

/* set_firsts computes FIRSTS for all nonterminals.
   See the algorithm in the book Compiler: principles, techniques, and tools,
   pp. 189
*/
int PEGram::set_firsts()

{
    int changed;
    do {
         GSymTbl::iterator giter; 
         for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
              if ((*giter).second->objtype() == GSym::NTERM) {
                    *(*giter).second->oldfirsts = *(*giter).second->firsts;
              }
         }
         RuleList::const_iterator iter;
         for (iter = rlist.begin(); iter != rlist.end(); iter++) {
              Rule *r = *iter;
              if (r->rhs->size() == 0) {   	/* step 2 */
                   r->lhs->firsts->insert(nil);
              } else {				/* step 3 */
                   /* 
                    * Remember whether "firsts" of left-hand symbol contain nil 
                    * Then we delete nil from "firsts" if it does not contain 
                    * nil in the first place.
                    * Since in step 2, nil should not be added into "firsts".
                    * unless all the firsts of the symbol contain nil
                    * We do so to make the code simplier. 
                    */
                   int nonil = r->lhs->firsts->find(nil)==r->lhs->firsts->end();
                   
                   GSymList::iterator iter1 = r->rhs->begin();
                   /* step 2: merge the firsts with that of (*iter) */
                   r->lhs->firsts->insert((*iter1)->firsts->begin(),
                                           (*iter1)->firsts->end());
                   while (iter1 != r->rhs->end() &&
                       (*iter1)->firsts->find(nil) != (*iter1)->firsts->end()) 
                   {    
                        iter1++;  
                        /* step 2: merge the firsts with that of (*iter) */
                        if (iter1 != r->rhs->end())
                            r->lhs->firsts->insert((*iter1)->firsts->begin(),
                                                   (*iter1)->firsts->end());
                   }
                   /* 
                    * Remove the nil symbol if "firsts" does not contain nil 
                    * in the first place && not firsts of all rhs symbol 
                    * contain nil 
                    */
                   if (nonil && iter1 != r->rhs->end())
                        r->lhs->firsts->erase(nil);
              }
         } 
         changed = 0;
         for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
              if ((*giter).second->objtype() == GSym::NTERM) {
                   if(*(*giter).second->oldfirsts != *(*giter).second->firsts) {
                        changed = 1; 
                        break;
                   };
              }
         }
    } while (changed == 1);
}

/** print the FIRSTS set, for debugging */
int PEGram::print_firsts(ostream& out=cerr) 

{
    GSymTbl::iterator giter; 
    out << "FIRSTS\n";
    out << "---------------------------\n";
    for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
         GSymSet::iterator iter;
         out << *(*giter).second << ":\t";
         for (iter = (*giter).second->firsts->begin();
              iter != (*giter).second->firsts->end();
              iter++) 
         {
              out << **iter << "  ";          
         }
         out << "\n";
    }

}

/* 
 ** set_lasts computer LASTS(sym), a set of terminals, for all nonterminals
 *  LASTS(sym) is defined as if (sym -*-> abc...y) then y is  in LASTS(sym);
 */
int PEGram::set_lasts()

{
     int changed;
     do {
         changed = 0;
         RuleList::const_iterator iter;
         for (iter = rlist.begin(); iter != rlist.end(); iter++) {
             Rule *r = *iter;
             if (r->lastSym() != r->rhs->rend()) {
                 GSym *sym = *(r->lastSym());
                 GSymSet savelast = *(r->lhs->lasts);
                 r->lhs->lasts->insert(sym->lasts->begin(), sym->lasts->end());
                 if (savelast != *(r->lhs->lasts))
                     changed = 1;
             }
         };
     } while (changed == 1);

};

/** print tht LASTS set */
int PEGram::print_lasts(ostream& out=cerr) 

{
    GSymTbl::iterator giter; 
    out << "LASTS\n";
    out << "---------------------------\n";
    for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
         GSymSet::iterator iter;
         out << *(*giter).second << ":\t";
         for (iter = (*giter).second->lasts->begin();
              iter != (*giter).second->lasts->end();
              iter++) 
         {
              out << **iter << "  ";          
         }
         out << "\n";
    }
}


/* 
 *  set_follows computes FOLLOWS(sym) for all nonterminals.
 *  See the algorithm in the book Compiler: principles, techniques, and tools,
 *  pp. 189
 */

int PEGram::set_follows()

{
    /* 
     * step 1: add eof to follows(S), the start symbol
     */
     start->follows->insert(eof); 

    /* 
     * step 2:
     */
     RuleList::const_iterator iter;
     for (iter = rlist.begin(); iter != rlist.end(); iter++) {
          Rule *r = *iter;
          GSymList::iterator iter1;
          for (iter1 = r->rhs->begin(); iter1 != r->rhs->end(); iter1++) {
                if ((*iter1)->objtype() == GSym::NTERM) {
                     GSymList::iterator iter2 = iter1;
                     GSymList::iterator iter3 = iter1;
                     iter3++;
                     GSymSet f1;
                     do {
                          if (++iter2 != r->rhs->end()) 
                              f1.insert( (*iter2)->firsts->begin(),
                                         (*iter2)->firsts->end()
                                       );
                     } while (iter2 != r->rhs->end() && 
                                    (*iter2)->firsts->find(nil) != 
                                       (*iter2)->firsts->end()
                             );

                     /* step 2 */
                     (*iter1)->follows->insert(f1.begin(), f1.end());
                     /* step 3 */
                     if (f1.find(nil) != f1.end() || iter3 == r->rhs->end())
                          (*iter1)->follows->insert( r->lhs->follows->begin(),
                                                     r->lhs->follows->end()
                                                   );
                }
                
          }

     }
}

int PEGram::print_follows(ostream& out=cerr) 

{
	GSymTbl::iterator giter; 
	out << "FOLLOWS\n";
    out << "---------------------------\n";
    for (giter = symtbl.begin(); giter != symtbl.end(); giter++) {
	    if ((*giter).second->objtype() == GSym::NTERM) {
		    GSymSet::iterator iter;
		 out << *(*giter).second << ":\t";
	      for (iter = (*giter).second->follows->begin();
			      iter != (*giter).second->follows->end();
			      iter++) 
	      {
		      out << **iter << "  ";          
	      }
	      out << "\n";
	 }
    }
};

int PEGram::set_parsetable()

{
    // allocate the parse table;
    // ptable[nt->id - nterm - nvt][t->id] contains the rule for (nt, t)
    ptable = new RuleSetPtr[nvars];
    int i,j;
    for (i=0; i<nvars; i++) {
	 ptable[i] = new RuleSet[nterm+nvt];
    }
    RuleList::const_iterator iter;
    for (iter = rlist.begin(); iter != rlist.end(); iter++) {
         Rule *r = *iter;
	 NTerm *left = (*iter)->lhs;
         GSymList::iterator iter1;
         if (r->rhs->size() != 0) {
              GSymSet frhs;
              iter1 = r->rhs->begin();
              frhs.insert((*iter1)->firsts->begin(), (*iter1)->firsts->end());
              while (iter1 != r->rhs->end() &&
                       (*iter1)->firsts->find(nil) != (*iter1)->firsts->end()) 
              {    
                   iter1++;  
                   if (iter1 != r->rhs->end())
                        frhs.insert((*iter1)->firsts->begin(), 
                                    (*iter1)->firsts->end());
              }
              GSymSet::iterator iter2;
              for (iter2 = frhs.begin(); iter2 != frhs.end(); iter2++) {
                   // cerr << left->id << ", " << (*iter2)->id << ":\t" << *r << "\n"; 
	           ptable[left->id - nterm - nvt ][(*iter2)->id].insert(r);
              }
              
         }
         if (r->rhs->size() == 0 || iter1 == r->rhs->end()) {
              /* follows */
              GSymSet::iterator iter2;
              for (iter2 = left->follows->begin(); 
                   iter2 != left->follows->end();
                   iter2++) 
              {
                   // cerr << "follow" << left->id << ", " << (*iter2)->id << ":\t" << *r << "\n"; 
	           ptable[left->id - nterm - nvt][(*iter2)->id].insert(r);
              }
         }
    }
    for (i=0; i<nvars; i++) {
         for (j=0; j<nterm+1; j++) {
             RuleSet::iterator iter3;
             if (ptable[i][j].size() > 1) {
                  cerr << "conflict on " << i + nterm + nvt 
                       << ", " << j << "\n" ;
             }
             if (dflag & DF_GRAM_PTABLE) { 
               for (iter3 = ptable[i][j].begin(); iter3 != ptable[i][j].end();
                  iter3++) {
                  cerr << "[" << i + nterm + nvt << ", " << j << "]\t" ;
                  cerr << **iter3 << "\n";
               }
             }
         }
    }
};

int PEGram::set_optimized_parsetable()

{
    int i,j;
    int change;
    // initialize opt_ptable
    opt_ptable = new Rule **[nvars];
    for (i=0; i<nvars; i++) {
         opt_ptable[i] = new Rule *[nterm+nvt];
         for (j=0; j < nterm+nvt; j++) {
              if (ptable[i][j].empty()) {
                   opt_ptable[i][j] = NULL;
              } else {
                   opt_ptable[i][j] = new Rule(**ptable[i][j].begin());
              }
         }
    }
    // step 2
    do {
       change = 0;
       for (i=0; i<nvars; i++) {
          for (j=0; j<nterm + nvt; j++) {
              Rule *r = opt_ptable[i][j];
              if (r != NULL && r->firstSym() != r->rhs->end() &&
                  (*r->firstSym())->objtype() == GSym::NTERM &&
                  (*r->firstSym())->id != r->lhs->id  ) 
              {
                   Rule *r1 = opt_ptable[r->rhs->front()->id - nterm - nvt][j];
                   if (r1 != NULL) {
                        GSymList::iterator iter = r->firstSym();
                        r->rhs->insert(iter, r1->rhs->begin(), r1->rhs->end());
                        r->rhs->erase(iter);
                        change = 1;
                   }
              }
          }
       } 
    } while (change == 1);
    if (dflag & DF_GRAM_OPTABLE) { 
      for (i=0; i<nvars; i++) {
         for (j=0; j<nterm+nvt; j++) {
             if (opt_ptable[i][j] != NULL) {
                  cerr << "[" << *idtbl[i+nterm+nvt] << ", " << j << "]\t" ;
                  cerr << *opt_ptable[i][j] << "\n";
             }
         }
      }
    }
}

/** Generate code */
source *PEGram::gencode()

{
	source *src = new source;
	src->gen_comment("Information on the PEGram");
	return src;
}

/** Propagate the $attributes reference in the actions to the
 *  corresponding Audit events.  We need to go through the reduce table
 *  to see the corresponding terminals which acceptance will cause the
 *  execution of the action. Then, we identify the audit events associated
 *  withe the terminals
 */
int PEGram::propagate_attr_ref()

{
     RuleList::const_iterator iter;
     for (iter = rlist.begin(); iter != rlist.end(); iter++) {
          Rule *r = *iter;
          GSymList::iterator giter;
          GSym *psym = NULL;		// holding the last encountered sym
          for (giter = r->rhs->begin(); giter != r->rhs->end(); giter++) {
               if ((*giter)->objtype() == GSym::ACTION) {
                   Action *act = (Action *) (*giter);
                   if (psym == NULL) {
                   } else if (psym->objtype() == GSym::NTERM){
                       GSymSet::iterator siter;
                       for (siter = psym->lasts->begin();
                            siter != psym->lasts->end(); siter++) {
                            tdtbl[(*siter)->name]->addAttrRef(act->sVarRefTbl);
                       }
                   } else if (psym->objtype() == GSym::TERM){
                       tdtbl[psym->name]->addAttrRef(act->sVarRefTbl);
                   }
               } else {
                   psym = *giter;
               };
          };
     };
}

int PEGram::check_attr_ref() 

{
	TermDefTbl::iterator iter;
	for (iter = tdtbl.begin(); iter != tdtbl.end(); iter++) {
		if ((*iter).second->chkAttrRef() == 0) nerr++; 
	}
}

source PEGram::getsrc_fromfile(const char *fname)

{
     source src;
     FILE *fp = fopen(fname, "r");
     char buf[256], *s;
     if (fp == NULL) {
         fprintf(stderr, "Fatal Error: %s not found\n", fname);
         exit(1);
     }
     while ((s = fgets(buf, 256, fp)) != NULL) {
          src << s; 
     };
     return src;
};

source PEGram::gen_define()

{
      int i;
      source src; 
      /* Generate constant about the grammar */
      src << def_comments; 
      src << "#define\tAZ_NAME    " << '"' << name << '"' << "\n";
      src << "#define\tNTERMS     " << nterm << "\n";
      src << "#define\tNVARS      " << nvars << "\n";
      src << "#define\tNACTS      " << nactions << "\n";
      src << "#define\tNSYMS      " << nterm + nvars << "\n";
      src << "#define\tNITEMS     " << nterm + nvars + nactions << "\n";
      src << "#define\tNO_OF_SUBFUNCS   " << nvars + alist.size() + 1 
          << "\n\n"; 

      /*
       * Generate ID definition of GSyms and Actions
       * In the generated code, each grammar symbol (GSyms) is assigned an 
       * id for referencing them in the run-time stack.
       */
      src.gen_comment("Grammar Symbols Definition");
      for (i=0;i<nvars + nterm + nvt;i++) {
           if (idtbl[i] == NULL) {
                cerr << "idtbl[" << i << "] == NULL\n";
                exit(0);
           }
           src << "#define " << idtbl[i]->cname << "\t" << idtbl[i]->id << "\n";
      }
      GSymList::iterator giter;
      for (giter = alist.begin(); giter != alist.end(); giter++) {
           src << "#define " << (*giter)->cname << "\t" << (*giter)->id << "\n";
      };
      src << "#define START_NTERM\t" << start->cname << "\n";
      
      return src;
}


/** check attributes referenced and post error message */
int TermDef::chkAttrRef()

{
     char buf[256];
     if (texpr == NULL || texpr->etype() != Expr::TermDef) {
          cerr << "fatal error #9\n";
          return 0;
     }

     TermDefExpr *e = (TermDefExpr *) texpr;
     if (e->type == 1) { /* positive def */
     	Expr *el = e->elist;
     	while (el != NULL) {
         		if (el->etype() == Expr::EvSel) {
     			VarRefTbl::iterator iter;
     			for (iter = atbl.begin(); iter != atbl.end(); iter++) {
					string attr = (*iter).first;
					int lineno = (*iter).second;
              			EvSelExpr *se = (EvSelExpr *) el;
              			if (se->aevent.find(attr) == NULL) {
                   			cerr << "line " << lineno << ": Invalid attribute "
                        			<< attr << " for event " 
							<< se->aevent.getSysOpName()
                        			<< " reference by terminal " << name << "\n";
                   			return 0;
					}
              		}
         		}
         		el = el->next;
     	}
     } else {	/* negative def, ~(....) */
     	SEvent *sev = curr_module->find("common");
     	VarRefTbl::iterator iter;
     	for (iter = atbl.begin(); iter != atbl.end(); iter++) {
			string 	attr 	= (*iter).first;
			int 		lineno 	= (*iter).second;
     		if (sev == NULL || !sev->find(attr)) {
				cerr << "line " << lineno << ": Invalid attribute "
					<< attr << " reference by terminal " << name << "\n";
     		}
     	}
	}
     return 1; 
};

/** Get the list of attributes that will be referenced
 *  in a token function associated with e
 */
AttrList TermDef::getAttrRefList(EvSelExpr *e)

{
	AttrList alist;
	VarRefTbl::iterator iter;
	SModule *m = e->aevent.sevent->m; 

	StrSet sset;
	/* obtain the set of attributes that will be referenced */  
	for (iter = atbl.begin(); iter != atbl.end(); iter++) {
		sset.insert((*iter).first);
	}	
	for (iter = e->svartbl->begin(); iter != e->svartbl->end(); iter++) {
		sset.insert((*iter).first);
	}

	StrSet::iterator siter;	
	for (siter = sset.begin(); siter != sset.end(); siter++) {
		alist.push_back( e->aevent.find(*siter) );
	}
	return alist;
}
