/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 *  Author: Calvin Ko
 *  Email : calvin_ko@nai.com
 *
 *  $Id: gram.h,v 1.26 2001/02/23 19:02:03 stevek Exp $
 */

#ifndef _KGRAM_H
#define _KGRAM_H
#include <libio.h>
#include <iostream.h>
#include <strstream.h>

#include "conf.h"

#include <list.h>
#include <set.h>
#include <map.h>
#include <multimap.h>

#ifdef WR_WINNT
#include <iostream>
#endif

#include "smodule.h"
#include "type.h"
#include "expr.h"

class Rule;
class GSym;
class Term;
struct ltgsym;
class TermDef;
class PEGram;


typedef list<Rule *> 		RuleList;
typedef list<string> 		StrList;
typedef list<const char *> 	CStrList;
typedef list<GSym *>  		GSymList;

typedef set<string>			StrSet;
typedef set<GSym *, ltgsym>  	GSymSet;
typedef set<TermDef *>		TDefSet;
typedef set<Rule *> 		RuleSet;
typedef RuleSet 			*RuleSetPtr;

typedef map<string, TermDef *> 	TermDefTbl;
typedef map<string, Term *> 	TermTbl;
typedef multimap<string, GSym *> GSymTbl;

// class for grammer symbol: base class of term and nterm
class GSym {
public:
     enum 	tcode {TERM, NTERM, ACTION};
     int 		id;
     string 	name;
     string	cname;
     GSymSet   *firsts, *oldfirsts;
     GSymSet   *follows, *lasts;
     short      used, defined;
     GSym() {};
     GSym(const string& n) {
          name = n;
          used = 0;
          defined = 0;
     };
     virtual tcode objtype() const = 0;
     virtual ostream& print(ostream& out=cerr) const = 0;
     virtual int operator ==(const GSym& s) = 0;
};

struct ltgsym {
     bool operator () (GSym *s1, GSym *s2) const {
          return s1->name < s2->name;
     }
};

inline ostream& operator << (ostream& out, const GSym& t) {
     return t.print(out);
};

class Action : public GSym {
public:
     Expr 	*elist;
     source 	*src;
     int  	lineno;		/* starting line number */
     VarRefTbl *sVarRefTbl;	/* a multimap that contain the names of the special 
						 * var ($var) reference (def + use) in the block
                               * and the line number of each reference
                               */
     Action (source *s=NULL, int aid=0, VarRefTbl *t=NULL, int l=0) {
          ostrstream ss;
          ss << "Action_" << aid << "_L" << l; 
          name = "Action";
          cname = ss.str();
          src = s;
          sVarRefTbl = t;
          lineno = l; 
          id = aid;
     }

     tcode objtype() const {
          return ACTION;
     }

     ostream& print(ostream& out=cerr) const {
          out << "{ }\n";
     };

     int operator == (const GSym& s) {
          return s.objtype() == ACTION;
     }

};

class NTerm : public GSym {
public:
     Expr      *elist;
     RuleList  derive;     
     NTerm (Expr *el) { 
	  // assume el points to an IdentExpr
          elist = el; 
          name = string(((IdentExpr *) elist)->name);
          cname = "NTERM_" + name;
     };
     ostream& print(ostream& out=cerr) const {
         Expr *e;
         out << '<' ;
         for (e = elist; e != NULL; e = e->next) {
              if (e != elist) out << ", ";
              out << *e;
         }
         out << '>';
     }
     tcode objtype() const {
          return NTERM;
     }
     int operator ==(const GSym& s) {
         if (s.objtype() == NTERM) {
              Expr *e = elist, *e1 = ((NTerm&)s).elist; 
              while (e != NULL && e1 != NULL) {
                   if (!(*e == *e1))
                        return 0;
                   e = e->next;     
                   e1 = e1->next;
              };
              return e == NULL && e1 == NULL;
         } 
         return 0;
     }
};

class Term : public GSym {
public:
     CStrList *mvars;
     Term(const string& n, CStrList *l=NULL) : GSym(n) {
          mvars = l != NULL ? l : new CStrList;
          cname = "TERM_" + name;
          firsts = new GSymSet;
          firsts->insert(this);
          lasts = new GSymSet;
          lasts->insert(this);
     };
     ostream& print(ostream& out=cerr) const {
          out << name;
     }
     tcode objtype() const {
          return TERM;
     }
     int operator ==(const GSym& s) {
          return s.objtype() == TERM && name == s.name;
     }
};

class Rule {
public:
     int   	id;			/* Rule ID 					*/
     NTerm 	*lhs;		/* left hand side symbols 		*/
     GSymList 	*rhs;   		/* list of GSyms in right-hand side */
     Rule(NTerm *t, GSymList *l = NULL) {
          // Inv: t and l != NULL
          lhs = t;
          rhs = l != NULL ? l : new GSymList;
     };
     Rule(const Rule& r) {
          id = r.id;
          lhs = r.lhs;
          rhs = new GSymList(r.rhs->begin(), r.rhs->end());
     }
     ostream& print(ostream& out=cerr) const {
          GSymList::const_iterator iter;
          out << "R" << id << ":\t";
          out << *lhs << " -> ";
          for (iter = rhs->begin(); iter != rhs->end(); iter++) {
               out << " " << **iter;
          } 
          out << ";";
          return out;
     }

     int hasfirstsym(GSym *s) {
          GSymList::iterator iter = rhs->begin();
          while (iter != rhs->end() && (*iter) != s) 
               iter++;
          return iter != rhs->end();
     }
     // Assuem rhs is not empty
     GSymList::iterator firstSym() {
          GSymList::iterator iter = rhs->begin();
          while (iter != rhs->end() && (*iter)->objtype() == GSym::ACTION) 
               iter++; 
          return iter;
     }

     GSymList::reverse_iterator lastSym() {
          GSymList::reverse_iterator riter = rhs->rbegin();
          while (riter != rhs->rend() && (*riter)->objtype() == GSym::ACTION) 
               riter++;
          return riter;
     }
};
typedef Rule *RulePtr;
typedef Rule **RulePtrPtr;

inline ostream& operator << (ostream& out, const Rule& r) {
     return r.print(out);
};

/** Class for storing definition of a terminal */
class TermDef {
	int		used;		/* Is the termdef used in the grammar 	*/
	VarRefTbl atbl;		/* attributes referenced 			*/
public:
     string 	name;		/* Name of the terminal 	*/
     Expr 	*texpr;		/* Expression 			*/
     PEGram 	*g;			/* pointer back to the PEGram containing the term */

     TermDef(const string& n) {
         name = n;
     };

     source  	*gen_tokenfunc();

     /** add info on attributes being referenced */
     int	addAttrRef(VarRefTbl *t) {
     	atbl.insert(t->begin(), t->end());
     };

	/** check whether attributes that will be referenced in subsequenct 
	  * action blocks are valid 
	  */
     int		chkAttrRef();

     AttrList	getAttrRefList(EvSelExpr *e);
};

/** class for holding a selection expression */
class Selector {
public:
     string host;
     string user;
     string prog;
     string proc;
     AEvSet *evset;
     Selector(string h, string u, string pg, string p, AEvSet *s) {
         host = h;
         user = u;
         prog = pg;
         proc = p;
         evset = s;
     };
};

typedef list<Selector *> SelList;

class PEGram {
public:
     string 	name;		// name of the Grammar
     int 		nrule;		// number of rules
     int 		nterm;		// number of terms
     int 		nvars;		// number of non terms
     int		nactions;		// number of actions
     int    	nvt;			// number of virtual terms
                              // should be the start id of the terminals

	int		nerr;		// no of errors
     NTerm 	*start;		// Start symbol
     GSymTbl 	symtbl;		// Table of all Grammer Symbols
     RuleList 	rlist; 		// List of Production Rules
     GSym 	**idtbl; 		// Map Gsym ID to Gsym
     GSymList	alist;		// list of actions
     source	envdecl;		// declaration of env vars
     source	lenvdecl;		// declaration of local env vars
     source	envinit;		// initialization of env vars
     source	lenvinit;		// initialization of local env vars
     SelList	*selector;	// trace selectors;
     source	init;		// initialization
     source	fini;		// finalization
     source	error;		// error

     Term 	*nil, *eof, *err;
     TermDefTbl tdtbl;			// Defintions of terminals
     RuleSet 	**ptable;			// parse table
     Rule    	***opt_ptable;		// optimized parse table
     GSymSet 	*nilfirsts;		// a symbol set containing just nil
     char 	cbuf[64];			// a temporary buffer

     PEGram(const char *n);

     /** For adding simple terminal.
      * if no symbol named s exist in GSymtbl, creat a new terminal,
      * in the GSymtbl and return it.
      * Otherwise, if a terminal named s is already in the Gsymtbl, return it
      * if a nonterminal named s is already int he Gsymtbl, return NULL
      */
     GSym *add_term(const string& s) {
         GSymTbl::iterator iter = symtbl.find(s); 
         if (iter == symtbl.end()) {
              Term *t = new Term(s);
              t->id = ++nterm + nvt - 1;
              symtbl.insert(make_pair(s, t));
              return t;
         } else {
              return (*iter).second->objtype() == GSym::TERM ?
                          (*iter).second : NULL;
         }
     }  

     /** add a termdef to the specification 
	 */
     int add_termdef(TermDef *td) {
           td->g = this;
           return tdtbl.insert(make_pair(td->name, td)).second;
     }

     char *gen_term_name(string n="Token") {
           ostrstream ss;
           ss << n << "_" << nterm + nvt; 
           strcpy(cbuf, ss.str());
           return cbuf;
     }

     GSym *get_term(const char *s) {
         GSymTbl::iterator iter = symtbl.find(s); 
         return iter == symtbl.end() || 
             (*iter).second->objtype() == GSym::NTERM ? NULL : (*iter).second;
     }
     
     NTerm *add_nterm(Expr *e) {
          string tname = ((IdentExpr *) e)->name;
          GSymTbl::iterator iter = symtbl.find(tname);
          if (iter == symtbl.end()) {
               NTerm *t = new NTerm(e);
               t->id = ++nvars;
               t->firsts = new GSymSet;
               t->oldfirsts = new GSymSet;
               t->follows = new GSymSet;
               t->lasts = new GSymSet;
               symtbl.insert(make_pair(tname, t));
               return t;
          } else if ((*iter).second->objtype() == GSym::NTERM) {
               NTerm *t = new NTerm(e);
               t->id = (*iter).second->id;
               t->firsts = (*iter).second->firsts;
               t->oldfirsts = (*iter).second->oldfirsts;
               t->follows = (*iter).second->follows;
               t->lasts = (*iter).second->lasts;
               while (iter != symtbl.end()) {
                    if (*(*iter).second == *t) {
                         delete t;
                         return (NTerm *) (*iter).second;
                    }
                    iter++;
               }
               symtbl.insert(make_pair(tname, t));
               return t;
          } else { // a terminal with the same name already exist
               return NULL;
          }
     }

     GSym *add_action(source *s, VarRefTbl *vt, int l=0) {
          // name = "Action";
          Action *t = new Action(s, nactions++, vt, l);
          t->firsts = nilfirsts;
          t->oldfirsts = nilfirsts;
          alist.push_back(t);
          return t;
     }

     void addrule(Rule *r) {
          r->id = ++nrule;
          rlist.push_back(r);     
     } 

     void printrule(ostream& out=cerr) {
          RuleList::const_iterator iter;
          for (iter = rlist.begin(); iter != rlist.end(); iter++) {
              out << **iter << "\n";
          } 
     }
	int process();

private:
     int 	check();
     int 	set_firsts();
     int 	set_follows();
     int 	set_lasts();
     int 	set_initcode();
     int 	set_parsetable();
     int 	set_optimized_parsetable();
     int	propagate_attr_ref();

public:
	source getsrc_fromfile(const char *);

	/* for debugging */
     int 	print_firsts(ostream& out=cerr);
     int 	print_follows(ostream& out=cerr);
     int 	print_lasts(ostream& out=cerr);
	int 	print_varref(ostream& out=cerr);
     
     virtual source 	*gencode();

     source 	gen_define();
	source 	*gen_tokenfuncs();
	source 	gen_tokenfunc(TermDef *, source *);
	int 		check_attr_ref();	

};

#endif /* _KGRAM_H */
