/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: pegc.cc,v 1.14 2001/04/05 18:40:24 stevek Exp $
 *
 * Author: Calvin Ko
 *
 * Purpose: Source for the main routine of the PE-Grammar compiler.
 *
 */


#include <stdio.h>
#include <libio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <iostream.h>
#include <string>

#include "conf.h"

#include "pegc.h"
#define _CC "cc"
#define WR_UNIX

#ifdef WR_UNIX
#define PREPROCESSFLAGS "-I. -DWR_IN_WRAPC=1 -DWR_KERNEL_MODULE=1 -x c -E"
#endif
#ifdef WR_WINNT
#define PREPROCESSFLAGS "-nologo -DWR_IN_WRAPC=1 -DWR_KERNEL_MODULE=1 -DWSSLINKAGE=IMPORT -D_WIN32_WINNT=0x0400 -D_INC_EXCPT -D_UNICODE -DUNICODE /MT /TC /EP"
#endif

extern "C" {
const char *pathfind( const char *path, const char *name, const char *mode );
}

int wr_parse_prototypes(FILE *fp);

int dflag = 0;				// debug flag
int debugCode = 0;			// generate debug code?

string target = "wdl";		// target output platform.
char *outfilename = NULL; 	// Leave as NULL for wrapc.out.c

static string ptargetflag = "-DPEG_WDL";

void usage() 

{
     cerr << "pegc [-I incdir] [-Dmacro[=defn]] [-g] [-v] [-d dflag] "
  	  << "wrspec-file\n";
     cerr << "  -g generate debugging code \n";
     cerr << "  -d dflag --  Print Basic Grammar Info      : 16 \n";
     cerr << "            Print Grammar Symbols         : 32 \n";
     cerr << "            Print Grammar Rules           : 64 \n";
     cerr << "            Print Parse Table             : 128 \n"; 
     cerr << "            Print Optimized Parse Table   : 256 \n";
     cerr << "            Print Terminal Definitions    : 512 \n";
     exit(1);
}

int
main(int argc, char **argv)
{
	int error = 0;
	int fPreprocess = 0;
	FILE *fp;
	char buff[1024];
	char *out;
	char *suf;
	char *include_dir = strdup("");
	char *define_list= strdup("");
	char *old_incdir;
	char *old_deflist;
	const char *cc;
	extern char *optarg;
	extern int optind;
	int ch, len;
	while ((ch = getopt(argc, argv, "d:t:I:D:vgE")) != EOF)
	switch(ch) {
	case 'd':		/* debug flag */
		dflag |= atoi(optarg);
		break;
	case 'I':		/* Include directories */
          old_incdir = include_dir;
          len = strlen(old_incdir) + strlen(optarg)+strlen("-I ")
              + 2;
          include_dir = (char*)malloc(len);
          snprintf(include_dir, len, "%s -I %s",old_incdir,optarg);
          free(old_incdir);
          break;
     case 'D':		/* Define list */
          old_deflist = define_list;
          len = strlen(old_deflist) + strlen(optarg)+strlen("-I ")
              + 2;
          define_list = (char*)malloc(len);
          snprintf(define_list, len, "%s -D %s",old_deflist,optarg);
          free(old_deflist);
          break;
	case 'g': 	/* generate debug code */
        	debugCode = 1;	
          break;	
     case 'E':		/* Preprocess 	*/
          fPreprocess = 1;
          break;
	case 't':		/* Target: wdl, wmo, caz, lkm */
		target = optarg;
		if (target == "wdl")
			ptargetflag = " -DPEG_WDL ";
		else if (target == "cpp")
			ptargetflag = " -DPEG_GENERIC ";
		else if (target == "java")
			ptargetflag = " -DPEG_GENERIC ";
		else if (target == "wmo")
			ptargetflag = " -DPEG_WMO ";
		else if (target == "lkm")
			ptargetflag = " -DPEG_KLM ";
		else
			ptargetflag = "";
		break;
     case 'v':		/* version	*/
          fprintf(stderr, "TIS PEGram Compiler version 0.5\n");
          usage();
     case '?':
     default:
          usage();
	}
     argc -= optind;
     argv += optind;

	if (argc < 1) {
		usage();
	}

	if( ( ( cc = getenv( "CC" ) ) == NULL ) &&
	    ( ( cc = pathfind( getenv( "PATH" ), "gcc", "rx" ) ) == NULL ) &&
	    ( ( cc = pathfind( getenv( "PATH" ), "cc", "rx" ) ) == NULL ) &&
            ( ( cc = _CC ) == NULL ) )
	{
		fprintf( stderr, "ERROR unable to determine compiler\n" );
          free(include_dir);
          free(define_list);
		exit( ENOENT );
	}

	while (argc) {
		snprintf(buff, 1024, "%s %s %s %s %s %s", cc,
                        define_list, include_dir, PREPROCESSFLAGS, ptargetflag.c_str(), *argv);

		if (!(fp = popen(buff,"r"))) {
          	fprintf(stderr, "Couldn't open file %s\n", *argv);
               exit(1);
		}
		if (fPreprocess) {
          	size_t count;
          	char bigbuf[2048];
          	/* Dump it all to standard out and quit */
          	while ((count = read(fileno(fp),bigbuf,sizeof bigbuf)))
				write(fileno(stdout),bigbuf,count);
          	exit(0);
          }
		out = strdup(*argv);
		suf = strrchr(out, '.');
		if (suf != NULL && (strcmp(suf, ".peg") == 0)) {
			if (target == "wdl") {
				strcpy(suf, ".wr");
			} else if (target == "cpp") {
				strcpy(suf, ".cc");
			}
               outfilename = strdup(out);
		} else {
		  	fprintf(stderr, "Warning: File name %s does not end in .peg\n", out);
		}

		free(out);
		error |= wr_parse_prototypes(fp);
		argv++, argc--;
		pclose(fp);
	}

     free(include_dir);
     free(define_list);
	return(error);
}
