/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 *  smodule.h: contain definitions of classes for holding events
 * 
 *  Author: Calvin Ko (calvin_ko@nai.com)
 *
 *  $Id: smodule.h,v 1.18 2001/02/23 19:02:03 stevek Exp $
 */

#ifndef _KO_SMODULE_H
#define _KO_SMODULE_H
#include <map.h>
#include <string>
#include "type.h"

/** Class for holding an attributes of a system operation */
class Attr {
public:
     char   *name;
     Type   *type;
     string tname;
     StrSet init_code;		/* initialization code, generic 		*/
     StrSet initpre_code;	/* initialization code for pre event 	*/
     StrSet initpostA_code;	/* initialization code for post event that
						 * need to perform before the operation	*/
     StrSet initpostB_code;	/* initialization code for post event  	*/
	source *fcode;			/* for holding the member function code */

     Attr (const char *n, Type *t) {
          name = strdup(n);
          type = t;
		tname = "env->" + string(n);
     };

	/** Get the initialization code for pre event 	*/
	StrSet& getPreCode() {
		if (!initpre_code.empty())
			return initpre_code;
		else
			return init_code;
	}

	/** Get the initialization code for a post event that need to
      *  be performed before the operation 			
 	 */
	StrSet& getPostACode() {
		return initpostA_code;
	}

	/** Get the initialization code for a post event that need to
	 * be performed after the operation
 	 */
	StrSet& getPostBCode() {
		if (!initpostB_code.empty())
			return initpostB_code;
		return init_code;
	}

};

struct ltstr {
     bool operator()(const char* s1, const char* s2) const {
         return strcmp(s1, s2) < 0;
     }
};

typedef map<string,  Attr *, less<string> > AttrTbl;
typedef list<Attr *> AttrList;
typedef list<const char *> CStrList;

class SModule;

/** Class for holding system events, defined in <system>.mod */
class SEvent {
public:
     string name;			/* Name of the event */
     SModule *m;			/* pointer to the module containing the event */
     AttrTbl atbl;

     /** SEvent constructor, the name could be of the form
      * name^, name+, name-, or name
      */
     SEvent (const char *n) {
          name = n;
          m = NULL;
     }

     /** Insert an attribute. Return 1 if successful, otherwise -1 */
     int insert(Attr *a) {
          if (atbl.insert(make_pair(a->name, a)).second)
               return 1;
          else
               return -1;
     }

     /** find the attibute named s and return a reference to it */
     Attr *find(string s) {
          AttrTbl::iterator iter = atbl.find(s);
          if (iter == atbl.end())
              return NULL;
          else
              return (*iter).second;
     }
     /** find the attibute named s and return a reference to it */
     Attr *operator [] (string s) {
          return find(s);
     }
};


/** Audit Event */
class AEvent {
public:
     enum Flag {All, Pre, Post, Succ, Fail};
     Flag	flag;
     SEvent     *sevent;
     AEvent(SEvent *s=NULL, Flag f=Succ) {
          flag = f;
          sevent = s;
     };

     int operator==(const AEvent& e) const {
          return flag == e.flag && sevent == e.sevent;
     }
     int operator<(const AEvent& e) const {
          int i = (sevent->name > e.sevent->name);
          return i > 0 ? i : (flag < e.flag); 
     }
     AEvent& operator=(const AEvent& e) {
          flag = e.flag;
          sevent = e.sevent;
          return *this;
     }

     /** Return true if Pre event */
     int	isPreEvent() {
     	return flag == All || flag == Pre;
     }

     /** Return true if Post event */
    	int 	isPostEvent() {
    		return flag == All || flag == Post || flag == Succ || flag == Fail;
    	}

     /** find attribute */
     Attr	*find(const string& attr) const {
     	return sevent->find(attr);
     }

     string getName() const {
     	string s = sevent->name;
		switch(flag) {
		case Pre:		return s + '^';
		case Succ:	return s + '+';
		case Fail:	return s + '-';
		case Post:	return s;
		default:		return s;
		}
	}
	
	string getSysOpName() const {
		return sevent->name;
	}

	/** print itself */
	ostream& print(ostream& out) const {
		out << sevent->name;
		switch(flag) {
		case Pre:		out << '^';
					return out;
		case Succ:	out << '+';
					return out;
		case Fail:	out << '-';
					return out;
		case Post:	return out;
		}
	}

	/** operator print itself */
	friend ostream& operator << (ostream& out, const AEvent&);
	
	
};

/** Audit Event Set */
class AEvSet {
public:
     set<AEvent>      es;
     /* 1 means successful event, 0 means failed event */
     AEvSet() {
     };

	/** Insert a new audit event */
     int insert(SEvent *ev, AEvent::Flag f=AEvent::Succ) {
         es.insert(AEvent(ev, f)); 
     }

	/** Insert a new audit event */
     int insert(AEvent e) {
         es.insert(e); 
     }

     /** get the first event name in the set, used for term name generation */
     string get1stName() {
     	if (!es.empty()) {
     		return (*es.begin()).sevent->name;
     	} else {
     		return "NULL";
     	}
     }

     /** check attr ref */
     int chkAttrRef(const string& attr, string& event) {
		set<AEvent>::iterator it;
		for (it = es.begin(); it != es.end(); it++) {
			if ((*it).find(attr) == NULL) {
				event = (*it).getName();
				return 0;
			}
		}
		return 1;
     }

	/** equal(==) operator */
	int operator == (const AEvSet& s) {
		return es == s.es;
	}
     typedef set<AEvent>::iterator iterator;
     iterator begin() {
          return es.begin();
     }
     iterator end() {
          return es.end();
     }

	/** print itself */
	ostream& print(ostream& out) const {
		set<AEvent>::iterator it;
		out << '(';
		for (it = es.begin(); it != es.end(); it++) {
			if (it != es.begin()) out << ", ";
			out << (*it);	
		}
		out << ')';
	}
	
	/** operator print itself */
	friend ostream& operator << (ostream& out, const AEvSet&);

	
};

typedef map<string, SEvent *, less<string> > EvTbl;

/** System Audit Interface Module */
class SModule {
public:
     string name;
     EvTbl  evtbl;
     SModule (string n) {
          name = n;
		SEvent *e = new SEvent("*");
		e->m = this;
          evtbl.insert(make_pair(e->name, e));
     }

	/**Insert an SEvent into the Module. 
	 * Return 1 if successful, 0 otherwise 
	 */
     int insert(SEvent *e) {
          if (evtbl.insert(make_pair(e->name, e)).second) {
               e->m = this;
               return 1;
          } else
               return 0;
     }

	/** Find an SEvent. Result NULL if not found */
     SEvent *find(string s) {
          EvTbl::iterator iter = evtbl.find(s);
          if (iter == evtbl.end())
              return NULL;
          else
              return (*iter).second;
     }
    
	/** Find an attribute, Return NULL if not found */
     Attr *findAttr(string s) {
          Attr *a;
          EvTbl::iterator iter;
          for (iter = evtbl.begin(); iter != evtbl.end(); iter++) {
               a = (*iter).second->find(s);
               if (a != NULL) return a;
          }
          return NULL;
     } 
};

typedef map<string, SModule *, less<string> > ModTbl;

inline ostream& operator<<(ostream& out, const AEvent& a) 
{
	return a.print(out);
}

inline ostream& operator<<(ostream& out, const AEvSet& s) 
{
	return s.print(out);
}

#endif /* _KO_SMODULE_H */

