/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 *  $Id: source.cc,v 1.14 2001/04/05 18:40:24 stevek Exp $
 *
 * Purpose: Object to generate and store C-code.
 *	Source objects are generated at the leaves of the parse
 *	tree, and are merged as rules are reduced to be a single source
 * 	object when parsing completes.  The object can then dump its C text.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include <iostream.h>
#include <strstream.h>

#include "conf.h"

#include "source.h"
#include "expr.h"

source::source() 
{
     head = tail = cur = NULL;
     type = NULL;
     vname = NULL;
     num = 0;
}

source::source(const source& s) 

{
     if (s.head == NULL) {
         head = tail = cur = NULL;
     } else {
         head = new srcrec(s.head->txt);
         tail = cur = head;
         srcrec *srec = s.head;
         while (srec->next != NULL) {
              srec = srec->next;
              tail->next = new srcrec(srec->txt); 
              tail = tail->next;
         }
     }
     type = NULL;
     vname = NULL;
     num = 0;
}

source::source(const char *s, int n=0) 

{
     if (s != NULL && *s != '\0')
          head = tail = new srcrec(s);
     else
          head = tail = NULL;
     cur = NULL;
     type = NULL;
     vname = NULL;
     num = n;

}

source::source(const char *s, Type *t) 

{
     if (s != NULL && *s != '\0')
          head = tail = new srcrec(s);
     else
          head = tail = NULL;
     cur = NULL;
     type = t;
     vname = NULL;
     num = 0;

}

source::~source()
{
     while (head != NULL) {
          cur = head;
          head = head->next;
          delete cur;
     };
}

source source::dup()                                                                            
{
     return *this;
}

int source::is_empty()

{
     return (head == NULL);
}

source& source::operator<<(const source& s)

{
     if (s.head == NULL)
          return *this;
     if (head == NULL) {
          *this = s;
     } else {
          srcrec *srec = s.head;
          tail->next = new srcrec(s.head->txt);
          tail = tail->next;
          while (srec->next != NULL) {
               srec = srec->next;
               tail->next = new srcrec(srec->txt); 
               tail = tail->next;
          }
     }
     return *this;
}

source &source::operator=(const source& s) 

{
     if (s.head == NULL) {
         head = tail = cur = NULL;
     } else {
         head = new srcrec(s.head->txt);
         tail = cur = head;
         srcrec *srec = s.head;
         while (srec->next != NULL) {
              srec = srec->next;
              tail->next = new srcrec(srec->txt); 
              tail = tail->next;
         }
     }
     type = NULL;
     vname = NULL;
     num = 0;
     return *this;
};


void source::make_header(source *link_decl)
{
}


/* Append some ascii text; probably should use the SLIST macro. */
source& source::append(const char *txt)

{
    /* playing safe */
    if (txt == NULL || *txt == '\0')
         return *this;
    if (head == NULL) {
         head = tail = new srcrec;
         head->txt = strdup(txt);
         head->next = NULL;
    } else {
         struct srcrec *tmp = new srcrec;
         tmp->txt = strdup(txt);
         tail->next = tmp;
         tail = tmp;
         tail->next = NULL;
         tmp = NULL;
    }
    return *this;
}

source& source::append(char c)

{
     char sbuf[8];
     sbuf[0] = c;
     sbuf[1] = '\0';
     return append(sbuf);
}

void source::prepend(const char *txt)
{
    if (txt == NULL || *txt == '\0')
        return;
    if (head == NULL) {
        append(txt);
    } else {
        struct srcrec *tmp = new srcrec;
        tmp->txt = strdup(txt);
        tmp->next = head;
        head = tmp;
        tmp = NULL;
    }
}

/* append an integer */
source& source::append(int n)

{
     char buf[64];
     sprintf(buf, "%d", n);
     return append(buf);
}

/* append an unsigned integer */
source& source::append(unsigned int n)

{
     ostrstream oss;
     oss << n;
     return append(oss.str());
}

/* append a long integer */
source& source::append(long n)

{
     ostrstream oss;
     oss << n;
     return append(oss.str());
}

/* append an unsigned long */
source& source::append(unsigned long n)

{
     ostrstream oss;
     oss << n;
     return append(oss.str());
}

/* append another source obj */
source& source::append(source *src)

{
     if (src == NULL || src->head == NULL)
          return *this;
     if (head == NULL) {
          head = src->head;
          tail = src->tail;
     } else {
          tail->next = src->head;
          tail = src->tail;
     }
     src->head = src->tail = src->cur = NULL;	/* empty second object */
     return *this;
}

 

/* append another expression obj */
source& source::append(Expr *e)

{
    return append(e->code);
}

/* append a string */
source& source::append(const string& s)

{
    return append(s.c_str());
}

int source::operator==(char *s)

{
     char *s1 = all();
     int i = !strcmp(s, s1);
     free(s1);
     return i;
}

void source::rewind()
{
	cur = head;
}

/* indent the text by the space given in space */
void source::indent(const char *space="     ")

{
     struct srcrec *tmp = head;
     // insert leading indent space for the first line
     if (tmp != NULL) {
           head = new srcrec;
           head->txt = strdup(space);
           head->next = tmp;
     };
     while (tmp != tail) {
          // insert indent space after each \n */
          if (tmp->txt[strlen(tmp->txt) - 1] == '\n') {
               struct srcrec *srec = new srcrec;
               srec->next = tmp->next;
               srec->txt = strdup(space);
               tmp->next = srec;
               tmp = srec->next;
          } else 
               tmp = tmp->next;
     }
};

/* Return true if there is more text to retrieve via next(). */
bool source::more_text()
{
	if (cur) return true;
	else return false;
}



/* Return next text string in the object, and step forward. */
char *source::next()
{
	if (!cur) return(NULL);
	struct srcrec *tmp = cur;
	cur = cur->next;
	return(tmp->txt);
}


/* Return a single text string containing everything in the object */
char *source::all()
{
  char *tmp;
  char *ret;
  size_t len = 0;
  size_t offset = 0;

  rewind();
  while (more_text()) 
    len += strlen(next());
  ret = (char *) malloc(len + 1);
  if (ret == NULL)
    return NULL;
  rewind();
  while (more_text()) {
    tmp = next();
    len = strlen(tmp);
    strncpy(ret + offset, tmp, len);
    offset += len;
  }
  ret[offset] = '\0';
  return ret;
}

int source::rec_count() 

{   int i = 0;
    rewind();
    while (more_text()) {
        next();
        i++;
    }
    return i;
}

bool source::find(char *txt)
{
  bool rtn = false;

  rewind();
  while(more_text()) {
    if(strcmp(cur->txt, txt) == 0) {
      rtn = true;
      break;
    }
    cur = cur->next;
  }
  return rtn;
}

// insert txt after cur pointer;
void source::insert(char *txt)

{
     if (txt == NULL || *txt == '\0') return;
     if (cur == NULL) {
          struct srcrec *tmp = head;
          head = new srcrec;
          head->txt = strdup(txt);
          head->next = tmp;
     } else if (cur == tail) {
          struct srcrec *tmp = new srcrec;
          tmp->txt = strdup(txt);
          tmp->next = NULL;
          cur->next = tmp; 
          tail = tmp; 
     } else {
          struct srcrec *tmp = new srcrec;
          tmp->txt = strdup(txt);
          tmp->next = cur->next;;
          cur->next = tmp; 
     }
};

void source::remove()
{
  srcrec *prev;

  if (!cur)
    return;

  if (head == cur) {
    prev = NULL;
    head = head->next;
    if (tail == cur)
      tail = head;
  }
  else {
    for (prev = head; prev->next != cur; prev = prev->next);
    prev->next = cur->next;
    if (tail == cur)
      tail = prev;
  }
  if (cur->txt)
    free(cur->txt);
  free(cur);
  if (prev)
    cur = prev->next;
  else
    cur = head;
}


char *source::text()
{
  if (!cur)
    return NULL;
  else
    return cur->txt;
}


void source::gen_tbd()
{
  append ("\n/* Code to be added TBD */\n");
}


void source::gen_tbd(char *txt)
{
  append ("\n/* Code to be added TBD for ");
  append (txt);
  append (" */\n");
}


void source::gen_debug_code()
{
	append("\n/* Test driver code. */\n#include \"debug.c\"\n");
}


void source::gen_ascii(char *str)
{
        append("\n#define MYNAME PD_MAKESTR(\"");
	append(str);
	append("\")\n");
	append("\n");
        append("static pd_char_t *WRLINK_MOD_NAME = MYNAME;\n");
        append("static pd_char_t *myname=MYNAME;\n\n");

        append("pd_char_t EXPORT *get_mod_name()        \n");
        append("{                                  \n");
        append("  return WRLINK_MOD_NAME;          \n");
        append("}                                  \n\n");
}

void source::gen_comment(char *str, char *str1=NULL)

{
	append("/* ");
        append(str);
        append(str1);
        append(" */\n");
}

void source::gen_nterm_funcdecl(const string& name)

{
      append("static int "); 
      append(name);
      append("(ThParser *tp, int token, int *accept, Env *env, LocalEnv *lenv)\n");
}


extern int table_elem_count;
extern VarStore vstore;
extern const char *current_table;
void wdlerror(const char *s, const char *s1 = "");
int wdlwarning(const char *s, const char *s1 = "");


class srcstream : strstream {
      srcstream() : strstream() {};
};
