/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: auditd.c,v 1.11 2001/02/23 19:02:04 stevek Exp $
 *
 * Author: Calvin Ko 
 *
 * Wrapper audit event viewer
 *
 * This application registers itself with the WSS, sets up a
 * raw event handler, and displays events as they are posted.
 */

#include "wr_conf.h"

#ifndef LINKAGE
#define LINKAGE IMPORT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#ifdef WR_WINNT
#include "getopt.h"
#else
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#endif /* WR_WINNT */

#include <eventfs.h>
#include <libeventfs.h>
#include <libwrv.h>

/* For _PATH_EVENTFS */
#include EFS_SYSTEM_HDR

void errorMessage( char *msg );

#define MAXBUFFERLEN 2048
char *binToAscii(char *buffer, unsigned long buffer_length)
{
    static char output[MAXBUFFERLEN]; /* Truncate if too long */
    size_t offset_in, offset_out;

    /* If the character is ascii, print it.
     * Else print '\xxx' where xxx is the octal representation.
     */
    for (offset_in=0, offset_out=0 ;
            ( offset_in < buffer_length ) && ( offset_out < (MAXBUFFERLEN-1));
            ++offset_in ) {
        if (isprint((int)buffer[offset_in])
                || buffer[offset_in] == '\n'
                || buffer[offset_in] == '\t') {
            output[offset_out] = buffer[offset_in];
            ++offset_out;
        } else {
            _sntprintf( output+offset_out,MAXBUFFERLEN - offset_out - 1,
                    PD_MAKESTR("\\%3.3o"), (int)buffer[offset_in] );
            offset_out += 4;
        }
    }

    if (offset_out >= MAXBUFFERLEN) {
        /* indicate so in the output */
        output[MAXBUFFERLEN - 4] = '.';
        output[MAXBUFFERLEN - 3] = '.';
        output[MAXBUFFERLEN - 2] = '.';
    }

    output[offset_out] = '\0';

    return output;
}


#ifdef WR_UNIX
#define errorMessage(msg) perror(msg)
#else
void errorMessage( char *msg )
{
   LPVOID lpMsgBuf;

   FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                  FORMAT_MESSAGE_FROM_SYSTEM |
                  FORMAT_MESSAGE_IGNORE_INSERTS,
                  NULL, GetLastError(),
                  MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
                  (LPTSTR) &lpMsgBuf, 0, NULL );
   _tprintf( _TEXT("%s: %s\n"), msg, (LPCTSTR) lpMsgBuf );
   LocalFree( lpMsgBuf );
}
#endif

            
void dumpEvent(eventfs_event_stream_t stream, eventfs_event_t event)
{
    /* Print out group name, type name, buffer. */
    u_long              mask;
#ifdef DEBUG_AUDIT
    int                 i;
    const char         *name;
#endif
    eventfs_event_group_t   group;

    mask = eventfsEventGetTypes(event);
    group = eventfsEventStreamGetGroup(stream);

#ifdef DEBUG_AUDIT
    printf("Event: group = %-20.20s Mask = %-8.8lo Size = %ld\n",
            eventfsEventGroupGetName(group), mask, eventfsEventGetSize(event));
    printf("Types:\n");

    for (i = 0; i < 32; ++i) {
        if (mask & 1<<i) {
            name = eventfsEventGroupGetTypeName(group,1<<i);
            if (name)
                printf("\t%s\n",name);
            else
                printf("\tinvalid type\n");
        }
    }
    printf("Buffer:\n");
#endif
    fputs(binToAscii(eventfsEventGetData(event), eventfsEventGetSize(event)),stdout);

}

int main( int argc, char *argv[] )

{
    
    u_long 	mask 	= 0;
    int		ch;

    eventfs_event_group_t 	group;
    eventfs_event_stream_t 	stream;
    eventfs_event_t		event;
    while ((ch = getopt(argc, argv, "pahi:")) != -1) {
        switch(ch) {
        case 'a':
            mask = 0xFFFFFFFF;
            break;
	case 'i':
	    mask |= (1 << atoi(optarg));
	    break;
        case 'h':
            fprintf(stderr, "auditd [-help] [-a] [-i num]* \n");
            fprintf(stderr, "\t-a:\tDisplay all audit events\n");
            fprintf(stderr, "\t-i num:\tDisplay audit events in queue num (0-32)\
n");
            exit(0);
       default:
            fprintf(stderr, "See auditd -h for help\n");
            exit(0);
       }
    }

    /*
     * Register with the WSS
     */
    if( wrvInit() < 0 ) {
         perror( "wrvInit" );
         exit( 0 );
    }

    /*
     * Open the group
     */                                                                 
    group = eventfsEventGetGroup(WREV_AUDIT_GROUP);
    if (group == 0) {
        perror("eventfsEventGetGroup("WREV_AUDIT_GROUP")");
        exit(1);
    }

    /*
     * Open the stream
     */
    if ( (stream = eventfsEventGroupGetStream(group, mask)) == NULL) {
        errorMessage( "eventfsEventGroupGetStream" );
        return EXIT_FAILURE;
    }

   /*
     * Sleep forever...
     */
    printf( PD_MAKESTR("Masks = %04X\n"), (unsigned int)mask );
    _tprintf( PD_MAKESTR("Waiting for events...\n") );
    for(;;)
    {
        if( (event = eventfsEventStreamGetNextEvent(stream)) != 0 )
        {
            dumpEvent( stream, event );
            eventfsEventRelease( event );
        }
        /* This only happens on a significant malfunction. */
        else
            break;
    }

    return EXIT_FAILURE;
    
}

