/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: Data.cc,v 1.9 2001/04/05 18:40:28 stevek Exp $
 * 
 * File: Data.cc
 * 
 * Author: Douglas Kilpatrick
 * 
 * Purpose: Classes to encapuslate the data of the sequence information
 * 
 * Classes:
 *      Sequence: a sequence of system call numbers for a specific process
 *      Tree: a pseudo-N-ary tree to hold possible sequences
 *      DB: an encapsulation of different tree's for different processes
 */

#include "wr_conf.h"

#ifndef DATA_H
#include "Data.h"
#endif

#include <iostream.h>
#include <fstream.h>

#include <iomanip.h>

int Sequence::get_length()
{
    return l.size();
}

void Sequence::push(int value)
{
    l.push_back(value);
}

int Sequence::pop()
{
    int ret_val;

    ret_val = l.front();

    l.pop_front();

    return ret_val;
}

// Copy constructor.
Sequence::Sequence(Sequence &orig)
{
    pdi = orig.pdi;
    copy(orig.l.begin(),orig.l.end(), back_inserter(l));
}

bool BranchNode::match(Sequence *seq)
{
    if (seq->is_empty()) {
        cerr << "DB error" << endl;
        return false;
    }
    int index = seq->pop();
    syscallNum_to_tree_t::iterator it;
    it = syscallNum_to_tree.find(index);
    if (it == syscallNum_to_tree.end())
        return false;
    return ((*it).second)->match(seq);
}

bool BranchNode::merge(Sequence *seq)
{
    int index = seq->pop();
    SyscallTree *subtree = syscallNum_to_tree[index];
    if (subtree == NULL && seq->is_empty()) {
        subtree = new LeafNode();
        syscallNum_to_tree[index] = subtree;
    } else if (seq->is_empty()) {
        return false;
    } else if (subtree == NULL) {
        subtree = new BranchNode();
        syscallNum_to_tree[index] = subtree;
    }
    return subtree->merge(seq);
}

bool BranchNode::import(istream &infile, int depth)
{
    bool error = false;
    while (infile && !error) {
        int syscall_num;
        SyscallTree *tree;

        infile >> syscall_num;
        if (syscall_num == -1)
            break;

        if (depth > 0) {
            tree = new BranchNode();
        } else {
            tree = new LeafNode();
        }
        syscallNum_to_tree[syscall_num] = tree;
        error = tree->import(infile , depth - 1);
    }

    return error;
}

void BranchNode::export_to_stream(ostream &outfile, int depth)
{
    syscallNum_to_tree_t::iterator it;
    it = syscallNum_to_tree.begin();

    while (it != syscallNum_to_tree.end()) {
        SyscallTree *tree;
        outfile << (*it).first << " ";
        tree = (*it).second;
        tree->export_to_stream(outfile, depth + 1);
        if (depth == 0)
            outfile << endl;
        ++it;
    }
    outfile << "-1    ";
}

BranchNode::~BranchNode()
{
    // iterate through the map and nuke all the trees.
    syscallNum_to_tree_t::iterator it, temp;

    it = syscallNum_to_tree.begin();

    while (it != syscallNum_to_tree.end()) {
        temp = it;
        ++it;

        delete (*temp).second;
        syscallNum_to_tree.erase(temp);
    }
}

BranchNode *BranchNode::clone()
{
    BranchNode *ret_val = new BranchNode;
    // iterate through the map and nuke all the trees.
    syscallNum_to_tree_t::iterator it;

    it = syscallNum_to_tree.begin();

    while (it != syscallNum_to_tree.end()) {
        ret_val->syscallNum_to_tree[(*it).first] = (*it).second->clone();
        ++it;
    }
    return ret_val;
}

void BranchNode::export_fast(int head, int depth, fast_db& db) {
    syscallNum_to_tree_t::iterator it = syscallNum_to_tree.begin();
    while (it != syscallNum_to_tree.end()) {
        db[head][(*it).first] |= 1<<depth;
        (*it).second->export_fast(head, depth + 1, db);
        ++it;
    }
}

void BranchNode::export_fast(fast_db& db) {
    syscallNum_to_tree_t::iterator it = syscallNum_to_tree.begin();
    while (it != syscallNum_to_tree.end()) {
        db[(*it).first][(*it).first] |= 1;
        (*it).second->export_fast((*it).first, 1, db);
        ++it;
    }
}



bool LeafNode::match(Sequence *)
{
    return true;
}

bool LeafNode::merge(Sequence *)
{
    return true;
}

LeafNode *LeafNode::clone()
{
    return new LeafNode;
}

SequenceStore::~SequenceStore()
{
    // iterate through the map and nuke all the trees.
    pdi_to_tree_t::iterator it, temp;

    it = pdi_to_tree.begin();

    while (it != pdi_to_tree.end()) {
        temp = it;
        ++it;

        delete (*temp).second;
        pdi_to_tree.erase(temp);
    }
}

pdi_to_tree_t *SequenceStore::export_map()
{
    pdi_to_tree_t *ret_val = new pdi_to_tree_t;
    pdi_to_tree_t::iterator it = pdi_to_tree.begin();

    while (it != pdi_to_tree.end()) {
        (*ret_val)[(*it).first] = ((*it).second)->clone();
        ++it;
    }
    return ret_val;
}

void SequenceStore::export_fast(char *filename, int pdi)
{
    /*
     * Format:
     * array of [syscall count][syscall count][depth] of bool.
     * where db[syscall 1][syscall 2][depth x] == true means that
     * syscall 2 can follow syscall 1 in the xth position.  This does
     * allow the possiblity of false negatives.
     * 
     * Actually, to make matters reasonably fast, depth is coded
     * as a bitvector.  If you limit yourself to a depth of 8, you
     * can fit that axis in a single byte.  If you are on a system with
     * < 256 syscalls, you can do the whole possible table in a 64k array.
     * 
     * So that's what we do.
     */
    
    fast_db db;

    SyscallTree *tree = pdi_to_tree[pdi];
    if (!tree) return;
    
    bzero(db, sizeof db);
    tree->export_fast(db);

    ofstream outFile(filename, ios::out);
    outFile.write(db, sizeof db);
}


bool ROSequenceStore::process(Sequence *seq)
{
    // Clone the sequence to avoid changing the original.
    Sequence copy(*seq);
    int pdi = copy.get_pdi();
    SyscallTree *tree = pdi_to_tree[pdi];
    if (!tree)
        return true;
    return !(tree->match(&copy));
}

#ifdef MIN
#undef MIN
#endif
template < class T >
static inline T & MIN(T & a, T & b)
{
    return (a < b) ? a : b;
}

static int audit_kern_analyze(fast_db &db, Sequence &seq) {
    const int DEPTH = 8;
    int min_ln = DEPTH; /* Start at max for algorythmic cleanliness */
    char ln_matrix[DEPTH+1][DEPTH+1];
    unsigned char sequence[DEPTH];
    int i, j, test_head;

    /* Onetime init sequence array */
    i = 0;
    while (!seq.is_empty()) {
        sequence[i] = seq.pop();
        ++i;
    }

    /* We got here, so there has to be at least one mismatch.
     * The lowest Levenstein distance possible should be 1, so
     * if we have a match at that level, take it and stop looking.
     */
    for (test_head = 0; min_ln > 1 && test_head < 255 ; ++test_head) {
        /* If db[head][head] == 0, then this potential head syscall never
         * actually starts a sequence.  Save time, continue
         */
        if (!(db[test_head][test_head] & 1)) {
            continue;
        }

        /* Init matrix */
        bzero(ln_matrix, sizeof ln_matrix);
        for (i = 1; i < DEPTH+1 ; ++i)
            ln_matrix[i][0] = i;
        for (i = 1; i < DEPTH+1 ; ++i)
            ln_matrix[0][i] = i;

        for (i=1; i < (DEPTH+1) ; ++i) {
            for (j=1; j < (DEPTH+1) ; ++j) {
                int dNW = ln_matrix[i-1][j-1];
                int dW  = ln_matrix[i  ][j-1] + 1;
                int dN  = ln_matrix[i-1][j  ] + 1;

                dNW += ((db[test_head][sequence[i-1]])&(1<<(j-1))) ? 0 : 1;
                
                ln_matrix[i][j] = MIN(dNW,MIN(dW,dN));
            }
        }
        /* Read the levenstein number out of the matrix */
        if (ln_matrix[DEPTH][DEPTH] < min_ln)
            min_ln = ln_matrix[DEPTH][DEPTH];
    }

    return min_ln;
}


void ROSequenceStore::analyze(Sequence *seq)
{
    Sequence copy(*seq);
    int pdi = copy.get_pdi();
    int min_ln;
    
    // Hack hack...
    // To test out the fast analysis techniques from the kernel wrapper,
    // we first do a fast-export to get the equivilent DB, then we reuse
    // the wrapper's code.
    
    SyscallTree *tree = pdi_to_tree[pdi];
    fast_db db;
    bzero(db, sizeof db);
    tree->export_fast(db);

    min_ln = audit_kern_analyze(db, copy);
    cerr << "Alert!.  Mismatch.  Levenshtein distance = " << min_ln << endl;
}

void ROSequenceStore::import(istream &inFile)
{
    bool error = false;
    // Format same as WOSequenceStore::export
    // First depth, then tree structure.  "-1" token means "back up one level"

    inFile >> depth;
    while (inFile && !error) {
        int pdi;
        SyscallTree *tree;

        inFile >> pdi;
        if (pdi == -1)
            break;

        if (depth > 0) {
            tree = new BranchNode();
        } else {
            tree = new LeafNode();
        }
        pdi_to_tree[pdi] = tree;
        error = tree->import(inFile, depth - 1 );
    }
}

bool WOSequenceStore::process(Sequence *seq)
{
    int pdi = seq->get_pdi();
    SyscallTree *tree = pdi_to_tree[pdi];
    if (!tree) {
        tree = new BranchNode(); // assumption
        pdi_to_tree[pdi] = tree;
    }

    return !(tree->merge(seq));
}

void WOSequenceStore::analyze(Sequence *)
{
    // Just a place holder here... Will be called if process fails, which
    // should only happen on malformed input
}

void WOSequenceStore::export_to_stream(ostream &outfile)
{
    outfile << depth << endl;
    pdi_to_tree_t::iterator it;
    it = pdi_to_tree.begin();

    while (it != pdi_to_tree.end()) {
        outfile << (*it).first << " ";
        (*it).second->export_to_stream(outfile, 0);
        ++it;
        outfile << endl;
    }
    outfile << "-1" << endl;
}

int Name_To_PDI::get_pdi(char *name) {
    if (name_to_pdi.find(name) == name_to_pdi.end())
        return -1;
    return name_to_pdi[name];
}

void Name_To_PDI::export_to_stream(ostream &output) {
    name_to_pdi_t::iterator it;
    it = name_to_pdi.begin();
    while (it != name_to_pdi.end()) {
        output << (*it).second << " " << (*it).first << endl;
        ++it;
    }
    output << "-1" << endl;
}

void Name_To_PDI::import(istream &input) {
    int temp;
    char buf[1024];

    while (input) {
        input >> temp;
        if (temp < 0)
            break;
        input >> setw(1024) >> buf;
        buf[1023] = '\0';
        name_to_pdi[strdup(buf)] = temp;
    }
}

Name_To_PDI::~Name_To_PDI()
{
    name_to_pdi_t::iterator it;
    it = name_to_pdi.begin();
    while (it != name_to_pdi.end()) {
        free((*it).first);
        ++it;
    }
}

void DB::import(char *filename)
{
    ifstream inFile(filename, ios::in);
    // File format is name-to-pdi, then trees.
    if (name_map) delete name_map;
    if (store) delete store;
    fRead = false;
    name_map = new Name_To_PDI;
    name_map->import(inFile);
    store = new ROSequenceStore();
    ((ROSequenceStore*)store)->import(inFile);
}

void DB::export_to_stream(char *filename)
{
    if (!fRead)
        return;
    ofstream outfile(filename,ios::out);
    // File format is name-to-pdi, then trees.
    name_map->export_to_stream(outfile);
    ((WOSequenceStore*)store)->export_to_stream(outfile);
}

void DB::export_fast(char *filename, char *process_name)
{
    int pdi = name_to_pdi(process_name);
    if (pdi < 0) return;

    store->export_fast(filename, pdi);
}

int DB::name_to_pdi(char *name) {
    static int counter = 0;
    int retval = name_map->get_pdi(name);
    if (fRead && (retval < 0)) {
        name_map->add(name, counter);
        retval = counter;
        counter++;
    }
    return retval;
}

bool DB::process(Sequence *seq) {
    static int counter = 0;
    ++counter;
    if (fRead && !(counter % 0x00010000)) {
        char buf[1024];
        sprintf(buf,".depth-%d.db",get_depth());
        export_to_stream(buf);
    }
    return store->process(seq);
}

DB::DB(int depth) {
    name_map = new Name_To_PDI();
    store = new WOSequenceStore(depth);
    fRead = true;
}
