/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: Data.h,v 1.8 2001/02/23 19:02:05 stevek Exp $
 * 
 * File: Data.h
 * 
 * Author: Douglas Kilpatrick
 * 
 * Purpose: Classes to encapuslate the data of the sequence information
 * 
 * Classes:
 *      Sequence: a sequence of system call numbers for a specific process
 *      Tree: a pseudo-N-ary tree to hold possible sequences
 *      DataStore: an encapsulation of different tree's for different processes
 */

#ifndef DATA_H
#define DATA_H

#ifdef WR_BROKEN_STL
#include "template/list"
#include "template/map"
#else
#include <list>
#include <map>
#endif

class Sequence {
    private:
        int             pdi;            // Process Data Identifier
        list<int>       l;
    public:
        int             get_length();
        void            push(int value);   // Queue meaning
        int             pop();             // Queue meaning
        inline int      get_pdi() {return pdi;}
        inline bool     is_empty() {return l.empty();}
        Sequence(int p) {pdi = p;};
        Sequence(Sequence &);
};

typedef unsigned char fast_db[255][255];
class SyscallTree {
    public:
        virtual bool match(Sequence *) = 0;
        virtual bool merge(Sequence *) = 0;
        virtual SyscallTree *clone() = 0;
        SyscallTree() { };
        virtual ~SyscallTree() { };
        virtual bool import(istream &stream, int depth) = 0;
        virtual void export_to_stream(ostream &stream, int depth) = 0;
        virtual void export_fast(int head, int depth, fast_db& db) = 0;
        virtual void export_fast(fast_db& db) = 0; // assumes this is head
};

typedef map<int, SyscallTree*, less<int>, alloc> pdi_to_tree_t;
typedef map<int, SyscallTree*, less<int>, alloc> syscallNum_to_tree_t;

class BranchNode: public SyscallTree {
    private:
        syscallNum_to_tree_t syscallNum_to_tree;
    public:
        virtual bool match(Sequence *);
        virtual bool merge(Sequence *);
        virtual BranchNode *clone();
        BranchNode() { };
        virtual ~BranchNode();
        virtual bool import(istream &stream, int depth);
        virtual void export_to_stream(ostream &stream, int depth);
        virtual void export_fast(int head, int depth, fast_db& db);
        virtual void export_fast(fast_db& db);
};

class LeafNode : public SyscallTree {
    public:
        virtual bool match(Sequence *);
        virtual bool merge(Sequence *);
        virtual LeafNode *clone();
        LeafNode() { };
        virtual ~LeafNode() { };
        inline virtual bool import(istream &stream, int depth) {
            int temp;
            stream >> temp;
            if (temp != -1 || depth >= 0) {
                return true;
            }
            return false;
        };
        inline virtual void export_to_stream(ostream &stream, int) {
            stream << "-1  ";
        };
        virtual void export_fast(int, int, fast_db&) { };
        virtual void export_fast(fast_db& ) { abort(); };
};

class SequenceStore {
    protected:
        pdi_to_tree_t   pdi_to_tree;
        int             depth;
    public:
        // Returns true on alert
        virtual bool process(Sequence *) = 0;
        virtual void analyze(Sequence *) = 0;
        int get_depth() { return depth; };
        SequenceStore(int d) {depth = d; }
        SequenceStore() { depth = 0; }
        virtual ~SequenceStore();
        pdi_to_tree_t *export_map();
        void export_fast(char *filename, int pdi);
};

class ROSequenceStore : public SequenceStore {
    public:
        virtual bool process(Sequence *);
        virtual void analyze(Sequence *);
        ROSequenceStore() { };
        virtual ~ROSequenceStore() { };
        void import(istream &);
};

class WOSequenceStore : public SequenceStore {
    public:
        virtual bool process(Sequence *);
        virtual void analyze(Sequence *);
        WOSequenceStore(int d) : SequenceStore(d) { };
        virtual ~WOSequenceStore() { };
        void export_to_stream(ostream &);
};

struct str_less : public binary_function<char *,char*, bool> {
    bool operator()(const char* x, const char* y)
        const { return (strcmp(x,y)<0);}
};

typedef map<char *,int,str_less,alloc> name_to_pdi_t;

class Name_To_PDI {
    private:
        name_to_pdi_t name_to_pdi;
    public:
        inline void add(char *name, int pdi) {name_to_pdi[strdup(name)] = pdi;};
        int get_pdi(char *name);
        void export_to_stream(ostream &);
        void import(istream &);
        ~Name_To_PDI();
};

class DB {
    private:
        Name_To_PDI *name_map;
        SequenceStore *store;
        bool fRead;
    public:
        void import(char *filename);
        void export_to_stream(char *filename);
        void export_fast(char *filename, char *name);
        
        int name_to_pdi(char *name);
        bool process(Sequence *seq);
        inline void analyze(Sequence *seq) { store->analyze(seq); };
        inline int get_depth() { return store->get_depth(); };
        
        DB(int depth);
        ~DB() { delete name_map; delete store; };
};
#endif

