/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: Message.cc,v 1.8 2001/04/05 18:40:28 stevek Exp $
 * 
 * File: Message.cxx
 * 
 * Author: Douglas Kilpatrick
 * 
 * Purpose: Message classes to encapsulate talking to the wrappers
 * event stream.
 * 
 * Classes:
 * 
 *      Source: data source
 *      EventSource: Source subclass that talks to the event manager.
 *      
 *      Message: root message class.
 *      execMessage: subclass for process creation
 *      exitMessage: subclass for process termination
 *      syscallMessage: subclass for syscall execution
 *      quitMessage: subclass for wrapper shutdown
 *
 */

#include "wr_conf.h"

#ifndef MESSAGE_H
#include "Message.h"
#endif

/* Used for strdup() only so far. */
#include <string.h>

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

#include "libeventfs.h"
#include "libwrv.h"


/*
 * Messages First.
 */

/*
 * execMessage
 */
int execMessage::process( DB *db, pid_to_pdi_t *pid_to_pdi, pid_to_seq_t *)
{
    int pdi;
    pdi = db->name_to_pdi(name);
    if (pdi < 0)
        return 0;

    (*pid_to_pdi)[pid] = pdi;

    return 0;
}

execMessage::execMessage(int process_id, const char* process_name)
{
    pid = process_id;
    name = strdup(process_name);
}

/*
 * exitMessage
 */
int exitMessage::process( DB *, pid_to_pdi_t *pid_to_pdi, pid_to_seq_t *)
{
    if (pid_to_pdi->find(pid) == pid_to_pdi->end())
        return 0;

    int pdi = (*pid_to_pdi)[pid];

    pid_to_pdi->erase(pdi);

    return 0;
}

/*
 * syscallMessage
 */
int syscallMessage::process(
        DB *db,
        pid_to_pdi_t *pid_to_pdi,
        pid_to_seq_t *pid_to_seq)
{
    if (pid_to_pdi->find(pid) == pid_to_pdi->end())
        return 0;
    int pdi = (*pid_to_pdi)[pid];

    list<Sequence*> *l = (*pid_to_seq)[pid];
    if (l == NULL) {
        l = new list<Sequence*>;
        (*pid_to_seq)[pid] = l;
    }

    Sequence *seq = new Sequence(pdi);
    l->push_back(seq);

    // iterate over the list, adding "syscall_num" to the end of all
    // sequences.

    list<Sequence*>::iterator it = l->begin();
    while (it != l->end()) {
        (*it)->push(syscall_num);
        ++it;
    }

    // iterate over the list, finding all sequences of length == depth
    // Er well... Since we always add the new sequence to the end of the
    // list, we only have to check to see if the first sequence in the
    // list is ready.

    seq = l->front();

    if (seq->get_length() == db->get_depth()) {
        if (db->process(seq))
            db->analyze(seq);
        l->pop_front();
        delete seq;
    }

    return 0;
}

syscallMessage::syscallMessage(int process_id, int syscall)
{
    pid = process_id;
    syscall_num = syscall;
}

/*
 * quitMessage
 */
int quitMessage::process( DB *, pid_to_pdi_t *, pid_to_seq_t *)
{
    return 1; // Which will kick us out of the loop.
}

/*
 * Sources next
 */
EventSource::EventSource(int num)
{
    eventfs_event_group_t group = eventfsEventGetGroup(WREV_AUDIT_GROUP);
    if (group == 0) {
        perror("eventfsEventGetGroup("WREV_AUDIT_GROUP")");
        exit(1);
    }
    this->stream = eventfsEventGroupGetStream(group,1<<num);
    if (stream == 0) {
        perror("eventfsEventGroup::GetStream()");
        exit(1);
    }
}

EventSource::~EventSource()
{
    eventfs_event_group_t group = eventfsEventStreamGetGroup(stream);

    eventfsEventStreamRelease(stream);
    eventfsEventGroupRelease(group);
}

Message *EventSource::next_message()
{
    Message            *retval;

    char               *buffer;
    unsigned long       buffer_length;
    int                 command;


    /* eventfsEventStreamGetNextEvent tries pretty hard.  If it
     * gives up its not something we can handle.
     */
    eventfs_event_t e;
   
    e = eventfsEventStreamGetNextEvent(stream);
    if (e == 0) {
        perror("eventfsEventStream::GetNextEvent()");
        return new quitMessage();
    }
    if (eventfsEventGetSize(e) < sizeof(int)) {
        perror("eventfsEvent::GetSize()");
        eventfsEventRelease( e );
        return new quitMessage();
    }

    // Create the new message
    // Translate to a new message type.

    command = *((int*)eventfsEventGetData(e));

    buffer = (char *)eventfsEventGetData(e) + sizeof(int);
    buffer_length = eventfsEventGetSize(e) - sizeof(int);

    switch ( command ) {
        int pid, syscall;
    default:
    case 0:
        retval = new quitMessage();
        break;
    case 1:
        if (buffer_length > 0) {
            pid = *((long*)buffer);
            buffer += sizeof(long);
            // Safe, because execMessage clones the string
            // of course, a bad buffer still hoses us.
            retval = new execMessage(pid,buffer);
        } else {
            retval = new quitMessage();
        }
        break;
    case 2:
        if (buffer_length > 0) {
            pid = *((long*)buffer);
            retval = new exitMessage(pid);
        } else {
            retval = new quitMessage();
        }
        break;
    case 3:
        if (buffer_length > 0) {
            pid = *((long*)buffer);
            buffer += sizeof(long);
            syscall = *((long*)buffer);
            retval = new syscallMessage(pid,syscall);
        } else {
            retval = new quitMessage();
        }
        break;
    }

    eventfsEventRelease( e );

    return retval;
}

FileSource::FileSource(char *filename)
{
    ifstream *inFile = new ifstream(filename, ios::in);
    input_stream = inFile;
}

FileSource::~FileSource()
{
    delete input_stream;
}

Message *FileSource::next_message()
{
    int command, pdi, syscall;
    char buf[1024];

    if (!(*input_stream)) return new quitMessage();
    (*input_stream) >> command;
    if (!(*input_stream)) return new quitMessage();
    switch(command) {
    default:
    case 0:
        return new quitMessage();
    case 1:
        (*input_stream) >> pdi;
        if (!(*input_stream)) return new quitMessage();
        (*input_stream) >> setw(1024) >> buf;
        if (!(*input_stream)) return new quitMessage();
        buf[1023] = '\0';
        return new execMessage(pdi,buf);
    case 2:
        (*input_stream) >> pdi;
        if (!(*input_stream)) return new quitMessage();
        return new exitMessage(pdi);
    case 3:
        (*input_stream) >> pdi;
        if (!(*input_stream)) return new quitMessage();
        (*input_stream) >> syscall;
        if (!(*input_stream)) return new quitMessage();
        return new syscallMessage(pdi,syscall);
    }
}


