/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ari.ch
 *
 * Author:  Tim Fraser
 *
 * This characterization defines the "Audit Record Interface" between
 * the concrete Audit Record Preprocessor (arp.wr) wrapper and other
 * abstract wrappers.  These wrappers cooperate to implement intrusion
 * detection functionality based on the method described in:
 *
 *    Koral Ilgun, Richard A. Kemmerer, and Philip A. Porras, "State
 *    Transition Analysis:  A Rule-Based Intrusion Detection Approach,"
 *    IEEE Transactions on Software Engineering, Vol. 21, No 3, March 1995.
 * 
 * The Audit Record Interface consists of a single "Audit Record"
 * event type.  The parameters associated with Audit Record events
 * correspond to the USTAT Audit Record Structure described in the
 * paper, above, in figure 7, page 191.
 *
 * Ideally, the Audit Record Interface would create an abstraction
 * barrier between the concrete (OS platform-specific) Audit Record
 * Preprocessor wrapper and all the cooperating abstract (OS
 * platform-independent) wrappers that would hide all
 * platform-specific details.  Unfortunately, such hiding is difficult
 * considering the significant differences between the basic
 * abstractions found in UNIX and NT.  The USTAT Audit Record
 * Structure is probably too UNIX-oritented an abstraction for our
 * purposes, but it seems like a good starting place, at least.
 *
 */


/* The following enumerated type defines the possible values of the  *
 * action field of the Audit Record event.  It is based on the USTAT *
 * action list presented in the Ilgun paper, table III, page 191.    *   
 * This should be an enumerated type, but the WDB only groks ints,   *
 * and type casting in WQL insert statements is problematic.  So,    *
 * I've used preprocessor constants and a u_int type, instead.       */

#define ARI_ACTION_INVALID      0
#define ARI_ACTION_READ         1
#define ARI_ACTION_WRITE        2
#define ARI_ACTION_CREATE       3
#define ARI_ACTION_DELETE       4
#define ARI_ACTION_EXECUTE      5
#define ARI_ACTION_EXIT         6
#define ARI_ACTION_MODIFYOWNER  7
#define ARI_ACTION_MODIFYPERM   8
#define ARI_ACTION_RENAME       9
#define ARI_ACTION_LINK        10


/* This is the interface characterization itself.  I've used generic  *
 * types below, theorizing that this would abstract away the          *
 * platform-specific types.  However, this act defeats the compiler's *
 * type-checking, and may not be a win overall.                       *
 *                                                                    *
 * Presently, there is a practical limit on the number of fields an   *
 * abstract event can have (about 8).  Since the USTAT Audit Record   *
 * Structure requires more than 8 fields, I have encapsulated all of  *
 * the fields in a single structure of type audit_record_fields_t.    *
 * Each audit_record abstract event contains a single field that is a *
 * pointer to an instance of the audit_record_t structure.            */

typedef struct audit_record_fields {
  unsigned long uid;       /* real user ID */
  unsigned long euid;      /* user's effective uid */
  unsigned long gid;       /* user's group ID */
  unsigned int  action;    /* action (operation) ID */
  unsigned long time;      /* time in secs since UNIX epoch */
  unsigned long pid;       /* process ID */
  char *        oname;     /* object name */
  unsigned long perm;      /* perms associated with object */
  unsigned long ooid;      /* object owner ID */
  unsigned long ogid;      /* object group ID */
  unsigned long oid;       /* object ID */
  unsigned long devid;     /* device ID */
  unsigned long fsid;      /* file system ID */
  void *        tname;     /* target name for link actions */
} audit_record_fields_t;

typedef struct proc_rec {
  unsigned long uid;       /* real user ID */
  unsigned long euid;      /* user's effective uid */
  unsigned long gid;       /* user's group ID */
  unsigned long pid;       /* process ID */
} pproc_t;

typedef struct file_rec {
  char *        pname;     /* object name */
  unsigned long perm;      /* perms associated with object */
  unsigned long ouid;      /* object owner ID */
  unsigned long ogid;      /* object group ID */
  unsigned long oid;       /* object ID */
  unsigned long devid;     /* device ID */
  unsigned long fsid;      /* file system ID */
} file_t;

characterize ari {

  void audit_record( audit_record_fields_t *p_fields{p_fields} );

} /* end of characterize ari */

characterize absari {
  // void  invalid();
  void  readop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  writeop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  creatop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  delop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  execop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  exitop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  chownop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  chmodop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file});
  void  renameop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file}, char *tname{nterm, tofile});
  void  linkop(int a{opcode}, long time{time}, pproc_t *proc{proc}, file_t *f{file}, char *tname{nterm, tofile});
}

