/*! \file linker_aout_machdep_i386.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker_aout_machdep_i386.c,v 1.5 2001/02/23 19:02:09 stevek Exp $
 *
 * \brief   System-dependant linker functions for a.out modules
 * \author  Steve Kiernan
 * \version $Revision: 1.5 $
 * \date    $Date: 2001/02/23 19:02:09 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_linker.h>

#include "linker_aout.h"


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

u_int32_t __LDPGSZ = 4096;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int
_linker_aout_machdep_supported( wss_linker_file_t *lf )
{
   linker_aout_data_t *aoutdata = (linker_aout_data_t *)lf->linker_file_data;

   if( N_GETMID(aoutdata->ex) != MID_I386 )
      return ENOEXEC;

   return 0;
}


/*  _linker_aout_machdep_relocate */
/*! \brief Relocate symbols within a.out segment
 * 
 *  \param lf      Pointer to linker file structure
 *  \param segment Pointer to segment data
 *  \param type    New type to assign to symbol
 *  \param r       Pointer to relocation data for symbol
 *
 *  \return 0 on success,
 *  \return EFAULT if a necessary opcode linkage is unimplemented or
 *                 a non-32bit relocation is encountered
 */
int
_linker_aout_machdep_relocate( wss_linker_file_t *lf, pd_byte_t *segment,
                               int type, AOut_Rel *r )
{
   linker_aout_data_t *aoutdata = (linker_aout_data_t *)lf->linker_file_data;
   u_int32_t           soffset  = (u_int32_t)r->r_address;
   u_int32_t           storaddr = (u_int32_t)((char *)segment + soffset);
   u_int32_t           snum     = r->r_symbolnum;
   u_int32_t           setvalue = 0;

#ifdef DEBUG_AOUT_LINKER
   wsi_printf( PD_MAKESTR("R[%d]: 0x%x [ pcrel:%d, length:%d, extern:%d, "
                           "baserel:%d, jmptable:%d, relative:%d, copy:%d ]\n"),
               r->r_symbolnum, r->r_address, r->r_pcrel, r->r_length,
               r->r_extern, r->r_baserel, r->r_jmptable, r->r_relative,
               r->r_copy );
#endif /* DEBUG_AOUT_LINKER */

   if( r->r_pcrel )
   {
      /*
       * PC-relative relocation
       */
      u_int32_t value = (u_int32_t)((aoutdata->symtab + snum)->n_value);
      (aoutdata->symtab + snum)->n_type &= ~N_TYPE;
      (aoutdata->symtab + snum)->n_type |= type;

      setvalue = (u_int32_t)( value - ( (u_int32_t)segment + soffset ) -
                              ( 1 << r->r_length ) );

      switch( r->r_length )
      {
         case 0:
            *((u_int8_t *)storaddr) = setvalue;
            break;

         case 1:
            *((u_int16_t *)storaddr) = setvalue;
            break;

         case 2:
            *((u_int32_t *)storaddr) = setvalue;
            break;
      }
   }
   else if( !r->r_pcrel && !r->r_baserel && !r->r_jmptable &&
            !r->r_relative && !r->r_copy )
   {
      /*
       * Absolute relocation
       */
      union reloc_value
      {
         u_int8_t v8;
         u_int16_t   v16;
         u_int32_t   v32;
      }
      rval;

      if( r->r_extern )
      {
         /*
          * Most likely a .bss value
          */
         switch( r->r_length )
         {
            case 2:
               rval.v32 = (u_int32_t)( (aoutdata->symtab + snum)->n_value ) +
                          *((u_int32_t *)storaddr);
               break;

            default:
               wsi_printf( PD_MAKESTR("Unable to handle non-32bit "
                                       "relocations (2^%d bytes requested)\n"),
                           r->r_length );
               return EFAULT;
         }
      }
      else
      {
         /*
          * .data or .text value
          */
         switch( r->r_length )
         {
            case 2:
               rval.v32 = (u_int32_t)( lf->linker_file_addr.addr32 +
                                       *((u_int32_t *)storaddr) );
               break;

            default:
               wsi_printf( PD_MAKESTR("Unable to handle non-32bit "
                                       "relocations (2^%d bytes requested)\n"),
                           r->r_length );
               return EFAULT;
         }
      }

      /*
       * Set relocated address
       */
      switch( r->r_length )
      {
         case 0:
            *((u_int8_t *)storaddr) = rval.v8;
            break;

         case 1:
            *((u_int16_t *)storaddr) = rval.v16;
            break;

         case 2:
            *((u_int32_t *)storaddr) = rval.v32;
            break;
      }
         
   }
   else
   {
      wsi_printf( PD_MAKESTR("Unimplemented opcode linkage: 0x%x + 0x%x "
                              "[%d]\n"),
                  (u_int)segment, (u_int)soffset, snum );
      return EFAULT;
   }

   return 0;
}

