/*! \file linker_aout.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker_aout.c,v 1.9 2001/02/23 19:02:08 stevek Exp $
 *
 * \brief   Support for loading a.out format modules
 * \author  Steve Kiernan
 * \version $Revision: 1.9 $
 * \date    $Date: 2001/02/23 19:02:08 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/types.h>

#include <wss_file.h>
#include <wss_malloc.h>
#include <wss_vm.h>

#include "linker_aout.h"


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#define AOUT_DATA_RELOC_PASS1  0x01
#define AOUT_DATA_RELOC_DONE   0x02


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

static int _linker_aout_load           __P(( const pd_char_t *,
                                             wss_linker_file_t * ));
static int _linker_aout_lookup_symbol  __P(( wss_linker_file_t *,
                                             const pd_char_t *,
                                             wss_linker_addr_t * ));
static int _linker_aout_export_symbols __P(( wss_linker_file_t * ));
static int _linker_aout_finalize       __P(( wss_linker_file_t * ));
static int _linker_aout_unload         __P(( wss_linker_file_t * ));


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static wss_linker_ops_t _linker_aout_ops =
{
   _linker_aout_load
};

static wss_linker_file_ops_t _linker_aout_file_ops =
{
   _linker_aout_lookup_symbol,
   _linker_aout_export_symbols,
   _linker_aout_finalize,
   _linker_aout_unload
};

static const pd_char_t *_linker_aout_description = PD_MAKESTR("a.out");


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * Lookup a single symbol in the binary's symbol table and return the
 * symbol's address
 */
static int
_linker_aout_lookup_symbol( wss_linker_file_t *lf, const pd_char_t *symbol,
                            wss_linker_addr_t *addr )
{
   int                  i;
   int                  entries;
   linker_aout_data_t  *aoutdata = (linker_aout_data_t *)lf->linker_file_data;
   char                *symname;

   if( ( addr == NULL ) || ( symbol == NULL ) )
      return EINVAL;

   /* If the symbol table has been exported, use it */
   if( lf->linker_file_symbols != NULL )
   {
      wss_symbol_t *syment;
      for( syment = lf->linker_file_symbols; syment && syment->name; syment++ )
      {
         if( !wsi_strcmp( syment->name, symbol ) )
         {
            addr->addr32 = syment->addr;
            return 0;
         }
      }
   }

   if( aoutdata == NULL )
      return EINVAL;

   symname = wss_malloc_string( strlen( symbol ) + 1 );
   if( !symname )
      return ENOMEM;

   wsi_sprintf( symname, PD_MAKESTR("_%s"), symbol );

   entries = aoutdata->ex->a_syms / sizeof(AOut_Sym);
   for( i = 0; i < entries; i++ )
   {
      AOut_Sym *s = aoutdata->symtab + i;

      /* Skip debugger symbols */
      if( IS_DEBUGGER_SYMBOL(s->n_type) )
         continue;

      if( !wsi_strcmp( aoutdata->strtab + s->n_strx, symname ) )
      {
         bzero( addr, sizeof(*addr) );
         addr->addr32 = (u_int32_t)s->n_value;
         wss_free( symname );
         return 0;
      }
   }

   wss_free( symname );
   return ENOENT;
}


static int
_linker_aout_resolve_data( wss_linker_file_t *lf, const pd_char_t *symbol,
                           wss_linker_addr_t *addr )
{
   linker_aout_data_t *aoutdata = (linker_aout_data_t *)lf->linker_file_data;
   pd_char_t         *symname;
   int                 entries;
   int                 i;

   if( aoutdata == NULL )
      return EINVAL;

   symname = wss_malloc_string( strlen( symbol ) + 1 );
   if( !symname )
      return ENOMEM;

   wsi_sprintf( symname, PD_MAKESTR("_%s"), symbol );

   entries = aoutdata->ex->a_syms / sizeof(AOut_Sym);
   for( i = 0; i < entries; i++ )
   {
      AOut_Sym *s = aoutdata->symtab + i;

      /* Skip debugger symbols */
      if( IS_DEBUGGER_SYMBOL(s->n_type) )
         continue;

      if( !wsi_strcmp( aoutdata->strtab + s->n_strx, symname ) )
      { 
         int relocations = aoutdata->ex->a_drsize / sizeof(AOut_Rel);
         for( i = 0; i < relocations; i++ )
         {
            AOut_Rel  *r   = aoutdata->reldata + i;
            u_int32_t  loc = s->n_value;

            wsi_printf( PD_MAKESTR("0x%x + 0x%x = 0x%x [0x%x]\n"),
                        (u_int32_t)aoutdata->data, (u_int32_t)r->r_address,
                        (u_int32_t)( aoutdata->data + r->r_address ), loc );

            if( (u_int32_t)( aoutdata->data + r->r_address ) == loc )
            {
               bzero( addr, sizeof(*addr) );
               addr->addr32 = lf->linker_file_addr.addr32 + *(u_int32_t *)loc;
               wsi_printf( PD_MAKESTR("0x%x + 0x%x = 0x%x\n"),
                           lf->linker_file_addr.addr32, *(u_int32_t *)loc,
                           addr->addr32 );
               wss_free( symname );
               return 0;
            }
         }
         break;
      }
   }

   wss_free( symname );
   return ENOENT;
}


/*
 * Resolve and relocate symbols used/contained in the binary
 */
static int
_linker_aout_relocate_symbols( wss_linker_file_t *lf )
{
   linker_aout_data_t *aoutdata    = (linker_aout_data_t *)lf->linker_file_data;
   int                 nrelocated  = 0;
   int                 nresolved   = 0;
   int                 nunresolved = 0;
   register int        entries     = aoutdata->ex->a_syms / sizeof(AOut_Sym);
   register int        i;

   if( ( aoutdata->flags & AOUT_DATA_RELOC_DONE ) != 0 )
      return 0;

   /* Resolve and compute relocated symbols */
   for( i = 0; i < entries; i++ )
   {
      AOut_Sym *s = aoutdata->symtab + i;

      if( IS_DEBUGGER_SYMBOL(s->n_type) ) continue;

      switch( SYMBOL_TYPE(s->n_type) )
      {
         case N_UNDF:
         {
            /* Only process after first reloc pass */
            if( ( aoutdata->flags & AOUT_DATA_RELOC_PASS1 ) != 0 )
            {
               wss_linker_addr_t  addr;
               const pd_char_t  *name = aoutdata->strtab + s->n_strx;

               if( name && *name != PD_MAKESTR('_') ) break;

               if( wss_linker_search( lf, name + 1, &addr ) == 0 )
               {
                   s->n_value = addr.addr32;
                   nresolved++;
               }
               else
               {
                  wsi_printf( PD_MAKESTR("a.out Linker: Unresolved symbol "
                                          "'%s'\n"), name );
                  nunresolved++;
               }
            }
            break;
         }

         case N_TEXT:
         case N_DATA:
         case N_BSS:
            /* Only process during first reloc pass */
            if( ( aoutdata->flags & AOUT_DATA_RELOC_PASS1 ) == 0 )
            {
               /*
                * .text, .data, and .bss relocations
                * need to be offset by relocation base
                * address
                */
               s->n_value += lf->linker_file_addr.addr32;
               nrelocated++;
            }
            break;

         default:
            break;
      }
   }

   if( ( aoutdata->flags & AOUT_DATA_RELOC_PASS1 ) == 0 )
      aoutdata->flags |= AOUT_DATA_RELOC_PASS1;
   else if( ( aoutdata->flags & AOUT_DATA_RELOC_DONE ) == 0 )
      aoutdata->flags |= AOUT_DATA_RELOC_DONE;

#ifndef WARN_UNRESOLVED
   if( nunresolved > 0 )
   {
      wsi_printf( PD_MAKESTR("a.out Linker: [%d unresolved symbol(s)] Unable")
                  PD_MAKESTR(" to load module.\n"), nunresolved );
      return ENOENT;
   }
#endif

#ifdef DEBUG_AOUT_LINKER
   if( nrelocated > 0 )
      wsi_printf( PD_MAKESTR("a.out Linker: Relocated %d symbol(s)\n"),
                  nrelocated );
   if( nresolved > 0 )
      wsi_printf( PD_MAKESTR("a.out Linker: Resolved %d symbol(s)\n"),
                  nresolved );
#endif /* DEBUG_AOUT_LINKER */

   return 0;
}


/*
 * Export library symbols for other modules to use
 */
static int
_linker_aout_export_symbols( wss_linker_file_t *lf )
{
   linker_aout_data_t *aoutdata = (linker_aout_data_t *)lf->linker_file_data;

   /* Just return if the symbol table has already been exported */
   if( lf->linker_file_symbols != NULL )
      return 0;

   /* Only export the symbol table if all relocations have been completed */
   if( ( aoutdata->flags & AOUT_DATA_RELOC_DONE ) != 0 )
   {
      int         entries       = aoutdata->ex->a_syms / sizeof(AOut_Sym);
      int         exported_syms = 0;
      int         i;

#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: %d candidates for exportable ")
                  PD_MAKESTR("symbols in table\n"),
                  entries );
#endif /* DEBUG_AOUT_LINKER */

      for( i = 1; i < entries; i++ )
      {
         AOut_Sym *s = aoutdata->symtab + i;

         /* Calculate size of exported symbol table */
         if( !IS_DEBUGGER_SYMBOL( s->n_type ) && ( s->n_type & N_EXT ) &&
             ( SYMBOL_TYPE(s->n_type) == N_TEXT ) &&
             ( *(char *)( aoutdata->strtab + s->n_strx ) == '_' ) )
            exported_syms++;
      }

      /* Allocate memory for symbol table */
      lf->linker_file_symbols =
         (wss_symbol_t *)wss_malloc( ( exported_syms + 1 ) *
                                     sizeof(wss_symbol_t) );
      bzero( lf->linker_file_symbols,
             ( exported_syms + 1 ) * sizeof(wss_symbol_t) );

      if( !exported_syms )
      {
         wsi_printf( PD_MAKESTR("a.out Linker: Warning -- No symbols ")
                     PD_MAKESTR("exported by library '%s'\n"),
                     lf->linker_file_name );
      }
      else
      {
         /* Construct symbol table entries for exported functions */
         for( i = 1; i < entries; i++ )
         {
            AOut_Sym *s = aoutdata->symtab + i;

            if( !IS_DEBUGGER_SYMBOL( s->n_type ) && ( s->n_type & N_EXT ) &&
                ( SYMBOL_TYPE(s->n_type) == N_TEXT ) &&
                ( *(char *)( aoutdata->strtab + s->n_strx ) == '_' ) )
            {
               wss_symbol_t *exsym = lf->linker_file_symbols + --exported_syms;
               pd_char_t   *symname =
                  (pd_char_t *)( (pd_byte_t *)aoutdata->strtab + s->n_strx );

               exsym->name = wss_strdup( symname );
               exsym->addr = (u_long)s->n_value;
               exsym->count = 0;

               if( !exported_syms )
                  break;
            }
         }
      }
   }
   else
      return EINVAL;

   return 0;
}


static int
_linker_aout_relocate_segment( wss_linker_file_t *lf,
                               linker_aout_reloc_t segtype )
{
   int                 relocations;
   int                 i;
   int                 type;
   AOut_Rel           *rel;
   pd_byte_t         *segment;
   linker_aout_data_t *aoutdata = (linker_aout_data_t *)lf->linker_file_data;
#ifdef DEBUG_AOUT_LINKER
   const pd_char_t   *sname;
#endif /* DEBUG_AOUT_LINKER */
   

   switch( segtype )
   {
      case AOUT_RELOC_TEXT:
         rel = aoutdata->reltext;
         segment = aoutdata->text;
         relocations = aoutdata->ex->a_trsize / sizeof(AOut_Rel);
         type = N_TEXT;
#ifdef DEBUG_AOUT_LINKER
         sname = PD_MAKESTR(".text");
#endif /* DEBUG_AOUT_LINKER */
         break;

      case AOUT_RELOC_DATA:
         rel = aoutdata->reldata;
         segment = aoutdata->data;
         relocations = aoutdata->ex->a_drsize / sizeof(AOut_Rel);
         type = N_DATA;
#ifdef DEBUG_AOUT_LINKER
         sname = PD_MAKESTR(".data");
#endif /* DEBUG_AOUT_LINKER */
         break;

      default:
         return EINVAL;
   }

   if( ( rel == NULL ) || ( segment == NULL ) )
      return EINVAL;

#ifdef DEBUG_AOUT_LINKER
   wsi_printf( PD_MAKESTR("a.out Linker: Relocating %d references in %s\n"),
               relocations, sname );
#endif /* DEBUG_AOUT_LINKER */
   for( i = 0; i < relocations; i++ )
   {
      int       error;
      AOut_Rel *r = rel + i;
      AOut_Sym *s = aoutdata->symtab + r->r_symbolnum;

      if( ( ( ( aoutdata->flags & AOUT_DATA_RELOC_DONE ) == 0 ) &&
            ( SYMBOL_TYPE(s->n_type) == N_DATA ) &&
            ( ( s->n_type & N_EXT ) == 0 ) ) ||
          ( ( ( aoutdata->flags & AOUT_DATA_RELOC_DONE ) != 0 ) &&
            ( ( SYMBOL_TYPE(s->n_type) != N_DATA ) ||
              ( ( s->n_type & N_EXT ) != 0 ) ) ) )
      {
         error = _linker_aout_machdep_relocate( lf, segment, type, rel + i );
         if( error )
            return error;
      }
   }

   return 0;
}


/*
 * Process .text and .data section relocations
 */
static int
_linker_aout_relocate_module( wss_linker_file_t *lf )
{
   linker_aout_data_t *aoutdata = (linker_aout_data_t *)lf->linker_file_data;
   int                 error;

   if( aoutdata->reltext )
   {
      error = _linker_aout_relocate_segment( lf, AOUT_RELOC_TEXT );
      if( error )
         return error;
   }
   if( aoutdata->reldata )
   {
      error = _linker_aout_relocate_segment( lf, AOUT_RELOC_DATA );
      if( error )
         return error;
   }

   return 0;
}


/*  _linker_aout_load */
/*! \brief Load an a.out format module
 *
 *  Resolves and relocates symbols within the .text, .data, and .bss
 *  sections.
 *
 *  The module does not depend on the file image in the filesystem once it is
 *  finalized.
 *
 *  \param filename Pointer to string containing the name of the file to load
 *  \param result   Pointer to linker file structure to fill
 *
 *  \return 0 if successfully loaded,
 *  \return ENOENT if unable to resolve symbols,
 *  \return EEXIST if the module is already loaded,
 *  \return EINVAL if the module is of an unknown type
 */                                                               
int EXPORT
_linker_aout_load( const pd_char_t *filename, wss_linker_file_t *result )
{
   wss_file_t          file;
   int                 size;
   pd_byte_t         *bin_image;
   pd_byte_t         *incore_image;
   AOut_Exec          *ex;
   linker_aout_data_t *aoutdata;
   wss_linker_addr_t   addr;
   size_t              memsize;
   u_int32_t           offset;
   int                 entries;
   int                 error   = 0;
   register int        i;

   /* Offsets */
   u_int32_t           txtoff;
   u_int32_t           datoff;
   u_int32_t           symtab;
   u_int32_t           strtab;
   u_int32_t           reltext;
   u_int32_t           reldata;

   /* Make sure result pointer is non-NULL */
   if( result == NULL )
   {
#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: result argument is NULL\n") );
#endif /* DEBUG_AOUT_LINKER */
      return EINVAL;
   }

#ifdef DEBUG_AOUT_LINKER
   wsi_printf( PD_MAKESTR("a.out Linker: Processing '%s'...\n"), filename );
#endif /* DEBUG_AOUT_LINKER */

   result->linker_file_ops = &_linker_aout_file_ops;

   file = wss_file_open( filename, WSS_FILE_READ | WSS_FILE_ROOTPRIV );
   if( file == WSS_FILE_INVALID )
      return ENOENT;

   size = wss_file_get_size( file );

   if( size < sizeof(AOut_Exec) )
   {
      wss_file_close( file );
      wss_file_rele( file );
      return ENOEXEC;
   }

   bin_image = wss_file_read_alloc( file, &size );
   wss_file_close( file );

   if( bin_image == NULL )
      return ENOMEM;

   result->linker_file_image.addr32 = (u_int32_t)bin_image;
   result->linker_file_size = size;

   ex = (AOut_Exec *)bin_image;

   /* Check for magic number */
   if( N_BADMAG( ex ) )
   {
#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: '%s' is not a.out\n"), filename );
#endif /* DEBUG_AOUT_LINKER */
      wss_free( bin_image );
      return ENOEXEC;
   }

   aoutdata = (linker_aout_data_t *)wss_malloc( sizeof(linker_aout_data_t) );
   if( aoutdata == NULL )
   {
      wss_free( bin_image );
      return ENOMEM;
   }

   bzero( aoutdata, sizeof(*aoutdata) );
   result->linker_file_data = (void *)aoutdata;
   aoutdata->ex = ex;

   if( _linker_aout_machdep_supported( result ) )
   {
#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: a.out module not compiled for ")
                  PD_MAKESTR("this processor architecture\n") );
#endif /* DEBUG_AOUT_LINKER */
      wss_free( bin_image );
      wss_free( aoutdata );
      return ENOEXEC;
   }

   txtoff  = (u_int32_t)N_TXTOFF(ex);
   datoff  = (u_int32_t)N_DATOFF(ex);
   symtab  = (u_int32_t)N_SYMOFF(ex);
   strtab  = (u_int32_t)N_STROFF(ex);
   reltext = (u_int32_t)N_RELOFF(ex);
   reldata = (u_int32_t)N_ALIGN(ex, N_RELOFF(ex) + ex->a_trsize);

   aoutdata->symtab = (AOut_Sym *)( bin_image + symtab );

   /* Relocate .bss entries */
   offset = N_BSSOFF(ex);
   entries = ex->a_syms / sizeof(AOut_Sym);
   for( i = 0; i < entries; i++ )
   {
      AOut_Sym *s = (AOut_Sym *)aoutdata->symtab + i;

      if( IS_DEBUGGER_SYMBOL(s->n_type) ) continue;

      if( ( SYMBOL_TYPE(s->n_type) == N_UNDF ) &&
          ( s->n_value != 0x0 ) )
      {
         u_int32_t size = s->n_value;

         s->n_value = offset - sizeof(AOut_Exec);
         offset += size;
         s->n_type &= ~N_TYPE;
         s->n_type |= N_BSS;
      }
   }
   ex->a_bss = offset - N_BSSOFF(ex);

   memsize = ex->a_text + ex->a_data + ex->a_bss;

   /* Allocate a block of memory to hold the in-image module */
   incore_image = (pd_byte_t *)wss_malloc( ex->a_text + ex->a_data +
                                            ex->a_bss );
   if( incore_image == NULL )
   {
      wss_free( bin_image );
      wss_free( aoutdata );
      return ENOMEM;
   }

   bzero( incore_image, memsize );
#ifdef DEBUG_AOUT_LINKER
   wsi_printf( PD_MAKESTR("a.out Linker: Allocated in-core image @ 0x%x\n"),
               (u_int)incore_image );
#endif /* DEBUG_AOUT_LINKER */
   result->linker_file_addr.addr32 = (u_int32_t)incore_image;

   /* Copy text and data segments and zero bss */
   bcopy( (pd_byte_t *)bin_image + txtoff, incore_image,
          ex->a_text + ex->a_data );
   if( ex->a_bss )
      bzero( (pd_byte_t *)incore_image + ex->a_text + ex->a_data, ex->a_bss );

   aoutdata->strtab = (pd_char_t *)( bin_image + strtab );
   aoutdata->text = incore_image + txtoff - sizeof(AOut_Exec);
   aoutdata->reltext = (AOut_Rel *)( bin_image + reltext );
   aoutdata->data = incore_image + datoff - sizeof(AOut_Exec);
   aoutdata->reldata = (AOut_Rel *)( bin_image + reldata );

#ifdef DEBUG_AOUT_LINKER
   wsi_printf( PD_MAKESTR(" strtab = 0x%x\n"), (u_int32_t)aoutdata->strtab );
   wsi_printf( PD_MAKESTR("   text = 0x%x\n"), (u_int32_t)aoutdata->text );
   wsi_printf( PD_MAKESTR("  [size = 0x%x]\n"), (u_int32_t)aoutdata->ex->a_text );
   wsi_printf( PD_MAKESTR("reltext = 0x%x\n"), (u_int32_t)aoutdata->reltext );
   wsi_printf( PD_MAKESTR("   data = 0x%x\n"), (u_int32_t)aoutdata->data );
   wsi_printf( PD_MAKESTR("  [size = 0x%x]\n"), (u_int32_t)aoutdata->ex->a_data );
   wsi_printf( PD_MAKESTR("reldata = 0x%x\n"), (u_int32_t)aoutdata->reldata );
#endif /* DEBUG_AOUT_LINKER */

   /* Relocate module symbols (ignore commons) */
   if( ( error = _linker_aout_relocate_symbols( result ) ) )
   {
#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: Unable to relocate symbols\n") );
#endif /* DEBUG_AOUT_LINKER */
      wss_free( bin_image );
      wss_free( aoutdata );
      return error;
   }

   /* Resolve name and dependencies */
   error = _linker_aout_resolve_data( result, WRLINKNAME_MOD_DEPENDENCIES,
                                      &addr );
   if( error )
   {
#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: Unable to determine symbol ")
                  PD_MAKESTR("'%s'\n"), WRLINKNAME_MOD_DEPENDENCIES );
#endif /* DEBUG_AOUT_LINKER */
      wss_free( bin_image );
      wss_free( aoutdata );
      return error;
   }

   result->linker_file_dependencies = (const pd_char_t *)addr.addr;

   /* Relocate module symbols (resolving undefined symbols) */
   if( ( error = _linker_aout_relocate_symbols( result ) ) )
   {
#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: Unable to relocate symbols\n") );
#endif /* DEBUG_AOUT_LINKER */
      wss_free( bin_image );
      wss_free( aoutdata );
      return error;
   }

   /* Relocate module relative to relocation base address */
   if( ( error = _linker_aout_relocate_module( result ) ) )
   {
#ifdef DEBUG_AOUT_LINKER
      wsi_printf( PD_MAKESTR("a.out Linker: Unable to relocate a.out ")
                  PD_MAKESTR("module\n") );
#endif /* DEBUG_AOUT_LINKER */
      wss_free( bin_image );
      wss_free( aoutdata );
      return error;
   }

   return error;
}

static int
_linker_aout_finalize( wss_linker_file_t *lf )
{
   if( lf->linker_file_data != NULL )
   {
      wss_free( lf->linker_file_data );
      lf->linker_file_data = NULL;
   }
   if( lf->linker_file_image.addr != NULL )
   {
      wss_free( lf->linker_file_image.addr );
      bzero( &lf->linker_file_image, sizeof(wss_linker_addr_t) );
   }

   return 0;
}

static int
_linker_aout_unload( wss_linker_file_t *lf )
{
   if( lf == NULL )
      return EINVAL;

   return wss_linker_file_destroy( lf );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

void
linker_aout_init()
{
   wss_linker_register( _linker_aout_description, &_linker_aout_ops );
}

