/*! \file linker_crit.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker_crit.c,v 1.38.2.1 2001/07/11 19:31:33 stevek Exp $
 *
 * \brief   Support for loading criteria modules
 * \author  Steve Kiernan
 * \version $Revision: 1.38.2.1 $
 * \date    $Date: 2001/07/11 19:31:33 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wsi_types.h>

#include <wss_file.h>
#include <wss_malloc.h>
#include <wss_linker.h>
#include <wss_module.h>

#include "linker_crit.h"

#ifdef USER_MODE_TEST
#  ifdef WR_FREEBSD
#    include <sys/param.h>
#    include <sys/proc.h>
#    include <sys/tty.h>
#    define dev2udev(a) ((udev_t)0)
#    define umajor(a) major(a)
#    define uminor(a) minor(a)
#  endif
#  ifdef WR_LINUX
/* Bogus to get this to compile. */
struct task_struct {
    pid_t       pid;
    struct task_struct *p_pptr;
    uid_t uid, euid, gid, egid;
};
#  endif
#  ifdef WR_SOLARIS
#    include <values.h>
#    define INT_MAX MAXINT
#  endif
#else  /* USER_MODE_TEST */
#  ifdef WR_FREEBSD
#    include <machine/limits.h>
#    include <sys/conf.h>
#    include <sys/systm.h>
#    include <sys/tty.h>
#  endif
#  ifdef WR_LINUX
#    include <linux/sched.h>
#  endif /* WR_LINUX */
#endif /* USER_MODE_TEST */

#ifdef WR_WINNT
#  include <limits.h>
#endif


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#define BUILTINS_SIZE ( sizeof(_linker_crit_builtins) / sizeof(pd_char_t *) )
#define OPS_SIZE      ( sizeof(_linker_crit_op_names) / sizeof(pd_char_t *) )
#define POSIX_SIZE    ( sizeof(_linker_crit_posix_charclass) / \
                        sizeof(pd_char_t *) )

#if   defined(WR_FREEBSD)
#define BIVAL_PID  wrp->wrp_p->p_pid
#define BIVAL_PPID wrp->wrp_p->p_oppid
#define BIVAL_UID  wrp->wrp_p->p_cred->p_ruid
#define BIVAL_EUID wrp->wrp_p->p_cred->p_svuid
#define BIVAL_GID  wrp->wrp_p->p_cred->p_rgid
#define BIVAL_EGID wrp->wrp_p->p_cred->p_svgid
#ifdef USER_MODE_TEST
#define BIVAL_TTY  0
#else
#define BIVAL_TTY  ((wrp->wrp_p->p_pgrp && \
                     wrp->wrp_p->p_pgrp->pg_session->s_ttyp)? \
                    wrp->wrp_p->p_pgrp->pg_session->s_ttyp->t_dev->si_udev:0)
#endif
#elif defined(WR_SOLARIS)
#define BIVAL_PID  wrp->wrp_p->p_pidp->pid_id
#define BIVAL_PPID wrp->wrp_p->p_ppid
#define BIVAL_UID  wrp->wrp_p->p_cred->cr_ruid
#define BIVAL_EUID wrp->wrp_p->p_cred->cr_uid
#define BIVAL_GID  wrp->wrp_p->p_cred->cr_rgid
#define BIVAL_EGID wrp->wrp_p->p_cred->cr_gid
#define BIVAL_TTY  0
#elif defined(WR_LINUX)
#define BIVAL_PID  wrp->wrp_p->pid
#define BIVAL_PPID wrp->wrp_p->p_pptr->pid
#define BIVAL_UID  wrp->wrp_p->uid
#define BIVAL_EUID wrp->wrp_p->euid
#define BIVAL_GID  wrp->wrp_p->gid
#define BIVAL_EGID wrp->wrp_p->egid
#ifdef USER_MODE_TEST
#define BIVAL_TTY  0
#else
#define BIVAL_TTY  (wrp->wrp_p->tty?kdev_t_to_nr(wrp->wrp_p->tty->device):0)
#endif
#endif


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

static int _linker_crit_load                    __P(( const pd_char_t *,
                                                      wss_linker_file_t * ));
static int _linker_crit_lookup_symbol           __P(( wss_linker_file_t *,
                                                      const pd_char_t *,
                                                      wss_linker_addr_t * ));
static int _linker_crit_export_symbols          __P(( wss_linker_file_t * ));
static int _linker_crit_finalize                __P(( wss_linker_file_t * ));
static int _linker_crit_unload                  __P(( wss_linker_file_t * ));


static linker_crit_compat_t
   _linker_crit_compat_arg           __P(( wss_linker_file_t *, crit_id_t ));
static void _linker_crit_entrypoint  __P(( wss_proc_t *, wss_crit_t * ));
static pd_char_t *
   _linker_crit_expression_to_string __P(( wss_linker_file_t *,
                                           crit_expression_t * ));
static pd_char_t *
   _linker_crit_literal_to_string    __P(( wss_linker_file_t *,
                                           crit_literal_t * ));
static pd_char_t *
   _linker_crit_regex_to_string      __P(( wss_linker_file_t *,
                                           crit_regex_t * ));
static pd_char_t *
   _linker_crit_regex_elem_to_string __P(( wss_linker_file_t *,
                                           crit_regex_t *,
                                           crit_regex_node_t * ));
static crit_expression_t *
   _linker_crit_lookup_expression    __P(( wss_linker_file_t *, crit_id_t ));
static crit_literal_t *
   _linker_crit_lookup_literal       __P(( wss_linker_file_t *, crit_id_t ));
static crit_regex_t *
   _linker_crit_lookup_regex         __P(( wss_linker_file_t *, crit_id_t ));
static crit_regex_node_t *
   _linker_crit_lookup_regex_elem    __P(( wss_linker_file_t *,
                                           crit_regex_t *, crit_id_t ));
static pd_bool_t
   _linker_crit_run_expression       __P(( wss_linker_file_t *,
                                           wss_proc_t *, crit_expression_t * ));
static pd_bool_t
   _linker_crit_run_regex            __P(( wss_linker_file_t *lf,
                                           pd_char_t *, crit_regex_t *re ));
static int *
_linker_crit_match_regex_elem        __P(( wss_linker_file_t *lf,
                                           pd_char_t *str, crit_regex_t *re,
                                           crit_regex_node_t *elem,
                                           unsigned int index ));


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static wss_linker_ops_t _linker_crit_ops =
{
   _linker_crit_load
};

static wss_linker_file_ops_t _linker_crit_file_ops =
{
   _linker_crit_lookup_symbol,
   _linker_crit_export_symbols,
   _linker_crit_finalize,
   _linker_crit_unload
};

static const pd_char_t *_linker_crit_description = PD_MAKESTR("Criteria");


static const pd_char_t *_linker_crit_act_name   = PD_MAKESTR("activation");
static const pd_char_t *_linker_crit_deact_name = PD_MAKESTR("deactivation");

static const pd_char_t *_linker_crit_builtins[] =
{
   PD_MAKESTR("prog"),
   PD_MAKESTR("pprog"),
   PD_MAKESTR("pid"),
   PD_MAKESTR("ppid"),
   PD_MAKESTR("cwd"),
   PD_MAKESTR("user"),
   PD_MAKESTR("uid"),
   PD_MAKESTR("euid"),
   PD_MAKESTR("gid"),
   PD_MAKESTR("egid"),
   PD_MAKESTR("tty")
};

static const linker_crit_compat_t _linker_crit_builtins_compat[] =
{
   LCC_STRING,     /* prog */
   LCC_STRING,     /* pprog */
#ifdef WR_UNIX
   LCC_INT,        /* pid */
   LCC_INT,        /* ppid */
#else
   LCC_UNKNOWN,    /* pid */
   LCC_UNKNOWN,    /* ppid */
#endif
   LCC_STRING,     /* cwd */
#ifdef WR_UNIX
   LCC_UNKNOWN,    /* user */
   LCC_INT,        /* uid */
   LCC_INT,        /* euid */
   LCC_INT,        /* gid */
   LCC_INT,        /* egid */
   LCC_INT         /* tty */
#else
   LCC_STRING,     /* user */
   LCC_UNKNOWN,    /* uid */
   LCC_UNKNOWN,    /* euid */
   LCC_UNKNOWN,    /* gid */
   LCC_UNKNOWN,    /* egid */
   LCC_UNKNOWN     /* tty */
#endif
};

static const pd_char_t *_linker_crit_op_names[] =
{
   PD_MAKESTR(""),
   PD_MAKESTR("&&"),
   PD_MAKESTR("||"),
   PD_MAKESTR("!"),
   PD_MAKESTR("<"),
   PD_MAKESTR(">"),
   PD_MAKESTR("<="),
   PD_MAKESTR(">="),
   PD_MAKESTR("!="),
   PD_MAKESTR("=="),
   PD_MAKESTR("=~"),
   PD_MAKESTR("suid("),
   PD_MAKESTR("sgid("),
   PD_MAKESTR("==>")
};

static const pd_char_t *_linker_crit_posix_charclass[] =
{
   PD_MAKESTR("alnum"),
   PD_MAKESTR("alpha"),
   PD_MAKESTR("blank"),
   PD_MAKESTR("cntrl"),
   PD_MAKESTR("digit"),
   PD_MAKESTR("graph"),
   PD_MAKESTR("lower"),
   PD_MAKESTR("print"),
   PD_MAKESTR("punct"),
   PD_MAKESTR("space"),
   PD_MAKESTR("upper"),
   PD_MAKESTR("xdigit")
};


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * Activate/deactivate entrypoint to execute the criteria expression
 */
static void
_linker_crit_entrypoint( wss_proc_t *wrp, wss_crit_t *crit )
{
   wss_linker_file_t  *lf       = crit->crit_module->wrm_linked;
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;
   crit_expression_t  *exp;

   exp = _linker_crit_lookup_expression( lf, critdata->hdr->start );
   (void) _linker_crit_run_expression( lf, wrp, exp );
}

static pd_char_t *
_linker_crit_builtin_string( wss_linker_file_t *lf, wss_proc_t *wrp,
                             crit_id_t arg )
{
   switch( arg )
   {
      case CTOK_PROG:
         return wrp->wrp_path;

      case CTOK_PPROG:
         return wrp->wrp_oldpath;

      case CTOK_CWD:
         return wrp->wrp_cwd;

#ifdef WR_WINNT
      case CTOK_USER:
         return wrp->wrp_p->p_user_name;
#endif /* WR_WINNT */

      default:
         break;
   }

   return NULL;
}

static int
_linker_crit_builtin_int( wss_linker_file_t *lf, wss_proc_t *wrp,
                          crit_id_t arg )
{
#ifdef WR_UNIX
   switch( arg )
   {
      case CTOK_PID:
         return BIVAL_PID;

      case CTOK_PPID:
         return BIVAL_PPID;

      case CTOK_UID:
         return BIVAL_UID;

      case CTOK_EUID:
         return BIVAL_EUID;

      case CTOK_GID:
         return BIVAL_GID;

      case CTOK_EGID:
         return BIVAL_EGID;

      case CTOK_TTY:
         return BIVAL_TTY;

      default:
         break;
   }
#endif /* WR_UNIX */

   return 0;
}

static pd_bool_t
_linker_crit_run_expression( wss_linker_file_t *lf, wss_proc_t *wrp,
                             crit_expression_t *exp )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;

#ifdef DEBUG_CRIT_LINKER
   wsi_printf( PD_MAKESTR("Criteria Linker: exp[0x%x]( 0x%x, 0x%x )\n"),
               exp->op, exp->arg1, exp->arg2 );
#endif /* DEBUG_CRIT_LINKER */

   switch( exp->op )
   {
      case CO_AND:
      case CO_OR:
      {
         crit_expression_t *left, *right;
         pd_bool_t         bleft, bright;
         pd_bool_t         result;

         left = _linker_crit_lookup_expression( lf, exp->arg1 );
         right = _linker_crit_lookup_expression( lf, exp->arg2 );

         bleft = _linker_crit_run_expression( lf, wrp, left );
         bright = _linker_crit_run_expression( lf, wrp, left );

         if( exp->op == CO_AND )
            result = ( bleft && bright ) ? TRUE : FALSE;
         else
            result = ( bleft || bright ) ? TRUE : FALSE;

         return result;
      }

      case CO_NOT:
      {
         crit_expression_t *left;
         pd_bool_t result;
         
         left = _linker_crit_lookup_expression( lf, exp->arg1 );
         result =
            _linker_crit_run_expression( lf, wrp, left ) ? FALSE : TRUE;

         return result;
      }

      case CO_LT:
      case CO_GT:
      case CO_LTE:
      case CO_GTE:
      case CO_NE:
      case CO_EQ:
      {
         crit_literal_t       *left;
         crit_literal_t       *right;
         linker_crit_compat_t  type;
         pd_bool_t            result = FALSE;
         pd_char_t           *lstr   = NULL;
         pd_char_t           *rstr   = NULL;
         int                   lint   = 0;
         int                   rint   = 0;

         type = _linker_crit_compat_arg( lf, exp->arg1 );

         /* Resolve the left side */
         if( exp->arg1 & CRIT_BUILTIN_MASK )
         {
            if( type == LCC_STRING )
               lstr = _linker_crit_builtin_string( lf, wrp, exp->arg1 );
            else
               lint = _linker_crit_builtin_int( lf, wrp, exp->arg1 );
         }
         else
         {
            left = _linker_crit_lookup_literal( lf, exp->arg1 );

            if( type == LCC_STRING )
               lstr = _linker_crit_literal_to_string( lf, left );
            else if( type == LCC_INT )
               lint = left->ival;
         }

         /* Resolve the right side */
         if( exp->arg2 & CRIT_BUILTIN_MASK )
         {
            if( type == LCC_STRING )
               rstr = _linker_crit_builtin_string( lf, wrp, exp->arg2 );
            else
               rint = _linker_crit_builtin_int( lf, wrp, exp->arg2 );
         }
         else
         {
            right = _linker_crit_lookup_literal( lf, exp->arg2 );

            if( type == LCC_STRING )
               rstr = _linker_crit_literal_to_string( lf, right );
            else if( type == LCC_INT )
               rint = right->ival;
         }

         /* Compare the strings */
         if( type == LCC_STRING )
            lint = wsi_strcmp( lstr, rstr );

         switch( exp->op )
         {
            case CO_LT:
               result = ( lint < rint ) ? TRUE : FALSE;
               break;
            case CO_GT:
               result = ( lint > rint ) ? TRUE : FALSE;
               break;
            case CO_LTE:
               result = ( lint <= rint ) ? TRUE : FALSE;
               break;
            case CO_GTE:
               result = ( lint >= rint ) ? TRUE : FALSE;
               break;
            case CO_NE:
               result = ( lint != rint ) ? TRUE : FALSE;
               break;
            case CO_EQ:
               result = ( lint == rint ) ? TRUE : FALSE;
               break;
            default:
               break;
         }

         if( ( ( exp->arg1 & CRIT_BUILTIN_MASK ) == 0 ) && lstr )
            wss_free( lstr );
         if( ( ( exp->arg2 & CRIT_BUILTIN_MASK ) == 0 ) && rstr )
            wss_free( rstr );

         return result;
      }

      case CO_RE:
      {
         linker_crit_compat_t  type;
         crit_regex_t         *re;
         pd_char_t           *lstr = NULL;
         pd_bool_t            retc;
         
         type = _linker_crit_compat_arg( lf, exp->arg1 );

         /* Resolve the left side */
         if( exp->arg1 & CRIT_BUILTIN_MASK )
         {
            if( type == LCC_STRING )
               lstr = _linker_crit_builtin_string( lf, wrp, exp->arg1 );
            else
               return FALSE;
         }
         else
         {
            crit_literal_t *left = _linker_crit_lookup_literal( lf, exp->arg1 );

            if( type == LCC_STRING )
               lstr = _linker_crit_literal_to_string( lf, left );
            else
               return FALSE;
         }

         re = _linker_crit_lookup_regex( lf, exp->arg2 );

         retc = _linker_crit_run_regex( lf, lstr, re );

         return retc;
      }

      case CO_SUID:
      case CO_SGID:
         break;

      case CO_ARROW:
      {
         crit_expression_t *left;
         pd_bool_t result;
         
         left = _linker_crit_lookup_expression( lf, exp->arg1 );
         result = _linker_crit_run_expression( lf, wrp, left );

         if( result )
         {
            if( critdata->hdr->activation )
            {
#ifdef DEBUG_CRIT_LINKER
               wsi_printf( PD_MAKESTR("Criteria Linker: Adding wrapper '%s'\n"),
                           critdata->implist[0] );
#endif /* DEBUG_CRIT_LINKER */
               wss_proc_add_wrapper_blind( wrp, critdata->implist[0] );
            }
            else
            {
#ifdef DEBUG_CRIT_LINKER
               wsi_printf( PD_MAKESTR("Criteria Linker: Removing wrapper '%s'\n"),
                           critdata->implist[0] );
#endif /* DEBUG_CRIT_LINKER */
               wss_proc_del_wrapper( wrp, critdata->implist[0] );
            }
         }
         break;
      }

      default:
         break;
   }

   return FALSE;
}

static pd_bool_t
_linker_crit_run_regex( wss_linker_file_t *lf, pd_char_t *str,
                        crit_regex_t *re )
{
   crit_regex_node_t *elem;
   pd_bool_t         rbool = FALSE;

#ifdef DEBUG_CRIT_LINKER
   wsi_printf( PD_MAKESTR("Criteria Linker: regex 0x%x, string \"%s\"\n"),
               (unsigned int)re, str );
#endif /* DEBUG_CRIT_LINKER */

   elem = _linker_crit_lookup_regex_elem( lf, re, re->regex );
   if( elem && str )
   {
      unsigned int i;
      
      for( i = 0; i < wsi_strlen(str); i++ )
      {
         int *ivec = _linker_crit_match_regex_elem( lf, str, re, elem, i );
         if( ivec )
         {
#ifdef DEBUG_CRIT_LINKER
            wsi_printf( PD_MAKESTR("Criteria Linker: regex 0x%x matched\n"),
                        (unsigned int)re );
#endif /* DEBUG_CRIT_LINKER */

            rbool = TRUE;
            wss_free( ivec );
            break;
         }
      }
   }

   return rbool;
}

static int *
_linker_crit_match_regex_elem( wss_linker_file_t *lf, pd_char_t *str,
                               crit_regex_t *re, crit_regex_node_t *elem,
                               unsigned int index )
{
   int *ivec = NULL;

   if( elem == NULL )
   {
      ivec = wss_malloc( sizeof(int) * 2 );
      ivec[0] = 1;
      ivec[1] = index;
      return ivec;
   }
   else if( index > wsi_strlen(str) )
      return NULL;

   switch( elem->op )
   {
      case RE_EXP:
      {
         crit_regex_node_t *exelem =
            _linker_crit_lookup_regex_elem( lf, re, elem->data );

         if( exelem )
            ivec = _linker_crit_match_regex_elem( lf, str, re, exelem, index );

         break;
      }

      case RE_CHAR:
      {
         crit_literal_t *lit = _linker_crit_lookup_literal( lf, elem->data );

         linker_crit_compat_t type = _linker_crit_compat_arg( lf, elem->data );
         if( lit && ( type == LCC_STRING ) )
         {
            pd_char_t *match = _linker_crit_literal_to_string( lf, lit );

            if( match )
            {
               register int i;
               register int matchlen = wsi_strlen( match );

               for( i = 0; i < matchlen; i++ )
               {
                  pd_char_t ch = str[index+i];
                  if( ( ch == PD_MAKESTR('\0') ) ||
                      ( ( (elem->data & CRIT_RE_INSENS)?wsi_tolower(ch):ch ) !=
                        match[i] ) )
                     break;
               }

               wss_free( match );

               if( i == matchlen )
               {
                  crit_regex_node_t *nextelem =
                     _linker_crit_lookup_regex_elem( lf, re, elem->next );

                  ivec = _linker_crit_match_regex_elem( lf, str, re,
                                                        nextelem,
                                                        index + matchlen );

                  if( ( elem->next & CRIT_RE_UNCLE ) && ivec )
                  {
                     wss_free( ivec );

                     ivec = wss_malloc( sizeof(int) * 2 );
                     ivec[0] = 1;
                     ivec[1] = index + matchlen;
                  }
               }
            }
         }
         break;
      }

      case RE_ANY:
      {
         pd_char_t ch = str[index];
         if( ( !(elem->data & 0x02) && ( ch == PD_MAKESTR('\n') ) ) ||
             ( (elem->data & 0x01) && ( ch == PD_MAKESTR('\0') ) ) )
            break;
         else
         {
            crit_regex_node_t *nextelem =
               _linker_crit_lookup_regex_elem( lf, re, elem->next );

            ivec = _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                                  index + 1 );
            if( ( elem->next & CRIT_RE_UNCLE ) && ivec )
            {
               wss_free( ivec );

               ivec = wss_malloc( sizeof(int) * 2 );
               ivec[0] = 1;
               ivec[1] = index + 1;
            }
         }

         break;
      }

      case RE_END:
      {
         pd_char_t         ch;
         crit_regex_node_t *nextelem;
         
         ch = str[index];
         nextelem = _linker_crit_lookup_regex_elem( lf, re, elem->next );

         if( ( ch == PD_MAKESTR('\0') ) ||
             ( elem->data && ( ch == PD_MAKESTR('\n') ) ) )
         {
            ivec = _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                                  index );
            if( ( elem->next & CRIT_RE_UNCLE ) && ivec )
            {
               wss_free( ivec );

               ivec = wss_malloc( sizeof(int) * 2 );
               ivec[0] = 1;
               ivec[1] = index;
            }
         }
         break;
      }

      case RE_ONEOF:
      {
         int               *possibles;
         int               *doables;
         crit_regex_node_t *tk;
         crit_regex_node_t *nextelem;
         int                i;
         
         possibles = wss_malloc( sizeof(int) * ( CRIT_ARG_MASK + 1 ) );
         possibles[0] = 0;

         for( tk = _linker_crit_lookup_regex_elem( lf, re, elem->oneofnext );
              tk;
              tk = _linker_crit_lookup_regex_elem( lf, re, tk->oneofnext ) )
         {
            int *newivec =
               _linker_crit_match_regex_elem( lf, str, re, tk, index );
            if( newivec )
            {
               if( elem->data & CRIT_RE_NEGATE )
               {
                  wss_free( newivec );
                  wss_free( possibles );
                  break;
               }

               for( i = 1; i <= newivec[0]; i++ )
                  possibles[++possibles[0]] = newivec[i];

               wss_free( newivec );
            }
         }

         doables = wss_malloc( sizeof(int) * ( CRIT_ARG_MASK + 1 ) );
         doables[0] = 0;

         nextelem = _linker_crit_lookup_regex_elem( lf, re, elem->next );

         for( i = 1; i <= possibles[0]; i++ )
         {
            int *newivec =
               _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                              possibles[i] );
            if( newivec )
            {
               if( elem->next & CRIT_RE_UNCLE )
                  doables[++doables[0]] = possibles[i];
               else
               {
                  int j;
                  for( j = 1; j <= newivec[0]; j++ )
                     doables[++doables[0]] = newivec[j];
               }
               wss_free( newivec );
            }
         }
         wss_free( possibles );

         if( doables[0] > 0 )
         {
            if( elem->data & CRIT_RE_NEGATE )
            {
               wss_free( doables );
               return NULL;
            }
            else
               ivec = doables;
         }
         else
         {
            wss_free( doables );
            if( elem->data & CRIT_RE_NEGATE )
            {
               ivec = _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                                     index + 1 );
               if( ( elem->next & CRIT_RE_UNCLE ) && ivec )
               {
                  wss_free( ivec );

                  ivec = wss_malloc( sizeof(int) * 2 );
                  ivec[0] = 1;
                  ivec[1] = index + 1;
               }
            }
         }

         break;
      }

      case RE_POSIX:
      {
         pd_char_t         ch;
         crit_regex_node_t *nextelem;
         pd_bool_t         rbool = FALSE;
         
         ch = str[index];
         if( ch == '\0' )
            return NULL;

         nextelem = _linker_crit_lookup_regex_elem( lf, re, elem->next );

         switch( elem->data & CRIT_ARG_MASK )
         {
            case RE_ALNUM:
            case RE_ALPHA:
            case RE_BLANK:
            case RE_CNTRL:
            case RE_DIGIT:
            case RE_GRAPH:
            case RE_LOWER:
            case RE_PRINT:
            case RE_PUNCT:
            case RE_SPACE:
            case RE_UPPER:
            case RE_XDIGIT:
               break;
         }

         if( elem->data & CRIT_RE_NEGATE )
            rbool = rbool?FALSE:TRUE;
         if( rbool )
         {
            ivec = _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                                  index + 1 );
            if( ( elem->next & CRIT_RE_UNCLE ) && ivec )
            {
               wss_free( ivec );

               ivec = wss_malloc( sizeof(int) * 2 );
               ivec[0] = 1;
               ivec[1] = index + 1;
            }
         }
         break;
      }

      case RE_RANGE:
      {
         crit_regex_node_t *dataelem =
            _linker_crit_lookup_regex_elem( lf, re, elem->data );
         if( dataelem && ( dataelem->op == RE_NOP ) )
         {
            pd_char_t         lo       = dataelem->next;
            pd_char_t         hi       = dataelem->oneofnext;
            pd_char_t         ch       = str[index];
            crit_regex_node_t *nextelem =
               _linker_crit_lookup_regex_elem( lf, re, elem->next );

            if( elem->data & CRIT_RE_INSENS )
               ch = wsi_tolower(ch);

            ivec = ( ( ch >= lo ) && ( ch <= hi ) )?
                      _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                                     index + 1 ) : NULL;
            if( ( elem->next & CRIT_RE_UNCLE ) && ivec )
            {
               wss_free( ivec );

               ivec = wss_malloc( sizeof(int) * 2 );
               ivec[0] = 1;
               ivec[1] = index + 1;
            }
         }
         break;
      }

      case RE_REPEAT:
      {
         crit_regex_node_t *dataelem =
            _linker_crit_lookup_regex_elem( lf, re, elem->data );

         if( dataelem && ( dataelem->op == RE_NOP ) )
         {
            int                min     = dataelem->next;
            int                max     = dataelem->oneofnext;
            crit_regex_node_t *repelem =
               _linker_crit_lookup_regex_elem( lf, re, dataelem->data );
            if( repelem )
            {
               int                nrepeats;
               int                posindex;
               int                i;
               int               *newivec;
               pd_bool_t         stingy;
               crit_regex_node_t *nextelem;
               union
               {
                  int i;
                  int *iptr;
               }
               *pos;

               nrepeats = 0;
               stingy = ( dataelem->data & CRIT_RE_STINGY )?TRUE:FALSE;
               nextelem = _linker_crit_lookup_regex_elem( lf, re, elem->next );

               newivec = wss_malloc( sizeof(int) * ( CRIT_ARG_MASK + 1 ) );
               newivec[0] = 1;
               newivec[1] = index;

               pos = wss_malloc( sizeof(*pos) * ( CRIT_ARG_MASK + 1 ) );
               pos[0].i = 0;

               do
               {
                  int *doables;

                  pos[++pos[0].i].iptr = newivec;

                  if( stingy && ( nrepeats >= min ) )
                     for( i = 1; i <= newivec[0]; i++ )
                     {
                        int *s =
                           _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                                          newivec[i] );
                        if( s )
                        {
                           int j;

                           if( elem->next & CRIT_RE_UNCLE )
                           {
                              wss_free( s );

                              s = wss_malloc( sizeof(int) * 2 );
                              s[0] = 1;
                              s[1] = newivec[i];
                           }

                           for( j = 1; j <= pos[0].i; j++ )
                              wss_free( pos[j].iptr );
                           wss_free( pos );

                           return s;
                        }
                     }

                  doables = wss_malloc( sizeof(int) * ( CRIT_ARG_MASK + 1 ) );
                  doables[0] = 0;

                  for( i = 1; i <= newivec[0]; i++ )
                  {
                     int *thisresult =
                        _linker_crit_match_regex_elem( lf, str, re, repelem,
                                                       newivec[i] );
                     if( thisresult )
                     {
                        int j;

                        for( j = 1; j <= thisresult[0]; j++ )
                           doables[++doables[0]] = thisresult[j];
                        wss_free( thisresult );
                     }
                  }

                  if( doables[0] == 0 )
                  {
                     wss_free( doables );
                     break;
                  }

                  newivec = doables;
               }
               while( nrepeats++ < max );

               if( nrepeats < min )
               {
                  for( i = 1; i <= pos[0].i; i++ )
                     wss_free( pos[i].iptr );
                  wss_free( pos );
                  return NULL;
               }

               posindex = pos[0].i;
               while( posindex > min )
               {
                  newivec = pos[posindex--].iptr;
                  for( i = 1; i <= newivec[0]; i++ )
                  {
                     int *doneindex =
                        _linker_crit_match_regex_elem( lf, str, re, nextelem,
                                                       newivec[i] );
                     if( doneindex )
                     {
                        int j;

                        if( elem->next & CRIT_RE_UNCLE )
                        {
                           wss_free( doneindex );

                           doneindex = wss_malloc( sizeof(int) * 2 );
                           doneindex[0] = 1;
                           doneindex[1] = newivec[i];
                        }

                        for( j = 1; j <= pos[0].i; j++ )
                           wss_free( pos[j].iptr );
                        wss_free( pos );

                        return doneindex;
                     }
                  }
               }

               for( i = 1; i <= pos[0].i; i++ )
                  wss_free( pos[i].iptr );
               wss_free( pos );
            }
         }
         break;
      }

      case RE_START:
      {
         pd_char_t         ch;
         crit_regex_node_t *nextelem;
         
         ch = str[index];
         nextelem = _linker_crit_lookup_regex_elem( lf, re, elem->next );

         if( index == 0 )
         ivec = _linker_crit_match_regex_elem( lf, str, re, nextelem, index );

         if( ( elem->next & CRIT_RE_UNCLE ) && ivec )
         {
            wss_free( ivec );

            ivec = wss_malloc( sizeof(int) * 2 );
            ivec[0] = 1;
            ivec[1] = index;
         }
         break;
      }

      default:
         break;
   }

   return ivec;
}


static crit_expression_t *
_linker_crit_lookup_expression( wss_linker_file_t *lf, crit_id_t id )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;

   id &= CRIT_ARG_MASK;

   if( id > critdata->numexps )
      return NULL;
   else
      return critdata->exptab + id;
}

static crit_regex_t *
_linker_crit_lookup_regex( wss_linker_file_t *lf, crit_id_t id )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;

   id &= CRIT_ARG_MASK;

   if( id > critdata->numregex )
      return NULL;
   else
      return critdata->regextab + id;
}

static crit_regex_node_t *
_linker_crit_lookup_regex_elem( wss_linker_file_t *lf, crit_regex_t *re,
                                crit_id_t id )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;

   id &= CRIT_ARG_MASK;

   if( id > re->regexsize )
      return NULL;
   else
      return critdata->regexelems + re->regex + id;
}

static crit_literal_t *
_linker_crit_lookup_literal( wss_linker_file_t *lf, crit_id_t id )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;

   id &= CRIT_ARG_MASK;

   if( id > critdata->numlits )
      return NULL;
   else
      return critdata->littab + id;
}

static pd_char_t *
_linker_crit_arg_to_string( wss_linker_file_t *lf, crit_id_t arg )
{
   if( arg & CRIT_OP_MASK )
   {
      crit_expression_t *exp = _linker_crit_lookup_expression( lf, arg );
      if( exp )
      {
         pd_char_t *str = _linker_crit_expression_to_string( lf, exp );
         if( str )
         {
            pd_char_t *retstr = wss_malloc_string( wsi_strlen(str) + 2 );
            wsi_sprintf( retstr, PD_MAKESTR("(%s)"), str );
            wss_free( str );
            return retstr;
         }
      }
   }
   else if( arg & CRIT_BUILTIN_MASK )
   {
      arg &= CRIT_ARG_MASK;
      if( arg < BUILTINS_SIZE )
         return wss_strdup(_linker_crit_builtins[arg]);
   }
   else if( arg & CRIT_RE_MASK )
   {
      crit_regex_t *re = _linker_crit_lookup_regex( lf, arg );
      if( re )
      {
         pd_char_t *str = _linker_crit_regex_to_string( lf, re );
         if( str )
         {
            pd_char_t *retstr = wss_malloc_string( wsi_strlen(str) + 2 );
            wsi_sprintf( retstr, PD_MAKESTR("|%s|"), str );
            wss_free( str );
            return retstr;
         }
      }
   }
   else
   {
      crit_literal_t *lit = _linker_crit_lookup_literal( lf, arg );
      if( lit )
         return _linker_crit_literal_to_string( lf, lit );
   }

   return NULL;
}

static linker_crit_compat_t
_linker_crit_compat_arg( wss_linker_file_t *lf, crit_id_t arg )
{
   if( arg & CRIT_OP_MASK )
      return LCC_EXP;
   else if( arg & CRIT_BUILTIN_MASK )
   {
      arg &= CRIT_ARG_MASK;
      if( arg < BUILTINS_SIZE )
         return _linker_crit_builtins_compat[arg];
   }
   else if( arg & CRIT_RE_MASK )
   {
      return LCC_RE;
   }
   else
   {
      crit_literal_t *lit = _linker_crit_lookup_literal( lf, arg );
      if( lit )
      {
         switch( lit->type )
         {
            case CT_INT:
            case CT_LABEL:
               return LCC_INT;

            case CT_STRING:
               return LCC_STRING;

            default:
               break;
         }
      }
   }

   return LCC_UNKNOWN;
}

static pd_char_t *
_linker_crit_expression_to_string( wss_linker_file_t *lf,
                                   crit_expression_t *exp )
{
   pd_char_t *str = NULL;

   if( exp->seen )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: recursive criteria\n") );
#endif
      return NULL;
   }

   exp->seen = TRUE;

   if( exp->op < OPS_SIZE )
   {
      const pd_char_t      *op   = _linker_crit_op_names[exp->op];
      pd_char_t            *arg1 = _linker_crit_arg_to_string( lf, exp->arg1 );
      linker_crit_compat_t  lcc1 = _linker_crit_compat_arg( lf, exp->arg1 );
      pd_char_t            *arg2 = NULL;
      linker_crit_compat_t  lcc2 = LCC_UNKNOWN;

      if( exp->op != CO_NOT )
      {
         arg2 = _linker_crit_arg_to_string( lf, exp->arg2 );
         lcc2 = _linker_crit_compat_arg( lf, exp->arg2 );
      }

      switch( exp->op )
      {
         case CO_AND:
         case CO_OR:
         case CO_LT:
         case CO_GT:
         case CO_LTE:
         case CO_GTE:
         case CO_NE:
         case CO_EQ:
            if( arg1 && arg2 && ( lcc1 == lcc2 ) && ( lcc1 != LCC_UNKNOWN ) )
            {
               str = wss_malloc_string( wsi_strlen(op) + wsi_strlen(arg1) +
                                        wsi_strlen(arg2) + 2 );
               wsi_sprintf( str, PD_MAKESTR("%s %s %s"), arg1, op, arg2 );
            }
            break;

         case CO_NOT:
            if( arg1 && !arg2 && ( lcc1 != LCC_UNKNOWN ) )
            {
               str = wss_malloc_string( wsi_strlen(op) + wsi_strlen(arg1) );
               wsi_sprintf( str, PD_MAKESTR("%s %s"), op, arg1 );
            }
            break;

         case CO_RE:
            if( arg1 && arg2 && ( lcc1 == LCC_STRING ) && ( lcc2 == LCC_RE ) )
            {
               str = wss_malloc_string( wsi_strlen(op) + wsi_strlen(arg1) +
                                        wsi_strlen(arg2) + 2 );
               wsi_sprintf( str, PD_MAKESTR("%s %s %s"), arg1, op, arg2 );
            }
            break;

         case CO_ARROW:
            if( arg1 && arg2 && ( lcc1 == LCC_EXP ) && ( lcc2 == LCC_STRING ) )
            {
               str = wss_malloc_string( wsi_strlen(op) + wsi_strlen(arg1) +
                                        wsi_strlen(arg2) + 2 );
               wsi_sprintf( str, PD_MAKESTR("%s %s %s"), arg1, op, arg2 );
            }
            break;

         case CO_SUID:
         case CO_SGID:
            if( arg1 && !arg2 && ( lcc1 == LCC_INT ) )
            {
               str = wss_malloc_string( wsi_strlen(op) + wsi_strlen(arg1) + 1 );
               wsi_sprintf( str, PD_MAKESTR("%s%s)"), op, arg1 );
            }
            else if( !arg1 && !arg2 )
            {
               str = wss_malloc_string( wsi_strlen(op) + 1 );
               wsi_sprintf( str, PD_MAKESTR("%s)"), op );
            }
            break;

         default:
            break;
      }

      if( arg1 ) wss_free( arg1 );
      if( arg2 ) wss_free( arg2 );
   }

   return str;
}

static pd_char_t *
_linker_crit_literal_to_string( wss_linker_file_t *lf, crit_literal_t *lit )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;
   pd_char_t         *str      = NULL;

   switch( lit->type )
   {
      case CT_INT:
      {
         pd_char_t buf[32];
         wsi_sprintf( buf, PD_MAKESTR("%d"), lit->ival );

         str = wss_strdup( buf );
         break;
      }

      case CT_STRING:
      case CT_LABEL:
      {
         if( lit->soff < critdata->hdr->strsize )
            str = wss_strdup( (pd_char_t *)( (pd_byte_t *)critdata->strtab +
                                              lit->soff ) );
         break;
      }

      default:
         break;
   }

   return str;
}

static pd_char_t *
_linker_crit_regex_to_string( wss_linker_file_t *lf, crit_regex_t *re )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;
   crit_regex_node_t  *elem     = critdata->regexelems + re->regex;

   return _linker_crit_regex_elem_to_string( lf, re, elem );
}

static pd_char_t *
_linker_crit_regex_elem_to_string( wss_linker_file_t *lf, crit_regex_t *re,
                                   crit_regex_node_t *elem )
{
   pd_char_t *str = NULL;

   if( elem == NULL )
      return str;

   switch( elem->op )
   {
      case RE_EXP:
      {
         crit_regex_node_t *exelem =
            _linker_crit_lookup_regex_elem( lf, re, elem->data );

         if( exelem )
         {
            pd_char_t *exstr =
               _linker_crit_regex_elem_to_string( lf, re, exelem );

            if( exstr )
            {
               str = wss_malloc_string( wsi_strlen(exstr) + 2 );
               wsi_sprintf( str, PD_MAKESTR("(%s)"), exstr );
               wss_free( exstr );
            }
         }
         break;
      }

      case RE_CHAR:
      {
         crit_literal_t *lit = _linker_crit_lookup_literal( lf, elem->data );

         if( lit )
            str = _linker_crit_literal_to_string( lf, lit );
         break;
      }

      case RE_ANY:
      {
         str = wss_malloc_string( 1 );
         wsi_strcpy( str, PD_MAKESTR(".") );
         break;
      }

      case RE_END:
      {
         str = wss_malloc_string( 1 );
         wsi_strcpy( str, PD_MAKESTR("$") );
         break;
      }

      case RE_ONEOF:
      {
         crit_regex_node_t *oneofelem = elem;
         while( oneofelem )
         {
            oneofelem =
               _linker_crit_lookup_regex_elem( lf, re, oneofelem->oneofnext );
            if( oneofelem )
            {
               pd_char_t *nextstr =
                  _linker_crit_regex_elem_to_string( lf, re, oneofelem );
               if( nextstr )
               {
                  if( str )
                  {
                     pd_char_t *newstr =
                        wss_malloc_string( wsi_strlen(str) +
                                           wsi_strlen(nextstr) +
                                           ((oneofelem->data & CRIT_RE_NEGATE)?
                                              0:1) );
                     wsi_sprintf( newstr, PD_MAKESTR("%s%s%s"), str,
                                  ((oneofelem->data & CRIT_RE_NEGATE)?
                                      PD_MAKESTR(""):PD_MAKESTR("|")),
                                  nextstr );
                     wss_free( str );
                     wss_free( nextstr );
                     str = newstr;
                  }
                  else
                     str = nextstr;
               }
               else
               {
                  wss_free( str );
                  str = NULL;
               }
            }
         }

         if( str )
         {
            pd_char_t *newstr =
               wss_malloc_string( wsi_strlen(str) +
                                  ((elem->data & CRIT_RE_NEGATE)?3:4) );
            if( elem->data & CRIT_RE_NEGATE )
               wsi_sprintf( newstr, PD_MAKESTR("[^%s]"), str );
            else
               wsi_sprintf( newstr, PD_MAKESTR("(?:%s)"), str );
            wss_free( str );
            str = newstr;
         }
         break;
      }

      case RE_POSIX:
      {
         crit_id_t id = elem->data & CRIT_ARG_MASK;
         if( id < POSIX_SIZE )
         {
            str =
               wss_malloc_string( wsi_strlen(_linker_crit_posix_charclass[id]) +
                                  4 + ((elem->data & CRIT_RE_NEGATE)?1:0) );
            wsi_sprintf( str, PD_MAKESTR("%s[:%s:]"),
                         ((elem->data & CRIT_RE_NEGATE)?
                            PD_MAKESTR(""):PD_MAKESTR("^")),
                         _linker_crit_posix_charclass[id] );
         }
         break;
      }

      case RE_RANGE:
      {
         crit_regex_node_t *dataelem =
            _linker_crit_lookup_regex_elem( lf, re, elem->data );
         if( dataelem && ( dataelem->op == RE_NOP ) )
         {
            str = wss_malloc_string(3);
            wsi_sprintf( str, PD_MAKESTR("%c-%c"), dataelem->next,
                         dataelem->oneofnext );
         }
         break;
      }

      case RE_REPEAT:
      {
         crit_regex_node_t *dataelem =
            _linker_crit_lookup_regex_elem( lf, re, elem->data );
         if( dataelem && ( dataelem->op == RE_NOP ) )
         {
            int                min     = dataelem->next;
            int                max     = dataelem->oneofnext;
            crit_regex_node_t *repelem =
               _linker_crit_lookup_regex_elem( lf, re, dataelem->data );
            if( repelem )
            {
               pd_char_t *repstr =
                  _linker_crit_regex_elem_to_string( lf, re, repelem );
               if( repstr )
               {
                  pd_bool_t stingy = ( dataelem->data & CRIT_RE_STINGY )?
                                      TRUE : FALSE;
                  if( ( max == INT_MAX ) && ( min <= 1 ) )
                  {
                     str = wss_malloc_string( wsi_strlen( repstr ) +
                                              (stingy?1:0) + 3 );
                     wsi_sprintf( str, PD_MAKESTR("(%s)%c%s"), repstr,
                                  (( min == 0 )?PD_MAKESTR('*')
                                               :PD_MAKESTR('+')),
                                  stingy?PD_MAKESTR("?"):PD_MAKESTR("") );
                  }
                  else if( ( min == 0 ) && ( max == 1 ) )
                  {
                     str = wss_malloc_string( wsi_strlen( repstr ) +
                                              (stingy?1:0) + 3 );
                     wsi_sprintf( str, PD_MAKESTR("(%s)?%s"), repstr,
                                  (stingy?PD_MAKESTR("?"):PD_MAKESTR("")) );
                  }
                  else
                  {
                     pd_char_t *minstr = wss_malloc_string(32);
                     size_t      ssize  = 4 + wsi_strlen( repstr ) +
                                          ( stingy?1:0 );

                     wsi_sprintf( minstr, PD_MAKESTR("%d"), min );
                     ssize += wsi_strlen( minstr );

                     if( max > min )
                     {
                        ssize++;
                        if( max != INT_MAX )
                        {
                           pd_char_t *maxstr = wss_malloc_string(32);
                           wsi_sprintf( maxstr, PD_MAKESTR("%d"), max );

                           str = wss_malloc_string( ssize +
                                                    wsi_strlen(maxstr) );
                           wsi_sprintf( str, PD_MAKESTR("(%s){%s,%s}%s"),
                                        repstr, minstr, maxstr,
                                        (stingy?PD_MAKESTR("?")
                                               :PD_MAKESTR("")) );
                           wss_free( maxstr );
                        }
                        else
                        {
                           str = wss_malloc_string( ssize );
                           wsi_sprintf( str, PD_MAKESTR("(%s){%s,}%s"),
                                        repstr, minstr,
                                        (stingy?PD_MAKESTR("?")
                                               :PD_MAKESTR("")) );
                        }
                     }
                     else
                     {
                        str = wss_malloc( ssize );
                        wsi_sprintf( str, PD_MAKESTR("(%s){%s}%s"),
                                     repstr, minstr,
                                     (stingy?PD_MAKESTR("?")
                                            :PD_MAKESTR("")) );
                     }

                     wss_free( minstr );
                  }
                  wss_free( repstr );
               }
            }
         }
         break;
      }

      case RE_START:
      {
         str = wss_malloc_string( 1 );
         wsi_strcpy( str, PD_MAKESTR("^") );
         break;
      }

      default:
         break;
   }

   if( str && elem->next && !( elem->next & CRIT_RE_UNCLE ) )
   {
      crit_regex_node_t *nextelem =
         _linker_crit_lookup_regex_elem( lf, re, elem->next );

      if( nextelem )
      {
         pd_char_t *nextstr =
            _linker_crit_regex_elem_to_string( lf, re, nextelem );
         if( nextstr )
         {
            pd_char_t *newstr =
               wss_malloc_string( wsi_strlen(str) + wsi_strlen(nextstr) );
            wsi_sprintf( newstr, PD_MAKESTR("%s%s"), str, nextstr );
            wss_free( str );
            wss_free( nextstr );
            str = newstr;
         }
         else
         {
            wss_free( str );
            str = NULL;
         }
      }
   }
   
   return str;
}

#ifdef DEBUG_CRIT_LINKER
static void
_linker_crit_dump_literals( wss_linker_file_t *lf )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;
   crit_id_t           i;

   wsi_printf( PD_MAKESTR("Literals:\n") );

   for( i = 0; i < critdata->numlits; i++ )
   {
      crit_literal_t *lit = critdata->littab + i;

      switch( lit->type )
      {
         case CT_INT:
            wsi_printf( PD_MAKESTR("0x%04x | I %d\n"), i, lit->ival );
            break;

         case CT_STRING:
            wsi_printf( PD_MAKESTR("0x%04x | S 0x%x\n"), i, lit->soff );
            break;

         case CT_LABEL:
            wsi_printf( PD_MAKESTR("0x%04x | I %d  S 0x%x\n"), i,
                        lit->ival, lit->soff );
            break;

         default:
            wsi_printf( PD_MAKESTR("0x%04x | <INVALID>\n"), i );
            break;
      }
   }
}

static void
_linker_crit_dump_expressions( wss_linker_file_t *lf )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;
   crit_id_t           i;

   wsi_printf( PD_MAKESTR("Expressions:\n") );

   for( i = 0; i < critdata->numexps; i++ )
   {
      crit_expression_t *exp = critdata->exptab + i;
      char      a1c, a2c;
      crit_id_t a1id, a2id;

      a1c = a2c = 'L';
      a1id = exp->arg1;
      a2id = exp->arg2;
      if( a1id == 0xffffffff )
         a1c = 'N';
      else if( a1id & CRIT_OP_MASK )
      {
         a1c = 'E';
         a1id &= CRIT_ARG_MASK;
      }
      else if( a1id & CRIT_RE_MASK )
      {
         a1c = 'R';
         a1id &= CRIT_ARG_MASK;
      }
      else if( a1id & CRIT_BUILTIN_MASK )
      {
         a1c = 'B';
         a1id &= CRIT_ARG_MASK;
      }

      if( a2id == 0xffffffff )
         a2c = 'N';
      else if( a2id & CRIT_OP_MASK )
      {
         a2c = 'E';
         a2id &= CRIT_ARG_MASK;
      }
      else if( a2id & CRIT_RE_MASK )
      {
         a2c = 'R';
         a2id &= CRIT_ARG_MASK;
      }
      else if( a2id & CRIT_BUILTIN_MASK )
      {
         a2c = 'B';
         a2id &= CRIT_ARG_MASK;
      }

      switch( exp->op )
      {
         case CO_EQ:
            wsi_printf( PD_MAKESTR("0x%04x | %c 0x%04x == %c 0x%04x\n"),
                        i, a1c, a1id, a2c, a2id );
            break;

         case CO_ARROW:
            wsi_printf( PD_MAKESTR("0x%04x | %c 0x%04x => %c 0x%04x\n"),
                        i, a1c, a1id, a2c, a2id );
            break;

         case CO_OR:
            wsi_printf( PD_MAKESTR("0x%04x | %c 0x%04x || %c 0x%04x\n"),
                        i, a1c, a1id, a2c, a2id );
            break;

         case CO_AND:
            wsi_printf( PD_MAKESTR("0x%04x | %c 0x%04x && %c 0x%04x\n"),
                        i, a1c, a1id, a2c, a2id );
            break;

         case CO_RE:
            wsi_printf( PD_MAKESTR("0x%04x | %c 0x%04x =~ %c 0x%04x\n"),
                        i, a1c, a1id, a2c, a2id );
            continue;

         default:
            wsi_printf( PD_MAKESTR("0x%04x | <INVALID>\n"), i );
            continue;
      }

   }
}
#endif /* DEBUG_CRIT_LINKER */


/*
 * Lookup a single symbol in the binary's symbol table and return the
 * symbol structure
 */
static int
_linker_crit_lookup_symbol( wss_linker_file_t *lf, const pd_char_t *symbol,
                            wss_linker_addr_t *addr )
{
   linker_crit_data_t *critdata = (linker_crit_data_t *)lf->linker_file_data;

   if( ( addr == NULL ) || ( symbol == NULL ) )
      return EINVAL;

   if( !wsi_strcmp( symbol, PD_MAKESTR(WRLINKNAME_MOD_TYPE) ) )
   {
      if( critdata->hdr->activation )
         addr->addr = (void *) &_linker_crit_act_name;
      else
         addr->addr = (void *) &_linker_crit_deact_name;

      return 0;
   }
   if( !wsi_strcmp( symbol, PD_MAKESTR(WRLINKNAME_MOD) ) )
   {
      addr->addr = &(critdata->name);
      return 0;
   }
   if( ( critdata->hdr->activation == TRUE ) &&
       !wsi_strcmp( symbol, PD_MAKESTR("activate") ) )
   {
      addr->addr = _linker_crit_entrypoint;
      return 0;
   }
   if( ( critdata->hdr->activation == FALSE ) &&
       !wsi_strcmp( symbol, PD_MAKESTR("deactivate") ) )
   {
      addr->addr = _linker_crit_entrypoint;
      return 0;
   }
   if( !wsi_strcmp( symbol, PD_MAKESTR("imp_list") ) )
   {
      addr->addr = &(critdata->implist);
      return 0;
   }

   return ENOENT;
}


/*
 * Export library symbols for other modules to use
 */
static int
_linker_crit_export_symbols( wss_linker_file_t *lf )
{
   return 0;
}


/*  _linker_crit_load */
/*! \brief Load a criteria module
 *
 *  The module does not depend on the file image in the filesystem once it is
 *  finalized.
 *
 *  \param filename Pointer to string containing the name of the file to load
 *  \param result   Pointer to linker file structure to fill
 *
 *  \return 0 if successfully loaded,
 *  \return ENOENT if unable to resolve symbols,
 *  \return EEXIST if the module is already loaded,
 *  \return EINVAL if the module is of an unknown type
 */                                                               
static int
_linker_crit_load( const pd_char_t *filename, wss_linker_file_t *result )
{
   wss_file_t          file;
   size_t              size;
   void               *bin_image;
   crit_header_t      *header;
   crit_expression_t  *exp;
   crit_literal_t     *lit;
   crit_regex_t       *re;
   linker_crit_data_t *critdata;
   crit_id_t           i;
   size_t              elemsize;

   /* Make sure result pointer is non-NULL */
   if( result == NULL )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: result argument is NULL\n") );
#endif /* DEBUG_CRIT_LINKER */
      return EINVAL;
   }
   
#ifdef DEBUG_CRIT_LINKER
   wsi_printf( PD_MAKESTR("Criteria Linker: Processing '%s'...\n"), filename );
#endif /* DEBUG_CRIT_LINKER */
   
   file = wss_file_open( filename, WSS_FILE_READ | WSS_FILE_ROOTPRIV );
   if( file == WSS_FILE_INVALID )
      return ENOENT;

   size = wss_file_get_size( file );
   if( size < sizeof(crit_header_t) )
   {
      wss_file_close( file );
      wss_file_rele( file );
      return ENOEXEC;
   }

   result->linker_file_size = size;

   bin_image = wss_file_read_alloc( file, &size );
   wss_file_close( file );

   if( bin_image == NULL )
   {
      return ENOMEM;
   }

   header = (crit_header_t *)bin_image;

   if( header->magic != CRIT_MAGIC )
   {
      wss_free( bin_image );
      return ENOEXEC;
   }

   if( header->version != CRIT_VERSION )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Mismatched version %d, ")
                  PD_MAKESTR("expected %d\n"), header->version, CRIT_VERSION );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      return ENOEXEC;
   }

   if( ( header->strings > size ) ||
       ( ( header->strings + header->strsize ) > size ) ||
       ( header->literals > size ) ||
       ( ( header->literals + header->litsize ) > size ) ||
       ( header->litsize % sizeof(crit_literal_t) ) ||
       ( header->expression > size ) ||
       ( ( header->expression + header->expsize ) > size ) ||
       ( header->expsize % sizeof(crit_expression_t) ) )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Possible corrupt module\n") );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      return EINVAL;
   }

   critdata = (linker_crit_data_t *)wss_malloc( sizeof(linker_crit_data_t) );
   if( critdata == NULL )
   {
      wss_free( bin_image );
      return ENOMEM;
   }

   bzero( critdata, sizeof(*critdata) );
   critdata->hdr = header;
   critdata->strtab = (pd_char_t *)( (pd_byte_t *)bin_image +
                                      header->strings );
   critdata->littab = (crit_literal_t *)( (pd_byte_t *)bin_image +
                                          header->literals );
   critdata->numlits = header->litsize / sizeof(crit_literal_t);
   critdata->exptab = (crit_expression_t *)( (pd_byte_t *)bin_image +
                                             header->expression );
   critdata->numexps = header->expsize / sizeof(crit_expression_t);
   critdata->regextab = (crit_regex_t *)( (pd_byte_t *)bin_image +
                                          header->regextab );
   critdata->numregex = header->regextabsize / sizeof(crit_regex_t);
   critdata->regexelems = (crit_regex_node_t *)( critdata->regextab +
                                                 critdata->numregex );
   critdata->numregexelems = ( header->strings -
                               ( header->regextab + header->regextabsize ) ) /
                             sizeof(crit_regex_node_t);
   critdata->implist[0] = NULL;
   critdata->implist[1] = NULL;

   if( header->start > critdata->numexps )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Possible corrupt module\n") );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      wss_free( critdata );
      return EINVAL;
   }

#ifdef DEBUG_CRIT_LINKER
   wsi_printf( PD_MAKESTR("Criteria Linker: Header Information\n") );
   wsi_printf( PD_MAKESTR("-----------------------------------\n") );
   wsi_printf( PD_MAKESTR("    String Table Offset: 0x%x\n"),
               header->strings );
   wsi_printf( PD_MAKESTR("      String Table Size: 0x%x\n"),
               header->strsize );
   wsi_printf( PD_MAKESTR("   Literal Table Offset: 0x%x\n"),
               header->literals );
   wsi_printf( PD_MAKESTR("     Literal Table Size: 0x%x\n"),
               header->litsize );
   wsi_printf( PD_MAKESTR("Expression Table Offset: 0x%x\n"),
               header->expression );
   wsi_printf( PD_MAKESTR("  Expression Table Size: 0x%x\n"),
               header->expsize );
   wsi_printf( PD_MAKESTR("     Regex Table Offset: 0x%x\n"),
               header->regextab );
   wsi_printf( PD_MAKESTR("       Regex Table Size: 0x%x\n"),
               header->regextabsize );
   wsi_printf( PD_MAKESTR("       Start Expression: 0x%x\n"),
               header->start );
   wsi_printf( PD_MAKESTR("-----------------------------------\n") );
#endif /* DEBUG_CRIT_LINKER */

   /* Make sure regex element tables are within bounds */
   elemsize = header->strings - ( header->regextab + header->regextabsize );
   if( elemsize % sizeof(crit_regex_node_t) )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Possible corrupt module\n") );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      wss_free( critdata );
      return EINVAL;
   }
   else
   {
      for( i = 0; i < critdata->numregex; i++ )
      {
         re = critdata->regextab + i;
         if( re->regex >= critdata->numregexelems )
         {
#ifdef DEBUG_CRIT_LINKER
            wsi_printf( PD_MAKESTR("Criteria Linker: Possible corrupt ")
                        PD_MAKESTR("module\n") );
#endif /* DEBUG_CRIT_LINKER */
            wss_free( bin_image );
            wss_free( critdata );
            return EINVAL;
         }
      }
   }

   result->linker_file_data = (void *)critdata;
   result->linker_file_ops = &_linker_crit_file_ops;
   result->linker_file_image.addr = (void *)bin_image;

#ifdef DEBUG_CRIT_LINKER
   _linker_crit_dump_literals( result );
   _linker_crit_dump_expressions( result );
#endif /* DEBUG_CRIT_LINKER */

   /* Clear the seen flag on all expressions */
   for( i = 0; i < critdata->numexps; i++ )
   {
      exp = critdata->exptab + i;
      exp->seen = FALSE;
   }

   /* Find the start expression */
   exp = _linker_crit_lookup_expression( result, critdata->hdr->start );

   if( exp == NULL )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Unable to find start ")
                  PD_MAKESTR("expression\n") );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      wss_free( critdata );

      result->linker_file_data = NULL;
      result->linker_file_ops = NULL;
      result->linker_file_image.addr = NULL;

      return EINVAL;
   }

   if( exp->op != CO_ARROW )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Invalid start expression\n") );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      wss_free( critdata );

      result->linker_file_data = NULL;
      result->linker_file_ops = NULL;
      result->linker_file_image.addr = NULL;

      return EINVAL;
   }

   critdata->name = _linker_crit_expression_to_string( result, exp );
   if( critdata->name == NULL )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Invalid criteria string\n") );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      wss_free( critdata );

      result->linker_file_data = NULL;
      result->linker_file_ops = NULL;
      result->linker_file_image.addr = NULL;

      return EINVAL;
   }

   /* construct import list based on literal of arrow expression */
   lit = _linker_crit_lookup_literal( result, exp->arg2 );
   if( lit == NULL )
   {
#ifdef DEBUG_CRIT_LINKER
      wsi_printf( PD_MAKESTR("Criteria Linker: Invalid import list\n") );
#endif /* DEBUG_CRIT_LINKER */
      wss_free( bin_image );
      wss_free( critdata );

      result->linker_file_data = NULL;
      result->linker_file_ops = NULL;
      result->linker_file_image.addr = NULL;

      return EINVAL;
   }
   critdata->implist[0] = _linker_crit_literal_to_string( result, lit );

   return 0;
}

static int
_linker_crit_finalize( wss_linker_file_t *lf )
{
   return 0;
}

static int
_linker_crit_unload( wss_linker_file_t *lf )
{
   linker_crit_data_t *critdata;

   if( lf == NULL )
      return EINVAL;

   critdata = (linker_crit_data_t *)lf->linker_file_data;
   if( critdata )
   {
      if( critdata->implist[0] )
         wss_free( critdata->implist[0] );
      if( critdata->name )
         wss_free( critdata->name );
   }

   return wss_linker_file_destroy( lf );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

void EXPORT
linker_crit_init()
{
   wss_linker_register( _linker_crit_description, &_linker_crit_ops );
}

