/*! \file linker_elf32.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker_elf32.c,v 1.15 2001/06/13 13:30:52 stevek Exp $
 *
 * \brief   Support for loading ELF-32 format modules
 * \author  Steve Kiernan
 * \version $Revision: 1.15 $
 * \date    $Date: 2001/06/13 13:30:52 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wsi_types.h>

#include <wss_file.h>
#include <wss_malloc.h>

#include "linker_elf32.h"


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#define MAX_PHDR  32

#define ELF32_DATA_RELOC_PASS1  0x01
#define ELF32_DATA_RELOC_DONE   0x02


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

static int _linker_elf32_load           __P(( const pd_char_t *,
                                              wss_linker_file_t * ));
static int _linker_elf32_lookup_symbol  __P(( wss_linker_file_t *,
                                              const pd_char_t *,
                                              wss_linker_addr_t * ));
static int _linker_elf32_export_symbols __P(( wss_linker_file_t * ));
static int _linker_elf32_finalize       __P(( wss_linker_file_t * ));
static int _linker_elf32_unload         __P(( wss_linker_file_t * ));


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static wss_linker_ops_t _linker_elf32_ops =
{
   _linker_elf32_load
};

static wss_linker_file_ops_t _linker_elf32_file_ops =
{
   _linker_elf32_lookup_symbol,
   _linker_elf32_export_symbols,
   _linker_elf32_finalize,
   _linker_elf32_unload
};

static const pd_char_t *_linker_elf32_description = PD_MAKESTR("ELF-32");


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * Lookup a single symbol in the binary's symbol table and return the
 * symbol structure
 */
static int
_linker_elf32_lookup_symbol( wss_linker_file_t *lf, const pd_char_t *symbol,
                             wss_linker_addr_t *addr )
{
   int                  i;
   int                  entries;
   linker_elf32_data_t *elfdata = (linker_elf32_data_t *)lf->linker_file_data;
   Elf32_Sym           *sym;
   void                *str;

   if( ( addr == NULL ) || ( symbol == NULL ) )
      return EINVAL;

   /* If the symbol table has been exported, use it */
   if( lf->linker_file_symbols != NULL )
   {
      wss_symbol_t *syment;
      for( syment = lf->linker_file_symbols; syment && syment->name; syment++ )
      {
         if( !wsi_strcmp( syment->name, symbol ) )
         {
            addr->addr32 = syment->addr;
            return 0;
         }
      }
   }

   if( elfdata == NULL )
      return EINVAL;

   /* Otherwise, search the symbol table in the image */
   if( elfdata->symtab->sh_type != SHT_SYMTAB )
      return EINVAL;
   
   sym = (Elf32_Sym *)( (pd_byte_t *)(lf->linker_file_image.addr) +
                        elfdata->symtab->sh_offset );
   str = (void *)( (pd_byte_t *)(lf->linker_file_image.addr) +
                   elfdata->strtab->sh_offset );

   entries = elfdata->symtab->sh_size / sizeof(Elf32_Sym);
   for( i = 1; i < entries; i++ )
   {
      pd_char_t *tsym = (pd_char_t *)((pd_byte_t *)str + (sym+i)->st_name);
      if( !wsi_strcmp( tsym, symbol ) )
      {
         sym += i;

         for( i = 1;
              ( i < elfdata->ehdr->e_shnum ) && ( sym->st_shndx != i );
              i++ )
         ;

         if( sym->st_shndx == i )
         {
            bzero( addr, sizeof(*addr) );
            addr->addr32 = (u_int32_t)sym->st_value;
            return 0;
         }
      }
   }
   return ENOENT;
}


/*
 * Resolve and relocate symbols used/contained in the binary
 */
static int
_linker_elf32_relocate_symbols( wss_linker_file_t *lf )
{
   register void       *bin_image   = (void *)lf->linker_file_image.addr32;
   linker_elf32_data_t *elfdata     = (linker_elf32_data_t *)lf->linker_file_data;
   Elf32_Sym           *sym         = (Elf32_Sym *)( (pd_byte_t *)bin_image +
                                                     elfdata->symtab->sh_offset );
   int                  entries     = elfdata->symtab->sh_size /
                                      sizeof(Elf32_Sym);
   int                  nrelocated  = 0;
   int                  nresolved   = 0;
   int                  nunresolved = 0;
   register int         i;

   if( ( elfdata->flags & ELF32_DATA_RELOC_DONE ) != 0 )
      return 0;

   /* Resolve and compute relocated symbols */
   for( i = 1; i < entries; i++ )
   {
      Elf32_Sym *s = sym + i;
      switch( s->st_shndx )
      {
         case SHN_ABS:
            break;

         case SHN_UNDEF:
         case SHN_COMMON:
         {
            /* Only process after first reloc pass */
            if( ( elfdata->flags & ELF32_DATA_RELOC_PASS1 ) != 0 )
            {
               wss_linker_addr_t addr;
               pd_char_t *str = (pd_char_t *)( (pd_byte_t *)bin_image +
                                                 elfdata->strtab->sh_offset ) +
                                 s->st_name;

               if( wss_linker_search( lf, str, &addr ) == 0 )
               {
                  s->st_value = addr.addr32;
                  nresolved++;
               }
               else
               {
                  wsi_printf( PD_MAKESTR("ELF-32 Linker: Unresolved symbol "
                                          "'%s'\n"), str );
                  nunresolved++;
               }
            }
            break;
         }

         default:
         {
            /* Only process during first reloc pass */
            if( ( elfdata->flags & ELF32_DATA_RELOC_PASS1 ) == 0 )
            {
               switch( ELF32_ST_TYPE(s->st_info) )
               {
                  case STT_SECTION:
                     s->st_value = ((Elf32_Shdr *)((pd_byte_t *)bin_image +
                                                   elfdata->ehdr->e_shoff) +
                                   s->st_shndx)->sh_addr;
#ifdef DEBUG_ELF32_LINKER
                     wsi_printf( PD_MAKESTR("ELF-32 Linker: Set section[%d] @ "
                                             "0x%x\n"), s->st_shndx,
                                 (u_int)s->st_value );
#endif /* DEBUG_ELF32_LINKER */
                     break;

                  case STT_FUNC:
                     s->st_value +=
                        (u_long)( (pd_byte_t *)lf->linker_file_addr.addr32 +
                                  elfdata->text->sh_offset );
                     break;

                  case STT_OBJECT:
                     s->st_value += ((Elf32_Shdr *)((pd_byte_t *)bin_image +
                                                    elfdata->ehdr->e_shoff) +
                                    s->st_shndx)->sh_addr;
                     break;

                  case STT_NOTYPE:
                     break;

                  default:
                     wsi_printf( PD_MAKESTR("ELF-32 Linker: Non-function or "
                                             "section encountered: 0x%x\n"),
                                 ELF32_ST_TYPE(s->st_info) );
                     s->st_value += lf->linker_file_addr.addr32;
                     break;
               }
               nrelocated++;
            }
            break;
         }
      }
   }

   if( ( elfdata->flags & ELF32_DATA_RELOC_PASS1 ) == 0 )
      elfdata->flags |= ELF32_DATA_RELOC_PASS1;
   else if( ( elfdata->flags & ELF32_DATA_RELOC_DONE ) == 0 )
      elfdata->flags |= ELF32_DATA_RELOC_DONE;

#ifndef WARN_UNRESOLVED
   if( nunresolved > 0 )
   {
      wsi_printf( PD_MAKESTR("ELF-32 Linker: [%d unresolved symbol(s)] Unable"
                              " to load module.\n"),
                  nunresolved );
      return ENOENT;
   }
#endif

#ifdef DEBUG_ELF32_LINKER
   if( nrelocated > 0 )
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Relocated %d symbol(s)\n"),
                  nrelocated );
   if( nresolved > 0 )
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Resolved %d symbol(s)\n"),
                  nresolved );
#ifdef DEBUG_DUMP_SYMBOLS
   _wss_elf_dump_symbols( bin_image, symtab, strtab );
#endif /* DEBUG_DUMPSYMBOLS */
#endif /* DEBUG_ELF32_LINKER */

   return 0;
}


/*
 * Export library symbols for other modules to use
 */
static int
_linker_elf32_export_symbols( wss_linker_file_t *lf )
{
   linker_elf32_data_t *elfdata   = (linker_elf32_data_t *)lf->linker_file_data;
   register pd_byte_t *bin_image = lf->linker_file_image.addr;

   /* Just return if the symbol table has already been exported */
   if( lf->linker_file_symbols != NULL )
      return 0;

   /* Do not export the symbol table if all relocations have not completed */
   if( ( elfdata->flags & ELF32_DATA_RELOC_DONE ) == 0 )
      return EINVAL;

   if( elfdata->symtab->sh_type == SHT_SYMTAB )
   {
      Elf32_Sym  *sym           = (Elf32_Sym *)( bin_image +
                                                 elfdata->symtab->sh_offset );
      pd_byte_t *str           = (void *)( bin_image +
                                            elfdata->strtab->sh_offset);
      int         entries       = elfdata->symtab->sh_size /
                                  elfdata->symtab->sh_entsize;
      int         exported_syms = 0;
      int         i;

#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: %d candidates for exportable "
                              "symbols in table\n"),
                  entries );
#endif /* DEBUG_ELF32_LINKER */

      for( i = 1; i < entries; i++ )
      {
         Elf32_Sym *s = sym + i;

         /* Calculate size of exported symbol table */
         if( ( ELF32_ST_BIND( s->st_info ) == STB_GLOBAL ) &&
             ( ELF32_ST_TYPE( s->st_info ) == STT_FUNC ) )
            exported_syms++;
      }

      /* Allocate memory for symbol table */
      lf->linker_file_symbols =
         (wss_symbol_t *)wss_malloc( ( exported_syms + 1 ) *
                                     sizeof(wss_symbol_t) );
      bzero( lf->linker_file_symbols,
             ( exported_syms + 1 ) * sizeof(wss_symbol_t) );

      if( !exported_syms )
      {
         wsi_printf( PD_MAKESTR("ELF-32 Linker: Warning -- No symbols "
                                 "exported by library '%s'\n"),
                     lf->linker_file_name );
      }
      else
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR("ELF-32 Linker: Exporting %d symbols\n"),
                     exported_syms );
#endif /* DEBUG_ELF32_LINKER */

         /* Construct symbol table entries for exported functions */
         for( i = 1; (exported_syms >= 0) && (i < entries); i++ )
         {
            Elf32_Sym *s = sym + i;

            if( ( ELF32_ST_BIND( s->st_info ) == STB_GLOBAL ) &&
                ( ELF32_ST_TYPE( s->st_info ) == STT_FUNC ) )
            {
               wss_symbol_t *exsym = lf->linker_file_symbols + --exported_syms;
               pd_char_t   *symname = (pd_char_t *)( str + s->st_name );

               exsym->name = wss_strdup( symname );
               exsym->addr = (u_long)s->st_value;
               exsym->count = 0;
            }
         }
      }
   }
   else
      return EINVAL;

   return 0;
}


static int
_linker_elf32_relocate_section( wss_linker_file_t *lf,
                                linker_elf32_reloc_t sectype )
{
   pd_byte_t          *bin_image   = (pd_byte_t *)lf->linker_file_image.addr;
   int                  relocations;
   int                  i;
   Elf32_Shdr          *relshdr;
   Elf32_Shdr          *shdr;
   int                  type = 0;
   linker_elf32_data_t *elfdata = (linker_elf32_data_t *)lf->linker_file_data;
#ifdef DEBUG_ELF32_LINKER
   const pd_char_t             *sname;
#endif /* DEBUG_ELF32_LINKER */

   switch( sectype )
   {
      case ELF32_RELOC_TEXT_RELA:
         relshdr = elfdata->relatext;
         shdr = elfdata->text;
         type = SHT_RELA;
#ifdef DEBUG_ELF32_LINKER
         sname = PD_MAKESTR(".text");
#endif /* DEBUG_ELF32_LINKER */
         break;

      case ELF32_RELOC_TEXT_REL:
         relshdr = elfdata->reltext;
         shdr = elfdata->text;
         type = SHT_REL;
#ifdef DEBUG_ELF32_LINKER
         sname = PD_MAKESTR(".text");
#endif /* DEBUG_ELF32_LINKER */
         break;

      case ELF32_RELOC_DATA_RELA:
         relshdr = elfdata->reladata;
         shdr = elfdata->data;
         type = SHT_RELA;
#ifdef DEBUG_ELF32_LINKER
         sname = PD_MAKESTR(".data");
#endif /* DEBUG_ELF32_LINKER */
         break;

      case ELF32_RELOC_DATA_REL:
         relshdr = elfdata->reldata;
         shdr = elfdata->data;
         type = SHT_REL;
#ifdef DEBUG_ELF32_LINKER
         sname = PD_MAKESTR(".data");
#endif /* DEBUG_ELF32_LINKER */
         break;

      case ELF32_RELOC_RODATA_RELA:
         relshdr = elfdata->relarodata;
         shdr = elfdata->rodata;
         type = SHT_RELA;
#ifdef DEBUG_ELF32_LINKER
         sname = PD_MAKESTR(".rodata");
#endif /* DEBUG_ELF32_LINKER */
         break;

      case ELF32_RELOC_RODATA_REL:
         relshdr = elfdata->relrodata;
         shdr = elfdata->rodata;
         type = SHT_REL;
#ifdef DEBUG_ELF32_LINKER
         sname = PD_MAKESTR(".rodata");
#endif /* DEBUG_ELF32_LINKER */
         break;

      default:
         return EINVAL;
   }

   if( ( relshdr == NULL ) || ( shdr == NULL ) )
      return EINVAL;

   relocations = relshdr->sh_size / relshdr->sh_entsize;
#ifdef DEBUG_ELF32_LINKER
   wsi_printf( PD_MAKESTR("ELF-32 Linker: Relocating %d references in %s\n"),
               relocations, sname );
#endif /* DEBUG_ELF32_LINKER */
   for( i = 0; i < relocations; i++ )
   {
      Elf32_Rel *r       = (Elf32_Rel *)( bin_image + relshdr->sh_offset +
                                          relshdr->sh_entsize * i );
      void      *secaddr = (void *)( (pd_byte_t *)lf->linker_file_addr.addr32 +
                                     shdr->sh_offset );
      int        error   = _linker_elf32_machdep_relocate( lf, secaddr, type,
                                                           r );

      if( error )
         return error;
   }

   return 0;
}


/*
 * Process .text, .data, and .rodata section relocations
 */
static int
_linker_elf32_relocate_module( wss_linker_file_t *lf )
{
   linker_elf32_data_t *elfdata = (linker_elf32_data_t *)lf->linker_file_data;
   int                  error;

   if( elfdata->reltext )
   {
      error = _linker_elf32_relocate_section( lf, ELF32_RELOC_TEXT_REL );
      if( error )
         return error;
   }
   if( elfdata->relatext )
   {
      error = _linker_elf32_relocate_section( lf, ELF32_RELOC_TEXT_RELA );
      if( error )
         return error;
   }
   if( elfdata->reldata )
   {
      error = _linker_elf32_relocate_section( lf, ELF32_RELOC_DATA_REL );
      if( error )
         return error;
   }
   if( elfdata->reladata )
   {
      error = _linker_elf32_relocate_section( lf, ELF32_RELOC_DATA_RELA );
      if( error )
         return error;
   }
   if( elfdata->relrodata )
   {
      error = _linker_elf32_relocate_section( lf, ELF32_RELOC_RODATA_REL );
      if( error )
         return error;
   }
   if( elfdata->relarodata )
   {
      error = _linker_elf32_relocate_section( lf, ELF32_RELOC_RODATA_RELA );
      if( error )
         return error;
   }

   return 0;
}


/*
 * Resolve a pointer value in the module and return it in addr
 */
static int
_linker_elf32_resolve_pointer( wss_linker_file_t *lf, void *ptr,
                               wss_linker_addr_t *addr )
{
   pd_byte_t           *bin_image   = (pd_byte_t *)lf->linker_file_image.addr;
   int                  relocations;
   int                  i;
   Elf32_Shdr          *relshdr     = NULL;
   Elf32_Shdr          *shdr        = NULL;
   int                  type        = 0;
   int                  pass;
   linker_elf32_data_t *elfdata = (linker_elf32_data_t *)lf->linker_file_data;
#ifdef DEBUG_ELF32_LINKER
   const pd_char_t     *sname;
#endif /* DEBUG_ELF32_LINKER */
   int                  error;

   if( ( ptr == NULL ) || ( addr == NULL ) )
      return EINVAL;

   for( pass = 0; pass < 6; pass++ )
   {
      switch( pass )
      {
         case 0:
            if( elfdata->reltext )
            {
               relshdr = elfdata->reltext;
               shdr = elfdata->text;
               type = SHT_REL;
#ifdef DEBUG_ELF32_LINKER
               sname = PD_MAKESTR(".text");
#endif /* DEBUG_ELF32_LINKER */
               break;
            }
            pass++;

         case 1:
            if( elfdata->relatext )
            {
               relshdr = elfdata->relatext;
               shdr = elfdata->text;
               type = SHT_RELA;
#ifdef DEBUG_ELF32_LINKER
               sname = PD_MAKESTR(".text");
#endif /* DEBUG_ELF32_LINKER */
               break;
            }
            pass++;

         case 2:
            if( elfdata->reldata )
            {
               relshdr = elfdata->reldata;
               shdr = elfdata->data;
               type = SHT_REL;
#ifdef DEBUG_ELF32_LINKER
               sname = PD_MAKESTR(".data");
#endif /* DEBUG_ELF32_LINKER */
               break;
            }
            pass++;

         case 3:
            if( elfdata->reladata )
            {
               relshdr = elfdata->reladata;
               shdr = elfdata->data;
               type = SHT_RELA;
#ifdef DEBUG_ELF32_LINKER
               sname = PD_MAKESTR(".data");
#endif /* DEBUG_ELF32_LINKER */
               break;
            }
	    pass++;

         case 4:
            if( elfdata->relrodata )
            {
               relshdr = elfdata->relrodata;
               shdr = elfdata->rodata;
               type = SHT_REL;
#ifdef DEBUG_ELF32_LINKER
               sname = PD_MAKESTR(".rodata");
#endif /* DEBUG_ELF32_LINKER */
               break;
            }
            pass++;

         case 5:
            if( elfdata->relarodata )
            {
               relshdr = elfdata->relarodata;
               shdr = elfdata->rodata;
               type = SHT_RELA;
#ifdef DEBUG_ELF32_LINKER
               sname = PD_MAKESTR(".rodata");
#endif /* DEBUG_ELF32_LINKER */
               break;
            }
            return ENOENT;
      }

      if( ( relshdr == NULL ) || ( shdr == NULL ) )
         return EINVAL;

      relocations = relshdr->sh_size / relshdr->sh_entsize;
      for( i = 0; i < relocations; i++ )
      {
         Elf32_Rel *r       = (Elf32_Rel *)( bin_image + relshdr->sh_offset +
                                             relshdr->sh_entsize * i );
         void      *secaddr = (void *)((pd_byte_t *)lf->linker_file_addr.addr32 +
                                       shdr->sh_offset);
         u_int32_t storaddr = (u_int32_t)((char *)secaddr + r->r_offset);

         if( storaddr == (u_int32_t) ptr )
         {
            u_int32_t prevalue = *((u_int32_t *)storaddr);
            error = _linker_elf32_machdep_relocate( lf, secaddr, type, r );
            if( error == 0 )
               addr->addr32 = *((u_int32_t *)storaddr);
            *((u_int32_t *)storaddr) = prevalue;
            return( error );
         }
      }
   }

   return ENOENT;
}



/*  _linker_elf32_load */
/*! \brief Load an ELF-32 format module
 *
 *  Resolves and relocates symbols within the .text, .rodata, .data, and .bss
 *  sections.
 *
 *  The module does not depend on the file image in the filesystem once it is
 *  finalized.
 *
 *  \param filename Pointer to string containing the name of the file to load
 *  \param result   Pointer to linker file structure to fill
 *
 *  \return 0 if successfully loaded,
 *  \return ENOENT if unable to resolve symbols,
 *  \return EEXIST if the module is already loaded,
 *  \return EINVAL if the module is of an unknown type
 */                                                               
int EXPORT
_linker_elf32_load( const pd_char_t *filename, wss_linker_file_t *result )
{
   wss_file_t           file;
   size_t               size;
   void                *bin_image;
   void                *incore_image;
   linker_elf32_data_t *elfdata;
   Elf32_Word           memsize;
   Elf32_Word           offset;
   Elf32_Word           fixed;
   Elf32_Ehdr          *ehdr;
   wss_linker_addr_t    addr;
   int                  entries;
   int                  error        = 0;
   register int         i;

   /*
    * Section indicies
    */
   int                  text         = 0;
   int                  rodata       = 0;
   int                  data         = 0;
   int                  bss          = 0;
   int                  shstrtab     = 0;
   int                  strtab       = 0;
   int                  symtab       = 0;
   int                  relatext     = 0;
   int                  reltext      = 0;
   int                  reladata     = 0;
   int                  reldata      = 0;
   int                  relarodata   = 0;
   int                  relrodata    = 0;

   /* Make sure result pointer is non-NULL */
   if( result == NULL )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: result argument is NULL\n") );
#endif /* DEBUG_ELF32_LINKER */
      return EINVAL;
   }

#ifdef DEBUG_ELF32_LINKER
   wsi_printf( PD_MAKESTR("ELF-32 Linker: Processing '%s'...\n"), filename );
#endif /* DEBUG_ELF32_LINKER */

   result->linker_file_ops = &_linker_elf32_file_ops;

   file = wss_file_open( filename, WSS_FILE_READ | WSS_FILE_ROOTPRIV );
   if( file == WSS_FILE_INVALID )
      return ENOENT;

   size = wss_file_get_size( file );
   if( size < sizeof(Elf32_Ehdr) )
   {
      wss_file_close( file );
      wss_file_rele( file );
      return ENOEXEC;
   }

   bin_image = wss_file_read_alloc( file, &size );
   wss_file_close( file );

   if( bin_image == NULL )
   {
      return ENOMEM;
   }

   result->linker_file_image.addr32 = (u_int32_t)bin_image;
   result->linker_file_size = size;

   ehdr = (Elf32_Ehdr *)bin_image;

   if( ( ehdr->e_ident[EI_MAG0] != ELFMAG0 ) ||
       ( ehdr->e_ident[EI_MAG1] != ELFMAG1 ) ||
       ( ehdr->e_ident[EI_MAG2] != ELFMAG2 ) ||
       ( ehdr->e_ident[EI_MAG3] != ELFMAG3 ) )
   {
      wss_free( bin_image );
      return ENOEXEC;
   }

   if( ehdr->e_type != ET_REL )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Warning: ELF module not "
                              "relocatable\n") );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      return ENOEXEC;
   }

   if( ehdr->e_phnum > MAX_PHDR )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Warning: Possible corrupt ELF "
                              "file\n") );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      return ENOEXEC;
   }

   elfdata = (linker_elf32_data_t *)wss_malloc( sizeof(linker_elf32_data_t) );
   if( elfdata == NULL )
   {
      wss_free( bin_image );
      return ENOMEM;
   }

   bzero( elfdata, sizeof(*elfdata) );
   result->linker_file_data = (void *)elfdata;
   elfdata->ehdr = ehdr;
   elfdata->shdr = (Elf32_Shdr *)( (pd_byte_t *)ehdr + ehdr->e_shoff );

   if( _linker_elf32_machdep_supported( result ) )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Module not compiled for this "
                              "processor architecture\n") );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      wss_free( elfdata );
      return ENOEXEC;
   }



   /* Determine .shstrtab section */
   for( i = 1; i < ehdr->e_shnum; i++ )
   {
      Elf32_Shdr *ptr = elfdata->shdr + i;

      ptr->sh_addr = (Elf32_Addr)( (pd_byte_t *)bin_image +
                                    ptr->sh_offset );

      if( !shstrtab &&
          ( i == ehdr->e_shstrndx ) &&
          ( ptr->sh_type == SHT_STRTAB ) &&
          ( ptr->sh_flags == 0 ) )
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR(".shstrtab = 0x%x (0x%x) 0x%x bytes\n"),
                     (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                         ptr->sh_offset),
                     (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
         shstrtab = i;
         break;
      }
   }

   if( !shstrtab )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Unable to determine section ")
                  PD_MAKESTR("header string table from ELF module\n") );
#endif /* DEBUG_ELF32_LINKER */
      return EINVAL;
   }

   /* Determine relevant sections */
   for( i = 1; i < ehdr->e_shnum; i++ )
   {
      Elf32_Shdr *ptr    = elfdata->shdr + i;
      char       *shname;

      ptr->sh_addr = (Elf32_Addr)( (pd_byte_t *)bin_image +
                                    ptr->sh_offset );
      shname = (char *)( (pd_byte_t *)bin_image +
                         (elfdata->shdr + shstrtab)->sh_offset + ptr->sh_name );

      if( !text && !strcmp( shname, ".text" ) &&
          ( ptr->sh_type == SHT_PROGBITS ) &&
          ( ptr->sh_flags & SHF_EXECINSTR ) )
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR(".text = 0x%x (0x%x) 0x%x bytes\n"),
                     (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                         ptr->sh_offset),
                     (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
         text = i;
      }
      else if( !rodata && !strcmp( shname, ".rodata" ) &&
          ( ptr->sh_type == SHT_PROGBITS ) &&
          ( ptr->sh_flags & SHF_ALLOC ) )
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR(".rodata = 0x%x (0x%x) 0x%x bytes\n"),
                     (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                         ptr->sh_offset),
                     (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
         rodata = i;
      }
      else if( !data && !strcmp( shname, ".data" ) &&
          ( ptr->sh_type == SHT_PROGBITS ) &&
          ( ( ptr->sh_flags & (SHF_WRITE|SHF_ALLOC) ) ==
             (SHF_WRITE|SHF_ALLOC) ) )
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR(".data = 0x%x (0x%x) 0x%x bytes\n"),
                     (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                         ptr->sh_offset),
                     (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
         data = i;
      }
      else if( !bss && !strcmp( shname, ".bss" ) &&
          ( ptr->sh_type == SHT_NOBITS ) &&
          ( ptr->sh_flags & SHF_ALLOC ) )
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR(".bss = 0x%x (0x%x) 0x%x bytes\n"),
                     (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                         ptr->sh_offset),
                     (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
         bss = i;
      }
      else if( !strtab && !strcmp( shname, ".strtab" ) &&
          ( i != ehdr->e_shstrndx ) &&
          ( ptr->sh_type == SHT_STRTAB ) &&
          ( ptr->sh_flags == 0 ) )
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR(".strtab = 0x%x (0x%x) 0x%x bytes\n"),
                     (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                         ptr->sh_offset),
                     (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
         strtab = i;
      }
      else if( !symtab && !strcmp( shname, ".symtab" ) &&
          ( ptr->sh_type == SHT_SYMTAB ) &&
          ( ptr->sh_flags == 0 ) )
      {
#ifdef DEBUG_ELF32_LINKER
         wsi_printf( PD_MAKESTR(".symtab = 0x%x (0x%x) 0x%x bytes\n"),
                     (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                         ptr->sh_offset),
                     (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
         symtab = i;
      }
      else if( !relatext && !strcmp( shname, ".rela.text" ) &&
               ( ptr->sh_type == SHT_RELA ) &&
               ( ptr->sh_info == text ) )
      {
#ifdef DEBUG_ELF32_LINKER
            wsi_printf( PD_MAKESTR(".rela.text = 0x%x (0x%x) 0x%x bytes\n"),
                        (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                            ptr->sh_offset),
                        (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
            relatext = i;
      }
      else if( !reltext && !strcmp( shname, ".rel.text" ) &&
               ( ptr->sh_type == SHT_REL ) &&
               ( ptr->sh_info == text ) )
      {
#ifdef DEBUG_ELF32_LINKER
            wsi_printf( PD_MAKESTR(".rel.text = 0x%x (0x%x) 0x%x bytes\n"),
                        (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                            ptr->sh_offset),
                        (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
            reltext = i;
      }
      else if( !reladata && !strcmp( shname, ".rela.data" ) &&
               ( ptr->sh_type == SHT_RELA ) &&
               ( ptr->sh_info == data ) )
      {
#ifdef DEBUG_ELF32_LINKER
            wsi_printf( PD_MAKESTR(".rela.data = 0x%x (0x%x) 0x%x bytes\n"),
                        (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                            ptr->sh_offset),
                        (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
            reladata = i;
      }
      else if( !reldata && !strcmp( shname, ".rel.data" ) &&
               ( ptr->sh_type == SHT_REL ) &&
               ( ptr->sh_info == data ) )
      {
#ifdef DEBUG_ELF32_LINKER
            wsi_printf( PD_MAKESTR(".rel.data = 0x%x (0x%x) 0x%x bytes\n"),
                        (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                            ptr->sh_offset),
                        (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
            reldata = i;
      }
      else if( !relarodata && !strcmp( shname, ".rela.rodata" ) &&
               ( ptr->sh_type == SHT_RELA ) &&
               ( ptr->sh_info == rodata ) )
      {
#ifdef DEBUG_ELF32_LINKER
            wsi_printf( PD_MAKESTR(".rela.rodata = 0x%x (0x%x) 0x%x bytes\n"),
                        (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                            ptr->sh_offset),
                        (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
            relarodata = i;
      }
      else if( !relrodata && !strcmp( shname, ".rel.rodata" ) &&
               ( ptr->sh_type == SHT_REL ) &&
               ( ptr->sh_info == rodata ) )
      {
#ifdef DEBUG_ELF32_LINKER
            wsi_printf( PD_MAKESTR(".rel.rodata = 0x%x (0x%x) 0x%x bytes\n"),
                        (u_int)ptr, (u_int)((pd_byte_t *)bin_image +
                                            ptr->sh_offset),
                        (u_int)ptr->sh_size );
#endif /* DEBUG_ELF32_LINKER */
            relrodata = i;
      }
#ifdef DEBUG_ELF32_LINKER
      else
      {
         wsi_printf( PD_MAKESTR("[Skipped section #%d: %s]\n"), i, shname );
      }
#endif /* DEBUG_ELF32_LINKER */
   }

   if( !text || !strtab || !symtab || ( !reladata && !reldata) || !rodata ||
       !data )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Unable to determine necessary ")
                  PD_MAKESTR("sections from ELF module\n") );
#endif /* DEBUG_ELF32_LINKER */
      return EINVAL;
   }


   /* Determine memory necessary for in-core image */
   elfdata->symtab = elfdata->shdr + symtab;
   memsize = (elfdata->shdr + text)->sh_size;
   if( rodata )
   {
      /* Align on zero-nibble boundary */
      if( memsize & 0xf ) memsize = (memsize & ~0xf) + 0x10;
      (elfdata->shdr + rodata)->sh_addr = memsize;
      memsize += (elfdata->shdr + rodata)->sh_size;
   }
   if( data )
   {
      /* Align on zero-nibble boundary */
      if( memsize & 0xf ) memsize = (memsize & ~0xf) + 0x10;
      (elfdata->shdr + data)->sh_addr = memsize;
      memsize += (elfdata->shdr + data)->sh_size;
   }

   if( bss )
   {
      /* Align on zero-nible boundary */
      if( memsize & 0xf ) memsize = (memsize & ~0xf) + 0x10;
      (elfdata->shdr + bss)->sh_addr = memsize;
      memsize += (elfdata->shdr + bss)->sh_size;
   }

   /* Align on zero-nibble boundary */
   if( memsize & 0xf ) memsize = (memsize & ~0xf) + 0x10;
   fixed = memsize;

   entries = (elfdata->shdr + symtab)->sh_size / sizeof(Elf32_Sym);
   for( i = 1; i < entries; i++ )
   {
      Elf32_Sym *elfsym = (Elf32_Sym *)( (pd_byte_t *)bin_image +
                                         elfdata->symtab->sh_offset ) + i;

      if( elfsym->st_shndx == SHN_COMMON )
         memsize += elfsym->st_size;
   }


   /* Allocate memory for in-core image */
   incore_image = wss_malloc( memsize );
   if( incore_image == NULL )
   {
      wss_free( bin_image );
      wss_free( elfdata );
      return ENOMEM;
   }

   bzero( incore_image, memsize );
#ifdef DEBUG_ELF32_LINKER
   wsi_printf( PD_MAKESTR("ELF-32 Linker: Allocated in-core image @ 0x%x - "
                           "0x%x\n"),
               (u_int)incore_image,
               (u_int)((pd_byte_t *)incore_image + memsize - 1) );
#endif /* DEBUG_ELF32_LINKER */
   result->linker_file_addr.addr32 = (u_int32_t)incore_image;

   /* Relocate COMMON .bss entries */
   offset = (Elf32_Word)incore_image + fixed;
   for( i = 1; i < entries; i++ )
   {
      Elf32_Sym *elfsym = (Elf32_Sym *)((pd_byte_t *)bin_image +
                                        elfdata->symtab->sh_offset) + i;

      if( elfsym->st_shndx == SHN_COMMON )
      {
         elfsym->st_shndx = SHN_ABS;
         elfsym->st_value = offset;
         offset += elfsym->st_size;
      }
   }


   /* Copy .text, .rodata, and .data sections
      Zero .bss section */
   bcopy( (pd_byte_t *)bin_image + (elfdata->shdr + text)->sh_offset,
          incore_image, (elfdata->shdr + text)->sh_size );
   (elfdata->shdr + text)->sh_offset = 0;
   (elfdata->shdr + text)->sh_addr = (Elf32_Addr)incore_image;
   if( rodata )
   {
      bcopy( (pd_byte_t *)bin_image + (elfdata->shdr + rodata)->sh_offset,
             (pd_byte_t *)incore_image + (elfdata->shdr + rodata)->sh_addr,
             (elfdata->shdr + rodata)->sh_size );
      (elfdata->shdr + rodata)->sh_offset = (elfdata->shdr + rodata)->sh_addr;
      (elfdata->shdr + rodata)->sh_addr += (Elf32_Addr)incore_image;
   }
   if( data )
   {
      bcopy( (pd_byte_t *)bin_image + (elfdata->shdr + data)->sh_offset,
             (pd_byte_t *)incore_image + (elfdata->shdr + data)->sh_addr,
             (elfdata->shdr + data)->sh_size );
      (elfdata->shdr + data)->sh_offset = (elfdata->shdr + data)->sh_addr;
      (elfdata->shdr + data)->sh_addr += (Elf32_Addr)incore_image;
   }
   if( bss )
   {
      (elfdata->shdr + bss)->sh_offset = (elfdata->shdr + bss)->sh_addr;
      (elfdata->shdr + bss)->sh_addr += (Elf32_Addr)incore_image;
   }

   elfdata->strtab = elfdata->shdr + strtab;
   elfdata->text = elfdata->shdr + text;
   elfdata->relatext = relatext?(elfdata->shdr + relatext):NULL;
   elfdata->reltext = reltext?(elfdata->shdr + reltext):NULL;
   elfdata->rodata = elfdata->shdr + rodata;
   elfdata->relarodata = relarodata?(elfdata->shdr + relarodata):NULL;
   elfdata->relrodata = relrodata?(elfdata->shdr + relrodata):NULL;
   elfdata->data = elfdata->shdr + data;
   elfdata->reladata = reladata?(elfdata->shdr + reladata):NULL;
   elfdata->reldata = reldata?(elfdata->shdr + reldata):NULL;

   /* Relocate module symbols (ignore commons) */
   if( ( error = _linker_elf32_relocate_symbols( result ) ) )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Unable to relocate symbols\n") );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      wss_free( elfdata );
      return error;
   }

   /* Resolve dependencies */
   error = _linker_elf32_lookup_symbol( result, WRLINKNAME_MOD_DEPENDENCIES,
                                        &addr );
   if( error )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Unable to determine symbol "
                              "'%s'\n"), WRLINKNAME_MOD_DEPENDENCIES );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      wss_free( elfdata );
      return error;
   }

   if( ( error = _linker_elf32_resolve_pointer( result, addr.addr, &addr ) ) )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Unable to resolve pointer to")
                  PD_MAKESTR(" dependencies list (%d)\n"), error );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      wss_free( elfdata );
      return error;
   }

   result->linker_file_dependencies = (const pd_char_t *)addr.addr32;

#ifdef DEBUG_ELF32_LINKER
   if( addr.addr32 && *(char *)addr.addr32 )
      wsi_printf( PD_MAKESTR("ELF-32 Linker: First dependency = '%s'\n"),
                  result->linker_file_dependencies );
   else
      wsi_printf( PD_MAKESTR("ELF-32 Linker: No dependencies\n") );
#endif /* DEBUG_ELF32_LINKER */

   /* Construct the dependency list */
   if( ( error = wss_linker_construct_dependencies( result ) ) )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Not all dependencies loaded\n") );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      wss_free( elfdata );
      return error;
   }

   /* Relocate module symbols (resolving commons) */
   if( ( error = _linker_elf32_relocate_symbols( result ) ) )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Unable to relocate symbols\n") );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      wss_free( elfdata );
      return error;
   }

   /* Relocate module relative to relocation base address */
   if( ( error = _linker_elf32_relocate_module( result ) ) )
   {
#ifdef DEBUG_ELF32_LINKER
      wsi_printf( PD_MAKESTR("ELF-32 Linker: Unable to relocate ELF module\n") );
#endif /* DEBUG_ELF32_LINKER */
      wss_free( bin_image );
      wss_free( elfdata );
      return error;
   }

   return error;
}

static int
_linker_elf32_finalize( wss_linker_file_t *lf )
{
   if( lf->linker_file_data != NULL )
   {
      wss_free( lf->linker_file_data );
      lf->linker_file_data = NULL;
   }
   if( lf->linker_file_image.addr != NULL )
   {
      wss_free( lf->linker_file_image.addr );
      bzero( &lf->linker_file_image, sizeof(wss_linker_addr_t) );
   }

   return 0;
}

static int
_linker_elf32_unload( wss_linker_file_t *lf )
{
   if( lf == NULL )
      return EINVAL;

   return wss_linker_file_destroy( lf );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

void
linker_elf32_init()
{
   wss_linker_register( _linker_elf32_description, &_linker_elf32_ops );
}

