/*! \file linker_elf32.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker_elf32.h,v 1.6 2001/02/23 19:02:10 stevek Exp $
 *
 * \brief   Prototypes and structures used by the Elf-32 module linker
 * \author  Steve Kiernan
 * \version $Revision: 1.6 $
 * \date    $Date: 2001/02/23 19:02:10 $
 */

#ifndef __LINKER_ELF32_H
#define __LINKER_ELF32_H	1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif

#ifndef __WSS_LINKER_H
#include <wss_linker.h>
#endif

#include "linker_elf_common.h"


/***************************************************************************\
 * Public macros                                                           *
\***************************************************************************/

/* Macros for accessing the fields of r_info. */
#define ELF32_R_SYM(info)   ((info) >> 8)
#define ELF32_R_TYPE(info)  ((unsigned char)(info))

/* Macros for accessing the fields of st_info. */
#define ELF32_ST_BIND(info) ((info) >> 4)
#define ELF32_ST_TYPE(info) ((info) & 0xf)


/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/

/*
 * ELF types for 32-bit architectures
 */
typedef u_int32_t            Elf32_Addr;
typedef u_int16_t            Elf32_Half;
typedef u_int32_t            Elf32_Off;
typedef int32_t              Elf32_Sword;
typedef u_int32_t            Elf32_Word;
typedef u_int32_t            Elf32_Size;
typedef struct _Elf32_Ehdr   Elf32_Ehdr;
typedef struct _Elf32_Phdr   Elf32_Phdr;
typedef struct _Elf32_Shdr   Elf32_Shdr;
typedef struct _Elf32_Rel        Elf32_Rel;
typedef struct _Elf32_Rela       Elf32_Rela;
typedef struct _Elf32_Sym        Elf32_Sym;
typedef enum _linker_elf32_reloc linker_elf32_reloc_t;
typedef struct linker_elf32_data linker_elf32_data_t;


/*
 * Linker structures
 */
struct linker_elf32_data
{
   Elf32_Ehdr *ehdr;
   Elf32_Shdr *shdr;
   Elf32_Shdr *rodata;
   Elf32_Shdr *relrodata;
   Elf32_Shdr *relarodata;
   Elf32_Shdr *data;
   Elf32_Shdr *reldata;
   Elf32_Shdr *reladata;
   Elf32_Shdr *text;
   Elf32_Shdr *reltext;
   Elf32_Shdr *relatext;
   Elf32_Shdr *symtab;
   Elf32_Shdr *strtab;
   int         flags;
};

enum _linker_elf32_reloc
{
   ELF32_RELOC_UNKNOWN,
   ELF32_RELOC_TEXT_REL,
   ELF32_RELOC_TEXT_RELA,
   ELF32_RELOC_DATA_REL,
   ELF32_RELOC_DATA_RELA,
   ELF32_RELOC_RODATA_REL,
   ELF32_RELOC_RODATA_RELA
};


/*
 * ELF header
 */
struct _Elf32_Ehdr
{
   unsigned char e_ident[EI_NIDENT]; /* File identification */
   Elf32_Half    e_type;             /* File type */
   Elf32_Half    e_machine;          /* Machine architecture */
   Elf32_Word    e_version;          /* ELF format version */
   Elf32_Addr    e_entry;            /* Entry point */
   Elf32_Off     e_phoff;            /* Program header file offset */
   Elf32_Off     e_shoff;            /* Section header file offset */
   Elf32_Word    e_flags;            /* Architecture-specific flags */
   Elf32_Half    e_ehsize;           /* Size of ELF header in bytes */
   Elf32_Half    e_phentsize;        /* Size of program header entry */
   Elf32_Half    e_phnum;            /* Number of program header entries */
   Elf32_Half    e_shentsize;        /* Size of section header entry */
   Elf32_Half    e_shnum;            /* Number of section header entries */
   Elf32_Half    e_shstrndx;         /* Section name strings section */
};


/*
 * Program header
 */
struct _Elf32_Phdr
{
   Elf32_Word    p_type;             /* Entry type */
   Elf32_Off     p_offset;           /* File offset of contents */
   Elf32_Addr    p_vaddr;            /* Virtual address in memory image */
   Elf32_Addr    p_paddr;            /* Physical address */
   Elf32_Size    p_filesz;           /* Size of contents in file */
   Elf32_Size    p_memsz;            /* Size of contents in memory */
   Elf32_Word    p_flags;            /* Access permission flags */
   Elf32_Size    p_align;            /* Alignment in memory and file */
};


/*
 * Section header.
 */
struct _Elf32_Shdr
{
   Elf32_Word    sh_name;            /* Section name */
   Elf32_Word    sh_type;            /* Section type */
   Elf32_Word    sh_flags;           /* Section flags */
   Elf32_Addr    sh_addr;            /* Address in memory image */
   Elf32_Off     sh_offset;          /* Offset in file */
   Elf32_Size    sh_size;            /* Size in bytes */
   Elf32_Word    sh_link;            /* Index of a related section */
   Elf32_Word    sh_info;            /* Depends on section type */
   Elf32_Size    sh_addralign;       /* Alignment in bytes */
   Elf32_Size    sh_entsize;         /* Size of each entry in section */
};


/*
 * Relocation entries
 */
/* Relocations that don't need an addend field */
struct _Elf32_Rel
{
   Elf32_Addr    r_offset;           /* Location to be relocated */
   Elf32_Word    r_info;             /* Relocation type and symbol index */
};

/* Relocations that need an addend field */
struct _Elf32_Rela
{
   Elf32_Addr    r_offset;           /* Location to be relocated */
   Elf32_Word    r_info;             /* Relocation type and symbol index */
   Elf32_Sword   r_addend;           /* Addend */
};


/*
 * Symbol table entries.
 */
struct _Elf32_Sym
{
   Elf32_Word    st_name;            /* String table index of name */
   Elf32_Addr    st_value;           /* Symbol value */
   Elf32_Size    st_size;            /* Size of associated object */
   unsigned char st_info;            /* Type and binding information */
   unsigned char st_other;           /* Reserved */
   Elf32_Half    st_shndx;           /* Section index of symbol */
};


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int WSSLINKAGE _linker_elf32_machdep_relocate  __P(( wss_linker_file_t *,
                                                     void *, int,
                                                     Elf32_Rel * ));
int WSSLINKAGE _linker_elf32_machdep_supported __P(( wss_linker_file_t * ));

#endif /* __LINKER_ELF32_H */

