/*! \file linker_elf32_machdep_sparc.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker_elf32_machdep_sparc.c,v 1.5 2001/02/23 19:02:10 stevek Exp $
 *
 * \brief   System-dependent linker functions for Elf-32 modules
 * \author  Steve Kiernan
 * \version $Revision: 1.5 $
 * \date    $Date: 2001/02/23 19:02:10 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_linker.h>

#include "linker_elf_common.h"
#include "linker_elf32.h"


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#define	R_SPARC_NONE		0	/* relocation type */
#define	R_SPARC_8		1
#define	R_SPARC_16		2
#define	R_SPARC_32		3
#define	R_SPARC_DISP8		4
#define	R_SPARC_DISP16		5
#define	R_SPARC_DISP32		6
#define	R_SPARC_WDISP30		7
#define	R_SPARC_WDISP22		8
#define	R_SPARC_HI22		9
#define	R_SPARC_22		10
#define	R_SPARC_13		11
#define	R_SPARC_LO10		12
#define	R_SPARC_GOT10		13
#define	R_SPARC_GOT13		14
#define	R_SPARC_GOT22		15
#define	R_SPARC_PC10		16
#define	R_SPARC_PC22		17
#define	R_SPARC_WPLT30		18
#define	R_SPARC_COPY		19
#define	R_SPARC_GLOB_DAT	20
#define	R_SPARC_JMP_SLOT	21
#define	R_SPARC_RELATIVE	22
#define	R_SPARC_UA32		23
#define	R_SPARC_PLT32		24
#define	R_SPARC_HIPLT22		25
#define	R_SPARC_LOPLT10		26
#define	R_SPARC_PCPLT32		27
#define	R_SPARC_PCPLT22		28
#define	R_SPARC_PCPLT10		29
#define	R_SPARC_10		30
#define	R_SPARC_11		31
#define	R_SPARC_64		32
#define	R_SPARC_OLO10		33
#define	R_SPARC_HH22		34
#define	R_SPARC_HM10		35
#define	R_SPARC_LM22		36
#define	R_SPARC_PC_HH22		37
#define	R_SPARC_PC_HM10		38
#define	R_SPARC_PC_LM22		39
#define	R_SPARC_WDISP16		40
#define	R_SPARC_WDISP19		41
#define	R_SPARC_GLOB_JMP	42
#define	R_SPARC_7		43
#define	R_SPARC_5		44
#define	R_SPARC_6		45
#define	R_SPARC_NUM		46


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int
_linker_elf32_machdep_supported( wss_linker_file_t *lf )
{
   linker_elf32_data_t *elfdata = (linker_elf32_data_t *)lf->linker_file_data;

   if( elfdata->ehdr->e_machine != EM_SPARC )
      return ENOEXEC;

   return 0;
}


/*  wsi_elf_relocate_section */
/*! \brief Relocate symbols within ELF-32 section 
 * 
 *  \param lf      Pointer to linker file structure
 *  \param secaddr Pointer to section data
 *  \param type    Type of relocation (with or without addend)
 *  \param r       Pointer to relocation data for symbol
 *
 *  \return 0 on success,
 *  \return EFAULT if a necessary opcode linkage is unimplemented
 */
int
_linker_elf32_machdep_relocate( wss_linker_file_t *lf, void *secaddr,
                                int type, Elf32_Rel *r )
{
   linker_elf32_data_t *elfdata;
   u_int32_t            soffset = r->r_offset;
   u_int32_t            snum = ELF32_R_SYM(r->r_info);
   u_char               stype = ELF32_R_TYPE(r->r_info);
   Elf32_Sym           *sym;
   u_int32_t            addend;
   u_int32_t            value;
   u_int32_t            setvalue = 0;
   u_int32_t            storaddr = (u_int32_t)((char*)secaddr + soffset);

   elfdata = (linker_elf32_data_t *)lf->linker_file_data;

   if( type == SHT_RELA )
      addend = (u_int32_t)((Elf32_Rela *)r)->r_addend;
   else if( type == SHT_REL )
      addend = *((u_int32_t *)storaddr);
   else
      return EINVAL;
   sym = (Elf32_Sym *)( lf->linker_file_image.addr32 +
                        elfdata->symtab->sh_offset ) + snum;
   value = (u_int32_t)sym->st_value + addend;

#ifdef DEBUG_ELF32_LINKER
   wsi_printf( PD_MAKESTR("R[%d](0x%x + 0x%x) = 0x%x + 0x%x (0x%x)\n"), snum,
               (u_int)secaddr, (u_int)soffset, (u_int)sym->st_value,
               (u_int)addend, (u_int)value );
#endif /* DEBUG_ELF32_LINKER */
   switch( stype )
   {
      case R_SPARC_HI22:
      case R_SPARC_WDISP22:
         setvalue = (value >> 10);
         break;

      case R_SPARC_LO10:
         setvalue = (value & 0x00003ff);
         break;

      case R_SPARC_WDISP30:
         setvalue = ((value - (u_int32_t)secaddr - soffset) >> 2);
         break;

      case R_SPARC_32:
         setvalue = value;
         break;

      default:
         wsi_printf( PD_MAKESTR("ELF-32 SPARC Linker: Unimplemented opcode "
                                 "linkage: %d @ 0x%x + 0x%x [0x%x#%d]\n"),
                     stype, (u_int)secaddr, (u_int)soffset, setvalue, snum );

         return EFAULT;
   }

   *((u_int32_t *)storaddr) |= setvalue;

   return 0;
}

