#!/usr/local/bin/perl -w
# 
# Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
# All rights reserved.
#
# This file is part of the NAI Generic Software Wrappers Toolkit
# (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
# distribution.
#
# The GSWTK is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The GSWTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: genstubs.pl,v 1.74 2001/04/05 18:40:36 stevek Exp $
#
# Perl 5 script for generating characterization file skeleton and
# LWM stub functions for interception from Windows NT header file(s).
#

require 5.003;

use Config;
use strict;
use Getopt::Long;


###########################
# Configuration variables #
###########################
my $follow_redirect = 0;        # Set to 1 to include redirected DLLs
my $show_proto_add = 0;         # Set to 1 to show when prototypes are added
my $show_type_add = 0;          # Set to 1 to show when types are added
my $show_processed_files = 0;   # Set to 1 to show files as they are processed
my $search_include_dirs = 0;    # Set to 1 to search the include directories
my $show_missing_argtype = 0;   # Set to 1 to ignore missing argtypes

my $hooks_per_table = 500;      # Set to the number of hooks to be generated
                                #   per hook table

my $header = 'WINDOWS.H';       # Set to default header to parse


###############
# Directories #
###############
my $sysdir = $ENV{'systemroot'} || die "Set SYSTEMROOT variable in environment";
my $incdir = $ENV{'include'} || die "Set INCLUDE variable in environment";
$sysdir .= "\\system32";


#############################
# Determine Windows version #
#############################
my $winvers;

open( WINVER, 'ver |' ) || die "Unable to determine OS version";
while( <WINVER> ) {
   /^.*Version ([0-9]+)\.[0-9]+.*$/ || next;
   $winvers = $1;
   last;
}
close( WINVER );

die "Unable to determine OS version" if !$winvers;


###########################
# Print usage information #
###########################

sub doHelp {
   print STDERR <<EOF;
Usage: $0 [Flags] [DLLs]

   Flags:
       -I, --include    Path of include directories (semi-colon delimited)
       -H, --header     Header file to parse (if --search is not specified)
       -r, --redirect   Follow symbols through redirected DLLs
       -f, --showfiles  Show files as they are processed
       -p, --showproto  Show when prototypes are added
       -t, --showtype   Show when types are added
       -s, --search     Search the header files in the include directories
                          for function prototypes
       -a, --noargtype  Show missing argtypes in characterization 
       -v, --winvers    Set version of Windows to generate stubs for
                          (currently 4 and 5 are supported)

   DLLs:
       List of DLLs to characterize and generate stubs for (defauts to
       kernel32.dll, advapi32.dll if none are specified)

EOF
   exit 0;
}


##############################
# Process command-line flags #
##############################
my $outfile;

GetOptions('redirect!', \$follow_redirect,
           'r!', \$follow_redirect,
           'showfiles!', \$show_processed_files,
           'f!', \$show_processed_files,
           'showproto!', \$show_proto_add,
           'p!', \$show_proto_add,
           'showtype!', \$show_type_add,
           't!', \$show_type_add,
           'search!', \$search_include_dirs,
           's!', \$search_include_dirs,
           'o=s', \$outfile,
           'include=s', \$incdir,
           'I=s', \$incdir,
           'header=s', \$header,
           'H=s', \&header,
           'winvers=i', \$winvers,
           'v=i', \$winvers,
           'a!', \$show_missing_argtype,
           'noargtype!', \$show_missing_argtype,
           'help', \&doHelp,
           'h', \&doHelp);

for( $winvers ) {
   if( /^4$/ ) {
      last;
   }
   if( /^5$/ ) {
      last;
   }
   die "Version $winvers of Windows not supported by the Toolkit.";
}

print "[Generating Windows ${winvers}.x stubs and characterization]\n";

if( $outfile ) {
   open( OUTFILE, ">$outfile" ) || die "$outfile: $!";
   select( OUTFILE );
   $| = 1;
}


#############################################
# DLLs to search for symbols                #
#   (default to kernel32.dll, advapi32.dll) #
#############################################
my @dlls = qw( kernel32.dll advapi32.dll );
@dlls = @ARGV if @ARGV;


##################
# Various locals #
##################
my @errors;
my @symtabs;
my @prototypes;
my $dll;
my $nm = $Config{'nm'} . ' ' . $Config{'nm_opt'};


########################
# Flags for prototypes #
########################
my $PROTO_RETTYPE = 0x01;
my $PROTO_OPTYPE  = 0x02;
my $PROTO_ARGTYPE = 0x04;
my $PROTO_MASK    = ($PROTO_RETTYPE | $PROTO_OPTYPE | $PROTO_ARGTYPE);


#########################
# Utilitity subroutines #
#########################
sub contained {
   my $mydlls = shift;
   my $entry = shift;

   foreach( @$mydlls ) { return 1 if $_ eq $entry; }
   0;
}

sub longNameToShortName {
   my @long = split( /\\/, shift );
   my $short = shift @long;

   my $part;
   foreach $part ( @long )
   {
      my $cmd = $ENV{'comspec'} . ' /C DIR /X "' . $short .
                (($short=~/:$/)?"\\":"") . '" |';
      open( LTOS, $cmd ) ||
         die "Unable to convert long name to short";
      while( <LTOS> ) {
         /^[0-9\/]{8,10}\s+[0-9:ap]{6}\s{6,8}(<DIR>\s{9}|[0-9,\s]{14})\s(\S*)\s+(\S.*)$/;
         next if !$3;
         next if( ( lc($3) ne lc($part) ) && ( lc($2) ne lc($part) ) );
         $short .= "\\" . (($2)?$2:$3);
         last;
      }
      close( LTOS );
   }

   $short;
}


###################
# Lexer variables #
###################
my @tokens;
my $delimiters = '\{\}\(\)\-\=\+\[\]\;\<\>\,\.\/\?\:\^\&\|\*\%\!';
my $skipping_header = 0;
my $incomment = 0;
my $lexfile;
my @processed_headers;
my @local_processed_headers;
my %translation;


#####################
# Lexer subroutines #
#####################
sub tokenize {
   my $line = shift;
   my $tok;

   # Process #line
   if( $line =~ /^\s*#line\s+[0-9]+\s+"(.*)"\s*$/ ) {
      $skipping_header = 0;
      $lexfile = lc($1);
      $lexfile =~ s/\\+/\\/g;
      $translation{$lexfile} = longNameToShortName( $lexfile )
         if !exists $translation{$lexfile};
      $lexfile = $translation{$lexfile};

      $skipping_header = 1, return if contained( \@processed_headers, $lexfile );
      if( !contained( \@local_processed_headers, $lexfile ) ) {
         print ">> Processing $lexfile\n" if( $show_processed_files );
         push @local_processed_headers, $lexfile;
      }
      return;
   }

   # Skip until next #line if we've seen this header block before
   return if $skipping_header;

   # Trim #.*
   return if $line =~ /^\s*#.*$/;

   # Trim in-line comments
   $line =~ s/\/\*.*\*\///g;

   # Trim C++ comments
   $line =~ s/\/\/.*$//;

   # Handle C-style comments
   if( !$incomment && $line =~ /^(.*)\/\*/ ) {
      $incomment = 1;
      $line = $1;
   }

   if( $incomment && $line =~ /\*\/(.*)$/ ) {
      $incomment = 0;
      $line = $1;
   }
   return if( $incomment );

   # Whitespace-pad delimiters
   $line =~ s/([$delimiters])/ $1 /g;

   # Compact whitespace
   $line =~ s/\n\r//g;
   $line =~ s/\s+/ /g;
   $line =~ s/^\s|\s$//;

   # Skip all-whitespace lines
   return if $line !~ /\S/;

   # Split line into tokens
   @tokens = split(/\s/, $line);
}

sub nextToken {
   my $descriptor = shift;

   while( !@tokens ) {
      my $line = <$descriptor>;
      return undef if( eof($descriptor) );

      tokenize( $line );
   }

   my $token = shift @tokens;
   $token;
}

sub openLexer {
   my $file = shift;
   local *descriptor;

   my $preproc = $Config{'cpp'} . ' -D_WIN32_WINNT=0x0' . $winvers . '00 "' .
                 $file . '" > genstubs-out.' . $$;
   `$preproc`;
   open( *descriptor, "genstubs-out.$$" ) || die "$file: $!";

   *descriptor;
}

sub closeLexer {
   my $descriptor = shift;

   close $descriptor;

   unlink "genstubs-out.$$";
}


#############################
# Extract symbols from DLLs #
#############################
foreach $dll ( @dlls ) {
   print "Searching '" . $dll . "'...\n";

   my $hdr_done = 0;
   my %syms;
   if( open( NM, "$nm $sysdir\\$dll |" ) ) {
      while( <NM> ) {
         $hdr_done = 1, next if /^\s+ordinal\s+hint\s+RVA\s+name\s*$/;
         last                if /^\s*Summary/;
         next                unless $hdr_done;
         next                unless /\S/;

         /^\s+((([0-9]*)\s+([0-9A-F]*)\s+([0-9A-F]*)\s+(\S+)\s*)|(.*\(forwarded to (\S+)\.(\S+).*\)\s*))$/;

         if( $follow_redirect ) {
            push( @dlls, $8 . ".dll" ), next
               if( $8 && !contained( \@dlls, $8 . ".dll" ) );
         }

         # skip redirects
         next if $7;

         #             ordinal, header, prototype
         $syms{$6} = [ $3,      undef,  undef ];
      }
      close( NM );
   }
   my $mykeys = keys %syms;
   push @symtabs, [ $dll, $mykeys, \%syms ]; 
}


###########################################
# Print out number of symbols per library #
###########################################
print "Search done.\n";
foreach( @symtabs ) {
   print "DLL: " . $_->[0] . " has " . $_->[1] . " exported symbols\n";
}


####################
# Parser variables #
####################
my @eatsemi    = qw( extern );
my @eatbrace   = qw( inline _inline __inline );

my $eatsemimatch = join('|', map "^$_\$", @eatsemi);
my $eatbracematch = join('|', map "^$_\$", @eatbrace);

my $types;

my $transNoArgName = {
   'ABORTPROC'               => 'ABORTPROC abortProc',
   'BOOL'                    => 'BOOL bVal',
   'COLORREF'                => 'COLORREF colorRef',
   'DRAWSTATEPROC'           => 'DRAWSTATEPROC drawStateProc',
   'DWORD'                   => 'DWORD dwVal',
   'ENHMFENUMPROC'           => 'ENHMFENUMPROC enhmfeNumProc',
   'FLOAT'                   => 'FLOAT fVal',
   'FONTENUMPROCA'           => 'FONTENUMPROCA fontEnumProcA',
   'FONTENUMPROCW'           => 'FONTENUMPROCW fontEnumProcW',
   'GOBJENUMPROC'            => 'GOBJENUMPROC gobjEnumProc',
   'HANDLE'                  => 'HANDLE hVal',
   'HBITMAP'                 => 'HBITMAP hBitmap',
   'HBRUSH'                  => 'HBRUSH hBrush',
   'HCOLORSPACE'             => 'HCOLORSPACE hColorSpace',
   'HCURSOR'                 => 'HCURSOR hCursor',
   'HDC'                     => 'HDC hDC',
   'HENHMETAFILE'            => 'HENHMETAFILE hEnhMetaFile',
   'HGDIOBJ'                 => 'HGDIOBJ hGdiObj',
   'HGLOBAL'                 => 'HGLOBAL hGlobal',
   'HINSTANCE'               => 'HINSTANCE hInstance',
   'HMENU'                   => 'HMENU hMenu',
   'HMETAFILE'               => 'HMETAFILE hMetaFile',
   'HPALETTE'                => 'HPALETTE hPalette',
   'HRGN'                    => 'HRGN hRgn',
   'HWND'                    => 'HWND hWnd',
   'ICMENUMPROCA'            => 'ICMENUMPROCW icmEnumProcA',
   'ICMENUMPROCW'            => 'ICMENUMPROCW icmEnumProcW',
   'int'                     => 'int iVal',
   'INT'                     => 'int iVal',
   'LINEDDAPROC'             => 'LINEDDAPROC lineDdaProc',
   'LONG'                    => 'LONG lVal',
   'LPABC'                   => 'LPABC lpAbc',
   'LPABCFLOAT'              => 'LPABCFLOAT lpAbcFloat',
   'LPACCEL'                 => 'LPACCEL lpAccel',
   'LPARAM'                  => 'LPARAM lParam',
   'LPBITMAPINFO'            => 'LPBITMAPINFO lpBitmapInfo',
   'LPBYTE'                  => 'LPBYTE lpByte',
   'LPCMENUITEMINFOA'        => 'LPCMENUITEMINFOA lpcMenuItemInfoA',
   'LPCMENUITEMINFOW'        => 'LPCMENUITEMINFOW lpcMenuItemInfoW',
   'LPCOLORADJUSTMENT'       => 'LPCOLORADJUSTMENT lpColorAdjustment',
   'LPCSCROLLINFO'           => 'LPCSCROLLINFO lpcScrollInfo',
   'LPCSTR'                  => 'LPCSTR lpcStr',
   'LPCWSTR'                 => 'LPCWSTR lpcWStr',
   'LPDRAWTEXTPARAMS'        => 'LPDRAWTEXTPARAMS lpDrawTextParams',
   'LPDWORD'                 => 'LPDWORD lpDWord',
   'LPENHMETAHEADER'         => 'LPENHMETAHEADER lpEnhMetaHeader',
   'LPGCP_RESULTSA'          => 'LPGCP_RESULTSA lpGcpResultsA',
   'LPGCP_RESULTSW'          => 'LPGCP_RESULTSW lpGcpResultsW',
   'LPGLYPHMETRICS'          => 'LPGLYPHMETRICS lpGlyphMetrics',
   'LPHANDLETABLE'           => 'LPHANDLETABLE lpHandleTable',
   'LPINT'                   => 'LPINT lpInt',
   'LPKERNINGPAIR'           => 'LPKERNINGPAIR lpKerningPair',
   'LPLOGCOLORSPACEA'        => 'LPLOGCOLORSPACEA lpLogColorSpaceA',
   'LPLOGCOLORSPACEW'        => 'LPLOGCOLORSPACEW lpLogColorSpaceW',
   'LPLOGFONTA'              => 'LPLOGFONTA lpLogFontA',
   'LPLOGFONTW'              => 'LPLOGFONTW lpLogFontW',
   'LPMENUITEMINFOA'         => 'LPMENUITEMINFOA lpMenuItemInfoA',
   'LPMENUITEMINFOW'         => 'LPMENUITEMINFOW lpMenuItemInfoW',
   'LPMETARECORD'            => 'LPMETARECORD lpMetaRecord',
   'LPOUTLINETEXTMETRICA'    => 'LPOUTLINETEXTMETRICA lpOutlineTextMetricA',
   'LPOUTLINETEXTMETRICW'    => 'LPOUTLINETEXTMETRICW lpOutlineTextMetricW',
   'LPPALETTEENTRY'          => 'LPPALETTEENTRY lpPaletteEntry',
   'LPPIXELFORMATDESCRIPTOR' => 'LPPIXELFORMATDESCRIPTOR lpPixelFormatDescriptor',
   'LPPOINT'                 => 'LPPOINT lpPoint',
   'LPRASTERIZER_STATUS'     => 'LPRASTERIZER_STATUS lpRasterizerStatus',
   'LPRECT'                  => 'LPRECT lpRect',
   'LPRGNDATA'               => 'LPRGNDATA lpRgnData',
   'LPSCROLLINFO'            => 'LPSCROLLINFO lpScrollInfo',
   'LPSIZE'                  => 'LPSIZE lpSize',
   'LPSTR'                   => 'LPSTR lpStr',
   'LPTEXTMETRICA'           => 'LPTEXTMETRICA lpTextMetricA',
   'LPTEXTMETRICW'           => 'LPTEXTMETRICW lpTextMetricW',
   'LPTPMPARAMS'             => 'LPTPMPARAMS lptPMParams',
   'LPMSGBOXPARAMSA'         => 'LPMSGBOXPARAMSA lpMsgBoxParamsA',
   'LPMSGBOXPARAMSW'         => 'LPMSGBOXPARAMSW lpMsgBoxParamsW',
   'LPVOID'                  => 'LPVOID lpVoid',
   'LPWNDCLASSEXA'           => 'LPWNDCLASSEXA lpWndClassExA',
   'LPWNDCLASSEXW'           => 'LPWNDCLASSEXW lpWndClassExW',
   'LPWSTR'                  => 'LPWSTR lpWStr',
   'LPXFORM'                 => 'LPXFORM lpXForm',
   'MFENUMPROC'              => 'MFENUMPROC mfEnumProc',
   'PFLOAT'                  => 'PFLOAT pFloat',
   'POINT'                   => 'POINT pnt',
   'UINT'                    => 'UINT uiVal',
   'WPARAM'                  => 'WPARAM wParam'
};

my @pointerStructs = qw(
   COORD
   LARGE_INTEGER
   LATENCY_TIME
);
my %pointerStructs;
map { $pointerStructs{$_} = undef } @pointerStructs;

my $uaptype = 'LPVOID';



#################################
# Search headers for prototypes #
#################################
print "Processing prototypes...\n";
foreach $MAIN::dir ( split( ';', $incdir ) ) {
   opendir( DIR, $MAIN::dir ) || die "$MAIN::dir: $!";

   sub processPrototypes
   {
      my( $dirent ) = @_;

      return if( lc($dirent) !~ /\.h\s*$/ );

      my $headername = lc( $MAIN::dir . '\\' . $dirent );
      $headername =~ s/\\+/\\/g;

      $translation{$headername} = longNameToShortName($headername)
         if !exists $translation{$headername};
      $headername = $translation{$headername};

      return if( ! -f $headername ||
                 contained( \@processed_headers, $headername ) );

      # Reset counters, we're in a new header
      $incomment = 0;
      $skipping_header = 0;

      # Hash for types
      $types = { 'int'      => 1,
                 'long'     => 1,
                 'short'    => 1,
                 'char'     => 1,
                 'float'    => 1,
                 'double'   => 1,
                 'signed'   => 1,
                 'unsigned' => 1,
                 'const'    => 1,
                 'register' => 1,
                 'volatile' => 1,
                 'void'     => 1 };

      sub addType {
         my $type = shift;

         ( print "Warning: typedef for $type defined multiple times\n" )
            if( $show_type_add && exists $types->{$type} );

         $types->{$type} = 1;
         ( print "Added new type '$type'\n" ) if( $show_type_add );
      }

      print "Searching '$headername'...\n";
      my $d = openLexer( $headername );

      PROCESS: while(1) {
         my $tok = nextToken($d);
         last if eof($d);

         my $depth = 0;
         if( $tok eq 'typedef' ) {
            my $last_tok;

            my $inbrace;
            my $parens;
            my $parensets;
            TYPEDEF: while( !eof($d) ) {
               $tok = nextToken($d); last PROCESS if( eof($d) );
               if( $tok eq '{' ) { $depth++ if( !$depth++ ); }
               $depth-- if $tok eq '}';
               last TYPEDEF if( ( !$depth && $tok eq ';' ) || $depth == 1 );
               $parens++ if $tok eq '(';
               $parens--, $parensets++ if $tok eq ')';
               $last_tok = $tok if $parens && !$parensets;
               next TYPEDEF if $parens;
               $inbrace++ if $tok eq '[';
               $inbrace-- if $tok eq ']';
               next TYPEDEF if $inbrace;
               next TYPEDEF if !$depth && $tok =~ /[$delimiters]/;

               $last_tok = $tok;
            }

            if( $tok eq ';' )
            {
               addType( $last_tok );
            }
            else
            {
               TYPENAMES: while( !eof($d) ) {
                  $tok = nextToken($d); last PROCESS if( eof($d) );
                  last TYPENAMES if $tok eq ';';
                  $inbrace++ if $tok eq '[';
                  $inbrace-- if $tok eq ']';
                  next TYPENAMES if $inbrace;
                  next TYPENAMES if $tok =~ /[$delimiters]/;
                  addType( $tok );
               }
            }
            next;
         }

         if( $tok =~ /$eatsemimatch/ ) {
            EATSEMI: while( !eof($d) ) {
               $tok = nextToken($d); last if( eof($d) );
               $depth++ if( $tok eq '{' );
               $depth-- if( $tok eq '}' );
               last EATSEMI if( $tok eq ';' && !$depth );
            }
            next;
         }

         if( $tok =~ /$eatbracematch/ ) {
            EATBRACE: while( !eof($d) ) {
               $tok = nextToken($d); last PROCESS if( eof($d) );
               if( $tok eq '{' ) { $depth++ if( !$depth++ ); }
               $depth-- if( $tok eq '}' );
               last EATBRACE if( $depth == 1 );
            }
            next;
         }

         # Look for exported function declarations
         if( $tok eq '__declspec' ) {
            $tok = nextToken($d); last PROCESS if( eof($d) );

            if( $tok eq '(' ) {
               $tok = nextToken($d); last PROCESS if( eof($d) );
               if( $tok eq 'dllimport' ) {
                  $tok = nextToken($d); last PROCESS if( eof($d) );

                  my $rettype = '';
                  my $declstate = 0;
                  while( !eof($d) ) {
                     $tok = nextToken($d); last PROCESS if( eof($d) );
                     last if $tok eq '__stdcall';
                     if( $tok eq '__declspec' ) {
                        # Eat the next three tokens
                        $tok = nextToken($d); last PROCESS if( eof($d) );
                        $tok = nextToken($d); last PROCESS if( eof($d) );
                        $tok = nextToken($d); last PROCESS if( eof($d) );
                        next;
                     }
                     $rettype .= ' ' if length($rettype);
                     $rettype .= $tok;
                  }

                  my $fname = nextToken($d); last PROCESS if( eof($d) );
                  $tok = nextToken($d); last PROCESS if( eof($d) );

#                  print ">>> $fname\n";

                  if( $tok eq '(' ) {
                     my @funcargs;
                     my $farg = '';
                     my $narg = 0;
                     while( !eof($d) ) {
                        $tok = nextToken($d); last PROCESS if( eof($d) );
                        last if $tok eq ')';
                        if( $tok eq ',' ) {
                           $narg++;
                           if( length($farg) && $farg ne 'void' ) {
                              $farg = $transNoArgName->{$farg} . $narg
                                 if exists $transNoArgName->{$farg};
                              push @funcargs, $farg;
                           }
                           $farg = '';
                           next;
                        }
                        $farg .= ' ' if length($farg);
                        $farg .= $tok;
                     }

                     $narg++;
                     if( length($farg) && $farg ne 'void' ) {
                        $farg = $transNoArgName->{$farg} . $narg
                           if exists $transNoArgName->{$farg};
                        push @funcargs, $farg;
                     }

                     if( $tok eq ')' ) {
                        # Make sure it's one we care about
                        my $symtab;
                        foreach $symtab ( @symtabs ) {
                           if( exists $symtab->[2]->{$fname} ) {
                              last if( defined $symtab->[2]->{$fname}->[1] );

                              # Add the prototype
                              # [ name, return type, array of function args,
                              #   name of DLL, flags, ordinal ]
                              my $proto = [ $fname, $rettype, \@funcargs,
                                            $symtab->[0], 0,
                                            $symtab->[2]->{$fname}->[0] ];
                              push @prototypes, $proto;

                              $symtab->[2]->{$fname}->[1] = $lexfile;
                              $symtab->[2]->{$fname}->[2] = $proto;

                              ( print "Added prototype for $fname\n" )
                                 if( $show_proto_add );
                              last;
                           }
                        }
                     }
                  }
               }
            }

            # Eat 'till semicolon
            while( !eof($d) ) {
               $tok = nextToken($d); last PROCESS if( eof($d) );
               last if( $tok eq ';' );
            }

            next;
         }
      }

      closeLexer($d);

      # Push headers processed by this pass onto global processed list
      foreach( @local_processed_headers ) {
         push @processed_headers, $_;
      }
      undef @local_processed_headers;
   }

   if( $search_include_dirs ) {
      my $dirent;
      while( $dirent = readdir( DIR ) ) {
         processPrototypes( $dirent );
      }
   }
   else {
      processPrototypes( $header );
   }

   closedir( DIR );
}
print "Processing done.\n";


##################################
# Generate characterization file #
##################################
print "Generating characterization file...\n";

#
# List of calls to ignore when generating stubs
#
my @czIgnoreCalls = qw(
   lstrcatA lstrcatW lstrcmpA lstrcmpW lstrcmpiA lstrcmpiW
   lstrcpyA lstrcpyW lstrcpynA lstrcpynW lstrlenA lstrlenW
   inet_ntoa
);

my %czIgnoreCalls;
map { $czIgnoreCalls{$_} = undef } @czIgnoreCalls;


#
# Map for functions which should have ops ignored
#
# Note: (See $czArgAttribs below)
#       The format for an attribute below which should be ignored is:
#
#          <attrib>:ignore <ignore list number>
#
my $czIgnoreOps = {
   '1' => [ 'FormatMessage', 'GetAtomName', 'GetComputerName',
            'GetEnvironmentVariable', 'GetLogicalDriveStrings',
            'GlobalGetAtomName' ],
   '2' => [ 'ReportEvent' ]
};


#
# Map for functions for which op-types can be determined by partial name
# matching
#
# Note: the following entries are regexp match strings and associated
#       op-types which are iteratively-matched against the function name
#
my @czRegexpMatchOps = (
   '^.*Ace$'                               => 'aclop',
   '^.*(Access|Privilege).*'               => 'aclop',
   '^.*Self$'                              => 'aclop',
   '^.*Atom.*'                             => 'atomop',
   '^.*Console.*'                          => 'consoleop',
   '^Crypt.*'                              => 'cryptop',
   '^.*Encrypt.*'                          => 'cryptop',
   '^.*(Date|Calendar).*'                  => 'dateop',
   '^.*Time(Formats|Zone.*|Adjustment)?$'  => 'dateop',
   '^.*Debug.*'                            => 'debugop',
   '^.*Environment.*'                      => 'envop',
   '^.*LastError.*'                        => 'errorop',
   '^.*Exception.*'                        => 'exop',
   '^.*Directory.*'                        => 'fileop',
   '^.*FileWith.*'                         => 'fileop',
   '^.*(Temp)?File(Ex|Mapping|Name|Apis.*)?$' => 'fileop',
   '^.*(Temp)?Path(Name)?$'                => 'fileop',
   '^.*Get(File|.*Drive).*'                => 'fileop',
   '^_[hl](read|write|close|open|creat|lseek).*' => 'fileop',
   '^RegisterEvent.*'                      => 'eventop',
   '^.*Performance.*'                      => 'hardwareop',
   '^.*DosDevice.*'                        => 'hardwareop',
   '^.*HwProfile$'                         => 'hardwareop',
   '^.*JobObject.*'                        => 'jobobjectop',
   '^Get.*CP$'                             => 'localeop',
   '^.*(LCID|LangID)$'                     => 'localeop',
   '^.*Mailslot.*'                         => 'mailslotop, ipcop',
   '^Tls(Alloc|Free)$'                     => 'memop, threadop',
   '^.*(Alloc|Free|Compact)(Ex)?$'         => 'memop',
   '^.*IsBad.*Ptr$'                        => 'memop',
   '^(Global|Heap|Virtual).*'              => 'memop',
   '^.*WriteWatch.*'                       => 'memop',
   '^.*PhysicalPages.*'                    => 'memop',
   '^.*NamedPipeHandleState.*'             => 'pipeop, ipcop',
   '^CreateNamedPipe.*'                    => 'pipeop, ipcop',
   '^.*Pipe.*'                             => 'pipeop, ipcop',
   '^.*Power.*'                            => 'powerop',
   '^.*SystemResume.*'                     => 'powerop',
   '^.*Wakeup.*'                           => 'powerop',
   '^.*ThreadExecutionState.*'             => 'powerop',
   '^CreateProcess.*'                      => 'processop, execop',
   '^.*Process.*'                          => 'processop',
   '^.*CommandLine$'                       => 'processop',
   '^.*Processor.*'                        => 'processorop',
   '^(Get|Write)(Private)?Profile.*'       => 'registryop',
   '^RegisterService.*'                    => 'serviceop',
   '^Reg.*'                                => 'registryop',
   '^.*Resource.*'                         => 'resourceop',
   '^WSA.*Service.*'                       => 'socketop, serviceop',
   '^.*(Service|SCManager).*'              => 'serviceop',
   '^.*(Sid|AccountName).*'                => 'sidop',
   '^get(host|proto|serv).*'               => 'socketop', # <--.
   '^[hn]to[nh][ls]$'                      => 'socketop', # <--+-- Really true?
   '^WSA.*'                                => 'socketop', # <--'
   '^Is(DBCS|Text).*'                      => 'stringop',
   '^.*Message$'                           => 'stringop',
   '^.*StringType(Ex)?$'                   => 'stringop',
   '^Interlocked.*'                        => 'syncop',
   '^.*(Mutex|Semaphore|CriticalSection|Event).*' => 'syncop:ignore 2',
   '^WaitFor.*Object.*'                    => 'syncop',
   '^.*APC.*'                              => 'syncop',
   '^.*ComputerName.*'                     => 'sysinfoop',
   '^.*UserName.*'                         => 'sysinfoop',
   '^.*SystemInfo.*'                       => 'sysinfoop',
   '^.*Version.*'                          => 'sysinfoop',
   '^.*(Fiber|Thread|WorkItem).*'          => 'threadop',
   '^Tls.*'                                => 'threadop',
   '^.*TimerQueue.*'                       => 'timerqueueop',
   '^.*(Timer|Tick.*)$'                    => 'timerop',
   '^.*Token.*'                            => 'tokenop',
   '^.*Volume.*'                           => 'volumeop'
);

# Coerce array form of hash into associative array
my( %czRegexpMatchOpsMap ) = @czRegexpMatchOps;
my $czRegexpMatchOpsMap = \%czRegexpMatchOpsMap;


#
# List of calls which map to native ops
#
my @czNativeOps = qw( 
   CreateMailSlot CreateNamedPipe CloseHandle DuplicateHandle RegCreateKey
   RegOpenKey RegDeleteKey RegDeleteValue RegEnumKey RegEnumKeyEx RegEnumValue
   RegFlushKey RegNotifyChangeKeyValue RegQueryKey RegQueryMultipleValues
   RegQueryValue RegQueryValueEx RegReplaceKey RegSaveKey RegRestoreKey
   RegSetValue RegLoadKey RegUnloadKey AccessCheck AccessCheckAuditAlarm
   AdjustTokenGroups AdjustTokenPrivileges ObjectCloseAuditAlarm CreateToken
   ObjectDeleteAuditAlarm DuplicateToken DuplicateTokenEx
   ImpersonateLoggedOnUser ObjectOpenAuditAlarm OpenProcessToken
   OpenThreadToken PrivilegeCheck ObjectPrivilegeAuditAlarm
   PrivilegedServiceAuditAlarm GetTokenInformation GetUserObjectSecurity
   GetPrivateObjectSecurity SetTokenInformation SetUserObjectSecurity
   SetrivateSecurityObject CreateProcess CreateThread Sleep SleepEx
   GetThreadContext SetThreadContext OpenProcess OpenThread GetProcessTimes
   GetProcessVersion GetProcessWorkingSetSize GetProcessPriorityBoost
   GetProcessAffinityMask GetPriorityClass GetProcessShutdownParameters
   GetThreadTimes GetThreadPriority GetThreadPriorityBoost QueueUserApc
   ResumeThread SetProcessAffinityMask SetPriorityClass SetProcessPriorityBoost
   SetProcessShutdownParameters SetProcessWorkingSetSize SetThreadAffinityMask
   SetThreadIdealProcessor SetThreadPriority SetThreadPriorityBoost
   SuspendThread TerminateProcess TerminateThread SwitchToThread AddAtom
   DeleteAtom FindAtom GetAtomName RaiseException SetErrorMode GetLocaleInfo
   SetLocaleInfo GetEnvironmentVariable SetEnvironmentVariable
   CancelWaitableTimer CreateWaitableTimer OpenWaitableTimer
   QueryPerformanceCounter QueryPerformanceFrequency GetSystemTime
   SetSystemTime GetTickCount CreateEvent OpenEvent PulseEvent ResetEvent
   SetEvent CreateMutex OpenMutex ReleaseMutex CreateSemaphore OpenSemaphore
   ReleaseSemaphore WaitForMultipleObjects WaitForMultipleObjectsEx
   WaitForSingleObject WaitForSingleObjectEx VirtualAlloc VirtualAllocEx
   VirtualFree VirtualFreeEx VirtualQuery VirtualQueryEx VirtualProtect
   VirtualProtectEx VirtualLock VirtualUnlock ReadProcessMemory
   WriteProcessMemory FlushViewOfFile CreateFileMapping OpenFileMapping
   MapViewOfFile UnmapViewOfFile CancelIo CreateFile CreateDirectory
   CreateDirectoryEx CreateIoCompletionPort DeleteFile DeviceIoControl
   FlushFileBuffers DeviceIoControl LockFile LockFileEx UnlockFile
   FindFirstChangeNotification FindNextChangeNotification OpenFile
   GetFileAttributesEx FindFirstFile FindFirstFileEx FindNextFile
   GetShortPathName GetLongPathName GetFullPathName GetFileType GetFileSize
   GetFileTime SetEndOfFile SetFileAttributes SetNamedPipeHandleState
   SetMailslotInfo GetDiskFreeSpace GetDriveType SetVolumeLabel ReadFile
   ReadFileEx WriteFile WriteFileEx ReadFileScatter WriteFileGather
   AllocateLocallyUniqueId ExitWindows
);

my %czNativeOps;
map { $czNativeOps{$_} = undef } @czNativeOps;

#
# Map for characterizing function op-types
#
my $czPrimitives = {
   'LPCWSTR' =>
   {
      'lpPath'                => 'searchop',
      'SubsystemName'         => 'aclop',
      'lpFileName'            => 'fileop',
      'lpPathName'            => 'fileop',
      'lpRootPathName'        => 'fileop',
      'lpDirectoryName'       => 'fileop',
      'lpApplicationName'     => 'fileop',
      'lpUNCServerName'       => 'fileop',
      'lpLibFileName'         => 'moduleop',
      'lpModuleName'          => 'moduleop',
      'lpComputerName'        => 'systemop',
      'lpDef'                 => 'dcbop',
      'lpszName'              => 'commconfigop'
   },
   'LPWSTR' =>
   {
      'lpMachineName'         => 'systemop',
      'lpszUsername'          => 'tokenop'
   },
   'LPCSTR' =>
   {
      'lpModuleName'          => 'moduleop',
      'lpCmdLine'             => 'processop'
   },
   'HANDLE' =>
   {
      'hFile'                 => 'fileop',
      'hProcess'              => 'processop',
      'hFileMappingObject'    => 'memop',
      'hThread'               => 'threadop',
      'hHeap'                 => 'memop',
      'hEvent'                => 'syncop, ipcop',
      'hEventLog'             => 'eventlogop',
      'hChangeHandle'         => 'fileop',
      '*'                     => 'handleop'
   },
   'DWORD' =>
   {
      'dwExitCode'            => 'processop',
      'GrantedAccess'         => 'aclop',
      'dwMapFlags'            => 'stringop',
      'nStdHandle'            => 'fileop',
      'dwFreq'                => 'hardwareop',
      'dwMilliseconds'        => 'processop'
   },
   'PDWORD' =>
   {
      'AccessMask'            => 'aclop'
   },
   'UINT' =>
   {
      'uExitCode'             => 'processop',
      'CodePage'              => 'localeop',
      'uNumber'               => 'processop',
      'uAction'               => 'processop',
      'uMode'                 => 'systemop'
   },
   'HKEY' =>
   {
      'hKey'                  => 'registryop, handleop'
   },
   'HWND' =>
   {
      'hWnd'                  => 'windowop'
   },
   'HGLOBAL' =>
   {
      'hMem'                  => 'memop, handleop'
   },
   'HLOCAL' =>
   {
      'hMem'                  => 'memop, handleop'
   },
   'SOCKET' =>
   {
      's'                     => 'fdop'
   },
   'PSECURITY_DESCRIPTOR' =>
   {
      'pSecurityDescriptor'   => 'aclop',
      'SecurityDescriptor'    => 'aclop',
      'ParentDescriptor'      => 'aclop',
      'ObjectDescriptor'      => 'aclop',
      'pSelfRelativeSecurityDescriptor' => 'aclop',
      'pAbsoluteSecurityDescriptor' => 'aclop'
   },
   'PSECURITY_DESCRIPTOR *' =>
   {
      'ObjectDescriptor'      => 'aclop'
   },
   'SECURITY_INFORMATION' =>
   {
      'SecurityInformation'   => 'aclop'
   },
   'LPSTARTUPINFOW' =>
   {
      'lpStartupInfo'         => 'processop'
   },
   'PACL' =>
   {
      'pAcl'                  => 'aclop'
   },
   'HCRYPTPROV' =>
   {
      'hProv'                 => 'handleop'
   },
   'HCRYPTKEY' =>
   {
      'hKey'                  => 'handleop'
   },
   'HCRYPTHASH' =>
   {
      'hHash'                 => 'handleop'
   },
   'SC_HANDLE' =>
   {
      'hSCObject'             => 'serviceop, handleop',
      'hService'              => 'serviceop, handleop',
      'hSCManager'            => 'serviceop, handleop'
   },
   'LCID' =>
   {
      'Locale'                => 'localeop'
   },
   'LPSECURITY_ATTRIBUTES' =>
   {
      'lpMutexAttributes'     => 'syncop',
      'lpEventAttributes'     => 'syncop'
   },
   'int' =>
   {
      'af'                    => 'socketop',
      'nfds'                  => 'socketop',
      'ExitCode'              => 'processop'
   },
   'SOCKET' =>
   {
      '*'                     => 'socketop'
   },
   'const char *' =>
   {
      'cp'                    => 'socketop'
   },
   'HMODULE' =>
   {
      '*'                     => 'moduleop'
   },
   'LPCVOID' =>
   {
      'pMem'                  => 'memop'
   },
   'PVOID' =>
   {
      'pvContext'             => 'fileop, cryptop'
   },
   'CODEPAGE_ENUMPROCW' =>
   {
      'lpCodePageEnumProc'    => 'localeop'
   },
   'LOCALE_ENUMPROCW' =>
   {
      'lpLocaleEnumProc'      => 'localeop'
   },
   'BOOL' =>
   {
      'BootAcceptable'        => 'systemop, serviceop'
   },
   'PFE_EXPORT_FUNC' =>
   {
      'pfExportCallback'        => 'fileop, cryptop'
   },
   'PFE_IMPORT_FUNC' =>
   {
      'pfImportCallback'        => 'fileop, cryptop'
   }
};



#
# Map for characterizing function arguments
#
my $czArgAttribs = {
   'LPCSTR' =>
   {
      'lpExistingFileName'   => 'path',
      'lpFileName'           => 'path',
      'lpNewFileName'        => 'destpath',
      'lpPathName'           => 'path',
      'lpTemplateDirectory'  => 'path',
      '*'                    => 'nterm'
   },
   'LPCWSTR' =>
   {
      'lpExistingFileName'   => 'path',
      'lpFileName'           => 'path',
      'lpNewFileName'        => 'destpath',
      'lpPathName'           => 'path',
      'lpTemplateDirectory'  => 'path',
      'lpApplicationName'    => 'path',
      '*'                    => 'nterm'
   },
   'LPWSTR' =>
   {
      'lpBuffer'             => 'outpath, nterm:ignore 1'
   },
   'HANDLE' =>
   {
      '*'                    => 'handle, fd'
   },
   'HLOCAL' =>
   {
      '*'                    => 'memory, handle, fd'
   },
   'HGLOBAL' =>
   {
      '*'                    => 'memory, handle, fd'
   },
   'HWND' =>
   {
      '*'                    => 'window, handle, fd'
   },
   'LPDWORD' =>
   {
      '*'                    => 'out'
   },
   'LPOVERLAPPED' =>
   {
      '*'                    => 'inout'
   },
   'LPVOID' =>
   {
      'lpFileInformation'    => 'out',
      'lpInBuffer'           => 'out',
      'lpStruct'             => 'out',
      'lpTapeInformation'    => 'out',
      'lpBuffer'             => 'out:copy'
   },
   'LPCVOID' =>
   {
      'lpBuffer'             => ':copy'
   },
   'LPVOID *' =>
   {
      'lpContext'            => 'inout'
   },
   'DWORD' =>
   {
      'nNumberOfBytesToRead' => 'nbytes',
      'nNumberOfBytesToWrite' => 'nbytes',
      'nNumberOfCharsToRead' => 'nbytes',
      'nBufferSize'          => 'nbytes',
      'nBufferLength'        => 'nbytes',
      'nSize'                => 'nbytes',
      'cbBufSize'            => 'nbytes'
   },
   'long' =>
   {
      'lBytes'               => 'nbytes'
   },
   'UINT' =>
   {
      'uBytes'               => 'nbytes'
   },
   'int' =>
   {
      'af'                   => 'domain',
      'type'                 => 'socktype',
      'len'                  => 'nbytes'
   },
   'fd_set *' =>
   {
      '*'                    => 'fdset'
   },
   'SOCKET' =>
   {
      '*'                    => 'socket, fd'
   },
   'struct sockaddr *' =>
   {
      '*'                    => 'sockadr'
   },
   'const struct sockaddr *' =>
   {
      '*'                    => 'sockadr'
   },
   'const char *' =>
   {
      'buf'                  => 'iobuf'
   },
   'HKEY' =>
   {
      'hKey'                 => 'handle'
   },
   'PACL' =>
   {
      'pAcl'                 => 'acl'
   },
   'HCRYPTPROV' =>
   {
      'hProv'                => 'cryptprovider, handle'
   },
   'HCRYPTKEY' =>
   {
      'hKey'                 => 'cryptkey, handle'
   },
   'HCRYPTHASH' =>
   {
      'hHash'                => 'crypthash, handle'
   },
   'SC_HANDLE' =>
   {
      '*'                    => 'handle'
   },
   'PSID' =>
   {
      '*'                    => 'sid'
   }
};

#
# Map for characterizing function return types
#
my $czTypeMap = {
   'HANDLE'                       => 'handleret, HANDLE_RET',
   'BOOL'                         => 'boolret, BOOL_RET',
   'VOID'                         => 'NO_RET',
   'void'                         => 'NO_RET',
   'long'                         => 'errret, ZERO_RET',
   'LONG'                         => 'errret, ZERO_RET',
   'u_long'                       => 'u_longret',
   'unsigned long'                => 'u_longret',
   'u_short'                      => 'u_shortret',
   'HMODULE'                      => 'moduleret, NONNULL_RET',
   'PVOID'                        => 'pointerret, NONNULL_RET',
   'LPVOID'                       => 'pointerret, NONNULL_RET',
   'UINT'                         => 'sizeret, NONZERO_RET',
   'HLOCAL'                       => 'localret, NONNULL_RET',
   'HGLOBAL'                      => 'globalret, NONNULL_RET',
   'int'                          => 'sizeret,VAL_RET',
   'INT'                          => 'errret, NONZERO_RET',
   'ATOM'                         => 'atomret, NONZERO_RET',
   'LCID'                         => 'localeret, NOERR_RET',
   'HRSRC'                        => 'resourceret, NONNULL_RET',
   'FARPROC'                      => 'pointerret, NONNULL_RET',
   'LANGID'                       => 'langret, NONZERO_RET',
   'LPTOP_LEVEL_EXCEPTION_FILTER' => 'NOERR_RET',
   'HFILE'                        => 'handleret, HANDLE_RET',
   'DWORD'                        => 'NOERR_RET',
   'LPSTR'                        => 'NOERR_RET',
   'LPWSTR'                       => 'NOERR_RET',
   'COORD'                        => 'NOERR_RET',
   'struct hostent *'             => 'hostentret',
   'struct protoent *'            => 'protoentret',
   'struct servent *'             => 'serventret',
   'SC_HANDLE'                    => 'sc_handleret, SC_HANDLE_RET',
   'PSID_IDENTIFIER_AUTHORITY'    => 'psid_identifier_authorityret, PSID_IDENTIFIER_AUTHORITY_RET',
   'PDWORD'                       => 'pdwordret',
   'PUCHAR'                       => 'pucharret',
   'SC_LOCK'                      => 'sc_lockret',
   'SERVICE_STATUS_HANDLE'        => 'service_status_handleret, SERVICE_STATUS_HANDLE_RET',
   'SOCKET'                       => 'fdret, VAL_RET',
   'WSAEVENT'                     => 'wsaeventret, WSAEVENT_RET'
};

open( CHARACTERIZE, '>../wr.include/winnt' . $winvers . '.ch' ) ||
   die 'winnt' . $winvers . ".ch: $!";
select( CHARACTERIZE );

print <<EOF;
/* 
 * \$Id\$
 */

#if 0
#define WR_KERNEL_MODULE
#include "wr_conf.h"
#endif

#define WR_WIN32 1

includes{
#ifndef _WR_DEFS_INCLUDED
#include "wr_conf.h"
#endif

#include <windows.h>
};

#include "wr_errname.h"

#define WR_LOCAL win32

characterize WR_LOCAL {

int errnum{
       errnum{
          vmap(WR_SUCC,         _),
          vmap(WR_FAIL,         _),
          vmap(WR_BADPERM,      ERROR_ACCESS_DENIED),    /* !!! no equivalent to super user */
          vmap(WR_NOFILE,       ERROR_FILE_NOT_FOUND),
          vmap(WR_NOPROC,       ERROR_NOT_FOUND),        /* !!! probably not correct, can't find exact complement to process not found */
          vmap(WR_IOERR,        ERROR_IO_DEVICE),        /* !!! seems to be best match */
          vmap(WR_ARG2LONG,     ERROR_BAD_LENGTH),       /* !!! probably not correct */
          vmap(WR_BADFD,        ERROR_INVALID_HANDLE),   /* !!! wider in scope than wrapper error */
          vmap(WR_NOCHILD,      ERROR_WAIT_NO_CHILDREN), /* !!! not really a good match */ 
          vmap(WR_BADACCES,     ERROR_ACCESS_DENIED),    /* !!! probably shouldn't be the same as WR_BADPERM */
          vmap(WR_FILEEXIST,    ERROR_FILE_EXISTS),
          vmap(WR_BADARG,       ERROR_BAD_ARGUMENTS),	 /* !!! may not be a good match, might be too specific on nt */
          vmap(WR_FILEBSY,      ERROR_LOCK_VIOLATION),   /* !!! might not be correct match */
          vmap(WR_FILE2BIG,     _),                      /* !!! can't find equivalent on NT */
          vmap(WR_NOSPC,        ERROR_DISK_FULL),
          vmap(WR_FNAME2LONG,   ERROR_BUFFER_OVERFLOW)	 /* !!! seems odd, but that's what the docs say */
       }
   };

/* 
 * syscalls that return 0 on success and -1 on failure
 */
#define STD_RET wrret{vmap(WR_SUCC,0), vmap(*,-1)}
/* 
 * syscalls that return some values on success and -1 on failure
 */
#define VAL_RET wrret{vmap(WR_SUCC,_), vmap(*,-1)}
/* 
 * syscalls that are always successful, and no return value is 
 * reserved to indicate an error. 
 */
#define NOERR_RET wrret
   
/*
 * calls that return handles
 */
					/* Too Much Knowledge encoded here */
#define HANDLE_RET wrret{vmap(WR_SUCC,_), vmap(*,-1)}

/*
 * calls that return a boolean
 */
#define BOOL_RET wrret{vmap(WR_SUCC,TRUE), vmap(*,FALSE)}

/* 
 * calls that return some values on success and 0 on failure
 */
#define NONZERO_RET wrret{vmap(WR_SUCC,_), vmap(*,0)}

/* 
 * calls that return zero on success and something else on failure
 */
#define ZERO_RET wrret{vmap(WR_SUCC,0), vmap(*,_)}

/* 
 * calls that return non-null on success and null on failure
 */
#define NONNULL_RET wrret{vmap(WR_SUCC,_), vmap(*,0)}

EOF

my $symtab;
my $nsyms;
foreach $symtab ( @symtabs ) {
   my $symbols = $symtab->[2];
   foreach ( sort keys %$symbols ) {
      if( $symbols->{$_}->[1] ) {
         my $sym = $_;
         /^(.*)W$/;
         my $base = $1;

         next if exists $czIgnoreCalls{$sym};
         next unless /W$/;
         next if $base eq 'GetEnvironmentStrings'; # Hack for MS hack
         next if $symbols->{$base . 'A'}->[1];

         if( $nsyms++ == 0 ) {
            print <<EOF;

/*
 * Macros for wide-format functions without ASCII-format complements
 */
EOF
         }
         print <<EOF;
#define $base $sym
EOF
      }
   }
}

print <<EOF;


/*
 * Function call characterizations
 */

EOF

my $proto;
foreach $proto ( sort { $a->[0] cmp $b->[0] } @prototypes ) {
   my( $fname, $rettype, $funcargs ) = @$proto;

   next if $fname =~ /A$/;
   next if $fname eq 'GetEnvironmentStrings'; # Hack for MS hack
   ( $proto->[4] |= $PROTO_MASK ), next if exists $czIgnoreCalls{$fname};

   print $rettype.' ';
   (print '{'.$czTypeMap->{$rettype}.'}'), $proto->[4] |= $PROTO_RETTYPE
      if exists $czTypeMap->{$rettype};
   print "\n";

   my $nbytesarg;
   foreach( @$funcargs ) {
      my $line = $_;
      /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
      my( $argtype, $argname, $arraytokens ) = ( $1, $3, $4 );

      if( exists $czArgAttribs->{$argtype} ) {
         my $cztype = $czArgAttribs->{$argtype};
         if( ( exists $cztype->{$argname} && $cztype->{$argname} eq 'nbytes' ) ||
             ( exists $cztype->{'*'} && $cztype->{'*'} eq 'nbytes' ) ) {
            $nbytesarg = $argname;
            last;
         }
      }
   }

   my $firstdone = 0;
   foreach( @$funcargs ) {
      my $line = $_;
      /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
      my( $argtype, $argname, $arraytokens ) = ( $1, $3, $4 );

      push( @errors, "Invalid type/name combination: $line" ), next if !$argtype;

      my $basename = $fname;
      $basename =~ s/W$//;

      if( $firstdone ) {
         print ',';
      }
      else {
         my $optype;

         my $regexp;
         foreach $regexp ( @czRegexpMatchOps ) {
            next unless $regexp =~ /^\^/;
            if( $basename =~ /$regexp/ ) {
               my $pendingoptype = $czRegexpMatchOpsMap->{$regexp};

               if( $pendingoptype =~ /^(.*):ignore (\S+)/ ) {
                  $pendingoptype = $1;
                  my $op;

                  foreach $op ( @{$czIgnoreOps->{$2}} ) {
                     undef $pendingoptype, last if( $op eq $basename );
                  }
               }

               if( $pendingoptype ) {
                  $optype .= ', ' if defined $optype && length($optype);
                  $optype .= $pendingoptype;
               }

               last;
            }
         }

         if( exists $czNativeOps{$basename} ) {
            $optype .= ', ' if defined $optype && length($optype);
            $optype .= 'nativeop';
         }

         if( exists $czPrimitives->{$argtype} ) {
            my $cztype = $czPrimitives->{$argtype};
            if( exists $cztype->{$argname} ) {
               my $pendingoptype = $cztype->{$argname};

               if( $pendingoptype =~ /^(.*):ignore (\S+)/ ) {
                  $pendingoptype = $1;
                  my $op;

                  foreach $op ( @{$czIgnoreOps->{$2}} ) {
                     undef $pendingoptype, last if( $op eq $basename );
                  }
               }

               if( $pendingoptype ) {
                  $optype .= ', ' if defined $optype && length($optype);
                  $optype .= $pendingoptype;
               }
            }
            if( exists $cztype->{'*'} ) {
               my $pendingoptype = $cztype->{'*'};

               if( $pendingoptype =~ /^(.*):ignore (\S+)/ ) {
                  $pendingoptype = $1;
                  my $op;

                  foreach $op ( @{$czIgnoreOps->{$2}} ) {
                     undef $pendingoptype, last if( $op eq $basename );
                  }
               }

               if( $pendingoptype ) {
                  $optype .= ', ' if defined $optype && length($optype);
                  $optype .= $pendingoptype;
               }
            }
         }


         # At this point we should have something with an op-type (hopefully)
         $proto->[4] |= $PROTO_OPTYPE if( defined $optype && length($optype) );

         if( $rettype eq 'HANDLE' ) {
            $optype .= ', ' if defined $optype && length($optype);
            $optype .= 'handleretop, fdretop'
         }

         if( $rettype eq 'SOCKET' ) {
            $optype .= ', ' if defined $optype && length($optype);
            $optype .= 'fdretop'
         }

         if( $fname =~ /W$/ ) {
            $optype .= ', ' if defined $optype && length($optype);
            $optype .= 'unicode';
         }

         if( defined $optype && length($optype) ) {
            my $last;
            my @ops;

            foreach( sort split /, /, $optype ) {
               push @ops, $_ if !$last || $_ ne $last;
               $last = $_;
            }
            $optype = join ', ', @ops;
         }

         my $fpre = $basename;
         $fpre .= '{'.$optype.'}' if( defined $optype && length($optype) );
         $fpre .= '(';
         print $fpre;

         $firstdone = 1;
      }

      print "\n\t\t$argtype";
      print ' *' if exists $pointerStructs{$argtype}; # Hack
      print " $argname";
      print '_' if $argname eq 'type';        # Hack for "type" argnames
      print $arraytokens if $arraytokens;

      print " {$argname";
      print '_' if $argname eq 'type';        # Hack for "type" argnames
      if( exists $czArgAttribs->{$argtype} ) {
         my $cztype = $czArgAttribs->{$argtype};
         if( exists $cztype->{$argname} ) {
            my $argtype = $cztype->{$argname};
            if( $argtype =~ /^(.*):ignore (\S+)/ ) {
               $argtype = $1;
               my $op;

               foreach $op ( @{$czIgnoreOps->{$2}} ) {
                  undef $argtype, last if( $op eq $basename );
               }
            }
            elsif( $argtype =~ /^(.*):copy/ ) {
               if( $nbytesarg ) {
                  $argtype = "copy($argname,$nbytesarg)";
                  $argtype .= ", $1" if $1;
               }
               else {
                  die "No nbytes argument found for out buffer '" . $argname .
                      "' in '" . $fname . "'";
               }
            }
            (print ", $argtype"), $proto->[4] |= $PROTO_ARGTYPE
               if defined $argtype;
         }
         if( exists $cztype->{'*'} ) {
            (print ", " . $cztype->{'*'}), $proto->[4] |= $PROTO_ARGTYPE;
         }
      }
      print '}';
   }
   if( !@$funcargs ) {
      my $regexp;
      my $basename = $fname;
      $basename =~ s/W$//;
      print $basename;
      foreach $regexp ( @czRegexpMatchOps ) {
         next unless $regexp =~ /^\^/;
         ( print '{'.$czRegexpMatchOpsMap->{$regexp}.'}' ),
            $proto->[4] |= $PROTO_OPTYPE, last if $basename =~ /$regexp/;
      }
      print '( void ';

      $proto->[4] |= $PROTO_ARGTYPE;
   }
   else {
      print "\n";
   }
   print ");\n\n";

   # Ignore missing argtypes in characterization if flag is set
   $proto->[4] |= $PROTO_ARGTYPE unless $show_missing_argtype;
}

print <<EOF;

}

EOF

select( STDOUT );
close( CHARACTERIZE );
print "Characterization done.\n";


###############################
# Generate interception stubs #
###############################
print "Generating interception stubs...\n";
open( STUBHDR, '>lwmstubs' . $winvers . '.h' ) ||
   die 'lwmstubs' . $winvers . ".h: $!";
select( STUBHDR );

print <<EOF;
/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * lwmstubs${winvers}.h
 *
 * \$Id: genstubs.pl,v 1.74 2001/04/05 18:40:36 stevek Exp $
 *
 * Note: This file is automatically generated by genstubs.pl
 *       Do not modify.
 */

#ifndef __LWMSTUBS_H
#define __LWMSTUBS_H 1

#include "hook.h"

#ifdef __cplusplus
extern "C" {
#endif

extern const int    k_NUM_HOOKED_FUNCS;
extern HOOKFUNCDESC g_stToHookFuncs[];
extern PROC         g_pOrigFuncs[];

/*
 * Intercepted function prototypes
 */
EOF

foreach( sort { $a->[0] cmp $b->[0] } @prototypes ) {
   my( $fname, $rettype, $funcargs ) = @$_;

   next if exists $czIgnoreCalls{$fname};

   my $fpre = "$rettype WINAPI LWM_$fname(";
   print $fpre;

   my $length = 0;
   my $prelen = 0;
   foreach( @$funcargs ) {
      if( $prelen ) { $length = $prelen; printf(",\n%*s", $length, ''); }
      else          { $prelen = length($fpre); }
      print " $_";
   }
   if( $length )    { printf("\n%*s", $length - 2, ''); }
   if( !@$funcargs ) { print ' void'; }
   print " );\n\n";
}

print <<EOF;

#ifdef __cplusplus
}
#endif

#endif /* !__LWMSTUBS_H */

EOF

select( STDOUT );
close( STUBHDR );

my $nentries = 0;

foreach( @prototypes ) {
   $nentries++ if !exists $czIgnoreCalls{${$_}[0]};
}

open( STUBS, '>lwmstubs' . $winvers . '.cxx' ) ||
   die 'lwmstubs' . $winvers . ".cxx: $!";
select( STUBS );

print <<EOF;
/*
 *  Copyright (c) 1999, 2000 Network Associates, Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use are governed by the terms detailed in the
 *  license document ("LICENSE") included with the toolkit.
 *
 * lwmstubs${winvers}.cxx
 *
 * \$Id\$
 *
 * Note: This file is automatically generated by genstubs.pl
 *       Do not modify.
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include "hook.h"
#include "lwmstubs${winvers}.h"

#include <wss_process.h>
#include <wss_wrapper.h>
#include <wss_arg.h>
#include <wsi_os.h>

extern HOOKMAP        g_hookMap;
extern wss_proc      *g_wrp;
extern volatile BOOL  g_bTraceCalls;
extern volatile BOOL  g_bCountCalls;

#ifdef UNICODE
#undef GetEnvironmentStrings
#endif

// Modules which can be hooked
const char *k_MODULES_TO_HOOK[] =
{
EOF

foreach( @symtabs ) {
   print '   "' . $_->[0] . '",' . "\n" if $_->[1];
}

print <<EOF;
   ""
};

// The number of functions to possibly be hooked
const int k_NUM_HOOKED_FUNCS = $nentries;

// The master table of function names and the hook functions.
EOF

my $ntables;
$nentries = 0;
foreach( sort { $a->[0] cmp $b->[0] } @prototypes ) {
   my( $fname ) = @$_;

   if( ( $nentries % $hooks_per_table ) == 0 ) {
      $ntables++;
      print "};\n\n" if $nentries > 0;
      print 'HOOKFUNCDESC g_stToHookFuncs' . $ntables . "[] =\n{\n";
   }

   next if exists $czIgnoreCalls{$fname};

   if( ( $nentries++ % $hooks_per_table ) != 0 ) { print ",\n"; }
   print '   HOOKFUNCDESC( "' . $_->[0] . '", (PROC)LWM_' . $_->[0] .
         ', "' . $_->[3] . '", ' . $_->[5] . ' )';
}

print <<EOF;
};

HOOKFUNCDESC & lwmstubGetHookFuncEntry( int index )
{
EOF

foreach( 1 .. $ntables ) {
   if( $_ < $ntables ) {
      print <<EOF;
   if( ( index >= ( ( $_ - 1 ) * $hooks_per_table ) ) &&
       ( index < ( $_ * $hooks_per_table ) ) )
EOF
      print '      return g_stToHookFuncs' . $_ .
            "[ index - ($_ - 1) * $hooks_per_table ];\n";
   }
   else {
      print <<EOF;
   else if( ( index >= ( ( $_ - 1 ) * $hooks_per_table ) ) &&
       ( index < $nentries ) )
EOF
      print '      return g_stToHookFuncs' . $ntables .
            "[ index - ($_ - 1) * $hooks_per_table ];\n";

      print <<EOF;
   panic( "lwmstubGetHookFuncEntry: invalid index" );
   
EOF
   }
}
print <<EOF;
}

WSS_MUTEX_PROTECT(g_callcount);
volatile DWORD g_callcount = 0;

WSS_MUTEX_PROTECT(g_tracecount);
volatile DWORD g_tracecount = 0;

PROC g_pOrigFuncs[ k_NUM_HOOKED_FUNCS ] =
EOF

print "{";

foreach( 0 .. $#prototypes ) {
   next if exists $czIgnoreCalls{${$prototypes[$_]}[0]};

   print "\n  " if( ( $_ % 10 ) == 0 );
   print ' NULL';
   print ',' if( $_ < $#prototypes );
}

print <<EOF;

};

LPCWSTR _LWMStubs_ConvertASCIIToUnicode( LPCSTR pStr )
{
   if( pStr == NULL ) return NULL;

   DWORD strLen = strlen( pStr ) + 1;
   pd_char_t *wstr = (pd_char_t *)malloc( strLen * sizeof(pd_char_t) );
   MultiByteToWideChar( CP_ACP, 0, pStr, strLen, wstr,
                        strLen * sizeof(pd_char_t) );
   return wstr;
}

LPCSTR _LWMStubs_ConvertUnicodeToASCII( LPCWSTR pStr )
{
   if( pStr == NULL ) return NULL;

   DWORD strLen = _tcslen( pStr ) + 1;
   char *wstr = (char *)malloc( strLen * sizeof(pd_char_t) );
   WideCharToMultiByte( CP_ACP, 0, pStr, strLen, wstr,
                        strLen * sizeof(pd_char_t), NULL, NULL );
   return wstr;
}

EOF

foreach( 0 .. int( ( $nentries - 1 ) / $hooks_per_table ) ) {
   print <<EOF;
#include "lwmstubs${winvers}_$_.cxx"
EOF
}

select( STDOUT );
close( STUBS );

my @stubEnums = qw(
   GET_FILEEX_INFO_LEVELS
   FINDEX_INFO_LEVELS
   FINDEX_SEARCH_OPS
   WSAESETSERVICEOP
   SECURITY_IMPERSONATION_LEVEL
   TOKEN_TYPE
   ACL_INFORMATION_CLASS
   TOKEN_INFORMATION_CLASS
   AUDIT_EVENT_TYPE
   SC_ENUM_TYPE
   COMPUTER_NAME_FORMAT
   JOBOBJECTINFOCLASS
   SC_STATUS_TYPE
);
my %stubEnums;
map { $stubEnums{$_} = undef } @stubEnums;

#
# Map for function argument conversions
#
my $stubArgInfo = {
   'LPCSTR' =>
   {
      '*'             => 'instr'
   },
   'LPCWSTR' =>
   {
      '*'             => 'inwstr'
   },
   'LPSTR' =>
   {
      'lpCommandLine' => 'instr',
      'lpDst'         => 'outstr',
      'lpDestSrc'     => 'outstr',
      'lpBuffer'      => 'outstr'
   },
   'DWORD' =>
   {
      'nSize'         => 'size',
      'cchDest'       => 'size',
      'nBufferLength' => 'size'
   },
   'int' =>
   {
      'nBufferMax'    => 'size',
      'nSize'         => 'size'
   },
   'UINT' =>
   {
      'uSize'         => 'size'
   },
   'LPDWORD' =>
   {
      'nSize'         => 'size,special:(MAX_COMPUTERNAME_LENGTH + 1)'
   },
   'LPSTR *' =>
   {
      'lpFilePart'    => 'offsetptr'
   }
};

my $funcid = 1;
foreach( sort { $a->[0] cmp $b->[0] } @prototypes ) {
   my( $fname, $rettype, $funcargs ) = @$_;

   next if exists $czIgnoreCalls{$fname};

   if( ( ( $funcid - 1 ) % $hooks_per_table ) == 0 ) {
      my $part = int( ( $funcid - 1 ) / $hooks_per_table );

      select( STDOUT ), close( STUBS ) if $part;
      open( STUBS, '>lwmstubs' . $winvers . '_' . $part . '.cxx' ) ||
         die 'lwmstubs' . $winvers . '_' . $part . ".cxx: $!";
      select( STUBS );

      my $end = $funcid + $hooks_per_table - 1;
      $end = $nentries unless $end < $nentries;

      print <<EOF;
/*
 *  Copyright (c) 1999, 2000 Network Associates, Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use are governed by the terms detailed in the
 *  license document ("LICENSE") included with the toolkit.
 *
 * lwmstubs${winvers}_${part}.cxx
 *
 * \$Id\$
 *
 * Note: This file is automatically generated by genstubs.pl
 *       Do not modify.
 */

/****************************************************************************\
 *
 * Here are the hooked function replacements (#$funcid - #$end)
 *
\****************************************************************************/

EOF
   }

   my $fpre = "$rettype WINAPI LWM_$fname(";
   print $fpre;

   my $length = 0;
   my $prelen = 0;
   foreach( @$funcargs ) {
      if( $prelen ) { $length = $prelen; printf(",\n%*s", $length, ''); }
      else          { $prelen = length($fpre); }
      print " $_";
   }
   if( $length )    { printf("\n%*s", $length - 2, ''); }
   if( !@$funcargs ) { print ' void'; }

   my $type = 'Unknown';
   $type = 'Unicode' if( $fname =~ /W$/ );
   $type = 'ASCII' if( $fname =~ /A$/ );
   $type = 'ASCII' if( $type eq 'Unknown' );

   print <<EOF;
 )
{
EOF

   print "   $rettype callRet = {NULL};\n" if $rettype ne 'void';
   print "   void *callRet = {NULL};\n" if $rettype eq 'void';

   my $nfuncargs = ((@$funcargs)?@$funcargs:0);
   if( $nfuncargs > 0 ) {
      print <<EOF;
   $uaptype uapsave[$nfuncargs];
   $uaptype uap[$nfuncargs];
EOF
   }
   else {
      print "   $uaptype *uap = NULL;\n";
   }

   print <<EOF;

   static HOOKFUNCDESC *hook = NULL;

   if( !hook )
   {
      if( g_bCountCalls )
      {
         WSS_MUTEX_ENTER( g_callcount );
         wsi_printf( PD_MAKESTR("\%6d:\%-32s\%c"), g_callcount++,
                     PD_MAKESTR("$fname"),
                     (g_callcount \% 2)?PD_MAKESTR('\\n'):PD_MAKESTR(' ') );
         WSS_MUTEX_EXIT( g_callcount );
      }

      HOOKMAP::iterator hookIterator;
      hookIterator = g_hookMap.find("$fname");
      hook = &(*hookIterator).second;
   }

EOF

   if( @$funcargs ) {
      my $sizearg;

      # Find the size argument (if any)
      my $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'size' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'size' ) ) {
               $sizearg = $3;
               last;
            }
            elsif( ( exists $stubtype->{$3} &&
                     $stubtype->{$3} =~ /^size,special:(.+)$/ ) ||
                   ( exists $stubtype->{'*'} &&
                     $stubtype->{'*'} =~ /^size,special:(.+)$/ ) ) {
               $sizearg = $1;
               last;
            }
         }
         $narg++;
      }

      # Do conversions and/or mallocs (as necessary)
      $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         print '   uapsave[' . $narg . "] =\n";
         print '      uap[' . $narg++ . "] = ($uaptype) ";
         print '&' if exists $pointerStructs{$1};  # Hack

         my $argname = $3;

         push( @errors, "Invalid type/name combination: $_" ), next if !$argname;

         my $doneIt;
         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$argname} && $stubtype->{$argname} eq 'instr' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'instr' ) ) {
               print "((HIWORD($argname)==0)? ($uaptype) $argname : " if( $fname =~ /Resource|ProcAddress/ );
               print "($uaptype) _LWMStubs_ConvertASCIIToUnicode( (LPCSTR)$argname )";
               print ")" if( $fname =~ /Resource|ProcAddress/ );
               $doneIt = 1;
            }
            elsif( ( exists $stubtype->{$argname} && $stubtype->{$argname} eq 'outstr' ) ||
                   ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'outstr' ) ) {
               if( $sizearg ) {
                  print "(($argname==NULL)?NULL:(($sizearg>0)?";
                  print "malloc($sizearg * sizeof(pd_char_t)):$argname))";
                  $doneIt = 1;
               }
               else {
                  die "No size argument found for out string '" . $argname .
                      "' in '" . $fname . "'";
               }
            }
            elsif( $type eq 'ASCII' &&
                   ( ( exists $stubtype->{$argname} && $stubtype->{$argname} eq 'size' ) ||
                     ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'size' ) ) ) {
               print "( $argname * sizeof(pd_char_t) )";
               $doneIt = 1;
            }
         }
         print $argname if !$doneIt;
         print ";\n";
      }

      print <<EOF;
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
EOF
      print "      wsi_printf( PD_MAKESTR(\"\%6d:\%s \%s(";

      foreach( @$funcargs ) {
         print ' 0x%x';

         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$3} && (($stubtype->{$3} eq 'instr') ||
                                             ($stubtype->{$3} eq 'inwstr')
                                             )) ||
                ( exists $stubtype->{'*'} && (($stubtype->{'*'} eq 'instr') ||
                                              ($stubtype->{'*'} eq 'inwstr')
                                              ))) {
               print ':\"%s\"';
            }
         }
      }
      print ' )"),';
      print "\n                  g_tracecount++,";
      print "\n                  PD_MAKESTR(\"$rettype\"),";
      print "\n                  PD_MAKESTR(\"$fname\"),";

      $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         print ',' if $narg;
         print "\n" . '                  uap[' . $narg . ']';
         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$3} && (($stubtype->{$3} eq 'instr') ||
                                             ($stubtype->{$3} eq 'inwstr')
                                             )) ||
                ( exists $stubtype->{'*'} && (($stubtype->{'*'} eq 'instr') ||
                                              ($stubtype->{'*'} eq 'inwstr')
                                              ))) {
               print ",\n                  (LPCWSTR) ";
               print '((HIWORD(uap[' . $narg . '])==0)? PD_MAKESTR("<resource id>") : ' if( $fname =~ /Resource/ );
               print '((HIWORD(uap[' . $narg . '])==0)? PD_MAKESTR("<ordinal>") : ' if( $fname =~ /ProcAddress/ );
               print 'uap[' . $narg . ']';
               print ')' if( $fname =~ /Resource|ProcAddress/ );
            }
         }
         $narg++;
      }

      print <<EOF;

                );
      WSS_MUTEX_EXIT( g_tracecount );
   }
EOF

   }
   else
   {
      print <<EOF;
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
      wsi_printf( PD_MAKESTR("\%6d:\%s %s()"), g_tracecount++,
                  PD_MAKESTR("$rettype"), PD_MAKESTR("$fname") );
      WSS_MUTEX_EXIT( g_tracecount );
   }
EOF
   }

   print <<EOF;

   int err = 0;
   wss_wrapper_data data = { 0, 0 };
   data.errnum = &err;

   int retval = hook->callDispatchPre( /*space*/ 0, $funcid, (void **)uap, 
                                       &callRet, g_wrp, &data,
                                       /*args_in_user_space*/ 1 );

EOF

   foreach( @$funcargs ) {
      /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
      if( exists $stubArgInfo->{$1} ) {
         my $stubtype = $stubArgInfo->{$1};
         if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'offsetptr' ) ||
             ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'offsetptr' ) ) {
            print <<EOF;
   DWORD $3__offset = 0;
EOF
         }
      }
   }

   print <<EOF;

   if( !retval )
   {
EOF

   if( $type eq 'ASCII' ) {
      if( $nfuncargs > 0 ) {
         print <<EOF;
      $uaptype uapascii[$nfuncargs];
EOF
      }

      my $sizearg;

      # Find the size argument (if any)
      my $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'size' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'size' ) ) {
               $sizearg = $3;
               last;
            }
            elsif( ( exists $stubtype->{$3} &&
                     $stubtype->{$3} =~ /^size,special:(.+)$/ ) ||
                   ( exists $stubtype->{'*'} &&
                     $stubtype->{'*'} =~ /^size,special:(.+)$/ ) ) {
               $sizearg = $1;
               last;
            }
         }
         $narg++;
      }

      $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;

         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'instr' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'instr' ) ) {
               my $argname = $3;
               print "      uapascii[$narg] =\n         ";
               print "((HIWORD($argname)==0)? ($uaptype) $argname : " if( $fname =~ /Resource|ProcAddress/ );
               print "($uaptype) _LWMStubs_ConvertUnicodeToASCII( (LPCWSTR)uap[$narg] )";
               print ")" if( $fname =~ /Resource|ProcAddress/ );
               print ";\n";
               $narg++;
               next;
            }
            elsif( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'size' ) ||
                   ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'size' ) ) {
               print <<EOF;
      uapascii[$narg] = ($uaptype) $sizearg;
EOF
               $narg++;
               next;
            }
         }

         print <<EOF;
      uapascii[$narg] = uap[$narg];
EOF
         $narg++;
      }
   }

   $fpre = "      callRet = $fname(" if $rettype ne 'void';
   $fpre = "      $fname(" if $rettype eq 'void';
   print $fpre;

   my $narg = 0;
   $length = $prelen = 0;
   foreach( @$funcargs ) {
      if( $prelen ) { $length = $prelen; printf(",\n%*s", $length, ''); }
      else          { $prelen = length($fpre); }

      /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;

      if( exists $pointerStructs{$1} ) {
         # Hack
         print $3;
         $narg++;
         next;
      }

      print ' (' . $1;
      print ' *' if $4;
      print ') ';
      print '((DWORD)' if exists $stubEnums{$1};
      print 'uap';
      print 'ascii' if $type eq 'ASCII';
      print '[' . $narg++ . ']';
      print ')' if exists $stubEnums{$1};
   }
   if( $length )    { printf("\n%*s", $length - 2, ''); }

   print <<EOF;
);

      if( g_bTraceCalls )
EOF

   if( $rettype ne 'void' )
   {
      print "         wsi_printf( PD_MAKESTR(\" ==> ";
      if( $rettype eq 'BOOL' ) {
         print "%s\\n\"), callRet ? PD_MAKESTR(\"TRUE\") : PD_MAKESTR(\"FALSE\")";
      }
      else {
         print "0x\%1x\\n\"), callRet";
      }
      print " );\n\n";
   }
   else
   {
      print <<EOF;
      wsi_printf( PD_MAKESTR(\"\\n\") );

EOF
   }

   my $hasoffsets;
   if( $type eq 'ASCII' ) {
      my $sizearg;

      # Find the size argument (if any)
      my $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'offsetptr' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'offsetptr' ) ) {
               $hasoffsets++;
            }
            elsif( !$sizearg ) {
               if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'size' ) ||
                   ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'size' ) ) {
                  $sizearg = $3;
               }
               elsif( ( exists $stubtype->{$3} &&
                        $stubtype->{$3} =~ /^size,special:(.+)$/ ) ||
                      ( exists $stubtype->{'*'} &&
                        $stubtype->{'*'} =~ /^size,special:(.+)$/ ) ) {
                  $sizearg = $1;
               }
            }
         }
         $narg++;
      }

      $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         my( $argtype, $argname ) = ( $1, $3 );

         if( exists $stubArgInfo->{$argtype} ) {
            my $stubtype = $stubArgInfo->{$argtype};
            if( ( exists $stubtype->{$argname} && $stubtype->{$argname} eq 'instr' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'instr' ) ) {
               print "      if( uapascii[$narg] ";
               print "&& ( HIWORD(uapascii[$narg]) != 0 ) " if( $fname =~ /Resource|ProcAddress/ );
               print <<EOF;
)
         free( (PVOID)uapascii[$narg] );
EOF
            }
            elsif( ( exists $stubtype->{$argname} && $stubtype->{$argname} eq 'outstr' ) ||
                   ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'outstr' ) ) {
               if( $sizearg ) {
                  print <<EOF;
      if( uapascii[$narg] )
      {
         pd_char_t *replace = (pd_char_t *)malloc($sizearg * sizeof(pd_char_t));
EOF
                  if( $hasoffsets ) {
                     my $offarg;
                     foreach $offarg ( @$funcargs ) {
                        $offarg =~ /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;

                        if( exists $stubArgInfo->{$1} ) {
                           my $stubtype = $stubArgInfo->{$1};
                           if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'offsetptr' ) ||
                               ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'offsetptr' ) ) {
                              print <<EOF;
         if( $3 != NULL ) $3__offset = (DWORD)*$3 - (DWORD)uapascii[$narg];
EOF
                           }
                        }
                     }
                  }
                  print <<EOF;
         MultiByteToWideChar( CP_ACP, 0, (LPCSTR)uapascii[$narg], $sizearg,
                              replace, $sizearg * sizeof(pd_char_t) );
         free( (PVOID)uapascii[$narg] );
         uap[$narg] = ($uaptype) replace;
      }
EOF
               }
               else {
                  die "No size argument found for out string '" . $argname .
                      "' in '" . $fname . "'";
               }
            }
         }
         $narg++;
      }
   }

   print <<EOF;

      retval = hook->callDispatchPost( /*space*/ 0, $funcid, (void **)uap, 
                                       &callRet, g_wrp, &data,
                                       /*args_in_user_space*/ 1 );
   }

EOF

   $funcid++;

   if( @$funcargs ) {
      my $sizearg;

      # Find the size argument (if any)
      $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         if( exists $stubArgInfo->{$1} ) {
            my $stubtype = $stubArgInfo->{$1};
            if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'size' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'size' ) ) {
               $sizearg = $3;
               last;
            }
            elsif( ( exists $stubtype->{$3} &&
                     $stubtype->{$3} =~ /^size,special:(.+)$/ ) ||
                   ( exists $stubtype->{'*'} &&
                     $stubtype->{'*'} =~ /^size,special:(.+)$/ ) ) {
               $sizearg = $1;
               last;
            }
         }
         $narg++;
      }

      # Do conversions and/or frees (as necessary)
      $narg = 0;
      foreach( @$funcargs ) {
         /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;
         my( $argtype, $argname ) = ( $1, $3 );

         if( exists $stubArgInfo->{$argtype} ) {
            my $stubtype = $stubArgInfo->{$argtype};
            if( ( exists $stubtype->{$argname} && $stubtype->{$argname} eq 'instr' ) ||
                ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'instr' ) ) {
               print "   if( uapsave[$narg] ";
               print "&& ( HIWORD(uapsave[$narg]) != 0 ) " if( $fname =~ /Resource|ProcAddress/ );
               print <<EOF;
)
      free( (PVOID)uapsave[$narg] );
EOF
            }
            elsif( $type eq 'ASCII' &&
                   ( ( exists $stubtype->{$argname} && $stubtype->{$argname} eq 'outstr' ) ||
                     ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'outstr' ) ) ) {
               if( $sizearg ) {
                  print <<EOF;
   if( uap[$narg] )
   {
      WideCharToMultiByte( CP_ACP, 0, (LPCWSTR)uap[$narg],
                           $sizearg * sizeof(pd_char_t),
                           (char *)$argname, $sizearg, NULL, NULL );
      free( (PVOID)uap[$narg] );
EOF
                  if( $hasoffsets ) {
                     my $offarg;
                     foreach $offarg ( @$funcargs ) {
                        $offarg =~ /^(\S+(\s+[^\[\s]+)*)\s+(\S+)(\s+\[\s+\]\s*)?$/;

                        if( exists $stubArgInfo->{$1} ) {
                           my $stubtype = $stubArgInfo->{$1};
                           if( ( exists $stubtype->{$3} && $stubtype->{$3} eq 'offsetptr' ) ||
                               ( exists $stubtype->{'*'} && $stubtype->{'*'} eq 'offsetptr' ) ) {
                              print <<EOF;
      if( $3 != NULL ) *$3 = $argname + $3__offset;
EOF
                           }
                        }
                     }
                  }
                  print <<EOF;
   }
EOF
               }
               else {
                  die "No size argument found for out string '" . $3 .
                      "' in '" . $fname . "'";
               }
            }
         }
         $narg++;
      }
   }

   if( $nfuncargs > 0 ) {
      print <<EOF;

   /* Clean up arg replacement */
   wss_arg_donesyscall( g_wrp, wsi_os_curthreadid(), 0, $funcid, (void **)uap );

EOF
   }

   if( $rettype ne 'void' ) {
      print <<EOF;
   if( !retval )
   {
      return callRet;
   }
   else
   {
      // we should really decompose the relevant return code!!!
      return callRet;
   }
EOF
   }

   print <<EOF;
}

EOF
}
select( STDOUT );
close( STUBS );
print "Stubs done.\n";


##########################
# Generate symbols table #
##########################
print "Generating symbols table...\n";
open( SYMSTAB, '>../wr.lib/libwdl/syms_nt' . $winvers . '.c' ) ||
   die 'syms_nt' . $winvers . ".c: $!";
select( SYMSTAB );

print <<EOF;
/*
 *  Copyright (c) 1999, 2000 Network Associates, Incorporated
 *  All rights reserved.
 *
 *  Redistribution and use are governed by the terms detailed in the
 *  license document ("LICENSE") included with the toolkit.
 *
 * syms_nt${winvers}.c
 *
 * \$Id\$
 *
 * Note: This file is automatically generated by genstubs.pl
 *       Do not modify.
 */

#include <wr_conf.h>

LPTSTR winnt_syscallnames[] =
{
EOF

foreach $symtab ( @symtabs ) {
   my $symbols = $symtab->[2];
   foreach ( sort keys %$symbols ) {
      if( $symbols->{$_}->[1] ) {
         my $sym = $_;

         next if exists $czIgnoreCalls{$sym};

         # Test for W and A versions
         if( /[WA]$/ ) {
            $sym =~ s/.$//;

            next if( /A$/ && $symbols->{$sym . 'W'}->[1] );
         }
         else {
            next if( $symbols->{$sym . 'W'}->[1] ||
                     $symbols->{$sym . 'A'}->[1] );
         }

         print '   _TEXT("' . $sym;
         print 'W' if $symbols->{$sym . 'W'}->[1];
         print '"),' . "\n";
      }
   }
}

print <<EOF;
   NULL
};

EOF
select( STDOUT );
close( SYMSTAB );
print "Symbols table done.\n";


#############################
# Display unmatched symbols #
#############################
my $missing;
my $intercepted;
foreach $symtab ( @symtabs ) {
   my $symbols = $symtab->[2];
   foreach ( sort keys %$symbols ) {
      if( ! $symbols->{$_}->[1] ) {
         $intercepted++, next if /[WA]$/;

         # Test for W and A versions
         next if( $symbols->{$_ . 'W'}->[1] && $symbols->{$_ . 'A'}->[1] );
         $intercepted++, next
            if( $symbols->{$_ . 'W'}->[1] || $symbols->{$_ . 'A'}->[1] );

         print "\nPrototypes unavailable for the following symbols:\n"
            if !$missing++;

         printf("%-39s", $_);
         print ((($missing % 2 )== 0)?"\n":" ");
      }
   }
}

print "\n\n$missing symbol(s) unmatched.\n" if $missing;
print "\n$intercepted symbol(s) intercepted.\n" if $intercepted;


###################################################
# Display functions with missing characterization #
###################################################

my $czMissing;
my $czArgMissing = 0;
my $czOpMissing = 0;
my $czRetMissing = 0;

foreach( sort { $a->[0] cmp $b->[0] } @prototypes ) {
   next if $_->[0] =~ /A$/;
   next if $_->[0] eq 'GetEnvironmentStrings'; # Hack for MS hack

   if( ( $_->[4] & ( $PROTO_MASK ) ) != $PROTO_MASK ) {
      print "\nPrototypes with incomplete characterization:\n"
         if !$czMissing++;

      printf("%-39s", $_->[0]);
      printf("  %-7s", ($_->[4] & $PROTO_RETTYPE)?'':'RETTYPE');
      printf("  %-7s", ($_->[4] & $PROTO_OPTYPE)?'':'OPTYPE');
      printf("  %-7s", ($_->[4] & $PROTO_ARGTYPE)?'':'ARGTYPE');
      print "\n";

      $czRetMissing++ unless $_->[4] & $PROTO_RETTYPE;
      $czOpMissing++ unless $_->[4] & $PROTO_OPTYPE;
      $czArgMissing++ unless $_->[4] & $PROTO_ARGTYPE;
   }
}

if( $czMissing ) {
   print <<EOF


$czMissing function(s) with incomplete characterization:

      Missing return type: $czRetMissing
   Missing operation type: $czOpMissing
    Missing argument type: $czArgMissing

EOF
}


##################
# Display errors #
##################

if( @errors ) {
   my $numErrors = @errors;
   print "\n\nEncountered $numErrors error(s) during processing:\n";
   foreach( @errors ) {
      print "  $_\n";
   }
}

close( OUTFILE ) if( $outfile );

1;
