/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: gwmcomm.cxx,v 1.45 2001/04/05 18:40:36 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Routines to communicate with the Global Wrapper Manager service
 */

#include <wr_conf.h>

#include <wsi_types.h>

#include <windows.h>
#include <process.h>
#include <tchar.h>
#include <direct.h>
#include <Lmcons.h>

#include <wss_malloc.h>
#include <wss_mutex.h>
#include <wss_cv.h>
#include <wss_process.h>

#include "gwmcomm.h"
#include "wdbm.h"

#include <vector>

using namespace std;

TCHAR g_progName[_MAX_PATH];
extern proc_t g_proc;

/***************************************************************************\
 * Define Structures                                                       *
\***************************************************************************/

// public class for mapping packet ids to wait handles
//
struct IdMap
{
   DWORD                      dwID;
   DWORD                      dwThreadId;
   HANDLE                     hWait;
   LPPIPESRV_COMM_RESP_HEADER pCommHdr;
   PVOID                      pBuf;

   IdMap( DWORD id, DWORD tid )
   {
      this->dwID = id;
      this->dwThreadId = tid;
      this->hWait = CreateEvent( NULL, TRUE, FALSE, NULL );
      this->pCommHdr = NULL;
      this->pBuf = NULL;
   }

   ~IdMap() { CloseHandle( this->hWait ); }

   operator==( IdMap &A ) { return this->dwID == A.dwID; }
};


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern volatile BOOL g_asyncOk;


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

HANDLE            g_hPipe = INVALID_HANDLE_VALUE;
vector<IdMap *>   g_vIdMaps;
WSS_MUTEX_PROTECT(g_vIdMaps);
HANDLE            g_hReadThread = NULL;
WSS_MUTEX_PROTECT(g_criticalThread);
WSS_MUTEX_PROTECT(g_criticalRead);
WSS_MUTEX_PROTECT(g_criticalWrite);
HANDLE            g_hReadEvent;
HANDLE            g_hWriteEvent;
int               g_syncReads = 0;
WSS_MUTEX_PROTECT(g_syncReads);
WSS_CV_PROTECT   (g_syncReads);
volatile BOOL     g_bNoThrMsgs = FALSE;
volatile BOOL     g_bNoProcMsgs = FALSE;
volatile BOOL     g_bTraceCalls = FALSE;
volatile BOOL     g_bCountCalls = FALSE;


/***************************************************************************\
 * Forward Declarations                                                    *
\***************************************************************************/

BOOL   gwmcommConnect              ( void );
BOOL   gwmcommGetPacket            ( DWORD, LPPIPESRV_COMM_RESP_HEADER *, PVOID * );
HANDLE gwmcommGetWaitHandle        ( DWORD );
BOOL   gwmcommProcessPacket        ( LPPIPESRV_COMM_RESP_HEADER, PVOID );
BOOL   gwmcommWaitForPacketResponse( LPPIPESRV_COMM_RESP_HEADER *, PVOID * );


/***************************************************************************\
 * Private Functions                                                       *
\***************************************************************************/

static int
_gwmcommSendMessage( const pd_char_t *format, va_list ap )
{
   // If GWM does not want thread messages, just return TRUE
   if( g_bNoThrMsgs ) return TRUE;

   // Construct packet for message
   PIPESRV_COMM_RESP_HEADER    commHdr = { LWM_COMM_MESSAGE, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) +
                                           ( MAXPATHLEN * sizeof(pd_char_t) ) };
   LPPIPESRV_COMM_RESP_HEADER  pCommHdr = &commHdr;
   pd_byte_t                  *buf      =
      (pd_byte_t *)wss_malloc( MAXPATHLEN * sizeof(pd_char_t) +
                                sizeof(DWORD) );

   *((DWORD *)buf) = GetCurrentThreadId();
   int retval = _vsntprintf( (pd_char_t *)(buf + sizeof(DWORD)), MAXPATHLEN,
                             format, ap );

   gwmcommWritePacket( (LPPIPESRV_COMM_HEADER)&commHdr, buf );
   wss_free( buf );

   return retval;
}


/***************************************************************************\
 * Public Functions                                                        *
\***************************************************************************/

void
gwmcommSetup()
{
   WSS_MUTEX_INIT( g_criticalThread );
   WSS_MUTEX_INIT( g_criticalRead );
   WSS_MUTEX_INIT( g_criticalWrite );

   g_hReadEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
   g_hWriteEvent = CreateEvent( NULL, TRUE, FALSE, NULL );

   WSS_MUTEX_INIT(g_vIdMaps);

   WSS_MUTEX_INIT(g_syncReads);
   WSS_CV_INIT(g_syncReads);

   wsi_printf_fn = _gwmcommSendMessage;
}

void
gwmcommShutdown()
{
   wsi_printf_fn = NULL;

   WSS_CV_DESTROY(g_syncReads);
   WSS_MUTEX_DESTROY(g_syncReads);

   WSS_MUTEX_DESTROY(g_vIdMaps);

   CloseHandle(g_hWriteEvent);
   CloseHandle(g_hReadEvent);

   WSS_MUTEX_DESTROY(g_criticalWrite);
   WSS_MUTEX_DESTROY(g_criticalRead);
   WSS_MUTEX_DESTROY(g_criticalThread);
}

BOOL
gwmcommWritePacket( LPPIPESRV_COMM_HEADER pCommHdr,
                    PVOID pData )
{
   OVERLAPPED os;
   DWORD      dwWait;
   DWORD      dwDatalen;
   DWORD      cbWritten;
   DWORD      dwID;
   BOOL       bRet;

   if( g_hPipe == INVALID_HANDLE_VALUE )
      if( !gwmcommConnect() )
         return FALSE;
   
   dwDatalen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_HEADER);

   WSS_MUTEX_ENTER( g_criticalWrite );

   // init the overlapped structure
   //
   memset( &os, 0, sizeof(os) );
   os.hEvent = g_hWriteEvent;
   ResetEvent( g_hWriteEvent );

   // add an id to the packet header
   //
   dwID = GetTickCount() << GWM_COMM_ID_SHIFT;
   pCommHdr->commandId |= ( dwID & GWM_COMM_ID_MASK );

   // send out the header...
   //
   bRet = WriteFile( g_hPipe, pCommHdr, sizeof(PIPESRV_COMM_HEADER), &cbWritten,
                     &os );
   if( !bRet )
   {
      switch( GetLastError() )
      {
         case ERROR_IO_PENDING:
            dwWait = WaitForSingleObject( g_hWriteEvent, INFINITE );
            switch( dwWait )
            {
               case WAIT_OBJECT_0:
                  break;

               case WAIT_FAILED:
               default:
                  WSS_MUTEX_EXIT( g_criticalWrite );
                  return FALSE;
            }
            break;

         default:
            WSS_MUTEX_EXIT( g_criticalWrite );
            return FALSE;
      }
   }

   // send out the data...
   //
   if( dwDatalen > 0 )
   {
      // init the overlapped structure
      //
      memset( &os, 0, sizeof(os) );
      os.hEvent = g_hWriteEvent;
      ResetEvent( g_hWriteEvent );

      bRet = WriteFile( g_hPipe, pData, dwDatalen, &cbWritten, &os );
      if( !bRet )
      {
         switch( GetLastError() )
         {
            case ERROR_IO_PENDING:
               dwWait = WaitForSingleObject( g_hWriteEvent, INFINITE );
               switch( dwWait )
               {
                  case WAIT_OBJECT_0:
                     break;

                  case WAIT_FAILED:
                  default:
                     WSS_MUTEX_EXIT( g_criticalWrite );
                     return FALSE;
               }
               break;

            default:
               WSS_MUTEX_EXIT( g_criticalWrite );
               return FALSE;
         }
      }
   }

   WSS_MUTEX_EXIT( g_criticalWrite );
   return TRUE;
}

BOOL
gwmcommConnect()
{
   g_hPipe = CreateFile( GWM_LWM_PIPE, ( GENERIC_READ | GENERIC_WRITE ), 0,
                         NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL );

   if( g_hPipe == INVALID_HANDLE_VALUE )
   {
      if( GetLastError() != ERROR_PIPE_BUSY )
         return FALSE;
      if( !WaitNamedPipe( GWM_LWM_PIPE, 5000 ) )
         return FALSE;
   }

   return TRUE;
}

BOOL
gwmcommReadPacket( LPPIPESRV_COMM_RESP_HEADER pCommHdr,
                   PVOID *pBuf )
{
   OVERLAPPED os;
   DWORD      dwWait;
   DWORD      cbRead;
   BOOL       bRet;

   WSS_MUTEX_ENTER( g_criticalRead );

   // init the overlapped structure
   //
   memset( &os, 0, sizeof(os) );
   os.hEvent = g_hReadEvent;
   ResetEvent( g_hReadEvent );

   bRet = ReadFile( g_hPipe, pCommHdr, sizeof(PIPESRV_COMM_RESP_HEADER), &cbRead,
                    &os );
   if( !bRet )
   {
      switch( GetLastError() )
      {
         case ERROR_IO_PENDING:
            dwWait = WaitForSingleObject( g_hReadEvent, INFINITE );
            switch( dwWait )
            {
               case WAIT_OBJECT_0:
                  bRet = TRUE;
                  break;

               default:
                  break;
            }
            break;

         default:
            break;
      }
   }

   BOOL bProcRet = FALSE;

   if( bRet )
   {
      DWORD buflen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_RESP_HEADER);

      if( buflen )
      {
         *pBuf = wss_malloc( buflen );

         // init the overlapped structure
         //
         memset( &os, 0, sizeof(os) );
         os.hEvent = g_hReadEvent;
         ResetEvent( g_hReadEvent );

         bRet = ReadFile( g_hPipe, *pBuf, buflen, &cbRead, &os );
         if( !bRet )
         {
            switch( GetLastError() )
            {
               case ERROR_IO_PENDING:
                  dwWait = WaitForSingleObject( g_hReadEvent, INFINITE );
                  switch( dwWait )
                  {
                     case WAIT_OBJECT_0:
                        bRet = TRUE;
                        break;

                     default:
                        break;
                  }
                  break;

               default:
                  break;
            }
         }
      }
      else
         *pBuf = NULL;

      if( bRet )
         bProcRet = gwmcommProcessPacket( pCommHdr, *pBuf );
   }

   if( !bRet )
   {
      if( pCommHdr )
         wss_free( pCommHdr );

      if( pBuf && *pBuf )
         wss_free( *pBuf );
   }

   WSS_MUTEX_EXIT( g_criticalRead );

   return( bRet?bProcRet:bRet );
}

DWORD WINAPI
gwmcommWorkThread( PVOID args )
{
   // if we can't connect assume the gwm isn't running
   //
   if( g_hPipe == INVALID_HANDLE_VALUE )
      if( gwmcommConnect() == FALSE )
         return 0;

   // since we're running, set flag to signify async I/O is available
   //
   g_asyncOk = TRUE;

   // wait for all synchronous reads to finish
   //
   WSS_MUTEX_ENTER( g_syncReads );
   while( g_syncReads > 0 )
      WSS_CV_WAIT( g_syncReads );
   WSS_MUTEX_EXIT( g_syncReads );

   // process incoming packets
   //
   while( 1 )
   {
      PVOID                  pBuf     = NULL;
      LPPIPESRV_COMM_RESP_HEADER pCommHdr =
         (LPPIPESRV_COMM_RESP_HEADER)wss_malloc( sizeof(PIPESRV_COMM_RESP_HEADER) );

      if( !gwmcommReadPacket( pCommHdr, &pBuf ) )
      {
#ifdef DEBUG_GWM_DEATH
         MessageBox( 0, TEXT("killing gwmcommWorkThread due to GWM death"),
                     TEXT("LWM"), MB_OK | MB_ICONSTOP );
#endif /* DEBUG_GWM_DEATH */
         break;
      }
   }

   // restore wsi_printf to default
   //
   wsi_printf_fn = NULL;

   // shutdown the WDBM (remove hooks)
   //
   WDBMShutdown();

   // close all wait handles for id maps
   //
   WSS_MUTEX_ENTER( g_vIdMaps );
   for( vector<IdMap *>::iterator i = g_vIdMaps.begin();
        i != g_vIdMaps.end();
        i++ )
   {
      IdMap *idMap = (*i);
      CloseHandle( idMap->hWait );
   }
   WSS_MUTEX_EXIT( g_vIdMaps );

   return 0;
}

BOOL
gwmcommWaitForPacketResponse( LPPIPESRV_COMM_RESP_HEADER *pCommHdr,
                              PVOID *pPBuf )
{
   DWORD  dwID;
   HANDLE hWait;
   DWORD  dwWait;

   dwID = GWM_COMM_ID( **pCommHdr );
   hWait = gwmcommGetWaitHandle( dwID );

dowait:
   dwWait = WaitForSingleObject( hWait, GWM_WAIT_TIMEOUT );
   switch( dwWait )
   {
      case WAIT_TIMEOUT:
      {
         PIPESRV_COMM_RESP_HEADER commHdr = { LWM_COMM_SYNC_RESEND, _getpid(),
                                          GetCurrentThreadId(),
#ifdef UNICODE
                                          GWM_COMM_UNICODE,
#else
                                          0,
#endif
                                          sizeof(PIPESRV_COMM_HEADER) +
                                          sizeof(dwID) };
         if( gwmcommWritePacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                                 &dwID ) )
            goto dowait;
         else
            return FALSE;
      }

      case WAIT_OBJECT_0:
         break;

      default:
         return FALSE;
   }

   gwmcommGetPacket( dwID, pCommHdr, pPBuf );

   return TRUE;
}

//
// Assumes lock on g_vIdMaps is held
//
IdMap *
gwmcommFindID( DWORD dwID,
               DWORD dwThreadId,
               BOOL bRemove,
               BOOL bCreate )
{
   for( vector<IdMap *>::iterator i = g_vIdMaps.begin();
        i != g_vIdMaps.end();
        i++ )
   {
      if( ( dwID == (*i)->dwID ) && ( dwThreadId == (*i)->dwThreadId ) )
      {
         IdMap *idMap = (*i);

         if( bRemove )
            g_vIdMaps.erase( i );

         return idMap;
      }
   }

   if( bCreate )
   {
      IdMap *idMap = new IdMap( dwID, dwThreadId );

      if( !bRemove )
         g_vIdMaps.push_back( idMap );

      return idMap;
   }

   return NULL;
}

HANDLE
gwmcommGetWaitHandle( DWORD dwID )
{
   WSS_MUTEX_ENTER( g_vIdMaps );
   IdMap *idMap = gwmcommFindID( dwID, GetCurrentThreadId(), FALSE, TRUE );

   HANDLE hMap = idMap->hWait;
   WSS_MUTEX_EXIT( g_vIdMaps );
   return hMap;
}

BOOL
gwmcommGetPacket( DWORD dwID,
                  LPPIPESRV_COMM_RESP_HEADER *pCommHdr,
                  PVOID *pPBuf )
{
   WSS_MUTEX_ENTER( g_vIdMaps );
   IdMap *idMap = gwmcommFindID( dwID, GetCurrentThreadId(), TRUE, FALSE );
   WSS_MUTEX_EXIT( g_vIdMaps );

   if( idMap == NULL )
      return FALSE;

   *pCommHdr = idMap->pCommHdr;
   *pPBuf = idMap->pBuf;

   delete idMap;

   return TRUE;
}

BOOL
gwmcommProcessPacket( LPPIPESRV_COMM_RESP_HEADER pCommHdr,
                      PVOID pBuf )
{
   DWORD dwID = GWM_COMM_ID( *pCommHdr );

   switch( GWM_COMM_CMD( *pCommHdr ) )
   {
      case LWM_COMM_EVENT_DROP_TABLE:
         if( pCommHdr->pktlen != sizeof(PIPESRV_COMM_RESP_HEADER) + sizeof(u_long) )
            goto default_handling;

         WDBMHandleDropTableEvent( *((u_long *)pBuf) );
         break;

      case LWM_COMM_DEBUG_PROC_MESSAGES:
         if( pCommHdr->pktlen != sizeof(PIPESRV_COMM_RESP_HEADER) + sizeof(BOOL) )
            goto default_handling;

         g_bNoProcMsgs = *((BOOL *)pBuf);

         break;

      case LWM_COMM_DEBUG_THR_MESSAGES:
         if( pCommHdr->pktlen != sizeof(PIPESRV_COMM_RESP_HEADER) + sizeof(BOOL) )
            goto default_handling;

         g_bNoThrMsgs = *((BOOL *)pBuf);

         break;

      case LWM_COMM_DEBUG_TRACE_CALLS:
         if( pCommHdr->pktlen != sizeof(PIPESRV_COMM_RESP_HEADER) + sizeof(BOOL) )
            goto default_handling;

         g_bTraceCalls = *((BOOL *)pBuf);

         break;

      case LWM_COMM_DEBUG_COUNT_CALLS:
         if( pCommHdr->pktlen != sizeof(PIPESRV_COMM_RESP_HEADER) + sizeof(BOOL) )
            goto default_handling;

         g_bCountCalls = *((BOOL *)pBuf);

         break;

      default:
default_handling:
         WSS_MUTEX_ENTER( g_vIdMaps );
         IdMap *idMap = gwmcommFindID( dwID, pCommHdr->tid, FALSE, TRUE );
         if( idMap != NULL )
         {
            idMap->pCommHdr = pCommHdr;
            idMap->pBuf = pBuf;

            HANDLE hWait = idMap->hWait;
            WSS_MUTEX_EXIT( g_vIdMaps );

            SetEvent( hWait );
         }
         else
         {
#ifdef DEBUG_GWMCOMM
            // debug message stating dropped unknown packet
#endif /* DEBUG_GWMCOMM */
            wss_free( pCommHdr );
            wss_free( pBuf );
            WSS_MUTEX_EXIT( g_vIdMaps );

            return FALSE;
         }
   }
   return TRUE;
}

int
gwmcommTransactPacket( LPPIPESRV_COMM_HEADER pSendHeader,
                       PVOID pSendData,
                       LPPIPESRV_COMM_RESP_HEADER *pRecvHeader,
                       PVOID *pRecvData,
                       PINT pError /* currently unused */ )
{
   if( !gwmcommWritePacket( pSendHeader, pSendData ) )
      return GWMCOMM_TRANSACT_WRITE_FAILED;

   if( g_asyncOk )
   {
      if( !gwmcommWaitForPacketResponse( pRecvHeader, pRecvData ) )
         return GWMCOMM_TRANSACT_READ_FAILED;
   }
   else
   {
      DWORD  dwID  = GWM_COMM_ID( *pSendHeader );
      HANDLE hWait = gwmcommGetWaitHandle( dwID );

      WSS_MUTEX_ENTER( g_syncReads );
      g_syncReads++;
      WSS_MUTEX_EXIT( g_syncReads );

      // Since we are synchronous, we need to keep processing packets
      // until we see the one we want
      if( WaitForSingleObject( hWait, 0L ) == WAIT_TIMEOUT )
      {
         while( 1 )
         {
            PVOID                  pBuf     = NULL;
            LPPIPESRV_COMM_RESP_HEADER pCommHdr =
               (LPPIPESRV_COMM_RESP_HEADER)wss_malloc( sizeof(PIPESRV_COMM_RESP_HEADER) );

            BOOL bRet = gwmcommReadPacket( pCommHdr, &pBuf );

            if( bRet == FALSE )
               return GWMCOMM_TRANSACT_READ_FAILED;

            // Waiting on the wait handle should return WAIT_OBJECT_0 if
            // we've received the packet we want
            if( WaitForSingleObject( hWait, 0L ) == WAIT_OBJECT_0 )
               break;
         }
      }
      gwmcommGetPacket( dwID, pRecvHeader, pRecvData );

      WSS_MUTEX_ENTER( g_syncReads );
      g_syncReads--;
      WSS_MUTEX_EXIT( g_syncReads );
      WSS_CV_SIGNAL( g_syncReads );
   }

   return GWMCOMM_TRANSACT_SUCCESS;
}


// LWM->GWM packet commands
//
BOOL
gwmcommRequestWrappers( DWORD *pFlags, LPTSTR *pBuf )
{
   PIPESRV_COMM_RESP_HEADER   commHdr = { LWM_COMM_REQUEST_WRAPPERS, _getpid(),
                                          GetCurrentThreadId(),
#ifdef UNICODE
                                          GWM_COMM_UNICODE,
#else
                                          0,
#endif
                                          sizeof(PIPESRV_COMM_HEADER) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;
   BOOL                       bRet;
   LPTSTR                     cwd;
   LPTSTR                     argv0 = NULL;
   PVOID                      data;

   // determine program name
#ifdef UNICODE
   LPTSTR cmdLine = GetCommandLine();

   if( cmdLine )
   {
      // eat whitespace
      for( LPTSTR ptr = cmdLine;
           ( *ptr == _TEXT(' ') ) || ( *ptr == _TEXT('\t') );
           ptr++, cmdLine++ );

      BOOL bInQuote  = FALSE;
      int  lastQuote = 0;

      for( int cmdlen = 0; *cmdLine; cmdLine++, cmdlen++ )
      {
         if( *cmdLine == _TEXT('"') )
         {
            if( !bInQuote && !cmdlen ) ptr++, cmdlen--;
            lastQuote = cmdlen;
            bInQuote = !bInQuote;
         }
         else if( *cmdLine == _TEXT('\\') )
            cmdLine++, cmdlen++;
         else if( !bInQuote &&
                  ( *cmdLine == _TEXT(' ') ) || ( *cmdLine == _TEXT('\t') ) )
         {
            if( lastQuote == ( cmdlen - 1 ) )
               cmdlen--;
            break;
         }
      }

      if( !*cmdLine && ( lastQuote == ( cmdlen - 1 ) ) )
         cmdlen--;

      argv0 = (LPTSTR)LocalAlloc( LPTR, ( cmdlen + 1 ) * sizeof(TCHAR) );
      memcpy( argv0, ptr, cmdlen * sizeof(TCHAR) );
   }
#else
   if( __argv )
      argv0 = __argv[0];
#endif

   if( argv0 == NULL )
      return FALSE;

   // determine CWD
   cwd = _tgetcwd( NULL, MAXPATHLEN );
   if( cwd == NULL )
   {
#ifdef UNICODE
      LocalFree( argv0 );
#endif
      return FALSE;
   }

   LPTSTR filePart;
   DWORD  dwLen;
   BOOL   allocedPath = FALSE;
   LPTSTR progPath    = argv0;
   int    cwdSize     = ( _tcslen( cwd ) + 1 ) * sizeof(TCHAR);

   int len = _tcslen( argv0 );

   if( ( ( len < 2 ) || ( progPath[1] != _TEXT(':') ) ) &&
       ( _tcschr( progPath, _TEXT('\\') ) != NULL ) )
   {
      progPath = (LPTSTR)wss_malloc( ( len + 1 ) * sizeof(TCHAR) + cwdSize );
      allocedPath = TRUE;

      _stprintf( progPath, _TEXT("%s%s"), cwd, argv0 );
   }

   dwLen = SearchPath( NULL, argv0, _TEXT(".exe"), _MAX_PATH, g_progName, &filePart );

   if( dwLen == 0 )
   {
      free( cwd );
#ifdef UNICODE
      LocalFree( argv0 );
#endif /* UNICODE */
      return FALSE;
   }

   // compute sizes for each of CWD and path
   int progSize = ( _tcslen( g_progName ) + 1 ) * sizeof(TCHAR);

   // add data size to packet length
   commHdr.pktlen += progSize + cwdSize;

   // allocate a data buffer
   data = wss_malloc( commHdr.pktlen - sizeof(PIPESRV_COMM_HEADER) );

   memcpy( data, g_progName, progSize );
   memcpy( (char *)data + progSize, cwd, cwdSize );

   free( cwd );

#ifdef UNICODE
   // free argv0
   LocalFree( argv0 );
#endif

   // write data packet
   bRet = gwmcommWritePacket( (LPPIPESRV_COMM_HEADER)&commHdr, data );

   // free memory used for data buffer
   wss_free( data );

   // handle error condition from gwmcommWritePacket
   if( !bRet )
      return FALSE;

   // since we're in the loader, we have to synchronously wait for
   // a response ourselves, as the worker thread won't be ready to
   // run until we're out of it.
   DWORD                      dwID       = GWM_COMM_ID( *pCommHdr );
   HANDLE                     hWait      = gwmcommGetWaitHandle( dwID );
   LPPIPESRV_COMM_RESP_HEADER pInCommHdr =
      (LPPIPESRV_COMM_RESP_HEADER)wss_malloc( sizeof(PIPESRV_COMM_RESP_HEADER) );

   bRet = gwmcommReadPacket( pInCommHdr, (PVOID *)pBuf );

   gwmcommGetPacket( dwID, &pCommHdr, (PVOID *)pBuf );

   if( pCommHdr->pktlen >= ( sizeof(PIPESRV_COMM_RESP_HEADER) + sizeof(DWORD) +
                             ((UNLEN + 1) * sizeof( pd_char_t ))) )
   {
      pd_byte_t  *pData = (pd_byte_t *)*pBuf;

      g_proc.p_user_name = wss_strdup((pd_char_t *)pData);
      pData += ((UNLEN + 1) * sizeof( pd_char_t ));
      if( pFlags )
         *pFlags = *((DWORD *)pData);

      if( pCommHdr->pktlen == ( sizeof(PIPESRV_COMM_RESP_HEADER) +
                                sizeof(DWORD) ) )
         *((pd_byte_t **)pBuf) = NULL;
      else
         *((pd_byte_t **)pBuf) = (pd_byte_t *)pData + sizeof(DWORD);
   }
   else
   {
      *pFlags = 0;
      *((pd_byte_t **)pBuf) = NULL;
   }

   return bRet;
}

BOOL
gwmcommSendCreatingChild()
{
   PIPESRV_COMM_HEADER commHdr = { LWM_COMM_CREATING_CHILD, _getpid(),
                                   GetCurrentThreadId(),
#ifdef UNICODE
                                   GWM_COMM_UNICODE,
#else
                                   0,
#endif
                                   sizeof(PIPESRV_COMM_HEADER) };

   return gwmcommWritePacket( &commHdr, NULL );
}

BOOL
gwmcommSendCreatedChild( DWORD pid )
{
   PIPESRV_COMM_HEADER commHdr = { LWM_COMM_CREATED_CHILD, _getpid(),
                                   GetCurrentThreadId(),
#ifdef UNICODE
                                   GWM_COMM_UNICODE,
#else
                                   0,
#endif
                                   sizeof(PIPESRV_COMM_HEADER) +
                                   sizeof(DWORD) };

   return gwmcommWritePacket( &commHdr, &pid );
}

