/*! \file lwm.cxx
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 *  $Id: lwm.cxx,v 1.69 2001/04/05 18:40:37 stevek Exp $
 *
 *  \brief   Local Wrapper Manager
 *  \author  Larry Spector, Steve Kiernan
 *  \version $Revision: 1.69 $
 *  \date    $Date: 2001/04/05 18:40:37 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <windows.h>
#include <tchar.h>
#include <ole2.h>
#include <process.h>

// The fast BSTR helper class.
#include "msjdbg.h"
#include "FastBSTR.h"
#include "lwm.h"
#include "hook.h"
#include "module.h"
#include "gwmcomm.h"
#include LWMSTUBS_H
#include "wdbm.h"

#include <wss_init.h>
#include <wss_arg.h>
#include <wss_log.h>
#include <wss_malloc.h>
#include <wss_module.h>
#include <wss_mutex.h>
#include <wss_path.h>
#include <wss_process.h>
#include <wss_wrapper.h>

#include <linker_pe.h>

#include <vector>
#include <algorithm>
#include <string>

#define MAXBUFLEN 1024

// Because of the STL stuff, the standard namespace must be used.
using namespace std ;


/*//////////////////////////////////////////////////////////////////////
                             Externals
//////////////////////////////////////////////////////////////////////*/

extern DWORD WINAPI     gwmcommWorkThread      ( PVOID );
extern HOOKFUNCDESC & lwmstubGetHookFuncEntry( int index );

extern HANDLE             g_hReadThread;
extern WSS_MUTEX_PROTECT( g_callcount );
extern volatile DWORD g_callcount;
extern WSS_MUTEX_PROTECT( g_tracecount );
extern volatile DWORD g_tracecount;
extern volatile BOOL      g_bNoThrMsgs;
extern volatile BOOL      g_bNoProcMsgs;
extern volatile BOOL      g_bTraceCalls;
extern volatile BOOL      g_bCountCalls;

extern TCHAR g_progName[_MAX_PATH];

extern LPCWSTR _LWMStubs_ConvertASCIIToUnicode( LPCSTR pStr );
extern LPCSTR _LWMStubs_ConvertUnicodeToASCII( LPCWSTR pStr );

extern "C" {
int IMPORT wsi_export_init( void );
}

/*//////////////////////////////////////////////////////////////////////
                             File Constants
//////////////////////////////////////////////////////////////////////*/
// The initial size used for the BSTR wrapper class. 
const int k_INITIAL_BSTR_SIZE       = 2048 ;
// The modules where all the functions to hook are located. 
extern const char *k_MODULES_TO_HOOK[];

// The initial size of the module array to allocate. 
const int k_INITIAL_MODARRAYSIZE    = 250 ;
// The number of functions we always hook.
const int k_NUM_ALWAYS_HOOK_FUNCS = 8;

/*//////////////////////////////////////////////////////////////////////
                          File Only Prototypes
//////////////////////////////////////////////////////////////////////*/
// The function that does all the hooking. 
static BOOL lwmHookFuncs ( HOOKLIST hooklist );

// Determines if the HINSTANCE is a DLL that is hookable.
static BOOL lwmIsHookableDLL ( HINSTANCE hInst ) ;

// Determines if the HINSTANCE is a DLL that is hooked.
static BOOL lwmIsHookedDLL ( HINSTANCE hInst ) ;

// Handles getting the loaded module list.
static HMODULE * lwmAllocateLoadedModuleList ( LPUINT puiCount ) ;

static DWORD WINAPI lwmAsyncThread( PVOID args );

VOID WINAPI    LWM_Always_ExitProcess   ( UINT );
HMODULE WINAPI LWM_Always_LoadLibraryA  ( LPCSTR );
HMODULE WINAPI LWM_Always_LoadLibraryW  ( LPCWSTR );
HMODULE WINAPI LWM_Always_LoadLibraryExA( LPCSTR, HANDLE, DWORD );
HMODULE WINAPI LWM_Always_LoadLibraryExW( LPCWSTR, HANDLE, DWORD );
BOOL WINAPI    LWM_Always_CreateProcessW( LPCWSTR, LPWSTR,
                                          LPSECURITY_ATTRIBUTES,
                                          LPSECURITY_ATTRIBUTES,
                                          BOOL, DWORD, LPVOID, LPCWSTR,
                                          LPSTARTUPINFOW,
                                          LPPROCESS_INFORMATION );
BOOL WINAPI    LWM_Always_CreateProcessA( LPCSTR, LPSTR,
                                          LPSECURITY_ATTRIBUTES,
                                          LPSECURITY_ATTRIBUTES,
                                          BOOL, DWORD, LPVOID, LPCSTR,
                                          LPSTARTUPINFOA,
                                          LPPROCESS_INFORMATION );
FARPROC WINAPI LWM_Always_GetProcAddress( HMODULE, LPCSTR );

// The master table of functions names and the hook functions.  This
//  is read only for this file.
static HOOKFUNCDESC g_AlwaysHookFuncs[ k_NUM_ALWAYS_HOOK_FUNCS ] =
{
  HOOKFUNCDESC( "ExitProcess" ,    (PROC)LWM_Always_ExitProcess, "kernel32.dll"),
  HOOKFUNCDESC( "LoadLibraryA" ,   (PROC)LWM_Always_LoadLibraryA, "kernel32.dll"),
  HOOKFUNCDESC( "LoadLibraryExA" , (PROC)LWM_Always_LoadLibraryExA, "kernel32.dll"),
  HOOKFUNCDESC( "LoadLibraryExW" , (PROC)LWM_Always_LoadLibraryExW, "kernel32.dll"),
  HOOKFUNCDESC( "LoadLibraryW" ,   (PROC)LWM_Always_LoadLibraryW, "kernel32.dll"),
  HOOKFUNCDESC( "CreateProcessW" , (PROC)LWM_Always_CreateProcessW, "kernel32.dll"),
  HOOKFUNCDESC( "CreateProcessA" , (PROC)LWM_Always_CreateProcessA, "kernel32.dll"),
  HOOKFUNCDESC( "GetProcAddress",  (PROC)LWM_Always_GetProcAddress, "kernel32.dll")
} ;


/*//////////////////////////////////////////////////////////////////////
                         File Global Variables
//////////////////////////////////////////////////////////////////////*/
// The flag that indicates if the DLL is completely initialized. 
static BOOL             g_bInitialized      = FALSE ; 
// The HINSTANCE for this DLL so it does not get hooked. 
static HINSTANCE        g_hInstance         = NULL  ;

// The fast BSTR class.
static CFastBSTR        g_cGlobalBSTR               ;
// The vector that will hold the HMODULES that have been patched.
static vector<HMODULE>  g_vHookedModules            ;


// This is the map of names to HOOKFUNCDESC structures,
// it contains all functions that CAN be hooked.
HOOKMAP g_hookMap;
// This is the list of Hooks that WILL be put in place
HOOKLIST g_hooklist;

// Flag for signifying when wrappers have been activated
volatile BOOL     g_asyncOk = FALSE;

pd_char_t g_mutexName[MAXPATHLEN];
pd_char_t g_cvName[MAXPATHLEN];

proc_t g_proc = {0, 0};
wss_proc         *g_wrp = NULL;


BOOL APIENTRY DllMain( HINSTANCE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID /*lpReserved*/
  )
{
   switch( ul_reason_for_call )
   {
      case DLL_PROCESS_ATTACH:
         g_hInstance = hModule;

         DisableThreadLibraryCalls( g_hInstance );

         gwmcommSetup();

         if( lwmQueryActivation( hModule ) )
            lwmInit();
         else
            gwmcommShutdown();
         break;

      case DLL_PROCESS_DETACH:
         if( g_wrp )
         {
            g_asyncOk = FALSE;
            wss_proc_exit( &g_proc );
         }
         break;

      default:
         break;
   }
   return TRUE;
}


/*  panic */
/*! \brief Handle critical WSS errors
 *
 *  \warning When a panic occurs, there is something seriously wrong
 *           which needs to be fixed.
 *
 *  \param format printf-style format string
 */
void EXPORT
panic( const char *format, ... )
{
   va_list va;
   char msg[MAXBUFLEN];
   int nchars;

   va_start( va, format );
   nchars = _snprintf( msg, MAXBUFLEN, format, va );
   va_end( va );

   MessageBoxA( 0, msg, "Wrapper Message",
                ( MB_OK | MB_ICONEXCLAMATION | MB_DEFAULT_DESKTOP_ONLY ) );
}


BOOL lwmSetup ( void )
{
  /* Initialize the wss */
  wss_init( WSS_INIT_NO_DB | WSS_INIT_NO_CONF );

  linker_pe_init();
  wsi_export_init();

  g_proc.p_pid = _getpid();
  g_proc.p_ppid = 0;
  g_wrp = wss_proc_new(&g_proc);
  g_wrp->wrp_path = wss_path_abspath( g_progName );

  /*
   * Generate the program name, full path should have at least
   * one '\' character.  Program name is everything after the
   * last '\'.
   */
  int i = wsi_strlen( g_wrp->wrp_path );
  int len;

  for( len = 0, i = wsi_strlen( g_wrp->wrp_path );
       ( i > 0 ) && ( g_wrp->wrp_path[i] != WSS_FILE_SEPARATOR );
       i--, len++ )
     continue;

  /*
   * Check (for sanity, etc)
   */
  if( ( i == 0 ) && ( g_wrp->wrp_path[0] != WSS_FILE_SEPARATOR ) )
  {
     wsi_printf( PD_MAKESTR("wsi_execve: Pathname `%s' didn't have a '%c' ")
                 PD_MAKESTR("in it\n"), g_wrp->wrp_path, WSS_FILE_SEPARATOR );
     g_wrp->wrp_progname = NULL;
  }
  else
  {
     g_wrp->wrp_progname = wss_malloc_string( len );
     wsi_strncpy( g_wrp->wrp_progname, &( g_wrp->wrp_path[i + 1] ), len );
  }

  for( i = 0; i < k_NUM_HOOKED_FUNCS; i++ )
     g_hookMap.insert( HOOKMAP::value_type(lwmstubGetHookFuncEntry( i ).szFunc, 
                                           lwmstubGetHookFuncEntry( i )) );

  for( int j = 0; j < k_NUM_ALWAYS_HOOK_FUNCS; j++ )
     g_hooklist.push_back(&g_AlwaysHookFuncs[j]);

  WDBMInitialize();

  WSS_MUTEX_INIT( g_callcount );
  WSS_MUTEX_INIT( g_tracecount );

  return TRUE;
}

BOOL lwmQueryActivation(HINSTANCE hModule)
{
  TCHAR   IniFile[] = TEXT("c:\\lwm.ini"); 
  TCHAR   buf[_MAX_PATH]; 
  int     areEqual=0;

  HMODULE h= GetModuleHandle (NULL);

  GetModuleFileName (h, buf, sizeof(buf));

  /*
   * First check to see if wrappers are turned on at all
   */
  UINT active = GetPrivateProfileInt( TEXT("LWM"),
                                      TEXT("wrap"),
                                      NULL,
                                      IniFile );


/*
 *  call into the service to find what wrappers to load
 */
  if( active )
  {
     DWORD  flags = 0;
     LPTSTR lpszWrappers;
     if( !gwmcommRequestWrappers( &flags, &lpszWrappers ) ) 
     {
#ifdef DEBUG_LWM
        MessageBox( 0,
                    _T("gwmcommRequestWrappers failed in lwmQueryActivation"), 
                    _T("LWM!"), MB_OK | MB_ICONSTOP );
#endif
        return FALSE;
     }
     else
     {
        if( flags & PDNF_NOTHRMSGS )
           g_bNoThrMsgs = 1;
        if( flags & PDNF_NOPROCMSGS )
           g_bNoProcMsgs = 1;
        if( flags & PDNF_TRACECALLS )
           g_bTraceCalls = 1;
        if( flags & PDNF_COUNTCALLS )
           g_bCountCalls = 1;

        if( lpszWrappers && *lpszWrappers)
        {
          lwmSetup();
          while( *lpszWrappers )
          {
            lwmLoadWrapper( lpszWrappers );
            
            lpszWrappers += wsi_strlen( lpszWrappers ) + 1;
          }
        }
        else {
          return FALSE;
        }
     }
  }
  return active;
}

typedef pd_char_t *(* get_mod_nameFunc)(void);

/* wr_* declarations */
typedef CHAR **(* wr_getsymbolsFunc)(void);
typedef int (* wr_activateFunc) ( struct wss_proc *wrp, 
                                  struct wss_wrapper_data *data);



/*
 * lwmLoadWrapper()
 * For the named wrapper...
 * 1- load the correct wrapper DLL
 * 2- call wr_getsymbols (returns pre/post strings containing API 
 *                     names for interception)
 * 3- setup data structures to handle interception for the returned APIs 
 * 4- call wr_activate 
 */
BOOL lwmLoadWrapper( LPCTSTR wrapperName )
{
  int        error;
  wss_mod_t *wrm   = wss_mod_new( (pd_char_t *)wrapperName, &error );

  if( wrm )
  {
    wss_linker_file_t      *lf   = wrm->wrm_linked;
    wss_linker_file_ops_t  *ops  = lf->linker_file_ops;
    wss_linker_addr_t       wr_getsymbols;

    if( wss_linker_lookup_multiple( lf, FALSE,
                                    PD_MAKESTR("wr_getsymbols"),
                                    &wr_getsymbols,
                                    NULL ) )
       return FALSE;

    // Set up the system-dependent module information
    wsi_mod_t *wrmi = (wsi_mod_t *)wrm->wrm_opaque;
    wrmi->module = GetModuleHandle( wrapperName );

    // construct a new wrapper instance
    wss_wrapper_instance_t *wi;
    wss_proc_add_wrapper( g_wrp, lf->linker_file_name, &wi );
    wi->wi_data = NULL;

    CHAR **symbols = ((wr_getsymbolsFunc)wr_getsymbols.addr)();

    CHAR *symbolName;
    HOOKMAP::iterator hookIterator;

    int len = 0;
    BOOL usePre = 0, usePost = 0;

    /* decompose these symbol names, and mark them for hooking... */
    symbolName=symbols[0];
    if (!strcmp(symbolName, "PREPOST")) {
      usePre=TRUE;
      usePost=TRUE;
    } else if (!strcmp(symbolName, "PRE")) {
      usePre=TRUE;
    } else if(!strcmp(symbolName, "POST")) {
      usePost=TRUE;
    }

    for (int i=1; *symbols[i]!=0; i++) {
      symbolName=symbols[i];

      hookIterator=g_hookMap.find(symbolName);
      if (hookIterator != g_hookMap.end()){
        if (usePre)
          (*hookIterator).second.addPreFunc(wi,wi->wi_wrapper->w_dispatch_pre,i);
        if (usePost)
          (*hookIterator).second.addPostFunc(wi,wi->wi_wrapper->w_dispatch_post,i);
        g_hooklist.push_back(&(*hookIterator).second);
      }
      len = strlen(symbolName) - 1;
      if (!strcmp(symbolName+len, "W")) {
        if (strcmp(symbolName, "GetEnvironmentStringsW"))
          sprintf(symbolName+len, "A");
        else
          sprintf(symbolName+len, "\0");
        hookIterator=g_hookMap.find(symbolName);
        if (hookIterator != g_hookMap.end()){
          if (usePre)
            (*hookIterator).second.addPreFunc(wi,wi->wi_wrapper->w_dispatch_pre,i);
          if (usePost)
            (*hookIterator).second.addPostFunc(wi,wi->wi_wrapper->w_dispatch_post,i);
          g_hooklist.push_back(&(*hookIterator).second);
        }
        sprintf(symbolName+len, "W");
      }
    }
    wss_wrapper_instance_rele( wi );

    // remove any duplicate list entries
//    g_hooklist.unique();

    return TRUE;
  }

  return FALSE;
}


BOOL lwmInit ( void )
{
  // Just make sure the assumptions are correct.
  ASSERT ( NULL != g_hInstance ) ;

    // If the library has already been initialized, then just return
    //  TRUE.
  if ( TRUE == g_bInitialized )
  {
    SetLastErrorEx ( ERROR_ALREADY_INITIALIZED , SLE_ERROR ) ;
    return ( TRUE ) ;
  }

  // Get the global BSTR wrapper class cranked up.
  g_cGlobalBSTR.Allocate ( k_INITIAL_BSTR_SIZE ) ;

  // Add this DLL to the list of DLLs that have been hooked so that 
  // it does not ever get hooked.  This DLL has the only non-hooked 
  // versions of the imported functions. 
  g_vHookedModules.push_back ( g_hInstance ) ;

#ifdef _ALPHA_
  // Is the FX32 DLL in the process space?  If a debugger is present, 
  // then FX32 is not injected. 
  HINSTANCE hInstTemp = GetModuleHandle ( "FX32AGNT.DLL" ) ; 
  if ( NULL != hInstTemp ) {
    g_vHookedModules.push_back ( hInstTemp ) ;
  }
#endif  // _ALPHA_

  // OK, everything is kopacetic so all of the functions can be
  //  hooked.
  BOOL bRet = lwmHookFuncs ( g_hooklist);
  ASSERT ( TRUE == bRet ) ;

  // We really should handle error conditions
  if( wss_wrapper_call_activation( g_wrp ) )
     wsi_printf( PD_MAKESTR("Activate returned an error!\n") );

  // Start worker thread for communications with GWM
  g_hReadThread = CreateThread( NULL, 0, gwmcommWorkThread, NULL, 0,
                                NULL );

  if( g_hReadThread == NULL ) 
  {
#ifdef DEBUG_LWM
     MessageBox( 0, TEXT("Unable to spawn worker thread!"), TEXT("LWM!"),
                 MB_OK | MB_ICONSTOP );
#endif
  }
  else
     SetThreadPriority( g_hReadThread, THREAD_PRIORITY_ABOVE_NORMAL );

  g_bInitialized = TRUE ;

  // All OK, Jumpmaster!
  SetLastError ( ERROR_SUCCESS ) ;

  return ( bRet ) ;
}


static BOOL lwmHookFuncs ( HOOKLIST hooklist)
{

  if (hooklist.size()>0) {
    // Get the loaded modules.

    // The pointer to the HMODULEs in this process.
    HMODULE * pModArray = NULL ;
    // The real number of entries in the module array returned from
    //  lwmAllocateLoadedModuleList.
    UINT uiModCount = 0 ;

    // Now that the list of modules is in pModArray, run through the
    //  list hooking all the calls needed.
    pModArray = lwmAllocateLoadedModuleList ( &uiModCount ) ;
    ASSERT ( NULL != pModArray ) ;

    if ( NULL == pModArray )
    {
      return ( FALSE ) ;
    }

    // We really do _NOT_ want to hook these
    HMODULE hUs = GetModuleHandle( _TEXT("lwm.dll") );
    HMODULE hWSS = GetModuleHandle( _TEXT("wss.dll") );
    HMODULE hWrImp = GetModuleHandle( _TEXT("wrimp.dll") );

    // Loop through each module and hook the functions.
    for ( UINT i = 0 ; i < uiModCount ; i++ )
    {
      ASSERT ( NULL != pModArray[ i ] ) ;

      // Is this a module that is hookable?
      if ( TRUE == lwmIsHookableDLL ( pModArray [ i ] ) )
      {
         // We _REALLY_ do not want to hook ourself
         if( ( pModArray[ i ] == hUs ) ||
             ( pModArray[ i ] == hWSS ) ||
             ( pModArray[ i ] == hWrImp ) )
            continue;

         // Make sure we're not hooking the wrappers
         WSS_MUTEX_ENTER( wss_module_list );
         for( wss_mod_t *wrm = wss_module_list; wrm; wrm = wrm->wrm_next )
         {
            wsi_mod_t *wrmi = (wsi_mod_t *)wrm->wrm_opaque;
            if( wrmi->module == pModArray[ i ] )
            {
               wss_mod_ref( wrm );
               break;
            }
         }
         WSS_MUTEX_EXIT( wss_module_list );
         if( wrm )
         {
            wss_mod_rele( wrm );
            continue;
         }

#ifdef _DEBUG
        TCHAR szFile[ MAX_PATH ] ;
        GetModuleFileName ( pModArray[ i ] , szFile , MAX_PATH ) ;
        OutputDebugString( PD_MAKESTR("Looking to hook: "));
        OutputDebugString( szFile ) ;
        OutputDebugString( PD_MAKESTR("\n"));
#endif

        UINT uiHooked = 0 ;

        HookImportedFunctionsByName ( pModArray [ i ]   ,
                                      k_MODULES_TO_HOOK ,
                                      hooklist        ,
                                      &uiHooked          ) ;
        // Stick the current module in the skip list no matter if 
        // it failed. 
        g_vHookedModules.push_back ( pModArray [ i ] ) ;

        OutputDebugString( PD_MAKESTR("    Hooked a total of "));
        _RPT1(_CRT_WARN, "%d", uiHooked); 
        OutputDebugString( PD_MAKESTR(" functions\n"));

      }
    }

    // Free the memory that was allocated earlier.
    free ( pModArray ) ;
  }
  // All OK, Jumpmaster!
  return ( TRUE ) ;
}

BOOL ShutdownLWM ( void )
{
  // Just make sure the assumptions are correct.
  ASSERT ( NULL != g_hInstance ) ;
  ASSERT ( TRUE == g_bInitialized ) ;

    // Don't call this if it has not been initialized.
  if ( FALSE == g_bInitialized )
  {
    return ( FALSE ) ;
  }

  // Get rid of the memory used by the BSTR class.
  g_cGlobalBSTR.Free ( ) ;

  // reverse the process by setting pProc == origProc then
  // (un)hooking everything.
  HOOKLIST::iterator i;
  for ( i=g_hooklist.begin( ); 
        i !=g_hooklist.end() ; i++ )
  {
    (*i)->pProc=(*i)->origProc;
  }

  // Clear out the list of modules that have been hooked.  This way
  //  the HookTraceSrvHelperFuncs will treat everything as fresh and 
  // unhook them.
  g_vHookedModules.clear ( );
  g_vHookedModules.push_back ( g_hInstance ) ;

#ifdef _ALPHA_
  // Is the FX32 DLL in the process space?  If a debugger is present, 
  // then FX32 is not injected. 
  HINSTANCE hInstTemp = GetModuleHandle ( "FX32AGNT.DLL" ) ; 
  if ( NULL != hInstTemp ) {
    g_vHookedModules.push_back ( hInstTemp ) ;
  }
#endif  // _ALPHA_

  // This does the unhooking.
  BOOL bRet = lwmHookFuncs ( g_hooklist);

  // Reset outselves to a known state.
  g_bInitialized = FALSE ;
  memset ( g_pOrigFuncs , NULL , sizeof (PROC) * k_NUM_HOOKED_FUNCS );
  g_vHookedModules.clear ( );

  // All OK, Jumpmaster!
  SetLastError ( ERROR_SUCCESS ) ;
  return ( bRet ) ;
}


static HMODULE * lwmAllocateLoadedModuleList ( LPUINT puiCount )
{
  ASSERT ( NULL != puiCount ) ;
  ASSERT ( FALSE == IsBadReadPtr ( puiCount , sizeof ( LPUINT ) ) ) ; 
  if ( TRUE == IsBadReadPtr ( puiCount , sizeof ( LPUINT ) ) ) 
  {
    return ( NULL ) ;
  }

  // The HMODULE array.
  HMODULE * pModArray ;

  pModArray =
    (HMODULE*)malloc ( k_INITIAL_MODARRAYSIZE * sizeof ( HMODULE ) );

  ASSERT ( NULL != pModArray ) ;
  if ( NULL == pModArray )
  {
    return ( FALSE ) ;
  }

  // Bugslaying tip:  Set the memory to a known value that we can
  //  assert against later.
  memset ( pModArray ,
           NULL      ,
           k_INITIAL_MODARRAYSIZE * sizeof ( HMODULE ) ) ;

  // bRet holds BOOLEAN return values.
  BOOL bRet = GetLoadedModules ( GetCurrentProcessId ( ) ,
                                 k_INITIAL_MODARRAYSIZE  ,
                                 pModArray               ,
                                 puiCount                 ) ;
  // Save off the last error so that the assert can still fire and
  //  not change the value.
  DWORD dwLastError = GetLastError ( ) ;
  ASSERT ( TRUE == bRet ) ;

  // Check the return to see if we need to allocate a new buffer and
  //  try and load the values again.
  if ( FALSE == bRet )
  {
    // If the error was anything other than the buffer being too
    //  small, then there is serious trouble going on.
    ASSERT ( ERROR_INSUFFICIENT_BUFFER == dwLastError ) ;
    if ( ERROR_INSUFFICIENT_BUFFER != dwLastError )
    {
      return ( NULL ) ;
    }

    // Check that a decent value was returned for the real count.
    ASSERT ( *puiCount < k_INITIAL_MODARRAYSIZE ) ;
    if ( *puiCount >= k_INITIAL_MODARRAYSIZE )
    {
      return ( FALSE ) ;
    }

    // Get rid of the last buffer.
    free ( pModArray ) ;

    // Allocate a larger array.
    pModArray =
      (HMODULE*)malloc ( k_INITIAL_MODARRAYSIZE *
                         ( sizeof ( HMODULE ) * 2 ) ) ;

    ASSERT ( NULL != pModArray ) ;
    if ( NULL == pModArray )
    {
      return ( FALSE ) ;
    }

    // Bugslaying tip:  Set the memory to a known value that we can
    //  assert against later.
    memset ( pModArray ,
             NULL      ,
             k_INITIAL_MODARRAYSIZE * sizeof ( HMODULE ) * 2 ) ;

    // Try one last time to get the module list.  If there are 512
    //  total DLLs in an executable, the user needs more than help
    //  tracing!
    bRet = GetLoadedModules ( GetCurrentProcessId ( ) ,
                              k_INITIAL_MODARRAYSIZE  ,
                              pModArray               ,
                              puiCount                 ) ;

    // If it failed again, just bail now.
    ASSERT ( TRUE == bRet ) ;
    if ( FALSE == bRet )
    {
      free ( pModArray ) ;
      return ( NULL ) ;
    }
  }
  // All OK, Jumpmaster!
  return ( pModArray ) ;
}

static BOOL lwmIsHookableDLL ( HINSTANCE hInst )
{
  ASSERT ( NULL != hInst ) ;

  // TODO TODO
  //  Right now, this only looks for the bare minimum.  To speed up
  //  life this needs to get a little smarter and start skipping out
  //  on many of the system DLLs.

  // Do some minor checking to see if this is a valid HINSTANCE.
  //  To really make sure, I should actually look at the PE header.
  ASSERT ( FALSE == IsBadReadPtr ( hInst, sizeof ( IMAGE_DOS_HEADER ) ) );
  if ( TRUE == IsBadReadPtr ( hInst , sizeof ( IMAGE_DOS_HEADER ) ) ) 
  {
    return ( FALSE ) ;
  }

  // Loop through the list of DLLs already processed.  If it is in
  //  there, then return FALSE so it is not patched.
  for ( vector<HMODULE>::iterator i = g_vHookedModules.begin ( ) ;
        i != g_vHookedModules.end ( )                            ;
        i++                                     )
  {
    if ( hInst == *i )
    {
      return ( FALSE ) ;
    }
  }

  // Hook Away.
  return ( TRUE ) ;
}

static BOOL lwmIsHookedDLL ( HINSTANCE hInst )
{
  if ( NULL == hInst ) return FALSE;

  // Do some minor checking to see if this is a valid HINSTANCE.
  //  To really make sure, I should actually look at the PE header.
  ASSERT ( FALSE == IsBadReadPtr ( hInst, sizeof ( IMAGE_DOS_HEADER ) ) );
  if ( TRUE == IsBadReadPtr ( hInst , sizeof ( IMAGE_DOS_HEADER ) ) ) 
  {
    return ( FALSE ) ;
  }

  // Loop through the list of DLLs already processed.  If it is in
  //  there, then return TRUE
  for ( vector<HMODULE>::iterator i = g_vHookedModules.begin ( ) ;
        i != g_vHookedModules.end ( )                            ;
        i++                                     )
  {
    if ( hInst == *i )
    {
      return ( TRUE ) ;
    }
  }

  return ( FALSE ) ;
}

/*********************************************************************
 *
 *     Here are the Hooked function replacements
 *
 *********************************************************************/

VOID WINAPI
LWM_Always_ExitProcess ( UINT uExitCode )
{
  // If this library is still loaded after ExitProcess, it can get
  //  into states where the message loop and other supports are
  //  already gone.
  // To get around any problems, I just hook ExitProcess and shutdown 
  // as the last act of the EXE.  This means that there are some 
  //  ODS that could be lost, primarily those that are in DllMain 
  // DLL_PROCESS_DETACH processing and in static destructor 
  //  processing. Serious bummer. 
  if ( TRUE == g_bInitialized ) {
    _CrtDumpMemoryLeaks ( ) ;
    ShutdownLWM ( ) ;
  }
  ExitProcess ( uExitCode ) ;
}


FARPROC WINAPI
LWM_Always_GetProcAddress ( HMODULE hModule, LPCSTR lpProcName )
{
  if( g_bTraceCalls )
  {
    TCHAR   buf[_MAX_PATH]; 
    DWORD retVal = GetModuleFileName(hModule, buf, _MAX_PATH);

    LPVOID tmpProcName = 
      (LPVOID)((HIWORD(lpProcName)==0)? 
               (LPVOID) lpProcName : 
               (LPVOID) _LWMStubs_ConvertASCIIToUnicode( (LPCSTR)lpProcName ));

    WSS_MUTEX_ENTER( g_tracecount );
    wsi_printf( PD_MAKESTR("%6d:%s %s( 0x%x:\"%s\" 0x%x:\"%s\" )"),
                g_tracecount++,
                PD_MAKESTR("FARPROC"),
                PD_MAKESTR("LWM_Always_GetProcAddress"),
                hModule,
                buf,
                lpProcName,
                (LPCWSTR) ((HIWORD(lpProcName)==0)? 
                           PD_MAKESTR("<ordinal>") : tmpProcName)
               );
    WSS_MUTEX_EXIT( g_tracecount );
  }

  FARPROC pProc = NULL;

  if( lwmIsHookedDLL( hModule ) )
  {
    if (HIWORD(lpProcName)) {
      HOOKLIST::iterator i;
      for ( i=g_hooklist.begin( ); i !=g_hooklist.end() ; i++ )
      {
        if ( 0 == strcmpi ( (*i)->szFunc , lpProcName) &&
             ((*i)->getDllHandle() == hModule)) 
        {
          pProc = (FARPROC)((*i)->getProcAddr());
          break;
        }
      }
    } 
    else {
      // TBD- Handle functions that are imported by ordinal value.
      //      For now, let them fall through.
    }
  }
  if( pProc == NULL )
     pProc = GetProcAddress ( hModule, lpProcName );

  if( g_bTraceCalls )
    wsi_printf( PD_MAKESTR(" ==> 0x%1x\n"), pProc );

  return pProc;
}


HMODULE WINAPI
LWM_Always_LoadLibraryA( LPCSTR lpLibFileName )
{
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
      wsi_printf( PD_MAKESTR("%6d:%s %s( 0x%x:\"%s\" )"),
                  g_tracecount++,
                  PD_MAKESTR("HMODULE"),
                  PD_MAKESTR("LWM_ALWAYS_LoadLibraryA"),
                  lpLibFileName,
                  lpLibFileName
                );
      WSS_MUTEX_EXIT( g_tracecount );
   }

  HMODULE hRet = LWM_LoadLibraryA ( lpLibFileName ) ;

  if( g_bTraceCalls )
    wsi_printf( PD_MAKESTR(" ==> 0x%1x\n"), hRet );

  if ( TRUE == g_bInitialized )
  {
    // Patch anything that this module might have.  We need to do
    //  all the modules again, because you never know what else this
    //  single module brought in.
    BOOL bRet = TRUE ;
    if ( NULL != hRet )
    {
      bRet = lwmHookFuncs ( g_hooklist);
      ASSERT ( TRUE == bRet ) ;
    }

#ifdef DEBUG_LWM
    if ( FALSE == bRet )
    {
      // Since there is not much we can do here, just let the user
      //  know.
      MessageBox ( GetForegroundWindow ( )  ,
                   _T("lwmHookFuncs failed in LWM_LoadLibraryA"),
                   _T ( "LWM!" )  ,
                   MB_OK  | MB_ICONSTOP       ) ;
    }
#endif
  }
  return ( hRet ) ;
}

HMODULE WINAPI
LWM_Always_LoadLibraryW( LPCWSTR lpLibFileName )
{
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
      wsi_printf( PD_MAKESTR("%6d:%s %s( 0x%x:\"%s\" )"),
                  g_tracecount++,
                  PD_MAKESTR("HMODULE"),
                  PD_MAKESTR("LWM_ALWAYS_LoadLibraryW"),
                  lpLibFileName,
                  lpLibFileName
                );
      WSS_MUTEX_EXIT( g_tracecount );
   }

  HMODULE hRet = LWM_LoadLibraryW ( lpLibFileName ) ;

  if( g_bTraceCalls )
    wsi_printf( PD_MAKESTR(" ==> 0x%1x\n"), hRet );

  if ( TRUE == g_bInitialized )
  {
    // Patch anything that this module might have.  We need to do
    //  all the modules again, because you never know what else this 
    // single module brought in. 
    BOOL bRet = TRUE ; 
    if ( NULL != hRet ) {
      bRet = lwmHookFuncs (g_hooklist);
      ASSERT ( TRUE == bRet ) ;
    }

#ifdef DEBUG_LWM
    if ( FALSE == bRet ) {
      // Since there is not much we can do here, just let the user
      //  know.
      MessageBox ( GetForegroundWindow ( )  ,
                   _T("lwmHookFuncs failed in LWM_LoadLibraryW") ,
                   _T ( "LWM!" )  ,
                   MB_OK  | MB_ICONSTOP       ) ;
    }
#endif
  }
  return ( hRet ) ;
}


HMODULE WINAPI
LWM_Always_LoadLibraryExA( LPCSTR lpLibFileName,
                           HANDLE hFile,
                           DWORD  dwFlags
                         )
{
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
      wsi_printf( PD_MAKESTR("%6d:%s %s( 0x%x:\"%s\" 0x%x 0x%x )"),
                  g_tracecount++,
                  PD_MAKESTR("HMODULE"),
                  PD_MAKESTR("LWM_ALWAYS_LoadLibraryExA"),
                  lpLibFileName,
                  lpLibFileName,
                  hFile,
                  dwFlags
                );
      WSS_MUTEX_EXIT( g_tracecount );
   }

  // Always do the load.
  HMODULE hRet = LWM_LoadLibraryExA ( lpLibFileName , hFile , dwFlags ) ;

  if( g_bTraceCalls )
    wsi_printf( PD_MAKESTR(" ==> 0x%1x\n"), hRet );

  if ( TRUE == g_bInitialized )
  {
    // If the dwFlags value is anything but LOAD_LIBRARY_AS_DATAFILE
    //  then I will patch away.
    if ( LOAD_LIBRARY_AS_DATAFILE != dwFlags )
    {
      // Patch anything that this module might have.  We need to
      //  do all the modules again, because you never know what
      //  else this single module brought in.
      BOOL bRet = TRUE ;
      if ( NULL != hRet )
      {
        bRet = lwmHookFuncs ( g_hooklist );
        ASSERT ( TRUE == bRet ) ;
      }

#ifdef DEBUG_LWM
      if ( FALSE == bRet )
      {
        // Since there is not much we can do here, just let the
        //  user know.
        MessageBox ( GetForegroundWindow ( )  ,
                     _T("lwmHookFuncs failed in LWM_LoadLibraryExA"), 
                     _T ("LWM!")  , MB_OK  | MB_ICONSTOP ) ;
      }
#endif
    }
  }
  return ( hRet ) ;
}

HMODULE WINAPI
LWM_Always_LoadLibraryExW( LPCWSTR lpLibFileName,
                           HANDLE  hFile,
                           DWORD   dwFlags
                         )
{
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
      wsi_printf( PD_MAKESTR("%6d:%s %s( 0x%x:\"%s\" 0x%x 0x%x )"),
                  g_tracecount++,
                  PD_MAKESTR("HMODULE"),
                  PD_MAKESTR("LWM_ALWAYS_LoadLibraryExW"),
                  lpLibFileName,
                  lpLibFileName,
                  hFile,
                  dwFlags
                );
      WSS_MUTEX_EXIT( g_tracecount );
   }

  // Always do the load.
  HMODULE hRet = LWM_LoadLibraryExW ( lpLibFileName , hFile , dwFlags ) ;

  if( g_bTraceCalls )
    wsi_printf( PD_MAKESTR(" ==> 0x%1x\n"), hRet );

  if ( TRUE == g_bInitialized )
  {
    // If the dwFlags value is anything but LOAD_LIBRARY_AS_DATAFILE
    //  then I will patch away.
    if ( LOAD_LIBRARY_AS_DATAFILE != dwFlags )
    {
      BOOL bRet = TRUE ;
      if ( NULL != hRet )
      {
        bRet = lwmHookFuncs (g_hooklist);
        ASSERT ( TRUE == bRet ) ;
      }

#ifdef DEBUG_LWM
      if ( FALSE == bRet )
      {
        // Since there is not much we can do here, just let the
        //  user know.
        MessageBox ( GetForegroundWindow ( )  ,
                     _T("lwmHookFuncs failed in LWM_LoadLibraryExW"), 
                     _T ( "LWM!" )  , MB_OK  | MB_ICONSTOP       ) ;
      }
#endif
    }
  }
  return ( hRet ) ;
}

BOOL WINAPI
LWM_Always_CreateProcessA( LPCSTR lpApplicationName,
                           LPSTR lpCommandLine,
                           LPSECURITY_ATTRIBUTES lpProcessAttributes,
                           LPSECURITY_ATTRIBUTES lpThreadAttributes,
                           BOOL bInheritHandles,
                           DWORD dwCreationFlags,
                           LPVOID lpEnvironment,
                           LPCSTR lpCurrentDirectory,
                           LPSTARTUPINFOA lpStartupInfo,
                           LPPROCESS_INFORMATION lpProcessInformation
                         )
{
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
      wsi_printf( PD_MAKESTR("%6d:%s %s( 0x%x:\"%s\" 0x%x:\"%s\" 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x:\"%s\" 0x%x 0x%x )"),
                  g_tracecount++,
                  PD_MAKESTR("BOOL"),
                  PD_MAKESTR("LWM_Always_CreateProcessA"),
                  lpApplicationName,
                  lpApplicationName,
                  lpCommandLine,
                  lpCommandLine,
                  lpProcessAttributes,
                  lpThreadAttributes,
                  bInheritHandles,
                  dwCreationFlags,
                  lpEnvironment,
                  lpCurrentDirectory,
                  lpCurrentDirectory,
                  lpStartupInfo,
                  lpProcessInformation
                );
      WSS_MUTEX_EXIT( g_tracecount );
   }

   gwmcommSendCreatingChild();

   BOOL hRet = LWM_CreateProcessA( lpApplicationName, lpCommandLine,
                               lpProcessAttributes, lpThreadAttributes,
                               bInheritHandles, dwCreationFlags, lpEnvironment,
                               lpCurrentDirectory, lpStartupInfo,
                               lpProcessInformation );

   gwmcommSendCreatedChild( hRet ? lpProcessInformation->dwProcessId : 0 );

   if( g_bTraceCalls )
     wsi_printf( PD_MAKESTR(" ==> %s\n"), hRet ? PD_MAKESTR("TRUE") : PD_MAKESTR("FALSE") );

   return hRet;
}

BOOL WINAPI
LWM_Always_CreateProcessW( LPCWSTR lpApplicationName,
                           LPWSTR lpCommandLine,
                           LPSECURITY_ATTRIBUTES lpProcessAttributes,
                           LPSECURITY_ATTRIBUTES lpThreadAttributes,
                           BOOL bInheritHandles,
                           DWORD dwCreationFlags,
                           LPVOID lpEnvironment,
                           LPCWSTR lpCurrentDirectory,
                           LPSTARTUPINFOW lpStartupInfo,
                           LPPROCESS_INFORMATION lpProcessInformation
                         )
{
   if( g_bTraceCalls )
   {
      WSS_MUTEX_ENTER( g_tracecount );
      wsi_printf( PD_MAKESTR("%6d:%s %s( 0x%x:\"%s\" 0x%x:\"%s\" 0x%x 0x%x 0x%x 0x%x 0x%x 0x%x:\"%s\" 0x%x 0x%x )"),
                  g_tracecount++,
                  PD_MAKESTR("BOOL"),
                  PD_MAKESTR("LWM_Always_CreateProcessA"),
                  lpApplicationName,
                  lpApplicationName,
                  lpCommandLine,
                  lpCommandLine,
                  lpProcessAttributes,
                  lpThreadAttributes,
                  bInheritHandles,
                  dwCreationFlags,
                  lpEnvironment,
                  lpCurrentDirectory,
                  lpCurrentDirectory,
                  lpStartupInfo,
                  lpProcessInformation
                );
      WSS_MUTEX_EXIT( g_tracecount );
   }

   gwmcommSendCreatingChild();

   BOOL hRet = LWM_CreateProcessW( lpApplicationName, lpCommandLine,
                               lpProcessAttributes, lpThreadAttributes,
                               bInheritHandles, dwCreationFlags, lpEnvironment,
                               lpCurrentDirectory, lpStartupInfo,
                               lpProcessInformation );

   gwmcommSendCreatedChild( hRet ? lpProcessInformation->dwProcessId : 0 );

   if( g_bTraceCalls )
     wsi_printf( PD_MAKESTR(" ==> %s\n"), hRet ? PD_MAKESTR("TRUE") : PD_MAKESTR("FALSE") );

   return hRet;
}

