/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wdbm.cxx,v 1.28 2001/04/05 18:40:47 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: The Wrapper DB Manager
 *
 *          -- DB access routines for the Local Wrapper Manager
 *
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include <wrerrno.h>

#include <wss_malloc.h>
#include <wss_db.h>

#include <wss_mutex.h>

#include "gwmcomm.h"
#include "wdbm.h"
#include "hook.h"


/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/

extern "C" {
void      IMPORT  free_col_args                   __P(( wqcd_t, int ));
void      IMPORT  free_row_args                   __P(( wqrow_t, wqcd_t,
                                                        int ));
void      IMPORT  wss_db_free_criteria            __P(( wqsx_t ));
void      IMPORT  wss_db_free_wqusl               __P(( wqusl_t ));
pd_char_t IMPORT *wss_db_create_actual_table_name __P(( pd_char_t *, int,
                                                        pd_char_t * ));
}



/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

static int    WDBMCreateTable     ( pd_char_t *wrapper_name, int instance_id,
                                    pd_char_t *table_name, int free_col_defs,
                                    wqcd_t col_defs, int col_count,
                                    u_long *table_id );
static int    WDBMDropTable       ( u_long table_id );
static int    WDBMInsertTableRow  ( u_long table_id, int free_vals,
                                    wqrow_t vals );
static int    WDBMSelectTableRows ( u_long table_id, int free_scrit,
                                    wqsx_t scrit, int take_first, int order_by,
                                    int desc, wqtab_t out_table );
static int    WDBMDeleteTableRows ( u_long table_id, int free_scrit,
                                    wqsx_t scrit, int take_first );
static int    WDBMUpdateTableRows ( u_long table_id, int free_scripts,
                                    wqsx_t scrit, int take_first,
                                    wqusl_t set_script );
static u_long WDBMFindTableByNames( pd_char_t *wrapper_name, int instance_id,
                                    pd_char_t *table_name );

static void   WDBMAuditPost       ( u_long types, const void *buffer,
                                    size_t length);

/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static DB_HANDLERS g_dbHandlers =
{
   WDBMCreateTable,
   WDBMDropTable,
   WDBMInsertTableRow,
   WDBMSelectTableRows,
   WDBMDeleteTableRows,
   WDBMUpdateTableRows,
   WDBMFindTableByNames,
   
   WDBMAuditPost
};

static LPDB_CACHE g_dbCache = NULL; // we cache the database table information
                                    // (only id, name, column defs) locally
                                    // to assist in db calls.
WSS_MUTEX_PROTECT(g_dbCache);


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static int
_WDBMComputeExprValMemoryRequirements( sxval_t val )
{
   if( val )
   {
      int memReq = sizeof(sxval);

      switch( val->which )
      {
         case literal:
            if( ( val->dtype == pd_char_t_t ) && ( val->v.l.cval ) )
               memReq += ( wsi_strlen(val->v.l.cval) + 1 ) *
                         sizeof(pd_char_t);
            break;

         case field:
            break;

         case variable:
            // Not implemented yet.
            break;
      }

      return memReq;
   }
   else
      return 0;
}

static int
_WDBMComputeScritMemoryRequirements( wqsx_t scrit )
{
   if( scrit )
   {
      int memReq = sizeof(wqsx);

      switch( scrit->which )
      {
         case term:
            memReq += sizeof(tx);
            if( scrit->un.exp->left )
               memReq +=
                  _WDBMComputeExprValMemoryRequirements( scrit->un.exp->left );
            if( scrit->un.exp->right )
               memReq +=
                  _WDBMComputeExprValMemoryRequirements( scrit->un.exp->right );
            break;

         case nonterm:
            memReq += sizeof(ntx);
            if( scrit->un.children->left )
               memReq += _WDBMComputeScritMemoryRequirements( scrit->un.children->left );
            if( scrit->un.children->right )
               memReq += _WDBMComputeScritMemoryRequirements( scrit->un.children->right );
            break;
      }

      return memReq;
   }
   else
      return 0;
}

static int
_WDBMComputeUpdateExprMemoryRequirements( wquexp_t expr )
{
   if( expr )
   {
      int memReq = sizeof(wquexp);

      switch( expr->which )
      {
         case term:
            if( expr->w.terminal )
               memReq +=
                  _WDBMComputeExprValMemoryRequirements( expr->w.terminal );
            break;

         case nonterm:
            if( expr->w.e.left )
               memReq +=
                  _WDBMComputeUpdateExprMemoryRequirements( expr->w.e.left );
            if( expr->w.e.right )
               memReq +=
                  _WDBMComputeUpdateExprMemoryRequirements( expr->w.e.right );
            break;
      }

      return memReq;
   }
   else
      return 0;
}

static int
_WDBMComputeSetScriptMemoryRequirements( wqusl_t script )
{
   int memReq = 0;

   for( ; script; script = script->next )
   {
      memReq += sizeof(wqusl);

      if( script->left )
         memReq += _WDBMComputeExprValMemoryRequirements( script->left );
      if( script->right )
         memReq += _WDBMComputeUpdateExprMemoryRequirements( script->right );
   }

   return memReq;
}

static void
_WDBMCopyExprValToLinearBuffer( sxval_t val, pd_byte_t *data, DWORD *index,
                                int *memLeft )
{
   if( val )
   {
      sxval_t bufVal = (sxval_t)data;
      *index += sizeof(sxval);
      *memLeft -= sizeof(sxval);
      data += sizeof(sxval);

      if( *memLeft < 0 )
         panic( "Insufficient memory to copy expression value to linear "
                "buffer!" );

      memcpy( bufVal, val, sizeof(sxval) );

      switch( val->which )
      {
         case literal:
            if( ( val->dtype == pd_char_t_t ) && ( val->v.l.cval ) )
            {
               bufVal->v.l.cval = (pd_char_t *)*index;

               int len = ( wsi_strlen(val->v.l.cval) + 1 ) *
                         sizeof(pd_char_t);

               pd_char_t *str = (pd_char_t *)data;
               *index += len;
               *memLeft -= len;
               data += len;

               if( *memLeft < 0 )
                  panic( "Insufficient memory to copy string value to linear "
                         "buffer!" );

               memcpy( str, val->v.l.cval, len );
            }
            break;

         case field:
            break;

         case variable:
            // Not implemented yet.
            bufVal->v.v.ptr = NULL;
            break;
      }
   }
}

static void
_WDBMCopyScritToLinearBuffer( wqsx_t scrit, pd_byte_t *data, DWORD *index,
                              int *memLeft )
{
   if( scrit )
   {
      wqsx_t bufScrit = (wqsx_t)data;
      *index += sizeof(wqsx);
      *memLeft -= sizeof(wqsx);
      data += sizeof(wqsx);

      if( *memLeft < 0 )
         panic( "Insufficient memory to copy scrit to linear buffer!" );

      memcpy( bufScrit, scrit, sizeof(wqsx) );

      switch( scrit->which )
      {
         case term:
         {
            bufScrit->un.exp = (tx_t)*index;

            tx_t bufTx = (tx_t)data;
            *index += sizeof(tx);
            *memLeft -= sizeof(tx);
            data += sizeof(tx);

            if( *memLeft < 0 )
               panic( "Insufficient memory to copy terminal expression to "
                      "linear buffer!" );

            memcpy( bufTx, scrit->un.exp, sizeof(tx) );

            if( scrit->un.exp->left )
            {
               bufTx->left = (sxval_t)*index;
               _WDBMCopyExprValToLinearBuffer( scrit->un.exp->left, data,
                                               index, memLeft );
               data += ( *index - (DWORD)bufTx->left );
            }

            if( scrit->un.exp->right )
            {
               bufTx->right = (sxval_t)*index;
               _WDBMCopyExprValToLinearBuffer( scrit->un.exp->right, data,
                                               index, memLeft );
            }

            break;
         }

         case nonterm:
         {
            bufScrit->un.children = (ntx_t)*index;

            ntx_t bufNtx = (ntx_t)data;
            *index += sizeof(ntx);
            *memLeft -= sizeof(ntx);
            data += sizeof(ntx);

            if( *memLeft < 0 )
               panic( "Insufficient memory to copy non-terminal expression "
                      "to linear buffer!" );

            memcpy( bufNtx, scrit->un.children, sizeof(ntx) );

            if( scrit->un.children->left )
            {
               bufNtx->left = (wqsx_t)*index;
               _WDBMCopyScritToLinearBuffer( scrit->un.children->left, data,
                                             index, memLeft );
               data += ( *index - (DWORD)bufNtx->left );
            }

            if( scrit->un.children->right )
            {
               bufNtx->right = (wqsx_t)*index;
               _WDBMCopyScritToLinearBuffer( scrit->un.children->right, data,
                                             index, memLeft );
            }

            break;
         }
      }
   }
}

static void
_WDBMCopyUpdateExprToLinearBuffer( wquexp_t expr, pd_byte_t *data,
                                   DWORD *index, int *memLeft )
{
   if( expr )
   {
      wquexp_t nexpr = (wquexp_t)data;
      *index += sizeof(wquexp);
      *memLeft -= sizeof(wquexp);
      data += sizeof(wquexp);

      if( *memLeft < 0 )
         panic( "Insufficient memory to copy update expression to linear "
                "buffer!" );

      memcpy( nexpr, expr, sizeof(wquexp) );

      switch( expr->which )
      {
         case term:
            if( expr->w.terminal )
            {
               nexpr->w.terminal = (sxval_t)*index;

               _WDBMCopyExprValToLinearBuffer( expr->w.terminal, data,
                                               index, memLeft );
            }
            break;

         case nonterm:
            if( expr->w.e.left )
            {
               nexpr->w.e.left = (wquexp_t)*index;
               _WDBMCopyUpdateExprToLinearBuffer( expr->w.e.left, data,
                                                  index, memLeft );
               data += ( *index - (DWORD)nexpr->w.e.left );
            }

            if( expr->w.e.right )
            {
               nexpr->w.e.right = (wquexp_t)*index;
               _WDBMCopyUpdateExprToLinearBuffer( expr->w.e.right,
                                                  data, index, memLeft );
            }
            break;
      }
   }
}

static void
_WDBMCopySetScriptToLinearBuffer( wqusl_t set_script, pd_byte_t *data,
                                  DWORD *index, int *memLeft )
{
   if( set_script )
   {
      wqusl_t lastSetScript = NULL;

      for( ; set_script; set_script = set_script->next )
      {
         wqusl_t bufSetScript = (wqusl_t)data;
         if( lastSetScript )
            lastSetScript->next = (wqusl_t)*index;
         lastSetScript = bufSetScript;
         *index += sizeof(wqusl);
         *memLeft -= sizeof(wqusl);
         data += sizeof(wqusl);

         if( *memLeft < 0 )
            panic( "Insufficient memory to copy set script to linear buffer!" );

         memcpy( bufSetScript, set_script, sizeof(wqusl) );

         if( set_script->left )
         {
            bufSetScript->left = (sxval_t)*index;

            _WDBMCopyExprValToLinearBuffer( set_script->left, data, index,
                                            memLeft );
            data += ( *index - (DWORD)bufSetScript->left );
         }

         if( set_script->right )
         {
            bufSetScript->right = (wquexp_t)*index;
            _WDBMCopyUpdateExprToLinearBuffer( set_script->right, data,
                                               index, memLeft );
            data += ( *index - (DWORD)bufSetScript->right );
         }
      }
   }
}

static LPDB_CACHE
WDBMFindTable( u_long id )
{
   LPDB_CACHE cache;

   WSS_MUTEX_ENTER( g_dbCache );
   for( cache = g_dbCache; cache; cache = cache->next )
      if( cache->id == id )
         break;
   if( cache )
      cache->refs++;
   WSS_MUTEX_EXIT( g_dbCache );

   return cache;
}

static LPDB_CACHE
WDBMFindTableByName( pd_char_t *name )
{
   LPDB_CACHE cache;

   WSS_MUTEX_ENTER( g_dbCache );
   for( cache = g_dbCache; cache; cache = cache->next )
      if( !wsi_strcmp( cache->name, name ) )
         break;
   if( cache )
      cache->refs++;
   WSS_MUTEX_EXIT( g_dbCache );

   return cache;
}

static void
WDBMLockCacheEntry( LPDB_CACHE cache )
{
   WSS_MUTEX_ENTER( cache->id );
}

static void
WDBMUnlockCacheEntry( LPDB_CACHE cache )
{
   WSS_MUTEX_EXIT( cache->id );
}

static void
WDBMReleaseCacheEntry( LPDB_CACHE cache )
{
   WSS_MUTEX_ENTER( g_dbCache );

   cache->refs--;
   if( !cache->refs )
   {
      *cache->prev = cache->next;
      WSS_MUTEX_DESTROY( cache->id );
      wss_free( cache );
   }
   else if( cache->refs < 0 )
      wsi_printf( PD_MAKESTR("Warning: released one-too-many cache entry ")
                  PD_MAKESTR("references for table '%s'\n"), cache->name );

   WSS_MUTEX_EXIT( g_dbCache );
}

static LPDB_CACHE
WDBMRequestTable( u_long id )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_REQUEST_TABLE, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) +
                                           sizeof(u_long) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   LPDB_CACHE recvData = WDBMFindTable( id );
   if( recvData != NULL )
      return recvData;

   WSS_MUTEX_ENTER( g_dbCache );

   if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                              (PVOID)&id,
                              &pCommHdr,
                              (LPVOID *)&recvData,
                              NULL ) != GWMCOMM_TRANSACT_SUCCESS )
      return NULL;

   if( recvData )
   {
      recvData->name = (pd_char_t *)( (pd_byte_t *)recvData +
                                       (DWORD)recvData->name );
      recvData->colDefs = (wqcd_t)( (pd_byte_t *)recvData +
                                    (DWORD)recvData->colDefs );
      wqcd_t cd;
      int i;
      for( i = 0; i < recvData->colCount; i++ )
      {
         cd = recvData->colDefs + i;
         cd->name = (pd_char_t *)( (pd_byte_t *)recvData +
                                    (DWORD)cd->name );
      }

      WSS_MUTEX_INIT( recvData->id );
      recvData->refs = 2;

      recvData->prev = &g_dbCache;
      recvData->next = g_dbCache;
      g_dbCache = recvData;
   }

   WSS_MUTEX_EXIT( g_dbCache );

   return recvData;
}

static int
WDBMCreateTable( pd_char_t *wrapper_name, int instance_id,
                 pd_char_t *table_name, int free_col_defs, wqcd_t col_defs,
                 int col_count, u_long *table_id )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_CREATE_TABLE, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   // Determine size of data block to allocate
   int wnameLen = wsi_strlen( wrapper_name ) + 1;
   int tnameLen = wsi_strlen( table_name ) + 1;

   int dataLen = sizeof(struct create_block) +
                 ( wnameLen * sizeof(pd_char_t) ) +
                 ( tnameLen * sizeof(pd_char_t) ) +
                 ( sizeof(wqcd) * col_count );

   for( int i = 0; i < col_count; i++ )
      dataLen += ( wsi_strlen( (col_defs + i)->name ) + 1 ) *
                 sizeof(pd_char_t);

   // Construct data block
   struct create_block *sendData = (struct create_block *)wss_malloc( dataLen );
   sendData->wname = (pd_char_t *)(sendData + 1);
   sendData->instance = instance_id;
   sendData->tname = sendData->wname + wnameLen;
   sendData->count = col_count;
   sendData->cd = (wqcd_t)(sendData->tname + tnameLen );

   memcpy( sendData->wname, wrapper_name, wnameLen * sizeof(pd_char_t) );
   sendData->wname = (pd_char_t *)( (pd_byte_t *)sendData->wname -
                                     (UINT)sendData );

   memcpy( sendData->tname, table_name, tnameLen * sizeof(pd_char_t) );
   sendData->tname = (pd_char_t *)( (pd_byte_t *)sendData->tname -
                                     (UINT)sendData );

   memcpy( sendData->cd, col_defs, col_count * sizeof(wqcd) );

   pd_char_t *names = (pd_char_t *)(sendData->cd + col_count);

   for( i = 0; i < col_count; i++ )
   {
      wqcd_t cd      = col_defs + i;
      int    nameLen = wsi_strlen( cd->name ) + 1;

      memcpy( names, cd->name, nameLen * sizeof(pd_char_t) );
      (sendData->cd + i)->name = (pd_char_t *)( (pd_byte_t *)names -
                                                 (UINT)sendData );
      names += nameLen;
   }
   sendData->cd = (wqcd_t)( (pd_byte_t *)sendData->cd -
                            (UINT)sendData );

   // Release column definitions if flagged to do so
   if( free_col_defs )
      free_col_args( col_defs, col_count );

   // Transact packet
   commHdr.pktlen += dataLen;

   PVOID recvData = NULL;
   if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                              (PVOID)sendData,
                              &pCommHdr,
                              &recvData,
                              NULL ) != GWMCOMM_TRANSACT_SUCCESS )
      return -WRE_ERROR;

   // Decompose packet data
   int retval;
   if( pCommHdr->pktlen != ( sizeof(commHdr) + sizeof(u_long) ) )
      retval = -WRE_INVALID_PARAMETER;
   else
   {
      retval = pCommHdr->errorCode;

      if( ( retval >= 0 ) && table_id )
         *table_id = *((u_long *)recvData);
   }

   if( recvData )
      wss_free( recvData );

   return retval;
}

static int
WDBMDropTable( u_long table_id )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_DROP_TABLE, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) +
                                           sizeof(u_long) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   PVOID recvData = NULL;
   if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                              (PVOID)&table_id,
                              &pCommHdr,
                              &recvData,
                              NULL ) != GWMCOMM_TRANSACT_SUCCESS )
      return -WRE_ERROR;

   int retval;
   if( pCommHdr->pktlen < sizeof(commHdr) )
      retval = -WRE_INVALID_PARAMETER;
   else
      retval = pCommHdr->errorCode;

   if( recvData )
      wss_free( recvData );

   return retval;
}

static int
WDBMInsertTableRow( u_long table_id, int free_vals, wqrow_t in_rows )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_INSERT_TABLE_ROW, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   LPDB_CACHE cache = WDBMRequestTable( table_id );
   if( cache == NULL )
      return -WRE_NOTFOUND;

   WDBMLockCacheEntry( cache );

   int dataLen  = sizeof(struct insert_block);
   int rowCount = 0;
   for( wqrow_t row = in_rows; row; row = row->next, rowCount++ )
   {
      wqval_t val;
      wqcd_t  cd;
      int     i;

      dataLen += sizeof(wqrow) + ( cache->colCount * sizeof(wqval) );

      for( cd = cache->colDefs, val = row->vals, i = 0;
           i < cache->colCount;
           cd++, val++, i++ )
         if( cd->dtype == pd_char_t_t )
            dataLen += ( wsi_strlen( val->v.cval ) + 1 ) * sizeof(pd_char_t);
   }

   // Construct data block
   struct insert_block *sendData = (struct insert_block *)wss_malloc( dataLen );
   sendData->id = table_id;
   sendData->rowCount = rowCount;

   wqrow_t     nrow       = (wqrow_t)( sendData + 1 );
   wqval_t     val_offset = (wqval_t)( nrow + rowCount );
   pd_char_t *str_offset = (pd_char_t *)( val_offset +
                                            ( rowCount * cache->colCount ) );
   for( row = in_rows; row; row = row->next, nrow++ )
   {
      memcpy( nrow, row, sizeof(wqrow) );
          if( nrow->next )
                nrow->next = (wqrow_t)( (pd_byte_t *)(nrow + 1) - (DWORD)sendData );
      nrow->vals = (wqval_t)( (pd_byte_t *)val_offset - (DWORD)sendData );

      for( int i = 0; i < cache->colCount; i++ )
      {
         memcpy( (PVOID)(val_offset + i), (PVOID)(row->vals + i),
                 sizeof(wqval) );

         if( (cache->colDefs + i)->dtype == pd_char_t_t )
         {
            int len = ( wsi_strlen( (row->vals + i)->v.cval ) + 1 );

            memcpy( str_offset, (row->vals + i)->v.cval,
                    len * sizeof(pd_char_t) );
            (val_offset + i)->v.cval =
               (pd_char_t *)( (pd_byte_t *)str_offset - (DWORD)sendData );
            str_offset += len;
         }
      }
      val_offset += cache->colCount;
   }

   // Release rows if flagged to do so
   if( free_vals )
      free_row_args( in_rows, cache->colDefs, cache->colCount );

   WDBMUnlockCacheEntry( cache );
   WDBMReleaseCacheEntry( cache );

   // Transact packet
   commHdr.pktlen += dataLen;

   PVOID recvData = NULL;
   if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                              (PVOID)sendData,
                              &pCommHdr,
                              &recvData,
                              NULL ) != GWMCOMM_TRANSACT_SUCCESS )
      return -WRE_ERROR;

   // Decompose packet data
   int retval;
   if( pCommHdr->pktlen != sizeof(commHdr) )
      retval = -WRE_INVALID_PARAMETER;
   else
      retval = pCommHdr->errorCode;

   if( recvData )
      wss_free( recvData );

   return retval;
}

static int
WDBMSelectTableRows( u_long table_id, int free_scrit, wqsx_t scrit,
                     int take_first, int order_by, int desc,
                     wqtab_t out_table )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_SELECT_TABLE_ROWS, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   int memReq = _WDBMComputeScritMemoryRequirements( scrit ) +
                sizeof(struct select_block);
   struct select_block *sendData = (struct select_block *)wss_malloc( memReq );

   sendData->id = table_id;
   sendData->takeFirst = take_first;
   sendData->orderBy = order_by;
   sendData->desc = desc;

   DWORD index   = sizeof(struct select_block);
   int   memLeft = memReq - index;
   _WDBMCopyScritToLinearBuffer( scrit, (pd_byte_t *)(sendData + 1),
                                 &index, &memLeft );

   if( free_scrit )
      wss_db_free_criteria( scrit );

   // Transact packet
   commHdr.pktlen += memReq;

   PVOID recvData = NULL;
   if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                              (PVOID)sendData,
                              &pCommHdr,
                              &recvData,
                              NULL ) != GWMCOMM_TRANSACT_SUCCESS )
      return -WRE_ERROR;

   // Decompose packet data
   int retval;
   if( pCommHdr->pktlen < sizeof(commHdr) )
      retval = -WRE_INVALID_PARAMETER;
   else
   {
      retval = pCommHdr->errorCode;

      if( out_table && retval )
      {
         if( recvData )
         {
            wqtab_t tab = (wqtab_t)recvData;
            memcpy( out_table, tab, sizeof(wqtab) );

            tab->col_defs = (wqcd_t)( (pd_byte_t *)tab + (DWORD)tab->col_defs );

            // Copy out the column definitions
            out_table->col_defs = (wqcd_t)wss_malloc( sizeof(wqcd) *
                                                      tab->col_count );
            memcpy( out_table->col_defs, tab->col_defs,
                    sizeof(wqcd) * tab->col_count );
            for( int i = 0; i < tab->col_count; i++ )
            {
               wqcd_t cd = tab->col_defs + i;
               wqcd_t ncd = out_table->col_defs + i;

               if( cd->name )
               {
                  cd->name = (pd_char_t *)( (pd_byte_t *)tab +
                                             (DWORD)cd->name );

                  int len = ( wsi_strlen(cd->name) + 1 ) * sizeof(pd_char_t);
                  ncd->name = (pd_char_t *)wss_malloc( len );
                  memcpy( ncd->name, cd->name, len );
               }
               else
                  ncd->name = NULL;
            }

            // Copy out the rows
            tab->rows = (wqrow_t)( (pd_byte_t *)tab + (DWORD)tab->rows );
            tab->tail = (wqrow_t)( (pd_byte_t *)tab + (DWORD)tab->rows );

            wqrow_t *last = &(out_table->rows);
            wqrow_t  nrow;
            for( wqrow_t row = tab->rows;
                 row;
                 row = (wqrow_t)( row->next?((pd_byte_t *)tab + (DWORD)row->next )
                                           :NULL ),
                    last = &(nrow->next) )
            {
               *last = nrow = (wqrow_t)wss_malloc( sizeof(wqrow) );
               memcpy( nrow, row, sizeof(wqrow) );

               row->vals = (wqval_t)( (pd_byte_t *)tab + (DWORD)row->vals );
               nrow->vals = (wqval_t)wss_malloc( sizeof(wqval) * tab->col_count );

               memcpy( nrow->vals, row->vals, sizeof(wqval) * tab->col_count );

               for( i = 0; i < tab->col_count; i++ )
               {
                  wqcd_t cd = tab->col_defs + i;
                  wqval_t val  = row->vals + i;
                  wqval_t nval = nrow->vals + i;

                  if( cd->dtype == pd_char_t_t )
                  {
                     val->v.cval = (pd_char_t *)( (pd_byte_t *)tab +
                                                   (DWORD)val->v.cval );
                     int len = ( wsi_strlen(val->v.cval) + 1 ) *
                               sizeof(pd_char_t);
                     nval->v.cval = (pd_char_t *)wss_malloc( len );
                     memcpy( nval->v.cval, val->v.cval, len );
                  }
               }
            }
         }
         else
         {
            out_table->col_defs = NULL;
            out_table->rows = NULL;
            out_table->col_count = 0;
            out_table->max_rows = 0;
            out_table->tail = NULL;
            out_table->next = NULL;
         }
      }
   }

   if( recvData )
      wss_free( recvData );

   return retval;
}

static int
WDBMDeleteTableRows( u_long table_id, int free_scrit, wqsx_t scrit,
                     int take_first )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_DELETE_TABLE_ROWS, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   int memReq = _WDBMComputeScritMemoryRequirements( scrit ) +
                sizeof(struct delete_block);
   struct delete_block *sendData = (struct delete_block *)wss_malloc( memReq );

   sendData->id = table_id;
   sendData->takeFirst = take_first;

   DWORD index   = sizeof(struct delete_block);
   int   memLeft = memReq - index;
   _WDBMCopyScritToLinearBuffer( scrit, (pd_byte_t *)(sendData + 1),
                                 &index, &memLeft );

   if( free_scrit )
      wss_db_free_criteria( scrit );

   // Transact packet
   commHdr.pktlen += memReq;

   PVOID recvData = NULL;
   if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                              (PVOID)sendData,
                              &pCommHdr,
                              &recvData,
                              NULL ) != GWMCOMM_TRANSACT_SUCCESS )
      return -WRE_ERROR;

   // Decompose packet data
   int retval;
   if( pCommHdr->pktlen != sizeof(commHdr) )
      retval = -WRE_INVALID_PARAMETER;
   else
      retval = pCommHdr->errorCode;

   if( recvData )
      wss_free( recvData );

   return retval;
}

static int
WDBMUpdateTableRows( u_long table_id, int free_scripts, wqsx_t scrit,
                     int take_first, wqusl_t set_script )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_UPDATE_TABLE_ROWS, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   int memReq = _WDBMComputeScritMemoryRequirements( scrit ) +
                sizeof(struct update_block) +
                _WDBMComputeSetScriptMemoryRequirements( set_script );

   struct update_block *sendData = (struct update_block *)wss_malloc( memReq );

   sendData->id = table_id;
   sendData->takeFirst = take_first;

   DWORD index   = sizeof(struct update_block);
   int   memLeft = memReq - index;
   _WDBMCopyScritToLinearBuffer( scrit, (pd_byte_t *)sendData + index,
                                 &index, &memLeft );

   sendData->setScript = (wqusl_t)index;
   wqusl_t debugScript = (wqusl_t)( (pd_byte_t *)sendData + index );
   _WDBMCopySetScriptToLinearBuffer( set_script, (pd_byte_t *)sendData + index,
                                     &index, &memLeft );

   if( free_scripts )
   {
      wss_db_free_criteria( scrit );
      wss_db_free_wqusl( set_script );
   }

   // Transact packet
   commHdr.pktlen += memReq;

   PVOID recvData = NULL;
   if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                              (PVOID)sendData,
                              &pCommHdr,
                              &recvData,
                              NULL ) != GWMCOMM_TRANSACT_SUCCESS )
      return -WRE_ERROR;

   // Decompose packet data
   int retval;
   if( pCommHdr->pktlen != sizeof(commHdr) )
      retval = -WRE_INVALID_PARAMETER;
   else
      retval = pCommHdr->errorCode;

   if( recvData )
      wss_free( recvData );

   return retval;
}

static u_long
WDBMFindTableByNames( pd_char_t *wrapper_name, int instance_id,
                      pd_char_t *table_name )
{
   PIPESRV_COMM_RESP_HEADER   commHdr  = { LWM_COMM_FIND_TABLE, _getpid(),
                                           GetCurrentThreadId(),
#ifdef UNICODE
                                           GWM_COMM_UNICODE,
#else
                                           0,
#endif
                                           sizeof(PIPESRV_COMM_HEADER) };
   LPPIPESRV_COMM_RESP_HEADER pCommHdr = &commHdr;

   u_long retval = 0;

   pd_char_t *fullName = wss_db_create_actual_table_name( wrapper_name,
                                                           instance_id,
                                                           table_name );
   if( fullName )
   {
      LPDB_CACHE cache = WDBMFindTableByName( fullName );
      if( cache )
      {
         WDBMLockCacheEntry( cache );
         retval = cache->id;
         WDBMUnlockCacheEntry( cache );
         WDBMReleaseCacheEntry( cache );
      }
      else
      {
         // Transact packet
         commHdr.pktlen += ( wsi_strlen(fullName) + 1 ) * sizeof(pd_char_t);

         PVOID recvData = NULL;
         if( gwmcommTransactPacket( (LPPIPESRV_COMM_HEADER)&commHdr,
                                    (PVOID)fullName,
                                    &pCommHdr,
                                    &recvData,
                                    NULL ) != GWMCOMM_TRANSACT_SUCCESS )
            return 0;

         // Decompose packet data
         if( pCommHdr->pktlen != sizeof(commHdr) )
         {
            if( pCommHdr->pktlen != sizeof(commHdr) + sizeof(int) )
               retval = 0;
            else
               retval = *(int *)recvData;
         }
         else
            retval = pCommHdr->errorCode;

         if( recvData )
            wss_free( recvData );
      }
   }

   return retval;
}

static void
WDBMAuditPost( u_long types, const void *buffer, size_t length)
{
   PIPESRV_COMM_HEADER commHdr = { LWM_COMM_AUDIT_POST, _getpid(),
                                   GetCurrentThreadId(),
#ifdef UNICODE
                                   GWM_COMM_UNICODE,
#else
                                   0,
#endif
                                   0 };
   struct event_block *packet;

   // Determine size of data block to allocate
   // '-1' comes from the size of the buffer at the end of the packet (1 char)
   commHdr.pktlen = sizeof(*packet)-1 + length;

   // Construct data block, simple part first.
   packet              = (struct event_block*)wss_malloc( commHdr.pktlen );

   packet->types       = types;
   packet->length      = length;
   // Copy in the opaque
   memcpy( packet->buffer, buffer, length );

   // Write packet
   gwmcommWritePacket( &commHdr, (PVOID)packet );

   if( packet )
      wss_free( packet );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

BOOL
WDBMInitialize( void )
{
   WSS_MUTEX_INIT( g_dbCache );
   return WRImpRegisterDBHandlers( &g_dbHandlers );
}

BOOL
WDBMShutdown( void )
{
   WSS_MUTEX_ENTER( g_dbCache );
   BOOL bRet = WRImpRegisterDBHandlers( NULL );
   WSS_MUTEX_EXIT( g_dbCache );
   WSS_MUTEX_DESTROY( g_dbCache );
   return bRet;
}

BOOL
WDBMHandleDropTableEvent( u_long table_id )
{
   LPDB_CACHE cache = WDBMFindTable( table_id );

   if( cache == FALSE )
      return FALSE;

   WDBMReleaseCacheEntry( cache );
   WDBMReleaseCacheEntry( cache );

   return TRUE;
}

