package Build::Component;

require 5.004;
use strict;
use vars qw(@EXPORT @ISA $VERSION);
use Carp;

$VERSION   = 0.001;
@ISA       = qw(Exporter);
@EXPORT    = qw(COMP_NAMES COMP_SETS
                COMP_GEN_INIT COMP_GEN_TEARDOWN COMP_GEN_FORCE
                getcompset compset generatesets);

my @sets = ();
my %sets = ();

sub COMP_GEN_INIT { 0x01; }
sub COMP_GEN_TEARDOWN { 0x02; }
sub COMP_GEN_FORCE { 0x04; }

sub COMP_NAMES { keys %sets; }
sub COMP_SETS  { values %sets; }

sub getcompset ($) {
   my $set = shift;
   croak 'Set "'.$set.'" does not exists' unless exists $sets{$set};
   $sets{$set};
}

sub compset {
   confess 'Usage: '.__PACKAGE__.'::compset <set name>, <symbol pattern>, <generator sub>'
      unless @_ == 3;
   my($set,$pattern,$generator) = @_;

   croak 'Set "'.$set.'" already exists' if exists $sets{$set};
   push( @sets, $sets{$set} = new Build::Component::Set( $set, $pattern, $generator ));
   $sets{$set};
}

sub generatesets {
   my $prefix;
   my $set;

   $prefix = $_[0].'_' if @_;

   # Generate the initialization and teardown routines for each set
   foreach $set ( @sets ) {
      my $name = $set->name;
      my $subname = lc($name);
      $subname =~ y/ /_/;
      my $init = '__compset_'.$prefix.$subname;
      my $teardown = '__compset_'.$prefix.'teardown_'.$subname;
      undef *generator;
      *generator = $set->generator;

      if( $set->known_symbols ) {
         print <<EOF;

/\* $name \*/
static int
$init( const pd_char_t \*group )
{
   int retc = 0;
EOF
         my @syms;
         my $sym;
         foreach $sym ( $set->known_symbols ) {
            push( @syms, $sym )
              if generator(COMP_GEN_INIT, 'group', 'retc', $set, $sym->name, $sym);
         }
         foreach $sym ( @syms ) {
            generator(COMP_GEN_INIT | COMP_GEN_FORCE, 'group', 'retc', $set,
                      $sym->name, $sym);
         }
         print <<EOF;
   return (retc);
}

/\* $name \*/
static int
$teardown( const pd_char_t \*group )
{
   int retc = 0;
EOF
         foreach $sym ( $set->known_symbols ) {
            generator(COMP_GEN_TEARDOWN | COMP_GEN_FORCE, 'group', 'retc', $set,
                      $sym->name, $sym);
         }
         print <<EOF;
   return (retc);
}
EOF
      }
   }

   # Generate the exported teardown routine which calls the various
   # set-specific teardown routines
   print <<EOF;

int
__gsw_compset_${prefix}teardown_routine( const pd_char_t \*group )
{
   int retc = 0;
EOF

   foreach $set ( reverse @sets ) {
      next unless $set->known_symbols;
      my $name = $set->name;
      my $teardown = '__compset_'.$prefix.'teardown_'.lc($name);
      $teardown =~ y/ /_/;
      print <<EOF;
   retc = $teardown(group);
   if( retc ) return retc;
EOF
   }

   print <<EOF;
   return (retc);
}
EOF

   # Generate the exported initialization routine which calls the
   # various set-specific initialization routines
   print <<EOF;

int
__gsw_compset_${prefix}init_routine( const pd_char_t \*group )
{
   int retc = 0;
EOF

   my $have_components;
   my $num_sets;
   my @donesets;
   foreach $set ( @sets ) {
      next unless $set->known_symbols;
      $num_sets++;
      my $name = $set->name;
      $have_components = 1 if $name eq 'Components';
      my $init = '__compset_'.$prefix.lc($name);
      $init =~ y/ /_/;
      print '   retc = '.$init.'(group);'."\n";
      foreach( reverse @donesets ) {
         next unless $_->known_symbols;
         my $teardown = '__compset_'.$prefix.'teardown_'.lc($_->name);
         $teardown =~ y/ /_/;
         print '   if( retc ) '.$teardown.'(group);'."\n";
      }
      print '   if( retc ) return (retc);'."\n";
      push( @donesets, $set );
   }

   if( $have_components ) {
      print <<EOF;

   /* Bootstrap the components which are not yet initialized */
   retc = gswk_component_initialize();
EOF
   }

   if( $num_sets ) {
      print <<EOF;
   if( retc ) __gsw_compset_${prefix}teardown_routine( group );
EOF
   }

   print <<EOF;
   return (retc);
}

/\* End of generated code \*/
EOF
}


package Build::Component::Set;

use strict;
use vars qw(@ISA $VERSION);
use Carp;
use BinFile::Symbol;

$VERSION = 0.001;
@ISA     = qw(Exporter);

sub new {
   my $this = shift;
   my $class = ref($this) || $this;
   confess 'Usage: '.__PACKAGE__.'->new( <set name>, <symbol pattern>, <generator sub> )' . ref($_[2])
      unless @_ == 3 && !ref($_[1]) && ref($_[2]) eq 'CODE';

   my($name,$pattern,$generator) = @_;
   my $self = {
      name      => $name,
      pattern   => $pattern,
      generator => $generator,
      symbols   => [],
      unknowns  => {},
      knowns    => {}
   };
   bless $self, $class;
   $self;
}

sub name ($) {
   my $this = shift;
   $this->{'name'};
}

sub generator ($) {
   my $this = shift;
   $this->{'generator'};
}

sub pattern ($) {
   my $this = shift;
   $this->{'pattern'};
}

sub symbols ($) {
   my $this = shift;
   @{$this->{'symbols'}};
}

sub known_names ($) {
   my $this = shift;
   keys %{$this->{'knowns'}};
}

sub known_symbols ($) {
   my $this = shift;
   values %{$this->{'knowns'}};
}

sub unknown_names ($) {
   my $this = shift;
   keys %{$this->{'unknowns'}};
}

sub unknown_symbols ($) {
   my $this = shift;
   values %{$this->{'unknowns'}};
}

sub addsymbol ($$) {
   my $this = shift;
   my $sym = shift;
   croak 'Symbol argument must be a class of BinFile::Symbol'
      unless $sym->isa('BinFile::Symbol');
   my $name = $sym->name;
   push( @{$this->{'symbols'}}, $sym );
   if( !exists $this->{'knowns'}->{$name} ) {
      if( $sym->isUnknown ) {
         $this->{'unknowns'}->{$name} = $sym;
      }
      else {
         $this->{'knowns'}->{$name} = $sym;
         delete $this->{'unknowns'}->{$name};
      }
   }
   $sym;
}

1;
