#!/usr/local/bin/perl -w
# 
# Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
# All rights reserved.
#
# This file is part of the NAI Generic Software Wrappers Toolkit
# (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
# distribution.
#
# The GSWTK is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The GSWTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: ntdllconst.pl,v 1.17 2001/04/05 18:40:52 stevek Exp $
#
# Author: Steve Kiernan
#
# Purpose: Script to extract service IDs for NT native calls from ntdll.dll
#          utilizing a disassembly-codescan approach
#

require 5.004;
use strict;
use CPU::x86::Opcodes qw(:FIELDS);
use BinFile::PE;
use IO::Seekable;
use Fcntl;
use BinFile::PE::Directory::Export;
use Config;

# Functions that we want to find the Service IDs
my $functions = "^Zw";

# Flag to enable/disable disassembly output
my $disassemble = 0;

# Split PATH into parts
my @searchpath = split($Config{'path_sep'}, $ENV{'PATH'});

die "Usage: $0 <outfile>" if !@ARGV;
my $header = $ARGV[0];


# Unpack a native value to Perl numeric form
sub unpackNative {
   my( $bytes, $packed ) = @_;

   # Unpack value to native format
   my $value;
   foreach( $bytes ) {
      /^1$/ && do { $value = unpack('C', $packed); last; };
      /^2$/ && do { $value = unpack('n', $packed); last; };
      /^3$/ && do { my @pp = unpack('C3', $packed);
                    push(@pp, 0);
                    $value = unpack('N', pack('C4', @pp));
                    last; };
      /^4$/ && do { $value = unpack('N', $packed); last; };
   }
   $value;
}

# Convert a bit vector to number
sub convertBitVector {
   my( $bytes, $bitvec ) = @_;

   my $packed = pack('B*', $bitvec);

   unpackNative( $bytes, $packed );
}


my %lookup;
foreach( @x86_opcodes ) {
   my( $bytes, $mask, $pattern ) = ( $_->[2], $_->[3], $_->[4] );

   $mask = convertBitVector( $bytes, $mask );
   $pattern = convertBitVector( $bytes, $pattern );

   $lookup{$bytes} = {} unless exists $lookup{$bytes};
   $lookup{$bytes}{$mask} = {} unless exists $lookup{$bytes}{$mask};
   $lookup{$bytes}{$mask}{$pattern} = $_;
}

sub openFile {
   my( $filename ) = @_;
   my $fp;

   foreach( @searchpath ) {
      my $name = $_ . '\\' . $filename;
      return \*fp if( sysopen( *fp, $name, O_RDONLY|eval('O_BINARY') ) );
   }
}

my $fp = openFile( "ntdll.dll" );
die "ntdll.dll: $!" if !$fp;
my $dll = new BinFile::PE( \*fp );

# Extract a value of the appropriate byte-length
sub extractValue {
   my( $ref, $offset, $bytes ) = @_;
   my %mach = %$ref;
   my @bytes = @{$mach{'Text'}};

   $offset += $mach{'PC'};

   # Pack value based on number of bytes being scanned
   my $packed = pack( 'C'.$bytes, @bytes[$offset .. ($offset + $bytes - 1)] );

   unpackNative( $bytes, $packed );
}

# Extract a value of the appropriate byte-length in Little-Endian order
sub extractEndianValue {
   my( $ref, $offset, $bytes ) = @_;
   my %mach = %$ref;
   my @bytes = @{$mach{'Text'}};

   $offset += $mach{'PC'};

   # Pack value based on number of bytes being scanned
   my $packed = pack( 'C'.$bytes, @bytes[$offset .. ($offset + $bytes - 1)] );

   unpackNative( $bytes, $packed );
}

# Generate mask and required shift
sub makeMask {
   my( $bchar, $bits ) = @_;

   my @bits = map(chr, unpack("C*", $bits));

   my( $mask, $shift, $bit ) = (0,0,0);
   foreach $bit ( @bits ) {
      $shift++;
      $mask <<= 1;
      $shift = 0, $mask |= 1 if $bit eq $bchar;
   }

   my $bytes = @bits;
   $bytes /= 8;

   [ $bytes, $shift, $mask ];
}

# Grab value based on bitmask and shift
sub grabValue {
   my( $ref, $bytes, $pairref ) = @_;

   return 0 if( !defined($pairref) );

   my( $shift, $mask ) = @$pairref;

   my $value = extractValue( $ref, 0, $bytes );
   $value &= convertBitVector( $bytes, $mask );
   $value >>= $shift;
   $value;
}

# Update opcode attributes
sub updateOpcode {
   my( $ref, $entry ) = @_;
   my %mach = %$ref;

   my $bytes = $entry->[2];

   my $pairref = $entry->[5]->{'RegType'}; 
   $mach{'Opcode'}{'RegType'}    =  grabValue( $ref, $bytes, $pairref );

   $pairref = $entry->[5]->{'SegmentReg'};
   $mach{'Opcode'}{'SegmentReg'} =  grabValue( $ref, $bytes, $pairref );

   $pairref = $entry->[5]->{'OpcodeFunc'};
   $mach{'Opcode'}{'OpcodeFunc'} =  grabValue( $ref, $bytes, $pairref );

   $pairref = $entry->[5]->{'IndexFunc'};
   $mach{'Opcode'}{'IndexFunc'} =   grabValue( $ref, $bytes, $pairref );

   if( defined($entry->[5]->{'RegType'}) ) {
      $mach{'Opcode'}{'IndexReg'} = undef;

      $pairref = $entry->[5]->{'Reg'};
      $mach{'Opcode'}{'Reg'} =      grabValue( $ref, $bytes, $pairref );
   }
   else {
      $mach{'Opcode'}{'Reg'} = undef;

      $pairref = $entry->[5]->{'IndexReg'};
      $mach{'Opcode'}{'IndexReg'} = grabValue( $ref, $bytes, $pairref );
   }
}

# Get an argument
sub getArgument {
   my( $ref, $arg ) = @_;
   my %mach = %$ref;
   my @bytes = @{$mach{'Text'}};

   foreach( $arg ) {
      /^Acc$/ &&
         do {
            if( $mach{'Opcode'}{'RegType'} ) { return [ 'EAX', 0 ]; }
            else { return [ 'AL', 0 ]; }
         };

      /^MemOfs$/ &&
         do {
            my @rb = @bytes[$mach{'NPC'} .. $mach{'NPC'}+1];
            my $addr;
            $addr = $mach{'Segment'}{'Selector'} if $mach{'Segment'}{'Byte'};
            $addr .= sprintf('0x%x', unpack('v', pack( 'C2', @rb ) ));
            return [ $addr, 4 ];
         };

      /^RegWord$/ &&
         do {
            my $ireg = $mach{'Opcode'}{'IndexReg'};
            my $bits = substr(unpack('B8', pack('C', $ireg)),5);
            return [ $x86_index_registers{$bits}, 0 ];
         };

      /^Reg$/ &&
         do {
            my $w = $mach{'Opcode'}{'RegType'};
            if( defined $w ) {
               my $reg = $mach{'Opcode'}{'Reg'};
               my $bits = substr(unpack('B8', pack('C', $reg)),5);
               return [ $x86_registers{$w}{$bits}, 0 ];
            }
            else {
               my $ireg = $mach{'Opcode'}{'IndexReg'};
               my $bits = substr(unpack('B8', pack('C', $ireg)),5);
               return [ $x86_index_registers{$bits}, 0 ];
            }
         };

      /^(Mem|MemWord)$/ &&
         do {
            my $memtype = $_;
#            print STDERR "OpcodeFunc: ".$mach{'Opcode'}{'OpcodeFunc'}."\n";
            foreach( $mach{'Opcode'}{'OpcodeFunc'} ) {
               /0/ &&
                  do {
                     if( $mach{'Opcode'}{'IndexFunc'} == 0x06 ) {
                        # Displacement
                        last;
                     }

                     my $offset = $mach{'NPC'} - $mach{'PC'};
                     my $nbytes = $mach{'Opcode'}{'IndexFunc'} - 1;
                     my $value = extractEndianValue( $ref, $offset, $nbytes );
                     my $addr;
                     $addr = $mach{'Segment'}{'Selector'}
                        if $mach{'Segment'}{'Byte'};
                     $addr .= sprintf('0x%x', $value);
                     return [ $addr, $nbytes ];
                  };
               /1/ &&
                  do {
                     my $ifunc = $mach{'Opcode'}{'IndexFunc'};
                     my $bits = substr(unpack('B8', pack('C', $ifunc)),5);
                     my $indexing = $x86_index_registers{$bits};
                     my $addr = sprintf('0x%x', $bytes[$mach{'NPC'}]);
                     return [ $addr.'('.$indexing.')', 1 ];
                  };
               /2/ &&
                  do {
                     my $ifunc = $mach{'Opcode'}{'IndexFunc'};
                     my $bits = substr(unpack('B8', pack('C', $ifunc)),5);
                     my $indexing = $x86_index_registers{$bits};
                     my $nbytes = 2;
                     $nbytes *= 2 if $memtype eq 'MemWord';
                     my @rb = @bytes[$mach{'NPC'} .. $mach{'NPC'}+$nbytes-1];
                     my $addr = sprintf('0x%x',
                                        unpack('v', pack('C'.$nbytes, @rb)));
                     return [ $addr.'('.$indexing.')', $nbytes ];
                  };
               /3/ &&
                  do {
                     my $ifunc = $mach{'Opcode'}{'IndexFunc'};
                     my $bits = substr(unpack('B8', pack('C', $ifunc)),5);
                     return [ $x86_index_registers{$bits}, 0 ];
                  };
            }
         };

      /^(Imm8|Short)$/ &&
         do { return [ sprintf('0x%x', $bytes[$mach{'NPC'}]), 1 ]; };

      /^(Imm16|imm NEAR)$/ &&
         do {
            my @rb = @bytes[$mach{'NPC'} .. $mach{'NPC'}+1];
            return [ sprintf('0x%x', unpack('v', pack('C2', @rb))), 2 ];
         };

      /^(Imm32|Near)$/ &&
         do {
            my @rb = @bytes[$mach{'NPC'} .. $mach{'NPC'}+3];
            return [ sprintf('0x%x', unpack('V', pack('C4', @rb))), 4 ];
         };
   }
   undef;
}

   
# Scan for system service codes
sub doScan {
   my( $name, $size, $bytesref, $serviceidsref ) = @_;
   my @bytes = @$bytesref;

   my %mach = ( 'PC'      => 0,
                'NPC'     => 0,
                'Segment' => { 'Byte'     => 0x00,
                               'Selector' => ''
                             },
                'Opcode'  => { 'IndexFunc'   => undef,
                               'IndexReg'    => undef,
                               'Reg'         => undef,
                               'RegFunction' => undef,
                               'RegType'     => undef,
                               'SegmentReg'  => undef
                             },
                'Text'    => $bytesref
              );

   my $eax;

   CHECK: for( $mach{'PC'} = 0;
               $mach{'PC'} < $size;
               $mach{'PC'} = $mach{'NPC'} ) {
      my $bytes;
      foreach $bytes ( sort keys %lookup ) {

         # Pack value based on number of bytes being scanned
         my $packed = pack( 'C'.$bytes,
                            @bytes[$mach{'PC'} .. ($mach{'PC'} + $bytes - 1)] );

         my $value = unpackNative( $bytes, $packed );

         # Check through all the masks for this byte-length
         my $mask;
         foreach $mask ( sort keys %{$lookup{$bytes}} ) {
            my $masked = $value & $mask;

            # Compare masked value to pattern
            my $pattern;
            foreach $pattern ( keys %{$lookup{$bytes}{$mask}} ) {
               next unless $masked == $pattern;

               # We have a winner, process the opcode
               my $entry = $lookup{$bytes}{$mask}{$pattern};
               $mach{'NPC'} = $mach{'PC'} + $bytes;

               # push selector identifier for next opcode
               $mach{'Segment'}{'Byte'} = $bytes[$mach{'PC'}],
               $mach{'Segment'}{'Selector'} = $entry->[0],
               next CHECK
                  if( $entry->[0] =~ /:$/ );

               my $end = $bytes-1;

               # Update opcode attributes
               updateOpcode( \%mach, $entry );
#               print STDERR $entry->[0].' '.$entry->[1]. "\n";

               # Split argument byte identifiers
               my @ids = split(',', $entry->[1]);
               my $dest = getArgument( \%mach, $ids[0] ) if @ids;
               $mach{'NPC'} += $dest->[1], $end += $dest->[1] if $dest;
               my $src = getArgument( \%mach, $ids[1] ) if @ids == 2;
               $mach{'NPC'} += $src->[1], $end += $src->[1] if $src;

               if( $disassemble ) {
                  print STDERR "\t";
                  my @pb = @bytes[$mach{'PC'} .. ($mach{'PC'}+$end)];
                  unshift( @pb, $mach{'Segment'}{'Byte'} )
                     if $mach{'Segment'}{'Byte'};
                  foreach( @pb ) {
                     printf(STDERR "%2.2x ", $_);
                  }
                  my $size = @pb;
                  foreach( 1 .. (7-$size) ) { print STDERR "   "; }

                  print STDERR ': ' . $entry->[0] . ' ';
                  print STDERR $src->[0].', ' if $src;
                  print STDERR $dest->[0] if $dest;
#                  print STDERR "\t[".$entry->[1].']';
                  print STDERR "\n";
               }

               # Look for setting EAX via a MOV
               if( $entry->[0] eq 'MOV' && $dest->[0] eq 'EAX' ) {
                  $eax = $src->[0];
               }

               # Look for INT 0x2e, we have the code if EAX is set
               if( $eax && $entry->[0] eq 'INT' && $dest->[0] eq '0x2e' ) {
                  push( @$serviceidsref, [ $name, $eax ] );

                  # Skip out, we've found what we're looking for
                  last CHECK;
               }

               $mach{'Segment'}{'Byte'} = 0x00;
               next CHECK;
            }
         }
      }
      # Unknown byte... let's just skip out
      printf(STDERR "\t%2.2x : UNKNOWN\n", $bytes[$mach{'PC'}]);
      $mach{'NPC'} = $size
   }
}


# Sort the functions in ascending order
my $key;
my $directory = $dll->optheader->directory->{'Export'};
my %keys = %{$directory->{'Instance'}->symbols->{'Addresses'}};
my @sortedKeys = sort { $keys{$a} <=> $keys{$b} } keys %keys;

# Open the output file and start scanning
print 'ntdll.dll: Searching for service IDs'."\n";

my $nkeys = @sortedKeys;
my( $inc, $index );
my @serviceids;
LOOP: for( $index = 0; $index < $nkeys; $index += $inc ) {
   $inc = 1;

   # Process only the functions we're interested in
   my $name = $sortedKeys[$index];
   next if $name !~ /$functions/;

   # Determine the starting offset
   my $rva = $keys{$sortedKeys[$index]};
   my $start = $dll->coff->findSectionByRVA($rva);
   next LOOP if( !$start || (($index + 1) == $nkeys) );

   # Determine the ending offset
   my $endrva = $keys{$sortedKeys[$index+$inc]};
   my $end = $dll->coff->findSectionByRVA($endrva);

   if( $start == $end ) {
      $inc++;
      $endrva = $keys{$sortedKeys[$index+$inc]};
      $end = $dll->coff->findSectionByRVA($endrva);
   }

   printf( STDERR "%d/%d: Reading from 0x%x(%d:0x%x) to 0x%x(%d:0x%x) [%s]\n",
           $index, $nkeys, $start, $index, $rva, $end, $index+$inc, $endrva,
           $name)
      if $disassemble;
   sysseek( *fp, $start, SEEK_SET ) || die;

   # Read the code block
   my $size = $end - $start;

   my $buf;
   my $nread = sysread( *fp, $buf, $size );

   # Unpack block into individual bytes
   my @bytes = unpack('C'.$size, $buf);

   # Start scanning code block
   doScan( $name, $size, \@bytes, \@serviceids );
}

# Write service IDs to output header
my $nids = @serviceids;
print 'ntdll.dll: Writing '.$nids.' service IDs to output header file'."\n";
open( HEADER, ">$header" ) || die $header.": $!";
select( HEADER );
$| = 1;

my $topheader = '__'.uc($header);
$topheader =~ s/\./_/g;

print <<EOF;
/*
 * $header
 *
 * File automatically generated by $0
 * Do not modify!
 */

#ifndef $topheader
#define $topheader 1

EOF

foreach( sort { $a->[0] cmp $b->[0] } @serviceids ) {
   print '#define SERVICEID_'.$_->[0].' '.$_->[1]."\n";
}

print <<EOF;

#endif /* $topheader */

EOF

close( HEADER );

