/*! \file linker.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker.c,v 1.18 2001/06/18 15:35:19 stevek Exp $
 *
 * \brief   A program for testing the linkers outside of kernel space
 * \author  Steve Kiernan
 * \version $Revision: 1.18 $
 * \date    $Date: 2001/06/18 15:35:19 $
 *
 *
 */


#include <wr_conf.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#ifdef WR_UNIX
#include <dirent.h>
#include <dlfcn.h>
#include <unistd.h>
#endif /* WR_UNIX */
#include <errno.h>

#include <wss_iface.h>
#include <wsi_iface.h>
#include <wss_module.h>
#include <wss_linker.h>

#ifdef WR_FREEBSD
#include <sys/param.h>
#include <sys/proc.h>
#endif /* WR_FREEBSD */
#ifdef WR_LINUX
/* Evil Evil */
struct task_struct {
    pid_t       pid;
    struct {
        pid_t   pid;
    } p_pptr[1];
    uid_t       uid, euid;
    gid_t       gid, egid;
};

struct proc {
    struct task_struct head;
};
#endif

#include "rtld.h"

const pd_char_t *linkroot =
#ifdef TOPDIR
   PD_MAKESTR(TOPDIR)
#else
   PD_MAKESTR("../..")
#endif
   PD_MAKESTR("/linkers");

struct linker_info
{
   wss_linker_t       *linker;
   void               *objentry;
   struct linker_info *next;
}
* linker_info_list = NULL;

extern wss_linker_t IMPORT *wss_linker_list;


int ProbeLinkers();


void
linker_dump()
{
   struct linker_info *link;

   if( linker_info_list )
   {
      _tprintf( PD_MAKESTR("\n[ Linker Description ] [ Load Address ] [ Map Size ]\n") );
      for( link = linker_info_list; link; link = link->next )
         _tprintf( PD_MAKESTR("  %-18.18s     0x%08x       0x%x\n"),
                   link->linker->linker_description,
#if defined(__FreeBSD__) && ( __FreeBSD_version >= 300000 )
                   (unsigned int)((Obj_Entry *)link->objentry)->mapbase,
                   (unsigned int)((Obj_Entry *)link->objentry)->mapsize
#else /* __FreeBSD__ && __FreeBSD_version >= 300000 */
                   0U, 0U
#endif /* __FreeBSD__ && __FreeBSD_version >= 300000 */
                 );
   }
}

void
linker_dump_chain()
{
#if defined(__FreeBSD__) && __FreeBSD_version >= 300000
   Obj_Entry *ent = (Obj_Entry *)dlopen( NULL, RTLD_LAZY );

   if( ent )
   {
      _tprintf( PD_MAKESTR("\n[ Filename                                    ] [ Load Address ] [ Map Size ]\n") );
      for( ; ent; ent = ent->next )
          _tprintf( PD_MAKESTR("  %-43.43s     0x%08x       0x%x\n"),
                    ent->path, (unsigned int)ent->mapbase,
                    (unsigned int)ent->mapsize );
   }
#endif /* __FreeBSD__ && __FreeBSD_version >= 300000 */
}

#ifdef WR_SOLARIS
unsigned
int wsi_urem( unsigned int a, unsigned int b )
{
   return a % b;
}

int
wsi_rem( int a, int b )
{
   return a % b;
}

unsigned int
wsi_udiv( unsigned int a, unsigned int b )
{
   return a / b;
}

int
wsi_div( int a, int b )
{
   return a / b;
}

wss_symbol_t wsi_export_table[] =
{
   WSS_EXPORT_ENTRY(".urem", (u_long)wsi_urem),
   WSS_EXPORT_ENTRY(".udiv", (u_long)wsi_udiv),
   WSS_EXPORT_ENTRY(".rem",  (u_long)wsi_rem),
   WSS_EXPORT_ENTRY(".div",  (u_long)wsi_div),
   { NULL, 0, 0 }
};
#endif /* WR_SOLARIS */

int
main( int argc, char *argv[] )
{
   int nlinkers = 0;

   wss_linker_init();

#ifdef WR_SOLARIS
   wss_linker_add_symbols( "*WSI SOLARIS SPARC*", wsi_export_table );
#endif /* WR_SOLARIS */

   _tprintf( PD_MAKESTR("Probing for available linkers...\n") );
   nlinkers = ProbeLinkers();

   if( nlinkers )
   {
#if defined(WR_WINNT)
      pd_char_t arg0[256];

      MultiByteToWideChar( CP_ACP, 0, (LPCSTR)argv[0], strlen(argv[0]) + 1,
                           arg0, 256);
#elif defined(WR_UNIX)
      pd_char_t *arg0 = argv[0];
#endif
      linker_dump_chain();

      if( argc >= 2 )
      {
         int i;

         for( i = 1; i < argc; i++ )
         {
            wss_linker_file_t lf;
            int               error;

#if defined(WR_WINNT)
            pd_char_t arg[2048];

            /* Convert argv param to Unicode */
            MultiByteToWideChar( CP_ACP, 0, (LPCSTR)argv[i],
                                 strlen(argv[i]) + 1, arg, 2048 );
#elif defined(WR_UNIX)
            pd_char_t *arg = argv[i];
#endif

            wss_linker_file_init( &lf );
            error = wss_linker_load( arg, &lf );

            if( error )
            {
               errno = error;
               perror( "wss_linker_load" );
               return error;
            }
            else
            {
               wss_linker_addr_t addr;

               _tprintf( PD_MAKESTR("Linked '%s'.\n"), arg );

               error = (lf.linker_file_ops->lookup_symbol)( &lf, PD_MAKESTR(WRLINKNAME_MOD),
                                                            &addr );
               if( error )
               {
                  _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                            strerror(error) );
                  wss_linker_remove_file( &lf );
                  continue;
               }
               else
                  _tprintf( PD_MAKESTR("name: \"%s\" (0x%x)\n"),
                            *(pd_char_t **)addr.addr, addr.addr32 );
               error =
                  (lf.linker_file_ops->lookup_symbol)( &lf, PD_MAKESTR(WRLINKNAME_MOD_TYPE),
                                                       &addr );
               if( error )
               {
                  _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                            strerror(error) );
                  wss_linker_remove_file( &lf );
                  continue;
               }
               else
                  _tprintf( PD_MAKESTR("type: \"%s\" (0x%x)\n"),
                            *(pd_char_t **)addr.addr, addr.addr32 );

/***************************
 * W R A P P E R   T Y P E *
 ***************************/
               if( !_tcscmp( *(pd_char_t **)addr.addr,
                             PD_MAKESTR("wrapper") ) )
               {
                  error =
                     (lf.linker_file_ops->lookup_symbol)( &lf, PD_MAKESTR(WRLINKNAME_MOD_DEPENDENCIES),
                                                          &addr );
                  if( error )
                  {
                     _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                               strerror(error) );
                     wss_linker_remove_file( &lf );
                     continue;
                  }
                  else
                     _tprintf( PD_MAKESTR("dependencies: \"%s\" (0x%x)\n"),
                               *(pd_char_t **)addr.addr, addr.addr32 );

                  error = (lf.linker_file_ops->export_symbols)( &lf );
                  if( error )
                  {
                     _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                               strerror(error) );
                     wss_linker_remove_file( &lf );
                     continue;
                  }

                  if( lf.linker_file_symbols )
                  {
                     int j, nsym;

                     _tprintf( PD_MAKESTR("\nSymbol Dump") );
                     _tprintf( PD_MAKESTR("\n===========\n") );
                     _tprintf( PD_MAKESTR("[ # ] [             Name             ] [   Value  ]\n") );
                     for( nsym = 0; (lf.linker_file_symbols + nsym)->name; nsym++ )
                        continue;

                     for( j = 1; j <= nsym; j++ )
                         _tprintf( PD_MAKESTR(" %3d   %-30.30s   0x%08x\n"), j,
                                   (lf.linker_file_symbols + (nsym - j))->name,
                                   (u_int32_t)(lf.linker_file_symbols +
                                      (nsym - j))->addr );
                  }

                  error =
                     (lf.linker_file_ops->lookup_symbol)( &lf, PD_MAKESTR("wr_install"),
                                                          &addr );
                  if( error )
                  {
                     _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                               strerror(error) );
                     wss_linker_remove_file( &lf );
                     continue;
                  }
                  else
                  {
                     _tprintf( PD_MAKESTR("Initializing...") );
                     fflush( stdout );
                     ((void(*)(void))(addr.addr))();
                     _tprintf( PD_MAKESTR(".done.\n") );

                     _tprintf( PD_MAKESTR("Releasing last reference\n") );
                     wss_linker_remove_file( &lf );
                  }
               }
/***************************
 * L I B R A R Y   T Y P E *
 ***************************/
               if( !_tcscmp( *(pd_char_t **)addr.addr,
                             PD_MAKESTR("library") ) )
               {
                  error =
                     (lf.linker_file_ops->lookup_symbol)( &lf, PD_MAKESTR(WRLINKNAME_MOD_DEPENDENCIES),
                                                          &addr );
                  if( error )
                  {
                     _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                               strerror(error) );
                     wss_linker_remove_file( &lf );
                     continue;
                  }
                  else
                     _tprintf( PD_MAKESTR("dependencies: \"%s\" (0x%x)\n"),
                               *(pd_char_t **)addr.addr, addr.addr32 );

                  error = (lf.linker_file_ops->export_symbols)( &lf );
                  if( error )
                  {
                     _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                               strerror(error) );
                     wss_linker_remove_file( &lf );
                     continue;
                  }

                  if( lf.linker_file_symbols )
                  {
                     int j, nsym;

                     _tprintf( PD_MAKESTR("\nSymbol Dump") );
                     _tprintf( PD_MAKESTR("\n===========\n") );
                     _tprintf( PD_MAKESTR("[ # ] [             Name             ] [   Value  ]\n") );
                     for( nsym = 0; (lf.linker_file_symbols + nsym)->name; nsym++ )
                        continue;

                     for( j = 1; j <= nsym; j++ )
                         _tprintf( PD_MAKESTR(" %3d   %-30.30s   0x%08x\n"), j,
                                   (lf.linker_file_symbols + (nsym - j))->name,
                                   (u_int32_t)(lf.linker_file_symbols +
                                      (nsym - j))->addr );
                  }

                  _tprintf( PD_MAKESTR("Releasing last reference\n") );
                  wss_linker_remove_file( &lf );
               }
/***************************************************
 * A C T I V A T I O N   C R I T E R I A   T Y P E *
 ***************************************************/
               else if( !_tcscmp( *(pd_char_t **)addr.addr,
                                  PD_MAKESTR("activation") ) )
               {
                  error =
                     (lf.linker_file_ops->lookup_symbol)( &lf, PD_MAKESTR("imp_list"),
                                                          &addr );
                  if( error )
                  {
                     _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                               strerror(error) );
                     wss_linker_remove_file( &lf );
                     continue;
                  }
                  else
                  {
                     char **imp = (char **)addr.addr;

                     if( *imp )
                     {
                        _tprintf( PD_MAKESTR("Dependencies:\n") );
                        while( *imp )
                           _tprintf( PD_MAKESTR("%s\n"), *imp++ );
                        _tprintf( PD_MAKESTR("--\n") );
                     }
                  }

                  error =
                     (lf.linker_file_ops->lookup_symbol)( &lf,
                                                          PD_MAKESTR("activate"),
                                                          &addr );
                  if( error )
                  {
                     _tprintf( PD_MAKESTR("Error: %d (%s)\n"), error,
                               strerror(error) );
                     wss_linker_remove_file( &lf );
                     continue;
                  }
                  else
                  {
                     wss_proc_t   wrp;
                     wss_crit_t   crit;
                     wss_mod_t    mod;
                     struct proc  p;
#if defined(WR_FREEBSD)
                     struct pcred cred;
#elif defined(WR_SOLARIS)
                     struct pid   pid;
                     struct cred  cred;
#endif /* WR_FREEBSD || WR_SOLARIS */
                     void       (*activate) __P(( wss_proc_t *, wss_crit_t * ));

                     wss_ref_init( &mod.wrm_refcount );
                     mod.wrm_linked = &lf;

                     wss_ref_init( &crit.crit_refcount );
                     crit.crit_module = &mod;

                     wss_ref_init( &wrp.wrp_refcount );
#if defined(WR_FREEBSD) || defined(WR_SOLARIS)
                     wrp.wrp_cwd = getenv("PWD");
#elif defined(WR_WINNT)
                     wrp.wrp_cwd = _tgetenv(PD_MAKESTR("PWD"));
#endif
                     wrp.wrp_wi_list = NULL;
                     wrp.wrp_arg_list = NULL;
                     wrp.wrp_pid = getpid();
                     wrp.wrp_path = arg0;
                     wrp.wrp_oldpath = PD_MAKESTR("/sbin/init");
                     wrp.wrp_progname = PD_MAKESTR("linker");
#ifdef WR_LINUX
                     wrp.wrp_p = &p.head;
#else
                     wrp.wrp_p = &p;
#endif

#ifdef WR_SOLARIS
                     p.p_pidp = &pid;
#endif /* WR_SOLARIS */

#ifndef WR_LINUX
                     p.p_pid = getpid();
#else
                     p.head.pid = getpid();
#endif
#ifdef WR_FREEBSD
                     p.p_pgrp = NULL;
                     p.p_oppid = getppid();
                     p.p_cred = &cred;

                     cred.p_ruid = getuid();
                     cred.p_svuid = geteuid();
                     cred.p_rgid = getgid();
                     cred.p_svgid = getegid();
#elif defined(WR_SOLARIS)
                     p.p_cred = &cred;

                     cred.cr_ruid = getuid();
                     cred.cr_suid = geteuid();
                     cred.cr_rgid = getgid();
                     cred.cr_sgid = getegid();
#endif /* WR_FREEBSD || WR_SOLARIS */

                     _tprintf( PD_MAKESTR("Activating...\n") );
                     fflush( stdout );
                     activate = addr.addr;
                     (*activate)( &wrp, &crit );
                     _tprintf( PD_MAKESTR("Done.\n") );

                     _tprintf( PD_MAKESTR("Releasing last reference\n") );
                     wss_linker_remove_file( &lf );
                  }
               }
            }
         }
      }
   }
   else
      _tprintf( PD_MAKESTR("No linkers available.\n") );

   return 0;
}

#ifdef WR_UNIX
int
ProbeLinkers()
{
   DIR *dp;
   int  nlinkers = 0;

   if( ( dp = opendir( linkroot ) ) != NULL )
   {
      struct dirent *dirent;

      while( ( dirent = readdir(dp) ) != NULL )
      {
         char buf[2048];

         if(
#ifdef WR_FREEBSD
             ( dirent->d_type == DT_DIR ) && ( dirent->d_namlen > 0 ) &&
#endif /* WR_FREEBSD */
             ( *dirent->d_name != '.' ) )
         {
            struct stat sb;

            snprintf( buf, 2048, "%s/%s/liblinker_%s.so.0.1", linkroot,
                      dirent->d_name, dirent->d_name );
            if( stat( buf, &sb ) >= 0 )
            {
               void *dlm;

               printf( "  %s ", dirent->d_name );
               fflush( stdout );

               dlm = dlopen( buf, RTLD_LAZY );
               if( dlm == NULL )
                  printf( "[Failed]\n     %s\n", dlerror() );
               else
               {
                  void *sym = dlsym( dlm, "_linker_init" );
                  if( sym == NULL )
                     printf( "[Loaded]\n" );
                  else
                  {
                     struct linker_info *link;

                     ((void(*)(void))sym)();
                     printf( "[Ok]\n" );

                     link = malloc( sizeof(*link) );
                     link->linker = wss_linker_list;
                     link->objentry = dlm;
                     link->next = linker_info_list;
                     linker_info_list = link;

                     nlinkers++;
                  }
               }
            }
         }
      }
   }
   return nlinkers;
}
#endif

#ifdef WR_WINNT
int
ProbeLinkers()
{
   WIN32_FIND_DATA fd;
   HANDLE          hSearch;
   pd_char_t      rootpath[2048];
   unsigned int    i;
   int             nlinkers = 0;

   /* Convert Unix-style paths to Windows */
   _sntprintf( rootpath, 2048, PD_MAKESTR("%s/*"), linkroot );
   for( i = 0; i < _tcslen(rootpath); i++ )
      if( rootpath[i] == PD_MAKESTR('/') )
         rootpath[i] = PD_MAKESTR('\\');

   /* Iterate through linker directories */
   hSearch = FindFirstFile( rootpath, &fd );
   if( hSearch != INVALID_HANDLE_VALUE )
   {
      rootpath[_tcslen(rootpath)-2] = PD_MAKESTR('\0');
      do
      {
         if( *fd.cFileName != PD_MAKESTR('.') )
         {
            pd_char_t   buf[2048];
            struct _stat sb;
            _sntprintf( buf, 2048, PD_MAKESTR("%s\\%s\\linker_%s.dll"),
                        rootpath, fd.cFileName, fd.cFileName );

            /* If we can stat the file and it's not a directory,
               record it as a match */
            if( ( _tstat( buf, &sb ) >= 0 ) && ( sb.st_mode != _S_IFDIR ) )
            {
               HANDLE hDll;

               _tprintf( PD_MAKESTR("  %s "), fd.cFileName );

               hDll = LoadLibrary( buf );

               if( hDll == NULL )
               {
                  LPVOID lpMsgBuf;

                  FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | 
                                 FORMAT_MESSAGE_FROM_SYSTEM |
                                 FORMAT_MESSAGE_IGNORE_INSERTS,
                                 NULL, GetLastError(),
                                 MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                                 (LPTSTR) &lpMsgBuf, 0, NULL );

                  _tprintf( PD_MAKESTR("[Failed]\n     %s\n"), lpMsgBuf );

                  LocalFree( lpMsgBuf );
               }
               else
               {
                  void *sym = GetProcAddress( hDll, "_LinkerInit" );
                  if( sym == NULL )
                     _tprintf( PD_MAKESTR("[Loaded]\n") );
                  else
                  {
                     struct linker_info *link;

                     ((void(*)(void))sym)();
                     _tprintf( PD_MAKESTR("[Ok]\n") );

                     link = malloc( sizeof(*link) );
                     link->linker = wss_linker_list;
                     link->objentry = hDll;
                     link->next = linker_info_list;
                     linker_info_list = link;

                     nlinkers++;
                  }
               }
            }
         }
      }
      while( FindNextFile( hSearch, &fd ) );

      FindClose( hSearch );
   }
   return nlinkers;
}
#endif

