/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: linker_file.c,v 1.7 2001/02/23 19:02:14 stevek Exp $
 */

#include <wr_conf.h>

#ifdef WR_SOLARIS
#include <unistd.h>
#include <fcntl.h>
#elif defined(WR_LINUX)
#include <errno.h>
#include <unistd.h>
#endif /* WR_SOLARIS */

#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <wss_file.h>
#include <wss_malloc.h>
#include <wss_mutex.h>

extern void panic __P(( pd_char_t *, ... ));


/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

/*! \brief File management structure */
struct wsi_file
{
   WSS_MUTEX_PROTECT(f);          /*!< Mutex to protect the file      */
   int               f_refcount;  /*!< Number of references to file   */
   int               f_fd;        /*!< File descriptor                */
   void             *f_mapAddr;   /*!< Address where file is mapped   */
   size_t            f_mapSize;   /*!< Size of mapped memory area     */
   pd_offset_t      f_offset;    /*!< Seek offset into file          */
};


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_file_open */
/*! \brief Construct wsi_file based on a path and open it
 *
 *  \param path Pathname of file to open                           
 *  \param mode Mode flags with which to open the file
 *  
 *  \return Handle to open file on success,
 *  \return \b WSS_FILE_INVALID on failure
 */
wss_file_t
wsi_file_open( const pd_char_t *path, wss_file_mode_t mode )
{
   int               err      = 0;
   struct wsi_file  *file;
   int               filemode = 0;
   pd_char_t       *abspath;
   int               fd;

#ifdef DEBUG_FILE
   _tprintf( PD_MAKESTR("wsi_file_open( '%s', 0x%x )\n"), path, mode );
#endif

   if( mode & WSS_FILE_READ )
   {
      if( mode & WSS_FILE_WRITE )
         filemode |= O_RDWR;
      else
         filemode |= O_RDONLY;
   }
   else if( mode & WSS_FILE_WRITE )
      filemode |= O_WRONLY;
   if( mode & WSS_FILE_APPEND )
      filemode |= O_APPEND;
   if( mode & WSS_FILE_TRUNC )
      filemode |= O_TRUNC;
   if( mode & WSS_FILE_CREAT )
      filemode |= O_CREAT;

   /* Make sure the path is in absolute terms */
   abspath = (pd_char_t *)path;

   if( abspath == NULL )
      return NULL;

   fd = open( abspath, filemode );
   if( fd < 0 )
      err = errno;

#ifdef DEBUG_FILE
   _tprintf( PD_MAKESTR("create result %d\n"), fd );
#endif

   if( err == 0 )
   {
      file = (struct wsi_file *)wss_malloc( sizeof(struct wsi_file) );

      if( file )
      {
         WSS_MUTEX_INIT( file->f );
         file->f_refcount = 1;
         file->f_fd = fd;
         file->f_offset = 0;
         file->f_mapAddr = NULL;
         file->f_mapSize = 0;
      }
      else
         close( fd );
   }
   else
      file = NULL;

#ifdef DEBUG_FILE
   _tprintf( PD_MAKESTR("Returning 0x%x\n"), (unsigned int)file );
#endif
   return( (void *)file );
}


/*  wsi_file_free */
/*! \brief Really close the file and free structure memory
 * 
 *  \param handle Handle to an open file
 */
void
wsi_file_free( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   assert( file != NULL );
   assert( file->f_refcount == 1 );

   if( file )
   {
      if( file->f_fd >= 0 )
         close( file->f_fd );

      WSS_MUTEX_DESTROY( file->f );
      wss_free( file );
   }
}


/*  wsi_file_close */
/*! \brief Close previously opened file
 *
 *  \param handle Handle to an open file
 */
void
wsi_file_close( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      if( file->f_mapAddr )
         wsi_file_unmap( handle );

      wsi_file_rele( handle );
   }
}


/*  wsi_file_ref */
/*! \brief Acquire reference to file sturcture
 *
 *  \param handle Handle to an open file
 */
void
wsi_file_ref( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file->f_refcount < 1 )
      panic( PD_MAKESTR("wsi_file_ref: reference count for 0x%x was %d < 1"),
             (unsigned int)file, file->f_refcount );

   WSS_MUTEX_ENTER( file->f );
   ++file->f_refcount;
   WSS_MUTEX_EXIT( file->f );
}

                               
/*  wsi_file_rele */
/*! \brief Release reference to file sturcture
 *
 *  \param handle Handle to an open file
 */
void
wsi_file_rele( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file->f_refcount < 1 )
      panic( PD_MAKESTR("wsi_file_rele: reference count for 0x%x was %d < 1"),
             (unsigned int)file, file->f_refcount );

   WSS_MUTEX_ENTER( file->f );
   if(file->f_refcount > 1)
   {
       --file->f_refcount;
       WSS_MUTEX_EXIT( file->f );
   }
   else
      wsi_file_free( handle );
}


/*  wsi_file_write */
/*! \brief Write data to open file
 *
 *  \param handle Handle to an open file
 *  \param buffer Pointer to buffer containing data to be written
 *  \param len    Number of bytes of data in buffer
 *
 *  \return number of bytes written on sucess,
 *  \return -EINVAL if \a handle is \b WSS_FILE_INVALID
 *  \return negated error code for errors associated with write(2)
 * 
 *  \sa write(2) man page
 */
int
wsi_file_write( wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file && file->f_fd )
   {
      int cnt;

      cnt = write( file->f_fd, buffer, len );
      if( cnt >= 0 )
         file->f_offset += cnt;

      return( (cnt<0)?-errno:cnt );
   }
   else
      return -EINVAL;
}


/*  wsi_file_read */
/*! \brief Read data from open file
 *
 *  \param handle Handle to an open file
 *  \param buffer Pointer to buffer which will contain read data
 *  \param len    Number of bytes of data to read
 *
 *  \return number of bytes read on sucess,
 *  \return -EINVAL if \a handle is \b WSS_FILE_INVALID
 *  \return negated error code for errors associated with read(2)
 *    
 *  \sa read(2) man page
 */
int
wsi_file_read( wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file && file->f_fd )
   {
      int cnt;

      cnt = read( file->f_fd, buffer, len );
      if( cnt >= 0 )
         file->f_offset += cnt;

      return( (cnt<0)?-errno:cnt );
   }
   else
      return -EINVAL;
}


/*  wsi_file_seek */
/*! \brief Seek withing an open file
 *
 *  \param handle Handle to an open file
 *  \param offset Offset into file
 *  \param whence Where to seek from:
 *                   \b WSS_FILE_SEEK_END - seek from end of file
 *                   \b WSS_FILE_SEEK_CUR - seek from current position in file
 *                   \b WSS_FILE_SEEK_SET - set current position in file
 *
 *  \return 0 on success,
 *  \return EINVAL if \a handle is \b WSS_FILE_INVALID or \a whence is not one
 *                 of the three valid paramters above
 */
int
wsi_file_seek( wss_file_t handle, pd_offset_t offset, int whence )
{
   struct wsi_file *file = (struct wsi_file *)handle;
   int              err = 0;

   if( file && file->f_fd )
   {
      switch( whence )
      {
         case WSS_FILE_SEEK_END:
         {
            int size = wsi_file_get_size( file, &err );
            if( !err )
               file->f_offset = size + offset;
            lseek( file->f_fd, offset, SEEK_END );
            break;
         }

         case WSS_FILE_SEEK_CUR:
            file->f_offset += offset;
            lseek( file->f_fd, offset, SEEK_CUR );
            break;

         case WSS_FILE_SEEK_SET:
            file->f_offset = offset;
            lseek( file->f_fd, offset, SEEK_SET );
            break;

         default:
            err = EINVAL;
      }
   }
   else
      err = EINVAL;

   return err;
}


/*  wsi_file_get_size */
/*! \brief Return the size of the file
 *
 *  \param handle Handle to an open file
 *  \param error  If non-NULL, pointer to memory to store error code
 *
 *  \return Size of file on success,
 *  \return 0 on failure
 */
size_t
wsi_file_get_size( wss_file_t handle, int *error )
{
   struct wsi_file *file = (struct wsi_file *)handle;

#ifdef DEBUG_FILE
   _tprintf( PD_MAKESTR("wsi_file_get_size( 0x%x )\n"), (unsigned int)file );
#endif

   if( file )
   {
      if( file->f_fd )
      {
         struct stat st;
         int         err;

         err = fstat( file->f_fd, &st );
#ifdef DEBUG_FILE
         _tprintf( PD_MAKESTR("fstat returned %d\n"), err );
#endif
         if( err < 0 )
         {
            *error = errno;
            return 0;
         }
         else
            return st.st_size;
      }
      else if( file->f_mapAddr )
         return file->f_mapSize;
   }

   if( error )
      *error = EINVAL;

   return 0;
}

/*  wsi_file_get_map_addr */
/*! \brief Return the address where the file is mapped
 * 
 *  \param handle Handle to an open file
 *
 *  \return Pointer to address where file is mapped on success,
 *  \return 0 on failure 
 */
void *
wsi_file_get_map_addr( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;
   void            *addr = NULL;

   if( file )
      addr = file->f_mapAddr;

   return addr;
}


/*  wsi_file_map */
/*! \brief Map an open file into the kernel
 *
 *  \param handle  Handle to an open file
 *  \param size    Number of bytes to map
 *  \param mapmode Mode flags to map file
 *  \param pos     Offset into file to map
 *
 *  \return 0 on success,
 *  \return EINVAL if \a handle is \b WSS_FILE_INVALID,
 *  \return error code for errors associated with mmap(2)
 */
int
wsi_file_map( wss_file_t handle, size_t size, int mapmode, pd_offset_t pos )
{
   struct wsi_file *file  = (struct wsi_file *)handle;
   int              error = EINVAL;

   if( file && file->f_fd )
   {
      int prot = 0;

      if( wsi_file_get_size( file, NULL ) < size )
      {
         char *bb = "xxx";

         _tprintf( PD_MAKESTR("WSS: creating %d byte memory-mapped file.\n"),
                   size );

         wsi_file_seek( file, size - 1, WSS_FILE_SEEK_SET );
         wsi_file_write( file, bb, 1 );
      }

      if( mapmode & WSS_FILE_MAP_WRITE )
         prot |= PROT_WRITE;

      if( mapmode & WSS_FILE_MAP_READ )
         prot |= PROT_READ;

      file->f_mapAddr = mmap( NULL, size, prot, MAP_SHARED, file->f_fd, pos );
      if( file->f_mapAddr != MAP_FAILED )
      {
         file->f_mapSize = size;
         close( file->f_fd );

         /* Disassociate file, it's mapped now */
         file->f_fd = -1;
      }
      else
         error = errno;
   }

   return error;
}


/*  wsi_file_unmap */
/*! \brief Unmap a memory-mapped file
 *  
 *  \param handle Handle to mapped file  
 */ 
void
wsi_file_unmap( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file && file->f_mapAddr )
   {
      munmap( file->f_mapAddr, file->f_mapSize );
      file->f_mapAddr = NULL;
      file->f_mapSize = 0;
   }
}

