/*! \file wsi_efsinterface.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: eventfs_interface.c,v 1.6 2001/02/23 19:02:16 stevek Exp $
 *
 * \brief   Interface to EventFS
 * \author  Steve Kiernan
 * \version $Revision: 1.6 $
 * \date    $Date: 2001/02/23 19:02:16 $
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <eventfs.h>

#include <pdpi/pi_mutex.h>

#include <errno.h>

#include <pipesrv.h>


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static eventfs_interface_t *g_pIFace = NULL;
PD_MUTEX_PROTECT           (EFSLINKAGE, g_pIFace);


/***************************************************************************\
 * EventFS stubs                                                           *
\***************************************************************************/

int EXPORT
eventfs_init()
{
   if( g_pIFace && g_pIFace->Init )
      return (*g_pIFace->Init)();

   return (0);
}

int EXPORT
eventfs_uninit()
{
   if( g_pIFace && g_pIFace->Uninit )
      return (*g_pIFace->Uninit)();

   return (0);
}

int EXPORT
eventfs_construct_stream( const pd_char_t *group, u_long mask,
                          pd_char_t **retname )
{
   if( g_pIFace && g_pIFace->ConstructStream )
      return (*g_pIFace->ConstructStream)( group, mask, retname );

   return (EINVAL);
}

int EXPORT
eventfs_destroy_group( const pd_char_t *group )
{
   if( g_pIFace && g_pIFace->DestroyGroup )
      return (*g_pIFace->DestroyGroup)( group );

   return (EINVAL);
}

int EXPORT
eventfs_destroy_stream( const pd_char_t *group, const pd_char_t *stream )
{
   if( g_pIFace && g_pIFace->DestroyStream )
      return (*g_pIFace->DestroyStream)( group, stream );

   return (EINVAL);
}

int EXPORT
eventfs_register_group( const pd_char_t *group )
{
   if( g_pIFace && g_pIFace->RegisterGroup )
      return (*g_pIFace->RegisterGroup)( group );

   return (EINVAL);
}

int EXPORT
eventfs_register_event_type( const pd_char_t *group,
                             const pd_char_t *event_type,
                             u_long *type_return )
{
   if( g_pIFace && g_pIFace->RegisterEventType )
      return (*g_pIFace->RegisterEventType)( group, event_type, type_return );

   return (EINVAL);
}

int EXPORT
eventfs_register_stream( const pd_char_t *group, const pd_char_t *stream,
                         u_long mask, pd_bool_t transient )
{
   if( g_pIFace && g_pIFace->RegisterStream )
      return (*g_pIFace->RegisterStream)( group, stream, mask, transient );

   return (EINVAL);
}

int EXPORT
eventfs_get_event_mask( const pd_char_t *group, const pd_char_t *type_name,
                        u_long *type )
{
   if( g_pIFace && g_pIFace->GetEventMask )
      return (*g_pIFace->GetEventMask)( group, type_name, type );

   return (EINVAL);
}

int EXPORT
eventfs_post_event( const pd_char_t *group, const pd_char_t *type,
                    size_t size, void *data )
{
   if( g_pIFace && g_pIFace->PostEvent )
      return (*g_pIFace->PostEvent)( group, type, size, data );

   return (EINVAL);
}

int EXPORT
eventfs_post_event_by_types( const pd_char_t *group, u_long types,
                             size_t size, const void *data )
{
   if( g_pIFace && g_pIFace->PostEventByTypes )
      return (*g_pIFace->PostEventByTypes)( group, types, size, data );

   return (EINVAL);
}

int EXPORT
eventfs_flush_stream( const pd_char_t *group, const pd_char_t *stream )
{
   if( g_pIFace && g_pIFace->FlushStream)
      return (*g_pIFace->FlushStream)( group, stream );

   return (EINVAL);
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

void EXPORT
eventfs_interface_init()
{
   PD_MUTEX_INIT( EFSLINKAGE, g_pIFace );
}

void EXPORT
eventfs_interface_uninit()
{
   PD_MUTEX_DESTROY( EFSLINKAGE, g_pIFace );
}


/*  eventfs_interface_register */
/*! \brief Register a set of EventFS functions
 */
BOOL EXPORT
eventfs_interface_register( eventfs_interface_t *iface )
{
   PD_MUTEX_ENTER( EFSLINKAGE, g_pIFace );
   if( !g_pIFace || ( iface == NULL ) )
   {
      g_pIFace = iface;
      PD_MUTEX_EXIT( EFSLINKAGE, g_pIFace );
      return (TRUE);
   }
   PD_MUTEX_EXIT( EFSLINKAGE, g_pIFace );
   return (FALSE);
}

void EXPORT
panic( char *format, ... )
{
   va_list va;
   va_start( va, format );
   vprintf( format, va );
   va_end( va );

   if( g_hPipeSrvStopEvent != INVALID_HANDLE_VALUE )
      SetEvent( g_hPipeSrvStopEvent );
}

