/*! \file pd_cv.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: pd_cv.c,v 1.6 2001/02/23 19:02:16 stevek Exp $
 *
 * \brief   Implementation of condition variables on Windows NT
 * \author  Steve Kiernan
 * \version $Revision: 1.6 $
 * \date    $Date: 2001/02/23 19:02:16 $
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <pdpi/pi_mutex.h>
#include <pdpi/pi_cv.h>


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  pd_cv_init */
/*! \brief Initialize condition variable
 *
 *  \param hCV Pointer to a condition variable handle
 */
void EXPORT
pd_cv_init( HANDLE *hCV )
{
   *hCV = CreateEvent( NULL, FALSE, FALSE, NULL );
}


/*  pd_cv_destroy */
/*! \brief Destroy condition variable
 *
 *  \param hCV Handle to a condition variable
 */
void EXPORT
pd_cv_destroy( HANDLE hCV )
{
   CloseHandle( hCV );
}


/*  pd_cv_wait */
/*! \brief Wait on a condition variable
 *
 *  \param hCV   Handle to a condition variable
 *  \param mutex Handle to a mutex
 */
#ifdef DEBUG_MUTEX
void EXPORT
pd_cv_wait( pd_cv_t hCV, pd_mutex_t *mutex, const pd_char_t *file, int line )
#else
void EXPORT
pd_cv_wait( pd_cv_t hCV, pd_mutex_t mutex )
#endif /* DEBUG_MUTEX */
{
#ifdef DEBUG_MUTEX
   pd_mutex_exit( mutex, file, line );
#else
   pd_mutex_exit( mutex );
#endif /* DEBUG_MUTEX */
   while( WaitForSingleObject( hCV, INFINITE ) != WAIT_OBJECT_0 );
#ifdef DEBUG_MUTEX
   pd_mutex_enter( mutex, file, line );
#else
   pd_mutex_enter( mutex );
#endif /* DEBUG_MUTEX */
}

/*  pd_cv_signal */
/*! \brief Signal a condition variable to wake up a single waiter
 *
 *  \param hCV Handle to a condition variable
 */
void EXPORT
pd_cv_signal( HANDLE hCV )
{
   PulseEvent( hCV );
}

/*  pd_cv_broadcast */
/*! \brief Signal a condition variable to wake up all waiters
 *
 *  \param hCV Handle to a condition variable
 */
void EXPORT
pd_cv_broadcast( HANDLE hCV )
{
   SetEvent( hCV );
}

