/*! \file pd_mutex.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: pd_mutex.c,v 1.6 2001/02/23 19:02:16 stevek Exp $
 *
 * \brief   Implementation of mutexes on Windows NT
 * \author  Steve Kiernan
 * \version $Revision: 1.6 $
 * \date    $Date: 2001/02/23 19:02:16 $
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <pdpi/pi_mutex.h>

#ifdef DEBUG_MUTEX
static BOOL paniced = FALSE;
#endif /* DEBUG_MUTEX */


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  pd_mutex_init */
/*! \brief Initialize mutex
 *
 *  \param mutex Pointer to mutex handle
 */
#ifdef DEBUG_MUTEX
void EXPORT
pd_mutex_init( pd_mutex_t *mutex, const pd_char_t *file, int line )
{
   mutex->hMutex = CreateMutex( NULL, FALSE, NULL );
   mutex->refCount = 0;
   _tprintf( _TEXT("pd_mutex_init(0x%x): 0x%x @ %s:%d\n"), mutex,
             mutex->hMutex, file, line );
}
#else
void EXPORT
pd_mutex_init( pd_mutex_t *mutex )
{
   *mutex = CreateMutex( NULL, FALSE, NULL );
}
#endif /* DEBUG_MUTEX */


/*  pd_mutex_destroy */
/*! \brief Destory mutex
 *
 *  \param mutex Handle to mutex
 */
#ifdef DEBUG_MUTEX
void EXPORT
pd_mutex_destroy( pd_mutex_t *mutex, const pd_char_t *file, int line )
{
   _tprintf( _TEXT("pd_mutex_destroy(0x%x): 0x%x @ %s:%d\n"), mutex,
             mutex->hMutex, file, line );
   CloseHandle( mutex->hMutex );
   mutex->hMutex = INVALID_HANDLE_VALUE;
   mutex->refCount = 0xffffffff;
}
#else
void EXPORT
pd_mutex_destroy( pd_mutex_t mutex )
{
   CloseHandle( mutex );
}
#endif /* DEBUG_MUTEX */


/*  pd_mutex_enter */
/*! \brief Acquire lock on mutex
 *
 *  \param mutex Handle to mutex
 */
#ifdef DEBUG_MUTEX
void EXPORT
pd_mutex_enter( pd_mutex_t *mutex, const pd_char_t *file, int line )
{
   DWORD dwRet;

   mutex->refCount++;
   _tprintf( _TEXT("pd_mutex_enter(0x%x): 0x%x [%ld] @ %s:%d\n"), mutex,
             mutex->hMutex, mutex->refCount, file, line );
   while( ( dwRet = WaitForSingleObject( mutex->hMutex, INFINITE ) ) != WAIT_OBJECT_0 )
      if( dwRet == WAIT_FAILED )
      {
         if( paniced )
            break;
         else
         {
            paniced = TRUE;
            panic( "pd_mutex_enter: wait failed for mutex 0x%x @ %s:%d\n",
                   mutex->hMutex, file, line );
         }
      }
}
#else
void EXPORT
pd_mutex_enter( pd_mutex_t mutex )
{
   while( WaitForSingleObject( mutex, INFINITE ) != WAIT_OBJECT_0 );
}
#endif /* DEBUG_MUTEX */


/*  pd_mutex_exit */
/*! \brief Release lock held on mutex
 *
 *  \param mutex Handle to mutex
 */
#ifdef DEBUG_MUTEX
void EXPORT
pd_mutex_exit( pd_mutex_t *mutex, const pd_char_t *file, int line )
{
   mutex->refCount--;
   _tprintf( _TEXT("pd_mutex_exit(0x%x): 0x%x [%ld] @ %s:%d\n"), mutex,
             mutex->hMutex, mutex->refCount, file, line );
   ReleaseMutex( mutex->hMutex );
}
#else
void EXPORT
pd_mutex_exit( pd_mutex_t mutex )
{
   ReleaseMutex( mutex );
}
#endif /* DEBUG_MUTEX */

