/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: pipesrv.cxx,v 1.8 2001/04/05 18:40:58 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: The Global Wrapper Manager service
 *
 *          This Windows NT service manages the global state of wrappers.
 *          It communicates with both wrapped processes' local wrapper
 *          manager instances and WSS management clients via a set of
 *          named pipes.
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>

#include "pipesrv.h"


// prototypes
//
static void PipeWorkerThread( void * );


// exported globals
//
extern "C" {
HANDLE EXPORT g_hPipeSrvStopEvent = INVALID_HANDLE_VALUE;
BOOL   EXPORT g_bPipeSrvDebugging = FALSE;
}



////////////////////////////////////////////////////////////////////////////
// Class methods
//
EXPORT _PIPESRV_QUEUE_INFO::_PIPESRV_QUEUE_INFO() : qActive(), qAvailable()
{
}

EXPORT _PIPESRV_QUEUE_INFO::~_PIPESRV_QUEUE_INFO()
{
   EnterCriticalSection( &critSec );
   while( !qActive.empty() )
   {
      LPPIPESRV_PIPE_INFO pPipe = qActive.front();
      qActive.pop();
      while( pPipe->thread )
         WaitForSingleObject( hEvents[PIPESRV_WORKER_DIED], INFINITE );
      delete pPipe;
   }

   while( !qAvailable.empty() )
   {
      LPPIPESRV_PIPE_INFO pPipe = qAvailable.front();
      qAvailable.pop();
      while( pPipe->thread )
         WaitForSingleObject( hEvents[PIPESRV_WORKER_DIED], INFINITE );
      delete pPipe;
   }
   LeaveCriticalSection( &critSec );

   CloseHandle(hEvents[PIPESRV_PIPE_AVAILABLE_EVENT]);
   CloseHandle(hEvents[PIPESRV_PIPE_ACTIVE_EVENT]);
   CloseHandle(hEvents[PIPESRV_WORKER_DIED]);
   CloseHandle(hEvents[PIPESRV_ACCEPTOR_DIED]);
   DeleteCriticalSection( &critSec );
}

_PIPESRV_PIPE_INFO::~_PIPESRV_PIPE_INFO()
{
   CloseHandle(hPipe);
   CloseHandle(hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT]);
   CloseHandle(hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT]);

   PIPESRV_LAST_PACKETS::iterator it = pLastPackets->begin();
   while( it != pLastPackets->end() )
   {
      LPPIPESRV_LAST_PACKET_DATA pPacket = (*it).second;
      pLastPackets->erase( it );
      delete pPacket;
      it = pLastPackets->begin();
   }
   delete pLastPackets;

   PD_MUTEX_DESTROY( PDLINKAGE, pLastPackets );
   PD_MUTEX_DESTROY( PDLINKAGE, readPipe );
   PD_MUTEX_DESTROY( PDLINKAGE, writePipe );

   free( pSD );
}


////////////////////////////////////////////////////////////////////////////
// Public functions
//
BOOL PipeSrvInitPipe( LPPIPESRV_PIPE_INFO pInfo )
{
   pInfo->pData = NULL;

   // create the event objects used in overlapped i/o
   //
   pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] = CreateEvent( NULL, TRUE, FALSE, NULL );

   if( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] == NULL )
      return FALSE;

   pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] = CreateEvent( NULL, TRUE, FALSE, NULL );

   if( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] == NULL )
   {
      CloseHandle( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] );

      pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] = INVALID_HANDLE_VALUE;

      return FALSE;
   }

   pInfo->pLastPackets = new PIPESRV_LAST_PACKETS();
   if( pInfo->pLastPackets == NULL )
   {
      CloseHandle( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] );
      CloseHandle( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] );

      pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] = INVALID_HANDLE_VALUE;
      pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] = INVALID_HANDLE_VALUE;

      return FALSE;
   }

   PD_MUTEX_INIT( PDLINKAGE, pInfo->pLastPackets );
   PD_MUTEX_INIT( PDLINKAGE, pInfo->readPipe );
   PD_MUTEX_INIT( PDLINKAGE, pInfo->writePipe );

   // create a security descriptor that allows anyone to write to
   //  the pipe...
   //
   pInfo->pSD = (PSECURITY_DESCRIPTOR)malloc( SECURITY_DESCRIPTOR_MIN_LENGTH );

   // add a NULL disc. ACL to the security descriptor.
   //
   if( ( pInfo->pSD == NULL ) ||
       !InitializeSecurityDescriptor( pInfo->pSD,
                                      SECURITY_DESCRIPTOR_REVISION ) ||
       !SetSecurityDescriptorDacl( pInfo->pSD, TRUE, (PACL) NULL, FALSE ) )
   {
      CloseHandle( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] );
      CloseHandle( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] );
      delete pInfo->pLastPackets;
      free( pInfo->pSD );

      pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] = INVALID_HANDLE_VALUE;
      pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] = INVALID_HANDLE_VALUE;
      pInfo->pLastPackets = NULL;
      pInfo->pSD = NULL;

      return FALSE;
   }


   pInfo->sa.nLength = sizeof(pInfo->sa);
   pInfo->sa.lpSecurityDescriptor = pInfo->pSD;
   pInfo->sa.bInheritHandle = TRUE;

   // open our named pipe...
   //
   pInfo->hPipe =
      CreateNamedPipe( pInfo->pPipeSrvQueueInfo->lpszPipeName,
                       ( FILE_FLAG_OVERLAPPED |
                         ( (pInfo->pPipeSrvQueueInfo->bReadOnly)?
                           PIPE_ACCESS_OUTBOUND : PIPE_ACCESS_DUPLEX ) ),
                       ( PIPE_TYPE_MESSAGE | PIPE_READMODE_MESSAGE | PIPE_WAIT ),
                       PIPE_UNLIMITED_INSTANCES, 0, 0, 1000, &pInfo->sa );

   if( pInfo->hPipe == INVALID_HANDLE_VALUE )
   {
      CloseHandle( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] );
      CloseHandle( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] );
      delete pInfo->pLastPackets;
      free( pInfo->pSD );

      pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] = INVALID_HANDLE_VALUE;
      pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] = INVALID_HANDLE_VALUE;
      pInfo->pLastPackets = NULL;
      pInfo->pSD = NULL;

      return FALSE;
   }

   return TRUE;
}

void EXPORT PipeSrvAcceptThread( void *args )
{
   LPPIPESRV_QUEUE_INFO pqInfo = (LPPIPESRV_QUEUE_INFO)args;
   LPPIPESRV_PIPE_INFO  pInfo;
   BOOL                 cnpRet;
   OVERLAPPED           os;
   DWORD                dwWait;

   if( g_bPipeSrvDebugging )
      _tprintf( _TEXT("AcceptThread[%d]: running for pipe %s\n"),
                GetCurrentThreadId(), pqInfo->lpszPipeName );

   while( 1 )
   {
      pInfo = NULL;

      // see if there are any pipe instances available
      //
      if( pqInfo->qAvailable.empty() )
      {
         // allocate a new pipe instance and initialize
         //
         pInfo = new PIPESRV_PIPE_INFO;
         pInfo->thread = 0;
         pInfo->hPipe = INVALID_HANDLE_VALUE;
         pInfo->hEvents[0] = g_hPipeSrvStopEvent;
         pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] = 0;
         pInfo->hEvents[PIPESRV_OVERLAPPED_IO_WRITE_EVENT] = 0;
         pInfo->pBuf = NULL;
         pInfo->pSD = NULL;
         memset( &(pInfo->sa), 0, sizeof(pInfo->sa) );
         pInfo->pPipeSrvQueueInfo = pqInfo;

         if( !PipeSrvInitPipe( pInfo ) )
         {
            if( g_bPipeSrvDebugging )
            {
               LPVOID lpMsgBuf;

               FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                              FORMAT_MESSAGE_FROM_SYSTEM |
                              FORMAT_MESSAGE_IGNORE_INSERTS,
                              NULL, GetLastError(),
                              MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
                              (LPTSTR) &lpMsgBuf, 0, NULL );

               _tprintf( _TEXT("AcceptThread[%d]: unable to init pipe %s: %s\n"),
                         GetCurrentThreadId(), pqInfo->lpszPipeName,
                         (LPCTSTR)lpMsgBuf );
               LocalFree( lpMsgBuf );
            }

            SetEvent( pqInfo->hEvents[0] );
            goto acceptend;
         }

         EnterCriticalSection( &pqInfo->critSec );
         pInfo->nPipe = pqInfo->nInstances++;
         LeaveCriticalSection( &pqInfo->critSec );

         if( g_bPipeSrvDebugging )
            _tprintf( _TEXT("AcceptThread[%d]: created pipe instance %x\n"),
                      GetCurrentThreadId(), pInfo->hPipe );
      }

      // if we reach here and we still don't have a pipe,
      // there must be one on the available queue
      //
      if( pInfo == NULL )
      {
         EnterCriticalSection( &pqInfo->critSec );
         pInfo = pqInfo->qAvailable.front();
         pqInfo->qAvailable.pop();
         LeaveCriticalSection( &pqInfo->critSec );
      }

      // init overlapped structure
      memset( &os, 0, sizeof(os) );
      os.hEvent = pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT];
      ResetEvent( pInfo->hEvents[PIPESRV_OVERLAPPED_IO_READ_EVENT] );

      if( g_bPipeSrvDebugging )
         _tprintf( _TEXT("AcceptThread[%d]: waiting for connect\n"),
                   GetCurrentThreadId() );

      // wait for a connection
      cnpRet = ConnectNamedPipe( pInfo->hPipe, &os );

      if( !cnpRet && ( GetLastError() == ERROR_IO_PENDING ) )
      {
         dwWait = WaitForMultipleObjects( 2, pInfo->hEvents, FALSE, INFINITE );

         // not overlapped i/o event or server stop signaled, exit thread
         if( dwWait != (WAIT_OBJECT_0 + PIPESRV_OVERLAPPED_IO_READ_EVENT) )
         {
            delete pInfo;
            goto acceptend;
         }
      }

      if( g_bPipeSrvDebugging )
         _tprintf( _TEXT("AcceptThread[%d]: connection started on pipe %x\n"),
                   GetCurrentThreadId(), pInfo->hPipe );

      // place pipe on active queue
      //
      EnterCriticalSection( &pqInfo->critSec );
      pqInfo->nActive++;
      pqInfo->qActive.push( pInfo );

      if( g_bPipeSrvDebugging )
         _tprintf( _TEXT("AcceptThread[%d]: calling open\n"),
                   GetCurrentThreadId() );

      if( pqInfo->fOpen && !(pqInfo->fOpen)( pInfo ) )
      {
         DisconnectNamedPipe( pInfo->hPipe );
         pqInfo->qActive.pop();
         pqInfo->qAvailable.push( pInfo );
         pqInfo->nActive--;

         SetEvent( pqInfo->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] );
         LeaveCriticalSection( &pqInfo->critSec );

         if( g_bPipeSrvDebugging )
            _tprintf( _TEXT("AcceptThread[%d]: cancelling connection\n"),
                      GetCurrentThreadId() );
         continue;
      }

      // signal thread that a pipe is available for processing
      //
      if( g_bPipeSrvDebugging )
         _tprintf( _TEXT("AcceptThread[%d]: signalling worker threads\n"),
                   GetCurrentThreadId() );
      LONG lCount = 0;
      BOOL bRet = ReleaseSemaphore( pqInfo->hEvents[PIPESRV_PIPE_ACTIVE_EVENT], 1,
                                    &lCount );
      if( g_bPipeSrvDebugging && !bRet )
      {
         LPVOID lpMsgBuf;

         FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                        FORMAT_MESSAGE_FROM_SYSTEM |
                        FORMAT_MESSAGE_IGNORE_INSERTS,
                        NULL, GetLastError(),
                        MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
                        (LPTSTR) &lpMsgBuf, 0, NULL );

         _tprintf( _TEXT("AcceptThread[%d]: %s\n"),
                   GetCurrentThreadId(), (LPCTSTR)lpMsgBuf );
         LocalFree( lpMsgBuf );
      }

      LeaveCriticalSection( &pqInfo->critSec );

      // check to see if we need more threads
      //
      if( ( pqInfo->nActive > pqInfo->nThreads )
        )
      {
         // spawn a new thread
         //
         EnterCriticalSection( &pqInfo->critSec );
         pqInfo->nThreads++;
         LeaveCriticalSection( &pqInfo->critSec );

         unsigned long tid = _beginthread( PipeWorkerThread, 0, args );

         if( g_bPipeSrvDebugging )
            _tprintf( _TEXT("AcceptThread[%d]: spawned new thread %d\n"),
                      GetCurrentThreadId(), tid );
      }
   }

acceptend:
   pqInfo->thread = 0;
   SetEvent( pqInfo->hEvents[PIPESRV_ACCEPTOR_DIED] );

   if( g_bPipeSrvDebugging )
      _tprintf( _TEXT("AcceptThread[%d]: ending\n"), GetCurrentThreadId() );
   _endthread();
}

static void PipeWorkerThread( void *args )
{
   LPPIPESRV_QUEUE_INFO pqInfo = (LPPIPESRV_QUEUE_INFO)args;
   LPPIPESRV_PIPE_INFO  pInfo;
   DWORD                dwWait;
   BOOL                 bDeactivate = TRUE;

   while( 1 )
   {
      HANDLE hEvents[2];

      hEvents[0] = pqInfo->hEvents[0];
      hEvents[1] = pqInfo->hEvents[PIPESRV_PIPE_ACTIVE_EVENT];

beginning:
      if( g_bPipeSrvDebugging )
         _tprintf( _TEXT("WorkerThread[%d]: waiting for active pipe\n"),
                   GetCurrentThreadId() );

      // wait for a pipe to be available
      //
      dwWait = WaitForMultipleObjects( 2, hEvents, FALSE, INFINITE );

      // not pipe active event or server stop signaled, exit thread
      if( dwWait != ( WAIT_OBJECT_0 + 1 ) )
         goto workerend;

      EnterCriticalSection( &pqInfo->critSec );
      if( pqInfo->qActive.empty() )
      {
         LeaveCriticalSection( &pqInfo->critSec );
         goto beginning;
      }
      pInfo = pqInfo->qActive.front();
      pqInfo->qActive.pop();
      LeaveCriticalSection( &pqInfo->critSec );

      // process data
      //
      if( pqInfo->fProcessing )
      {
         if( g_bPipeSrvDebugging )
            _tprintf( _TEXT("WorkerThread[%d]: processing pipe %x\n"),
                      GetCurrentThreadId(), pInfo->hPipe );
         bDeactivate = (*pqInfo->fProcessing)( pInfo );
      }

      // place pipe on proper queue
      //
      EnterCriticalSection( &pqInfo->critSec );
      if( bDeactivate )
      {
         if( g_bPipeSrvDebugging )
            _tprintf( _TEXT("WorkerThread[%d]: disconnecting pipe %x\n"),
                      GetCurrentThreadId(), pInfo->hPipe );

         // let specific close-time processing occur
         //
         if( pqInfo->fClose )
            (*pqInfo->fClose)( pInfo );

         // pipe is slated for deactivation, disconnect client
         DisconnectNamedPipe( pInfo->hPipe );

         pqInfo->qAvailable.push( pInfo );
         pqInfo->nActive--;

         SetEvent( pqInfo->hEvents[PIPESRV_PIPE_AVAILABLE_EVENT] );
      }
      else
      {
         pqInfo->qActive.push( pInfo );
         ReleaseSemaphore( pqInfo->hEvents[PIPESRV_PIPE_ACTIVE_EVENT], 1, NULL );
      }
      LeaveCriticalSection( &pqInfo->critSec );
   }

workerend:
   pInfo->thread = 0;
   SetEvent( pqInfo->hEvents[PIPESRV_WORKER_DIED] );

   if( g_bPipeSrvDebugging )
      _tprintf( _TEXT("WokerThread[%d]: ending\n"), GetCurrentThreadId() );

   _endthread();
}


