/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: EventGroupEvent.java,v 1.15 2001/04/05 18:40:59 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

package com.eventfs;

import java.io.*;
import java.util.*;
import java.util.zip.Adler32;
import com.endian.*;

public class EventGroupEvent extends EventObject
{
	// Private data
	private   int       mask;
	private   int       size;
	private   int       chksum;
	private   byte[]    data;
	private   Hashtable notes = new Hashtable();


	// Protected data
	protected int       dataOffset;


	// Constructors
	public EventGroupEvent( EventGroup group ) { super( group ); }
	public EventGroupEvent( EventGroup group, int newMask )
	{
		this(group);
		mask = newMask;
	}

	// Public methods
	public void addNote( Object key, Object val ) { notes.put(key, val); }
	public void clearNotes() { notes.clear(); }
	public EventGroup getEventGroup() { return (EventGroup)getSource(); }
	public int getMask() { return mask; }
	public Object getNote( Object key ) { return notes.get(key); }
	public int getSize() { return size; }

	public static EventGroupEvent readEvent( EventGroup eg, InputStream is ) throws Exception
	{
		EndianDataInputStream edis = new EndianDataInputStream( is );

		int maskIn = edis.readInt();
		int sizeIn = edis.readInt();
		int chksum = edis.readInt();

		EventGroupEvent event = eg.createEvent( maskIn );
		if( sizeIn > 0 )
		{
			event.setChecksum( chksum );
			event.setSize( sizeIn );
			event.readData( edis );
		}

		return event;
	}

	public void removeNote( Object key ) { notes.remove(key); }

	public void setChecksum( int newChecksum )
	{
		chksum = newChecksum;
	}

	public void setMask( int newMask ) throws IllegalArgumentException
	{
		if( mask > 0 )
			throw new IllegalArgumentException( "Cannot change mask of an event after it has been set" );
		mask = newMask;
	}

	public void setSize( int newSize ) throws IllegalArgumentException
	{
		if( newSize < 0 ) newSize = 0;
		if( size > 0 )
			throw new IllegalArgumentException( "Cannot change size of an event after it has been set" );
		size = newSize;
	}

	public int readData( InputStream is ) throws IOException, CorruptEventException
	{
		if( size <= 0 ) return 0;
		data = new byte[size];
		int nread = is.read( data, 0, size );

		// Do Adler-32 checksum test
		Adler32 adler = new Adler32();
		adler.update(data);
		if( (int)adler.getValue() != chksum )
			throw new CorruptEventException( this, "Checksum mismatch [" + chksum + " != " + adler.getValue() + "]" );

		return nread;
	}

	public int readData( InputStream is, int newSize ) throws IOException, CorruptEventException
	{
		setSize( newSize );
		return readData( is );
	}

	public byte[] getData() { return data; }

	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append(super.toString());
		sb.append("[mask ").append(Integer.toHexString(mask));
		sb.append(", chksum ").append(Integer.toHexString(chksum));
		sb.append(", size ").append(Integer.toHexString(size));
		sb.append("]");
		return sb.toString();
	}
}

