/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: EventStream.java,v 1.20 2001/04/05 18:40:59 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

package com.eventfs;

import java.io.*;
import java.util.*;

public class EventStream implements Runnable, EventGroupEventListener
{
	// Private data
	private String         name;
	private EventManager   em;
	private EventTypeTable eventTypes;
	private EventGroup     group;
	private Thread         eventThread;
	private Object         owner;

	// Protected data
	protected Vector       eventListeners = new Vector();


	// Constructors
	public EventStream( EventGroup eg ) throws InvalidEventStreamException
	{
		group = eg;
		try
		{
			em = EventManager.getEventManager();
		}
		catch( ClassNotFoundException cnfe )
		{
			throw new InvalidEventStreamException(this);
		}
	}

	public EventStream( EventStream es ) throws InvalidEventStreamException
	{
		this(es.group);
		eventTypes = new EventTypeTable( es.eventTypes );
	}

	public EventStream( EventGroup eg, String streamName ) throws InvalidEventStreamException
	{
		this(eg);
		try
		{
			accessStream( streamName );
		}
		catch( InvalidStatFileException isfe )
		{
			throw new InvalidEventStreamException(this, "Bad or missing stats file");
		}
	}


	// Public methods
	public void accessStream( String streamName ) throws InvalidEventStreamException,InvalidStatFileException
	{
		name = new String(streamName);
		try
		{
			eventTypes = new EventTypeTable(em.openStreamStats(this));
		}
		catch( FileNotFoundException fnfe )
		{
			throw new InvalidEventStreamException( this, "Unable to open event stream" );
		}

		if( group.manages( eventTypes ) == false )
			throw new InvalidEventStreamException( this, "Event tables not synchronized with group" );
	}

	public synchronized void addEventListener( EventGroupEventListener egel )
	{
		eventListeners.addElement( egel );
	}

	public synchronized void eventGroupEventOccurred( EventGroupEvent e ) throws EventException
	{
		if( !manages( e ) )
			throw new UnhandledEventException( e );
		for( Enumeration enum = eventListeners.elements();
		     enum.hasMoreElements(); )
			((EventGroupEventListener)enum.nextElement()).eventGroupEventOccurred( e );
	}

	public EventGroup getEventGroup() { return group; }
	public String getName() { return name; }
	public Object getOwner() { return owner; }

	public boolean manages( EventTypeTable ett )
	{
		return eventTypes.contains( ett );
	}

	public boolean manages( EventGroupEvent ege )
	{
		if( ege.getEventGroup() != group )
		{
//			System.err.println("Groups mismatched");
			return false;
		}
		else if( eventTypes.lookupTag( ege.getMask() ) == null )
		{
//			System.err.println("Unable to find tag for '" + ege.getMask() +  "'");
//			System.err.println(eventTypes.toString());
			return false;
		}
		else
			return true;
	}

	public InputStream open() throws FileNotFoundException
	{
		return em.openStream(this);
	}

	public synchronized void removeEventListener( EventGroupEventListener egel )
	{
		eventListeners.removeElement( egel );
	}

	public void listen() throws RuntimeException
	{
		if( eventThread != null )
			throw new RuntimeException("Stream already listening");

		// Start the thread
		eventThread = new Thread( this );
		eventThread.start();
	}

	public void setOwner( Object o ) { owner = o; }

	public void stop()
	{
		Thread t = eventThread;
		eventThread = null;
		t.interrupt();
	}

	public void run()
	{
		Thread myThread = Thread.currentThread();

		try
		{
			InputStream is = this.open();

			try
			{
				while( eventThread == myThread )
				{
					EventGroupEvent ege = EventGroupEvent.readEvent( group, is );
					eventGroupEventOccurred( ege );
				}
			}
			catch( UnhandledEventException uee )
			{
				uee.printStackTrace( System.err );
			}
			catch( Exception e )
			{
				e.printStackTrace( System.err );
			}

			try { is.close(); } catch( IOException ioe ) {}
		}
		catch( FileNotFoundException fnfe )
		{
			fnfe.printStackTrace( System.err );
		}
	}

	public String toString() { return "EventStream: " + name; }
}

