/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: EventTypeTable.java,v 1.16 2001/04/05 18:40:59 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

package com.eventfs;

import java.io.*;
import java.util.*;

public class EventTypeTable
{
	private static Character idStart = new Character('[');
	private static Character idEnd   = new Character(']');
	private String           id;
	private Hashtable        maskMap;
	private Hashtable        tagMap;

	// Constructors
	public EventTypeTable( EventTypeTable ett )
	{
		for( Enumeration e = ett.maskMap.keys(); e.hasMoreElements(); )
		{
			Object o = e.nextElement();
			maskMap.put( o, ett.maskMap.get( o ) );
		}

		for( Enumeration e = ett.tagMap.keys(); e.hasMoreElements(); )
		{
			Object o = e.nextElement();
			tagMap.put( o, ett.tagMap.get( o ) );
		}
	}

	public EventTypeTable( InputStream is ) throws InvalidStatFileException
	{
		maskMap = new Hashtable();
		tagMap = new Hashtable();

		accessTypes( is );
	}


	// Private methods
	private void throwException( StreamTokenizer st, String msg ) throws InvalidStatFileException
	{
		StringBuffer sb = new StringBuffer();
		switch( st.ttype )
		{
			case StreamTokenizer.TT_WORD:
				sb.append('"').append(st.sval).append('"');
				break;
			case StreamTokenizer.TT_EOL:
				sb.append("<EOL>");
				break;
			case StreamTokenizer.TT_EOF:
				sb.append("<EOF>");
				break;
			default:
				sb.append("'").append(st.ttype).append("'");
				break;
		}
		sb.append(": ").append(msg);
		throw new InvalidStatFileException(sb.toString());
	}


	// Public methods
	public void accessTypes( InputStream is ) throws InvalidStatFileException
	{
		maskMap.clear();
		tagMap.clear();

		// Read types from stream
		String encoding = "ASCII";
		if( System.getProperty("os.name").startsWith("Windows") )
			encoding = "UnicodeLittle";

		InputStreamReader isr;
		try { isr = new InputStreamReader(is, encoding); }
		catch( UnsupportedEncodingException uee )
		{ throw new InvalidStatFileException(uee.toString()); }

		Reader r = new BufferedReader(isr);
		StreamTokenizer st = new StreamTokenizer(r);
		st.resetSyntax();
		st.wordChars( 'a', 'z' );
		st.wordChars( 'A', 'Z' );
		st.wordChars( '0', '9' );
		st.wordChars( 128 + 32, 255 );
		st.whitespaceChars( 0, ' ' );
		st.eolIsSignificant( true );
		int tt;
		boolean processing = false;
		try
		{
			while( st.nextToken() != StreamTokenizer.TT_EOF )
			{
				if( processing )
				{
					// Read type entry
					if( st.ttype != StreamTokenizer.TT_WORD )
						throwException(st, "Invalid mask entry");
					Integer mask = new Integer(st.sval);
					if( mask.intValue() <= 0 )
						throwException(st, "Invalid mask entry");
					StringBuffer sb = new StringBuffer();
					while( st.nextToken() != StreamTokenizer.TT_EOL &&
					       st.ttype != StreamTokenizer.TT_EOF )
					{
						if( sb.length() > 0 ) sb.append(" ");
						if( st.ttype == StreamTokenizer.TT_NUMBER )
							sb.append( new Double(st.nval).intValue() );
						else if( st.ttype == StreamTokenizer.TT_WORD )
							sb.append( st.sval );
						else
							sb.append( st.ttype );
					}
					String tag = sb.toString();
					maskMap.put( mask, tag );
					tagMap.put( tag, mask );
					if( st.ttype != StreamTokenizer.TT_EOL )
						throwException(st, "Missing EOL where expected");
				}
				else
				{
					// Read type table header
					if( st.ttype != EventTypeTable.idStart.charValue() )
						throwException(st, "Missing or invalid ID start");
					StringBuffer sb = new StringBuffer();
					while( st.nextToken() != StreamTokenizer.TT_EOL &&
					       st.ttype != StreamTokenizer.TT_EOF &&
					       st.ttype != EventTypeTable.idEnd.charValue() )
					{
						if( sb.length() > 0 ) sb.append(" ");
						if( st.ttype == StreamTokenizer.TT_NUMBER )
							sb.append( new Double(st.nval).intValue() );
						else if( st.ttype == StreamTokenizer.TT_WORD )
							sb.append( st.sval );
						else
							sb.append( st.ttype );
					}
					id = new String(sb.toString());
					if( st.ttype != EventTypeTable.idEnd.charValue() )
						throwException(st, "Missing or invalid ID end");
					st.nextToken();
					if( st.ttype != StreamTokenizer.TT_EOL )
						throwException(st, "Missing EOL where expected");
					processing = true;
				}
			}
			is.close();
		}
		catch( IOException ioe )
		{
			if( !System.getProperty("os.name").startsWith("Windows") )
				throwException(st, ioe.toString());
			else
				try { is.close(); }
				catch( IOException ioe2 )
				{ throwException(st, ioe2.toString()); }
		}
	}

	public boolean contains( EventTypeTable ett )
	{
		for( Enumeration e = ett.tagMap.keys(); e.hasMoreElements(); )
		{
			String key = (String)e.nextElement();
			Integer val = ett.lookupMask(key);
			if( val == null ||
			    val.intValue() != lookupMask(key).intValue() )
				return false;
		}

		for( Enumeration e = ett.maskMap.keys(); e.hasMoreElements(); )
		{
			Integer key = (Integer)e.nextElement();
			String val = ett.lookupTag( key );
			if( val == null || !val.equals(lookupTag(key)) )
				return false;
		}
		return true;
	}

	public Integer lookupMask( String tag )
	{
		return (Integer)tagMap.get( tag );
	}

	public String lookupTag( Integer mask )
	{
		return (String)maskMap.get( mask );
	}

	public String lookupTag( int mask )
	{
		return lookupTag( new Integer( mask ) );
	}

	public int size() { return maskMap.size(); }

	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append("EventTypeTable: ");
		if( id != null )
		{
			sb.append("[").append(id).append("]");
			sb.append(System.getProperty("line.separator"));
			for( Enumeration e = tagMap.keys(); e.hasMoreElements(); )
			{
				String key = (String)e.nextElement();
				sb.append(key).append("\t\t").append(tagMap.get(key));
				sb.append(System.getProperty("line.separator"));
			}
		}
		else
		{
			sb.append("<invalid>");
		}
		return sb.toString();
	}
}

