/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: com_eventfs_ut.c,v 1.12 2001/02/23 19:02:17 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

#include <pi_defs.h>

#include <sys/module.h>
#include <sys/linker.h>
#include <sys/sysproto.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/unistd.h>
#include <sys/proc.h>
#include <sys/sysctl.h>
#if __FreeBSD_version >= 500000
#include <sys/kthread.h>
#endif


#include <eventfs.h>

static int test_mod_event __P(( module_t, int, void * ));

static moduledata_t test_mod =
{
   "com.eventfs unittest module",
   test_mod_event,
   NULL,
};
DECLARE_MODULE( com_eventfs_ut, test_mod, SI_SUB_KTHREAD_IDLE,
                SI_ORDER_MIDDLE );

#if __FreeBSD_version >= 500000
MODULE_VERSION(com_eventfs_ut, 1);
MODULE_DEPEND(com_eventfs_ut, eventfs, 1, 1, 1);
#endif

#define EFSGROUP      "java event test"

#define NULL_TYPE     "null"
#define PERIODIC_TYPE "periodic"
#define RANDOM_TYPE   "random"


/* Event type masks */
static u_long nulltype;
static u_long periodictype;
static u_long randomtype;


/* Kernel thread control variables */
static int test_run = TRUE;
static int test_initialized = TRUE;
static int tests_running = 0;

/* Sleep/wait variables */
static int test_periodic_wakeup;
static int test_random_wakeup;
static int test_waitinit;


/*
 * Periodic timer kernel thread
 */
static void test_periodic __P(( void ));
static struct proc *test_periodic_proc;
static struct kproc_desc test_periodic_kp =
{
   "periodic UT",
   test_periodic,
   &test_periodic_proc
};

#if __FreeBSD_version < 500000
SYSINIT_KT(test_periodic, SI_SUB_KTHREAD_IDLE, SI_ORDER_FIRST, kproc_start,
           &test_periodic_kp)
#else
SYSINIT(test_periodic, SI_SUB_KTHREAD_IDLE, SI_ORDER_FIRST, kproc_start,
        &test_periodic_kp)
#endif


static u_long periodic_count = 0;
static int periodic_interval = 20;

#if __FreeBSD_version < 500011
static int sysctl_vfs_eventfs_periodicinterval SYSCTL_HANDLER_ARGS;
#else
static int sysctl_vfs_eventfs_periodicinterval(SYSCTL_HANDLER_ARGS);
#endif

SYSCTL_DECL(_vfs_eventfs);
/* sysctl: vfs.eventfs.periodicinterval */
SYSCTL_PROC(_vfs_eventfs, OID_AUTO, periodicinterval, CTLTYPE_INT|CTLFLAG_RW,
            &periodic_interval, 0, sysctl_vfs_eventfs_periodicinterval, "I",
            "");


/*
 * Random timer kernel thread
 */
static void test_random __P(( void ));
static struct proc *test_random_proc;
static struct kproc_desc test_random_kp =
{
   "random UT",
   test_random,
   &test_random_proc
};

#if __FreeBSD_version < 500000
SYSINIT_KT(test_random, SI_SUB_KTHREAD_IDLE, SI_ORDER_FIRST, kproc_start,
           &test_random_kp)
#else
SYSINIT(test_random, SI_SUB_KTHREAD_IDLE, SI_ORDER_FIRST, kproc_start,
        &test_random_kp)
#endif

#define MAX_RAND_INTERVAL  120
#define MAX_RAND_PAYLOADSZ 8000

static int test_fini();
static int test_init();


static int
test_init()
{
   int    error;

   error = eventfs_register_group( EFSGROUP );
   if( !error )
      error = eventfs_register_event_type( EFSGROUP, NULL_TYPE, &nulltype );
   if( !error )
      error = eventfs_register_event_type( EFSGROUP, PERIODIC_TYPE, &periodictype );
   if( !error )
      error = eventfs_register_event_type( EFSGROUP, RANDOM_TYPE, &randomtype );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "raw",
                                       nulltype | periodictype | randomtype,
                                       FALSE );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "timed",
                                       periodictype | randomtype,
                                       FALSE );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "random", randomtype,
                                       FALSE );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "periodic", randomtype,
                                       FALSE );
   if( error )
      test_run = FALSE;

   test_initialized = TRUE;
   wakeup( &test_waitinit );

   if( error )
   {
      int fini_error = test_fini();
      if( fini_error )
         error = fini_error;
   }

   return (error);
}


static int
test_fini()
{
   int              error = 0;
   struct wait_args wa;

   test_run = FALSE;

   while( ( error == 0 ) && ( tests_running > 0 ) )
   {
      wakeup( &test_periodic_wakeup );
      wakeup( &test_random_wakeup );
      if( tests_running > 0 )
         tsleep( &tests_running, PUSER, "unload", 0 );
   }

   /* Wait for periodic thread to die */
   wa.pid = test_periodic_proc->p_pid;
   wa.status = NULL;
   wa.options = 0;
   wa.rusage = NULL;
   wait4( &proc0, &wa );

   /* Wait for random thread to die */
   wa.pid = test_random_proc->p_pid;
   wa.status = NULL;
   wa.options = 0;
   wa.rusage = NULL;
   wait4( &proc0, &wa );

   /* Destroy EventFS group */
   if( !error ) error = eventfs_destroy_group( EFSGROUP );

   return (error);
}

static void
test_periodic()
{
   tests_running++;

   /* Wait for initialization of EventFS group and streams */
   while( test_initialized == FALSE )
      tsleep( &test_waitinit, PUSER, "efsinit", 0 );

   while( test_run )
   {
      /* Sleep for interval */
      tsleep( &test_periodic_wakeup, PUSER, "sleep", hz * periodic_interval );
      test_periodic_wakeup = 0;

      /* Increment the count */
      periodic_count++;

      /* Post event to stream */
#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Posting periodic event #%d\n", periodic_count );
#endif /* EFS_TEST_DIAGNOSTICS */
      eventfs_post_event_by_types( EFSGROUP, periodictype, sizeof(u_long),
                                   &periodic_count );
   }

   --tests_running;
   wakeup( &tests_running );

   /* Exit the thread */
   kthread_exit( 0 );
}

static void
test_random()
{
   tests_running++;

   /* Wait for initialization of EventFS group and streams */
   while( test_initialized == FALSE )
      tsleep( &test_waitinit, PUSER, "efsinit", 0 );

   while( test_run )
   {
      u_char       *payload;
      unsigned int  size;
      int           random_interval;
      int           i;

      random_interval = (random() % MAX_RAND_INTERVAL) + 1;

#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Next random event in %d seconds\n", random_interval );
#endif /* EFS_TEST_DIAGNOSTICS */

      /* Sleep for interval */
      tsleep( &test_random_wakeup, PUSER, "sleep", hz * random_interval );
      test_random_wakeup = 0;

      /* Generate random size payload */
      size = (random() % MAX_RAND_PAYLOADSZ) + 1;
      MALLOC(payload, u_char *, size, M_TEMP, M_WAITOK);
      for( i = 0; i < size; i++ )
         payload[i] = (u_char)(random() % 256);

      /* Post event to stream */
#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Posting random event with payload of %d bytes\n", size );
#endif /* EFS_TEST_DIAGNOSTICS */
      eventfs_post_event_by_types( EFSGROUP, randomtype, size, payload );

      FREE(payload, M_TEMP);
   }

   --tests_running;
   wakeup( &tests_running );

   /* Exit the thread */
   kthread_exit( 0 );
}

static int
test_mod_event( module_t mod, int type, void *data )
{
   int error = 0;

   switch( type )
   {
      case MOD_UNLOAD:
         error = test_fini();
         break;

      case MOD_LOAD:
         error = test_init();
         break;

      default:
         break;
   }
   return (error);
}


#if __FreeBSD_version < 500011
static int
sysctl_vfs_eventfs_periodicinterval SYSCTL_HANDLER_ARGS
#else
static int
sysctl_vfs_eventfs_periodicinterval(SYSCTL_HANDLER_ARGS)
#endif
{
   int error = sysctl_handle_int(oidp, oidp->oid_arg1, oidp->oid_arg2, req);
   if( !error )
      wakeup( &test_periodic_wakeup );
   return error;
}

