/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: com_eventfs_ut_linux.c,v 1.7 2001/02/23 19:02:17 stevek Exp $
 *
 * Author: 
 *
 */

#define __KERNEL_SYSCALLS__

static int errno; /* Hack Hack */

#include <pi_defs.h>

#include <eventfs.h>

#include <linux/version.h>
#include <linux/random.h>
#include <linux/unistd.h>


#define EFSGROUP      "java event test"

#define NULL_TYPE     "null"
#define PERIODIC_TYPE "periodic"
#define RANDOM_TYPE   "random"

#define PERIODIC_EV_COUNT 50

/* Event type masks */
static u_long nulltype;
static u_long periodictype;
static u_long randomtype;


/*
 * Periodic timer kernel thread
 */
static u_long periodic_count = 0;
static int periodic_interval = 200;

/*
 * Pid's
 */
static int periodic_pid = 0;
static int random_pid = 0;

/*
 * wait_queue.  Sleep on this.
 */
static struct wait_queue       *sleep_obj = 0;
/*
 * long interruptable_sleep_on_timeout(struct wait_queue **p, long timeout);
 * wake_up_interruptable(struct wait_queue **p);
 */

/*
 * Random timer kernel thread
 */
#define MAX_RAND_INTERVAL  400
#define MAX_RAND_PAYLOADSZ 4000

static int done = 0;

static int periodic_main(void *arg)
{
    while (periodic_count < (PERIODIC_EV_COUNT + 1) && !done) {
        periodic_count++;

        /* Post event to stream */
#ifdef EFS_TEST_DIAGNOSTICS
        printk(KERN_NOTICE"Posting periodic event. mask=%2ld,val=%3ld\n",
                periodictype, periodic_count);
#endif
        eventfs_post_event_by_types( EFSGROUP, periodictype, sizeof(u_long),
                &periodic_count );

        /* Sleep for periodic_interval */
        interruptible_sleep_on_timeout(&sleep_obj, periodic_interval);
    }
    if (!done) {
        done = 1;
        wake_up_interruptible(&sleep_obj); /* Wake up random */
    }

    return 0;
}

static int random_main(void *arg)
{
    u_char             *payload;
    unsigned int        size;
    int                 random_byte;
    int                 random_interval;

    while (!done) {
        /* Generate random size payload */
        /* Size = ?? */
        get_random_bytes(&size, 2);             /* Evil endian cast. */
        size = (size % MAX_RAND_PAYLOADSZ) + 1;
        /* Value = ??.  Doing the whole buffer random takes too long */
        get_random_bytes(&random_byte,1);       /* Evil endian cast. */
        /* Allocate and set values */
        payload = kmalloc(size, GFP_KERNEL);
        memset(payload,random_byte,size);

#ifdef EFS_TEST_DIAGNOSTICS
        printk(KERN_NOTICE"Posting random event.   mask=%2ld,val=%2x,size=%x\n",
                randomtype, random_byte, size);
#endif
        /* Post event to stream */
        eventfs_post_event_by_types( EFSGROUP, randomtype, size, payload );

        kfree(payload);

        get_random_bytes(&random_interval, 2);
        random_interval = (random_interval % MAX_RAND_INTERVAL) + 10;

        interruptible_sleep_on_timeout(&sleep_obj, random_interval);
    }
    return 0;
}


int init_module(void)
{
    int                  error;

    error = eventfs_register_group( EFSGROUP );
    if( !error )
        error = eventfs_register_event_type( EFSGROUP, NULL_TYPE,
                &nulltype );
    if( !error )
        error = eventfs_register_event_type( EFSGROUP, PERIODIC_TYPE,
                &periodictype );
    if( !error )
        error = eventfs_register_event_type( EFSGROUP, RANDOM_TYPE,
                &randomtype );
#ifdef EFS_TEST_DIAGNOSTICS
    printk(KERN_NOTICE"event types are null:%ld,periodic:%ld,random:%ld\n",
            nulltype, periodictype, randomtype);
#endif

    if( !error )
        error = eventfs_register_stream( EFSGROUP, "raw",
                nulltype | periodictype | randomtype,
                FALSE );
    if( !error )
        error = eventfs_register_stream( EFSGROUP, "timed",
                periodictype | randomtype,
                FALSE );
    if( !error )
        error = eventfs_register_stream( EFSGROUP, "random", randomtype,
                FALSE );
    if( !error )
        error = eventfs_register_stream( EFSGROUP, "periodic", randomtype,
                FALSE );

    /* Create and start the threads. */
    if ( !error )
        periodic_pid = kernel_thread(periodic_main, NULL, 0);

    if ( !error && periodic_pid > 0 )
        random_pid   = kernel_thread(random_main,   NULL, 0);

    if( error )
        (void) eventfs_destroy_group( EFSGROUP );

    return (error);
}


int cleanup_module(void)
{
   int          error = eventfs_destroy_group( EFSGROUP );
   sigset_t     tmpsig;

   /* Ok, block here a while. */

   /* First, set the "quit now" flag.
    */
   done = 1;

   /* Second, wake everyone up.
    */
   wake_up_interruptible(&sleep_obj);

   /* Third, wait for the threads.  Signal code stolen from kmod.c */

   spin_lock_irq(&current->sigmask_lock);
   tmpsig = current->blocked;
   siginitsetinv(&current->blocked, sigmask(SIGKILL)|sigmask(SIGSTOP));
   recalc_sigpending(current);
   spin_unlock_irq(&current->sigmask_lock);

   if (periodic_pid > 0)
       waitpid(periodic_pid, NULL, __WCLONE);
   if (random_pid > 0)
       waitpid(random_pid, NULL, __WCLONE);

   spin_lock_irq(&current->sigmask_lock);
   current->blocked = tmpsig;
   recalc_sigpending(current);
   spin_unlock_irq(&current->sigmask_lock);

   return (error);
}

