/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: com_eventfs_ut.c,v 1.13 2001/02/23 19:02:18 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

#include <pi_defs.h>

#include <sys/systm.h>
#include <sys/modctl.h>
#include <sys/ksynch.h>
#include <sys/var.h>
#include <sys/ddi.h>

#include <eventfs.h>

extern struct mod_ops mod_miscops;

char _depends_on[] = KMOD_DEPENDS;

static struct modlmisc modlmisc =
{
   &mod_miscops,
   WSS_MAKESTR("com.eventfs unittest module")
};

static struct modlinkage modlinkage =
{
   MODREV_1,
   {
      &modlmisc,
      NULL
   }
};

#define EFSGROUP "java event test"

#define NULL_TYPE     "null"
#define PERIODIC_TYPE "periodic"
#define RANDOM_TYPE   "random"


/* Event type masks */
static u_long nulltype;
static u_long periodictype;
static u_long randomtype;


/* Kernel thread control variables */
static int test_initialized = FALSE;
static int test_run         = TRUE;
static int tests_running    = 0;


/* Sleep/wait variables */
static kmutex_t   test_initialized_mu;
static kmutex_t   tests_running_mu;
static kmutex_t   periodic_wakeup_mu;
static kmutex_t   random_wakeup_mu;
static kcondvar_t test_initialized_cv;
static kcondvar_t tests_running_cv;
static kcondvar_t periodic_wakeup_cv;
static kcondvar_t random_wakeup_cv;


/*
 * Periodic timer kernel thread
 */
static u_long periodic_count    = 0;
static int    periodic_interval = 20;


/*
 * Random timer kernel thread
 */
#define MAX_RAND_INTERVAL  120
#define MAX_RAND_PAYLOADSZ (PAGESIZE - 1)


static void test_periodic __P(( void * ));
static void test_random   __P(( void * ));
static int  random        __P(( void ));


int
_init()
{
   int error;

   error = eventfs_register_group( EFSGROUP );
   if( !error )
      error = eventfs_register_event_type( EFSGROUP, "null", &nulltype );
   if( !error )
      error = eventfs_register_event_type( EFSGROUP, "periodic", &periodictype );
   if( !error )
      error = eventfs_register_event_type( EFSGROUP, "random", &randomtype );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "raw",
                                       nulltype | periodictype | randomtype,
                                       FALSE );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "timed",
                                       periodictype | randomtype,
                                       FALSE );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "random", randomtype,
                                       FALSE );
   if( !error )
      error = eventfs_register_stream( EFSGROUP, "periodic", randomtype,
                                       FALSE );

   if( error )
   {
      (void) eventfs_destroy_group( EFSGROUP );
      return (error);
   }

   mutex_init( &test_initialized_mu, NULL, MUTEX_DRIVER, NULL );
   mutex_init( &tests_running_mu, NULL, MUTEX_DRIVER, NULL );
   mutex_init( &periodic_wakeup_mu, NULL, MUTEX_DRIVER, NULL );
   mutex_init( &random_wakeup_mu, NULL, MUTEX_DRIVER, NULL );
   cv_init( &test_initialized_cv, NULL, CV_DRIVER, NULL );
   cv_init( &tests_running_cv, NULL, CV_DRIVER, NULL );
   cv_init( &periodic_wakeup_cv, NULL, CV_DRIVER, NULL );
   cv_init( &random_wakeup_cv, NULL, CV_DRIVER, NULL );

   thread_create((caddr_t)NULL, 0, test_periodic, (caddr_t)NULL, 0, &p0,
                 TS_RUN, (v.v_maxsyspri - 4));
   thread_create((caddr_t)NULL, 0, test_random, (caddr_t)NULL, 0, &p0,
                 TS_RUN, (v.v_maxsyspri - 4));

   error = mod_install( &modlinkage );

   printf("Registered types:\n");
   printf("      null : %d\n", nulltype);
   printf("  periodic : %d\n", periodictype);
   printf("    random : %d\n", randomtype);

   mutex_enter( &test_initialized_mu );
   test_initialized = TRUE;
   cv_broadcast( &test_initialized_cv );
   mutex_exit( &test_initialized_mu );

   return (error);
}

int
_fini()
{
   int error = 0;

   test_run = FALSE;

   while( ( error == 0 ) && ( tests_running > 0 ) )
   {
      mutex_enter( &periodic_wakeup_mu );
      cv_broadcast( &periodic_wakeup_cv );
      mutex_exit( &periodic_wakeup_mu );

      mutex_enter( &random_wakeup_mu );
      cv_broadcast( &random_wakeup_cv );
      mutex_exit( &random_wakeup_mu );

      mutex_enter( &tests_running_mu );
      if( tests_running > 0 )
         cv_wait( &tests_running_cv, &tests_running_mu );
      mutex_exit( &tests_running_mu );
   }

   if( !error ) error = eventfs_destroy_group( EFSGROUP );
   if( !error ) error = mod_remove( &modlinkage );
   return (error);
}

int
_info( struct modinfo *modinfop )
{
   int error = mod_info( &modlinkage, modinfop );

   printf( "com_eventfs_ut: _info: module loaded at address 0x%x, "
           "size 0x%x\n", (unsigned int) modinfop->mi_base, modinfop->mi_size );

   return error;
}


/*
 * Not a real PRNG, but good enough for this testing
 */
static int
random()
{
   static int rand_val = 1;

   clock_t cur_ticks;
   drv_getparm( LBOLT, &cur_ticks );

   rand_val <<= 1;
   rand_val += cur_ticks;
   if( rand_val < 0 ) rand_val = -rand_val;
   return (rand_val);
}

static void
test_periodic( void *arg )
{
   mutex_enter( &tests_running_mu );
   tests_running++;
   mutex_exit( &tests_running_mu );

   mutex_enter( &test_initialized_mu );
   while( test_initialized == FALSE )
      cv_wait( &test_initialized_cv, &test_initialized_mu );
   mutex_exit( &test_initialized_mu );

   mutex_enter( &periodic_wakeup_mu );
   while( test_run )
   {
      clock_t cur_ticks, timeout;

      drv_getparm( LBOLT, &cur_ticks );
      timeout = cur_ticks + drv_usectohz( periodic_interval * 1000000 );

      /* Sleep for interval */
      cv_timedwait( &periodic_wakeup_cv, &periodic_wakeup_mu, timeout );

      /* Increment the count */
      periodic_count++;

      /* Post event to stream */
#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Posting periodic event #%d (type %d)\n", periodic_count,
              periodictype );
#endif /* EFS_TEST_DIAGNOSTICS */
      eventfs_post_event_by_types( EFSGROUP, periodictype, sizeof(u_long),
                                   &periodic_count );
   }
   mutex_exit( &periodic_wakeup_mu );

   mutex_enter( &tests_running_mu );
   --tests_running;
   cv_broadcast( &tests_running_cv );
   mutex_exit( &tests_running_mu );

   /* Exit the thread */
   thread_exit();
}

static void
test_random( void * arg )
{
   mutex_enter( &tests_running_mu );
   tests_running++;
   mutex_exit( &tests_running_mu );

   mutex_enter( &test_initialized_mu );
   while( test_initialized == FALSE )
      cv_wait( &test_initialized_cv, &test_initialized_mu );
   mutex_exit( &test_initialized_mu );

   mutex_enter( &periodic_wakeup_mu );
   while( test_run )
   {
      u_char       *payload;
      unsigned int  size;
      int           random_interval;
      int           i;
      clock_t       cur_ticks, timeout;

      random_interval = (random() % MAX_RAND_INTERVAL) + 1;
#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Next random event in %d seconds\n", random_interval );
#endif /* EFS_TEST_DIAGNOSTICS */

      drv_getparm( LBOLT, &cur_ticks );
      timeout = cur_ticks + drv_usectohz( random_interval * 1000000 );

      /* Sleep for interval */
      cv_timedwait( &periodic_wakeup_cv, &periodic_wakeup_mu, timeout );

      /* Generate random size payload */
      size = (random() % MAX_RAND_PAYLOADSZ) + 1;
#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Allocating payload of %d bytes\n", size );
#endif /* EFS_TEST_DIAGNOSTICS */
      payload = (u_char *)kmem_alloc(size, KM_SLEEP);
      for( i = 0; i < size; i++ )
         payload[i] = (u_char)(random() % 256);

      /* Post event to stream */
#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Posting random event with payload of %d bytes\n", size );
#endif /* EFS_TEST_DIAGNOSTICS */
      eventfs_post_event_by_types( EFSGROUP, randomtype, size, payload );

      kmem_free(payload, size);
   }
   mutex_exit( &periodic_wakeup_mu );

   mutex_enter( &tests_running_mu );
   --tests_running;
   cv_broadcast( &tests_running_cv );
   mutex_exit( &tests_running_mu );

   /* Exit the thread */
   thread_exit();
}

