/*! \file eventfs.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: eventfs.h,v 1.37 2001/02/23 19:02:15 stevek Exp $
 *
 * \brief   Header file for the eventfs extension
 * \author  Steve Kiernan
 * \version $Revision: 1.37 $
 * \date    $Date: 2001/02/23 19:02:15 $
 */

#ifndef __EVENTFS_H
#define __EVENTFS_H 1

#include <pdpi/pi_defs.h>
#include <pdpi/pi_mutex.h>

#include <pdpi/queue.h>


/***************************************************************************\
 * Define structures and types                                             *
\***************************************************************************/

struct eventfs_event
{
   u_long efsevent_type;
   size_t efsevent_size;
   u_long efsevent_chksum;
};

#ifdef PD_IN_KERNEL
struct eventfs_entry
{
   LIST_ENTRY(eventfs_entry) efsent_link;
   PD_MUTEX_PROTECT        ( EFSLINKAGE, efsent );
   int                       efsent_refcount;
   struct eventfs_event      efsent_event;
};

struct eventfs_stream_entry
{
   LIST_ENTRY(eventfs_stream_entry)  efstrent_link;
   PD_MUTEX_PROTECT                ( EFSLINKAGE, efstrent );
   int                               efstrent_refcount;
   struct eventfs_entry             *efstrent_entry;
};

struct eventfs_stream
{
   LIST_ENTRY(eventfs_stream)                          efstream_link;

   PD_MUTEX_PROTECT                                  ( EFSLINKAGE, efstream );
   int                                                 efstream_refcount;
   pd_char_t                                          *efstream_name;
   long                                                efstream_namelen;
   u_int                                               efstream_flags;
#define EFS_FLG_BLOCKED   0x01
#define EFS_FLG_GOING     0x02
#define EFS_FLG_TRANSIENT 0x04
   u_int16_t                                           efstream_id;
   u_long                                              efstream_mask;
   struct eventfs_group                               *efstream_group;
   struct timespec                                     efstream_ctime;
   PD_LIST_HEAD(efstream_head, eventfs_stream_entry)   efstream_entries;
   struct eventfs_stream_entry                        *efstream_tail;
   void                                               *efstream_data;
};
#define EVENTFS_NSTREAMS 16383

#define EFGROUP_NTYPES (sizeof(u_long)*8)
struct eventfs_group
{
   int                                          efgroup_refcount;
   pd_char_t                                   *efgroup_name;
   long                                         efgroup_namelen;
   u_int8_t                                     efgroup_id;
   u_long                                       efgroup_eventmask;
   struct timespec                              efgroup_ctime;
   PD_MUTEX_PROTECT                           ( EFSLINKAGE, efgroup );
   const pd_char_t                             *efgroup_types[EFGROUP_NTYPES];
   PD_LIST_HEAD(efgroup_head, eventfs_stream)   efgroup_streams;
   int                                          efgroup_nstreams;
   int                                          efgroup_nextstreamid;
   void                                        *efgroup_data;
};
#define EVENTFS_NGROUPS 127
extern struct eventfs_group *eventfs_groups[EVENTFS_NGROUPS];
extern int eventfs_ngroups;


/***************************************************************************\
 * Define macros                                                           *
\***************************************************************************/

#define EFIND_NONE   -1
#define EFIND_EXACT  0
#define EFIND_NEAR   1

#ifdef _EFS
#define EFSNAME(a)     EFSLINKAGE a
#define EFSPURENAME(a) a
#else
#define EFSNAME(a)     LINKAGE a
#define EFSPURENAME(a) a
#endif

#define EFSCHKSUM(c,e) adler32((c), (pd_byte_t *)(e) + sizeof(*e), \
                               (e)->efsent_event.efsevent_size)

#define EFS_DEBUG_GROUP		0x0001
#define EFS_DEBUG_STREAM	0x0002
#define EFS_DEBUG_EVENT		0x0004
#define EFS_DEBUG_MD_MASK	0xfff0


/***************************************************************************\
 * Include system header                                                   *
\***************************************************************************/

#include EFS_SYSTEM_HDR


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

/* Refcount structure control methods
 * ref:     add a reference
 * rele:    release a reference
 * destroy: Mark as going-away.
 */
int eventfs_stream_ref      (struct eventfs_stream *);
int eventfs_stream_rele     (struct eventfs_stream *);

int eventfs_strent_ref      (struct eventfs_stream_entry *);
int eventfs_strent_rele     (struct eventfs_stream_entry *);

int eventfs_entry_ref       (struct eventfs_entry  *);
int eventfs_entry_rele      (struct eventfs_entry  *);

int eventfs_group_ref       (struct eventfs_group *);
int eventfs_group_rele      (struct eventfs_group *);

/* Public function available to other kernel modules */
int  EFSNAME(eventfs_init)                ( void );
int  EFSNAME(eventfs_construct_stream)    ( const pd_char_t * , u_long,
                                            pd_char_t **);
int  EFSNAME(eventfs_destroy_group)       ( const pd_char_t * );
int  EFSNAME(eventfs_destroy_stream)      ( const pd_char_t *,
                                            const pd_char_t * );
int  EFSNAME(eventfs_flush_stream)        ( const pd_char_t *,
                                            const pd_char_t * );
void EFSNAME(eventfs_free_stream_name)    ( pd_char_t * );
int  EFSNAME(eventfs_get_event_mask)      ( const pd_char_t *,
                                            const pd_char_t *, u_long * );
int  EFSNAME(eventfs_post_event)          ( const pd_char_t *,
                                            const pd_char_t *, size_t, void * );
int  EFSNAME(eventfs_post_event_by_types) ( const pd_char_t *, u_long, size_t,
                                            const void *);
int  EFSNAME(eventfs_register_event_type) ( const pd_char_t *,
                                            const pd_char_t *, u_long * );
int  EFSNAME(eventfs_register_group)      ( const pd_char_t * );
int  EFSNAME(eventfs_register_stream)     ( const pd_char_t *,
                                            const pd_char_t *, u_long,
                                            pd_bool_t );
int  EFSNAME(eventfs_uninit)              ( void );

#ifdef __cplusplus
}
#endif

#endif /* PD_IN_KERNEL */

#endif /* __EVENTFS_H */

