/*! \file eventfs.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: eventfs_common.c,v 1.81 2001/06/21 20:22:23 stevek Exp $
 *
 * \brief   Common routines for the eventfs extension
 * \author  Steve Kiernan
 * \version $Revision: 1.81 $
 * \date    $Date: 2001/06/21 20:22:23 $
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <pdpi/pi_mutex.h>

#include <eventfs.h>
#include <eventfs_machine.h>

#include <errno.h>

#include <zlib.h>


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

struct eventfs_group *eventfs_groups[EVENTFS_NGROUPS];
int                   eventfs_ngroups = 0;
PD_MUTEX_PROTECT    ( EFSLINKAGE, eventfs_groups );
u_int                 eventfs_debug_mask = 0xffff;


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

static struct eventfs_group  *
_eventfs_find_group                 ( const pd_char_t *, pd_bool_t );
static u_long _eventfs_find_type    ( struct eventfs_group *,
                                      const pd_char_t * );
static struct eventfs_stream *
_eventfs_find_stream                ( struct eventfs_group *,
                                      const pd_char_t *, pd_bool_t );
static struct eventfs_stream *
_eventfs_find_stream_with_types     ( struct eventfs_group *, u_long, int *,
                                      pd_bool_t );

static void
_eventfs_post_event_to_stream_unsafe( struct eventfs_stream *,
                                      struct eventfs_entry * );
static int
_eventfs_post_event_by_types        ( struct eventfs_group *, u_long, size_t,
                                      const void * );
static void _eventfs_flush_stream   ( struct eventfs_stream * );


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static PD_LIST_HEAD(efsent_head, eventfs_entry) _events;
static PD_MUTEX_PROTECT( EFSLINKAGE, _events );


/***************************************************************************\
 * Refcount functions                                                      *
\***************************************************************************/

/*
 * Groups:
 * 
 * Groups are reference counted only so that ref-counted still-open
 * streams can work while the group is nuked beneath it. (work == not crash
 * the system),
 * 
 * Groups are created with a ref-count of 1, and are referenced by each
 * stream that is a member of the group.  The "destroy_group" eventfs
 * call dereferences the group, thus leading to its destruction once
 * all the streams it has quit.
 */
int
eventfs_group_ref( struct eventfs_group *eg )
{
   int retval = 0;

   PD_MUTEX_ENTER(EFSLINKAGE, eg->efgroup);
   if( eg->efgroup_refcount > 0 )
      retval = ++eg->efgroup_refcount;
   PD_MUTEX_EXIT(EFSLINKAGE, eg->efgroup);

   return (retval);
}

static int
eventfs_group_rele_unsafe( struct eventfs_group *eg )
{
    int retval = --eg->efgroup_refcount;
    if( retval < 1 )
    {
        /* Nuke it */
#ifdef EVENTFS_DIAGNOSTICS
        if (eg->efgroup_streams.lh_first != NULL)
            panic("eventfs: destroying in-use group '%s'",eg->efgroup_name);
#endif /* EVENTFS_DIAGNOSTICS */
        eg->efgroup_refcount = -1;
        PD_MUTEX_DESTROY( EFSLINKAGE, eg->efgroup );
        EFS_FREE( eg->efgroup_name );
        EFS_FREE( eg );

        return 0;
    }
    return (retval);
}

int
eventfs_group_rele( struct eventfs_group *eg )
{
    int retval;
    PD_MUTEX_ENTER(EFSLINKAGE, eg->efgroup);
    retval = eventfs_group_rele_unsafe( eg );
    if( retval > 0 )
       PD_MUTEX_EXIT(EFSLINKAGE, eg->efgroup);
    return retval;
}


/* Assumes stream comes in locked.  Does the actual de-alloc
 * of memory if its safe to do so.
 */
static void
_eventfs_stream_free_unsafe( struct eventfs_stream *es )
{
   eventfs_strent_rele( es->efstream_tail );

   /* Release our reference of the group */
   eventfs_group_rele( es->efstream_group );

   EFS_FREE( es->efstream_name );
   PD_MUTEX_DESTROY( EFSLINKAGE, es->efstream );
   EFS_FREE( es );
}

static void
_eventfs_stream_destroy_unsafe( struct eventfs_stream *es )
{
   struct eventfs_group *gr;

   es->efstream_flags |= EFS_FLG_GOING;

   /* Wake up any sleeping listeners and let them quit. */
   eventfs_machine_destroy_stream( es );

   gr = es->efstream_group;

   _eventfs_flush_stream( es );

#ifdef EVENTFS_DIAGNOSTICS
   if( es->efstream_entries.lh_first != es->efstream_tail )
       panic("eventfs: stream_destroy_unsafe called with non-empty event list");
#endif /* EVENTFS_DIAGNOSTICS */

   PD_MUTEX_ENTER( EFSLINKAGE, gr->efgroup );
   LIST_REMOVE( es, efstream_link );
   PD_MUTEX_EXIT( EFSLINKAGE, gr->efgroup );
}

int
eventfs_stream_ref( struct eventfs_stream *es )
{
   int retval = 0;

   PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
   if( ( (es->efstream_flags & EFS_FLG_GOING) == 0 ) &&
       ( es->efstream_refcount > 0 ) )
      retval = ++es->efstream_refcount;
   PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );

   return (retval);
}

int
eventfs_stream_rele_unsafe( struct eventfs_stream *es )
{
   int retval = --es->efstream_refcount;

   /* Nuke it? */
   if( retval < 1 )
   {
      if( es->efstream_flags & EFS_FLG_GOING )
      {
         _eventfs_stream_free_unsafe( es );
         /* No need to unlock, stream is destroyed */
      }
#ifdef EVENTFS_DIAGNOSTICS
      else
         panic("eventfs: eventfs_stream_rele called on non-going stream");
#endif /* EVENTFS_DIAGNOSTICS */
      return 0;
   }
   else if( ( retval == 1 ) &&
            ( (es->efstream_flags & EFS_FLG_TRANSIENT) != 0 ) )
   {
      /* Handle IOCTL-created streams */
      _eventfs_stream_destroy_unsafe( es );
      return eventfs_stream_rele_unsafe( es );
   }
   return (retval);
}

int
eventfs_stream_rele( struct eventfs_stream *es )
{
   int retval;
   PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
   retval = eventfs_stream_rele_unsafe( es );
   if( retval > 0 )
      PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );
   return retval;
}

static int
eventfs_stream_destroy( struct eventfs_stream *es )
{
   int f_prog;

   /*
    * is valid to "destroy" a stream that is in use.  The
    * useage parts have to be woken up and deregister themselves.
    */
   PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
   if( es->efstream_flags & EFS_FLG_GOING )
   {
      PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );
      return EBUSY;
   }

   _eventfs_stream_destroy_unsafe( es );

   /* Hack:  If the stream is a "Programatically created" stream
    * (ie, was not created by IOCTL in response for specific mask)
    * then give it an extra release here.
    * 
    * Possible race not an issue because no opens on the stream
    * will succeed since its going away.  (right, machine-dep parts?)
    */
   f_prog = (es->efstream_flags & EFS_FLG_TRANSIENT) ? FALSE : TRUE;

   if( !f_prog || ( eventfs_stream_rele_unsafe( es ) > 0 ) )
      PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );

   return (0);
}


int
eventfs_entry_rele( struct eventfs_entry *efsent )
{
   int retval;

   PD_MUTEX_ENTER(EFSLINKAGE, efsent->efsent);
   retval = --efsent->efsent_refcount;
   if( retval <= 1 )
   {
      /* Nuke it */
      efsent->efsent_refcount = -1;
      PD_MUTEX_ENTER( EFSLINKAGE, _events );
      LIST_REMOVE( efsent, efsent_link );
      PD_MUTEX_EXIT( EFSLINKAGE, _events );
      PD_MUTEX_DESTROY( EFSLINKAGE, efsent->efsent );
      EFS_FREE( efsent );

      return (0);
   }
   PD_MUTEX_EXIT(EFSLINKAGE, efsent->efsent);
   return (retval);
}

int
eventfs_entry_ref( struct eventfs_entry *efsent )
{
   int retval = 0;

   PD_MUTEX_ENTER( EFSLINKAGE, efsent->efsent );
   if( efsent->efsent_refcount > 0 )
      retval = ++efsent->efsent_refcount;
   PD_MUTEX_EXIT( EFSLINKAGE, efsent->efsent );

   return (retval);
}

int
eventfs_strent_rele( struct eventfs_stream_entry *estrent )
{
   int retval;

   PD_MUTEX_ENTER(EFSLINKAGE, estrent->efstrent);
   retval = --estrent->efstrent_refcount;
   if( retval <= 1 )
   {
      struct eventfs_entry *efsent = estrent->efstrent_entry;

      if( efsent )
         eventfs_entry_rele( efsent );

      /* Nuke it */
      estrent->efstrent_refcount = -1;
      LIST_REMOVE( estrent, efstrent_link );
      PD_MUTEX_DESTROY( EFSLINKAGE, estrent->efstrent );
      EFS_FREE( estrent );

      return (0);
   }
   PD_MUTEX_EXIT(EFSLINKAGE, estrent->efstrent);
   return (retval);
}

int
eventfs_strent_ref( struct eventfs_stream_entry *estrent )
{
   int retval = 0;

   PD_MUTEX_ENTER( EFSLINKAGE, estrent->efstrent );
   if( estrent->efstrent_refcount > 0 )
      retval = ++estrent->efstrent_refcount;
   PD_MUTEX_EXIT( EFSLINKAGE, estrent->efstrent );

   return (retval);
}


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

void
#ifdef __GNUC__
__inline__
#endif /* __GNUC__ */
_eventfs_checkpoint_stream_unsafe( struct eventfs_stream *es )
{
   if( es->efstream_tail->efstrent_link.le_next != NULL )
      panic("checkpoint_stream: non-NULL next pointer in sentinel");
   if( es->efstream_tail->efstrent_entry != NULL )
      panic("checkpoint_stream: non-NULL entry pointer in sentinel");
}

struct eventfs_group *
_eventfs_find_group( const pd_char_t *group, pd_bool_t ref )
{
   int i;

   PD_MUTEX_ENTER( EFSLINKAGE, eventfs_groups );
   for( i = 0; i < EVENTFS_NGROUPS; i++ )
      if( eventfs_groups[i] &&
          !_tcscmp( group, eventfs_groups[i]->efgroup_name ) )
      {
         if( ref && !eventfs_group_ref( eventfs_groups[i] ) )
            continue;
         PD_MUTEX_EXIT( EFSLINKAGE, eventfs_groups );
         return eventfs_groups[i];
      }
   PD_MUTEX_EXIT( EFSLINKAGE, eventfs_groups );

   return (NULL);
}

u_long
_eventfs_find_type( struct eventfs_group *eg, const pd_char_t *type )
{
   u_long ul;

   /* Make sure type does not already exist */
   for( ul = 0; ul < EFGROUP_NTYPES; ul++ )
      if( eg->efgroup_types[ul] &&
          !_tcscmp( eg->efgroup_types[ul], type ) )
         return (1<<ul);

   return (0);
}

struct eventfs_stream *
_eventfs_find_stream( struct eventfs_group *eg, const pd_char_t *stream,
                      pd_bool_t ref )
{
   struct eventfs_stream *es;

   PD_MUTEX_ENTER( EFSLINKAGE, eg->efgroup );
   for( es = eg->efgroup_streams.lh_first; es; es = es->efstream_link.le_next )
      if( !_tcscmp( stream, es->efstream_name ) )
      {
         if( ref && !eventfs_stream_ref( es ) )
            es = NULL;
         PD_MUTEX_EXIT( EFSLINKAGE, eg->efgroup );
         return es;
      }

   PD_MUTEX_EXIT( EFSLINKAGE, eg->efgroup );
   return (NULL);
}

struct eventfs_stream *
_eventfs_find_stream_with_types( struct eventfs_group *eg, u_long types,
                                 int *bounds, pd_bool_t ref )
{
   struct eventfs_stream *es;
   struct eventfs_stream *match  = NULL;
   pd_bool_t              exact  = TRUE;
   int                    status = EFIND_NONE;

   if( bounds && ( *bounds == EFIND_NEAR ) )
      exact = FALSE;

   PD_MUTEX_ENTER( EFSLINKAGE, eg->efgroup );
   for( es = eg->efgroup_streams.lh_first; es; es = es->efstream_link.le_next )
      if( ( es->efstream_mask & types ) == types )
      {
         if( ( es->efstream_mask != types ) && exact )
            continue;

         match = es;
         if( es->efstream_mask == types )
         {
            status = EFIND_EXACT;
            break;
         }
         status = EFIND_NEAR;
      }

   if( match && ref && !eventfs_stream_ref( match ) )
   {
      match = NULL;
      status = EFIND_NONE;
   }
   PD_MUTEX_EXIT( EFSLINKAGE, eg->efgroup );

   if( bounds )
      *bounds = status;

   return (match);
}

int
_eventfs_post_event_by_types( struct eventfs_group *eg, u_long types,
                              size_t size, const void *data )
{
   /* Arugments have been checked before entry here */
   struct eventfs_stream *es;
   struct eventfs_entry  *efsent = NULL;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_EVENT )
      pd_printf( PD_MAKESTR("eventfs: Posting event( 0x%x, 0x%x, 0x%x, 0x%x )\n"),
                  (u_int) eg, (u_int) types, size, (u_int) data );
#endif /* EVENTFS_DIAGNOSTICS */

   PD_MUTEX_ENTER( EFSLINKAGE, eg->efgroup );
   for( es = eg->efgroup_streams.lh_first; es; es = es->efstream_link.le_next )
   {
      if( ( ( types & es->efstream_mask ) != 0 ) &&
          ( es->efstream_refcount > 1 ) &&
          ( ( es->efstream_flags & EFS_FLG_BLOCKED ) == 0 ) )
      {
         /* If we have not created an entry, create it now */
         if( efsent == NULL )
         {
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_EVENT )
               pd_printf( PD_MAKESTR("eventfs: Created event entry (type 0x%x)\n"),
                           (unsigned int)types );
#endif /* EVENTFS_DIAGNOSTICS */

            EFS_MALLOC( efsent, struct eventfs_entry *,
                        sizeof(struct eventfs_entry) + size );
            efsent->efsent_refcount = 1;
            PD_MUTEX_INIT( EFSLINKAGE, efsent->efsent );
            efsent->efsent_event.efsevent_type = types;
            efsent->efsent_event.efsevent_size = size;
            if( data )
            {
               efsent->efsent_event.efsevent_chksum = adler32( 1L, data, size );
               bcopy( (char *)data, (char *)efsent + sizeof(*efsent), size );
            }
            else
               efsent->efsent_event.efsevent_chksum = 0;
            PD_MUTEX_ENTER( EFSLINKAGE, _events );
            LIST_INSERT_HEAD( &_events, efsent, efsent_link );
            PD_MUTEX_EXIT( EFSLINKAGE, _events );
         }

         /* Post event to the stream */
         PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
         _eventfs_post_event_to_stream_unsafe( es, efsent );
         PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );
      }
   }
   PD_MUTEX_EXIT( EFSLINKAGE, eg->efgroup );

   return (0);
}

void
_eventfs_post_event_to_stream_unsafe( struct eventfs_stream *es,
                                      struct eventfs_entry *efsent )
{
   struct eventfs_stream_entry *newestrent;
   struct eventfs_stream_entry *estrent;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_STREAM|EFS_DEBUG_EVENT) )
      pd_printf( PD_MAKESTR("eventfs: Posting event to stream %s::%s\n"),
                  es->efstream_group->efgroup_name, es->efstream_name );
#endif /* EVENTFS_DIAGNOSTICS */

#ifdef EVENTFS_DIAGNOSTICS
   _eventfs_checkpoint_stream_unsafe( es );
#endif /* EVENTFS_DIAGNOSTICS */

   estrent = es->efstream_tail;

   /*
    * Construct a new terminating node
    *
    * The terminating node inherits the reference count of the
    * new event node in order to keep track of open-stream references.
    */
   EFS_MALLOC( newestrent, struct eventfs_stream_entry *,
               sizeof(struct eventfs_stream_entry) );
   newestrent->efstrent_refcount = estrent->efstrent_refcount;
   PD_MUTEX_INIT( EFSLINKAGE, newestrent->efstrent );
   newestrent->efstrent_entry = NULL;

   /* Move the sentenial pointer and
    * remove the refcount from the stream */
   es->efstream_tail = newestrent;
   --estrent->efstrent_refcount;
   /* Push the new terminating node on the stream */
   LIST_INSERT_AFTER( estrent, newestrent, efstrent_link );

   /* Set the entry data for the node */
   efsent->efsent_refcount++;
   estrent->efstrent_entry = efsent;

   /* Call system-dependent function */
   eventfs_machine_post_event_to_stream( es, efsent );

#ifdef EVENTFS_DIAGNOSTICS
   _eventfs_checkpoint_stream_unsafe( es );
#endif /* EVENTFS_DIAGNOSTICS */
}

void
_eventfs_flush_stream( struct eventfs_stream *es )
{
   struct eventfs_stream_entry *estrent;
#ifdef EVENTFS_DIAGNOSTICS
   struct eventfs_stream_entry *oestrent = NULL;
#endif /* EVENTFS_DIAGNOSTICS */

   /* Make sure system-dependent part flushes references first */
   eventfs_machine_flush( es );


   while( ( estrent = es->efstream_entries.lh_first ) != es->efstream_tail )
   {
      struct eventfs_entry *efsent;
  
#ifdef EVENTFS_DIAGNOSTICS
      if( estrent == NULL )
         panic( "encountered NULL entry before sentinel, last entry 0x%x",
                (unsigned int) oestrent );
#endif /* EVENTFS_DIAGNOSTICS */
      PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
      LIST_REMOVE( estrent, efstrent_link );
      PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );

      efsent = estrent->efstrent_entry;
      if( efsent )
         eventfs_entry_rele( efsent );
      PD_MUTEX_DESTROY( EFSLINKAGE, estrent->efstrent );
      EFS_FREE( estrent );

#ifdef EVENTFS_DIAGNOSTICS
      oestrent = estrent;
#endif /* EVENTFS_DIAGNOSTICS */
   }
}

void
_eventfs_close_stream_unsafe( struct eventfs_stream *es,
                              struct eventfs_stream_entry *efstrent )
{
   struct eventfs_stream_entry *next = efstrent->efstrent_link.le_next;
   struct eventfs_stream_entry *last = NULL;

   do
   {
      next = efstrent->efstrent_link.le_next;

      eventfs_strent_rele( efstrent );

      last = efstrent;
      efstrent = next;
   }
   while( efstrent != NULL );

#ifdef EVENTFS_DIAGNOSTICS
   if( last != es->efstream_tail )
      panic("Did not stop at sentinel");
   if( es->efstream_entries.lh_first == NULL )
      panic("Stream entry list missing sentinel");
#endif /* EVENTFS_DIAGNOSTICS */

   eventfs_stream_rele_unsafe( es );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int
EFSNAME(eventfs_init)()
{
   LIST_INIT(&_events);
   PD_MUTEX_INIT( EFSLINKAGE, _events );
   PD_MUTEX_INIT( EFSLINKAGE, eventfs_groups );

   bzero( &eventfs_groups, sizeof(eventfs_groups) );

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask )
      pd_printf( PD_MAKESTR("eventfs_init: calling eventfs_machine_init()\n") );
#endif /* EVENTFS_DIAGNOSTICS */

   return (eventfs_machine_init());
}

int
EFSNAME(eventfs_uninit)()
{
   int retval = eventfs_machine_uninit();
   PD_MUTEX_DESTROY( EFSLINKAGE, eventfs_groups );
   PD_MUTEX_DESTROY( EFSLINKAGE, _events );
   return retval;
}

int
EFSNAME(eventfs_register_group)( const pd_char_t *group )
{
   struct eventfs_group *eg;
   int                   i;
   int                   openslot = EVENTFS_NGROUPS;

   if( group == NULL )
      return (EINVAL);

   PD_MUTEX_ENTER( EFSLINKAGE, eventfs_groups );

   for( i = 0; i < EVENTFS_NGROUPS; i++ )
   {
      if( eventfs_groups[i] == NULL )
      {
         if( openslot == EVENTFS_NGROUPS )
            openslot = i;
      }
      else if( !_tcscmp( group, eventfs_groups[i]->efgroup_name ) )
      {
         PD_MUTEX_EXIT( EFSLINKAGE, eventfs_groups );
         return (EEXIST);
      }
   }

   if( openslot == EVENTFS_NGROUPS )
   {
      PD_MUTEX_EXIT( EFSLINKAGE, eventfs_groups );
      return (ENOSPC);
   }

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_GROUP )
      pd_printf( PD_MAKESTR("eventfs_register_group(\"%s\")\n"), group );
#endif /* EVENTFS_DIAGNOSTICS */

   EFS_MALLOC( eg, struct eventfs_group *, sizeof(struct eventfs_group) );
   eg->efgroup_refcount = 1;
   eg->efgroup_eventmask = 0;
   eg->efgroup_data = NULL;
   eg->efgroup_namelen = _tcslen(group);
   EFS_MALLOC( eg->efgroup_name, pd_char_t *,
               sizeof(pd_char_t) * (eg->efgroup_namelen + 1) );
   bcopy( group, eg->efgroup_name, eg->efgroup_namelen * sizeof(pd_char_t) );
   eg->efgroup_name[eg->efgroup_namelen] = PD_MAKESTR('\0');
   bzero( &eg->efgroup_ctime, sizeof(struct timespec) );
   bzero( (void*)(&eg->efgroup_types), sizeof(eg->efgroup_types) );
   LIST_INIT( &eg->efgroup_streams );
   PD_MUTEX_INIT( EFSLINKAGE, eg->efgroup );
   eg->efgroup_nstreams = 0;
   eg->efgroup_nextstreamid = 0;
   eg->efgroup_id = openslot;
   eventfs_groups[openslot] = eg;
   eventfs_ngroups++;

   /* before we unlock: */
   eventfs_machine_register_group(eg);

   PD_MUTEX_EXIT( EFSLINKAGE, eventfs_groups );

   return (0);
}

int
EFSNAME(eventfs_register_event_type)( const pd_char_t *group,
                                      const pd_char_t *event_type,
                                      u_long *type_return )
{
   struct eventfs_group *eg;
   u_long                ul;

   if( ( group == NULL ) || ( event_type == NULL ) || ( type_return == NULL ) )
      return (EINVAL);

   /* Find the group structure */
   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( _eventfs_find_type( eg, event_type ) )
   {
      eventfs_group_rele( eg );
      return (EEXIST);
   }

   /* Find open bit in map */
   for( ul = 0; ul < EFGROUP_NTYPES; ul++ )
      if( eg->efgroup_types[ul] == NULL )
         break;

   if( ul == EFGROUP_NTYPES )
   {
      eventfs_group_rele( eg );
      return (ENOSPC);
   }

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_EVENT )
      pd_printf( PD_MAKESTR("eventfs_register_event_type(\"%s\", \"%s\") = %ld\n"),
                  group, event_type, ul );
#endif /* EVENTFS_DIAGNOSTICS */

   /* Set event type name */
   eg->efgroup_types[ul] = event_type;

   /* Add bit to mask */
   eg->efgroup_eventmask |= 1<<ul;

   /* Return bitmap to caller */
   *type_return = 1<<ul;

   eventfs_group_rele( eg );

   return (0);
}

int
EFSNAME(eventfs_register_stream)( const pd_char_t *group,
                                  const pd_char_t *stream,
                                  u_long mask, pd_bool_t transient )
{
   struct eventfs_group        *eg;
   struct eventfs_stream       *es;
   struct eventfs_stream_entry *estrent;

   if( ( group == NULL ) || ( stream == NULL ) )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( _eventfs_find_stream( eg, stream, FALSE ) )
   {
      eventfs_group_rele( eg );
      return (EEXIST);
   }

   if( eg->efgroup_nextstreamid >= EVENTFS_NSTREAMS )
   {
      eventfs_group_rele( eg );
      return (ENOSPC);
   }

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_STREAM )
      pd_printf( PD_MAKESTR("eventfs_register_stream(\"%s\", \"%s\", 0x%x)\n"),
                  group, stream, (unsigned int)mask );
#endif /* EVENTFS_DIAGNOSTICS */

   EFS_MALLOC( es, struct eventfs_stream *, sizeof(struct eventfs_stream) );
   es->efstream_refcount = 1;
   es->efstream_data = NULL;
   es->efstream_namelen = _tcslen(stream);
   EFS_MALLOC( es->efstream_name, pd_char_t *,
               sizeof(pd_char_t) * (es->efstream_namelen + 1) );
   bcopy( stream, es->efstream_name, es->efstream_namelen * sizeof(pd_char_t) );
   es->efstream_name[es->efstream_namelen] = PD_MAKESTR('\0');
   es->efstream_id = eg->efgroup_nextstreamid++;
   es->efstream_group = eg;               /* Referenced pointer to group */
   es->efstream_flags = (transient)?EFS_FLG_TRANSIENT:0;
   es->efstream_mask = mask & eg->efgroup_eventmask;
   bzero( &es->efstream_ctime, sizeof(es->efstream_ctime) );
   PD_MUTEX_INIT( EFSLINKAGE, es->efstream );
   LIST_INIT( &es->efstream_entries );
   EFS_MALLOC( estrent, struct eventfs_stream_entry *,
               sizeof(struct eventfs_stream_entry) );
   PD_MUTEX_INIT( EFSLINKAGE, estrent->efstrent );
   estrent->efstrent_refcount = 2; /* on list + tail ptr = 2 */
   estrent->efstrent_entry = NULL;
   es->efstream_tail = estrent;
   LIST_INSERT_HEAD( &(es->efstream_entries), estrent, efstrent_link );
   eg->efgroup_nstreams++;
   LIST_INSERT_HEAD( &(eg->efgroup_streams), es, efstream_link );

   /* before we unlock: */
   eventfs_machine_register_stream(es);

   /* Don't unlock the group, we have a reference to it */
   return 0;
}

int
EFSNAME(eventfs_post_event)( const pd_char_t *group, const pd_char_t *type,
                             size_t size, void *data )
{
   struct eventfs_group *eg;
   u_long                ul;
   int                   error;

   if( ( group == NULL ) || ( type == NULL ) )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( ( ul = _eventfs_find_type( eg, type ) ) == 0 )
   {
      eventfs_group_rele( eg );
      return (EINVAL);
   }

   /* ul is already a mask. */
   error = _eventfs_post_event_by_types( eg, ul, size, data );

   eventfs_group_rele( eg );

   return error;
}

int
EFSNAME(eventfs_post_event_by_types)( const pd_char_t *group, u_long types,
                                      size_t size, const void *data )
{
   struct eventfs_group *eg;
   int                   error;

   if( group == NULL )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( ( types & eg->efgroup_eventmask ) != types )
   {
      eventfs_group_rele( eg );
      return (EINVAL);
   }

   error = _eventfs_post_event_by_types( eg, types, size, data );

   eventfs_group_rele( eg );

   return error;
}

int
EFSNAME(eventfs_flush_stream)( const pd_char_t *group,
                               const pd_char_t *stream )
{
   struct eventfs_group        *eg;
   struct eventfs_stream       *es;

   if( ( group == NULL ) || ( stream == NULL ) )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( ( es = _eventfs_find_stream( eg, stream, TRUE ) ) == NULL )
   {
      eventfs_group_rele( eg );
      return (EINVAL);
   }

   eventfs_group_rele( eg );

   _eventfs_flush_stream( es );

   eventfs_stream_rele( es );

   return (0);
}

int
EFSNAME(eventfs_get_event_mask)( const pd_char_t *group,
                                 const pd_char_t *type_name, u_long *type )
{
   struct eventfs_group *eg;

   if( ( group == NULL ) || ( type_name == NULL ) || ( type == NULL ) )
      return (EINVAL);

   /* Find the group structure */
   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   *type = _eventfs_find_type( eg, type_name );

   eventfs_group_rele( eg );

   return (0);
}

int
EFSNAME(eventfs_block_stream)( const pd_char_t *group,
                               const pd_char_t *stream )
{
   struct eventfs_group        *eg;
   struct eventfs_stream       *es;

   if( ( group == NULL ) || ( stream == NULL ) )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( ( es = _eventfs_find_stream( eg, stream, TRUE ) ) == NULL )
   {
      eventfs_group_rele( eg );
      return (EINVAL);
   }

   eventfs_group_rele( eg );
   es->efstream_flags |= EFS_FLG_BLOCKED;
   eventfs_stream_rele( es );

   return (0);
}

int
EFSNAME(eventfs_unblock_stream)( const pd_char_t *group,
                                 const pd_char_t *stream )
{
   struct eventfs_group        *eg;
   struct eventfs_stream       *es;

   if( ( group == NULL ) || ( stream == NULL ) )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( ( es = _eventfs_find_stream( eg, stream, TRUE ) ) == NULL )
   {
      eventfs_group_rele( eg );
      return (EINVAL);
   }

   eventfs_group_rele( eg );
   es->efstream_flags &= ~EFS_FLG_BLOCKED;
   eventfs_stream_rele( es );

   return (0);
}

int
EFSNAME(eventfs_destroy_group)( const pd_char_t *group )
{
   struct eventfs_group                       *eg;
   struct eventfs_stream                      *es;
   int                                         i;
   PD_LIST_HEAD(streams_head, eventfs_stream)  stream_list;

   if( group == NULL )
      return EINVAL;

   /* Normally we'd use _eventfs_find_group, but we want to keep
    * the lock on the list for a bit afterwards.
    */
   eg = NULL;
   PD_MUTEX_ENTER( EFSLINKAGE, eventfs_groups );
   for( i = 0 ; i < EVENTFS_NGROUPS; ++i )
   {
      if( eventfs_groups[i] &&
          !_tcscmp( group,eventfs_groups[i]->efgroup_name ) )
      {
         /* Found it */
         eg = eventfs_groups[i];
         if( !eventfs_group_ref( eg ) )
            continue;
         break;
      }
   }

   if( eg == NULL )
      return EINVAL;

   /* Detach the group entry from the list */
   eventfs_groups[i] = NULL;

   /* Acquire lock on group */
   PD_MUTEX_ENTER( EFSLINKAGE, eg->efgroup );

   /* Ok, unlock the global list now. */
   PD_MUTEX_EXIT( EFSLINKAGE, eventfs_groups );

   LIST_INIT(&stream_list);

   /* Shanghai the stream list */
   if( eg->efgroup_streams.lh_first )
   {
      stream_list.lh_first = eg->efgroup_streams.lh_first;
      stream_list.lh_first->efstream_link.le_prev = &stream_list.lh_first;
      eg->efgroup_streams.lh_first = NULL;
   }

   PD_MUTEX_EXIT( EFSLINKAGE, eg->efgroup );

   /* Tell streams to go bye-bye */
   while ( ( es = stream_list.lh_first) )
      eventfs_stream_destroy( es );

   /* Re-acquire lock on group */
   PD_MUTEX_ENTER( EFSLINKAGE, eg->efgroup );

   /* Let the machine-specific part get a shot */
   eventfs_machine_destroy_group(eg);

   /* Release lock on group */
   PD_MUTEX_EXIT( EFSLINKAGE, eg->efgroup );
   
   /* Release the reference we acquired above */
   eventfs_group_rele( eg );
   /* Release the reference it was created with */
   eventfs_group_rele( eg );

   return (0);
}

int
EFSNAME(eventfs_destroy_stream)( const pd_char_t *group,
                                 const pd_char_t *stream )
{
   struct eventfs_group        *eg;
   struct eventfs_stream       *es;

   if( ( group == NULL ) || ( stream == NULL ) )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( ( es = _eventfs_find_stream( eg, stream, TRUE ) ) == NULL )
   {
      eventfs_group_rele( eg );
      return (EINVAL);
   }

   eventfs_group_rele( eg );
   eventfs_stream_destroy( es );
   eventfs_stream_rele( es );

   return (0);
}

int
EFSNAME(eventfs_construct_stream)( const pd_char_t *group, u_long mask,
                                   pd_char_t **retname )
{
   struct eventfs_group  *eg;
   struct eventfs_stream *es;
   pd_char_t             *name = NULL;

   if( ( group == NULL ) || ( mask == 0 ) )
      return (EINVAL);

   if( ( eg = _eventfs_find_group( group, TRUE ) ) == NULL )
      return (EINVAL);

   if( mask & ~( eg->efgroup_eventmask ) )
   {
      eventfs_group_rele( eg );
      return (EINVAL);
   }

   /* Search for stream with the exact types specified */
   es = _eventfs_find_stream_with_types( eg, mask, NULL, TRUE );

   if( es == NULL )
   {
      /* One does not exist, create a temporary stream */
      EFS_MALLOC( name, pd_char_t *, sizeof(pd_char_t) * 16 );
      eventfs_sprintf( name, PD_MAKESTR("stream0%x"), (unsigned int)mask );

      if( EFSNAME(eventfs_register_stream)( eg->efgroup_name, name, mask,
                                            TRUE ) )
      {
         EFS_FREE( name );
         name = NULL;
      }
   }
   else
   {
      /* One does exist, return its name */
      EFS_MALLOC( name, pd_char_t *,
                  sizeof(pd_char_t) * (es->efstream_namelen + 1) );
      bcopy( es->efstream_name, name, es->efstream_namelen * sizeof(pd_char_t) );
      name[es->efstream_namelen] = PD_MAKESTR('\0');
      eventfs_stream_rele( es );
   }

   eventfs_group_rele( eg );

   if( retname )
      *retname = name;
   else if( name )
      EFS_FREE( name );

   return (0);
}

void
EFSNAME(eventfs_free_stream_name)( pd_char_t *name )
{
   if( name )
      EFS_FREE( name );
}

