/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: eventfs_subr.c,v 1.36 2001/02/23 19:02:18 stevek Exp $
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <eventfs.h>
#include <eventfs_machine.h>
#include "eventfs_freebsd.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/vmmeter.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <sys/malloc.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/dirent.h>
#include <sys/resource.h>
#include <sys/fcntl.h>
#include <sys/poll.h>
#include <sys/signalvar.h>
#include <sys/sysctl.h>

#if __FreeBSD_version >= 500000
MODULE_VERSION(eventfs,1);
#endif

SYSCTL_NODE(_vfs, OID_AUTO, eventfs, CTLFLAG_RW, 0, "EventFS");
/* sysctl: vfs.eventfs.debugmask */
#if __FreeBSD_version < 500000
SYSCTL_INT(_vfs_eventfs, OID_AUTO, debugmask, CTLFLAG_RW,
           &eventfs_debug_mask, 0, "Debugging message mask");
#else
SYSCTL_UINT(_vfs_eventfs, OID_AUTO, debugmask, CTLFLAG_RW,
            &eventfs_debug_mask, 0, "Debugging message mask");
#endif

#define NEFCACHE 4
#define EF_NHASH(ix) (&efhashtbl[(ix) & efhash])
static LIST_HEAD(efhashhead, eventfs_opennode) *efhashtbl = NULL;
static u_long efhash = 0;

#define EFL_WANT   0x01
#define EFL_LOCKED 0x02
static int efcache_lock;

extern void _eventfs_close_stream_unsafe( struct eventfs_stream *,
                                          struct eventfs_stream_entry * );

int
eventfs_machine_init()
{
   efhashtbl = hashinit(NEFCACHE, M_CACHE, &efhash);

   return (0);
}

int  eventfs_machine_uninit() { return 0; }
void eventfs_machine_destroy_group( struct eventfs_group *eg ) {}
void eventfs_machine_destroy_stream( struct eventfs_stream *es ) {}
void eventfs_machine_register_group( struct eventfs_group *eg ) {}
void eventfs_machine_register_stream( struct eventfs_stream *es ) {}

void
eventfs_machine_flush( struct eventfs_stream *es )
{
   struct vnode *vp = NULL;
   int           i;

   for( i = 0; i < efhash; i++ )
   {
      struct eventfs_opennode *efsopn;
      struct efhashhead       *ef     = &efhashtbl[i];

      for( efsopn = ef->lh_first;
           efsopn;
           efsopn = efsopn->efsopen_hash.le_next )
      {
         if( ( efsopn->efsopen_node->efs_type == Estream ) &&
             ( efsopn->efsopen_node->un.efs_stream == es ) )
         {
            struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;

            if( vp == NULL )
               vp = efsopn->efsopen_node->efs_node;
   
            /* Decrement the reference count, release memory on last release */
            if( --efsopn->efsopen_refcount <= 1 )
            {
               PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
               _eventfs_close_stream_unsafe( es, efstrent );
               PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );
               LIST_REMOVE( efsopn, efsopen_hash );
               FREE( efsopn, M_TEMP );
            }

            efsopn->efsopen_curentry = es->efstream_tail;
         }
      }
   }

   /* Wake up any readers blocked on the stream */
   if( vp )
      wakeup( (caddr_t)vp );
}

void
eventfs_machine_post_event_to_stream( struct eventfs_stream *es,
                                      struct eventfs_entry *eent )
{
   if( es->efstream_data )
   {
      struct vnode *vp = (struct vnode *)es->efstream_data;

#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_STREAM )
         printf( "eventfs_machine_post_event_to_stream: posting\n" );
#endif /* EVENTFS_DIAGNOSTICS */

      /* Wake up any readers blocked on the stream */
      wakeup( (caddr_t)vp );

      /* Signal asynchronous reads */
      eventfs_sigasync( es );
   }
}

int
eventfs_check_stream( struct eventfs_node *efs )
{
   switch( efs->efs_type )
   {
      case Esdata:
      case Esstats:
         /* Stream has been destroyed and is going away */
         if( efs->un.efs_stream->efstream_flags & EFS_FLG_GOING )
         {
            int error = (vn_close( efs->efs_node, 0, curproc->p_cred->pc_ucred,
                                   curproc ));
            if( error )
               return error;
            return (EINTR);
         }
         break;

      default:
         break;
   }
   return (0);
}

int
eventfs_doopen( struct eventfs_node *efs, struct proc *p )
{
   struct eventfs_opennode *efsopn;
   struct efhashhead       *ef;
   int                      error;

   if( ( error = eventfs_check_stream( efs ) ) )
      return error;

   ef = EF_NHASH(curproc->p_pid);
loop:
   for( efsopn = ef->lh_first;
        efsopn && ( efsopn->efsopen_node != efs );
        efsopn = efsopn->efsopen_hash.le_next )
      continue;

   if( efsopn == NULL )
   {
      if( efcache_lock & EFL_LOCKED )
      {
         efcache_lock |= EFL_WANT;
         (void) tsleep( (caddr_t) &efcache_lock, PINOD, "efopen", 0 );
         goto loop;
      }
      efcache_lock |= EFL_LOCKED;

      /* Add a new opennode to the hash table */
      MALLOC( efsopn, struct eventfs_opennode *,
              sizeof(struct eventfs_opennode), M_TEMP, M_WAITOK );
      efsopn->efsopen_refcount = 1;
      efsopn->efsopen_node = efs;
      efsopn->efsopen_state = 0;
      efsopn->efsopen_sigio = NULL;
      efsopn->efsopen_offset = 0;
      efsopn->efsopen_curentry = efs->un.efs_stream->efstream_tail;
      efsopn->efsopen_curentry->efstrent_refcount++;
      efs->un.efs_stream->efstream_refcount++;

#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_OPEN )
         printf( "eventfs_doopen: \"%s\", refcount 0x%x, hash 0x%x\n",
                 efs->un.efs_stream->efstream_name,
                 (u_int) efs->un.efs_stream->efstream_refcount,
                 (u_int) (curproc->p_pid & efhash) );
#endif /* EVENTFS_DIAGNOSTICS */

      LIST_INSERT_HEAD( ef, efsopn, efsopen_hash );

      efcache_lock &= ~EFL_LOCKED;
      if( efcache_lock & EFL_WANT )
      {
         efcache_lock &= ~EFL_WANT;
         wakeup( (caddr_t) &efcache_lock );
      }
   }
   else
   {
      /* Increment the reference count */
      efsopn->efsopen_refcount++;
   }

   return (0);
}

int
eventfs_doclose( struct eventfs_node *efs, struct proc *p )
{
   struct eventfs_opennode *efsopn;

   efsopn = eventfs_lookup_opennode( p, efs );
   if( efsopn == NULL )
      panic( "eventfs_close: no matching eventfs_opennode" );
   else
   {
      struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;

      /* Decrement the reference count, release memory on last release */
      if( --efsopn->efsopen_refcount <= 1 )
      {
         PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
         _eventfs_close_stream_unsafe( efsopn->efsopen_node->un.efs_stream, efstrent );
         PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );
         LIST_REMOVE( efsopn, efsopen_hash );
         FREE( efsopn, M_TEMP );
      }
   }

   return (0);
}

int
eventfs_dodata( struct eventfs_node *efs, struct uio *uio,
                unsigned int *retnread, pd_bool_t *eof )
{
   struct eventfs_opennode *efsopn;
   int                      rval   = 0;
   unsigned int             nread  = 0;

   if( ( retnread == NULL ) || ( eof == NULL ) )
      return (EINVAL);

   efsopn = eventfs_lookup_opennode( curproc, efs );
   if( efsopn == NULL )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
         printf( "eventfs_dodata: unable to find open stream\n" );
#endif /* EVENTFS_DIAGNOSTICS */
      rval = EBADF;
   }
   else
   {
      /* We found the open stream */
      int off = efsopn->efsopen_offset;
      int seekoff = efsopn->efsopen_offset;
      int offmax = uio->uio_offset + efsopn->efsopen_offset;

      /* Seek into the stream */
      while( off < offmax )
      {
         struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;
         struct eventfs_entry        *efsent   = efstrent->efstrent_entry;

         if( efsent == NULL )
         {
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
               printf( "eventfs_dodata: no events pending\n" );
#endif /* EVENTFS_DIAGNOSTICS */
            *eof = TRUE;
            break;
         }

         if( ( off + sizeof(struct eventfs_event) ) > offmax )
         {
            seekoff = off + sizeof(struct eventfs_event) - offmax;
            break;
         }
         else
         {
            off += sizeof(struct eventfs_event);

            if( ( off + efsent->efsent_event.efsevent_size ) > offmax )
            {
               seekoff = off + efsent->efsent_event.efsevent_size -
                         offmax;
            }
            else
            {
               /* Flush this packet and move on */
               off += efsent->efsent_event.efsevent_size;
               efsopn->efsopen_curentry = efstrent->efstrent_link.le_next;

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_dodata: flushing packet\n" );
#endif /* EVENTFS_DIAGNOSTICS */

               eventfs_strent_rele( efstrent );
            }
         }
      }

      /* Read data from stream */
      if( off >= offmax )
      {
         while( ( rval == 0 ) && uio->uio_resid )
         {
            struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;
            struct eventfs_entry        *efsent   = efstrent->efstrent_entry;

            if( efsent == NULL )
            {
#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_dodata: no events pending\n" );
#endif /* EVENTFS_DIAGNOSTICS */
               *eof = TRUE;
               break;
            }
         
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
               printf( "eventfs_dodata: uio_resid = %d\n", uio->uio_resid );
#endif /* EVENTFS_DIAGNOSTICS */

            if( seekoff < sizeof(struct eventfs_event) )
            {
               /* Read in event head */
               int xlen = sizeof(struct eventfs_event) - seekoff;

               xlen = min( xlen, uio->uio_resid );

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_dodata: reading event head (offset %d, len %d)\n",
                          seekoff, xlen );
#endif /* EVENTFS_DIAGNOSTICS */

               if( xlen <= 0 )
               {
#ifdef EVENTFS_DIAGNOSTICS
                  if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                     printf( "eventfs_dodata: no space left in buffer\n" );
#endif /* EVENTFS_DIAGNOSTICS */

                  rval = 0;
                  break;
               }

               /* Write out event head chunk to requestor */
               rval = uiomove( (char *) (&efsent->efsent_event) + seekoff, xlen,
                               uio );
               if( rval )
                  break;

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_dodata: adjusting offset by %d\n", xlen );
#endif /* EVENTFS_DIAGNOSTICS */
               nread += xlen;
               seekoff += xlen;
            }
            else
            {
               int xlen;

               off = seekoff - sizeof(struct eventfs_event);

               /* Read in event data packet */
               xlen = efsent->efsent_event.efsevent_size - off;

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_dodata: reading event data packet\n" );
#endif /* EVENTFS_DIAGNOSTICS */

               if( xlen > 0 )
               {
                  /* Write out event data packet chunk to requestor */
                  rval = uiomove( (char *) (efsent + 1) + off, xlen, uio );
                  if( rval )
                     break;

                  nread += xlen;
                  seekoff += xlen;
               }

               if( seekoff >= ( efsent->efsent_event.efsevent_size +
                                sizeof(struct eventfs_event) ) )
               {
                  /* Flush this packet and move on */
                  seekoff = 0;
                  efsopn->efsopen_curentry = efstrent->efstrent_link.le_next;

#ifdef EVENTFS_DIAGNOSTICS
                  if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                     printf( "eventfs_dodata: flushing packet\n" );
#endif /* EVENTFS_DIAGNOSTICS */

                  eventfs_strent_rele( efstrent );
               }
            }
         }

         /* Adjust offset to be relative to remaining events in the stream */
         uio->uio_offset = 0;
         efsopn->efsopen_offset = seekoff;
      }
   }

   *retnread = nread;
   return (rval);
}


int
eventfs_dostats( struct eventfs_node *efs, struct uio *uio )
{
   struct eventfs_group *eg;
   char                  sbuf[PAGE_SIZE];
   u_long                mask;
   int                   ndx = 0;
   int                   i;

   if( uio->uio_rw != UIO_READ )
      return (EOPNOTSUPP);

   if( uio->uio_offset != 0 )
      return (0);

   switch( efs->efs_type )
   {
      case Esstats:
         eg   = efs->un.efs_stream->efstream_group;
         mask = efs->un.efs_stream->efstream_mask;
         ndx = snprintf( sbuf, PAGE_SIZE - 1, "[%s::%s]\n",
                         eg->efgroup_name, efs->un.efs_stream->efstream_name );
         break;

      case Egstats:
         eg   = efs->un.efs_group;
         mask = efs->un.efs_group->efgroup_eventmask;
         ndx = snprintf( sbuf, PAGE_SIZE - 1, "[%s]\n", eg->efgroup_name );
         break;

      default:
         return (EOPNOTSUPP);
   }


   for( i = 0; ( i < EFGROUP_NTYPES ) && ( ndx < (PAGE_SIZE - 1) ); i++ )
   {
      if( eg->efgroup_types[i] && ( mask & (((u_long)1)<<i) ) )
         ndx += snprintf( sbuf + ndx, PAGE_SIZE - ndx - 1, "%lu\t%s\n",
                          ((u_long)1<<i), eg->efgroup_types[i] );
   }
   sbuf[PAGE_SIZE-1] = '\0';

   ndx = min( ndx, uio->uio_resid );

   return (uiomove( sbuf, ndx, uio ));
}

int
eventfs_dopoll( struct eventfs_node *efs, struct vnode *vp, struct proc *p,
                int events )
{
   struct eventfs_opennode *efsopn;
   int                      rval   = 0;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_POLL )
      printf( "eventfs_poll( 0x%x, 0x%x )\n", (u_int) efs, (u_int) events );
#endif /* EVENTFS_DIAGNOSTICS */

   if( ( events & (POLLIN|POLLRDNORM) ) == 0 )
      return (0);

   efsopn = eventfs_lookup_opennode( curproc, efs );
   if( efsopn == NULL )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_POLL )
         printf( "eventfs_poll: unable to find open stream\n" );
#endif /* EVENTFS_DIAGNOSTICS */
   }
   else if( efs->efs_type == Esdata )
   {
      /* We found the open stream */
      struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;
      struct eventfs_entry        *efsent   = efstrent->efstrent_entry;

      if( efsent == NULL )
      {
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_POLL )
            printf( "eventfs_poll: no events pending\n" );
#endif /* EVENTFS_DIAGNOSTICS */
      }
      else
         rval |= events & (POLLIN|POLLRDNORM);
   }

   if( !rval )
      rval = vn_pollrecord( vp, p, events );

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_POLL )
      printf( "eventfs_poll: returning %d\n", rval );
#endif /* EVENTFS_DIAGNOSTICS */

   return (rval);
}

void
eventfs_sigasync( struct eventfs_stream *es )
{
   struct vnode *vp  = (struct vnode *)es->efstream_data;
   int           i;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_ASYNC )
      printf( "eventfs_sigasync: 0x%x\n", (u_int) es );
#endif /* EVENTFS_DIAGNOSTICS */

   /* Notify polling processes of event */
   VN_POLLEVENT( vp, POLLRDNORM );

   /* Notify SIGIO handlers */
   for( i = 0; i < NEFCACHE; i++ )
   {
      struct eventfs_opennode *efsopn;
      struct efhashhead       *ef     = &efhashtbl[i];

      for( efsopn = ef->lh_first;
           efsopn;
           efsopn = efsopn->efsopen_hash.le_next )
         if( ( efsopn->efsopen_node->efs_type == Esdata ) &&
             ( efsopn->efsopen_node->un.efs_stream == es ) &&
             ( efsopn->efsopen_state & EFS_NODE_ASYNC ) &&
             ( efsopn->efsopen_sigio ) )
         {
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_ASYNC )
               printf( "eventfs_sigasync: SIGIO to 0x%x\n",
                       (u_int) efsopn->efsopen_sigio );
#endif /* EVENTFS_DIAGNOSTICS */
            pgsigio( efsopn->efsopen_sigio, SIGIO, 0 );
         }
   }
}

struct eventfs_opennode *
eventfs_lookup_opennode( struct proc *p, struct eventfs_node *efs )
{
   struct efhashhead       *ef;
   struct eventfs_opennode *efsopn;

   ef = EF_NHASH(p->p_pid);
   for( efsopn = ef->lh_first;
        efsopn && ( efsopn->efsopen_node != efs );
        efsopn = efsopn->efsopen_hash.le_next )
      continue;

   return efsopn;
}


