/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: eventfs_vnops.c,v 1.30 2001/02/23 19:02:18 stevek Exp $
 *
 * Event filesystem (/event)
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <eventfs.h>
#include <eventfs_machine.h>
#include "eventfs_freebsd.h"
#include "eventfs_ioctl.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/vmmeter.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <sys/malloc.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/namei.h>
#include <sys/dirent.h>
#include <sys/resource.h>
#include <sys/fcntl.h>
#include <sys/poll.h>
#include <sys/filio.h>

static int eventfs_access   __P(( struct vop_access_args *ap ));
static int eventfs_badop    __P(( void ));
static int eventfs_getattr  __P(( struct vop_getattr_args *ap ));
static int eventfs_inactive __P(( struct vop_inactive_args *ap ));
static int eventfs_ioctl    __P(( struct vop_ioctl_args *ap ));
static int eventfs_lookup   __P(( struct vop_lookup_args *ap ));
static int eventfs_poll     __P(( struct vop_poll_args *ap ));
static int eventfs_print    __P(( struct vop_print_args *ap ));
static int eventfs_read     __P(( struct vop_read_args *ap ));
static int eventfs_readdir  __P(( struct vop_readdir_args *ap ));
static int eventfs_reclaim  __P(( struct vop_reclaim_args *ap ));
static int eventfs_setattr  __P(( struct vop_setattr_args *ap ));

extern int eventfs_doopen   __P(( struct eventfs_node *, struct proc * ));
extern int eventfs_doclose  __P(( struct eventfs_node *, struct proc * ));
extern int eventfs_dostats  __P(( struct eventfs_node *, struct uio * ));
extern int eventfs_dodata   __P(( struct eventfs_node *, struct uio *,
                                  unsigned int *, pd_bool_t * ));
extern int eventfs_dopoll   __P(( struct eventfs_node *, struct vnode *,
                                  struct proc *, int ));

static struct stream_target
{
   u_char   st_type;
   u_char   st_namelen;
   char    *st_name;
   efstype  st_efstype;
}
stream_targets[] =
{
#define N(s) sizeof(s)-1, s
   { DT_DIR, N("."),     Estream },
   { DT_DIR, N(".."),    Egroup },
   { DT_REG, N("data"),  Esdata },
   { DT_REG, N("stats"), Esstats }
#undef N
};
static const int nstream_targets =
   sizeof(stream_targets) / sizeof(stream_targets[0]);

static struct group_target
{
   u_char   gt_type;
   u_char   gt_namelen;
   char    *gt_name;
   efstype  gt_efstype;
}
group_targets[] =
{
#define N(s) sizeof(s)-1, s
   { DT_DIR, N("."),     Egroup },
   { DT_DIR, N(".."),    Eroot },
   { DT_REG, N("stats"), Egstats }
#undef N
};
static const int ngroup_targets =
   sizeof(group_targets) / sizeof(group_targets[0]);

#define EFL_WANT   0x01
#define EFL_LOCKED 0x02
static int efsvplock;

static struct eventfs_node *efshead;

int
eventfs_allocvp( mp, vpp, strgrp, efs_type )
   struct mount  *mp;
   struct vnode **vpp;
   void          *strgrp;
   efstype        efs_type;
{
   struct proc          *p     = curproc;
   struct eventfs_node  *efs;
   struct vnode         *vp;
   struct eventfs_node **ep;
   int                   error;

loop:
   for( efs = efshead; efs; efs = efs->efs_next )
   {
      vp = EVENTFSTOV(efs);
      if( ( efs->efs_type == efs_type ) && ( vp->v_mount == mp ) )
      {
         switch( efs_type )
         {
            case Eroot:
               efs->efs_fileno = EVENTFS_FILENO(0, 0, 0, Eroot);
               break;

            case Egroup:
               if( efs->un.efs_group != (struct eventfs_group *)strgrp )
                  continue;

               efs->efs_fileno = EVENTFS_FILENO(efs->un.efs_group->efgroup_id,
                                                0, 0, Egroup);
               break;

            case Egstats:
               if( efs->un.efs_group != (struct eventfs_group *)strgrp )
                  continue;

               efs->efs_fileno = EVENTFS_FILENO(efs->un.efs_group->efgroup_id,
                                                0, 1, Egstats);
               break;

            case Estream:
               if( efs->un.efs_stream != (struct eventfs_stream *)strgrp )
                  continue;

               efs->efs_fileno =
                  EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                                 efs->un.efs_stream->efstream_id, 0, Estream);
               break;

            case Esdata:
               if( efs->un.efs_stream != (struct eventfs_stream *)strgrp )
                  continue;

               efs->efs_fileno =
                  EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                                 efs->un.efs_stream->efstream_id, 1, Esdata);
               break;

            case Esstats:
               if( efs->un.efs_stream != (struct eventfs_stream *)strgrp )
                  continue;

               efs->efs_fileno =
                  EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                                 efs->un.efs_stream->efstream_id, 2, Esstats);
               break;

            default:
               panic( "eventfs_allocvp" );
         }

         if( vget( vp, 0 , p ) )
            goto loop;

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
            printf( "eventfs_allocvp: returning vnode (0x%x)\n", (u_int) vp );
#endif /* EVENTFS_DIAGNOSTICS */

         *vpp = vp;
         return (0);
      }
   }

   if( efsvplock & EFL_LOCKED )
   {
      efsvplock |= EFL_WANT;
      (void) tsleep( (caddr_t) &efsvplock, PINOD, "efsavp", 0 );
      goto loop;
   }
   efsvplock |= EFL_LOCKED;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_allocvp: allocating structures\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   MALLOC( efs, struct eventfs_node *, sizeof(struct eventfs_node), M_TEMP,
           M_WAITOK );

   if( ( error = getnewvnode( VT_EVENTFS, mp, eventfs_vnodeop_p, vpp ) ) )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
         printf( "eventfs_allocvp: error %d in getnewvnode\n", error );
#endif /* EVENTFS_DIAGNOSTICS */
      FREE( efs, M_TEMP );
      goto out;
   }

   vp = *vpp;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_allocvp: setting values for vnode (0x%x)\n", (u_int) vp );
#endif /* EVENTFS_DIAGNOSTICS */

   vp->v_data = efs;
   efs->efs_lockowner = 0;
   efs->efs_next = NULL;
   efs->efs_type = efs_type;
   switch( efs_type )
   {
      case Egroup:
         efs->un.efs_group = (struct eventfs_group *) strgrp;
         vp->v_type = VDIR;
         break;

      case Egstats:
         efs->un.efs_group = (struct eventfs_group *) strgrp;
         vp->v_type = VREG;
         break;

      case Estream:
         efs->un.efs_stream = (struct eventfs_stream *) strgrp;
         vp->v_type = VDIR;
         break;

      case Esdata:
      {
         struct eventfs_stream *es = (struct eventfs_stream *) strgrp;
         efs->un.efs_stream = es;
         es->efstream_data = (void *) vp;
         vp->v_type = VFIFO;
         break;
      }

      case Esstats:
         efs->un.efs_stream = (struct eventfs_stream *) strgrp;
         vp->v_type = VREG;
         break;

      case Eroot:      /* /event = dr-xr-xr-x */
         efs->un.efs_group = NULL;
         vp->v_type = VDIR;
         vp->v_flag = VROOT;
         break;

      default:
         panic( "eventfs_allocvp" );
         break;
   }
   efs->efs_node = vp;

   /* Append to end of list */
   for( ep = &efshead; *ep; ep = &(*ep)->efs_next )
      continue;
   *ep = efs;

out:
   efsvplock &= ~EFL_LOCKED;

   if( efsvplock & EFL_WANT )
   {
      efsvplock &= ~EFL_WANT;
      wakeup( (caddr_t) &efsvplock );
   }

   return (error);
}


static int
eventfs_open(ap)
   struct vop_open_args /* {
      struct vnode *a_vp;
      int           a_mode;
      struct ucred *a_cred;
      struct proc  *a_p;
   } */ *ap;
{
   struct eventfs_node     *efs = VTOEVENTFS(ap->a_vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_open\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( ap->a_mode & FWRITE )
      return (EPERM);

   if( efs->efs_type == Esdata )
      return eventfs_doopen( efs, ap->a_p );

   return (0);
}


static int
eventfs_close(ap)
   struct vop_close_args /* {
      struct vnode *a_vp;
      int           a_fflag;
      struct ucred *a_cred;
      struct proc  *a_p;
   } */ *ap;
{
   struct eventfs_node     *efs = VTOEVENTFS(ap->a_vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_close\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( efs->efs_type == Esdata )
      eventfs_doclose( efs, ap->a_p );

   return (0);
}


/*
 * vp is the current namei directory
 * ndp is the name to locate in that directory...
 */
static int
eventfs_lookup(ap)
   struct vop_lookup_args /* {
      struct vnode * a_dvp;
      struct vnode ** a_vpp;
      struct componentname * a_cnp;
   } */ *ap;
{
   struct componentname  *cnp   = ap->a_cnp;
   struct vnode         **vpp   = ap->a_vpp;
   struct vnode          *dvp   = ap->a_dvp;
   char                  *pname = cnp->cn_nameptr;
   struct eventfs_node   *efs;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_lookup\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   *vpp = NULL;

   if( ( cnp->cn_nameiop == DELETE ) || ( cnp->cn_nameiop == RENAME ) )
      return (EROFS);

   if( ( cnp->cn_namelen == 1 ) && ( *pname == '.' ) )
   {
      *vpp = dvp;
      VREF(dvp);
      return (0);
   }

   efs = VTOEVENTFS(dvp);

   switch( efs->efs_type )
   {
      case Eroot:
      {
         struct eventfs_group *eg = NULL;
         int                   i;

         if( cnp->cn_flags & ISDOTDOT )
            return (EIO);

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf( "eventfs_lookup(Eroot): %s [%ld]\n", pname, cnp->cn_namelen );
#endif /* EVENTFS_DIAGNOSTICS */

         for( i = 0; i < EVENTFS_NGROUPS; i++ )
         {
            eg = eventfs_groups[i];
            if( eg == NULL )
               continue;

            if( ( cnp->cn_namelen == eg->efgroup_namelen ) &
                ( bcmp( eg->efgroup_name, pname, cnp->cn_namelen ) == 0 ) )
               return (eventfs_allocvp( dvp->v_mount, vpp, eg, Egroup ));
         }

         break;
      }

      case Egroup:
      {
         struct eventfs_stream *es;
         struct group_target   *gt;
         int                    i;

         if( cnp->cn_flags & ISDOTDOT )
         {
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf( "eventfs_lookup(Egroup): ..\n" );
#endif /* EVENTFS_DIAGNOSTICS */
            return (eventfs_root(dvp->v_mount, vpp));
         }

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_lookup(Egroup): %s [%ld]\n", pname, cnp->cn_namelen );
#endif /* EVENTFS_DIAGNOSTICS */

         for( es = efs->un.efs_group->efgroup_streams.lh_first;
              es;
              es = es->efstream_link.le_next )
         {
            if( ( cnp->cn_namelen == es->efstream_namelen ) &&
                ( bcmp( es->efstream_name, pname, cnp->cn_namelen ) == 0 ) )
               return (eventfs_allocvp( dvp->v_mount, vpp, es, Estream ));
         }

         for( gt = group_targets, i = 0; i < ngroup_targets; gt++, i++ )
         {
            if( ( cnp->cn_namelen == gt->gt_namelen ) &&
                ( bcmp( gt->gt_name, pname, cnp->cn_namelen ) == 0 ) )
               return (eventfs_allocvp( dvp->v_mount, vpp, efs->un.efs_group,
                                        gt->gt_efstype ));
         }

         break;
      }

      case Estream:
      {
         struct stream_target *st;
         int                   i;

         if( cnp->cn_flags & ISDOTDOT )
            return (eventfs_allocvp( dvp->v_mount, vpp,
                                     efs->un.efs_stream->efstream_group,
                                     Egroup ));

         for( st = stream_targets, i = 0; i < nstream_targets; st++, i++ )
         {
            if( ( cnp->cn_namelen == st->st_namelen ) &&
                ( bcmp( st->st_name, pname, cnp->cn_namelen ) == 0 ) )
               return (eventfs_allocvp( dvp->v_mount, vpp, efs->un.efs_stream,
                                        st->st_efstype ));
         }

         break;
      }

      default:
         return (ENOTDIR);
   }

   return (cnp->cn_nameiop == LOOKUP ? ENOENT : EROFS);
}

static int
eventfs_access(ap)
   struct vop_access_args /* {
      struct vnode *a_vp;
      int  a_mode;
      struct ucred *a_cred;
      struct proc *a_p;
   } */ *ap;
{
   struct vattr *vap;
   struct vattr  vattr;
   int            error;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_access\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   /* If super-user, you have access */
   if( ap->a_cred->cr_uid == 0 )
      return (0);

   vap = &vattr;
   error = VOP_GETATTR( ap->a_vp, vap, ap->a_cred, ap->a_p );
   if( error )
      return error;

   /*
    * Access check is based on only one of owner, group, public.
    * If not owner, then check group. If not a member of the
    * group, then check public access.
    */
   if( ap->a_cred->cr_uid != vap->va_uid )
   {
      gid_t *gp;
      int    i;

      ap->a_mode >>= 3;
      gp = ap->a_cred->cr_groups;
      for( i = 0; i < ap->a_cred->cr_ngroups; i++, gp++ )
         if( vap->va_gid == *gp )
            goto found;
      ap->a_mode >>= 3;
found:
      ;
   }

   if( ( vap->va_mode & ap->a_mode ) == ap->a_mode )
      return (0);

   return (EACCES);
}

static int
eventfs_getattr(ap)
   struct vop_getattr_args /* {
      struct vnode *a_vp;
      struct vattr *a_vap;
      struct ucred *a_cred;
      struct proc *a_p;
   } */ *ap;
{
   struct eventfs_node *efs = VTOEVENTFS(ap->a_vp);
   struct vattr        *vap = ap->a_vap;

   /* Start by zeroing out the attributes */
   VATTR_NULL(vap);

   /* Next do all the common fields */
   vap->va_type = ap->a_vp->v_type;
   vap->va_fileid = efs->efs_fileno;
   vap->va_uid = 0;
   vap->va_gid = 0;
   vap->va_flags = 0;
   vap->va_blocksize = PAGE_SIZE;
   vap->va_bytes = vap->va_size = 0;
   nanotime(&vap->va_ctime);
   vap->va_atime = vap->va_mtime = vap->va_ctime;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_getattr: vnode 0x%x, type 0x%x\n", (u_int) ap->a_vp,
              vap->va_type );
#endif /* EVENTFS_DIAGNOSTICS */

   /* Now do the object specific fields */
   switch( efs->efs_type )
   {
      case Eroot:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Eroot)\n" );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD|VEXEC) |
                        (VREAD|VEXEC) >> 3 |
                        (VREAD|VEXEC) >> 6;
         vap->va_nlink = 1;
         vap->va_size = vap->va_bytes = DEV_BSIZE;
         break;

      case Egroup:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Egroup[%s])\n",
                    efs->un.efs_group->efgroup_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD|VEXEC) |
                        (VREAD|VEXEC) >> 3 |
                        (VREAD|VEXEC) >> 6;
         vap->va_nlink = 1 + efs->un.efs_group->efgroup_nstreams;
         vap->va_size = vap->va_bytes = DEV_BSIZE;
         break;

      case Egstats:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Egstats[%s])\n",
                    efs->un.efs_group->efgroup_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD) |
                        (VREAD) >> 3 |
                        (VREAD) >> 6;
         vap->va_nlink = 1;
         break;

      case Estream:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Estream[%s])\n",
                    efs->un.efs_stream->efstream_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD|VEXEC) |
                        (VREAD|VEXEC) >> 3 |
                        (VREAD|VEXEC) >> 6;
         vap->va_nlink = 3;
         vap->va_size = vap->va_bytes = DEV_BSIZE;
         break;

      case Esdata:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Esdata[%s])\n",
                    efs->un.efs_stream->efstream_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD) |
                        (VREAD) >> 3 |
                        (VREAD) >> 6;
         vap->va_nlink = efs->un.efs_stream->efstream_refcount;
         break;

      case Esstats:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Esstats[%s])\n",
                    efs->un.efs_stream->efstream_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD) |
                        (VREAD) >> 3 |
                        (VREAD) >> 6;
         vap->va_nlink = 1;
         break;

      default:
         panic( "eventfs_getattr: invalid type on eventfs node (0x%x)",
                efs->efs_type );
   }

   return 0;
}

static int
eventfs_setattr(ap)
   struct vop_setattr_args /* {
      struct vnode *a_vp;
      struct vattr *a_vap;
      struct ucred *a_cred;
      struct proc  *a_p;
   } */ *ap;
{
   if( ap->a_vap->va_flags != VNOVAL )
      return (EOPNOTSUPP);

   return (0);
}

static int
eventfs_read(ap)
   struct vop_read_args /* {
      struct vnode *a_vp;
      struct uio *a_uio;
      int  a_ioflag;
      struct ucred *a_cred;
   } */ *ap;
{
   int                      rval   = 0;
   struct vnode            *vp     = ap->a_vp;
   struct uio              *uio    = ap->a_uio;
   struct eventfs_node     *efs    = VTOEVENTFS(vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_READ) )
      printf( "eventfs_read\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   /* Acquire lock */
   while( efs->efs_lockowner )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_READ) )
         printf( "eventfs_read: sleeping on lock\n" );
#endif /* EVENTFS_DIAGNOSTICS */
      tsleep( &efs->efs_lockowner, PRIBIO, "efslck", 0 );
   }
   efs->efs_lockowner = curproc->p_pid;

   /* Check for stream teardown */
   if( ( rval = eventfs_check_stream( efs ) ) == 0 )
   {
      switch( efs->efs_type )
      {
         case Esdata:
         {
            pd_bool_t    eof    = FALSE;
            unsigned int nread  = 0;

refetch:
            rval = eventfs_dodata( efs, uio, &nread, &eof );

            /* If blocking allowed, sleep pending data */
            if( eof && ( ( ap->a_ioflag & IO_NDELAY ) == 0 ) && !nread )
            {
               /* Release lock */
               efs->efs_lockowner = 0;
               wakeup( &efs->efs_lockowner );

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_READ) )
                  printf( "eventfs_read: blocking (0x%x)\n",
                          (u_int) nread );
#endif /* EVENTFS_DIAGNOSTICS */
               /* Sleep on the vnode */
               if( ( rval = tsleep( (caddr_t)vp, PCATCH, "efsrd", 0 ) ) )
                  return (rval);
#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_READ) )
                  printf( "eventfs_read: awoke from blocking\n" );
#endif /* EVENTFS_DIAGNOSTICS */

               /* Acquire lock */
               while( efs->efs_lockowner )
               {
#ifdef EVENTFS_DIAGNOSTICS
                  if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_READ) )
                     printf( "eventfs_read: sleeping on lock\n" );
#endif /* EVENTFS_DIAGNOSTICS */
                  tsleep( &efs->efs_lockowner, PRIBIO, "efslck", 0 );
               }
               efs->efs_lockowner = curproc->p_pid;

               /* Check for stream teardown */
               if( ( rval = eventfs_check_stream( efs ) ) == 0 )
                  goto refetch;
            }
            /* If non-blocking and no data available, return EAGAIN */
            else if( eof && ( ap->a_ioflag & IO_NDELAY ) && !nread )
               rval = EAGAIN;

            break;
         }

         case Esstats:
         case Egstats:
            rval = eventfs_dostats( efs, uio );
            break;

         default:
            rval = EOPNOTSUPP;
            break;
      }
   }

   /* Release lock */
   efs->efs_lockowner = 0;
   wakeup( &efs->efs_lockowner );

   return (rval);
}

static int
eventfs_readdir(ap)
   struct vop_readdir_args /* {
      struct vnode *a_vp;
      struct uio *a_uio;
      struct ucred *a_cred;
      int *a_eofflag;
      int *a_ncookies;
      u_long **a_cookies;
   } */ *ap;
{
   struct uio          *uio   = ap->a_uio;
   struct dirent        d;
   struct dirent       *dp    = &d;
   struct eventfs_node *efs;
   int                  error = 0;
   int                  ecnt  = 0;
   int                  scnt  = 0;
   int                  i;
   int                  off;

   if( ap->a_vp->v_type != VDIR )
      return (ENOTDIR);

   off = (int)uio->uio_offset;
   if( ( off != uio->uio_offset ) || ( off < 0 ) ||
       ( ( (u_int)off % UIO_MX ) != 0 ) || ( uio->uio_resid < UIO_MX ) )
      return (EINVAL);

   i = (u_int)off / UIO_MX;

   efs = VTOEVENTFS(ap->a_vp);
   switch( efs->efs_type )
   {
      case Egroup:
      {
         struct group_target   *gt;
         struct eventfs_stream *es;

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_readdir( Egroup[%s], %d )\n",
                    efs->un.efs_group->efgroup_name, i );
#endif /* EVENTFS_DIAGNOSTICS */

         for( gt = &group_targets[i]; 
              ( uio->uio_resid >= UIO_MX ) && ( i < ngroup_targets );
              gt++, i++ )
         {
            dp->d_reclen = UIO_MX;
            dp->d_fileno = efs->efs_fileno;
            dp->d_namlen = gt->gt_namelen;
            bcopy( gt->gt_name, dp->d_name, gt->gt_namelen + 1 );
            dp->d_type = gt->gt_type;

            error = uiomove( (caddr_t) dp, UIO_MX, uio );
         }

         for( es = efs->un.efs_group->efgroup_streams.lh_first;
              es && ( uio->uio_resid >= UIO_MX );
              i++, ecnt++ )
         {
            while( ecnt < i )
            {
               ecnt++;
               if( ecnt > ngroup_targets )
                  es = es->efstream_link.le_next;
               if( es == NULL )
                  goto done;
            }

            if( es == NULL )
               goto done;

            dp->d_reclen = UIO_MX;
            dp->d_fileno = efs->efs_fileno;
            dp->d_namlen = es->efstream_namelen;
            bcopy( es->efstream_name, dp->d_name, dp->d_namlen + 1 );
            dp->d_type = DT_REG;
            es = es->efstream_link.le_next;

            error = uiomove( (caddr_t) dp, UIO_MX, uio );
         }
done:
         break;
      }

      case Eroot:
         for( ;
              ( ( i < 2 ) || ( scnt < EVENTFS_NGROUPS ) ) &&
                 ( uio->uio_resid >= UIO_MX );
              i++, ecnt++ )
         {

#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf( "eventfs_readdir( Eroot, %d )\n", i );
#endif /* EVENTFS_DIAGNOSTICS */
         
            switch( i )
            {
               case 0:      /* `.' */
               case 1:      /* `..' */
                  dp->d_reclen = UIO_MX;
                  dp->d_fileno = efs->efs_fileno;
                  dp->d_namlen = i + 1;
                  bcopy( "..", dp->d_name, dp->d_namlen );
                  dp->d_name[i + 1] = '\0';
                  dp->d_type = DT_DIR;
                  break;

               default:
                  while( ecnt < i )
                  {
                     if( ecnt >= 2 )
                        scnt++;
                     while( eventfs_groups[scnt] == NULL )
                        if( ++scnt >= EVENTFS_NGROUPS )
                           goto done;
                     ecnt++;
                  }

                  if( eventfs_groups[scnt] == NULL )
                     goto done;

                  dp->d_reclen = UIO_MX;
                  dp->d_fileno = efs->efs_fileno;
                  dp->d_namlen = eventfs_groups[scnt]->efgroup_namelen;
                  bcopy( eventfs_groups[scnt]->efgroup_name, dp->d_name,
                         dp->d_namlen + 1 );
                  dp->d_type = DT_DIR;
                  scnt++;
                  break;
            }

            error = uiomove( (caddr_t) dp, UIO_MX, uio );
         }
         break;

      case Estream:
      {
         struct stream_target *st;

         for( st = &stream_targets[i]; 
              ( uio->uio_resid >= UIO_MX ) && ( i < nstream_targets );
              st++, i++ )
         {
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf( "eventfs_readdir( Estream[%s], %d )\n",
                       efs->un.efs_stream->efstream_name, i );
#endif /* EVENTFS_DIAGNOSTICS */
            dp->d_reclen = UIO_MX;
            dp->d_fileno = efs->efs_fileno;
            dp->d_namlen = st->st_namelen;
            bcopy( st->st_name, dp->d_name, st->st_namelen + 1 );
            dp->d_type = st->st_type;

            error = uiomove( (caddr_t) dp, UIO_MX, uio );
         }
         break;
      }

      default:
         return (ENOTDIR);
   }

   uio->uio_offset = i * UIO_MX;

   return error;
}


static int
eventfs_ioctl(ap)
   struct vop_ioctl_args /* {
      struct vnoe *a_vp;
      struct proc *a_p;
      u_long       a_command;
      caddr_t      a_data;
   } */ *ap;
{
   int error = ENOTTY;

#ifdef EVENTFS_DIAGNOSTICS
    if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
       printf( "eventfs_ioctl( %lu )\n", ap->a_command );
#endif /* EVENTFS_DIAGNOSTICS */

   switch( ap->a_command )
   {
      case EFSIOCREQ:
      {
         struct eventfs_node *efs = VTOEVENTFS(ap->a_vp);

         if( efs->efs_type == Egstats )
         {
            struct eventfs_request *req = (struct eventfs_request *)ap->a_data;
            char *name = NULL;

            error = eventfs_construct_stream( efs->un.efs_group->efgroup_name,
                                              req->mask, &name );
            if( !error )
            {
               int len = strlen( name );
#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
                  printf( "eventfs_ioctl: name '%s'\n", name );
#endif /* EVENTFS_DIAGNOSTICS */
               bcopy( name, req->name, len );
               req->name[len] = '\0';
               EFS_FREE( name );
            }
         }
         break;
      }

      case FIONBIO:
         error = 0;
         break;

      case FIOASYNC:
      {
         struct eventfs_opennode *efsopn;
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf("eventfs_icotl: FIOASYNC\n");
#endif /* EVENTFS_DIAGNOSTICS */
         efsopn = eventfs_lookup_opennode( ap->a_p, VTOEVENTFS(ap->a_vp) );
         if( efsopn )
         {
            if( *(int *)ap->a_data )
               efsopn->efsopen_state |= EFS_NODE_ASYNC;
            else
               efsopn->efsopen_state &= ~EFS_NODE_ASYNC;
            error = 0;
         }
         else
            error = EBADF;
         break;
      }

      case FIONREAD:
      {
         struct eventfs_opennode *efsopn;
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf("eventfs_icotl: FIONREAD\n");
#endif /* EVENTFS_DIAGNOSTICS */
         efsopn = eventfs_lookup_opennode( ap->a_p, VTOEVENTFS(ap->a_vp) );
         if( efsopn )
         {
            struct eventfs_stream       *es;
            struct eventfs_stream_entry *efstrent;
            int                          cnt;

            es = efsopn->efsopen_node->un.efs_stream;
            cnt = 0;

            PD_MUTEX_ENTER(EFSLINKAGE, es->efstream);
            for( efstrent = es->efstream_entries.lh_first;
                 efstrent != es->efstream_tail;
                 efstrent = efstrent->efstrent_link.le_next )
               cnt += efstrent->efstrent_entry->efsent_event.efsevent_size +
                      sizeof(struct eventfs_event);
            PD_MUTEX_EXIT(EFSLINKAGE, es->efstream);
            *(int *)ap->a_data = cnt;
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf("eventfs_icotl: FIONREAD: %d bytes\n", cnt);
#endif /* EVENTFS_DIAGNOSTICS */
            error = 0;
         }
         else
            error = EBADF;
         break;
      }

      case FIOSETOWN:
      {
         struct eventfs_opennode *efsopn;
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf("eventfs_icotl: FIOSETOWN\n");
#endif /* EVENTFS_DIAGNOSTICS */
         efsopn = eventfs_lookup_opennode( ap->a_p, VTOEVENTFS(ap->a_vp) );
         if( efsopn )
            error = fsetown(*(int *)ap->a_data, &efsopn->efsopen_sigio );
         else
            error = EBADF;
         break;
      }

      case FIOGETOWN:
      {
         struct eventfs_opennode *efsopn;
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf("eventfs_icotl: FIOGETOWN\n");
#endif /* EVENTFS_DIAGNOSTICS */
         efsopn = eventfs_lookup_opennode( ap->a_p, VTOEVENTFS(ap->a_vp) );
         if( efsopn )
         {
            *(int *)ap->a_data = fgetown( efsopn->efsopen_sigio );
            error = 0;
         }
         else
            error = EBADF;
         break;
      }

      default:
   }

   return (error);
}


static int
eventfs_inactive(ap)
   struct vop_inactive_args /* {
      struct vnode *a_vp;
      struct proc *a_p;
   } */ *ap;
{
   struct vnode *vp = ap->a_vp;

   VOP_UNLOCK( vp, 0, ap->a_p );

   return (0);
}

static int
eventfs_reclaim(ap)
   struct vop_reclaim_args /* {
      struct vnode *a_vp;
   } */ *ap;
{
   struct vnode         *vp    = ap->a_vp;
   struct eventfs_node  *efs   = VTOEVENTFS(vp);
   struct eventfs_node **efspp;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_reclaim\n" );
#endif /* EVENTFS_DIAGNOSTICS */
   for( efspp = &efshead; *efspp; efspp = &(*efspp)->efs_next )
   {
      if( *efspp == efs )
      {
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
            printf( "eventfs_reclaim: reclaiming vnode\n" );
#endif /* EVENTFS_DIAGNOSTICS */
         if( efs->efs_type == Estream )
         {
            struct eventfs_stream *es = efs->un.efs_stream;
            es->efstream_data = NULL;
         }

         *efspp = efs->efs_next;
         break;
      }
   }

   if( vp->v_data )
   {
      FREE( vp->v_data, M_TEMP );
      vp->v_data = NULL;
   }

   return (0);
}


/*
 * Poll open stream for available data
 */
static int
eventfs_poll(ap)
   struct vop_poll_args /* {
      struct vnode *a_vp;
      int           a_events;
      struct ucred *a_cred;
      struct proc  *a_p;
   } */ *ap;
{
   struct vnode        *vp  = ap->a_vp;
   struct eventfs_node *efs = VTOEVENTFS(vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_poll\n" );
#endif /* EVENTFS_DIAGNOSTICS */
   if( eventfs_check_stream( efs ) )
      return (POLLHUP); /* EOF */

   return eventfs_dopoll( efs, vp, ap->a_p, ap->a_events );
}


/*
 * Print out the contents of a eventfs vnode.
 */
static int
eventfs_print(ap)
   struct vop_print_args /* {
      struct vnode *a_vp;
   } */ *ap;
{
   struct eventfs_node *efs = VTOEVENTFS(ap->a_vp);

   printf("tag VT_EVENTFS, ");
   switch( efs->efs_type )
   {
      case Estream:
         printf( "stream %s\n", efs->un.efs_stream->efstream_name );
         break;

      case Egroup:
         printf( "group %s\n", efs->un.efs_group->efgroup_name );
         break;

      case Eroot:
         printf( "root\n" );
         break;

      default:
         panic( "eventfs_print: unknown type encountered" );
   }

   return (0);
}

/*
 * EventFS "should never get here" operation
 */
static int
eventfs_badop()
{
   return (EIO);
}


vop_t **eventfs_vnodeop_p;
static struct vnodeopv_entry_desc eventfs_vnodeop_entries[] =
{
   { &vop_default_desc,  (vop_t *) vop_defaultop },
#if PD_FREEBSD < 4
   { &vop_abortop_desc,  (vop_t *) eventfs_badop },
#endif /* PD_FREEBSD < 4 */
   { &vop_access_desc,   (vop_t *) eventfs_access },
   { &vop_advlock_desc,  (vop_t *) eventfs_badop },
   { &vop_bmap_desc,     (vop_t *) eventfs_badop },
   { &vop_close_desc,    (vop_t *) eventfs_close },
   { &vop_create_desc,   (vop_t *) eventfs_badop },
   { &vop_getattr_desc,  (vop_t *) eventfs_getattr },
   { &vop_inactive_desc, (vop_t *) eventfs_inactive },
   { &vop_link_desc,     (vop_t *) eventfs_badop },
   { &vop_lookup_desc,   (vop_t *) eventfs_lookup },
   { &vop_mkdir_desc,    (vop_t *) eventfs_badop },
   { &vop_mknod_desc,    (vop_t *) eventfs_badop },
   { &vop_open_desc,     (vop_t *) eventfs_open },
   { &vop_pathconf_desc, (vop_t *) vop_stdpathconf },
   { &vop_print_desc,    (vop_t *) eventfs_print },
   { &vop_read_desc,     (vop_t *) eventfs_read },
   { &vop_readdir_desc,  (vop_t *) eventfs_readdir },
   { &vop_readlink_desc, (vop_t *) eventfs_badop },
   { &vop_reclaim_desc,  (vop_t *) eventfs_reclaim },
   { &vop_remove_desc,   (vop_t *) eventfs_badop },
   { &vop_rmdir_desc,    (vop_t *) eventfs_badop },
   { &vop_setattr_desc,  (vop_t *) eventfs_setattr },
   { &vop_symlink_desc,  (vop_t *) eventfs_badop },
   { &vop_write_desc,    (vop_t *) eventfs_badop },
   { &vop_ioctl_desc,    (vop_t *) eventfs_ioctl },
   { &vop_poll_desc,     (vop_t *) eventfs_poll },
   { NULL, NULL }
};

static struct vnodeopv_desc eventfs_vnodeop_opv_desc =
   { &eventfs_vnodeop_p, eventfs_vnodeop_entries };

VNODEOP_SET(eventfs_vnodeop_opv_desc);

