/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: efs_test.c,v 1.15 2001/06/22 00:15:08 stevek Exp $
 */

#include <pdpi/pi_defs.h>

#include <sys/module.h>
#include <sys/linker.h>
#include <sys/sysproto.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/unistd.h>
#include <sys/proc.h>
#include <sys/sysctl.h>
#if ( ( __FreeBSD_version >= 420000 ) && ( __FreeBSD_version < 500000 ) ) || \
    ( __FreeBSD_version >= 500000 )
#include <sys/kthread.h>
#endif

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/vm_extern.h>
#include <vm/pmap.h>
#include <vm/vm_map.h>
#include <sys/shm.h>

#include <eventfs.h>

static int efs_test_mod_event __P(( module_t, int, void * ));
static int efs_test_newproc   __P(( void ));

static moduledata_t efs_test_mod =
{
   "EventFS test module",
   efs_test_mod_event,
   NULL,
};
DECLARE_MODULE( efs_test, efs_test_mod, SI_SUB_KTHREAD_IDLE,
                SI_ORDER_MIDDLE );

#if __FreeBSD_version >= 500000
MODULE_VERSION(efs_test, 1);
MODULE_DEPEND(efs_test, eventfs, 1, 1, 1);
#endif


static int efs_test_run = TRUE;
static int efs_test_stopped = FALSE;

int efs_test_wakeup;


static int
efs_test_mod_event( module_t mod, int type, void *data )
{
   int error = 0;

   switch( type )
   {
      case MOD_UNLOAD:
         efs_test_run = 0;
#ifdef EFS_TEST_DIAGNOSTICS
         printf( "Unloading...\n" );
#endif /* EFS_TEST_DIAGNOSTICS */
         while( ( error == 0 ) && ( efs_test_stopped == FALSE ) )
         {
            wakeup( &efs_test_wakeup );
            if( efs_test_stopped == FALSE )
               tsleep( &efs_test_stopped, PUSER, "unload", 0 );
         }
         printf( "Test stopped\n" );
         if( ( error = eventfs_destroy_group( "efstest" ) ) )
            panic( "eventfs_destroy_group failed with error code %d", error );
         break;

      case MOD_LOAD:
      default:
         break;
   }
   return (error);
}


static int efs_test_interval = 30;

#if (__FreeBSD_version < 420000) || \
    ( (__FreeBSD_version >= 500000) && (__FreeBSD_version < 500011) )
static int sysctl_vfs_eventfs_testinterval SYSCTL_HANDLER_ARGS;
#else
static int sysctl_vfs_eventfs_testinterval(SYSCTL_HANDLER_ARGS);
#endif

SYSCTL_NODE(_vfs, OID_AUTO, eventfs, CTLFLAG_RW, 0, "EventFS");
/* sysctl: vfs.eventfs.testinterval */
SYSCTL_PROC(_vfs_eventfs, OID_AUTO, testinterval, CTLTYPE_INT|CTLFLAG_RW,
            &efs_test_interval, 0, sysctl_vfs_eventfs_testinterval, "I", "");


static void efs_test __P(( void ));
static struct proc *efstproc;
static struct kproc_desc efst_kp =
{
   "efstest",
   efs_test,
   &efstproc
};

#if ( __FreeBSD_version < 420000 )
SYSINIT_KT(efs_test, SI_SUB_KTHREAD_IDLE, SI_ORDER_FIRST, kproc_start,
           &efst_kp)
#else
SYSINIT(efs_test, SI_SUB_KTHREAD_IDLE, SI_ORDER_FIRST, kproc_start,
        &efst_kp)
#endif

static void
efs_test()
{
   int    error;
   u_long ticktype;
   u_long tocktype;
   u_long clocktype;

   /* Register event stream group (efstest) */
#ifdef EFS_TEST_DIAGNOSTICS
   printf( "Registering group.\n" );
#endif /* EFS_TEST_DIAGNOSTICS */

   if( ( error = eventfs_register_group( "efstest" ) ) )
      panic( "eventfs_register_group failed with error code %d", error );

   /* Register event type with group (tick) */
#ifdef EFS_TEST_DIAGNOSTICS
   printf( "Registering event types.\n" );
#endif /* EFS_TEST_DIAGNOSTICS */

   if( ( error = eventfs_register_event_type( "efstest", "tick", &ticktype ) ) )
      panic( "eventfs_register_event_type failed with error code %d", error );

   if( ( error = eventfs_register_event_type( "efstest", "tock", &tocktype ) ) )
      panic( "eventfs_register_event_type failed with error code %d", error );

   if( ( error = eventfs_register_event_type( "efstest", "clock", &tocktype ) ) )
      panic( "eventfs_register_event_type failed with error code %d", error );

   /* Register event stream (raw) */
#ifdef EFS_TEST_DIAGNOSTICS
   printf( "Registering stream.\n" );
#endif /* EFS_TEST_DIAGNOSTICS */

   if( ( error = eventfs_register_stream( "efstest", "raw", ticktype, FALSE ) ) )
      panic( "eventfs_register_stream failed with error code %d", error );

#ifdef EFS_TEST_DIAGNOSTICS
   printf( "Reving up for thread work..\n" );
#endif /* EFS_TEST_DIAGNOSTICS */

   while(efs_test_run)
   {
      tsleep( &efs_test_wakeup, PUSER, "efstest", hz * efs_test_interval );
      efs_test_wakeup = 0;

      /* Post event to stream */
      eventfs_post_event_by_types( "efstest", ticktype, 6, "Tick\n" );
#ifdef EFS_TEST_DIAGNOSTICS
      printf( "Tick\n" );
#endif /* EFS_TEST_DIAGNOSTICS */
   }

#ifdef EFS_TEST_DIAGNOSTICS
   printf( "Stopping test\n" );
#endif /* EFS_TEST_DIAGNOSTICS */
   efs_test_stopped = TRUE;
   wakeup( &efs_test_stopped );

   exit1( efstproc, 0 );
}

#if ( __FreeBSD_version < 420000 ) || \
    ( ( __FreeBSD_version >= 500000 ) && ( __FreeBSD_version < 500011 ) )
static int
sysctl_vfs_eventfs_testinterval SYSCTL_HANDLER_ARGS
#else
static int
sysctl_vfs_eventfs_testinterval(SYSCTL_HANDLER_ARGS)
#endif
{
   int error = sysctl_handle_int(oidp, oidp->oid_arg1, oidp->oid_arg2, req);
   if( !error )
      wakeup( &efs_test_wakeup );
   return error;
}

