/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: eventfs.c,v 1.67 2001/02/23 19:02:19 stevek Exp $
 *
 * Author: Douglas Kilpatrick
 *
 * Event FileSystem interface library
 *
 */

#define LINKAGE IMPORT

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <ctype.h>
#include <fcntl.h>

#include <eventfs.h>
#include EFS_SYSTEM_HDR

#undef LINKAGE
#define LINKAGE EXPORT
#include "libeventfs.h"

#ifdef PD_UNIX
#  include <unistd.h>
#  include <fcntl.h>
#  define LIB_ERROR errno
#  define Read(f,b,l,c) read(f,b,l)
#  define Open(p,f,pfdo) (((*(pfdo)) = open(p,f)),(*(pfdo))>=0)
#  define Close(f) close(f)
#endif
#ifdef PD_WINNT
#  include <io.h>
#  include <fcntl.h>
#  define LIB_ERROR GetLastError()
int Read(pd_fd_obj fd, void *buf, size_t length, pd_bool_t convert)
{
   int        read_length;
   OVERLAPPED os;
   BOOL       bRet;
   DWORD      dwWait;
#ifdef UNICODE
#  define readBuf buf
#else
   PVOID      readBuf;
#endif

   memset( &os, 0, sizeof(os) );
   os.hEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
   ResetEvent( os.hEvent );

#ifndef UNICODE
   readBuf = (PVOID) malloc( length * sizeof(WCHAR) );
   memset( readBuf, 0, length * sizeof(WCHAR) );
#else
   memset( readBuf, 0, length );
#endif

heartbeat:
   bRet = ReadFile( fd, readBuf, length, NULL, &os );
   if( !bRet )
   {
      DWORD error = GetLastError();
      switch( error )
      {
         case ERROR_IO_PENDING:
            dwWait = WaitForSingleObject( os.hEvent, INFINITE );
            switch( dwWait )
            {
               case WAIT_OBJECT_0:
                  bRet = TRUE;
                  break;

               default:
                  break;
            }
            break;

         case ERROR_PIPE_NOT_CONNECTED:
            read_length = 0;
            break;

         default:
            read_length = -1;
            break;
      }
   }

   if( bRet )
   {
      GetOverlappedResult( fd, &os, &read_length, TRUE );
      if( read_length == 0 )
         goto heartbeat;
   }

   CloseHandle( os.hEvent );

#ifndef UNICODE
   if( read_length >= 0 )
   {
      if( convert )
         read_length = WideCharToMultiByte( CP_ACP, 0, (LPCWSTR) readBuf,
                                            read_length, (LPSTR) buf, length,
                                            NULL, NULL );
      else
         memcpy( buf, readBuf, read_length );
   }
   free( readBuf );
#endif

   return (read_length);
}

int Write(pd_fd_obj fd, void *buf, size_t length)
{
   int         write_length;
   OVERLAPPED os;
   BOOL       bRet;
   DWORD      dwWait;
#ifdef UNICODE
#  define writeBuf buf
#else
   PVOID      writeBuf;
#endif

   memset( &os, 0, sizeof(os) );
   os.hEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
   ResetEvent( os.hEvent );

#ifndef UNICODE
   writeBuf = (PVOID) malloc( length * sizeof(WCHAR) );
   memset( writeBuf, 0, length * sizeof(WCHAR) );

   MultiByteToWideChar( CP_ACP, 0, (LPCSTR) buf, length, (LPWSTR) writeBuf,
                        length );
#endif

   bRet = WriteFile( fd, buf, length, &write_length, &os );
   if( !bRet )
   {
      DWORD error = GetLastError();
      switch( error )
      {
         case ERROR_IO_PENDING:
            dwWait = WaitForSingleObject( os.hEvent, INFINITE );
            switch( dwWait )
            {
               case WAIT_OBJECT_0:
                  bRet = TRUE;
                  break;

               default:
                  break;
            }
            GetOverlappedResult( fd, &os, &write_length, TRUE );
            break;

         case ERROR_PIPE_NOT_CONNECTED:
            write_length = 0;
            break;

         default:
            write_length = -1;
            break;
      }
   }
   else
      GetOverlappedResult( fd, &os, &write_length, TRUE );

   CloseHandle( os.hEvent );

#ifndef UNICODE
   free( writeBuf );
#endif


   return (write_length);
}

BOOL Open(LPTSTR path, int flags, pd_fd_obj *fd)
{
   pd_fd_obj newfd;

   if( fd == NULL )
      return FALSE;

   newfd = CreateFile( path, flags, 0, NULL, OPEN_EXISTING,
                       FILE_FLAG_OVERLAPPED, NULL );

   if( newfd == INVALID_HANDLE_VALUE )
   {
      if( GetLastError() != ERROR_PIPE_BUSY )
         return FALSE;
      if( !WaitNamedPipe( path, 20000 ) )
         return FALSE;
   }

   *fd = newfd;
   return TRUE;
}
#define Close(f) CloseHandle(f)
#endif

#if 0
#ifdef PD_WINNT
#include "getopt.h"
#endif

/*
 * Internal globals
 */
static char *stream_file = "";
#endif


/*
 * Private functions
 */

/*
 * Frees a typelist
 */
static void typelist_release(eventfs_event_group_typelist *typelist)
{
    int i;

    i = typelist->count;
    for (i = typelist->count; i > 0; --i) {
        if (typelist->list[i-1].name != NULL)
            free(typelist->list[i-1].name);
    }
    free(typelist);
}

/*
 * Constructs a eventfs_event_group_typelist for the group named
 *
 * On error returns NULL
 */
static eventfs_event_group_typelist *
parseStatsFile(const LPTSTR gr_name, const LPTSTR str_name)
{
    TCHAR                       path[MAXPATHLEN];
    TCHAR                       buf[1024];
    int                         count, linecount;
    pd_fd_obj                   fd;
    int                         rval;
    int                         open_flags;

    eventfs_event_group_typelist   *typelist = NULL;
    /* parser building variables */
    TCHAR                       namebuf[80-8+1]; /* limit to one line */
    int                         tempmask;
    int                         state, i;

    /* Open the stats file for the group */
    if (str_name != NULL) {
        snprintf(path, sizeof(path)/sizeof(path[0]), "%s%c%s%c%s%c%s",
                _PATH_EVENTFS, FILESEP, gr_name, FILESEP, str_name,
                FILESEP, "stats");
    } else {
        snprintf(path, sizeof(path)/sizeof(path[0]), "%s%c%s%c%s",
                _PATH_EVENTFS, FILESEP, gr_name, FILESEP, "stats");
    }
    path[MAXPATHLEN-1] = '\0';

#ifdef PD_WINNT
    open_flags = GENERIC_READ;
#else
    open_flags = O_RDONLY;
#endif
    if (!Open(path, open_flags, &fd))
        return NULL;
    
    /*
     * Pass 1, count lines.
     */
    linecount = 0;
    do {
        rval = Read(fd, buf, sizeof(buf), TRUE);
        if (rval > 0) {
            LPTSTR iter = strchr(buf, '\n');
            while (iter) {
                ++linecount;
                iter = strchr(iter+1, '\n');
            }
        }
    } while (rval > 0);
    /* The stream stats file has an extra line at the top
     * for "[streamname]\n"
     */
    --linecount;

    /* Count has the number of lines.  In a group file, each line
     * is a type.
     *
     * Statys file is in form of
     * <number>\t<typename>\n
     *
     * Rewind the file.  Close()&Open() to simplify requirements on eventfs
     */
    Close(fd);
    if( linecount <= 0 )
            return NULL;

    if (!Open(path, open_flags, &fd))
        return NULL;

    /* Allocate space for the typelist */
    typelist = malloc(sizeof(*typelist)+sizeof(typelist->list[0])*linecount);
    if (typelist == NULL)
        return NULL;
    /* Null initialize the string points
     * so we can safely nuke them in release
     */
    memset(typelist, 0, sizeof(*typelist)+sizeof(typelist->list[0])*linecount);
    typelist->count = linecount;

    /* Regex to parse file line*/
    /* The lines look like:
     * "([0-9][0-9]*)\t([a-zA-Z][a-zA-Z0-9 ]*)"
     * I was using regex, but its not on NT so I did a small handrolled
     * parser instead
     */

    linecount   = 0;
    count       = 0;
    tempmask    = 0;
    i           = 0;
    state       = 3; /* Start in header-eater state. */

    while  ( (rval = Read(fd, buf, sizeof(buf), TRUE)) > 0 ) {
        int c;
        count = 0;

        while (count < rval) {
            c = buf[count++];
            if( c == '\0' ) break;

            switch(state) {
            case 0:
                /* numeater state. */
                if (!isdigit(c)) {
                    state = 1;
                    typelist->list[linecount].mask = tempmask;
                    tempmask = 0;
                } else {
                    /* Assuming ascii here. */
                    tempmask = tempmask * 10 + (c - '0');
                }
                break;
            case 1:
                /* space eater state. */
                if (isspace(c)) {
                    break;
                } else {
                    state = 2;
                    // Deal with the character now.
                }
            case 2:
                /* name state */
                if ((isalnum(c) || isspace(c)) && c != '\n') {
                    if (i < sizeof(namebuf))
                        namebuf[i++] = c;
                } else {
                    namebuf[i] = '\0';
                    typelist->list[linecount].name = strdup(namebuf);
                    i = 0;
                    linecount++;
                    /* Restart */
                    state = 0;
                }
                break;
            case 3:
                /* header state */
                if (c == '\n') {
                    state = 0;
                }
            }
        }
    }

    if (rval < 0) { /* Error reading. */
        typelist_release(typelist);
        typelist = NULL;
    }

    Close(fd);

    return typelist;
}

/*
 * Privately exported functions
 */
void EXPORT
_eventfsInit( void )
{
#if 0
    char        ch;

    while ( (ch = getopt(*pargc, *pargv, "f:" ) ) != -1 )
    {
        switch(ch)
        {
        case 'f':
            free(stream_file);
            stream_file = strdup(optarg);
            /* Move all of argv down one,
             * decrement argc.
             */
            break;
        default:
            break;
        }
    }
#endif
}

void EXPORT
_eventfsShutdown()
{
    /*
     * Hope our shutdown can all be done by
     * the OS, cause we ain't keeping around the
     * data to do it here.
     */
}



/*
 * Public functions
 */

/* eventfsEvent functions */
eventfs_event_group_t EXPORT
eventfsEventGetGroup( const LPTSTR name )
{
    eventfs_event_group_t           group;
    eventfs_event_group_typelist   *typelist;
    unsigned int                i;
    u_long                      mask;

    typelist = parseStatsFile(name, NULL);
    /* If the typelist is null, assume no such group */

    if (typelist == NULL)
        return NULL;

    group = malloc(sizeof(*group));

    if (group == NULL) {
        typelist_release(typelist);
        return NULL;
    }

    mask = 0;
    /* Cheat.  Assume all types are consecutive. */
    for (i = 0; i < typelist->count; i++) {
        mask |= 1<<i;
    }
    group->mask       = mask;
    group->name       = strdup(name);
    group->list       = typelist;

    return group;
}

/* eventfsEvent functions */
eventfs_event_stream_t EXPORT
eventfsEventGetStreamByFile( const LPTSTR path )
{
    eventfs_event_group_t           group;
    eventfs_event_stream_t          stream;
    eventfs_event_group_typelist   *typelist;
    int                             open_flags;
    pd_fd_obj                       fd;

    /* This is a big "fakeit" operation here.
     *
     * Create a typelist.  Meaningless.
     * Create a group.     Meaningless.
     * Create a stream.    FD has meaning, nothing else does.
     */
    typelist = malloc(sizeof(*typelist)+sizeof(typelist->list[0])*1);
    if (typelist == NULL)
        goto exit;

    typelist->count = 1;
    typelist->list[1].mask = 1;
    if ((typelist->list[1].name = strdup("fake")) == NULL)
        goto exit1;

    if ((group = malloc(sizeof(*group))) == NULL)
        goto exit1;
    group->mask       = 1;
    group->list       = typelist;
    if ((group->name = strdup("fake")) == NULL)
        goto exit2;

#ifdef PD_WINNT
    open_flags = GENERIC_READ | GENERIC_WRITE;
#else
    open_flags = O_RDONLY;
#endif
    if (!Open(path, open_flags, &fd))
        goto exit3;

    if ((stream = malloc(sizeof(*stream))) == NULL)
        goto exit4;
    stream->group       = group;
    stream->stream      = fd;
    stream->mask        = 1;

    return stream;

exit4: Close(fd);
exit3: free(group->name);
exit2: free(group);
exit1: typelist_release(typelist);
exit:  return NULL;
}



/* eventfsEventGroup functions */

eventfs_event_stream EXPORT *
eventfsEventGroupGetStream( eventfs_event_group_t group, u_long mask )
{
    /*
     * IOCTL on the stats file for the group.
     *
     * descriminator == EFSIOCREQ
     * arg1 == struct {u_long mask, char name[MAXPATHLEN]};
     * mask is an in parameter.  Mask of returned stream.
     * name is an out parameter.  Name of new? stream.
     */

    /*
     * Open stats file of group.
     * ioctl.
     * open data file of stream.
     * construct stream object.
     * return stream object.
     */
    eventfs_event_stream_t      stream = NULL;
    pd_fd_obj                   fd;
    int                         rval;
    TCHAR                       path[MAXPATHLEN];
    struct eventfs_request      req;
    int                         open_flags;

    mask &= group->mask;

    snprintf(path, sizeof(path)/sizeof(path[0]), "%s%c%s%c%s",
            _PATH_EVENTFS, FILESEP, group->name, FILESEP,
#ifdef PD_WINNT
            "ioctl"
#else
            "stats"
#endif
            );
#ifdef PD_WINNT
    open_flags = GENERIC_READ | GENERIC_WRITE;
#else
    open_flags = O_RDONLY;
#endif
    path[sizeof(path)-1] = '\0';

    if (!Open(path, open_flags, &fd))
        return NULL;
 
    req.mask = mask;
    
#ifdef PD_WINNT
    rval = Write(fd, &(req.mask), sizeof(req.mask));
    if (rval < 0) {
        Close(fd);
        return NULL;
    }
    rval = Read(fd, req.name, sizeof(req.name), TRUE);
    if (rval < 0) {
        Close(fd);
        return NULL;
    }
#else
    rval = ioctl(fd, EFSIOCREQ, &req);
    if (rval < 0) {
        Close(fd);
        return NULL;
    }
#endif
    Close(fd); /* Close the group */

    snprintf(path, sizeof(path)/sizeof(path[0]), "%s%c%s%c%s%c%s",
            _PATH_EVENTFS, FILESEP, group->name,
            FILESEP, req.name, FILESEP, "data");
    path[sizeof(path)-1] = '\0';

#ifdef PD_WINNT
    open_flags = GENERIC_READ;
#endif
    if (!Open(path, open_flags, &fd))
        return NULL;

    stream = malloc(sizeof(*stream));
    if (stream == NULL) {
        Close(fd);
        return NULL;
    }

    stream->group     = group;
    stream->stream    = fd;
    stream->mask      = mask;
    return stream;
}

eventfs_event_stream EXPORT *
eventfsEventGroupGetStreamByName( eventfs_event_group_t group, LPTSTR stream_name)
{
    /*
     * open data file of stream.
     * construct stream object.
     * return stream object.
     */
    eventfs_event_stream_t      stream = NULL;
    pd_fd_obj                   fd;
    TCHAR                       path[MAXPATHLEN];
    eventfs_event_group_typelist   *typelist;
    unsigned long               mask;
    unsigned int                i;
    int                         open_flags;

    snprintf(path, sizeof(path)/sizeof(path[0]), "%s%c%s%c%s%c%s",
            _PATH_EVENTFS, FILESEP, group->name, FILESEP, stream_name,
            FILESEP, "data");

    path[sizeof(path)-1] = '\0';

#ifdef PD_WINNT
    open_flags = GENERIC_READ;
#else
    open_flags = O_RDONLY;
#endif
    if (!Open(path, open_flags, &fd)) /* Open the stream data file */
        return NULL;

    typelist = parseStatsFile( group->name, stream_name);

    if (typelist == NULL) {
        Close(fd);
        return NULL;
    }

    for (mask = 0, i = 0; i < typelist->count; i++) {
        mask |= typelist->list[i].mask;
    }

    stream = malloc(sizeof(*stream));
    if (stream == NULL) {
        Close(fd);
        return NULL;
    }

    stream->group     = group;
    stream->stream    = fd;
    stream->mask      = mask;

    typelist_release(typelist);
    return stream;
}


const LPTSTR EXPORT
eventfsEventGroupGetTypeName( eventfs_event_group_t group, u_long mask )
{
    int         i;

    /* Stupid check */
    if (group == NULL || !(mask & group->mask) || group->list == NULL)
        return NULL;

    for (i = group->list->count; i > 0 ; --i) {
        if (mask & group->list->list[i-1].mask)
            break;
    }
    if (i > 0) 
        return group->list->list[i-1].name;
    else
        return NULL;
        /* Abort()?? We should never get here. */
}

u_long EXPORT
eventfsEventGroupGetType( eventfs_event_group_t group, LPTSTR type_name)
{
    int         i;

    /* Stupid check */
    if (group == NULL || group->list == NULL)
        return 0;

    for (i = group->list->count; i > 0 ; --i) {
        if (!strcmp(group->list->list[i-1].name, type_name))
            break;
    }
    if (i > 0) 
        return group->list->list[i-1].mask;
    else
        return 0;
}

#if WRV_EVENT_POST
void EXPORT
eventfsEventGroupPostEvent( eventfs_event_group_t group, eventfs_event_t event )
{
#ifndef PD_WINNT
#warning TBD: group::PostEvent(event)
#endif
    return;
}
#endif

void EXPORT
eventfsEventGroupRelease( eventfs_event_group_t group )
{
    if (!group)
        return;

    if (group->list)
        typelist_release(group->list);

    free(group);
    return;
}

#if WRV_EVENT_PROCESS
void EXPORT
eventfsEventStreamProcess( eventfs_event_stream_t stream, volatile int* quit_var)
{
#warning TBD: stream::Process(quit_var)
    return;
}
#endif

#ifdef PD_SOLARIS
#define SETSIZE(size,start)                                                  \
do {                                                                         \
    memcpy(&(size), &(((eventfs_event_t)(start))->efsevent_size),            \
            sizeof((size)));                                                 \
    (size) += sizeof(eventfs_event);                                         \
} while (0)
#else
#define SETSIZE(size,start)                                                  \
do {                                                                         \
        (size) = sizeof(eventfs_event)+                                      \
            ((eventfs_event_t)(start))->efsevent_size;                       \
} while (0)
#endif

eventfs_event_t EXPORT
eventfsEventStreamGetNextEvent( eventfs_event_stream_t stream)
{
    /* Ok.
     * 1: This function is NOT threadsafe!
     * 2: This function should be callable from different
     *    threads, just not at the same time.
     * 3: It isn't re-entrant or signal safe.
     * 
     * We keep a buffer to do reasonable sized reads from the
     * stream.  keep a 20 page buffer, always read one page.  If
     * alignmen prohibits, move unparsed section (incomplete events)
     * back to start of buffer.
     * 
     * Assumptions: No single event is bigger than 20 pages.
     * I should be able to write it to handle 21 page events, but it
     * would be a minor pain, and its userspace, so memory is cheap.
     */

    /* PAGE_SIZE not global define, so just use 1024 */
#define WRVEV_PAGE_SIZE 1024
    static unsigned char        buf[WRVEV_PAGE_SIZE * 21];
    static unsigned char       *start = 0;
    static size_t               data_size = 0;

    eventfs_event_t             retval = 0;
    ssize_t                     count;
    ssize_t                     size = 0;
    
#ifdef DEBUG_WRVEVENT
    /* statistics */
    static int                  fFirst = 0;
    static int                  splitEvents = 0;
    static int                  totalEvents = 0;
#endif

    /* Sanity checking */
    if (stream == NULL) {
        errno = EINVAL;
        return NULL;
    }
    /* I dont know how to check that a fd is valid. */

    /*
     * If we do not have a full event in the buffer
     *     if we can not read a page
     *         move the buffer contents to start
     *         reset start and end
     *     read a page
     * parse an event
     * reset end
     * return the event.
     */
    /* Special case first read for sanity */
    if (start == NULL || data_size == 0) {
        start = buf;
        do {
            data_size = Read(stream->stream, buf, WRVEV_PAGE_SIZE, FALSE);
        } while(data_size < 0 && errno == EAGAIN);
        if (data_size <= 0) {
            errno = EIO;
            return NULL;
        }
#ifdef DEBUG_WRVEVENT
        fFirst = 1;
#endif
    }
    /* Originally: While we dont have a full event.
     * Split due to SPARC issues.
     *
     * If we have a full header, get size from the header.
     * While we either dont have a header
     *     or we dont have the full event
     *   read in another page of events.
     *   if we have a full header, get size from the header.
     */
    if (data_size >= sizeof(*retval)) {
        SETSIZE(size,start);
        if (size > sizeof(buf))
            goto event_too_big;
    }
    while ( (data_size < sizeof(*retval)) || (data_size < size) ) {
        /* Do we need to move the data in the buffer? */
        if ( start != buf
        && (start + (data_size + WRVEV_PAGE_SIZE)) > (buf+sizeof(buf))) {
            /* memmove is allowed to overlap. */
            memmove(buf, start, data_size);
            start = buf;
        }
        if ((start + (data_size + WRVEV_PAGE_SIZE)) > (buf+sizeof(buf))) {
            /* If this is still true, we be hosed. */
            goto event_too_big;
        }

#ifdef DEBUG_WRVEVENT
        splitEvents++;
#endif

        /* Read in a a page */
        count = Read(stream->stream, start+data_size, WRVEV_PAGE_SIZE, FALSE);
        if (count <= 0)
            return NULL;

        if (data_size >= sizeof(*retval)) {
            SETSIZE(size,start);
            if (size > sizeof(buf))
                goto event_too_big;
        }

        data_size += count;
    }
    /* We have at least one event in the buffer.  We may have more,
     * but if so we can handle them on the next call to this function.
     */
    /* Parse the event */
    retval = malloc(size);
    if (!retval) {
        errno = ENOMEM;
        return NULL;
    }

    memcpy(retval, start, size);
    start += size;
    data_size -= size;

#ifdef DEBUG_WRVEVENT
    ++totalEvents;
#endif

    return retval;

event_too_big:
#ifdef DEBUG_WRVEVENT
    fprintf(stderr,
            "Event of unusual size.  %d. %u/%lu %d/%d\n",
            fFirst, data_size, size,
            splitEvents, totalEvents);
#endif
    errno = ERANGE;
    return NULL;
}

void EXPORT 
eventfsEventStreamRelease( eventfs_event_stream_t stream)
{
    if (stream == NULL)
        return;

    Close(stream->stream);

    free(stream);
}

/*
 * Macros:
 * eventfsEventStream::GetTypelist()
 * eventfsEventStream::GetWaitObject()
 * eventfsEventStream::GetHandler()
 */


/* eventfsEvent */
eventfs_event_t EXPORT
eventfsEventClone( eventfs_event_t event )
{
    eventfs_event_t         clone;
    if (event == NULL)
        return NULL;

    clone = malloc(sizeof(*clone) + event->efsevent_size);
    /* No pointers in an event. */
    memcpy(clone,event,sizeof(*clone) + event->efsevent_size);
    return clone;
}

void EXPORT
eventfsEventRelease( eventfs_event_t event )
{
    if (event == NULL)
        return;

    free(event);
}

/*
 * Macros
 *      eventfsEventGetData(event)
 *      eventfsEventGetSize(event)
 *      eventfsEventGetTypes(event)
 */
