/*! \file eventfs_proc.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: eventfs_proc.c,v 1.32 2001/06/11 19:15:15 stevek Exp $
 *
 * \brief   Linux implimentation of eventfs via /proc
 * \author  Douglas Kilpatrick
 * \version $Revision: 1.32 $
 * \date    $Date: 2001/06/11 19:15:15 $
 */                                                                             

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <eventfs.h>
#include <eventfs_machine.h>
#include "eventfs_linux.h"

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>

/* External variables we use */
extern struct proc_dir_entry proc_root;

static struct proc_dir_entry *eventfs_proc_dir;

struct linux_machine_data {
    struct proc_dir_entry      *streamdir;
    struct eventfs_stream      *evstream;
#if PD_LINUX < 0x02004
    struct wait_queue          *waitq;
#else
    wait_queue_head_t           waitq;
#endif
    struct fasync_struct       *sigq;
};


#ifndef MIN
#define MIN(a,b) (((b)<(a))?(b):(a))
#endif
#ifndef MAX
#define MAX(a,b) (((b)<(a))?(a):(b))
#endif

#if 0

static int linux_eventfs_data_read(
        char *page, char **start, off_t off, int count, int *eof, void *data
        );
static int linux_eventfs_stats_read(
        char *page, char **start, off_t off, int count, int *eof, void *data
        );
#error Attempt to use brain damaged interface.  Stop that!

/* 
 * Below we have a not-so-brief explaination of the default proc file
 * read interface.  This is all so everyone understands why I don't use it.
 *
 * The prototype for a read function looks like:
 * static int linux_eventfs_data_read(
 * char *page, char **start, off_t off, int count, int *eof, void *data
 * );
 *
 * This function is set up to be used in three different ways:
 *
 * 1) I don't know what offsets are, and I don't care.  My output is always
 * less than 1 page in size.
 *
 * I write my output to page, return the number of bytes I wrote.
 * I set *eof = 1.
 * I completely ignore start, off, and count.
 *
 * The function that called me will start reading my output according
 * to the offset, and increment offset accordingly (*loff += retval-loff)
 *
 * 2) I know what offsets are, but its a simple output stream so offsets will
 * always grow.
 *
 * I write my output to page, starting my output according to off.
 * I take care not to write more than one page (4k?) to page, but I
 * ignore count.  I set (start = page), and return the number of bytes
 * I wrote.  If its the end of the file, I set eof.
 *
 * The function that called me will increment offset by my return value.
 *
 * 3) I know what offsets are, and I want to manipulate them, but I don't
 *    ever want to set it to <= 0.  (and this covers?)
 *
 *  I write my output to page, starting my output according to off.
 *  I take care not to write more than one page (4k) to page, but I
 *  ignore count.  I set ( (long)start = new_offset), and return
 *  the number of bytes I wrote.  If it is the end of the file, I set
 *  eof.
 *
 *  The function that called me will set offset to start, cast to a long.
 *
 *  Now in case you wonder how the controlling function distinguishes between
 *  case 2 and case 3, its not by any flags or anything like that.  The test is
 *
 *  if (start < page) {} 
 *
 *  They use a pointer comparison!
 *
 *  And if that was not ugly enough, there is documentation that suggests you
 *  set start to a static buffer if you don't want to risk overflowing the 
 *  passed in buffer...
 *
 *  Will someone explain to me
 *  1: How linux is as stable as it is?
 *  2: Why this function takes a "count" argument?  No one uses it...
 *  3: What brain damaned lemur on crack came up with this?
 *
 *  Now with that out of the way, lets pretend we didn't see that and use our
 *  own inode operations.
 */

#else /* SANE_DRIVER_FUNCTIONS */

/* Declare driver functions */
/* Data file */
static int linux_eventfs_open(struct inode *, struct file *);
static int linux_eventfs_release(struct inode *, struct file *);
static ssize_t linux_eventfs_data_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        );
static loff_t linux_eventfs_lseek(struct file *file, loff_t offset, int orig);
static unsigned int linux_eventfs_poll(struct file *, poll_table *);
static int linux_eventfs_fasync(int fd, struct file *file, int fOn);
/* Stream stats */
static ssize_t linux_eventfs_stats_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        );
/* Group statis */
static ssize_t linux_eventfs_gr_stats_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        );
static int linux_eventfs_ioctl(struct inode *, struct file *, uint, ulong);
/* Group control file */
static int linux_eventfs_ctl_open(struct inode *, struct file *);
static int linux_eventfs_ctl_release(struct inode *, struct file *);
static ssize_t linux_eventfs_ctl_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        );
static ssize_t linux_eventfs_ctl_write(
        struct file *file, const char *u_buf, size_t length, loff_t *offset
        );

/* Declare driver fileop structures */

static struct file_operations DataFileOps = 
{
#if PD_LINUX == 0x02004
    NULL,                       /* owner                */
#endif
    linux_eventfs_lseek,        /* lseek                */
    linux_eventfs_data_read,    /* read from file       */
    NULL,                       /* write to file        */
    NULL,                       /* readdir              */
    linux_eventfs_poll,         /* poll                 */
    NULL,                       /* ioctl                */
    NULL,                       /* mmap                 */
    linux_eventfs_open,         /* open                 */
    NULL,                       /* flush                */
    linux_eventfs_release,      /* close                */
    NULL,                       /* fsync                */
    linux_eventfs_fasync        /* fasync               */
    /* etc.  Skip */
};
static struct inode_operations DataInodeOps =
{
#if PD_LINUX < 0x02004
    &DataFileOps
#endif
    /*etc.  Skip */
};

static struct file_operations StatsFileOps = 
{
#if PD_LINUX == 0x02004
    NULL,                       /* owner                */
#endif
    NULL,                       /* lseek                */
    linux_eventfs_stats_read,   /* read from file       */
    /* etc.  Skip */
};
static struct inode_operations StatsInodeOps =
{
#if PD_LINUX < 0x02004
    &StatsFileOps
#endif
    /*etc.  Skip */
};

static struct file_operations GroupStatsFileOps = 
{
#if PD_LINUX == 0x02004
    NULL,                       /* owner                */
#endif
    NULL,                       /* lseek                */
    linux_eventfs_gr_stats_read,/* read from file       */
    NULL,                       /* write to file        */
    NULL,                       /* readdir              */
    NULL,                       /* poll                 */
    linux_eventfs_ioctl         /* ioctl                */
    /* etc.  Skip */
};
static struct inode_operations GroupStatsInodeOps =
{
#if PD_LINUX < 0x02004
    &GroupStatsFileOps
#endif
    /* etc.  Skip */
};

static struct file_operations GroupCtlFileOps = 
{
#if PD_LINUX == 0x02004
    NULL,                       /* owner                */
#endif
    NULL,                       /* lseek                */
    linux_eventfs_ctl_read,     /* read from file       */
    linux_eventfs_ctl_write,    /* write to file        */
    NULL,                       /* readdir              */
    NULL,                       /* poll                 */
    NULL,                       /* ioctl                */
    NULL,                       /* mmap                 */
    linux_eventfs_ctl_open,     /* open                 */
    NULL,                       /* flush                */
    linux_eventfs_ctl_release   /* close                */
    /* etc.  Skip */
};
static struct inode_operations GroupCtlInodeOps =
{
#if PD_LINUX < 0x02004
    &GroupCtlFileOps
#endif
    /* etc.  Skip */
};

#endif

int
eventfs_machine_init()
{
#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Creating /proc/eventfs\n");
#endif
    /* dir = create_proc_entry(name, S_IFDIR, parent */
    eventfs_proc_dir = create_proc_entry("eventfs", S_IFDIR, 0);
       
    return 0;
}

int eventfs_machine_uninit()
{
    remove_proc_entry("eventfs", &proc_root);
    return 0;
}

int eventfs_init(void);

int init_module(void)
{
#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Eventfs module loading\n");
#endif
    return eventfs_init();
}

void cleanup_module(void)
{
#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Eventfs module unloading\n");
#endif
    eventfs_uninit();
}

void eventfs_machine_flush(struct eventfs_stream *es)
{
    /*
     * empty event stream.
     * to clean up offset pointers or such.
     */
#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Eventfs module flushing\n");
#endif
#warning TBD: flush: close/release anyone holding that stream
    return;
}

void eventfs_machine_destroy_group(struct eventfs_group *eg)
{
#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Removing group %s\n",eg->efgroup_name);
#endif
    /* I hope its this easy */
#warning TBD: destroy group: handle orphans
    remove_proc_entry("stats", eg->efgroup_data);
    remove_proc_entry(eg->efgroup_name, eventfs_proc_dir);
}

void eventfs_machine_destroy_stream(struct eventfs_stream *es)
{
    struct linux_machine_data *lmd;
#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Removing stream %s/%s\n",
            es->efstream_group->efgroup_name,
            es->efstream_name);
#endif
    lmd = es->efstream_data;

    /* Wake up anyone blocked. */
    wake_up_interruptible(&(lmd->waitq));

#warning Destroy_Stream: race condition on EFL_FLG_GOING
    /* Ping-pong to a different waitq? */

    /* I hope its this easy */
    remove_proc_entry("data", lmd->streamdir);
    remove_proc_entry("stats", lmd->streamdir);

    remove_proc_entry(es->efstream_name, es->efstream_group->efgroup_data);
    kfree(lmd);
    es->efstream_data = 0;
}

void eventfs_machine_register_group(struct eventfs_group *eg)
{
    struct proc_dir_entry *ent;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Adding group %s\n", eg->efgroup_name);
#endif

    ent = create_proc_entry(eg->efgroup_name, S_IFDIR, eventfs_proc_dir);
    ent->data = (void*)eg; /* data comes back to us... */
    eg->efgroup_data = ent; /* and now the circle is complete... */

    /* Stats file */
    ent = create_proc_entry("stats",0,ent);
#if PD_LINUX < 0x02004
    ent->ops = &GroupStatsInodeOps;
#else
    ent->proc_iops = &GroupStatsInodeOps;
    ent->proc_fops = &GroupStatsFileOps;
#endif
    ent->data = eg;

    /* Control file */
    ent = create_proc_entry("control",0,ent);
#if PD_LINUX < 0x02004
    ent->ops = &GroupCtlInodeOps;
#else
    ent->proc_iops = &GroupCtlInodeOps;
    ent->proc_fops = &GroupCtlFileOps;
#endif
    ent->data = eg;

    /* 
     * Pretend that's it.  I assume streams will be registered
     * seperately.
     */
}

void eventfs_machine_register_stream(struct eventfs_stream *es)
{
    struct proc_dir_entry *evstream;
    struct proc_dir_entry *evstream_file;
    struct linux_machine_data *lmd;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Adding stream %s/%s\n",
            es->efstream_group->efgroup_name, es->efstream_name);
#endif
    /* Create the stream directory */
    evstream = create_proc_entry( es->efstream_name, S_IFDIR,
            es->efstream_group->efgroup_data);
    /* Instead of the normal tight loop, I have to add an intermediary. */
    lmd            = kmalloc(sizeof(*lmd),GFP_KERNEL);
#if PD_LINUX < 0x02004
    lmd->waitq     = 0;
#else
    init_waitqueue_head(&(lmd->waitq));
#endif
    lmd->sigq      = 0;
    lmd->streamdir = evstream;
    lmd->evstream  = es;

    evstream->data = lmd;
    es->efstream_data = lmd;

    /* And the two actual files */
    /* Data file */
    evstream_file = create_proc_entry("data",
            S_IFIFO|S_IRUSR|S_IRGRP|S_IROTH,evstream);
#if PD_LINUX < 0x02004
    evstream_file->ops = &DataInodeOps;
#else
    evstream_file->proc_iops = &DataInodeOps;
    evstream_file->proc_fops = &DataFileOps;
#endif
    evstream_file->data = lmd;
    /* Stats file */
    evstream_file = create_proc_entry("stats",0,evstream);
#if PD_LINUX < 0x02004
    evstream_file->ops = &StatsInodeOps;
#else
    evstream_file->proc_iops = &StatsInodeOps;
    evstream_file->proc_fops = &StatsFileOps;
#endif
    evstream_file->data = lmd;

}

void eventfs_machine_post_event_to_stream(
        struct eventfs_stream *stream, struct eventfs_entry *entry
        )
{
    struct linux_machine_data          *lmd;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Posting event to %s/%s\n",
            stream->efstream_group->efgroup_name, stream->efstream_name);
#endif
    lmd = stream->efstream_data;

    /* Wake up anyone blocked. */
    wake_up_interruptible(&(lmd->waitq));
    if (lmd->sigq)
#ifdef PD_LINUX_KILL_FASYNC_NEED_3_ARGS
#if PD_LINUX < 0x02004
        kill_fasync(lmd->sigq, SIGIO, SI_SIGIO);
#else
        kill_fasync(&(lmd->sigq), SIGIO, SI_SIGIO);
#endif
#else
        kill_fasync(lmd->sigq, SIGIO);
#endif
}

/*
 * driver Functions that impliment the reads.
 * Its a pipe... We don't support seek here.
 *
 * We do support lseek forward if you want to skip events
 */
static ssize_t linux_eventfs_data_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        )
{
    struct eventfs_event               *event;
    struct eventfs_stream              *stream;
    struct eventfs_stream_entry        *stream_iter;
    struct linux_machine_data          *lmd;
    int                                 count;
    int                                 off;

    /* No pread. */
    if (offset != &file->f_pos)
        return -ESPIPE;

    if (!length) return 0;

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == linux_machine_data *          */
    lmd = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;
    stream = lmd->evstream;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Reading from stream %s/%s\n",
            stream->efstream_group->efgroup_name, stream->efstream_name);
#endif

    /* Are we going away? */
    if (!stream || stream->efstream_flags & EFS_FLG_GOING)
        return 0;

    /* file->private_data is our iterator.  If its not set, we didn't
     * get here through open.  FDGB.
     */
    stream_iter = file->private_data;
    if (stream_iter == NULL)
        return 0;

    /* We return 0 to mean EOF.  Since this is a stream, that doesn't
     * happen much.
     */

    count = 0;

    /* Stream->stream_entry_list->stream_entry.
     * "eventfs_stream_entry" is the list type.  Each stream_entry
     * has one entry, which is the actual event type.  (There is also
     * a global list of events.  I'm gonna hope I don't need to know
     * about that.)
     *
     * Stream has a list of stream_entry objects, which each have a single
     * entry object, which encapsulates a event object.
     *
     * the list of stream_entry objects is always finished by a sentinal
     * object.  The sentinal at the end does not have a entry object.
     */
    /*
     * 0) Block?
     *
     * 1) Finish up any partial events.
     * 2) Write events until we hit the sentinal
     */
#define STREAM_EMPTY(iter) ((iter)->efstrent_entry == NULL)

    /* We can not be in the middle of the sentinal event.  FDGB */
    assert(*offset == 0 || !STREAM_EMPTY(stream_iter));

    if (STREAM_EMPTY(stream_iter)) {
        /* No data to write.  Lets block. */
        if (file->f_flags & O_NONBLOCK)
            return -EAGAIN; /* Or not, if they don't want to */

        /* Three things can cause us to stop blocking:
         * 1) An event, in which case we can drop out the
         *    bottom of this loop and write data.
         * 2) a signal.  Restart.
         * 3) Going bye-bye.  Return EOF.
         */
#ifdef EVENTFS_DIAGNOSTICS
        printk(KERN_WARNING"Blocking on stream\n");
#endif
        while(STREAM_EMPTY(stream_iter) && !signal_pending(current)
                && !(stream->efstream_flags & EFS_FLG_GOING)) {
            interruptible_sleep_on(&(lmd->waitq));
        }
#ifdef EVENTFS_DIAGNOSTICS
        printk(KERN_WARNING"Woke on stream\n");
#endif
        /* Signal? */
        if (signal_pending(current))
            return -ERESTARTSYS;

        /* Going ByeBye? */
        if (stream->efstream_flags & EFS_FLG_GOING)
            return 0;

        /* Else, we have data.  So lets write it. */
    }
    off = 0;

    event       = &(stream_iter->efstrent_entry->efsent_event);
#ifdef EVENTFS_DIAGNOSTICS
    if (event->efsevent_size > 1024) {
        printk(KERN_WARNING"Large event being posted.  YMMV.\n");
    }
#endif

    /* Write out events */
    while ( count < length) {
        int                             wholesize;
        int                             writesize;
        int                             rval;
        struct eventfs_stream_entry    *old;

        wholesize = sizeof(*event)+event->efsevent_size-*offset;
        writesize = MIN( (length - count), wholesize);

#ifdef EVENTFS_DIAGNOSTICS
        printk(KERN_WARNING"Writing out data from stream\n");
#endif
        rval = copy_to_user(u_buf+count,((char*)event)+*offset,writesize);
        count += writesize;

        if (rval)
            return -EFAULT;

        /* We done?  If we couldn't write the whole thing, we
         * are.
         */
        if (writesize < wholesize) {
            *offset += writesize;
            return count;
        }

        /* So the next one can't be a partial */
        *offset = 0;

        /* Increment iterator */
        /* Don't "up" the new refcount.  The sentenial inherits ref counts,
         * so it already did that for us.
         */
        old = stream_iter;
        stream_iter = stream_iter->efstrent_link.le_next;
        file->private_data = stream_iter;
        eventfs_strent_rele(old);

        if (STREAM_EMPTY(stream_iter)) {
            return count;
        }

        event = &(stream_iter->efstrent_entry->efsent_event);

#ifdef EVENTFS_DIAGNOSTICS
        if (event->efsevent_size > 1024) {
            printk(KERN_WARNING"Large event being posted.  YMMV.\n");
        }
#endif
    }
    return count;
}

/* Drive function to impliment poll (select/poll/etc framework */
static unsigned int linux_eventfs_poll(struct file *file, poll_table *wait)
{
    struct eventfs_stream              *stream;
    struct eventfs_stream_entry        *stream_iter;
    struct linux_machine_data          *lmd;
    unsigned int                        mask;

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == linux_machine_data *          */
    lmd = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;
    stream = lmd->evstream;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Polling stream %s/%s\n",
            stream->efstream_group->efgroup_name, stream->efstream_name);
#endif

    mask = 0;

    /* Are we going away? */
    if (!stream || stream->efstream_flags & EFS_FLG_GOING)
        return POLLHUP; /* EOF.  Let em read to get the 0 */

    /* file->private_data is our iterator.  If its not set, we didn't
     * get here through open.  FDGB.
     */
    stream_iter = file->private_data;
    if (stream_iter == NULL)
        return POLLERR; /* Let em read to get the error */

    /* Put myself on the wait Q with the poll_wait function so that
     * the process will get awoken when I during event_post
     *
     * Apparently, magic happens outside this function to remove it
     * if unneeded or sleep if needed.
     */
    poll_wait(file,&(lmd->waitq),wait);

    if (!STREAM_EMPTY(stream_iter))
        mask |= POLLIN|POLLRDNORM;

    return mask;
}

#ifndef SEEK_END
#define SEEK_END 2
#endif
#ifndef SEEK_CUR
#define SEEK_CUR 1
#endif
#ifndef SEEK_SET
#define SEEK_SET 0
#endif
static loff_t linux_eventfs_lseek(struct file *file, loff_t offset, int orig)
{
    struct eventfs_event               *event;
    struct eventfs_stream              *stream;
    struct eventfs_stream_entry        *stream_iter;
    struct linux_machine_data          *lmd;
 
    loff_t                              retval;
    loff_t                              off;

    switch (orig) {
    default:
        return -EINVAL;
    case SEEK_SET:
        /* In stream? */
        return -EINVAL;
    case SEEK_END:
        /* From end */
        /* This is a pita.  I have no idea how far from the end I am... */
        return -EINVAL;
    case SEEK_CUR:
        /* From current */
        if (offset < 0)
            return -EINVAL;

        if (offset == 0) {
            /* Noop? */
            return file->f_pos;
        }

        /* Seek past any events they want to skip.  Hopefully this would
         * actually take place in the lseek operation instead of here, 
         * but I can't be sure.
         *
         * If they hit the end, reset offset to 0, and EAGAIN
         */

        /* file->f_dentry->d_inode == struct inode *                */
        /* inode->u.generic_ip     == struct proc_dir_entry *       */
        /* dp->data                == linux_machine_data *          */
        lmd = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                    ->u.generic_ip))->data;
        stream = lmd->evstream;

#ifdef EVENTFS_DIAGNOSTICS
        printk(KERN_WARNING"Seeking on stream %s/%s\n",
            stream->efstream_group->efgroup_name, stream->efstream_name);
#endif
        /* file->private_data is our iterator.  If its not set, we didn't
         * get here through open.  FDGB.
         */
        stream_iter = file->private_data;
        if (stream_iter == NULL)
            return -EBADF;

        off     = 0;
        retval  = 0;
        event   = &(stream_iter->efstrent_entry->efsent_event);

        while (off < offset && !STREAM_EMPTY(stream_iter)) {
            struct eventfs_stream_entry *old;
            /* Skip any events.  We reset offset */

            if ((off + sizeof(*event) + event->efsevent_size ) > offset ) {
                /* If we need to finish this event, stop here. */
                retval = off + sizeof(*event) + event->efsevent_size - offset;
                break;
            }

            off += sizeof(*event) + event->efsevent_size;

            /* Play fast and loose with ref-counting since we don't support
             * SMP anyway.
             */
            old = stream_iter;
            stream_iter = stream_iter->efstrent_link.le_next;
            eventfs_strent_ref(stream_iter);
            file->private_data = stream_iter;
            eventfs_strent_rele(old);

            event = &(stream_iter->efstrent_entry->efsent_event);
        }

        file->f_pos = retval;
        return retval;
    }
}

/* Danger Danger!  No length checking! */
/* Format:
 * 1 \t     event_name\n
 * 4 \t     event_name\n
 *
 * Only lists the events in "mask"
 */
static ssize_t stats_read(char *buf, char *printname, u_long mask,
                            struct eventfs_group *group)
{
    ssize_t     count;
    int         i;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Reading stats on group %s with mask %x\n",
            group->efgroup_name, (unsigned int) mask );
#endif

    count = sprintf( buf, "[%s]\n", printname );
    for (i = 0; ( i < EFGROUP_NTYPES) && (count < PAGE_SIZE-1) ; ++i) {
        if (group->efgroup_types[i] && (mask & (((u_long)1)<<i) ) ) {
            count += sprintf( buf + count, "%lu\t%s\n",
                    (((u_long)1)<<i), group->efgroup_types[i]);
        }
    }

    return count;
}

static ssize_t linux_eventfs_stats_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        )
{
    struct eventfs_stream      *stream;
    struct linux_machine_data  *lmd;
    size_t                      count;
    char                       *buf;

    if (*offset > PAGE_SIZE)
        return 0;

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == linux_machine_data *          */
    lmd = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;
    stream = lmd->evstream;

    /* Format:
     * [stream_name]\n
     * 1 \t     event_name\n
     * 4 \t     event_name\n
     */

    buf = (char*)__get_free_page(GFP_KERNEL);
    if (!buf)
        return -ENOMEM;

    count = stats_read(buf,stream->efstream_name, stream->efstream_mask,
                        stream->efstream_group);

    /* Now write the correct portion out! */
    length = MAX(MIN(length, count-*offset),0);
    if (length) {
        copy_to_user(u_buf,buf+*offset,length);
        *offset += length;
    }
    free_page((unsigned long)buf);

    return length;
}
static ssize_t linux_eventfs_gr_stats_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        )
{
    struct eventfs_group       *group;
    size_t                      count;
    char                       *buf;

    if (*offset > PAGE_SIZE)
        return 0;

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == struct eventfs_group *        */
    group = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;

    /* Format:
     * 1 \t     event_name\n
     * 4 \t     event_name\n
     */

    buf = (char*)__get_free_page(GFP_KERNEL);
    if (!buf)
        return -ENOMEM;

    count = stats_read(buf,group->efgroup_name, 0xFFFFFFFF,group);

    /* Now write the correct portion out! */
    length = MAX(MIN(length, count-*offset),0);
    if (length) {
        copy_to_user(u_buf,buf+*offset,length);
        *offset += length;
    }
    free_page((unsigned long)buf);

    return length;
}

static int linux_eventfs_open(struct inode *inode, struct file *file)
{
    /* Basically should have to do very little asside from setting up
     * the private_data pointer to point to the iterator.
     * Iterator starts at the sentenial (tail).
     */
    struct linux_machine_data          *lmd;
    struct eventfs_stream              *stream;
    struct eventfs_stream_entry        *stream_iter;
    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == linux_machine_data *          */
    lmd = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;
    if (lmd == NULL)
        return -EBADF;

    stream = lmd->evstream;

    if (!stream || stream->efstream_flags & EFS_FLG_GOING)
        return -EBADF;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Opening stream %s/%s\n",
            stream->efstream_group->efgroup_name, stream->efstream_name);
#endif
    /* Initialize iterator */
    stream_iter = stream->efstream_tail;
    eventfs_strent_ref(stream_iter);
    file->private_data = stream_iter;
    /* Reference the stream to let generic eventfs know someone cares. */
    eventfs_stream_ref(stream);

    return 0;
}
static int linux_eventfs_release(struct inode *inode, struct file *file)
{
    struct linux_machine_data          *lmd;
    struct eventfs_stream_entry        *stream_iter;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Closing some stream\n");
#endif

    /* Remove any IO-using waitors */
    linux_eventfs_fasync(-1,file,0);

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == linux_machine_data *          */
    lmd = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;
    if (lmd == NULL)
        return -EBADF;

    /* Clean up the iterator.
     */
    stream_iter = file->private_data;
    if (stream_iter == NULL) {
        printk(KERN_WARNING"Eventfs: Invalid data in filep\n");
        return 0;
    }

    eventfs_strent_rele(stream_iter);
    file->private_data = NULL;

    /* Dereference the stream to let generic eventfs know we don't care
     * anymore
     */
    eventfs_stream_rele(lmd->evstream);

    return 0;
}
static int linux_eventfs_fasync(int fd, struct file *file, int fOn)
{
    struct linux_machine_data          *lmd;
    struct fasync_struct               *fa;
    struct fasync_struct               *fna = NULL;
    struct fasync_struct              **prev;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Toggling fasync status on stream\n");
#endif

    if (fOn) {
        fna = (struct fasync_struct*)kmalloc(sizeof(*fna),GFP_KERNEL);
        if (fna == NULL)
            return -ENOMEM;
    }

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == linux_machine_data *          */
    lmd = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;

    prev = &(lmd->sigq);

    for (fa=*prev; fa != NULL; prev=&fa->fa_next, fa=*prev)
        if (fa->fa_file == file)
            break;

    if (fOn) {
        if (fa != NULL) {
            fa->fa_fd = fd;
            kfree(fna);
        }
        fna->fa_file = file;
        fna->fa_fd = fd;
        fna->magic = FASYNC_MAGIC;
        fna->fa_next = lmd->sigq;
        lmd->sigq = fna;
    } else {
        if (fa != NULL) {
            *prev = fa->fa_next;
            kfree(fa);
        }
    }

    return 0;
}

static int linux_eventfs_ioctl_create_stream(
        struct eventfs_group *group, struct eventfs_request *u_req)
{
    int                         retval;
    unsigned long               mask;
    char                       *name;

    retval = copy_from_user(&mask, &(u_req->mask), sizeof(mask));
    if (retval)
        return retval;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"New? stream requested: mask = %x\n",(uint)mask);
#endif

    retval = eventfs_construct_stream(group->efgroup_name,mask,&name);
    if (retval)
        return retval;

    retval = copy_to_user(&(u_req->name),name,MAXNAMELEN);
    kfree(name);

    return retval;
}

static int linux_eventfs_ioctl(
        struct inode *inode, struct file *file, uint which, ulong arg)
{
    struct eventfs_group       *group;

    if (which == (int)EFSIOCREQ) {
        /* file->f_dentry->d_inode == struct inode *                */
        /* inode->u.generic_ip     == struct proc_dir_entry *       */
        /* dp->data                == struct eventfs_group *        */
        group = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                    ->u.generic_ip))->data;

        if (!group)
            return -EINVAL;
        return linux_eventfs_ioctl_create_stream(group,
                (struct eventfs_request *)arg);
    } else {
        return -EINVAL;
    }
}

static int linux_eventfs_ctl_open(struct inode *inode, struct file *file)
{
    /* Basically should have to do very little asside from setting up
     * the private_data pointer to point to the request data.
     */
    struct eventfs_group               *eg;
    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == eventfs_group *               */
    eg = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;
    if (!eg)
        return -EBADF;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Opening group control-file %s\n",
            eg->efgroup_name);
#endif
    /* Initialize */
    file->private_data = NULL;

    /* Reference the group to let generic eventfs know someone cares. */
    eventfs_group_ref(eg);

    return 0;
}
static int linux_eventfs_ctl_release(struct inode *inode, struct file *file)
{
    struct eventfs_group               *eg;

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"Closing some control file\n");
#endif

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == eventfs_group *               */
    eg = ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;
    if (!eg)
        return -EBADF;

    /* Clean up the output field. */
    if (file->private_data != NULL)
        kfree(file->private_data);

    file->private_data = NULL;

    /* Dereference the group to let generic eventfs know we don't care
     * anymore
     */
    eventfs_group_rele(eg);

    return 0;
}

static ssize_t linux_eventfs_ctl_read(
        struct file *file, char *u_buf, size_t length, loff_t *offset
        )
{
    struct eventfs_group       *eg;
    size_t                      count;
    char                       *buf;

    if (*offset > PAGE_SIZE)
        return 0;

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == struct eventfs_group *        */
    eg= ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;

    if (!eg)
        return -EBADF;

    if (file->private_data == NULL)
        return -EINVAL;

    buf = (char*)__get_free_page(GFP_KERNEL);
    if (!buf)
        return -ENOMEM;

    strncpy(buf, file->private_data, PAGE_SIZE-1);
    buf[PAGE_SIZE-1] = '\0';

    count = strlen(buf);

    /* Now write the correct portion out! */
    length = MAX(MIN(length, count-*offset),0);
    if (length) {
        copy_to_user(u_buf,buf+*offset,length);
        *offset += length;
    }
    free_page((unsigned long)buf);

    if (*offset >= count) {
        kfree(file->private_data);
        file->private_data = NULL;
        *offset = 0;
    }

    return length;
}

static ssize_t linux_eventfs_ctl_write(
        struct file *file, const char *u_buf, size_t length, loff_t *offset
        )
{
    struct eventfs_group       *eg;
    char                       *scratch;
    char                       *name;
    unsigned long               mask;
    int                         i;

    if (*offset != 0)
        return -EINVAL;

    /* file->f_dentry->d_inode == struct inode *                */
    /* inode->u.generic_ip     == struct proc_dir_entry *       */
    /* dp->data                == struct eventfs_group *        */
    eg= ((struct proc_dir_entry*)((file->f_dentry->d_inode)
                                        ->u.generic_ip))->data;

    if (!eg)
        return -EBADF;

    /* One write, one read.  In that order. */
    if (file->private_data != NULL)
        return -EINVAL;

    /* Input is "12345" 
     * I'm a lazy bastard, so num goes until a non-numerical character,
     * is only base-10.
     *
     * ceil(log10(2^32)) +1 for nul = 11
     */

    scratch = kmalloc(11,GFP_KERNEL);
    if (!scratch)
        return -ENOMEM;
    scratch[10] = '\0';

    mask = 0;

    copy_from_user(scratch, u_buf, 10);

    for (i = 0; i < 10; ++i) {
        if (scratch[i] < '0' || scratch[i] > '9')
            break;
        mask *= 10;
        mask += scratch[i] - '0';
    }

#ifdef EVENTFS_DIAGNOSTICS
    printk(KERN_WARNING"New? stream requested: mask = %x\n",(uint)mask);
#endif

    i = eventfs_construct_stream(eg->efgroup_name,mask,&name);
    kfree(scratch);

    if (i)
        return -i;
    else {
        file->private_data = name;
        return length; /* fakeit. */
    }
}
