/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: eventfs_subr.c,v 1.30 2001/02/23 19:02:20 stevek Exp $
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <sys/kmem.h>
#include <sys/mutex.h>
#include <sys/vfs.h>
#include <sys/vnode.h>

#include <eventfs.h>
#include <eventfs_machine.h>
#include "eventfs_solaris.h"

#define NEFCACHE 4
#define EF_NHASH(ix) (&efhashtbl[(ix) & efhash])
static PD_LIST_HEAD(efhashhead, eventfs_opennode) *efhashtbl = NULL;
static u_long efhash = 0;

static kmutex_t efcache_lock;

extern void _eventfs_close_stream_unsafe( struct eventfs_stream *,
                                          struct eventfs_stream_entry * );

void *
hashinit( int elements, u_long *hashmask )
{
   long                            hashsize;
   PD_LIST_HEAD(generic, generic) *hashtbl;
   u_long                           size;
   int                              i;
 
   if( elements <= 0 )
      panic( "hashinit: bad elements" ); 

   for( hashsize = 1; hashsize <= elements; hashsize <<= 1 )
      continue;

   hashsize >>= 1;
   size = (u_long) ( hashsize * sizeof(*hashtbl) ) + sizeof(u_long);
   hashtbl = kmem_alloc( size, KM_SLEEP );

   *(u_long *)hashtbl = size;
   hashtbl = (void *)( (u_long *)hashtbl - 1 );

   for( i = 0; i < hashsize; i++ )
      LIST_INIT( &hashtbl[i] );

   *hashmask = hashsize - 1;
   return (hashtbl);
}

static size_t
strnlen( const pd_char_t *s, size_t n )
{
   int i;
   for( i = 0 ; i < n ; i++ )
      if( s[i] == '\0' )
         break;
   return (i);
}

int
efs_snprintf( pd_char_t *s, size_t n, const pd_char_t *format, ... )
{
   va_list ap;
   const pd_char_t *retval;

   va_start( ap, format );
   retval = vsprintf_len( n, s, format, ap );
   va_end( ap );

   return retval == 0 ? -1 : strnlen( s, n );
}

int
eventfs_machine_init()
{
#ifdef EVENTFS_DIAGNOSTICS
   printf( "eventfs_machine_init: Initializing hash table\n" );
#endif /* EVENTFS_DIAGNOSTICS */
   efhashtbl = hashinit( NEFCACHE, &efhash );
#ifdef EVENTFS_DIAGNOSTICS
   printf( "eventfs_machine_init: Done\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (0);
}

int  eventfs_machine_uninit() { return 0; }

void eventfs_machine_destroy_group( struct eventfs_group *eg ) {}
void eventfs_machine_destroy_stream( struct eventfs_stream *es ) {}
void eventfs_machine_register_group( struct eventfs_group *eg ) {}
void eventfs_machine_register_stream( struct eventfs_stream *es ) {}

void
eventfs_machine_flush( struct eventfs_stream *es )
{
   struct eventfs_node *node = NULL;
   int                  i;

   for( i = 0; i < efhash; i++ )
   {
      struct eventfs_opennode *efsopn;
      struct efhashhead       *ef     = &efhashtbl[i];

      for( efsopn = ef->lh_first;
           efsopn;
           efsopn = efsopn->efsopen_hash.le_next )
      {
         if( ( efsopn->efsopen_node->efs_type == Estream ) &&
             ( efsopn->efsopen_node->un.efs_stream == es ) )
         {
            struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;
            if( node == NULL )
               node = efsopn->efsopen_node;

            /* Decrement the reference count, release memory */
            if( --efsopn->efsopen_refcount <= 1 )
            {
               PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
               _eventfs_close_stream_unsafe( es, efstrent );
               PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );
               LIST_REMOVE( efsopn, efsopen_hash );
               kmem_free( efsopn, sizeof(*efsopn) );
            }

            efsopn->efsopen_curentry = es->efstream_tail;
         }
      }
   }

   /* Wake up any readers blocked on the stream */
   if( node )
      cv_broadcast( &(node->efs_cv) );
}

void
eventfs_machine_post_event_to_stream( struct eventfs_stream *es,
                                      struct eventfs_entry *eent )
{
   if( es->efstream_data )
   {
      struct eventfs_node *node = (struct eventfs_node *)es->efstream_data;

#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_STREAM )
         printf( "eventfs_machine_post_event_to_stream: posting\n" );
#endif /* EVENTFS_DIAGNOSTICS */

      cv_broadcast( &(node->efs_cv) );
   }
}

int
eventfs_check_stream( struct eventfs_node *efs )
{
   switch( efs->efs_type )
   {
      case Esdata:
      case Esstats:
         /* Stream has been destroyed and is going away */
         if( efs->un.efs_stream->efstream_flags & EFS_FLG_GOING )
         {
            int error = (VOP_CLOSE( efs->efs_node, 8193, 1, 0, NULL ));
            if( error )
               return error;
            return (EINTR);
         }
         break;

      default:
         break;
   }
   return (0);
}

int
eventfs_getflags( struct eventfs_node *efs )
{
   struct eventfs_opennode *efsopn;
   struct efhashhead       *ef;

   ef = EF_NHASH(curproc->p_pid);
   for( efsopn = ef->lh_first;
        efsopn && ( efsopn->efsopen_node != efs );
        efsopn = efsopn->efsopen_hash.le_next )
      continue;

   if( efsopn )
      return efsopn->efsopen_flags;
   return 0;
}

void
eventfs_setflags( struct eventfs_node *efs, int flags )
{
   struct eventfs_opennode *efsopn;
   struct efhashhead       *ef;

   ef = EF_NHASH(curproc->p_pid);
   for( efsopn = ef->lh_first;
        efsopn && ( efsopn->efsopen_node != efs );
        efsopn = efsopn->efsopen_hash.le_next )
      continue;

   if( efsopn )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
         printf( "eventfs_setflags: 0x%x\n", flags );
#endif /* EVENTFS_DIAGNOSTICS */
      efsopn->efsopen_flags = flags;
   }
}

int
eventfs_doopen( struct eventfs_node *efs, struct proc *p )
{
   struct eventfs_opennode *efsopn;
   struct efhashhead       *ef;
   int                      error;

   if( ( error = eventfs_check_stream( efs ) ) )
      return error;

   ef = EF_NHASH(p->p_pid);
   for( efsopn = ef->lh_first;
        efsopn && ( efsopn->efsopen_node != efs );
        efsopn = efsopn->efsopen_hash.le_next )
      continue;

   if( efsopn == NULL )
   {
      mutex_enter( &efcache_lock );

      /* Add a new opennode to the hash table */
      efsopn = (struct eventfs_opennode *)kmem_zalloc( sizeof(*efsopn),
                                                       KM_SLEEP );
      efsopn->efsopen_refcount = 1;
      efsopn->efsopen_node = efs;
      efsopn->efsopen_offset = 0;
      efsopn->efsopen_curentry = efs->un.efs_stream->efstream_tail;
      efsopn->efsopen_curentry->efstrent_refcount++;
      efs->un.efs_stream->efstream_refcount++;

#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_OPEN )
         printf( "eventfs_doopen: \"%s\", refcount 0x%x, hash 0x%x\n",
                 efs->un.efs_stream->efstream_name,
                 (u_int) efs->un.efs_stream->efstream_refcount,
                 (u_int) (curproc->p_pid & efhash) );
#endif /* EVENTFS_DIAGNOSTICS */

      LIST_INSERT_HEAD( ef, efsopn, efsopen_hash );

      mutex_exit( &efcache_lock );
   }
   else
   {
      /* Increment the reference count */
      efsopn->efsopen_refcount++;
   }

   return (0);
}

int
eventfs_doclose( struct eventfs_node *efs, struct proc *p )
{
   struct eventfs_opennode *efsopn;

   efsopn = eventfs_lookup_opennode( p, efs );
   if( efsopn == NULL )
      panic( "eventfs_close: no matching eventfs_opennode" );
   else
   {
      struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;

      /* Decrement the reference count, release memory on last release */
      if( --efsopn->efsopen_refcount <= 1 )
      {
         struct eventfs_stream *es = efsopn->efsopen_node->un.efs_stream;
         PD_MUTEX_ENTER( EFSLINKAGE, es->efstream );
         _eventfs_close_stream_unsafe( es, efstrent );
         PD_MUTEX_EXIT( EFSLINKAGE, es->efstream );
         LIST_REMOVE( efsopn, efsopen_hash );
         kmem_free( efsopn, sizeof(*efsopn) );
      }
   }

   return (0);
}


int
eventfs_dodata( struct eventfs_node *efs, struct uio *uio,
                unsigned int *retnread, pd_bool_t *eof )
{
   struct eventfs_opennode *efsopn;
   int                      rval   = 0;
   unsigned int             nread  = 0;

   if( ( retnread == NULL ) || ( eof == NULL ) )
      return (EINVAL);

   efsopn = eventfs_lookup_opennode( curproc, efs );
   if( efsopn == NULL )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
         printf( "eventfs_read: unable to find open stream\n" );
#endif /* EVENTFS_DIAGNOSTICS */
      rval = EBADF;
   }
   else
   {
      /* We found the open stream */
      int off = efsopn->efsopen_offset;
      int seekoff = efsopn->efsopen_offset;
      int offmax = uio->uio_offset + efsopn->efsopen_offset;

      /* Seek into the stream */
      while( off < offmax )
      {
         struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;
         struct eventfs_entry        *efsent   = efstrent->efstrent_entry;

         if( efsent == NULL )
         {
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
               printf( "eventfs_read: no events pending\n" );
#endif /* EVENTFS_DIAGNOSTICS */
            *eof = TRUE;
            break;
         }

         if( ( off + sizeof(struct eventfs_event) ) > offmax )
         {
            seekoff = off + sizeof(struct eventfs_event) - offmax;
            break;
         }
         else
         {
            off += sizeof(struct eventfs_event);

            if( ( off + efsent->efsent_event.efsevent_size ) > offmax )
            {
               seekoff = off + efsent->efsent_event.efsevent_size - offmax;
            }
            else
            {
               /* Flush this packet and move on */
               off += efsent->efsent_event.efsevent_size;
               efsopn->efsopen_curentry = efstrent->efstrent_link.le_next;

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_read: flushing packet\n" );
#endif /* EVENTFS_DIAGNOSTICS */

               eventfs_strent_rele( efstrent );
            }
         }
      }

      /* Read data from stream */
      if( off >= offmax )
      {
         while( ( rval == 0 ) && uio->uio_resid )
         {
            struct eventfs_stream_entry *efstrent = efsopn->efsopen_curentry;
            struct eventfs_entry        *efsent   = efstrent->efstrent_entry;

            if( efsent == NULL )
            {
#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_read: no events pending\n" );
#endif /* EVENTFS_DIAGNOSTICS */
               *eof = TRUE;
               break;
            }
         
            if( seekoff < sizeof(struct eventfs_event) )
            {
               /* Read in event head */
               int xlen = sizeof(struct eventfs_event) - seekoff;

               xlen = min( xlen, uio->uio_resid );

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_read: reading event head (offset %d, len %d)\n",
                          seekoff, xlen );
#endif /* EVENTFS_DIAGNOSTICS */

               if( xlen <= 0 )
               {
#ifdef EVENTFS_DIAGNOSTICS
                  if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                     printf( "eventfs_read: no space left in buffer\n" );
#endif /* EVENTFS_DIAGNOSTICS */

                  rval = 0;
                  break;
               }

               /* Write out event head chunk to requestor */
               rval = uiomove( (char *) (&efsent->efsent_event) + seekoff, xlen,
                               UIO_READ, uio );
               if( rval )
                  break;

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_read: adjusting offset by %d\n", xlen );
#endif /* EVENTFS_DIAGNOSTICS */
               nread += xlen;
               seekoff += xlen;
            }
            else
            {
               int xlen;

               off = seekoff - sizeof(struct eventfs_event);

               /* Read in event data packet */
               xlen = efsent->efsent_event.efsevent_size - off;

#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                  printf( "eventfs_read: reading event data packet\n" );
#endif /* EVENTFS_DIAGNOSTICS */

               if( xlen > 0 )
               {
                  /* Write out event data packet chunk to requestor */
                  rval = uiomove( (char *) (efsent + 1) + off, xlen, UIO_READ,
                                  uio );
                  if( rval )
                     break;

                  nread += xlen;
                  seekoff += xlen;
               }

               if( seekoff >= ( efsent->efsent_event.efsevent_size +
                                sizeof(struct eventfs_event) ) )
               {
                  /* Flush this packet and move on */
                  seekoff = 0;
                  efsopn->efsopen_curentry = efstrent->efstrent_link.le_next;

#ifdef EVENTFS_DIAGNOSTICS
                  if( eventfs_debug_mask & EFS_DEBUG_MD_DATA )
                     printf( "eventfs_read: flushing packet\n" );
#endif /* EVENTFS_DIAGNOSTICS */

                  eventfs_strent_rele( efstrent );
               }
            }
         }

         /* Adjust offset to be relative to remaining events in the stream */
         uio->uio_offset = 0;
         efsopn->efsopen_offset = seekoff;
      }
   }

   *retnread = nread;
   return (rval);
}


int
eventfs_dostats( struct eventfs_node *efs, struct uio *uio )
{
   struct eventfs_group *eg;
   char                 *sbuf;
   u_long                mask;
   int                   error;
   int                   ndx;
   int                   i;

   if( ( error = eventfs_check_stream( efs ) ) )
      return error;

   if( uio->uio_offset != 0 )
      return (0);

   sbuf = kmem_zalloc( PAGESIZE, KM_SLEEP );

   switch( efs->efs_type )
   {
      case Esstats:
         eg   = efs->un.efs_stream->efstream_group;
         mask = efs->un.efs_stream->efstream_mask;
         ndx = efs_snprintf( sbuf, PAGESIZE - 1, "[%s::%s]\n",
                             eg->efgroup_name,
                             efs->un.efs_stream->efstream_name );
         break;

      case Egstats:
         eg   = efs->un.efs_group;
         mask = efs->un.efs_group->efgroup_eventmask;
         ndx = efs_snprintf( sbuf, PAGESIZE - 1, "[%s]\n", eg->efgroup_name );
         break;

      default:
         return (EOPNOTSUPP);
   }

   for( i = 0; ( i < EFGROUP_NTYPES ) && ( ndx < (PAGESIZE - 1) ); i++ )
   {
      if( eg->efgroup_types[i] && ( mask & (((u_long)1)<<i) ) )
         ndx += efs_snprintf( sbuf + ndx, PAGESIZE - ndx - 1, "%lu\t%s\n",
                              (((u_long)1)<<i), eg->efgroup_types[i] );
   }
   sbuf[PAGESIZE-1] = '\0';

   ndx = min( ndx, uio->uio_resid );

   error = (uiomove( sbuf, ndx, UIO_READ, uio ));
   kmem_free( sbuf, PAGESIZE );
   return error;
}

struct eventfs_opennode *
eventfs_lookup_opennode( struct proc *p, struct eventfs_node *efs )
{
   struct efhashhead       *ef;
   struct eventfs_opennode *efsopn;

   ef = EF_NHASH(p->p_pid);
   for( efsopn = ef->lh_first;
        efsopn && ( efsopn->efsopen_node != efs );
        efsopn = efsopn->efsopen_hash.le_next )
      continue;

   return efsopn;
}

