/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: eventfs_vfsops.c,v 1.9 2001/02/23 19:02:20 stevek Exp $
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <sys/vnode.h>
#include <sys/vfs.h>
#include <sys/mount.h>
#include <sys/mutex.h>
#include <sys/time.h>
#include <sys/ddi.h>
#include <sys/mkdev.h>
#include <sys/errno.h>

#include <eventfs.h>
#include <eventfs_machine.h>
#include "eventfs_solaris.h"

extern struct vfsops eventfs_vfsops;
static int eventfs_fstype;
int eventfs_major;
int eventfs_minor;
kmutex_t eventfs_minor_lock;

extern int  eventfs_init    __P(( void ));
extern int  eventfs_keep    __P(( void ));
extern int  eventfs_release __P(( void ));
extern void eventfs_vn_init __P(( void ));

int
eventfs_solaris_init( struct vfssw *vswp, int fstype )
{
#ifdef EVENTFS_DIAGNOSTICS
   printf( "eventfs_solaris_init: fstype = %d\n", fstype );
#endif /* EVENTFS_DIAGNOSTICS */

   eventfs_fstype = fstype;

   vswp->vsw_vfsops = &eventfs_vfsops;

   mutex_init( &eventfs_minor_lock, "eventfs minor lock", MUTEX_DEFAULT, NULL );

   eventfs_vn_init();

   if( ( eventfs_major = getudev() ) == -1 )
   {
      cmn_err( CE_WARN,
               "eventfs: eventfs_init: can't get unique device number" );

      eventfs_major = 0;
   }
   eventfs_minor = 0;

#ifdef EVENTFS_DIAGNOSTICS
   printf( "eventfs: major = %d, calling eventfs_init()\n", eventfs_major );
#endif /* EVENTFS_DIAGNOSTICS */

   return (eventfs_init());
}

int
eventfs_mount( struct vfs *vfsp, struct vnode *mvp, struct mounta *uap,
               struct cred *cred )
{
   dev_t         eventfs_dev;
   struct vnode *rootvp;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_mount\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( !suser( cred ) )
      return (EPERM);

   if( mvp->v_type != VDIR )
      return (ENOTDIR);

   eventfs_keep();

#if 0
   mutex_enter( &mvp->v_lock );
   if( ( mvp->v_count != 1 ) || ( mvp->v_flag & VROOT ) )
   {
      mutex_exit( &mvp->v_lock );
      return (EBUSY);
   }
   mutex_exit( &mvp->v_lock );
#endif

   if( !uap )
   {
      eventfs_release();
      return (EINVAL);
   }

   if( ( uap->flags & MS_REMOUNT ) ||
       ( uap->flags & MS_OVERLAY ) )
   {
      eventfs_release();
      return (EOPNOTSUPP);
   }

   vfsp->vfs_flag |= VFS_RDONLY | VFS_NOTRUNC | VFS_NOSUID;
   vfsp->vfs_bsize = mvp->v_vfsp->vfs_bsize;
   vfsp->vfs_fstype = eventfs_fstype;
   mutex_enter( &eventfs_minor_lock );
   do
   {
      eventfs_minor = (eventfs_minor + 1) & MAXMIN;
      eventfs_dev = makedevice( eventfs_major, eventfs_minor );
   }
   while( vfs_devsearch( eventfs_dev ) );
   mutex_exit( &eventfs_minor_lock );
   vfsp->vfs_dev = eventfs_dev;
   vfsp->vfs_fsid.val[0] = eventfs_dev;
   vfsp->vfs_fsid.val[1] = eventfs_fstype;
   vfsp->vfs_bcount = 0;
   eventfs_root( vfsp, &rootvp );
   vfsp->vfs_data = (void *) rootvp;
   vfsp->vfs_vnodecovered = mvp;
   mvp->v_vfsmountedhere = vfsp;

   return (0);
}

int
eventfs_unmount( struct vfs *vfsp, struct cred *cred )
{
   struct vnode *rootvp;
   struct vnode *mvp;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_unmount\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( !suser( cred ) )
      return (EPERM);

   rootvp = (struct vnode *)vfsp->vfs_data;

   if( !rootvp )
   {
      eventfs_release();
      return (0);
   }

   if( rootvp->v_count != 1 )
      return (EBUSY);

   mvp = vfsp->vfs_vnodecovered;
   mvp->v_vfsmountedhere = NULL;

   VN_RELE( rootvp );

   eventfs_release();

   return (0);
}

int
eventfs_root( struct vfs *vfsp, struct vnode **vpp )
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_root\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (eventfs_allocvp( vfsp, vpp, 0, Eroot ));
}

int
eventfs_statvfs( struct vfs *vfsp,
#ifdef _LARGEFILE64_SOURCE
                 struct statvfs64 *sbp )
#else
                 struct statvfs *sbp )
#endif
{
   int i;
   int nfiles = eventfs_ngroups;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_statvfs\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   sbp->f_bsize = PAGESIZE;
   sbp->f_frsize = PAGESIZE;
   sbp->f_blocks = 0;
   sbp->f_bfree = 0;
   sbp->f_bavail = 0;

   if( nfiles )
      for( i = 0; i < EVENTFS_NGROUPS; i++ )
         if( eventfs_groups[i] )
            nfiles += (eventfs_groups[i]->efgroup_nstreams * 3);

   sbp->f_files = nfiles;
   sbp->f_ffree = 0;
   sbp->f_favail = 0;
   sbp->f_fsid = 0xa1a55a1a;
   strcpy( sbp->f_basetype, "eventfs" );
   sbp->f_flag = ST_NOTRUNC | ST_RDONLY | ST_NOSUID;
   sbp->f_namemax = 256;
   sbp->f_fstr[0] = '\0';

   return (0);
}

int
eventfs_sync( struct vfs *vfsp, short flag, struct cred *cred )
{
   return (0);
}

int
eventfs_vget( struct vfs *vfsp, struct vnode **vpp, struct fid *fidp )
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_vget\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (ENOSYS);
}

int
eventfs_mountroot( struct vfs *vfsp, enum whymountroot reason )
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_mountroot\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (ENOSYS);
}

int
eventfs_swapvp( struct vfs *vfsp, struct vnode **vpp, char *path )
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_swapvp\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (ENOSYS);
}

