/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: eventfs_vnops.c,v 1.16 2001/02/23 19:02:20 stevek Exp $
 */

/*
 * Event filesystem (/event)
 */

#include <configuration.h>
#include <pdpi/pi_defs.h>

#include <sys/condvar.h>
#include <sys/dirent.h>
#include <sys/file.h>
#include <sys/fs_subr.h>
#include <sys/kmem.h>
#include <sys/mutex.h>
#include <sys/pathname.h>
#include <sys/vfs.h>
#include <sys/vnode.h>

#include <eventfs.h>
#include <eventfs_machine.h>
#include "eventfs_solaris.h"
#include "eventfs_ioctl.h"

#ifdef _LARGEFILE64_SOURCE
#define DIRENTLEN \
   (((int)(&((dirent64_t *)0)->d_name) + 1 + (NAMELEN) + 7) & ~ 7)
#else /* _LARGEFILE64_SOURCE */
#define DIRENTLEN \
   (((int)(&((dirent_t *)0)->d_name) + 1 + (NAMELEN) + 3) & ~ 3)
#endif /* _LARGEFILE64_SOURCE */

static int  eventfs_access   __P(( struct vnode *, int, int, struct cred * ));
static int  eventfs_close    __P(( struct vnode *, int, int, offset_t,
                                   struct cred * ));
static int  eventfs_getattr  __P(( struct vnode *, struct vattr *, int,
                                   struct cred * ));
static void eventfs_inactive __P(( struct vnode *, struct cred * ));
static int  eventfs_lookup   __P(( struct vnode *, char *, struct vnode **,
                                   struct pathname *, int, struct vnode *,
                                   struct cred * ));
static int  eventfs_open     __P(( struct vnode **, int, struct cred * ));
static int  eventfs_read     __P(( struct vnode *, struct uio *, int,
                                   struct cred * ));
static int  eventfs_readdir  __P(( struct vnode *, struct uio *, struct cred *,
                                   int * ));
static void eventfs_rwlock   __P(( struct vnode *, int ));
static void eventfs_rwunlock __P(( struct vnode *, int ));
static int  eventfs_setattr  __P(( struct vnode *, struct vattr *, int,
                                   struct cred * ));
static int  eventfs_setfl    __P(( struct vnode *, int, int, struct cred * ));

extern int eventfs_doopen    __P(( struct eventfs_node *, struct proc * ));
extern int eventfs_doclose   __P(( struct eventfs_node *, struct proc * ));
extern int eventfs_dostats   __P(( struct eventfs_node *, struct uio * ));
extern int eventfs_dodata    __P(( struct eventfs_node *, struct uio *,
                                   unsigned int *, pd_bool_t * ));

static int eventfs_badop();

static int eventfs_addmap();
static int eventfs_create();
static int eventfs_delmap();
static int eventfs_dispose();
static int eventfs_fid();
static int eventfs_fsync();
static int eventfs_getsecattr();
static int eventfs_ioctl();
static int eventfs_link();
static int eventfs_map();
static int eventfs_mkdir();
static int eventfs_poll();
static int eventfs_realvp();
static int eventfs_remove();
static int eventfs_rename();
static int eventfs_rmdir();
static int eventfs_setsecattr();
static int eventfs_seek();
static int eventfs_space();
static int eventfs_write();

struct vnodeops eventfs_vnodeop =
{
   eventfs_open,                      /* vop_open */
   eventfs_close,                     /* vop_close */
   eventfs_read,                      /* vop_read */
   eventfs_write,                     /* vop_write */
   eventfs_ioctl,                     /* vop_ioctl */
   eventfs_setfl,                     /* vop_setfl */
   eventfs_getattr,                   /* vop_getattr */
   eventfs_setattr,                   /* vop_setattr */
   eventfs_access,                    /* vop_access */
   eventfs_lookup,                    /* vop_lookup */
   eventfs_create,                    /* vop_create */
   eventfs_remove,                    /* vop_remove */
   eventfs_link,                      /* vop_link */
   eventfs_rename,                    /* vop_rename */
   eventfs_mkdir,                     /* vop_mkdir */
   eventfs_rmdir,                     /* vop_rmdir */
   eventfs_readdir,                   /* vop_readdir */
   eventfs_badop,                     /* vop_symlink */
   eventfs_badop,                     /* vop_readlink */
   eventfs_fsync,                     /* vop_fsync */
   eventfs_inactive,                  /* vop_inactive */
   eventfs_fid,                       /* vop_fid */
   eventfs_rwlock,                    /* vop_rwlock */
   eventfs_rwunlock,                  /* vop_rwunlock */
   eventfs_seek,                      /* vop_seek */
   eventfs_badop,                     /* vop_cmp */
   eventfs_badop,                     /* vop_frlock */
   eventfs_space,                     /* vop_space */
   eventfs_realvp,                    /* vop_realvp */
   eventfs_badop,                     /* vop_getpage */
   eventfs_badop,                     /* vop_putpage */
   (void *) eventfs_map,              /* vop_map */
   (void *) eventfs_addmap,           /* vop_addmap */
   eventfs_delmap,                    /* vop_delmap */
   (void *) eventfs_poll,             /* vop_poll */
   eventfs_badop,                     /* vop_dump */
   fs_pathconf,                       /* vop_pathconf */
   eventfs_badop,                     /* vop_pageio */
   eventfs_badop,                     /* vop_dumpctl */
   (void *) eventfs_dispose,          /* vop_dispose */
   eventfs_setsecattr,                /* vop_setsecattr */
   eventfs_getsecattr                 /* vop_getsecattr */
};

static struct stream_target
{
   u_char   st_namelen;
   char    *st_name;
   efstype  st_efstype;
}
stream_targets[] =
{
#define N(s) sizeof(s)-1, s
   { N("."),     Estream },
   { N(".."),    Egroup },
   { N("data"),  Esdata },
   { N("stats"), Esstats }
#undef N
};
static const int nstream_targets =
   sizeof(stream_targets) / sizeof(stream_targets[0]);

static struct group_target
{
   u_char   gt_namelen;
   char    *gt_name;
   efstype  gt_efstype;
}
group_targets[] =
{
#define N(s) sizeof(s)-1, s
   { N("."),     Egroup },
   { N(".."),    Eroot },
   { N("stats"), Egstats }
#undef N
};
static const int ngroup_targets =
   sizeof(group_targets) / sizeof(group_targets[0]);

static kmutex_t efsvplock;

static struct eventfs_node *efshead;

int eventfs_allocvp __P(( struct vfs *, struct vnode **, void *, efstype ));

void
eventfs_vn_init()
{
   mutex_init( &efsvplock, "eventfs vnode lock", MUTEX_DEFAULT, NULL );
}


int
eventfs_allocvp( struct vfs *vfsp, struct vnode **vpp, void *strgrp,
                 efstype efs_type )
{
   struct eventfs_node  *efs;
   struct vnode         *vp;
   struct eventfs_node **ep;

   mutex_enter( &efsvplock );

   for( efs = efshead; efs; efs = efs->efs_next )
   {
      vp = EVENTFSTOV(efs);
      if( ( efs->efs_type == efs_type ) && ( vp->v_vfsp == vfsp ) )
      {
         switch( efs_type )
         {
            case Eroot:
               break;

            case Egroup:
               if( efs->un.efs_group != (struct eventfs_group *)strgrp )
                  continue;

               break;

            case Egstats:
               if( efs->un.efs_group != (struct eventfs_group *)strgrp )
                  continue;

               break;

            case Estream:
               if( efs->un.efs_stream != (struct eventfs_stream *)strgrp )
                  continue;

               break;

            case Esdata:
               if( efs->un.efs_stream != (struct eventfs_stream *)strgrp )
                  continue;

               break;

            case Esstats:
               if( efs->un.efs_stream != (struct eventfs_stream *)strgrp )
                  continue;

               break;

            default:
               panic( "eventfs_allocvp" );
         }

         VN_HOLD( vp );

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
            printf( "eventfs_allocvp: returning vnode (0x%x)\n", (u_int) vp );
#endif /* EVENTFS_DIAGNOSTICS */

         *vpp = vp;
         mutex_exit( &efsvplock );
         return (0);
      }
   }

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_allocvp: allocating structures\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   efs = (struct eventfs_node *) kmem_zalloc( sizeof(struct eventfs_node),
                                              KM_SLEEP );

   vp = (struct vnode *) kmem_zalloc( sizeof(*vp), KM_SLEEP );
   VN_INIT( vp, vfsp, VNON, 0 );
   vp->v_op = &eventfs_vnodeop;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VFS )
      printf( "eventfs_allocvp: setting values for vnode (0x%x)\n", (u_int) vp );
#endif /* EVENTFS_DIAGNOSTICS */

   vp->v_data = (void *) efs;
   efs->efs_next = NULL;
   efs->efs_type = efs_type;
   mutex_init( &efs->efs_lock, "eventfs node lock", MUTEX_DEFAULT, NULL );
   cv_init( &efs->efs_cv, "eventfs node condvar", CV_DRIVER, NULL );

   switch( efs_type )
   {
      case Egroup:
         efs->un.efs_group = (struct eventfs_group *) strgrp;
         efs->efs_nodeid = EVENTFS_FILENO(efs->un.efs_group->efgroup_id, 0, 0,
                                          Egroup);
         vp->v_type = VDIR;
         break;

      case Egstats:
         efs->un.efs_group = (struct eventfs_group *) strgrp;
         efs->efs_nodeid = EVENTFS_FILENO(efs->un.efs_group->efgroup_id,
                                          0, 1, Egstats);
         vp->v_type = VREG;
         break;

      case Estream:
         efs->un.efs_stream = (struct eventfs_stream *) strgrp;
         efs->efs_nodeid =
            EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                           efs->un.efs_stream->efstream_id, 0, Estream);
         vp->v_type = VDIR;
         break;

      case Esdata:
      {
         struct eventfs_stream *es = (struct eventfs_stream *) strgrp;
         efs->un.efs_stream = es;
         es->efstream_data = (void *) efs;
         efs->efs_nodeid =
            EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                           efs->un.efs_stream->efstream_id, 1, Esdata);
         vp->v_type = VREG;
         break;
      }

      case Esstats:
         efs->un.efs_stream = (struct eventfs_stream *) strgrp;
         efs->efs_nodeid =
            EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                           efs->un.efs_stream->efstream_id, 2, Esstats);
         vp->v_type = VREG;
         break;

      case Eroot:      /* /event = dr-xr-xr-x */
         efs->efs_nodeid = EVENTFS_FILENO(0, 0, 0, Eroot);
         efs->un.efs_group = NULL;
         vp->v_type = VDIR;
         vp->v_flag = VROOT;
         break;

      default:
         panic( "eventfs_allocvp" );
         break;
   }
   efs->efs_node = vp;

   /* Append to end of list */
   for( ep = &efshead; *ep; ep = &(*ep)->efs_next )
      continue;
   *ep = efs;

   *vpp = vp;

   mutex_exit( &efsvplock );

   return (0);
}

static void
eventfs_rwlock( struct vnode *vp, int enter_type )
{
   struct eventfs_node *efs = VTOEVENTFS(vp);

   if( efs )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS)
         printf( "eventfs_rwlock(%d)\n", enter_type );
#endif /* EVENTFS_DIAGNOSTICS */

      mutex_enter( &efs->efs_lock );
   }
#ifdef EVENTFS_DIAGNOSTICS
   else if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfw_rwlock: no eventfs_node found for vnode 0x%x\n",
              (int) vp );
#endif /* EVENTFS_DIAGNOSTICS */
}

static void
eventfs_rwunlock( struct vnode *vp, int enter_type )
{
   struct eventfs_node *efs = VTOEVENTFS(vp);

   if( efs )
   {
#ifdef EVENTFS_DIAGNOSTICS
      if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
         printf( "eventfs_rwunlock(%d)\n", enter_type );
#endif /* EVENTFS_DIAGNOSTICS */

      mutex_exit( &efs->efs_lock );
   }
#ifdef EVENTFS_DIAGNOSTICS
   else if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfw_rwunlock: no eventfs_node found for vnode 0x%x\n",
              (int) vp );
#endif /* EVENTFS_DIAGNOSTICS */
}


static int
eventfs_open( struct vnode **vpp, int flags, struct cred *cred )
{
   struct eventfs_node *efs = VTOEVENTFS(*vpp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_open\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( flags & FWRITE )
      return (EPERM);

   if( efs->efs_type == Esdata )
      eventfs_doopen( efs, curproc );

   return (0);
}


static int
eventfs_close( struct vnode *vp, int flags, int count, offset_t offset,
               struct cred *cred )
{
   struct eventfs_node *efs = VTOEVENTFS(vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_close\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( efs->efs_type == Esdata )
      eventfs_doclose( efs, curproc );

   return (0);
}


/*
 * vp is the current namei directory
 * ndp is the name to locate in that directory...
 */
static int
eventfs_lookup( struct vnode *dvp, char *lookup, struct vnode **vpp,
                struct pathname *pnp, int flags, struct vnode *rdir,
                struct cred *cred )
{
   struct eventfs_node *efs;
   struct vfs          *vfsp;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_lookup\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   *vpp = NULL;

   if( !strcmp( lookup, "." ) )
   {
      *vpp = dvp;
      VN_HOLD(dvp);
      return (0);
   }

   vfsp = dvp->v_vfsp;
   efs = VTOEVENTFS(dvp);

   switch( efs->efs_type )
   {
      case Eroot:
      {
         struct eventfs_group *eg = NULL;
         int                   i;

         if( !strcmp( lookup, ".." ) )
            return (EIO);

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_lookup(Eroot): %s\n", lookup );
#endif /* EVENTFS_DIAGNOSTICS */

         for( i = 0; i < EVENTFS_NGROUPS; i++ )
         {
            eg = eventfs_groups[i];
            if( eg == NULL )
               continue;

            if( ( strlen(lookup) == eg->efgroup_namelen ) &
                ( !strncmp( eg->efgroup_name, lookup, eg->efgroup_namelen ) ) )
               return (eventfs_allocvp( vfsp, vpp, eg, Egroup ));
         }

         break;
      }

      case Egroup:
      {
         struct eventfs_stream *es;
         struct group_target   *gt;
         int                    i;

         if( !strcmp( lookup, ".." ) )
         {
#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf( "eventfs_lookup(Egroup): ..\n" );
#endif /* EVENTFS_DIAGNOSTICS */
            return (eventfs_root( vfsp, vpp ));
         }

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_lookup(Egroup): %s\n", lookup );
#endif /* EVENTFS_DIAGNOSTICS */

         for( es = efs->un.efs_group->efgroup_streams.lh_first;
              es;
              es = es->efstream_link.le_next )
         {
            if( ( strlen( lookup ) == es->efstream_namelen ) &&
                ( !strncmp( es->efstream_name, lookup,
                            es->efstream_namelen ) ) )
               return (eventfs_allocvp( vfsp, vpp, es, Estream ));
         }

         for( gt = group_targets, i = 0; i < ngroup_targets; gt++, i++ )
         {
            if( ( strlen( lookup ) == gt->gt_namelen ) &&
                ( !strncmp( gt->gt_name, lookup, gt->gt_namelen ) ) )
               return (eventfs_allocvp( vfsp, vpp, efs->un.efs_group,
                                        gt->gt_efstype ));
         }

         break;
      }

      case Estream:
      {
         struct stream_target *st;
         int                   i;

         if( !strcmp( lookup, ".." ) )
            return (eventfs_allocvp( vfsp, vpp,
                                     efs->un.efs_stream->efstream_group,
                                     Egroup ));

         for( st = stream_targets, i = 0; i < nstream_targets; st++, i++ )
         {
            if( ( strlen( lookup ) == st->st_namelen ) &&
                ( !strncmp( st->st_name, lookup, st->st_namelen ) ) )
               return (eventfs_allocvp( vfsp, vpp, efs->un.efs_stream,
                                        st->st_efstype ));
         }

         break;
      }

      default:
         return (ENOTDIR);
   }

   return (ENOENT);
}

static int
eventfs_access( struct vnode *vp, int mode, int flags, struct cred *cred )
{
   struct vattr *vap;
   struct vattr  vattr;
   int           error;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_access\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   /* If super-user, you have access */
   if( suser( cred ) )
      return (0);

   vap = &vattr;
   error = VOP_GETATTR( vp, vap, flags, cred );
   if( error )
      return error;

   /*
    * Access check is based on only one of owner, group, public.
    * If not owner, then check group. If not a member of the
    * group, then check public access.
    */
   if( cred->cr_uid != vap->va_uid )
   {
      gid_t *gp;
      int    i;

      mode >>= 3;
      gp = cred->cr_groups;
      for( i = 0; i < cred->cr_ngroups; i++, gp++ )
         if( vap->va_gid == *gp )
            goto found;
      mode >>= 3;
found:
      ;
   }

   if( ( vap->va_mode & mode ) == mode )
      return (0);

   return (EACCES);
}

static int
eventfs_getattr( struct vnode *vp, struct vattr *vap, int flags,
                 struct cred *cred )
{
   struct eventfs_node *efs = VTOEVENTFS(vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs: eventfs_getattr\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   /* Start by zeroing out the attributes */
   bzero( vap, sizeof( *vap ) );

   /* Next do all the common fields */
   vap->va_type = vp->v_type;
   vap->va_nodeid = efs->efs_nodeid;
   vap->va_uid = 0;
   vap->va_gid = 0;
   vap->va_blksize = PAGESIZE;
   vap->va_size = 0;
   vap->va_atime = vap->va_mtime = vap->va_ctime = tod_get();

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_getattr: vnode 0x%x, type 0x%x, id 0x%x\n", (u_int) vp,
              vap->va_type, efs->efs_nodeid );
#endif /* EVENTFS_DIAGNOSTICS */

   /* Now do the object specific fields */
   switch( efs->efs_type )
   {
      case Eroot:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Eroot)\n" );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD|VEXEC) |
                        (VREAD|VEXEC) >> 3 |
                        (VREAD|VEXEC) >> 6;
         vap->va_nlink = 1;
         vap->va_size = DEV_BSIZE;
         break;

      case Egroup:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Egroup[%s])\n",
                    efs->un.efs_group->efgroup_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD|VEXEC) |
                        (VREAD|VEXEC) >> 3 |
                        (VREAD|VEXEC) >> 6;
         vap->va_nlink = 1 + efs->un.efs_group->efgroup_nstreams;
         vap->va_size = DEV_BSIZE;
         break;

      case Egstats:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Egstats[%s])\n",
                    efs->un.efs_group->efgroup_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD) |
                        (VREAD) >> 3 |
                        (VREAD) >> 6;
         vap->va_nlink = 1;
         break;

      case Estream:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Estream[%s])\n",
                    efs->un.efs_stream->efstream_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD|VEXEC) |
                        (VREAD|VEXEC) >> 3 |
                        (VREAD|VEXEC) >> 6;
         vap->va_nlink = 3;
         vap->va_size = DEV_BSIZE;
         break;

      case Esdata:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Esdata[%s])\n",
                    efs->un.efs_stream->efstream_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD) |
                        (VREAD) >> 3 |
                        (VREAD) >> 6;
         vap->va_nlink = efs->un.efs_stream->efstream_refcount;
         break;

      case Esstats:
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_getattr(Esstats[%s])\n",
                    efs->un.efs_stream->efstream_name );
#endif /* EVENTFS_DIAGNOSTICS */
         vap->va_mode = (VREAD) |
                        (VREAD) >> 3 |
                        (VREAD) >> 6;
         vap->va_nlink = 1;
         break;

      default:
         panic( "eventfs_getattr: invalid type on eventfs node (0x%x)",
                efs->efs_type );
   }

   return 0;
}

static int
eventfs_setattr( struct vnode *vp, struct vattr *vap, int flags,
                 struct cred *cred )
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs: eventfs_setattr\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( vap->va_mask != AT_NOSET )
      return (EOPNOTSUPP);

   return (0);
}

static int
eventfs_setfl( struct vnode *vp, int f, int a, struct cred *cred )
{
   struct eventfs_node *efs = VTOEVENTFS(vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs: eventfs_setfl( 0x%x, 0x%x )\n", f, a );
#endif /* EVENTFS_DIAGNOSTICS */

   if( efs == NULL )
      return (EIO);

   eventfs_setflags( efs, a );

   return (0);
}

static int
eventfs_read( struct vnode *vp, struct uio *uio, int ioflag, struct cred *cred )
{
   int                      rval   = 0;
   struct eventfs_node     *efs    = VTOEVENTFS(vp);

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_READ) )
      printf( "eventfs_read\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   if( ( rval = eventfs_check_stream( efs ) ) == 0 )
   {
      switch( efs->efs_type )
      {
         case Esdata:
         {
            pd_bool_t    eof   = FALSE;
            unsigned int nread = 0;
            int          flags;

refetch:
            flags = eventfs_getflags( efs );
            rval = eventfs_dodata( efs, uio, &nread, &eof );
            /* If blocking allowed, sleep pending data */
            if( eof && ( ( flags & O_NDELAY ) == 0 ) && !nread )
            {
#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_READ) )
                  printf( "eventfs_read: blocking (0x%x)\n", (u_int) nread );
#endif /* EVENTFS_DIAGNOSTICS */

               /* Wait for on condition variable */
               if( !cv_wait_sig( &(efs->efs_cv), &(efs->efs_lock) ) )
                  return (EINTR);

               /* Check for stream teardown */
               if( ( rval = eventfs_check_stream( efs ) ) == 0 )
                  goto refetch;
            }
            /* If non-blocking and no data available, return EAGAIN */
            else if( eof && ( flags & O_NDELAY ) && !nread )
               rval = EAGAIN;

            /* If blocking allowed, sleep pending data */
            break;
         }

         case Esstats:
         case Egstats:
            rval = eventfs_dostats( efs, uio );
            break;

         default:
            rval = EOPNOTSUPP;
            break;
      }
   }

   return (rval);
}

static int
eventfs_readdir( struct vnode *vp, struct uio *uio, struct cred *cred,
                 int *eofp )
{
   unsigned char        d[DIRENTLEN];
#ifdef _LARGEFILE64_SOURCE
   dirent64_t          *dp    = (dirent64_t *)d;
#else /* _LARGEFILE64_SOURCE */
   dirent_t            *dp    = (dirent_t *)d;
#endif /* _LARGEFILE64_SOURCE */
   struct eventfs_node *efs;
   int                  error = 0;
   int                  ecnt  = 0;
   int                  scnt  = 0;
   int                  i;
   int                  off;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs: eventfs_readdir (resid = %d, eofp = 0x%x)\n",
              uio->uio_resid, (int) eofp );
#endif /* EVENTFS_DIAGNOSTICS */

   if( vp->v_type != VDIR )
      return (ENOTDIR);

   off = (int)uio->uio_offset;
   if( ( off != uio->uio_offset ) || ( off < 0 ) ||
       ( ( (u_int)off % DIRENTLEN ) != 0 ) || ( uio->uio_resid < DIRENTLEN ) )
      return (EINVAL);

   i = (u_int)off / DIRENTLEN;

   efs = VTOEVENTFS(vp);
   switch( efs->efs_type )
   {
      case Egroup:
      {
         struct group_target   *gt;
         struct eventfs_stream *es;

#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_readdir( Egroup[%s], %d )\n",
                    efs->un.efs_group->efgroup_name, i );
#endif /* EVENTFS_DIAGNOSTICS */

         for( gt = &group_targets[i]; 
              ( uio->uio_resid >= DIRENTLEN ) && ( i < ngroup_targets );
              gt++, i++ )
         {
            bzero( dp, DIRENTLEN );
            dp->d_reclen = DIRENTLEN;
            dp->d_off = DIRENTLEN;

            switch( gt->gt_efstype )
            {
               case Eroot:
               case Egroup:
                  dp->d_ino = 2;
                  break;

               case Egstats:
                  dp->d_ino = EVENTFS_FILENO(efs->un.efs_group->efgroup_id,
                                             0, 1, Egstats);
                  break;

               default:
                  break;
            }
            bcopy( gt->gt_name, dp->d_name, gt->gt_namelen + 1 );

            error = uiomove( (caddr_t) dp, DIRENTLEN, UIO_READ, uio );
         }

         for( es = efs->un.efs_group->efgroup_streams.lh_first;
              es && ( uio->uio_resid >= DIRENTLEN );
              i++, ecnt++ )
         {
            bzero( dp, DIRENTLEN );
            dp->d_reclen = DIRENTLEN;
            dp->d_off = DIRENTLEN;

            while( ecnt < i )
            {
               ecnt++;
               if( ecnt > ngroup_targets )
                  es = es->efstream_link.le_next;
               if( es == NULL )
                  goto done;
            }

            if( es == NULL )
               goto done;

            dp->d_ino = EVENTFS_FILENO(es->efstream_group->efgroup_id,
                                       es->efstream_id, 0, Estream);
            bcopy( es->efstream_name, dp->d_name, es->efstream_namelen + 1 );
            es = es->efstream_link.le_next;

            error = uiomove( (caddr_t) dp, DIRENTLEN, UIO_READ, uio );
         }
done:
         break;
      }

      case Eroot:
         for( ;
              ( ( i < 2 ) || ( scnt < EVENTFS_NGROUPS ) ) &&
                 ( uio->uio_resid >= DIRENTLEN );
              i++, ecnt++ )
         {
            bzero( dp, DIRENTLEN );
            dp->d_reclen = DIRENTLEN;
            dp->d_off = DIRENTLEN;

#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf( "eventfs_readdir( Eroot, %d, %d )\n", i,
                       (int) uio->uio_offset );
#endif /* EVENTFS_DIAGNOSTICS */
         
            switch( i )
            {
               case 0:      /* `.' */
               case 1:      /* `..' */
                  dp->d_ino = 2;
                  bcopy( "..", dp->d_name, i + 1 );
                  *(dp->d_name + i + 1) = '\0';
                  break;

               default:
                  while( ecnt < i )
                  {
                     if( ecnt >= 2 )
                        scnt++;
                     while( eventfs_groups[scnt] == NULL )
                        if( ++scnt >= EVENTFS_NGROUPS )
                        {
#ifdef EVENTFS_DIAGNOSTICS
                           if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
                              printf( "eventfs_readdir: reached end of groups\n" );
#endif /* EVENTFS_DIAGNOSTICS */

                           goto done;
                        }
                     ecnt++;
                  }

                  if( eventfs_groups[scnt] == NULL )
                  {
#ifdef EVENTFS_DIAGNOSTICS
                     if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
                        printf( "eventfs_readdir: group pointer null\n" );
#endif /* EVENTFS_DIAGNOSTICS */

                     goto done;
                  }

#ifdef EVENTFS_DIAGNOSTICS
                  if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
                     printf( "eventfs_readdir: adding group \"%s\"\n",
                             eventfs_groups[scnt]->efgroup_name );
#endif /* EVENTFS_DIAGNOSTICS */

                  dp->d_ino =
                     EVENTFS_FILENO( eventfs_groups[scnt]->efgroup_id, 0, 0,
                                     Egroup);
                  bcopy( eventfs_groups[scnt]->efgroup_name, dp->d_name,
                         eventfs_groups[scnt]->efgroup_namelen + 1 );
                  scnt++;
                  break;
            }

            error = uiomove( (caddr_t) dp, DIRENTLEN, UIO_READ, uio );
         }
         break;

      case Estream:
      {
         struct stream_target *st;

         for( st = &stream_targets[i]; 
              ( uio->uio_resid >= DIRENTLEN ) && ( i < nstream_targets );
              st++, i++ )
         {
            bzero( dp, DIRENTLEN );
            dp->d_reclen = DIRENTLEN;
            dp->d_off = DIRENTLEN;

#ifdef EVENTFS_DIAGNOSTICS
            if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
               printf( "eventfs_readdir( Estream[%s], %d )\n",
                       efs->un.efs_stream->efstream_name, i );
#endif /* EVENTFS_DIAGNOSTICS */
            switch( st->st_efstype )
            {
               case Estream:
               case Egroup:
                  dp->d_ino = 2;
                  break;

               case Esdata:
                  dp->d_ino =
                  EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                                 efs->un.efs_stream->efstream_id, 1, Esdata);
                  break;

               case Esstats:
                  dp->d_ino =
                  EVENTFS_FILENO(efs->un.efs_stream->efstream_group->efgroup_id,
                                 efs->un.efs_stream->efstream_id, 2, Esstats);
                  break;

               default:
                  break;
            }
            bcopy( st->st_name, dp->d_name, st->st_namelen + 1 );

            error = uiomove( (caddr_t) dp, DIRENTLEN, UIO_READ, uio );
         }

         break;
      }

      default:
         return (ENOTDIR);
   }

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs: eventfs_readdir: returning %d\n", error );
#endif /* EVENTFS_DIAGNOSTICS */

   if( error == 0 )
      if( eofp )
      {
         *eofp = ( uio->uio_resid >= DIRENTLEN ) ? 1 : 0;
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs: eventfs_readdir: eof set to %d\n", *eofp );
#endif /* EVENTFS_DIAGNOSTICS */
      }

   return error;
}

static void
eventfs_inactive( struct vnode *vp, struct cred *cred )
{
   struct eventfs_node  *efs   = VTOEVENTFS(vp);
   struct eventfs_node **efspp;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_inactive\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   for( efspp = &efshead; *efspp; efspp = &(*efspp)->efs_next )
   {
      if( *efspp == efs )
      {
#ifdef EVENTFS_DIAGNOSTICS
         if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
            printf( "eventfs_inactive: reclaiming vnode\n" );
#endif /* EVENTFS_DIAGNOSTICS */
         *efspp = efs->efs_next;
         break;
      }
   }

   if( vp->v_data )
   {
      mutex_destroy( &efs->efs_lock );
      cv_destroy( &efs->efs_cv );

      kmem_free( vp->v_data, sizeof(struct eventfs_node) );
      vp->v_data = NULL;
   }
}

static int
eventfs_ioctl( struct vnode *vp, int command, intptr_t data, int flag, 
               struct cred *cred, int *result )
{
   int error = ENOTTY;

#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
      printf( "eventfs_ioctl\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   switch( command )
   {
      case EFSIOCREQ:
      {
         struct eventfs_node *efs = VTOEVENTFS(vp);

         if( efs->efs_type == Egstats )
         {
            struct eventfs_request *req  = (struct eventfs_request *)data;
            char                   *name = NULL;
            unsigned long           mask;

            if( copyin( (caddr_t) &(req->mask), (caddr_t) &mask, sizeof(mask) ) )
               error = EFAULT;
            else
               error = eventfs_construct_stream( efs->un.efs_group->efgroup_name,
                                                 mask, &name );
            if( !error )
            {
               pd_byte_t  terminator = 0;
               int        len        = strlen( name );
#ifdef EVENTFS_DIAGNOSTICS
               if( eventfs_debug_mask & EFS_DEBUG_MD_VNOPS )
                  printf( "eventfs_ioctl: name '%s'\n", name );
#endif /* EVENTFS_DIAGNOSTICS */
               if( copyout( (caddr_t) name, (caddr_t) req->name, len ) ||
                   copyout( (caddr_t) &terminator,
                            (caddr_t) ( (pd_byte_t *) req->name + len ), 1 ) )
                  error = EFAULT;
               EFS_FREE( name );
            }
         }
         break;
      }

      default:
   }

   if( result )
      *result = error;

   return (error);
}



/*
 * EventFS "should never get here" operation
 */
static int
eventfs_badop()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_badop\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_addmap()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_addmap [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_create()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_create [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_delmap()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_delmap [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_dispose()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_dispose [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_fid()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_fid [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_fsync()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_fsync [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_getsecattr()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_getsecattr [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_link()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_link [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_map()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_map [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_mkdir()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_mkdir [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_poll()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_poll [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_realvp()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_realvp [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_remove()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_remove [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_rename()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_rename [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_rmdir()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_rmdir [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_seek()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_seek [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_setsecattr()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_setsecattr [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_space()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_space [BAD]\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

static int
eventfs_write()
{
#ifdef EVENTFS_DIAGNOSTICS
   if( eventfs_debug_mask & (EFS_DEBUG_MD_VNOPS|EFS_DEBUG_MD_BADOPS) )
      printf( "eventfs_write\n" );
#endif /* EVENTFS_DIAGNOSTICS */

   return (EIO);
}

