/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: sysctl.c,v 1.5 2001/02/23 19:02:22 stevek Exp $
 */

#include <stdio.h>
#include <strings.h>
#include <kvm.h>
#include <errno.h>
#include <fcntl.h>

typedef enum
{
   MIB_INT,
   MIB_UINT
}
mib_type_t;

static struct mib_entry
{
   const char *mib_name;
   mib_type_t  mib_type;
   const char *mib_symbol;
}
mib_map[] =
{
   { "vfs.eventfs.debugmask", MIB_UINT, "eventfs_debug_mask" }
};

#define NUM_MIBS (sizeof(mib_map) / sizeof(struct mib_entry))

struct mib_write_request
{
   struct mib_entry *mib;
   union
   {
      int   ival;
      u_int uival;
   }
   un;
   struct mib_write_request *next;
}
*mib_write_requests = NULL;

static boolean_t   f_noname = B_FALSE;
static int         nentries = 0;

static const char *progname;


/* Forward declarations */
static int               queue_write_request( const char * );
static struct mib_entry *find_mib           ( const char * );


int
main( int argc, char *argv[] )
{
   int                       ch, i, error;
   kvm_t                    *kd;
   struct nlist             *nl, *cnl;
   struct mib_write_request *req;

   progname = (const char *)argv[0];

   while( ( ch = getopt( argc, argv, "nw:" ) ) != -1 )
   {
      switch( ch )
      {
         case 'w':
            error = queue_write_request(optarg);
            if( error ) return (error);
            break;

         case 'n':
            f_noname = B_TRUE;
            break;

         default:
            fprintf( stderr, "%s: Unknown option '%c'\n", argv[0], ch );
            return EINVAL;
      }
   }

   argc -= optind;
   argv += optind;

   if( (nentries+argc) <= 0 )
      return 0;


   /* Validate the read requests */
   for( i = 0; i < argc; i++ )
      if( find_mib( argv[i] ) == NULL )
      {
         fprintf( stderr, "%s: Invalid MIB '%s'\n", progname, argv[i] );
         return ENOENT;
      }


   /* Open the kernel descriptor */
   kd = kvm_open( NULL, NULL, NULL,
                  ((mib_write_requests == NULL)?O_RDONLY:O_RDWR),
                  (char *)progname );
   if( kd == NULL ) return EPERM;


   /* Build up the nlist */
   cnl = nl = (struct nlist *)malloc(sizeof(*nl) * (nentries+argc+1));
   for( req = mib_write_requests; req; req = req->next )
      (cnl++)->n_name = (char *)req->mib->mib_symbol;

   for( i = 0; i < argc; i++ )
   {
      struct mib_entry *mib = find_mib( argv[i] );
      (cnl++)->n_name = (char *)mib->mib_symbol;
   }
   cnl->n_name = NULL;


   /* Complete the nlist */
   if( kvm_nlist( kd, nl ) < 0 )
   {
      fprintf( stderr, "%s: %s\n", progname, strerror(errno) );
      return errno;
   }


   /* Process write requests */
   cnl = nl;
   while( mib_write_requests )
   {
      req = mib_write_requests;
      mib_write_requests = req->next;

      if( cnl->n_type == 0 )
      {
         fprintf( stderr, "%s: Unable to find '%s' in the kernel\n", progname,
                  req->mib->mib_name );
      }
      else
      {
         int retc = 0;
         switch( req->mib->mib_type )
         {
            case MIB_INT:
               retc = kvm_kwrite( kd, cnl->n_value, (char *)&(req->un.ival),
                                  sizeof(req->un.ival));
               break;

            case MIB_UINT:
               retc = kvm_kwrite( kd, cnl->n_value, (char *)&(req->un.uival),
                                  sizeof(req->un.uival));
               break;

            default:
               fprintf( stderr,
                        "%s: Internal error, unknown MIB type encountered\n",
                        progname );
               break;
         }

         if( retc < 0 )
            fprintf( stderr, "%s: '%s' - %s\n", progname, req->mib->mib_name,
                     strerror(error) );
      }

      free( req );

      cnl++;
   }


   /* Process read requests */
   for( i = 0; i < argc; i++ )
   {
      struct mib_entry *mib = find_mib( argv[i] );

      if( cnl->n_type == 0 )
      {
         fprintf( stderr, "%s: Unable to find '%s' in the kernel\n", progname,
                  mib->mib_name );
      }
      else
      {
         int retc = 0;
         switch( mib->mib_type )
         {
            case MIB_INT:
            {
               int ival;
               retc = kvm_kread( kd, cnl->n_value, (char *)&ival,
                                 sizeof(ival) );
               if( retc >= 0 )
               {
                  if( f_noname )
                     printf( "%d\n", ival );
                  else
                     printf( "%s: %d\n", mib->mib_name, ival );
               }
               break;
            }

            case MIB_UINT:
            {
               u_int uival;
               retc = kvm_kread( kd, cnl->n_value, (char *)&uival,
                                 sizeof(uival) );
               if( retc >= 0 )
               {
                  if( f_noname )
                     printf( "%u\n", uival );
                  else
                     printf( "%s: %u\n", mib->mib_name, uival );
               }
               break;
            }

            default:
               fprintf( stderr,
                        "%s: Internal error, unknown MIB type encountered\n",
                        progname );
               break;
         }

         if( retc < 0 )
            fprintf( stderr, "%s: '%s' - %s\n", progname, mib->mib_name,
                     strerror(error) );
      }

      cnl++;
   }

   return kvm_close( kd );
}

static struct mib_entry *
find_mib( const char *mibname )
{
   struct mib_entry *mib = NULL;
   int               i;

   for( i = 0; i < NUM_MIBS; i++ )
      if( !strcmp( mibname, mib_map[i].mib_name ) )
      {
         mib = mib_map + i;
         break;
      }

   return (mib);
}

static int
queue_write_request( const char *arg )
{
   struct mib_entry *mib = NULL;
   char             *eq  = strchr(arg, '=');
   int               i;

   if( eq == NULL )
   {
      fprintf( stderr, "%s: Missing '=' for write flag\n", progname );
      return EINVAL;
   }

   *eq++ = '\0';
   mib = find_mib( arg );
   if( !mib )
   {
      fprintf( stderr, "%s: Invalid MIB '%s'\n", progname, arg );
      return ENOENT;
   }
   else
   {
      struct mib_write_request *req;

      req = (struct mib_write_request *)malloc(sizeof(*req));
      req->mib = mib;

      switch( mib->mib_type )
      {
         case MIB_INT:
            if( *eq == '\0' )
            {
               fprintf( stderr, "%s: Missing integer value\n", progname );
               return EINVAL;
            }
            req->un.ival = strtol( eq, (char **)NULL, 10 );
            break;

         case MIB_UINT:
            if( *eq == '\0' )
            {
               fprintf( stderr, "%s: Missing unsigned integer value\n", progname );
               return EINVAL;
            }
            req->un.uival = strtoul( eq, (char **)NULL, 10 );
            break;

         default:
            fprintf( stderr,
                     "%s: Internal error, unknown MIB type encountered\n",
                     progname );
            return EFAULT;
            break;
      }

      req->next = mib_write_requests;
      mib_write_requests = req;
      nentries++;
   }

   return 0;
}

