/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: GridCell.java,v 1.8 2001/04/05 18:41:10 stevek Exp $
 */
package com.nailabs.ui;


/**
 * Class GridCell represents a cell in the Grid.
 * 
 */
import java.awt.*;

public class GridCell 
{
	protected Color         bgColor       = Color.white;
	protected Color         textColor     = Color.black;

	protected Color         hlBgColor     = Color.blue;
	protected Color         hlTextColor   = Color.white;

	protected Font          font          = new Font( "Dialog", Font.PLAIN, 12 );
	
	protected boolean       highlight     = false;
	private   Component     component;
	private   String        value;
	private   boolean       redraw        = true;
	private   boolean       editable      = true;
	private   boolean       resizable     = true;
	private   int           xSpan         = 1;
	private   int           ySpan         = 1;
	
	public static final int SPAN_X        = 0;
	public static final int SPAN_Y        = 1;

	public static final int BG_COLOR      = 0;
	public static final int TEXT_COLOR    = 1;
	public static final int HL_BG_COLOR   = 2;
	public static final int HL_TEXT_COLOR = 3;
	public static final int BORDER_COLOR  = 4;


	/**
	 * This method can be called if a component is displayed
	 * in the cell rather than the text.
	 * @param rect java.awt.Rectangle is area in which the component
	 * should be displayed.
	 */
	public void draw( Rectangle rect )
	{
		if( component != null )
			component.setBounds( rect.x,rect.y,rect.width,rect.height );
	}

	/**
	 * Call this method to draw the contents of the cell in the
	 * rectangle rect.
	 * @param rect java.awt.Rectangle 
	 * @param g java.awt.Graphics
	 */
	public void draw( Rectangle rect, Graphics g )
	{
		if( redraw )
		{
			//redraw only if content changed.

			FontMetrics fm ; //used for calculating size of the string
			int x;
			int y;

			g.setFont(font); //set font for this cell

			fm = g.getFontMetrics();
			//compute client area
			x = rect.x + 2;
			y = rect.y + ( rect.height + fm.getHeight() ) / 2 - fm.getDescent();
			//if cell to be highlighted, change colors.

			if( highlight )
			{
				//draw background
				g.setColor(hlBgColor);
				g.fillRect(rect.x,rect.y,rect.width,rect.height);
				//draw content 
				if( value != null )
				{
					g.setColor(hlTextColor);
					g.drawString(value,x,y);
				}
			}
			else
			{
				//draw back ground
				g.setColor(bgColor);
				g.fillRect(rect.x,rect.y,rect.width,rect.height);
				//draw content
				if( value != null )
				{
					g.setColor(textColor);
					g.drawString(value,x,y);
				}
			}

			redraw = false;
		}
	}

	/**
	 * Method getComponent returns the component attached with this
	 * cell.
	 * @return java.lang.String
	 */
	public Component getComponent() { return component; }

	/**
	 * method getSpan() returns the the span amount. i.e. this is not used
	 * in the Grid Class. Yet to be implemented
	 * @return int
	 * @param direction int
	 */
	public int getSpan( int direction )
	{
		if( direction == SPAN_X )
			return xSpan; //return how many hor. cell units.
		else if( direction == SPAN_Y )
			return ySpan; //return how many vertical cell units.
		else
			return 0;
	}

	/**
	 * method getValue returns value of the content of this cell.
	 * @return java.lang.String
	 */
	public String getValue() { return value; }

	/**
	 * method getWidth() returns width of the cell.
	 * This method is not implemented.
	 * @return int
	 * @param g java.awt.Graphics
	 */
	public int getWidth( Graphics g ) { return 0; }

	/**
	 * method highlight() set the highlight property of this cell.
	 * @param flag boolean . If flag is true, cell is highlighted Else
	 * cell is not highlighted.
	 */
	public void highlight( boolean flag ) { highlight = flag; }

	/**
	 * Call this method to set this cell for redraw.
	 * @param flag boolean
	 * 
	 */
	public void invalidate( boolean flag ) { redraw = flag; }

	/**
	 * method isEditable() returns the status of the cell whether
	 * editable or not.
	 * @return boolean
	 */
	public boolean isEditable() { return editable; }

	/**
	 * method isHighlighted() returns the status of the cell whether
	 * highlighted or not.
	 * @return boolean
	 */
	public boolean isHighlighted() { return highlight; }

	/**
	 * method isResizable() returns the status of the cell whether
	 * resizable or not.
	 * @return boolean
	 */
	public boolean isResizable() { return resizable; }

	/**
	 * Resets the value of the component.
	 */
	public void resetValue()
	{
		if( component instanceof TextField )
			((TextField)component).setText( this.value );
		else if( component instanceof Choice )
			((Choice)component).select( this.value );
		redraw = true;
	}

	/**
	 * method setColor sets the different color attributes of the cell.
	 * @param type	 int  the color type.
	 * @param color java.awt.Color 
	 *
	 */
	public void setColor( int type, Color color )
	{
		switch( type )
		{
			case Grid.TEXT_COLOR:
				textColor = color;
				break;
			case Grid.HL_TEXT_COLOR:
				hlTextColor = color;
				break;
			case Grid.BG_COLOR:
				bgColor = color;
				break;
			case Grid.HL_BG_COLOR:
				hlBgColor = color;
				break;
		}
	}

	/**
	 * Sets the component for this cell
	 * @param comp java.awt.Component is the new component for this cell.
	 */
	public void setComponent( Component comp ) { component = (Component)comp; }

	/**
	 * Makes the cell editable or not according to the value of flag.
	 * @param flag boolean if true, cell is editable or not.
	 */
	public void setEditable( boolean flag ) { editable = flag; }

	/**
	 * Sets font for the cell.
	 * @param font java.awt.Font is the new font for the cell.
	 */
	public void setFont( Font font ) { this.font = font; }

	/**
	 * Makes the cell resizable or not according to the value of flag.
	 * @param flag boolean if true, cell is resizable or not.
	 */
	public void setResizable( boolean flag ) { resizable = flag; }

	/**
	 * Sets the different span attributes for the cell. This method is
	 * not used in Grid. But you can use this method to extend the
	 * functionality of Grid.
	 * @param direction int specifies whether span is horizontal or vertical.
	 * values can be either SPAN_X or SPAN_Y.
	 * @param value int is the now of span units.
	 */
	public void setSpan( int direction, int value )
	{
		if( direction == SPAN_X )
			xSpan = value;
		else if( direction == SPAN_Y )
			ySpan = value;
	}

	/**
	 * Sets the value of the cell.
	 * @param value java.lang.String is the new value.
	 */
	public void setValue( String value )
	{
		this.value = value;

		if( component instanceof TextField )
			((TextField)component).setText( value );
		else if( component instanceof Choice )
			((Choice)component).select( value );
		redraw = true;
	}
}

