/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * $Id: GridRow.java,v 1.8 2001/04/05 18:41:10 stevek Exp $
 */
package com.nailabs.ui;


import java.awt.*;
import java.util.Vector;

public class GridRow
{
	protected Vector      contents;
	protected Font        font                = new Font( "Dialog", Font.PLAIN, 12 );
	protected FontMetrics fm;
	protected Rectangle   clipRect;
	protected int         x;
	protected int         y;
	protected int         rowWidth;
	protected int         rowHeight;
	protected int[]       xPos;
	protected int[]       colWidths;
	protected int         yPos                = 0;
	protected int         columnCount         = 0;
	protected boolean     widthFlag           = false;
	protected boolean     marginFlag          = true;
	protected boolean     showHorBorder       = true;
	protected boolean     showVerticalBorder  = true;
	protected boolean     highlight           = false;
	protected boolean     redraw              = true;
	protected Color       textColor           = Color.black;
	protected Color       bgColor             = Color.white;
	protected Color       hlTextColor         = Color.white;
	protected Color       hlBgColor           = Color.blue;
	protected Color       headerColor1        = Color.black;
	protected Color       headerColor2        = Color.gray;
	protected Color       headerColor3        = Color.white;
	protected Color       headerColor4        = Color.lightGray;
	protected Object      userData;

	public static final int NO_BORDER         = 0;
	public static final int	HORIZONTAL_BORDER = 1;
	public static final int	VERTICAL_BORDER   = 2;
	public static final int	BOTH_BORDER       = 3;



	/**
	 * GridRow Constructor
	 * @param cells java.util.Vector  is the content of this row
	 */
	public GridRow( Vector cells )
	{
		if( cells != null )
		{
			columnCount = cells.size();
			xPos = new int[100];
			contents = cells;
		}
	}

	/**
	 * GridRow constructor
	 * @param cells Vector is the content of row
	 * @param borderFlag int 
	 * @param marginFlag boolean
	 */
	public GridRow( Vector cells, int borderFlag, boolean marginFlag )
	{
		this( cells );
		this.marginFlag = marginFlag;
		switch( borderFlag )
		{
			case HORIZONTAL_BORDER:
				showHorBorder = true;
				showVerticalBorder = false;
				break;
			case VERTICAL_BORDER:
				showHorBorder = false;
				showVerticalBorder = true;
				break;
			case BOTH_BORDER:
				showHorBorder = true;
				showVerticalBorder = true;
				break;
			case NO_BORDER:
				showHorBorder = false;
				showVerticalBorder = false;
				break;
		}
	}

	/**
	 * add a new cell to the Row
	 * @param cell Grid.GridCell
	 */
	public void addCell( GridCell cell )
	{
		setProperties( cell );
		if( cell != null )
		{
			contents.addElement( cell );
			widthFlag = false;
		}
	}

	/**
	 * This method was created by a SmartGuide.
	 */
	protected void adjustSize( Object[] values )
	{
		int len = values.length;
		if( len > columnCount )
		{
			widthFlag = false;
			for( int i = columnCount; i < len; i++ )
			{
				GridCell cell = new GridCell();
				contents.addElement( cell );
			}
			columnCount = len;
		}
	}

	/**
	 * computes the start x position of each cell in this row
	 */
	protected void computeXpos()
	{
		int startX = x;
		for( int i=0; i <= columnCount + 1; i++ )
		{
			xPos[i] = startX;
			if( marginFlag || ( i > 0 ) )
				startX += colWidths[i];
		}
	}

	/**
	 * deletes cell at position index.
	 * @param cell Grid.GridCell
	 */
	public void deleteCell( int index )
	{
		if( index < columnCount )
		{
			contents.removeElementAt( index );
			widthFlag = false;
		}
	}

	/**
	 * draw the content of this row inside the Rectangle rect
	 * @param rect java.awt.Rectangle
	 * @param g java.awt.Graphics
	 */
	public void draw( Graphics g )
	{
		if( redraw && ( g != null ) && widthFlag )
		{
			g.setFont(font);

			Graphics cg;
			GridCell cell;
			int      charX,charY;
			int      startX,endX;

			clipRect = g.getClipBounds();
			rowHeight = clipRect.height;
			rowWidth = clipRect.width;

			x = clipRect.x;
			y = clipRect.y;

			computeXpos();
			drawMargin( g );

			if( fm == null )
				fm = g.getFontMetrics();

			for( int index = 1; index <= columnCount + 1; index++ )
			{
				startX = xPos[index] + 1;
				endX = xPos[index] + colWidths[index] - 1;

				if( showVerticalBorder )
				{
					g.setColor( headerColor2 );
					g.drawLine( xPos[index], y, xPos[index], y + rowHeight );
				}

				// do rest of the things here
				cell = (GridCell)contents.elementAt( index );
				if( cell.isHighlighted() == true )
					g.setColor( hlTextColor );
				else
					g.setColor( textColor );
				cell.draw( new Rectangle( startX, y, colWidths[index] - 1, rowHeight ), g );
			}
			drawBorder( g );
			redraw = false;
		}
	}

	/**
	 * draw border of this row
	 * @param g java.awt.Graphics
	 */
	protected void drawBorder( Graphics g )
	{
		int x1 = x + rowWidth;
		int y1 = y + rowHeight;

		//draw the outer border
		g.setColor( headerColor2 );		//gray
		if( showHorBorder )
		{
			g.drawLine( x, y, x1, y );	//upper line
		}

		if( showVerticalBorder )
		{
			g.drawLine( x, y, x, y1 );	//left line
			g.drawLine( x1, y, x1, y1 );	//right line
		}
	}

	/**
	 * draw margin of this row
	 * @param g java.awt.Graphics
	 */
	protected void drawMargin( Graphics g )
	{
		if( marginFlag )
		{
			int x0 	= x + 1;
			int y0 	= y + 1;
			int x1	= xPos[1]-1;
			int y1	= y + rowHeight - 1;

			g.setColor( headerColor4 );
			g.fillRect( x0, y0, x1, y1 );

			g.setColor( headerColor3 );
			g.drawLine( x0, y0, x1, y0 );
			g.drawLine( x0, y0, x0, y1 );

			g.setColor( headerColor2 );
			g.drawLine( x1, y0, x1, y1 );
			g.drawLine( x0, y1, x1, y1 );
		}

		if( highlight )
		{
			g.setColor( hlBgColor );
			g.fillRect( xPos[1], y, rowWidth, rowHeight );
		}
		else if( redraw )
		{
			g.setColor( Color.white );
			g.fillRect( xPos[1], y, rowWidth, rowHeight );
		}
	}

	/**
	 * get the cell at position index.
	 * @return java.lang.String
	 */
	public GridCell getCell( int index )
	{
		if( index <= columnCount )
		{
			GridCell cell = (GridCell)contents.elementAt( index );
			return cell;
		}
		else
			return null;
	}

	/**
	 * get the user data attached to this row
	 * @return java.lang.Object
	 */
	public Object getData() { return userData; }

	/**
	 * get the font for this row
	 * @return java.awt.Font
	 */
	public Font getFont() { return font; }

	/**
	 * get the text color of this row
	 * @return java.awt.Color
	 */
	public Color getTextColor() { return textColor; }

	/**
	 * enable highlighting.
	 * @param flag boolean  highlight the row if flag is true.
	 */
	public void highlight( boolean flag )
	{
		GridCell cell;
		highlight = flag;
		if( contents != null )
		{
			for( int i = 0; i < contents.size(); i++ )
			{
				cell = (GridCell)contents.elementAt( i );
				cell.highlight( flag );	
			}
		}
	}

	/**
	 * inserts a new cell at position int
	 * @param index int
	 * @param cell com.widget.grid.GridCell
	 */
	public void insertCell( int index, GridCell cell )
	{
		if( ( index <= columnCount ) && ( index >= 0 ) )
			contents.insertElementAt( cell, index );
	}

	/**
	 * marks this row for redraw
	 * @param flag boolean
	 */
	public void invalidate( boolean flag )
	{
		GridCell cell;
		redraw = flag;
		if( contents != null )
		{
			for( int i = 0; i < contents.size(); i++ )
			{
				cell = (GridCell)contents.elementAt( i );
				cell.invalidate( flag );
			}
		}
	}

	/**
	 *
	 */
	public void setBorder( int borderFlag )
	{
		switch( borderFlag )
		{
			case HORIZONTAL_BORDER:
				showHorBorder = true;
				showVerticalBorder = false;
				break;
			case VERTICAL_BORDER:
				showHorBorder = false;
				showVerticalBorder = true;
				break;
			case BOTH_BORDER:
				showHorBorder = true;
				showVerticalBorder = true;
				break;
			case NO_BORDER:
				showHorBorder = false;
				showVerticalBorder = false;
				break;
		}
	}

	/**
	 * replace cell at position index with cell
	 * @param index int is the position of the cel
	 * @param cell  GridCell is the new cell
	 */
	public void setCell( int index, GridCell cell )
	{
		if( index <= columnCount )
			contents.setElementAt( cell, index );
	}

	/**
	 * set color of this row
	 * @param int type is the color type
	 * @param color java.awt.Color
	 */
	public void setColor( int type, Color color )
	{
		switch( type )
		{
			case Grid.TEXT_COLOR:
				textColor = color;
				break;
			case Grid.HL_TEXT_COLOR:
				hlTextColor = color;
				break;
			case Grid.BG_COLOR:
				bgColor = color;
				break;
			case Grid.HL_BG_COLOR:
				hlBgColor = color;
				break;
		}
	}

	/**
	 * sets the no. of cells and width of each cell in this row.
	 * @param colWidths int[] width of cells
	 * @param columnCount int is the number of cells
	 */
	public void setColWidths( int[] colWidths, int columnCount )
	{
		if( ( colWidths != null ) && ( columnCount != 0 ) )
		{
			this.colWidths = colWidths;
			this.columnCount = columnCount;
			widthFlag = true;
		}
	}

	/**
	 * attaches a user data object to this row
	 * @param data java.lang.Object is the user data object
	 */
	public synchronized void setData( Object data ) { userData = data; }

	/**
	 * sets font of this row
	 * @param font java.awt.Font is the new font
	 */
	public void setFont( Font font )
	{
		GridCell cell;

		for( int i = 1; i < columnCount; i++ )
		{
			cell	= (GridCell)contents.elementAt( i );
			cell.setFont( font );
		}
	}

	/**
	 * @param boolean flag if true, draw border, otherwise don't
	 */
	public void setMargin( boolean flag ) { marginFlag = flag; }

	/**
	 * set properties of this row
	 */
	private void setProperties( GridCell cell )
	{
		cell.setColor( Grid.TEXT_COLOR, textColor );
		cell.setColor( Grid.HL_TEXT_COLOR, hlTextColor );
		cell.setColor( Grid.BG_COLOR, bgColor );
		cell.setColor( Grid.HL_BG_COLOR, hlBgColor );

		cell.setFont( font );
	}
}

