package BinFile::COFF::Section;
# 
# Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
# All rights reserved.
#
# This file is part of the NAI Generic Software Wrappers Toolkit
# (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
# distribution.
#
# The GSWTK is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The GSWTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: Section.pm,v 1.7 2001/04/05 18:41:12 stevek Exp $

require 5.004;
use strict;
use Carp;
use IO::Seekable;
use Data::Dumper;

BEGIN {
   use Exporter    ();
   use vars        qw(@EXPORT $VERSION $SECTION_SIZE);
   @EXPORT       = qw(loadCoffSection);
   $VERSION      = 0.001;
   $SECTION_SIZE = 40;
}

sub import { goto &Exporter::import }

use Class::Struct qw(struct);
struct 'BinFile::COFF::Section' => [
   name       => '$', # 8 bytes (unsigned char [8])  offset 0
   virtsize   => '$', # 4 bytes (unsigned long)      offset 8
   virtaddr   => '$', # 4 bytes (unsigned long)      offset 12
   rawsize    => '$', # 4 bytes (unsigned long)      offset 16
   rawoffset  => '$', # 4 bytes (unsigned long)      offset 20
   relocs     => '$', # 4 bytes (unsigned long)      offset 24
   lines      => '$', # 4 bytes (unsigned long)      offset 28
   nrelocs    => '$', # 2 bytes (unsigned short)     offset 32
   nlines     => '$', # 2 bytes (unsigned short)     offset 34
   charistics => '$'  # 4 bytes (unsigned long)      offset 36
];

my %section_characteristics = (
   0x00000020 => 'Text',
   0x00000040 => 'Data',
   0x00000080 => 'BSS',
   0x00000200 => 'Info',
   0x00000800 => 'Remove',
   0x00001000 => 'Comdat',
   0x00004000 => 'No Defer Speculative Exceptions',
   0x00008000 => 'Far Data',
   0x00020000 => 'Purgeable (16-bit)',
   0x00040000 => 'Locked',
   0x00080000 => 'Preload',
#  0x00100000 -> 0x00E00000 are alignment characteristics
   0x01000000 => 'Extended Reloc',
   0x02000000 => 'Discardable',
   0x04000000 => 'No Cache',
   0x08000000 => 'No Page',
   0x10000000 => 'Shared',
   0x20000000 => 'Executable',
   0x40000000 => 'Readable',
   0x80000000 => 'Writable'
);

sub loadCoffSection ($) {
   my $fd = shift;

   croak "Invalid argument" unless ref($fd);

   my $section;
   my $total = 0;
   while( $total < $SECTION_SIZE ) {
      my $nread = sysread( $fd, $section, $SECTION_SIZE, $total );
      croak "Unable to read COFF section: $!" unless $nread;
      $total += $nread;
   }
   croak "Unable to read COFF section: $!" if( $total < $SECTION_SIZE );

   my @parts = unpack('C8 VVVVVVvvV', $section);

   my $sh = new();
   $sh->[0]               = pack('C8', @parts[ (0..7) ]);
   ( $sh->[1], $sh->[2],
     $sh->[3], $sh->[4],
     $sh->[5], $sh->[6] ) = @parts[ (8..13) ];
   ( $sh->[7], $sh->[8] ) = @parts[ (14..15) ];
   $sh->[9]               = $parts[16];
   $sh;
}

sub dump ($) {
   my $this = shift;

   printf("[COFF Section]\n");
   printf("      name: %s\n", $this->name);
   printf("  virtsize: %8.8x\n", $this->virtsize);
   printf("  virtaddr: %8.8x\n", $this->virtaddr);
   printf("   rawsize: %8.8x\n", $this->rawsize);
   printf(" rawoffset: %8.8x\n", $this->rawoffset);
   printf("    relocs: %8.8x\n", $this->relocs);
   printf("     lines: %8.8x\n", $this->lines);
   printf("   nrelocs:     %4.4x\n", $this->nrelocs);
   printf("    nlines:     %4.4x\n", $this->nlines);
   printf("charistics: %8.8x", $this->charistics);

   my $ch = $this->charistics;
   my $nshown = 0;
   foreach( sort keys %section_characteristics ) {
      printf("%s%s%s",
             ($nshown % 3)?", "
                          :($nshown?"                       ":"   "),
             $section_characteristics{$_},
             (($nshown++ % 3) == 2)?"\n":"")
         if( $ch & $_ );
   }
   printf("\n") if( !$nshown || ($nshown % 3) );
   printf("                       Align %d-byte%s\n",
          2^(($ch & 0x00f00000)>>40), ($ch & 0x00100000)?"":"s")
      if( $ch & 0x00f00000 );
}

sub size { return $SECTION_SIZE; }

1;
