package BinFile::ELF;
# 
# Copyright (C) 2001  Network Associates, Incorporated (NAI)
# All rights reserved.
#
# This file is part of the NAI Generic Software Wrappers Toolkit
# (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
# distribution.
#
# The GSWTK is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The GSWTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: ELF.pm,v 1.13 2001/01/19 18:30:23 stevek Exp $

require 5.004;
use strict;
use vars qw($VERSION @ISA);
use Carp;
use BinFile;
use BinFile::ELF::Header;
use BinFile::ELF::Section;
use BinFile::ELF::StringTable;
use BinFile::ELF::SymbolTable;
use IO::Seekable;

$VERSION = 0.001;
@ISA     = qw(BinFile);

sub new {
   my $this = shift;
   my $class = ref($this) || $this;
   my $file = shift;
   my $options = @_?shift:0;
   my $self = {};
   bless $self, $class;
   $self->_initialize($file,$options);
}

sub _initialize {
   my $self = shift;
   my $file = shift;
   my $options = shift;

   # Initialize BinFile
   $self->SUPER::_initialize( $file, $options );

   my $fd;
   *fd = $self->{'Descriptor'};

   # read ELF header
   my $elfHeader   = loadElfHeader( \*fd );
   $self->{'ELF Header'}   = $elfHeader;

   # read ELF sections
   $self->{'ELF Sections'} = [];
   if( $elfHeader->shnum > 0 )
   {
      my $result = eval 'use '.$self->sectionclass;
      carp $@ if $@;
      sysseek( *fd, $elfHeader->shoff, SEEK_SET );
      foreach( 0 .. ($elfHeader->shnum - 1) ) {
         my $section = $self->sectionclass->new;
         $section->load($_,$self);
      }
   }

   # read ELF string tables
   $self->{'ELF String Tables'} = {};
   if( ($self->{'Options'} & BINFILE_LOADSTRINGS) && $elfHeader->shnum > 0 ) {
      my @sections = $self->elfsections;
      foreach( 0 .. ($elfHeader->shnum - 1) ) {
         my $section = $sections[$_];
         next unless $section->type == SHT_STRTAB;
         my $strtab = $self->strtabclass->new;
         $strtab->load($_,$self);
      }
   }
   $self->strtabclass->new->load($elfHeader->shstrndx,$self)
      unless $self->hdrstr;

   # read ELF symbol tables
   $self->{'ELF Symbol Tables'} = {};
   if( ($self->{'Options'} & BINFILE_LOADSYMS) && $elfHeader->shnum > 0 ) {
      my @sections = $self->elfsections;
      foreach( 0 .. ($elfHeader->shnum - 1) ) {
         my $section = $sections[$_];
         next unless $section->type == SHT_SYMTAB;
         my $symtab = $self->symtabclass->new;
         $symtab->load($_,$self);
      }
   }

   $self;
}

sub descriptor ($) {
   my $this = shift;
   $this->{'Descriptor'};
}

sub sectionclass ($) {
   my $this = shift;
   my $formatclass = $this->formatclass;
   return $formatclass.'::Section' if $formatclass;
   undef;
}

sub symtabclass ($) {
   my $this = shift;
   ref($this).'::SymbolTable';
}

sub strtabclass ($) {
   my $this = shift;
   ref($this).'::StringTable';
}

sub formatclass ($) {
   my $this = shift;
   return ref($this)."::ELF32" if $this->elfheader->class == ELFCLASS32;
   return ref($this)."::ELF64" if $this->elfheader->class == ELFCLASS64;
   undef;
}

sub format ($) {
   my $this = shift;
   return "ELF-32" if $this->elfheader->class == ELFCLASS32;
   return "ELF-64" if $this->elfheader->class == ELFCLASS64;
   "ELF-Unknown";
}

sub elfheader ($) {
   my $this = shift;
   $this->{'ELF Header'};
}

sub elfsections ($) {
   my $this = shift;
   my $sections = $this->{'ELF Sections'};
   ref($sections)?@$sections:$sections;
}

sub elfstrtabs ($) {
   my $this = shift;
   my $strtabs = $this->{'ELF String Tables'};
   $strtabs;
}

sub elfsymtabs ($) {
   my $this = shift;
   my $symtabs = $this->{'ELF Symbol Tables'};
   $symtabs;
}

sub addsection ($$) {
   my $this = shift;
   my $section = shift;
   croak 'Not an ELF section' unless $section->isa('BinFile::ELF::Section');

   push(@{$this->{'ELF Sections'}}, $section);
   $section;
}

sub addstrtab ($$) {
   my $this = shift;
   my $strtab = shift;
   croak 'Not an ELF string table'
      unless $strtab->isa('BinFile::ELF::StringTable');

   $this->{'ELF String Tables'}->{$strtab->section->index} = $strtab;
   $this->hdrstr($strtab)
      if $strtab->section->index == $this->elfheader->shstrndx;
   $strtab;
}

sub addsymtab ($$) {
   my $this = shift;
   my $symtab = shift;
   croak 'Not an ELF symbol table'
      unless $symtab->isa('BinFile::ELF::SymbolTable');

   $this->{'ELF Symbol Tables'}->{$symtab->section->index} = $symtab;
   $symtab;
}

sub search ($$) {
   my $this = shift;
   my $pattern = shift;
   my $symtab;
   my @results;
   foreach $symtab ( values %{$this->elfsymtabs} ) {
      foreach( $symtab->search($pattern) ) {
         push( @results, $_ );
      }
   }
   @results;
}

sub dump ($) {
   my $this = shift;
}

sub dumpheader ($) {
   my $this = shift;
   $this->SUPER::dumpheader;
   $this->elfheader->dump();
}

sub dumpsections ($) {
   my $this = shift;
   foreach( $this->elfsections ) {
      print '['.$this->format.' Section]'."\n";
      $_->dump($this->hdrstr);
   }
}

sub dumpstrtabs ($) {
   my $this = shift;
   while( my($key, $value) = each %{$this->elfstrtabs} ) {
      printf("Section 0x%2.2x: ", $key);
      $value->dump();
   }
}

sub dumpsymtabs ($) {
   my $this = shift;
   while( my($key, $value) = each %{$this->elfsymtabs} ) {
      printf("Section 0x%2.2x: ", $key);
      $value->dump();
   }
}

1;
