package BinFile::ELF::Header;
# 
# Copyright (C) 2001  Network Associates, Incorporated (NAI)
# All rights reserved.
#
# This file is part of the NAI Generic Software Wrappers Toolkit
# (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
# distribution.
#
# The GSWTK is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The GSWTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: Header.pm,v 1.6 2001/01/17 22:36:33 stevek Exp $
#
require 5.004;
use strict;
use Carp;
use IO::Seekable;

BEGIN {
   use Exporter   ();
   use vars       qw(@EXPORT $VERSION);
   @EXPORT      = qw(loadElfHeader
                     EI_MAG0 EI_MAG1 EI_MAG2 EI_MAG3 EI_CLASS EI_DATA
                     EI_VERSION EI_OSABI EI_ABIVERSION OLD_EI_BRAND
                     EI_PAD EI_NIDENT
                     ELFMAG0 ELFMAG1 ELFMAG2 ELFMAG3 ELFMAG
                     EV_NONE EV_CURRENT
                     ELFCLASSNONE ELFCLASS32 ELFCLASS64
                     ELFDATANONE ELFDATA2LSB ELFDATA2MSB
                     ELFOSABI_SYSV ELFOSABI_HPUX ELFOSABI_NETBSD
                     ELFOSABI_LINUX ELFOSABI_HURD ELFOSABI_86OPEN
                     ELFOSABI_SOLARIS ELFOSABI_MONTEREY ELFOSABI_IRIX
                     ELFOSABI_FREEBSD ELFOSABI_TRU64 ELFOSABI_ARM
                     ELFOSABI_STANDALONE
                     ET_NONE ET_REL ET_EXEC ET_DYN ET_CORE
                     EM_NONE EM_M32 EM_SPARC EM_386 EM_68K EM_88K EM_486
                     EM_860 EM_MIPS EM_MIPS_RS4_BE EM_SPARC64 EM_PARISC
                     EM_PPC EM_IA_64 EM_ALPHA);
   $VERSION     = 0.001;
}


# Indexes into the e_ident array
sub EI_MAG0 { 0; }         # Magic number, byte 0
sub EI_MAG1 { 1; }         # Magic number, byte 1
sub EI_MAG2 { 2; }         # Magic number, byte 2
sub EI_MAG3 { 3; }         # Magic number, byte 3
sub EI_CLASS { 4; }        # Class of machine
sub EI_DATA { 5; }         # Data format
sub EI_VERSION { 6; }      # ELF format version
sub EI_OSABI { 7; }        # Operating system / ABI identification
sub EI_ABIVERSION { 8; }   # ABI version
sub OLD_EI_BRAND { 8; }    # Start of architecture identification
sub EI_PAD { 9; }          # Start of padding (per SVR4 ABI)
sub EI_NIDENT { 16; }      # Size of e_ident array

# Values for the magic number bytes
sub ELFMAG0 { 0x7f; }      # 0x7f
sub ELFMAG1 { 0x45; }      # 'E'
sub ELFMAG2 { 0x4c; }      # 'L'
sub ELFMAG3 { 0x46; }      # 'F'

sub IS_ELF (@) {
   croak 'Magic array must contain at least '.((EI_MAG3)+1).' values'
      unless @_ > EI_MAG3;
   ($_[EI_MAG0] == ELFMAG0) && ($_[EI_MAG1] == ELFMAG1) &&
   ($_[EI_MAG2] == ELFMAG2) && ($_[EI_MAG3] == ELFMAG3);
}

# Values for e_ident[EI_VERSION] and e_version
sub EV_NONE { 0; }
sub EV_CURRENT { 1; }

# Values for e_ident[EI_CLASS]
sub ELFCLASSNONE { 0; }
sub ELFCLASS32 { 1; }
sub ELFCLASS64 { 2; }

# Values for e_ident[EI_DATA]
sub ELFDATANONE { 0; }
sub ELFDATA2LSB { 1; }
sub ELFDATA2MSB { 2; }

# Values for e_ident[EI_OSABI]
sub ELFOSABI_SYSV { 0; }
sub ELFOSABI_HPUX { 1; }
sub ELFOSABI_NETBSD { 2; }
sub ELFOSABI_LINUX { 3; }
sub ELFOSABI_HURD { 4; }
sub ELFOSABI_86OPEN { 5; }
sub ELFOSABI_SOLARIS { 6; }
sub ELFOSABI_MONTEREY { 7; }
sub ELFOSABI_IRIX { 8; }
sub ELFOSABI_FREEBSD { 9; }
sub ELFOSABI_TRU64 { 10; }
sub ELFOSABI_ARM { 97; }
sub ELFOSABI_STANDALONE { 255; }

# Values for e_type
sub ET_NONE { 0; }         # Unknown type.
sub ET_REL { 1; }          # Relocatable.
sub ET_EXEC { 2; }         # Executable.
sub ET_DYN { 3; }          # Shared object.
sub ET_CORE { 4; }         # Core file.

# Values for e_machine.
sub EM_NONE { 0; }         # Unknown machine.
sub EM_M32 { 1; }          # AT&T WE32100.
sub EM_SPARC { 2; }        # Sun SPARC.
sub EM_386 { 3; }          # Intel i386.
sub EM_68K { 4; }          # Motorola 68000.
sub EM_88K { 5; }          # Motorola 88000.
sub EM_486 { 6; }          # Intel i486.
sub EM_860 { 7; }          # Intel i860.
sub EM_MIPS { 8; }         # MIPS R3000 Big-Endian only
sub EM_MIPS_RS4_BE { 10; } # MIPS R4000 Big-Endian
sub EM_SPARC64 { 11; }     # SPARC v9 64-bit unoffical
sub EM_PARISC { 15; }      # HPPA
sub EM_PPC { 20; }         # PowerPC
sub EM_IA_64 { 50; }       # Intel IA-46 Processor
sub EM_ALPHA { 0x9026; }   # Alpha


sub import { goto &Exporter::import }

use Class::OOStruct qw(struct);
struct 'BinFile::ELF::Header' => [
   ident      => '@', # 4 bytes  (unsigned byte)     offset 0
   class      => '$', # 1 byte   (unsigned byte)     offset 4
   data       => '$', # 1 byte   (unsigned byte)     offset 5
   fmtversion => '$', # 1 byte   (unsigned byte)     offset 6
   osabi      => '$', # 1 byte   (unsigned byte)     offset 7
   brand      => '@', # 8 bytes  (unsigned byte)     offset 8
   type       => '$',
   machine    => '$',
   version    => '$',
   entry      => '$',
   phoff      => '$',
   shoff      => '$',
   flags      => '$',
   ehsize     => '$',
   phentsize  => '$',
   phnum      => '$',
   shentsize  => '$',
   shnum      => '$',
   shstrndx   => '$'
];


sub loadElfHeader ($) {
   my $fd = shift;

   croak "Invalid argument" unless ref($fd);
   my $bytes;
   my $total = 0;
   while( $total <= EI_CLASS ) {
      my $nread = sysread( $fd, $bytes, (EI_CLASS)+1, $total );
      croak "Unable to read ELF header" unless $nread;
      $total += $nread;
   }
   croak "Unable to read ELF header" if( $total <= EI_CLASS );

   my @parts = unpack('C'.((EI_CLASS)+1), $bytes);
   croak "Not an ELF file" unless IS_ELF @parts[(EI_MAG0..EI_MAG3)];

   my $header;
   if( $parts[EI_CLASS] == ELFCLASS32 ) {
      eval('use BinFile::ELF::ELF32::Header;'.
           '$header = loadElf32Header( $fd );');
      return $header if ref($header);
      print "ELF-32 Header creation failed: $@";
      croak "ELF-32 Header creation failed: $@";
   }

   if( $parts[EI_CLASS] == ELFCLASS64 ) {
      eval('use BinFile::ELF::ELF64::Header;'.
           '$header = loadElf64Header( $fd );');
      return $header if ref($header);
      croak "ELF-64 Header creation failed: $@";
   }

   croak "Unknown ELF class";
}

sub dump ($) {
   my $this = shift;

   printf("     ident: %2.2x%2.2x%2.2x%2.2x [%c%c%c%c]\n",
          @{$this->ident}[0..3], @{$this->ident}[0..3]);
   printf("     class:       %2.2x [32-bit]\n", $this->class);
   printf("      data:       %2.2x", $this->data);
   print(" [little-endian]\n") if $this->data == ELFDATA2LSB;
   print(" [big-endian]\n")    if $this->data == ELFDATA2MSB;
   print(" [UNKOWN]\n" )       if( ($this->data<ELFDATA2LSB) ||
                                   ($this->data>ELFDATA2MSB) );
   printf("fmtversion:       %2.2x", $this->fmtversion);
   print(" [current]\n" )      if $this->fmtversion == 0x01;
   print(" [UNKOWN]\n" )       if $this->fmtversion != 0x01;
   printf("     osabi:       %2.2x", $this->osabi);
   print(" [SysV]\n")          if $this->osabi == 0x00;
   print(" [HP-UX]\n")         if $this->osabi == 0x01;
   print(" [NETBSD]\n")        if $this->osabi == 0x02;
   print(" [Linux]\n")         if $this->osabi == 0x03;
   print(" [Hurd]\n")          if $this->osabi == 0x04;
   print(" [86Open]\n")        if $this->osabi == 0x05;
   print(" [Solaris]\n")       if $this->osabi == 0x06;
   print(" [Monterey]\n")      if $this->osabi == 0x07;
   print(" [IRIX]\n")          if $this->osabi == 0x08;
   print(" [FreeBSD]\n")       if $this->osabi == 0x09;
   print(" [TRU64]\n")         if $this->osabi == 0x0a;
   print(" [ARM]\n")           if $this->osabi == 0x61;
   print(" [Standalone]\n")    if $this->osabi == 0xff;
   print(" [UNKNOWN]\n")       if( ($this->osabi>0x0a) && ($this->osabi<0x61) ||
                                   ($this->osabi>0x61) && ($this->osabi<0xff) );
   printf("     brand: %2.2x%2.2x%2.2x%2.2x [%c%c%c%c]\n",
          @{$this->brand}[0..3], @{$this->brand}[0..3]);
   printf("            %2.2x%2.2x%2.2x%2.2x [%c%c%c%c]\n",
          @{$this->brand}[4..7], @{$this->brand}[4..7]);
   printf("      type:     %4.4x", $this->type);
   print(" [Relocatable]\n")   if $this->type == 0x01;
   print(" [Executable]\n")    if $this->type == 0x02;
   print(" [Shared Object]\n") if $this->type == 0x03;
   print(" [Core]\n")          if $this->type == 0x04;
   print(" [UNKNOWN]\n")       if( ($this->type == 0x00) ||
                                   ($this->type>0x04) );
   printf("   machine:     %4.4x", $this->machine);
   print(" [AT&T WE32100]\n")          if $this->machine == 0x01;
   print(" [Sun SPARC]\n")             if $this->machine == 0x02;
   print(" [Intel i386]\n")            if $this->machine == 0x03;
   print(" [Motorola 68000]\n")        if $this->machine == 0x04;
   print(" [Motorola 88000]\n")        if $this->machine == 0x05;
   print(" [Intel i486]\n")            if $this->machine == 0x06;
   print(" [Intel i860]\n")            if $this->machine == 0x07;
   print(" [MIPS R3000 Big-Endian]\n") if $this->machine == 0x08;
   print(" [MIPS R4000 Big-Endian]\n") if $this->machine == 0x0a;
   print(" [SPARC v9 64-bit]\n")       if $this->machine == 0x0b;
   print(" [HPPA]\n")                  if $this->machine == 0x0f;
   print(" [PowerPC]\n")               if $this->machine == 0x14;
   print(" [Intel IA-64]\n")           if $this->machine == 0x32;
   print(" [Alpha]\n")                 if $this->machine == 0x9026;
   print(" [UNKNOWN]\n")               if( ($this->machine==0x00) ||
                                           ($this->machine==0x09) ||
                                           ( ($this->machine>0x0b) &&
                                             ($this->machine<0x0f) ) ||
                                           ( ($this->machine>0x0f) &&
                                             ($this->machine<0x14) ) ||
                                           ( ($this->machine>0x14) &&
                                             ($this->machine<0x32) ) ||
                                           ($this->machine>0x9026) );
   printf("   version: %8.8x\n", $this->version);
   printf("     entry: %8.8x\n", $this->entry);
   printf("     phoff: %8.8x\n", $this->phoff);
   printf("     shoff: %8.8x\n", $this->shoff);
   printf("     flags: %8.8x\n", $this->flags);
   printf("    ehsize:     %4.4x\n", $this->ehsize);
   printf(" phentsize:     %4.4x\n", $this->phentsize);
   printf("     phnum:     %4.4x\n", $this->phnum);
   printf(" shentsize:     %4.4x\n", $this->shentsize);
   printf("     shnum:     %4.4x\n", $this->shnum);
   printf("  shstrndx:     %4.4x\n", $this->shstrndx);
};

1;
