package BinFile::PE::Directory::Export;
# 
# Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
# All rights reserved.
#
# This file is part of the NAI Generic Software Wrappers Toolkit
# (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
# distribution.
#
# The GSWTK is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The GSWTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: Export.pm,v 1.7 2001/04/05 18:41:14 stevek Exp $

require 5.004;
use strict;
use Carp;
use IO::Seekable;

BEGIN {
   use Exporter   ();
   use vars       qw(@EXPORT $VERSION $DIR_SIZE);
   @EXPORT      = qw(loadPEExportDirectory);
   $VERSION     = 0.001;
   $DIR_SIZE    = 40;
}

sub import { goto &Exporter::import }

use Class::Struct qw(struct);
struct 'BinFile::PE::Directory::Export' => [
   charistics   => '$', # 4 bytes  (unsigned long)  offset 0
   timestamp    => '$', # 4 bytes  (unsigned long)  offset 4
   majorvers    => '$', # 2 bytes  (unsigned short) offset 8
   minorvers    => '$', # 2 bytes  (unsigned short) offset 10
   name         => '$', # 4 bytes  (unsigned long)  offset 12
   base         => '$', # 4 bytes  (unsigned long)  offset 16
   nfuncs       => '$', # 4 bytes  (unsigned long)  offset 20
   nnames       => '$', # 4 bytes  (unsigned long)  offset 24
   addrfuncs    => '$', # 4 bytes  (unsigned long)  offset 28
   addrnames    => '$', # 4 bytes  (unsigned long)  offset 32
   addrnameords => '$', # 4 bytes  (unsigned long)  offset 36
   symbols      => '%'
];

sub loadPEExportDirectory {
   my $fd = shift;
   my( $coff, $section ) = @_;

   croak "Invalid argument" unless ref($fd) && ref($coff) && ref($section);

   my $offset = $coff->findSectionByRVA( $section->{'RVA'} );
   sysseek( $fd, $offset, SEEK_SET ) || croak;

   my $dir;
   my $total = 0;
   while( $total < $DIR_SIZE ) {
      my $nread = sysread( $fd, $dir, $DIR_SIZE, $total );
      croak "Unable to read PE Export directory" unless $nread;
      $total += $nread;
   }
   croak "Unable to read PE Export directory" if( $total < $DIR_SIZE );
   my @parts = unpack('VVvvVVVVVVV', $dir);
   my $ped = new();
   ($ped->[0],  $ped->[1] ) = @parts[ (0..1) ];
   ($ped->[2],  $ped->[3] ) = @parts[ (2..3) ];
   ($ped->[4],  $ped->[5],
    $ped->[6],  $ped->[7],
    $ped->[8],  $ped->[9],
    $ped->[10])             = @parts[ (4..10) ];

   $ped->symbols->{'Ordinals'} = {};
   $ped->symbols->{'Names'} = {};
   $ped->symbols->{'Addresses'} = {};
   $ped->symbols->{'RVA Addresses'} = {};

   # Read name RVAs
   my $index;
   my $addroff = $coff->findSectionByRVA($ped->addrnames);
   for( $index = 0; $index < $ped->nnames; $index++ ) {
      sysseek( $fd, $addroff + ( $index * 4 ), SEEK_SET ) || croak;
      my $buf;
      my $nread = sysread( $fd, $buf, 4 );
      croak "Unable to read name RVA at index $index" if $nread != 4;
      my( $namerva ) = unpack( 'V', $buf );

      # Read name
      my $nameoff = $coff->findSectionByRVA($namerva);
      sysseek( $fd, $nameoff, SEEK_SET ) || croak;
      my @parts;
      do {
         my $c;
         sysread( $fd, $c, 1 ) || croak;
         push( @parts, unpack('C', $c) );
      }
      while( $parts[$#parts] );
      croak "Unable to read symbol name" if !@parts;
      $ped->symbols->{'Names'}{$index} = pack('C'.($#parts), @parts);
   }

   # Read name ordinals
   $addroff = $coff->findSectionByRVA($ped->addrnameords);
   for( $index = 0; $index < $ped->nnames; $index++ ) {
      sysseek( $fd, $addroff + ( $index * 2 ), SEEK_SET ) || croak;
      my $buf;
      my $nread = sysread( $fd, $buf, 4 );
      croak "Unable to read name ordinal at index $index" if $nread != 4;
      my( $nameord ) = unpack( 'v', $buf );
      $ped->symbols->{'Ordinals'}{$nameord} = $index;
   }

   # Read addresses and match by ordinal
   $addroff = $coff->findSectionByRVA($ped->addrfuncs);
   for( $index = 0; $index < $ped->nfuncs; $index++ ) {
      sysseek( $fd, $addroff + ( $index * 4 ), SEEK_SET ) || croak;
      my $buf;
      my $nread = sysread( $fd, $buf, 4 );
      croak "Unable to read directory address RVA at index $index"
         if $nread != 4;
      my $addrrva = unpack( 'V', $buf );
      if( ( $addrrva < $offset ) ||
          ( $addrrva >= ( $offset + $section->{'Size'} ) ) ) {
         # Not forwarded, find the name
         my $name;
         if( exists $ped->symbols->{'Ordinals'}{$index} ) {
            my $ord = $ped->symbols->{'Ordinals'}{$index};
            $name = $ped->symbols->{'Names'}{$ord};
         }
         else {
            $name = '@' . $index;
         }
         $ped->symbols->{'Addresses'}{$name} = $addrrva;
         my $addr = $coff->findSectionByRVA($addrrva);
         $ped->symbols->{'RVA Addresses'}{$addr} = $name if $addr;
      }
   }

   $ped;
}

sub size { return $DIR_SIZE; }

sub dump ($) {
   my $this = shift;
   printf("[PE Export Directory]\n");
   printf("  charistics: %8.8x\n", $this->charistics);
   printf("   timestamp: %8.8x\n", $this->timestamp);
   printf("   majorvers:     %4.4x\n", $this->majorvers);
   printf("   minorvers:     %4.4x\n", $this->minorvers);
   printf("        name: %8.8x\n", $this->name);
   printf("        base: %8.8x\n", $this->base);
   printf("      nfuncs: %8.8x\n", $this->nfuncs);
   printf("      nnames: %8.8x\n", $this->nnames);
   printf("   addrfuncs: %8.8x\n", $this->addrfuncs);
   printf("   addrnames: %8.8x\n", $this->addrnames);
   printf("addrnameords: %8.8x\n", $this->addrnameords);
}

1;
