package BinFile::PE::OptionalHeader;
# 
# Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
# All rights reserved.
#
# This file is part of the NAI Generic Software Wrappers Toolkit
# (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
# distribution.
#
# The GSWTK is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# The GSWTK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: OptionalHeader.pm,v 1.9 2001/04/05 18:41:14 stevek Exp $

require 5.004;
use strict;
use Carp;
use IO::Seekable;

BEGIN {
   use Exporter   ();
   use vars       qw(@EXPORT $VERSION $HEADER_SIZE);
   use BinFile::COFF;
   use BinFile::PE::Directory::Export;
   use BinFile::PE::Directory::Import;
   @EXPORT      = qw(loadPEOptionalHeader);
   $VERSION     = 0.001;
   $HEADER_SIZE = 96;
}

sub import { goto &Exporter::import }

my @directory_entries = qw(
   Export    Import      Resource      Exception
   Security  BaseReloc   Debug         Architecture
   GlobalPtr TLS         LoadConfig    BoundImport
   IAT       DelayImport ComDescriptor
);

use Class::Struct qw(struct);
struct 'BinFile::PE::OptionalHeader' => [
   # COFF Fields
   magic         => '$', # 2 bytes  (unsigned short)      offset 0
   majorvers     => '$', # 1 byte   (unsigned char)       offset 2
   minorvers     => '$', # 1 byte   (unsigned char)       offset 3
   textsize      => '$', # 4 bytes  (unsigned long)       offset 4
   datasize      => '$', # 4 bytes  (unsigned long)       offset 8
   bsssize       => '$', # 4 bytes  (unsigned long)       offset 12
   entry         => '$', # 4 bytes  (unsigned long)       offset 16
   textbase      => '$', # 4 bytes  (unsigned long)       offset 20
   database      => '$', # 4 bytes  (unsigned long)       offset 24
   # Additional Fields
   imagebase     => '$', # 4 bytes  (unsigned long)       offset 28
   sectalign     => '$', # 4 bytes  (unsigned long)       offset 32
   pagealign     => '$', # 4 bytes  (unsigned long)       offset 36
   osversmajor   => '$', # 2 bytes  (unsigned short)      offset 40
   osversminor   => '$', # 2 bytes  (unsigned short)      offset 42
   imgversmajor  => '$', # 2 bytes  (unsigned short)      offset 44
   imgversminor  => '$', # 2 bytes  (unsigned short)      offset 46
   subversmajor  => '$', # 2 bytes  (unsigned short)      offset 48
   subversminor  => '$', # 2 bytes  (unsigned short)      offset 50
   reserved      => '$', # 4 bytes  (unsigned long)       offset 52
   imagesize     => '$', # 4 bytes  (unsigned long)       offset 56
   headersize    => '$', # 4 bytes  (unsigned long)       offset 60
   checksum      => '$', # 4 bytes  (unsigned long)       offset 64
   subsystem     => '$', # 2 bytes  (unsigned short)      offset 68
   dllcharistics => '$', # 2 bytes  (unsigned short)      offset 70
   stackreserve  => '$', # 4 bytes  (unsigned long)       offset 72
   stackcommit   => '$', # 4 bytes  (unsigned long)       offset 76
   heapreserve   => '$', # 4 bytes  (unsigned long)       offset 80
   heapcommit    => '$', # 4 bytes  (unsigned long)       offset 84
   loaderflags   => '$', # 4 bytes  (unsigned long)       offset 88
   nrvasize      => '$', # 4 bytes  (unsigned long)       offset 92
   directories   => '@',
   directory     => '%'
];

sub loadPEOptionalHeader {
   my $fd = shift;
   my $coff = shift;

   croak "Invalid argument" unless ref($fd) && ref($coff);

   # Read optional header
   my $header;
   my $total = 0;
   while( $total < $HEADER_SIZE ) {
      my $nread = sysread( $fd, $header, $HEADER_SIZE, $total );
      croak "Unable to read PE header" unless $nread;
      $total += $nread;
   }
   croak "Unable to read PE header" if( $total < $HEADER_SIZE );
   my @parts = unpack('v c1 c1 VVVVVV VVVvvvvvvVVVVvvVVVVVV', $header);
   my $deh = new();
   $deh->[0]                = $parts[0];
   ($deh->[1],  $deh->[2] ) = @parts[ (1..2) ];
   ($deh->[3],  $deh->[4],
    $deh->[5],  $deh->[6],
    $deh->[7],  $deh->[8],
    $deh->[9],  $deh->[10],
    $deh->[11])             = @parts[ (3..11) ];
   ($deh->[12], $deh->[13],
    $deh->[14], $deh->[15],
    $deh->[16], $deh->[17]) = @parts[ (12..17) ];
   ($deh->[18], $deh->[19],
    $deh->[20], $deh->[21]) = @parts[ (18..21) ];
   ($deh->[22], $deh->[23]) = @parts[ (22..23) ];
   ($deh->[24], $deh->[25],
    $deh->[26], $deh->[27],
    $deh->[28], $deh->[29]) = @parts[ (24..29) ];

   # Read RVA/size pairs
   if( $deh->nrvasize > 0 ) {
      my @pairs;
      foreach( 0 .. ($deh->nrvasize - 1) ) {
         my $rvasize;
         $total = 0;
         while( $total < 8 ) {
            my $nread = sysread( $fd, $rvasize, 8, $total );
            croak "Unable to read RVA/size pair" unless $nread;
            $total += $nread;
         }
         croak "Unable to read RVA/size pair" if( $total < 8 );
         my @pairs = unpack('VV', $rvasize);
         $deh->directories($_, \@pairs);
         $deh->directory->{$directory_entries[$_]} =
            { 'RVA'      => $pairs[0],
              'Size'     => $pairs[1],
              'Instance' => undef }
            if( defined($directory_entries[$_]) && $pairs[1] );

      }
   }

   # Read in directory entries
   if( $deh->nrvasize > 0 ) {
      foreach( 0 .. ($deh->nrvasize - 1) ) {
          my $name = $directory_entries[$_];
          next unless defined($name);
          if( exists $deh->directory->{$name} ) {
             my $dir = $deh->directory->{$name};
             foreach( $name ) {
               /^Export$/ &&
                  do {
                     my $offset = $coff->findSectionByRVA( $dir->{'RVA'} );
                     $deh->directory->{$name}->{'Instance'} =
                        loadPEExportDirectory( $fd, $coff, $dir ) if $offset;
                     last;
                  };

               /^Import$/ &&
                  do {
                     last;
                  };
            }
         }
      }
   }
  
   $deh;
}

sub size { return $HEADER_SIZE; }

sub dump ($) {
   my $this = shift;
   printf("[PE Optional Header]\n");
   printf("        magic:     %4.4x\n", $this->magic);
   printf("    majorvers:       %2.2x\n", $this->majorvers);
   printf("    minorvers:       %2.2x\n", $this->minorvers);
   printf("     textsize: %8.8x\n", $this->textsize);
   printf("     datasize: %8.8x\n", $this->datasize);
   printf("      bsssize: %8.8x\n", $this->bsssize);
   printf("        entry: %8.8x\n", $this->entry);
   printf("     textbase: %8.8x\n", $this->textbase);
   printf("     database: %8.8x\n", $this->database);
   printf("    imagebase: %8.8x\n", $this->imagebase);
   printf("    sectalign: %8.8x\n", $this->sectalign);
   printf("    pagealign: %8.8x\n", $this->pagealign);
   printf("  osversmajor:     %4.4x\n", $this->osversmajor);
   printf("  osversminor:     %4.4x\n", $this->osversminor);
   printf(" imgversmajor:     %4.4x\n", $this->imgversmajor);
   printf(" imgversminor:     %4.4x\n", $this->imgversminor);
   printf(" subversmajor:     %4.4x\n", $this->subversmajor);
   printf(" subversminor:     %4.4x\n", $this->subversminor);
   printf("     reserved: %8.8x\n", $this->reserved);
   printf("    imagesize: %8.8x\n", $this->imagesize);
   printf("   headersize: %8.8x\n", $this->headersize);
   printf("     checksum: %8.8x\n", $this->checksum);
   printf("    subsystem:     %4.4x\n", $this->subsystem);
   printf("dllcharistics:     %4.4x\n", $this->dllcharistics);
   printf(" stackreserve: %8.8x\n", $this->stackreserve);
   printf("  stackcommit: %8.8x\n", $this->stackcommit);
   printf("  heapreserve: %8.8x\n", $this->heapreserve);
   printf("   heapcommit: %8.8x\n", $this->heapcommit);
   printf("  loaderflags: %8.8x\n", $this->loaderflags);
   printf("     nrvasize: %8.8x\n", $this->nrvasize);

   if( $this->nrvasize > 0 ) {
      printf("\n  [Directories]\n");
      foreach( 0 .. ($this->nrvasize - 1) ) {
         my( $rva, $size ) = @{$this->directories($_)};
         printf("  [%13s] %3d. RVA: %8.8x  Size: %8.8x\n",
                (defined($directory_entries[$_]))?
                   $directory_entries[$_]:'UNKNOWN', $_+1, $rva, $size);
      }
   }
}

1;
