/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: dbi.cc,v 1.18 2001/04/05 18:41:16 stevek Exp $
 *
 * Author: Lee Badger
 *
 * Wrappers DB command line utility.
 *
 * A few notes about this program:
 *
 * Dbi creates a wss.db file if it doesn't exist.  The last parameter
 * to a dbi command is usually a string that dbi feeds to a parser.
 * The parser builds data structures that are then passed to the WSS
 * DBMS routines, which can be compiled either for kernel space (as in
 * the WSS) or for user space (as in this program).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include "wr_conf.h"

extern "C" {
#include "wss_db.h"
#include "wr_iface.h"
#include "wrerrno.h"
vm_offset_t wsi_vm_init(size_t size);
unsigned wr_db_space();
}

#include "dbi.h"

/*
 *	The wrperror stuff should be moved to a libwrap library.
 */
char *wrerror_strings[] = {
	"Everything is OK; success",
	"General, nondescript error",
	"Operation not permitted",
	"Cannot allocate memory",
	"Object not found",
	"Object still has references",
	"Object already exists",
	"Any of a variety of problems",
	"Field name doesn't exist",
	"wss.db exhausted",
	"Comparison of incomparable types",
	NULL
	};
#define	WR_NUM_ERRORS	10
void wr_perror(int error, char *subsys)
{
	if ((error < 0) || (error > WR_NUM_ERRORS)) {
		fprintf(stderr, "Error: unknown error code: %d\n", error);
	}
	else {
		fprintf(stderr, "%s error: %s\n", subsys,
					wrerror_strings[error]);
	}
}

/*
 * Print the usage information.
 */
int
usage()
{
	fprintf(stderr, "\
usage: dbi mem
       dbi create <table-name> '( <name> <type>, <name> <type>, ... )'
       dbi drop <table-name>
       dbi insert <table-name> '(<val>, <val>, ...), (<val>, <val>, ...), ...'
       dbi select <table-name> '<bool-expr>' e.g., '<field> = <val>'
       dbi delete
       dbi update
where <table-name> is: <wrapper> <id> <table>, or <id> <table>, or
<wrapper> <table> or <table>;
by default, <wrapper> is the environment var DEFWRAP, and tables are global.\n");
	exit(2);
}

#ifdef WR_SOLARIS
#define	WRDB_SIZE	16			/* in megabytes */
#else
#define	WRDB_SIZE	64			/* in megabytes */
#endif /* WR_SOLARIS */
#define	WRDB_BYTES	(WRDB_SIZE * 1048576)

void mem_cmd(char **av, int ac);
void tables_cmd(char *av[], int ac);
void create_cmd(char *av[], int ac);
void drop_cmd(char *av[], int ac);
void insert_cmd(char *av[], int ac);
void select_cmd(char *av[], int ac);
void delete_cmd(char *av[], int ac);
void update_cmd(char *av[], int ac);
char *get_table_name(char *av[], int ac, char **wr, int *id, char **tab);
void print_row(wqcd_t cd, wqrow_t row, int col_count);

/* Globals */
char *default_wrapper = NULL;
int  wrapper_id;
char *table_name = NULL;
char *wrapper_name = NULL;

int main(int argc, char **argv)
{
	wsi_vm_init(WRDB_BYTES);
	default_wrapper = getenv("DEFWRAP");

	if (argc < 2) {
		usage();
	}

	if (! strcmp(argv[1], "mem")) {
		mem_cmd(argv, argc);
	}
	else if (! strcmp(argv[1], "tables")) {
		tables_cmd(argv, argc);
	}
	else if (! strcmp(argv[1], "create")) {
		create_cmd(argv, argc);
	}
	else if (! strcmp(argv[1], "drop")) {
		drop_cmd(argv, argc);
	}
	else if (! strcmp(argv[1], "insert")) {
		insert_cmd(argv, argc);
	}
	else if (! strcmp(argv[1], "select")) {
		select_cmd(argv, argc);
	}
	else if (! strcmp(argv[1], "delete")) {
		delete_cmd(argv, argc);
	}
	else if (! strcmp(argv[1], "update")) {
		update_cmd(argv, argc);
	}
}

/*
 * Print the free space in the DB file.
 */
void mem_cmd(char **av, int ac)
{
	if (ac != 2) {
		usage();
		exit(2);
	}
	printf("wss.db(%d bytes free)\n", wr_db_space());
}

/*
 * Print the names of the tables in the DB.  If the "cols" option was
 * specified, print their schemas also.
 */
extern struct wql_table **wss_vm_root_table;
void tables_cmd(char *av[], int ac)
{
	wqtab_t tabs;

        tabs = *wss_vm_root_table?(wqtab_t)rel2abs(*wss_vm_root_table):NULL;
	while (tabs) {
		printf("%s\n", rel2abs(tabs->name));
		if ((ac == 3) && (!strcmp(av[2], "cols"))) {
			wqcd_t cd;
                        int i;
                        cd = (wqcd_t)rel2abs(tabs->col_defs);
                        for (i = 0 ; i < tabs->col_count ; i++) {
				char *dtypec;
                                switch(cd[i].dtype) {
				case char_t:
					dtypec = "char";
					break;
				case int_t:
					dtypec = "int";
					break;
				case float_t:
					dtypec = "float";
					break;
                                case none_t:
                                        dtypec = "none";
                                        break;
				}
				printf("\t(%s, %s, %d)\n",
				rel2abs(cd[i].name), dtypec, cd[i].max_len);
			}
		}
		tabs = (tabs->next)?(wqtab_t)rel2abs(tabs->next):NULL;
	}
}

/*
 * Create a table with the specified schema.
 */
void create_cmd(char *av[], int ac)
{
	char *tabname, *wr, *tab;
	int id;
	struct parse_result *res = (pres_t)malloc(sizeof(struct parse_result));

	if ((ac < 4) || (ac > 6)) {
		usage();
	}
	tabname = get_table_name(av, ac-1, &wr, &id, &tab);
	if (wr_parse_dbi_command(res, av[ac-1], create_c) != NULL) {
		fprintf(stderr, "parse error in create command\n");
		exit(2);
	}
	int error = wss_create(wr,id,tab,1,
                res->r.c.create,res->r.c.col_count,
                NULL);
	if (error) {
		wr_perror(-error, "dbi");
	}
}

/*
 * Drop a table from the DB.
 */
void drop_cmd(char *av[], int ac)
{
	char *tabname, *wr, *tab;
        unsigned long table_id;
	int id;

	if ((ac < 3) || (ac > 5)) {
		usage();
	}
	tabname = get_table_name(av, ac, &wr, &id, &tab);
        
        table_id = wss_db_find_table_by_names(wr,id,tab);
        if (!table_id) {
            wr_perror(WRE_NOTFOUND, "wss_db_find_table_by_names");
            return;
        }
	int error = wss_drop(table_id);
	if (error) {
		wr_perror(-error, "dbi");
	}
}

/*
 * Insert a row into a table.
 */
void insert_cmd(char *av[], int ac)
{
	char *tabname, *wr, *tab;
	int id;
        unsigned long table_id;
	struct parse_result *res = (pres_t)malloc(sizeof(struct parse_result));

	if ((ac < 4) || (ac > 6)) {
		usage();
	}
	tabname = get_table_name(av, ac-1, &wr, &id, &tab);
	if (wr_parse_dbi_command(res, av[ac-1], insert_c)) {
		fprintf(stderr, "parse error in insert command\n");
		exit(2);
	}
        table_id = wss_db_find_table_by_names(wr,id,tab);
        if (!table_id) {
            wr_perror(WRE_NOTFOUND, "wss_db_find_table_by_names");
            return;
        }
	int error = wss_insert(table_id, !NULL, res->r.insert);
	if (error) {
		wr_perror(-error, "dbi");
	}
}

/*
 * Print selected rows from a specified table on the standard output.
 */
void select_cmd(char *av[], int ac)
{
	char *tabname, *wr, *tab;
	int id;
        unsigned long table_id;
	struct parse_result *res = (pres_t)malloc(sizeof(struct parse_result));

	if ((ac < 4) || (ac > 6)) {
		usage();
	}
	tabname = get_table_name(av, ac-1, &wr, &id, &tab);
	wrapper_id = id;
	table_name = tab;
 	wrapper_name = wr;
	if (wr_parse_dbi_command(res, av[ac-1], select_c)) {
		fprintf(stderr, "parse error in select command\n");
		exit(2);
	}
        wqtab new_tab;

        table_id = wss_db_find_table_by_names(wr,id,tab);
        if (!table_id) {
            wr_perror(WRE_NOTFOUND, "wss_db_find_table_by_names");
            return;
        }
	int error = wss_select(table_id, 1,
			res->r.s.select, 0,
                        res->r.s.order->order_field,
			res->r.s.order->decending,
                        &new_tab);
	if (error < 0) {
		wr_perror(-error, "dbi");
	}
	else {
                wqrow_t rows = new_tab.rows;
		while (rows) {
			printf("\t");
                        print_row(new_tab.col_defs, rows, new_tab.col_count);
			rows = rows->next;
		}
	}
}

/*
 * Delete specified rows from a table.
 */
void delete_cmd(char *av[], int ac)
{
	int num;
        unsigned long table_id;

        table_id = wss_db_find_table_by_names("wrap",0,"companies");
        if (!table_id) {
            wr_perror(WRE_NOTFOUND, "wss_db_find_table_by_names");
            return;
        }
	num = wss_delete(table_id, 0, NULL, 0);

	printf("Removed %d rows\n"
                "(this is actually completely unimplimented.)\n", num);
}

/*
 * Update values in specified rows of a table.
 */
void update_cmd(char *av[], int ac)
{
    char *tabname, *wr, *tab;
    int id;
    unsigned long table_id;
    struct parse_result *res = (pres_t)malloc(sizeof(struct parse_result));

    if ((ac < 4) || (ac > 6)) {
        usage();
    }
    tabname = get_table_name(av, ac-1, &wr, &id, &tab);
    wrapper_id = id;
    table_name = tab;
    wrapper_name = wr;
    if (wr_parse_dbi_command(res, av[ac-1], update_c)) {
        fprintf(stderr, "parse error in update command\n");
        exit(2);
    }

    table_id = wss_db_find_table_by_names(wr,id,tab);
    if (!table_id) {
        wr_perror(WRE_NOTFOUND, "wss_db_find_table_by_names");
        return;
    }
    int error = wss_update(table_id, 1,
            res->r.u.select, 0,
            res->r.u.update);
    if (error < 0) {
        wr_perror(-error, "dbi");
    }
    else {
        wqtab new_tab;
        wqrow_t rows;
        wss_select(table_id, 0, NULL, 0,
                -1, 0, &new_tab);
        rows = new_tab.rows;
        while (rows) {
            printf("\t");
            print_row(new_tab.col_defs, rows, new_tab.col_count);
            rows = rows->next;
        }
    }
}


/*
 * Print a row.  Tab and row are absolute ptrs.
 */
void print_row(wqcd_t cd, wqrow_t row, int col_count)
{
        int i;

	/* printf("row: "); */
        for (i = 0 ; i < col_count ; i++) {
		switch(cd[i].dtype) {
		case char_t :
			printf("char(%s) ", row->vals[i].v.cval);
			break;
		case int_t :
			printf("int(%d) ", row->vals[i].v.ival);
			break;
		case float_t :
			printf("float(%g) ", row->vals[i].v.fval);
			break;
                case none_t :
                        printf("none ");
                        break;
		}
	}
	printf("\n");
}

/*
 * Examines command line arguments and constructs a "complete" table
 * name, which consists of a wrapper name concatinated with
 * (optionally) an integer (in the case of local tables) concatinated
 * with a "simple" table name.  The user may provide any of the
 * following:
 *
 *	<wrapper> <id>	<table>
 *	<wrapper> 	<table>
 *		  <id>	<table>
 *			<table>
 * 
 * However, if <wrapper> is not given, the environment variable
 * DEFWRAP must be defined.  If <id> is not given, the constructed
 * table name is simply <wrapper><table>.
 *
 * Returns a newly allocated merged name, and also returns (via
 * parameters) the separate components in newly allocated buffers.
 */
char *get_table_name(char *av[], int ac, char **wr, int *id, char **tab)
{
	/* User provided only the table name; use defaults. */
	if (ac == 3) {
		if (default_wrapper == NULL) {
			fprintf(stderr,
				"DEFWRAP must be defined to use shortcut\n");
				exit(2);
		}
		char *str = (char *)malloc(strlen(default_wrapper) +
				strlen("_0_") + strlen(av[2]) + 3);
		strcpy(str, default_wrapper);
		*wr = strdup(default_wrapper);
		strcat(str, "_0_");
		*id = 0;
		strcat(str, av[2]);
		*tab = strdup(av[2]);
		return(str);
	}

	/* User ommited either the wrapper name or the instance id. */
	if (ac == 4) {
		char *cp;
		int isnumber = !NULL;
		for (cp=av[2]; *cp; cp++) {
			if (!isdigit(*cp)) {
				isnumber = NULL;
			}
		}
		if (isnumber) {
			if (default_wrapper == NULL) {
			fprintf(stderr,
				"DEFWRAP must be defined to use shortcut\n");
				exit(2);
			}
			char *str = (char *)malloc(strlen(default_wrapper) +
				strlen(av[2]) + strlen(av[3]) + 5);
			strcpy(str, default_wrapper);
			*wr = strdup(default_wrapper);
			strcat(str, "_");
			strcat(str, av[2]);
			*id = atoi(av[2]);
			strcat(str, "_");
			strcat(str, av[3]);
			*tab = strdup(av[3]);
			return(str);
		}
		else {
			char *str = (char *)malloc(
					strlen(av[2]) + strlen("_0_") +
					strlen(av[3]) + 5);
			strcpy(str, av[2]);
			*wr = strdup(av[2]);
			strcat(str, "_0_");
			*id = 0;
			strcat(str, av[3]);
			*tab = strdup(av[3]);
			return(str);
		}
	}

	/* User provided everything: wrapper, instance id, and table. */
	if (ac == 5) {
		char *str = (char *)malloc(strlen(av[2]) + strlen(av[3]) +
				 strlen(av[4]) + 5);
		*str = NULL;
		strcpy(str, av[2]);
		*wr = strdup(av[2]);
		strcat(str, "_");
		strcat(str, av[3]);
		*id = atoi(av[3]);
		strcat(str, "_");
		strcat(str, av[4]);
		*tab = strdup(av[4]);
		return(str);
	}
        
        return NULL;
}
