/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: source.cc,v 1.6 2001/04/05 18:41:16 stevek Exp $
 *
 * Author: Lee Badger
 *
 * Purpose: TBD.  This file may get used for dbi.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <grp.h>
#include <pwd.h>
#include <string.h>
#include "source.h"

/* linked list of text */
struct srcrec {
	char *txt;
	struct srcrec *next;
};

source::source()
{
	head = tail = cur = NULL;
}

/* clean out */
source::~source()
{
	struct srcrec *tmp = head;
	struct srcrec *prev;

	while (tmp) {
		prev = tmp;
		tmp = tmp->next;
		free(prev->txt);
		delete(prev);
	}
}

/* Append some ascii text; probably should use the SLIST macro. */
void source::append(char *txt)
{
	if ((!head) && (!tail) ) {
		head = tail = new srcrec;
		head->txt = strdup(txt);
		head->next = NULL;
	}
	else {
		struct srcrec *tmp = new srcrec;
		tmp->txt = strdup(txt);
		tail->next = tmp;
		tail = tmp;
		tail->next = NULL;
	}
}

void source::append(char *txt1, char *txt2)
{
	append(txt1);
	append(txt2);
}

void source::append(char *txt1, char *txt2, char *txt3)
{
	append(txt1);
	append(txt2);
	append(txt3);
}

void source::append(char *txt1, char *txt2, char *txt3, char *txt4)
{
	append(txt1);
	append(txt2);
	append(txt3);
	append(txt4);
}

/* append another source obj */
void source::append(source *src)
{
	if (src->head == NULL) {
		return;
	}
	tail->next = src->head;
	tail = src->tail;
	src->head = src->tail = NULL;	/* empty second object */
}

void source::rewind()
{
	cur = head;
}

/* Return true if there is more text to retrieve via next(). */
int source::more_text()
{
	if (cur) return(!NULL);
	else return(NULL);
}

/* Return next text string in the object, and step forward. */
char *source::next()
{
	if (!cur) return(NULL);
	struct srcrec *tmp = cur;
	cur = cur->next;
	return(tmp->txt);
}

enum type_t { int_t, charstar_t };
struct sym_element {
	char *name;	/* in the activation criteria expression */
	type_t kind;
	char *field;	/* text to access proc struct field */
};

/*
 * Table used to translate references to WDL environment variable names
 * in activation criteria into references to kernel runtime structures
 * and fields.
 */
static struct sym_element sym_list[] = {
	/* Probably most useful useful terms. */
	{"user",	int_t,		"->p->p_cred->p_svuid"},
	{"prog", 	charstar_t,	"->pathbuf"},
	{"pprog",	charstar_t,	"->opathbuf"},
	{"cwd",		charstar_t,	"->cwd"},

	/* Less useful. */
	{"children",	int_t,		"TBD"},
	{"cputime",	int_t,		"->p->p_cpticks"},
	{"descriptors",	int_t,		"TBD"},
	{"memoryuse",	int_t,		"TBD"},
	{"niceval",	int_t,		"->p->p_nice"},

	/* Possibly not practical. */
	{"uid",		int_t, 		"->p->p_cred->p_ruid"},
	{"pid",		int_t,		"->p->p_pid"},
	{"ppid", 	int_t,		"->p->p_oppid"},
	{"euid",	int_t,		"->p->p_cred->p_svuid"},
	{"egid",	int_t,		"->p->p_cred->p_svgid"},
	{"gid",		int_t,		"->p->p_cred->p_rgid"},
	{"pgid",	int_t,		"->p->p_pgrp->pg_id"},

	/* and more to come ... */

	{NULL, (type_t)NULL, NULL} };

/* Return matching sym_element, if any. */
struct sym_element *is_reserved(char *name)
{
	struct sym_element *cur;
	for (cur = sym_list; cur->name && strcmp(name, cur->name); cur++) {
		/* empty */
	}
	if (cur->name) return(cur);
	else return(NULL);
}

struct op_map {
	enum operators_t op;
	char *int_name;
};

static struct op_map op_map[] = {
	{ not, " ! " },
	{ gt,  " > " },
	{ lt,  " < " },
	{ gte, " >= " },
	{ lte, " <= " },
	{ ne,  " != " },
	{ eq,  " == " },
	{ (enum operators_t) NULL, NULL} };

/* Return string version of a relational operator, or NULL. */
char *op_int_str(enum operators_t op)
{
	struct op_map *cur;
	for (cur = op_map; cur->int_name && (cur->op != op); cur++) {
		/* empty */
	}
	if (cur->int_name) return(cur->int_name);
	else return(NULL);
}

/*
 * First, notice input that refers to environment variables as opposed
 * to input that is a constant value.  Check to make sure that the
 * types of compared values match (char* or int).  Finally, generate
 * C-code that calls the right comparison function depending on the
 * type info.
 */
source *
gen_binary_comparison(char *one, enum operators_t op, char *two)
{
	struct sym_element *sym1 = is_reserved(one);
	struct sym_element *sym2 = is_reserved(two);

	if (sym1 && sym2 && (sym1->kind != sym2->kind)) {
		fprintf(stderr, "type mismatch\n");
		exit(2);
	}
	source *src = new source;
	struct sym_element *sym = NULL;
	if (sym1) sym = sym1;
	if (sym2) sym = sym2;

	if (sym == NULL) {
		/*
		 * The expression only uses constant values; not a meaningful
		 * activation criteria expression.  Error.
		 */
		char *buf = (char*)malloc(strlen(one) + strlen(two) +
			   strlen(op_int_str(op)) + strlen("(  )") + 1);
		sprintf(buf, "(%s %s %s)", one, op_int_str(op), two);
		fprintf(stderr, "Expression \"%s\" must reference a state variable; exiting...\n", buf);
		exit(2);
	}
	else if (!strcmp(one, "user")) {
		/*
		 * Translate "one" to a uid and generate int comparison.
		 * Note: if user names change, activation criteria have
		 * to be remade and reloaded.  If the getpwnam() call were
		 * made by the loading mechanism (in user space), it would
		 * only be necessary to reload activation criteria.  TBD.
		 */
		if ((op != eq) && (op != ne)) {
			fprintf(stderr,
				"invalid comparison operator for user attr\n");
			exit(2);
		}
		src->append("wrproc", sym1->field, op_int_str(op));
		struct passwd *pw;
		if (! (pw = getpwnam(two))) {
			fprintf(stderr, "invalid user name \"%s\"\n", two);
			exit(2);
		}
		char uid_buf[50];
		sprintf(uid_buf, "%d", pw->pw_uid);
		src->append(uid_buf);
	}
	else if (!strcmp(two, "user")) {
		if (op != eq) {
			fprintf(stderr,
				"invalid comparison operator for user attr\n");
			exit(2);
		}
		src->append("wrproc", sym2->field, op_int_str(op));
		struct passwd *pw;
		if (! (pw = getpwnam(one))) {
			fprintf(stderr, "invalid user name \"%s\"\n", one);
			exit(2);
		}
		char uid_buf[50];
		sprintf(uid_buf, "%d", pw->pw_uid);
		src->append(uid_buf);
	}

	/* more special case stuff goes here, then the simple field refs */

	else if (sym->kind == int_t) {
		if (sym1) {
			src->append("wrproc", sym1->field);
		}
		else {
			src->append(one);	/* pass through */
		}
		/* add int operator */
		src->append(op_int_str(op));
		if (sym2) {
			src->append("wrproc", sym2->field);
		}
		else {
			src->append(two);	/* pass through */
		}
	}
	else if (sym->kind == charstar_t) {
		src->append("(*wr_strcmp)(");
		if (sym1) {
			src->append("wrproc", sym1->field);
		}
		else {
			src->append("\"", one, "\"");	/* pass through */
		}
		src->append(", ");
		if (sym2) {
			src->append("wrproc", sym2->field);
		}
		else {
			src->append("\"", two, "\"");	/* pass through */
		}

		switch (op) {
		case gt:
			src->append(") > 0");
			break;
		case lt:
			src->append(") < 0");
			break;
		case gte:
			src->append(") >= 0");
			break;
		case lte:
			src->append(") <= 0");
			break;
		case ne:
			src->append(") != 0");
			break;
		case eq:
			src->append(") == 0");
			break;
		default:
			fprintf(stderr, "invalid binary operator");
			exit(2);
			break;
		}
	}
	return(src);
}

/*
 * Generate C code that adds a wrapper to the list of wrappers this
 * criterion should activate.
 */
source *
gen_arrow(source *expr, char *name, struct imp_head *imp_list)
{
	struct imp_entry *imp = (struct imp_entry*)malloc(
						sizeof(struct imp_entry));
	imp->name = name;
	SLIST_INSERT_HEAD(imp_list, imp, entries);

	source *tmp = new source;
	tmp->append("((");
	tmp->append(expr);
	tmp->append(") && wr_proc_add_wrapper(wrproc, \"");
	tmp->append(name);
	tmp->append("\", myhandle) || 1)");
	delete expr;
	return(tmp);
}

source *
gen_setuid(char *name) {
	source *s = new source;

	if (name == NULL) {
		s->append("(wrproc->setuid_to != -1)");
	}
	else {
		struct passwd *pw;
		if (! (pw = getpwnam(name))) {
			fprintf(stderr, "invalid user name \"%s\"\n", name);
			exit(2);
		}
		char uid_buf[50];
		sprintf(uid_buf, "%d", pw->pw_uid);
		s->append("(wrproc->setuid_to == ", uid_buf, ")");
	}
	return(s);
}

source *
gen_setgid(char *name) {
	source *s = new source;

	if (name == NULL) {
		s->append("(wrproc->setgid_to != -1)");
	}
	else {
		struct group *gr;
		if (! (gr = getgrnam(name))) {
			fprintf(stderr, "invalid group name \"%s\"\n", name);
			exit(2);
		}
		char gid_buf[50];
		sprintf(gid_buf, "%d", gr->gr_gid);
		s->append("(wrproc->setgid_to == ", gid_buf, ")");
	}
	return(s);
}

/*
 * Generate regular expression check code.  Need to decide whether to
 * put Hery Spencer's regex code in kernel.  For now, this function generates
 * code that returns TRUE on equality of strings.
 */
source *
gen_regex(char *str, char *regex) {
	source *s = new source;

	/* first, get the real value of str */
	struct sym_element *sym = is_reserved(str);
	if (! sym) {
		fprintf(stderr, "symbol \"%s\" is not an attribute name\n",
			str);
		exit(2);
	}
	s->append("(! (*wr_strcmp)(wrproc", sym->field, ", ");
	s->append("\"", regex, "\"))");

/*	printf("wrproc%s\n", sym->field);
	s->append("1");*/
	return(s);
}

void source::gen_import_list(struct imp_head *list)
{
	append("\
/* Wrappers referenced by this activation criteron. */ \n\
char *imp_list[] = {\n");

	imp_entry *ep = list->slh_first;
	for (; ep != NULL; ep = ep->entries.sle_next) {
		append("\t\"");
		append(ep->name);
		append("\",\n");
	}
	append("\tNULL\n};\n\n");
}

void source::gen_decision_function(source *expr)
{
	/* generate function header */
	append("\
/*									\n\
 * Decision function, called by WSS at exec.  Assumes list is		\n\
 * initialized; adds elements to it if wrappers should be		\n\
 * activated								\n\
 */									\n\
void activate(struct wr_proc *wrproc)					\n\
{									\n\
	if ( "); append(expr); append(" )				\n\
		/* do nothing */ ;					\n\
}\n");
}

void source::gen_debug_code()
{
	append("\n/* Test driver code. */\n#include \"debug.c\"\n");
}
