/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wr_vm.c,v 1.8 2001/02/23 19:02:22 stevek Exp $
 *
 * Author: Lee Badger
 *
 * User-mode vm subsystem for WQL development.  Chunks of code debugged
 *	here will get ported to the wss and wrapc.
 */

#include <stdio.h>
#include <stdlib.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <unistd.h>

#include "wr_conf.h"
#include "wss_db.h"
#include "dbi.h"

vm_offset_t wss_vm_data;
size_t wss_vm_size;
struct wql_table **wss_vm_root_table;

/*
 * The following mem malloc/free structures adapted from "The C
 * Programming Language" by Kernighan and Ritchie, chapter 8.
 */
typedef int ALIGN;
union header {
	struct {
		union header *next;	/* next free block */
		unsigned size;		/* size of this free block, in units */
	} s;
	ALIGN x;	/* force alignment of blocks */
};
typedef union header HEADER;

/*
 * The first word of the vm region points to the root table.  The next
 * structure is for wss_vm_malloc/wss_vm_free.
 */
struct wsi_vm_header {
	struct wql_table *root_table;	/* relative address of first table  */
	unsigned free_bytes;
	HEADER base;
} *wsi_vm_head;

/*
 * Allocate a region of VM paged to a file, of the given size.  Create
 * the file, if necessary.  The file size is set by lseek'ing past the
 * end of the (initially) empty file (much faster than filling it with
 * writes; hope that ports to kernel space).  If the file already
 * exists, simply map it.
 *
 * On success, returns a pointer to the allocated region,
 * On failure, returns NULL
 *		an existing wss.db file is the wrong size
 *		any file operation failed
 *		the mmap call failed
 */
vm_offset_t wsi_vm_init(size_t size)
{
	int fd;
	struct stat sb;
	off_t offset;
	char *fname = "wss.db";

	if (stat(fname, &sb)) {
		fd = open(fname, O_RDWR | O_CREAT, 0644);
		if (fd < 0) {
			perror("td:");
			return 0;
		}
		offset = lseek(fd, size-1, SEEK_SET);
		wss_vm_size = size;
		write(fd, "x", 1);
		/* printf("created(%d)\n", (int)wss_vm_size); */
	}
	else {
		fd = open(fname, O_RDWR);
		if (fd < 0) {
			perror("td");
			return 0;
		}
		wss_vm_size = sb.st_size;
		if (wss_vm_size != size) {
			printf("existing wss.db file wrong size!\n");
			return 0;
		}
		/* printf("found(%d)\n", (int)wss_vm_size); */
	}
	wss_vm_data = (vm_offset_t)mmap((caddr_t) 0, wss_vm_size,
			(PROT_READ|PROT_WRITE), MAP_SHARED, fd, 0);
	if ((int)wss_vm_data == (int)MAP_FAILED) {
		printf("map failed!\n");
		return 0;
	}
	wsi_vm_head = (struct wsi_vm_header*) wss_vm_data;
	wss_vm_root_table = &wsi_vm_head->root_table;
	/* printf("wss.db(%d bytes free)\n", wsi_vm_head->free_bytes ?
		 (int)wsi_vm_head->free_bytes : 
		 (wss_vm_size - sizeof(struct wsi_vm_header)) / sizeof(HEADER)
			* sizeof(HEADER)) ;  */

#ifdef	notdef
{
	void *wss_vm_malloc();
	char *x1, *x2, *x3;

printf("space(%d)\n", wsi_vm_head->free_bytes);
	x1 = (char*)wss_vm_malloc(40);
	x2 = (char*)wss_vm_malloc(400);
	x3 = (char*)wss_vm_malloc(4000);
printf("space(%d)\n", wsi_vm_head->free_bytes);
	wss_vm_free((void*)x2);
	wss_vm_free((void*)x1);

	wss_vm_free((void*)x3);
printf("space(%d)\n", wsi_vm_head->free_bytes);
}
#endif

	return(wss_vm_data);
}

/*
 * Synchronously write the vm region to disk.
 *
 * Returns NULL on success, !NULL on failure.
 */
int wsi_vm_sync()
{
	if (-1 == msync((caddr_t)wss_vm_data, 0, MS_INVALIDATE)) {
		printf("msync failed");
		return(-1);
	}
	return 0;
}

/*
 * Unmap the VM region, forcing any updates to disk.
 *
 * Returns NULL on success, !NULL on failure.
 */
int wsi_vm_shutdown()
{
	if (-1 == munmap((caddr_t)wss_vm_data, wss_vm_size)) {
		printf("munmap failed");
		return(-1);
	}
	return 0;
}

/* for debugging; prints the size of (absolute) memory gotten from
 * wss_vm_malloc() */
int psize(void *mem)
{
	HEADER *m;

	m = (HEADER*)mem - 1;
	return(m->s.size * sizeof(HEADER));
}

unsigned wr_db_space()
{
	return(wsi_vm_head->free_bytes ? (unsigned)wsi_vm_head->free_bytes : 
		 (wss_vm_size - sizeof(struct wsi_vm_header)) / sizeof(HEADER)
			* sizeof(HEADER));
}

/*
 * The following mem malloc/free code adapted from "The C Programming
 * Language" by Kernighan and Ritchie, chapter 8.
 */

static HEADER *allocp = NULL;	/* last allocated block, an absolute ptr */

/*
 * Like malloc(3), but allocates (absolute ptrs) from the VM region.
 * This routine maintains all internal pointers as relative addresses
 * within the VM region.  This means that memory allocated by DB
 * structures in the VM region will be preserved across runs of the
 * WSS.
 *
 * Returns NULL on failure (i.e., if no memory was available).
 */
void *wss_vm_malloc(unsigned nbytes)
{
	register HEADER *p, *q;
	register int nunits;
	register HEADER *base;

	/* compute the number of needed units */
	nunits = 1 + (nbytes + sizeof(HEADER) - 1) / sizeof(HEADER);

	/* 
	 * If this is the first malloc/free for this run of the wss,
	 * initialize the header.
	 */
	if ((q = allocp) == NULL) {
		base = &(wsi_vm_head->base);
		if (wsi_vm_head->base.s.next == NULL) {
			/* new wss.db file */
			base->s.next = (HEADER*)abs2rel(base);
			base->s.size =
				(wss_vm_size - sizeof(struct wsi_vm_header)) / 
					sizeof(HEADER);
			wsi_vm_head->free_bytes = wsi_vm_head->base.s.size *
				sizeof(HEADER);
		}
		allocp = q = base;
	}

	/* Find a free block: use first one big enough. */
	for (p=(HEADER*)rel2abs(q->s.next);
			; q=p, p=(HEADER*)rel2abs(p->s.next)) {

		if (p->s.size >= nunits) {		/* big enough */
			if (p->s.size == nunits) {	/* exactly */
				q->s.next = p->s.next;	/* both addrs rel */
			}
			else {
				p->s.size -= nunits;
				p += p->s.size;
				p->s.size = nunits;
			}
	/* printf("malloc(%x, %d, free(%d))\n", (int)(p+1), nunits*8,
			(int)wsi_vm_head->free_bytes); */

			wsi_vm_head->free_bytes -= nunits*sizeof(HEADER);
			allocp = q;
			return((void*) (p+1));
		}
		if (p == allocp) {	/* wrapped around free list*/
			return(NULL);	/* out of memory */
		}
	}
}

/*
 * Like free(3), but returns memory to the VM region.
 */
void wss_vm_free(void *addr)
{
	register HEADER *mem, *q;

	mem = (HEADER *)addr - 1;		/* point to header */

/* printf("free(%x, %d, free(%d))\n", (int)addr,
		mem->s.size*8, (int)wsi_vm_head->free_bytes); */

	/* 
	 * If this is the first malloc/free for this run of the wss,
	 * initialize the header.
	 */
	if (allocp == NULL) {
		allocp = &(wsi_vm_head->base);
		if (wsi_vm_head->base.s.next == NULL) {
			printf("BAD Error: trying to free from VM before any allocations\n");
			/* new wss.db file */
			allocp->s.next = (HEADER*)abs2rel(allocp);
			allocp->s.size =
				(wss_vm_size - sizeof(struct wsi_vm_header)) / 
					sizeof(HEADER);
			wsi_vm_head->free_bytes = wsi_vm_head->base.s.size *
				sizeof(HEADER);
		}
	}

	wsi_vm_head->free_bytes += mem->s.size * sizeof(HEADER);

	/* find free blocks adjacent to the free'd memory*/
	for (q=allocp;	!((mem>q) && (mem<(HEADER*)rel2abs(q->s.next)));
			q=(HEADER*)rel2abs(q->s.next)) {
		if ((q >= (HEADER*)rel2abs(q->s.next)) &&
		    ((mem > q) || (mem < (HEADER*)rel2abs(q->s.next)))) {
			break;		/* at one end or the other */
		}
	}
	if (mem + mem->s.size == (HEADER*)rel2abs(q->s.next)) {
		/* join to upper free block */
		mem->s.size += ((HEADER*)rel2abs(q->s.next))->s.size;
		mem->s.next = ((HEADER*)rel2abs(q->s.next))->s.next;
	}
	else {
		mem->s.next = q->s.next;
	}
	if (q + q->s.size == mem) {
		/* join to lower block */
		q->s.size += mem->s.size;
		q->s.next = mem->s.next;
	}
	else {
		q->s.next = (HEADER*)abs2rel(mem);
	}
	allocp = q;
}

void *wsi_malloc(size_t size)
{
    return malloc(size);
}

void wsi_free(void* ptr)
{
    free(ptr);
}
