/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: main.c,v 1.8 2001/02/23 19:02:23 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Debugging program for dumping the contents of a criteria file
 *
 */

#include <wr_conf.h>

#include <stdio.h>

#include <libre.h>
#include <libcrit.h>
#include <errno.h>

static const char *usage_str = "\
dumpcrit <file1> [<file2> ... <filen>]\n\
\n\
    Dump the contents of a criteria file.";

void
Usage()
{
   fprintf( stderr, "%s\n", usage_str );
}

int
DumpCriteria( const char *name )
{
   crit_header_t hdr;
   FILE *f;
   int nread, i;
   int nlit, nexp, nre;
   
   f = fopen( name, "rb" );
   if( f == NULL )
      return errno;

   nread = fread( &hdr, 1, sizeof(hdr), f );
   if( nread != sizeof(hdr) )
   {
      fprintf( stderr, "%s: not a criteria file\n", name );
      fclose( f );
      return 0;
   }

   if( hdr.magic != CRIT_MAGIC )
   {
      fprintf( stderr, "%s: not a criteria file (magic mismatch)\n", name );
      fclose( f );
      return 0;
   }

   printf( "[ %s ]\nVersion: %d\n   Type: %s\n", name, hdr.version,
           hdr.activation?"Activation":"Deactivation" );

   if( hdr.version != CRIT_VERSION )
   {
      printf( "Unable to dump remaining parts due to version mismatch\n" );
      fclose( f );
      return 0;
   }

   nlit = hdr.litsize / sizeof(crit_literal_t);
   nexp = hdr.expsize / sizeof(crit_expression_t);
   nre  = hdr.regextabsize / sizeof(crit_regex_t);

   printf( "\nStatistics:\n" );
   printf( "  %3d Literal(s)\n", nlit );
   printf( "  %3d Expression(s)\n", nexp );
   printf( "  %3d Regular Expression(s)\n", nre );

   if( nlit )
   {
      printf( "\nLiterals:\n [id#] stroff  value     type\n" );
      fseek( f, hdr.literals, SEEK_SET );
      for( i = 0; i < nlit; i++ )
      {
         crit_literal_t lit;
         const char *type;
         if( fread( &lit, 1, sizeof(lit), f ) != sizeof(lit) )
         {
            fclose( f );
            return errno;
         }

         switch( lit.type )
         {
            case CT_INT:
               type = "Integer";
               break;

            case CT_STRING:
               type = "String";
               break;

            case CT_LABEL:
               type = "Label";
               break;

            case CT_BUILTIN:
               type = "Builtin";
               break;

            case CT_NONE:
               type = "None";
               break;
         }

         printf( " [%3d] 0x%4.4x  0x%4.4x -- %s\n", lit.id, lit.soff, lit.ival,
                 type );
      }
   }

   if( nexp )
   {
      printf( "\nExpressions:\n [id#]  Arg #1      Arg #2\n" );
      fseek( f, hdr.expression, SEEK_SET );
      for( i = 0; i < nexp; i++ )
      {
         crit_expression_t exp;
         char a1t, a2t;
         const char *op;
         crit_id_t a1id, a2id;

         if( fread( &exp, 1, sizeof(exp), f ) != sizeof(exp) )
         {
            fclose( f );
            return errno;
         }

         a1t = a2t = 'L';
         op = "??";
         a1id = exp.arg1 & CRIT_ARG_MASK;
         a2id = exp.arg2 & CRIT_ARG_MASK;

         if( exp.arg1 & CRIT_OP_MASK )
            a1t = 'E';
         else if( exp.arg1 & CRIT_RE_MASK )
            a1t = 'R';
         if( exp.arg2 & CRIT_OP_MASK )
            a2t = 'E';
         else if( exp.arg2 & CRIT_RE_MASK )
            a2t = 'R';

         switch( exp.op )
         {
            case CO_AND:
               op = "&&";
               break;

            case CO_OR:
               op = "||"; 
               break;

            case CO_NOT:
               op = " !"; 
               break;

            case CO_LT:
               op = "<"; 
               break;

            case CO_GT:
               op = ">"; 
               break;

            case CO_LTE:
               op = "<="; 
               break;

            case CO_GTE:
               op = ">="; 
               break;

            case CO_NE:
               op = "!="; 
               break;

            case CO_EQ:
               op = "=="; 
               break;

            case CO_RE:
               op = "=~"; 
               break;

            case CO_SUID:
               op = "SU"; 
               break;

            case CO_SGID:
               op = "SG"; 
               break;

            case CO_ARROW:
               op = "==>"; 
               break;

            default:
               break;
         }

         printf( " [%3d] %c 0x%4.4x %-3.3s %c 0x%4.4x (0x%x)\n", i, a1t, a1id, op,
                 a2t, a2id, exp.op );
      }

      printf( "\n L - Literal   E - Expression   R - Regular Expression\n" );
   }

   if( nre )
   {
      printf( "\nRegular Expressions:\n" );
      for( i = 0; i < nre; i++ )
      {
         crit_regex_t re;
         unsigned int j;

         fseek( f, hdr.regextab + i * sizeof(re), SEEK_SET );
         if( fread( &re, 1, sizeof(re), f ) != sizeof(re) )
         {
            fclose( f );
            return errno;
         }

         fseek( f, hdr.regextab + nre * sizeof(re) +
                   re.regex * sizeof(crit_regex_node_t), SEEK_SET );
         printf( "\n  %3d: [id#] [  op  ] [   next   ] [ oneofnext] "
                 "[   data   ]\n", i );
         for( j = 0; j < re.regexsize; j++ )
         {
            crit_regex_node_t  relem;
            const char        *opname;

            if( fread( &relem, 1, sizeof(relem), f ) != sizeof(relem) )
            {
               fclose( f );
               return errno;
            }

            switch( relem.op )
            {
               case RE_NOP:
                  opname = "NOP";
                  break;

               case RE_EXP:
                  opname = "EXP";
                  break;

               case RE_CHAR:
                  opname = "CHAR";
                  break;

               case RE_ANY:
                  opname = "ANY";
                  break;

               case RE_END:
                  opname = "END";
                  break;

               case RE_ONEOF:
                  opname = "ONEOF";
                  break;

               case RE_POSIX:
                  opname = "POSIX";
                  break;

               case RE_RANGE:
                  opname = "RANGE";
                  break;

               case RE_REPEAT:
                  opname = "REPEAT";
                  break;

               case RE_START:
                  opname = "START";
                  break;

               default:
                  break;
            }

            printf( "       [%3d]  %6s   0x%8.8x   0x%8.8x   0x%8.8x\n", j, 
                    opname, relem.next, relem.oneofnext, relem.data );
         }
      }
   }

   fclose( f );

   return 0;
}

int
main( int argc, char *argv[] )
{
   int i;

   if( argc <= 1 )
   {
      Usage();
      return 0;
   }

   for( i = 1; i < argc; i++ )
   {
      if( DumpCriteria( argv[i] ) )
      {
         perror( argv[i] );
         return 0;
      }
   }

   return 0;
}

