/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: pathfind.c,v 1.6 2001/02/23 19:02:23 stevek Exp $
 *
 * author: Steve Kiernan
 *
 * Purpose: Implementation of pathfind function as seen on Solaris
 *
 * DESCRIPTION
 *      pathfind() searches the directories named in <path> for  the
 *      file  <name>.  The directories named in <path> are separated
 *      by colons.  <mode> is a string of option letters chosen from
 *      the set [rwxfbcdpugks]:
 * 
 *           Letter    Meaning
 *           r         readable
 *           w         writable
 *           x         executable
 *           f         normal file
 *           b         block special
 *           c         character special
 *           d         directory
 *           p         FIFO (pipe)
 *           u         set user ID bit
 *           g         set group ID bit
 *           k         sticky bit
 *           s         size nonzero
 * 
 *      Options read, write, and execute are checked relative to the
 *      real (not the effective) user ID and group ID of the current
 *      process.
 * 
 *      If the file <name>, with all the  characteristics  specified
 *      by <mode>,  is found in any of the directories specified  by
 *      <path>,  then  pathfind() returns  a  pointer  to  a  string
 *      containing  the  member  of  <path>,  followed  by  a  slash
 *      character (/), followed by <name>.
 * 
 *      An empty <path> member is treated as the current  directory.
 *      / is  not prepended  at  the  occurrence of the first match;
 *      rather, the unadorned <name> is returned.
 */

#include "wr_conf.h"
#include <stdio.h>
#include <stdlib.h>

#ifdef WR_UNIX
#include <strings.h>
#include <unistd.h>
#include <sys/param.h>
#endif

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef WR_WINNT
#include <direct.h>
#endif


const char *pathfind( const char *path, const char *name, const char *mode )
{
   char *pathstr;
   char *tok;

#ifdef WR_UNIX
   mode_t flags = 0;
   int nzero = 0;

   for( ; mode && *mode; mode++ )
      switch( *mode )
      {
         case 'r': flags |= S_IROTH; break;
         case 'w': flags |= S_IWOTH; break;
         case 'x': flags |= S_IXOTH; break;
         case 'f': flags |= S_IFREG; break;
         case 'b': flags |= S_IFBLK; break;
         case 'c': flags |= S_IFCHR; break;
         case 'd': flags |= S_IFDIR; break;
         case 'p': flags |= S_IFIFO; break;
         case 'u': flags |= S_ISUID; break;
         case 'g': flags |= S_ISGID; break;
         case 'k': flags |= S_ISVTX; break;
         case 's': nzero = 1; break;
         default: errno = -EINVAL; return NULL;
      }

#endif

   pathstr = (char *)( (path!=NULL)?strdup( path ):getcwd( NULL, MAXPATHLEN ) );

   for( tok = strtok( pathstr, ":" ); tok; tok = strtok( NULL, ":" ) )
   {
      static char cpath[MAXPATHLEN];
      struct stat st;

#ifdef WR_UNIX
      mode_t maskMode = flags & ~0xfff;
#endif

      snprintf( cpath, MAXPATHLEN, "%s/%s", tok, name );
      if( stat( cpath, &st ) < 0 ) continue;

#ifdef WR_UNIX
      if( ( st.st_mode & maskMode ) == maskMode )
      {
         if( ( flags & S_IROTH ) &&
             !( st.st_mode & S_IROTH ) &&
             !( ( st.st_mode & S_IRUSR ) && ( st.st_uid == getuid() ) ) &&
             !( ( st.st_mode & S_IRGRP ) && ( st.st_uid == getgid() ) ) )
            continue;
         if( ( flags & S_IWOTH ) &&
             !( st.st_mode & S_IWOTH ) &&
             !( ( st.st_mode & S_IWUSR ) && ( st.st_uid == getuid() ) ) &&
             !( ( st.st_mode & S_IWGRP ) && ( st.st_uid == getgid() ) ) )
            continue;
         if( ( flags & S_IXOTH ) &&
             !( st.st_mode & S_IXOTH ) &&
             !( ( st.st_mode & S_IXUSR ) && ( st.st_uid == getuid() ) ) &&
             !( ( st.st_mode & S_IXGRP ) && ( st.st_uid == getgid() ) ) )
            continue;
      }

#endif

      free( pathstr );
      return( (const char *)((path==NULL)?name:cpath) );
   }

   free( pathstr );
   return NULL;
}

