/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wreview.c,v 1.30 2001/02/23 19:02:23 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Wrapper Viewer event viewer
 *
 * This application registers itself with the WSS, sets up a
 * raw event handler, and displays events as they are posted.
 */

#include <wr_conf.h>

#ifndef LINKAGE
#define LINKAGE IMPORT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#ifdef WR_WINNT
#include "getopt.h"
#else
#include <signal.h>
#include <unistd.h>
#include <ctype.h>
#endif /* WR_WINNT */

#include <eventfs.h>
#include <libeventfs.h>
#include <libwrv.h>

/* For _PATH_EVENTFS */
#include EFS_SYSTEM_HDR

void errorMessage( char *msg );

#define MAXBUFFERLEN 2048
char *binToAscii(char *buffer, unsigned long buffer_length)
{
    static char output[MAXBUFFERLEN]; /* Truncate if too long */
    size_t offset_in, offset_out;

    /* If the character is ascii, print it.
     * Else print '\xxx' where xxx is the octal representation.
     */
    for (offset_in=0, offset_out=0 ;
            ( offset_in < buffer_length ) && ( offset_out < (MAXBUFFERLEN-1));
            ++offset_in ) {
        if (isprint((int)buffer[offset_in])
                || buffer[offset_in] == '\n'
                || buffer[offset_in] == '\t') {
            output[offset_out] = buffer[offset_in];
            ++offset_out;
        } else {
            _sntprintf( output+offset_out,MAXBUFFERLEN - offset_out - 1,
                        PD_MAKESTR("\\%3.3o"), (int)buffer[offset_in] );
            offset_out += 4;
        }
    }

    if (offset_out >= MAXBUFFERLEN) {
        /* indicate so in the output */
        output[MAXBUFFERLEN - 4] = '.';
        output[MAXBUFFERLEN - 3] = '.';
        output[MAXBUFFERLEN - 2] = '.';
    }

    output[offset_out] = '\0';

    return output;
}

#ifdef WR_UNIX
#define errorMessage(msg) perror(msg)
#else
void errorMessage( char *msg )
{
   LPVOID lpMsgBuf;

   FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                  FORMAT_MESSAGE_FROM_SYSTEM |
                  FORMAT_MESSAGE_IGNORE_INSERTS,
                  NULL, GetLastError(),
                  MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
                  (LPTSTR) &lpMsgBuf, 0, NULL );
   _tprintf( _TEXT("%s: %s\n"), msg, (LPCTSTR) lpMsgBuf );
   LocalFree( lpMsgBuf );
}
#endif

void dumpEvent(eventfs_event_stream_t stream, eventfs_event_t event)
{
    /* Print out group name, type name, buffer. */
    u_long              mask;
    int                 i;
    const char         *name;
    eventfs_event_group_t   group;

    mask = eventfsEventGetTypes(event);
    group = eventfsEventStreamGetGroup(stream);

    printf("Event: group = %-20.20s Mask = %-8.8lo Size = %d\n",
            eventfsEventGroupGetName(group), mask, eventfsEventGetSize(event));
    printf("Types:\n");

    for (i = 0; i < 32; ++i) {
        if (mask & 1<<i) {
            name = eventfsEventGroupGetTypeName(group,1<<i);
            if (name)
                printf("\t%s\n",name);
            else
                printf("\tinvalid type\n");
        }
    }
    printf("Buffer:\n");
    puts(
            binToAscii(
                eventfsEventGetData(event),
                eventfsEventGetSize(event)));

}

int main( int argc, char *argv[] )
{
    u_long              mask          = 0xFFFFFFFF;
    char               *group_name    = strdup(WREV_EFS_GROUP);
    char               *stream_name   = 0;
    char               *base_path     = strdup(_PATH_EVENTFS);
    int                 ch;
    /* Stream argument is not implimented yet */

    eventfs_event_group_t   group;
    eventfs_event_stream_t  stream;
    eventfs_event_t         event;

    while( ( ch = getopt( argc, argv, "b:g:m:s:" ) ) != -1 )
    {
        switch( ch )
        {
        case 'm':
            mask = atoi(optarg);
            break;
        case 'g':
            free(group_name);
            group_name = strdup(optarg);
            break;
        case 's':
            if (stream_name) free(stream_name);
            stream_name = strdup(optarg);
        case 'b':
            free(base_path);
            base_path = strdup(optarg);
            break;

        default:
            break;
        }
    }

    argc -= optind;
    argv += optind;

    if (stream_name) {
        char *pathname;
        int   grouplen, streamlen;
        grouplen = strlen(group_name); streamlen = strlen(stream_name);
        pathname = malloc(grouplen + 1 + streamlen + 1 + 4 + 1);
        strcpy(pathname, group_name);
        pathname[grouplen] = '/';
        strcpy(pathname+grouplen+1,stream_name);
        pathname[grouplen+1+streamlen] = '/';
        strcpy(pathname+grouplen+1+streamlen+1,"data");
        pathname[grouplen+1+streamlen+1+4] = '\0';
        stream = eventfsEventGetStreamByFile(pathname);
    } else {
        /*
         * Open the group
         */
        if( (group = eventfsEventGetGroup(group_name)) == NULL ) {
            errorMessage( "eventfsEventGetGroup" );
            return EXIT_FAILURE;
        }

        /*
         * Open the stream
         */
        if ( (stream = eventfsEventGroupGetStream(group, mask)) == NULL) {
            errorMessage( "eventfsEventGroupGetStream" );
            return EXIT_FAILURE;
        }
    }

    /*
     * Sleep forever...
     */
    _tprintf( PD_MAKESTR("Waiting for events...\n") );
    for(;;)
    {
        if( (event = eventfsEventStreamGetNextEvent(stream)) != 0 )
        {
            dumpEvent( stream, event );
            eventfsEventRelease( event );
        }
        /* This only happens on a significant malfunction. */
        else
            break;
    }

    return EXIT_FAILURE;
}

