/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wrselect.c,v 1.29 2001/06/13 13:25:30 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Program to return selected database information of the WSS.
 * 
 */

#define LINKAGE IMPORT

#include <wr_conf.h>

#include <stdarg.h>
#include <wsi_stdlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>
#include <time.h>
#ifndef WR_WINNT
#include <unistd.h>
#include <pwd.h>
#else
#include "getopt.h"
#endif

#include <wrappers.h>
#include <wrerrno.h>
#include "libwrv.h"

#define DEFAULT_BUFSIZE 1024
#define PAD_SIZE 512
#define DEFAULT_ROWSIZE 16
#define DEFAULT_COLSIZE 16

#define FORMAT_STRING_BLOCK(str,size) (str)+=(strlen(str)>(size))?(size):(strlen(str))

#define NOEXIT 0
#define EXIT 1

#define STRIP_FRONT 0x01
#define STRIP_BACK  0x02
#define STRIP_BOTH  (STRIP_FRONT | STRIP_BACK)


static int debug = 0;

#ifdef WR_WINNT
int
strcasecmp( LPTSTR s1, LPTSTR s2 )
{
   if( !s1 && !s2 ) return 0;
   if( !s1 ) return -1;
   if( !s2 ) return 1;
      
   while( *s1 && *s2 && ( tolower(*s1) == tolower(*s2) ) )
      s1++, s2++;

   return( tolower(*s1) - tolower(*s2) );
}
#endif /* WR_WINNT */


/*
 * Prints available tables; Deallocates rowlist when it's done.
 */
void
DisplayTables( wrv_table_id_t *tabid )
{
   wrv_table_id_t *id;

   _tprintf( PD_MAKESTR("Handle      Table\n") );
   _tprintf( PD_MAKESTR("----------  --------------------------------------------------\n") );
   for( id = tabid; id; id = id->next )
   {
      if( id->handle != 0 )
      {
         /* Normally would be "!= NULL"... but id->handle
          * is a pointer cast as a long type
          */

#ifdef WR_WINNT
         id->name = (pd_char_t *)( (pd_byte_t *)tabid + (DWORD)id->name );
#endif /* WR_WINNT */

         _tprintf( PD_MAKESTR("0x%8.8x  %s\n"), id->handle, id->name );
      }
      else
         break;

#ifdef WR_WINNT
      if( id->next )
         id->next = (wrv_table_id_t *)( (pd_byte_t *)tabid + (DWORD)id->next );
#endif /* WR_WINNT */
   }

   wrvDbReleaseTableIds( tabid );
}


/*
 * Prints column names in specified table; Deallocates rowlist when it's done.
 */
void
DisplayColumns( pd_char_t *name, wrdb_coldef_t coldef, int col_count )
{
   int i;

   printf( PD_MAKESTR("Columns in table %s\n\n"), name );
   printf( PD_MAKESTR("Type         Name                                  Max. Length\n") );
   printf( PD_MAKESTR("-----------  ------------------------------------- -----------\n") );
   for( i = 0 ; i < col_count; i++ )
   {
      const pd_char_t *type;
      switch( coldef[i].dtype )
      {
         case pd_char_t_t:
            type = PD_MAKESTR("String");
            break;
         case int_t:
            type = PD_MAKESTR("Integer");
            break;
         case float_t:
            type = PD_MAKESTR("Float");
            break;
         default:
            break;
      }
#ifdef WR_WINNT
      coldef[i].name = (pd_char_t *)( (pd_byte_t *)coldef +
                                       (DWORD)coldef[i].name );
#endif /* WR_WINNT */

      _tprintf( PD_MAKESTR("%-11s  %-37.37s %d\n"), type, coldef[i].name,
                coldef[i].max_len );
   }

   wrvDbReleaseColumnDefinition( coldef );
}


/*
 * Prints selected column data in specified table; Deallocates table when it's done.
 * 
 * colcount is the number of columns selected for viewing.
 * col_count is the number of columns in the original table.
 * Sorry for the inconvience.
 */
void
DisplayTableData( pd_char_t *name, int colcount, pd_char_t **columns,
                  wrdb_coldef_t coldef, wrdb_row_t rows )
{
   int csize[DEFAULT_COLSIZE];
   int corder[DEFAULT_COLSIZE];
   int i, r, col_count;
   wrdb_row_t row;

   printf( "Selected data from table %s\n\n", name );
        
   if( rows == NULL )
   {
      printf( "Empty\n");
      return;
   }
   col_count = ROW_NUMVALS(rows);

#ifdef WR_WINNT
   for( i = 0; i < col_count ; i++ )
   {
      coldef[i].name = (pd_char_t *)( (pd_byte_t *)coldef +
                                       (DWORD)coldef[i].name );
   }
#endif /* WR_WINNT */

   /*
    * Determine column sizes and headings; print headings
    */
   if( colcount == 0 )
   {
      for( i = 0; i < col_count ; i++ )
      {
         corder[i] = i;

         if( coldef[i].max_len == 0 )
            coldef[i].max_len = 10;
         csize[i] = (strlen(coldef[i].name)>(size_t)coldef[i].max_len)?
               strlen(coldef[i].name):coldef[i].max_len;
         if (i) printf("  ");

         printf("%-*.*s", csize[i], csize[i], coldef[i].name);
      }
      printf("\n");
      colcount = i;
   }
   else
   {
      int c;
      for( c = 0; c < colcount; c++ )
      {
         corder[c] = -1;
         csize[c] = 0;

         for( i = 0; i < col_count ; i++ )
         {
            if( strcasecmp(columns[c], coldef[i].name) == 0 )
            {
               corder[c] = i;
               csize[c] = (strlen(coldef[i].name)>(size_t)coldef[i].max_len)?
                      strlen(coldef[i].name):coldef[i].max_len;
               if (c) printf("  ");

               printf("%-*.*s", csize[c], csize[c], coldef[i].name);
               break;
            }
         }
      }
      printf("\n");
   }

   /*
    * Print dividing lines for headings
    */
   for( i = 0; i < colcount; i++ )
   {
      int c;

      if( i ) printf( "  " );
      for( c = 0; c < csize[i]; c++ )
         printf( "-" );
   }
   printf( "\n" );

   for( row = rows, r = 0; row; row = ROW_NEXT(row), r++ )
   {
      wrdb_value_t val;
      int c;

      val = (wrdb_value_t)(ROW(row)->vals);
#ifdef WR_WINNT
      val = (wrdb_value_t)( (pd_byte_t *)rows + (DWORD)val );
#endif /* WR_WINNT */

      for( c = 0; c < colcount; c++ )
      {
         char tmpbuf[64];
         if( c ) printf( "  " );
                        
         if( corder[c] < 0 || corder[c] >= col_count )
            continue;

         switch( coldef[corder[c]].dtype )
         {
            case pd_char_t_t:
            {
               int j;
               pd_char_t *s;
               j = corder[c];
               s = val[j].v.cval;
               j = csize[c];
#ifdef WR_WINNT
               s = (pd_char_t *)( (pd_byte_t *)rows + (DWORD)s );
#endif /* WR_WINNT */
               printf("%-*.*s", j, j, s);
               break;
            }
            case int_t:
               sprintf(tmpbuf, "%d", val[corder[c]].v.ival);
               printf("%*.*s", csize[c], csize[c], tmpbuf);
               break;
            case float_t:
               sprintf(tmpbuf, "%f", val[corder[c]].v.fval);
               printf("%*.*s", csize[c], csize[c], tmpbuf);
               break;
            case none_t:
            default:
               printf("%-*.*s", csize[c], csize[c], "none");
         }
      }

#ifdef WR_WINNT
      if( ROW(row)->next != NULL )
         ROW(row)->next = (struct wql_row *)( (pd_byte_t *)rows +
                                              (DWORD)ROW_NEXT(row) );
#endif /* WR_WINNT */

      printf( "\n" );
   }
}


/*
 * Prints the usage information and exits the program with an error.
 */
void
Usage( int doexit )
{
   printf("wrselect");
   printf("\t-T Show all available tables\n");
   printf("\t\t-c <table> Show column names and types for table <table>\n");
   printf("\t\t-d <table> Dump data from table <table>\n");
   printf("\t\t\t[Note: for -c and -d you can also use the table\n");
   printf("\t\t\t       handle in place of the table name]\n");
   printf("\t\t-h Print help (this message)\n\n");
   printf("\t\tWithout any parameters, wrselect runs in interactive mode,\n");
   printf("\t\twhich simulates an SQL terminal interface to the database.\n\n");
   printf("\t\t\tAvailable SQL commands are:\n");
   printf("\t\t\tSHOW TABLES;\n");
   printf("\t\t\tSHOW COLUMNS IN <table>;\n");
   printf("\t\t\tSELECT (*|<column>[,<column>]) FROM <table>;\n");
	printf("\t\t\tDELETE FROM <table>;\n");
	printf("\t\t\tINSERT INTO <table> VALUES (<value>[,<value>]);\n\n");
   printf("\t\t\tQUIT\n");
   printf("\t\t\tHELP\n");
   if( doexit )
      exit(-1);
}


/*
 *
 */
unsigned long
HexToLong( pd_char_t *str )
{
   unsigned long retval = 0;

   for( ; str && *str; str++ )
   {
      retval <<= 4;
      if( *str >= '0' && *str <= '9' )
         retval += *str - '0';
      else if( *str >= 'A' && *str <= 'F' )
         retval += *str - 'A' + 10;
      else if( *str >= 'a' && *str <= 'f' )
         retval += *str - 'a' + 10;
      else
         return retval;
   }

   return retval;
}


/*
 * Like the libc strsep, but continues until it finds the end of whitespace
 */
pd_char_t *
mystrsep( pd_char_t **str, pd_char_t *delims )
{
   int matched = 0;
   pd_char_t *ptr, *orig;
   
   orig = *str;

   for( ptr = orig; *ptr; ptr++ )
   {
      pd_char_t *d;
      for( d=delims; *d; d++ )
      {
         if( *ptr == *d )
         {
            *ptr = '\0';
            matched = 1;
            break;
         }
      }
      if( *ptr && matched ) break;
   }
   *str = ptr;
   return( orig );
}


/*
 * Provided to strip whitespace characters before and/or after
 * the specified delimiter character
 */
pd_char_t *
mystrstrip( pd_char_t **str, char delim, pd_char_t *whitespace, int dir )
{
   pd_char_t *ptr;

   for( ptr = *str; *ptr; ptr++ )
   {
      if( *ptr == delim )
      {
         if( dir & STRIP_FRONT )
         {
            pd_char_t *front;
            int done = 0;

            for( front = ptr - 1; !done && (front >= *str); front-- )
            {
               pd_char_t *d;
               int found = 0;
               for( d = whitespace; *d; d++ )
                  if( *front == *d )
                     found = 1;
               if( !found )
               {
                  done = 1;
                  break;
               }
            }

            if( done ) strcpy( ++front, ptr );
            ptr = front;
         }
         if( dir & STRIP_BACK )
         {
            pd_char_t *back;
            int done = 0;

            for( back = ptr + 1; !done && *back; back++ )
            {
               pd_char_t *d;
               int found = 0;
               for( d = whitespace; *d; d++ )
                  if( *back == *d )
                     found = 1;
               if( !found )
               {
                  done = 1;
                  break;
               }
            }

            if( done ) strcpy(ptr + 1, back);
         }
      }
   }
   return( *str );
}

/*
 * Interactive command processor
 */
void
Interactive()
{
   char cmd[DEFAULT_BUFSIZE];
   int quit = 0;
   pd_char_t *statement = NULL;
   int alloced = 0;

   while( !quit )
   {
      int eoc = 0;
      int eos = 0;
      int lines = 0;
      int errors = 0;
      pd_char_t *args;

      printf("wrselect> ");

      while( !eos )
      {
         pd_char_t *stmt;
         pd_char_t *ptr;

         fgets(cmd, DEFAULT_BUFSIZE, stdin);

         if( feof(stdin) )
         {
            eoc = quit = 1;
            break;
         }

         /* Trim leading and trailing whitespace */
         for (stmt = cmd; *stmt && ((*stmt == ' ') || (*stmt == '\t')); stmt++);
         for (ptr = (cmd+strlen(cmd)-1);
              (ptr >= stmt) && ((*ptr == ' ') || (*ptr == '\t') || (*ptr == '\n'));
              ptr--);
         *(ptr+1) = '\0';

         if( strlen(stmt) <= 0 )
         {
            if( lines==0 )
               printf("wrselect> ");
            else
               printf("%8d> ", lines);
            continue;
         }

         /*
          * Check for local commands
          */
         if( lines == 0 )
         {
            switch( tolower(*cmd) )
            {
               case 'q':
                  if( (strcasecmp(cmd, "quit") == 0) ||
                      (strcasecmp(cmd, "q") == 0) )
                  {
                     if( statement )
                        free((void *)statement);
                     return;
                  }
                  break;
               case 'h':
                  if( (strcasecmp(cmd, "help") == 0) ||
                      (strcasecmp(cmd, "h") == 0) )
                  {
                     Usage( NOEXIT );
                     eos = 1;
                     eoc = 1;
                     continue;
                  }
                  break;
               case 'd':
                  if( strcasecmp(cmd, "debug") == 0 )
                  {
                     debug = !debug;
                     printf("Debugging %sabled.\n", (debug)?"en":"dis");
                     eos = 1;
                     eoc = 1;
                     continue;
                  }
                  break;
            }
         }

         /* Check for end-of-statement character */
         if( *ptr == ';' )
         {
            *ptr = '\0';
            eos = 1;
         }

         /* Add line to statement */
         if( alloced == 0 )
         {
            alloced = sizeof(char)*(strlen(stmt)+1);
            statement = (pd_char_t *)malloc(alloced);
            bzero(statement, alloced);
         }
         else if( (strlen(statement)+strlen(stmt)) > (size_t)alloced )
         {
            alloced += sizeof(char)*(strlen(stmt)+1);
            statement = (pd_char_t *)realloc(statement,alloced);
         }
         if( lines && *stmt )
            strcat(statement, " ");
         strcat( statement, stmt );
         
         if( !eos )
            printf( "%8d> ", ++lines );
      }

      if( eoc ) continue;

      /*
       * Determine command (chop off the first 'word')
       */
      args = statement;
      mystrsep( &args, " \t" );

      if( debug )
         fprintf(stderr, "Command: %s\n", statement);

      switch( tolower(*statement) )
      {
         case 'd':
            /*
             * DELETE FROM <database> [WHERE <condition>]
             */
            if( strcasecmp(statement, "delete") == 0 )
            {
               pd_char_t      *database;
               wrv_table_id_t *id;
               wrdb_row_t      row;
               int             r;

               mystrstrip(&args, ',', " \t", STRIP_BOTH);
               database = mystrsep(&args, " \t");

               if (strcasecmp(database, "from") != 0) break;

               database = mystrsep(&args, " \t");
               if( strncmp(database, "0x", 2) == 0 )
               {
                  unsigned long handle = HexToLong( database+2 );
                  id = wrvDbGetTableIdByHandle( handle );
               }
               else
                  id = wrvDbGetTableIdByName( database );

               if( id == NULL )
               {
                  printf("Invalid table '%s'\n", database);
                  break;
               }

#ifdef WR_WINNT
               id->name = (pd_char_t *)( (pd_byte_t *)id + (DWORD)id->name );
#endif /* WR_WINNT */

               for( row = wrvDbGetRows( id->handle ), r = 0;
                    row;
                    row = ROW_NEXT(row), r++ )
                  wrvDbDelTableRow( id->handle, ROW_HANDLE(row) );

               printf("%d rows deleted.\n", r);

               wrvDbReleaseTableIds( id );
               break;
            }
            errors++;
            break;

         case 'i':
            /*
             * INSERT INTO <database> VALUES (<value>[,<value>])
             */
            if( strcasecmp(statement, "insert") == 0 )
            {
               pd_char_t      *values[DEFAULT_COLSIZE+1];
               pd_char_t      *value;
               pd_char_t      *valuelist;
               pd_char_t      *database;
               int             valcount = 0;
               int             size;
               wrv_table_id_t *id;
               wrdb_coldef_t   cd;

               mystrstrip(&args, ',', " \t", STRIP_BOTH);
               mystrstrip(&args, '(', " \t", STRIP_BOTH);
               mystrstrip(&args, ')', " \t", STRIP_BOTH);

               database = mystrsep(&args, " \t");

               if (strcasecmp(database, "into") != 0) break;

               database = mystrsep(&args, " \t");

               if( strncmp(database, "0x", 2) == 0 )
               {
                  unsigned long handle = HexToLong( database+2 );
                  id = wrvDbGetTableIdByHandle( handle );
               }
               else
                  id = wrvDbGetTableIdByName( database );

               if( id == NULL )
               {
                  printf("Invalid table '%s'\n", database);
                  break;
               }

#ifdef WR_WINNT
               id->name = (pd_char_t *)( (pd_byte_t *)id + (DWORD)id->name );
#endif /* WR_WINNT */

               valuelist = mystrsep(&args, " \t(");
               if (strcasecmp(valuelist, "values") != 0)
               {
                  wrvDbReleaseTableIds( id );
                  break;
               }

               cd = wrvDbGetColumnDefinition( id->handle, &size );
               if( cd == NULL )
               {
                  printf("Unable to get column definitions\n");
                  wrvDbReleaseTableIds( id );
                  break;
               }

               valuelist = mystrsep(&args, ")");

               do
               {
                  value = mystrsep(&valuelist, ",");

                  if( ++valcount > DEFAULT_COLSIZE ) break;

                  values[valcount-1] = value;
               }
               while( *valuelist != '\0' );
               values[valcount] = NULL;

               if( valcount != size )
               {
                  printf("Incorrect number of values, expecting %d\n", size);
                  wrvDbReleaseColumnDefinition( cd );
                  wrvDbReleaseTableIds( id );
                  break;
               }

               if( wrvDbAddTableRow( id->handle, values ) == 0 )
                  printf("Inserted 1 row.\n");
               else
                  errors++;
               wrvDbReleaseColumnDefinition( cd );
               wrvDbReleaseTableIds( id );
               break;
            }
            errors++;
            break;

         case 's':
            /*
             * SELECT [<column>[,<column>]] FROM <database>
             */
            if( strcasecmp(statement, "select") == 0 )
            {
               pd_char_t *columns[DEFAULT_COLSIZE];
               pd_char_t *column;
               pd_char_t *columnlist;
               int colcount = 0;
               int allcols = 0;
               wrv_table_id_t *id;

               mystrstrip(&args, ',', " \t", STRIP_BOTH);
               columnlist = mystrsep(&args, " \t");

               /*
                * Split required columns
                */
               do
               {
                  column = mystrsep(&columnlist, ",");

                  if( strcmp(column, "*") == 0 )
                  {
                     allcols = 1;
                     break;
                  }
                  if( ++colcount > DEFAULT_COLSIZE ) break;

                  columns[colcount-1] = column;
               }
               while( *columnlist != '\0' );

               if( debug )
                  fprintf(stderr, "Selecting %d potential columns\n", colcount);

               /*
                * Process the 'FROM <database>' statement
                */
               column = mystrsep(&args, " \t");

               if (strcasecmp(column, "from") != 0) break;

               if( strncmp(args, "0x", 2) == 0 )
               {
                  unsigned long handle = HexToLong( args+2 );
                  id = wrvDbGetTableIdByHandle( handle );
               }
               else
                  id = wrvDbGetTableIdByName( args );

               if( id == NULL )
               {
                  printf("Invalid table '%s'\n", args);
                  break;
               }

#ifdef WR_WINNT
               id->name = (pd_char_t *)( (pd_byte_t *)id + (DWORD)id->name );
#endif /* WR_WINNT */

               DisplayTableData( id->name, allcols?0:colcount, columns,
                                 wrvDbGetColumnDefinition( id->handle, NULL ),
                                 wrvDbGetRows( id->handle ) );
               break;
            }
            else
            /*
             * SHOW (TABLES|COLUMNS IN <database>)
             */
            if( strcasecmp(statement, "show") == 0 )
            {
               pd_char_t *whatshow;


               /*
                * Determine what to show (chop off next 'word')
                */
               whatshow = mystrsep(&args, " \t");

               if (debug)
                  fprintf(stderr, "What to show: %s\n", whatshow);

               switch( tolower(*whatshow) )
               {
                  case 't':
                     if( strcasecmp(whatshow, "tables") == 0 )
                     {
                        DisplayTables( wrvDbGetTableIds() );
                        goto jump_getout; /* Ugly, I know, but easier */
                     }
                     break;
                  case 'c':
                     if( strcasecmp(whatshow, "columns") == 0 )
                     {
                        wrv_table_id_t *id;
                        int             size;
                        wrdb_coldef_t   cd;
                        char           *inptr = mystrsep(&args, " \t");

                        if( strcasecmp(inptr, "in") != 0 ) break;

                        if( strncmp(args, "0x", 2) == 0 )
                        {
                           unsigned long handle = HexToLong( args+2 );
                           id = wrvDbGetTableIdByHandle( handle );
                        }
                        else
                           id = wrvDbGetTableIdByName( args );

                        if( id == NULL )
                        {
                           printf("Invalid table '%s'\n", args);
                           break;
                        }

#ifdef WR_WINNT
                        id->name = (pd_char_t *)( (pd_byte_t *)id +
                                                   (DWORD)id->name );
#endif /* WR_WINNT */

                        cd = wrvDbGetColumnDefinition( id->handle, &size );
                        DisplayColumns( id->name, cd, size );

                        goto jump_getout;
                     }
                     break;
                  default:
                     break;
               }
            }
            errors++;
            break;
         default:
            errors++;
            break;
      }

jump_getout:
      if (errors)
         fprintf(stderr, "There %s %d %s encountered with your statement\n",
            (errors>1)?"were":"was", errors, (errors>1)?"errors":"error");

      errors = 0;
      if (statement)
         bzero(statement, alloced);
   }
}


/*
 * The main routine doesn't do much except parse the arguments and
 * call the correct display routine.
 */
int
main( int argc, char **argv )
{
   int c;

   if( !wrvIsLoaded() )
   {
      fprintf( stderr, "wrselect: WSS not loaded.\n" );
      exit( 1 );
   }

   while( (c = getopt(argc, argv, "Thc:d:")) != EOF )
   {
      switch(c)
      {
         case 'T':
            DisplayTables( wrvDbGetTableIds() );
            break;
         case 'c':
         {
            wrv_table_id_t *id;
            wrdb_coldef_t   cd;
            int             size;

            if( strncmp(optarg, "0x", 2) == 0 )
            {
               unsigned long handle = HexToLong( optarg+2 );
               id = wrvDbGetTableIdByHandle( handle );
            }
            else
               id = wrvDbGetTableIdByName( optarg );

            if( id == NULL )
            {
               printf( "Invalid table '%s'\n", optarg );
               break;
            }

#ifdef WR_WINNT
            id->name = (pd_char_t *)( (pd_byte_t *)id + (DWORD)id->name );
#endif /* WR_WINNT */

            cd = wrvDbGetColumnDefinition( id->handle, &size );
            DisplayColumns( id->name, cd, size );
            break;
         }
         case 'd':
         {
            wrv_table_id_t *id;
            if( strncmp(optarg, "0x", 2) == 0 )
            {
               unsigned long handle = HexToLong( optarg+2 );
               id = wrvDbGetTableIdByHandle( handle );
            }
            else
               id = wrvDbGetTableIdByName( optarg );

            if( id == NULL )
            {
               printf( "Invalid table '%s'\n", optarg );
               break;
            }

#ifdef WR_WINNT
            id->name = (pd_char_t *)( (pd_byte_t *)id + (DWORD)id->name );
#endif /* WR_WINNT */

            DisplayTableData( id->name, 0, NULL,
                              wrvDbGetColumnDefinition( id->handle, NULL ),
            wrvDbGetRows( id->handle ) );
            break;
         }
         case 'h':
            Usage( EXIT );
            break;
         default:
            printf("default???");
            Usage( EXIT );
            break;
      }
   }

   if (argc < 2)
      Interactive();

   return(0);
}

