/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: MainFrame.java,v 1.18 2001/04/05 18:41:21 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import jclass.bwt.*;
import com.gswtk.ui.*;
import com.gswtk.viewer.*;
import com.gswtk.wrapper.event.*;

public class MainFrame extends Frame
{
	protected WRVPanelNotebook  notebook;
	protected Menu		    pageMenu;
	protected CriteriaBox	    critBox;
	protected ModuleCache       modCache;
	protected ModuleEventStream stream;

	// Constructors
	public MainFrame()
	{
		super( "WSS Wrapper Viewer" );

		// Construct the module cache
		modCache = new ModuleCache(true);

		// Build the menu bar
		Font menuFont = new Font( "Helvetica", Font.PLAIN, 12 );
		MenuBar menuBar = new MenuBar();
		menuBar.setFont( menuFont );

		// Set up the panel notebook
		notebook = new WRVPanelNotebook();
		notebook.setInsets( new Insets( 2, 2, 2, 2 ) );
		final DatabasePanel dbPanel = new DatabasePanel( notebook, menuBar );
		ModulesPanel modulesPanel = new ModulesPanel( notebook, menuBar, modCache );
		ActivationPanel activationPanel = new ActivationPanel( notebook, menuBar, modCache );
		DeactivationPanel deactivationPanel = new DeactivationPanel( notebook, menuBar, modCache );
		ProcessesPanel processesPanel = new ProcessesPanel( notebook );
		WrappersPanel wrappersPanel = new WrappersPanel( notebook, menuBar, modCache );
		LibrariesPanel librariesPanel = new LibrariesPanel( notebook, menuBar, modCache );
		AboutPanel aboutPanel = new AboutPanel( notebook );

		// Set up file menu
		MenuItem fileLoad = new MenuItem( "Load Wrapper", new MenuShortcut( 'L' ) );
		fileLoad.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) { loadWrapper(); }
		} );
		fileLoad.setFont( menuFont );

		MenuItem fileAdd = new MenuItem( "Add Criteria", new MenuShortcut( 'A' ) );
		fileAdd.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) { addCriteria(); }
		} );
		fileAdd.setFont( menuFont );

		MenuItem fileExit = new MenuItem( "Exit", new MenuShortcut( 'x' ) );
		fileExit.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) { dispose(); System.exit( 0 ); }
		} );
		fileExit.setFont( menuFont );

		Menu fileMenu = new Menu( "File" );
		fileMenu.setFont( menuFont );
		fileMenu.add( fileLoad );
		fileMenu.add( fileAdd );
		fileMenu.addSeparator();
		fileMenu.add( fileExit );

		// Set up menubar
		menuBar.add( fileMenu );

		// Set up frame
		setMenuBar( menuBar );
		add( notebook );

		// Add a window listener to listen for close events
		addWindowListener( new WindowAdapter()
		{
			public void windowClosing( WindowEvent e )
			{
				dispose();
				System.exit( 0 );
			}
		} );

		// Minor hack to get DatabasePanel to populate the menubar
		// when wrv starts up
		addComponentListener( new ComponentAdapter()
		{
			public void componentShown( ComponentEvent e )
			{
				dbPanel.setVisible( true );
				removeComponentListener( this );
			}
		} );

		// Pack the frame
		pack();

		dbPanel.setVisible( false );

		// We set the tab colors here because there is a bug in the
		// JClass package which doesn't keep the colors until the
		// notebook is layed out
		dbPanel.setTabColor( Color.red );
		dbPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		dbPanel.getTab().setShadowThickness( 2 );

		modulesPanel.setTabColor( Color.orange );
		modulesPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		modulesPanel.getTab().setShadowThickness( 2 );

		activationPanel.setTabColor( Color.yellow );
		activationPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		activationPanel.getTab().setShadowThickness( 2 );

		deactivationPanel.setTabColor( new Color( 0, 192, 178 ) );
		deactivationPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		deactivationPanel.getTab().setShadowThickness( 2 );

		processesPanel.setTabColor( Color.green );
		processesPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		processesPanel.getTab().setShadowThickness( 2 );

		wrappersPanel.setTabColor( Color.cyan );
		wrappersPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		wrappersPanel.getTab().setShadowThickness( 2 );

		librariesPanel.setTabColor( Color.blue );
		librariesPanel.setTabForeground( Color.white );
		librariesPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		librariesPanel.getTab().setShadowThickness( 2 );

		aboutPanel.setTabColor( Color.black );
		aboutPanel.setTabForeground( Color.white );
		aboutPanel.getTab().setShadowType( BWTEnum.SHADOW_FRAME_OUT );
		aboutPanel.getTab().setShadowThickness( 2 );

		// Tell the module cache to start listening for events
		try
		{
			stream = new ModuleEventStream();
			modCache.listen(stream);
			stream.listen();
		}
		catch( Exception e )
		{
			e.printStackTrace(System.err);
			System.exit(0);
		}
	}

	// Protected methods
	protected void loadWrapper()
	{
		Frame frame = new Frame();
		FileDialog loadDialog = new FileDialog( frame, "Load Wrapper", FileDialog.LOAD );
		loadDialog.setDirectory( System.getProperty( "user.dir" ) );
		loadDialog.setFilenameFilter( new FilenameFilter() {
			public boolean accept( File dir, String name ) { return( name.endsWith( ".o" ) ); }
		} );
		loadDialog.show();

		String loadFile = loadDialog.getDirectory() +
				  loadDialog.getFile();
		if( loadDialog.getFile() != null )
		{
			try
			{
				WrapperViewer.getGlobal().loadModule( loadFile );
			}
			catch( IOException ioe )
			{
				ioe.printStackTrace();
			}
		}
	}

	protected void addCriteria()
	{
		if( critBox == null ) critBox = new CriteriaBox( this );
		Point loc = getLocationOnScreen();
		Rectangle bounds = this.getBounds();
		Rectangle fbounds = critBox.getBounds();
		critBox.setLocation( loc.x + ( bounds.width - fbounds.width ) / 2,
				     loc.y + ( bounds.height - fbounds.height ) / 2 );
		critBox.show();
	}
}

