/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: ModulesPanel.java,v 1.18 2001/04/05 18:41:21 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import jclass.bwt.*;
import jclass.util.*;
import com.nailabs.ui.*;
import com.gswtk.wrapper.event.*;
import com.gswtk.viewer.*;
import com.gswtk.ui.*;

public class ModulesPanel extends WRVPanel implements ModuleEventListener
{
	public final static String[] columnNames =
	{
		"ID", "Module", "Type", "Load Address", "Size"
	};

	public final static int[]    columnSizes =
	{
		70, 210, 105, 175, 105
	};

	protected Grid               grid;
	protected Menu		     moduleMenu;
	protected ModuleCache        modCache;


	// Constructors
	public ModulesPanel( WRVPanelNotebook notebook, MenuBar menubar, ModuleCache cache )
	{
		super( notebook, "Loaded Modules", true );
		setMenuBar( menubar );
		setTabColor( Color.orange );
		modCache = cache;

		moduleMenu = new Menu( "Module" );
		moduleMenu.setFont( menubar.getFont() );

		final MenuItem moduleUnload = new MenuItem( "Unload", new MenuShortcut( 'U' ) );
		moduleUnload.addActionListener( new ActionListener() {
			public void actionPerformed( ActionEvent e ) { unloadSelected(); }
		} );
		moduleUnload.setFont( menubar.getFont() );
		moduleUnload.setEnabled( false );
		moduleMenu.add( moduleUnload );

		grid = new Grid( columnNames );
		grid.setDefaultColumnWidth( 100 );
		grid.setColumnWidths( columnSizes );
		grid.showMargin( false );
		grid.setEditable( false );
		grid.addItemListener( new ItemListener()
		{
			public void itemStateChanged( ItemEvent e )
			{
				if( e.getStateChange() == ItemEvent.SELECTED )
					moduleUnload.setEnabled( true );
				else
					moduleUnload.setEnabled( false );
			}
		} );

		Panel page = getPage();
		page.setLayout( new BorderLayout() );
		page.setBackground( Color.lightGray );
		page.add( grid, "Center" );

		addComponentListener( new ComponentAdapter()
		{
			public void componentHidden( ComponentEvent e )
			{
				MenuBar mbar = getMenuBar();
				mbar.remove( moduleMenu );
			}

			public void componentShown( ComponentEvent e )
			{
				MenuBar mbar = getMenuBar();
				mbar.add( moduleMenu );

				if( grid.getSelectedRow() != null )
					moduleUnload.setEnabled( true );
			}
		} );

		modCache.addModuleEventListener( this );

		updateList();
	}

	// Protected methods
	protected void handleDock()
	{
		MenuBar mbar = getMenuBar();
		mbar.remove( moduleMenu );

		super.handleDock();
	}

	protected void handleUndock()
	{
		MenuBar mbar = getMenuBar();
		mbar.remove( moduleMenu );

		super.handleUndock();
	}

	protected void updateList()
	{
		Enumeration e = modCache.elements();

		if( e != null )
		{
			grid.removeAll();

			while( e.hasMoreElements() )
				addRow( (Module)e.nextElement() );
		}
	}

	// Public methods
	public void finalize()
	{
		modCache.removeModuleEventListener( this );
	}

	public void addRow( Module m )
	{
		String[] values = new String[5];

		values[0] = Integer.toString( m.getHandle() );
		values[1] = m.getName();
		if( m instanceof Wrapper )
			values[2] = "wrapper";
		else if( m instanceof ActivationCriteria )
			values[2] = "activation";
		else if( m instanceof DeactivationCriteria )
			values[2] = "deactivation";
		else if( m instanceof Library )
			values[2] = "library";
		else
			values[2] = "unknown";

		if( m.getLoadAddr() == 0 )
			values[3] = "N/A";
		else
			values[3] = "0x" +
			            Long.toHexString( m.getLoadAddr() );
		values[4] = Integer.toString( m.getSize() );

		grid.addRow( values );
	}

	public void addRow( int id )
	{
		Module m = (Module)modCache.get( new Integer(id) );

		addRow( m );
	}

	public void removeRow( int id )
	{
		String strID = Integer.toString(id);

		for( int i = 1; i <= grid.getTotalRows(); i++ )
		{
			String[] items = grid.getRow( i );

			if( items[0].equals( strID ) )
			{
				grid.deleteRow( i );
				break;
			}
		}
	}

	public void moduleEventOccurred( ModuleEvent e )
	{
		int id = e.getModuleID();

		if( e instanceof ModuleLoadedEvent )
			addRow( id );
		else if( e instanceof ModuleUnloadedEvent )
			removeRow( id );
	}

	public void unloadSelected()
	{
		String[] item = grid.getSelectedRow();

		if( item != null )
		{
			int modId = Integer.valueOf( item[0] ).intValue();
			try { WrapperViewer.getGlobal().unloadModule( modId ); } catch( IOException ioe ) {}
		}
	}
}

