/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: DatabaseCache.java,v 1.18 2001/04/05 18:41:25 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

package com.gswtk.viewer;

import java.util.*;
import com.eventfs.*;
import com.gswtk.db.*;
import com.gswtk.wrapper.event.*;

public class DatabaseCache extends ObjectCache implements TableEventListener
{
	protected WrapperViewer       wrv;
	protected Vector	      eventListeners = new Vector();
	protected DatabaseEventStream des;

	// Constructors
	public DatabaseCache( WrapperViewer wv )
	{
		// do not fill the cache
		super(false);

		// store reference to WrapperViewer
		wrv = wv;

		// now fill the cache
		fill();

		// open database event stream
		try
		{
			des = new DatabaseEventStream();
		}
		catch( Exception e )
		{
			e.printStackTrace(System.err);
			System.exit(0);
		}
		des.addTableEventListener( this );
		des.listen();
	}


	// Public methods
	public void fill()
	{
		clear();

		// Populate the table hash with the available tables
		//	* Calling requestTables() returns an enumeration
		//	  of Table objects which only contain the table
		//	  name and handle
		for( Enumeration e = wrv.requestTables();
		     e != null && e.hasMoreElements(); )
		{
			Table table = (Table)e.nextElement();

			put( new Integer(table.getHandle()), table );
		}
	}

	public synchronized void tableEventOccurred( TableEvent e )
	{
		Integer id  = new Integer(e.getTableID());
		Table   tab = null;

		if( e instanceof CreateTableEvent )
		{
			tab = wrv.requestTableByHandle(id.intValue(), false);
			if( tab.getName().length() > 0 ) put( id, tab );
		}
		else if( e instanceof DropTableEvent )
		{
			tab = (Table)get( id );
			if( tab != null ) tab = (Table)remove( tab );
		}
		else if( e instanceof InsertRowEvent )
		{
			tab = (Table)get( id );

			if( ( tab != null ) && tab.isLoaded() )
			{
				try
				{
					wrv.requestRow( tab, ((InsertRowEvent)e).getRowID() );
				}
				catch( InvalidHandleException ihe )
				{
					// We don't worry about InvalidHandleException, since it
					//   probably means we're getting events for a table which
					//   was recently dropped (a few events after this one)
				}
			}
		}
		else if( e instanceof DeleteRowEvent )
		{
			tab = (Table)get( id );

			if( ( tab != null ) && tab.isLoaded() )
				tab.remove( new Row( ((DeleteRowEvent)e).getRowID() ) );
		}
		else if( e instanceof UpdateRowEvent )
		{
			tab = (Table)get( id );

			if( ( tab != null ) && tab.isLoaded() )
			{
				try
				{
					wrv.requestRow( tab, ((UpdateRowEvent)e).getRowID() );
				}
				catch( InvalidHandleException ihe )
				{
					// We don't worry about InvalidHandleException, since it
					//   probably means we're getting events for a table which
					//   was recently dropped (a few events after this one)
				}
			}
		}
		else
		{
			System.err.println( "Unknown event:\n" + e.toString() );
		}

		if( tab != null && tab.getName().length() > 0 )
		try
		{
			e.addNote( "table", tab );
			for( Enumeration enum = eventListeners.elements(); enum.hasMoreElements(); )
				((TableEventListener)enum.nextElement()).tableEventOccurred( e );
		}
		catch( Exception ex )
		{
			ex.printStackTrace( System.err );
		}
	}

	public synchronized void addEventListener( TableEventListener tel )
	{
		eventListeners.addElement( tel );
	}

	public synchronized void removeEventListener( TableEventListener tel )
	{
		eventListeners.removeElement( tel );
	}

	public synchronized Enumeration getTables() { return elements(); }
	public synchronized Table getTable( Integer id ) { return (Table)get(id); }
	public synchronized Table getTable( int handle ) { return (Table)get(new Integer(handle)); }

}

