/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: ModuleCache.java,v 1.16 2001/04/05 18:41:25 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 */

package com.gswtk.viewer;

import java.util.*;
import com.eventfs.*;
import com.gswtk.wrapper.event.*;

public class ModuleCache extends ModuleObjectCache implements ModuleEventListener
{
	// Public finals
	public static final int CACHE_ACTIVATION   = 1;
	public static final int CACHE_DEACTIVATION = 2;
	public static final int CACHE_CRITERIA     = CACHE_ACTIVATION | CACHE_DEACTIVATION;
	public static final int CACHE_WRAPPER      = 4;
	public static final int CACHE_LIBRARY      = 8;
	public static final int CACHE_ALL          = CACHE_CRITERIA | CACHE_WRAPPER | CACHE_LIBRARY;


	// Private data
	private ActivationCache   aCache;
	private DeactivationCache dCache;
	private WrapperCache      wCache;
	private LibraryCache      lCache;


	// Constructors
	public ModuleCache() { this(CACHE_ALL); }
	public ModuleCache( boolean fill ) { this(CACHE_ALL, fill); }
	public ModuleCache( int mask ) { this(mask, false); }

	public ModuleCache( int mask, boolean fill )
	{
		super();

		if( ( mask & CACHE_ACTIVATION ) != 0 )
		{
			aCache = new ActivationCache( false );
			aCache.addModuleEventListener( this );
		}
		if( ( mask & CACHE_DEACTIVATION ) != 0 )
		{
			dCache = new DeactivationCache( false );
			dCache.addModuleEventListener( this );
		}
		if( ( mask & CACHE_WRAPPER ) != 0 )
		{
			wCache = new WrapperCache( false );
			wCache.addModuleEventListener( this );
		}
		if( ( mask & CACHE_LIBRARY ) != 0 )
		{
			lCache = new LibraryCache( false );
			lCache.addModuleEventListener( this );
		}

		if( fill ) fill();
	}


	// Private methods
	private void fill( ModuleObjectCache moc )
	{
		if( moc != null )
		{
			moc.fill();
			for( Enumeration e = moc.elements();
			     e.hasMoreElements(); )
			{
				Module m = (Module)e.nextElement();
				put( new Integer(m.getHandle()), m );
			}
		}
	}

	public Module fill( ModuleObjectCache moc, int id )
	{
		if( moc != null )
		{
			Module m = moc.fill( id );
			put( new Integer(m.getHandle()), m );

			return m;
		}

		return null;
	}

	public Vector getModules( ModuleObjectCache moc, Vector v )
	{
		for( Enumeration e = moc.elements(); e.hasMoreElements(); )
			v.addElement( e.nextElement() );
		return v;
	}


	// Public methods
	public void addModuleEventListener( int mask, ModuleEventListener mel ) throws IllegalArgumentException
	{
		try
		{
			if( ( mask & CACHE_ACTIVATION ) != 0 )
				aCache.addModuleEventListener( mel );
			if( ( mask & CACHE_DEACTIVATION ) != 0 )
				dCache.addModuleEventListener( mel );
			if( ( mask & CACHE_WRAPPER ) != 0 )
				wCache.addModuleEventListener( mel );
			if( ( mask & CACHE_LIBRARY ) != 0 )
				lCache.addModuleEventListener( mel );
		}
		catch( NullPointerException npe )
		{
			throw new IllegalArgumentException("ModuleCache is not caching one of the specified module types");
		}
	}

	public final void fill()
	{
		clear();

		fill( aCache );
		fill( dCache );
		fill( wCache );
		fill( lCache );
	}

	public final Module fill( int id )
	{
		Module m = null;
		m = fill( aCache, id );
		if( m == null ) m = fill( dCache, id );
		if( m == null ) m = fill( wCache, id );
		if( m == null ) m = fill( lCache, id );
		return m;
	}

	public Enumeration getModules( int mask )
	{
		Vector v = new Vector();
		if( ( mask & CACHE_ACTIVATION ) != 0 )
			v = getModules( aCache, v );
		if( ( mask & CACHE_DEACTIVATION ) != 0 )
			v = getModules( dCache, v );
		if( ( mask & CACHE_WRAPPER ) != 0 )
			v = getModules( wCache, v );
		if( ( mask & CACHE_LIBRARY ) != 0 )
			v = getModules( lCache, v );
		return v.elements();
	}

	public void listen( ModuleEventStream stream ) throws Exception
	{
		if( aCache != null ) aCache.listen( stream );
		if( dCache != null ) dCache.listen( stream );
		if( wCache != null ) wCache.listen( stream );
		if( lCache != null ) lCache.listen( stream );

		mes = stream;
	}

	public void moduleEventOccurred( ModuleEvent me ) throws EventException
	{
		ObjectCache oc = (ObjectCache)me.getNote( "cache" );

		if( oc != null )
		{
			Integer id = new Integer(me.getModuleID());

			if( me instanceof ModuleLoadedEvent )
				put(id, oc.get(id));
			else if( me instanceof ModuleUnloadedEvent )
				remove(id);

			// Tell the ModuleObjectCache not to cache
			me.addNote( "nocache", oc );
		}

		// Let the ModuleObjectCache do its thing
		super.moduleEventOccurred( me );
	}

	public void removeModuleEventListener( int mask, ModuleEventListener mel ) throws IllegalArgumentException
	{
		try
		{
			if( ( mask & CACHE_ACTIVATION ) != 0 )
				aCache.removeModuleEventListener( mel );
			if( ( mask & CACHE_DEACTIVATION ) != 0 )
				dCache.removeModuleEventListener( mel );
			if( ( mask & CACHE_WRAPPER ) != 0 )
				wCache.removeModuleEventListener( mel );
			if( ( mask & CACHE_LIBRARY ) != 0 )
				lCache.removeModuleEventListener( mel );
		}
		catch( NullPointerException npe )
		{
			throw new IllegalArgumentException("ModuleCache is not caching one of the specified module types");
		}
	}
}

