%{
/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: criteria.y,v 1.22 2001/08/06 20:03:35 stevek Exp $
 *
 * author: Lee Badger
 *
 * Purpose: Parser for criteria.
 * 
 */

#ifndef LINKAGE
#define LINKAGE IMPORT
#endif

#include <wr_conf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#ifdef WR_UNIX
#include <sys/param.h>
#endif
#include <sys/stat.h>

#ifdef WR_BROKEN_STL
#define __PUT_STATIC_DATA_MEMBERS_HERE
#include "template/map"
#endif

#include <libwrmod.h>
#include <libre.h>

#undef LINKAGE
#define LINKAGE EXPORT
#include "libcrit.h"

#define TMPFILE_TEMPLATE "/tmp/wrcrit.XXXXXXXX." _CRITEXT

#ifndef WR_FREEBSD
extern "C" {
   int mkstemps(char *, int);
};
#endif

int criteriaerror(char *s);
int criterialex();
int criteriaparse();

extern char *criteria_strin;
extern char error_code[256];
int debug_flag = 0;

Processor *proc = NULL;

extern "C"
{
int errno;
}

extern "C" {
   int critCompile( char *critstr );
}

%}

%union {
   crit_op_t   op;
   Literal    *lit;
   Expression *exp;
   Regex      *regex;
}

%token   <lit>      IDENTIFIER
%token   <lit>      NUM
%token   <regex>    REGEX
%token              K_AND K_OR K_ARROW K_STRCMP
%token              K_GTE K_LTE K_NE K_EQ K_SETUID K_SETGID

%type <op> bin_op
%type <lit> value

%type <exp> rel_expr
%type <exp> bool_expr
%type <exp> setid_expr

%left K_AND K_OR '!'

%%
act_crit:   bool_expr K_ARROW IDENTIFIER
            {
               proc->GenArrow( $1, $3 );
            }

bool_expr:  '(' bool_expr ')'
            {
               $$ = $2;
            }
            | '!' bool_expr
            {
               $$ = new Expression( proc, CO_NOT, $2, NULL );
            }
            | bool_expr K_AND bool_expr
            {
               $$ = new Expression( proc, CO_AND, $1, $3 );
            }
            | bool_expr K_OR bool_expr
            {
               $$ = new Expression( proc, CO_OR, $1, $3 );
            }
            | rel_expr
            { $$ = $1; }
            | setid_expr
            { $$ = $1; }
            | NUM
            {
               $$ = new Expression( proc, ($1==0)?CO_NE:CO_EQ,
                                    new Literal( proc, 0 ),
                                    new Literal( proc, 0 ) );
            }

bin_op:     '>'
            {$$ = CO_GT;}
            | '<'
            {$$ = CO_LT;}
            | K_GTE
            {$$ = CO_GTE;}
            | K_LTE
            {$$ = CO_LTE;}
            | K_NE
            {$$ = CO_NE;}
            | K_EQ
            {$$ = CO_EQ;}

value:      IDENTIFIER
            {}
            | NUM
            {}

rel_expr:   value bin_op value
            {
               $$ = new Expression( proc, $2, $1, $3 );
            }
            | IDENTIFIER K_STRCMP '|' REGEX '|'
            {
               $$ = new Expression( proc, $1, $4 );
            }

setid_expr: K_SETUID
            {
               $$ = new Expression( proc, CO_SUID );
            }
            | K_SETUID '(' IDENTIFIER ')'
            {
               $$ = new Expression( proc, CO_SUID, $3 );
            }
            | K_SETGID
            {
               $$ = new Expression( proc, CO_SGID );
            }
            | K_SETGID '(' IDENTIFIER ')'
            {
               $$ = new Expression( proc, CO_SGID, $3 );
            }

%%

#include "lex.criteria.c"

int
criteriawrap()
{
   return 1;
}

/*
 * Parse the criteria.  Read the characters from the provided string.
 * On success, allocates a "Processor" object that contains the strings,
 * literals, and expressions that have been parsed.
 * Resets the lexer, so this function
 * can be called multiple times, perhaps with different input strings,
 * during the same program run.
 *
 * Returns 0 if the criterion is valid, error otherwise.
 */
int EXPORT
critParse( char *str, Processor **src )
{
   criteria_strin = str;

   proc = new Processor;

   *error_code = '\0';

   int err = criteriaparse();
   if( err )
      delete proc;
   else
      *src = proc;

   yyrestart( NULL );
   return err;
}

/*
 * When given a "Processor" object produced by wr_parse_criterion,
 * creates a temporary .wmo file containing the compiled contents of the
 * source object.  Does not delete the provided source object.  The
 * caller is responsible both for unlinking the .wmo file and for
 * free'ing the returned file name.
 * 
 * If output is non-NULL, it is used as the name of the output file.
 *
 * Returns the path of the temporary file on success, or NULL on
 * failure.
 */
char  EXPORT*
critMakeTempFile( Processor *srcObj, char *output, pd_bool_t addWrapper )
{
   char        *objName;
   struct stat  statb;
   FILE        *fp;

   if( output != NULL )
   {
      /* If it exists, return error. */
      stat( output, &statb );
      if( errno != ENOENT )
         return NULL;

      if( ( fp = fopen( output, "wb" ) ) == NULL )
      {
         fprintf( stderr, "ERROR opening `%s' for output\n", output );
         return NULL;
      }
      objName = output;
   }
   else
   {
      int fd;

      /* Get a tmp name and append the appropriate extension */
      objName = (char *) malloc( sizeof(TMPFILE_TEMPLATE) );
      strncpy( objName, TMPFILE_TEMPLATE, sizeof(TMPFILE_TEMPLATE) );

      fd = mkstemps( objName, sizeof(_CRITEXT) + 1 );
      if( fd < 0 )
         return NULL;

      if( ( fp = fdopen( fd, "wb" ) ) == NULL )
      {
         fprintf( stderr, "ERROR opening `%s' for output\n", objName );
         close( fd );
         free( objName );
         return NULL;
      }
   }

   srcObj->WriteOut( fp, addWrapper );
   fclose( fp );

   return( objName );
}

const char  EXPORT*
critParserError()
{
   return (const char *)error_code;
}

int EXPORT
critCompile( char *critStr, int addWrapper )
{
   Processor *srcObj;
   char *tmpName;
   int error;

   error = critParse( critStr, &srcObj );
   if( error != 0 )
      return error;

   tmpName = critMakeTempFile( srcObj, NULL, addWrapper );
   error = wrmodLoad( tmpName );
#ifndef DEBUG_CRITERIA
   unlink( tmpName );
   free( tmpName );
#endif

   return 0;
}

