/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: processor.cc,v 1.38 2001/06/23 00:24:28 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Criteria processor classes for generating criteria files
 *
 */

#ifndef LINKAGE
#define LINKAGE IMPORT
#endif

#include <wr_conf.h>
#include <wsi_types.h>

#ifdef WR_UNIX
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#endif

#ifdef WR_BROKEN_STL
#define EXTRACT(s) (s).chars()
#else
#define EXTRACT(s) (s).c_str()
#endif

#ifdef WR_WINNT
#pragma warning(disable: 4786)
#endif

#include "libre.h"

#undef LINKAGE
#define LINKAGE EXPORT

#include "libcrit.h"


typedef struct sym_element sym_element_t;
struct sym_element
{
   TCHAR       *name; 
   crit_id_t    token;
   crit_type_t  type;
};    

static sym_element_t sym_list[] =
{
   { PD_MAKESTR("prog"),   CTOK_PROG,  CT_STRING },
   { PD_MAKESTR("pprog"),  CTOK_PPROG, CT_STRING },
   { PD_MAKESTR("cwd"),    CTOK_CWD,   CT_STRING },
#ifdef WR_WINNT
   { PD_MAKESTR("user"),   CTOK_USER,  CT_STRING },
#else
   { PD_MAKESTR("user"),   CTOK_UID,   CT_INT    },
   { PD_MAKESTR("pid"),    CTOK_PID,   CT_INT    },
   { PD_MAKESTR("ppid"),   CTOK_PPID,  CT_INT    },
   { PD_MAKESTR("uid"),    CTOK_UID,   CT_INT    },
   { PD_MAKESTR("euid"),   CTOK_EUID,  CT_INT    },
   { PD_MAKESTR("gid"),    CTOK_GID,   CT_INT    },
   { PD_MAKESTR("egid"),   CTOK_GID,   CT_INT    },
   { PD_MAKESTR("tty"),    CTOK_TTY,   CT_INT    },
#endif
   { NULL,    0,          CT_NONE   }
};


//
// class Node
//
Node::Node( Processor *_proc )
{
   proc = _proc;
}

char *
Node::ToString()
{
   return strdup( "n/a" );
}


//
// class Literal
//
Literal::Literal( Processor *_proc, int ival ) : Node( _proc )
{
   Initialize( CT_INT );

   lit.ival = ival;
   lit.soff = CRIT_NOARG_VALUE;
}

Literal::Literal( Processor *_proc, char *str ) : Node( _proc )
{
   Initialize( CT_STRING );

   lit.soff = proc->RegisterString( str );
   lit.ival = CRIT_NOARG_VALUE;
}

Literal::~Literal()
{
   proc->UnregisterLiteral( this );
}

void
Literal::Initialize( crit_type_t _type )
{
   lit.type = _type;

   lit.id = proc->RegisterLiteral( this );
}

char *
Literal::ToString()
{
   switch( lit.type )
   {
      case CT_STRING:
         return strdup( EXTRACT(proc->FindString( lit.soff )) );

      case CT_LABEL:
      {
         char buf[256];
         sprintf( buf, "%s (%d)", EXTRACT(proc->FindString( lit.soff )),
                  lit.ival );
         return strdup( buf );
      }

      case CT_INT:
      {
         char buf[20];
         sprintf( buf, "%d", lit.ival );
         return strdup( buf );
      }

      default:
         break;
   }

   return Node::ToString();
}

crit_id_t
Literal::GetId()
{
   if( IsBuiltin() )
      return lit.soff;
   else
      return lit.id;
}

crit_type_t
Literal::GetType()
{
   return lit.type;
}

crit_type_t
Literal::GetCompatType()
{
   if( lit.type == CT_LABEL )
      return CT_INT;
   else
      return lit.type;
}

pd_bool_t
Literal::IsBuiltin()
{
   if( ( lit.soff != CRIT_NOARG_VALUE ) &&
       ( ( lit.soff & CRIT_BUILTIN_MASK ) != 0 ) )
      return TRUE;
   else
      return FALSE;
}

int
Literal::ToBuiltin()
{
   struct sym_element *cur;
   String              name;

   // Must be a string literal
   if( lit.type != CT_STRING )
      return EINVAL;

   // Look for already-converted state
   if( ( lit.soff & CRIT_BUILTIN_MASK ) != 0 )
      return EINVAL;

   name = proc->FindString( lit.soff );

   for( cur = sym_list; cur->name; cur++ )
      if( !strcmp( cur->name, EXTRACT(name) ) )
      {
         proc->UnregisterString( lit.soff );
         lit.type = cur->type;
         lit.soff = cur->token;
#ifdef DEBUG_CRIT_PROCESSOR
         fprintf( stderr, "Mutated string on literal 0x%x to 0x%x\n", lit.id,
                  lit.soff );
#endif /* DEBUG_CRIT_PROCESSOR */
      }

   return 0;
}

int
Literal::ToDevice()
{
#ifdef WR_UNIX
   String devname;

   switch( lit.type )
   {
      case CT_STRING:
      {
         char        path[MAXPATHLEN];
         struct stat st;

         // Change string to a label type
         lit.type = CT_LABEL;

         // Construct uid
         devname = proc->FindString( lit.soff );
         snprintf( path, MAXPATHLEN, "/dev/%s", EXTRACT(devname) );

         if( stat( path, &st ) )
         {
            fprintf( stderr, "Invalid device name \"%s\"\n", EXTRACT(devname) );
            return EINVAL;
         }
         lit.ival = st.st_rdev;
         break;
      }

      case CT_INT:
         // Allow 0 to specify a process with no controlling tty
         if( lit.ival == 0 )
            break;

      default:
         fprintf( stderr, "Invalid type in device lookup\n" );
         return EINVAL;
   }
#endif
   return 0;
}

int
Literal::ToUserId()
{
#ifdef WR_UNIX
   struct passwd *pw;
   String         name;

   switch( lit.type )
   {
      case CT_STRING:
         // Change string to a label type
         lit.type = CT_LABEL;

         // Construct uid
         name = proc->FindString( lit.soff );
         if( ( pw = getpwnam( EXTRACT(name) ) ) == NULL )
         {
            fprintf( stderr, "Invalid user name \"%s\"\n", EXTRACT(name) );
            return EINVAL;
         }
         lit.ival = (int) pw->pw_uid;
         break;

      case CT_INT:
      default:
         fprintf( stderr, "Invalid type in user lookup\n" );
         return EINVAL;
   }
#endif
   return 0;
}

int
Literal::ToGroupId()
{
#ifdef WR_UNIX
   struct group *gr;
   String        name;

   switch( lit.type )
   {
      case CT_STRING:
         // Change string to a label type
         lit.type = CT_LABEL;

         // Construct uid
         name = proc->FindString( lit.soff );
         if( ( gr = getgrnam( EXTRACT(name) ) ) == NULL )
         {
            fprintf( stderr, "Invalid group name \"%s\"\n", EXTRACT(name) );
            return EINVAL;
         }
         lit.ival = (int) gr->gr_gid;
         break;

      case CT_INT:
      default:
         fprintf( stderr, "Invalid type in group lookup\n" );
         return EINVAL;
   }
#endif
   return 0;
}

int
Literal::WriteOut( FILE *fp )
{
   return fwrite( &lit, sizeof(lit), 1, fp );
}


//
// class Expression
//

// arrow op
Expression::Expression( Processor *_proc, Expression *exp, Literal *lit )
   : Node( _proc )
{
   Initialize( CO_ARROW, exp->GetId(), lit->GetId() );
}

// setuid and setgid ops
Expression::Expression( Processor *_proc, crit_op_t op )
   : Node( _proc )
{
   switch( op )
   {
      case CO_SUID:
      case CO_SGID:
         break;

      default:
         fprintf( stderr, "Invalid arguments to expression\n" );
         exit( 2 );
   }

   Initialize( op, CRIT_NOARG_VALUE, CRIT_NOARG_VALUE );
}

// setuid and setgid ops with value
Expression::Expression( Processor *_proc, crit_op_t op, Literal *lit )
   : Node( _proc )
{
   if( lit == NULL )
   {
      fprintf( stderr, "Expression missing argument\n" );
      exit( 2 );
   }

   switch( op )
   {
      case CO_SUID:
         if( lit->ToUserId() )
            exit( 2 );
         break;

      case CO_SGID:
         if( lit->ToGroupId() )
            exit( 2 );
         break;

      default:
         fprintf( stderr, "Invalid arguments to expression\n" );
         exit( 2 );
   }

   Initialize( op, lit->GetId(), CRIT_NOARG_VALUE );
}

// regex op
Expression::Expression( Processor *_proc, Literal *lit, Regex *re )
   : Node( _proc )
{
   if( re == NULL )
   {
      fprintf( stderr, "Expression missing argument\n" );
      exit( 2 );
   }

   // Convert literal to builtin op if possible
   lit->ToBuiltin();

   // Make sure types match in comparison
   if( lit->GetCompatType() != CT_STRING )
   {
      fprintf( stderr, "Invalid arguments to expression (type mismatch)\n" );
      exit( 2 );
   }

   Initialize( CO_RE, lit->GetId(), re->GetId() );

   if( lit->IsBuiltin() )
      delete lit;
}

// comparison ops
Expression::Expression( Processor *_proc, crit_op_t op, Literal *lit1,
                        Literal *lit2 )
   : Node( _proc )
{
   switch( op )
   {
      case CO_GT:
      case CO_LT:
      case CO_GTE:
      case CO_LTE:
      case CO_NE:
      case CO_EQ:
         break;

      default:
         fprintf( stderr, "Invalid arguments to expression\n" );
         exit( 2 );
   }

   // Convert to builtin ops if possible
   lit1->ToBuiltin();
   lit2->ToBuiltin();

#ifdef WR_UNIX
   if( lit1->IsBuiltin() &&
       ( ( lit1->GetId() == CTOK_UID ) ||
         ( lit1->GetId() == CTOK_EUID ) ) )
      if( lit2->ToUserId() )
         exit( 2 );
   if( lit2->IsBuiltin() &&
       ( ( lit2->GetId() == CTOK_UID ) ||
         ( lit2->GetId() == CTOK_EUID ) ) )
      if( lit1->ToUserId() )
         exit( 2 );

   if( lit1->IsBuiltin() &&
       ( ( lit1->GetId() == CTOK_GID ) ||
         ( lit1->GetId() == CTOK_EGID ) ) )
      if( lit2->ToGroupId() )
         exit( 2 );
   if( lit2->IsBuiltin() &&
       ( ( lit2->GetId() == CTOK_GID ) ||
         ( lit2->GetId() == CTOK_EGID ) ) )
      if( lit1->ToGroupId() )
         exit( 2 );

   if( lit1->IsBuiltin() &&
       ( lit1->GetId() == CTOK_TTY ) )
      if( lit2->ToDevice() )
         exit( 2 );
   if( lit2->IsBuiltin() &&
       ( lit2->GetId() == CTOK_TTY ) )
      if( lit1->ToDevice() )
         exit( 2 );
#endif /* WR_UNIX */

   // Make sure types match in comparison
   if( lit1->GetCompatType() != lit2->GetCompatType() )
   {
      fprintf( stderr, "Invalid arguments to expression (type mismatch)\n" );
      exit( 2 );
   }

   Initialize( op, lit1->GetId(), lit2->GetId() );

   if( lit1->IsBuiltin() )
      delete lit1;
   if( lit2->IsBuiltin() )
      delete lit2;
}

// boolean ops
Expression::Expression( Processor *_proc, crit_op_t op, Expression *exp1,
                        Expression *exp2 )
   : Node( _proc )
{
   switch( op )
   {
      case CO_AND:
      case CO_OR:
         Initialize( op, exp1->GetId(), exp2->GetId() );
         break;

      case CO_NOT:
         Initialize( op, exp1->GetId(), (crit_id_t) 0xFFFFFFFF );
         break;

      default:
         fprintf( stderr, "Invalid arguments to expression\n" );
         exit( 2 );
   }
}


Expression::~Expression()
{
   proc->UnregisterExpression( this );
}

void
Expression::Initialize( crit_op_t op, crit_id_t arg1, crit_id_t arg2 )
{
   exp.op = op;
   exp.arg1 = arg1;
   exp.arg2 = arg2;
   exp.seen = FALSE;

   id = proc->RegisterExpression( this );
}

crit_id_t
Expression::GetId()
{
   return id;
}


int
Expression::WriteOut( FILE *fp )
{
   return fwrite( &exp, sizeof(exp), 1, fp );
}


//
// class Regex
//
Regex::Regex( Processor *_proc, char *exp )
   : Node( _proc ), elemmap(), elemnodemap()
{
   nextElement = 0;

   reToken = new regex();
   int error = reToken->re_construct( exp );
   if( error )
   {
      fprintf( stderr, "regular expression: error %d\n", error );
      exit( 2 );
   }
   id = proc->RegisterRegex( this );

   re.regex = 0;
   (void) GenerateStructures( reToken );
   re.regexsize = elemmap.size();
}

Regex::~Regex()
{
   proc->UnregisterRegex( this );
}

crit_id_t
Regex::GetId()
{
   return id;
}

crit_regex_elem_t *
Regex::GenerateStructures( regex_token *tok, crit_id_t *last )
{
   crit_regex_elem_t *relem = NULL;

   for( ; tok; tok = tok->re_walk() )
   {
      crit_regex_elem_t *elem = new crit_regex_elem_t;
      memset( elem, 0, sizeof(*elem) );
      if( !relem )
         relem = elem;
      elem->elem.op = tok->re_gettype();
      elem->node = tok;
      crit_id_t id = RegisterElement( elem );
      if( !tok->re_getnext() && tok->re_getuncle() )
      {
         elem->elem.next = CRIT_RE_UNCLE;
         elem->uncle = tok->re_getuncle();
      }
      else
      {
         elem->elem.next = CRIT_ARG_MASK;
      }

      switch( elem->elem.op )
      {
         case RE_START:
         {
            regex_start_token *stok = (regex_start_token *)tok;
            elem->elem.data = ( stok->re_canbenewline()?1:0 );
            break;
         }

         case RE_END:
         {
            regex_end_token *etok = (regex_end_token *)tok;
            elem->elem.data = ( etok->re_canbenewline()?1:0 );
            break;
         }

         case RE_ANY:
         {
            regex_any_token *atok = (regex_any_token *)tok;
            elem->elem.data = ( ( atok->re_canbenull()?0x01:0x00 ) |
                                ( atok->re_canbenewline()?0x02:0x00) );
            break;
         }

         case RE_POSIX:
         {
            regex_posix_token *ptok = (regex_posix_token *)tok;
            elem->elem.data = ( ( ptok->re_getposixtype() & CRIT_ARG_MASK ) |
                                ( ptok->re_isinsens()?CRIT_RE_INSENS:0 ) |
                                ( ptok->re_isnegated()?CRIT_RE_NEGATE:0 ) );
            break;
         }

         case RE_RANGE:
         {
            regex_range_token *rtok = (regex_range_token *)tok;

            // We construct a nop element which contains the data necessary
            //   for a range element (saves data space for other tokens)
            crit_regex_elem_t *dataelem = new crit_regex_elem_t;
            memset( dataelem, 0, sizeof(*dataelem) );
            dataelem->elem.op = RE_NOP;
            dataelem->elem.next = rtok->re_getlo();
            dataelem->elem.oneofnext = rtok->re_gethi();
            dataelem->elem.data = ( ( rtok->re_isinsens()?CRIT_RE_INSENS:0 ) |
                                    ( rtok->re_canstingy()?CRIT_RE_STINGY:0 ) );
            elem->elem.data = RegisterElement( dataelem );
            break;
         }

         case RE_REPEAT:
         {
            regex_repeated_token *rtok = (regex_repeated_token *)tok;

            // We construct a nop element which contains the data necessary
            //   for a repeat element (saves data space for other tokens)
            crit_regex_elem_t *dataelem = new crit_regex_elem_t;
            memset( dataelem, 0, sizeof(*dataelem) );
            dataelem->elem.op = RE_NOP;
            dataelem->elem.next = rtok->re_getmin();
            dataelem->elem.oneofnext = rtok->re_getmax();
            elem->elem.data = RegisterElement( dataelem );
            GenerateStructures( (*rtok)[0], &(dataelem->elem.data) );
            break;
         }

         case RE_CHAR:
         {
            regex_char_token *ctok = (regex_char_token *)tok;
            Literal *lit = new Literal( proc, (char *)ctok->re_getstr() );
            elem->elem.data = ( lit->GetId() |
                                (ctok->re_isinsens()?CRIT_RE_INSENS:0) );
            break;
         }

         case RE_EXP:
            GenerateStructures( (*tok)[0], &(elem->elem.data) );
            break;

         case RE_ONEOF:
         {
            regex_oneof_token *otok = (regex_oneof_token *)tok;
            crit_id_t *lastone = &(elem->elem.oneofnext);
            elem->elem.data = ( otok->re_isnegated()?CRIT_RE_NEGATE:0 );
            for( int i = 0; i < tok->re_children(); i++ )
            {
               crit_regex_elem_t *oneofelem =
                  GenerateStructures( (*tok)[i], lastone );
               if( oneofelem )
               {
                  oneofelem->elem.oneofnext = CRIT_ARG_MASK;
                  lastone = &(oneofelem->elem.oneofnext);
               }
               else
               {
                  fprintf( stderr, "Unable to generate structures for "
                           "children.  Stopped at %d of %d.\n", i,
                           tok->re_children() );
                  break;
               }
            }
            break;
         }

         default:
            fprintf( stderr, "Invalid token encountered in regex\n" );
            exit( 2 );
      }

      if( last )
         *last = (*last & ~CRIT_ARG_MASK) | (id & CRIT_ARG_MASK);
      last = &(elem->elem.next);

      if( elem->elem.next & CRIT_RE_UNCLE )
         break;
   }

   return relem;
}

crit_id_t
Regex::RegisterElement( crit_regex_elem_t *elem )
{
   crit_id_t elemid = nextElement++;

   if( elemid > CRIT_ARG_MASK )
   {
      fprintf( stderr, "No more than %d regular expression elements is "
               "allowed\n", CRIT_ARG_MASK );
      exit( 2 );
   }

   elem->id = elemid;
   elemmap[elemid] = elem;
   elemnodemap[elem->node] = elem;

#ifdef DEBUG_CRIT_PROCESSOR
   fprintf( stderr, "Registered regular expression #%d element #%d\n",
            ( id & CRIT_ARG_MASK ), elemid );
#endif /* DEBUG_CRIT_PROCESSOR */

   return elemid;
}

crit_regex_elem_t *
Regex::FindElement( crit_id_t elemid )
{
   return elemmap[elemid];
}

int
Regex::WriteOut( FILE *fp )
{
   int nwritten = 0;
   for( int i = 0; i < nextElement; i++ )
   {
      crit_regex_elem_t *elem = elemmap[i];
      int retc = fwrite( &(elem->elem), sizeof(elem->elem), 1, fp );
      if( retc < 0 )
         return retc;
      else
         nwritten += retc;
   }
   return nwritten;
}


//
// class Processor
//
Processor::Processor() : expmap(), litmap(), strmap(), regexmap()
{
   rootnode = NULL;

   nextExpression = 0;
   nextLiteral = 0;
   nextString = 0;
   nextRegex = 0;

   finalized = FALSE;
}

void
Processor::GenArrow( Expression *exp, Literal *lit )
{
   if( rootnode != NULL )
   {
      fprintf( stderr, "arrow part of expression already generated\n" );
      exit( 2 );
   }

   rootnode = new Expression( this, exp, lit );
}

int
Processor::WriteOut( FILE *fp, pd_bool_t act )
{
   if( rootnode == NULL )
   {
      fprintf( stderr, "attempted to write out incomplete criteria\n" );
      exit( 2 );
   }

   if( !finalized )
      Finalize();

   crit_id_t regexend = 0;
   if( nextRegex )
   {
      Regex *re = regexmap[nextRegex-1];
      regexend = ( re->re.regex + re->re.regexsize ) *
                 sizeof(crit_regex_node_t);
   }

   crit_header_t header;
   header.magic = CRIT_MAGIC;
   header.version = CRIT_VERSION;
   header.activation = act;
   header.literals = sizeof(header);
   header.litsize = nextLiteral * sizeof(crit_literal_t);
   header.expression = header.literals + header.litsize;
   header.expsize = nextExpression * sizeof(crit_expression_t);
   header.regextab = header.expression + header.expsize;
   header.regextabsize = nextRegex * sizeof(crit_regex);
   header.strings = header.regextab + header.regextabsize + regexend;
   header.strsize = nextString;
   header.start = rootnode->GetId() & CRIT_ARG_MASK;
   fwrite( &header, sizeof(header), 1, fp );

   // Write out the literals
   crit_id_t i;
   for( i = 0; i < nextLiteral; i++ )
   {
      Literal *lit = litmap[i];

      lit->WriteOut( fp );
   }

   // Write out expressions
   for( i = 0; i < nextExpression; i++ )
   {
      Expression *exp = expmap[i];

      exp->WriteOut( fp );
   }

   // Write out regular expressions table
   for( i = 0; i < nextRegex; i++ )
   {
      Regex *re = regexmap[i];
      fwrite( &re->re, sizeof(re->re), 1, fp );
   }

   // Write out regular expression elements
   for( i = 0; i < nextRegex; i++ )
   {
      Regex *re = regexmap[i];
      re->WriteOut( fp );
   }

   // Write out strings
   for( i = 0; i < nextString; i++ )
   {
      strmap_t::iterator it = strmap.find( i );
      if( it != strmap.end() )
      {
         String str = (*it).second;
#ifdef WR_WINNT
         WCHAR *widestr = (WCHAR *)new WCHAR[str.length() + 1];
         MultiByteToWideChar( CP_ACP, 0, EXTRACT(str), str.length() + 1,
                              widestr, ( str.length() + 1 ) * sizeof(WCHAR) );
         fwrite( widestr, ( str.length() + 1 ) * sizeof(WCHAR), 1, fp );
         delete widestr;
#else
         fwrite( EXTRACT(str), str.length() + 1, 1, fp );
#endif
      }
   }

   return 0;
}

crit_id_t
Processor::RegisterExpression( Expression *exp )
{
   crit_id_t id = nextExpression++;

   if( id > CRIT_ARG_MASK )
   {
      fprintf( stderr, "No more than %d expressions is allowed\n",
               CRIT_ARG_MASK );
      exit( 2 );
   }

   expmap[id] = exp;

#ifdef DEBUG_CRIT_PROCESSOR
   fprintf( stderr, "Registered expression #%d\n", id );
#endif /* DEBUG_CRIT_PROCESSOR */

   return ( id | CRIT_OP_MASK );
}

pd_bool_t
Processor::UnregisterExpression( Expression *exp )
{
   crit_id_t id = exp->GetId();
   expmap_t::iterator it = expmap.find( id & CRIT_ARG_MASK );
   if( it != expmap.end() )
   {
      expmap.erase( it );
      return TRUE;
   }
   return FALSE;
}

Expression *
Processor::FindExpression( crit_id_t id )
{
   if( ( id & CRIT_OP_MASK ) == 0 )
      return NULL;

   return expmap[id & CRIT_ARG_MASK];
}

crit_id_t
Processor::RegisterLiteral( Literal *lit )
{
   crit_id_t id = nextLiteral++;

   if( id > CRIT_ARG_MASK )
   {
      fprintf( stderr, "No more than %d literals is allowed\n",
               CRIT_ARG_MASK );
      exit( 2 );
   }

   litmap[id] = lit;

#ifdef DEBUG_CRIT_PROCESSOR
   fprintf( stderr, "Registered literal #%d\n", id );
#endif /* DEBUG_CRIT_PROCESSOR */

   return id;
}

pd_bool_t
Processor::UnregisterLiteral( Literal *lit )
{
   crit_id_t id = lit->lit.id;
   litmap_t::iterator it = litmap.find( id );
   if( it != litmap.end() )
   {
      litmap.erase( it );
      if( ( lit->lit.type == CT_STRING ) ||
          ( lit->lit.type == CT_LABEL ) )
         return UnregisterString( lit->lit.soff );
      return TRUE;
   }
   return FALSE;
}

Literal *
Processor::FindLiteral( crit_id_t id )
{
   if( ( id & ~CRIT_ARG_MASK ) != 0 )
      return NULL;

   return litmap[id];
}

crit_id_t
Processor::RegisterString( char *str )
{
   crit_id_t id = nextString++;

   if( id > CRIT_ARG_MASK )
   {
      fprintf( stderr, "No more than %d strings is allowed\n",
               CRIT_ARG_MASK );
      exit( 2 );
   }

   strmap[id] = strdup( str );

#ifdef DEBUG_CRIT_PROCESSOR
   fprintf( stderr, "Registered string #%d (%s)\n", id, str );
#endif /* DEBUG_CRIT_PROCESSOR */

   return id;
}

pd_bool_t
Processor::UnregisterString( crit_id_t id )
{
   strmap_t::iterator it = strmap.find( id );
   if( it != strmap.end() )
   {
      strmap.erase( it );
      return TRUE;
   }

   return FALSE;
}

String &
Processor::FindString( crit_id_t id )
{
   return strmap[id];
}

crit_id_t
Processor::RegisterRegex( Regex *re )
{
   crit_id_t id = nextRegex++;

   if( id > CRIT_ARG_MASK )
   {
      fprintf( stderr, "No more than %d regular expressions is allowed\n",
               CRIT_ARG_MASK );
      exit( 2 );
   }

   regexmap[id] = re;

#ifdef DEBUG_CRIT_PROCESSOR
   fprintf( stderr, "Registered regular expression #%d\n", id );
#endif /* DEBUG_CRIT_PROCESSOR */

   return ( id | CRIT_RE_MASK );
}

pd_bool_t
Processor::UnregisterRegex( Regex *re )
{
   crit_id_t id = re->GetId();
   regexmap_t::iterator it = regexmap.find( id & CRIT_ARG_MASK );
   if( it != regexmap.end() )
   {
      regexmap.erase( it );
      return TRUE;
   }
   return FALSE;
}

Regex *
Processor::FindRegex( crit_id_t id )
{
   if( ( id & CRIT_OP_MASK ) == 0 )
      return NULL;

   return regexmap[id & CRIT_ARG_MASK];
}


#ifdef WR_BROKEN_STL
typedef map< String, crit_id_t, less<String> >       compact_t;
typedef map< int, Literal *, less<int> >             intlits_t;
typedef map< crit_id_t, Literal *, less<crit_id_t> > strlits_t;
#else
typedef map< String, crit_id_t >    compact_t;
typedef map< int, Literal * >       intlits_t;
typedef map< crit_id_t, Literal * > strlits_t;
#endif

void
Processor::Finalize()
{
   if( finalized )
      return;

   finalized = TRUE;

   crit_id_t           i;
   crit_id_t           index;
   compact_t           compactstr;
   compact_t::iterator cit;

   // Generate compact string map
   for( i = 0, index = 0; i < nextString; i++ )
   {
      strmap_t::iterator strIt = strmap.find( i );

      if( strIt != strmap.end() )
      {
         String str = (*strIt).second;
         cit = compactstr.find( str );
         if( cit == compactstr.end() )
         {
            compactstr[str] = index;
#ifdef WR_WINNT
            index += ( str.length() + 1 ) * sizeof(WCHAR);
#else
            index += str.length() + 1;
#endif
         }
      }
   }

   // Adjust literal string offsets
   litmap_t::iterator litIt;
   for( litIt = litmap.begin(); litIt != litmap.end(); litIt++ )
   {
      Literal *lit = (*litIt).second;

#ifdef DEBUG_CRIT_PROCESSOR
      fprintf( stderr, "Finalize: processing literal 0x%x\n", lit->GetId() );
#endif /* DEBUG_CRIT_PROCESSOR */

      if( ( ( lit->GetType() == CT_STRING ) ||
            ( lit->GetType() == CT_LABEL ) ) &&
         !lit->IsBuiltin() )
      {
         String              str = strmap[lit->lit.soff];
         compact_t::iterator cit = compactstr.find( str );

         if( cit == compactstr.end() )
         {
            fprintf( stderr,
                     "Internal error #1: literal 0x%x missing compact string "
                     "0x%x\n", lit->GetId(), lit->lit.soff );
            exit( 2 );
         }

         lit->lit.soff = (*cit).second;
      }
   }

   // Copy compact string map to string map
#ifdef WR_BROKEN_STL
   strmap.erase( strmap.begin(), strmap.end() );
#else
   strmap.clear();
#endif
   for( cit = compactstr.begin(); cit != compactstr.end(); cit++ )
      strmap[(*cit).second] = (*cit).first;

   // Set nextString to size of string table
   nextString = index;

   // Generate compact maps of integer and string literals */
   intlits_t     intlits;
   strlits_t     strlits;
   reassignmap_t reassignmap;
   litmap_t      compactmap;

   for( i = 0, index = 0; i < nextLiteral; i++ )
   {
      Literal *lit;

      lit = FindLiteral( i );
      if( lit )
      {
         switch( lit->GetType() )
         {
            case CT_INT:
            case CT_LABEL:
            {
               intlits_t::iterator it = intlits.find( lit->lit.ival );
               if( it == intlits.end() )
               {
                  compactmap[index] = lit;
                  intlits[lit->lit.ival] = lit;
                  reassignmap[lit->GetId()] = index++;
               }
               else
               {
                  crit_id_t id = reassignmap[(*it).second->GetId()];
                  reassignmap[lit->GetId()] = id;
               }
               break;
            }

            case CT_STRING:
            {
               strlits_t::iterator it = strlits.find( lit->lit.soff );
               if( it == strlits.end() )
               {
                  compactmap[index] = lit;
                  strlits[lit->lit.soff] = lit;
                  reassignmap[lit->GetId()] = index++;
               }
               else
               {
                  crit_id_t id = reassignmap[(*it).second->GetId()];
                  reassignmap[lit->GetId()] = id;
               }
               break;
            }

            default:
               fprintf( stderr,
                        "Unexpected type encountered in finalization\n" );
         }
      }
   }

   // Regenerate the literal map from the compact map
#ifdef WR_BROKEN_STL
   litmap.erase( litmap.begin(), litmap.end() );
#else
   litmap.clear();
#endif
   litmap_t::iterator it;
   for( it = compactmap.begin(); it != compactmap.end(); it++ )
      litmap[(*it).first] = (*it).second;

   // Set nextLiteral to number of literals
   nextLiteral = index;

   // Remap identifiers in expressions to compact form
   expmap_t::iterator expit;
   for( expit = expmap.begin(); expit != expmap.end(); expit++ )
   {
      Expression *exp = (*expit).second;
      reassignmap_t::iterator rmit = reassignmap.find( exp->exp.arg1 );
      if( rmit != reassignmap.end() )
         exp->exp.arg1 = (*rmit).second;
      rmit = reassignmap.find( exp->exp.arg2 );
      if( rmit != reassignmap.end() )
         exp->exp.arg2 = (*rmit).second;
   }

   // Complete uncle node identifiers in regular expressions
   regexmap_t::iterator regexit;
   for( regexit = regexmap.begin(); regexit != regexmap.end(); regexit++ )
   {
      Regex *re = (*regexit).second;
      elemmap_t::iterator emit;
      for( emit = re->elemmap.begin(); emit != re->elemmap.end(); emit++ )
      {
         crit_regex_elem_t *elem = (*emit).second;
         if( ( elem->elem.op != RE_NOP ) &&
             ( elem->elem.next & CRIT_RE_UNCLE ) )
         {
            elemnodemap_t::iterator enmit = re->elemnodemap.find( elem->uncle );
            if( enmit != re->elemnodemap.end() )
               elem->elem.next |= (*enmit).second->id;
            else
            {
               fprintf( stderr, "Unable to find uncle node for element\n" );
               exit( 2 );
            }
         }
      }
   }

   // Remap identifiers in regular expressions to compact form
   for( regexit = regexmap.begin(); regexit != regexmap.end(); regexit++ )
   {
      Regex *re = (*regexit).second;
      elemmap_t::iterator emit;
      for( emit = re->elemmap.begin(); emit != re->elemmap.end(); emit++ )
      {
         crit_regex_elem_t *elem = (*emit).second;
         if( elem->elem.op == RE_CHAR )
         {
            reassignmap_t::iterator rmit =
               reassignmap.find( elem->elem.data & CRIT_ARG_MASK );
            if( rmit != reassignmap.end() )
               elem->elem.data = (elem->elem.data & ~CRIT_ARG_MASK) |
                                 (*rmit).second;
         }
      }
   }

   // Determine expressions table offsets
   crit_id_t offset = 0;
   for( i = 0; i < nextRegex; i++ )
   {
      Regex *re = regexmap[i];
      re->re.regex = offset;
      offset += re->re.regexsize;
   }
}

