/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: processor.h,v 1.25 2001/06/18 15:28:14 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Header file for criteria processor classes
 *
 */

#ifndef __PROCESSOR_H
#define __PROCESSOR_H 1

// This is an identifier was truncated to '255' characters
// warning that shows up simply by using string and map
// templates together. It can be ignored.
#ifdef WR_WINNT
#pragma warning(disable: 4786)
#endif

#ifdef WR_BROKEN_STL
#include "template/map"
#else
#include <map>
#endif

#include "wr_string.h"
#include "libre.h"

#ifndef WR_BROKEN_NAMESPACES
using namespace std;
#endif

#define CRIT_NOARG_VALUE 0xffff

typedef struct crit_regex_elem       crit_regex_elem_t;
typedef struct crit_node             crit_node_t;
typedef class Processor              Processor;
typedef class Node                   Node;
typedef class Literal                Literal;
typedef class Expression             Expression;
typedef class Regex                  Regex;
#ifdef WR_BROKEN_STL
typedef map< crit_id_t, String, less<crit_id_t> >              strmap_t;
typedef map< crit_id_t, Literal *, less<crit_id_t> >           litmap_t;
typedef map< crit_id_t, Expression *, less<crit_id_t> >        expmap_t;
typedef map< crit_id_t, Regex *, less<crit_id_t> >             regexmap_t;
typedef map< crit_id_t, crit_id_t, less<crit_id_t> >           reassignmap_t;
typedef map< crit_id_t, crit_regex_elem_t *, less<crit_id_t> > elemmap_t;
typedef map< regex_token *, crit_regex_elem_t *, less<crit_id_t> > elemnodemap_t;
#else
typedef map< crit_id_t, String >                  strmap_t;
typedef map< crit_id_t, Literal * >               litmap_t;
typedef map< crit_id_t, Expression * >            expmap_t;
typedef map< crit_id_t, Regex * >                 regexmap_t;
typedef map< crit_id_t, crit_id_t >               reassignmap_t;
typedef map< crit_id_t, crit_regex_elem_t * >     elemmap_t;
typedef map< regex_token *, crit_regex_elem_t * > elemnodemap_t;
#endif

struct crit_regex_elem
{
   crit_regex_node_t  elem;
   crit_id_t          id;
   regex_token       *node;
   regex_token       *uncle;
};

class Node
{
   protected:
      Processor *proc;

   public:
      Node( Processor * );

      virtual char      *ToString( void );
      virtual crit_id_t  GetId   ( void ) = 0;
};

class Literal : public Node
{
   protected:
      crit_literal_t lit;

      void         Initialize   ( crit_type_t );

   public:
      Literal( Processor *, int );
      Literal( Processor *, char * );
      virtual ~Literal();

      char        *ToString     ( void );
      crit_id_t    GetId        ( void );

      crit_type_t  GetType      ( void );
      crit_type_t  GetCompatType( void );

      pd_bool_t   IsBuiltin    ( void );

      int          ToUserId     ( void );
      int          ToGroupId    ( void );
      int          ToDevice     ( void );
      int          ToBuiltin    ( void );

      int          WriteOut     ( FILE * );

      friend Processor;
};

class Expression : public Node
{
   protected:
      crit_expression_t  exp;

   private:
      crit_id_t  id;

      void      Initialize( crit_op_t, crit_id_t, crit_id_t );

   public:
      Expression( Processor *, crit_op_t, Expression *, Expression * );
      Expression( Processor *, crit_op_t, Literal *, Literal * );
      Expression( Processor *, crit_op_t, Literal * );
      Expression( Processor *, crit_op_t );
      Expression( Processor *, Expression *, Literal * );
      Expression( Processor *, Literal *, Regex * );
      virtual ~Expression();

      crit_id_t GetId   ( void );
      crit_op_t GetOp   ( void );

      int       WriteOut( FILE * );

      friend Processor;
};

class Regex : public Node
{
   private:
      crit_regex_t  re;
      crit_id_t     id;
      regex        *reToken;
      int           nextElement;
      elemmap_t     elemmap;
      elemnodemap_t elemnodemap;

      crit_regex_elem_t *GenerateStructures( regex_token *tok,
                                             crit_id_t *last = NULL );
      crit_id_t          RegisterElement( crit_regex_elem_t * );
      crit_regex_elem_t *FindElement( crit_id_t );

   public:
      Regex( Processor *, char * );
      virtual ~Regex();

      crit_id_t GetId   ( void );

      int       WriteOut( FILE * );

      friend Processor;
};


class Processor
{
   private:
      Expression *rootnode;

      pd_bool_t  finalized;

      crit_id_t   nextExpression;
      crit_id_t   nextLiteral;
      crit_id_t   nextString;
      crit_id_t   nextRegex;

      expmap_t    expmap;
      litmap_t    litmap;
      strmap_t    strmap;
      regexmap_t  regexmap;

      void        Finalize            ( void );

   public:
      Processor();

      crit_id_t   RegisterExpression  ( Expression * );
      pd_bool_t  UnregisterExpression( Expression * );
      Expression *FindExpression      ( crit_id_t );

      crit_id_t   RegisterLiteral     ( Literal * );
      pd_bool_t  UnregisterLiteral   ( Literal * );
      Literal    *FindLiteral         ( crit_id_t );

      crit_id_t   RegisterString      ( char * );
      pd_bool_t  UnregisterString    ( crit_id_t );
      String     &FindString          ( crit_id_t );

      crit_id_t   RegisterRegex       ( Regex * );
      pd_bool_t  UnregisterRegex     ( Regex * );
      Regex      *FindRegex           ( crit_id_t );

      void        GenArrow            ( Expression *, Literal * );
      int         WriteOut            ( FILE *, pd_bool_t );
};

#endif /* __PROCESSOR_H */

