/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: gwmcomm.cxx,v 1.29 2001/04/05 18:41:32 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Routines to communicate with the Global Wrapper Manager service
 */

#include <wr_conf.h>
#include <pdpi/pi_defs.h>

#include "libgwmcomm.h"

#include <gwm.h>

#include <vector>

using namespace std;



/////////////////////////////////////////////////////////////////////////////////
// public class for mapping packet ids to wait handles
/////////////////////////////////////////////////////////////////////////////////
struct IdMap
{
   DWORD                      dwID;
   HANDLE                     hWait;
   LPPIPESRV_COMM_RESP_HEADER pCommHdr;
   PVOID                      pBuf;

   IdMap( DWORD id )
   {
      this->dwID = id;
      this->hWait = CreateEvent( NULL, TRUE, FALSE, NULL );
      this->pCommHdr = NULL;
      this->pBuf = NULL;
   }

   ~IdMap() { CloseHandle( this->hWait ); }

   BOOL operator==( IdMap &A ) { return this->dwID == A.dwID; }
};


/////////////////////////////////////////////////////////////////////////////////
// globals
/////////////////////////////////////////////////////////////////////////////////
HANDLE           g_hPipe = INVALID_HANDLE_VALUE;
HANDLE           g_hReadEvent = INVALID_HANDLE_VALUE;
HANDLE           g_hWriteEvent = INVALID_HANDLE_VALUE;
vector<IdMap *>  g_vIdMaps;
unsigned long    g_tid = 0;
CRITICAL_SECTION g_criticalThread;
CRITICAL_SECTION g_criticalRead;
CRITICAL_SECTION g_criticalWrite;
CRITICAL_SECTION g_criticalEventCallback;
CRITICAL_SECTION g_criticalIdMaps;

GWMCommEventCallbackFunc _gwmcommEventCB = NULL;

static volatile BOOL     g_initted = FALSE;


/////////////////////////////////////////////////////////////////////////////////
// forward declarations
/////////////////////////////////////////////////////////////////////////////////
BOOL   gwmcommConnect              ( void );
BOOL   gwmcommGetPacket            ( DWORD, LPPIPESRV_COMM_RESP_HEADER *,
                                     PVOID * );
HANDLE gwmcommGetWaitHandle        ( DWORD );
void   gwmcommProcessPacket        ( LPPIPESRV_COMM_RESP_HEADER, PVOID );
ULONG  gwmcommSpawnThread          ( void );
BOOL   gwmcommWaitForPacketResponse( DWORD, DWORD *, PVOID *, DWORD * );


/////////////////////////////////////////////////////////////////////////////////
// macros
/////////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG_CRITSEC
#define gwmcommEnterCriticalSection(a) \
   do { \
      _tprintf( PD_MAKESTR("EnterCriticalSection(%s:0x%x) @ %s:%d\n"), #a, &(a), \
                __FILE__, __LINE__ ); \
      EnterCriticalSection( &(a) ); \
   } while(0)
#define gwmcommLeaveCriticalSection(a) \
   do { \
      _tprintf( PD_MAKESTR("LeaveCriticalSection(%s:0x%x) @ %s:%d\n"), #a, &(a), \
                __FILE__, __LINE__ ); \
      LeaveCriticalSection( &(a) ); \
   } while(0)
#else /* DEBUG_CRITSEC */
#define gwmcommEnterCriticalSection(a) EnterCriticalSection( &(a) );
#define gwmcommLeaveCriticalSection(a) LeaveCriticalSection( &(a) );
#endif /* DEBUG_CRITSEC */


/////////////////////////////////////////////////////////////////////////////////
// private functions
/////////////////////////////////////////////////////////////////////////////////
BOOL
gwmcommConnect()
{
   g_hPipe = CreateFile( GWM_CLIENT_PIPE, ( GENERIC_READ | GENERIC_WRITE ), 0,
                         NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL |
                         FILE_FLAG_WRITE_THROUGH | FILE_FLAG_NO_BUFFERING |
                         FILE_FLAG_OVERLAPPED, NULL );

   if( g_hPipe == INVALID_HANDLE_VALUE )
   {
      if( GetLastError() != ERROR_PIPE_BUSY )
         return FALSE;
      if( !WaitNamedPipe( GWM_CLIENT_PIPE, 20000 ) )
         return FALSE;
   }

   return TRUE;
}

BOOL
gwmcommReadPacket( LPPIPESRV_COMM_RESP_HEADER pCommHdr, PVOID *pBuf )
{
   OVERLAPPED os;
   DWORD      dwWait;
   DWORD      cbRead;
   BOOL       bRet;

   *pBuf = NULL;

   gwmcommEnterCriticalSection( g_criticalRead );

   // set up the overlapped structure
   //
   memset( &os, 0, sizeof(os) );
   os.hEvent = g_hReadEvent;
   ResetEvent( g_hReadEvent );

#ifdef DEBUG_IO
   _tprintf( MAKESTR("ReadFile(1): waiting for %ld bytes\n"), sizeof(*pCommHdr) );
#endif /* DEBUG_IO */

   bRet = ReadFile( g_hPipe, pCommHdr, sizeof(*pCommHdr), &cbRead, &os );
   if( !bRet )
   {
      switch( GetLastError() )
      {
         case ERROR_IO_PENDING:
            dwWait = WaitForSingleObject( g_hReadEvent, INFINITE );
            switch( dwWait )
            {
               case WAIT_OBJECT_0:
                  bRet = TRUE;
                  break;
               default:
                  break;
            }
            break;
         default:
            break;
      }
   }

#ifdef DEBUG_IO
   if( bRet )
      _tprintf( MAKESTR("ReadFile(1): read %ld bytes\n"), sizeof(*pCommHdr) );
#endif /* DEBUG_IO */

   if( bRet )
   {
      DWORD buflen = pCommHdr->pktlen - sizeof(*pCommHdr);

      if( buflen )
      {
         *pBuf = malloc( buflen );

         // set up the overlapped structure
         //
         memset( &os, 0, sizeof(os) );
         os.hEvent = g_hReadEvent;
         ResetEvent( g_hReadEvent );

#ifdef DEBUG_IO
         _tprintf( MAKESTR("ReadFile(2): waiting for %ld bytes\n"), buflen );
#endif /* DEBUG_IO */

         bRet = ReadFile( g_hPipe, *pBuf, buflen, &cbRead, &os );
         if( !bRet )
         {
            switch( GetLastError() )
            {
               case ERROR_IO_PENDING:
                  dwWait = WaitForSingleObject( g_hReadEvent, INFINITE );
                  switch( dwWait )
                  {
                     case WAIT_OBJECT_0:
                        bRet = TRUE;
                        break;
                     default:
                        break;
                  }
                  break;
               default:
                  break;
            }
         }

#ifdef DEBUG_IO
         if( bRet )
            _tprintf( MAKESTR("ReadFile(2): read %ld bytes\n"), buflen );
#endif /* DEBUG_IO */
      }
   }

   if( bRet )
      gwmcommProcessPacket( pCommHdr, *pBuf );
   else
   {
      if( pCommHdr )
         free( pCommHdr );

      if( *pBuf )
         free( *pBuf );
   }

   gwmcommLeaveCriticalSection( g_criticalRead );

   return bRet;
}

void
gwmcommWorkThread( PVOID args )
{
   // if we can't connect assume the gwm isn't running
   //
   if( g_hPipe == INVALID_HANDLE_VALUE )
      if( gwmcommConnect() == FALSE )
      {
         _endthread();
         return;
      }

   // process incoming packets
   //
   while( 1 )
   {
      PVOID                  pBuf     = NULL;
      LPPIPESRV_COMM_RESP_HEADER pCommHdr =
         (LPPIPESRV_COMM_RESP_HEADER)malloc( sizeof(*pCommHdr) );

      gwmcommReadPacket( pCommHdr, &pBuf );
   }
}

IdMap *
gwmcommFindID( DWORD dwID, BOOL bRemove, BOOL bCreate )
{
   for( vector<IdMap *>::iterator i = g_vIdMaps.begin();
        i != g_vIdMaps.end();
        i++ )
   {
      if( dwID == (*i)->dwID )
      {
         IdMap *idMap = (*i);

         if( bRemove )
            g_vIdMaps.erase( i );

         return idMap;
      }
   }

   if( bCreate )
   {
      IdMap *idMap = new IdMap( dwID );

      if( !bRemove )
         g_vIdMaps.push_back( idMap );

      return idMap;
   }

   return NULL;
}

HANDLE
gwmcommGetWaitHandle( DWORD dwID )
{
   gwmcommEnterCriticalSection( g_criticalIdMaps );
   IdMap *idMap = gwmcommFindID( dwID, FALSE, TRUE );

   HANDLE hMap = idMap->hWait;
   gwmcommLeaveCriticalSection( g_criticalIdMaps );
   return hMap;
}

BOOL
gwmcommGetPacket( DWORD dwID, LPPIPESRV_COMM_RESP_HEADER *pCommHdr,
                  PVOID *pPBuf )
{
   gwmcommEnterCriticalSection( g_criticalIdMaps );
   IdMap *idMap = gwmcommFindID( dwID, TRUE, FALSE );
   gwmcommLeaveCriticalSection( g_criticalIdMaps );

   if( idMap == NULL )
      return FALSE;

   if( pCommHdr )
      *pCommHdr = idMap->pCommHdr;
   if( pPBuf )
      *pPBuf = idMap->pBuf;

   delete idMap;

   return TRUE;
}

void
gwmcommProcessPacket( LPPIPESRV_COMM_RESP_HEADER pCommHdr, PVOID pBuf )
{
   DWORD dwID = GWM_COMM_ID( *pCommHdr );

   switch( GWM_COMM_CMD( *pCommHdr ) )
   {
      case WRV_EVENT_POST:
         if( _gwmcommEventCB )
            (*_gwmcommEventCB)( pBuf, pCommHdr->pktlen -
                                      sizeof(PIPESRV_COMM_RESP_HEADER) );
         else
            free( pBuf );

         free( pCommHdr );

         break;
      default:
         gwmcommEnterCriticalSection( g_criticalIdMaps );
         IdMap *idMap = gwmcommFindID( dwID, FALSE, TRUE );
         if( idMap != NULL )
         {
            idMap->pCommHdr = pCommHdr;
            idMap->pBuf = pBuf;

            SetEvent( idMap->hWait );
#ifdef DEBUG_IO
            _tprintf( MAKESTR("gwmcommProcessPacket triggered event\n") );
#endif /* DEBUG_IO */
         }
         else
         {
#ifdef DEBUG_IO
            _tprintf( MAKESTR("Warning: gwmcommFindID failed for unknown packet\n") );
#endif /* DEBUG_IO */
            free( pCommHdr );
            free( pBuf );
         }
         gwmcommLeaveCriticalSection( g_criticalIdMaps );
   }
}

ULONG gwmcommSpawnThread()
{
   gwmcommEnterCriticalSection( g_criticalThread );
   if( g_tid == 0 )
      g_tid = _beginthread( gwmcommWorkThread, 0, NULL );
   gwmcommLeaveCriticalSection( g_criticalThread );

   return g_tid;
}

LPTSTR
gwmcommGetLastErrorText( LPTSTR lpszBuf, DWORD dwSize )
{
   DWORD dwRet;
   LPTSTR lpszTemp = NULL;

   dwRet = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                          FORMAT_MESSAGE_FROM_SYSTEM |
                          FORMAT_MESSAGE_ARGUMENT_ARRAY,
                          NULL,
                          GetLastError(),
                          LANG_NEUTRAL,
                          (LPTSTR)&lpszTemp,
                          0,
                          NULL );

   // supplied buffer is not long enough
   if( !dwRet || ( (long)dwSize < (long)dwRet+14 ) )
      lpszBuf[0] = PD_MAKESTR('\0');
   else
   {
      lpszTemp[lstrlen(lpszTemp)-2] = PD_MAKESTR('\0');
      _stprintf( lpszBuf, PD_MAKESTR("%s (0x%x)"), lpszTemp, GetLastError() );
   }

   if( lpszTemp )
      LocalFree( (HLOCAL)lpszTemp );

   return lpszBuf;
}


/////////////////////////////////////////////////////////////////////////////////
// public functions
/////////////////////////////////////////////////////////////////////////////////
void EXPORT
gwmcommSetup()
{
   if( !g_initted )
   {
      // TBD: change to BOOL return and return appropriately on error

      InitializeCriticalSection( &g_criticalThread );
      InitializeCriticalSection( &g_criticalRead );
      InitializeCriticalSection( &g_criticalWrite );
      InitializeCriticalSection( &g_criticalEventCallback );

      g_hReadEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
      g_hWriteEvent = CreateEvent( NULL, TRUE, FALSE, NULL );

      InitializeCriticalSection( &g_criticalIdMaps );

      g_initted = TRUE;
   }
}

DWORD EXPORT
gwmcommWritePacket( DWORD dwCommand, DWORD dwDatalen, PVOID pData )
{
   OVERLAPPED          os;
   DWORD               dwWait;
   PIPESRV_COMM_HEADER commHdr;
   DWORD               cbWritten;
   DWORD               dwID;
   BOOL                bRet;

   if( g_hPipe == INVALID_HANDLE_VALUE )
      if( !gwmcommConnect() )
         return 0;
   
   // set up the packet header
   //
   dwID = GetTickCount() << GWM_COMM_ID_SHIFT;
   commHdr.commandId = ( dwID & GWM_COMM_ID_MASK ) |
                       ( dwCommand & GWM_COMM_CMD_MASK );
#ifdef UNICODE
   commHdr.flags = GWM_COMM_UNICODE;
#else
   commHdr.flags = 0;
#endif /* UNICODE */
   commHdr.pid = _getpid();
   commHdr.tid = GetCurrentThreadId();
   commHdr.pktlen = sizeof(commHdr) + dwDatalen;

   gwmcommEnterCriticalSection( g_criticalWrite );

   // set up the overlapped structure
   //
   memset( &os, 0, sizeof(os) );
   os.hEvent = g_hWriteEvent;
   ResetEvent( g_hWriteEvent );

   // send out the header...
   //
#ifdef DEBUG_IO
   _tprintf( PD_MAKESTR("WriteFile(1): writing %ld bytes\n"), sizeof(commHdr) );
#endif /* DEBUG_IO */

   bRet = WriteFile( g_hPipe, &commHdr, sizeof(commHdr), &cbWritten, &os );
   if( !bRet )
   {
      switch( GetLastError() )
      {
         case ERROR_IO_PENDING:
            dwWait = WaitForSingleObject( g_hWriteEvent, INFINITE );
            switch( dwWait )
            {
               case WAIT_OBJECT_0:
                  break;
               case WAIT_FAILED:
               default:
                  gwmcommLeaveCriticalSection( g_criticalWrite );
                  return 0;
            }
            break;
         default:
            gwmcommLeaveCriticalSection( g_criticalWrite );
            return 0;
      }
   }

#ifdef DEBUG_IO
   _tprintf( PD_MAKESTR("WriteFile(1): wrote %ld bytes\n"), sizeof(commHdr) );
#endif /* DEBUG_IO */

   // send out the data...
   //
   if( dwDatalen > 0 )
   {
      // set up the overlapped structure
      //
      memset( &os, 0, sizeof(os) );
      os.hEvent = g_hWriteEvent;
      ResetEvent( g_hWriteEvent );

#ifdef DEBUG_IO
      _tprintf( PD_MAKESTR("WriteFile(2): writing %ld bytes\n"), dwDatalen );
#endif /* DEBUG_IO */

      bRet = WriteFile( g_hPipe, pData, dwDatalen, &cbWritten, &os );
      if( !bRet )
      {
         switch( GetLastError() )
         {
            case ERROR_IO_PENDING:
               dwWait = WaitForSingleObject( g_hWriteEvent, INFINITE );
               switch( dwWait )
               {
                  case WAIT_OBJECT_0:
                     break;
                  case WAIT_FAILED:
                  default:
                     gwmcommLeaveCriticalSection( g_criticalWrite );
                     return 0;
               }
               break;
            default:
               gwmcommLeaveCriticalSection( g_criticalWrite );
               return 0;
         }
      }

#ifdef DEBUG_IO
      _tprintf( PD_MAKESTR("WriteFile(2): wrote %ld bytes\n"), dwDatalen );
#endif /* DEBUG_IO */
   }

   gwmcommLeaveCriticalSection( g_criticalWrite );

   // spawn the work thread
   //
   gwmcommSpawnThread();

   return dwID;
}

BOOL EXPORT
gwmcommWaitForPacketResponse( DWORD dwID, DWORD *error, PVOID *pData, DWORD *pLen )
{
   LPPIPESRV_COMM_RESP_HEADER pCommHdr;
   HANDLE                     hWait;
   DWORD                      dwWait;

   hWait = gwmcommGetWaitHandle( dwID );

dowait:
   dwWait = WaitForSingleObject( hWait, GWM_WAIT_TIMEOUT );
   switch( dwWait )
   {
      case WAIT_TIMEOUT:
      {
         if( gwmcommWritePacket( WRV_COMM_SYNC_RESEND, sizeof(dwID), &dwID ) )
            goto dowait;
         else
            return FALSE;
      }

      case WAIT_OBJECT_0:
         break;

      default:
         return FALSE;
   }

   if( !gwmcommGetPacket( dwID, &pCommHdr, pData ) )
      return FALSE;

   if( pLen )
      *pLen = pCommHdr->pktlen - sizeof(PIPESRV_COMM_RESP_HEADER);
   if( error )
      *error = pCommHdr->errorCode;

   free( pCommHdr );

   return TRUE;
}

void EXPORT
gwmcommRegisterEventCallback( GWMCommEventCallbackFunc eventCB )
{
   gwmcommEnterCriticalSection( g_criticalEventCallback );
   _gwmcommEventCB = eventCB;
   gwmcommLeaveCriticalSection( g_criticalEventCallback );
}

void EXPORT
gwmcommFreeData( PVOID pBuf )
{
   if( pBuf )
      free( pBuf );
}

