/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: codegen.cc,v 1.92 2001/06/13 13:22:54 stevek Exp $
 *
 * Authors: Lee Badger, Mark Feldman, Ben Uecker
 *
 * Purpose: Object to generate and store C-code representing a wrapper.
 *	Source objects are generated at the leaves of the parse
 *	tree, and are merged as rules are reduced to be a single source
 * 	object when parsing completes.  The object can then dump its C text.
 */

#include "wr_conf.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libwdl.h"
#include "codegen.h"
#include "codefrags.h"
#include "event.h"
#include "expr.h"
#include "headers.h"

extern EvSet 		*event_set;
extern EvSubSysTbl   	essgrp;
extern EvSubSysTbl      esstbl;

/* linked list of text */
struct srcrec {
	char *txt;
	struct srcrec *next;
};


source::source()
{
	flags = 0;
	paren_count = 0;
	head = tail = cur = NULL;
        d_type = NULL;
        vname = NULL;
        label = NULL;
        offset = NULL;
}


int source::is_empty()
{
  return (head == NULL);
}


void source::make_header(source *link_decl)
{
  MAKE_HEADER(std_includes, src_header);
  append(link_decl);
  append(src_header1);
}


void source::gen_seq_step_funcs()
{
	append(seq_step_funcs_str);
}


/* Append some ascii text; probably should use the SLIST macro. */
void source::append(const char *txt)
{
  if (txt == NULL || *txt == '\0')
    return;
  
  if (head == NULL) {
    head = tail = new srcrec;
    head->txt = strdup(txt);
    head->next = NULL;
  }
  else {
    struct srcrec *tmp = new srcrec;
    tmp->txt = strdup(txt);
    tail->next = tmp;
    tail = tmp;
    tail->next = NULL;
    tmp = NULL;
  }
}
void source::prepend(const char *txt)
{
    if (txt == NULL || *txt == '\0')
        return;
    if (head == NULL) {
        append(txt);
    } else {
        struct srcrec *tmp = new srcrec;
        tmp->txt = strdup(txt);
        tmp->next = head;
        head = tmp;
        tmp = NULL;
    }
}

#ifdef	notyet
/* The following members just make adding strings more convenient. */
void source::append(const char *txt1, const char *txt2)
{
	append(txt1); append(txt2);
}
void source::append(const char *txt1, const char *txt2, const char *txt3)
{
	append(txt1); append(txt2); append(txt3);
}
void source::append(const char *txt1, const char *txt2, const char *txt3,
		    const char *txt4)
{
	append(txt1); append(txt2); append(txt3); append(txt4);
}
#endif

/* append an integer */
void source::append(int n)
{
  char s[22];  /* big enough for a 64-bit int, its sign, and a NULL */
  
  sprintf(s, "%d", n);
  append(s);
}


/* append another source obj */
void source::append(source *src)
{
  if (src == NULL || src->head == NULL)
    return;
  if (head == NULL) {
    head = src->head;
    tail = src->tail;
  }
  else {
    tail->next = src->head;
    tail = src->tail;
    src->head = src->tail = NULL;	/* empty second object */
  }
}

/* append another expression obj */
void source::append(Expr *e)

{
    append(e->code);
    return;
}

int source::operator==(char *s)

{
     char *s1 = all();
     int i = !strcmp(s, s1);
     free(s1);
     return i;
}


void source::rewind()
{
	cur = head;
}

/* indent the text by the space given in space */
void source::indent(const char *space)

{
     struct srcrec *tmp = head;
     // insert leading indent space for the first line
     if (tmp != NULL) {
           head = new srcrec;
           head->txt = strdup(space);
           head->next = tmp;
     };
     while (tmp != tail) {
          // insert indent space after each \n */
          if (tmp->txt[strlen(tmp->txt) - 1] == '\n') {
               struct srcrec *srec = new srcrec;
               srec->next = tmp->next;
               srec->txt = strdup(space);
               tmp->next = srec;
               tmp = srec->next;
          } else 
               tmp = tmp->next;
     }
};

/* Return true if there is more text to retrieve via next(). */
bool source::more_text()
{
	if (cur) return true;
	else return false;
}



/* Return next text string in the object, and step forward. */
char *source::next()
{
	if (!cur) return(NULL);
	struct srcrec *tmp = cur;
	cur = cur->next;
	return(tmp->txt);
}


/* Return a single text string containing everything in the object */
char *source::all()
{
  char *tmp;
  char *ret;
  size_t len = 0;
  size_t offset = 0;

  rewind();
  while (more_text()) 
    len += strlen(next());
  ret = (char *) malloc(len + 1);
  if (ret == NULL)
    return NULL;
  rewind();
  while (more_text()) {
    tmp = next();
    len = strlen(tmp);
    strncpy(ret + offset, tmp, len);
    offset += len;
  }
  ret[offset] = '\0';
  return ret;
}

int source::rec_count() 

{   int i = 0;
    rewind();
    while (more_text()) {
        next();
        i++;
    }
    return i;
}

bool source::find(char *txt)
{
  bool rtn = false;

  rewind();
  while(more_text()) {
    if(strcmp(cur->txt, txt) == 0) {
      rtn = true;
      break;
    }
    cur = cur->next;
  }
  return rtn;
}

// insert txt after cur pointer;
void source::insert(char *txt)

{
     if (txt == NULL || *txt == '\0') return;
     if (cur == NULL) {
          struct srcrec *tmp = head;
          head = new srcrec;
          head->txt = strdup(txt);
          head->next = tmp;
     } else if (cur == tail) {
          struct srcrec *tmp = new srcrec;
          tmp->txt = strdup(txt);
          tmp->next = NULL;
          cur->next = tmp; 
          tail = tmp; 
     } else {
          struct srcrec *tmp = new srcrec;
          tmp->txt = strdup(txt);
          tmp->next = cur->next;;
          cur->next = tmp; 
     }
};

void source::remove()
{
  srcrec *prev;

  if (!cur)
    return;

  if (head == cur) {
    prev = NULL;
    head = head->next;
    if (tail == cur)
      tail = head;
  }
  else {
    for (prev = head; prev->next != cur; prev = prev->next);
    prev->next = cur->next;
    if (tail == cur)
      tail = prev;
  }
  if (cur->txt)
    free(cur->txt);
  free(cur);
  if (prev)
    cur = prev->next;
  else
    cur = head;
}


char *source::text()
{
  if (!cur)
    return NULL;
  else
    return cur->txt;
}


void source::set_flags(int fl)
{
  flags = fl;
}


int source::get_flags()
{
  return flags;
}


void source::gen_tbd()
{
  append ("\n/* Code to be added TBD */\n");
}


void source::gen_tbd(char *txt)
{
  append ("\n/* Code to be added TBD for ");
  append (txt);
  append (" */\n");
}


void source::gen_debug_code()
{
	append("\n/* Test driver code. */\n#include \"debug.c\"\n");
}


void source::gen_ascii(char *str)
{
        append("\n#define MYNAME PD_MAKESTR(\"");
	append(str);
	append("\")\n");
	append("\n");
        append("static pd_char_t *WRLINK_MOD_NAME = MYNAME;\n");
        append("static pd_char_t *myname=MYNAME;\n\n");

        append("pd_char_t EXPORT **get_mod_name()\n");
        append("{\n");
        append("   return &WRLINK_MOD_NAME;\n");
        append("}\n\n");
}

void source::gen_comment(char *str, char *str1)

{
	append("/* ");
        append(str);
        append(str1);
        append(" */\n");
}


/*
 * Generate code to create a (global or local) table.  If a local
 * table is being created, the wss needs to export the get_pid()
 * function so the pid can be used to make local table names unique.
 *
 * Note: this function assumes that paren_count has been set on cols
 * by parsing code.
 */
void source::gen_wql_create(
        enum tabsc scope, char *tabname, source *cols, int count)
{
	append("wql_create(");
        switch (scope) {
        case global_table:
            append("PD_MAKESTR(\"\"), 0, ");
            break;
        case wrapper_table:
            append("myname, 0, ");		/* wrapper name */
            break;
        case instance_table:
            append("myname, _pid, ");
            break;
        }

	append("PD_MAKESTR(\"");
	append(tabname);
	append("\"), !NULL, ");
        /* Clone cols so we can use it again. */
	append(cols->all());			/* column definitions */

        SourceAppend3(this, ", ", count, ", &");
        if (scope == instance_table) {
            SourceAppend3(this,"WSS_GET_WIV_VAR(",tabname,")");
        } else {
            append(tabname);
        }
        append(");\n");
}

extern int table_elem_count;
extern VarStore vstore;
extern const char *current_table;
void wdlerror(const char *s, const char *s1 = "");
int wdlwarning(const char *s, const char *s1 = "");
void source::gen_wql_value(source *exp)
{
        VarNode *vn = NULL;

        if (current_table)
            vn = vstore[current_table];
        if (!vn || vn->type.tc != TC_DBTABLE || !vn->type.info)
            wdlerror("Unknown table.");
        TableTypeInfo *tti=(TableTypeInfo*)vn->type.info;
        /* Get the type for the current column */
        char buf[5];
        snprintf(buf, sizeof buf, "%d", table_elem_count);
        char *cp = (*tti->num2name)[buf];
        if (!cp)
            wdlerror("Invalid number of colums in insert.");
        Type *t = &((*tti->vtbl)[cp]->type);

        if (exp->d_type->is_strtype() && t->is_strtype() ) {
		append("pd_char_t_t, wr_strdup(");
		append(exp);
		append(")");
        } else if ((exp->d_type->is_inttype() || exp->d_type->is_floattype())
            && t->is_inttype()) {
		append("int_t, ");
		append(exp);
        } else if ((exp->d_type->is_floattype() || exp->d_type->is_inttype())
            && t->is_floattype()) {
		append("float_t, ");
		append(exp);
        } else {
                wdlerror("Invalid type in insert value list.");
        }

	paren_count = 1;
}


/*
 * Generate code to find a table.
 */
void source::gen_wql_find(enum tabsc scope, char *tabname)
{
    if (scope == instance_table) {
        SourceAppend3(this, "if (WSS_GET_WIV_VAR(",tabname,") == 0) {\n");
    } else {
        append("if (tabname == 0) {\n");
    }
    append("    wql_find_table_by_names(");
    switch (scope) {
    case global_table:
        append("\"\", 0, ");
        break;
    case instance_table: /* fall throught */
        append("myname, _pid, PD_MAKESTR(\"");
    case wrapper_table:
        append("myname,    0, PD_MAKESTR(\"");
        break;
    }
    append(tabname);
    append("\"));\n");
    append("}\n");
}

void source::gen_wql_insert(enum tabsc scope, char *tabname, source *rows)
{
	append("wql_insert(");
        if (scope == instance_table) {
            SourceAppend3(this,"WSS_GET_WIV_VAR(",tabname,")");
        } else {
            append(tabname);
        }
	append(", !NULL, ");
	append(rows);
	append("NULL");

	/* now close the nested function invocation */
	for (int i=1; i<=rows->paren_count+1; i++) {
		append(")");
	}
	append(";\n");
}


void source::gen_argrepl_code(source *lhs, source *rhs)

{
      source *sp = this;
      char *rhs_str = rhs->all();
      char *fnname;
      if (lhs->label->para->is_out()) {
          fnname = "(void)wr_arg_copy(";
      } else {
          fnname = "(void)wr_arg_replace(wrp, CUR_THREAD_ID, space, code, ";
      }
      if (lhs->label->para->is_nterm()) {
           SourceAppend8(sp, fnname, " uap, ", rhs_str, 
                             ", ", lhs->offset, ", sizeof(pd_char_t) * (wr_strlen(", rhs_str, 
                             ")+1));\n");
      } else if (lhs->label->para->d_type->is_inttype()) {
           SourceAppend6(sp, fnname, " uap, (void *)",
                             rhs_str, ", ", lhs->offset, ", 0);\n");
      } else if (lhs->label->para->is_ntermlist()) {
          /* This isn't safe for copying out...
           * Currently no syscalls use "*nterm" for "return" blocks.
           * If that changes, life sucks.
           */
           /* Order of business:
            * 1) Copy the vector to user space and point uap[index] to it.
            *    wr_arg_replace does this.
            * 2) Copy all the buffers pointed to into user space and point
            *    uap[index][i] to them.
            */
           source *sp1 = new source;
           sp1->append("{\n");
           /* eval rhs_str once to hopefully avoid multiple expansion
            * of side effects.
            */
           SourceAppend3(sp1, "char **spc = ", rhs_str, ";\n");
           SourceAppend1(sp1, "int lsize = ntermlistsize(spc);\n");
           SourceAppend1(sp1, "char **dpc;\n");
           SourceAppend1(sp1, "char *tmp;\n");
           SourceAppend4(sp1, fnname, " uap, (void *) spc, ",
                              lhs->offset, ",lsize + 1);\n");
           /* Get dpc now to get the new value */
           SourceAppend3(sp1, "dpc = ", lhs, ";\n");
           /* Copy new strings into user space and assign new ptr to
            * point to them.
            * This complicated by the fact that dpc is in user space,
            * and spc is in kernel space.  Have fun.
            * (Based on:
            *     int *user_i; int kernel_i = 3;
            * //  *user_i = kernel_i;  is done in a VM same manner by
            *     copyout(&kernel_i, user_i, sizeof(int));
            * )
            */
           sp1->append(
                   "while (*spc != NULL) {\n"
                   "    tmp = wr_curproc_strdup(*spc);\n"
                   "    wr_copyout(&tmp, dpc, sizeof(tmp));\n"
                   "    ++spc; ++dpc;\n"
                   "}\n"
                   "*dpc = NULL;\n" );
           sp1->append("}\n");
           sp1->indent("    ");
           sp->append(sp1);
      } else {
           /* The code for WrEvent::gen_argrepl_code also recognizes
            * out buffers.
            */
      }
}

void source::gen_argcopyin(mlabel *l, char *offset)

{
      source *body = new source;
      if (l->para->is_nterm()) {
           source *dst = new source;
           source *src = new source;
           SourceAppend3(dst, "uapk[ ", offset, " ]");
           SourceAppend3(src, "uap[ ", offset, " ]");
           body->append("int done;\n");
           SourceAppend5(body, "if (", src->all(), "==NULL)", dst->all(), "=NULL;\n");
           body->append("else {\n");
           SourceAppend2(body, dst->all(), " = wr_malloc(MAXPATHLEN*sizeof(pd_char_t));\n");
           SourceAppend5(body, "wr_copyinstr(", src, ", ", 
                                dst, ", MAXPATHLEN, &done);\n");
           body->append("}\n");
           delete dst; delete src;
      } else if (l->para->is_ntermlist()) {
           source *dst = new source;
           source *src = new source;
           SourceAppend3(src, "uap[ ", offset, " ]");
           SourceAppend3(dst, "uapk[ ", offset, " ]");
           SourceAppend4(body, dst, " = wr_dupin_ntermlist(", src, ");\n");
           delete dst; delete src;
      } else {
           EvSet::iterator iter;
           if (event_set->size() > 1) {
               for (iter = event_set->begin(); iter != event_set->end(); iter++)
               {
                    char *pn = l->data->all();
                    WrEvent *cev = essgrp.get_event(*iter);
                    source *sp1 = cev->gen_argcopyin_code(pn);
                    sp1->indent("    ");
                    int code = esstbl[cev->space]->get_code(*iter);
                    if (iter == event_set->begin()) {
                         SourceAppend3(body, "if (code == ", code, ") ");
                    } else {
                         SourceAppend3(body, "else if (code == ", code, ") ");
                    }
                    SourceAppend3(body, "{\n", sp1, "}\n");
               }
           } else {
               iter = event_set->begin();
               char *pn = l->data->all();
               WrEvent *cev = essgrp.get_event(*iter);
               source *sp1 =cev->gen_argcopyin_code(pn);
               body->append(sp1);
           }
      }
      body->indent("    ");
      SourceAppend5(this,
              "if (!args_in_userspace) {\n",
              "    wr_bcopy(uap, uapk, 10 * sizeof(void*));\n",
              "} else {\n", body, "}\n");
      return;
};

void source::gen_argcopyout(mlabel *l, char *offset)
{
      source *body = new source;
      if (l->para->is_nterm()) {
           source *dst = new source;
           source *src = new source;
           SourceAppend3(dst, "uap[ ", offset, " ]");
           SourceAppend3(src, "uapk[ ", offset, " ]");
           body->append("int done;\n");
           SourceAppend7(body, "wr_copyout(", src, ", ", dst->all(),
                               ", strlen(", dst, ") + 1, &done);\n");
           SourceAppend3(body, "wr_free(", src, ");\n");
           delete dst; delete src;
      } else if (l->para->is_ntermlist()) {
           wdlerror("ntermlist as output param not handled\n");
      } else {
           EvSet::iterator iter;
           if (event_set->size() > 1) {
               for (iter = event_set->begin(); iter != event_set->end(); iter++)
               {
                    char *pn = l->data->all();
                    WrEvent *cev = essgrp.get_event(*iter);
                    source *sp1 = cev->gen_argcopyout_code(pn);
                    sp1->indent("    ");
                    int code = esstbl[cev->space]->get_code(*iter);
                    if (iter == event_set->begin()) {
                         SourceAppend3(body, "if (code == ", code, ") ");
                    } else {
                         SourceAppend3(body, "else if (code == ", code, ") ");
                    }
                    SourceAppend3(body, "{\n", sp1, "}\n");
               }
           } else {
               iter = event_set->begin();
               char *pn = l->data->all();
               WrEvent *cev = essgrp.get_event(*iter);
               source *sp1 =cev->gen_argcopyout_code(pn);
               body->append(sp1);
           }
      }
      body->indent("    ");
      SourceAppend5(this,
              "if (!args_in_userspace) {\n",
              "    wr_bcopy(uapk, uap, 10 * sizeof(void*));\n",
              "} else {\n", body, "}\n");
      return;
}

void source::gen_argfree(mlabel *l, char *offset)
{
      source *body = new source;
      if (l->para->is_nterm()) {
           SourceAppend3(body, "wr_free(uapk[ ", offset , " ]);\n");
      } else if (l->para->is_ntermlist()) {
          SourceAppend3(body, "wr_free_ntermlist( uapk[ ", offset, " ] );\n" );
      } else {
          EvSet::iterator iter;
          if (event_set->size() > 1) {
              for (iter = event_set->begin(); iter != event_set->end(); iter++)
              {
                  char *pn = l->data->all();
                  WrEvent *cev = essgrp.get_event(*iter);
                  source *sp1 = cev->gen_argfree_code(pn);
                  sp1->indent("    ");
                  int code = esstbl[cev->space]->get_code(*iter);
                  if (iter == event_set->begin()) {
                      SourceAppend3(body, "if (code == ", code, ") ");
                  } else {
                      SourceAppend3(body, "else if (code == ", code, ") ");
                  }
                  SourceAppend3(body, "{\n", sp1, "}\n");
              }
          } else {
              iter = event_set->begin();
              char *pn = l->data->all();
              WrEvent *cev = essgrp.get_event(*iter);
              source *sp1 =cev->gen_argfree_code(pn);
              body->append(sp1);
          }
      }
      body->indent("    ");
      SourceAppend3(this, "if (!args_in_userspace); else {\n", body, "}\n");
      delete body;
      return;
}

void source::gen_argrepl(mlabel *l, char *offset, source *rhs)

{
      source *sp = new source;
      char *fnname = "(void)wr_arg_replace(wrp, CUR_THREAD_ID, space, code, ";
      if (l->para->is_nterm()) {
           source *src = new source;
           SourceAppend3(src, "((pd_char_t *) uapk[ ", offset, " ])")
           if (rhs == NULL) rhs = src;
           SourceAppend3(sp, "if( ", rhs->all(), "!=NULL ) {\n" );
           SourceAppend8(sp, fnname, "uap, ", rhs->all(), ", ", offset, 
                             ", sizeof(pd_char_t) * (wr_strlen(", rhs->all(), ")+1));\n");
           SourceAppend3(sp, "wr_free(", rhs, ");\n");
           sp->append("}\n");
      } else if (l->para->d_type->is_inttype()) {
           source *src = new source;
           SourceAppend3(src, "((int) uapk[ ", offset, " ])")
           if (rhs == NULL) rhs = src;
           SourceAppend6(sp, fnname, " uap, (void *)",
                             rhs, ", ", offset, ", 0);\n");
      } else if (l->para->is_ntermlist()) {
           source *src = new source;
           SourceAppend3(src, "((pd_char_t **) uapk[ ", offset, " ])")
           if (rhs == NULL) rhs = src;
           SourceAppend5(sp, "wr_replace_ntermlist(wrp, uap, ", rhs->all(), 
                             ", ", offset, ");\n");
           SourceAppend3(sp, "wr_free_ntermlist(", rhs, ");\n");
      } else {
           EvSet::iterator iter;
           source *src = new source;
           SourceAppend3(src, "uapk[ ", offset, " ]");
           if (rhs == NULL) rhs = src;
           if (event_set->size() > 1) {
               for (iter = event_set->begin(); iter != event_set->end(); iter++)
               {
                    char *pn = l->data->all();
                    WrEvent *cev = essgrp.get_event(*iter);
                    source *sp1 = cev->gen_argrepl_code(pn, rhs);
                    sp1->indent("    ");
                    int code = esstbl[cev->space]->get_code(*iter);
                    if (iter == event_set->begin()) {
                         SourceAppend3(sp, "if (code == ", code, ") ");
                    } else {
                         SourceAppend3(sp, "else if (code == ", code, ") ");
                    }
                    SourceAppend3(sp, "{\n", sp1, "}\n");
               }
           } else {
               iter = event_set->begin();
               char *pn = l->data->all();
               WrEvent *cev = essgrp.get_event(*iter);
               source *sp1 =cev->gen_argrepl_code(pn, rhs);
               // sp1->indent("    ");
               // SourceAppend3(sp, " {\n", sp1, "}\n");
               sp->append(sp1);
           }
      }
      sp->indent("    ");
      SourceAppend5(this,
              "if (!args_in_userspace) {\n",
              "    wr_log(PD_MAKESTR(\"arg_repl "
                          "not supported in sysspace.\\n\"));\n",
              "} else {\n", sp, "}\n");
}

void source::gen_wr_uninstall(source *supplied_body, source *dbcleanup)
{
    this->append("\n/* lifecycle function */\n");
    this->append("int EXPORT wr_uninstall ( )          \n");
    this->append("{                                    \n");
    this->append(supplied_body);
    this->append(dbcleanup);
    this->append("\n");
    this->append("    return WR_SUCCESS;\n}\n");
}

void source::gen_wr_install(source *init, source *supplied_body)
{
    init->indent("    ");
    this->append("\n/* lifecycle function */\n");
    this->append("int EXPORT wr_install ( int newhandle )\n{\n");
    this->append("    myhandle = newhandle;\n    {\n");
    this->append(init);
    this->append("    }\n");
    this->append(supplied_body);
    this->append("\n");
    this->append("    return WR_SUCCESS;\n}\n");
}

void source::gen_wr_lifecycle(const char*name, char *init,source *supplied)
{
    this->append("\n/* lifecycle function */\n");
    this->append("int EXPORT ");
    this->append(name);
    this->append("( struct wss_proc *wrp, struct wss_wrapper_data *data)\n{\n");
    this->append(init);
    this->append(supplied);
    this->append("\n");
    this->append("    return WR_SUCCESS;\n}\n");
}

typedef smap<source *> SrcTbl;

void source::gen_dispatchers(source *prelist, source *postlist)
{
    this->gen_comment("Dispatch routines.");
    for (int i = 0 ; i < 2; ++i) {
        SourceAppend8(this, 
                "int EXPORT dispatch_", i ? "post" : "pre",
                                         "(int space, int code, void *uap[],\n",
                "                        int *retval,struct wss_proc *wrp,\n",
                "                        struct wss_wrapper_data *data,\n",
                "                        int args_in_userspace)\n",
                "{\n",
                "    int ret = WR_SUCCESS;\n\n");
    
        SourceAppend3(this,
                i ? postlist : prelist, 
                "    return ret;\n",
                "}\n");
    }
}

