/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: expr.h,v 1.28 2001/02/23 19:02:28 stevek Exp $
 */

#ifndef _EXPR_H
#define _EXPR_H
#include <stdio.h>
#include <string.h>
#include "wr_string.h"
#include "codegen.h"
#include "type.h"
#include "event.h"
#ifdef WR_BROKEN_STL
#include "template/list"
#else
#include <list>
#endif

struct WrEvent;
class SpecIdentExpr;
class Expr {
public:
     enum IdType { Para, Ret, InVar};
     enum eType { Ident, SpecIdent, RecElm, ArrayElm, PtrToElm, PtrRef, 
                  Const, Binary, Unary, PostFix, Trinary, Assmt, List,
                  FuncCall, Decl };
     Type type;
     int  lvalue;
     source *code;
     Expr *next;
     Expr () {   
         code = new source;   
         next = NULL;
     };
     virtual ~Expr() {
     }
     virtual eType etype() const = 0;
     virtual source *get_src() { return NULL; };
     virtual const char *str() { return ""; };
     virtual list<SpecIdentExpr *>*get_siexpr() { 
          return new list<SpecIdentExpr*>; 
     };
     virtual int operator ==(const Expr &e) = 0;
     // return the first ident in a var ref, e.g. return a for a.b.c
     virtual const char *head() { return NULL; };
     virtual source *tail() {return NULL; };
     virtual Type compute_type(WrEvent *)   {   
          return Type(TC_ERR);    
     };
     virtual source *subexpr(const Expr& e) {
          return *this == e ? new source : NULL;
     }
/*     int operator !=(const Expr &e) {
          return ! (*this == e);
     };
*/
};

class UnaryExpr: public Expr {
public:
     Expr *e1;
     String op;
     UnaryExpr(const String& ops, Expr *expr1) : Expr() {
          op = ops;
          e1 = expr1;  
          lvalue = expr1->lvalue;
     }
     eType etype() const {   return (Unary);         }
     source *get_src()  {   
          source *sp = new source;
          source *sptemp = e1->get_src();
          sp->append("(");
          sp->append(sptemp);
          sp->append(")");
          delete sptemp;
          return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
          return e1->get_siexpr(); 
     };
     int operator ==(const Expr& e) {
          if (e.etype() == Unary) {
               UnaryExpr& be = (UnaryExpr&) e;
               return op == be.op && *e1 == *be.e1;
          } else
               return 0;
     }
};

class PostFixExpr : public Expr {
public:
     Expr *e1;
     String op;
     PostFixExpr(const String& ops, Expr *expr1) : Expr() {
          op = ops;
          e1 = expr1;  
          lvalue = expr1->lvalue;
     }
     eType etype() const {   return (PostFix);         }
     source *get_src()  {   
          source *sp = new source;
          source *sptemp = e1->get_src();
          sp->append("(");
          sp->append(sptemp);
#ifdef WR_BROKEN_STL
          sp->append(op.chars());
#else
          sp->append(op.data());
#endif
          sp->append(")");
          delete sptemp;
          return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
          return e1->get_siexpr(); 
     };
     int operator ==(const Expr& e) {
          if (e.etype() == PostFix) {
               PostFixExpr& be = (PostFixExpr&) e;
               return op == be.op && *e1 == *be.e1;
          } else
               return 0;
     }
};

class AssmtExpr: public Expr {
public:
     Expr *e1, *e2;
     String op;
     AssmtExpr(const String& ops, Expr *expr1, Expr *expr2) : Expr() {
          op = ops;
          e1 = expr1;  
          e2 = expr2;
     }
     eType etype() const {   return Assmt;         }
     source *get_src()   {   return code;	   }
     int operator ==(const Expr& e) {
          if (e.etype() == Assmt) {
               AssmtExpr& ae = (AssmtExpr& ) e;
               return op == ae.op && *e1 == *ae.e1 && *e2 == *ae.e2;
          } else
               return 0;
     }

     list<SpecIdentExpr *>*get_siexpr() { 
         list<SpecIdentExpr*> *l, *l2;
         l = e1->get_siexpr();
         if (l == NULL)
             l = e2->get_siexpr();
         else {
             l2 = e2->get_siexpr();
             if (l2 != NULL)
                 l->insert(l->end(),l2->begin(), l2->end());
             delete l2;
         }
         return l;
     };
};

class BinaryExpr: public Expr {
public:
     Expr *e1, *e2;
     String op;
     BinaryExpr(const String& ops, Expr *expr1, Expr *expr2) : Expr() {
          op = ops;
          e1 = expr1;  
          e2 = expr2;
          lvalue = 0;
     }
     eType etype() const {   return (Binary);         }
     source *get_src()  {   
          source *sp = new source;
          source *sptemp = e1->get_src();
          sp->append("(");
          sp->append(sptemp);
#ifdef WR_BROKEN_STL
          sp->append(op.chars());
#else
          sp->append(op.data());
#endif
          delete sptemp;
          sptemp = e2->get_src();
          sp->append(sptemp);
          delete sptemp;
          sp->append(")");
          return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
         list<SpecIdentExpr*> *l, *l2;
         l = e1->get_siexpr();
         if (l == NULL)
             l = e2->get_siexpr();
         else {
             l2 = e2->get_siexpr();
             if (l2 != NULL)
                 l->insert(l->end(),l2->begin(), l2->end());
             delete l2;
         }
         return l;
     };
     int operator ==(const Expr& e) {
          if (e.etype() == Binary) {
               BinaryExpr& be = (BinaryExpr& ) e;
               return op == be.op && *e1 == *be.e1 && *e2 == *be.e2;
          } else
               return 0;
     }
};

class TrinaryExpr: public Expr {
public:
     Expr *e1, *e2, *e3;
     String op;
     TrinaryExpr(const String& ops, Expr *expr1, Expr *expr2, Expr *expr3) 
     : Expr() 
     {
          op = ops; 
          e1 = expr1;
          e2 = expr2;
          e3 = expr3;
          lvalue = 0;
     }
     eType etype() const {   return (Trinary);         }
     source *get_src()  {   
          return code;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
         list<SpecIdentExpr*> *l, *l2;
         l = e1->get_siexpr();
         if (l == NULL)
             l = e2->get_siexpr();
         else {
             l2 = e2->get_siexpr();
             if (l2 != NULL)
                 l->insert(l->end(),l2->begin(), l2->end());
             delete l2;
         }
         if (l == NULL)
             l = e3->get_siexpr();
         else {
             l2 = e3->get_siexpr();
             if (l2 != NULL)
                 l->insert(l->end(), l2->begin(), l2->end());
             delete l2;
         }
         return l;
     }
     int operator ==(const Expr& e) {
          if (e.etype() == Trinary) {
               TrinaryExpr& te = (TrinaryExpr& ) e;
               return (op == te.op) && (*e1 == *te.e1) && *e2 == *te.e2 && *e3 == *te.e3;
          } else
               return 0;
     }
};

class DeclExpr: public Expr {
public:
     char *vname;
     source *d;
     Expr *e;
     DeclExpr(source *dsrc, Expr *expr, char *vn) : Expr() {
          vname = vn;
          d = dsrc;
          e = expr;  
     }
     eType etype() const {   return (Decl);         }
     source *get_src()  {   
          return code;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
          return e->get_siexpr(); 
     };
     int operator ==(const Expr& e) {
          if (e.etype() == Decl) {
               return 1;
          } else
               return 0;
     }
};

class ListExpr: public Expr {
     Expr *elist;
public:
     ListExpr(Expr *l) : Expr() {
          elist = l; 
     }
     eType etype() const {   return List;          }
     source *get_src()   {   return code;           }
     list<SpecIdentExpr *>*get_siexpr() { 
          if (elist != NULL)
               return elist->get_siexpr(); 
          else
               return NULL;
     };
     int operator ==(const Expr& e) {
           if (e.etype() == List)
                return 1;
           else
                return 0;
     }
};

class FuncCallExpr: public Expr {
public:
     Expr *func, *plist; 
     FuncCallExpr(Expr *f, Expr *elist) : Expr() {
         func = f;
         plist = elist; 
     }
     eType etype() const {   return FuncCall;       }
     source *get_src()   {   return code;           }
     list<SpecIdentExpr *>*get_siexpr() { 
         list<SpecIdentExpr*> *l, *l2;
         l = func->get_siexpr();
         if (l == NULL)
             if (plist != NULL)
                 l = plist->get_siexpr();
             else
                 return NULL;
         else {
             if (plist != NULL) {
                 l2 = plist->get_siexpr();
                 if (l2 != NULL)
                     l->insert(l->end(),l2->begin(), l2->end());
                 delete l2;
             } else
                 return NULL;
         }
         return l;
     }
     int operator ==(const Expr& e) {
           if (e.etype() == FuncCall)
                return 1;
           else
                return 0;
     }
};

class RecElmExpr: public Expr {
public:
     Expr *expr;
     char *name;
     RecElmExpr(Expr *e, const char *n) : Expr() {
         expr = e;
         name = strdup(n);
	 lvalue = 1;
     }
     eType etype() const {   return (RecElm);         }
     source *get_src()  {   
         source *sp = new source;
         SourceAppend3(sp, expr->get_src(), ".", name);
         return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
          return expr->get_siexpr(); 
     };
     int operator ==(const Expr& e) {
          if (e.etype() == RecElm) {
               const RecElmExpr& e1 = (const RecElmExpr&) e;
               return *expr == *e1.expr && !strcmp(name, e1.name);
          } else
               return 0;
     }
     const char *head() { 
          return expr->head();
     }
     source *tail() { 
          source *sp = expr->tail();
          SourceAppend2(sp, ".", name);
          return sp;
     }
     source *subexpr(const Expr& e) {
          if (*this == e) {
               return new source; 
          } else {
               source *sp = expr->subexpr(e); 
               if (sp != NULL) {
                    SourceAppend2(sp, ".", name);
               }
               return sp;
          }
     }
     Type compute_type(WrEvent *);
};

class ArrayElmExpr: public Expr {
public:
     Expr *expr, *iexpr;
     ArrayElmExpr(Expr *e, Expr *e1) : Expr() {
         expr  = e;
         iexpr = e1;
         lvalue = 1;
     }
     eType etype() const {   return (ArrayElm);         }
     source *get_src()  {   
         source *sp = new source;
         SourceAppend4(sp, expr->get_src(), "[", iexpr->get_src(), "]");
         return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
         list<SpecIdentExpr*> *l, *l2;
         l = expr->get_siexpr();
         if (l == NULL)
             l = iexpr->get_siexpr();
         else {
             l2 = iexpr->get_siexpr();
             if (l2 != NULL)
                 l->insert(l->end(),l2->begin(), l2->end());
             delete l2;
         }
         return l;
     };
     int operator ==(const Expr& e) {
          if (e.etype() == ArrayElm) {
               const ArrayElmExpr& e1 = (const ArrayElmExpr&) e;
               return *expr == *e1.expr && *iexpr == *e1.iexpr;
          } else
               return 0;
     }
     const char *head() { 
          return expr->head();
     }
     source *tail() { 
          source *sp = expr->tail();
          SourceAppend3(sp, "[", iexpr->get_src(), "]");
          return sp;
     }
     source *subexpr(const Expr& e) {
          if (*this == e) {
               return new source; 
          } else {
               source *sp = expr->subexpr(e); 
               if (sp != NULL) {
                    SourceAppend3(sp, "[", iexpr->get_src(), "]");
               }
               return sp;
          }
     }
     Type compute_type(WrEvent *);
     
};

class IdentExpr: public Expr {
public:
     char *name;
     IdentExpr(const char *n) : Expr() {
         name = strdup(n);
         lvalue = 1;
     }
     eType etype() const {   return (Ident);         }
     const char *str()   {   return name;            }
     source *get_src()  {   
         source *sp = new source;
         sp->append(name);
         return sp;
     }
     int operator ==(const Expr& e) {
          /* return e.etype() == Ident && 
                 ( !strcmp(name, ((const IdentExpr&) e).name) ||
                   !strcmp(((const IdentExpr&) e).name, "*") 
                   // allow wild card comparision
                 );
          */
          if (e.etype() == Ident) {
                 return !strcmp(name, ((const IdentExpr&) e).name);
          } else
          return 0;
     }
     const char *head() {
          return name;
     };
     source *tail() { 
          source *sp = new source;
          return sp;
     }
     Type compute_type(WrEvent *);
};

class SpecIdentExpr: public Expr {
public:
     char *name;
     char *offset;
     mlabel *label;
     IdType idtype;
     SpecIdentExpr(const char *n) : Expr() {
         name = strdup(n);
         lvalue = 1;
     }
     eType etype() const {   return (SpecIdent);         }
     source *get_src()  {   
         if (idtype == -1)
             return NULL; // Don't know what I am yet...

         source *sp = new source;
         switch(idtype) {
         case InVar:
             sp->append("*data->");
             sp->append(label->para->name);
             break;
         case Ret:
           /* 
            * Try to cast retval according to its real type.
            * On Unix its always an int.
            * On NT it can be many things.
            */
             sp->append("(");
             sp->append(type.get_name());
             sp->append(") ");
             sp->append("*retval");
             break;
         case Para:
             sp->append("((");
             sp->append(type.get_name());
             sp->append(") ");
             if (type.is_ptrtype()) {
                 sp->append("uapk[ ");
             } else {
                 sp->append("uap[ ");
             }
             sp->append(offset);
             sp->append(" ])");
             break;
         default:
             delete sp;
             return code;
         }
         return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
         list<SpecIdentExpr *>*l = new list<SpecIdentExpr *>;
         l->push_back(this);
         return l; 
     };
     int operator ==(const Expr& e) {
          if (e.etype() == SpecIdent) {
                 return !strcmp(name, ((const SpecIdentExpr&) e).name);
          } else
                 return 0;
     }
};


class PtrRefExpr: public Expr {
public:
     Expr *expr;
     PtrRefExpr(Expr *e) : Expr() {
         expr = e;
         lvalue = 1;
     };
     eType etype() const {   return (PtrRef);         }
     source *get_src()  {   
         source *sp = new source;
         SourceAppend2(sp, "*", expr->get_src());
         return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
          return expr->get_siexpr(); 
     };
     int operator ==(const Expr& e) {
          if (e.etype() == PtrRef) {
               const PtrRefExpr& e1 = (const PtrRefExpr&) e;
               return *expr == *e1.expr;
          } else
               return 0;
     }
     const char *head() {
          return expr->head();
     }
     source *tail() {
          return expr->tail();
     }
     Type compute_type(WrEvent *);
};

class PtrToElmExpr: public Expr {
public:
     Expr *expr;
     char *name;
     PtrToElmExpr(Expr *e, const char *n) : Expr() {
         expr = e;
         name = strdup(n);
         lvalue = 1;
     }
     eType etype() const {   return (PtrToElm);         }
     source *get_src()  {   
         source *sp = new source;
         SourceAppend3(sp, expr->get_src(), "->", name);
         return sp;
     }
     list<SpecIdentExpr *>*get_siexpr() { 
          return expr->get_siexpr(); 
     };
     int operator ==(const Expr& e) {
          if (e.etype() == PtrToElm) {
               const PtrToElmExpr& e1 = (const PtrToElmExpr&) e;
               if (*expr == *e1.expr) {
                    return (!strcmp(name, e1.name) || !strcmp(e1.name, "*"));
               } else
                    return 0;
          } else
               return 0;
     }
     const char *head() { 
          return expr->head();
     }
     source *tail() { 
          source *sp = expr->tail();
          if (!sp->is_empty()) {
              SourceAppend2(sp, "->", name);
          } else
              sp->append(name);
          return sp;
     }
     source *subexpr(const Expr& e) {
          if (*this == e) {
               return new source; 
          } else {
               source *sp = expr->subexpr(e); 
               if (sp != NULL) {
                    SourceAppend2(sp, "->", name);
               }
               return sp;
          }
     }
     Type compute_type(WrEvent *);
};

class ConstExpr : public Expr {
public:     
     char *val;
     ConstExpr(const char *v, Type t) : Expr() {   
         val = strdup(v);    
         type = t;
         lvalue = 0;
     }
     eType etype() const      {    return Const;      }
     source *get_src() {
         source *sp = new source;
         sp->append(val);
         return sp;
     }
     int operator ==(const Expr& e) {
          return e.etype() == Const && 
                 !strcmp( val, ((const ConstExpr&) e).val);
     }
     const char *head() { 
          return val;
     }
     source *tail() { 
          source *sp = new source;
          return sp;
     }
     Type compute_type(WrEvent *) {   return type;    }
};

#endif /* _EXPR_H */
