/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: smap.h,v 1.20 2001/02/23 19:02:28 stevek Exp $
 *
 * Author: Calvin Ko
 *
 * Purpose: This file contains a template smap<T>, which provides for
 *          fast retrival of values of another type T based on a string
 *          (char *) key
 *
 * Examples: typedef smap<int> scoretbl;
 *          itbl.insert("name1", 100);    --  insert score of "name1"
 *          itbl.insert("name2", 95);     --  insert score of "name2"
 *          int y = itbl[name1];          --  retrive the score of "name1"
 *                                        --  y should be 100
 *
 */

#ifndef __SMAP_H
#define __SMAP_H

#ifndef WR_DEFS_INCLUDED
#include "wr_conf.h"
#endif

#ifdef WR_BROKEN_STL
#ifndef __SGI_STL_PAIR_H
#include "template/pair.h"
#endif
#ifndef __SGI_STL_SET_H
#include "template/set.h"
#endif
#else
#ifdef WR_UNIX
#include <set.h>
#endif
#ifdef WR_WINNT
#include <set>
#include <iostream>
#endif
#endif

#ifndef WR_BROKEN_NAMESPACES
using namespace std;
#endif


#if 0
template <class T1, class T2>
struct pair {
    T1 first;
    T2 second;
    pair() {}
    pair(const T1& a, const T2& b) : first(a), second(b) {}
};
#endif

template< class T1, class T2 >
ostream& operator<<(ostream& s,const pair<T1, T2>& p)
{
    return s << '(' << p.first << ", "<< p.second << ')';
}

/* Ok, so I'm evil.   DSK */
template< class T1, class T2, class T3 >
ostream& operator<<(ostream& s,const set<T1,T2,T3>& p)
{
    copy(p.begin(),p.end(),
            ostream_iterator<T1>(s," "));
    return s;
}

template <class T> 
struct lnode {
  typedef lnode<T> myType;
     T val;
     myType *next;
     lnode()          {   next = NULL;                }
     lnode(T v)       {   val  = v;  next = NULL;     }
};

template <class T> 
class smap {
public:
     typedef T value_type;

private:
     struct tentry {
          char          *key;
          value_type     data;       
          struct tentry *next; 
          tentry() { key = NULL; data = NULL; next = NULL; }
          tentry(char *c, value_type d) { 
               key = c;
               data = d; 
               next = NULL; 
          }
     };
     
     tentry **htable;
     int tblsize;
  
     int hash(const char *s) {
          int v = 0;
          while (*s)
               v = (v + *s++) % tblsize;
          return v;
     }

public: 
     class iterator;
     smap(int hsize = 50) {
          int i;
          tblsize = hsize;
          htable = new tentry * [tblsize];
          for (i=0; i<tblsize; i++) 
               htable[i] = NULL; 
     }
     ~smap() {
          delete[] htable;
     }

private:
     value_type find (int row, const char *key) {
          tentry *p = htable[row];
          while (p != NULL) {
               if (strcmp(p->key, key) == 0)
                     return (p->data);
               p = p->next;
          }
          return 0;
     }

     int insert(int row, tentry *e) {
          if (htable[row] == NULL) {
                htable[row] = e; 
          } else {
               tentry *p = htable[row];
               while (p->next != NULL) p = p->next; 
               p->next = e; 
          }
          return 1;
     }

public:
     /* return 1 if successful, otherwise 0 */
     int insert(const char *key, value_type data) {
          int row = hash(key);
          if (!find(row, key)) {
               tentry *e = new tentry(strdup(key), data);
               insert(row, e); 
               return 1;
          } else
               return 0;
     }
 
     const value_type operator [] (const char *key) {
          return find(hash(key), key);
     }

     // there must be a val correspond to the key in the map
     int mod_value(char *key, value_type val) {
          int row = hash(key);
          tentry *p = htable[row];
          while (p != NULL) {
               if (strcmp(p->key, key) == 0) {
                    p->data = val;
                    return 1;
               }
               p = p->next;
          };
          return 0;
     }
     void set_value(char *key, value_type val) {
          int row = hash(key);
          tentry *p = htable[row];
          while (p != NULL) {
               if (strcmp(p->key, key) == 0) {
                    p->data = val;
                    return;
               }
               p = p->next;
          };
          insert(row, new tentry(strdup(key), val));
          return;
     }

     class iterator {
     friend class smap<T>;
          tentry **htable;
          int    tblsize;
          int    crow;  
          tentry *cp;
     public:
          iterator(tentry **t=NULL, int s=0, int row=0, tentry *p=NULL) {
               htable = t;
               tblsize = s;
               crow = row;
               cp   = p;   
          } 
          value_type operator*()          {   return cp->data;    }
          char *get_key()                 {   return cp->key;     } 
          iterator& operator=(iterator x) {
               htable = x.htable;
               tblsize = x.tblsize;
               crow = x.crow;
               cp   = x.cp;
               return *this;
          }
          iterator& operator++() {
               if (cp != NULL) {
                    cp = cp->next; 
                    while (cp == NULL && crow < tblsize - 1) {
                         cp = htable[++crow];
                    }
               };
               return *this;
          } 
          iterator operator++(int) {
               iterator tmp = *this;
               if (cp != NULL) {
                    cp = cp->next; 
                    while (cp == NULL && crow < tblsize - 1) {
                         cp = htable[++crow];
                    }
               };
               return tmp;
          } 
/*          friend bool operator==(const iterator& A, const iterator& B) {
              return A.cp == B.cp;
          }
          friend bool operator!=(const iterator& A, const iterator& B) {
              return A.cp != B.cp;
          }
*/ 
          bool operator!=(const iterator& A) const {
              return A.cp != cp;
          }
          bool operator==(const iterator& A) const {
              return A.cp == cp;
          }
     };

     iterator begin() {
          int i;
          for (i=0;i<tblsize;i++) {
               if (htable[i] != NULL)
                    return iterator(htable, tblsize, i, htable[i]);
          };
          return iterator(htable, tblsize);
     }

     iterator end() {
          return iterator(htable, tblsize);
     }
    
     int empty() {
          int i;
          for (i=0;i<tblsize;i++) {
               if (htable[i] != NULL)
                    return 0; 
          }
          return 1;
     }

    public:
     friend bool operator==(smap<value_type>& A, smap<value_type>& B) {
         return A.htable == B.htable;
     }
};

#endif /** __SMAP_H */
