/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wrmod.c,v 1.31 2001/02/23 19:02:29 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Library for manipulating wrappers modules.  Currently,
 * modules are either activation criteria or a wrapper.
 */

#include <wr_conf.h>

#ifdef LINKAGE
#undef LINKAGE
#endif
#define LINKAGE IMPORT

#ifdef WR_WINNT
#include <direct.h>
#else
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#endif /* WR_WINNT */

#include <errno.h>
#include <wrappers.h>

#ifdef WR_WINNT
#include "../libgwmcomm/libgwmcomm.h"
#endif /* WR_WINNT */

#undef LINKAGE
#define LINKAGE EXPORT
#include "libwrmod.h"

#ifdef WR_WINNT

/* Windows NT */
#define syscall(n,c,s,b) wrmodntProcessCommand(c,b,s)
#else /* WR_WINNT */
#define _tgetcwd(a,b) getcwd(a,b)
#if defined(WR_LINUX)
extern long int syscall __P ((long int __sysno, ...));
#else /* WR_LINUX */
extern int syscall( int callnum, ... );
#endif /* WR_LINUX */
#endif /* WR_WINNT */

void EXPORT
wrmodSetup()
{
#ifdef WR_WINNT
   gwmcommSetup();
#endif /* WR_WINNT */
}

/*
 * This function loads a wrapper or activation criteria module into
 * the WSS.  It returns -1 on failure and the module id on success.
 */
int EXPORT
wrmodLoad( const LPTSTR modobj )
{
   int          error;
   const TCHAR *p;
   void        *block = NULL;
   int          size;
   BOOL         bAlloced = FALSE;
   LPTSTR       cwd      = NULL;
#ifdef WR_WINNT
   HANDLE       hMod;
#else
   int          fd;
#endif /* WR_WINNT */
   
   p = _tcsrchr( modobj, PD_MAKESTR('.') );
   if( !p || ( _tcscmp( p, PD_MAKESTR(_WRAPEXT) ) &&
               _tcscmp( p, PD_MAKESTR(_CRITEXT) ) ) )
   {
      errno = EINVAL;
      return -1;
   }

   if( ( _tcslen( modobj ) <= 2 ) ||
#ifdef WR_WINNT
       ( ( modobj[1] != PD_MAKESTR(':') ) &&
         ( modobj[0] != PD_MAKESTR(FILESEP) ) ) 
#else
       ( modobj[0] != PD_MAKESTR(FILESEP) )
#endif
     )
   {
      int   len;
	  TCHAR sepString[2] = { PD_MAKESTR(FILESEP), 0 };

      cwd = _tgetcwd( NULL, MAXPATHLEN );

      /* Relative path */
      len = _tcslen(cwd) + _tcslen(modobj) + sizeof(TCHAR) * 2;

      /* Allocate memory for path */
      block = malloc( len );
      if( block == NULL )
      {
         errno = ENOMEM;
         return -1;
      }

      /* Construct full path to module */
      len = size = _tcslen(cwd);
      memcpy( block, (void *)cwd, len );
      memcpy( (LPTSTR)block + size, (void *)sepString, sizeof(TCHAR) );
      size++;
      len = _tcslen(modobj);
      memcpy( (LPTSTR)block + size, (void *)modobj, len );
      size += len;
      memset( (LPTSTR)block + size, 0, sizeof(TCHAR) );
      size++;

      size *= sizeof(TCHAR);

      /* Set allocated flag so memory is freed later */
      bAlloced = TRUE;
   }
#ifdef WR_WINNT
   else if( ( modobj[0] == PD_MAKESTR(FILESEP) ) &&
	        ( modobj[1] != PD_MAKESTR(FILESEP) ) )
   {
     int   len;
     
     cwd = _tgetcwd( NULL, MAXPATHLEN );
     
     /* Relative path */
     len = _tcslen(modobj) + sizeof(TCHAR) * 3;
     
      /* Allocate memory for path */
     block = malloc( len );
     if( block == NULL )
     {
       errno = ENOMEM;
       return -1;
     }
     size = 2 * sizeof(TCHAR);

     memset( block, 0, len );

     /* Construct full path to module */
     memcpy( block, (void *)cwd, size );
     memcpy( (LPTSTR)block + 2, (void *)modobj, len - size);

     /* Set allocated flag so memory is freed later */
     bAlloced = TRUE;

     size = len;
   }
#endif
   else
   {
      /* Absolute path */
      block = (void *)modobj;
      size = _tcslen( modobj ) + sizeof(TCHAR);
   }

#ifdef WR_WINNT
   if( ( hMod = CreateFile( (LPTSTR)block, 0, 0, NULL, OPEN_EXISTING,
                            FILE_ATTRIBUTE_NORMAL, NULL ) ) == 0 )
      return -1;

   CloseHandle( hMod );
#else
   if( ( fd = open( (LPTSTR)block, O_RDONLY, 0 ) ) == -1)
      return -1;

   close( fd );
#endif

   error = WR_wrctl( WRCTL_LOAD, size, block );

   if( bAlloced )
   {
      if( block )
         free( block );
      if( cwd )
         free( cwd );
   }

#ifndef WR_SOLARIS
   if( error < 0 )
   {
      errno = -error;
      error = -1;
   }
#endif /* !WR_SOLARIS */

   return( error );
}

/*
 * Interface to the WSS syscall to unload a module.
 */
int EXPORT
wrmodUnload( int id )
{
   int error;
#ifdef WR_WINNT
   error = WR_wrctl( WRCTL_UNLOAD, sizeof(id), &id );
#else
   error = WR_wrctl( WRCTL_UNLOAD, id, NULL );
#endif

#ifndef WR_SOLARIS
   if( error < 0 )
   {
      errno = -error;
      error = -1;
   }
#endif /* !WR_SOLARIS */

   return( error );
}

/*
 * Interface to the WSS syscall to enable/disable a module.
 */
int EXPORT
wrmodEnable( int id, int yesno )
{
#ifdef WR_WINNT
   int data[2] = { id, (yesno==0)?0:1 };

   int error = WR_wrctl( WRCTL_ENABLED, sizeof(data), data );
#else
   int error = WR_wrctl( WRCTL_ENABLED, id, (void *)((yesno==0)?0:1) );
#endif

#ifndef WR_SOLARIS
   if( error < 0 )
   {
      errno = -error;
      error = -1;
   }
#endif /* !WR_SOLARIS */

   return( error );
}

