/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wrv.c,v 1.53 2001/06/22 04:15:10 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Wrapper Viewer interface library
 *
 */

#define LINKAGE IMPORT

#include <wr_conf.h>

#include <wsi_types.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h> /* malloc, realloc */
#ifndef WR_WINNT
#include <unistd.h> /* syscall */
#include <string.h> /* bcopy, bzero */
#include <strings.h> /* bcopy, bzero */
#include <signal.h>
#endif

#include <wrappers.h>
#include <wss_db.h>
#include <wrerrno.h>

#ifdef WR_WINNT
#include <direct.h>
#include <tchar.h>

#include <wsi_utilities.h>
#include "../libgwmcomm/libgwmcomm.h"
#endif

#include "../libwrmod/libwrmod.h"

#undef LINKAGE
#define LINKAGE EXPORT
#include "libwrv.h"

#define FALSE     0
#define TRUE      1
#define DEFAULT_BUFSIZE 1024

#ifdef WR_FREEBSD
#  define LIB_ERROR  -error
#endif
#ifdef WR_SOLARIS
#  define LIB_ERROR  errno
#endif
#ifdef WR_LINUX
#  define LIB_ERROR  errno
#endif
#ifdef WR_WINNT
#  define LIB_ERROR  GetLastError()
#  define syscall(n,c,b,s) wrvntProcessCommand(c,b,s)
#endif


/*
 * Public functions
 */
int EXPORT
wrvInit()
{
   return 0;
}

int EXPORT
wrvShutdown()
{
    return 0;
}

static volatile int wssLoaded = TRUE;

#ifndef PD_NOSIGSYS
static void
_wrvLoadCheckSigHandler( int sig )
{
   wssLoaded = FALSE;
}

int EXPORT
wrvIsLoaded()
{
   struct sigaction sa;
   struct sigaction osa;

#ifdef DEBUG_WRV
   printf("Setting up signal handler\n");
#endif

   /* Set up signal handler for SIGSYS */
   sa.sa_handler = _wrvLoadCheckSigHandler;
   sigemptyset( &sa.sa_mask );
   sa.sa_flags = 0;
#ifdef DEBUG_WRV
   printf("Calling sigaction\n");
#endif
   sigaction( SIGSYS, &sa, &osa );
#ifdef DEBUG_WRV
   printf("Called sigaction\n");
#endif

   /* Get WRV interface version */
   (void) wrvGetVersion();

   /* Restore signal handler to default */
   sigaction( SIGSYS, &osa, NULL );

   return wssLoaded;
}
#else
int EXPORT
wrvIsLoaded()
{
   float vers;

#ifdef WR_WINNT
   gwmcommSetup();
#endif

   vers = wrvGetVersion();
   if( vers == 0.0 )
      wssLoaded = FALSE;

   return wssLoaded;
}
#endif /* !PD_NOSIGSYS */


float EXPORT
wrvGetVersion()
{
   int            error;
   static size_t  bufsize = sizeof(int) * 2;
   int           *version = NULL;
   float          retval;

   error = wrvCallSystemRecvData( WRV_GET_VERSION, (void **)&version,
                                  &bufsize );

   if( error < 0 )
      retval = 0.0f;
   else
      retval = (float)version[0] + (float)(version[1] / 100.0);
   
#ifdef WR_WINNT
   gwmcommFreeData(version);
#else
   free(version);
#endif
   return retval;
}

void EXPORT
wrvAdminDone()
{
   wrvCallSystemSendData( WRV_ADMIN_DONE, 0, 0 );
}

unsigned long EXPORT
wrvGetModuleIdByName( const char *name )
{
   int            error;
   static size_t  bufsize = sizeof(unsigned long);
   unsigned long *id = NULL;

   if( name == NULL ) return 0;

   bufsize = ( _tcslen(name) + 1 ) * sizeof(TCHAR);

   error = wrvCallSystemSendRecvData( WRV_GET_MODULE_ID_BY_NAME, (void *)name,
                                      bufsize, (void **)&id, NULL );

   if( error < 0 )
      return 0;
   else
      return *id;
}

wrv_buffer EXPORT *
wrvGetWrappedProcesses( int handle )
{
   wrv_h_t        data;
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   data.handle = handle;

   error = wrvCallSystemSendRecvData( WRV_GET_PROCS_BY_WRAPPER_ID, &data,
                                      sizeof(data), (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      if( buf )
#ifdef WR_WINNT
         gwmcommFreeData( (void *)buf );
#else
         free( (void *)buf );
#endif
      return NULL;
   }

   return buf;
}

wrv_buffer EXPORT *
wrvGetProcessByPID( pid_t pid )
{
   wrv_h_t        data;
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   data.handle = (u_long)pid;

   error = wrvCallSystemSendRecvData( WRV_GET_PROC_BY_PID, &data, sizeof(data),
                                      (void **)&buf, &bufsize );
   if( error < 0 )
   {
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }
  
   return buf;
}

wrv_buffer EXPORT *
wrvGetActiveWrappers()
{
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   error = wrvCallSystemRecvData( WRV_GET_ACTIVE_WRAPPERS, (void **)&buf,
                                  &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

wrv_buffer EXPORT *
wrvGetActivationCriteriaByHandle( int handle )
{
   wrv_h_t        data;
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   data.handle = handle;

   error = wrvCallSystemSendRecvData( WRV_GET_ACT_CRIT_BY_ID, &data,
                                      sizeof(data), (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

wrv_buffer EXPORT *
wrvGetDeactivationCriteriaByHandle( int handle )
{
   wrv_h_t        data;
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   data.handle = handle;

   error = wrvCallSystemSendRecvData( WRV_GET_DEACT_CRIT_BY_ID, &data,
                                      sizeof(data), (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

wrv_buffer EXPORT *
wrvGetLibraryByHandle( int handle )
{
   wrv_h_t        data;
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   data.handle = handle;

   error = wrvCallSystemSendRecvData( WRV_GET_LIBRARY_BY_ID, &data,
                                      sizeof(data), (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

wrv_buffer EXPORT *
wrvGetModuleByHandle( int handle )
{
   wrv_h_t        data;
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   data.handle = handle;

   error = wrvCallSystemSendRecvData( WRV_GET_MODULE_BY_ID, &data,
                                      sizeof(data), (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

wrv_buffer EXPORT *
wrvGetWrapperByHandle( int handle )
{
   wrv_h_t        data;
   wrv_buffer    *buf     = NULL;
   static size_t  bufsize = 0;
   int            error;

   data.handle = handle;

   error = wrvCallSystemSendRecvData( WRV_GET_WRAPPER_BY_ID, &data,
                                      sizeof(data), (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
#ifdef WR_WINNT
      gwmcommFreeData( (void *)buf );
#else
      free( (void *)buf );
#endif
      return NULL;
   }
   if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

int EXPORT
wrvLoadModule( const char *module )
{
   int handle = wrmodLoad( (LPTSTR)module );

   if( handle == 0 ) handle = -1;

   return handle;
}

void EXPORT
wrvReleaseActivationCriteria( void *actcrit )
{
#ifdef WR_WINNT
   gwmcommFreeData( actcrit );
#else
   free( actcrit );
#endif
}

void EXPORT
wrvReleaseDeactivationCriteria( void *deactcrit )
{
#ifdef WR_WINNT
   gwmcommFreeData( deactcrit );
#else
   free( deactcrit );
#endif
}

void EXPORT
wrvReleaseLibraries( void *libraries )
{
#ifdef WR_WINNT
   gwmcommFreeData( libraries );
#else
   free( libraries );
#endif
}

void EXPORT
wrvReleaseModules( void *modules )
{
#ifdef WR_WINNT
   gwmcommFreeData( modules );
#else
   free( modules );
#endif
}

void EXPORT
wrvReleaseProcesses( void *processes )
{
#ifdef WR_WINNT
   gwmcommFreeData( processes );
#else
   free( processes );
#endif
}

void EXPORT
wrvReleaseWrappers( void *wrappers )
{
#ifdef WR_WINNT
   gwmcommFreeData( wrappers );
#else
   free( wrappers );
#endif
}

void EXPORT
wrvReleaseActiveWrappers( void *wrappers )
{
#ifdef WR_WINNT
   gwmcommFreeData( wrappers );
#else
   free( wrappers );
#endif
}

void EXPORT
wrvReleaseWrappedProcesses( void *processes )
{
#ifdef WR_WINNT
   gwmcommFreeData( processes );
#else
   free( processes );
#endif
}

int EXPORT
wrvUnloadModule( int handle )
{
   return( wrmodUnload( handle ) );
}

