/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wrvdb.c,v 1.45 2001/02/23 19:02:29 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Wrapper Viewer interface library
 *
 */

#define LINKAGE IMPORT
#include <wr_conf.h>

#include <wsi_types.h>

#include <stdlib.h>
#include <errno.h>
#ifndef WR_WINNT
#include <unistd.h>
#include <strings.h>
#endif /* !WR_WINNT */
#include <string.h>

#include <wrappers.h>
#include <wss_db.h>

#undef LINKAGE
#define LINKAGE EXPORT
#include "libwrv.h"

#ifdef WR_WINNT
#include <wsi_utilities.h>
#include "../libgwmcomm/libgwmcomm.h"
#endif

#define DEFAULT_BUFSIZE 1024

#ifdef WR_FREEBSD
#  define LIB_ERROR  -error
#endif
#ifdef WR_SOLARIS
#  define LIB_ERROR  errno
#endif
#ifdef WR_LINUX
#  define LIB_ERROR  errno
#endif
#ifdef WR_WINNT
#  define LIB_ERROR  GetLastError()
#  define syscall(n,c,b,s) wrvntProcessCommand(c,b,s)
#endif


/*
 * Public functions
 */
wrdb_coldef_t
wrvDbGetColumnDefinition( unsigned long id, int *size )
{
   wrv_h_t       data;
   wrdb_coldef_t buf     = NULL;
   static size_t bufsize = 0;
   int           error;

   data.handle = id;

   error = wrvCallSystemSendRecvData( WRV_GET_TABLE_COLUMN_DEF, &data,
                                      sizeof(data), (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
      free( (void *)buf );
      return NULL;
   }
   else if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   if( ( error >= 0 ) && size )
      *size = error;

   return buf;
}

wrdb_row_t
wrvDbGetRowsByHandle( unsigned long thandle, unsigned long *handles, int size )
{
   wrv_h_t      *sendbuf;
   size_t        sendsize;
   wrdb_row_t    buf      = NULL;
   static size_t bufsize  = 0;
   int           error;

   sendsize = sizeof(wrv_h_t) + ( size + 1 ) * sizeof(u_long);
   sendbuf = (wrv_h_t *)malloc( sendsize );
   bzero( sendbuf, sendsize );

   sendbuf->handle = thandle;
   if( handles )
      bcopy( handles, sendbuf + 1, sendsize - sizeof(wrv_h_t) );
   error = wrvCallSystemSendRecvData( WRV_GET_TABLE_ROWS, sendbuf, sendsize,
                                      (void **)&buf, &bufsize );

   free( (void *)sendbuf );
   if( error < 0 )
   {
      errno = LIB_ERROR;
      free( (void *)buf );
      return NULL;
   }

   if( bufsize <= 0 )
   {
      if( buf )
         free( (void *)buf );
      return NULL;
   }

   return buf;
}

wrv_table_id_t * 
wrvDbGetTableIdsByHandle( unsigned long *handles, unsigned int size )
{
   u_long         *sendbuf  = NULL;
   size_t          sendsize = 0;
   wrv_table_id_t *buf      = NULL;
   static size_t   bufsize  = 0;
   int             error;

   sendsize = ( size + 1 ) * sizeof(u_long);
   sendbuf = (u_long *)malloc( sendsize );
   bzero( sendbuf, sendsize );

   if( handles )
      bcopy( handles, sendbuf, size * sizeof(u_long) );

   error = wrvCallSystemSendRecvData( WRV_GET_TABLE_HANDLES, sendbuf,
                                      sendsize, (void **)&buf, &bufsize );

   free( (void *)sendbuf );
   if( error < 0 )
   {
      errno = LIB_ERROR;
      free( (void *)buf );
      return NULL;
   }
   else if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

wrv_table_id_t *
wrvDbGetTableIdByName( const char *name )
{
   wrv_table_id_t *buf     = NULL;
   static size_t   bufsize;
   int             error;

   if( name == NULL )
   {
      errno = EINVAL;
      return NULL;
   }

   /* GetTableHandleByName returns the name as well as the id.  Go fig */
   bufsize = ( _tcslen(name) + 1 ) * sizeof(TCHAR) + sizeof(wrv_table_id_t);
   error = wrvCallSystemSendRecvData( WRV_GET_TABLE_HANDLE_BY_NAME,
                                      (void *)name,
                                      ( _tcslen(name) + 1 ) * sizeof(TCHAR),
                                      (void **)&buf, &bufsize );
   if( error < 0 )
   {
      errno = LIB_ERROR;
      free( (void *)buf );
      return NULL;
   }
   else if( bufsize <= 0 )
   {
      errno = EPERM;
      return NULL;
   }

   return buf;
}

int
wrvDbAddNullTableRow( unsigned long handle )
{
   wrv_h_t data;
   int     error;

   data.handle = handle;
   error = wrvCallSystemSendData( WRV_ADD_NULL_TABLE_ROW, &data, sizeof(data) );
   if( error < 0 )
   {
      errno = LIB_ERROR;
      return -1;
   }

   return 0;
}

/*
 * Add a row, with data, to the specified table.
 *
 * The fields parameter is a pointer to a NULL-terminated array
 * of pointers to strings.  Strings are in the same order as the
 * fields in the table are declared in the .wr file.
 */
int wrvDbAddTableRow( unsigned long handle, LPTSTR rows[]) 
{
    wrv_h_t    *data;
    int         error;
    size_t      size = 0;
    int         i;
    LPTSTR      offset;
    
    for (i = 0; rows[i] ; ++i)
        size += ( _tcslen(rows[i]) + 1 ) * sizeof(pd_char_t);

    size += sizeof(*data);
    
    data = malloc(size);

    data->handle = handle;

    for (offset = ((pd_byte_t *)data) + sizeof(*data), i = 0; rows[i]; ++i) {
        _tcscpy(offset, rows[i]);
        offset += _tcslen(rows[i]) + 1;
    }

    error = wrvCallSystemSendData( WRV_ADD_TABLE_ROW, data, size );
    if ( error < 0 ) {
        errno = LIB_ERROR;
        return -1;
    }
    return 0;
}


int
wrvDbDelTableRow( unsigned long thandle, unsigned long rhandle )
{
   wrv_tr_t data;
   int      error;

   data.thandle = thandle;
   data.rhandle = rhandle;
   error = wrvCallSystemSendData( WRV_DEL_TABLE_ROW, &data, sizeof(data) );
   if( error < 0 )
   {
      errno = LIB_ERROR;
      return -1;
   }

   return 0;
}

int
wrvDbSetTableRowField( unsigned long thandle, unsigned long rhandle,
                       const char *field, const char *value )
{
   wrv_trnd_t *data;
   size_t bufsize;
   int error;

   if( ( field == NULL ) || ( value == NULL ) )
   {
      errno = EINVAL;
      return -1;
   }

   bufsize = sizeof( *data ) + strlen( value ) + 1;
   data = (wrv_trnd_t *)malloc( bufsize );
   bzero( data, bufsize );
   data->thandle = thandle;
   data->rhandle = rhandle;
   strncpy( data->name, field, MAXNAMELEN );
   strcpy( (char *)(data + 1), value );

   error = wrvCallSystemSendData( WRV_SET_TABLE_ROW_FIELD, data, bufsize );

   free( data );
   if( error < 0 )
   {
      errno = LIB_ERROR;
      return -1;
   }

   return 0;
}


void
wrvDbReleaseColumnDefinition( wrdb_coldef_t cd )
{
#ifdef WR_WINNT
   gwmcommFreeData( (void *)cd );
#else
   free( (void *)cd );
#endif /* WR_WINNT */
}

void
wrvDbReleaseRows( wrdb_row_t row )
{
#ifdef WR_WINNT
   gwmcommFreeData( (void *)row );
#else
   free( (void *)row );
#endif /* WR_WINNT */
}

void
wrvDbReleaseTableIds( wrv_table_id_t *table )
{
#ifdef WR_WINNT
   gwmcommFreeData( (void *)table );
#else
   free( (void *)table );
#endif /* WR_WINNT */
}

