/*
 *  Copyright (c) 1999, 2000 Network Associates, Incorporated
 *  All rights reserved.
 *	 
 *  Redistribution and use are governed by the terms detailed in the
 *  license document ("LICENSE") included with the toolkit.
 *
 * $Id: Message.cc,v 1.9 2000/10/27 20:41:23 dougk Exp $
 * 
 * File: Message.cxx
 * 
 * Author: Douglas Kilpatrick
 * 
 * Purpose: Message classes to encapsulate talking to the wrappers
 * event stream.
 * 
 * Classes:
 * 
 *      Source: data source
 *      EventSource: Source subclass that talks to the event manager.
 *      
 *      Message: root message class.
 *      execMessage: subclass for process creation
 *      exitMessage: subclass for process termination
 *      syscallMessage: subclass for syscall execution
 *      quitMessage: subclass for wrapper shutdown
 *
 */

#include "wr_conf.h"

#ifndef MESSAGE_H
#include "Message.h"
#endif

#include "unistd.h"

/* Used for strdup() only so far. */
#include <string.h>

#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>

/* Types */
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>

#include "libeventfs.h"
#include "libwrv.h"
#include "../codes.h"

/* STL */
#include <string>
#include <vector>
#include <algo.h>

/*
 * Messages First.
 */

/* Duplicate defines between here and emerald.wr.  Sorry about that. */

#include "../packets.h"

/*
 * Message sub-classes.  Can create themselves from a header object
 * (actually from a whole packet.  C-style inheritance).  Can generate
 * Emerald events when toEmeraldEvent is called.
 */

#define EVENTCLASS(classname, headertype)                               \
class classname: public Message {                                       \
    private:                                                            \
        headertype      event_data;                                     \
    public:                                                             \
        virtual bool toEmeraldEvent(void);                              \
        ~classname() { };                                               \
        classname(header *in_data, size_t) {                            \
            event_data = *(headertype*)in_data;                         \
        };                                                              \
}                                                                       \

#define PATHEVENTCLASS(classname)                                       \
class classname: public Message {                                       \
    private:                                                            \
        path_data       event_data;                                     \
        string          pathname;                                       \
    public:                                                             \
        virtual bool toEmeraldEvent(void);                              \
        ~classname() {  };                                              \
        classname(header *in_data, size_t) {                            \
            event_data = *(path_data*)in_data;                          \
            pathname = ((path_data*)in_data)->pathname;                 \
        };                                                              \
}                                                                       \

#define DESTPATHEVENTCLASS(classname)                                   \
class classname: public Message {                                       \
    private:                                                            \
        path_data       event_data;                                     \
        string          pathname;                                       \
        string          destpathname;                                   \
    public:                                                             \
        virtual bool toEmeraldEvent(void);                              \
        ~classname() {  };                                              \
        classname(header *in_data, size_t) {                            \
            event_data = *(path_data*)in_data;                          \
            pathname = ((path_data*)in_data)->pathname;                 \
            destpathname = ((path_data*)in_data)->pathname              \
                        +pathname.length()+1;                           \
        };                                                              \
}                                                                       \


EVENTCLASS(             MProtectEvent,                  header);
EVENTCLASS(             ExitEvent,                      header);
EVENTCLASS(             ForkEvent,                      header);
EVENTCLASS(             AcceptEvent,                    audit_sock_data);
EVENTCLASS(             KillEvent,                      op_data);
EVENTCLASS(             SetreuidEvent,                  op_data);
PATHEVENTCLASS(         UTimeEvent                      );
PATHEVENTCLASS(         MKNodEvent                      );
PATHEVENTCLASS(         MKDirEvent                      );
PATHEVENTCLASS(         RMDirEvent                      );
PATHEVENTCLASS(         ChmodEvent                      );
PATHEVENTCLASS(         ChownEvent                      );
PATHEVENTCLASS(         CreatEvent                      );
PATHEVENTCLASS(         OpenREvent                      );
PATHEVENTCLASS(         OpenWEvent                      );
PATHEVENTCLASS(         OpenRWEvent                     );
PATHEVENTCLASS(         TruncateEvent                   );
PATHEVENTCLASS(         FLockEvent                      );
PATHEVENTCLASS(         UnlinkEvent                     );
PATHEVENTCLASS(         CoreEvent                       );
DESTPATHEVENTCLASS(     RenameEvent                     );
DESTPATHEVENTCLASS(     LinkEvent                       );
DESTPATHEVENTCLASS(     SymlinkEvent                    );

/* Unfortunately exec is a special case. */
class ExecEvent: public Message {
    private:
        header          event_data;
        string          path;
        vector<string>  argv;
        vector<string>  envp;
    public:
        virtual bool toEmeraldEvent(void);
        ~ExecEvent() { };
        ExecEvent(header *in_data, size_t) {
            event_data = *in_data;

            const char *str;
            const char *find;

            str         = ((char*)in_data)+sizeof(event_data);

            /* Path */
            path        = str;

            str        += path.length() + 1;
            find        = str;
            /* argc */
            while (*find != '\0') {
                while (*find != '\0')
                    ++find;
                argv.push_back(str);
                ++find;
                str = find;
            }
            ++find;
            str = find;
            /* envp */
            while (*find != '\0') {
                while (*find != '\0')
                    ++find;
                envp.push_back(str);
                ++find;
                str = find;
            }
        };
};


/* EMERALD: Define these */
bool MProtectEvent::toEmeraldEvent(void) { return false; }
bool ExitEvent::toEmeraldEvent(void) { return false; }
bool KillEvent::toEmeraldEvent(void) { return false; }
bool ForkEvent::toEmeraldEvent(void) { return false; }
static inline void print_string(string s) {
    cout << "    " << s << endl;
}
bool ExecEvent::toEmeraldEvent(void) {
#if 0
    cout << "Exec event" << endl;
    cout << "header:" << endl;
    cout << "  event  = " << event_data.event  << endl;
    cout << "  ruid   = " << event_data.ruid   << endl;
    cout << "  euid   = " << event_data.euid   << endl;
    cout << "  auid   = " << event_data.auid   << endl;
    cout << "  rgid   = " << event_data.rgid   << endl;
    cout << "  egid   = " << event_data.egid   << endl;
    cout << "  pid    = " << event_data.pid    << endl;
    cout << "  ppid   = " << event_data.ppid   << endl;
    cout << "  sid    = " << event_data.sid    << endl;
    cout << "  retval = " << event_data.retval << endl;
    cout << "  argv:" << endl;
    for_each(argv.begin(), argv.end(), print_string);
    cout << "  envp:" << endl;
    for_each(envp.begin(), envp.end(), print_string);
#endif
    return false;
}
bool CoreEvent::toEmeraldEvent(void) {
    cout << "Coredump event" << endl;
    cout << "header:" << endl;
    cout << "  event  = " << event_data.h.event  << endl;
    cout << "  ruid   = " << event_data.h.ruid   << endl;
    cout << "  euid   = " << event_data.h.euid   << endl;
    cout << "  auid   = " << event_data.h.auid   << endl;
    cout << "  rgid   = " << event_data.h.rgid   << endl;
    cout << "  egid   = " << event_data.h.egid   << endl;
    cout << "  pid    = " << event_data.h.pid    << endl;
    cout << "  ppid   = " << event_data.h.ppid   << endl;
    cout << "  sid    = " << event_data.h.sid    << endl;
    cout << "  retval = " << event_data.h.retval << endl;
    cout << "  owner  = " << event_data.file_owner      << endl;
    cout << "  group  = " << event_data.file_group      << endl;
    cout << "  mode   = " << event_data.file_mode       << endl;
    cout << "  device = " << event_data.file_device     << endl;
    cout << "  inode  = " << event_data.file_ino        << endl;
    cout << "  n_attr = " << event_data.new_attrib      << endl;
    cout << "  core dump path:" << pathname << endl;
    return false;
}
bool SetreuidEvent::toEmeraldEvent(void) {
    cout << "Setreuid event" << endl;
    cout << "header:" << endl;
    cout << "  event      = " << event_data.h.event     << endl;
    cout << "  ruid       = " << event_data.h.ruid      << endl;
    cout << "  euid       = " << event_data.h.euid      << endl;
    cout << "  auid       = " << event_data.h.auid      << endl;
    cout << "  rgid       = " << event_data.h.rgid      << endl;
    cout << "  egid       = " << event_data.h.egid      << endl;
    cout << "  pid        = " << event_data.h.pid       << endl;
    cout << "  ppid       = " << event_data.h.ppid      << endl;
    cout << "  sid        = " << event_data.h.sid       << endl;
    cout << "  retval     = " << event_data.h.retval    << endl;
    cout << "  target_pid = " << event_data.target_pid  << endl;
    cout << "  target_ruid= " << event_data.target_ruid << endl;
    cout << "  target_euid= " << event_data.target_euid << endl;
    cout << "  target_sig = " << event_data.target_signal<< endl;
    return false;
}
bool MKNodEvent::toEmeraldEvent(void) {
#if 0
    cout << "MKNod event" << endl;
    cout << "header:" << endl;
    cout << "  event  = " << event_data.h.event  << endl;
    cout << "  ruid   = " << event_data.h.ruid   << endl;
    cout << "  euid   = " << event_data.h.euid   << endl;
    cout << "  auid   = " << event_data.h.auid   << endl;
    cout << "  rgid   = " << event_data.h.rgid   << endl;
    cout << "  egid   = " << event_data.h.egid   << endl;
    cout << "  pid    = " << event_data.h.pid    << endl;
    cout << "  ppid   = " << event_data.h.ppid   << endl;
    cout << "  sid    = " << event_data.h.sid    << endl;
    cout << "  retval = " << event_data.h.retval << endl;
    cout << "  owner  = " << event_data.file_owner      << endl;
    cout << "  group  = " << event_data.file_group      << endl;
    cout << "  mode   = " << event_data.file_mode       << endl;
    cout << "  device = " << event_data.file_device     << endl;
    cout << "  inode  = " << event_data.file_ino        << endl;
    cout << "  n_attr = " << event_data.new_attrib      << endl;
    cout << "  Node path:" << pathname << endl;
#endif
    return false;
}
bool UnlinkEvent::toEmeraldEvent(void) {
#if 0
    cout << "Unlink event" << endl;
    cout << "header:" << endl;
    cout << "  event  = " << event_data.h.event  << endl;
    cout << "  ruid   = " << event_data.h.ruid   << endl;
    cout << "  euid   = " << event_data.h.euid   << endl;
    cout << "  auid   = " << event_data.h.auid   << endl;
    cout << "  rgid   = " << event_data.h.rgid   << endl;
    cout << "  egid   = " << event_data.h.egid   << endl;
    cout << "  pid    = " << event_data.h.pid    << endl;
    cout << "  ppid   = " << event_data.h.ppid   << endl;
    cout << "  sid    = " << event_data.h.sid    << endl;
    cout << "  retval = " << event_data.h.retval << endl;
    cout << "  owner  = " << event_data.file_owner      << endl;
    cout << "  group  = " << event_data.file_group      << endl;
    cout << "  mode   = " << event_data.file_mode       << endl;
    cout << "  device = " << event_data.file_device     << endl;
    cout << "  inode  = " << event_data.file_ino        << endl;
    cout << "  n_attr = " << event_data.new_attrib      << endl;
    cout << "  Unlinked path:" << pathname << endl;
#endif
    return false;
}


bool AcceptEvent::toEmeraldEvent(void) { return false; }
bool UTimeEvent::toEmeraldEvent(void) { return false; }
bool MKDirEvent::toEmeraldEvent(void) { return false; }
bool RMDirEvent::toEmeraldEvent(void) { return false; }
bool ChmodEvent::toEmeraldEvent(void) { return false; }
bool ChownEvent::toEmeraldEvent(void) { return false; }
bool CreatEvent::toEmeraldEvent(void) { return false; }
bool OpenREvent::toEmeraldEvent(void) { return false; }
bool OpenWEvent::toEmeraldEvent(void) { return false; }
bool OpenRWEvent::toEmeraldEvent(void) { return false; }
bool TruncateEvent::toEmeraldEvent(void) { return false; }
bool FLockEvent::toEmeraldEvent(void) { return false; }
bool RenameEvent::toEmeraldEvent(void) { return false; }
bool LinkEvent::toEmeraldEvent(void) { return false; }
bool SymlinkEvent::toEmeraldEvent(void) { return false; }


/*
 * MessageSource:
 * Class that acts as an object factory for message objects.  Cut and
 * paste code, should need little modification.
 *
 * Sorry this isn't done better, C++ sucks sometimes and I'm not willing
 * to hack in the workarounds on this timeframe.
 */
static struct timeval      tv = {0,0};
MessageSource::MessageSource(int &argc, char **& argv)
{
    char       *input_file      = strdup("");
    char        ch;

    this->fFile = false;
    gettimeofday(&tv,0);

    while ((ch = getopt(argc, argv, "f:")) != -1) {
        switch(ch) {
        case 'f':
            free(input_file);
            input_file = strdup(optarg);
            this->fFile = true;
            break;
        default:
            break;
        }
    }
    if (input_file[0] != '\0') {
        this->stream = eventfsEventGetStreamByFile(input_file);
    } else {
        eventfs_event_group_t group = eventfsEventGetGroup(WREV_AUDIT_GROUP);
        if (group == 0) {
            perror("eventfsEventGetGroup("WREV_AUDIT_GROUP")");
            exit(1);
        }
        this->stream = eventfsEventGroupGetStream(group,1<<BSM_AUDIT_NUMBER);
    }
    if (this->stream == 0) {
        perror("eventfsEventGroup::GetStream()");
        exit(1);
    }
}

MessageSource::~MessageSource()
{
    if (!fFile) {
        eventfs_event_group_t group = eventfsEventStreamGetGroup(stream);

        eventfsEventStreamRelease(stream);
        eventfsEventGroupRelease(group);
    }
    /* else {   // Its all fake, so don't risk nuking things.
     */
}

static void hexdump(eventfs_event_t e)
{
    header             *packet;

    cout << "Bogus packet." << endl;
    cout << "Size:  " << eventfsEventGetSize(e) << endl;
    cout << "Types: " << eventfsEventGetTypes(e) << endl;

    packet = (header*)eventfsEventGetData(e);
    cout << "header:" << endl;
    cout << "  event  = " << packet->event  << endl;
    cout << "  ruid   = " << packet->ruid   << endl;
    cout << "  euid   = " << packet->euid   << endl;
    cout << "  auid   = " << packet->auid   << endl;
    cout << "  rgid   = " << packet->rgid   << endl;
    cout << "  egid   = " << packet->egid   << endl;
    cout << "  pid    = " << packet->pid    << endl;
    cout << "  ppid   = " << packet->ppid   << endl;
    cout << "  sid    = " << packet->sid    << endl;
    cout << "  retval = " << packet->retval << endl;
    cout << "Data remaining: " <<
        eventfsEventGetSize(e) - offsetof(header, retval) - sizeof(packet->retval)
        << endl;
}

Message *MessageSource::Next()
{
    Message            *retval;

    header             *packet;


    /* eventfsEventStreamGetNextEvent tries pretty hard.  If it
     * gives up its not something we can handle.
     */
    eventfs_event_t e;
   
    e = eventfsEventStreamGetNextEvent(stream);
    if (e == 0) {
        perror("eventfsEventStream::GetNextEvent()");
        return NULL;
    }
    if (eventfsEventGetSize(e) < sizeof(header)) {
        perror("eventfsEvent::GetSize()");
        eventfsEventRelease( e );
        return NULL;
    }

    packet = (header *)eventfsEventGetData(e);

    if (packet->timestamp < tv.tv_sec) {
        hexdump(e);
    }

    switch ( packet->event ) {
    default:
        retval = NULL;
        break;
    case KERN_EXIT:
        retval = new ExitEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_FORK:
        retval = new ForkEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_EXECVE:
        retval = new ExecEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_KILL:
        retval = new KillEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_SETREUID:
        retval = new SetreuidEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_MPROTECT:
        retval = new MProtectEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_LINK:
        retval = new LinkEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_SYMLINK:
        retval = new SymlinkEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_UNLINK:
        retval = new UnlinkEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_CHMOD:
        retval = new ChmodEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_CHOWN:
        retval = new ChownEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_FLOCK:
        retval = new FLockEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_UTIME:
        retval = new UTimeEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_RENAME:
        retval = new RenameEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_MKNOD:
        retval = new MKNodEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_MKDIR:
        retval = new MKDirEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_RMDIR:
        retval = new RMDirEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_CREAT:
        retval = new CreatEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_TRUNCATE:
        retval = new TruncateEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_OPEN_R:
        retval = new OpenREvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_OPEN_W:
        retval = new OpenWEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_OPEN_RW:
        retval = new OpenRWEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_CORE:
        retval = new CoreEvent(packet, eventfsEventGetSize(e));
        break;
    case KERN_ACCEPT:
        retval = new AcceptEvent(packet, eventfsEventGetSize(e));
        break;
    }

    eventfsEventRelease( e );

    return retval;
}


