%{
/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: pol.y,v 1.13 2001/04/05 18:41:47 stevek Exp $
 *
 * author: Lee Badger
 *
 * Purpose: Parser for Napoleon-generated policy files.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <pwd.h>

#include "policy.h"

#include <sys/types.h>
#ifdef WR_UNIX
#include <sys/param.h>
#endif
#include <sys/stat.h>

int polerror(char *s);
int pollex();
int polparse();

unsigned int construct_inet_entry(int inet1, int inet2, int inet3, int inet4);

/* Global container for generated policy. */
policy *pol;

%}

%union {
	int			u_int_const;
	charlist_t		u_charlist;
	invoke_statement	*u_invoke;
	invoke_list_t		u_invoke_list;
	access_specifier_t	u_access_spec;
	sub_parse_t		u_sub_parse;
	sub_parselist_t		u_sub_parselist;
	subject			*u_sub;
	sub_list_t		u_sublist;
	service_tuple_t		u_service;
	service_tuple_list_t	u_service_list;
	hostperm_list_t		u_hostperm_list;
	char			*u_name;
}

%token   <u_name>      IDENTIFIER
%token   <u_name>      K_PATH
%token   <u_int_const> NUM
%token   <u_name>      REGEX
%token  K_AND K_OR 
%token	K_SUBJECT K_OBJECT K_INVOKE K_SYSCALL K_FILE K_NETWORK
%token	K_ALLOW K_DENY K_IF K_USER K_EXEC K_HOMEDIR


%type <u_charlist> syscall
%type <u_charlist> syscall_list
%type <u_access_spec> allow_deny_option
%type <u_sub_parse> sub_syscall
%type <u_sub_parse> sub_file
%type <u_sub_parse> sub_network
%type <u_sub_parselist> sub_constraint
%type <u_sub_parselist> sub_constraint_list
%type <u_sublist> subject_policy
%type <u_sublist> subject_policy_list
%type <u_charlist> path
%type <u_charlist> homedir
%type <u_charlist> path_list

%type <u_invoke_list> invoke
%type <u_invoke_list> invoke_list
%type <u_invoke> invoke_clause
%type <u_invoke> invoke_expr
%type <u_charlist> invoke_user
%type <u_charlist> user_list
%type <u_charlist> user
%type <u_charlist> invoke_exec
%type <u_charlist> invoke_syscall

%type <u_service_list> service
%type <u_service_list> service_list
%type <u_int_const> port
%type <u_int_const> ip_addr
%type <u_int_const> octet

%type <u_hostperm_list> hostperm
%type <u_hostperm_list> hostperm_list

%left K_AND K_OR '!'

%%
pol:   subject_policy_list object_policy_list invoke_list
            {
		pol->add_subjects($1);
		pol->add_invoke_statements($3);
		/* pol now contains everything; we will traverse it
		 * to generate tables after polparse() returns. */
            }

/*----------------------------------------------------------------------
 *	Rules for subject_policy statements.
 */
subject_policy_list: subject_policy
	{
		$$ = $1;
	}
	|	subject_policy_list subject_policy
	{
		$$ = sub_list_merge($1, $2);
	}
	;

subject_policy: K_SUBJECT IDENTIFIER '{' sub_constraint_list '}' ';'
	{
		subject *sub = new subject();
		sub->subject_name = strdup($2);
		sub->add_constraints($4);
		sub_list_t list = new sub_list_s;
		list->append(sub);
		$$ = list;
	}

sub_constraint_list:	sub_constraint
	{	$$ = $1;	}
	|	sub_constraint_list sub_constraint
	{	$$ = sub_parselist_merge($1, $2);	}
	;

sub_constraint:	sub_syscall
	{
		/* put generic structure in a list */
		sub_parselist_t list = new sub_parselist_s;
		list->append($1);
		$$ = list;
	}
	|
	sub_file
	{	/* put generic structure in a list */
		sub_parselist_t list = new sub_parselist_s;
		list->append($1);
		$$ = list;
	}
	|
	sub_network
	{
		/* put generic structure in a list */
		sub_parselist_t list = new sub_parselist_s;
		list->append($1);
		$$ = list;
	}

/*----------------------------------------------------------------------
 *	Syscall clauses.
 */
sub_syscall:	K_SYSCALL allow_deny_option syscall_list ';'
	{
		/* build and return a generic structure, not a list */
		sub_parse_t prec = new sub_parse_s;
		prec->kind = "syscall";
		prec->sys_allow_deny = $2;
		prec->syscall_names = $3;
		$$ = prec;
	}

allow_deny_option:	K_ALLOW
	{	$$ = allow;	}
	| K_DENY
	{	$$ = deny;	}

syscall_list:	syscall
	{	$$ = $1;	}
	| syscall_list ',' syscall
	{	$$ = charlist_merge($1, $3);	}
	;

syscall:	IDENTIFIER
	{
		charlist_t list = new charlist_s;
		list->append(strdup($1));
		$$ = list;
	}

/*----------------------------------------------------------------------
 *	File clauses.
 */
sub_file:	K_FILE allow_deny_option IDENTIFIER path_list ';'
	{
		/* build and return a generic structure, not a list */
		sub_parse_t prec = new sub_parse_s;
		prec->kind = "file";
		prec->file_allow_deny = $2;
		if (valid_mode($3)) {
			prec->access_mode = strdup($3);
		}
		else {
			printf("invalid access mode (%s)\n", $3);
			exit(1);
		}
		prec->paths = $4;
		$$ = prec;
	}

path_list: path
	{	$$ = $1;	}
	| path_list '>' path
	{	$$ = charlist_merge($1, $3);	}



path: K_PATH
	{
		charlist_t list = new charlist_s;
		list->append(strdup($1));
		$$ = list;
	}
	| homedir
	{ $$ = $1; }

homedir: K_HOMEDIR '{' IDENTIFIER '}' K_PATH
	{
		char *dir;
		struct passwd *pwd = getpwnam($3);

		/*
		 * Expand a user's home directory.  If the user does not
		 * exist, generate a bogus homedir to allow napload to be
		 * run for testing without errors.
		 */
		if (pwd) {
			dir = (char*) malloc(strlen(pwd->pw_dir) +
				strlen($5) + 1);
			sprintf(dir, "%s%s", pwd->pw_dir, $5);
		}
		else {
			dir = (char*) malloc(strlen("/bogus/") +
				strlen($3) + strlen("-homedir/") +
				strlen($5) + 1);
			sprintf(dir, "/bogus/%s-homedir%s", $3, $5);
		}
		charlist_t list = new charlist_s;
		list->append(dir);
		$$ = list;
	}

/*----------------------------------------------------------------------
 *	Network clauses.
 */
	sub_network:	K_NETWORK allow_deny_option hostperm_list ';'
	{
		/* build and return a generic structure, not a list */
		sub_parse_t prec = new sub_parse_s;
		prec->kind = "network";
		prec->network_allow_deny = $2;
		prec->hostperms = $3;
		$$ = prec;
	}

hostperm_list: hostperm
	{	$$ = $1;	}
	| hostperm_list ',' hostperm
	{	$$ = hostperm_list_merge($1, $3);	}
	;

hostperm: '(' ip_addr ',' service_list ')'
	{
		hostperm_t host = new hostperm_s;
		host->ip_addr = $2;
		host->services = $4;
		hostperm_list_t list = new hostperm_list_s;
		list->append(host);
		$$ = list;
	}

octet: NUM
       | '*' {
           $$ = 255;
       }

ip_addr: octet '.' octet '.' octet '.' octet 
	{
		$$ = construct_inet_entry($1, $3, $5, $7);
	}

service_list: service
	{	$$ = $1;	}
	| service_list ',' service
	{	$$ = service_tuple_list_merge($1, $3);	}
	;

service: '(' IDENTIFIER ',' port ',' port ')'
	{
		service_tuple_t tuple = new service_tuple_s;
		tuple->transport = strdup($2);
		tuple->local_port = $4;
		tuple->remote_port = $6;
		service_tuple_list_t list = new service_tuple_list_s;
		list->append(tuple);
		$$ = list;
	}

port: NUM
	{	$$ = $1;	}
	| '*'
	{	$$ = -1;	}

/*----------------------------------------------------------------------
 *	Rules for object_policy statements.
 */

object_policy_list:	object_policy_list object_policy
	|
	;

object_policy:	K_OBJECT IDENTIFIER '{' '}' ';'
	{
		printf("saw object(%s)\n", $2);
	}

/*----------------------------------------------------------------------
 *	Rules for invoke statements.
 */
invoke_list:	invoke
	{ $$ = $1; }
	| invoke_list invoke
	{ $$ = invoke_list_merge($1, $2); }
	;

invoke:	K_INVOKE IDENTIFIER '{' IDENTIFIER K_IF invoke_expr '}' ';'
	{
		/* Finish building the invoke_statement object, and put it
		 * in a list. */
		invoke_statement *inv = $6;
		inv->invoke_name = strdup($2);
		inv->subject = strdup($4);
		invoke_list_t list = new invoke_list_s;
		list->append(inv);
		$$ = list;
	}

invoke_expr:	invoke_clause
	{ $$ = $1; }
	|	'(' invoke_clause ')'
	{ $$ = $2; }
	|	invoke_expr K_AND invoke_expr
	{
		/* merge two invoke_statment objects together */
		invoke_statement *inv1 = $1;
		invoke_statement *inv2 = $3;

		inv1->exec_list = charlist_merge(inv1->exec_list,
			inv2->exec_list);
		inv1->user_list = charlist_merge(inv1->user_list,
			inv2->user_list);
		inv1->syscall_list = charlist_merge(inv1->syscall_list,
			inv2->syscall_list);
                // see checking message
		// delete inv2;
		$$ = inv1;
	}

invoke_clause: invoke_user
	{
		invoke_statement *invoke = new invoke_statement;
                // see checking message
		// delete(invoke->user_list);
		invoke->user_list = $1;
		$$ = invoke;
	}
	| invoke_exec
	{
		invoke_statement *invoke = new invoke_statement;
                // see checking message
		// delete(invoke->exec_list);
		invoke->exec_list = $1;
		$$ = invoke;
	}
	| invoke_syscall
	{
		invoke_statement *invoke = new invoke_statement;
                // see checking message
		// delete(invoke->syscall_list);
		invoke->syscall_list = $1;
		$$ = invoke;
	}

invoke_user: K_USER user_list
	{ $$ = $2; }

user_list: user
	{	$$ = $1;	}
	| user_list ',' user
	{	$$ = charlist_merge($1, $3);	}

user:	IDENTIFIER
	{
		charlist_t list = new charlist_s;
		list->append(strdup($1));
		$$ = list;
	}

invoke_exec: K_EXEC path_list
	{ $$ = $2; }

invoke_syscall: K_SYSCALL syscall_list
	{ $$ = $2; }

%%

#include "lex.pol.c"

int
polwrap()
{
   return 1;
}


/*
 * Set the input file from which lex reads, create an object to hold
 * the parsed policy, and invoke yacc.
 */
policy *parse_policy_file(FILE *fp)
{
	yyin = fp;
	pol = new policy;

	int err = polparse();
	if (err) {
		return(NULL);
	}
	else {
		return(pol);
	}
}

/*
 * Purpose: Allocates an internet entry.
 *
 * Returns: a newly allocated structure containing an inet address.
 */
unsigned int construct_inet_entry(
                int inet1,      /* IN, components of a dotted inet address */
                int inet2,      /* ...                                     */
                int inet3,      /* ...                                     */
                int inet4)      /* ...                                     */
{
        unsigned int addr = 0;
	/*
	 * From the routing legacy, root net stuff goes in the LSB
	 */
        addr |= (inet4 & 0xFF) << 24;
        addr |= (inet3 & 0xFF) << 16;
        addr |= (inet2 & 0xFF) << 8;
        addr |= (inet1 & 0xFF);

        return(addr);
}
