/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: policy.cc,v 1.17 2001/04/05 18:41:47 stevek Exp $
 *
 * Author: Lee Badger
 *
 * Purpose: Objects to hold a policy as we parse, and then write it out.
 */

#include "policy.h"

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <pwd.h>
#include <assert.h>

/*----------------------------------------------------------------------
 * Helper functions.
 */

/*
 * Merge two lists of char* records, and return the merge.  The second list
 * is deleted.
 */
charlist_t charlist_merge(charlist_t one, charlist_t two)
{
        assert(two!=one);

	char *name;
	two->reset_iterator();
	while ((name = two->next())) {
		one->append(name);
	}
        // see checking message for reasoning
	// delete two;
	return(one);
}

/*
 * Merge two lists of sub_parse_t records, and return the merge.
 */
sub_parselist_t sub_parselist_merge(sub_parselist_t one, sub_parselist_t two)
{
        assert(two!=one);

	sub_parse_t rec;
	two->reset_iterator();
	while ((rec = two->next())) {
		one->append(rec);
	}
        // see checking message for reasoning
	// delete two;
	return(one);
}

/*
 * Merge two lists of subject* records, and return the merge.
 */
sub_list_t sub_list_merge(sub_list_t one, sub_list_t two)
{
        assert(two!=one);

	subject *rec;
	two->reset_iterator();
	while ((rec = two->next())) {
		one->insert_ordered(rec);
	}
        // see checking message for reasoning
	// delete two;
	return(one);
}

/*
 * Merge two lists of invoke_statement* records, and return the merge.
 */
invoke_list_t invoke_list_merge(invoke_list_t one, invoke_list_t two)
{
        assert(two!=one);

	invoke_statement *rec;
	two->reset_iterator();
	while ((rec = two->next())) {
		one->append(rec);
	}
        // see checking message for reasoning
	// delete two;
	return(one);
}

service_tuple_list_t service_tuple_list_merge(service_tuple_list_t one,
		service_tuple_list_t two)
{
        assert(two!=one);

	service_tuple_t tuple;
	two->reset_iterator();
	while ((tuple = two->next())) {
		one->append(tuple);
	}
        // see checking message for reasoning
	// delete two;
	return(one);
}

hostperm_list_t hostperm_list_merge(hostperm_list_t one,
		hostperm_list_t two)
{
        assert(two!=one);

	hostperm_t rec;
	two->reset_iterator();
	while ((rec = two->next())) {
		one->append(rec);
	}
        // see checking message for reasoning
	// delete two;
	return(one);
}

/*
 * A mode must be comprised of only "r" "x" and "w"
 */
int valid_mode(char *mode_str)
{
	return(strlen(mode_str) == strspn(mode_str, "rwx"));
}

/*----------------------------------------------------------------------
 * Object methods.
 */

subject::subject()
{
	has_syscalls = 0;
	has_files = 0;
	has_networks = 0;

	syscall_names = new UT_list<char*>;
	syscall_allow_deny = invalid;
	file_statements = new sub_parselist_s;
	network_statements = new sub_parselist_s;
}

/*
 * Accepts an unordered list of syscall, file, and network constraints,
 * and fills in the subject object.
 */
void subject::add_constraints(sub_parselist_t list)
{
	list->reset_iterator();
	sub_parse_t rec;
	while ((rec = list->next())) {
		if (!strcmp("syscall", rec->kind)) {
			has_syscalls = 1;
			if (syscall_allow_deny == invalid) {
				syscall_allow_deny = rec->sys_allow_deny;
			}
			else if (syscall_allow_deny != rec->sys_allow_deny) {
				printf(
	"system call allow/deny conflict in subject \"%s\"\n", name());
				exit(1);
			}

			syscall_names = charlist_merge(syscall_names,
						rec->syscall_names);
			list->remove(rec);
                        // see checking message for reasoning
			// delete rec;
		}
		else if (!strcmp("file", rec->kind)) {
			list->remove(rec);
			has_files = 1;
			file_statements->append(rec);
		}
		else if (!strcmp("network", rec->kind)) {
			list->remove(rec);
			has_networks = 1;
			network_statements->append(rec);
		}
	}
}

/*
 * Initialize an invoke statement.  An empty list means there are no such
 * clauses in the invoke statement.
 */
invoke_statement::invoke_statement()
{
	invoke_name = NULL;
	subject = NULL;
	exec_list = new charlist_s;
	user_list = new charlist_s;
	syscall_list = new charlist_s;
}

void invoke_statement::print(int)
{
}

/*
 * Print out the policy state.
 */
void policy::print()
{
	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		printf("SubjectPolicy(%s)\n", sub->name());
		if (sub->has_syscalls) {
			UT_list<char*> *list = sub->syscall_names;
			list->reset_iterator();
			char *sysname;
			printf("\tSyscalls ");
			if (sub->syscall_allow_deny == allow) {
				printf("[allowed] ");
			}
			else {
				printf("[denied] ");
			}
			while ((sysname = list->next())) {
				printf("%s ", sysname);
			}
			printf("\n");
		}

		if (sub->has_files) {
			printf("\tFiles\n");
			sub_parselist_t file_list = sub->file_statements;
			file_list->reset_iterator();
			sub_parse_t file_rec;
			while ((file_rec = file_list->next())) {
				printf("\t\tfile ");
				if (file_rec->file_allow_deny == allow) {
					printf("allow ");
				}
				else {
					printf("deny ");
				}
				printf("%s ", file_rec->access_mode);
				charlist_t paths = file_rec->paths;
				paths->reset_iterator();
				char *path;
				while ((path = paths->next())) {
					printf("%s ", path);
				}
				printf("\n");
			}
		}

		if (sub->has_networks) {
			printf("\tNetwork\n");
			sub_parselist_t list = sub->network_statements;
			list->reset_iterator();
			sub_parse_t rec;
			while ((rec = list->next())) {
				printf("\t\tnetwork ");
				if (rec->network_allow_deny == allow) {
					printf("allow ");
				}
				else {
					printf("deny ");
				}
				hostperm_list_t host_list = rec->hostperms;

				host_list->reset_iterator();
				hostperm_t hostperm;
				while ((hostperm = host_list->next())) {
					printf("%x ", hostperm->ip_addr);

					service_tuple_list_t slist =
							hostperm->services;
					slist->reset_iterator();
					service_tuple_t tup;
					while ((tup = slist->next())) {
						printf("(%s, %d, %d) ",
							tup->transport,
							tup->local_port,
							tup->remote_port);
					}
				}
				printf("\n");
			}
		}
	}

	invoke_list_t inv_list = invoke_statements;
	inv_list->reset_iterator();
	invoke_statement *inv;
	while ((inv = inv_list->next())) {
		printf("Invoke(%s) To %s if ", inv->name(), inv->subject);
		if (! inv->exec_list->empty()) {
			printf("exec(");
			charlist_t list = inv->exec_list;
			list->reset_iterator();
			char *file;
			while ((file = list->next())) {
				printf("%s ", file);
			}
			printf(")");
		}
		printf(" ");
		if (! inv->user_list->empty()) {
			printf("user(");
			charlist_t list = inv->user_list;
			list->reset_iterator();
			char *user;
			while ((user = list->next())) {
				printf("%s ", user);
			}
			printf(")");
		}

		printf("\n");
	}
}

/*
 * Write the policy object out in database tables in files in human-readable
 * format:
 *	sub_syscalls
 *	sub_files
 *	sub_addrs
 *	sub_ports
 *
 *	inv_progs
 */
void policy::write_files()
{
	write_sub_syscalls();
	write_sub_files();
	write_sub_addrs();
	write_sub_ports();

	write_inv_progs();
	write_inv_users();
}

/* this struct must exactly match the one in systable.cc, awk-generated */
struct systable {
	char *name;
	int num;
};

#define	MAX_SYSCALL	256	/* make this a multiple of 8 */
#define	SYSBUF_LEN	(MAX_SYSCALL/8)

int sysname_to_offset(char *name)
{
	extern struct systable systable[];
	for (int i=0; i<MAX_SYSCALL&&systable[i].name; i++) {
		if (!strcmp(name, systable[i].name)) {
			return(systable[i].num);
		}
	}
	fprintf(stderr, "sysname %s doesn't exist, exiting...\n", name);
	exit(1);
}

/*
 * Within each byte, the bits are indexed least significant to most.
 */
void set_vector_bit(char buf[], int index)
{
	int byte = index / 8;
	int bit = index % 8;

	buf[byte] |= (1 << bit);
}

/*
 * Write the sub_syscalls table into a file of that name.nap in the current
 * directory.
 */
void policy::write_sub_syscalls()
{
	int fd = open("sub_syscalls.nap", O_RDWR|O_CREAT|O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "create of sub_syscalls.nap failed\n");
		exit(1);
	}

	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		if (! sub->has_syscalls) continue;

		char line[160];
		sprintf(line, "%s\t", sub->name());
		write(fd, line, strlen(line));

		if (sub->syscall_allow_deny == allow) {
			write(fd, "allow\t", strlen("allow\t"));
		}
		else {
			write(fd, "deny\t", strlen("deny\t"));
		}

		/* Construct the mask and write it out. */
		char buf[SYSBUF_LEN];
		bzero(buf, SYSBUF_LEN);

		sub->syscall_names->reset_iterator();
		char *name;
		while ((name = sub->syscall_names->next())) {
			int offset = sysname_to_offset(name);
			set_vector_bit(buf, offset);
		}

		char printbuf[SYSBUF_LEN*2];
		for (int i=0; i<SYSBUF_LEN; i++) {
			sprintf(&printbuf[i*2], "%2x", (int)buf[i]);
		}
		write(fd, printbuf, SYSBUF_LEN*2);
		write(fd, "\n", strlen("\n"));
	}
	close(fd);
}


/*
 * Write the sub_files table into a file of that name.nap in the current
 * directory.
 */
void policy::write_sub_files()
{
	int fd = open("sub_files.nap", O_RDWR|O_CREAT|O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "create of sub_files.nap failed\n");
		exit(1);
	}

	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		if (! sub->has_files) continue;

		sub_parselist_t rlist = sub->file_statements;
		sub_parse_t rec;
		rlist->reset_iterator();
		while ((rec = rlist->next())) {
			char line[160];
			sprintf(line, "%s\t%s\t%s\t", sub->name(), 
				rec->file_allow_deny == allow ?
				"allow" : "deny", rec->access_mode);
			rec->paths->reset_iterator();
			char *path;
			while ((path = rec->paths->next())) {
				write(fd, line, strlen(line));

				write(fd, path, strlen(path));
				write(fd, "\n", strlen("\n"));
			}
		}
	}
	close(fd);
}

/*
 * Write the sub_addrs table into a file of that name.nap in the current
 * directory.
 */
void policy::write_sub_addrs()
{
	int fd = open("sub_addrs.nap", O_RDWR|O_CREAT|O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "create of sub_addrs.nap failed\n");
		exit(1);
	}

	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		if (! sub->has_networks) continue;

		sub_parselist_t rlist = sub->network_statements;
		sub_parse_t rec;
		rlist->reset_iterator();
		while ((rec = rlist->next())) {

			rec->hostperms->reset_iterator();
			hostperm_t host;
			while ((host = rec->hostperms->next())) {
				char line[160];
				sprintf(line, "%s\t%s\t%x\n", sub->name(), 
					rec->network_allow_deny ?
							"allow" : "deny",
					host->ip_addr);
				write(fd, line, strlen(line));
			}
		}
	}
	close(fd);
}

/*
 * Write the sub_ports table into a file of that name.nap in the current
 * directory.
 */
void policy::write_sub_ports()
{
	int fd = open("sub_ports.nap", O_RDWR|O_CREAT|O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "create of sub_ports.nap failed\n");
		exit(1);
	}

	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		if (! sub->has_networks) continue;

		sub_parselist_t rlist = sub->network_statements;
		sub_parse_t rec;
		rlist->reset_iterator();
		while ((rec = rlist->next())) {

			rec->hostperms->reset_iterator();
			hostperm_t host;
			while ((host = rec->hostperms->next())) {

				host->services->reset_iterator();
				service_tuple_t service;
				while ((service = host->services->next())) {
					char line[160];

					sprintf(line, "%s\t%x\t%s\t%d\t%d\n",
						sub->name(), 
						host->ip_addr,
						service->transport,
						service->local_port,
						service->remote_port);
					write(fd, line, strlen(line));
				}
			}
		}
	}
	close(fd);
}

/*
 * Write the inv_progs table into a file of that name.nap in the current
 * directory.
 */
void policy::write_inv_progs()
{
	int fd = open("inv_progs.nap", O_RDWR|O_CREAT|O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "create of inv_progs.nap failed\n");
		exit(1);
	}

	invoke_statement *inv;
	invoke_statements->reset_iterator();
	while ((inv = invoke_statements->next())) {
		inv->exec_list->reset_iterator();
		char *path;
		while ((path = inv->exec_list->next())) {
			char buf[4096];
			sprintf(buf, "%s\t%s\t%s\n", inv->name(),
				inv->subject, path);
			write(fd, buf, strlen(buf));
		}
	}
	close(fd);
}

/*
 * Write the inv_users table into a file of that name.nap in the current
 * directory.
 */
void policy::write_inv_users()
{
	int fd = open("inv_users.nap", O_RDWR|O_CREAT|O_TRUNC,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	if (fd == -1) {
		fprintf(stderr, "create of inv_users.nap failed\n");
		exit(1);
	}

	invoke_statement *inv;
	invoke_statements->reset_iterator();
	while ((inv = invoke_statements->next())) {
		inv->user_list->reset_iterator();
		char *user;
		while ((user = inv->user_list->next())) {
			char buf[4096];
			sprintf(buf, "%s\t%s\t%s\n", inv->name(),
				inv->subject, user);
			write(fd, buf, strlen(buf));
		}
	}
	close(fd);
}

#ifndef	IN_SANDBOX

/*
 * Placeholder function for when this source is not built in a wrappers
 * distribution.
 */
void policy::load_kernel()
{
	printf(
"This copy of napload has been compiled outside of the wrappers\n\
sandbox environment.  It therefore does not have support linked in\n\
for loading information into the wrappers kernel subsystem.\n");
}

#else

#include "libwrv.h"

/*
 * Fill the sub_syscalls table.
 */
void policy::load_sub_syscalls(unsigned long table_handle)
{
	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		char *new_row[] = { "policy", "verb", "vector", NULL };

		if (! sub->has_syscalls) continue;

		new_row[0] = sub->name();

		if (sub->syscall_allow_deny == allow) {
			new_row[1] = "allow";
		}
		else {
			new_row[1] = "deny";
		}

		/* Construct the mask. */
		char buf[SYSBUF_LEN];
		bzero(buf, SYSBUF_LEN);
		sub->syscall_names->reset_iterator();
		char *name;
		while ((name = sub->syscall_names->next())) {
			int offset = sysname_to_offset(name);
			set_vector_bit(buf, offset);
		}
		char printbuf[SYSBUF_LEN*2];
		for (int i=0; i<SYSBUF_LEN; i++) {
			sprintf(&printbuf[i*2], "%2x", (int)buf[i]);
		}
		new_row[2] = printbuf;

		wrvDbAddTableRow(table_handle, new_row);
	}
}

/*
 * Fill the sub_files table.
 */
void policy::load_sub_files(unsigned long table_handle)
{
	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		if (! sub->has_files) continue;

		sub_parselist_t rlist = sub->file_statements;
		sub_parse_t rec;
		rlist->reset_iterator();
		while ((rec = rlist->next())) {
			char *new_row[] =
				{ "policy", "verb", "mode", "file", NULL };

			new_row[0] = sub->name();
			if (rec->file_allow_deny == allow) {
				new_row[1] = "allow";
			}
			else {
				new_row[1] = "deny";
			}
			new_row[2] = rec->access_mode;
			rec->paths->reset_iterator();
			char *path;
			while ((path = rec->paths->next())) {
				new_row[3] = path;
				wrvDbAddTableRow(table_handle, new_row);
			}
		}
	}
}

/*
 * Fill the sub_addrs table.
 */
void policy::load_sub_addrs(unsigned long table_handle)
{
	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		if (! sub->has_networks) continue;

		sub_parselist_t rlist = sub->network_statements;
		sub_parse_t rec;
		rlist->reset_iterator();
		while ((rec = rlist->next())) {

			rec->hostperms->reset_iterator();
			hostperm_t host;
			while ((host = rec->hostperms->next())) {
				char *new_row[] =
					{ "policy", "verb", "addr", NULL };

				new_row[0] = sub->name();
				if (rec->network_allow_deny == allow) {
					new_row[1] = "allow";
				}
				else {
					new_row[1] = "deny";
				}
				char line[160];
				sprintf(line, "%d", host->ip_addr);
				new_row[2] = line;
				wrvDbAddTableRow(table_handle, new_row);
			}
		}
	}
}

/*
 * Fill the sub_ports table.
 */
void policy::load_sub_ports(unsigned long table_handle)
{
	subjects->reset_iterator();
	subject *sub;
	while ((sub = subjects->next())) {
		if (! sub->has_networks) continue;

		sub_parselist_t rlist = sub->network_statements;
		sub_parse_t rec;
		rlist->reset_iterator();
		while ((rec = rlist->next())) {

			rec->hostperms->reset_iterator();
			hostperm_t host;
			while ((host = rec->hostperms->next())) {

				host->services->reset_iterator();
				service_tuple_t service;
				while ((service = host->services->next())) {
					char *new_row[] =
		{ "policy", "addr", "transport", "l_port", "r_port", NULL };

					new_row[0] = sub->name();

					char line1[160];
					sprintf(line1, "%d", host->ip_addr);
					new_row[1] = line1;
					new_row[2] = service->transport;

					char line2[160];
					sprintf(line2, "%d",
							service->local_port);
					new_row[3] = line2;

					char line3[160];
					sprintf(line3, "%d",
							service->remote_port);
					new_row[4] = line3;

					wrvDbAddTableRow(table_handle,new_row);
				}
			}
		}
	}
}

/*
 * Fill the inv_progs table.
 */
void policy::load_inv_progs(unsigned long table_handle)
{
	invoke_statement *inv;
	invoke_statements->reset_iterator();
	while ((inv = invoke_statements->next())) {
		inv->exec_list->reset_iterator();
		char *path;
		while ((path = inv->exec_list->next())) {
			char *new_row[] = { "id", "policy", "program", NULL };

			new_row[0] = inv->name();
			new_row[1] = inv->subject;
			new_row[2] = path;
			wrvDbAddTableRow(table_handle, new_row);
		}
	}
}

/*
 * Fill the inv_users table.
 */
void policy::load_inv_users(unsigned long table_handle)
{
	invoke_statement *inv;
	invoke_statements->reset_iterator();
	while ((inv = invoke_statements->next())) {
		inv->user_list->reset_iterator();
		char *user;
		while ((user = inv->user_list->next())) {
                        char buf[11]; /* ceil((2^32 log 10)) + 1 */
                                      // I can write C in any language...
			char *new_row[] = { "id", "policy", "user", NULL };

                        struct passwd *pwent = getpwnam(user);
                        if (pwent)
                            sprintf(buf,"%d",pwent->pw_uid);
                        else
                            sprintf(buf, "-1");

			new_row[0] = inv->name();
			new_row[1] = inv->subject;
			new_row[2] = buf;
			wrvDbAddTableRow(table_handle, new_row);
		}
	}
}

/*
 * Write table information to the tables in the WSS.  Assumes that the
 * napenforce wrapper has already created the tables.
 */
static void emptyTable(unsigned long thandle)
{
    wrdb_row_t rows, ri;

    rows = wrvDbGetRows(thandle);

    if (!rows)
        return;

    for (ri = rows; ri != NULL ;ri = ROW_NEXT(ri) ) {
        wrvDbDelTableRow(thandle, ri->handle);
    }

    wrvDbReleaseRows(rows);
}

void policy::load_kernel()
{
	wrv_table_id_t *tabs = (wrv_table_id_t*) NULL;

	tabs = wrvDbGetTableIds();
	while (tabs) {
		if (tabs->handle != 0) {
			/* printf("%s ", tabs->name); */

			if (!strcmp(tabs->name, "napenforce.sub_syscalls")) {
                                emptyTable(tabs->handle);
				load_sub_syscalls(tabs->handle);
			}
			else if (!strcmp(tabs->name, "napenforce.sub_files")) {
                                emptyTable(tabs->handle);
				load_sub_files(tabs->handle);
			}
			else if (!strcmp(tabs->name, "napenforce.sub_addrs")) {
                                emptyTable(tabs->handle);
				load_sub_addrs(tabs->handle);
			}
			else if (!strcmp(tabs->name, "napenforce.sub_ports")) {
                                emptyTable(tabs->handle);
				load_sub_ports(tabs->handle);
			}
			else if (!strcmp(tabs->name, "napenforce.inv_progs")) {
                                emptyTable(tabs->handle);
				load_inv_progs(tabs->handle);
			}
			else if (!strcmp(tabs->name, "napenforce.inv_users")) {
                                emptyTable(tabs->handle);
				load_inv_users(tabs->handle);
			}


			/* TBD.  Add more load_* functions.  Two more are
			   needed for the networking info. */

		}
		else {
			printf("oops, null handle\n");
		}
		tabs = tabs->next;
	}
}

#endif
