/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: policy.h,v 1.11 2001/02/23 19:02:30 stevek Exp $
 *
 * Author: Lee Badger
 *
 * Purpose: Objects to hold a policy as we parse, consistency-check it,
 *	and then write it out.
 */

#ifndef _POLICY_H_
#define _POLICY_H_

#include <stdio.h>
#include "utils.h"


/*----------------------------------------------------------------------
 * Simple types.
 */
typedef enum access_specifier { invalid, allow, deny } access_specifier_t;
typedef UT_list<char*> charlist_s, *charlist_t;

typedef	struct service_tuple {
	char *transport;
	int remote_port;
	int local_port;
} service_tuple_s, *service_tuple_t;
typedef	UT_list<service_tuple_t> service_tuple_list_s, *service_tuple_list_t;

typedef struct hostperm {
	int ip_addr;
	service_tuple_list_t services;	
} hostperm_s, *hostperm_t;
typedef	UT_list<hostperm_t> hostperm_list_s, *hostperm_list_t;

/* really should contain a union */
typedef struct sub_parse {
	char *kind;

	/* fields for a syscall constraint */
	access_specifier_t sys_allow_deny;
	charlist_t syscall_names;

	/* fields for a file constraint */
	access_specifier_t file_allow_deny;
	char *access_mode;
	charlist_t paths;

	/* fields for a network constraint */
	access_specifier_t network_allow_deny;
	hostperm_list_t hostperms;
} sub_parse_s, *sub_parse_t;
typedef UT_list<sub_parse_t> sub_parselist_s, *sub_parselist_t;


/*----------------------------------------------------------------------
 * Objects.
 */

/*
 * Represents a subject policy.
 */
class subject {
public:
	char *subject_name;
	void add_constraints(sub_parselist_t recs);
	subject();
	~subject() { delete syscall_names; }
	void print(int) {}
	char *name() { return(subject_name); }

	/* syscall constraints */
	int has_syscalls;
	access_specifier_t syscall_allow_deny;
	UT_list<char*> *syscall_names;

	/* file constraints */
	int has_files;
	sub_parselist_t file_statements;

	/* network constraints */
	int has_networks;
	sub_parselist_t network_statements;
};

/*
 * Represents an invoke statement.
 */
class invoke_statement {
public:
	char *invoke_name;
	char *name() { return(invoke_name); }
	void print(int level);
	char *subject;
	charlist_t exec_list;
	charlist_t user_list;
	charlist_t syscall_list;
	invoke_statement();
};

typedef UT_named_list<subject*> sub_list_s, *sub_list_t;
typedef UT_named_list<invoke_statement*> invoke_list_s, *invoke_list_t;


/*----------------------------------------------------------------------
 * Helper functions.
 */
charlist_t charlist_merge(charlist_t one, charlist_t two);
sub_parselist_t sub_parselist_merge(sub_parselist_t one, sub_parselist_t two);
sub_list_t sub_list_merge(sub_list_t one, sub_list_t two);
invoke_list_t invoke_list_merge(invoke_list_t one, invoke_list_t two);
int valid_mode(char *mode_str);
service_tuple_list_t service_tuple_list_merge(service_tuple_list_t one,
		service_tuple_list_t two);
hostperm_list_t hostperm_list_merge(hostperm_list_t one,
		hostperm_list_t two);

/*
 * Container for the entire policy that Napoleon has generated.
 */
class policy {
public:
	void add_subjects(sub_list_t subs) { subjects = subs; }
	void add_invoke_statements(invoke_list_t invs)
				{ invoke_statements = invs; }
	void print();
	void write_files();
	void load_kernel();
private:
	sub_list_t subjects;
	invoke_list_t invoke_statements;
	void write_sub_syscalls();
	void write_sub_files();
	void write_sub_addrs();
	void write_sub_ports();
	void write_inv_progs();
	void write_inv_users();
	void load_sub_syscalls(unsigned long table_handle);
	void load_sub_files(unsigned long table_handle);
	void load_sub_addrs(unsigned long table_handle);
	void load_sub_ports(unsigned long table_handle);
	void load_inv_progs(unsigned long table_handle);
	void load_inv_users(unsigned long table_handle);
};

#endif _POLICY_H_
