/*! \file linker_pe.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: linker_pe.c,v 1.9 2001/02/23 19:02:30 stevek Exp $
 *
 * \brief   Support for loading PE format modules
 * \author  Steve Kiernan
 * \version $Revision: 1.9 $
 * \date    $Date: 2001/02/23 19:02:30 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_file.h>
#include <wss_malloc.h>
#include <wss_linker.h>
#include <wss_module.h>

#include <linker_pe.h>


/***************************************************************************\
 * Private structures and types                                            *
\***************************************************************************/

typedef pd_char_t **(*get_imp_list_func)        ( void );
typedef pd_char_t  *(*get_mod_name_func)        ( void );
typedef pd_char_t  *(*get_mod_type_func)        ( void );
typedef pd_char_t  *(*get_mod_dependencies_func)( void );

typedef struct linker_pe_namefn linker_pe_namefn_t;
typedef struct linker_pe_data   linker_pe_data_t;

struct linker_pe_namefn
{
   const pd_char_t *name;
   const pd_char_t *fnname;
};

struct linker_pe_data
{
   HINSTANCE         hInstance;
};


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#define IMAGE_NT_SIGNATURE 0x00004550


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

static int _linker_pe_load           __P(( const pd_char_t *,
                                           wss_linker_file_t * ));
static int _linker_pe_lookup_symbol  __P(( wss_linker_file_t *,
                                           const pd_char_t *,
                                           wss_linker_addr_t * ));
static int _linker_pe_export_symbols __P(( wss_linker_file_t * ));
static int _linker_pe_finalize       __P(( wss_linker_file_t * ));
static int _linker_pe_unload         __P(( wss_linker_file_t * ));


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static wss_linker_ops_t _linker_pe_ops =
{
   _linker_pe_load
};

static wss_linker_file_ops_t _linker_pe_file_ops =
{
   _linker_pe_lookup_symbol,
   _linker_pe_export_symbols,
   _linker_pe_finalize,
   _linker_pe_unload
};

static const pd_char_t *_linker_pe_description = PD_MAKESTR("PE");

static linker_pe_namefn_t _linker_pe_name_to_func[] =
{
   { PD_MAKESTR(WRLINKNAME_MOD),
     PD_MAKESTR("get_mod_name") },
   { PD_MAKESTR(WRLINKNAME_MOD_TYPE),
     PD_MAKESTR("get_mod_type") },
   { PD_MAKESTR(WRLINKNAME_MOD_DEPENDENCIES),
     PD_MAKESTR("get_mod_dependencies") },
   { PD_MAKESTR("imp_list"),
     PD_MAKESTR("get_imp_list") },
   { NULL,
     NULL }
};


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static char *
_linker_pe_wide_to_multibyte( const pd_char_t *str )
{
#ifdef UNICODE
   char *outstr = NULL;

   if( str )
   {
      DWORD strLen = wsi_strlen( str ) + 1;

      outstr = (char *)malloc( strLen * sizeof(pd_char_t) );
      WideCharToMultiByte( CP_ACP, 0, str, strLen, outstr,
                           strLen * sizeof(pd_char_t), NULL, NULL );
   }
   return outstr;
#else /* UNICODE */
   return str;
#endif /* UNICODE */
}

static void
_linker_pe_free_multibyte( char *str )
{
#ifdef UNICODE
   if( str )
      wss_free( str );
#endif /* UNICODE */
}

/*
 * Lookup a single symbol in the binary's symbol table and return the
 * symbol's address
 */
static int
_linker_pe_lookup_symbol( wss_linker_file_t *lf, const pd_char_t *symbol,
                          wss_linker_addr_t *addr )
{
   linker_pe_data_t *pedata  = (linker_pe_data_t *)lf->linker_file_data;
   BOOL              callfn  = FALSE;
   char             *symname;
   void             *fnaddr;
   int               i;

   if( ( addr == NULL ) || ( symbol == NULL ) )
      return EINVAL;

   for( i = 0; _linker_pe_name_to_func[i].name; i++ )
   {
      if( !wsi_strcmp( _linker_pe_name_to_func[i].name, symbol ) )
      {
         symbol = _linker_pe_name_to_func[i].fnname;
         callfn = TRUE;
         break;
      }
   }

   symname = _linker_pe_wide_to_multibyte( symbol );

   fnaddr = GetProcAddress( pedata->hInstance, symname );

   _linker_pe_free_multibyte( symname );

   if( callfn && fnaddr )
      fnaddr = ((void *(*)())fnaddr)();

   if( fnaddr )
   {
      addr->addr = fnaddr;
      return 0;
   }

   return ENOENT;
}


/*
 * Export library symbols for other modules to use
 */
static int
_linker_pe_export_symbols( wss_linker_file_t *lf )
{
   return 0;
}


/*  _linker_pe_load */
/*! \brief Load a PE format module
 *
 *  Uses LoadLibrary to load the module.  It does \b NOT validate the
 *  symbols before it loads the module.  This is something that should be
 *  done in a production version, but since we are in user space, there are
 *  worse things to worry about.
 *
 *  \param filename Pointer to string containing the name of the file to load
 *  \param result   Pointer to linker file structure to fill
 *
 *  \return 0 if successfully loaded,
 *  \return ENOENT if unable to resolve symbols,
 *  \return EEXIST if the module is already loaded,
 *  \return EINVAL if the module is of an unknown type
 */                                                               
static int
_linker_pe_load( const pd_char_t *filename, wss_linker_file_t *result )
{
   wss_file_t          file;
   UCHAR               dossig[2];
   DWORD               sig;
   linker_pe_data_t   *pedata;
   wss_linker_addr_t   addr;
   int                 error     = 0;


   /* Make sure result pointer is non-NULL */
   if( result == NULL )
   {
#ifdef DEBUG_PE_LINKER
      wsi_printf( PD_MAKESTR("PE Linker: result argument is NULL\n") );
#endif /* DEBUG_PE_LINKER */
      return EINVAL;
   }

#ifdef DEBUG_PE_LINKER
   wsi_printf( PD_MAKESTR("PE Linker: Loading '%s'...\n"), filename );
#endif /* DEBUG_PE_LINKER */

   result->linker_file_ops = &_linker_pe_file_ops;

   file = wss_file_open( filename, WSS_FILE_READ | WSS_FILE_ROOTPRIV );
   if( file == WSS_FILE_INVALID )
      return ENOENT;

   result->linker_file_size = wss_file_get_size( file );

   wss_file_seek( file, 0, WSS_FILE_SEEK_SET );
   if( !wss_file_read( file, (pd_byte_t *)dossig, 2 ) )
   {
      wss_file_close( file );
      return ENOEXEC;
   }

   /* Validate the DOS stub */
   if( ( dossig[0] != 'M' ) || ( dossig[1] != 'Z' ) )
   {
      wss_file_close( file );
      return ENOEXEC;
   }

   /* Find signature offset */
   wss_file_seek( file, 60, WSS_FILE_SEEK_SET );
   if( !wss_file_read( file, (pd_byte_t *)&sig, 4 ) )
   {
      wss_file_close( file );
      return ENOEXEC;
   }

   /* Read signature */
   wss_file_seek( file, sig, WSS_FILE_SEEK_SET );
   if( !wss_file_read( file, (pd_byte_t *)&sig, 4 ) )
   {
      wss_file_close( file );
      return ENOEXEC;
   }

   wss_file_close( file );
   if( sig != IMAGE_NT_SIGNATURE )
      return ENOEXEC;

   pedata = (linker_pe_data_t *)wss_malloc( sizeof(linker_pe_data_t) );
   if( pedata == NULL )
      return ENOMEM;

   bzero( pedata, sizeof(*pedata) );
   result->linker_file_data = (void *)pedata;

   pedata->hInstance = LoadLibrary( filename );

   if( pedata->hInstance == NULL )
   {
      wss_free( pedata );
      return ENOEXEC;
   }

   error = _linker_pe_lookup_symbol( result,
                                     PD_MAKESTR(WRLINKNAME_MOD_DEPENDENCIES),
                                     &addr );
   if( error )
   {
#ifdef DEBUG_PE_LINKER
      wsi_printf( PD_MAKESTR("PE Linker: Unable to determine symbol ")
                  PD_MAKESTR("'%s'\n"),
                  PD_MAKESTR(WRLINKNAME_MOD_DEPENDENCIES) );
#endif /* DEBUG_PE_LINKER */
      wss_free( pedata );
      CloseHandle( pedata->hInstance );
      return error;
   }

   result->linker_file_dependencies = (const pd_char_t *)addr.addr;

   return 0;
}

static int
_linker_pe_finalize( wss_linker_file_t *lf )
{
   return 0;
}

static int
_linker_pe_unload( wss_linker_file_t *lf )
{
   if( lf == NULL )
      return EINVAL;

   assert( WSS_REF_COUNT(lf->linker_file_refcount) == 1 );

   if( lf->linker_file_data )
   {
      linker_pe_data_t *pedata = (linker_pe_data_t *)lf->linker_file_data;

      FreeLibrary( pedata->hInstance );
   }

   return wss_linker_file_destroy( lf );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

void EXPORT
linker_pe_init()
{
   wss_linker_register( _linker_pe_description, &_linker_pe_ops );
}

