/*! \file wsi_exec.c
 *
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_exec.c,v 1.16 2001/02/23 19:02:30 stevek Exp $
 *
 * \brief   Handle new program execution wrapper setup
 * \author  Steve Kiernan
 * \version $Revision: 1.16 $
 * \date    $Date: 2001/02/23 19:02:30 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_path.h>
#include <wss_malloc.h>
#include <wss_process.h>
#include <wss_criteria.h>
#include <wss_wrapper.h>
#include <wsi_exec.h>
#include <wsi_file.h>


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_exec */
/*! \brief Handle setup of wrappers for a newly executing program
 *
 *  \param p        Pointer to the process structure
 *  \param cwd      Pointer to the string containing the current working
 *                  directory of the process
 *  \param progPath Path name of the executing program
 */
BOOL EXPORT
wsi_exec( proc_t *p, pd_char_t *cwd, pd_char_t *progPath )
{
   wss_proc_t *wrp;
   BOOL        wr_run_wrapped;
   BOOL        has_wrappers;
   int         len;
   int         i;


   /*
    * Look up the current process in the wrappers proc
    * structure. If it wasn't found, this means there was either
    * a problem, or this is a process that was running before the
    * wrappers module was loaded into the kernel.  In the first
    * case, this is a big problem.  In the second case, we can
    * probably work around it by creating the structure at this
    * time, but until I'm possitive, I'm leaving this as a TBD
    */
   wrp = wss_pfind( p->p_pid );
   if( !wrp )
   {
      wsi_printf( PD_MAKESTR("wsi_exec: can't find pid: p->p_pid %d\n"),
                  (int)p->p_pid );
      wrp = wss_proc_new( p );
   }

   if( wrp->wrp_oldpath )
      wss_free( wrp->wrp_oldpath );

   wrp->wrp_oldpath = wrp->wrp_path; 
   wrp->wrp_path = wss_strdup( progPath );
   WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
   wrp->wrp_cwd = wss_strdup( cwd );
   WSI_RW_EXIT( wrp->wrp_cwd );

   /* Clear out progname if its currently set */
   if( wrp->wrp_progname != NULL )
      wss_free( wrp->wrp_progname );

   /*
    * Generate the program name, full path should have at least
    * one '\' character.  Program name is everything after the
    * last '\'.
    */
   for( len = 0, i = wsi_strlen( wrp->wrp_path );
        ( i > 0 ) && ( wrp->wrp_path[i] != WSS_FILE_SEPARATOR );
        i--, len++ );

   /*
    * Check (for sanity, etc)
    */
   if( ( i == 0 ) && ( wrp->wrp_path[0] != WSS_FILE_SEPARATOR ) )
   {
      wsi_printf( PD_MAKESTR("wsi_execve: Pathname `%s' didn't have a '%c' in it\n"),
                  wrp->wrp_path, WSS_FILE_SEPARATOR );
      wrp->wrp_progname = NULL;
   }
   else
   {
      wrp->wrp_progname = wss_malloc_string( len );
      wsi_strncpy( wrp->wrp_progname, &( wrp->wrp_path[i + 1] ), len + 1 );
   }

   /*
    * We may now check the activation criteria to determine
    * whether or not to run this process wrapped.
    */
   wss_crit_check_deactivation( wrp );
   wr_run_wrapped = wss_crit_check_activation( wrp );

   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   has_wrappers = (wrp->wrp_wi_list == NULL) ? FALSE : TRUE;
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   if( has_wrappers )
   {
   }

   wss_proc_rele( wrp );

   return( wr_run_wrapped );
}

