/*! \file wsi_file.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_file.c,v 1.22 2001/02/23 19:02:30 stevek Exp $
 *
 * \brief   Routines to read files in the filesystem
 * \author  Steve Kiernan
 * \version $Revision: 1.22 $
 * \date    $Date: 2001/02/23 19:02:30 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_mutex.h>

#include <wsi_file.h>
#include <wss_file.h>


/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

struct wsi_file
{
   WSS_MUTEX_PROTECT(f);
   int               f_refcount;
   HANDLE            f_hFile;
   HANDLE            f_hMap;
   PVOID             f_pMapAddr;
};


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_file_open */
/*! \brief Open or create a file in the filesystem
 *
 *  \param filename Pointer to string containing the file name
 *  \param mode     Mode with which to open the file
 *
 *  \return Handle for successfully opened file,
 *  \return \b WSS_FILE_INVALID on failure
 */
wss_file_t EXPORT
wsi_file_open( const pd_char_t *filename, wss_file_mode_t mode )
{
   DWORD            access = ( (mode & WSS_FILE_READ)?GENERIC_READ:0 ) |
                             ( (mode & WSS_FILE_WRITE)?GENERIC_WRITE:0 );
   DWORD            creat  = ( (mode & WSS_FILE_CREAT)?OPEN_ALWAYS:0 );
   HANDLE           hFile  = INVALID_HANDLE_VALUE;
   struct wsi_file *file   = WSS_FILE_INVALID;

   if( !creat )
      creat = OPEN_EXISTING;

   hFile = CreateFile( filename, access, 0, NULL, creat,
                              FILE_ATTRIBUTE_NORMAL, NULL );
   if( ( hFile != INVALID_HANDLE_VALUE ) &&
       ( ( mode & WSS_FILE_TRUNC ) != 0 ) )
   {
      CloseHandle( hFile );
      hFile = CreateFile( filename, access, 0, NULL,
                          creat | TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL,
                          NULL );
   }
   if( hFile != INVALID_HANDLE_VALUE )
   {
      file = (struct wsi_file *)wss_malloc( sizeof(*file) );

      if( file )
      {
         WSS_MUTEX_INIT( file->f );
         file->f_refcount = 1;
         file->f_hFile = hFile;
         file->f_hMap = INVALID_HANDLE_VALUE;
         file->f_pMapAddr = NULL;
      }
      else
      {
         file = WSS_FILE_INVALID;
         CloseHandle( hFile );
      }
   }

   return( (void *)file );
}

/*  wsi_file_free */
/*! \brief Release resources used by file
 *
 *  \param handle Handle to a file
 */
void EXPORT
wsi_file_free( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   assert( file != NULL );
   assert( file->f_refcount == 1 );

   if( file )
   {
      if( file->f_hFile )
         CloseHandle( file->f_hFile );

      WSS_MUTEX_DESTROY( file->f );
      wss_free( file );
   }
}

/*  wsi_file_close */
/*! \brief Close previously-opened file
 *
 *  \param handle Handle to open file
 */
void EXPORT
wsi_file_close( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      if( file->f_hMap != INVALID_HANDLE_VALUE )
         wsi_file_unmap( handle );

      wsi_file_rele( handle );
   }
}

/*  wsi_file_ref */
/*! \brief Acquire reference to file structure
 *
 *  \param handle Handle to a file
 */
void
wsi_file_ref( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file->f_refcount < 1 )
      panic( "wsi_file_ref: reference count for 0x%x was %d < 1", file,
             file->f_refcount );

   WSS_MUTEX_ENTER( file->f );
   ++file->f_refcount;
   WSS_MUTEX_EXIT( file->f );
}

                               
/*  wsi_file_rele */
/*! \brief Release reference to file sturcture
 *
 *  \param handle Handle to a file
 */
void
wsi_file_rele( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file->f_refcount < 1 )
      panic( "wsi_file_rele: reference count for 0x%x was %d < 1", file,
             file->f_refcount );

   WSS_MUTEX_ENTER( file->f );
   if(file->f_refcount > 1)
   {
       --file->f_refcount;
       WSS_MUTEX_EXIT( file->f );
   }
   else
      wsi_file_free(file);
}


/*  wsi_file_read */
/*! \brief Read bytes from opened file into specified buffer
 *
 *  \param handle Handle to a previously-opened file
 *  \param buffer Pointer to buffer which will contain read data
 *  \param len    Number of bytes to read
 *
 *  \return Number of bytes read on success,
 *  \return -1 on failure
 */
int EXPORT
wsi_file_read( wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      DWORD nRead;
      BOOL  bRet;

      WSS_MUTEX_ENTER( file->f );
      bRet = ReadFile( file->f_hFile, (LPVOID)buffer, len, &nRead, NULL );
      WSS_MUTEX_EXIT( file->f );

      return( (bRet)?nRead:-1 );
   }

   SetLastError( ERROR_INVALID_PARAMETER );
   return -1;
}


/*  wsi_file_write */
/*! \brief Write bytes from the specified buffer to opened file
 *
 *  \param handle Handle to a previously-opened file
 *  \param buffer Pointer to buffer containing data to be written
 *  \param len    Number of bytes to write out
 *
 *  \return Number of bytes written to file on success,
 *  \return -1 on failure
 */
int EXPORT
wsi_file_write( wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      DWORD nWritten;
      BOOL  bRet;

      WSS_MUTEX_ENTER( file->f );
      bRet = WriteFile( file->f_hFile, (LPCVOID)buffer, len, &nWritten, NULL );
      WSS_MUTEX_EXIT( file->f );

      return( (bRet)?nWritten:-1 );
   }

   SetLastError( ERROR_INVALID_PARAMETER );
   return -1;
}

/*  wsi_file_seek */
/*! \brief Repositions the file offset pointer
 *
 *  This is equivalent to the C library function lseek(2)
 *
 *  \param handle Handle of open file
 *  \param offset Offset to seek within file
 *  \param whence Where to start seeking from (beginning, current position,
 *                or end of file)
 *
 *  \return 0 on success,
 *  \return EINVAL if an invalid argument is given
 */
int EXPORT
wsi_file_seek( wss_file_t handle, pd_offset_t offset, int whence )
{
   struct wsi_file *file  = (struct wsi_file *)handle;
   int              error = EINVAL;

   if( file )
   {
      DWORD winnt_whence;
      switch( whence )
      {
         case WSS_FILE_SEEK_END:
            winnt_whence = FILE_END;
            break;
         case WSS_FILE_SEEK_SET:
            winnt_whence = FILE_BEGIN;
            break;
         case WSS_FILE_SEEK_CUR:
            winnt_whence = FILE_CURRENT;
            break;
         default:
            return EINVAL;
      }

      WSS_MUTEX_ENTER( file->f );

      if( SetFilePointer( file->f_hFile, offset, NULL, winnt_whence ) != 0xffffffff )
         error = 0;

      WSS_MUTEX_EXIT( file->f );
   }

   return error;
}


/*  wsi_file_get_size */
/*! \brief Get size of the open file
 *
 *  \param handle Handle of open file
 *  \param error  If non-NULL, pointer to memory where error code should be
 *                returned
 *
 *  \return Size of file on success,
 *  \return 0 on failure
 */
size_t EXPORT
wsi_file_get_size( wss_file_t handle, int *error )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      DWORD dwSize;

      WSS_MUTEX_ENTER( file->f );

      if( ( dwSize = GetFileSize( file->f_hFile, NULL ) ) == 0xffffffff )
      {
         if( error )
            *error = (int)GetLastError();
         dwSize = 0;
      }

      WSS_MUTEX_EXIT( file->f );
      return dwSize;
   }

   SetLastError( ERROR_INVALID_PARAMETER );
   if( error )
      *error = ERROR_INVALID_PARAMETER;
   return 0;
}


/*  wsi_file_get_map_addr */
/*! \brief Get memory-mapped address of mapped file
 *
 *  \param handle Handle of open file
 *
 *  \return Pointer to beginning of mapped view of file on success,
 *  \return NULL otherwise
 */
void EXPORT *
wsi_file_get_map_addr( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;
   PVOID            addr = NULL;

   if( file )
   {
      WSS_MUTEX_ENTER( file->f );
      addr = file->f_pMapAddr;
      WSS_MUTEX_EXIT( file->f );
   }

   return addr;
}


/*  wsi_file_map */
/*! \brief Map opened file into memory space
 *
 *  \param handle  Handle of open file
 *  \param size    Number of bytes to map in (some systems will round-up
 *                 to a page boundary)
 *  \param mapmode Mode to map file view
 *  \param pos     Offset into file to begin mapping
 *
 *  \return 0 on success,
 *  \return error code on failure
 */
int EXPORT
wsi_file_map( wss_file_t handle, size_t size, int mapmode, pd_offset_t pos )
{
   struct wsi_file *file  = (struct wsi_file *)handle;
   int              error = EINVAL;

   if( file )
   {
      DWORD prot = 0;

      WSS_MUTEX_ENTER( file->f );
      if( mapmode & WSS_FILE_MAP_WRITE )
         prot = PAGE_READWRITE;
      else if( mapmode & WSS_FILE_MAP_READ )
         prot = PAGE_READONLY;

      prot |= SEC_COMMIT;

      file->f_hMap = CreateFileMapping( file->f_hFile, NULL, prot, 0, size, NULL );
      if( file->f_hMap != INVALID_HANDLE_VALUE )
      {
         DWORD access = 0;

         if( ( mapmode & WSS_FILE_MAP_ALL ) == WSS_FILE_MAP_ALL )
            access = FILE_MAP_ALL_ACCESS;
         else if( ( mapmode & WSS_FILE_MAP_READ ) == WSS_FILE_MAP_READ )
            access = FILE_MAP_READ;
         else if( ( mapmode & WSS_FILE_MAP_WRITE ) == WSS_FILE_MAP_WRITE )
            access = FILE_MAP_WRITE;

         file->f_pMapAddr = MapViewOfFile( file->f_hMap, access, 0, pos, size );
         if( file->f_pMapAddr == NULL )
         {
            CloseHandle( file->f_hMap );
            file->f_hMap = INVALID_HANDLE_VALUE;
         }
         else
            error = 0;
      }

      WSS_MUTEX_EXIT( file->f );
   }

   return error;
}

/*  wsi_file_unmap */
/*! \brief Unmaps a file previously mapped into the WSS memory space
 *
 *  \param handle Handle to mapped file
 */
void EXPORT
wsi_file_unmap( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      WSS_MUTEX_ENTER( file->f );

      if( file->f_pMapAddr != NULL )
      {
         UnmapViewOfFile( file->f_pMapAddr );
         file->f_pMapAddr = NULL;
      }

      if( file->f_hMap != INVALID_HANDLE_VALUE )
      {
         CloseHandle( file->f_hMap );
         file->f_hMap = INVALID_HANDLE_VALUE;
      }

      WSS_MUTEX_EXIT( file->f );
   }
}

