/*! \file wsi_init.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_init.c,v 1.12 2001/02/23 19:02:30 stevek Exp $
 *
 * \brief   Read in WSS configuration file in system-dependent fashion
 * \author  Steve Kiernan
 * \version $Revision: 1.12 $
 * \date    $Date: 2001/02/23 19:02:30 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_debug.h>
#include <wss_file.h>
#include <wss_init.h>


/************************************************************************\
 * Private globals                                                      *
\************************************************************************/

wss_init_entry_t wsi_init_locals[] =
{
   { NULL, 0, NULL }
};


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static int
_wsi_init_process_entry( wss_init_entry_t *entry, DWORD dwType,
                         pd_byte_t *data )
{
   pd_char_t *value = (pd_char_t *)data;

   switch( entry->type )
   {
      case WSS_INIT_TYPE_STRING:
      {
         if( dwType == REG_SZ )
         {
            pd_char_t **str = (pd_char_t **)entry->storage;

            if( *str )
               wss_free( *str );
            *str = wss_strdup( value );
#ifdef DEBUG_INIT
            wsi_debug_log( PD_MAKESTR("Adding string entry '%s': %s\n"),
                           entry->key, value );
#endif
            return TRUE;
         }
         break;
      }

      case WSS_INIT_TYPE_INT:
      {
         if( dwType == REG_SZ )
         {
            int val;

            for( val = 0;
                 *value && ( *value >= '0' ) && ( *value <= '9' );
                 value++ )
            {
               val *= 10;
               val += ( *value - '0' );
            }

            *((int *)entry->storage) = val;
#ifdef DEBUG_INIT
            wss_debug_log( PD_MAKESTR("Adding int entry '%s': %s\n"),
                           entry->key, value );
#endif
            return TRUE;
         }
         if( dwType == REG_DWORD )
         {
            DWORD val = *(DWORD *)data;

            *((int *)entry->storage) = val;
#ifdef DEBUG_INIT
            wss_debug_log( PD_MAKESTR("Adding int entry '%s': %d\n"),
                           entry->key, val );
#endif
            return TRUE;
         }
         break;
      }

      case WSS_INIT_TYPE_BOOL:
         if( dwType == REG_SZ )
         {
            if( !wsi_strcmp( value, PD_MAKESTR("TRUE") ) ||
                !wsi_strcmp( value, PD_MAKESTR("true") ) ||
                !wsi_strcmp( value, PD_MAKESTR("YES") ) ||
                !wsi_strcmp( value, PD_MAKESTR("yes") ) )
            {
               *((int *)entry->storage) = 1;
#ifdef DEBUG_INIT
               wss_debug_log( PD_MAKESTR("Adding bool entry '%s': %s\n"),
                              entry->key, value );
#endif
               return TRUE;
            }

            if( !wsi_strcmp( value, PD_MAKESTR("FALSE") ) ||
                !wsi_strcmp( value, PD_MAKESTR("false") ) ||
                !wsi_strcmp( value, PD_MAKESTR("NO") ) ||
                !wsi_strcmp( value, PD_MAKESTR("no") ) )
            {
               *((int *)entry->storage) = 0;
#ifdef DEBUG_INIT
               wss_debug_log( PD_MAKESTR("Adding bool entry '%s': %s\n"),
                              entry->key, value );
#endif
               return TRUE;
            }
         }
         if( dwType == REG_DWORD )
         {
            DWORD bit = *(DWORD *)data;

            if( bit )
            {
               *((int *)entry->storage) = 1;
#ifdef DEBUG_INIT
               wss_debug_log( PD_MAKESTR("Adding bool entry '%s': TRUE\n"),
                              entry->key );
#endif
               return TRUE;
            }
            else
            {
               *((int *)entry->storage) = 0;
#ifdef DEBUG_INIT
               wss_debug_log( PD_MAKESTR("Adding bool entry '%s': FALSE\n"),
                              entry->key );
#endif
               return TRUE;
            }
         }
         break;

      default:
         break;
   }

   return FALSE;
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_init_read_conf */
/*! \brief Read configuration file from filesystem
 */
void EXPORT
wsi_init_read_conf( void )
{

   HKEY hWrappers;

#ifdef DEBUG_INIT
   wss_debug_log( PD_MAKESTR("wsi_init_read_conf: Opening key '%s'\n"),
                  PD_MAKESTR(WSS_REG_KEY) );
#endif /* DEBUG_INIT */

   if( RegOpenKey( HKEY_LOCAL_MACHINE, PD_MAKESTR(WSS_REG_KEY),
                   &hWrappers ) == ERROR_SUCCESS )
   {
      pd_char_t keyName[MAX_PATH];
      pd_byte_t valStr[MAX_PATH];
      DWORD      i        = 0;
      DWORD      dwSize   = MAX_PATH;
      DWORD      cbData   = MAX_PATH;
      DWORD      dwType;

#ifdef DEBUG_INIT
      wss_debug_log( PD_MAKESTR("wsi_init_read_conf: Enumerating...\n") );
#endif /* DEBUG_INIT */

      while( RegEnumValue( hWrappers, i, keyName, &dwSize, NULL,
                           &dwType, valStr, &cbData ) == ERROR_SUCCESS )
      {
         wss_init_entry_t *init_entry;

#ifdef DEBUG_INIT
         wss_debug_log( PD_MAKESTR("wsi_init_read_conf: Checking '%s'...\n"),
                        keyName );
#endif /* DEBUG_INIT */

         /* Search for entries in configuration table(s) */
         for( init_entry = wss_init_globals;
              ( init_entry->key != NULL );
              init_entry++ )
         {
#ifdef DEBUG_INIT
            wss_debug_log( PD_MAKESTR("wsi_init_read_conf: '%s' == '%s'\n"),
                           keyName, init_entry->key );
#endif /* DEBUG_INIT */
            if( !_tcsicmp( init_entry->key, keyName ) &&
                _wsi_init_process_entry( init_entry, dwType, valStr ) )
               goto foundit;
         }

         for( init_entry = wsi_init_locals;
              ( init_entry->key != NULL );
              init_entry++ )
         {
#ifdef DEBUG_INIT
            wss_debug_log( PD_MAKESTR("wsi_init_read_conf: '%s' == '%s'\n"),
                           keyName, init_entry->key );
#endif /* DEBUG_INIT */
            if( !_tcsicmp( init_entry->key, keyName ) &&
                _wsi_init_process_entry( init_entry, dwType, valStr ) )
               goto foundit;
         }

foundit:
         i++;
         dwSize = MAX_PATH;
         cbData = MAX_PATH;
      }

      RegCloseKey( hWrappers );
   }
}

