/*! \file wsi_malloc.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_malloc.c,v 1.13 2001/02/23 19:02:30 stevek Exp $
 *
 * \brief   Platform specific part of the malloc routines.
 * \author  Douglas Kilpatrick, Steve Kiernan
 * \version $Revision: 1.13 $
 * \date    $Date: 2001/02/23 19:02:30 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

void *wsi_proc_malloc __P((struct proc *p, int size));
void  wsi_proc_free   __P((struct proc *p, void *addr, int size));


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_malloc */
/*! \brief Allocate memory of at least specified size
 *
 *  \param size Number of bytes to allocate
 *
 *  \return Pointer to block of memory on success,
 *  \return NULL on allocation failure
 */
void EXPORT *
wsi_malloc( size_t size )
{
   return malloc( size );
}

/*  wsi_free */
/*! \brief Free allocated memory
 *
 *  \param ptr Pointer to allocated memory
 */
void EXPORT
wsi_free( void *ptr )
{
   free( ptr );
}


/*  wsi_curproc_malloc */
/*! \brief Allocates memory in the process space of the current process
 *
 *  This gets used to temporarily store replaced syscall parameters.
 *  The size and the address of the memory allocated here will be passed
 *  into free, so the allocation mechanism may be simplified.
 *
 *  \param size Number of bytes to allocate
 *
 *  \return Pointer to block of memory on success,
 *  \return NULL on allocation failure 
 */
void EXPORT *
wsi_curproc_malloc( size_t size )
{
   return malloc( size );
}


/*  wsi_curproc_malloc_string */
/*! \brief Allocates enough memory to contain specified number of characters
 *
 *  This function handles allocating memory for Unicode characters on
 *  Unicode-enabled systems.
 *
 *  \param nchars Number of characters to allocate memory for
 *
 *  \return Pointer to block of memory on success,
 *  \return NULL on allocation failure
 */
pd_char_t EXPORT *
wsi_curproc_malloc_string( size_t nchars )
{
  return (pd_char_t *) malloc( nchars * sizeof( pd_char_t ));
}

/*  wsi_curproc_free */
/*! \brief Frees the address from the address space of the current process
 *
 *  \param addr Pointer to allocated memory
 *  \param size Number of bytes allocated
 */
void EXPORT
wsi_curproc_free( void *addr, size_t size )
{
  free( addr );
}

/*  wsi_curproc_strdup */
/*! \brief Similar to strdup, but allocates memory from the address space of
 *         the current process
 *
 *  \param str Pointer to string to duplicate
 *
 *  \return Pointer to copy of string referenced by \a str on success,
 *  \return NULL on allocation failure
 */
pd_char_t EXPORT *
wsi_curproc_strdup( const pd_char_t *str )
{
   return _tcsdup( str );
}

