/*! \file wsi_rw.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_rw.c,v 1.8 2001/02/23 19:02:30 stevek Exp $
 *
 * \brief   Implementation of reader/writer locks on NT
 * \author  Steve Kiernan
 * \version $Revision: 1.8 $
 * \date    $Date: 2001/02/23 19:02:30 $
 */


#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wsi_rw.h>


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_rw_init */
/*! \brief Initialize a reader/writer lock
 *
 *  \param rw Pointer to reader/writer lock
 */
void EXPORT
wsi_rw_init( wsi_rw_t *rw )
{
   assert( rw != NULL );

   rw->rw_active = 0;
   rw->rw_reads = 0;
   rw->rw_writes = 0;
   WSS_MUTEX_INIT( rw->rw );
   WSS_CV_INIT( rw->rw_read );
   WSS_CV_INIT( rw->rw_write );
}


/*  wsi_rw_destroy */
/*! \brief Destroy a reader/writer lock
 *
 *  \param rw Pointer to reader/writer lock
 */
void EXPORT
wsi_rw_destroy( wsi_rw_t *rw )
{
   assert( rw != NULL );
   assert( rw->rw_active == 0 );

   WSS_MUTEX_DESTROY( rw->rw );
   WSS_CV_DESTROY( rw->rw_read );
   WSS_CV_DESTROY( rw->rw_write );
}

/*  wsi_rw_enter */
/*! \brief Acquire a reader/writer lock
 *
 *  \param rw   Pointer to reader/writer lock
 *  \param type Type of lock to acquire (\b RW_READER or \b RW_WRITER)
 */
void EXPORT
wsi_rw_enter( wsi_rw_t *rw, wsi_rw_type_t type )
{
   assert( rw != NULL );
   assert( ( type == RW_READER ) || ( type == RW_WRITER ) );

   WSS_MUTEX_ENTER( rw->rw );
   switch( type )
   {
      case RW_READER:
         rw->rw_reads++;

         /* Prevent writer starvation */
         if( rw->rw_writes > 0 )
#ifdef DEBUG_MUTEX
            pd_cv_wait( rw->rw_read_cv, &(rw->rw_lock),
                        PD_MAKESTR(__FILE__), __LINE__ );
#else
            pd_cv_wait( rw->rw_read_cv, rw->rw_lock );
#endif /* DEBUG_MUTEX */
         /* Someone has writer lock */
         while( rw->rw_active < 0 )
#ifdef DEBUG_MUTEX
            pd_cv_wait( rw->rw_read_cv, &(rw->rw_lock),
                        PD_MAKESTR(__FILE__), __LINE__ );
#else
            pd_cv_wait( rw->rw_read_cv, rw->rw_lock );
#endif /* DEBUG_MUTEX */

         rw->rw_active++;
         rw->rw_reads--;

         break;

      case RW_WRITER:
         rw->rw_writes++;

         while( rw->rw_active )
#ifdef DEBUG_MUTEX
            pd_cv_wait( rw->rw_read_cv, &(rw->rw_lock),
                        PD_MAKESTR(__FILE__), __LINE__ );
#else
            pd_cv_wait( rw->rw_read_cv, rw->rw_lock );
#endif /* DEBUG_MUTEX */

         rw->rw_writes--;
         rw->rw_active = -1;

         break;
   }

   WSS_MUTEX_EXIT( rw->rw );
}

/*  wsi_rw_exit */
/*! \brief Release hold on reader/writer lock
 *
 *  \param rw Pointer to reader/writer lock
 */
void EXPORT
wsi_rw_exit( wsi_rw_t *rw )
{
   assert( rw != NULL );

   WSS_MUTEX_ENTER( rw->rw );

   if( rw->rw_active < 0 )
   {
      BOOL wake;

      /* Writer lock */
      rw->rw_active = 0;
      wake = (rw->rw_reads == 0) ? FALSE : TRUE;
      WSS_MUTEX_EXIT( rw->rw );

      if( wake )
         WSS_CV_BROADCAST( rw->rw_read );
      else
         WSS_CV_SIGNAL( rw->rw_write );
   }
   else if( rw->rw_active > 0 )
   {
      rw->rw_active--;
      if( rw->rw_active == 0 )
      {
         WSS_MUTEX_EXIT( rw->rw );
         WSS_CV_SIGNAL( rw->rw_write );
      }
      else
      {
         WSS_MUTEX_EXIT( rw->rw );
      }
   }
   else
      panic( "wsi_rw_exit: rwlock 0x%x not obtained by thread 0x%x\n",
             rw, GetCurrentThread() );
}

/*  wsi_rw_downgrade */
/*! \brief Downgrade a writer lock to a reader lock
 *
 *  \param rw Pointer to reader/writer lock
 */
void EXPORT
wsi_rw_downgrade( wsi_rw_t *rw )
{
   BOOL wake;

   WSS_MUTEX_ENTER( rw->rw );
   rw->rw_active = 1;
   wake = (rw->rw_reads == 0) ? FALSE : TRUE;
   WSS_MUTEX_EXIT( rw->rw );

   if( wake )
      WSS_CV_BROADCAST( rw->rw_read );
}

