/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_utilities.c,v 1.27 2001/02/23 19:02:30 stevek Exp $
 * 
 * Author: Doug Kilpatrick and others.
 * 
 * Purpose: Repository of usefull functions that might be missing.
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#ifdef PD_IN_KERNEL
#include <wss_malloc.h>
#endif /* PD_IN_KERNEL */

#include <stdio.h>

#define MAXBUFLEN 1024


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/*! Pointer to printf function to be used by wsi_printf() and wsi_vprintf() */
wsi_printf_func EXPORT wsi_printf_fn = NULL;


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static int
_wsi_default_printf( const pd_char_t *format, va_list ap )
{
   TCHAR msg[MAXBUFLEN];
   int nchars;

   nchars = _vsntprintf( msg, MAXBUFLEN, format, ap );

#ifdef DEBUG_PRINTF
   MessageBox( 0, msg, PD_MAKESTR("Wrapper Message"),
               ( MB_OK | MB_ICONEXCLAMATION | MB_DEFAULT_DESKTOP_ONLY ) );
#endif

   return nchars;
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  bcopy */
/*! \brief Copy specified number of bytes from one buffer to another
 *
 *  \param src  Pointer to source buffer
 *  \param dest Pointer to destination buffer
 *  \param len  Number of bytes to copy
 */
void EXPORT
bcopy( const void *src, void *dest, size_t len )
{
   memcpy( dest, src, len );
}

/*  bzero */
/*! \brief Write specified number of zero-value bytes to a buffer
 *
 *  \param b   Buffer to clear
 *  \param len Number of bytes to clear
 */
void EXPORT
bzero( void *b, size_t len )
{
   (void) memset( b, 0, len );
}

/*  copyin */
/*! \brief Copy specified number of bytes from user space to kernel space
 *
 *  Note: this is just a memory copy on the Windows NT port of wrappers,
 *        since we are in user space always
 *
 *  \param uaddr Pointer to user space buffer
 *  \param kaddr Pointer to kernel space buffer
 *  \param len   Number of bytes to copy
 *
 *  \return 0 always
 */
int EXPORT
copyin( void *uaddr, void *kaddr, size_t len )
{
   memcpy( kaddr, uaddr, len );

   return 0;
}

/*  copyout */
/*! \brief Copy specified number of bytes from kernel space to user space
 *
 *  Note: this is just a memory copy on the Windows NT port of wrappers,
 *        since we are in user space always
 *
 *  \param kaddr Pointer to kernel space buffer
 *  \param uaddr Pointer to user space buffer
 *  \param len   Number of bytes to copy
 *
 *  \return 0 always
 */
int EXPORT
copyout( void *kaddr, void *uaddr, size_t len )
{
   memcpy( uaddr, kaddr, len );

   return 0;
}

/*  copyinstr */
/*! \brief Copy a NULL-terminated string, at most the specified number of
 *         bytes from user space to kernel space
 *
 *  Note: this is just a memory copy on the Windows NT port of wrappers,
 *        since we are in user space always
 *
 *  \param uaddr Pointer to user space buffer
 *  \param kaddr Pointer to kernel space buffer
 *  \param len   Number of bytes to copy
 *  \param done  Pointer to memory in which to store number of bytes
 *               actually copied
 *
 *  \return 0 always
 */
int EXPORT
copyinstr( const void *uaddr, void *kaddr, size_t len, size_t *done )
{
   pd_char_t *result;
   result = _tcsncpy( (pd_char_t *)kaddr, (const pd_char_t *)uaddr, len );
   if( result && done )
      *done = (size_t)( ( _tcslen(result) + 1 ) * sizeof(pd_char_t) );
   return 0;
}

/*  wsi_vprintf */
/*! \brief Version of vprint which uses a user-defined printf function
 *
 *  \param format printf-style format string
 *  \param ap     Variable-length argument list
 *
 *  \return Number of characters printed
 */
int EXPORT
wsi_vprintf( const pd_char_t *format, va_list ap )
{
   int retval;

   if( wsi_printf_fn )
      retval = (*wsi_printf_fn)( format, ap );
   else
      retval = _wsi_default_printf( format, ap );

   return retval;
}

/*  wsi_printf */
/*! \brief Version of print which uses a user-defined printf function
 *
 *  \param format printf-style format string
 *
 *  \return Number of characters printed
 */
int EXPORT
wsi_printf( const pd_char_t *format, ... )
{
   va_list ap;
   int     retval;

   va_start( ap, format );
   retval = wsi_vprintf( format, ap );
   va_end( ap );

   return retval;
}


/*  wsi_toupper */
/*! \brief Implementation of toupper
 *
 *  \param c Character to convert to uppercase
 *
 *  \return Character converted to uppercase
 */
pd_char_t EXPORT
wsi_toupper( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('a') ) && ( c <= PD_MAKESTR('z') ) )?
            c - PD_MAKESTR('a') + PD_MAKESTR('A') :
            c;
}

/*  wsi_tolower */
/*! \brief Implementation of tolower
 *
 *  \param c Character to convert to lowercase
 *
 *  \return Character converted to lowercase
 */
pd_char_t EXPORT
wsi_tolower( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('A') ) && ( c <= PD_MAKESTR('Z') ) )?
            c - PD_MAKESTR('A') + PD_MAKESTR('a') :
            c;
}

#ifdef PD_IN_KERNEL
/*  wsi_strunicode */
/*! \brief Convert an ASCII string to Unicode
 *
 *  \param str Pointer to ASCII string to convert
 *
 *  \return Pointer to Unicode version of string
 */
pd_char_t EXPORT *
wsi_strunicode( const char *astr, int *size )
{
#ifdef UNICODE
   pd_char_t *ustr = NULL;

   if( astr )
   {
      DWORD asize = strlen( astr ) + 1;
      DWORD usize = asize * sizeof(pd_char_t);

      ustr = (pd_char_t *)wss_malloc( usize );

      MultiByteToWideChar( CP_ACP, 0, astr, asize, ustr, usize );

      if( size )
         *size = _tcslen( ustr );
   }
   else if( size )
      *size = 0;

   return ustr;
#else
   if( size ) *size = ( astr == NULL ) ? 0 : _tcslen( astr );
   return ( astr == NULL ) ? NULL : wss_strdup( astr );
#endif
}

/*  wsi_strascii */
/*! \brief Convert a Unicode string to ASCII
 *
 *  \param str Pointer to Unicode string to convert
 *
 *  \return Pointer to ASCII version of string
 */
char EXPORT *
wsi_strascii( const pd_char_t *ustr, int *size )
{
#ifdef UNICODE
   char *astr = NULL;

   if( ustr )
   {
      DWORD allocsz = _tcslen( ustr ) + 1;
      astr = (char *)wss_malloc( allocsz );

      WideCharToMultiByte( CP_ACP, 0, ustr, allocsz, astr, allocsz, NULL,
                           NULL );

      if( size )
         *size = strlen( astr );
   }
   else if( size )
      *size = 0;

   return astr;
#else
   if( size ) *size = ( ustr == NULL ) ? 0 : _tcslen( ustr );
   return ( ustr == NULL ) ? NULL : wss_strdup( ustr );
#endif
}
#endif /* PD_IN_KERNEL */
