/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: loader.c,v 1.34 2001/02/23 19:02:34 stevek Exp $
 *
 * Author: Chris Vance, Steve Kiernan
 *
 * Purpose: Load/unload the WSS from the kernel.
 * 
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/sysproto.h>
#include <sys/exec.h>
#include <sys/sysent.h>
#include <sys/module.h>
#include <sys/linker.h>
#include <sys/syscall.h>

#if defined(KERNEL) || defined(_KERNEL)
#include <sys/kernel.h>
#include <sys/sysctl.h>
#endif /* KERNEL || _KERNEL */

#include <wss_init.h>
#include <wss_log.h>
#include <wss_module.h>
#include <wss_vm.h>
#include <wsi_syscalls.h>


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

/* KLD */
static int wsi_mod_event       __P(( module_t mod, int type, void *data ));
void       wsi_unload_deferred __P(( void ));


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/* KLD */
static moduledata_t wsi_mod =
{
   "wss",
   wsi_mod_event,
   NULL,
};
DECLARE_MODULE( wss, wsi_mod, SI_SUB_KTHREAD_IDLE, SI_ORDER_MIDDLE );

int wsi_version = 1;
int wsi_unloading = 0;


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

/*
 * The next set of defines assumes a standard naming scheme for the
 * WSS syscall replacement functions.
 */

/*
 * Prototype the replacement syscall with name "name."
 */
#define WR_DEFINE_REPLACEMENT(name, numargs)          \
   struct sysent orig_ ## name ##_sysent;          \
   extern int wsi_ ## name  __P((struct proc *p, void *uap)); \
   static struct sysent new_ ## name ## _sysent = {numargs, wsi_ ## name }

/*
 * Prototype the new syscall with name "name."
 */
#define WR_DEFINE_NEW(name, numargs)            \
   struct sysent orig_ ## name ##_sysent;       \
   static struct sysent new_ ## name ## _sysent = {numargs, \
      (sy_call_t *)name }

/*
 * Save the original syscall and copy in the entry for the replacement.
 */
#define WR_SYS_REPLACE(name)                 \
   bcopy(&sysent[SYS_ ##name ], &orig_ ## name ## _sysent,     \
         sizeof(struct sysent));             \
   bcopy(&new_ ## name ## _sysent, &sysent[SYS_ ## name],      \
         sizeof(struct sysent))
/*
 * Save the original syscall and copy in the new syscall.
 */
#define WR_SYS_ADD(name)                  \
   bcopy(&sysent[SYS_ ##name ], &orig_ ## name ## _sysent,  \
         sizeof(struct sysent));             \
   bcopy(&new_ ## name ## _sysent, &sysent[SYS_ ## name],      \
         sizeof(struct sysent))

/*
 * Put the original syscall entry back in the table.
 */
#define WR_SYS_FIX(name)                  \
   bcopy(&orig_ ## name ## _sysent, &sysent[SYS_ ## name],  \
         sizeof(struct sysent))
/*
 * Put the original entry back.  It _should_ be lkmnosys.
 */
#define WR_SYS_REMOVE(name)                  \
   bcopy(&orig_ ## name ## _sysent, &sysent[SYS_ ## name],  \
         sizeof(struct sysent))


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

/*
 * Syscalls replaced by the WSS.
 */
WR_DEFINE_REPLACEMENT(execve,3);
WR_DEFINE_REPLACEMENT(chdir,1);
WR_DEFINE_REPLACEMENT(fchdir,1);
WR_DEFINE_REPLACEMENT(rename,2);

/*
 * Syscalls added by the WSS.
 */
WR_DEFINE_NEW(wrctl,3);
WR_DEFINE_NEW(wrv,4);


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * This is the lkm load function.  It initializes the WSS, replaces
 * some system calls, adds an exit function, adds a new system call,
 * and prepares for the loading of wrappers (last one is still TBD).
 */
int EXPORT
wsi_load( module_t mod )
{
   int error;

   if( ( error = wss_init( WSS_INIT_DB | WSS_INIT_CONF ) ) )
      return error;

   /* Store the old syscall entry, and replace it with the new one */
   WR_SYS_REPLACE(execve);
   WR_SYS_REPLACE(chdir);
   WR_SYS_REPLACE(fchdir);
   WR_SYS_REPLACE(rename);

   /*
    * Add the new syscall for module status.  TBD: should
    * probably check to make sure we aren't blindly overwriting a
    * previously added syscall.
    */
   WR_SYS_ADD(wrctl);
   WR_SYS_ADD(wrv);

   (void)at_exit( wss_proc_exit );
   (void)at_fork( wss_proc_fork );

   /*
    * If we make it to here, print message on console.
    */
   printf( "\nSuccessfully loaded WSS module\n" );

   return 0;
}

/*
 * This is the module unload function.  It should clean up anything
 * that was done as a result of the module load and execution.  Right
 * now it cleans up our wrapper-specific process data structure,
 * terminates loggin, and puts back the correct syscall dispatch
 * entries.
 */
int EXPORT
wsi_unload()
{
   if( wsi_unloading == 2 )
   {
      printf( "\nSuccessfully unloaded WSS module!\n" );
      return(0);
   }

   /* Flag that we're unloading the LKM */
   wsi_unloading = 1;

   /* Clean up all the data structures that have been created */
   wss_proc_shutdown();

   if( wss_proc_num_in_syscall == 0 )
   {
      wss_crit_shutdown();
      wss_mod_unload_all();
      wss_wrapper_shutdown(); /* Cleanup the wrapper module subsystem */
      (void)rm_at_exit(wss_proc_exit); /* Remove our exit function */
   }
   (void)rm_at_fork(wss_proc_fork); /* Remove our fork function */

   if( wss_proc_num_in_syscall == 0 )
   {
      wss_vm_shutdown();
   }

   /* We now put the original system call back in place */
   WR_SYS_FIX(execve);
   WR_SYS_FIX(chdir);
   WR_SYS_FIX(fchdir);
   WR_SYS_FIX(rename);

   if( wss_proc_num_in_syscall == 0 )
   {
      wss_log_shutdown();
   }

   /* Remove the syscalls that were added */
   WR_SYS_REMOVE(wrctl);
   WR_SYS_REMOVE(wrv);

   /* if we make it to here, print message on console*/
   if( wss_proc_num_in_syscall == 0 )
      printf( "\nSuccessfully unloaded WSS module!\n" );
   else
      printf( "\nDeferring unload for wrapped procs\n" );

   return( wss_proc_num_in_syscall );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*
 * This is the WSI shutdown function.
 */
int
wsi_shutdown()
{
   /*
    * Okay, let's do it... shutdown time!
    */
   wsi_printf( PD_MAKESTR("wsi_shutdown: WSS shutting down.\n") );
   wsi_printf( PD_MAKESTR("[Note: It is normal to see a single process which has orphaned syscalls in the\n") );
   wsi_printf( PD_MAKESTR("       following messages, it is the process which performed the shutdown\n") );
   wsi_printf( PD_MAKESTR("       command.]\n") );

   /*
    * Replace our syscall with the original one
    */
   WR_SYS_FIX( execve );
   WR_SYS_FIX( chdir );
   WR_SYS_FIX( fchdir );
   WR_SYS_FIX( rename );

   /*
    * Unhook the syscalls that were added
    */
   WR_SYS_REMOVE( wrctl );
   WR_SYS_REMOVE( wrv );

   wss_proc_shutdown();

   /*
    * Unload all modules
    */
   wss_crit_shutdown();
   wss_mod_unload_all();
   wss_wrapper_shutdown();
   (void) rm_at_exit( wss_proc_exit );
   (void) rm_at_fork( wss_proc_fork );

   /*
    * Shutdown mapped vm space
    */
   wss_vm_shutdown();

   /*
    * Close the log file
    */
   wss_log_shutdown();

   return 0;
}

/*
 * This is called when an unload is deferred and the last wrapped
 * process has exited its wrapped syscall
 */
void
wsi_unload_deferred()
{
   /* If there are no more processes in system calls, finish unload prep */
   if( wss_proc_num_in_syscall == 0 )
   {
      wss_crit_shutdown();
      wss_mod_unload_all();
      wss_wrapper_shutdown(); /* Cleanup the wrapper module subsystem */
      (void)rm_at_exit(wss_proc_exit); /* Remove our exit function */
      wss_vm_shutdown();
      wss_log_shutdown();
      wsi_unloading = 2;

      printf( "\nWSS ready for unload, run modunload once more!\n" );
   }
   else
      printf( "\nDeferred unload called with processes pending!\n" );
}

/*
 * This is the main routine (KLD version)
 */
static int
wsi_mod_event( module_t mod, int type, void *data )
{
   switch( type )
   {
      case MOD_LOAD:
         return wsi_load( mod );
      case MOD_UNLOAD:
         return wsi_unload();
      default:
         break;
   }
   return 0;
}

