/*! \file wsi_dir.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_dir.c,v 1.23 2001/06/11 17:26:06 stevek Exp $
 *
 * \brief   Replacement system calls for chdir, fchdir, and rename
 * \author  Chris Vance, Doug Kilpatrick, Steve Kiernan
 * \version $Revision: 1.23 $
 * \date    $Date: 2001/06/11 17:26:06 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/sysent.h>

#include <wss_mutex.h>

#include <wss_malloc.h>
#include <wss_path.h>
#include <wss_strings.h>
#include <wss_process.h>

#include <sys/vnode.h>
#include <sys/proc.h>

#include <sys/sysproto.h>
#include <sys/filedesc.h>
#include <sys/proc.h>
#include <sys/namei.h>
#include <sys/stat.h>
#include <sys/malloc.h>
#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>
#include <dirent.h>



/***************************************************************************\
 * Definitions for the replacement system calls                            *
\***************************************************************************/

extern struct sysent orig_chdir_sysent;
extern struct sysent orig_fchdir_sysent;
extern struct sysent orig_rename_sysent;


/***************************************************************************\
 * Private structures and types                                            *
\***************************************************************************/

struct rename_uap
{
   char *from;
   char *to;
};

struct chdir_uap
{
   char *path;
};

struct fchdir_uap
{
   int fd;
};


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_rename */
/*! \brief Replacement syscall for rename to handle CWD caching
 *
 *  \param p      Pointer to process structure
 *  \param uap    Pointer to arguments structure
 *  \param retval Pointer to storage for return value
 *
 *  \sa rename(2) man page for return values
 */
int
wsi_rename( proc_t *p, register struct rename_uap *uap, int *retval )
{
   int         error;
   char       *from  = NULL;
   wss_proc_t *wrp;

   /*
    * If the source isn't a directory, we don't care about it at all.
    */
   if( wsi_path_is_dir( p, uap->from ) == NULL )
      return( ( *(sy_call_t *)orig_rename_sysent.sy_call )( p, uap ) );

   /*
    * Canonicalize the incoming pathnames.  Allocated memory!.
    */
   WSS_MUTEX_ENTER( wss_proc_list );
   wrp = wss_proc_list.lh_first;
   WSS_MUTEX_EXIT( wss_proc_list );

   if( wrp )
      from = wsi_path_canonicalize( p, uap->from, FALSE );
       
   /*
    * Call the original rename syscall.
    */
   error = ( *(sy_call_t *)orig_rename_sysent.sy_call )( p, uap );
       
   /*
    * Continue if the original syscall succeeded.
    */
   if( from == NULL ) return error;
       
   if( error != 0 )
   {
      wss_free( from );
      return( error );
   }
       
   WSS_MUTEX_ENTER( wss_proc_list );
   for( wrp = wss_proc_list.lh_first; wrp; wrp = wrp->wrp_p_list.le_next )
   {
      if( wrp->wrp_cwd )
      {
         WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER )
         if( wrp->wrp_cwd && wss_strstr( from, wrp->wrp_cwd ) )
         {
            if( wrp->wrp_p == p )
            {
               char *tmp = wrp->wrp_cwd;
               wrp->wrp_cwd = wss_path_cwd( wrp );
               if( tmp )
                  wss_free( tmp );
            }
            else
            {
               WSS_MUTEX_ENTER( wrp->wrp );
               wrp->wrp_flags |= WRPF_GET_CWD;
               WSS_MUTEX_EXIT( wrp->wrp );
            }
         }
         WSI_RW_EXIT( wrp->wrp_cwd );
      }
   }
   WSS_MUTEX_EXIT( wss_proc_list );

   wss_free( from ); /* We shouldn't need this anymore. */
       
   return error;
}      


/*  wsi_chdir */
/*! \brief Replacement syscall for chdir to handle CWD caching
 *
 *  \param p      Pointer to process structure
 *  \param uap    Pointer to arguments structure
 *  \param retval Pointer to storage for return value
 *
 *  \sa chdir(2) man page for return values
 */
int
wsi_chdir( proc_t *p, register struct chdir_uap *uap, int *retval )
{
   int         error;
   wss_proc_t *wrp;
   char       *tmp;
   
   /*
    * Call the original chdir syscall.
    */
   error = ( *(sy_call_t *)orig_chdir_sysent.sy_call )( p, uap );

   /*
    * Continue if the original syscall succeeded.
    */
   if( error != 0 )
      return error;

   wrp = wss_pfind( p->p_pid );

   /*
    * Only care about wrapped processes.
    */
   if( !wrp )
      return error;
   if( wrp->wrp_flags & WRPF_WRAPPED )
   {
      WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
      tmp = wrp->wrp_cwd;
      wrp->wrp_cwd = wss_path_cwd( wrp );
      WSI_RW_EXIT( wrp->wrp_cwd );

      if( tmp )
         wss_free( tmp );
   }

   wss_proc_rele( wrp );

   return error;
}


/*  wsi_fchdir */
/*! \brief Replacement syscall for fchdir to handle CWD caching
 *
 *  \param p      Pointer to process structure
 *  \param uap    Pointer to arguments structure
 *  \param retval Pointer to storage for return value
 *
 *  \sa fchdir(2) man page for return values
 */
int
wsi_fchdir( proc_t *p, register struct fchdir_uap *uap, int *retval )
{
   int         error;
   wss_proc_t *wrp;
   char       *tmp;

   /*
    * Call the original chdir syscall.
    */
   error = ( *(sy_call_t *)orig_fchdir_sysent.sy_call )( p, uap );

   /*
    * Continue if the original syscall succeeded.
    */
   if( error != 0 )
      return error;

   wrp = wss_pfind( p->p_pid );

   /*
    * Only care about wrapped processes.
    */
   if( !wrp )
      return error;
   if( wrp->wrp_flags & WRPF_WRAPPED )
   {
      WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
      tmp = wrp->wrp_cwd;
      wrp->wrp_cwd = wss_path_cwd( wrp );
      WSI_RW_EXIT( wrp->wrp_cwd );
      if( tmp )
         wss_free( tmp );
   }

   wss_proc_rele( wrp );

   return error;
}

