/*! \file wsi_dispatch.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_dispatch.c,v 1.36 2001/06/21 21:28:50 stevek Exp $
 *
 * \brief   This is the main dispatch routine for all syscalls
 * \author  Chris Vance, Steve Kiernan
 * \version $Revision: 1.36 $
 * \date    $Date: 2001/06/21 21:28:50 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/param.h>
#include <sys/sysent.h>
#include <sys/resource.h>
#include <sys/syscall.h>
#include <sys/proc.h>

#if defined(DEBUG_REFS) && !defined(DISABLE_PROC_DEBUG_REFS)
#define DISABLE_PROC_DEBUG_REFS 1
#endif /* DEBUG_REFS && !DISABLE_PROC_DEBUG_REFS */

#include <wss_mutex.h>

#include <wss_module.h>
#include <wss_path.h>
#include <wss_malloc.h>
#include <wss_arg.h>
#include <wsi_syscalls.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern int wsi_unloading;


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

int wsi_dispatch __P((int space, pd_syscall_id_t code, proc_t *p, void *uap,
                      int retval[]));


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_dispatch */
/*! \brief Syscall dispatch routine
 *
 * The entire syscall dispatch table was replaced with a version
 * (called wr_aout_sysvec) where each syscall has an intermediate
 * function that calls this dispatch routine.  Parameters are the same
 * as the original syscall, with the addition of a space and code parameter.
 *
 * This function does a lot.  At this point we are gauranteed that the
 * process is wrapped, but we need to determine what (if anything) to
 * do before and after the syscall.
 * 
 * We need to know:
 *    What wrappers are active for this process.
 * For each active wrapper, are there pre/post events?
 * For each pre/post event, conditionally run the action.
 *
 * \param space  Event space identifier
 * \param code   Index into the syscall table
 * \param p      Pointer to process structure
 * \param uap    Pointer to arguments structure
 * \param retval Pointer to storage for return value
 *
 * \return The value returned by the original call or a pre/post operation
 */
int
wsi_dispatch( int space, pd_syscall_id_t code, proc_t *p, void *uap,
              int *retval )
{
   wss_wrapper_instance_t  *wi, *prev_wi;
   wss_proc_t              *wrp;
   struct sysent           *callp=0;    /* points to the true system call */
   int                      err2;
   int                      err = 0;
   wss_wrapper_data_t       wrd = {0,0,0};
   int                      did_truncate = 0;

   retval = p->p_retval;

   wrd.errnum = &err;     

   wrp = wss_pfind( p->p_pid );
   if( !wrp )
   {
      wsi_printf( PD_MAKESTR("wsi_dispatch:: ERROR, no wss_proc!, pid %d\n"),
                  (int)p->p_pid );
      if( !space )
      {
         callp = &p->p_sysent->sv_table[code];
         return( (*callp->sy_call)( p, uap ) );
      }
      else
         return 0;
   }

   /* !space: the system event space (syscalls) is always space 0 */
   if (!space)
   {
      struct sysentvec *sysentvec = WSI_PROC_OPAQUE(wrp)->sysent;
      if( sysentvec )
         callp = sysentvec->sv_table + code;
      else
         panic("process %d (0x%x) without replacement sysentvec",
               p->p_pid, (unsigned int) p);
   }

   /*
    * Call pre functions for each wrapper.
    */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( prev_wi = NULL, wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
        wi;
        prev_wi = wi, wi = wi->wi_next )
   {
      if( wi->wi_wrapper->w_dispatch_pre )
      {
         wrd.wiv = wi->wi_data;
         err2 = (*wi->wi_wrapper->w_dispatch_pre)( space, code, uap, retval,
                                                   wrp, &wrd, (space == 0) );

         /* only uses the lower 2 bit of the return val */
         if( ( err2 & WR_MASK1 ) == WR_DENY )
         {
            did_truncate = 1;
            /* break; */
            goto exit;
         }
      }
   } 
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   /*
    * Call the actual syscall.
    */ 
   if( !space )
   {
      WSS_MUTEX_ENTER( wrp->wrp );
      wrp->wrp_nsyscall++;
      WSS_MUTEX_EXIT( wrp->wrp );

      wss_proc_rele( wrp );

      err = (*callp->sy_call)( p, uap );

      /* find the wrp again, since we might be the child of a fork! */
      wrp = wss_pfind( curproc->p_pid );

      if( !wrp )
         panic( "wsi_dispatch: unable to find wrp after syscall" );
      else
      {
         WSS_MUTEX_ENTER( wrp->wrp );
         wrp->wrp_nsyscall--;
         WSS_MUTEX_EXIT( wrp->wrp );
      }

      switch( err )
      {
         case 0:
            /*
             * Reinitialize proc pointer `p' as it may be different
             * if this is a child returning from fork syscall.
             */
            p = curproc;

            WSS_MUTEX_ENTER( wrp->wrp );
            if( wrp->wrp_flags & WRPF_GET_CWD )
            {
               pd_char_t *tmp;
               tmp = wrp->wrp_cwd;
               WSS_MUTEX_EXIT( wrp->wrp );

               wrp->wrp_cwd = wss_path_cwd( wrp );
               if( tmp )
                  wss_free( tmp );

               WSS_MUTEX_ENTER( wrp->wrp );
               wrp->wrp_flags &= ~WRPF_GET_CWD;
            }
            WSS_MUTEX_EXIT( wrp->wrp );
            break;

         case ERESTART:
         case EJUSTRETURN:
            break;

         default:
            /*
             * TBD: Fix This!  -1 is not the correct error return
             * for all syscalls.  -1 is a perfectly valid pointer!
             */
            /* *retval = -1; */
            break;
      }
   }

   /*
    * Call post functions for each wrapper.  We perform the
    * "truncated onion method" for failure handling.  Basically
    * if there was a failure, we start with the wrapper that
    * cause the error and work our way back to the first wrapper
    * called.  In this way, after a PRE event fails, no more PRE
    * events run.  POST events start with the failing one and
    * continue with POST events for all prior PREs.
    */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );

   if( !did_truncate )
   {
      /* find the wi again, since we might be the child of a fork! */
      /* reaquire */
      wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
      while( wi && wi->wi_next )
         wi = wi->wi_next;
   }
   else
      wi = prev_wi;

   for( ; wi; wi = wi->wi_prev )
   {
      if( wi->wi_wrapper->w_dispatch_post )
      {
         wrd.wiv = wi->wi_data;
         err2 = (*wi->wi_wrapper->w_dispatch_post)( space, code, uap, retval,
                                                    wrp, &wrd, (space == 0) );
      }
   }

exit:
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   /* No threads, so use 0 for tid.  system space is 0*/
   wss_arg_donesyscall( wrp, 0, 0, code, uap );

   if( wsi_unloading )
   {
      /*
       * If we're slated to unload the WSS LKM, restore the original vector
       */
      if( wrp->wrp_nsyscall == 0 ) wsi_proc_restore_sysent( wrp, TRUE );
   }

   wss_proc_rele( wrp );

   return err; /* err is from syscall execution or set by the pre */
}


/*  wsi_syscall_name */
/*! \brief Return syscall name based on code argument
 *  
 *  \param code Syscall indentifier
 *  
 *  \return Pointer to syscall name
 */
const pd_char_t *
wsi_syscall_name( pd_syscall_id_t code )
{
   return freebsd_syscallnames[code];
}

