/*! \file wsi_execve.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_execve.c,v 1.32 2001/06/21 21:27:04 stevek Exp $
 *
 * \brief   A replacement for the execve syscall.
 * \author  Chris Vance, Steve Kiernan
 * \version $Revision: 1.32 $
 * \date    $Date: 2001/06/21 21:27:04 $
 * 
 * This does some wrappers-relevant work and then passes control
 * off to the original execve system call.
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/sysent.h>

#include <wss_mutex.h>

#include <wss_path.h>
#include <wss_malloc.h>
#include <wss_process.h>
#include <wss_criteria.h>
#include <wss_wrapper.h>
#include <wss_log.h>
#include <wsi_execve.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern struct sysent orig_execve_sysent;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_execve */ 
/*! \brief  The replacement function for the execve system call.
 * 
 *  It does some wrappers-specific stuff, and then calls the original
 *  execve syscall.  Of importance is the need to look up the full
 *  pathname for the program being executed.  This should only need to
 *  be done if the exec was a success.  With any luck, most of the
 *  processing it does has been cached during the similar lookup in the
 *  actual syscall.  The program name is also available in argv[0]
 *  (relative to the cwd) and is stored in p->p_comm by the execve
 *  syscall (only the first 16 chars of the binary name - no path).
 *
 *  \param p      Pointer to process structure
 *  \param uap    Pointer to the call arugment structure
 *  \param retval Pointer to memory where return value should be stored
 *
 *  \return Value returned by original execve call
 */
int
wsi_execve( proc_t *p, struct execve_uap *uap, int *retval )
{
   wss_proc_t *wrp;
   int         err;
   int         act_error;
   int         wr_run_wrapped;
   pd_char_t *path;
   int         len;
   int         i;
   pd_char_t *tmp;

   retval = p->p_retval;


   /*
    * Look up the current process in the wrappers proc
    * structure. If it wasn't found, this means there was either
    * a problem, or this is a process that was running before the
    * wrappers module was loaded into the kernel.  In the first
    * case, this is a big problem.  In the second case, we can
    * probably work around it by creating the structure at this
    * time.
    */
   wrp = wss_pfind( p->p_pid );
   if( !wrp )
   {
      wsi_printf( PD_MAKESTR("wsi_execve: can't find pid: p->p_pid %d\n"),
                  (int)p->p_pid );

      wrp = wss_proc_new( p );
   }

   /*
    * wrp was tested for validity early on.  Make space and copy
    * the path, first saving the current one in an 'old' pathbuf
    */
   if( wrp->wrp_oldpath )
      wss_free( wrp->wrp_oldpath );

   WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
   tmp = wrp->wrp_cwd;
   wrp->wrp_cwd = wsi_path_cwd( wrp );
   WSI_RW_EXIT( wrp->wrp_cwd );
   if( tmp )
      wss_free( tmp );

   path = wss_malloc( MAXPATHLEN );
   copyinstr( uap->path, path, MAXPATHLEN, &len );
   /* wsi_printf("EXEC before path = %s\n", path); */
   wrp->wrp_oldpath = wrp->wrp_path; 
   WSI_RW_ENTER( wrp->wrp_cwd, RW_READER );
   wrp->wrp_path = wss_path_rel_to_abspath( wrp->wrp_cwd, path );
   WSI_RW_EXIT( wrp->wrp_cwd );
   wss_free( path );

   if( wrp->wrp_path == NULL )
   {
      wss_proc_rele( wrp );

      /* file was not found, return ENOENT */
      return ENOENT;
   }
      
   /* Clear out progname if its currently set */
   if( wrp->wrp_progname != NULL )
   {
      pd_char_t *progname = wrp->wrp_progname;
      wrp->wrp_progname = NULL;
      wss_free( progname );
   }

   /*
    * Generate the program name, full path should have at least
    * one '/' character.  Program name is everything after the
    * last '/'.
    */
   for( len = 0, i = wsi_strlen( wrp->wrp_path );
        ( i > 0 ) && ( wrp->wrp_path[i] != WSS_FILE_SEPARATOR );
        i--, len++ );

   /*
    * Check (for sanity, etc)
    */
   if( ( i == 0 ) && ( wrp->wrp_path[0] != WSS_FILE_SEPARATOR ) )
   {
      wsi_printf( PD_MAKESTR("wsi_execve: Pathname `%s' didn't have a '%c' in it\n"),
                  wrp->wrp_path, WSS_FILE_SEPARATOR );
      wrp->wrp_progname = NULL;
   }
   else
   {
      wrp->wrp_progname = wss_malloc_string( len );
      bcopy( &( wrp->wrp_path[i + 1] ), wrp->wrp_progname,
             len * sizeof(pd_char_t));
   }

   /* Release our reference to the wrapper process structure */
   wss_proc_rele( wrp );

   /*
    * Next, call the original syscall.
    */ 
   err = (*(sy_call_t *)orig_execve_sysent.sy_call)( p, uap );

   /* Find the wrapper process structure again */
   wrp = wss_pfind( p->p_pid );
   if( !wrp )
      panic( "wsi_execve: unable to find wrp after syscall" );

   /* If there was no error in exec and this is not a kernel thread
    * process, run through activation checks
    */
   if( ( err == 0 ) && ( ( p->p_flag & P_KTHREADP ) == 0 ) )
   {
      /* zero from execve syscall indicates success? */
      int has_wrappers;

      /*
       * We may now check the activation criteria to determine
       * whether or not to run this process wrapped.
       */
      wss_crit_check_deactivation( wrp );
      wr_run_wrapped = wss_crit_check_activation( wrp );

      WSS_MUTEX_ENTER( wrp->wrp_wi_list );
      has_wrappers = (wrp->wrp_wi_list == NULL) ? FALSE : TRUE;
      WSS_MUTEX_EXIT( wrp->wrp_wi_list );

      if( has_wrappers )
      {
         err = wsi_proc_replace_sysent( wrp );
         if( err )
         {
            wsi_printf( PD_MAKESTR("GSWTK: Error %d replacing sysent for")
                        PD_MAKESTR(" PID %d, terminating\n"), err, p->p_pid );
            wss_proc_rele( wrp );
            return (err);
         }

         /*
          * Now is a convenient place to call the
          * wrapper-specific wr_activation() functions.
          */
         act_error = wss_wrapper_call_activation( wrp );
         if( act_error != 0 )
         {
            /*
             * TBD: Is there anything else we should
             * do here, like kill the process?
             */
            wsi_printf( PD_MAKESTR("Activate returned an error!\n") );
         }
      }
   }

#ifdef LOG_WRAPPED
   (void)wss_log_printf( PD_MAKESTR("Running %s %swrapped\n"), wrp->wrp_path,
                         ( wrp->wrp_wi_list ? PD_MAKESTR("")
                                            : PD_MAKESTR("un") ) );
#endif /* LOG_WRAPPED */

   wss_proc_rele( wrp );

   return( err );
}

