/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_fileutil.c,v 1.17 2001/06/11 17:24:42 stevek Exp $
 *
 * Author: Calvin Ko
 *
 * Purpose: Contain functions for network and files
 *
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include <sys/types.h>
#include <sys/filedesc.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/kernel.h>
#include <sys/mbuf.h>
#include <sys/namei.h>
#include <sys/protosw.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/vnode.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <wsi_fileutil.h>

/*
 * Copies information from the stat pointer to the identical wr_stat
 * pointer.  Maybe "memcpy" would be faster?
 */
static void cvstat(struct wr_stat *wrsb, struct stat *sb)
{
        wrsb->st_dev = sb->st_dev;
	wrsb->st_ino = sb->st_ino;
	wrsb->st_mode = sb->st_mode;
	wrsb->st_nlink = sb->st_nlink;
	wrsb->st_uid = sb->st_uid;
	wrsb->st_gid = sb->st_gid;
	wrsb->st_rdev = sb->st_rdev;
	wrsb->st_atime = sb->st_atime;
	wrsb->st_atimespec = sb->st_atimespec;
	wrsb->st_mtime = sb->st_mtime;
	wrsb->st_mtimespec = sb->st_mtimespec;
	wrsb->st_ctime = sb->st_ctime;
	wrsb->st_ctimespec = sb->st_ctimespec;
	wrsb->st_size = sb->st_size;
	wrsb->st_gen = sb->st_gen; 
        /* memcpy(wrsb, sb, sizeof(struct stat)); */
        return;
}

int wsi_fileutil_stat(const char *path, struct wr_stat *wrsb)

{
        struct stat sb;
        struct nameidata nd;
        int saveuid;

        NDINIT(&nd, LOOKUP, FOLLOW | LOCKLEAF, UIO_SYSSPACE, path, curproc);
        wrsb->error = namei(&nd);
        if (wrsb->error) 
                return (wrsb->error);
        saveuid = curproc->p_ucred->cr_uid; /* a hack to make curproc as root */
        curproc->p_ucred->cr_uid = 0; 
        wrsb->error = vn_stat(nd.ni_vp, &sb, curproc);
        curproc->p_ucred->cr_uid = saveuid;
        vput(nd.ni_vp);
        cvstat(wrsb, &sb);
        return wrsb->error;
}

int wsi_fileutil_lstat(const char *path, struct wr_stat *wrsb)
{
        struct vnode *vp;
        struct stat sb;
        struct nameidata nd;

        NDINIT(&nd, LOOKUP, FOLLOW | LOCKLEAF, UIO_SYSSPACE, path, curproc);
        wrsb->error = namei(&nd);
        if (wrsb->error) 
                return (wrsb->error);
        vp = nd.ni_vp;
        wrsb->error = vn_stat(vp, &sb, curproc);
        if (vp->v_type == VLNK) 
                sb.st_mode |= S_IFLNK | ACCESSPERMS;    /* 0777 */

        vput(vp);
        cvstat(wrsb, &sb);
        return wrsb->error;
}

int wsi_fileutil_fstat(unsigned int fd, struct wr_stat *wrsb)
{
        register struct filedesc *fdp = curproc->p_fd;
        register struct file *fp;
        struct stat ub;
        int saveuid;

        if (fd >= fdp->fd_nfiles || (fp = fdp->fd_ofiles[fd]) == NULL)
                return (EBADF);
        switch (fp->f_type) {

        case DTYPE_FIFO:
        case DTYPE_VNODE:
                /* a hack to make curproc as root */
                saveuid = curproc->p_ucred->cr_uid; 
                curproc->p_ucred->cr_uid = 0; 
                wrsb->error = vn_stat((struct vnode *)fp->f_data, &ub, curproc);
                curproc->p_ucred->cr_uid = saveuid;
        	cvstat(wrsb, &ub);
                break;

        case DTYPE_SOCKET:
#if WR_FREEBSD < 4
                wrsb->error = soo_stat((struct socket *)fp->f_data, &ub);
#else
                wrsb->error = soo_stat(fp, &ub, curproc);
#endif
        	cvstat(wrsb, &ub);
                break;

        case DTYPE_PIPE:
                wrsb->error = pipe_stat((struct pipe *)fp->f_data, &ub);
        	cvstat(wrsb, &ub);
                break;
        default:
                return -1;
        }
        return (wrsb->error);
}

int wsi_fileutil_fdtype(unsigned int fd)
{
        register struct filedesc *fdp = curproc->p_fd;
        register struct file *fp;
        if (fd >= fdp->fd_nfiles || (fp = fdp->fd_ofiles[fd]) == NULL)
              return (-1);
        switch (fp->f_type) {
        case DTYPE_FIFO  :    return WR_DTYPE_FIFO;
        case DTYPE_VNODE :    return WR_DTYPE_VNODE;
        case DTYPE_SOCKET:    return WR_DTYPE_SOCKET;
        case DTYPE_PIPE  :    return WR_DTYPE_PIPE;
        } 
        return -1;
};

int wsi_fileutil_fdfflag(unsigned int fd)
{
        register struct filedesc *fdp = curproc->p_fd;
        register struct file *fp;
        if (fd >= fdp->fd_nfiles || (fp = fdp->fd_ofiles[fd]) == NULL)
              return (-1);
        return fp->f_flag;
}

/*
 *	Network Related Stuff
 */

static const char *af_name[AF_MAX] = {
        "AF_UNSPEC", 
        "AF_LOCAL", 
        "AF_INET", 
        "AF_IMPLINK", 
        "AF_PUP", 
        "AF_CHAOS",  
        "AF_NS", 
        "AF_ISO", 
        "AF_ECMA", 
        "AF_DATAKIT", 
        "AF_CCITT", 
        "AF_SNA", 
        "AF_DECnet", 
        "AF_DLI", 
        "AF_LAT", 
        "AF_HYLINK", 
        "AF_APPLETALK", 
        "AF_ROUTE", 
        "AF_LINK", 
        "pseudo_AF_XTP", 
        "AF_COIP", 
        "pseudo_AF_RTIP", 
        "AF_IPX",  
        "AF_SIP", 
        "pseudo_AF_PIP", 
        "AF_ISDN", 
        "AF_E164", 
        "pseudo_AF_KEY",  
        "AF_INET6"      
};

static const char *af_unknown = "AF_UNKNOWN";
static const char *msg_noaddr = "No address";
static const char *msg_fail = "Fail";

/* Buffer must be at least 16 long.  We don't check length here */
static void wr_netitoa(long inaddr, char *netbuf)
{
     unsigned char *ucp = (unsigned char *) &inaddr;
     sprintf(netbuf, "%d.%d.%d.%d",
                        ucp[0] & 0xff,
                        ucp[1] & 0xff,
                        ucp[2] & 0xff,
                        ucp[3] & 0xff);
}

const char *wsi_netutil_get_family(const struct sockaddr *saddr, int len)
{
     if (saddr != 0 && len >= 16 && saddr->sa_family < AF_MAX) {
          return af_name[saddr->sa_family];
     } else {
          return af_unknown;
     }
}

#define asdf_RETURN(s)                                          \
    data_len = strlen((s)) + 1;                                 \
    if (!buffer || buffer_length < data_len)                    \
        return -1;                                              \
    strcpy(buffer,(s));                                         \
    return 0

int wsi_netutil_get_addr(
        const struct sockaddr *saddr,
        int len,
        char *buffer,
        int buffer_length)
{
    int data_len;
    char netbuf[16];
    if (saddr != 0 && len >= 16 && saddr->sa_family < AF_MAX) {
        switch (saddr->sa_family) {
        case AF_UNIX:
            asdf_RETURN((char*)saddr->sa_data);
        case AF_INET:
            wr_netitoa(((struct sockaddr_in*)saddr)->sin_addr.s_addr,netbuf);
            data_len = strlen(netbuf) /* "255.255.255.255" */
                + 2 /* "::" */ + 5 /* "32768" */ + 1 /* '\0' */;
            if (!buffer || buffer_length < data_len)
                return -1;
            sprintf(buffer, "%s::%d",
                    netbuf,
                    ntohs(((struct sockaddr_in*)saddr)->sin_port)); 
            return 0;
        default:
            asdf_RETURN(msg_fail);
        }
    } else {
        asdf_RETURN(msg_noaddr);
    }
}    

int wsi_netutil_ntohs(short i)
{
     return ntohs(i);
};

int wsi_netutil_getsockname(int fd, struct sockaddr *s, int *len)
{
        struct file *fp;
        register struct socket *so;
        struct mbuf *m; 
        int error;

        fp = curproc->p_fd->fd_ofiles[fd];
        if( fp->f_type != DTYPE_SOCKET )
		return (EINVAL);
        so = (struct socket *)fp->f_data;
        m = m_getclr(M_WAIT, MT_SONAME);
        if (m == NULL) 
                return (ENOBUFS);
        error = (*so->so_proto->pr_usrreqs->pru_sockaddr)(so, m);
        if (!error) {
                *len = m->m_len;
                bcopy(mtod(m, caddr_t), (caddr_t)s, (u_int)*len); 
        }
        m_freem(m);
        return error;
}


int wsi_netutil_getpeername(int fd, struct sockaddr *s, int *len)
{
        struct file *fp;
        register struct socket *so;
        struct mbuf *m; 
        int error;

        fp = curproc->p_fd->fd_ofiles[fd];
        if( fp->f_type != DTYPE_SOCKET )
		return (EINVAL);
        so = (struct socket *)fp->f_data;
        if ((so->so_state & (SS_ISCONNECTED|SS_ISCONFIRMING)) == 0)
                return (ENOTCONN);
        m = m_getclr(M_WAIT, MT_SONAME); 
        if (m == NULL)
                return (ENOBUFS);
        error = (*so->so_proto->pr_usrreqs->pru_peeraddr)(so, m);
        if (!error) {
                *len = m->m_len;
                bcopy(mtod(m, caddr_t), (caddr_t)s, (u_int)*len); 
        }
        m_freem(m);
        return (error);
}

char *wsi_gethostname()

{
	return hostname;
};

