/*! \file wsi_init.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_init.c,v 1.9 2001/02/23 19:02:34 stevek Exp $
 *
 * \brief   Read in WSS configuration file in system-dependent fashion
 * \author  Steve Kiernan
 * \version $Revision: 1.9 $
 * \date    $Date: 2001/02/23 19:02:34 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_file.h>
#include <wss_init.h>


/***************************************************************************\
 * Macros                                                                  *
\***************************************************************************/

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif


/************************************************************************\
 * Private globals                                                      *
\************************************************************************/

wss_init_entry_t wsi_init_locals[] =
{
   { NULL, 0, NULL }
};


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static int
_wsi_init_process_entry( wss_init_entry_t *entry, const char *value )
{
   switch( entry->type )
   {
      case WSS_INIT_TYPE_STRING:
      {
         pd_char_t **str = (pd_char_t **)entry->storage;

         if( *str )
            wss_free( *str );
         *str = wss_strdup( value );
#ifdef DEBUG_INIT
         wsi_printf( PD_MAKESTR("Adding string entry '%s': %s\n"), entry->key,
                     value );
#endif
         return TRUE;
      }

      case WSS_INIT_TYPE_INT:
      {
         int val;

         for( val = 0;
              *value && ( *value >= '0' ) && ( *value <= '9' );
              value++ )
         {
            val *= 10;
            val += ( *value - '0' );
         }

         *((int *)entry->storage) = val;
#ifdef DEBUG_INIT
         wsi_printf( PD_MAKESTR("Adding int entry '%s': %s\n"), entry->key,
                     value );
#endif
         return TRUE;
      }

      case WSS_INIT_TYPE_BOOL:
         if( !strcmp( value, "TRUE" ) || !strcmp( value, "true" ) ||
             !strcmp( value, "YES" ) || !strcmp( value, "yes" ) )
         {
            *((int *)entry->storage) = 1;
#ifdef DEBUG_INIT
            wsi_printf( PD_MAKESTR("Adding bool entry '%s': %s\n"), entry->key,
                        value );
#endif
            return TRUE;
         }
         if( !strcmp( value, "FALSE" ) || !strcmp( value, "false" ) ||
             !strcmp( value, "NO" ) || !strcmp( value, "no" ) )
         {
            *((int *)entry->storage) = 0;
#ifdef DEBUG_INIT
            wsi_printf( PD_MAKESTR("Adding bool entry '%s': %s\n"), entry->key,
                        value );
#endif
            return TRUE;
         }
         break;

      default:
         break;
   }
   return FALSE;
}


static void
_wsi_init_parse_entry( char *entry, int line )
{
   char *eq;

   /* Eat white space */
   for( ; ( *entry == ' ' ) || ( *entry == '\t' ); entry++ );

   /* Skip comment or blank line */
   if( ( *entry == '#' ) || ( *entry == '\0' ) ) return;

   for( eq = entry; *eq && *eq != '='; eq++ );

   if( *eq == '=' )
   {
      wss_init_entry_t *init_entry;
      char             *entend = eq++;
      char             *eqend = eq;

      /* Eat trailing whitespace */
      while( ( entend > entry ) &&
             ( ( *entend == ' ' ) || ( *entend == '\t' ) ||
               ( *entend == '=' ) ) )
         *entend-- = '\0';

      /* Eat leading whitespace */
      for( ; ( *eq == ' ' ) || ( *eq == '\t' ); eq++ );

      /* Find assignment end and chop off comments */
      for( ; *eqend && (*eqend != '=' ); eqend++ );
      *eqend = '\0';

      /* Eat trailing whitespace */
      while( ( eqend > eq ) &&
             ( ( *eqend == ' ' ) || ( *eqend == '\t' ) || ( *eqend == '=' ) ) )
         *eqend-- = '\0';

      /* Search for entries in configuration table(s) */
      for( init_entry = wss_init_globals;
           ( init_entry->key != NULL );
           init_entry++ )
         if( !strcmp( init_entry->key, entry ) &&
             _wsi_init_process_entry( init_entry, eq ) )
            return;

      for( init_entry = wsi_init_locals;
           ( init_entry->key != NULL );
           init_entry++ )
         if( !strcmp( init_entry->key, entry ) &&
             _wsi_init_process_entry( init_entry, eq ) )
            return;
   }

   wsi_printf( PD_MAKESTR("wsi_init: Invalid entry at line %d\n"), line );
}

static int
_wsi_init_parse( char *buffer, int len, int *line )
{
   int index = 0;
   int start = 0;

   for( ; index < len; index++ )
   {
      if( buffer[index] == '\n' )
      {
         buffer[index] = '\0';
         _wsi_init_parse_entry( buffer + start, *line );
         start = index + 1;
         *line += 1; /* no ++ because gcc complains value not used if we do */
      }
   }

   /* Shift the remaining part up */
   if( start && ( start < len ) )
      strncpy( buffer, buffer + start, len - start );

   return( len - start );
}

/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_init_read_conf */
/*! \brief Open configuration file and read configuration data to initialize
 *         the WSS
 */
void
wsi_init_read_conf( void )
{
   wss_file_t file;
   char       buffer[MAXPATHLEN];
   int        num_to_read;
   int        offset = 0;
   int        nbytes;
   int        line = 0;

   if( !( file = wss_file_open( WSS_INIT_FILE,
                                WSS_FILE_READ | WSS_FILE_ROOTPRIV ) ) )
      return;
   
   for( num_to_read = wss_file_get_size( file );
        num_to_read > 0;
        num_to_read -= nbytes )
   {
      nbytes = MIN( MAXPATHLEN, num_to_read ) - offset;

      if( wsi_file_read( file, buffer + offset, nbytes ) < 0 )
         break;

      if( ( offset = _wsi_init_parse( buffer, nbytes, &line ) ) == MAXPATHLEN )
      {
         wsi_printf( PD_MAKESTR("wsi_init_read_conf: possible corrupt ")
                     PD_MAKESTR("configuration file, line longer than max ")
                     PD_MAKESTR("(%d)\n"), MAXPATHLEN );
         break;
      }
   }

   wss_file_close( file );
}

