/*! \file wsi_malloc.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_malloc.c,v 1.20 2001/02/23 19:02:34 stevek Exp $
 *
 * \brief   Platform specific part of the malloc routines
 * \author  Douglas Kilpatrick, Steve Kiernan
 * \version $Revision: 1.20 $
 * \date    $Date: 2001/02/23 19:02:34 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>
#include <wsi_malloc.h>

#include <sys/types.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/mman.h>
#include <sys/lock.h>

/* for VM types */
#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/vm_extern.h>
#if WR_FREEBSD < 4
#include <vm/vm_prot.h>
#endif
#include <vm/pmap.h>
#include <vm/vm_map.h>


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

void *wsi_proc_malloc      __P(( proc_t *p, int size ));
void  wsi_malloc_proc_free __P(( proc_t *p, void *addr, int size ));

/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

#ifdef MALLOC_DEFINE
MALLOC_DEFINE(M_WSI, "wsi", "WSI allocated memory");
#endif /* MALLOC_DEFINE */


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_malloc */
/*! \brief Allocate memory of at least specified size
 *  
 *  \param size Number of bytes to allocate
 *
 *  \return Pointer to allocated memory on success,
 *  \return NULL on error
 */
void *
wsi_malloc( size_t size )
{
   char *buf;
#ifdef WSI_MALLOC_FENCE
   size += 2*sizeof(long) + sizeof(size_t);
#endif
   MALLOC( buf, char *, size, M_WSI, M_WAITOK );
#ifdef WSI_MALLOC_FENCE
   *((size_t*)buf) = size;
   *((long*)(buf+sizeof(size_t))) = 0xdeadbeef;
   /* Unaligned access... */
   *((long*)(buf+size-sizeof(long))) = 0xdeadbeef;

   buf += sizeof(size_t) + sizeof(long);
#endif
   return( (void *)buf );
}


/*  wsi_free */
/*! \brief Free allocated memory
 *
 *  \param ptr Pointer to allocated memory block
 */
void
wsi_free( void *ptr )
{
#ifdef WSI_MALLOC_FENCE
    size_t size;
#endif

    if (!ptr)
        return;

#ifdef WSI_MALLOC_FENCE
    ptr = (char*)ptr - (sizeof(size_t) + sizeof(long));
    size = *(size_t *)ptr;

    if (*((long*)((char*)ptr+sizeof(size_t))) != 0xdeadbeef)
        panic("low fence overwritten: ptr=%p,size=%d,low=%lx",
                ptr,
                size-sizeof(size_t)-2*sizeof(long),
                *((long*)((char*)ptr+sizeof(size_t))) );
    if (*((long*)((char*)ptr+size-sizeof(long))) != 0xdeadbeef)
        panic("high fence overwritten: ptr=%p,size=%d,high=%lx",
                ptr,
                size-sizeof(size_t)-2*sizeof(long),
                *((long*)((char*)ptr+size-sizeof(long))) );
#endif
   FREE( ptr, M_WSI );
}


/*  wsi_curproc_malloc */
/*! \brief Allocate a chunk of memory in the vmspace of the current process
 *
 *  \param size Number of bytes to allocate
 *
 *  \return Pointer to allocated memory on success,
 *  \return NULL on error
 */
void *
wsi_curproc_malloc( size_t size )
{
   void *addr;

   addr = wsi_proc_malloc( curproc, size );
   if( addr == NULL )
      printf( "wss_curproc_malloc failed!\n" );

   return addr;
}


/*  wsi_proc_malloc */
/*! \brief Allocate a chunk of memory in the vmspace of a process
 *
 *  \param p    Pointer to process structure
 *  \param size Number of bytes to allocate
 *
 *  \warning In the future this should probably be redesigned.  Eventually,
 *           we will probably want to allocate a single piece of memory and
 *           make allocations from there.  This will allow us to lock down
 *           those pages so they may not be modified by the process from
 *           user-space. Otherwise, there may be succeptibility in
 *           multi-threaded systems.  We are using this memory primarily to
 *           hold replaced syscall parameters.  We do not want the user-space
 *           program to be able to modify anything we store here.
 *
 *  \return Pointer to allocated memory on success,
 *  \return NULL on error
 */
void *
wsi_proc_malloc( proc_t *p, int size )
{
   vm_offset_t address, addr;
   vm_size_t len = size;

   address = round_page( (vm_offset_t)p->p_vmspace->vm_daddr + MAXDSIZ );
   len = round_page( size );

   addr = vm_mmap( &p->p_vmspace->vm_map, &address, len, VM_PROT_DEFAULT,
                   VM_PROT_ALL, MAP_ANON, 0, 0 );
   if( addr == KERN_SUCCESS )
   {
      addr = address; /* Gets changed if needed to be */
   }
   else
   {
      addr = 0;
      printf( "wsi_proc_malloc failed!\n" );
   }
   return (void*)addr;
}


/*  wsi_curproc_free */
/*! \brief Frees the address from the vmspace of the current process
 *
 *  \param addr Pointer to allocated memory block
 *  \param size Size of block in bytes
 */ 
void
wsi_curproc_free( void *addr, int size )
{
   wsi_malloc_proc_free( curproc, addr, size );
}


/*  wsi_malloc_proc_free */
/*! \brief Frees the address from the vmspace of a process
 *
 *  \param p    Pointer to process structure
 *  \param addr Pointer to allocated memory block
 *  \param size Size of block in bytes
 */ 
void
wsi_malloc_proc_free( proc_t *p, void *addr, int size )
{
   /* Stupid checks. */
   vm_offset_t address = (vm_offset_t)addr;
   vm_size_t len = size;
   int pageoff;

   pageoff = (address & PAGE_MASK);
   address -= pageoff;
   len += pageoff;
   len = (vm_size_t)round_page( len );
   if( ( len == 0 ) || ( address+len < address ) )
      return;

   vm_map_remove( &p->p_vmspace->vm_map, address, address+len );
}


/*  wsi_curproc_stdup */
/*! \brief Similar to strdup, but allocates memory from the vmspace of the
 *         current process
 *
 *  \param str Pointer to string to duplicate
 *
 *  \return Pointer to duplicated string within the process' vmspace on success,
 *  \return NULL on failure
 */
pd_char_t *
wsi_curproc_strdup( const pd_char_t *str )
{
   pd_char_t *buf;
   buf = wsi_curproc_malloc_string( wsi_strlen( str ) );
   if( buf )
      wsi_strcpy( buf, str );
   return( buf );
}

