/*! \file wsi_path.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_path.c,v 1.43 2001/06/22 00:18:15 stevek Exp $
 *
 * \brief   Functions to determine the paths in system-dependent fashion
 * \author  Steve Kiernan
 * \version $Revision: 1.43 $
 * \date    $Date: 2001/06/22 00:18:15 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/sysctl.h>
#include <sys/mount.h>
#include <sys/vnode.h>
#include <sys/namei.h>
#include <sys/malloc.h>
#include <sys/sysproto.h>
#include <sys/proc.h>
#include <sys/filedesc.h>
#if __FreeBSD_version >= 430000
#include <sys/fnv_hash.h>
#endif

#include <wss_malloc.h>
#include <wss_process.h>
#include <wss_path.h>

#define MIN(a,b) ((a)<(b)?(a):(b))


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_path_get_cwd_as_string */
/*! \brief Determine the absolute path of the current working directory
 *
 *  \param p Pointer to process structure
 *
 *  \warning The caller is responsible for freeing the string.
 *           The WSS deallocation routine should be used.
 *
 *  \return Pointer to string containing the abolute path of the
 *          current working directory of the process on success,
 *  \return NULL otherwise (which could indicate that the V-node has
 *          been unlinked)
 */
pd_char_t *
wsi_path_get_cwd_as_string( proc_t const *p )
{
   return wsi_path_from_vnode( p, p->p_fd->fd_cdir );
}


/*  wsi_path_cwd */
/*! \brief Get the current working directory for a process
 *       
 *  Simple front-end to the wsi_path_get_cwd_as_string() routine.
 * 
 *  \param wrp Pointer to a wrapper process structure
 * 
 *  \return Pointer to string containing the current working directory of
 *          the process
 */
pd_char_t *
wsi_path_cwd( wss_proc_t const *wrp )
{
   if( !wrp )
      return NULL;
   return wsi_path_get_cwd_as_string( wrp->wrp_p );
}


/*  wsi_path_canonicalize */
/*! \brief Make a canonical path from the pathname given
 *
 *  \param p     Pointer to process structure
 *  \param path  Pointer to string containing path
 *  \param kspace Boolean flag, if true path is in kernel space
 *                otherwise path is in user space
 *
 *  \warning Space is allocated!
 *
 *  \return Pointer to string containing canonical path on success,
 *  \return NULL otherwise
 */
pd_char_t *
wsi_path_canonicalize( proc_t const *p, pd_char_t *path, pd_bool_t kspace )
{
   struct vnode *vp;
   pd_char_t    *outpath;

   vp = wsi_path_to_vnode( p, path, kspace );
   if( vp == NULL ) return NULL;

   outpath = wsi_path_from_vnode( p, vp );
   vput( vp );
   return (outpath);
}


/*  wsi_path_is_dir */
/*! \brief Determine if path is a directory
 *
 *  \param p    Pointer to process structure
 *  \param path Pointer to string containing the path
 *
 *  \return TRUE if the path is a directory,
 *  \return NULL otherwise
 */
int
wsi_path_is_dir( proc_t const *p, pd_char_t *path )
{
   struct vnode *vp;
   int           ret = 0;

   vp = wsi_path_to_vnode( p, path, TRUE );
   
   if( vp == NULL )
      return 0;

   if( vp->v_type == VDIR )
      ret = 1;

   vput( vp );
   
   return ret;
}


/*  wsi_path_abspath */
/*! \brief Converts the given path to absolute terms
 *
 *  \b Note:
 *      The path parameter passed in is in kernel space.
 *      The return value is in kernel space.
 *      This function is made available to wrappers.
 *
 *  \param path Pointer to string containing the path
 *
 *  \warning Wrappers can currently misuse this
 *
 *  \return Pointer to string containing the absolute path
 */
pd_char_t *
wsi_path_abspath( pd_char_t *path )
{
   struct vnode *vp;
   pd_char_t    *scratch;
   pd_char_t    *endp;
   pd_char_t    *tail;
   pd_char_t    *outpath = NULL;

   if( path == NULL ) return NULL;

   /* Get the vnode for the specified path */
   vp = wsi_path_to_vnode( curproc, path, TRUE );

   /* If we get a match, convert it to absolute path and return */
   if( vp )
   {
      outpath = wsi_path_from_vnode( curproc, vp );
      vput( vp );
      return (outpath);
   }

   /* Hmm... not an actual file, so let's work on this a bit */
   scratch = wss_malloc( MAXPATHLEN * sizeof(pd_char_t) );
   wsi_strncpy( scratch, path, MAXPATHLEN );
   scratch[MAXPATHLEN-1] = PD_MAKESTR('\0');
   tail = endp = scratch + wsi_strlen( scratch ) - 1;

   /* Keep choping off parts of paths until we find a valid one */
   while( ( vp == NULL ) && ( endp > scratch ) )
   {
      while( ( endp > scratch ) && ( *endp != WSS_FILE_SEPARATOR ) )
         endp--;
      *endp-- = PD_MAKESTR('\0');
      vp = wsi_path_to_vnode( curproc, scratch, TRUE );
   }

   /* We have a valid part, let's piece this path back together */
   if( vp )
   {
      int    index;
      size_t remaining;
      size_t len;
      size_t nbytes;

      index = (endp - scratch) + sizeof(pd_char_t);
      remaining = (tail - endp);

      outpath = wsi_path_from_vnode( curproc, vp );
      vput( vp );

      len = wsi_strlen( outpath );
      nbytes = MIN(remaining, (MAXPATHLEN - len - 1) * sizeof(pd_char_t));

      memcpy( outpath + len, (pd_byte_t *)path + index, nbytes );
      memset( (pd_byte_t *)(outpath + len) + nbytes, 0, sizeof(pd_char_t) );
   }

   /* Free the scratch space */
   wss_free( scratch );

   /* Return the resulting path (or NULL) */
   return (outpath);
}


/*  wsi_path_to_vnode */
/*! \brief Get the Vnode for the pathname
 *
 *  \param p      Pointer to process structure
 *  \param path   Pointer to string containing the path 
 *  \param kspace Boolean flag, if true path is in kernel space
 *                otherwise path is in user space
 * 
 *  \return Pointer to Vnode related to the path on success,
 *  \return NULL if the path does not exist
 */
struct vnode *
wsi_path_to_vnode( proc_t const *p, pd_char_t *path, pd_bool_t kspace )
{
   struct nameidata  nd;
   struct vnode     *vp = NULL;

   NDINIT( &nd, LOOKUP, LOCKLEAF | FOLLOW, kspace?UIO_SYSSPACE:UIO_USERSPACE,
           path, (proc_t *)p );

   if( namei( &nd ) == 0 )
      vp = nd.ni_vp;

   return vp;
}


/* wsi_path_from_vnode adapted from FreeBSD 4.3 version of vfs_cache.c */

/*
 * Copyright (c) 1989, 1993, 1995
 *      The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Poul-Henning Kamp of the FreeBSD Project.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * This structure describes the elements in the cache of recent
 * names looked up by namei.
 */

struct  namecache {
        LIST_ENTRY(namecache) nc_hash;  /* hash chain */
        LIST_ENTRY(namecache) nc_src;   /* source vnode list */
        TAILQ_ENTRY(namecache) nc_dst;  /* destination vnode list */
        struct  vnode *nc_dvp;          /* vnode of parent of name */
        struct  vnode *nc_vp;           /* vnode the name refers to */
        u_char  nc_flag;                /* flag bits */
        u_char  nc_nlen;                /* length of name */
        char    nc_name[0];             /* segment name */
};

pd_char_t *
wsi_path_from_vnode( proc_t const *p, struct vnode *vnodep )
{
   pd_char_t *bp, *cbp, *buf;
   int i, slash_prefixed;
   struct filedesc *fdp;
   struct namecache *ncp;
   struct vnode *vp;

   if( vnodep == NULL ) return NULL;

   buf = wsi_malloc( MAXPATHLEN * sizeof(pd_char_t) );
   bp = buf + MAXPATHLEN - 1;
   *bp = '\0';
   fdp = p->p_fd;
   slash_prefixed = 0;
   for( vp = vnodep; vp != fdp->fd_rdir && vp != rootvnode; )
   {
      if( vp->v_flag & VROOT )
      {
         if( vp->v_mount == NULL )
         {
            /* forced unmount */
            wsi_free( buf );
            return NULL;
         }
         vp = vp->v_mount->mnt_vnodecovered;
         continue;
      }
      if( ( vp != vnodep ) && ( vp->v_dd->v_id != vp->v_ddid ) )
      {
         wsi_free( buf );
         return NULL;
      }
      ncp = TAILQ_FIRST( &vp->v_cache_dst );
      if( !ncp )
      {
         wsi_free( buf );
         return NULL;
      }
      if( ( vp != vnodep ) && ( ncp->nc_dvp != vp->v_dd ) )
      {
         wsi_free( buf );
         return NULL;
      }
      for( i = ncp->nc_nlen - 1; i >= 0; i-- )
      {
         if( bp == buf )
         {
            wsi_free( buf );
            return NULL;
         }
         *--bp = ncp->nc_name[i];
      }
      if( bp == buf )
      {
         wsi_free( buf );
         return NULL;
      }
      *--bp = '/';
      slash_prefixed = 1;
      vp = ncp->nc_dvp;
   }
   if( !slash_prefixed )
   {
      if( bp == buf )
      {
         wsi_free( buf );
         return NULL;
      }
      *--bp = '/';
   }

   for( cbp = buf; ( *cbp++ = *bp++ ); )
      ;

   return (buf);
}

