/*! \file wsi_process.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_process.c,v 1.29 2001/06/22 01:33:37 stevek Exp $
 *
 * \brief   System-dependent process management functions
 * \author  Steve Kiernan
 * \version $Revision: 1.29 $
 * \date    $Date: 2001/06/22 01:33:37 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_process.h>
#include <wss_malloc.h>
#include <wss_path.h>

#include <sys/vnode.h>
#include <sys/sysent.h>
#include <sys/proc.h>
#include <sys/sysproto.h>
#include <sys/filedesc.h>
#include <sys/proc.h>
#include <sys/namei.h>
#include <sys/stat.h>
#include <sys/malloc.h>
#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>
#include <dirent.h>


/***************************************************************************\
 * External structures                                                     *
\***************************************************************************/

extern struct sysent wr_sysent[];


/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/

extern void wsi_unload_deferred __P(( void ));
extern int  wsi_shutdown        __P(( void ));


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_proc_replace_sysent */
/*! \brief Replace the sysent for the specified process
 *
 *  \param wrp Pointer to wrapper process structure
 */
int
wsi_proc_replace_sysent( wss_proc_t *wrp )
{
   struct sysentvec *sysvec = wss_malloc(sizeof(*sysvec));
   if( sysvec == NULL )
      return ENOMEM;

   memcpy( sysvec, wrp->wrp_p->p_sysent, sizeof(*sysvec) );
   sysvec->sv_table = wr_sysent;

   WSI_PROC_OPAQUE(wrp)->sysent = wrp->wrp_p->p_sysent;
   wrp->wrp_p->p_sysent = sysvec;
   wrp->wrp_flags |= WRPF_WRAPPED;

#ifdef DEBUG_WSS_PROC
   wsi_printf(PD_MAKESTR("wsi_proc_replace_sysent: PID %d, replaced 0x%x ")
              PD_MAKESTR("with 0x%x\n"), wrp->wrp_p->p_pid,
              (unsigned int) WSI_PROC_OPAQUE(wrp)->sysent->sv_table,
              (unsigned int) wr_sysent);
#endif /* DEBUG_WSS_PROC */

   return 0;
}


/*  wsi_proc_restore_sysent */
/*! \brief Restore the sysent for the specified process
 *
 *  \param wrp Pointer to wrapper process structure
 */
void
wsi_proc_restore_sysent( wss_proc_t *wrp, pd_bool_t unloading )
{
   if( wrp->wrp_flags & WRPF_WRAPPED )
   {
      struct sysentvec *sysvec;

#ifdef DEBUG_WSS_PROC
      wsi_printf( PD_MAKESTR("putting p->p_sysent back for program %s, pid %d\n"),
                  wrp->wrp_path, wrp->wrp_p->p_pid );
#endif /* DEBUG_WSS_PROC */
      sysvec = wrp->wrp_p->p_sysent;
      if( WSI_PROC_OPAQUE(wrp)->sysent == NULL )
         panic("wsi_proc_restore_sysent: restoration called for process %d "
               "(0x%x) with NULL sysent entry in opaque data\n",
               wrp->wrp_p->p_pid, (unsigned int) wrp->wrp_p);
   
      wrp->wrp_p->p_sysent = WSI_PROC_OPAQUE(wrp)->sysent;
      wss_free( sysvec );

      wrp->wrp_flags &= ~WRPF_WRAPPED;

      if( unloading )
      {
         if( --wss_proc_num_in_syscall == 0 )
            wsi_unload_deferred();
#ifdef DEBUG_UNLOAD
         else
            wsi_printf( PD_MAKESTR("waiting for %d processes...\n"),
                        wss_proc_num_in_syscall );
#endif /* DEBUG_UNLOAD */
      }
   }
}


/*  wsi_proc_force_syscall_exit */
/*! \brief Force specified process to reissue its syscalls
 *
 *  This function posts a SIGINT signal to the process to force any
 *  syscalls to be reissued.
 *    
 *  \param wrp Pointer to wrapper process structure
 */      
void
wsi_proc_force_syscall_exit( wss_proc_t *wrp )
{
   psignal( wrp->wrp_p, SIGINT );
}


/*  wsi_proc_new */
/*! \brief Create the system-dependent opaque data part of the wss_proc 
 *         structure
 *  
 *  \param wrp Pointer to wrapper process structure
 */
void
wsi_proc_new( wss_proc_t *wrp )
{
   WSI_PROC_OPAQUE(wrp) = wss_malloc(sizeof(wsi_proc_t));
   memset( WSI_PROC_OPAQUE(wrp), 0, sizeof(wsi_proc_t) );

   if( wrp->wrp_p->p_sysent->sv_table == NULL )
      panic("Process %d (0x%x) with NULL sysent table\n", wrp->wrp_p->p_pid,
            (unsigned int) wrp->wrp_p);
}


/*  wsi_proc_free */
/*  \brief Free the system-dependent opaque data part of the wss_proc structure
 *
 *  We use this function to determine if init (PID 1) is exiting.  If it is,
 *  then the system is shutting down, so we call wsi_shutdown() to clean up
 *  resources used by the WSS and prevent corruption in the database.
 *
 *  \param wrp Pointer to wrapper process structure
 */
void
wsi_proc_free( wss_proc_t *wrp )
{
   wsi_proc_restore_sysent( wrp, FALSE );

   wss_free( WSI_PROC_OPAQUE(wrp) );

   /* If init is dying, we're outta here */
   if( wrp->wrp_p->p_pid == 1 )
   {
      wsi_printf( PD_MAKESTR("wsi_proc_free: init process exited, cleaning up WSS structures...\n") );

      wsi_shutdown();
   }
}


/*  wsi_proc_fork */
/*! \brief Handling copying sysent for child process on fork
 *
 *  \param wrp_parent Pointer to parent wrapper process structure
 *  \param wrp        Pointer to child wrapper process structure
 */
void
wsi_proc_fork( wss_proc_t *wrp_parent, wss_proc_t *wrp )
{
   if( WSI_PROC_OPAQUE(wrp_parent)->sysent )
   {
      struct sysentvec *sysentvec = wss_malloc(sizeof(*sysentvec));
      if( sysentvec == NULL )
      {
         /* Restore the sysent and kill the child */
         wrp->wrp_p->p_sysent = WSI_PROC_OPAQUE(wrp_parent)->sysent;
         killproc( wrp->wrp_p, "resource exhaustion" );
         return;
      }

      memcpy( sysentvec, wrp_parent->wrp_p->p_sysent, sizeof(*sysentvec) );
      wrp->wrp_p->p_sysent = sysentvec;
      WSI_PROC_OPAQUE(wrp)->sysent = WSI_PROC_OPAQUE(wrp_parent)->sysent;
   }
}

