/*! \file wsi_utilties.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_utilities.c,v 1.18 2001/02/23 19:02:34 stevek Exp $
 * 
 * \brief   Repository of useful functions which might be missing.
 * \author  Doug Kilpatrick and others.
 * \version $Revision: 1.18 $
 * \date    $Date: 2001/02/23 19:02:34 $
 */

#include <wr_conf.h>

#if !defined(USER_MODE_TEST) && !defined(DBINTD_INTERFACE)
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#endif

#include <wsi_stdlib.h>


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

#ifdef PD_BZERO_IS_FN_PTR
#ifdef bzero
#undef bzero
#endif
#endif /* PD_BZERO_IS_FN_PTR */

/*  wsi_bzero */
/*! \brief Implementation of bzero
 *
 *  \param buf Pointer to buffer
 *  \param len Number of bytes in buffer
 */
void
wsi_bzero( void *buf, size_t len )
{
    bzero( buf, len );
}


/*  wsi_strlen */
/*! \brief Implementation of strlen
 * 
 *  \param s Pointer to string
 * 
 *  \return Number of characters in string
 */
size_t
wsi_strlen( const pd_char_t *s )
{
   return strlen( s );
}


/*  wsi_strcmp */
/*! \brief Implementation of strcmp
 *
 *  \param s1 Pointer to first string
 *  \param s2 Pointer to second string
 *
 *  \return Integer difference of strings
 */
int
wsi_strcmp( const pd_char_t *s1, const pd_char_t *s2 )
{
   return strcmp( s1, s2 );
}


/*  wsi_strncmp */
/*! \brief Implementation of strncmp
 *
 *  \param s1  Pointer to first string
 *  \param s2  Pointer to second string
 *  \param len Number of characters to compare
 * 
 *  \return Integer difference of strings 
 */
int
wsi_strncmp( const pd_char_t *s1, const pd_char_t *s2, size_t len )
{
   return strncmp( s1, s2, len );
}


/*  wsi_strcpy */
/*! \brief Implementation of strcpy
 *
 *  \param dst Pointer to destination buffer
 *  \param src Pointer to source string
 *
 *  \return Pointer to destination buffer
 */
pd_char_t *
wsi_strcpy( pd_char_t *dst, const pd_char_t *src )
{
   return strcpy( dst, src );
}


/*  wsi_strncpy */
/*! \brief Implementation of strncpy
 *
 *  \param dst Pointer to destination buffer
 *  \param src Pointer to source string
 *  \param len Maximum number of characters to copy
 *
 *  \return Pointer to destination buffer
 */
pd_char_t *
wsi_strncpy( pd_char_t *dst, const pd_char_t *src, size_t len )
{
   return strncpy( dst, src, len );
}


/*  wsi_strcat */
/*! \brief Implementation of strcat
 *
 *  \param dst Pointer to destination buffer
 *  \param src Pointer to source string
 *
 *  \return Pointer to destination buffer
 */
pd_char_t *
wsi_strcat( pd_char_t *dst, const pd_char_t *src )
{
   return strcat( dst, src );
}


/*  wsi_vsprintf */
/*! \brief Implementation of vsprintf 
 *  
 *  This function is provided for consistency in user-space and kernel-space.
 *  On FreeBSD, the kernel-space implementation of vsprintf (kvprintf)
 *  requires a radix and does not NULL-terminal the buffer.
 *
 *  \param s      Pointer to destination buffer
 *  \param format Pointer to format string
 *  \param ap     Variable arguments
 *
 *  \return Number of characters written to buffer
 */
int
wsi_vsprintf( pd_char_t *buf, const pd_char_t *format, va_list ap )
{
   int     ret;
   /* va_list ap;

   va_start( ap, format );
   */
#if !defined(USER_MODE_TEST) && !defined(DBINTD_INTERFACE)
   ret = kvprintf( format, NULL, buf, 10, ap );
   buf[ret] = PD_MAKESTR('\0');
#else
   ret = vsprintf( buf, format, ap );
#endif
   /* va_end( ap ); */

   return ret;
}


/*  wsi_toupper */
/*! \brief Implementation of toupper
 *
 *  \param c Character to convert to uppercase
 *
 *  \return Character converted to uppercase
 */
pd_char_t EXPORT
wsi_toupper( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('a') ) && ( c <= PD_MAKESTR('z') ) )?
            c - PD_MAKESTR('a') + PD_MAKESTR('A') :
            c;
}

/*  wsi_tolower */
/*! \brief Implementation of tolower
 *
 *  \param c Character to convert to lowercase
 *
 *  \return Character converted to lowercase
 */
pd_char_t EXPORT
wsi_tolower( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('A') ) && ( c <= PD_MAKESTR('Z') ) )?
            c - PD_MAKESTR('A') + PD_MAKESTR('a') :
            c;
}


#if !defined(USER_MODE_TEST) && !defined(DBINTD_INTERFACE)
/* wsi_util_sleep */
/*! \brief Implementation of tsleep for wrappers.
 *
 *  \param cookie Arbitrary 32 bit handle to identify the event.
 *
 *  \return 1 if interrupted, 0 if awakened.
 *
 * Puts the thread/process to sleep until Wakeup is called on the
 * same cookie.
 */
int EXPORT
wsi_util_sleep( long cookie )
{
    int retval;

    /* use PCATCH to allow signals.  timeout = 0 to not timeout */
    retval = tsleep((void*)cookie, PLOCK|PCATCH, "WRPSLP", 0);

    if (retval == EINTR || retval == ERESTART) /* Signals */
        return 1;
    /* if (retval == 0) */
    return 0;
}

void EXPORT
wsi_util_wakeup( long cookie )
{
    /* Not as a macro so that wss_exports has something to link against. */
    wakeup((void*)cookie);
}
#endif /* USER_MODE_TEST */
