/*
 * Copyright (C) 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_i386_hook.c,v 1.27 2001/04/25 21:08:26 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: This file contains the implementation of the runtime hooking
 *          of linux system calls routines.
 *          (Very Eeeevil code grubbing going on here.)
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>
#include <wsi_syscalls.h>
#include <wsi_hook.h>

#include <sys/errno.h>

#include <asm/unistd.h>

#include <linux/sys.h>
#include <linux/smp_lock.h>
#include <linux/net.h>

#include "wsi_intercept.h"

#ifdef __SMP__
#error SMP is not currently supported (do a little work hooking free_uid in release and things will be closer)
#endif


/***************************************************************************\
 * Macros                                                                  *
\***************************************************************************/

/*
 * Intel(tm) 80x86 opcodes
 */
#define PUSH_EBX           0x53
#define ADD_SIGN_RM_IMM    0x83
#define REL_MOV            0x8b
#define MOV_IMM_TO_RM      0xc7
#define NEAR_CALL          0xe8
#define JMP_DISP_REL_NEAR  0xe9
#define JMP_DISP_REL_SHORT 0xeb
#define JMP_ABS_IND        0xff


/*
 * MOD-R/M byte
 */
#define MOD_MASK 0x03
#define REG_MAKS 0x1c
#define RM_MASK  0xc0

#define MODRM_MOD(n) ((n) & 0x3)
#define MODRM_REG(n) (((n) & 0x07) << 2)
#define MODRM_RM(n) (((n) & 0x07) << 5)


/*
 * Bit patterns for registers
 */
#define BIT_REG_EAX 0x00        /* AL, AX, EAX, MM0 */
#define BIT_REG_ECX 0x01        /* CL, CX, ECX, MM1 */
#define BIT_REG_EDX 0x02        /* DL, DX, EDX, MM2 */
#define BIT_REG_EBX 0x03        /* BL, BX, EBX, MM3 */
#define BIT_REG_ESP 0x04        /* AH, SP, ESP, MM4 */
#define BIT_REG_EBP 0x05        /* CH, BP, EBP, MM5 */
#define BIT_REG_ESI 0x06        /* DH, SI, ESI, MM6 */
#define BIT_REG_EDI 0x07        /* BH, DI, EDI, MM7 */


/*
 * R/M32
 */
#define RM32_ESP        0xc4


/*
 * Debugging macros
 */
#ifdef DEBUG_HOOK
#  define DEBUG_HOOK_PRINT(msg) wsi_printf msg
#else
#  define DEBUG_HOOK_PRINT(msg)
#endif /* DEBUG_HOOK */

#ifdef DEBUG_NO_REPLACEMENT
#  define DEBUG_DO_REPLACE(store,value)
#  define DEBUG_DO_RESTORE_PAIR(name)
#else
#  define DEBUG_DO_REPLACE(store,value) do { *(store) = value; } while(0)
#  define DEBUG_DO_RESTORE_PAIR(name) \
      do { \
         if(STORAGEADDR(name)) \
            DEBUG_DO_REPLACE(STORAGEADDR(name), \
                             (caddr_t)((unsigned int)ORIGINAL(name) - \
                                       ((unsigned int)STORAGEADDR(name) + \
                                                      4))); \
      } while(0)
#endif /* DEBUG_NO_REPLACEMENT */


/*
 * Macro for call storage in saved_sysent
 */
#ifdef DEBUG_NO_REPLACEMENT
#  define DEBUG_DO_STORE(name)
#else
#  define DEBUG_DO_STORE(name) \
      do { \
         extern syscallptr saved_sysent[]; \
         if( ORIGINAL(name) ) \
            saved_sysent[__NR_ ## name] = (syscallptr) ORIGINAL(name); \
      } while(0)
#endif


/*
 * Macro for call replacement and storage code block used by routines below
 */
void DO_REPLACEMENT(pd_byte_t *,caddr_t *,caddr_t **,caddr_t *,caddr_t);
#define DO_REPLACEMENT(ip,loc,addrstore,origstore,replace) \
   do { \
      caddr_t   __relcall = *loc; \
      int       __offset; \
      caddr_t   __replacement = replace; \
      caddr_t **__addrstore = addrstore; \
      caddr_t  *__origstore = origstore; \
      DEBUG_HOOK_PRINT((PD_MAKESTR("Call to 0x%x\n"), \
                        (unsigned int)(ip + (int)__relcall))); \
      *__addrstore = loc; \
      *__origstore = (caddr_t)((unsigned int)ip + (unsigned int)__relcall); \
      __offset = (int)((unsigned int)__replacement - (unsigned int)ip); \
      DEBUG_DO_REPLACE(loc,(void *)__offset); \
      DEBUG_HOOK_PRINT((PD_MAKESTR("Replaced with 0x%x\n"), \
                        (unsigned int)(ip + __offset))); \
   } while(0)


/*
 * Macro to call DO_REPLACEMENT with a replacement pair
 */
#define DO_REPLACEMENT_PAIR(ip,loc,name) \
   DO_REPLACEMENT(ip,loc, &(STORAGEADDR(name)), &(ORIGINAL(name)), \
                  REPLACEMENT(name))
 

/***************************************************************************\
 * System call replacement pairs                                           *
\***************************************************************************/

/* For exit (stage 1 exit routine) */
DEFINE_REPLACEMENT(exit_thread);

/* For wait4 (stage 2 exit routine) */
DEFINE_REPLACEMENT(free_uid);

/* For exec */
DEFINE_REPLACEMENT(do_execve);

/* For fork */
DEFINE_REPLACEMENT(wake_up_process);

/* For splitting socketcall */
DEFINE_REPLACEMENT(getsockopt);
DEFINE_REPLACEMENT(recvfrom);
DEFINE_REPLACEMENT(socket);
DEFINE_REPLACEMENT(connect);
DEFINE_REPLACEMENT(send);
DEFINE_REPLACEMENT(getpeername);
DEFINE_REPLACEMENT(sendto);
DEFINE_REPLACEMENT(setsockopt);
DEFINE_REPLACEMENT(listen);
DEFINE_REPLACEMENT(accept);
DEFINE_REPLACEMENT(recv);
DEFINE_REPLACEMENT(getsockname);
DEFINE_REPLACEMENT(sendmsg);
DEFINE_REPLACEMENT(socketpair);
DEFINE_REPLACEMENT(recvmsg);
DEFINE_REPLACEMENT(shutdown);
DEFINE_REPLACEMENT(bind);


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern char _stext, _etext;
extern syscallptr sys_call_table[];


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static int
wsi_i386_hook_scan_table( caddr_t *table, int index, caddr_t replacement,
                          caddr_t *original, caddr_t **addr,
                          const pd_char_t *name )
{
   pd_byte_t *ptr;
   int        pc;
   caddr_t   *calladdr;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for %s at 0x%x\n"), name,
               (unsigned int) table[index-1] );
#endif /* DEBUG_HOOK */

   ptr = (pd_byte_t *)table[index-1];
   pc = 0;

#define SCAN_LIMIT 0x20
   for( ;
        (pc < SCAN_LIMIT) && !(
           /**** call <name> ****/
           (*ptr == NEAR_CALL) &&
           /**** jmp <end> ****/
           ( (*(ptr+5) == JMP_DISP_REL_NEAR) ||
             (*(ptr+5) == JMP_DISP_REL_SHORT) ||
           /**** addl <imm>,%esp ****/
             ( (*(ptr+5) == ADD_SIGN_RM_IMM) &&
               (*(ptr+6) == RM32_ESP) ) ||
           /**** movl <off>(%ebp),reg ****/
             ( (*(ptr+5) == REL_MOV ) &&
               ((*(ptr+6) & (MOD_MASK|RM_MASK)) ==
                   (MODRM_MOD(0x01)|MODRM_RM(BIT_REG_EBP)) ) ) )
        );
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to hook '%s'\n"), name );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }

#undef SCAN_LIMIT
   pc++, ptr++;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   DO_REPLACEMENT(ptr,calladdr,addr,original,replacement);

#ifndef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Hook[%s]: 0x%x\n"), name, (unsigned int) *original );
#endif /* !DEBUG_HOOK */

   return 0;
}

static int
wsi_i386_hook_sys_socketcall( void )
{
   pd_byte_t *ptr;
   caddr_t   *table;
   int        pc;
   pd_byte_t  modrm;
   pd_byte_t  sib;

   pc = 0;
   ptr = (pd_byte_t *)sys_call_table[__NR_socketcall];

#ifdef DEBUG_HOOK
   wsi_printf("sys_socketcall @ 0x%x\n", (unsigned int)ptr);
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x200
doscan:
   /*
    * Scan for a 'mov' followed by a relative jump
    */
   for( ; (pc < SCAN_LIMIT) && (*ptr != REL_MOV) && (*(ptr+4) != JMP_ABS_IND);
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
      return ENOENT;

#undef SCAN_LIMIT
   pc++, ptr++;

   /*
    * We expect the jump to be near and have a SIB
    */
   modrm = *(ptr+4);
   if( ( modrm & 0x38 ) == 0x28 )
   {
#ifdef DEBUG_HOOK_VERBOSE
      /* Far call (maybe) */
      wsi_printf( PD_MAKESTR("Far call relative jump encountered in ")
                  PD_MAKESTR("sys_socketcall hooking, continuing scan\n") );
#endif /* DEBUG_HOOK_VERBOSE */
      goto doscan;
   }

   if( ( modrm & 0xc7) != 0x04 )
   {
#ifdef DEBUG_HOOK_VERBOSE
       /* Not a SIB-modified effective address (maybe) */
      wsi_printf( PD_MAKESTR("Non SIB-modified effective address encountered ")
                  PD_MAKESTR("in sys_socketcall hooking, continuing scan\n") );
#endif /* DEBUG_HOOK_VERBOSE */
      goto doscan;
   }
   pc+=5, ptr+=5;
   sib = *ptr;

   pc++, ptr++;

   /*
    * The table is address is absolute, so we just grab it unmodified
    */
   table = *((caddr_t **)ptr);
#ifdef DEBUG_HOOK
   wsi_printf(PD_MAKESTR("Table at addr: 0x%x\n"), (unsigned int)table);
#endif /* DEBUG_HOOK */

   /*
    * Now we start the fun of hooking all the individual parts
    */
#define HOOK_SOCKETCALL(name, ndx) \
   do { \
      int _retc; \
      _retc = wsi_i386_hook_scan_table( table, (ndx), \
                                        REPLACEMENT(name), &(ORIGINAL(name)), \
                                        &(STORAGEADDR(name)), \
                                        PD_MAKESTR(#name) ); \
      if( _retc ) return _retc; \
   } while(0)

   HOOK_SOCKETCALL(getsockopt, SYS_GETSOCKOPT);
   HOOK_SOCKETCALL(recvfrom, SYS_RECVFROM);
   HOOK_SOCKETCALL(socket, SYS_SOCKET);
   HOOK_SOCKETCALL(connect, SYS_CONNECT);
   HOOK_SOCKETCALL(send, SYS_SEND);
   HOOK_SOCKETCALL(getpeername, SYS_GETPEERNAME);
   HOOK_SOCKETCALL(sendto, SYS_SENDTO);
   HOOK_SOCKETCALL(setsockopt, SYS_SETSOCKOPT);
   HOOK_SOCKETCALL(listen, SYS_LISTEN);
   HOOK_SOCKETCALL(accept, SYS_ACCEPT);
   HOOK_SOCKETCALL(recv, SYS_RECV);
   HOOK_SOCKETCALL(getsockname, SYS_GETSOCKNAME);
   HOOK_SOCKETCALL(sendmsg, SYS_SENDMSG);
   HOOK_SOCKETCALL(socketpair, SYS_SOCKETPAIR);
   HOOK_SOCKETCALL(recvmsg, SYS_RECVMSG);
   HOOK_SOCKETCALL(shutdown, SYS_SHUTDOWN);
   HOOK_SOCKETCALL(bind, SYS_BIND);
#undef HOOK_SOCKETCALL

   return 0;
}


#if PD_LINUX < 0x02004
static int
wsi_i386_hook_do_fork( pd_byte_t *ptr )
{
   int        pc;
   int        next_run;
   int        prev_run;
   caddr_t   *calladdr;

   pc = 0;

   next_run = offsetof(struct task_struct, next_run);
   prev_run = offsetof(struct task_struct, prev_run);

   /*
    * (2) in do_fork, look for the setting of p->next_run and
    *     p->prev_run to NULL and the next call found after that should be
    *     the call to wake_up_process
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for next_run and prev_run set to NULL ")
               PD_MAKESTR("at 0x%x\n"), (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x0F00
   for( ;
        (pc < SCAN_LIMIT) &&
           /**** movl $0x0,next_run(%r) ****/
           ( (*ptr != MOV_IMM_TO_RM) || (*(ptr + 2) != next_run) ||
               (*((int *)(ptr + 3)) != 0) ||
           /**** movl $0x0,prev_run(%r) ****/
             (*(ptr + 7) != MOV_IMM_TO_RM) || (*(ptr + 9) != prev_run) ||
               (*((int *)(ptr + 10)) != 0) );
        pc++, ptr++ )
   ;
   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add fork hook [next_run/prev_run]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }

   pc += 14, ptr += 14;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for wake_up_process at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call wake_up_process ****/
           (*ptr != NEAR_CALL);
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add fork hook [wake_up_process]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc++, ptr++;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   DO_REPLACEMENT_PAIR(ptr,calladdr,wake_up_process);

#ifndef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Hook[wake_up_process]: 0x%x\n"),
               (unsigned int) ORIGINAL(wake_up_process) );
#endif /* !DEBUG_HOOK */

   return 0;
}
#elif PD_LINUX == 0x02004 /* Linux 2.4 kernel */
static int
wsi_i386_hook_do_fork( pd_byte_t *ptr )
{
   int        pc;
   caddr_t   *calladdr;

   pc = 0;

   /*
    * (2) in do_fork, look for the call to send_sig, the next call after it
    *     should be wake_up_process
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for send_sig at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x0F00
   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call send_sig ****/
           ( (*ptr != NEAR_CALL) ||
             ((ptr + 5 + *((int *)(ptr + 1))) != (void *)send_sig) );
        pc++, ptr++ )
   ;
   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add fork hook [send_sig]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }

   pc += 5, ptr += 5;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for wake_up_process at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call wake_up_process ****/
           (*ptr != NEAR_CALL);
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add fork hook [wake_up_process]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc++, ptr++;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   DO_REPLACEMENT_PAIR(ptr,calladdr,wake_up_process);

#ifndef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Hook[wake_up_process]: 0x%x\n"),
               (unsigned int) ORIGINAL(wake_up_process) );
#endif /* !DEBUG_HOOK */

   return 0;
}
#endif

static int
wsi_i386_hook_sys_fork( void )
{
   int        pc;
   pd_byte_t *ptr;
   caddr_t   *calladdr;
   int        offset;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("-=- hook fork -=-\n") );
#endif /* DEBUG_HOOK */

   pc = 0;
   ptr = (pd_byte_t *)sys_call_table[__NR_fork];

   /*
    * For the wake_up_process hook in do_fork, we need to do two searches:
    *
    * (1) in sys_fork, find call to do_fork, which should be first
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for do_fork at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x20
   for( ; (pc < SCAN_LIMIT) && (*ptr != NEAR_CALL); pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add fork hook [do_fork]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc++, ptr++;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   offset = *((int *)calladdr); 
   ptr += offset;

   return wsi_i386_hook_do_fork( ptr );
}


static int
wsi_i386_hook_sys_exec( void )
{
   int        pc;
   pd_byte_t *ptr;
   caddr_t   *calladdr;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("-=- hook execve -=-\n") );
#endif /* DEBUG_HOOK */

   pc = 0;
   ptr = (pd_byte_t *)sys_call_table[__NR_execve];

   /*
    * For the do_execve hook in sys_execve, just look for the call to
    * do_execve, which is exported
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for do_execve at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x50
   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call do_execve ****/
           ( (*ptr != NEAR_CALL) ||
             ((ptr + 5 + *((int *)(ptr + 1))) != (void *)do_execve) );
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add execve hook [do_execve]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc++, ptr++;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   DO_REPLACEMENT_PAIR(ptr,calladdr,do_execve);

#ifndef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Hook[do_execve]: 0x%x\n"),
               (unsigned int) ORIGINAL(do_execve) );
#endif /* !DEBUG_HOOK */

   return 0;
}

static int
wsi_i386_hook_sys_exit( void )
{
   int        pc;
   pd_byte_t *ptr;
   caddr_t   *calladdr;
   int        offset;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("-=- hook exit -=-\n") );
#endif /* DEBUG_HOOK */

   pc = 0;
   ptr = (pd_byte_t *)sys_call_table[__NR_exit];

   /*
    * For the exit_thread hook in do_exit, we need to do two searches:
    * (1) in sys_exit, find call to do_exit, which should be first
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for do_exit at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x20
   for( ; (pc < SCAN_LIMIT) && (*ptr != NEAR_CALL); pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add exit hook [do_exit]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc++, ptr++;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   offset = *((int *)calladdr); 
   ptr += offset;
   pc = 0;

   /*
    * (2) in do_exit, look for a call to flush_signals, which is exported,
    *     and the next call found after that should be the call to exit_thread
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for flush_signals at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x500
   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call flush_signals ****/
           ( (*ptr != NEAR_CALL) ||
             ((ptr + 5 + *((int *)(ptr + 1))) != (void *)flush_signals) );
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add exit hook [flush_signals]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
   pc += 5, ptr += 5;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for exit_thread at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

   for( ; (pc < SCAN_LIMIT) && (*ptr != NEAR_CALL); pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add exit hook [exit_thread]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc++, ptr++;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   DO_REPLACEMENT_PAIR(ptr,calladdr,exit_thread);

#ifndef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Hook[exit_thread]: 0x%x\n"),
               (unsigned int) ORIGINAL(exit_thread) );
#endif /* !DEBUG_HOOK */

   return 0;
}

static int
wsi_i386_hook_sys_wait4( void )
{
   int        pc;
   pd_byte_t *ptr;
   caddr_t   *calladdr;
   int        offset;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("-=- hook wait4 -=-\n") );
#endif /* DEBUG_HOOK */

   pc = 0;
   ptr = (pd_byte_t *)sys_call_table[__NR_wait4];

   /*
    * For the free_uid hook in release, we need to do two searches:
    * (1) in sys_wait4, find call to notify_parent, which is exported, and the
    *     next call found after that should be the call to release
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for notify_parent at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x300
   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call notify_parent ****/
           ( (*ptr != NEAR_CALL) ||
             ((ptr + 5 + *((int *)(ptr + 1))) != (void *)notify_parent) );
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add wait4 hook [notify_parent]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
   pc += 5, ptr += 5;

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for release at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call release ****/
           (*(ptr + 1) != NEAR_CALL);
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add wait4 hook [release]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc += 2, ptr += 2;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   offset = *((int *)calladdr); 
   ptr += offset;
   pc = 0;


   /*
    * (2) in release, the first call found should be free_uid (unless SMP is
    *     enabled, then we would have to do a little more work, but right
    *     now we are not supporting SMP)
    */

#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Looking for free_uid at 0x%x\n"),
               (unsigned int) ptr );
#endif /* DEBUG_HOOK */

#define SCAN_LIMIT 0x20
   for( ;
        (pc < SCAN_LIMIT) &&
           /**** call free_uid ****/
           (*(ptr + 1) != NEAR_CALL);
        pc++, ptr++ )
   ;

   if( pc >= SCAN_LIMIT )
   {
#ifndef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("Unable to add wait4 hook [free_uid]\n") );
#endif /* DEBUG_HOOK */
      return ENOENT;
   }
#undef SCAN_LIMIT
   pc += 2, ptr += 2;

   calladdr = (caddr_t *)ptr;
   pc += 4, ptr += 4;

   DO_REPLACEMENT_PAIR(ptr,calladdr,free_uid);

#ifndef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("Hook[free_uid]: 0x%x\n"),
               (unsigned int) ORIGINAL(free_uid) );
#endif /* !DEBUG_HOOK */

   return 0;
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int
wsi_hook_init( void )
{
   int retc = 0;

   lock_kernel();

   retc = wsi_i386_hook_sys_fork();
   if( retc ) goto done_hook;
   retc = wsi_i386_hook_sys_exec();
   if( retc ) goto done_hook;
   retc = wsi_i386_hook_sys_exit();
   if( retc ) goto done_hook;
   retc = wsi_i386_hook_sys_wait4();
   if( retc ) goto done_hook;
   retc = wsi_i386_hook_sys_socketcall();

   DEBUG_DO_STORE(getsockopt);
   DEBUG_DO_STORE(recvfrom);
   DEBUG_DO_STORE(socket);
   DEBUG_DO_STORE(connect);
   DEBUG_DO_STORE(send);
   DEBUG_DO_STORE(getpeername);
   DEBUG_DO_STORE(sendto);
   DEBUG_DO_STORE(setsockopt);
   DEBUG_DO_STORE(listen);
   DEBUG_DO_STORE(accept);
   DEBUG_DO_STORE(recv);
   DEBUG_DO_STORE(getsockname);
   DEBUG_DO_STORE(sendmsg);
   DEBUG_DO_STORE(socketpair);
   DEBUG_DO_STORE(recvmsg);
   DEBUG_DO_STORE(shutdown);
   DEBUG_DO_STORE(bind);

done_hook:
   unlock_kernel();

   return retc;
}

int
wsi_hook_shutdown( void )
{
   lock_kernel();

   DEBUG_DO_RESTORE_PAIR(free_uid);
   DEBUG_DO_RESTORE_PAIR(exit_thread);
   DEBUG_DO_RESTORE_PAIR(do_execve);
   DEBUG_DO_RESTORE_PAIR(wake_up_process);
   DEBUG_DO_RESTORE_PAIR(getsockopt);
   DEBUG_DO_RESTORE_PAIR(recvfrom);
   DEBUG_DO_RESTORE_PAIR(socket);
   DEBUG_DO_RESTORE_PAIR(connect);
   DEBUG_DO_RESTORE_PAIR(send);
   DEBUG_DO_RESTORE_PAIR(getpeername);
   DEBUG_DO_RESTORE_PAIR(sendto);
   DEBUG_DO_RESTORE_PAIR(setsockopt);
   DEBUG_DO_RESTORE_PAIR(listen);
   DEBUG_DO_RESTORE_PAIR(accept);
   DEBUG_DO_RESTORE_PAIR(recv);
   DEBUG_DO_RESTORE_PAIR(getsockname);
   DEBUG_DO_RESTORE_PAIR(sendmsg);
   DEBUG_DO_RESTORE_PAIR(socketpair);
   DEBUG_DO_RESTORE_PAIR(recvmsg);
   DEBUG_DO_RESTORE_PAIR(shutdown);
   DEBUG_DO_RESTORE_PAIR(bind);

   unlock_kernel();

   return 0;
}

