/* 
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 ***************************************************************************
 *
 * binpatch.c
 *
 * A utility for patching function calls within the linux kernel
 * during run-time.  In this case, "patching" means replacing all
 * calls to some function old_foo() with calls to some other function
 * new_foo().  Both functions must take exactly the same arguments.
 * On each call, this utiltity patches all calls to old_foo() made by
 * one specified function, rather than all calls to old_foo() in the
 * entire binary.  So, with multiple calls, you can replace all calls
 * to old_foo() made by caller_one() with calls to new_foo(), and then
 * replace all calls to old_foo() made by caller_two() with calls to
 * even_newer_foo().
 *
 * So far, this code works only for "Direct Withing Segment" calls on
 * the i386.
 *
 ***************************************************************************/

#include "binpatch.h"  /* This file must be constructed by binpatch.pl */

#define FUNCTION_CALL 232  /* Direct Within Segment i386 CALL */

/* binpatch()
 *
 * in:     caller_start - the address at which the calling function 
 *                        starts.
 *         caller_end   - the address of the last byte of the calling 
 *                        function.
 *         old_callee   - the address of the callee function we want to
 *                        replace.
 *         new_callee   - the address of the function to replace 
 *                        `old_callee'.
 * out:    nothing
 * return: number of calls patched.
 *
 *
 */

int binpatch( unsigned char *caller_start, unsigned char *caller_end,
	      unsigned int old_callee, unsigned int new_callee ) {

  unsigned char *p_byte_in_text;  /* traverses text of calling function */
  int *p_offset_to_callee;        /* points to callee offsets */
  int num_patches_made = 0;       /* number of patches made */

  /* Scan through every byte in the calling function, that is, every *
   * byte from `caller_start' to `caller_end' inclusive, looking for *
   * the CPU-architecture-dependent bit pattern that indicates a     *
   * function call.                                                  */

  for( p_byte_in_text = caller_start; p_byte_in_text <= caller_end;
       p_byte_in_text++ ) {

    /* If `p_byte_in_text' points to a CALL instruction, we must   *
     * examine it further to determine whether or not it is a call *
     * to `old_callee'.                                            */
    
    if( *p_byte_in_text == FUNCTION_CALL ) {

      /* Yes, `p_byte_in_text' points to a function call.  The four
       * bytes after `p_byte_in_text contain the offset to the callee
       * function.  This offset is calculated as follows:
       * ( p_byte_in_text + 5 ) + ( signed offset ) == ( callee function )
       *
       * Place the offset into `*p_offset_to_callee' and determine if
       * is a call to `old_callee'.
       */

      p_offset_to_callee = (int *)(p_byte_in_text + 1);

      /* Addresses in the kernel are small.  Otherwise, we might have *
       * some overflow problems with the sign bit.                    */
      if( ( ( (int)(p_byte_in_text) + 5 ) + *p_offset_to_callee ) ==
	  old_callee ) {
	
	/* Yes, this is a call to `old_callee'.  We must replace the *
	 * offset at `p_offset_to_callee' to point to `new_callee'.  */

	*p_offset_to_callee = ( (int)(new_callee)  -
				( (int)(p_byte_in_text) + 5 ) );
	num_patches_made++;

      } /* if call to `old_callee' */

    } /* if call */

  } /* for all bytes in calling function's text */

  return( num_patches_made );

} /* binpatch() */


