/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_custom.c,v 1.7 2001/02/22 23:50:11 stevek Exp $
 *
 * Author: Douglas Kilpatrick
 *
 * Purpose: This file contains the implementation of the custom
 *          replacement linux system call routines.
 */

#include <wr_conf.h>

#include <linux/sched.h>

#include <wss_process.h>
#include <wsi_stdlib.h>
#include <wss_malloc.h>
#include <wss_path.h>
#include <wss_criteria.h>
#include <wss_wrapper.h>
#ifdef LOG_WRAPPED
#include <wss_log.h>
#endif

#include <wsi_syscalls.h>
#include <wsi_hook.h>


DECLARE_REPLACEMENT(wake_up_process);
DECLARE_REPLACEMENT(do_execve);
DECLARE_REPLACEMENT(free_uid);
DECLARE_REPLACEMENT(exit_thread);


/***************************************************************************\
 * Custom system call replacement functions                                *
\***************************************************************************/

/* To call the original function, you have to use the intermediate variable
 * to hold the function pointer.  Smells like an egcs bug.
 *
 * wake_up_process is a "FASTCALL", which means the first few arguments
 * all go via registers.  For some reason I don't understand, this is
 * not the normal calling convention.  To handle the hook properly, we
 * have to duplicate all that.
 */
void FASTCALL(wsi_custom_wake_up_process( struct task_struct *));
void
wsi_custom_wake_up_process( struct task_struct *p)
{
    __attribute__((regparm(3))) void (*volatile orig)(struct task_struct *) =
        ((void (*)(struct task_struct*))(ORIGINAL(wake_up_process)));
    wss_proc_fork(current, p, 0);

    orig(p);
}

int
wsi_custom_do_execve( char *filename, char **argv, char **envp,
                      struct pt_regs *regs )
{
    int (*volatile orig)(char*,char**,char**,struct pt_regs*) =
        ((int (*)(char *,char**,char**,struct pt_regs*))(ORIGINAL(do_execve)));

    struct wss_proc *wrp;
    proc_t          *p;
    int              err;
    int              act_error;
    int              wr_run_wrapped;
    pd_char_t       *tmp;
    int              len;
    int              i;

    /*
     * Look up the current process in the wrappers proc
     * structure. If it wasn't found, this means there was either
     * a problem, or this is a process that was running before the
     * wrappers module was loaded into the kernel.  In the first
     * case, this is a big problem.  In the second case, we can
     * probably work around it by creating the structure at this
     * time, but until I'm possitive, I'm leaving this as a TBD
     */
    p = current;
    wrp = wss_pfind( wsi_proc_pid(p) );
    if( !wrp ) {
        wsi_printf(
                PD_MAKESTR("wsi_custom_do_execve: can't find pid: p->pid %d\n"),
                wsi_proc_pid(current) );

        wrp = wss_proc_new( p );
    }

    /*
     * wrp was tested for validity early on.  Make space and copy
     * the path, first saving the current one in an 'old' pathbuf
     */
    if( wrp->wrp_oldpath )
        wss_free( wrp->wrp_oldpath );

    WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
    tmp = wrp->wrp_cwd;
    wrp->wrp_cwd = wss_path_cwd( wrp );
    if (!wrp->wrp_cwd) wrp->wrp_cwd = wss_strdup(PD_MAKESTR("/"));
    WSI_RW_EXIT( wrp->wrp_cwd );
    if( tmp )
        wss_free(tmp);

    wrp->wrp_oldpath = wrp->wrp_path; 
    WSI_RW_ENTER( wrp->wrp_cwd, RW_READER );
    wrp->wrp_path = wss_path_rel_to_abspath( wrp->wrp_cwd, filename);
    WSI_RW_EXIT( wrp->wrp_cwd );

    /* Clear out progname if its currently set */
    if( wrp->wrp_progname != NULL )
        wss_free( wrp->wrp_progname );

    /*
     * Generate the program name, full path should have at least
     * one '/' character.  Program name is everything after the
     * last '/'.
     */
    for( len = 0, i = wsi_strlen( wrp->wrp_path );
            ( i > 0 ) && ( wrp->wrp_path[i] != '/' );
            i--, len++ );

    /*
     * Check (for sanity, etc)
     */
    if( ( i == 0 ) && ( wrp->wrp_path[0] != '/' ) ) {
        wsi_printf(
                PD_MAKESTR(
                    "wsi_custom_do_execve: Pathname `%s' didn't have a '/' in it\n"),
                wrp->wrp_path );
        wrp->wrp_progname = NULL;
    } else {
        wrp->wrp_progname = (pd_char_t *)wss_malloc_string( len );
        bcopy( &( wrp->wrp_path[i + 1] ), wrp->wrp_progname,
                len * sizeof(pd_char_t) );
    }

    /*
     * Next, call the original syscall.
     */ 

#if 0
    wsi_printf(PD_MAKESTR("wsi_custom_do_execve: original syscall.\n"));
#endif

    err = orig(filename, argv, envp, regs);

#if 0
    wsi_printf(PD_MAKESTR("wsi_custom_do_execve: after orig syscall.\n"));
#endif

    /* zero from execve syscall indicates success. */
    if( err >= 0 ) { 
        int has_wrappers;
        /*
         * We may now check the activation criteria to determine
         * whether or not to run this process wrapped.
         */
#if 0
        wsi_printf("calling check_deact\n");
#endif
        wss_crit_check_deactivation( wrp );
#if 0
        wsi_printf("calling check_act\n");
#endif
        wr_run_wrapped = wss_crit_check_activation( wrp );

#if 0
        wsi_printf(PD_MAKESTR("wsi_custom_do_execve: run_wrapped = %d\n"),
                wr_run_wrapped);
#endif
        WSS_MUTEX_ENTER( wrp->wrp_wi_list );
        has_wrappers = (wrp->wrp_wi_list == NULL) ? FALSE : TRUE;
        WSS_MUTEX_EXIT( wrp->wrp_wi_list );

        if( has_wrappers ) {
            /*
             * Now is a convenient place to call the
             * wrapper-specific wr_activation() functions.
             */
            act_error = wss_wrapper_call_activation( wrp );
            if( act_error != 0 ) {
                /*
                 * TBD: Is there anything else we should
                 * do here, like kill the process?
                 */
                wsi_printf( PD_MAKESTR("Activate returned an error!\n") );
            }
        }
    }

#ifdef LOG_WRAPPED
    (void)wss_log_printf(
                         PD_MAKESTR("Running %s %swrapped\n"),
                         filename,
                         ( wrp->wrp_wi_list?PD_MAKESTR(""):PD_MAKESTR("un")));
#endif /* LOG_WRAPPED */

    wss_proc_rele( wrp );

#if 0
    wsi_printf("Exiting exec.\n");
#endif
    return err;
}

void
wsi_custom_free_uid(struct task_struct *p)
{
    void (* volatile orig)(struct task_struct *) =
        ((void(*)(struct task_struct *))(ORIGINAL(free_uid)));

    wss_proc_exit_stage2( wsi_proc_pid(p) );

    orig(p);
}

void
wsi_custom_exit_thread(void)
{
    void (*volatile orig)(void) = ((void(*)(void))(ORIGINAL(exit_thread)));
    /* TBD: Thread issues.  Rethink for linux model */
    /* Call the stage 1 exit code */
    wss_proc_exit_stage1( current );

    orig();
}

