/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_db.c,v 1.9 2001/02/23 19:02:37 stevek Exp $
 *
 * Author: Tim Fraser
 *
 * This is the system-specific db file.  This handles flushing the DB to 
 * file since our mmap-implimentation that the DB uses is completely fake.
 */

#include <wr_conf.h>
#include <wss_db.h>
#include <wsi_stdlib.h>
#include <wss_file.h>
#include <wsi_file.h>
#include <wss_vm.h>

#include <linux/unistd.h>

/* flush_block()
 *
 * in:     p_first - address of the first byte to write to the filesystem.
 *         p_last  - address of the last  byte to write to the filesystem.
 * out:    nothing
 * return: nothing
 *
 * This is a convenience function, allowing us to write a series of
 * bytes to the filesystem by specifying the addresses of the first
 * and last byte.
 *
 */
static void flush_block(wss_vmspace_t *space, char *p_first, char *p_last )
{
    size_t      size;
    size_t      count;
    int         retval;
    off_t       offset;

    offset = (off_t)abs2rel(space, p_first);
    size        = (size_t)((p_last - p_first) + 1);

#ifdef DEBUG_DBFLUSH
    wsi_printf( "flushing %p - %p inclusive: size = %u/%x, offset = %d/%x\n",
            p_first, p_last,
            size, size,
            (int)offset, (unsigned int)offset);
#endif

    retval = wsi_file_seek(
            space->vms_file,
            offset,
            WSS_FILE_SEEK_SET );
    if (retval < 0) {
        wsi_printf("seek failed: %d\n", retval);
        return;
    }

    count       = 0;
    while (count < size) {
        count += wsi_file_write( space->vms_file, p_first + count, size-count);
    }
} /* flush_block() */


/*
 * Algo:
 * !) Assumes free list is sorted
 *
 * flush from start of DB to first free block.
 * From from end of free-block to start of next free-block
 * Flush from end of last free-block to end of DB file.
 *
 * Ignore the actual free list data, because vmck recreates it anyway.
 */
void wsi_db_flush(wss_vmspace_t *space)
{
    char       *first_block;
    char       *block;
    char       *next_block;
    char       *end_ptr;

#ifdef DEBUG_DBFLUSH
    wsi_printf("Flushing head structure.\n");
#endif
    /* Prolog:
     *  Start to first-free.  This will get all the header information.
     */
    first_block = wss_vm_first_free_block(space);
    if (!first_block)
        return;

    flush_block(space, (char*)space->vms_head, first_block - 1);

#ifdef DEBUG_DBFLUSH
    wsi_printf("Flushing used data.\n");
#endif
    /*
     * Main loop
     *  From end of freeblock to start of next
     */
    block = first_block;
    next_block = wss_vm_next_free_block(space, block);

    while (next_block && next_block != first_block) {
        flush_block( space, block+wss_vm_free_block_size(block), next_block-1);

        block           = next_block;
        next_block      = wss_vm_next_free_block(space,block);
    }

    /*
     * Epilog:
     *  Next block is bogus, so write from block to end.
     */
    
#ifdef DEBUG_DBFLUSH
    wsi_printf("Flushing end data.\n");
#endif
#if 1
    end_ptr = (char*)space->vms_head + space->vms_size - 1;
#else
    end_ptr = (char*)(((struct wsi_file *)space->vms_file)->mem_addr) + 
                     (((struct wsi_file *)space->vms_file)->size) - 1;
#endif
    if (block + wss_vm_free_block_size(block) < end_ptr) {
        flush_block(space, block + wss_vm_free_block_size(block), end_ptr);
    }

#ifdef DEBUG_DBFLUSH
    wsi_printf("Flush of DB complete.\n");
#endif
}
