/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_file.c,v 1.14 2001/04/23 20:58:28 stevek Exp $
 *
 * Author: Linda Thomas, Steve Kiernan
 *         Linux hacks resembling a port - Tim Fraser
 *
 * Purpose: This file contains file i/o functions.
 *
 * Notes on the Linux port:
 *
 * This file contains functions to do file I/O to and from buffers in
 * kernel memory.  Two kinds of I/O are "supported":
 * 1) Non-memory-mapped file I/O via open [read|write]+ close.
 * 2) Simulated memory-mapped file I/O via
 *    open map <memory twiddling> unmap close.
 *
 * Unlike BSD, Linux does not provide pageable kernel memory, and
 * cannot mmap files into kernel memory.  So, these functions
 * simulated mmap functionality by allocating a non-pageable buffer in
 * kernel memory and copying the file into it on mmap, and by writing
 * the data back out and freeing the buffer on unmap.  Note that there
 * is no real backing store for files "mapped" by this simulated
 * memory-mapping technique: no data is written back to the filesystem
 * until the file is unmapped.
 *
 * Unlike BSD, Linux does not provide a UIO abstraction on top of the
 * vnode layer which allows the caller to request filesystem I/O to
 * and from kernel-space buffers rather than user-space buffers.  In
 * fact, the copyin/copyout operations are hard-coded into the vnode
 * operations in Linux.  So, in order to do filesystem I/O to and from
 * kernel buffers, these function must fool with the kernel's memory
 * protection mechanism in a rather strange way.  See note #1 below.
 *
 */

#include <wsi_types.h>      /* for pd_offset_t and kernel types */
#include <wsi_file.h>       /* our exported interface */
#include <wss_file.h>       /* for WSS_FILE_* flags */
#include <wsi_stdlib.h>
#include <wss_malloc.h>


#include <linux/smp_lock.h> /* for lock_kernel() and unlock_kernel() */
#include <linux/mm.h>       /* for GFP_KERNEL */
#include <linux/fs.h>       /* for struct inode, IS_ERR macro */
#include <linux/vmalloc.h>  /* for vmalloc(), vfree() */
#include <asm/fcntl.h>      /* for file open access flag defines */
#include <asm/uaccess.h>    /* for get_fs(), set_fs(), KERNEL_DS */

/* not exactly sure about the correct version number. */
#if (PD_LINUX < 0x02004) && (LINUX_VERSION_CODE > KERNEL_VERSION(2,2,13))
#define EVENT global_event
#else
#define EVENT event
#endif

#if (PD_LINUX < 0x02004)
#define file_count(f)        ((f)->f_count)
#define get_file(f)          (((f)->f_count)++)
#define dec_and_test_file(f) ((file_count(f) > 1)?((f)->f_count--),FALSE:TRUE)
#else
#define dec_and_test_file(f) (atomic_dec_and_test(&(f)->f_count))
#endif

/* Note 1:
 *
 * Re: get_fs(), set_fs(), copy_from/to_user(), and kernel file I/O.
 *
 * In Linux, the write vnode operation is hard-coded to do a
 * copy_from_user() on the source buffer.  To prevent a malicious
 * application program from tricking the kernel into copying data
 * from arbitrary kernel data structures, copy_from_user() checks
 * that the address of the source buffer is located in user-space,
 * not kernel-space.  It uses the `addr_limit' field of the current
 * process's task_struct to determine where the boundary between
 * kernel-space and user-space is (usually 0xBFFFFFFF on 32bit cpus).
 *
 * Our source buffer resides in kernel-space, so we must disable
 * copy_from_user()'s address check.  We do this by temporarily
 * setting the current process's `addr_limit' to the maximum
 * possible address (0xFFFFFFFF) for the duration of the write
 * operation.  This effectively makes copy_from_user()'s check
 * believe that the entire virtual memory range is user-space, and
 * the check will always pass.  We use get_fs() and set_fs() to
 * cause the change, and to restore the original values later.
 *
 * Note that this action temporarily disables part of the kernel's
 * protection mechanism for the kernel thread using the modified
 * task_struct.  Other threads should not be affected.  Proving that
 * the thread is unable to misuse its temporarily increased
 * privileges is non-trivial, since (I believe) that interrupts may
 * cause it to execute instructions other than those explicitily
 * listed between the set_fs() calls.
 *
 * See <asm/uaccess.h>.
 *
 */



/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

/* The BSD version of this code used the standard BSD file structure
 * (renamed "wsi_file") to represent files.  Since the standard BSD
 * kernel mechanisms were sufficient to accomplish kernel/file I/O and
 * mmap'ing, the BSD version had no need to add new fields to the
 * existing data structures, and the standard file structure worked
 * just fine.  Unfortunatley, the Linux kernel can't map a file into
 * kernel memory, so we're going to have to simulate this
 * functionality using the more primitive kernel/file I/O operations
 * that Linux does provide.  
 *
 * This need for simulation presents us with two choices: (1) use the
 * existing Linux data structures (specifically `struct
 * vm_area_struct') to support our simulated mmap'ing, or (2) invent
 * some new data structures.  I'm too much of a coward to use the
 * existing ones - the entries representing our simulated mmaps would
 * be visible to the existing kernel code, which might become confused
 * by these seemingly invalid entries.  So I'm going to take option 2
 * and invent a new data structure to support our simulation, as
 * described below:
 *
 *
 * We use the `wsi_file' structure to represent a file opened for
 * kernel I/O.  It encapsulates a reference to the normal Linux
 * `struct file' to represent the file, and adds some extra fields to
 * support our kernel-space mmap simulation.  The BSD version of this
 * structure included a mutex lock field.  The linux version does not,
 * since all of the I/O functions aquire the global kernel lock before
 * operating on these structures.
 *
 * We use the normal kernel mechanisms to manage the allocation and
 * deallocation of the `struct file' part of this structure.  We
 * manage the allocation and deallocation of the encapsulating `struct
 * wsi_file' structures ourselves.
 *
 * Note that these functions allow you to "mmap" a portion of a file
 * beginning at a specified offset (`file_offset') and running for a
 * specified length (`size').  Neither the offset nor the length may
 * change while the portion of the file is mapped.
 * 
 * Also note that *NIX filesystem semantics will not be preserved if
 * you separately map overlapping portions of the same file - whoever
 * writes last will elminate the data written by whoever writes first.
 *
 * You must acquire the global kernel lock before using instances of
 * this struct.
 * */

typedef struct wsi_file
{
  struct file *p_file;      /* file structure representing actual file */
  void        *mem_addr;    /* We've "mapped" the file to this kern address */
  size_t       size;        /* The mapped portion of the file is `size' long */
  off_t        file_offset; /* Mapped portion is this many bytes into file */
} *wsi_file_t;


/***************************************************************************\
 * Private functions                                                        *
\***************************************************************************/


/* wsi_file_free()
 *
 * in:     handle - a pointer to a wsi_file struct representing an
 *                  open file.
 * out:    handle - frees wsi_file struct.
 * return: nothing
 *
 * Really close the file
 *
 * The wsi_file_rele() function calls this one to free the resources
 * used to represent a file once its reference count drops below 1.
 *
 * This function closes an open file and frees the `wss_file_t'
 * structure that used to represent it.  It will yack if the file
 * indicated by `handle' not currently open by exactly one process.
 * It will also yack if the file is currently "mmap"d by the simulated
 * memory mapping routines contained in this file.
 *
 * This function assumes the caller is holding the kernel lock.
 *
 * 
 */

void
wsi_file_free( wss_file_t handle )
{
   wsi_file_t p_wsi_file  = (wsi_file_t)handle; 

   assert( p_wsi_file != NULL );
   assert( p_wsi_file->p_file != NULL );
   assert( p_wsi_file->mem_addr == 0 );
   assert( p_wsi_file->size == 0 );
   assert( p_wsi_file->file_offset == 0 );
   assert( file_count( p_wsi_file->p_file ) == 1 );   /* f_refcount in BSD */


   filp_close( p_wsi_file->p_file, NULL );
   wss_free( p_wsi_file );
} /* wsi_file_free() */


/*
 * wsi_file_ref()
 *
 * in:     handle - a pointer to a wsi_file struct representing an open file.
 * out:    handle - wsi_file struct with reference count incremented.
 * return: nothing
 *
 * Description:
 *    Acquire reference to file sturcture
 *
 *    Increases the reference count of file indicated by `handle'.
 *    This reference count is the same one used by the kernel to count
 *    the users of a given file.  For most operations, such as
 *    wsi_file_open(), the kernel will increment the reference count
 *    correctly.  This function allows the WSI to increment the
 *    reference count in cases where the kernel won't do it for us.
 *    This function assumes the caller is holding the kernel lock.
 */

void
wsi_file_ref( wss_file_t handle )
{
   wsi_file_t p_wsi_file = (struct wsi_file *)handle;
   
   if( file_count( p_wsi_file->p_file ) < 1 )
     panic( "wsi_file_ref: reference count for 0x%x was %d < 1", 
	    (unsigned int)p_wsi_file,
	    file_count( p_wsi_file->p_file ) );

   get_file( p_wsi_file->p_file );

} /* wsi_file_ref() */

                               
/*
 * wsi_file_rele()
 *
 * in:     handle - a pointer to a wsi_file struct representing an open file.
 * out:    handle - wsi_file struct with reference count decremented, or
 *                  frees the wsi_file struct
 * return: nothing
 *
 * Description:
 *    Release reference to file sturcture
 *
 *    Decreases the reference count of file indicated by `handle'.
 *
 *    If the reference count falls below 1, this function assumes
 *    all processes are done using the file and calls wsi_file_free()
 *    to close the file and free the wsi_file structure used to
 *    represent it.
 *
 *    This function assumes the caller is holding the kernel lock.
 */

void
wsi_file_rele( wss_file_t handle )
{
   struct wsi_file *p_wsi_file = (struct wsi_file *)handle;

   if( file_count( p_wsi_file->p_file ) < 1 )
      panic( "wsi_file_rele: reference count for 0x%x was %d < 1",
	     (unsigned int)p_wsi_file,
	     file_count( p_wsi_file->p_file ) );

   if( dec_and_test_file( p_wsi_file->p_file ) )
      wsi_file_free( handle );

} /* wsi_file_rele() */


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/* wsi_file_open()
 *
 * in:     path - pathname of file.
 *         mode - permissions we want for the opened file, 
 *                defined as WSS_FILE_* macros in wss_file.h .
 * out:    nothing
 * return: pointer to a wss_file structure corresponding to `path' or
 *         NULL if unable to open `path'.
 *
 * Construct wsi_file based on a path and open it.
 *
 * This function allocates a wsi_file structure.  
 * 
 * This function expects its arguments to reside in kernel-space.
 *
 * This function aquires and releases the kernel lock.  This function
 * may block on disk I/O while holding this lock.
 */

wss_file_t
wsi_file_open( const pd_char_t *path, wss_file_mode_t mode )
{
   int filemode = 0;            /* linux equivalent of `mode' flags */
   struct file *p_file;         /* file corresponding to `path' */
   wsi_file_t p_wsi_file;       /* wsi representation of `path's file */
   int err;                     /* holds error return values */

#ifdef DEBUG_FILE
   wsi_printf( PD_MAKESTR("wsi_file_open( '%s', 0x%x )\n"), path, mode );
#endif

   /* Convert the WSS-specific flags in `mode' to kernel-specific *
    * flags in `filemode'.                                        */
    
   if( mode & WSS_FILE_READ )
   {
      if( mode & WSS_FILE_WRITE )
         filemode |= O_RDWR;
      else
         filemode |= O_RDONLY;
   }
   else if( mode & WSS_FILE_WRITE )
      filemode |= O_WRONLY;
   if( mode & WSS_FILE_APPEND )
      filemode |= O_APPEND;
   if( mode & WSS_FILE_TRUNC )
      filemode |= O_TRUNC;
   if( mode & WSS_FILE_CREAT )
      filemode |= O_CREAT;

   /* Open the file corresponding to `path'. This should give us a *
    * `struct file' structure representing the file.               */

   lock_kernel(); {
     /* Ask the kernel to open the file and increment its reference count. */
     p_file = filp_open( path, filemode, 0x00 );
   } unlock_kernel();

   if( IS_ERR( p_file ) ) {
     err = PTR_ERR( p_file );
#ifdef DEBUG_FILE
     wsi_printf( PD_MAKESTR("create result %d\n"), err );
#endif
     return( NULL );
   }

   /* Allocate a new `struct wss_file' to represent the newly opened *
    * file and return it to the caller.                              */
   if( !( p_wsi_file = (wsi_file_t)wss_malloc( sizeof( struct wsi_file )) ) ) {
     lock_kernel(); {
       filp_close( p_file, NULL );
     } unlock_kernel();
     return( NULL );
   }

   /* We won't lock here because, since we are just creating   *
    * `p_wsi_file' now, we're the only one who knows about it, *
    * and there can't be any problems resulting from sharing.  */
   p_wsi_file->p_file = p_file;
   p_wsi_file->mem_addr = 0;
   p_wsi_file->size = 0;
   p_wsi_file->file_offset = 0;

   return( (wss_file_t)p_wsi_file );

} /* wsi_file_open() */



/* wsi_file_close()
 *
 * in:     handle - pointer to a wsi_file struct representing open file
 * out:    handle - unmaps file, decrements reference count, may free
 *                  the wsi_file structure.
 * return: nothing
 *
 *    Close file previously opened
 *
 *    Unmaps the file indicated by `handle', and decreases its
 *    reference count by one.  If the reference count falls below
 *    1, this function may free the wsi_file struct indicated by `handle'.
 *
 *    This function acquires and releases the kernel lock.
 */

void
wsi_file_close( wss_file_t handle )
{
   struct wsi_file *p_file = (struct wsi_file *)handle;

   if( p_file ) {
      if( p_file->mem_addr )
         wsi_file_unmap( handle );

      lock_kernel(); {
	wsi_file_rele( handle );
      } unlock_kernel();
   }
}



/* wsi_file_write()
 *
 * in:     handle - pointer to a wsi_file struct representing an open file.
 *         buffer - buffer of data to write to the file indicated by `handle'.
 *         len    - number of bytes to write from `buffer'.
 * out:    handle - the act of writing to the open file updates our current
 *                  position in the open file.
 * return: value        condition
 *         -----        ---------
 *         >= 0         number of bytes written
 *         <  0         error code * (-1)
 *
 * Write data to open file
 *
 * This function partially disables the kernel's protection mechanisms.
 * See note 1, above.
 *
 * This function aquires and releases the kernel lock.  This function
 * may block on disk I/O while holding this lock.
 *
 */

int wsi_file_write( wss_file_t handle, pd_byte_t *buffer, size_t len ) {

  struct file *p_file = ((struct wsi_file *)handle)->p_file;    /* the file */
  mm_segment_t max_address;  /* max virtual address useable by current proc */
  int ret_val;               /* holds error return values */

  if( !p_file ) {            /* sanity check */
    return( -EINVAL );
  }

  lock_kernel(); {

    max_address = get_fs();  /* See note 1, above. */
    set_fs( KERNEL_DS );

    ret_val = p_file->f_op->write( p_file, buffer, len, &(p_file->f_pos) );

    set_fs( max_address );   /* See note 1, above. */
    
  } unlock_kernel();

  return( ret_val );

} /* wsi_file_write() */


/* wsi_file_read()
 *
 * in:     handle - pointer to a wsi_file struct representing an open file.
 *         len    - number of bytes to read from the file indicated 
 *                  by `handle'.
 * out:    handle - the act of reading from the file updates our current
 *                  position in the file.
 *         buffer - will hold the data read from the file.
 * return: value       condition
 *         -----       ---------
 *         >=0         number of bytes read
 *         < 0         error code * -1         
 *
 * Read data from open file
 *
 * This function partially disables the kernel's protection mechanisms.
 * See note 1, above.
 *
 * This function aquires and releases the kernel lock.  This function
 * may block on disk I/O while holding this lock.
 *
 */

int
wsi_file_read( wss_file_t handle, pd_byte_t *buffer, size_t len )
{

  struct file *p_file = ((struct wsi_file *)handle)->p_file;    /* the file */
  mm_segment_t max_address;  /* max virtual address useable by current proc */
  int ret_val;               /* holds error return values */

  if( !p_file ) {            /* sanity check */
    return( -EINVAL );
  }

  lock_kernel(); {

    max_address = get_fs();  /* See note 1, above. */
    set_fs( KERNEL_DS );

    ret_val = p_file->f_op->read( p_file, buffer, len, &(p_file->f_pos) );

    set_fs( max_address );   /* See note 1, above. */
    
  } unlock_kernel();

  return( ret_val );

} /* wsi_file_read() */




/* wsi_file_seek()
 *
 * in:     handle - pointer to wsi_file struct representing an open file.
 *         offset - desired offset relative to `whence'.
 *         whence - indicates point of reference for `offset'.
 * out:    handle - this function changes current offset in file.
 * return: the new offset, or -EINVAL on error.
 *
 * This function changes the current offset in the file indicated by
 * `handle'.  It's arguments are similar to those of the lseek() C function.
 *
 * This function acquires and releases the kernel lock.
 *
 */

int
wsi_file_seek( wss_file_t handle, pd_offset_t offset, int whence )
{
    loff_t              retval;
    struct file        *file;
    mm_segment_t        max_address;

    if (!handle)
        return -EINVAL;

    file = ((struct wsi_file *)handle)->p_file;   /* the file */

    if( !file || whence > 2 || whence < 0)
        return -EINVAL;

    lock_kernel();
    if (file->f_op && file->f_op->llseek) {
        max_address = get_fs();
        set_fs( KERNEL_DS );

        /* We luck out in that the WSS_FILE_SEEK_* macros are the
         * same values the linux kernel uses
         */
        retval = file->f_op->llseek(file,offset,whence);

        set_fs( max_address );
    } else {
        /* Default implimentation.  Cut and paste from fs/read_write.c:16
         */
        switch (whence) {
        case WSS_FILE_SEEK_END:
            offset += file->f_dentry->d_inode->i_size;
            break;
        case WSS_FILE_SEEK_CUR:
            offset += file->f_pos;
        /* default: WSS_FILE_SEEK_SET */
        }
        retval = -EINVAL;
        if (offset >= 0) {
            if (offset != file->f_pos) {
                file->f_pos = offset;
                file->f_reada = 0;
                file->f_version = ++EVENT;
            }
            retval = offset;
        }
    }
    unlock_kernel();

    return retval;
}

/* wsi_file_get_size()
 *
 * in:     handle - pointer to a wsi_file struct representing an open file.
 * out:    error  - set to -EINVAL on error, 0 otherwise.
 * return: length of file in bytes, or 0 on error.
 *
 * Return the size of the file
 *
 * This function acquires and releases the kernel lock.
 *
 */

size_t
wsi_file_get_size( wss_file_t handle, int *error )
{
  struct file *p_file = ((struct wsi_file *)handle)->p_file;    /* the file */
  off_t size;               /* file size */

#ifdef DEBUG_FILE
  wsi_printf("wsi_file_get_size: handle = %p, p_file = %p\n", handle,p_file);
#endif
  if( !p_file ) {            /* sanity check */
    if (error)
      *error = -EINVAL;
    return( 0 );
  }

  lock_kernel(); {  
      size = p_file->f_dentry->d_inode->i_size;
  } unlock_kernel();

  if (error)
    *error = 0;
  return( size );

} /* wsi_file_get_size() */


/* wsi_file_get_map_addr()
 *
 * in:     handle - a pointer to a wsi_file struct representing an open
 *                  and "mapped" file.
 * out:    nothing
 * return: address at which the file indicated by `handle' is mapped,
 *         or 0 if it is not mapped.
 *
 * This function acquires and releases the kernel lock.
 */

void *
wsi_file_get_map_addr( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;
   void            *addr = NULL;

   if( !file ) {
     return( NULL );
   }

   lock_kernel(); {
     addr = file->mem_addr;
   } unlock_kernel();

   return addr;
}


/* wsi_file_map()
 *
 * in:     handle - a pointer to a wsi_file struct representing an open file.
 *         size   - the size of the portion of the file to map.  Note that
 *                  the size of this portion cannot change while it is
 *                  mapped.
 *         mapmode - ignored.
 *         pos     - the portion to map begins at the file offset `pos'.
 * out:    handle  - the function fills in the wsi_file fields related 
 *                   to memory mapping.
 * return  value    condition
 *         -----    ---------
 *           0      mapping is OK.
 *          < 0     error value
 *
 * This function simulates the ability to memory-map a portion of a file.
 * It implements the following algorithm:
 *
 * 1.  Allocate a chunk of memory to hold the "mapped" portion of the file.
 *     The chunk should be `size' bytes big.
 * 2.  Read `size' bytes of the file indicated by `handle' into the chunk
 *     of memory allocated in step 1, beginning `pos' bytes from the start
 *     of the file.
 * 3.  Record the location and size of the memory chunk, and the `pos' file
 *     offset in the wsi_file structure indicated by `handle'.
 *
 * Note that the return value behavior of this function is not the same
 * as the original BSD version of this function.  The original version
 * returned the address of the memory chunk (an unsigned value) or an
 * error code (a signed, negative value) on error.  Unfortunately, the
 * addresses in kernel memory are so large that the highest bit is often
 * set - this can interpreted as a negative integer and mistaken for
 * an error code.
 *
 * This function aquires and releases the kernel lock.  This function
 * may block on disk I/O while holding this lock.
 *
 */

int
wsi_file_map( wss_file_t handle, size_t size, int mapmode, pd_offset_t pos )
{
   struct wsi_file *file  = (struct wsi_file *)handle;
   void *buffer;             /* file portion "mapped" into this buffer */
   int ret_val;              /* holds function return values */

   /* 1.  Allocate a chunk of memory to hold the "mapped" portion of *
    *     the file.  The chunk should be `size' bytes big.           */

   if( !( buffer = (void *)vmalloc( size ) ) ) {
     return( -ENOMEM );
   }

   memset(buffer, 0, size); /* Zero-init buffer */
   /* 2.  Read `size' bytes of the file indicated by `handle' into the *
    *     chunk of memory allocated in step 1, beginning `pos' bytes   *
    *     from the start of the file.                                  */

   ret_val = wsi_file_seek( handle, pos, WSS_FILE_SEEK_SET );
   if( pos != ret_val ) {
     vfree( buffer );
     return( ret_val );
   }
   
   ret_val = wsi_file_read( handle, buffer, size );
   if( ret_val < 0 ) {
     vfree( buffer );
     return( ret_val );
   }


   /* 3.  Record the location and size of the memory chunk, and the *
    *     `pos' file offset in the wsi_file structure indicated by  *
    *	  `handle'.                                                 */
   lock_kernel(); {
     file->mem_addr = buffer;
     file->size = size;
     file->file_offset = pos;
   } unlock_kernel();

   return( 0 );

} /* wsi_file_map() */



/* wsi_file_unmap()
 *
 * in:     handle - a pointer to a wsi_file struct representing an open file.
 * out:    handle - this function zeroes the fields related to "mapping".
 * return: nothing
 *
 *     This function simulates the un-memory-mapping of a file.  It
 * writes the data in the memory buffer indicated by `handle' back out
 * to the file, frees the buffer, and zeroes out `handle's
 * "mapping"-related fields.
 *
 * This function aquires and releases the kernel lock.  This function
 * may block on disk I/O while holding this lock.
 *
 */

void
wsi_file_unmap( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( !file ) {       /* sanity check */
     return;
   }

   /* Make sure file is "mapped". */
   lock_kernel(); {
     if( !(file->mem_addr) ) {
       unlock_kernel();
       return;
     }
   } unlock_kernel();

   /* Write data in `file->mem_addr' buffer back out to the filesystem. */
   wsi_file_seek( handle, file->file_offset, WSS_FILE_SEEK_SET );
   wsi_file_write( handle, file->mem_addr, file->size );

   /* *Free The Mallocs!* */
   vfree( file->mem_addr );

   /* clear the "mapping"-related fields */
   lock_kernel(); {
     file->mem_addr = 0;
     file->size = 0;
     file->file_offset = 0;
   } unlock_kernel();

} /* wsi_file_unmap() */


