/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_fileutil.c,v 1.18 2001/04/23 20:59:02 stevek Exp $
 *
 * Author: Calvin Ko
 *
 * Purpose: Contain functions for network and files
 *
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>
#include <wsi_fileutil.h>
#include <wsi_path.h>
#include <linux/kdev_t.h> /* for kdev_t_to_nr  */
#include <linux/mm.h>     /* for GFP_KERNEL    */
#include <linux/fs.h>     /* for namei, lnamei */
#include <linux/file.h>

#define MAX_SOCK_ADDR 128 /* per the linux kernel */

/* stolen from linux/stat.h for fdtype ease */
#define S_IFMT  00170000
#define S_IFSOCK 0140000
#define S_IFLNK  0120000
#define S_IFREG  0100000
#define S_IFBLK  0060000
#define S_IFDIR  0040000
#define S_IFCHR  0020000
#define S_IFIFO  0010000
/* end stolen defs */

#warning the stability and accuracy of these defs is hereby questioned

static void istat(struct inode *from, struct wr_stat *to) {
   if (from && to) { /* sanity */ 
      to->st_dev               = kdev_t_to_nr(from->i_dev);
      to->st_ino               = from->i_ino;
      to->st_mode        = from->i_mode;
      to->st_nlink             = from->i_nlink;
      to->st_uid               = from->i_uid;
      to->st_gid               = from->i_gid;
      to->st_rdev              = kdev_t_to_nr(from->i_rdev);
      /* here's where it gets iffy.... */
      to->st_atimespec.tv_sec  = from->i_atime;
      to->st_atimespec.tv_nsec = 0; /* might be the wrong answer */
      to->st_mtimespec.tv_sec  = from->i_mtime;
      to->st_mtimespec.tv_nsec = 0; /*             "             */
      to->st_ctimespec.tv_sec  = from->i_ctime;
      to->st_ctimespec.tv_nsec = 0; /*             "             */
      /* things are hunky-dory again */
      to->st_size              = from->i_size;
      to->st_gen               = from->i_generation;
   }
}

static int wsi_stat(const char *path, struct wr_stat *sb, unsigned int flags)
{
   struct dentry *d; /* we'll get the inode from this sucker */
   if (path && sb) {
#if PD_LINUX < 0x02004
      d = lookup_dentry((char *)path, NULL, 0); 
#else
      d = lookup_one((char *)path, NULL);
#endif
      if (!IS_ERR(d) && d && d->d_inode) {
         istat(d->d_inode, sb);
         return 0; /* everything is okey-dokey */
      }
   }
   return -1; /* ERROR */
}

int wsi_fileutil_stat(const char *path, struct wr_stat *sb)
{
    return wsi_stat(path, sb, 0);
}

int wsi_fileutil_lstat(const char *path, struct wr_stat *sb)
{
    return wsi_stat(path, sb, LOOKUP_FOLLOW);
}

int wsi_fileutil_fstat(unsigned int fd,  struct wr_stat *sb)
{
    int retval = -1;
    struct file *f =  NULL;
    f = fget(fd);
    if (f && sb && f->f_dentry && f->f_dentry->d_inode) {
        istat(f->f_dentry->d_inode, sb);
        retval = 0;
       fput(f);
    }
    return retval; 
}

int wsi_fileutil_fdtype(unsigned int fd) {
   int retval = -1; /* this should never be right */
   struct file *f = NULL;
   f = fget(fd);
   if (f && f->f_dentry && f->f_dentry->d_inode && f->f_dentry->d_inode->i_mode) {
      /* this mimics a macro-fied test in linux/stat.h */
      switch (f->f_dentry->d_inode->i_mode & S_IFMT) {
         case S_IFIFO:
            retval = WR_DTYPE_FIFO;
            break;
         case S_IFSOCK:
            retval =  WR_DTYPE_SOCKET;
            break;
         case S_IFDIR: 
         case S_IFREG:
         case S_IFLNK: /* will these ever  */
         case S_IFBLK: /* be defined in a  */
         case S_IFCHR: /* rational manner? */
            retval = WR_DTYPE_VNODE;
            break;
         default:
            retval = WR_DTYPE_OTHER;   
            break;
      }
      fput(f);
   }
   return retval; 
}

int wsi_fileutil_fdfflag(unsigned int fd) {
   int retval = -1; /* can't be right! f_flags is unsigned. */
   struct file *f = NULL;   
   f = fget(fd) ;
   if (f && f->f_flags) retval = f->f_flags;
   fput(f);
   return retval; 
}

/*
 * Network Related Stuff
 */

static const char *af_name[AF_MAX+1] = {
   "AF_UNSPEC",                    /*      0 */
   "AF_UNIX",                      /*      1 */
   "AF_INET",                      /*      2 */
   "AF_AX25",                      /*      3 */
   "AF_IPX",                       /*      4 */
   "AF_APPLETALK",                 /*      5 */
   "AF_NETROM",                    /*      6 */
   "AF_BRIDGE",                    /*      7 */
   "AF_ATMPVC",                    /*      8 */
   "AF_X25",                       /*      9 */
   "AF_INET6",                     /*     10 */
   "AF_ROSE",                      /*     11 */
   "AF_DECnet",                    /*     12 */
   "AF_NETBEUI",                   /*     13 */
   "AF_SECURITY",                  /*     14 */
   "pseudo_AF_KEY",                /*     15 */
   "AF_NETLINK",                   /*     16 */
   "AF_PACKET",                    /*     17 */
   "AF_ASH",                       /*     18 */
   "AF_ECONET",                    /*     19 */
   "AF_ATMSVC",                    /*     20 */
   "unused",                       /*     21 */
   "AF_SNA",                       /*     22 */
   "AF_IRDA",                      /*     23 */
   "unused",                       /*     24 */
   "unused",                       /*     25 */
   "unused",                       /*     26 */
   "unused",                       /*     27 */
   "unused",                       /*     28 */
   "unused",                       /*     29 */
   "unused",                       /*     30 */
   "unused"                        /*     31 */
};

static const char *msg_noaddr = "No address";
static const char *msg_fail = "Fail";

/**** THIS MAY BE COMPLETELY WRONG, NEEDS CHECKING ****/
/* can anyone verify this for endian order under linux/intel? */
static void wr_netitoa(long inaddr, char *netbuf) { 
   unsigned char *ucp = (unsigned char *)(&inaddr);
   sprintf(netbuf, "%d.%d.%d.%d",
      ucp[0] & 0xff,
      ucp[1] & 0xff,
      ucp[2] & 0xff,
      ucp[3] & 0xff);
}

int wsi_netutil_get_addr(const struct sockaddr *saddr, 
      int len, char *buffer, int buffer_length) {
   char netbuf[3+1+3+1+3+1+3+1]; /* 255.255.255.255\0 */
   int retval = 0;
   int datalen = 0; 
   /* excessive sanity checking */
   if (saddr && buffer && (saddr->sa_family <= AF_MAX)) {
      datalen = strlen((char *)saddr->sa_data)+1;
      switch (saddr->sa_family) {
         case AF_UNIX:
         if (buffer_length < datalen)
               retval = -1;
            strncpy(buffer, saddr->sa_data, datalen);
         break;
         case AF_INET:
         wr_netitoa(((struct sockaddr_in *)saddr)->sin_addr.s_addr, 
            netbuf);
         datalen = strlen(netbuf);
         if (buffer_length < datalen) retval = -1;
         sprintf(buffer, "%s::%d", netbuf, 
            ntohs(((struct sockaddr_in *)saddr)->sin_port));
         break;
         default: /* failed */
         datalen = strlen(msg_fail) + 1;   
         if (buffer_length < datalen)
            retval = -1;
         strncpy(buffer, msg_fail, datalen);
      }
   } else { /* there is no address, buffer, or the address is bogus */
      strncpy(buffer, msg_noaddr, buffer_length);
      retval = -1;
   }
   return retval;
}

const char *wsi_netutil_get_family(const struct sockaddr *saddr, int len) {
   if (saddr && (len>=16)) {
      if (saddr->sa_family <= AF_MAX)  
         return af_name[saddr->sa_family];
      return "AF_UNKNOWN"; /* unknon family? error? */
   }
   return NULL; /* definitely an error */
}

#if PD_LINUX < 0x02004
/* helper for socket functions. */
static struct socket *sockfd_lookup(int fd, int *err)
{
   struct file   *f;
   struct inode  *i;
   struct socket *s = NULL;

   f = fget(fd);
   if( !f )
   {
      *err = -EBADF;
      goto donelookup;
   }

   i = f->f_dentry->d_inode;
   if( !i || !i->i_sock )
   {
      *err = -ENOTSOCK;
      fput(f);
      goto donelookup;
   }

   s = &i->u.socket_i;
   wsi_printf("inode = %p sock = %p\n", i, s);
   if( s->file != f )
   {
      /* socket file changed */
      s->file = f;
   }

donelookup:
   return (s); 
}
#endif /* PD_LINUX < 0x02004 */

static __inline__ void sockfd_put( struct socket *s )
{
   fput(s->file);
}

int
wsi_netutil_getsockname( int fd, struct sockaddr *s, int *len )
{
   char           addr[MAX_SOCK_ADDR];
   struct socket *sock;
   int            ret  = -EINVAL;

   if( s && len )
   {
      sock = sockfd_lookup( fd, &ret );
      if( sock && sock->ops && sock->ops->getname )
      {
         if( !sock->ops->getname( sock, (struct sockaddr *)addr, len, 0 ) )
         {
            bcopy(addr, s, *len); 
            ret = 0;
         }
         sockfd_put( sock );
      }
   }
   return (ret);
}

int
wsi_netutil_getpeername( int fd, struct sockaddr *s, int *len )
{
   char           addr[MAX_SOCK_ADDR];
   struct socket *sock;
   int            ret  = -EINVAL;

   if( s && len )
   {
      sock = sockfd_lookup( fd, &ret );
      if( sock && sock->ops && sock->ops->getname )
      {
         if( !sock->ops->getname( sock, (struct sockaddr *)addr, len, 1 ) )
         {
            bcopy( addr, s, *len );
            ret = 0;
         }
         sockfd_put( sock );
      }
   }
   return (ret);
}

unsigned short
wsi_netutil_ntohs( short x )
{
   return ntohs(x);
}

