/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_path.c,v 1.14 2001/04/23 21:15:46 stevek Exp $
 *
 * Authors: Tim Fraser, Douglas Kilpatrick
 *
 * Purpose: This file contains functions used to determine a processes current
 *          working directory.
 */

#include <wr_conf.h>
#include <wsi_types.h>
#include <wss_process.h>
#include <wss_path.h>
#include <wss_malloc.h>
#include <wsi_stdlib.h>

#include <linux/module.h>      /* for module macros */
#include <linux/kernel.h>      /* for kernel utility functions */
#include <linux/limits.h>      /* for PATH_MAX */
#include <linux/sched.h>       /* for struct task_struct, current */
#include <linux/mm.h>          /* for GFP_KERNEL */
#include <linux/fs.h>          /* for open_namei(), struct inode */
#include <linux/string.h>      /* for string handling functions */


/*
 * Internal macros
 */

#define WSS_FILE_SEPARATOR '/'

/* wsi_path_abspath()
 *
 * in:     path_s - a path.
 * out:    nothing
 * return: An absolute path in canonical form which is equivalent to 
 *         `path_s' in the context of the `current' process, or NULL
 *         if we cannot construct such a path from `path_s'.
 *
 * This function may return a pointer to a string.  If it does, the 
 * string's memory has been dynamically allocated, and must be freed
 * by the caller.  Example:
 *
 * char *result_s;
 * if( result_s = wsi_path_abspath( "/usr/bin/orbital" ) ) {
 *    ... do what you need to do with result_s ...
 *    wss_free( result_s );
 * }
 *
 * Note that this function always allocates PATH_MAX space for the
 * result, regardless of the actual length of the result string.
 * We may want to fix this resource-hogging behavior if callers
 * like to keep these result strings hanging around for a long time.
 *
 * Note that this function canonicalizes only those parts of `path_s'
 * that correspond to existing filesystem objects.  So, assuming the
 * /usr/local/ directory is empty, the result for input
 * "/usr/local/../local/foo/../foo/bar" is
 * "/usr/local/foo/../foo/bar", not
 * "/usr/local/foo/bar".
 *
 * Although this is ugly, it's the right thing, since the meaning of
 * the ".." that remains in the result depends on whether or not
 * /usr/local/foo is meant to be a link.
 * 
 */
pd_char_t *
wsi_path_abspath( const pd_char_t *path_s )
{
   char *scratch_s;                 /* scratch space */
   char *p_prefix_end;              /* end of the existing part of `path_s' */
   char *d_path_output_s;  /* d_path()'s output, somewhere in scratch space */
   char *result_s = NULL;           /* our result goes here */
   char *p_end;                     /* used to test for separators */
   size_t result_len;
   size_t path_len;
   size_t path_output_len;
   size_t chunk_len;
#if PD_LINUX < 0x02004
   struct dentry *p_dentry = NULL;  /* leafmost existing node of `path_s'   */
#else
   struct nameidata nd;
#endif

   if( path_s == NULL ) return NULL;

   /* Get a page of scratch space in `scratch_s' and copy `path_s' there. */
   scratch_s = (char *)__get_free_page(GFP_KERNEL);
   strncpy( scratch_s, path_s, PATH_MAX );
   chunk_len = path_len = strnlen( scratch_s, PATH_MAX );

   /*
    *  STEP 1:
    *
    * Find the longest prefix of `scratch_s' that describes an existing
    * filesystem object.  We use open_namei() to determine whether or
    * not the filesystem object named by `scratch_s' describes an
    * existing object or not.  We begin by giving open_namei() the full
    * `scratch_s' string.  If open_namei() rejects the full string, we
    * shorten `scratch_s' by removing its last
    * WSS_FILE_SEPARATOR-delimited component, and try again.
    * 
    *  Eventually, either open_namei() will accept `scratch_s' or we
    *  will trim `scratch_s' down to nothing, which open_namei() will
    *  accept as a relative reference to the `current' process's
    *  current working directory.
    *
    *  Once open_namei() has accepted `scratch_s', we can use Linux's
    *  dcache to get the equivalent absolute path in canonical form.
    *  Note that the result of this step is merely to get the existing
    *  part of `path_s' into absolute canonical form.  The last few
    *  components of `path_s' may not describe filesystem components
    *  which exist, and we'll have to deal with them later.
    */
   p_prefix_end = scratch_s + path_len;

   while( 1 )
   {
      /* Try to lookup the prefix currently in `scratch_s' */
#if PD_LINUX < 0x02004
      p_dentry = open_namei( scratch_s, 0x00, 0x00 );
      if( !IS_ERR( p_dentry ) )
         break;
#else
      if( !open_namei( scratch_s, 0x00, 0x00, &nd ) )
         break;
#endif /* PD_LINUX < 0x02004 */

      /* If we've arrived here, then the current prefix was not found. *
       * Shorten the current prefix and try again.                     */
      while( ( *p_prefix_end != WSS_FILE_SEPARATOR ) &&
             ( p_prefix_end > scratch_s ) )
         chunk_len--, p_prefix_end--;

      *p_prefix_end = '\0';
   } /* while `scratch_s' has not been accepted by open_namei() */


   /* open_namei() has told us that (the possibly shortened)
    * `scratch_s' corresponds to the `p_dentry' dcache entry.  Ask the
    * dcache what the absolute canonical name for `p_dentry' is using
    * d_path().  Note that d_path() ignores the contents of scratch_s,
    * and overwrites this space as it uses it for its own purposes.
    */
#if PD_LINUX < 0x02004
   d_path_output_s = d_path( p_dentry, scratch_s, PAGE_SIZE );
#else
   d_path_output_s = d_path( nd.dentry, nd.mnt, scratch_s, PAGE_SIZE );
#endif
   if( !d_path_output_s )
      goto out;

   path_output_len = strnlen( d_path_output_s, PATH_MAX );
   result_len = path_output_len + ( path_len - chunk_len ) + 2;

   /*
    * STEP 2:
    *
    * Take the non-existent part of `path_s' and append it to the
    * cannonical absolute part that we discovered in step 1.
    */

   /* `d_path_output_s' now contains the beginning of our result
    * string.  Copy it to `result_s' for safe keeping.
    */
   if( !( result_s = (char *)wss_malloc( result_len ) ) )
      goto out;

   strncpy( result_s, d_path_output_s, result_len );
  
   /* Now we need to get the non-existing part of `path_s' that we     
    * trimmed off in step 1.  Copy `path_s' back into `scratch_s'.     
    * Step 1 should have left `p_prefix_end' pointing at the beginning 
    * of the part of `path_s' that we want.  Append the part of        
    * `scratch_s' that begins at `p_prefix_end' to `result_s'.         
    *
    * Appending is a little complicated, because we must ensure that
    * there are the correct number of WSS_FILE_SEPARATOR characters
    * between the `result_s' and `scratch_s' parts.  We use the
    * following algorithm:
    *
    * IF the part in `scratch_s' is a zero-length string, we're done.
    * ELSE we make sure that the `result_s' part ends with a 
    * WSS_FILE_SEPARATOR, and the `scratch_s' part does not begin with
    * one.
    */
   strncpy( scratch_s, path_s, PATH_MAX );  

   /* If the `p_prefix_end' part is not zero-length, then we have to
    * make sure the WSS_FILE_SEPARATORS are correct and do concatenation.
    */
   if( *p_prefix_end )
   {
      /* Make sure that the part already in `result_s' ends with a
       * WSS_FILE_SEPARATOR.
       */
      p_end = result_s + path_output_len;
      if( *(p_end - 1) != WSS_FILE_SEPARATOR )
      {
         *p_end++ = WSS_FILE_SEPARATOR;
         *p_end = '\0';
         path_output_len++;
      }

      /* Make sure that the part starting at `p_prefix_end' does not
       * begin with a WSS_FILE_SEPARATOR.
       */
      while( *p_prefix_end == WSS_FILE_SEPARATOR )
         p_prefix_end++;

      /* append the `p_prefix_end' part with the part already in `result_s'. */
      strncpy( p_end, p_prefix_end, result_len - path_output_len );

   } /* if `p_prefix_end' part is not zero length */

out:
   free_page( (unsigned long)scratch_s );
#if PD_LINUX < 0x02004
   dput( p_dentry );  /* We're done with `p_dentry' now.  Thank you. */
#else
   dput( nd.dentry );
#endif
   return (result_s);
}

pd_char_t *
wsi_path_cwd( const wss_proc_t *wrp )
{
   char *tmp, *path;
   
   if( !wrp || !wrp->wrp_p || !wrp->wrp_p->fs )
      return NULL;

   tmp = (char*)__get_free_page(GFP_KERNEL);

#if PD_LINUX < 0x02004
   path = d_path( wrp->wrp_p->fs->pwd, tmp, PAGE_SIZE );
#else
   path = d_path( wrp->wrp_p->fs->pwd, wrp->wrp_p->fs->pwdmnt, tmp, PAGE_SIZE );
#endif

   /* Get our own copy to pass out. */
   path = wss_strdup(path);
    
   free_page( (unsigned long)tmp );

   return (path);
}

