/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_utilities.c,v 1.17 2001/02/23 19:02:37 stevek Exp $
 * 
 * Author: Doug Kilpatrick and others.
 * 
 * Purpose: Repository of usefull functions that might be missing.
 */

#define SUPPRESS_MACROS

#include <wr_conf.h>
#include <wsi_stdlib.h>
#include <wsi_utilities.h>
#include <wss_utilities.h>
#include <wss_mutex.h>
#include <wss_malloc.h>

#ifdef USER_MODE_TEST
#include <string.h>
#endif


/***************************************************************************\
 * Private structures and types                                            *
\***************************************************************************/


#warning wr_wait/sleep are Reentrant, not SMP safe
/*
 * SMP: semaphore?
 */
struct wait_node
{
    wss_util_rbnode_t   parent;
    long                cookie;
    int                 count;
    struct wait_queue  *wq;
};

typedef struct wait_node wait_node_t;


#ifndef USER_MODE_TEST
/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static wss_util_rbtree_t *wsi_wait_tree = NULL;
WSS_MUTEX_PROTECT       ( wsi_wait_tree );


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static int
waitnode_compfunc(void *n1, void *n2)
{
    wait_node_t *node1 = n1;
    wait_node_t *node2 = n2;
    return (
            (node1->cookie <  node2->cookie) ? -1 :
            (node1->cookie == node2->cookie) ?  0 : 1 );
}

static void *
waitnode_clonefunc(void *n)
{
    wait_node_t *retval = wss_malloc(sizeof(*retval));
    wss_util_rbnode_init(&(retval->parent));
    if (!n) {
        retval->cookie  = 0;
        retval->wq      = 0;
        retval->count   = 0;
    } else {
        wait_node_t *orig = n;
        retval->cookie  = orig->cookie;
        retval->wq      = orig->wq;
        retval->count   = orig->count;
    }
    return retval;
}

static void
waitnode_free(void *n)
{
    wait_node_t *node = n;
    if (node) {
        /* Don't fiddle with wq's. */
        wss_free(node);
    }
}

#endif

/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

typedef int             cmp_t __P((const void *, const void *));
static inline char     *med3 __P((char *, char *, char *, cmp_t *));
static inline void      swapfunc __P((char *, char *, int, int));

#define min(a, b)      ((a) < (b) ? (a) : (b))

/*
 * Qsort routine from Bentley & McIlroy's "Engineering a Sort Function".
 */
#define swapcode(TYPE, parmi, parmj, n) {              \
       long i = (n) / sizeof (TYPE);                   \
       register TYPE *pi = (TYPE *) (parmi);           \
       register TYPE *pj = (TYPE *) (parmj);           \
       do {                                            \
               register TYPE   t = *pi;                \
               *pi++ = *pj;                            \
               *pj++ = t;                              \
        } while (--i > 0);                             \
}

#define SWAPINIT(a, es) swaptype = ((char *)a - (char *)0) % sizeof(long) || \
       es % sizeof(long) ? 2 : es == sizeof(long)? 0 : 1;

static inline void
swapfunc(char *a, char *b, int n, int swaptype)
{
       if(swaptype <= 1)
               swapcode(long, a, b, n)
       else
               swapcode(char, a, b, n)
}

#define swap(a, b)                                     \
       if (swaptype == 0) {                            \
               long t = *(long *)(a);                  \
               *(long *)(a) = *(long *)(b);            \
               *(long *)(b) = t;                       \
       } else                                          \
               swapfunc(a, b, es, swaptype)

#define vecswap(a, b, n)       if ((n) > 0) swapfunc(a, b, n, swaptype)

static inline char *
med3(char *a, char *b, char *c, cmp_t *cmp)
{
       return cmp(a, b) < 0 ?
              (cmp(b, c) < 0 ? b : (cmp(a, c) < 0 ? c : a ))
              :(cmp(b, c) > 0 ? b : (cmp(a, c) < 0 ? a : c ));
}

void
qsort(void *a, size_t n, size_t es, cmp_t *cmp)
{
       char *pa, *pb, *pc, *pd, *pl, *pm, *pn;
       int d, r, swaptype, swap_cnt;

loop:  SWAPINIT(a, es);
       swap_cnt = 0;
       if (n < 7) {
               for (pm = (char *)a + es; pm < (char *)a + n * es; pm += es)
                       for (pl = pm; pl > (char *)a && cmp(pl - es, pl) > 0;
                            pl -= es)
                               swap(pl, pl - es);
               return;
       }
       pm = (char *)a + (n / 2) * es;
       if (n > 7) {
               pl = a;
               pn = (char *)a + (n - 1) * es;
               if (n > 40) {
                       d = (n / 8) * es;
                       pl = med3(pl, pl + d, pl + 2 * d, cmp);
                       pm = med3(pm - d, pm, pm + d, cmp);
                       pn = med3(pn - 2 * d, pn - d, pn, cmp);
               }
               pm = med3(pl, pm, pn, cmp);
       }
       swap(a, pm);
       pa = pb = (char *)a + es;

       pc = pd = (char *)a + (n - 1) * es;
       for (;;) {
               while (pb <= pc && (r = cmp(pb, a)) <= 0) {
                       if (r == 0) {
                               swap_cnt = 1;
                               swap(pa, pb);
                               pa += es;
                       }
                       pb += es;
               }
               while (pb <= pc && (r = cmp(pc, a)) >= 0) {
                       if (r == 0) {
                               swap_cnt = 1;
                               swap(pc, pd);
                               pd -= es;
                       }
                       pc -= es;
               }
               if (pb > pc)
                       break;
               swap(pb, pc);
               swap_cnt = 1;
               pb += es;
               pc -= es;
       }
       if (swap_cnt == 0) {  /* Switch to insertion sort */
               for (pm = (char *)a + es; pm < (char *)a + n * es; pm += es)
                       for (pl = pm; pl > (char *)a && cmp(pl - es, pl) > 0;
                            pl -= es)
                               swap(pl, pl - es);
               return;
       }

       pn = (char *)a + n * es;
       r = min(pa - (char *)a, pb - pa);
       vecswap(a, pb - r, r);
       r = min(pd - pc, pn - pd - es);
       vecswap(pb, pn - r, r);
       if ((r = pb - pa) > es)
               qsort(a, r / es, es, cmp);
       if ((r = pd - pc) > es) {
               /* Iterate rather than recurse to save stack space */
               a = pn - r;
               n = r / es;
               goto loop;
       }
/*             qsort(pn - r, r / es, es, cmp);*/                               
}

int copyin(const void *user, void *driver, int n)
{
#ifdef USER_MODE_TEST
    memcpy(driver,user,n);
    return 0;
#else
    if (copy_from_user(driver, user, n))
        return EFAULT;
    return 0;
#endif
}

int copyout(const void *driver, void *user, int n)
{
#ifdef USER_MODE_TEST
    memcpy(user,driver,n);
    return 0;
#else
    if (copy_to_user(user, driver, n))
        return EFAULT;
    return 0;
#endif
}

/*
 * copyinstr
 *
 * strncpy from userspace to kernel space.  len is maximum lenght.
 * If I do it in terms of copyin, its gonna be slow as hell, so I'm
 * gonna punt.
 */
int copyinstr(const char *user, char *driver, int len, size_t *pi_done)
{
#ifdef USER_MODE_TEST
    strncpy(driver,user,len);
    ((char*)driver)[len-1] = '\0';
    *pi_done = strlen(driver);
    return 0;
#else
    char ch;
    const char *u_ch = user;
    int  i = 0, retval;

    if (len <= 0 || driver == NULL)
        return EINVAL;

    do {
        retval = get_user(ch, u_ch + i);
    } while (!retval && (driver[i] = ch) != '\0' && ++i < len);
    driver[len-1] = '\0';
    if (pi_done)
        *pi_done = strnlen(driver, len);
    return (retval ? EFAULT : 0);
#endif
}

#ifndef USER_MODE_TEST
void bcopy(const void *src, void *dest, int n)
{
    memcpy(dest, src, n);
}

void bzero(void *b, int n)
{
    memset(b, 0, n);
}
#endif

char *wsi_strcat(char *dest, const char *src)
{
    return strcat(dest, src);
}

int wsi_strcmp(const char *s1, const char *s2)
{
    return strcmp(s1,s2);
}

int wsi_strncmp(const char *s1, const char *s2, size_t n)
{
    return strncmp(s1,s2,n);
}

size_t wsi_strlen(const char *s)
{
    return strlen(s);
}

char *wsi_strcpy(char *dest, const char *src)
{
    return strcpy(dest,src);
}

char *wsi_strncpy(char *dest, const char *src, size_t n)
{
    return strncpy(dest,src,n);
}

/*  wsi_toupper */
/*! \brief Implementation of toupper
 *
 *  \param c Character to convert to uppercase
 *
 *  \return Character converted to uppercase
 */
pd_char_t wsi_toupper( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('a') ) && ( c <= PD_MAKESTR('z') ) )?
            c - PD_MAKESTR('a') + PD_MAKESTR('A') :
            c;
}

/*  wsi_tolower */
/*! \brief Implementation of tolower
 *
 *  \param c Character to convert to lowercase
 *
 *  \return Character converted to lowercase
 */
pd_char_t wsi_tolower( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('A') ) && ( c <= PD_MAKESTR('Z') ) )?
            c - PD_MAKESTR('A') + PD_MAKESTR('a') :
            c;
}

#ifndef USER_MODE_TEST
void wsi_util_init()
{
    WSS_MUTEX_INIT(wsi_wait_tree);
}

void wsi_util_shutdown()
{
    if( wsi_wait_tree )
    {
       WSS_MUTEX_ENTER(wsi_wait_tree);
       if( wsi_wait_tree )
       {
          wss_util_rbtree_free(wsi_wait_tree);
          wsi_wait_tree = NULL;
          WSS_MUTEX_EXIT(wsi_wait_tree);
       }
    }

    WSS_MUTEX_DESTROY(wsi_wait_tree);
}

/*
 * 0 for booring wake up.
 * !0 for signaled.
 */
int wsi_util_sleep(long cookie)
{
    wait_node_t *new_node;
    wait_node_t *result;
    wait_node_t  temp;

    WSS_MUTEX_ENTER(wsi_wait_tree);
    if (!wsi_wait_tree)
        wsi_wait_tree = wss_util_rbtree_new( waitnode_compfunc,
                                             waitnode_clonefunc,
                                             waitnode_free );
    wss_util_rbnode_init(&(temp.parent));
    temp.cookie = cookie; temp.wq = 0; temp.count = 0;
    new_node = (wait_node_t*)wss_util_rbtree_find(wsi_wait_tree,
                                                    &(temp.parent));

    if (!new_node) {
        new_node = waitnode_clonefunc(&temp);
        wss_util_rbtree_insert(wsi_wait_tree, &(new_node->parent));
    }
    WSS_MUTEX_EXIT(wsi_wait_tree);
    new_node->count++;
    interruptible_sleep_on(&(new_node->wq));
    new_node->count--;
    WSS_MUTEX_ENTER(wsi_wait_tree);
    /* Honestly, we should have a background thread to clean this out every
     * few seconds instead of doing it constantly.  Ug.
     */
    /* May be unnecessary.  I can't remember if in a RB tree, you are
     * guarenteed that the node object never mutates.  To be safe, re-get
     * the node.
     */
    new_node = (wait_node_t*)wss_util_rbtree_find(wsi_wait_tree,
                                                    &(temp.parent));
    if (new_node->count == 0) {
        result = (wait_node_t*)wss_util_rbtree_delete(wsi_wait_tree,
                                                    &(new_node->parent));
        /* assert(result == new_node); *//* Clone of?? */
        waitnode_free(result);
    }
    WSS_MUTEX_EXIT(wsi_wait_tree);

    return signal_pending(current);
}

void wsi_util_wakeup(long cookie)
{
    wait_node_t *result;
    wait_node_t  temp;

    WSS_MUTEX_ENTER(wsi_wait_tree);
    if (!wsi_wait_tree)
        goto done; /* Obviously no one to wake up. */

    wss_util_rbnode_init(&(temp.parent));
    temp.cookie = cookie; temp.wq = 0; temp.count = 0;
    result = (wait_node_t*)wss_util_rbtree_find(wsi_wait_tree,
                                                    &(temp.parent));

    if (!result)
        goto done; /* No one to wake up. */

    wake_up_interruptible(&(result->wq));

done:
    WSS_MUTEX_EXIT(wsi_wait_tree);
}
#endif

