/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_wrctl.c,v 1.13 2001/02/23 19:02:37 stevek Exp $
 *
 * Author: Steve Kiernan
 *
 * Purpose: Implementation of the wrctl system call
 * 
 */

#include <wr_conf.h>

#include <wss_mutex.h>

#include <wss_process.h>

#include <wsi_syscalls.h>

#include <wss_module.h>

#include <wss_malloc.h>

#include <wsi_stdlib.h>

#include <wsi_errno.h>

#include <wss_wrvevent.h>

/* Imported from elsewhere */
extern wss_linker_t            *wss_linker_list;
WSS_MUTEX_EXPORT               (wss_linker_list);

extern wss_linker_file_t        wss_linker_file_list;
WSS_MUTEX_EXPORT               (wss_linker_file_list);

extern void wss_malloc_dump(void);

/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*
 * wrctl syscall
 */
asmlinkage
int wrctl( int command, int id, void *buffer )
{
    struct wss_proc *wrp;
    int error = 0;

    /*
     * Find wss_proc structure for current process
     */
    if( ( wrp = wss_pfind( current->pid ) ) ) {
        /* If we're wrapped, make it look like this syscall isn't available */
        WSS_MUTEX_ENTER(wrp->wrp_wi_list);
        if( wrp->wrp_wi_list )
            error = -ENOSYS;
        WSS_MUTEX_EXIT(wrp->wrp_wi_list);
        if (error) goto end;
    }

#ifdef SUSER_ONLY
    if( !suser() ) {
        error = -EPERM;
        goto end;
    }
#endif


    switch( command ) {
    case WRCTL_LOAD:
        {
            /* Load wrapper module */
            wss_mod_t  *wrm;
            pd_char_t *filename;
            int         error;

#ifdef DEBUG_WRCTL
            wsi_printf( PD_MAKESTR("Allocating block\n") );
#endif
            if( (filename = (pd_char_t *)wss_malloc( id ) ) == NULL ) {
                return -EAGAIN;
            }
            copyin( buffer, filename, id );

            /*
             * Create and add a new wrapper module information
             * structure to the global list.
             */
#ifdef DEBUG_WRCTL
            wsi_printf( PD_MAKESTR("Allocating module block\n") );
#endif
            if( ( wrm = wss_mod_new( filename, &error ) ) == NULL ) {
                wss_free(filename);

                return -error;
            }

            /*
             * Setup the wrapper module structure
             */
            wrm->wrm_uid = current->p_pptr->uid;

#ifdef DEBUG_WRCTL
            wsi_printf( PD_MAKESTR("Module #%d '%s' loaded\n"), wrm->wrm_id,
                    wrm->wrm_linked->linker_file_name );
#endif

            return wrm->wrm_id;
        }
    case WRCTL_UNLOAD:
        {
            /* Unload wrapper module */
            int        type = 0;
            wss_mod_t *wrm;
            int        error;

            /* Find the module */
            if( ( wrm = wss_mod_find( id ) ) != NULL ) {
                type = wrm->wrm_type;
                wss_mod_rele( wrm );
            }

            /* Delete the module */
            error = wss_mod_delete( id );

            if( error ) {
                return -error;
            } else {
                /*
                 * Post module unloaded event
                 */
                wss_wrvevent_post_modunloaded( id, type );

                return id;
            }
            return 0;
        }
    case WRCTL_ENABLED:
    #warning WRCTL_ENABLED: TBD
        return 0;

    case WRCTL_STAT_LINKERS:
        {
            wss_linker_t *linker;
            int           num;

            WSS_MUTEX_ENTER( wss_linker_list );
            for( linker = wss_linker_list, num = 0;
                    linker;
                    linker = linker->linker_next, num++ )
                ;

            wsi_printf( "[ %d Linker%s Registered ]\n", num, (num==1)?"":"s" );
            for( linker = wss_linker_list; linker;
                    linker = linker->linker_next ) {
                wsi_printf( "%s @ 0x%x\n", linker->linker_description,
                        (u_int32_t)linker );
            }
            WSS_MUTEX_EXIT( wss_linker_list );

            return 0;
        }

    case WRCTL_STAT_REFS:
        {
            if( id & ( WRCTL_STAT_REFS_PROCESSES |
                        WRCTL_STAT_REFS_WRAPPER_INSTANCES ) )
            {
                wss_proc_t *wrp;

                WSS_MUTEX_ENTER( wss_proc_list );
                wsi_printf( "\n[ Process References ]\n" );
                for( wrp = wss_proc_list.lh_first; wrp;
                        wrp = wrp->wrp_p_list.le_next ) {
                    wsi_printf( "%-50s = %d reference%s\n", wrp->wrp_progname,
                            WSS_REF_COUNT(wrp->wrp_refcount),
                            (WSS_REF_COUNT(wrp->wrp_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
                    if( id & WRCTL_STAT_REFS_PROCESSES_DUMP_CHAIN ) {
                        wss_ref_t *ref;

                        for( ref = wrp->wrp_refcount.ref_next; ref;
                                ref = ref->ref_next )
                            wsi_printf( "   %s:%d\n", ref->un.st.ref_file,
                                    ref->un.st.ref_line );
                    }
#endif /* DEBUG_REFS */

                    if( id & WRCTL_STAT_REFS_WRAPPER_INSTANCES ) {
                        wss_wrapper_instance_t *wi;

                        if( wrp->wrp_wi_list ) {
                            wsi_printf( "   === Wrapper Instances ===\n" );
                            for( wi = (wss_wrapper_instance_t *)
                                    wrp->wrp_wi_list; wi; wi = wi->wi_next ) {
                                wsi_printf( "   %-47s = %d reference%s\n",
                                        wi->wi_wrapper->w_module->
                                            wrm_linked->linker_file_name,
                                        WSS_REF_COUNT(wi->wi_refcount),
                                        (WSS_REF_COUNT(
                                            wi->wi_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
                                if( id &
                                WRCTL_STAT_REFS_WRAPPER_INSTANCES_DUMP_CHAIN ) {
                                    wss_ref_t *ref;

                                    for( ref = wi->wi_refcount.ref_next; ref;
                                            ref = ref->ref_next )
                                        wsi_printf( "   %s:%d\n",
                                                ref->un.st.ref_file,
                                                ref->un.st.ref_line );
                                }
#endif /* DEBUG_REFS */
                            }
                        }
                    }
                }
                WSS_MUTEX_EXIT( wss_proc_list );
            }

            if( id & WRCTL_STAT_REFS_MODULES ) {
                wss_mod_t *wrm;

                WSS_MUTEX_ENTER( wss_module_list );
                wsi_printf( "\n[ Module References ]\n" );
                for( wrm = wss_module_list; wrm; wrm = wrm->wrm_next ) {
                    wsi_printf( "%-50s = %d reference%s\n",
                            wrm->wrm_linked->linker_file_name,
                            WSS_REF_COUNT(wrm->wrm_refcount),
                            (WSS_REF_COUNT(wrm->wrm_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
                    if( id & WRCTL_STAT_REFS_MODULES_DUMP_CHAIN ) {
                        wss_ref_t *ref;

                        for( ref = wrm->wrm_refcount.ref_next; ref;
                                ref = ref->ref_next )
                            wsi_printf( "   %s:%d\n", ref->un.st.ref_file,
                                    ref->un.st.ref_line );
                    }
#endif /* DEBUG_REFS */
                }
                WSS_MUTEX_EXIT( wss_module_list );
            }

            if( id & WRCTL_STAT_REFS_WRAPPERS ) {
                wss_wrapper_t *w;

                WSS_MUTEX_ENTER( wss_wrapper_list );
                wsi_printf( "\n[ Wrapper References ]\n" );
                for( w = wss_wrapper_list; w; w = w->w_next ) {
                    wsi_printf( "%-50s = %d reference%s\n",
                            w->w_module->wrm_linked->linker_file_name,
                            WSS_REF_COUNT(w->w_refcount),
                            (WSS_REF_COUNT(w->w_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
                    if( id & WRCTL_STAT_REFS_WRAPPERS_DUMP_CHAIN ) {
                        wss_ref_t *ref;

                        for( ref = w->w_refcount.ref_next; ref;
                                ref = ref->ref_next )
                            wsi_printf( "   %s:%d\n", ref->un.st.ref_file,
                                    ref->un.st.ref_line );
                    }
#endif /* DEBUG_REFS */
                }
                WSS_MUTEX_EXIT( wss_wrapper_list );
            }

            if( id & WRCTL_STAT_REFS_LINKER_FILES ) {
                wss_linker_file_t *lf;

                WSS_MUTEX_ENTER( wss_linker_file_list );
                wsi_printf( "\n[ Linker File References ]\n" );
                for( lf = &wss_linker_file_list; lf;
                        lf = lf->linker_file_next ) {
                    wsi_printf( "%-50s = %d reference%s\n",
                            lf->linker_file_name,
                            WSS_REF_COUNT(lf->linker_file_refcount),
                            (WSS_REF_COUNT(
                                    lf->linker_file_refcount)==1)?"":"s");
                }
                WSS_MUTEX_EXIT( wss_linker_file_list );
            }
        }
        return 0;

    case WRCTL_STAT_MEMORY:
#ifdef MALLOC_DEBUG 
        wss_malloc_dump();
        return 0;
#else /* MALLOC_DEBUG */
        return EOPNOTSUPP;
#endif /* MALLOC_DEBUG */

    default:
        return -EINVAL;
    }
    
    return -EINVAL;
end:
    wss_proc_rele(wrp);
    return error;
}

