/*! \file wsi_wrv.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_wrv.c,v 1.27 2001/06/23 00:07:20 stevek Exp $
 *
 * \brief   Implementation of the wrv system call
 * \author  Steve Kiernan, Linda Thomas
 * \version $Revision: 1.27 $
 * \date    $Date: 2001/06/23 00:07:20 $
 */
#include <wr_conf.h>

#include <libwrv.h>

#ifndef __WSI_STDLIB_H
#include <wsi_stdlib.h>
#endif

#ifndef __WSI_TYPES_H
#include <wsi_types.h>
#endif

#ifndef __WSS_MALLOC_H
#include <wss_malloc.h>
#endif

#ifndef __WSS_WRAPPER_H
#include <wss_wrapper.h>
#endif

#ifndef __WSS_CRITERIA_H
#include <wss_criteria.h>
#endif

#ifndef __WSS_MODULE_H
#include <wss_module.h>
#endif

#ifndef __WSS_LIBRARY_H
#include <wss_library.h>
#endif

#ifndef __WSI_SYSCALLS_H
#include <wsi_syscalls.h>
#endif

#ifndef __WSS_VM_H
#include <wss_vm.h>
#endif

#ifndef __WSS_WRV_H
#include <wss_wrv.h>
#endif

#ifndef __WSS_DB_H
#include <wss_db.h>
#endif

#ifndef __WSI_TIME_H
#include <wsi_time.h>
#endif

#ifndef __WSS_DISPATCH_H
#include <wss_dispatch.h>
#endif

#ifndef __WSS_UTILITIES_H
#include <wss_utilities.h>
#endif

#include <wrerrno.h>


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

/*! \brief Major version number of the WRV interface */
#define WRV_VERSION_MAJOR  1

/*! \brief Minor version number of the WRV interface */
#define WRV_VERSION_MINOR  1


/***************************************************************************\
 * Private globals                                                         *
 \***************************************************************************/

static int _versionBuffer[] = { WRV_VERSION_MAJOR, WRV_VERSION_MINOR };


/***************************************************************************\
 * Private structures                                                      *
 \***************************************************************************/

/*******************************************************************/

/*! \brief Buffer construction structure
 *
 *  The buffer will be layed out like this:
 *
 *  size_t  # rows in array
 *  size_t  # cols in array
 *  pd_char_t *  array[0][0]
 *  pd_char_t *  array[0][1]
 *  pd_char_t *  array[0][2]
 *  ...
 *  pd_char_t *  array[0][#cols-1]
 *  pd_char_t *  array[1][0]
 *  pd_char_t *  array[1][1]
 *  ...
 *  pd_char_t *  array[1][#cols-1]
 *  ...
 *  pd_char_t *  array[#rows-1][0]
 *  pd_char_t *  array[#rows-1][1]
 *  ...
 *  pd_char_t *  array[#rows-1][#cols-1]
 *
 *  The remainder of the buffer contains the strings the array
 *  elements point to.
 *
 *  Because of the way data is allocated, it is best to allocate
 *  each element only once.  The order can be arbitrary, but
 *  changing an element will lose access to chunk of buffer it was
 *  previously pointing to.
 *
 */
struct bufferBuilder
{
    void        *buffer;     /*!< Pointer to buffer where all the data goes */
    ssize_t      bufferSize; /*!< Size of buffer in bytes */
    size_t       numRows;    /*!< Number of rows in buffer */
    size_t       numCols;    /*!< Number of columns in buffer */
    pd_char_t **array;      /*!< Pointer to beginning of array portion */
    pd_char_t  *data;       /*!< Pointer to beginning of data portion */
    int          offset;     /*!< Offset of where next data should get put */
    int          remaining;  /*!< Amount of space left free */
    ssize_t      needed;     /*!< Size the buffer needs to be */  
};


/***************************************************************************\
 * External functions                                                      *
 \***************************************************************************/

extern void    wss_db_set_row_value_from_string ( wqcd_t, wqrow_t, int,
            pd_char_t * );
extern wqtab_t wss_db_find_table_by_name_r      ( pd_char_t * );
extern int     wss_db_insert_nulls              ( u_long );
extern int     wss_db_delete_by_handle          ( u_long, u_long );
extern int     wsi_sysevents_getevspace         (void);


/***************************************************************************\
 * Private functions                                                       *
 \***************************************************************************/

/*
 * Validate the buffer
 */
static int checkBuffer( void *pbuffer, ssize_t bufferSize )
{
    /* Or maybe we should just return 0, since copy_to_user doesn't
     * really need this to have been called.
     */
    return verify_area(VERIFY_WRITE,pbuffer,bufferSize);
}


static int initBuffer(
        struct bufferBuilder *builder,
        void *buffer, ssize_t bufferSize,
        size_t nRows, size_t nCols )
{
    int rval   = -1;

    builder->buffer     = buffer;
    builder->bufferSize = bufferSize;
    builder->numRows    = nRows;
    builder->numCols    = nCols;
    builder->remaining  = 0;

    if( ( rval = checkBuffer( buffer, bufferSize ) ) == 0 ) {
        if( bufferSize < ( ( 2 * sizeof(size_t) )+ 
                    ( nRows * nCols )* sizeof(pd_char_t *) ) )
        {
            /* buffer not large enough to hold the array framework */
            builder->needed = ( 2 * sizeof(size_t) ) +
                ( nRows * nCols )* sizeof(pd_char_t *);
            rval = -EAGAIN;
        } else {
            size_t *sizes;
            pd_char_t   *nullbuf  = NULL;
            int     row;
            int     col;

            sizes = (size_t *)buffer;
            copyout( &nRows, &sizes[0], sizeof(nRows) );
            copyout( &nCols, &sizes[1], sizeof(nCols) );

            builder->array = (pd_char_t **)&sizes[2];
            for( row = 0; row < nRows; row++ ) {
                for( col = 0; col < nCols; col++ ) {
                    copyout( &nullbuf, &builder->array[row*nCols+col],
                            sizeof(pd_char_t *) );
                }
            }

            builder->data = (pd_char_t *)(&builder->array[nRows * nCols]);
            builder->offset = 0;
            builder->remaining = bufferSize - ( 2 * sizeof(size_t) ) -
                ( nRows * nCols ) * sizeof(pd_char_t *);
            builder->needed = ( 2 * sizeof(size_t) ) *
                ( nRows * nCols ) * sizeof(pd_char_t *);

            rval = 0;
        }
    }

    return rval;
}

static int setString(
        struct bufferBuilder *builder, int row, int col, const pd_char_t *s )
{
    int rval = 0;

#ifdef DEBUG_WRV
    wsi_printf(PD_MAKESTR("string is '%s'\n"), s);
#endif

    if( ( row >= builder->numRows ) || ( col >= builder->numCols ) ) {
        wsi_printf( PD_MAKESTR(
                    "wsi_wrv: attempt to set out of range, "
                    " limit %d, %d, actual %d, %d\n"),
                builder->numRows, builder->numCols, row, col );
        rval = -E2BIG;
    } else {
        if( s == NULL ) {
            pd_char_t *nullbuf = NULL;
            copyout( &nullbuf, &builder->array[row*builder->numCols+col],
                    sizeof(pd_char_t *) );
        } else {
            int len = wsi_strlen( s )+ sizeof(pd_char_t);

            builder->needed += len;
            if( builder->remaining < len )
            {
#ifdef DEBUG_WRV
                wsi_printf( PD_MAKESTR("buffer too small to hold '%s'\n"), s );
#endif
                rval = -EAGAIN;
            } else {
                pd_char_t *p = (pd_char_t *)&builder->data[builder->offset];
                copyout( &p, &builder->array[row*builder->numCols+col], 
                        sizeof(pd_char_t *) );
                copyout( s, &builder->data[builder->offset], len );
                builder->offset += len;
                builder->remaining -= len;
            }
        }
    }

    return rval;
}

static int setInt( struct bufferBuilder *builder, int row, int col, int i )
{
    pd_char_t ibuf[64];

    wsi_sprintf( ibuf, PD_MAKESTR("%d"), i );
    return setString( builder, row, col, ibuf );
}

static int setLong( struct bufferBuilder *builder, int row, int col, long l )
{
    pd_char_t lbuf[64];

    wsi_sprintf( lbuf, PD_MAKESTR("%ld"), l );
    return setString( builder, row, col, lbuf );
}

static int setPtr( struct bufferBuilder *builder, int row, int col, void *p )
{
    pd_char_t pbuf[64];

    wsi_sprintf( pbuf, PD_MAKESTR("%p"), p );
    return setString( builder, row, col, pbuf );
}


/********************************************************************/

static int _wrv_get_active_wrappers( pd_char_t *buf, ssize_t *bufferSize )
{
    struct bufferBuilder  builder;
    wss_wrapper_t        *wr;
    int                   i;
    int                   rval   = 0;

    WSS_MUTEX_ENTER( wss_wrapper_list );
    for( wr = wss_wrapper_list, i = 0; wr; wr = wr->w_next )
        if( wr->w_procs )
            i++;
    WSS_MUTEX_EXIT( wss_wrapper_list );

    rval = initBuffer( &builder, buf, *bufferSize, i, 2 );

    WSS_MUTEX_ENTER( wss_wrapper_list );
    for( wr = wss_wrapper_list; wr; wr = wr->w_next ) {
        if( wr->w_procs ) {
            rval = setString( &builder, i, 0,
                    wr->w_module->wrm_linked->linker_file_name);
            rval = setInt( &builder, i, 1, wr->w_procs );
        }
    }
    WSS_MUTEX_EXIT( wss_wrapper_list );

    if( builder.needed > *bufferSize )
        *bufferSize = builder.needed;

    return rval;
}

/********************************************************************/

static int
_wrv_get_criteria_by_id( int command, u_long id, pd_char_t *buf,
        ssize_t *bufferSize )
{
    struct bufferBuilder  builder;
    wss_crit_t           *act;
    wss_crit_t           *head_ptr;
    int                   i;
    int                   rval = 0;
    int                   row;

    if( command == WRV_GET_ACT_CRIT_BY_ID ) {
        WSS_MUTEX_ENTER( wss_act_crit_list );
        head_ptr = wss_act_crit_list;
    } else {
        WSS_MUTEX_ENTER( wss_deact_crit_list );
        head_ptr = wss_deact_crit_list;
    }

    if( id ) {
        for( act = head_ptr;
                act && act->crit_module->wrm_id != id;
                act = act->crit_next );

        if( act == NULL ) {
            if( command == WRV_GET_ACT_CRIT_BY_ID )
                WSS_MUTEX_EXIT( wss_act_crit_list );
            else
                WSS_MUTEX_EXIT( wss_deact_crit_list );

            rval = -ENOENT;
            return rval;
        }

        rval = initBuffer( &builder, buf, *bufferSize, 1, 7 );

    } else {
        for( act = head_ptr, i = 0; act; act = act->crit_next, i++ );

        rval = initBuffer( &builder, buf, *bufferSize, i, 7 );
    }

    for( act = head_ptr, row = 0; act; act = act->crit_next ) {
        if( ( id == 0 ) || ( id == act->crit_module->wrm_id ) ) {
            rval = setInt( &builder, row, 0, act->crit_module->wrm_id );
            rval = setString( &builder, row, 1,
                    act->crit_module->wrm_linked->linker_file_name);
            rval = setString( &builder, row, 2,
                    act->crit_status == WRACT_ENABLE
                    ? PD_MAKESTR("enabled")
                    : PD_MAKESTR("disabled") );
            rval = setInt( &builder, row, 3, act->crit_module->wrm_uid );
            rval = setInt( &builder, row, 4, act->crit_module->wrm_loadtime );
            rval = setPtr( &builder, row, 5,
                           ( void * )act->crit_module->wrm_linked->linker_file_addr.addr );
            rval = setLong( &builder, row, 6,
                    act->crit_module->wrm_linked->linker_file_size);
            row++;
        }
    }

    if( command == WRV_GET_ACT_CRIT_BY_ID )
        WSS_MUTEX_EXIT( wss_act_crit_list );
    else
        WSS_MUTEX_EXIT( wss_deact_crit_list );

    if( builder.needed > *bufferSize )
        *bufferSize = builder.needed;

    return rval;
}


/********************************************************************/

static int _wrv_get_library_by_id(
        u_long id, pd_char_t *buf, ssize_t *bufferSize )
{
    struct bufferBuilder  builder;
    wss_lib_t           *lp;
    wss_symbol_t        *symtab;
    pd_char_t          *symbuf;
    pd_char_t          *symbufptr;
    int                  symbuflen;
    int                  symlen;
    int                  sym;
    int                  symbufsize = 2048;
    int                  rval       = 0;
    int                  i;
    int                  count;

    if( id ) {
        lp = wss_lib_find( id );
        count = id + 1;

        if( lp == NULL )
            return -ENOENT;

        wss_lib_rele( lp );

        rval = initBuffer( &builder, buf, *bufferSize, 1, 8 );
    } else {
        for( count = 0; wss_lib_find( count )!= NULL; count++ )
            ;

        rval = initBuffer( &builder, buf, *bufferSize, count, 8 );
    }

    for( i = 0; i < count; i++ ) {
        if( ( id == 0 ) || ( id == i ) ) {
            lp = wss_lib_find( i );

            rval = setInt( &builder, i, 0, lp->lib_module->wrm_id );
            rval = setString( &builder, i, 1,
                    lp->lib_module->wrm_linked->linker_file_name);
            rval = setInt( &builder, i, 2,
                    WSS_REF_COUNT(lp->lib_module->wrm_refcount) - 2 );
            rval = setInt( &builder, i, 3, lp->lib_module->wrm_uid );
            rval = setLong( &builder, i, 4, lp->lib_module->wrm_loadtime );
            rval = setPtr( &builder, i, 5,
                    ( void * )lp->lib_module->wrm_linked->linker_file_addr.addr );
            rval = setLong( &builder, i, 6,
                    lp->lib_module->wrm_linked->linker_file_size);

            symbuf = wss_malloc( symbufsize );
            if( symbuf == NULL ) {
                wss_lib_rele( lp );

                rval = -ENOMEM;
                break;
            }

            symbufptr = symbuf;
            symbuflen = 0;
            for( symtab = lp->lib_symtab, sym = 0;
                    symtab[sym].name != NULL;
                    sym++ )
            {
                symlen = wsi_strlen( symtab[sym].name )+ 1;

                if( symbuflen + symlen >= symbufsize ) {
                    pd_char_t *tmp = wss_malloc( symbufsize + 6000 );

                    bcopy( symbuf, tmp, symbufsize );
                    symbuf = tmp;
                    symbufptr = symbuf + symbuflen;
                }

                wsi_strcpy( symbufptr, symtab[sym].name );
                wsi_strcat( symbufptr, PD_MAKESTR(",") );
                symbuflen += symlen;
            }

            wss_lib_rele( lp );

            rval = setString( &builder, i, 6, symbuf );
            wss_free( symbuf );
        }
    }

    if( builder.needed > *bufferSize )
        *bufferSize = builder.needed;

    return rval;
}


/********************************************************************/

static int
_wrv_get_module_by_id( u_long id, pd_char_t *buf, ssize_t *bufferSize )
{
    struct bufferBuilder  builder;
    wss_mod_t            *wrm;
    int                   i;
    int                   rval = 0;
    int                   row;

#ifdef DEBUG_WRV
    wsi_printf( PD_MAKESTR("_getModuleById( %lu, %p, %d\n"), id, buf, *bufferSize );
#endif

    if( id ) {
        wrm = wss_mod_find( id );

        if( wrm == NULL )
            rval = -ENOENT;
        else {
            wss_mod_rele( wrm );
            rval = initBuffer( &builder, buf, *bufferSize, 1, 5 );
        }
    } else {
        WSS_MUTEX_ENTER( wss_module_list );
        for( wrm = wss_module_list, i = 0;
                wrm;
                wrm = wrm->wrm_next, i++ )
            ;
        WSS_MUTEX_EXIT( wss_module_list );

        rval = initBuffer( &builder, buf, *bufferSize, i, 5 );
    }

    WSS_MUTEX_ENTER( wss_module_list );
    for( wrm = wss_module_list, row = 0; wrm; wrm = wrm->wrm_next ) {

        if( ( id == 0 ) || ( id == wrm->wrm_id ) ) {

            rval = setInt( &builder, row, 0, wrm->wrm_id );
            rval = setString( &builder, row, 1,
                    wrm->wrm_linked->linker_file_name );

            if( wrm->wrm_type == WR_MOD_ACTIVATION )
                rval = setString( &builder, row, 2, PD_MAKESTR("activation") );
            else if( wrm->wrm_type == WR_MOD_DEACTIVATION )
                rval = setString( &builder, row, 2,PD_MAKESTR("deactivation"));
            else if( wrm->wrm_type == WR_MOD_WRAPPER )
                rval = setString( &builder, row, 2, PD_MAKESTR("wrapper") );
            else if( wrm->wrm_type == WR_MOD_LIBRARY )
                rval = setString( &builder, row, 2, PD_MAKESTR("library") );
            else
                rval = setString( &builder, row, 2, PD_MAKESTR("unknown") );

            rval = setPtr( &builder, row, 3,
                    ( void * )wrm->wrm_linked->linker_file_addr.addr );
            rval = setLong( &builder, row, 4,
                    wrm->wrm_linked->linker_file_size );

            row++;

            if( id != 0 )
                break;
        }
    }
    WSS_MUTEX_EXIT( wss_module_list );

    if( builder.needed > *bufferSize )
        *bufferSize = builder.needed;

    return rval;
}


/********************************************************************/
static int _wrv_get_module_id_by_name( pd_char_t *buf, ssize_t *bufferSize )
{
    wss_mod_t  *wrm;
    int         rval = 0;
    int         temp;

    if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 ) {

        pd_char_t *buffer = (pd_char_t *)wss_malloc( *bufferSize );

        if( buffer != NULL ) {
            copyin( buf, buffer, *bufferSize );

            WSS_MUTEX_ENTER( wss_module_list );
            for( wrm = wss_module_list; 
                    wrm && wrm->wrm_linked &&wrm->wrm_linked->linker_file_name&&
                        wsi_strcmp( wrm->wrm_linked->linker_file_name, buffer );
                    wrm = wrm->wrm_next )
                ;

            if( wrm )
                copyout( buf, &wrm->wrm_id, sizeof(wrm->wrm_id));
            else {
                temp = 0;
                copyout( buf, &temp, sizeof(int));
            }
            WSS_MUTEX_EXIT( wss_module_list );
        } else
            rval = -ENOMEM;
    }

    return rval;
}

/********************************************************************/
static int
_wrv_get_proc_by_pid( u_long id, pd_char_t *buf, ssize_t *bufferSize )
{
    struct wss_proc      *wrp;
    struct bufferBuilder  builder;
    int                   rval     = 0;
    int                   i;
    int                   row;

    if( id ) {
        wrp = wss_pfind( id );
        if( wrp ) {
            wss_proc_rele( wrp );
            rval = initBuffer( &builder, buf, *bufferSize, 1, 7 );
        } else
            rval = -ENOENT;
    } else {
        WSS_MUTEX_ENTER( wss_proc_list );
        for( wrp = wss_proc_list.lh_first, i = 0;
                wrp;
                wrp = wrp->wrp_p_list.le_next, i++ )
            ;
        WSS_MUTEX_EXIT( wss_proc_list );

        rval = initBuffer( &builder, buf, *bufferSize, i, 7 );
    }

    WSS_MUTEX_ENTER( wss_proc_list );
    for( wrp = wss_proc_list.lh_first, row = 0;
            wrp;
            wrp = wrp->wrp_p_list.le_next )
    {
        pid_t pid = wrp->wrp_pid;

        if( ( pid == id ) || ( id == 0 ) ) {
            wss_wrapper_instance_t *wi;
            int                     size = 0;

            WSS_MUTEX_ENTER( wrp->wrp_wi_list );
            for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
                    wi;
                    wi = wi->wi_next )
                if( wi->wi_wrapper )
                    size += wsi_strlen(wi->wi_wrapper->w_module->
                                        wrm_linked->linker_file_name )+ 1;

            rval = setInt( &builder, row, 0, pid );
            rval = setString( &builder, row, 1, wrp->wrp_progname );
            rval = setInt( &builder, row, 2, wrp->wrp_p->uid );
            rval = setLong( &builder, row, 3, 
                    wsi_time_get_current(NULL) - wrp->wrp_p->times.tms_utime);
            rval = setString( &builder, row, 4, wrp->wrp_path );
            WSI_RW_ENTER( wrp->wrp_cwd, RW_READER );
            rval = setString( &builder, row, 5, wrp->wrp_cwd );
            WSI_RW_EXIT( wrp->wrp_cwd );

            if( size ) {
                pd_char_t *buf;
                buf = (pd_char_t *)wss_malloc( size );
                if( buf == NULL ) {
                    WSS_MUTEX_EXIT( wrp->wrp_wi_list );
                    rval = -ENOMEM;
                    break;
                }
                bzero( buf, size );
                for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
                        wi
                        ; wi = wi->wi_next )
                {
                    if( wi->wi_wrapper ) {
                        if( *buf )
                            wsi_strcat( buf, PD_MAKESTR(",") );
                        wsi_strcat( buf, wi->wi_wrapper->w_module->
                                            wrm_linked->linker_file_name );
                    }
                }
                rval = setString( &builder, 0, 6, buf );
                wss_free( buf );
            } else
                rval = setString( &builder, 0, 6, PD_MAKESTR("") );
            WSS_MUTEX_EXIT( wrp->wrp_wi_list );

            wss_proc_rele( wrp );

            row++;

            if( id == pid )
                break;
        }
    }
    WSS_MUTEX_EXIT( wss_proc_list );

    if( builder.needed > *bufferSize )
        *bufferSize = builder.needed;

    return rval;
}


/********************************************************************/
static int
_wrv_get_procs_by_wrapper_id( u_long id, pd_char_t *buf, ssize_t *bufferSize )
{
    struct wss_proc      *wrp;
    struct bufferBuilder  builder;
    int                   rval     = 0;
    int                   i;
    int                   row;

    i = 0;

    if( id ) {
        WSS_MUTEX_ENTER( wss_proc_list );
        for( wrp = wss_proc_list.lh_first; wrp; wrp = wrp->wrp_p_list.le_next) {
            wss_wrapper_instance_t *wi;

            WSS_MUTEX_ENTER( wrp->wrp_wi_list );
            for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
                    wi;
                    wi = wi->wi_next )
                if( wi->wi_wrapper && (wi->wi_wrapper->w_module->wrm_id == id))
                    i++;
            WSS_MUTEX_EXIT( wrp->wrp_wi_list );
        }
        WSS_MUTEX_EXIT( wss_proc_list );

        if( i == 0 )
            return -ENOENT;
    } else {
        WSS_MUTEX_ENTER( wss_proc_list );
        for( wrp = wss_proc_list.lh_first;
                wrp;
                wrp = wrp->wrp_p_list.le_next, i++ )
            ;
        WSS_MUTEX_EXIT( wss_proc_list );
    }

    rval = initBuffer( &builder, buf, *bufferSize, i, 6 );

    WSS_MUTEX_ENTER( wss_proc_list );
    for( wrp = wss_proc_list.lh_first, row = 0;
            wrp;
            wrp = wrp->wrp_p_list.le_next )
    {
        wss_wrapper_instance_t *wi;
        int                     size  = 0;
        int                     found = 0;

        WSS_MUTEX_ENTER( wrp->wrp_wi_list );
        for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
                wi;
                wi = wi->wi_next )
        {
            if( wi->wi_wrapper ) {
                size += wsi_strlen( wi->wi_wrapper->w_module->wrm_linked->
                                    linker_file_name )+ 1;
                if( id && ( wi->wi_wrapper->w_module->wrm_id == id ) )
                    found++;
            }
        }
        WSS_MUTEX_EXIT( wrp->wrp_wi_list );

        if( id && !found )
            continue;

        rval = setInt( &builder, row, 0, wrp->wrp_pid );
        rval = setString( &builder, row, 1, wrp->wrp_progname );
        rval = setInt( &builder, row, 2, wrp->wrp_p->uid );
        rval = setString( &builder, row, 3, wrp->wrp_path );
        WSI_RW_ENTER( wrp->wrp_cwd, RW_READER );
        rval = setString( &builder, row, 4, wrp->wrp_cwd );
        WSI_RW_EXIT( wrp->wrp_cwd );

        if( size ) {

            pd_char_t *buf;
            buf = (pd_char_t *)wss_malloc( size );
            if( buf == NULL ) {
                rval = -ENOMEM;
                break;
            }
            bzero( buf, size );
            WSS_MUTEX_ENTER( wrp->wrp_wi_list );
            for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
                    wi;
                    wi = wi->wi_next )
            {
                if( wi->wi_wrapper ) {
                    if( *buf )
                        wsi_strcat( buf, PD_MAKESTR(",") );
                    wsi_strcat( buf, wi->wi_wrapper->w_module->
                                    wrm_linked->linker_file_name );
                }
            }
            WSS_MUTEX_EXIT( wrp->wrp_wi_list );
            rval = setString( &builder, row, 5, buf );
            wss_free( buf );
        } else
            rval = setString( &builder, row, 5, PD_MAKESTR("") );

        row++;
    }
    WSS_MUTEX_EXIT( wss_proc_list );

    if( builder.needed > *bufferSize )
        *bufferSize = builder.needed;

    return rval;
}


/********************************************************************/
/*
 * Return the column names, data type, and max length
 * for the specified table
 */

static int 
_wrv_get_table_column_def( u_long handle, pd_char_t *buf, ssize_t *bufferSize )
{
    int  rval  = 0;
    int         i;
    wqcd_t      col_defs;
    wqtab_t     tab;
    int         record_count = 0;
    int         offset = 0;

    tab = wss_db_find_table_by_handle( handle );
    if( tab == NULL ) {
#ifdef DEBUG_WRV
        wsi_printf( PD_MAKESTR("No such table found.\n") );
#endif
        return -ENOENT;
    }

    if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 ) {
        record_count = tab->col_count;
#ifdef DEBUG_WRV
        wsi_printf( PD_MAKESTR("Table has %d columns.\n"),tab->col_count );
#endif

        offset = sizeof(wqcd) * record_count;
        if( offset > *bufferSize ) {
#ifdef DEBUG_WRV
            wsi_printf(PD_MAKESTR("No space for wqcd array.\n"));
#endif
            rval = -EAGAIN;
        }

        col_defs = (wqcd_t)( ( tab->col_defs )?
                DB_REL2ABS(tab->col_defs):NULL );

        for( i = 0; i < record_count ; i++ ) {
            wqcd_t cd = col_defs + i;
            wqcd_t ncd = (wqcd_t)buf + i;
            pd_char_t *name = DB_REL2ABS(cd->name);
            int str_len;

            if( !rval ) {
#ifdef DEBUG_WRV
                int error =
#endif
                    copyout( cd, ncd, sizeof(wqcd) );
#ifdef DEBUG_WRV
                wsi_printf(PD_MAKESTR("copyout 1 returned %d\n"), error );
#endif
            }


            str_len = wsi_strlen( name )+ 1;
            if( !rval && ( ( offset + str_len ) > *bufferSize ) ) {
#ifdef DEBUG_WRV
                wsi_printf(PD_MAKESTR("No space for string.\n"));
#endif
                rval = -EAGAIN;
            }
            if( !rval ) {
                int error1, error2;
                error1 = copyout( name, buf + offset, str_len );
                name = buf + offset;
                error2 = copyout( &name, &ncd->name, sizeof(name) );
#ifdef DEBUG_WRV
                wsi_printf(PD_MAKESTR("copyout 2 returned %d\n"), error1 );
                wsi_printf(PD_MAKESTR("copyout 3 returned %d\n"), error2 );
                name = cd->name ? DB_REL2ABS(cd->name): NULL;
                wsi_printf( PD_MAKESTR("Name %s.\n"),name?name:PD_MAKESTR("( null )") );
#endif
            }
            offset += str_len;
        }
    }
#ifdef DEBUG_WRV
    else
        wsi_printf( PD_MAKESTR("buffer failed onwership test.\n"));
#endif

    *bufferSize = offset;
    if( rval >= 0 )
        rval = record_count;
#ifdef DEBUG_WRV
    else if( rval == -EAGAIN )
        wsi_printf( PD_MAKESTR("No memory.\n") );
#endif

    /* find_table_by_handle returns a locked table. */
    wss_db_unlock_table( tab );
    return rval;
}


/********************************************************************/

    static int 
_wrv_get_table_handle_by_name( pd_char_t *buf, ssize_t *bufferSize )
{
    int rval;
    int        offset = 0;

    /*
     * Return the table id pair for the named table
     */
    wqtab_t tab;

    if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 ) {
        pd_char_t *buffer = (pd_char_t *)wss_malloc( *bufferSize );

        if( buffer == NULL ) {
            rval = -ENOMEM;
        } else {
            copyin( buf, buffer, *bufferSize );

            tab = wss_db_find_table_by_name( buffer );

            if( tab == NULL )
                rval = -ENOENT;
            else {
                wss_wrv_table_id_t *nid = (wss_wrv_table_id_t *)buf;
                pd_char_t               *name;
                size_t                    len;

                offset = sizeof(wss_wrv_table_id_t);

                if( offset > *bufferSize )
                    rval = -EAGAIN;

                if( !rval ) {
                    u_long handle = (u_long)DB_ABS2REL(tab);
                    copyout( &handle, &nid->handle, sizeof(u_long) );
                }

                name = DB_REL2ABS(tab->name);
                len = wsi_strlen( name ) + 1;
                if( offset + len  > *bufferSize )
                    rval = -EAGAIN;

                if( !rval ) {
                    void *addr = buf + offset;

                    copyout( name, buf + offset, len );
                    copyout( &addr, &nid->name, sizeof(void *) );
                }
                offset += len;

                wss_db_unlock_table( tab );
            }
        }
    }

    *bufferSize = offset;
#ifdef DEBUG_WRV
    if( rval == -EAGAIN )
        wsi_printf( PD_MAKESTR("No memory.\n") );
#endif

    return rval;
}


/********************************************************************/

/*
 * Return the table ids for requested tables
 */

static int 
_wrv_get_table_handles( pd_char_t *buf, ssize_t *bufferSize )
{
    int                 rval         = 0;
    u_long             *handles;
    unsigned int        size;
    wss_wrv_table_id_t *prev         = NULL;
    wss_wrv_table_id_t *nid;
    u_long              rel          = 0;
    int                 hindex       = 0;
    int                 i;
    int                 record_count = 0;
    int                 offset       = 0;
    wqtab_t             tab;

#ifdef DEBUG_WRV
    wsi_printf( PD_MAKESTR("In _getTableHandles.\n") );
#endif

    if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 ) {
        pd_char_t *buffer = (pd_char_t *)wss_malloc( *bufferSize );

        if( buffer == NULL ) {
            rval = -ENOMEM;
        } else {
            int err;

            err = copyin( buf, buffer, *bufferSize );
#ifdef DEBUG_WRV
            wsi_printf( PD_MAKESTR("Copyin returned %d: bufferSize = %d\n"),err, *bufferSize );
#endif

            /*
             * Determine number of table handles passed in
             */
            for( handles = (u_long *)buffer, size = 0; *handles;
                    handles++, size++ )
                ;

#ifdef DEBUG_WRV
            wsi_printf( PD_MAKESTR("size = %d\n"),size );
#endif

            /*
             * Allocate space for copying handles and copy them
             */
            if( size > 0 ) {
                handles = (u_long *)wss_malloc( sizeof(u_long) * size );
                if( handles == NULL )
                    rval = -ENOMEM;
                else
                    bzero( handles, sizeof(u_long) * size );
            }

            for( tab = (wqtab_t)((DBHEAD->db_tables)?
                        DB_REL2ABS(DBHEAD->db_tables):NULL);
                    tab;
                    tab = (wqtab_t)((tab->next)?
                        DB_REL2ABS(tab->next):NULL))
            {
                if( ( rval != -ENOMEM ) && ( size > 0 ) ) {
                    /*
                     * We only want the specified tables
                     */
                    rel = (u_long)DB_ABS2REL(tab);
                    for( i = 0;
                            ( i < size ) && ( *( (u_long *)buffer + i )!= rel );
                            i++ )
                        ;

                    if( i >= size )
                        continue;
                    handles[i] = *( (u_long *)buffer + i );
                }
                record_count++;
            }
#ifdef DEBUG_WRV
            wsi_printf( PD_MAKESTR("record_count = %d\n"),record_count );
#endif

            offset = sizeof(wss_wrv_table_id_t) * record_count;
            if( offset > *bufferSize )
                rval = -EAGAIN;

            if( size > 0 ) {
                if( ( rval != -ENOMEM ) && ( record_count > 0 ) ) {
                    /*
                     * Find first valid handle
                     */
                    for( hindex = 0; handles[hindex] == 0; hindex++ );

                    tab = (wqtab_t)DB_REL2ABS(handles[hindex]);
                    rel = handles[hindex];
                } else
                    tab = NULL;
            } else {
                tab = (wqtab_t)( (DBHEAD->db_tables)?
                        DB_REL2ABS(DBHEAD->db_tables):NULL );
#ifdef DEBUG_WRV
                if( *wss_vm_root_table == NULL )
                    wsi_printf( PD_MAKESTR("No tables yet.\n") );
#endif
                rel = (u_long)(DBHEAD->db_tables);
            }

            for( i = 0; tab && ( i < record_count ); i++ ) {
                pd_char_t *name;

                nid = (wss_wrv_table_id_t *)buf + i;
                if( !rval && prev ) {
                    void *ptr = nid;

                    copyout( &ptr, &prev->next, sizeof(ptr) );
                }

                prev = nid;

                if( !rval ) {
                    /*
                     * We'll use the relative address as the handle
                     */
                    u_long handle = (u_long)DB_ABS2REL(tab);
                    err = copyout( &handle, &nid->handle, sizeof(handle) );
#ifdef DEBUG_WRV
                    wsi_printf( PD_MAKESTR("copyout 1 returned %d\n"),err );
#endif
                }

                /*
                 * Copy the table name into the id pair
                 */
                name = DB_REL2ABS(tab->name);

                if( ( offset + wsi_strlen( name )+ 1 )> *bufferSize )
                    rval = -EAGAIN;

                if( !rval ) {
                    void *addr;
                    err = copyout( name, buf + offset, wsi_strlen( name )+ 1 );
#ifdef DEBUG_WRV
                    wsi_printf( PD_MAKESTR("copyout 2 returned %d\n"),err );
#endif
                    addr = buf + offset;
                    err = copyout( &addr, &nid->name, sizeof(void *) );
#ifdef DEBUG_WRV
                    wsi_printf( PD_MAKESTR("copyout 4 returned %d\n"),err );
#endif
                }
                offset += wsi_strlen( name )+ 1;

                /*
                 * Get next table to construct id
                 */
                if( size > 0 ) {
                    if( !rval && ( ( i + 1 ) < record_count ) ) {
                        /*
                         * Find next valid handle
                         */
                        for( ; handles[hindex] == 0; hindex++ )
                            ;

                        tab = ( wqtab_t )
                            DB_REL2ABS(handles[hindex]);
                        rel = handles[hindex];
                    } else
                        tab = NULL;
                } else {
                    rel = (u_long)tab->next;
                    tab = (wqtab_t)( ( tab->next )?
                            DB_REL2ABS(tab->next):NULL );
                }

                if( !rval ) {
                    void *next = NULL;
                    if( tab )
                        next = nid + 1;
                    err = copyout( &next, &nid->next, sizeof(void *) );
#ifdef DEBUG_WRV
                    wsi_printf( PD_MAKESTR("copyout 3 returned %d\n"),err );
#endif
                }
            }

            /*
             * Free memory used by handles array
             */
            if( ( size > 0 ) && handles )
                wss_free( (void *)handles );
        }
    }

    *bufferSize = offset;
    if( rval >= 0 )
        rval = record_count;
#ifdef DEBUG_WRV
    else if( rval == -EAGAIN )
        wsi_printf( PD_MAKESTR("No memory.\n") );
#endif

    return rval;
}

/********************************************************************/

/*
 * Return the requested rows for the specified table
 */

static int 
_wrv_get_table_rows( u_long handle, pd_char_t *ubuf, ssize_t *bufferSize )
{
    int                 rval = 0;
    int                 record_count = 0;
    int                 offset=0, string_offset = 0;
    int                 i;
    u_long              *handles;
    unsigned int        size;
    wqrow_t             row;
    int                 hindex = 0;
    u_long              rel = 0;
    wqtab_t             tab;

    if( ( tab = wss_db_find_table_by_handle( handle ) ) == NULL )
        return -ENOENT;

    /*
     * Determine number of table handles passed in
     */
#ifdef DEBUG_WRV
    wsi_printf( PD_MAKESTR("_wrv_get_table_rows\n") );
#endif /* DEBUG_WRV */
    if( ( rval = checkBuffer( ubuf, *bufferSize ) ) == 0 ) {
        pd_char_t *kbuf = (pd_char_t *)wss_malloc( *bufferSize );

        if( !kbuf )
            rval = -ENOMEM;
        else {
            copyin( ubuf, kbuf, *bufferSize );

            for(    handles = (u_long *)kbuf + 1, size = 0; 
                    *handles;
                    handles++, size++ )
                ;

            /*
             * Allocate space for copying handles and copy them
             */
            if( size > 0 ) {
                handles = (u_long *)wss_malloc( sizeof(u_long) * size );
                if( handles == NULL )
                    rval = -ENOMEM;
                else
                    bzero( handles, sizeof(u_long) * size );
            }

            for(    row = (wqrow_t)( ( tab->rows )?
                        DB_REL2ABS(tab->rows):NULL ); 
                    row;
                    row = (wqrow_t)( ( row->next )?
                        DB_REL2ABS(row->next):NULL ) )
            {
                u_long rel = (u_long)DB_ABS2REL(row);

                if( !rval && size ) {
                    /*
                     * We only want the specified rows
                     */
                    int i;
                    for( i = 0;
                            ( i < size ) && ( *( (u_long *)kbuf + i ) != rel );
                            i++ );
                    if( i >= size )
                        continue;

                    handles[i] = *( (u_long *)kbuf + i );
                }

                record_count++;
            }

            /* offset starts at the beginning of the table data. */
            offset = sizeof(wss_wrv_db_row_t) * record_count;

            /* string_offset starts at the beginning of the string table */
            string_offset = offset + sizeof(wqval)*tab->col_count*record_count;

            if ( string_offset > *bufferSize )
                rval = -EAGAIN;

            if( size > 0 ) {
                if( !rval && ( record_count > 0 ) ) {
                    /*
                     * Find first valid handle
                     */
                    for( hindex = 0; handles[hindex] == 0; hindex++ );

                    row = (wqrow_t)DB_REL2ABS(handles[hindex]);
                    rel = handles[hindex];
                } else
                    row = NULL;
            } else {
                rel = (u_long)tab->rows;
                row = (wqrow_t)( ( tab->rows )?
                        DB_REL2ABS(tab->rows):NULL );
            }

            for( i = 0; row && ( i < record_count ); i++ ) {
                wss_wrv_db_row_t *nrow = (wss_wrv_db_row_t *)ubuf + i;
                wqval_t           nvalues;
                wqcd_t            cd;
                wqval_t           val, val_list;
                int               k;
                void             *next;

                if( !rval )
                    copyout( &rel, &nrow->handle, sizeof(rel));

                /*
                 * Determine number of values in row
                 */
                if( !rval )
                    copyout(
                            &tab->col_count,
                            &nrow->numvals,
                            sizeof(tab->col_count));

                nvalues = (wqval_t)( ubuf + offset );
                offset += sizeof(wqval) * tab->col_count;

                /*
                 * Copy row values
                 */
                cd = (wqcd_t)( ( tab->col_defs )?
                        DB_REL2ABS(tab->col_defs):NULL);
                val_list = (wqval_t)( ( row->vals )?
                        DB_REL2ABS(row->vals):NULL );

                for( k = 0; k < tab->col_count; k++ ) {
                    wqval_t nval = nvalues + k;
                    val = val_list + k;

                    if( !rval )
                        copyout( val, nval, sizeof(wqval) );

                    if( cd[k].dtype == pd_char_t_t ) {
                        /*
                         * Check available space for string value
                         */
                        pd_char_t *value = DB_REL2ABS(val->v.cval);

                        if( !rval
                        && ((string_offset+wsi_strlen(value)+1) > *bufferSize))
                            rval = -EAGAIN;

                        if( !rval ) {
                            char *cval = ubuf + string_offset;

                            copyout( value,
                                    ubuf+string_offset, wsi_strlen(value)+1);
                            copyout( &cval, &nval->v.cval, sizeof(cval) );
                        }
                        string_offset += wsi_strlen( value )+ 1;
                    }
                }

                if( !rval )
                    copyout( &nvalues, &( ROW( nrow )->vals ), sizeof(nvalues));

                /*
                 * Get next row to be copied
                 */
                if( size > 0 ) {
                    if( !rval && ( ( i + 1 ) < record_count ) ) {
                        /*
                         * Find next valid handle
                         */
                        for( ; handles[hindex] == 0; hindex++ );

                        row = (wqrow_t)DB_REL2ABS(handles[hindex]);
                        rel = handles[hindex];
                    } else
                        row = NULL;
                } else {
                    rel = (u_long)row->next;
                    row = (wqrow_t)( ( row->next )?
                            DB_REL2ABS(row->next):NULL );
                }

                next = NULL;
                if( row )
                    next = nrow + 1;

                if( !rval )
                    copyout( &next, &ROW( nrow )->next, sizeof(next) );
            }

            /*
             * Free memory used by handles array
             */
            if( ( size > 0 ) && handles )
                wss_free( ( void * )handles );
            wss_free( kbuf );
        }
    }
#if 0
    else
        rval == -EFAULT;
#endif

    *bufferSize = string_offset;
    if( rval >= 0 )
        rval = record_count;
#ifdef DEBUG_WRV
    else if( rval == -EAGAIN )
        wsi_printf( PD_MAKESTR("No memory.\n") );
#endif

    /* find_table_by_handle returns a locked table. */
    wss_db_unlock_table( tab );

    return rval;
}

/********************************************************************/
/* Extracts a row from char *[] array.  
 * 
 * Buffer is assumed to be null terminated at least once.  Can be
 * malformed other than that.
 */
static int
_wrv_add_table_row( u_long thandle, pd_char_t *buffer, size_t size,
                    wqrow_t *ret )
{
   int         i;
   wqrow_t     temp = NULL;
   wqtab_t     tab  = NULL;
   wqval_t     vals = NULL;
   int         rval = 0;
   pd_char_t *end  = buffer + size;

   tab = wss_db_find_table_by_handle(thandle);
   if( tab == NULL )
      rval = -EINVAL;

   if( !rval )
   {
      temp = wss_malloc(sizeof(*temp));
      if( !temp )
         rval = -ENOMEM;
   }

   if( !rval )
   {
      vals = wss_malloc(sizeof(*vals) * tab->col_count);
      if( !temp )
      {
         wss_free(temp);
         rval = -ENOMEM;
      }

      /* zero default */
      bzero( vals, sizeof(*vals) * tab->col_count );
      temp->next = NULL;
      temp->vals = vals;

      /* For each column, get a value. */
      for( i = 0; ( i < tab->col_count ) && ( buffer < end ); i++ )
      {
         wqcd_t abs_col_defs = (wqcd_t)DB_REL2ABS(tab->col_defs);
         switch( abs_col_defs[i].dtype )
         {
            case pd_char_t_t:
               vals[i].v.cval = wss_strdup( buffer );
               break;

            case int_t:
               vals[i].v.ival = wss_atoi( buffer );
               break;

            case float_t:
               vals[i].v.fval = wss_atof( buffer );

            case none_t:
            default:
               wsi_printf( PD_MAKESTR("%s:%d: none_t typed column found!\n"),
                           __FILE__, __LINE__ );
               break;
         }

         buffer += wsi_strlen( buffer ) + 1;
      }

      /* Did we not get enough fields? */
      if( i < tab->col_count )
      {
         for( ; i >= 0; --i )
         {
            if( tab->col_defs[i].dtype == pd_char_t_t )
               wss_free( vals[i].v.cval );
         }
         wss_free( vals );
         wss_free( temp );
         rval = -EINVAL;
      }

      if( !rval )
         *ret = temp;
   }

   /* find_table_by_handle returns a locked table. */
   if( tab )
      wss_db_unlock_table( tab );

   return rval;
}


/********************************************************************/

static int 
_wrv_get_version( pd_char_t *buf, ssize_t *bufferSize )
{
    int rval;

    if( *bufferSize < 2 * sizeof(int) ) {
        *bufferSize = 2 * sizeof(int);
        rval = -EINVAL;
    } else {
        copyout( _versionBuffer, buf, 2 * sizeof(int) );
        rval = 0;
    }

    return rval;
}


/********************************************************************/
static int
_wrv_get_wrapper_by_id( u_long id, pd_char_t *buf, ssize_t *bufferSize )
{
    wss_wrapper_t        *w;
    struct bufferBuilder  builder;
    int                   i;
    int                   rval = 0;
    int                   row;

#ifdef DEBUG_WRV
    wsi_printf( PD_MAKESTR("_getWrapperById( %lu, %p, %d )\n"),
            id, buf, *bufferSize );
#endif

    if( id ) {
        WSS_MUTEX_ENTER( wss_wrapper_list );
        for(    w = wss_wrapper_list;
                w && w->w_module->wrm_id != id;
                w = w->w_next)
            ;
        WSS_MUTEX_EXIT( wss_wrapper_list );

        if( w == NULL )
            rval = -ENOENT;
        else
            rval = initBuffer( &builder, buf, *bufferSize, 1, 7 );
    } else {
        WSS_MUTEX_ENTER( wss_wrapper_list );
        for( w = wss_wrapper_list, i = 0; w; w = w->w_next, i++ );
        WSS_MUTEX_EXIT( wss_wrapper_list );
        rval = initBuffer( &builder, buf, *bufferSize, i, 7 );
    }

    WSS_MUTEX_ENTER( wss_wrapper_list );
    for( w = wss_wrapper_list, row = 0; w; w = w->w_next ) {
        if( ( id == 0 ) || ( id == w->w_module->wrm_id ) ) {
            rval = setInt( &builder, row, 0, w->w_module->wrm_id );
            rval = setString( &builder, row, 1,
                              w->w_module->wrm_linked->linker_file_name );
            rval = setInt( &builder, row, 2,
                           WSS_REF_COUNT(w->w_module->wrm_refcount) - 2 );
            rval = setInt( &builder, row, 3, w->w_module->wrm_uid );
            rval = setLong( &builder, row, 4, w->w_module->wrm_loadtime );
            rval = setPtr( &builder, row, 5,
                           ( void * )w->w_module->wrm_linked->
                                linker_file_addr.addr );
            rval = setLong( &builder, row, 6, w->w_module->wrm_linked->
                                        linker_file_size );
            row++;
        }
    }
    WSS_MUTEX_EXIT( wss_wrapper_list );

    if( builder.needed > *bufferSize )
        *bufferSize = builder.needed;

    return rval;
}


/********************************************************************/
static int
_wrv_set_table_row_field( u_long htable, u_long hrow, pd_char_t *field,
        pd_char_t *value )
{
    int rval = -ENOENT;
    wqrow_t    row;
    wqtab_t    tab;
    wqcd_t     cd;
    int        max;

    /*
     * Validate the table handle
     */
    if( !( tab = wss_db_find_table_by_handle( htable ) ) )
        return rval;

    max = tab->col_count;

    /*
     * Make sure we have some column definitions
     */
    if( ( cd = (wqcd_t)( tab->col_defs?
                    DB_REL2ABS(tab->col_defs):NULL ) ) ) {
        /*
         * Validate row handle
         */
        for(    row = (wqrow_t)( tab->rows?
                    DB_REL2ABS(tab->rows):NULL );
                row && ( (u_long)DB_ABS2REL(row)!= hrow );
                row = (wqrow_t)( row->next?
                    DB_REL2ABS(row->next):NULL ) )
            ;

        if( row ) {
            int i;

            /*
             * Find field
             */
            for(    i = 0;
                    ( i < max ) && wsi_strcmp(
                                (pd_char_t *)DB_REL2ABS(cd[i].name), field );
                    i++ )
                ;

            if( i < max ) {
                wss_proc_t     *wrp;
                /*
                 * Tell WSS to set the value based on the type
                 */
                wss_db_set_row_value_from_string( cd, row, i, value );

                /*
                 * Post update row event
                 */
                wss_wrvevent_post_updaterow( tab, row );

                /*
                 * Post system event
                 */
                wrp = wss_pfind(0);
                if (wrp) {
                    wss_dispatch(wsi_sysevents_getevspace(), 2, wrp, 0,
                            1, DB_REL2ABS(tab->name));
                    wss_proc_rele(wrp);
                }
                rval = 0;
            }
        }
    }

    /* find_table_by_handle returns a locked table. */
    wss_db_unlock_table( tab );

    return rval;
}

/****************************************************************************\
 * Public functions                                                         *
\****************************************************************************/

/*  wrv */
/*! \brief wrv syscall
 *
 *  \param command Identifier of command
 *                   Can be one of the following:
 *                      \b WRV_GET_ACT_CRIT_BY_ID
 *                      \b WRV_GET_DEACT_CRIT_BY_ID
 *                      \b WRV_GET_ACTIVE_WRAPPERS
 *                      \b WRV_GET_LIBRARY_BY_ID
 *                      \b WRV_GET_MODULE_BY_ID
 *                      \b WRV_GET_MODULE_ID_BY_NAME
 *                      \b WRV_GET_PROC_BY_PID
 *                      \b WRV_GET_PROCS_BY_WRAPPER_ID
 *                      \b WRV_GET_TABLE_COLUMN_DEF
 *                      \b WRV_GET_TABLE_HANDLE_BY_NAME
 *                      \b WRV_GET_TABLE_HANDLES
 *                      \b WRV_GET_TABLE_ROWS
 *                      \b WRV_SET_TABLE_ROW_FIELD
 *                      \b WRV_ADD_TABLE_ROW
 *                      \b WRV_DEL_TABLE_ROW
 *                      \b WRV_GET_VERSION
 *                      \b WRV_GET_WRAPPER_BY_ID
 *  \param args    Pointer to argument structure
 *  \param bufsize Pointer to size of buffer passed in and/or needed
 *
 *  \return Return value on success,
 *  \return EPERM if \b SUSER_ONLY is defined and user is not root,
 *  \return ENOTSUP if the command is not supported,
 *  \return ENOMEM if we have exhausted virtual memory,
 *  \return EINVAL if an invalid parameter is passed,
 *  \return EAGAIN if passed-in buffer is too small,
 *  \return E2BIG  if more rows or columns are added than was allocated,
 *  \return EFAULT if the passed-in buffer is not in user-space
 */
asmlinkage
int wrv( int command, void *args, size_t *bufsize )
{
    int              rval       = 0;
    ssize_t          bufferSize = 0;
    struct wss_proc *wrp;

#ifdef SUSER_ONLY
    if( !suser( ) )
        return -EPERM;
#endif

    if( ( wrp = wss_pfind( wsi_proc_get_current_pid() ) ) ) {
        int badsys = 0;
        if( wrp->wrp_wi_list )
            badsys = 1;

        wss_proc_rele( wrp );

        if( badsys )
            return -ENOSYS;
    }

    if( bufsize && (rval = copyin(bufsize, &bufferSize, sizeof(bufferSize))))
        return rval;

    switch( command )
    {
    case WRV_GET_ACT_CRIT_BY_ID:
    case WRV_GET_DEACT_CRIT_BY_ID:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_criteria_by_id( command, data.handle,
                        (pd_char_t *)args, &bufferSize );
            else
                rval = -rval;
            break;
        }

    case WRV_GET_ACTIVE_WRAPPERS:
        rval = _wrv_get_active_wrappers( (pd_char_t *)args, &bufferSize );
        break;

    case WRV_GET_LIBRARY_BY_ID:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_library_by_id( data.handle, (pd_char_t *)args,
                        &bufferSize );
            else
                rval = -rval;

            break;
        }

    case WRV_GET_MODULE_BY_ID:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_module_by_id( data.handle,(pd_char_t *)args,
                        &bufferSize );
            else
                rval = -rval;

            break;
        }

    case WRV_GET_MODULE_ID_BY_NAME:
        rval = _wrv_get_module_id_by_name( (pd_char_t *)args, &bufferSize );
        break;

    case WRV_GET_PROC_BY_PID:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_proc_by_pid( data.handle,(pd_char_t *)args,
                        &bufferSize );
            else
                rval = -rval;

            break;
        }

    case WRV_GET_PROCS_BY_WRAPPER_ID:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_procs_by_wrapper_id( data.handle,
                        (pd_char_t *)args, &bufferSize );
            else
                rval = -rval;

            break;
        }

    case WRV_GET_TABLE_COLUMN_DEF:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_table_column_def( data.handle,
                        (pd_char_t *)args, &bufferSize );
            else
                rval = -rval;

            break;
        }

    case WRV_GET_TABLE_HANDLE_BY_NAME:
        rval = _wrv_get_table_handle_by_name( (pd_char_t *)args, &bufferSize );
        break;

    case WRV_GET_TABLE_HANDLES:
        rval = _wrv_get_table_handles( (pd_char_t *)args, &bufferSize );
        break;

    case WRV_GET_TABLE_ROWS:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_table_rows( data.handle,
                        (pd_char_t *)args, &bufferSize );
            else
                rval = -rval;

            break;
        }

    case WRV_SET_TABLE_ROW_FIELD:
        {
            wrv_trnd_t id;
            pd_char_t *data;
            int dsize = bufferSize - sizeof(id);

            id.name[MAXNAMELEN-1] = '\0';
            if( dsize > 0 ) {
                data = wss_malloc( dsize );
                if( data == NULL ) {
                    rval = ENOMEM;
                    break;
                }
            } else {
                data = NULL;
                rval = EINVAL;
            }

            if( !( rval = copyin( args, &id, sizeof(id) ) ) &&
                    data &&
                    !( rval = copyin( (pd_byte_t *)args + sizeof(id),
                            data, dsize ) ) )
                rval = _wrv_set_table_row_field( id.thandle, id.rhandle,
                        id.name, data );
            else
                rval = -rval;

            if( data )
                wss_free( data );

            break;
        }

    case WRV_ADD_NULL_TABLE_ROW:
        {
            wrv_h_t data;
            if( !( rval = copyin( args, &data, sizeof( data ) ) ) )
                rval = wss_db_insert_nulls( data.handle );
            else
                rval = -rval;

            break;
        }

    case WRV_ADD_TABLE_ROW:
        {
            wrv_h_t     id;
            wqrow_t     row;
            pd_char_t *buffer;
            size_t      dsize    = bufferSize - sizeof(id) + 1;

            if( sizeof(id) >= bufferSize ) {
                rval = -EINVAL;
                break;
            }
            if( ( buffer = wss_malloc( dsize ) ) == NULL ) {
                rval = -ENOMEM;
                break;
            }
            if( ( rval = copyin( args, &id, sizeof(id) ) ) ||
                    ( rval = copyin( (pd_byte_t *)args + sizeof(id), buffer,
                                    dsize ) ) )
            {
                rval = -rval;
                break;
            }
            buffer[dsize - 1] = PD_MAKESTR('\0');
            if( !( rval = _wrv_add_table_row( id.handle, buffer, dsize, &row)))
                switch( wss_insert( id.handle, 1, row ) )
                {
                case 0:
                    break;

                case -WRE_NOMEM:
                    rval = -ENOMEM;
                    break;

                case -WRE_NOTFOUND:
                    rval = -ENOENT;
                    break;

                case -WRE_INVALID_PARAMETER:
                default:
                    rval = -EINVAL;
                    break;
                }

            wss_free(buffer);
            break;
        }

    case WRV_DEL_TABLE_ROW:
        {
            wrv_tr_t data;
            if( !( rval = copyin( args, &data, sizeof( data ) ) ) )
                rval = wss_db_delete_by_handle( data.thandle, data.rhandle );
            else
                rval = -rval;

            break;
        }

    case WRV_GET_VERSION:
        rval = _wrv_get_version( (pd_char_t *)args, &bufferSize );
        break;

    case WRV_GET_WRAPPER_BY_ID:
        {
            wrv_h_t data;

            if( !( rval = copyin( args, &data, sizeof(data) ) ) )
                rval = _wrv_get_wrapper_by_id( data.handle,(pd_char_t *)args,
                        &bufferSize );
            else
                rval = -rval;

            break;
        }

    case WRV_ADMIN_DONE:
       wss_crit_admin_done = TRUE;
       break;

    default:
        rval = -EINVAL;
    }

    copyout( &bufferSize, bufsize, sizeof(bufferSize) );

#ifdef DEBUG_WRV
    wsi_printf( PD_MAKESTR("wrv retval: %d\n"),rval );
#endif
    return rval;
}

