/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_execve.c,v 1.5 2001/02/23 19:02:40 stevek Exp $
 *
 * Author: Chris Vance, Steve Kiernan
 *
 * Purpose: This file contains a replacement for the execve syscall.
 *          It does some wrappers-relevant work and then passes control
 *          off to the original execve system call.
 */

#include <wr_conf.h>

/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/* 
 * This is the replacement function for the execve system call.  It
 * does some wrappers-specific stuff, and then calls the original
 * execve syscall.  Of importance is the need to look up the full
 * pathname for the program being executed.  This should only need to
 * be done if the exec was a success.  With any luck, most of the
 * processing it does has been cached during the similar lookup in the
 * actual syscall.  The program name is also available in argv[0]
 * (relative to the cwd) and is stored in p->p_comm by the execve
 * syscall (only the first 16 chars of the binary name - no path).
 * 
 * After the execve call, run "wss_crit_check_deactivation()" and 
 * "wss_crit_check_activation()", to stop any wrappers that should be
 * stopped and start any wrappers that should be started.
 * 
 * Then run "wss_wrapper_call_activation()" to call the wr_activate
 * functions of the wrappers.
 */
int
wsi_execve( struct execve_uap *uap, rval_t *retval )
{
   struct wss_proc   *wrp;
   struct proc       *p;
   int                err;
   int                act_error;
   char              *tmp;
   char              *path;
   int                len;
   int                i;


   /*
    * Look up the current process in the wrappers proc structure. If it
    * wasn't found, this means this is a process that was running before
    * the wrappers module was loaded into the kernel.  In the this case,
    * we can probably work around it by creating the structure at this time.
    */
   p = curproc;
   wrp = wss_pfind_r( p->p_pid );
   if( !wrp )
   {
      printf( "wsi_execve: can't find pid: p->p_pid %d\n", (int)p->p_pid );

      wrp = wss_proc_new( p );
   }

   /*
    * wrp was tested for validity early on.  Make space and copy
    * the path, first saving the current one in an 'old' pathbuf
    */
   if( wrp->opathbuf )
      wss_free( wrp->opathbuf );

   tmp = wrp->cwd;
   wrp->cwd = wss_path_cwd(wrp);
   if (tmp)
      wss_free(tmp);

   path = wss_malloc( MAXPATHLEN );

   /* Usually will be a "copyinstr" */
   len = strncpy( path, (char *)uap->path, MAXPATHLEN );

   wrp->opathbuf = wrp->pathbuf; 
   wrp->pathbuf = wss_path_rel_to_abspath( wrp->cwd, path );
   wss_free( path );

   /* Clear out progname if its currently set */
   if( wrp->progname != NULL )
      wss_free( wrp->progname );

   /*
    * Generate the program name, full path should have at least
    * one '/' character.  Program name is everything after the
    * last '/'.
    */
   len = 0;
   i = strlen( wrp->pathbuf );
   for( len = 0, i = strlen( wrp->pathbuf );
        ( i > 0 ) && ( wrp->pathbuf[i] != '/' );
        i--, len++ );

   /*
    * Check (for sanity, etc)
    */
   if( ( i == 0 ) && ( wrp->pathbuf[0] != '/' ) )
   {
      printf( "wsi_execve: Pathname `%s' didn't have a '/' in it\n",
              wrp->pathbuf );
      wrp->progname = NULL;
   }
   else
   {
      wrp->progname = (char *)wss_malloc( len );
      bcopy( &( wrp->pathbuf[i + 1] ), wrp->progname, len );
   }


   /*
    * Next, call the original syscall.
    */ 
   err = (*orig_execve_sysent.sy_call)( uap, retval );

   /* Verify the execve call succeeded */
   if( retval->r_val == 0 )
   {
      /* zero from execve syscall indicates success? */
      int has_wrappers;

      /*
       * We may now check the activation criteria to determine
       * whether or not to run this process wrapped.
       */
      wss_crit_check_deactivation( wrp );
      wss_crit_check_activation( wrp );

      has_wrappers = (wrp->wrappers==NULL)?0:1;

      if( has_wrappers )
      {
         /*
          * Now is a convenient place to call the
          * wrapper-specific wr_activation() functions.
          */
         act_error = wss_wrapper_call_activation( p, wrp );
         if( act_error != 0 )
         {
            printf( "Activate returned an error!\n" );
         }
      }
   }

   return( err );
}

