/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_wrv.c,v 1.9 2001/02/23 19:02:40 stevek Exp $
 *
 * Author: Steve Kiernan, Linda Thomas
 *
 * Purpose: Implementation of the wrv system call
 * 
 */

#include <wr_conf.h>

#include <libwrv.h>


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

/*
 * Version of the WRV interface
 */
#define WRV_VERSION_MAJOR  1
#define WRV_VERSION_MINOR  1


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static int _versionBuffer[] = { WRV_VERSION_MAJOR, WRV_VERSION_MINOR };


/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

/*******************************************************************/

/*
 *  Buffer construction
 *
 *  These buffer macros will lay the buffer out like this:
 *
 *  size_t  # rows in array
 *  size_t  # cols in array
 *  char *  array[0][0]
 *  char *  array[0][1]
 *  char *  array[0][2]
 *  ...
 *  char *  array[0][#cols-1]
 *  char *  array[1][0]
 *  char *  array[1][1]
 *  ...
 *  char *  array[1][#cols-1]
 *  ...
 *  char *  array[#rows-1][0]
 *  char *  array[#rows-1][1]
 *  ...
 *  char *  array[#rows-1][#cols-1]
 *  The remainder of the buffer contains the strings the array
 *  elements point to.
 *
 *  Because of the way data is allocated, it is best to allocate
 *  each element only once.  The order can be arbitrary, but
 *  changing an element will lose access to chunk of buffer it was
 *  previously pointing to.
 *
 */

struct bufferBuilder
{
   void      *buffer;         /* where all the data really goes */
   ssize_t    bufferSize;     /* in bytes */
   size_t     numRows;
   size_t     numCols;
   char     **array;          /* the beginning of the array portion of buffer */
   char      *data;           /* the beginning of the data portion of buffer */
   int        offset;         /* where next data should get put */
   int        remaining;      /* amount of space left free */
   ssize_t    needed;         /* size the buffer needs to be */  
};


/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/

extern void    wss_db_set_row_value_from_string __P(( wqcd_t, wqrow_t, int,
                                                      char * ));
extern wqtab_t find_table                       __P((char *));


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * Validate the buffer
 */
static int
checkBuffer( void *pbuffer, ssize_t bufferSize )
{
   return 0;
}


static int
initBuffer( struct bufferBuilder *builder, void *buffer, ssize_t bufferSize,
            size_t nRows, size_t nCols )
{
   int rval   = -1;

   builder->buffer     = buffer;
   builder->bufferSize = bufferSize;
   builder->numRows    = nRows;
   builder->numCols    = nCols;

   if( ( rval = checkBuffer( buffer, bufferSize ) ) == 0 )
   {
      if( bufferSize < ( ( 2 * sizeof(size_t) )+ 
                         ( nRows * nCols )* sizeof(char *) ) )
      {
         /* buffer not large enough to hold the array framework */
         builder->needed = ( 2 * sizeof(size_t) )*
                           ( nRows * nCols )* sizeof(char *);
         set_errno( ENOMEM );
         rval = -ENOMEM;
      }
      else
      {
         size_t *sizes;
         char   *nullbuf  = NULL;
         int     row;
         int     col;

         sizes = (size_t *)buffer;
         copyout( &nRows, &sizes[0], sizeof(nRows) );
         copyout( &nCols, &sizes[1], sizeof(nCols) );

         builder->array = (char **)&sizes[2];
         for( row = 0; row < nRows; row++ )
         {
            for( col = 0; col < nCols; col++ )
            {
               copyout( &nullbuf, &builder->array[row*nCols+col],
                        sizeof(char *) );
            }
         }

         builder->data = (char *)(&builder->array[nRows * nCols]);
         builder->offset = 0;
         builder->remaining = bufferSize - ( 2 * sizeof(size_t) ) -
                              ( nRows * nCols ) * sizeof(char *);
         builder->needed = ( 2 * sizeof(size_t) ) *
                           ( nRows * nCols ) * sizeof(char *);

         rval = 0;
      }
   }

   return rval;
}

static int
setString( struct bufferBuilder *builder, int row, int col, char *s )
{
   int rval = 0;

#ifdef DEBUG_WRV
   printf("string is '%s'\n", s);
#endif

   if( ( row >= builder->numRows ) || ( col >= builder->numCols ) )
   {
      printf( "wsi_wrv: attempt to set out of range, limit %d, %d, actual %d, "
              "%d\n", builder->numRows, builder->numCols, row, col );
      set_errno( E2BIG );
      rval = -E2BIG;
   }
   else
   {
      if( s == NULL )
      {
         char *nullbuf = NULL;
         copyout( &nullbuf, &builder->array[row*builder->numCols+col],
                  sizeof(char *) );
      }
      else
      {
         int len = strlen( s )+ 1;

         builder->needed += len;
         if( builder->remaining < len )
         {
#ifdef DEBUG_WRV
            printf( "buffer too small to hold '%s'\n", s );
#endif
            set_errno( ENOMEM );
            rval = -ENOMEM;
         }
         else
         {
            char *p = (char *)&builder->data[builder->offset];
            copyout( &p, &builder->array[row*builder->numCols+col], 
                     sizeof(char *) );
            copyout( s, &builder->data[builder->offset], len );
            builder->offset += len;
            builder->remaining -= len;
         }
      }
   }

   return rval;
}

static int
setInt( struct bufferBuilder *builder, int row, int col, int i )
{
   char ibuf[64];

   sprintf( ibuf, "%d", i );
   return setString( builder, row, col, ibuf );
}

static int
setLong( struct bufferBuilder *builder, int row, int col, long l )
{
   char lbuf[64];

   sprintf( lbuf, "%ld", l );
   return setString( builder, row, col, lbuf );
}

static int
setPtr( struct bufferBuilder *builder, int row, int col, void *p )
{
   char pbuf[64];

   sprintf( pbuf, "%p", p );
   return setString( builder, row, col, pbuf );
}


/********************************************************************/

static longlong_t
_wrv_get_active_wrappers( char *buf, ssize_t *bufferSize )
{
   struct bufferBuilder  builder;
   wss_wrapper_t        *wr;
   int                   i;
   int                   rval   = 0;

   WSS_MUTEX_ENTER( wrapper_list );
   for( wr = wrapper_list, i = 0; wr; wr = wr->next )
      if( wr->procs )
         i++;
   WSS_MUTEX_EXIT( wrapper_list );

   rval = initBuffer( &builder, buf, *bufferSize, i, 2 );

   WSS_MUTEX_ENTER( wrapper_list );
   for( wr = wrapper_list; wr; wr = wr->next )
   {
      if( wr->procs )
      {
         rval = setString( &builder, i, 0, wr->name );
         rval = setInt( &builder, i, 1, wr->procs );
      }
   }
   WSS_MUTEX_EXIT( wrapper_list );

   if( builder.needed > *bufferSize )
      *bufferSize = builder.needed;

   return rval;
}

/********************************************************************/

static longlong_t
_wrv_get_criteria_by_id( int command, u_long id, char *buf,
                         ssize_t *bufferSize )
{
   struct bufferBuilder  builder;
   wss_crit_t           *act;
   wss_crit_t           *head_ptr;
   int                   i;
   int                   rval = 0;
   int                   row;

   if( command == WRV_GET_ACT_CRIT_BY_ID )
      head_ptr = wss_act_crit_list;
   else
      head_ptr = wss_deact_crit_list;

   if( id )
   {
      for( act = head_ptr; act && act->id != id; act = act->next );

      if( act == NULL )
      {
         rval = -ENOENT;
         return rval;
      }

      rval = initBuffer( &builder, buf, *bufferSize, 1, 6 );
    
   }
   else
   {
      for( act = head_ptr, i = 0; act; act = act->next, i++ );

      rval = initBuffer( &builder, buf, *bufferSize, i, 6 );
   }

   for( act = head_ptr, row = 0; act; act = act->next )
   {
      if( ( id == 0 ) || ( id == act->id ) )
      {
         rval = setInt( &builder, row, 0, act->id );
         rval = setString( &builder, row, 1, act->name );
         rval = setString( &builder, row, 2,
                           act->status == WRACT_ENABLE ? "enabled" :
                                                         "disabled" );
         rval = setInt( &builder, row, 3, act->module->uid );
         rval = setInt( &builder, row, 4, act->module->loadtime );
         rval = setLong( &builder, row, 5, act->module->size );
         row++;
      }
   }

   return rval;
}


/********************************************************************/

static longlong_t
_wrv_get_library_by_id( u_long id, char *buf, ssize_t *bufferSize )
{
   struct bufferBuilder  builder;
   wss_lib_t            *lp;
   wss_symbol_t         *symtab;
   char                 *symbuf;
   char                 *symbufptr;
   int                  symbuflen;
   int                  symlen;
   int                  sym;
   int                  symbufsize = 2048;
   int                  rval       = 0;
   int                  i;
   int                  count;

   if( id )
   {
      lp = wss_lib_find( id );
      count = id + 1;

      if( lp == NULL )
         return -ENOENT;

      rval = initBuffer( &builder, buf, *bufferSize, 1, 7 );
   }
   else
   {
      for( count = 0; wss_lib_find( count )!= NULL; count++ );

      rval = initBuffer( &builder, buf, *bufferSize, count, 7 );
   }

   for( i = 0; i < count; i++ )
   {
      if( ( id == 0 ) || ( id == i ) )
      {
         lp = wss_lib_find( i );

         rval = setInt( &builder, i, 0, lp->id );
         rval = setString( &builder, i, 1, lp->name );
         rval = setInt( &builder, i, 2,
                        wss_idlist_count_ids( lp->module->used ) );
         rval = setInt( &builder, i, 3, lp->module->uid );
         rval = setLong( &builder, i, 4, lp->module->loadtime );
         rval = setLong( &builder, i, 5, lp->module->size );

         symbuf = wss_malloc( symbufsize );
         symbufptr = symbuf;
         symbuflen = 0;
         WSS_MUTEX_ENTER( lp->symtab );
         for( symtab = lp->symtab, sym = 0; symtab[sym].name != NULL; sym++ )
         {
            symlen = strlen( symtab[sym].name )+ 1;

            if( symbuflen + symlen >= symbufsize )
            {
               char *tmp = wss_malloc( symbufsize + 6000 );

               bcopy( symbuf, tmp, symbufsize );
               symbuf = tmp;
               symbufptr = symbuf + symbuflen;
            }

            strcpy( symbufptr, symtab[sym].name );
            strcat( symbufptr, "," );
            symbuflen += symlen;
         }
         WSS_MUTEX_EXIT( lp->symtab );

         rval = setString( &builder, i, 6, symbuf );
         wss_free( buf );
      }
   }

   if( builder.needed > *bufferSize )
      *bufferSize = builder.needed;

   return rval;
}


/********************************************************************/

static longlong_t
_wrv_get_module_by_id( u_long id, char *buf, ssize_t *bufferSize )
{
   struct bufferBuilder  builder;
   wss_mod_t            *mod;
   int                   i;
   int                   rval = 0;
   int                   row;

#ifdef DEBUG_WRV
   printf( "_getModuleById( %lu, %p, %d\n", id, buf, *bufferSize );
#endif

   if( id )
   {
      WSS_MUTEX_ENTER( wss_module_list );
      for( mod = wss_module_list; mod && mod->id != id; mod = mod->next );
      WSS_MUTEX_EXIT( wss_module_list );

      if( mod == NULL )
         rval = -ENOENT;
      else
         rval = initBuffer( &builder, buf, *bufferSize, 1, 5 );
   }
   else
   {
      WSS_MUTEX_ENTER( wss_module_list );
      for( mod = wss_module_list, i = 0; mod; mod = mod->next, i++ );
      WSS_MUTEX_EXIT( wss_module_list );

      rval = initBuffer( &builder, buf, *bufferSize, i, 5 );
   }

   WSS_MUTEX_ENTER( wss_module_list );
   for( mod = wss_module_list, row = 0; mod; mod = mod->next )
   {
      if( ( id == 0 ) || ( id == mod->id ) )
      {
         rval = setInt( &builder, row, 0, mod->id );
         rval = setString( &builder, row, 1, mod->name );

         if( mod->type == WR_MOD_ACTIVATION )
            rval = setString( &builder, row, 2, "activation" );
         else if( mod->type == WR_MOD_DEACTIVATION )
            rval = setString( &builder, row, 2, "deactivation" );
         else if( mod->type == WR_MOD_WRAPPER )
            rval = setString( &builder, row, 2, "wrapper" );
         else if( mod->type == WR_MOD_LIBRARY )
            rval = setString( &builder, row, 2, "library" );
         else
            rval = setString( &builder, row, 2, "unknown" );

         rval = setPtr( &builder, row, 3,( void * )mod->addr );
         rval = setLong( &builder, row, 4, mod->size );

         row++;

         if( id != 0 )
            break;
      }
   }
   WSS_MUTEX_EXIT( wss_module_list );

   if( builder.needed > *bufferSize )
      *bufferSize = builder.needed;

   return rval;
}


/********************************************************************/
static longlong_t
_wrv_get_module_id_by_name( char *buf, ssize_t *bufferSize )
{
   wss_mod_t *mod;
   int        rval = 0;

   if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 )
   {
      char *buffer = (char *)wss_malloc( *bufferSize );
      if( buffer != NULL )
      {
         copyin( buf, buffer, *bufferSize );

         WSS_MUTEX_ENTER( wss_module_list );
         for( mod = wss_module_list; 
              mod && mod->name && strcmp( mod->name, buffer ); 
              mod = mod->next )
              ;

         if( mod )
            suword( (int *)buf, mod->id );
         else
            suword( (int *)buf, 0 );
         WSS_MUTEX_EXIT( wss_module_list );
      }
      else
      {
         /* not sure what to return here because we can't */
         /* allocate kernel memory but returning ENOMEM will */
         /* make the app think it needs to reallocate the */
         /* buffer bigger */
         rval = -ENOMEM;
      }
   }

   return rval;
}

/********************************************************************/
static longlong_t
_wrv_get_proc_by_pid( u_long id, char *buf, ssize_t *bufferSize )
{
   struct wss_proc      *wrp;
   struct bufferBuilder  builder;
   int                   rval     = 0;
   int                   i;
   int                   row;

   if( id )
   {
      for( wrp = wss_allproc.lh_first;
           wrp && ( wrp->pid != id );
           wrp = wrp->p_list.le_next )
           ;

      if( wrp )
         rval = initBuffer( &builder, buf, *bufferSize, 1, 7 );
      else
         rval = -ENOENT;
   }
   else
   {
      for( wrp = wss_allproc.lh_first, i = 0; wrp; wrp = wrp->p_list.le_next, i++ );

      rval = initBuffer( &builder, buf, *bufferSize, i, 7 );
   }

   for( wrp = wss_allproc.lh_first, row = 0; wrp; wrp = wrp->p_list.le_next )
   {
      if( ( wrp->pid == id ) || ( id == 0 ) )
      {
         wss_wrapper_instance_t *wi;
         int                     size = 0;

         for( wi = wrp->wrappers; wi; wi = wi->next )
            if( wi->wrapper )
               size += strlen( wi->wrapper->name )+ 1;

         rval = setInt( &builder, row, 0, wrp->pid );
         rval = setString( &builder, row, 1, wrp->progname );
         rval = setInt( &builder, row, 2, wrp->p->p_cred->cr_ruid );
         rval = setLong( &builder, row, 3, 
                         wsi_time_get_current()- wrp->p->p_utime );
         rval = setString( &builder, row, 4, wrp->pathbuf );
         rval = setString( &builder, row, 5, wrp->cwd );

         if( size )
         {
            char *buf;
            buf = (char *)wss_malloc( size );
            bzero( buf, size );
            for( wi = wrp->wrappers; wi; wi = wi->next )
            {
               if( wi->wrapper )
               {
                  if( *buf )
                     strcat( buf, "," );
                  strcat( buf, wi->wrapper->name );
               }
            }
            rval = setString( &builder, 0, 6, buf );
            wss_free( buf );
         }
         else
            rval = setString( &builder, 0, 6, "" );

         row++;

         if( id == wrp->pid )
            break;
      }
   }

   if( builder.needed > *bufferSize )
      *bufferSize = builder.needed;

   return rval;
}


/********************************************************************/
static longlong_t
_wrv_get_procs_by_wrapper_id( u_long id, char *buf, ssize_t *bufferSize )
{
   struct wss_proc      *wrp;
   struct bufferBuilder  builder;
   int                   rval     = 0;
   int                   i;
   int                   row;

   i = 0;

   if( id )
   {
      for( wrp = wss_allproc.lh_first; wrp; wrp = wrp->p_list.le_next )
      {
         wss_wrapper_instance_t *wi;
         for( wi = wrp->wrappers; wi; wi = wi->next )
            if( wi->wrapper && ( wi->wrapper->id == id ) )
               i++;
      }

      if( i == 0 )
         return -ENOENT;
   }
   else
   {
      for( wrp = wss_allproc.lh_first; wrp; wrp = wrp->p_list.le_next, i++ );
   }

   rval = initBuffer( &builder, buf, *bufferSize, i, 6 );

   for( wrp = wss_allproc.lh_first, row = 0; wrp; wrp = wrp->p_list.le_next )
   {
      wss_wrapper_instance_t *wi;
      int                     size  = 0;
      int                     found = 0;

      for( wi = wrp->wrappers; wi; wi = wi->next )
      {
         if( wi->wrapper )
         {
            size += strlen( wi->wrapper->name )+ 1;
            if( id && ( wi->wrapper->id == id ) )
               found++;
         }
      }

      if( id && !found )
         continue;

      rval = setInt( &builder, row, 0, wrp->pid );
      rval = setString( &builder, row, 1, wrp->progname );
      rval = setInt( &builder, row, 2, wrp->p->p_cred->cr_ruid );
      rval = setString( &builder, row, 3, wrp->pathbuf );
      rval = setString( &builder, row, 4, wrp->cwd );

      if( size )
      {
         char *buf;
         buf = (char *)wss_malloc( size );
         bzero( buf, size );
         for( wi = wrp->wrappers; wi; wi = wi->next )
         {
            if( wi->wrapper )
            {
               if( *buf )
                  strcat( buf, "," );
               strcat( buf, wi->wrapper->name );
            }
         }
         rval = setString( &builder, row, 5, buf );
         wss_free( buf );
      }
      else
         rval = setString( &builder, row, 5, "" );

      row++;
   }

   if( builder.needed > *bufferSize )
      *bufferSize = builder.needed;

   return rval;
}


/********************************************************************/
/*
 * Return the column names, data type, and max length
 * for the specified table
 */

static longlong_t 
_wrv_get_table_column_def( u_long handle, char *buf, ssize_t *bufferSize )
{
   longlong_t  rval  = 0;
   int         i;
   wqcd_t      col_defs;
   wqtab_t     tab;
   int         record_count = 0;
   int         offset = 0;
   int         string_offset = 0;

   tab = wss_db_find_table_by_handle_r( handle );
   if( tab == NULL )
   {
#ifdef DEBUG_WRV
      printf( "No such table found.\n" );
#endif
      return -ENOENT;
   }

   if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 )
   {
      record_count = tab->col_count;
#ifdef DEBUG_WRV
      printf( "Table has %d columns.\n",tab->col_count );
#endif
    
      offset = sizeof(wqcd) * record_count;
      if( offset > *bufferSize )
         rval = -ENOMEM;

      col_defs = (wqcd_t)( ( tab->col_defs )?rel2abs(tab->col_defs):NULL );

      string_offset = offset + record_count * sizeof(wqcd);
      if( !rval && ( string_offset > *bufferSize ) )
         rval = -ENOMEM;

      for( i = 0; i < record_count ; i++ )
      {
         wqcd_t cd = col_defs + i;
         wqcd_t ncd = (wqcd_t)buf + i;
         char *name = rel2abs( cd->name );
         int str_len;

         if( !rval )
            copyout( cd, ncd, sizeof(wqcd) );

         str_len = strlen( name )+ 1;
         if( !rval && ( ( string_offset + str_len ) > *bufferSize ) )
            rval = -ENOMEM;
         if( !rval )
         {
            copyout( name, buf + string_offset, str_len );
            name = buf + string_offset;
            copyout( &name, &ncd->name, sizeof(name) );
#ifdef DEBUG_WRV
            name = cd->name ? rel2abs( cd->name ): NULL;
            printf( "Name %s.\n",name?name:"( null )" );
#endif
         }
         string_offset += str_len;
      }
   }

   if( rval == -ENOMEM )
   {
      *bufferSize = string_offset;
#ifdef DEBUG_WRV
      printf( "No memory.\n" );
#endif
   }
   else
      *bufferSize = record_count;

  /* find_table_by_handle_r returns a locked table. */
  wss_db_unlock_table( tab );
  return rval;
}


/********************************************************************/

static longlong_t 
_wrv_get_table_handle_by_name( char *buf, ssize_t *bufferSize )
{
   longlong_t rval;
   int        offset = 0;

   /*
    * Return the table id pair for the named table
    */
   wqtab_t tab;

   if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 )
   {
      char *buffer = (char *)wss_malloc( *bufferSize );

      if( buffer != NULL )
      {
         copyin( buf, buffer, *bufferSize );

         tab = find_table( buffer );

         if( tab == NULL )
            rval = -ENOENT;
         else
         {
            wss_wrv_table_id_t *nid = (wss_wrv_table_id_t *)buf;
            char               *name;

            offset = sizeof(wss_wrv_table_id_t);

            if( offset > *bufferSize )
               rval = -ENOMEM;

            if( !rval )
            {
               u_long handle = (u_long)abs2rel( tab );

               copyout( &handle, &nid->handle, sizeof(u_long) );
            }

            name = rel2abs( tab->name );
            if( ( offset + strlen( name )+ 1 )> *bufferSize )
               rval = -ENOMEM;

            if( !rval )
            {
               void *addr = buf + offset;

               copyout( name, buf + offset, strlen( name )+ 1 );
               copyout( &addr, &nid->name, sizeof(void *) );
            }
         }
      }
   }

   if( rval == -ENOMEM )
   {
      *bufferSize = offset;
#ifdef DEBUG_WRV
      printf( "No memory.\n" );
#endif
   }

   return rval;
}


/********************************************************************/

/*
 * Return the table ids for requested tables
 */

static longlong_t 
_wrv_get_table_handles( char *buf, ssize_t *bufferSize )
{
   longlong_t          rval         = 0;
   u_long             *handles;
   unsigned int        size;
   wss_wrv_table_id_t *prev         = NULL;
   wss_wrv_table_id_t *nid;
   u_long              rel          = 0;
   int                 hindex       = 0;
   int                 i;
   int                 record_count = 0;
   int                 offset       = 0;
   wqtab_t             tab;

#ifdef DEBUG_WRV
   printf( "In _getTableHandles.\n" );
#endif

   if( ( rval = checkBuffer( buf, *bufferSize ) ) == 0 )
   {
      char *buffer = (char *)wss_malloc( *bufferSize );

      if( buffer != NULL )
      {
         int err;

         err = copyin( buf, buffer, *bufferSize );
#ifdef DEBUG_WRV
         printf( "Copyin returned %d: bufferSize = %d\n",err, *bufferSize );
#endif

         /*
          * Determine number of table handles passed in
          */
         for( handles = (u_long *)buffer, size = 0; *handles;
              handles++, size++ )
              ;

#ifdef DEBUG_WRV
         printf( "size = %d\n",size );
#endif

         /*
          * Allocate space for copying handles and copy them
          */
         if( size > 0 )
         {
            handles = (u_long *)wss_malloc( sizeof(u_long) * size );
            bzero( handles, sizeof(u_long) * size );
         }

         for( tab= (wqtab_t)( ( *wss_vm_root_table )?rel2abs( *wss_vm_root_table ):NULL );
              tab; tab = (wqtab_t)( ( tab->next )?rel2abs( tab->next ):NULL ) )
         {
            if( size > 0 )
            {
               /*
                * We only want the specified tables
                */
               rel = (u_long)abs2rel( tab );
               for( i = 0;
                    ( i < size ) && ( *( (u_long *)buffer + i )!= rel );
                    i++ )
                    ;

               if( i >= size )
                  continue;
               handles[i] = *( (u_long *)buffer + i );
            }
            record_count++;
         }
#ifdef DEBUG_WRV
         printf( "record_count = %d\n",record_count );
#endif

         offset = sizeof(wss_wrv_table_id_t) * record_count;
         if( offset > *bufferSize )
            rval = -ENOMEM;

         if( size > 0 )
         {
            if( record_count > 0 )
            {
               /*
                * Find first valid handle
                */
               for( hindex = 0; handles[hindex] == 0; hindex++ );

               tab = (wqtab_t)rel2abs( handles[hindex] );
               rel = handles[hindex];
            }
            else
               tab = NULL;
         }
         else
         {
            tab= (wqtab_t)( ( *wss_vm_root_table )?rel2abs( *wss_vm_root_table ):NULL );
#ifdef DEBUG_WRV
            if( *wss_vm_root_table == NULL )
               printf( "No tables yet.\n" );
#endif
            rel = (u_long)( *wss_vm_root_table );
         }

         for( i = 0; tab && ( i < record_count ); i++ )
         {
            char *name;

            nid = (wss_wrv_table_id_t *)buf + i;
            if( !rval && prev )
            {
               void *ptr = nid;

               copyout( &ptr, &prev->next, sizeof(ptr) );
            }

            prev = nid;

            if( !rval )
            {
               /*
                * We'll use the relative address as the handle
                */
               u_long handle = (u_long)abs2rel( tab );
               err = copyout( &handle, &nid->handle, sizeof(handle) );
#ifdef DEBUG_WRV
               printf( "copyout 1 returned %d\n",err );
#endif
            }
   
            /*
             * Copy the table name into the id pair
             */
            name = rel2abs( tab->name );

            if( ( offset + strlen( name )+ 1 )> *bufferSize )
               rval = -ENOMEM;

            if( !rval )
            {
               void *addr;
               err = copyout( name, buf + offset, strlen( name )+ 1 );
#ifdef DEBUG_WRV
               printf( "copyout 2 returned %d\n",err );
#endif
               addr = buf + offset;
               err = copyout( &addr, &nid->name, sizeof(void *) );
#ifdef DEBUG_WRV
               printf( "copyout 4 returned %d\n",err );
#endif
            }
            offset += strlen( name )+ 1;

            /*
             * Get next table to construct id
             */
            if( size > 0 )
            {
               if( ( i + 1 )< record_count )
               {
                  /*
                   * Find next valid handle
                   */
                  for( ; handles[hindex] == 0; hindex++ );

                  tab = ( wqtab_t )rel2abs( handles[hindex] );
                  rel = handles[hindex];
               }
               else
                  tab = NULL;
            }
            else
            {
               rel = (u_long)tab->next;
               tab = (wqtab_t)( ( tab->next )?rel2abs( tab->next ):NULL );
            }

            if( !rval )
            {
               void *next = NULL;
               if( tab )
                  next = nid + 1;
               err = copyout( &next, &nid->next, sizeof(void *) );
#ifdef DEBUG_WRV
               printf( "copyout 3 returned %d\n",err );
#endif
            }
         }

         /*
          * Free memory used by handles array
          */
         if( size > 0 )
            wss_free( (void *)handles );
      }
   }

   if( rval == -ENOMEM )
   {
      *bufferSize = offset;
#ifdef DEBUG_WRV
      printf( "No memory.\n" );
#endif
   }
   else
      *bufferSize = record_count;

   return rval;
}

/********************************************************************/

/*
 * Return the requested rows for the specified table
 */

static longlong_t 
_wrv_get_table_rows( u_long handle, char *ubuf, ssize_t *bufferSize )
{
   int            rval = 0;
   int            record_count = 0;
   int            offset=0, string_offset = 0;
   int            i;
   u_long *handles;
   unsigned int   size;
   wqrow_t        row;
   int            hindex = 0;
   u_long  rel = 0;
   wqtab_t        tab;

   if( ( tab = wss_db_find_table_by_handle_r( handle ) ) == NULL )
      return -ENOENT;

   /*
    * Determine number of table handles passed in
    */
#ifdef DEBUG_WRV
   printf( "_wrv_get_table_rows\n" );
#endif /* DEBUG_WRV */
   if( ( rval = checkBuffer( ubuf, *bufferSize ) ) == 0 )
   {
      char *kbuf = (char *)wss_malloc( *bufferSize );

      if( kbuf )
      {
         copyin( ubuf, kbuf, *bufferSize );

         for( handles = (u_long *)kbuf + 1, size = 0; 
              *handles;
              handles++, size++ );

         /*
          * Allocate space for copying handles and copy them
          */
         if( size > 0 )
         {
            handles = (u_long *)wss_malloc( sizeof(u_long) * size );
            bzero( handles, sizeof(u_long) * size );
         }

         for( row = (wqrow_t)( ( tab->rows )?rel2abs( tab->rows ):NULL ); 
              row;
              row = (wqrow_t)( ( row->next )?rel2abs( row->next ):NULL ) )
         {
            u_long rel = (u_long)abs2rel( row );

            if( size )
            {
               /*
                * We only want the specified rows
                */
               int i;
               for( i = 0;
                   ( i < size ) && ( *( (u_long *)kbuf + i ) != rel );
                    i++ );
               if( i >= size )
                  continue;

               handles[i] = *( (u_long *)kbuf + i );
            }

            record_count++;
        }

        /* offset starts at the beginning of the table data. */
        offset = sizeof(wss_wrv_db_row_t) * record_count;

        /* string_offset starts at the beginning of the string table */
        string_offset = offset + sizeof(wqval) * tab->col_count * record_count;

        if( string_offset > *bufferSize )
           rval = -ENOMEM;
         
        if( size > 0 )
        {
           if( record_count > 0 )
           {
              /*
               * Find first valid handle
               */
              for( hindex = 0; handles[hindex] == 0; hindex++ );

              row = (wqrow_t)rel2abs( handles[hindex] );
              rel = handles[hindex];
           }
           else
              row = NULL;
        }
        else
        {
           rel = (u_long)tab->rows;
           row = (wqrow_t)( ( tab->rows )?rel2abs( tab->rows ):NULL );
        }

        for( i = 0; row && ( i < record_count ); i++ )
        {
           wss_wrv_db_row_t *nrow = (wss_wrv_db_row_t *)ubuf + i;
           wqval_t           nvalues;
           wqcd_t            cd;
           wqval_t           val, val_list;
           int               k;
           void             *next;

           if( !rval )
              suword( (int *)&nrow->handle, rel );

           /*
            * Determine number of values in row
            */
           if( !rval )
              suword( &nrow->numvals, tab->col_count );

           nvalues = (wqval_t)( ubuf + offset );
           offset += sizeof(wqval) * tab->col_count;

           /*
            * Copy row values
            */
           cd = (wqcd_t)( ( tab->col_defs )?rel2abs( tab->col_defs ):NULL );
           val_list = (wqval_t)( ( row->vals )?rel2abs( row->vals ):NULL );

           for( k = 0; k < tab->col_count; k++ )
           {
              wqval_t nval = nvalues + k;
              val = val_list + k;

              if( !rval )
                 copyout( val, nval, sizeof(wqval) );

              if( cd[k].dtype == pd_char_t_t )
              {
                 /*
                  * Check avaialable space for string value
                  */
                  char *value = rel2abs( val->v.cval );

                  if( !rval &&
                      ( ( string_offset + strlen( value )+ 1 ) > *bufferSize ) )
                     rval = -ENOMEM;

                  if( !rval )
                  {
                     char *cval = ubuf + string_offset;

                     copyout( value, ubuf + string_offset, strlen( value )+ 1 );
                     copyout( &cval, &nval->v.cval, sizeof(cval) );
                  }
                  string_offset += strlen( value )+ 1;
               }
            }

            if( !rval )
               copyout( &nvalues, &( ROW( nrow )->vals ), sizeof(nvalues) );

            /*
             * Get next row to be copied
             */
            if( size > 0 )
            {
               if( ( i + 1 )< record_count )
               {
                  /*
                   * Find next valid handle
                   */
                  for( ; handles[hindex] == 0; hindex++ );

                  row = (wqrow_t)rel2abs( handles[hindex] );
                  rel = handles[hindex];
               }
               else
                  row = NULL;
            }
            else
            {
               rel = (u_long)row->next;
               row = (wqrow_t)( ( row->next )?rel2abs( row->next ):NULL );
            }

            next = NULL;
            if( row )
               next = nrow + 1;

            if( !rval )
               copyout( &next, &ROW( nrow )->next, sizeof(next) );
         }

         /*
          * Free memory used by handles array
          */
         if( size > 0 )
            wss_free( ( void * )handles );
      }
   }
#if 0
   else
      rval == -EFAULT;
#endif

   if( rval == -ENOMEM )
   {
      *bufferSize = string_offset;
#ifdef DEBUG_WRV
       printf( "No memory.\n" );
#endif
   }
   else
      *bufferSize = record_count;

   /* find_table_by_handle_r returns a locked table. */
   wss_db_unlock_table( tab );

   return rval;
}


/********************************************************************/

static longlong_t 
_wrv_get_version( char *buf, ssize_t *bufferSize )
{
   longlong_t rval;

   if( *bufferSize < 2 * sizeof(int) )
   {
      *bufferSize = 2 * sizeof(int);
      set_errno( EINVAL );
      rval = -EINVAL;
   }
   else
   {
      copyout( _versionBuffer, buf, 2 * sizeof(int) );
      rval = 0;
   }

   return rval;
}


/********************************************************************/
static longlong_t
_wrv_get_wrapper_by_id( u_long id, char *buf, ssize_t *bufferSize )
{
   wss_wrapper_t        *wp;
   struct bufferBuilder  builder;
   int                   i;
   int                   rval = 0;
   int                   row;

#ifdef DEBUG_WRV
   printf( "_getWrapperyId( %lu, %p, %d\n", id, buf, *bufferSize );
#endif

   if( id )
   {
      WSS_MUTEX_ENTER( wrapper_list );
      for( wp = wrapper_list; wp && wp->id != id; wp = wp->next );
      WSS_MUTEX_EXIT( wrapper_list );

      if( wp == NULL )
         rval = -ENOENT;
      else
         rval = initBuffer( &builder, buf, *bufferSize, 1, 6 );

   }
   else
   {
      WSS_MUTEX_ENTER( wrapper_list );
      for( wp = wrapper_list, i = 0; wp; wp = wp->next, i++ );
      WSS_MUTEX_EXIT( wrapper_list );
      rval = initBuffer( &builder, buf, *bufferSize, i, 6 );
   }

   WSS_MUTEX_ENTER( wrapper_list );
   for( wp = wrapper_list, row = 0; wp; wp = wp->next, row++ )
   {
      if( ( id == 0 ) || ( id == wp->id ) )
      {
         setInt( &builder, row, 0, wp->id );
         setString( &builder, row, 1, wp->name );
         setInt( &builder, row, 2, wp->refcount );
         setInt( &builder, row, 3, wp->module->uid );
         setLong( &builder, row, 4, wp->module->loadtime );
         setLong( &builder, row, 5, wp->module->size );
      }
   }
   WSS_MUTEX_EXIT( wrapper_list );
    
   if( builder.needed > *bufferSize )
      *bufferSize = builder.needed;

   return rval;
}

/********************************************************************/
static longlong_t
_wrv_set_table_row_field( u_long htable, u_long hrow, char *field,
                          char *value )
{
   longlong_t rval = -ENOENT;
   wqrow_t    row;
   wqtab_t    tab;
   wqcd_t     cd;
   int        max;

   /*
    * Validate the table handle
    */
   if( !( tab = wss_db_find_table_by_handle_r( htable ) ) )
      return rval;

   max = tab->col_count;

   /*
    * Make sure we have some column definitions
    */
   if( ( cd = (wqcd_t)( tab->col_defs?rel2abs( tab->col_defs ):NULL ) ) )
   {
      /*
       * Validate row handle
       */
      for( row = (wqrow_t)( tab->rows?rel2abs( tab->rows ):NULL );
           row && ( (u_long)abs2rel( row )!= hrow );
           row = (wqrow_t)( row->next?rel2abs( row->next ):NULL ) );

      if( row )
      {
         int i;

         /*
          * Find field
          */
         for( i = 0;
             ( i < max ) && strcmp( (char *)rel2abs( cd[i].name ), field );
              i++ );

         if( i < max )
         {
            /*
             * Tell WSS to set the value based on the type
             */
            wss_db_set_row_value_from_string( cd, row, i, value );
            rval = 0;
         }
      }
   }

   /* find_table_by_handle_r returns a locked table. */
   wss_db_unlock_table( tab );

   return rval;
}


/****************************************************************************\
 * Public functions                                                         *
\****************************************************************************/

/*
 * wrv syscall
 */
longlong_t
wrv( int command, void *args, ssize_t *bufsize )
{
   int              rval       = 0;
   ssize_t          bufferSize = 0;
   rval_t           stupidunion;
   struct wss_proc *wrp;

#ifdef SUSER_ONLY
   if( !suser( ) )
   {
      set_errno( EPERM );

      stupidunion.r_val1 = -EPERM;
      stupidunion.r_val2 = 0;
      return stupidunion.r_vals;
   }
#endif

   if( ( wrp = wss_pfind( curproc->p_pid ) ) )
   {
      boolean_t badsys = B_FALSE;
      if( wrp->wrappers )
         badsys = B_TRUE;

      WSS_MUTEX_EXIT( wrp->wss_proc );

      if( badsys )
         return( (*nosys_ent.sy_callc)() );
   }

   if( bufsize &&
       ( rval = copyin( bufsize, &bufferSize, sizeof(bufferSize) ) ) )
   {
      stupidunion.r_val1 = rval;
      stupidunion.r_val2 = 0;
      return stupidunion.r_vals;
   }

   switch( command )
   {
      case WRV_GET_ACT_CRIT_BY_ID:
      case WRV_GET_DEACT_CRIT_BY_ID:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_criteria_by_id( command, data.handle,(char *)args,
                                            &bufferSize );
         else
            rval = -rval;

         break;
      }

      case WRV_GET_ACTIVE_WRAPPERS:
         rval = _wrv_get_active_wrappers( (char *)args, &bufferSize );
         break;

      case WRV_GET_LIBRARY_BY_ID:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_library_by_id( data.handle, (char *)args,
                                           &bufferSize );
         else
            rval = -rval;

         break;
      }

      case WRV_GET_MODULE_BY_ID:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_module_by_id( data.handle,(char *)args,
                                          &bufferSize );
         else
            rval = -rval;

         break;
      }

      case WRV_GET_MODULE_ID_BY_NAME:
         rval = _wrv_get_module_id_by_name( (char *)args, &bufferSize );
         break;

      case WRV_GET_PROC_BY_PID:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_proc_by_pid( data.handle,(char *)args,
                                         &bufferSize );
         else
            rval = -rval;

         break;
      }

      case WRV_GET_PROCS_BY_WRAPPER_ID:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_procs_by_wrapper_id( data.handle,(char *)args,
                                                 &bufferSize );
         else
            rval = -rval;

         break;
      }

      case WRV_GET_TABLE_COLUMN_DEF:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_table_column_def( data.handle,(char *)args,
                                              &bufferSize );
         else
            rval = -rval;

         break;
      }

      case WRV_GET_TABLE_HANDLE_BY_NAME:
         rval = _wrv_get_table_handle_by_name( (char *)args, &bufferSize );
         break;

      case WRV_GET_TABLE_HANDLES:
         rval = _wrv_get_table_handles( (char *)args, &bufferSize );
         break;

      case WRV_GET_TABLE_ROWS:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_table_rows( data.handle,(char *)args, &bufferSize );
         else
            rval = -rval;

         break;
      }

      case WRV_SET_TABLE_ROW_FIELD:
      {
         wrv_trnd_t id;
         char *data;
         int dsize = bufferSize - sizeof(id);

         id.name[MAXNAMELEN-1] = '\0';
         if( dsize > 0 )
            data = wss_malloc( dsize );
         else
         {
            data = NULL;
            rval = EINVAL;
         }

         if( !( rval = copyin( args, &id, sizeof(id) ) ) &&
             data && !( rval = copyin( args + sizeof(id), data, dsize ) ) )
            rval = _wrv_set_table_row_field( id.thandle, id.rhandle,
                                             id.name, data );
         else
            rval = -rval;

         if( data )
            wss_free( data );

         break;
      }

      case WRV_GET_VERSION:
         rval = _wrv_get_version( (char *)args, &bufferSize );
         break;

      case WRV_GET_WRAPPER_BY_ID:
      {
         wrv_h_t data;

         if( !( rval = copyin( args, &data, sizeof(data) ) ) )
            rval = _wrv_get_wrapper_by_id( data.handle,(char *)args,
                                           &bufferSize );
         else
            rval = -rval;

         break;
      }

      default:
         set_errno( ENOTSUP );
         rval = -ENOTSUP;
   }

   if( rval < 0 )
      set_errno( -rval );

   copyout( &bufferSize, bufsize, sizeof(bufferSize) );

   stupidunion.r_val1 = rval;
   stupidunion.r_val2 = 0;
#ifdef DEBUG_WRV
   printf( "wrv retval: %d\n",rval );
#endif
   return stupidunion.r_vals;
}

