/*! \file loader.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: loader.c,v 1.47 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Load/unload the WSS from the kernel.
 * \author  Steve Kiernan
 * \version $Revision: 1.47 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/param.h>
#include <sys/syscall.h>
#include <sys/systm.h>
#include <sys/modctl.h>
#include <sys/systeminfo.h>
#include <sys/processor.h>

#include <wss_init.h>
#include <wss_module.h>
#include <wss_process.h>
#include <wss_vm.h>
#include <wss_log.h>
#include <wsi_syscalls.h>
#include <wsi_hook.h>
#include <wsi_execve.h>
#include <wsi_dir.h>


/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/

extern int wsi_custom_hook_fork_and_exec();
extern int wsi_custom_hook_exit();
extern int wsi_custom_unhook_fork_and_exec();
extern int wsi_export_init();


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern struct syscallvec new_sysent[];
extern struct mod_ops    mod_miscops;
extern int               wsi_shutdown;
extern struct modlinkage modlinkage;


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

/*
 * The next set of defines assumes a standard naming scheme for the
 * WSS syscall replacement functions.
 */

/*
 * Prototype the replacement (or new) syscall with name "name."
 */
#define WR_DEFINE_REPLACEMENT(name, numargs)             \
   struct sysent orig_ ## name ##_sysent;                \
   static struct sysent new_ ## name ## _sysent =        \
      { numargs, 0, NULL, NULL, (longlong_t(*)())wsi_ ## name }

#define WR_DEFINE_REPLACEMENT_OLD(name, numargs)         \
   struct sysent orig_ ## name ##_sysent;                \
   static struct sysent new_ ## name ## _sysent =        \
      { numargs, 0, (int(*)())wsi_ ## name, NULL, NULL }

/*
 * Save the original syscall and copy in the entry for the replacement.
 */
#define WR_SYS_REPLACE(name)                                 \
   bcopy(&sysent[SYS_ ##name ], &orig_ ## name ## _sysent,   \
      sizeof(struct sysent));                                \
   bcopy(&new_ ## name ## _sysent, &sysent[SYS_ ## name],    \
      sizeof(struct sysent))

/*
 * Save the original syscall and copy in the entry for the replacement.
 * Special added work for old-style syscalls
 */
#define WR_SYS_REPLACE_OLD(name)                             \
   bcopy(&sysent[SYS_ ##name ], &orig_ ## name ## _sysent,   \
      sizeof(struct sysent));                                \
   bcopy(&sysent[SYS_ ##name ], &new_ ## name ## _sysent,    \
      sizeof(struct sysent));                                \
   new_ ## name ## _sysent.sy_call = (int(*)())wsi_ ## name; \
   bcopy(&new_ ## name ## _sysent, &sysent[SYS_ ## name],    \
      sizeof(struct sysent))

/*     
 * Save the original syscall and copy in the new syscall.
 */
#define WR_SYS_ADD(name)                                   \
   bcopy(&sysent[SYS_ ##name ], &orig_ ## name ## _sysent, \
      sizeof(struct sysent));                              \
   bcopy(&new_ ## name ## _sysent, &sysent[SYS_ ## name],  \
      sizeof(struct sysent))
       
/*
 * Put the original syscall entry back in the table.
 */
#define WR_SYS_FIX(name)                                   \
   bcopy(&orig_ ## name ## _sysent, &sysent[SYS_ ## name], \
      sizeof(struct sysent))

/*     
 * Put the original entry back.  It _should_ be lkmnosys.
 */
#define WR_SYS_REMOVE(name)                                \
   bcopy(&orig_ ## name ## _sysent, &sysent[SYS_ ## name], \
      sizeof(struct sysent))


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

struct modlmisc modlmisc =
{
   &mod_miscops,
   PD_MAKESTR("Wrapper Support Subsystem v" WR_VERSION_STRING)
};


/*
 * Syscalls augmented by the WSI
 */
WR_DEFINE_REPLACEMENT_OLD( execve, 3 );
WR_DEFINE_REPLACEMENT( chdir, 1 );
WR_DEFINE_REPLACEMENT( fchdir, 1 );
WR_DEFINE_REPLACEMENT( rename, 2 );

/*
 * Syscalls added by the WSI
 */
WR_DEFINE_REPLACEMENT( wrctl, 3 );
WR_DEFINE_REPLACEMENT( wrv, 4 );


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

struct sysent  saved_sysent[NSYSCALL];
id_t           unloadpid;
kthread_t     *unloadt;
int            wsi_unloading = 0;


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

#ifdef FIXED_UNLOAD
/*
 * Thread for unloading the WSS (does not completely work, yet)
 */
static void
wsi_fini_thread( void *arg )
{
   ksema_t *sem = (ksema_t *)arg;
   int nsyscall;

   WSS_MUTEX_ENTER_DEBUG( wss_proc_num_in_syscall );
   nsyscall = wss_proc_num_in_syscall;
   while( nsyscall > 0 )
   {
      WSS_MUTEX_ENTER_DEBUG( wss_allproc );

      if( wss_allproc.lh_first != NULL )
      {
         proc_t    *p;

         p = wss_allproc.lh_first->p;
         if( ( p->p_pid == unloadpid ) &&
             ( wss_allproc.lh_first->p_list.le_next != NULL ) )
         {
            wsi_printf( PD_MAKESTR("Choosing next proc\n") );
            p = wss_allproc.lh_first->p_list.le_next->p;
         }

         wsi_printf( PD_MAKESTR("Suspending threads in proc #%d (%d,%d)\n"), (int)p->p_pid,
                 ( p->p_pid == unloadpid )?1:0, 
                 ( wss_allproc.lh_first->p_list.le_next != NULL )?1:0 );
         if( wsi_proc_suspend( p ) == 0 )
         {
            wsi_printf( PD_MAKESTR("Resuming threads in proc #%d\n"), (int)p->p_pid );
            wsi_proc_continue( p );
         }
         else
            cmn_err( CE_WARN, PD_MAKESTR("Error during suspend of threads\n") );
      }
      WSS_MUTEX_EXIT_DEBUG( wss_allproc );

      WSS_CV_WAIT_DEBUG( wss_proc_num_in_syscall );
      nsyscall = wss_proc_num_in_syscall;
      wsi_printf( PD_MAKESTR("wsi_fini_thread: %d\n"), nsyscall );
   }
   WSS_MUTEX_EXIT_DEBUG( wss_proc_num_in_syscall );

   wss_crit_shutdown();
   wss_mod_unload_all();
   wss_wrapper_shutdown();
   wsi_hook_shutdown();
   wss_vm_shutdown();
   wss_log_shutdown();

   sema_v( sem );
   thread_exit();
}
#endif /* FIXED_UNLOAD */


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  _shutdown */
/*! \brief This is the WSI shutdown function
 *
 *  It is called when the init process (PID 0) exits because a system
 *  shutdown is initiated.  This function will clean up after the WSS
 *  so that the database file will not be corrupted by the shutdown.
 */
int
_shutdown()
{
   int ent;

   if( !wsi_shutdown )
   {
      /* This should _not_ happen */
      wsi_printf( PD_MAKESTR("_shutdown: shutdown flag not set, why?\n") );
      return EBUSY;
   }

   /*
    * Okay, let's do it... shutdown time!
    */
   wsi_printf( PD_MAKESTR("_shutdown: WSS shutting down.\n") );
   wsi_printf( PD_MAKESTR("[Note: It is normal to see a single process which has orphaned syscalls in the\n") );
   wsi_printf( PD_MAKESTR("       following messages, it is the process which performed the shutdown\n") );
   wsi_printf( PD_MAKESTR("       command.]\n") );

   /*
    * Replace our syscall with the original one
    */
   WR_SYS_FIX( execve );
   WR_SYS_FIX( chdir );
   WR_SYS_FIX( fchdir );
   WR_SYS_FIX( rename );

   /*
    * Unhook the syscalls that were added
    */
   WR_SYS_REMOVE( wrctl );
   WR_SYS_REMOVE( wrv );

#ifndef WR_DONT_SAVE_SYSENT
   /*
    * Restore original sysent
    */
   wsi_printf(PD_MAKESTR("Restoring sysent\n"));
   for( ent = 0; ent < NSYSCALL; ent++ )
   { 
      if( (void *)sysent[ent].sy_call != (void *)nosys_ent.sy_callc )
      {
         bcopy( (void *)&saved_sysent[ent], (void *)&sysent[ent],
                sizeof(struct sysent) );
      }
   } 
#endif /* WR_DONT_SAVE_SYSENT */

   wss_proc_shutdown();

   /*
    * Unload all modules
    */
   wss_crit_shutdown();
   wss_mod_unload_all();
   wss_wrapper_shutdown();

   /*
    * Remove OS hooks
    */
   wsi_hook_shutdown();

   /*
    * Shutdown mapped vm space
    */
   wss_vm_shutdown();

   /*
    * Close the log file
    */
   wss_log_shutdown();

   WSS_MUTEX_DESTROY( wss_proc_num_in_syscall );
   WSS_CV_DESTROY( wss_proc_num_in_syscall );

   return 0;
}


/*  wsi_load */
/*! \brief This is the lkm load function
 *
 *  It initializes the WSS, replaces some system calls, adds an exit function,
 *  adds two new system calls, and prepares for the loading of wrappers.
 */
int
wsi_load()
{
   int error;

   /*
    * Initialize the Wrapper SubSystem
    */
   if( ( error = wss_init( WSS_INIT_DB | WSS_INIT_CONF ) ) )
      return error;

   /*
    * Initialize the kernel hot-patch system
    */
   if( ( error = wsi_hook_init() ) )
      return error;

   /*
    * Perform intrusive hook for fork, fork1, and vfork, exec,
    * and exit syscalls
    */
   if( ( error = wsi_custom_hook_fork_and_exec() ) )
      return error;

   if( ( error = wsi_custom_hook_exit() ) )
      return error;

   if( ( error = wsi_export_init() ) )
      return error;

   if( ( error = mod_install( &modlinkage ) ) )
   {
      /*
       * Operating system module load failed, undo all the
       * work we performed above
       */
      wss_proc_shutdown();
      wss_crit_shutdown();
      wss_mod_unload_all();
      wss_wrapper_shutdown();
      wsi_hook_shutdown();
      wsi_custom_unhook_fork_and_exec();
      wss_vm_shutdown();
      wss_log_shutdown();
   }
   else
   {
#ifndef WR_DONT_SAVE_SYSENT
      int ent;
#endif /* WR_DONT_SAVE_SYSENT */

      /*
       * Replace old syscall entries with our new ones
       */
      WR_SYS_REPLACE_OLD( execve );
      WR_SYS_REPLACE( chdir );
      WR_SYS_REPLACE( fchdir );
      WR_SYS_REPLACE( rename );

#ifndef WR_DONT_SAVE_SYSENT
      wsi_printf(PD_MAKESTR("Saving sysent\n"));

      /*
       * Save sysent for later restore on unload
       *
       * Note: Unloading the WSS on Solaris does not yet work.
       *       We still need to find out how to force LWPs to restart
       *       their system calls
       */
      for( ent = 0; ent < NSYSCALL; ent++ )
      {
         if( (void *)sysent[ent].sy_call != (void *)nosys_ent.sy_callc )
         {
            bcopy( (void *)&sysent[ent], (void *)&saved_sysent[ent],
                   sizeof(struct sysent) );
            if( sysent[ent].sy_call != NULL )
               sysent[ent].sy_call = new_sysent[ent].sy_call;
            else
               sysent[ent].sy_callc = new_sysent[ent].sy_callc;
#ifdef DEBUG_PRINT_SYSCALLS
            wsi_printf(PD_MAKESTR("%s: %d\t"), solaris_syscallnames[ent], sysent[ent].sy_narg);
            if( sysent[ent].sy_call != NULL )
               wsi_printf(PD_MAKESTR("old style\n"));
            else if( sysent[ent].sy_callc != NULL )
               wsi_printf(PD_MAKESTR("argc style\n"));
#endif /* DEBUG_PRINT_SYSCALLS */
         }
      }
#endif /* WR_DONT_SAVE_SYSENT */

      /*
       * Hook into wrctl and wrv syscalls
       */
      WR_SYS_ADD( wrctl );
      WR_SYS_ADD( wrv );

      /*
       * If we make it to here, print message on console.
       */
      wsi_printf(PD_MAKESTR("\nSuccessfully loaded WSS module!\n"));
   }

   return error;
}

/*  wsi_unload */
/*! \brief This is the lkm unload function
 *
 *  It shuts down the WSS, restores the replaced system calls, removes the exit
 *  function, removes the added system calls, unloads wrapper modules, and
 *  forces LWPs to reissue their system calls.
 *
 *  \warning Does not work yet.  We still need to figure out how to get
 *           LWPs to reissue their system calls.
 */
int
wsi_unload()
{
#ifdef FIXED_UNLOAD
   int i;

   if( ( i = mod_remove( &modlinkage ) ) == 0 )
   {
#ifndef WR_DONT_SAVE_SYSENT
      int ent;
#endif /* WR_DONT_SAVE_SYSENT */

      /* Flag that we're unloading the module */
      wsi_unloading = 1;

      /*
       * Replace our syscall with the original one
       */
      WR_SYS_FIX( execve );
      WR_SYS_FIX( chdir );
      WR_SYS_FIX( fchdir );
      WR_SYS_FIX( rename );

      /*
       * Unhook the syscalls that were added
       */
      WR_SYS_REMOVE( wrctl );
      WR_SYS_REMOVE( wrv );

#ifndef WR_DONT_SAVE_SYSENT
      /*
       * Restore original sysent
       */
      wsi_printf(PD_MAKESTR("Restoring sysent\n"));
      for( ent = 0; ent < NSYSCALL; ent++ )
      {
         if( (void *)sysent[ent].sy_call != (void *)nosys_ent.sy_callc )
         {
            bcopy( (void *)&saved_sysent[ent], (void *)&sysent[ent],
                   sizeof(struct sysent) );
         }
      }
#endif /* WR_DONT_SAVE_SYSENT */
                
      wss_proc_shutdown();

      if( wss_proc_num_in_syscall == 0 )
      {
         /*
          * Unload all modules
          */
         wss_crit_shutdown();
         wss_mod_unload_all();
         wss_wrapper_shutdown();

         wsi_hook_shutdown();
      }
      else
         wsi_custom_unhook_fork_and_exec();

      if( wss_proc_num_in_syscall == 0 )
      {
         wss_vm_shutdown();
         wss_log_shutdown();
      }

      if( wss_proc_num_in_syscall != 0 )
      {
         ksema_t sem;
         sema_init( &sem, 0, NULL, SEMA_DEFAULT, NULL );

         unloadpid = curproc->p_pid;
         unloadt = curthread;
         wsi_printf( PD_MAKESTR("Creating unload thread for process #%d\n"), (int)unloadpid );
         if( thread_create( NULL, _defaultstksz, wsi_fini_thread,
                            (caddr_t)&sem, 0, &p0, TS_RUN, 99 ) == NULL )
            cmn_err( CE_PANIC, PD_MAKESTR("Unable to create thread for unload\n") );

         sema_p( &sem );
      }

      WSS_MUTEX_DESTROY( wss_proc_num_in_syscall );
      WSS_CV_DESTROY( wss_proc_num_in_syscall );

      /* if we make it to here, print message on console*/
      wsi_printf( PD_MAKESTR("\nSuccessfully unloaded WSS module!\n") );
   }

   return( i );
#else /* FIXED_UNLOAD */
   return EBUSY;
#endif /* FIXED_UNLOAD */
}

