/*! \file wsi_custom.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_custom.c,v 1.27 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Custom implementation of specific system calls
 * \author  Steve Kiernan
 * \version $Revision: 1.27 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/types.h>
#include <sys/bootconf.h>
#include <sys/kobj.h>
#include <sys/kobj_impl.h>
#include <sys/modctl.h>
#include <sys/systm.h>
#include <sys/mman.h>
#include <sys/uadmin.h>
#include <sys/poll.h>

#include <wss_process.h>
#include <wss_malloc.h>
#include <wsi_hook.h>
#include <wsi_execve.h>


/***************************************************************************\
 * External variables                                                      *
\***************************************************************************/

extern struct sysent saved_sysent[NSYSCALL];
extern int           wsi_unloading;
extern int           wsi_shutdown;


/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/


extern int        wsi_dispatch   __P(( int space, int call, void *uap,
                                       rval_t *retval ));
extern longlong_t wsi_dispatch_c __P(( int space, int call, int i0, int i1,
                                       int i2, int i3, int i4, int i5 ));
extern int        _shutdown      __P(( void ));


/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

/*! \brief thread-to-process mapping structure */
struct wsi_t2p
{
   kthread_t *t;         /*!< Pointer to kernel thread calling exit */
   pid_t      pid;       /*!< PID of exiting process                */
   struct wsi_t2p *next; /*!< Pointer to next mapping structure     */
};

/* uadmin syscall args */
struct uadmin_uap
{
   int cmd;
   int fcn;
   int mdep;
};


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static struct wsi_t2p    *wsi_t2p_list = NULL;
static WSS_MUTEX_PROTECT( wsi_t2p_list );


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/*! \brief Flag indicating that the system is shutting down */
int wsi_shutdown = 0;


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*  wsi_custom_pgjoin */
/*! \brief Custom replacement for pgjoin call in fork
 *
 *  This custom shim function provides a hook in the cfork call within
 *  Solaris so we can be notified when fork, fork1, and vfork calls occur.
 */
static void
wsi_custom_pgjoin( struct proc *child, struct pid *pid )
{
   /* Call pgjoin */
   pgjoin( child, pid );

   /* Call our fork hook */
   wss_proc_fork( curproc, child, 0 );
}

/*  wsi_custom_exitlwps */
/*! \brief Custom replacement for exitlwps call in exit
 *
 *  This custom shim function provides a hook early in the exit call within
 *  Solaris so we can start cleaning up wrappers-specific kernel data for the
 *  process.
 *
 *  We create a map of the thread calling the exit syscall to the exiting
 *  process' pid so we can clean up the process structure after it is no
 *  longer needed.  We need to keep a mapping of the exit-calling thread to
 *  the process pid, because the pid contained in process structure is no
 *  longer necessarily valid when the hook for the stage 2 exit routine is
 *  called.
 *
 *  After the mapping has been added to the map list, we call the
 *  wss_proc_exit_stage1 function to deactivate all the wrappers currently
 *  wrapping the process.
 */
static void
wsi_custom_exitlwps( int rcode )
{
   /* Construct a thread-to-process mapping for use by the stage 2
      exit code */
   struct wsi_t2p *t2p = wss_malloc( sizeof( *t2p ) );
   t2p->t = curthread;
   t2p->pid = curproc->p_pid;

   /* Add mapping to list */
   WSS_MUTEX_ENTER( wsi_t2p_list );
   t2p->next = wsi_t2p_list;
   wsi_t2p_list = t2p;
   WSS_MUTEX_EXIT( wsi_t2p_list );

   /* Call the stage 1 exit code */
   wss_proc_exit_stage1( curproc );

   /* Call exitlwps */
   exitlwps( rcode );
}


/*  wsi_custom_pollcleanup */
/*! \brief Custom replacement for pollcleanup call in exit
 *
 *  This custom shim function provides a hook later in the exit call within
 *  Solaris so we can finish cleaning up wrappers-specific kernel data for the
 *  process.
 *
 *  We search the list of thread-to-process maps (which we should find if
 *  we arrive here) and call wss_proc_exit_stage2 with the process id to clean
 *  up the remaining structures which still exist (mainly the wss_proc
 *  structure associated with the exiting process is dereferenced).
 */
static void
wsi_custom_pollcleanup( kthread_t *t )
{
   struct wsi_t2p *t2p, **last;

   /* Search for thread-to-process mapping and unlink the match from
      the list */
   WSS_MUTEX_ENTER( wsi_t2p_list );
   for( last = &wsi_t2p_list, t2p = wsi_t2p_list; t2p;
        last = &t2p->next, t2p = t2p->next )
   {
      if( t2p->t == t )
      {
         *last = t2p->next;
         break;
      }
   }
   WSS_MUTEX_EXIT( wsi_t2p_list );

   if( !t2p )
      wsi_printf( PD_MAKESTR("Warning: Unable to determine thread->pid mapping\n") );
   else
   {
      /* Call stage 2 exit code if we have a mapping (we should always
         have a mapping) */
      wss_proc_exit_stage2( t2p->pid );

      /* If init is dieing, we're outta here */
      if( t2p->pid == 1 )
      {
         wsi_printf( PD_MAKESTR("Init process exited, cleaning up WSS ")
                     PD_MAKESTR("structures...\n") );
         _shutdown();
      }
   }

   /* Call pollcleanup */
   pollcleanup( t );

   /* Free the thread-to-process mapping */
   if( t2p )
      wss_free( t2p );
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_custom_uadmin */
/*! \brief Custom uadmin syscall implementations
 *
 *  \param call   System call identifier
 *  \param uap    Pointer to the parameter structure for this call
 *  \param retval Pointer to the return value storage space
 */
int
wsi_custom_uadmin(int call, struct uadmin_uap *uap, rval_t *retval )
{
   int ret;

   /* If the user is calling for a shutdown or reboot, set the shutdown
      flag so we can handle the shutdown situation */
   if( ( ( uap->cmd == A_SHUTDOWN ) || ( uap->cmd == A_REBOOT ) ) &&
       suser( CRED() ) )
      wsi_shutdown = 1;

   /* Call the dispatch routine */
   ret = wsi_dispatch(0, call, uap, retval );

   /* Hmm... didn't happen, unflag the shutdown */
   if( ( uap->cmd == A_SHUTDOWN ) || ( uap->cmd == A_REBOOT ) )
      wsi_shutdown = 0;

   return ret;
}

/*  wsi_custom_hook_fork_and_exec */
/*! \brief Apply hot-patches to Solaris kernel for hooks to cfork and exec calls
 *
 *  \return 0 on success,
 *  \return -1 on failure
 */
int
wsi_custom_hook_fork_and_exec()
{
   int error;

   KERN_HOOK_BEGIN( fork_hooks )
      KERN_HOOK( pgjoin, wsi_custom_pgjoin )
   KERN_HOOK_END;

   KERN_HOOK_BEGIN( exec_hooks )
      KERN_HOOK( exece, wsi_execve )
   KERN_HOOK_END;

   if( ( error = wsi_hook_add( PD_MAKESTR("cfork"), fork_hooks ) ) != 0 ) return error;
   if( ( error = wsi_hook_add( PD_MAKESTR("exec"), exec_hooks ) ) != 0 ) return error;

   return 0;
}

/*  wsi_custom_unhook_fork_and_exec */
/*! \brief Remove hooks from cfork and exec calls
 *
 *  \return Number of hooks removed
 */
int
wsi_custom_unhook_fork_and_exec()
{
   int count;

   count = wsi_hook_remove( PD_MAKESTR("cfork") );
   count += wsi_hook_remove( PD_MAKESTR("exec") );

   return count;
}

/*  wsi_custom_hook_exit */
/*! \brief Apply hot-patches to Solaris kernel for hooks to exit call
 *
 *  \return 0 on success,
 *  \return -1 on failure
 */
int
wsi_custom_hook_exit()
{
   KERN_HOOK_BEGIN( exit_hooks )
      KERN_HOOK( exitlwps, wsi_custom_exitlwps )
      KERN_HOOK( pollcleanup, wsi_custom_pollcleanup )
   KERN_HOOK_END;

   return( wsi_hook_add( PD_MAKESTR("exit"), exit_hooks ) );
}

