/*! \file wsi_dir.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_dir.c,v 1.7 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Replacement system calls for chdir, fchdir, and rename
 * \author  Steve Kiernan
 * \version $Revision: 1.7 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_path.h>
#include <wss_strings.h>
#include <wss_process.h>


/***************************************************************************\
 * Definitions for the replacement system calls                            *
\***************************************************************************/

extern struct sysent orig_chdir_sysent;
extern struct sysent orig_fchdir_sysent;
extern struct sysent orig_rename_sysent;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_rename */
/*! \brief Replacement syscall for rename to handle CWD caching
 *
 *  \param from Pointer to string containing source path
 *  \param to   Pointer to string containing target path
 *
 *  \sa rename(2) man page for return values
 */
longlong_t
wsi_rename( char *from, char *to )
{
   wss_proc_t *wrp;
   pd_char_t *path;
   pd_char_t *absfrom;
   rval_t      stupidunion;
   int         len;
   longlong_t  retval;

   /*
    * If the source isn't a directory, we don't care about it at all.
    */
#ifdef DEBUG_DIR
   wsi_printf( PD_MAKESTR("Calling wsi_path_is_dir\n") );
#endif /* DEBUG_DIR */
   if( wsi_path_is_dir( from, UIO_USERSPACE ) == FALSE )
   {
#ifdef DEBUG_DIR
      wsi_printf( PD_MAKESTR("Calling original rename (after dir check\n") );
#endif /* DEBUG_DIR */
      return( ( *orig_rename_sysent.sy_callc )( from, to ) );
   }

   wrp = wss_pfind( curproc->p_pid );
   if( wrp == NULL )
   {
#ifdef DEBUG_DIR
      wsi_printf( PD_MAKESTR("Calling original rename (after pfind)\n") );
#endif /* DEBUG_DIR */
      return( ( *orig_rename_sysent.sy_callc )( from, to ) );
   }

   /*
    * Convert from path to absolute terms
    */
   path = (pd_char_t *)wss_malloc( MAXPATHLEN );
   copyinstr( (pd_char_t *)from, path, MAXPATHLEN, &len );

#ifdef DEBUG_DIR
   wsi_printf( PD_MAKESTR("rename: source = '%s'\n"), path );
#endif /* DEBUG_DIR */

   WSI_RW_ENTER( wrp->wrp_cwd, RW_READER );
   absfrom = wss_path_rel_to_abspath( wrp->wrp_cwd, path );
   WSI_RW_EXIT( wrp->wrp_cwd );

   wss_proc_rele( wrp );

   wss_free( path );

   /*
    * Call the original rename syscall.
    */
   retval = ( *orig_rename_sysent.sy_callc )( from, to );
   stupidunion.r_vals = retval;

   /*
    * Continue if the original syscall succeeded.
    */
   if( from == NULL )
      return retval;

   if( stupidunion.r_val1 != 0 )
   {
      wss_free( absfrom );
      return retval;
   }
       
   WSS_MUTEX_ENTER( wss_proc_list );
   for( wrp = wss_proc_list.lh_first; wrp; wrp = wrp->wrp_p_list.le_next )
   {
      if( wrp->wrp_cwd )
      {
         WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
         if( wrp->wrp_cwd && wss_strstr( absfrom, wrp->wrp_cwd ) )
         {
            if( wrp->wrp_p == curproc )
            {
               char *tmp = wrp->wrp_cwd;
               wrp->wrp_cwd = wss_path_cwd( wrp );
               if( tmp )
                  wss_free( tmp );
            }
            else
            {
               WSS_MUTEX_ENTER( wrp->wrp );
               wrp->wrp_flags |= WRPF_GET_CWD;
               WSS_MUTEX_EXIT( wrp->wrp );
            }
         }
         WSI_RW_EXIT( wrp->wrp_cwd );
      }
   }
   WSS_MUTEX_EXIT( wss_proc_list );

   wss_free( absfrom ); /* We shouldn't need this anymore. */
       
   return retval;
}      

/*  wsi_chdir */
/*! \brief Replacement syscall for chdir to handle CWD caching
 *
 *  \param path Pointer to string containing the path to change
 *
 *  \sa chdir(2) man page for return values
 */
longlong_t
wsi_chdir( char *path )
{
   wss_proc_t *wrp;
   char       *tmp;
   rval_t      stupidunion;
   longlong_t  retval;
   
   /*
    * Call the original chdir syscall.
    */
   retval = ( *orig_chdir_sysent.sy_callc )( path );
   stupidunion.r_vals = retval;

   /*
    * Continue if the original syscall succeeded.
    */
   if( stupidunion.r_val1 != 0 )
      return retval;

   wrp = wss_pfind( curproc->p_pid );

   /*
    * Only care about wrapped processes.
    */
   if( !wrp )
      return retval;

   WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
   tmp = wrp->wrp_cwd;
   wrp->wrp_cwd = wss_path_cwd( wrp );
   WSI_RW_EXIT( wrp->wrp_cwd );

   if( tmp )
      wss_free( tmp );

   wss_proc_rele( wrp );

   return retval;
}


/*  wsi_fchdir */
/*! \brief Replacement syscall for fchdir to handle CWD caching
 *
 *  \param fd     File descriptor to use as CWD
 *
 *  \sa fchdir(2) man page for return values
 */
longlong_t
wsi_fchdir( int fd )
{
   wss_proc_t *wrp;
   char       *tmp;
   rval_t      stupidunion;
   longlong_t  retval;

   /*
    * Call the original chdir syscall.
    */
   retval = ( *orig_fchdir_sysent.sy_callc )( fd );
   stupidunion.r_vals = retval;

   /*
    * Continue if the original syscall succeeded.
    */
   if( stupidunion.r_val1 != 0 )
      return retval;

   wrp = wss_pfind( curproc->p_pid );
   if( !wrp )
      return retval;

   WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
   tmp = wrp->wrp_cwd;
   wrp->wrp_cwd = wss_path_cwd( wrp );
   WSI_RW_EXIT( wrp->wrp_cwd );
   if( tmp )
      wss_free( tmp );

   wss_proc_rele( wrp );

   return retval;
}

