/*! \file wsi_dispatch.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_dispatch.c,v 1.45 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   This is the main dispatch routine for all syscalls -- Solaris style
 * \author  Steve Kiernan
 * \version $Revision: 1.45 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

/*
 * Disable reference debugging messages within this file.
 *
 * Trust me, you don't want them here.
 */
#ifndef DISABLE_PROC_DEBUG_REFS
#define DISABLE_PROC_DEBUG_REFS
#endif /* DISABLE_PROC_DEBUG_REFS */

#include <wsi_stdlib.h>

#include <sys/types.h>
#include <sys/systm.h>
#include <sys/syscall.h>

#include <wss_malloc.h>
#include <wss_criteria.h>
#include <wss_process.h>
#include <wss_module.h>
#include <wss_path.h>
#include <wss_arg.h>

#include <wsi_syscalls.h>
#include <wsi_os.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern struct sysent saved_sysent[NSYSCALL];
extern int           wsi_unloading;
extern int           wsi_shutdown;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_dispatch */
/*! \brief Old-style syscall dispatch routine
 *
 *  \param space  Event space identifier
 *  \param code   Call identifier
 *  \param uap    Pointer to the call argument structure
 *  \param retval Pointer to location to store the return value
 *
 *  \return The value returned by the original call
 */
int
wsi_dispatch( int space, pd_syscall_id_t code, void *uap, rval_t *retval )
{
   wss_proc_t              *wrp;
   wsi_proc_t              *wrpi;
   proc_t                  *p;
   struct sysent           *callp;      /* points to the true system call */
   int                      ret;
   int                      err;
   wsi_proc_dispatch_t     *dispatch;
   wsi_proc_dispatch_t     *prev_dispatch = NULL;
   wss_wrapper_dispatch_t   intercept;
   wss_wrapper_data_t       wrd = {0,0,0};
   int                      rval = 0;

   p = curproc;
   callp = &saved_sysent[code];
   wrd.errnum = &err;

   /* Find the wss_proc structure for this process */
   wrp = wss_pfind( p->p_pid );
   if( !wrp )
   {
      if( !space )
         return( (*callp->sy_call)( uap, retval ) );
      else
         return (0);
   }

   wrpi = (wsi_proc_t *)wrp->wrp_opaque;

   /* Call pre functions for each wrapper */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( dispatch = wrpi->dispatchers[code].list.lh_first;
        dispatch;
        prev_dispatch = dispatch, dispatch = dispatch->next.le_next )
   {
      if( dispatch->instance->wi_wrapper->w_dispatch_pre )
      {
         int inerror;

         intercept = dispatch->instance->wi_wrapper->w_dispatch_pre;
         wrd.wiv = dispatch->instance->wi_data;
#ifdef DEBUG_DISPATCH
         wsi_printf( PD_MAKESTR("%d: Calling pre [%d:%d]\n"), code,
                     wsi_proc_get_current_pid(), wsi_os_curthreadid() );
#endif
         inerror = (*intercept)( space, code, uap, &rval,
                                 wrp, &wrd, (space == 0) );
#ifdef DEBUG_DISPATCH
         wsi_printf( PD_MAKESTR("old-call %d: rval = %d [%d:%d]\n"), code,
                     rval, wsi_proc_get_current_pid(), wsi_os_curthreadid() );
#endif
         if( ( inerror & WR_MASK1 ) == WR_DENY )
         {
#ifdef DEBUG_DISPATCH
            wsi_printf( PD_MAKESTR("old-call %d: *wrd.errnum = %d [%d:%d]\n"),
                        code, *wrd.errnum, wsi_proc_get_current_pid(),
                        wsi_os_curthreadid() );
#endif
            retval->r_val1 = set_errno( *wrd.errnum );
            retval->r_val2 = *wrd.errnum;
            ret = err;

            goto os_exit;
         }
      }
   }
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   /* Call the actual syscall.  */ 
   WSS_MUTEX_ENTER( wrp->wrp );
   wrp->wrp_nsyscall++;
   WSS_MUTEX_EXIT( wrp->wrp );

   wss_proc_rele( wrp );

   ret = (*callp->sy_call)( uap, retval );


   /* Note: This probably isn't needed as fork doesn't return back
    *       through the syscall stack on Solaris, it just jumps back
    *       to userspace
    */
   p = curproc;
   wrp = wss_pfind( p->p_pid );
   if( !wrp )
      cmn_err( CE_PANIC,
              PD_MAKESTR("Unable to find process #%d after syscall #%d\n"),
              p->p_pid, code );
   /* End note */

   WSS_MUTEX_ENTER( wrp->wrp );
   wrp->wrp_nsyscall--;
   WSS_MUTEX_EXIT( wrp->wrp );


   /*
    * Call post functions for each wrapper.  We perform the
    * "truncated onion method" for failure handling.  Basically
    * if there was a failure, we start with the wrapper that
    * cause the error and work our way back to the first wrapper
    * called.  In this way, after a PRE event fails, no more PRE
    * events run.  POST events start with the failing one and
    * continue with POST events for all prior PREs.
    */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );

   /* Hmmm... race condition anyone?  What happens if the instance list
      changes sometime after we've done this? */
   if( dispatch == NULL )
      dispatch = prev_dispatch;
   for( ; dispatch;
        dispatch =
           ( dispatch == *(dispatch->next.le_prev) ) ?
              NULL :
              *(dispatch->next.le_prev) )
   {
      if( dispatch->instance->wi_wrapper->w_dispatch_post )
      {
         int inerror;

         intercept = dispatch->instance->wi_wrapper->w_dispatch_post;
#ifdef DEBUG_DISPATCH
         wsi_printf( PD_MAKESTR("old-call %d: Calling post [%d:%d]\n"), code,
                     wsi_proc_get_current_pid(), wsi_os_curthreadid() );
#endif
         wrd.wiv = dispatch->instance->wi_data;

         if( curthread->_tu._ts._t_post_sys )
             *wrd.errnum = curthread->t_lwp->lwp_errno;
         rval = retval->r_val1;

         inerror = (*intercept)( space, code, uap, &rval,
                                 wrp, &wrd, (space == 0));
         if( ( inerror & WR_MASK1 ) == WR_DENY )
         {
#ifdef DEBUG_DISPATCH
            wsi_printf( PD_MAKESTR("old-call %d: *wrd.errnum = %d [%d:%d]\n"),
                        code, *wrd.errnum, wsi_proc_get_current_pid(),
                        wsi_os_curthreadid() );
#endif
            retval->r_val1 = set_errno( *wrd.errnum );
            retval->r_val2 = *wrd.errnum;
            ret = err;

            goto os_exit;
         }
      }
   }
os_exit:
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   /* Clean up arg replacement */
   wss_arg_donesyscall( wrp, wsi_os_curthreadid(), 0, code, uap );

   /* If we're unloading, handle syscall reissuing properly */
   if( wsi_unloading )
   {
      if( wsi_proc_insyscall( wrp ) <= 0 )
      {
         WSS_MUTEX_ENTER_DEBUG( wss_proc_num_in_syscall );

         wss_proc_num_in_syscall--;
         WSS_CV_SIGNAL_DEBUG( wss_proc_num_in_syscall );

         WSS_MUTEX_EXIT_DEBUG( wss_proc_num_in_syscall );
      }
      else
         wsi_printf( PD_MAKESTR("pid %d, tid %d, syscall %d exiting\n"),
                     (int)curproc->p_pid, (int)curthread->t_tid, code );
   }

   wss_proc_rele( wrp );

   return( ret ); /* ret is from syscall execution or set by the pre */
}

/*  wsi_dispatch_c */
/*! \brief C-style syscall dispatch routine
 *
 *  \param space Event space identifier
 *  \param code  Call indentifier
 *  \param i1    Input register 1 value
 *  \param i2    Input register 2 value
 *  \param i3    Input register 3 value
 *  \param i4    Input register 4 value
 *  \param i5    Input register 5 value
 *  \param i6    Input register 6 value
 *
 *  \b Note:
 *      The i1-i6 stand for input registers i1-i5, plus the 6th register
 *      possibly used by syscalls which resides on the stack.  We use ints,
 *      since the generated assembly will use one register per int and do
 *      everything we need done properly.  Just trust me on this, it works
 *      correctly.
 *
 *  \return Return code from pre/post or original call
 */
longlong_t wsi_dispatch_c( int space, pd_syscall_id_t code,
        int i1, int i2, int i3, int i4, int i5, int i6 )
{
   wss_proc_t              *wrp;
   wsi_proc_t              *wrpi;
   proc_t                  *p;
   struct sysent           *callp;
   longlong_t               ret = 0;
   void                    *uap;
   int                      err;
   wsi_proc_dispatch_t     *dispatch;
   wsi_proc_dispatch_t     *prev_dispatch = NULL;
   wss_wrapper_dispatch_t   intercept;
   wss_wrapper_data_t       wrd = {0,0};
   int                      rval = 0;
#ifdef DEBUG_DISPATCH
   int                      fWrapped = 0;
#endif

   p = curproc;
   callp = &saved_sysent[code];
   wrd.errnum = &err;

   /* Find the wss_proc structure for this process */
   wrp = wss_pfind( p->p_pid );
   if( !wrp )
   {
      if( !space )
         return( (*callp->sy_callc)( i1, i2, i3, i4, i5, i6 ) );
      else
         return (0);
   }

   /* Construct uap argument structure for pre calls */
   uap = wss_malloc( sizeof( int ) * 6 );
   *((int *)uap) = i1;
   *((int *)uap + 1) = i2;
   *((int *)uap + 2) = i3;
   *((int *)uap + 3) = i4;
   *((int *)uap + 4) = i5;
   *((int *)uap + 5) = i6;

   wrpi = (wsi_proc_t *)wrp->wrp_opaque;

   /* Call pre functions for each wrapper */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( dispatch = wrpi->dispatchers[code].list.lh_first;
        dispatch;
        prev_dispatch = dispatch, dispatch = dispatch->next.le_next )
   {
      if( dispatch->instance->wi_wrapper->w_dispatch_pre )
      { 
         int inerror;

         intercept = dispatch->instance->wi_wrapper->w_dispatch_pre;
#ifdef DEBUG_DISPATCH
         wsi_printf( PD_MAKESTR("c-call %d: Calling pre [%d:%d]\n"), code,
                     wsi_proc_get_current_pid(), wsi_os_curthreadid() );
         fWrapped = 1;
#endif
         wrd.wiv = dispatch->instance->wi_data;
         inerror = (*intercept)( space, code, uap, &rval,
                                 wrp, &wrd, (space == 0) );
#ifdef DEBUG_DISPATCH
         wsi_printf( PD_MAKESTR("c-call %d: rval = %d [%d:%d]\n"), code, rval,
                     wsi_proc_get_current_pid(), wsi_os_curthreadid() );
#endif
         if( ( inerror & WR_MASK1 ) == WR_DENY )
         {
            rval_t retval;

#ifdef DEBUG_DISPATCH
            wsi_printf( PD_MAKESTR("c-call %d: *wrd.errnum = %d [%d:%d]\n"),
                        code, *wrd.errnum, wsi_proc_get_current_pid(),
                        wsi_os_curthreadid() );
#endif
            retval.r_val1 = set_errno( *wrd.errnum );
            retval.r_val2 = *wrd.errnum;
            ret = retval.r_vals;

            goto cs_exit;
         }
      } 
   }   
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );


   /* Replace syscall args with possible changes from pre */
   i1 = *((int *)uap);
   i2 = *((int *)uap + 1);
   i3 = *((int *)uap + 2);
   i4 = *((int *)uap + 3);
   i5 = *((int *)uap + 4);
   i6 = *((int *)uap + 5);


   /* Call the actual syscall. */ 
   WSS_MUTEX_ENTER( wrp->wrp );
   wrp->wrp_nsyscall++;
   WSS_MUTEX_EXIT( wrp->wrp );

   wss_proc_rele( wrp );

   ret = (*callp->sy_callc)(i1,i2,i3,i4,i5,i6);
#ifdef DEBUG_DISPATCH
   if( fWrapped )
   {
       rval_t retval;
       retval.r_vals = ret;
       wsi_printf( PD_MAKESTR("c-call %d: real call: ret = %d [%d:%d]\n"),
                   code, retval.r_val2, wsi_proc_get_current_pid(),
                   wsi_os_curthreadid() );
   }
#endif

   /* Note: This probably isn't needed as fork doesn't return back
    *       through the syscall stack on Solaris, it just jumps back
    *       to userspace
    */
   p = curproc;
   wrp = wss_pfind( curproc->p_pid );
   if( !wrp )
      cmn_err( CE_PANIC,
              PD_MAKESTR("Unable to find process #%d after syscall #%d\n"),
              p->p_pid, code );
   /* End note */

   WSS_MUTEX_ENTER( wrp->wrp );
   wrp->wrp_nsyscall--;
   WSS_MUTEX_EXIT( wrp->wrp );

   /*
    * Call post functions for each wrapper.  We perform the
    * "truncated onion method" for failure handling.  Basically
    * if there was a failure, we start with the wrapper that
    * cause the error and work our way back to the first wrapper
    * called.  In this way, after a PRE event fails, no more PRE
    * events run.  POST events start with the failing one and
    * continue with POST events for all prior PREs.
    */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   if( dispatch == NULL )
      dispatch = prev_dispatch;
   for( ; dispatch;
        dispatch =
           ( dispatch == *(dispatch->next.le_prev) ) ?
              NULL :
              *(dispatch->next.le_prev) )
   {
      if( dispatch->instance->wi_wrapper->w_dispatch_post )
      { 
         int    inerror;
         rval_t retval;

         intercept = dispatch->instance->wi_wrapper->w_dispatch_post;
#ifdef DEBUG_DISPATCH
         wsi_printf( PD_MAKESTR("c-call %d: Calling post [%d:%d]\n"), code,
                     wsi_proc_get_current_pid(), wsi_os_curthreadid() );
#endif
         wrd.wiv = dispatch->instance->wi_data;

         if( curthread->_tu._ts._t_post_sys )
             *wrd.errnum = curthread->t_lwp->lwp_errno;
         retval.r_vals = ret;
         rval = retval.r_val1;

         inerror = (*intercept)( space, code, uap, &rval,
                                wrp, &wrd, (space == 0) );

         if( ( inerror & WR_MASK1 ) == WR_DENY )
         {
             retval.r_val1 = set_errno( *wrd.errnum );
             retval.r_val2 = *wrd.errnum;
             ret = retval.r_vals;
#ifdef DEBUG_DISPATCH
             wsi_printf( PD_MAKESTR("%c-call d: Returning error code %d ")
                         PD_MAKESTR("[%d:%d]\n"), code, retval.r_val2,
                         wsi_proc_get_current_pid(), wsi_os_curthreadid() );
#endif

         }
      } 
   }
cs_exit:
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   /* Clean up arg replacement */
   wss_arg_donesyscall( wrp, wsi_os_curthreadid(), 0, code, uap );

   /* If we're unloading, handle syscall reissuing properly */
   if( wsi_unloading )
   {
      if( wsi_proc_insyscall( wrp ) <= 0 )
      {
         WSS_MUTEX_ENTER_DEBUG( wss_proc_num_in_syscall );

         wss_proc_num_in_syscall--;
         WSS_CV_SIGNAL_DEBUG( wss_proc_num_in_syscall );

         WSS_MUTEX_EXIT_DEBUG( wss_proc_num_in_syscall );
      }
      else
         wsi_printf( PD_MAKESTR("pid %d, tid %d, syscall %d exiting\n"), (int)curproc->p_pid,
                 (int)curthread->t_tid, code );
   }

   wss_proc_rele( wrp );

   wss_free( uap );

   return( ret );
}

/*  wsi_syscall_name */
/*! \brief Return syscall name based on code argument
 *
 *  \param code Syscall indentifier
 *
 *  \return Pointer to syscall name
 */
const pd_char_t *
wsi_syscall_name( pd_syscall_id_t code )
{
   return solaris_syscallnames[code];
}

