/*! \file wsi_execve.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_execve.c,v 1.27 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   A replacement for the execve syscall.
 * \author  Chris Vance, Steve Kiernan
 * \version $Revision: 1.27 $
 * \date    $Date: 2001/02/23 19:02:42 $
 *
 * This does some wrappers-relevant work and then passes control
 * off to the original execve system call (called exece in the
 * Solaris kernel).
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/syscall.h>

#include <wss_path.h>
#include <wss_malloc.h>
#include <wss_process.h>
#include <wss_criteria.h>
#include <wss_log.h>
#include <wss_wrapper.h>
#include <wsi_execve.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern struct sysent orig_execve_sysent;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_execve */
/*! \brief  The replacement function for the execve system call.
 *
 *  It does some wrappers-specific stuff, and then calls the original
 *  execve syscall.  Of importance is the need to look up the full
 *  pathname for the program being executed.  This should only need to
 *  be done if the exec was a success.  With any luck, most of the
 *  processing it does has been cached during the similar lookup in the
 *  actual syscall.  The program name is also available in argv[0]
 *  (relative to the cwd) and is stored in p->p_comm by the execve
 *  syscall (only the first 16 chars of the binary name - no path).
 *
 *  \param uap    Pointer to the call arugment structure
 *  \param retval Pointer to memory where return value should be stored
 *
 *  \return Value returned by original execve call
 */
int
wsi_execve( struct execve_uap *uap, rval_t *retval )
{
   struct wss_proc *wrp;
   struct proc     *p;
   int              err;
   int              act_error;
   int              wr_run_wrapped;
   pd_char_t      *tmp;
   pd_char_t      *path;
   int              len;
   int              i;


   /*
    * Look up the current process in the wrappers proc
    * structure. If it wasn't found, this means there was either
    * a problem, or this is a process that was running before the
    * wrappers module was loaded into the kernel.  In the first
    * case, this is a big problem.  In the second case, we can
    * probably work around it by creating the structure at this
    * time, but until I'm possitive, I'm leaving this as a TBD
    */
   p = curproc;
   wrp = wss_pfind( p->p_pid );
   if( !wrp )
   {
      wsi_printf( PD_MAKESTR("wsi_execve: can't find pid: p->p_pid %d\n"), (int)p->p_pid );

      wrp = wss_proc_new( p );
   }

   /*
    * wrp was tested for validity early on.  Make space and copy
    * the path, first saving the current one in an 'old' pathbuf
    */
   if( wrp->wrp_oldpath )
      wss_free( wrp->wrp_oldpath );

   WSI_RW_ENTER( wrp->wrp_cwd, RW_WRITER );
   tmp = wrp->wrp_cwd;
   wrp->wrp_cwd = wss_path_cwd( wrp );
   if( tmp )
      wss_free(tmp);
   WSI_RW_EXIT( wrp->wrp_cwd );

   path = wss_malloc( MAXPATHLEN );
   copyinstr( (pd_char_t *)uap->path, path, MAXPATHLEN, &len );
   wrp->wrp_oldpath = wrp->wrp_path; 
   WSI_RW_ENTER( wrp->wrp_cwd, RW_READER );
   wrp->wrp_path = wss_path_rel_to_abspath( wrp->wrp_cwd, path );
   WSI_RW_EXIT( wrp->wrp_cwd );
   wss_free( path );

   /* Clear out progname if its currently set */
   if( wrp->wrp_progname != NULL )
      wss_free( wrp->wrp_progname );

   /*
    * Generate the program name, full path should have at least
    * one '/' character.  Program name is everything after the
    * last '/'.
    */
   for( len = 0, i = wsi_strlen( wrp->wrp_path );
        ( i > 0 ) && ( wrp->wrp_path[i] != '/' );
        i--, len++ );

   /*
    * Check (for sanity, etc)
    */
   if( ( i == 0 ) && ( wrp->wrp_path[0] != '/' ) )
   {
      wsi_printf( PD_MAKESTR("wsi_execve: Pathname `%s' didn't have a '/' in it\n"),
              wrp->wrp_path );
      wrp->wrp_progname = NULL;
   }
   else
   {
      wrp->wrp_progname = (pd_char_t *)wss_malloc_string( len );
      bcopy( &( wrp->wrp_path[i + 1] ), wrp->wrp_progname,
             len * sizeof(pd_char_t) );
   }

   wss_proc_rele( wrp );

   /*
    * Next, call the original syscall.
    */ 
   err = (*orig_execve_sysent.sy_call)( uap, retval );

   wrp = wss_pfind( p->p_pid );
   if( !wrp )
      panic( "Unable to find wrapper process structure after exec" );

   /*
    * TBD: Again, we are checking for error incorrectly.  This
    * should probably be checking a combination of *retval and
    * err from the syscall.
    */
   if( retval->r_val1 == 0 )
   {
      /* zero from execve syscall indicates success? */
      int has_wrappers;

      /*
       * We may now check the activation criteria to determine
       * whether or not to run this process wrapped.
       */
      wss_crit_check_deactivation( wrp );
      wr_run_wrapped = wss_crit_check_activation( wrp );

      WSS_MUTEX_ENTER( wrp->wrp_wi_list );
      has_wrappers = (wrp->wrp_wi_list == NULL) ? FALSE : TRUE;
      WSS_MUTEX_EXIT( wrp->wrp_wi_list );

      if( has_wrappers )
      {
         /*
          * Now is a convenient place to call the
          * wrapper-specific wr_activation() functions.
          */
         act_error = wss_wrapper_call_activation( wrp );
         if( act_error != 0 )
         {
            /*
             * TBD: Is there anything else we should
             * do here, like kill the process?
             */
            wsi_printf( PD_MAKESTR("Activate returned an error!\n") );
         }
      }
   }

#ifdef LOG_WRAPPED
   (void)wss_log_printf( PD_MAKESTR("Running %s %swrapped\n"), wrp->wrp_path,
                         ( wrp->wrp_wi_list?PD_MAKESTR(""):PD_MAKESTR("un")) );
#endif /* LOG_WRAPPED */

   wss_proc_rele( wrp );

   return( err );
}

