/*! \file wsi_file.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_file.c,v 1.17 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   System-dependent file I/O functions
 * \author  Linda Thomas, Steve Kiernan
 * \version $Revision: 1.17 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <stdarg.h>
#include <sys/param.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/file.h>

#include <vm/as.h>

#include <wss_malloc.h>
#include <wss_ref.h>
#include <wss_file.h>


/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

/*! \brief File management structure */
struct wsi_file
{
   WSS_MUTEX_PROTECT(f);           /*!< Mutex to protect the file      */
   wss_ref_t         f_refcount;   /*!< Number of references to file   */
   struct vnode     *f_vp;         /*!< Pointer to V-node of file      */
   struct cred      *f_cr;         /*!< Pointer to credentials of file */
   void             *f_mapAddr;    /*!< Address where file is mapped   */
   size_t            f_mapSize;    /*!< Size of mapped memory area     */
   pd_offset_t      f_offset;     /*!< Seek offset into file          */
};


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_file_open */
/*! \brief Construct wsi_file based on a path and open it
 *
 *  \param path Pathname of file to open
 *  \param mode Mode flags with which to open the file
 *
 *  \return Handle to open file on success,
 *  \return \b WSS_FILE_INVALID on failure
 */
wss_file_t
wsi_file_open( const pd_char_t *path, wss_file_mode_t mode )
{
   struct vnode    *vp        = NULL;
   int              err;
   struct wsi_file *file      = NULL;
   int              filemode  = 0;
   cred_t          *old_creds = NULL;

#ifdef DEBUG_FILE
   wsi_printf( PD_MAKESTR("creating '%s'\n"), path );
#endif

   if( mode & WSS_FILE_READ )
      filemode |= FREAD;
   if( mode & WSS_FILE_WRITE )
      filemode |= FWRITE;
   if( mode & WSS_FILE_APPEND )
      filemode |= FAPPEND;
   if( mode & WSS_FILE_TRUNC )
      filemode |= FTRUNC;
   if( mode & WSS_FILE_CREAT )
      filemode |= FCREAT;

   if( mode & WSS_FILE_ROOTPRIV )
   {
      old_creds = curthread->t_cred;
      curthread->t_cred = kcred;
   }

   err = vn_open( (pd_char_t *)path, UIO_SYSSPACE, filemode,
                  S_IFREG | S_IREAD | S_IWRITE, &vp, CRCREAT );

   if( mode & WSS_FILE_ROOTPRIV )
      curthread->t_cred = old_creds;

#ifdef DEBUG_FILE
   wsi_printf( PD_MAKESTR("create result %d\n"), err );
#endif

   if( err == 0 )
   {
      file = (struct wsi_file *)wss_malloc( sizeof(struct wsi_file) );

      if( file )
      {
         WSS_MUTEX_INIT( file->f );
         wss_ref_init( &(file->f_refcount) );
         file->f_vp = vp;
         file->f_offset = 0;
         file->f_mapAddr = NULL;
         file->f_mapSize = 0;
         if( mode & WSS_FILE_ROOTPRIV )
            file->f_cr = kcred;
         else
            file->f_cr = curthread->t_cred;
      }
      else
         (void) VOP_CLOSE( vp, 0, 0, (offset_t) 0,
                           ( mode & WSS_FILE_ROOTPRIV ) ? kcred
                                                        : curthread->t_cred );
   }

   return( (void *)file );
}

/*  wsi_file_free */
/*! \brief Really close the file and free structure memory
 *
 *  \param handle Handle to an open file
 */
void
wsi_file_free( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   assert( file != NULL );
   assert( WSS_REF_COUNT(file->f_refcount) == 1 );

   WSS_REF_DEC(file->f_refcount);

   if( file )
   {
      if( file->f_vp )
      {
#if DEBUG_FILE
         int err = VOP_CLOSE( file->f_vp, 0, 0, (offset_t) 0, file->f_cr );
         wsi_printf( PD_MAKESTR("close of %p returned %d\n"), file->f_vp,
                     err );
#else
         (void) VOP_CLOSE( file->f_vp, 0, 0, (offset_t) 0, file->f_cr );
#endif
      }

      WSS_MUTEX_DESTROY( file->f );
      wss_ref_destroy( &(file->f_refcount) );
      wss_free( file );
   }
}

/*  wsi_file_close */
/*! \brief Close previously opened file
 *
 *  \param handle Handle to an open file
 */
void
wsi_file_close( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      if (file->f_mapAddr)
          wsi_file_unmap(handle);

      wsi_file_rele(handle);
   }
}


/*  wsi_file_ref */
/*! \brief Acquire reference to file sturcture
 *
 *  \param handle Handle to an open file
 */
void
wsi_file_ref( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( WSS_REF_COUNT(file->f_refcount) < 1 )
      panic( "wsi_file_ref: reference count for 0x%x was %d < 1", file,
             WSS_REF_COUNT(file->f_refcount) );

   WSS_MUTEX_ENTER( file->f );
   WSS_REF_INC(file->f_refcount);
   WSS_MUTEX_EXIT( file->f );
}

                               
/*  wsi_file_rele */
/*! \brief Release reference to file sturcture
 *
 *  \param handle Handle to an open file
 */
void
wsi_file_rele( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( WSS_REF_COUNT(file->f_refcount) < 1 )
      panic( "wsi_file_rele: reference count for 0x%x was %d < 1", file,
             WSS_REF_COUNT(file->f_refcount) );

   WSS_MUTEX_ENTER( file->f );
   if(WSS_REF_COUNT(file->f_refcount) > 1)
   {
       WSS_REF_DEC(file->f_refcount);
       WSS_MUTEX_EXIT( file->f );
   }
   else
      wsi_file_free(file);
}


/*  wsi_file_write */
/*! \brief Write data to open file
 *
 *  \param handle Handle to an open file
 *  \param buffer Pointer to buffer containing data to be written
 *  \param len    Number of bytes of data in buffer
 *
 *  \return number of bytes written on sucess,
 *  \return -EINVAL if \a handle is \b WSS_FILE_INVALID
 *  \return negated error code for errors associated with write(2)
 *
 *  \sa write(2) man page
 */
int
wsi_file_write( wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      struct uio       io;
      struct iovec     iovector;
      int              err = 0;

#ifdef DEBUG_FILE
      wsi_printf( PD_MAKESTR("file %p, buffer %p, len %d\n"), file, buffer,
                  len );
#endif

      iovector.iov_base = buffer;
      iovector.iov_len  = len;
  
      bzero( &io, sizeof(io) );
      io.uio_iov     = &iovector;
      io.uio_iovcnt  = 1;
      io.uio_fmode   = FWRITE | FSYNC | FDSYNC;
      io.uio_segflg  = UIO_SYSSPACE;
      io.uio_loffset = file->f_offset;
      io.uio_llimit  = file->f_offset + len;
      io.uio_resid   = len;

      VOP_RWLOCK( file->f_vp, RW_WRITER );
      err = VOP_WRITE( file->f_vp, &io, FKLYR, file->f_cr );
      VOP_RWUNLOCK( file->f_vp, RW_WRITER );

      if( err == 0 )
         file->f_offset += len;

#ifdef DEBUG_FILE
      wsi_printf( PD_MAKESTR("err = %d, resid = %lu\n"), err,
                  (u_long)io.uio_resid );
#endif

      return( ( err == 0 ) ? len : -err );
   }
   else
      return -EINVAL;
}

/*  wsi_file_read */
/*! \brief Read data from open file
 *
 *  \param handle Handle to an open file
 *  \param buffer Pointer to buffer which will contain read data
 *  \param len    Number of bytes of data to read 
 *
 *  \return number of bytes read on sucess,
 *  \return -EINVAL if \a handle is \b WSS_FILE_INVALID
 *  \return negated error code for errors associated with read(2)
 *
 *  \sa read(2) man page
 */
int
wsi_file_read( wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      struct uio       io;
      struct iovec     iovector;
      int              err = 0;

#ifdef DEBUG_FILE
      wsi_printf( PD_MAKESTR("file %p, buffer %p, len %d\n"), file, buffer,
                  len );
#endif

      iovector.iov_base = buffer;
      iovector.iov_len  = len;
  
      bzero( &io, sizeof(io) );
      io.uio_iov     = &iovector;
      io.uio_iovcnt  = 1;
      io.uio_fmode   = FREAD | FSYNC | FDSYNC;
      io.uio_segflg  = UIO_SYSSPACE;
      io.uio_loffset = file->f_offset;
      io.uio_llimit  = file->f_offset + len;
      io.uio_resid   = len;

      VOP_RWLOCK( file->f_vp, RW_READER );
      err = VOP_READ( file->f_vp, &io, FKLYR, file->f_cr );
      VOP_RWUNLOCK( file->f_vp, RW_READER );

      if( err == 0 )
         file->f_offset += len;

#ifdef DEBUG_FILE
      wsi_printf( PD_MAKESTR("err = %d, resid = %lu\n"), err,
                  (u_long)io.uio_resid );
#endif

      return( ( err == 0 ) ? len : -err );
   }
   else
      return -EINVAL;
}

/*  wsi_file_seek */
/*! \brief Seek withing an open file
 *
 *  \param handle Handle to an open file
 *  \param offset Offset into file
 *  \param whence Where to seek from:
 *                   \b WSS_FILE_SEEK_END - seek from end of file
 *                   \b WSS_FILE_SEEK_CUR - seek from current position in file
 *                   \b WSS_FILE_SEEK_SET - set current position in file
 *
 *  \return 0 on success,
 *  \return EINVAL if \a handle is \b WSS_FILE_INVALID or \a whence is not one
 *                 of the three valid paramters above
 */
int
wsi_file_seek( wss_file_t handle, pd_offset_t offset, int whence )
{
   struct wsi_file *file = (struct wsi_file *)handle;
   int              err = 0;

   if( file )
   {
      switch( whence )
      {
         case WSS_FILE_SEEK_END:
         {
            int size = wsi_file_get_size( file, &err );
            if( !err )
               file->f_offset = size + offset;
            break;
         }
         case WSS_FILE_SEEK_CUR:
            file->f_offset += offset;
            break;
         case WSS_FILE_SEEK_SET:
            file->f_offset = offset;
            break;
         default:
            err = EINVAL;
      }

#ifdef DEBUG_FILE
      wsi_printf( PD_MAKESTR("file size is %lu, err is %d\n"), (u_long)file->f_offset, err );
#endif
   }
   else
      err = EINVAL;

   return err;
}

/*  wsi_file_get_size */
/*! \brief Return the size of the file
 *
 *  \param handle Handle to an open file
 *  \param error  If non-NULL, pointer to memory to store error code
 *
 *  \return Size of file on success,
 *  \return 0 on failure
 */
size_t
wsi_file_get_size( wss_file_t handle, int *error )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file )
   {
      struct vattr     attributes;
      int              err;

      err = VOP_GETATTR( file->f_vp, &attributes, AT_SIZE, file->f_cr );
      if( error )
         *error = err;

      if( err )
         return 0;
      else
         return attributes.va_size;
   }
   else
   {
      if( error )
         *error = EINVAL;
      return 0;
   }
}


/*  wsi_file_get_map_addr */
/*! \brief Return the address where the file is mapped
 *
 *  \param handle Handle to an open file
 *
 *  \return Pointer to address where file is mapped on success,
 *  \return 0 on failure
 */
void *
wsi_file_get_map_addr( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;
   void            *addr = NULL;

   if( file )
      addr = file->f_mapAddr;

   return addr;
}


/*  wsi_file_map */
/*! \brief Map an open file into the kernel
 *
 *  \param handle  Handle to an open file
 *  \param size    Number of bytes to map
 *  \param mapmode Mode flags to map file
 *  \param pos     Offset into file to map
 *
 *  \return 0 on success,
 *  \return EINVAL if \a handle is \b WSS_FILE_INVALID,
 *  \return error code for errors associated with mmap(2)
 */
int
wsi_file_map( wss_file_t handle, size_t size, int mapmode, pd_offset_t pos )
{
   struct wsi_file *file  = (struct wsi_file *)handle;
   int              error = EINVAL;

   if( file )
   {
      unsigned int prot = 0;

      if( wsi_file_get_size( file, NULL ) < size )
      {
         char *bb = "xxx";

         wsi_printf( PD_MAKESTR("WSS: creating %d byte memory-mapped file.\n"),
                     size );

         wsi_file_seek( file, size - 1, WSS_FILE_SEEK_SET );
         wsi_file_write( file, bb, 1 );
      }

      if( mapmode & WSS_FILE_MAP_WRITE )
         prot |= PROT_WRITE;
      if( mapmode & WSS_FILE_MAP_READ )
         prot |= PROT_READ;

      prot |= PROT_USER;

      error = VOP_MAP( ((struct wsi_file *)handle)->f_vp, pos, &kas,
                       (caddr_t *)&(file->f_mapAddr), size, prot,
                       PROT_READ | PROT_WRITE | PROT_USER,
                       MAP_SHARED, ((struct wsi_file *)handle)->f_cr );

      if( error == 0 )
         file->f_mapSize = size;
   }

   return error;
}

/*  wsi_file_unmap */
/*! \brief Unmap a memory-mapped file
 *
 *  \param handle Handle to mapped file
 */
void
wsi_file_unmap( wss_file_t handle )
{
   struct wsi_file *file = (struct wsi_file *)handle;

   if( file && ( file->f_mapAddr != NULL ) )
   {
      as_unmap( &kas, file->f_mapAddr, file->f_mapSize );

      file->f_mapAddr = NULL;
      file->f_mapSize = 0;
   }
}

