/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_fileutil.c,v 1.11 2001/02/23 19:02:42 stevek Exp $
 *
 * Author: Douglas Kilpatrick
 *
 * Purpose: Contain functions for network and files
 *
 */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include <sys/pathname.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <sys/user.h>

#include <wsi_fileutil.h>
#include <wsi_path.h>

/* Takes a vnode and a pointer to a wr_stat structure.  Fills in the 
 * wrstat structure.
 */
static void vnstat(struct vnode *vn, struct wr_stat *target)
{
    struct vattr va;
    int flags = AT_ALL;
    int rc;

    rc = VOP_GETATTR(vn,&va,flags,kcred);
    /* What is "rc" ?? */
    
    /* Now copy the stuff in va into target */
    
    target->st_dev              = va.va_rdev;
    target->st_ino              = va.va_nodeid;
    target->st_mode             = va.va_mode;
    target->st_nlink            = va.va_nlink;
    target->st_uid              = va.va_uid;
    target->st_gid              = va.va_gid;
    target->st_rdev             = va.va_fsid;
    target->st_atimespec.tv_sec = va.va_atime.tv_sec;
    target->st_atimespec.tv_nsec= va.va_atime.tv_nsec;
    target->st_mtimespec.tv_sec = va.va_mtime.tv_sec;
    target->st_mtimespec.tv_nsec= va.va_mtime.tv_nsec;
    target->st_ctimespec.tv_sec = va.va_ctime.tv_sec;
    target->st_ctimespec.tv_nsec= va.va_ctime.tv_nsec;
    target->st_size             = va.va_size;
    target->st_gen              = va.va_vcode; /* questionable transform */
}

/*
 * Stats a file.  Follows any symlinks.
 * Fills in the structure passed in with the relevant data.  Does not
 * allocate anything.
 */
int wsi_fileutil_stat(const char *path, struct wr_stat *wrsb)
{
    struct vnode *vn;
    
    vn = wsi_path_to_vnode(path, UIO_SYSSPACE);
    
    if (vn == NULL)
        return -1;
    
    vnstat(vn,wrsb);
    return 0;
}
 
/*
 * Stats a file.  Does not follow any symlinks.
 * Fills in the structure passed in with the relevant data.  Does not
 * allocate anything.
 * 
 * The symlink part may be busted.  All I really did was remove the
 * 'FOLLOW' flag to the lookupname call.  That may prevent symlink->directory
 * stuff from working or something.
 */
int wsi_fileutil_lstat(const char *path, struct wr_stat *wrsb)
{
    struct vnode *dirvp;
    struct vnode *filevp;
    int           err;

    err = lookupname( (char *)path, UIO_SYSSPACE, 0, &dirvp, &filevp );
    if (!filevp)
        return err;

    vnstat(filevp,wrsb);
    
    return 0;
}

/*
 * Takes a file descriptor number, returns the vnode of that fd from
 * the current process.
 * 
 * Returns NULL on error.
 * 
 * Should deal with locking correctly.  Not 100% sure about that.
 */
static struct vnode *fd_to_vnode(unsigned int fd)
{
    struct vnode *vn = NULL;
   
    mutex_enter(&(u.u_flock));
    if (fd <= u.u_nofiles) {
        vn = u.u_flist[fd].uf_ofile->f_vnode;
    }
    mutex_exit(&(u.u_flock));
    
    return vn;
}

/*
 * Stats a file by file-descriptor number.
 * Probally follows symlinks.  (If the "open()" does the right thing)
 * Fills in the structure passed in with the relevant data.  Does not
 * allocate anything.
 */
int wsi_fileutil_fstat(unsigned int fd, struct wr_stat *wrsb)
{
    struct vnode *vn = fd_to_vnode(fd);
   
    if (!vn)
        return -1;

    vnstat(vn,wrsb);
    return 0;
}

/*
 * Returns the "type" of the file being opened.
 * Is a busted function since it was written on FreeBSD and just
 * supported FreeBSD concepts of file types.
 * 
 * IMHO, a group of "isfoo()" functions is a better approach.
 * wsi_fileutil_ispipe() would be true for both a PIPE and a FIFO
 * under BSD and true for a FIFO under solaris.  That prevents the
 * exact constant's used from being important.
 * TBD:
 */
int wsi_fileutil_fdtype(unsigned int fd)
{
    struct vnode *vn = fd_to_vnode(fd);
    
    if (!vn)
        return -1;
    
    
    switch (vn->v_type) {
    case VFIFO:
        return WR_DTYPE_FIFO;
    case VSOCK:
        return WR_DTYPE_SOCKET;
    case VREG:
    case VDIR:
        return WR_DTYPE_VNODE;
        /* Other types:
         * VNON
         * VBLK
         * VCHR
         * VLNK
         * VDOOR
         * VPROC
         * VBAD
         */
        /* Other DTYPE values:
         * WR_DTYPE_PIPE
         * 
         * I'm not sure what to map those too...
         */
    default:
        return WR_DTYPE_OTHER;
    } 
}

/*
 * returns -1 on error.  That can't be a true value since the v_flag
 * is a short.
 * 
 * vnode flags.
 * VROOT           0x01    root of its file system 
 * VNOCACHE        0x02    don't keep cache pages on vnode
 * VNOMAP          0x04    file cannot be mapped/faulted
 * VDUP            0x08    file should be dup'ed rather then opened
 * VNOSWAP         0x10    file cannot be used as virtual swap device
 * VNOMOUNT        0x20    file cannot be covered by mount
 * VISSWAP         0x40    vnode is being used for swap
 * VSWAPLIKE       0x80    vnode acts like swap (but may not be)
 */
int wsi_fileutil_fdfflag(unsigned int fd)
{
    struct vnode *vn = fd_to_vnode(fd);
    if (!vn)
        return -1;
    return vn->v_flag;
}



/*
 * Netutil.
 * Network Related Stuff
 * 
 * TBD: Move to a different file.
 */

static const char *af_name[AF_MAX + 1] = {
        "AF_UNSPEC", 
        "AF_UNIX",
        "AF_INET", 
        "AF_IMPLINK", 
        "AF_PUP", 
        "AF_CHAOS",  
        "AF_NS", 
        "AF_NBS",
        "AF_ECMA", 
        "AF_DATAKIT", 
        "AF_CCITT", 
        "AF_SNA", 
        "AF_DECnet", 
        "AF_DLI", 
        "AF_LAT", 
        "AF_HYLINK", 
        "AF_APPLETALK", 
        "AF_NIT",
        "AF_802",
        "AF_OSI",
        "AF_X25",
        "AF_OSINET",
        "AF_GOSIP",
        "AF_IPX",
        "AF_ROUTE",
        "AF_LINK"
};

static const char *af_unknown = "AF_UNKNOWN";
static const char *msg_noaddr = "No address";
static const char *msg_fail = "Fail";

/* Buffer must be at least 16 in length */
static void wr_netitoa(long inaddr, char *netbuf)
{
     unsigned char *ucp = (unsigned char *) &inaddr;
     sprintf(netbuf, "%d.%d.%d.%d",
                        ucp[0] & 0xff,
                        ucp[1] & 0xff,
                        ucp[2] & 0xff,
                        ucp[3] & 0xff);
}

const char *wsi_netutil_get_family(const struct sockaddr *saddr, int len)
{
     if (saddr != 0 && len >= 16 && saddr->sa_family < AF_MAX) {
          return af_name[saddr->sa_family];
     } else {
          return af_unknown;
     }
}

int wsi_netutil_get_addr(
        const struct sockaddr *saddr,
        int len,
        char *buffer,
        int buffer_length)
{
    int data_len;
    char netbuf[16];
    if (saddr != 0 && len >= 16 && saddr->sa_family < AF_MAX) 
    {
        switch (saddr->sa_family) {
        case AF_UNIX:
            data_len = strlen((char*) saddr->sa_data) + 1;
            if (!buffer || buffer_length < data_len)
                return -1;
            strcpy(buffer,saddr->sa_data);
            break;
        case AF_INET:
            wr_netitoa(((struct sockaddr_in*)saddr)->sin_addr.s_addr,netbuf);
            data_len = strlen(netbuf) /* "255.255.255.255" */
                + 2 /* "::" */ + 5 /* "32768" */ + 1 /* '\0' */;
            if (!buffer || buffer_length < data_len)
                return -1;
            sprintf(buffer, "%s::%d",
                    netbuf,
                    ntohs(((struct sockaddr_in*)saddr)->sin_port));
            break;
        default:
            data_len = strlen(msg_fail) + 1;
            if (!buffer || buffer_length < data_len)
                return -1;
            strcpy(buffer,msg_fail);
            break;
        }
    } else {
        data_len = strlen(msg_fail) + 1;
        if (!buffer || buffer_length < data_len)
            return -1;
        strcpy(buffer,msg_noaddr);
    }
    return 0;
}

int wsi_netutil_getsockname(int fd, struct sockaddr *s, int *len)
{
    return 0;
}

int wsi_netutil_getpeername(int fd, struct sockaddr *s, int *len)
{
    return 0;
}

#if 0
int wsi_getsockname(int fd, struct sockaddr *s, int *len)

{
        struct file *fp;
        register struct socket *so;
        struct mbuf *m; 
        int error;

        error = getsock(curproc->p_fd, fd, &fp); 
        if (error)  
                return (error);
        so = (struct socket *)fp->f_data;
        m = m_getclr(M_WAIT, MT_SONAME);
        if (m == NULL) 
                return (ENOBUFS);
        error = (*so->so_proto->pr_usrreqs->pru_sockaddr)(so, m);
        if (!error) {
                *len = m->m_len;
                bcopy(mtod(m, caddr_t), (caddr_t)s, (u_int)*len); 
        }
        m_freem(m);
        return error;
};

int wsi_getpeername(int fd, struct sockaddr *s, int *len)

{
        struct file *fp;
        register struct socket *so;
        struct mbuf *m; 
        int error;

        error = getsock(curproc->p_fd, fd, &fp);
        if (error)     
                return (error);
        so = (struct socket *)fp->f_data;
        if ((so->so_state & (SS_ISCONNECTED|SS_ISCONFIRMING)) == 0)
                return (ENOTCONN);
        m = m_getclr(M_WAIT, MT_SONAME); 
        if (m == NULL)
                return (ENOBUFS);
        error = (*so->so_proto->pr_usrreqs->pru_peeraddr)(so, m);
        if (!error) {
                *len = m->m_len;
                bcopy(mtod(m, caddr_t), (caddr_t)s, (u_int)*len); 
        }
        m_freem(m);
        return (error);
}

#endif

in_port_t wsi_netutil_ntohs (in_port_t i)

{
  return ntohs (i);
}
