/*! \file wsi_hook.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_hook.c,v 1.13 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Implementation of the Solaris kernel hot-patch hook routine
 * \author  Steve Kiernan
 * \version $Revision: 1.13 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <strings.h>
#include <sys/types.h>
#include <sys/bootconf.h>
#include <sys/kobj.h>
#include <sys/kobj_impl.h>

#include <wss_malloc.h>
#include <wss_process.h>
#include <wsi_hook.h>


/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/

extern int hot_patch_kernel_text( caddr_t addr, uint32_t val, int bytes );


/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

/*! \brief hook structure */
typedef struct _wsi_hook
{
   const pd_char_t *fname; /*!< Name of function to hook                   */
   caddr_t           addr;  /*!< Address of call to function                */
   uint32_t          value; /*!< Old value stored at location (for restore) */
   struct _wsi_hook *next;  /*!< Pointer to next hook structure in list     */
}
wsi_hook_t;


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static wsi_hook_t *_wsi_hook_hooks = NULL;


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

/*! \brief Mask for determining CALL instruction */
#define CALL_MASK 0x3fffffff
/*! \brief Opcode for CALL instruction on UltraSPARC processors */
#define CALL_OP   0x40000000


/***************************************************************************\
 * Public functions
\***************************************************************************/

/*  wsi_hook_add */
/*! \brief Add hot-patch kernel hooks based on symbol pair arguments
 *
 *  \param fname Pointer to name of function to add hooks
 *  \param syms  Array of symbol pairs
 *
 *  \return 0 on success,
 *  \return -1 on failure to find symbol(s)
 *
 *  \sa KERN_HOOK_BEGIN, KERN_HOOK, KERN_HOOK_SAVE, KERN_HOOK_END
 */
int
wsi_hook_add( const pd_char_t *fname, sym_pair_t *syms )
{
   Elf32_Sym func_sym, hook_sym;
   Elf32_Addr addr, end;
   int i;

   /*
    * Search the kernel for necessary symbols
    */
   bzero( &func_sym, sizeof(Elf32_Sym) );
   func_sym.st_value = kobj_getelfsym( (pd_char_t *)fname, NULL,
                                       (int *)&func_sym.st_size );
   if( func_sym.st_value == 0 )
   {
      wsi_printf( PD_MAKESTR("WSI Error: Unable to determine symbol '%s' for hooks\n"),
                  fname );
      return -1;
   }
#ifdef DEBUG_HOOK
   wsi_printf( PD_MAKESTR("%s = 0x%x\n"), fname, (unsigned int)func_sym.st_value );
#endif /* DEBUG_HOOK */

   for( i = 0; syms[i].callname != NULL; i++ )
   {
      bzero( &hook_sym, sizeof(Elf32_Sym) );
      hook_sym.st_value = kobj_getelfsym( (pd_char_t *)syms[i].callname, NULL,
                                          (int *)&hook_sym.st_size );
      if( hook_sym.st_value == 0 )
      {
         wsi_printf( PD_MAKESTR("WSI Error: Unable to determine symbol '%s' for hooks\n"),
                     syms[i].callname );
         return -1;
      }
      syms[i].orig = (caddr_t)hook_sym.st_value;
#ifdef DEBUG_HOOK
      wsi_printf( PD_MAKESTR("%s = 0x%x\n"), syms[i].callname, (uint32_t)syms[i].orig );
#endif /* DEBUG_HOOK */
   }

   addr = func_sym.st_value;
   end = addr + func_sym.st_size;

   /*
    * Search target function for entries we need to replace
    */
   while( addr < end )
   {
      Elf32_Word value = *((Elf32_Word *)addr);

      if( ( value & ~CALL_MASK ) == CALL_OP )
      {
         Elf32_Addr calladdr = (Elf32_Addr)( ( value & CALL_MASK ) << 2 ) +
                               addr;

         for( i = 0; syms[i].callname != NULL; i++ )
         {
            if( calladdr == (Elf32_Addr)syms[i].orig )
            {
               wsi_hook_t *hook = wss_malloc( sizeof( *hook ) );
               uint32_t offset = (uint32_t)(syms[i].replace - addr);
               
               /* Hot-patch the kernel with our replacement */
               hot_patch_kernel_text( (caddr_t)addr,
                                      (uint32_t)( CALL_OP | ( offset >> 2 ) ), 4 );
#ifdef DEBUG_HOOK
               wsi_printf( PD_MAKESTR("%s call @ 0x%x = 0x%x\n"), syms[i].callname,
                           (uint32_t)addr, (uint32_t)syms[i].replace );
#endif /* DEBUG_HOOK */
               if( syms[i].save != NULL )
                  *syms[i].save = (caddr_t)addr;
               hook->fname = fname;
               hook->addr = (caddr_t)addr;
               hook->value = value;
               hook->next = _wsi_hook_hooks;
               _wsi_hook_hooks = hook;
               syms[i].instances++;
               break;
            }
         }
      }
      addr += sizeof(addr);
   }

   return 0;
}


/*  wsi_hook_remove */
/*! \brief Remove hot-patch kernel hook and restore as original
 *
 *  \param fname Pointer to name of function to remove hooks from
 *
 *  \return Number of hooks removed
 */
int
wsi_hook_remove( const pd_char_t *fname )
{
   wsi_hook_t *hook, *next, **prev;
   int count = 0;

   for( hook = _wsi_hook_hooks, prev = &_wsi_hook_hooks; hook;
        prev = &hook->next, hook = next )
   {
      next = hook->next;

      if( !wsi_strcmp( fname, hook->fname ) )
      {
         hot_patch_kernel_text( hook->addr, hook->value, 4 );
         wss_free( hook );

         *prev = next;
         count++;
      }
   }

   return count;
}

/*  wsi_hook_init */
/*! \brief Initialize hot-patch kernel hook code
 *
 *  Place holder for an initialization function.  It may never be needed...
 *
 *  \return 0 always
 */
int
wsi_hook_init()
{
   /* Currently unused */
   return 0;
}

/*  wsi_hook_shutdown */
/*! \brief Shutdown hot-patch kernel hook code and restore original entries
 *
 *  \return 0 always
 */
int
wsi_hook_shutdown()
{
   wsi_hook_t *next;

   while( _wsi_hook_hooks )
   {
      next = _wsi_hook_hooks->next;

      hot_patch_kernel_text( _wsi_hook_hooks->addr, _wsi_hook_hooks->value, 4 );
      wss_free( _wsi_hook_hooks );

      _wsi_hook_hooks = next;
   }
   return 0;
}

