/*! \file wsi_malloc.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_malloc.c,v 1.15 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Platform specific part of the malloc routines
 * \author  Douglas Kilpatrick, Steve Kiernan
 * \version $Revision: 1.15 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/types.h>
#include <sys/kmem.h>

#include <sys/ddi.h>
#include <sys/pathname.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/proc.h>
#include <vm/as.h>


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

void *wsi_proc_malloc __P((struct proc *p, int size));
void  wsi_proc_free   __P((struct proc *p, void *addr, int size));


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_malloc */
/*! \brief Allocate memory of at least specified size
 *
 *  \param size Number of bytes to allocate
 *
 *  \return Pointer to allocated memory on success,
 *  \return NULL on error
 */
void *
wsi_malloc( size_t size )
{
   pd_byte_t *p = kmem_alloc( size + sizeof(double), KM_NOSLEEP );

   /* Store size of block for free() */
   *((size_t *)p) = size;
   p += sizeof(double);

   /* Return block */
   return p;
}

/*  wsi_free */
/*! \brief Free allocated memory
 *
 *  \param ptr Pointer to allocated memory block
 */
void
wsi_free( void *ptr )
{
   pd_byte_t *p = ptr;

   /* Adjust pointer by size of block data */
   p -= sizeof(double);

   /* Free memory */
   kmem_free( p, *((size_t *)p) + sizeof(double) );
}


/*  wsi_curproc_malloc */
/*! \brief Allocate a chunk of memory in the vmspace of the current process
 *
 *  \param size Number of bytes to allocate
 *
 *  \warning In the future this should probably be redesigned.  Eventually,
 *           we will probably want to allocate a single piece of memory and
 *           make allocations from there.  This will allow us to lock down
 *           those pages so they may not be modified by the process from
 *           user-space. Otherwise, there may be succeptibility in
 *           multi-threaded systems.  We are using this memory primarily to
 *           hold replaced syscall parameters.  We do not want the user-space
 *           program to be able to modify anything we store here.
 *
 *  \return Pointer to allocated memory on success,
 *  \return NULL on error
 */
void *
wsi_curproc_malloc( size_t size )
{
   vnode_t *file = NULL;
   vnode_t *vndir;
   int error;
   caddr_t addr;

   /* Round size up to nearest page. */
   size = ptob( btopr( size ) );
    
   /* Open /dev/zero.  mmap it. */
   lookupname( "/dev/zero", UIO_SYSSPACE, FOLLOW, &vndir, &file );
   if( file == NULL )
      return NULL;

   error = VOP_MAP( file, 0, curproc->p_as, &addr, size, PROT_ALL, PROT_ALL,
                    MAP_PRIVATE, kcred );
   if( error )
      return NULL;

   return addr;
}

/*  wsi_curproc_free */
/*! \brief Frees the address from the current process's vmspace.
 *
 *  \param addr Pointer to allocated memory block
 *  \param size Size of block in bytes
 */
void
wsi_curproc_free( void *addr, size_t size )
{
   if( addr )
   {
      size = ptob( btopr( size ) );
      as_unmap( curproc->p_as, (caddr_t)addr, size );
   }
}

/*  wsi_curproc_stdup */
/*! \brief Similar to strdup, but allocates memory from the vmspace of the
 *         current process
 *
 *  \param str Pointer to string to duplicate
 *
 *  \return Pointer to duplicated string within the process' vmspace on success,
 *  \return NULL on failure
 */
pd_char_t *
wsi_curproc_strdup( const pd_char_t *str )
{
   pd_char_t *buf;
   buf = (pd_char_t *)wsi_curproc_malloc( wsi_strlen( str ) + 1 );
   if( buf )
   {
      wsi_strcpy( buf, str );
   }
   return( buf );
}

