/*! \file wsi_path.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_path.c,v 1.30 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Functions to determine the paths in system-dependent fashion
 * \author  Linda Thomas
 * \version $Revision: 1.30 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <dirent.h>
#include <limits.h>

#include <sys/cred.h>
#include <sys/param.h>
#include <sys/pathname.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/user.h>
#include <sys/vfs.h>
#include <sys/vnode.h>

#include <wss_malloc.h>
#include <wsi_path.h>


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#define READSIZE MAXBSIZE


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * Purpose:  Read a block of data from a directory.
 *
 * Returns:  NULL on success, otherwise a VOP error.
 */

static int
_wsi_path_read_dir_block( struct vnode *vp, pd_byte_t *dir_buf, 
                         u_offset_t *offset, u_offset_t *length )
{
   struct uio auio;
   struct iovec aiov;
   int error;
   int eof = 0;

   aiov.iov_base = dir_buf;
   aiov.iov_len = READSIZE;
  
   bzero( &auio, sizeof auio );
   auio.uio_iov     = &aiov;
   auio.uio_iovcnt  = 1;
   auio.uio_fmode   = UIO_READ;
   auio.uio_segflg  = UIO_SYSSPACE;
   auio.uio_loffset = *offset;
   auio.uio_llimit  = *offset + READSIZE;
   auio.uio_resid   = READSIZE;

#ifdef DEBUG_PATH2
   wsi_printf( PD_MAKESTR("before offset = %lu, limit = %lu, resid = %lu\n"),
           (unsigned long) auio.uio_loffset, (unsigned long) auio.uio_llimit, 
           (unsigned long) auio.uio_resid);
#endif /* DEBUG_PATH2 */

   VOP_RWLOCK( vp, RW_READER );
   error = VOP_READDIR( vp, &auio, kcred, &eof );
   VOP_RWUNLOCK( vp, RW_READER );
   if( error != 0 )
   {
      wsi_printf( PD_MAKESTR("VOP_READDIR failed %d\n"), error );
   }
#ifdef DEBUG_PATH2
   else
   {
      wsi_printf( PD_MAKESTR("afteroffset = %lu, limit = %lu, resid = %lu, eof = %d\n"),
              (unsigned long) auio.uio_loffset, 
              (unsigned long) auio.uio_llimit, 
              (unsigned long) auio.uio_resid, eof );
   }
#endif /* DEBUG_PATH2 */

#ifdef DEBUG_PATH2
   {
      int i;
      for( i = 0; i < auio.uio_offset; i++ )
      {
         wsi_printf (PD_MAKESTR("%2x "), dir_buf[i]);
         if( ( dir_buf[i] >= 32 ) && ( dir_buf[i] <= 127 ) )
            wsi_printf( PD_MAKESTR("%c "), dir_buf[i] );
         else
            wsi_printf( PD_MAKESTR("  ") );
         if( ( i % 15 ) == 14 )
            wsi_printf( PD_MAKESTR("\n") );
      }
      wsi_printf( PD_MAKESTR("\n**********************\n") );
   }
#endif /* DEBUG_PATH2 */

   *offset = auio.uio_loffset;
   *length = READSIZE - auio.uio_resid;

   return eof ? -1: error;
}


static pd_char_t *
_wsi_path_get_node_name( struct vnode *vp, struct vnode *parent )
{
   u_offset_t offset;
   u_offset_t length;
   pd_char_t *dir_buf;
   int error;
   struct dirent64 *pos;
   struct dirent64 *oldpos;
   pd_char_t *s = NULL;
   int i;
   struct vnode *tempvp;
   int lookuperror = FALSE;
   int found       = FALSE;

   /* First, fail if this isn't a directory. */
   if( parent->v_type != VDIR )
   {
      return( NULL );
   }

   dir_buf = (pd_char_t *) wss_malloc( READSIZE );
   bzero( dir_buf, READSIZE );
   offset = 0;

   do
   {
      pos = (struct dirent64 *)dir_buf;
      oldpos = NULL;

      error = _wsi_path_read_dir_block( parent, dir_buf, &offset, &length );
      if( error > 0 )
      {
         wsi_printf( PD_MAKESTR("error is %d\n"), error );
         break;
      }

      /*
       * Loop through the block buffer and put each name (with a
       * newly malloc'd string) in the passed list.
       */
      while( !found &&
             ( (unsigned long)pos < (unsigned long)dir_buf + length ) )
      {
         i = VOP_LOOKUP( parent, pos->d_name, &tempvp, NULL, 0, NULL, kcred );
         if( i != 0 )
         {
            if( !lookuperror )
            {
               wsi_printf( PD_MAKESTR("last good was '%s', %lu\n"), oldpos->d_name,
                       (unsigned long) oldpos->d_reclen );
               wsi_printf( PD_MAKESTR("vop lookup '%s' failed i = %d\n"), pos->d_name, i );
               lookuperror = TRUE;
            }
         }
         else if( vp == tempvp )
         {
            /* found the match! */
            /*wsi_printf( PD_MAKESTR("found '%s'\n"), pos->d_name ); */
            s = wss_malloc( wsi_strlen( pos->d_name ) + 1 );
            wsi_strcpy( s, pos->d_name );
            found = TRUE;
         }
         else if( tempvp->v_vfsmountedhere != NULL )
         {
            i = traverse (&tempvp);
            if( vp == tempvp )
            {
               /*wsi_printf( PD_MAKESTR("found '%s' by traversing vnode\n"), pos->d_name );*/
               s = wss_malloc( wsi_strlen( pos->d_name ) + 1 );
               wsi_strcpy( s, pos->d_name );
               found = TRUE;
            }
         }

         oldpos = pos;
         if( pos->d_reclen == 0 )
            pos = (struct dirent64 *)( ((pd_byte_t *)pos) + sizeof(struct dirent64) );
         else
            pos = (struct dirent64 *)( ((pd_byte_t *)pos) + pos->d_reclen );
      }
   }
   while( error != -1 && !found );

   wss_free( dir_buf );
   return s;
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wsi_path_is_dir */
/*! \brief Determine if path is a directory
 *
 *  \param path Pointer to string containing the path
 *  \param uio  Enumeration stating where the path string resides in memory
 *              (\b UIO_SYSSPACE or \b UIO_USERSPACE)
 *
 *  \return TRUE if the path is a directory,
 *  \return NULL otherwise
 */
int
wsi_path_is_dir( pd_char_t *path, uio_seg_t uio )
{
   struct vnode *vp;
   int           ret = 0;

   vp = wsi_path_to_vnode( path, uio );
   
   if(!vp)
       return 0;

   if( vp->v_type == VDIR )
      ret = 1;

   return ret;
}


/*  wsi_path_vnode_to_path */
/*! \brief Determine path for given Vnode
 *
 *  \param vp Pointer to the Vnode
 *
 *  \return Pointer to string containing the path on success,
 *  \return NULL on failure
 */
pd_char_t *
wsi_path_vnode_to_path( struct vnode *vp )
{
   pd_char_t *name;
   int idx = 0;
   int length = 2;
   struct vnode *parentvp = NULL;
   pd_char_t *fnames[500];
   int i;

   bzero( fnames, 500 * sizeof (pd_char_t *) );

   while( vp != NULL )
   {
      if( ( vp != NULL ) && ( vp->v_flag & VROOT ) )
      {
         vp = vp->v_vfsp->vfs_vnodecovered;
         continue;
      }

      if( vp != NULL )
      {
         i = VOP_LOOKUP( vp, PD_MAKESTR(".."), &parentvp, NULL, LOOKUP_DIR, NULL, kcred );
         if( i != 0 )
         {
            /*wsi_printf( PD_MAKESTR("vop lookup .. failed i = %d\n"), i );*/
            return NULL;
         }
      }

      if( ( vp != NULL ) && ( parentvp != NULL ) )
      {
         name = _wsi_path_get_node_name( vp, parentvp );
         if( name != NULL )
         {
            length += wsi_strlen( name ) + 1;
            if( idx < 500 )
               fnames[idx++] = name;
         }
         vp = parentvp;
      }
   }

   name = wss_malloc( length );
   *name = '\0';

   if( idx != 0 )
   {
      while( --idx >= 0 )
      {
         wsi_strcat( name, PD_MAKESTR("/") );
         wsi_strcat( name, fnames[idx] );
         if( fnames[idx] != NULL )
            wss_free( fnames[idx] );
      }
   }
   else
      wsi_strcpy( name, PD_MAKESTR("/") );

   /*wsi_printf( PD_MAKESTR("name is '%s'\n"), name );*/

   return name;
}


/*  wsi_path_cwd */
/*! \brief Get the current working directory for a process
 *
 *  Simple front-end to the wr_vnode_to_path() routine.
 *
 *  \param wrp Pointer to a wrapper process structure
 *
 *  \return Pointer to string containing the current working directory of
 *          the process
 */
pd_char_t *
wsi_path_cwd( struct wss_proc *wrp )
{
  return wsi_path_vnode_to_path( wrp->wrp_p->p_user.u_cdir );
}


/*  wsi_path_abspath */
/*! \brief Converts the given path to absolute terms
 *
 *  \b Note:
 *      The path parameter passed in is in kernel space.
 *      The return value is in kernel space.
 *      This function is made available to wrappers.
 *
 *  \param path Pointer to string containing the path
 *
 *  \warning Wrappers can currently misuse this
 *
 *  \return Pointer to string containing the absolute path
 */
pd_char_t *
wsi_path_abspath( pd_char_t *path )
{
   pd_char_t *fullpath       = NULL;
   pd_char_t *dirpart        = NULL;
   pd_char_t *filepart       = NULL;
   pd_char_t *p;
   int   err;
   pd_char_t  c              = '\0';
   struct vnode *dirvp  = NULL;
   struct vnode *filevp = NULL;

   if( path == NULL )
      fullpath = NULL;
   else if( ( path[0] == '/' ) && ( path[1] == '\0' ) )
   {
      /* lookupname returns EINVAL on /, so special case it */
      fullpath = wss_malloc( 2 );
      fullpath[0] = '/';
      fullpath[1] = '\0';
   }
   else
   {
      err = lookupname( path, UIO_SYSSPACE, FOLLOW, &dirvp, &filevp );

      if( err == 0 )
      {
         if( filevp != NULL && dirvp != NULL )
         {
            /* we found the whole thing */
            if( filevp->v_type == VDIR )
               fullpath = wsi_path_vnode_to_path( filevp );
            else
            {
               dirpart = wsi_path_vnode_to_path( dirvp );
               filepart = _wsi_path_get_node_name( filevp, dirvp );
               if( ( dirpart != NULL ) && ( filepart != NULL ) )
               {
                  fullpath = wss_malloc( wsi_strlen( dirpart ) +
                                         wsi_strlen( filepart ) + 2 );
                  if( wsi_strcmp( dirpart, PD_MAKESTR("/") ) != 0 )
                     wsi_sprintf( fullpath, PD_MAKESTR("%s/%s"), dirpart, filepart );
                  else
                     wsi_sprintf( fullpath, PD_MAKESTR("/%s"), filepart );
               }
            }
         }
         else if( dirvp != NULL )
         {
            /* the last component was not found */
            dirpart = wsi_path_vnode_to_path( dirvp );
            p = path + wsi_strlen( path );
            while ( p != path && *p != '/' )
               p--;
            if ( *p == '/' )
               p++;
            fullpath = wss_malloc( wsi_strlen( dirpart ) + wsi_strlen( p ) + 2 );
            if( wsi_strcmp( dirpart, PD_MAKESTR("/") ) != 0 )
               wsi_sprintf( fullpath, PD_MAKESTR("%s/%s"), dirpart, p );
            else
               wsi_sprintf( fullpath, PD_MAKESTR("/%s"), p );
         }
         else
         {
            /* this should never happen */
            panic( PD_MAKESTR("wsi_path_abspath: both file and dir parts were null!\n") );
         }
      }
      else if( err == ENOENT )
      {
         /* some component before the leaf was not found */
         /* try to find out how much was valid */
         p = path + wsi_strlen( path );
         do
         {
            while( ( p != path ) && ( *p != '/' ) )
               p--;
            if( p != path )
            {
               c = *p;
               *p = '\0';
            }
            err = lookupname( path, UIO_SYSSPACE, FOLLOW, &dirvp, &filevp );
            if( p != path )
               *p-- = c;
         }
         while( err != 0 && p != path );
   
         if( err == 0 )
         {
            while( p != path && *p != '/' )
               p--;
            if ( *p == '/' )
               p++;
            dirpart = wsi_path_vnode_to_path( dirvp );
            fullpath = wss_malloc( wsi_strlen( dirpart ) + wsi_strlen( p ) + 2 );
            if( wsi_strcmp ( dirpart, PD_MAKESTR("/") ) != 0 )
               wsi_sprintf( fullpath, PD_MAKESTR("%s/%s"), dirpart, p );
            else
               wsi_sprintf( fullpath, PD_MAKESTR("/%s"), p );
         }
      }
      else
      {
#ifdef DEBUG_PATH
         /* something unforseen went wrong */
         wsi_printf( PD_MAKESTR("wsi_path_abspath: lookup of %s, error was %d\n"), path, err );
#endif /* DEBUG_PATH */
         fullpath = NULL;
      }
   }

   if( dirpart != NULL )
      wss_free( dirpart );

   if( filepart != NULL )
      wss_free( filepart );

   return fullpath;
}


/*  wsi_path_to_vnode */
/*! \brief Get the Vnode for the pathname
 *
 *  \param path Pointer to string containing the path
 *
 *  \return Pointer to Vnode related to the path on success,
 *  \return NULL if the path does not exist
 */
struct vnode *
wsi_path_to_vnode( const pd_char_t *path, uio_seg_t uio )
{
   struct vnode *dirvp  = NULL;
   struct vnode *filevp = NULL;
   int           err;

   err = lookupname( (pd_char_t *)path, uio, FOLLOW, &dirvp, &filevp );

   return filevp;
}

