/*! \file wsi_process.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_process.c,v 1.22 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   System-dependent process management functions
 * \author  Steve Kiernan
 * \version $Revision: 1.22 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/prsystm.h>
#include <sys/dditypes.h>
#include <sys/stream.h>
#include <sys/isdnio.h>
#include <sys/audiovar.h>
#include <sys/ddidmareq.h>
#include <sys/dbrireg.h>
#include <sys/dbrivar.h>
#include <sys/procfs.h>
#include <sys/cpuvar.h>
#include <sys/rwlock.h>
#include <sys/kmem.h>
#include <sys/wait.h>

#include <wss_malloc.h>
#include <wss_process.h>
#include <wss_path.h>

#include <wrerrno.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern kthread_t        *unloadt;
extern const pd_char_t *solaris_syscallnames[];


/***************************************************************************\
 * Public functions
\***************************************************************************/

/*  wsi_proc_new */
/*! \brief Create the system-dependent opaque data part of the wss_proc
 *         structure
 *
 *  \param wrp Pointer to wrapper process structure
 */
void
wsi_proc_new( struct wss_proc *wrp )
{
   wsi_proc_t *wrpi;
   int         i;
   
   wrpi = (wsi_proc_t *)wss_malloc( sizeof(wsi_proc_t) );
   wrp->wrp_opaque = (void *)wrpi;

   /* Initialize the dispatch lists */
   for( i = 0; i < NSYSCALL; i++ )
      LIST_INIT( &(wrpi->dispatchers[i].list) );
}

/*  wsi_proc_free */
/*  \brief Free the system-dependent opaque data part of the wss_proc structure
 *
 *  \param wrp Pointer to wrapper process structure
 */
void
wsi_proc_free( struct wss_proc *wrp )
{
   wsi_proc_t *wrpi;
   int         i;

   wrpi = (wsi_proc_t *)wrp->wrp_opaque;

   if( wrpi )
   {
      /* Clean up the dispatch lists */
      for( i = 0; i < NSYSCALL; i++ )
      {
         wsi_proc_dispatch_t *dispatch;

         while( wrpi->dispatchers[i].list.lh_first )
         {
            dispatch = wrpi->dispatchers[i].list.lh_first;
            LIST_REMOVE( dispatch, next );
            wss_free( dispatch );
         }
      }

      /* Free memory used by wsi_proc structure */
      wsi_free( wrp->wrp_opaque );
      wrp->wrp_opaque = NULL;
   }
}

/*  wsi_proc_insyscall */
/*! \brief Returns the number of LWPs within syscalls for the specified process
 *
 *  \param wrp Pointer to wrapper process structure
 *
 *  \return Number of LWPs within syscalls
 */
int
wsi_proc_insyscall( struct wss_proc *wrp )
{
   int insyscall;

   WSS_MUTEX_ENTER( wrp->wrp );
   insyscall = wrp->wrp_nsyscall;
   WSS_MUTEX_EXIT( wrp->wrp );

   return insyscall;
}

/*  wsi_proc_suspend_lwp */
/*! \brief Suspends the specified LWP
 *
 *  \param t     Pointer to kernel thread
 *  \param rcode Pointer to memory to hold the return code
 *
 *  \warning Doesn't quite do what we want it to do.  I was hoping it would
 *           force LWPs to reissue their syscalls when the LWP is resumed.  It
 *           doesn't do that, though.
 *
 *  \return 0 on success,
 *  \return EINTR if the thread was interrupted,
 *  \return ESRCH if the thread could not be found on the thread list
 */
int
wsi_proc_suspend_lwp( kthread_t *t, int *rcode )
{
   register proc_t *p;

   p = t->t_procp;
   *rcode = 0;
   t->t_proc_flag |= TP_HOLDLWP;

   if( t == curthread )
   {
      t->t_astflag = 1;
      return 0;
   }

   thread_lock( t );
   t->t_astflag = 1;

   if( ( t->t_state == TS_SLEEP ) && ( t->t_flag & T_WAKEABLE ) )
      setrun_locked( t );
   else if( ( t->t_state != TS_ONPROC ) && ( curthread->t_cpu != t->t_cpu ) )
      poke_cpu( t->t_cpu->cpu_id );

   do
   {
      if( ( t->t_state != TS_STOPPED ) &&
          ( (t->t_schedflag & (TS_CSTART|TS_UNPAUSE)) !=
            (TS_CSTART|TS_UNPAUSE) ) )
      {
         *rcode = t->t_sysnum;
         disp_lock_exit( t->t_lockp );
         return 0;
      }

      disp_lock_exit( t->t_lockp );

#if EXIT_ON_EXITLWPS
      if( p->p_flag & EXITLWPS )
         lwp_exit();
#endif /* EXIT_ON_EXITLWPS */

      if( cv_wait_sig( &p->p_holdlwps, &p->p_lock ) == 0 )
         return EINTR;

      if( idtot( p->p_tlist, t->t_tid ) == NULL )
         return ESRCH;

      thread_lock( t );
   }
   while( t->t_proc_flag & TP_HOLDLWP );

   return 0;
}

/*  wsi_proc_suspend */
/*! \brief Suspend the LWPs within the process
 *
 *  \param p Pointer to process structure
 *
 *  \return 0 on success,
 *  \return EINTR if a thread was interrupted,
 *  \return ESRCH if a thread could not be found on the thread list
 */
int
wsi_proc_suspend( struct proc *p )
{
   register kthread_t *t;

   WSS_MUTEX_ENTER_DEBUG( p->p );

   cmn_err( CE_NOTE, "Waiting for %d LWPs to exit from syscalls for proc #%d",
            p->p_lwprcnt, (int)p->p_pid );

   t = p->p_tlist;
   do
   {
      if( ( t->t_lwp != NULL ) && ( t->t_sysnum >= 0 ) )
      {
         register int err;
         int sysnum;

         cmn_err( CE_NOTE, "Thread 0x%x in syscall %s(%d), suspending...",
                  (uint32_t)t, solaris_syscallnames[t->t_sysnum],
                  t->t_sysnum );
         err = wsi_proc_suspend_lwp( t, &sysnum );
         if( err != 0 )
         {
            cmn_err( CE_WARN, "Unable to suspend thread 0x%x, error #%d",
                     (uint32_t)t, err );

            WSS_MUTEX_EXIT_DEBUG( p->p );
            return err;
         }
      }
      t = t->t_forw;
   }
   while( t->t_forw != p->p_tlist );

   cmn_err( CE_NOTE, "Proc #%d has %d LWPs in syscalls after stop",
            (int)p->p_pid, p->p_lwprcnt );

   WSS_MUTEX_EXIT_DEBUG( p->p );

   return 0;
}


/*  wsi_proc_continue_lwp */
/*! \brief Resume LWP execution
 *
 *  \param t Pointer to kernel thread
 */
void
wsi_proc_continue_lwp( kthread_t *t )
{
   register int holdlwp;
   register proc_t *p;

   holdlwp = t->t_proc_flag & TP_HOLDLWP;

   t->t_proc_flag &= ~TP_HOLDLWP;
   thread_lock( t );

   p = t->t_procp;
   if( ( t->t_state == T_LWPZOMB ) &&
       ((t->t_schedflag & (TS_CSTART|TS_UNPAUSE)) != (TS_CSTART|TS_UNPAUSE)) &&
       ( ( p->p_flag & (HOLDWATCH|HOLDFORK1|HOLDFORK) ) == 0 ) )
   {
      p->p_lwprcnt++;
      t->t_schedflag |= TS_CSTART;
      setrun_locked( t );
   }

   disp_lock_exit( t->t_lockp );

   if( holdlwp )
      cv_broadcast( &p->p_holdlwps );
}

/*  wsi_proc_continue */
/*! \brief Resume suspended LWPs in a process
 *
 *  \param p Pointer to process structure
 */
void
wsi_proc_continue( struct proc *p )
{
   register kthread_t *t;

   WSS_MUTEX_ENTER_DEBUG( p->p );

   t = p->p_tlist;
   do
   {
      if( t->t_lwp != NULL )
      {
         cmn_err( CE_NOTE, "Resuming thread 0x%x...", (uint32_t)t );
         wsi_proc_continue_lwp( t );
      }
      t = t->t_forw;
   }
   while( t->t_forw != p->p_tlist );

   WSS_MUTEX_EXIT_DEBUG( p->p );
}

/*  wsi_proc_force_syscall_exit */
/*! \brief Force threads within a process to exit syscalls
 *
 *  This is a placeholder for code that actually does make LWPs reissue
 *  their syscalls.  It's obviously TBD.
 *
 *  \param wrp Pointer to wrapper process structure
 */
void
wsi_proc_force_syscall_exit( struct wss_proc *wrp )
{
}

