/*! \file wsi_utilities.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_utilities.c,v 1.20 2001/02/23 19:02:42 stevek Exp $
 * 
 * \brief   Repository of useful functions which might be missing.
 * \author  Doug Kilpatrick and others.
 * \version $Revision: 1.20 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#ifdef USER_MODE_TEST
#include <stdarg.h>
#endif

#include <wsi_stdlib.h>

#include <sys/types.h>
#include <sys/acl.h>
#include <stdarg.h>
#include <sys/cmn_err.h>
#include <strings.h>


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

#ifndef USER_MODE_TEST
/*  qsort */
/*! \brief Implementation of quick-sort
 *
 *  \param base   Pointer to initial data member
 *  \param bel    Number of elements
 *  \param width  Size of each element
 *  \param compar Comparison function
 */
void
qsort( void *base, size_t bel, size_t width,
       int (*compar)(const void*, const void*) )
{
   ksort( base, bel, width, (int (*)(aclent_t *, aclent_t *))compar );
}
#endif /* !USER_MODE_TEST */

/*  strnlen */
/*! \brief Determine length of a string up to a specified number of characters
 *
 *  \param s Pointer to string
 *  \param n Maximum size of string
 *
 *  \return Number of characters in string
 */
size_t
strnlen( const pd_char_t *s, size_t n )
{
   int i;
   for( i = 0 ; i < n ; i++ )
      if(!(s[i])) /* Non portable hack */
         break;
   return i;
}

/*  wsi_vsprintf */
/*! \brief Implementation of vsprintf
 *
 *  This function is provided for consistency in user-space and kernel-space.
 *  On Solaris, the user-space vsprintf returns the number of characters
 *  written to the buffer.  In the kernel-space implementation, it does not.
 *
 *  \param s      Pointer to destination buffer
 *  \param format Pointer to format string
 *  \param ap     Variable arguments
 *
 *  \return Number of characters written to buffer
 */
int
wsi_vsprintf( pd_char_t *s, const pd_char_t *format, va_list ap)
{
#ifndef USER_MODE_TEST
   const pd_char_t *retval;
   retval = vsprintf( s, format, ap );
   return retval == 0 ? -1 : strlen( s );
#else
   int len;
   len = vsprintf( s, format, ap );
   return len;
#endif
}


/*  snprintf */
/*! \brief Implementation of snprintf
 *
 *  This function is provided for consistency in user-space and kernel-space.
 *
 *  \param s      Pointer to destination buffer
 *  \param n      Maximum number of characters which can be written to buffer
 *  \param format Pointer to format string
 *
 *  \return Number of characters written to buffer
 */
int
snprintf( pd_char_t *s, size_t n, const pd_char_t *format, ... )
{
   va_list ap;
#ifndef USER_MODE_TEST
   const pd_char_t *retval;
#else
   int len;
#endif

   va_start( ap, format );
#ifndef USER_MODE_TEST
   retval = vsprintf_len( n, s, format, ap );
#else
   len = vsnprintf( s, n, format, ap );
#endif
   va_end( ap );

#ifndef USER_MODE_TEST
   return retval == 0 ? -1 : strnlen( s, n );
#else
   return len;
#endif
}


/*  memcpy */
/*! \brief Copy bytes from one memory location to another
 *
 *  \param to    Destination location
 *  \param from  Source location
 *  \param count Number of bytes to copy
 *
 *  \return Pointer to destination location
 */
void *
memcpy( void *to, const void *from, size_t count )
{
    bcopy( from, to, count );
    return to;
}


/*  wsi_strlen */
/*! \brief Implementation of strlen
 *
 *  \param s Pointer to string
 *
 *  \return Number of characters in string
 */

size_t
wsi_strlen( const pd_char_t *s )
{
   return strlen( s );
}


/*  wsi_strcmp */
/*! \brief Implementation of strcmp
 *
 *  \param s1 Pointer to first string
 *  \param s2 Pointer to second string
 *
 *  \return Integer difference of strings
 */          
int
wsi_strcmp( const pd_char_t *s1, const pd_char_t *s2 )
{
   return strcmp( s1, s2 );
}

/*  wsi_strncmp */
/*! \brief Implementation of strncmp
 *
 *  \param s1  Pointer to first string
 *  \param s2  Pointer to second string
 *  \param len Number of characters to compare
 *
 *  \return Integer difference of strings
 */
int
wsi_strncmp( const pd_char_t *s1, const pd_char_t *s2, size_t len )
{
   return strncmp( s1, s2, len );
}


/*  wsi_strcpy */
/*! \brief Implementation of strcpy
 *
 *  \param dst Pointer to destination buffer
 *  \param src Pointer to source string
 *
 *  \return Pointer to destination buffer
 */
pd_char_t *
wsi_strcpy( pd_char_t *dst, const pd_char_t *src )
{
   return strcpy( dst, src );
}


/*  wsi_strncpy */
/*! \brief Implementation of strncpy
 *
 *  \param dst Pointer to destination buffer
 *  \param src Pointer to source string
 *  \param len Maximum number of characters to copy
 *
 *  \return Pointer to destination buffer
 */
pd_char_t *
wsi_strncpy( pd_char_t *dst, const pd_char_t *src, size_t len )
{
   return strncpy( dst, src, len );
}


/*  wsi_strcat */
/*! \brief Implementation of strcat
 *
 *  \param dst Pointer to destination buffer
 *  \param src Pointer to source string
 *
 *  \return Pointer to destination buffer
 */
pd_char_t *
wsi_strcat( pd_char_t *dst, const pd_char_t *src )
{
   return strcat( dst, src );
}


/*  wsi_toupper */
/*! \brief Implementation of toupper
 *
 *  \param c Character to convert to uppercase
 *
 *  \return Character converted to uppercase
 */
pd_char_t EXPORT
wsi_toupper( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('a') ) && ( c <= PD_MAKESTR('z') ) )?
            c - PD_MAKESTR('a') + PD_MAKESTR('A') :
            c;
}


/*  wsi_tolower */
/*! \brief Implementation of tolower
 *
 *  \param c Character to convert to lowercase
 *
 *  \return Character converted to lowercase
 */
pd_char_t EXPORT
wsi_tolower( const pd_char_t c )
{
   return ( ( c >= PD_MAKESTR('A') ) && ( c <= PD_MAKESTR('Z') ) )?
            c - PD_MAKESTR('A') + PD_MAKESTR('a') :
            c;
}

#if !defined(USER_MODE_TEST)
int EXPORT
wsi_util_sleep(long cookie)
{
    /* Not sure about much here.  Retval might be wrong.  I'm assuming
     * 0 gets returned on normal wakeup, like FreeBSD, so the &&1 is just
     * a normalizer.
     * 
     * The second arg I am assuming is a timeout, in which case 0 should
     * mean 'don't timeout'
     */
    return (sleep((void*)cookie, 0) && 1); /* timeout*? */
}

void EXPORT
wsi_util_wakeup(long cookie)
{
    wakeup((void*)cookie);
}
#endif /* USER_MODE_TEST */
