/*! \file wsi_wrapper.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wsi_wrapper.c,v 1.5 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Perform system-dependent actions on wrappers
 * \author  Steve Kiernan
 * \version $Revision: 1.5 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_wrapper.h>
#include <wss_module.h>
#include <wss_mutex.h>


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/


/*  wsi_wrapper_activate */
/*! \brief Handle system-dependent activation
 *
 *  \param wi Pointer to wrapper instance structure
 */
void
wsi_wrapper_activate( wss_proc_t *wrp, wss_wrapper_instance_t *wi )
{
   wsi_proc_t *wrpi;
   int         i;
   int        *intercepts;

   wrpi = (wsi_proc_t *)wrp->wrp_opaque;
   intercepts = (int *)wi->wi_wrapper->w_module->wrm_opaque;

   for( i = 0; intercepts[i]; i++ )
   {
      wsi_proc_dispatch_t *dispatch;
      wsi_proc_dispatch_t *last     = NULL;

      if( ( intercepts[i] < 1 ) || ( intercepts[i] >= NSYSCALL ) )
         break;

      for( dispatch = wrpi->dispatchers[intercepts[i]].list.lh_first;
           dispatch;
           last = dispatch, dispatch = dispatch->next.le_next )
      {
         if( dispatch->instance == wi )
         {
            dispatch->refcount++;
            break;
         }
      }

      if( dispatch == NULL )
      {
         dispatch = (wsi_proc_dispatch_t *)wss_malloc( sizeof(*dispatch) );
         wss_wrapper_instance_ref( wi );
         dispatch->instance = wi;
         dispatch->refcount = 1;

         if( last )
         {
            LIST_INSERT_AFTER( last, dispatch, next );
         }
         else
         {
            LIST_INSERT_HEAD( &(wrpi->dispatchers[intercepts[i]].list),
                              dispatch, next );
         }
      }
   }
}

/*  wsi_wrapper_deactivate */
/*! \brief Handle system-dependent deactivation
 *
 *  \param wi Pointer to wrapper instance structure
 */
void
wsi_wrapper_deactivate( wss_proc_t *wrp, wss_wrapper_instance_t *wi )
{
   wsi_proc_t *wrpi;
   int         i;
   int        *intercepts;

   wrpi = (wsi_proc_t *)wrp->wrp_opaque;
   intercepts = (int *)wi->wi_wrapper->w_module->wrm_opaque;

   for( i = 0; intercepts[i]; i++ )
   {
      wsi_proc_dispatch_t *dispatch;
      wsi_proc_dispatch_t *next;

      if( ( intercepts[i] < 1 ) || ( intercepts[i] >= NSYSCALL ) )
         break;

      for( dispatch = wrpi->dispatchers[intercepts[i]].list.lh_first;
           dispatch;
           dispatch = next )
      {
         next = dispatch->next.le_next;
         if( dispatch->instance == wi )
         {
            if( --dispatch->refcount == 0 )
            {
               LIST_REMOVE( dispatch, next );
               wss_wrapper_instance_rele( dispatch->instance );
               wss_free( dispatch );
            }

            break;
         }
      }
   }
}

/*  wsi_wrapper_duplicate */
/*! \brief Handle system-dependent duplication
 *
 *  \param wi Pointer to wrapper instance structure
 */
void
wsi_wrapper_duplicate( wss_proc_t *wrp, wss_wrapper_instance_t *wi )
{
   wsi_wrapper_activate( wrp, wi );
}

