/*! \file wsi_wrctl.c
 *
 * Copyright (c) 1999, 2000, 2001 Network Associates, Incorporated
 * All rights reserved.
 *    
 * Redistribution and use are governed by the terms detailed in the
 * license document ("LICENSE") included with the toolkit.
 *
 * $Id: wsi_wrctl.c,v 1.42 2001/02/23 19:02:42 stevek Exp $
 *
 * \brief   Implementation of the wrctl system call
 * \author  Steve Kiernan
 * \version $Revision: 1.42 $
 * \date    $Date: 2001/02/23 19:02:42 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <sys/systm.h>

#include <wss_malloc.h>
#include <wss_ref.h>
#include <wss_criteria.h>
#include <wss_linker.h>
#include <wss_module.h>
#include <wss_wrvevent.h>
#include <wsi_time.h>
#include <wsi_syscalls.h>


/***************************************************************************\
 * External globals                                                        * 
\***************************************************************************/

extern wss_linker_t      *wss_linker_list;
WSS_MUTEX_EXPORT         (wss_linker_list);
extern wss_linker_file_t  wss_linker_file_list;
WSS_MUTEX_EXPORT         (wss_linker_file_list);


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wrctl */
/*! \brief wrctl syscall
 *
 *  This syscall handles loading, unloading, and enabling/disabling wrapper
 *  modules.
 *
 *  \param command Identifier of command.
 *                    Can be one of the following:
 *                       \b WRCTL_LOAD,
 *                       \b WRCTL_UNLOAD,
 *                       \b WRCTL_ENABLED
 *  \param id      Identifier of module
 *  \param buffer  Pointer to buffer data (generic argument)
 *
 *  \return Return value on success,
 *  \return EPERM if \b SUSER_ONLY is defined and user is not root,
 *  \return ENOENT if the module is not found or a symbol cannot be resolved
 *                 in the ELF-32 module,
 *  \return EINVAL if an invalid parameter is encountered,
 *  \return EAGAIN if there is not enough memory for an allocation,
 *  \return ENOEXEC if the specified module is an unknown format,
 *  \return EEXIST if the module is already loaded,
 *  \return EFAULT if an opcode linkage is unimplemented
 */
longlong_t
wsi_wrctl( int command, int id, void *buffer )
{
   rval_t           rval;
   struct wss_proc *wrp;

#ifdef SUSER_ONLY
   if( !suser() )
   {
      set_errno( EPERM );

      rval.r_val1 = -EPERM;
      rval.r_val2 = 0;
      return rval;
   }
#endif

   /*
    * Find wss_proc structure for current process
    */
   if( ( wrp = wss_pfind( curproc->p_pid ) ) )
   {
      boolean_t badsys = B_FALSE;
      if( wrp->wrp_wi_list )
         badsys = B_TRUE;

      wss_proc_rele( wrp );

      /* If we're wrapped, make it look like this syscall isn't available */
      if( badsys )
         return( (*nosys_ent.sy_callc)() );
   }

   switch( command )
   {
      case WRCTL_LOAD:
      {
         /* Load wrapper module */
         wss_mod_t  *wrm;
         pd_char_t *filename;
         int         error;

#ifdef DEBUG_WRCTL
         wsi_printf( PD_MAKESTR("Allocating block\n") );
#endif
         if( (filename = (pd_char_t *)wss_malloc(id)) == NULL )
         {
            set_errno(EAGAIN);
            return -EAGAIN;
         }
         copyin( buffer, filename, id );

         /*
          * Create and add a new wrapper module information
          * structure to the global list.
          */
#ifdef DEBUG_WRCTL
         wsi_printf( PD_MAKESTR("Allocating module block\n") );
#endif
         if( (wrm = wss_mod_new( filename, &error ) ) == NULL )
         {
            wss_free(filename);
            set_errno(error);
            return -error;
         }

         /*
          * Setup the wrapper module structure
          */
#ifdef DEBUG_WRCTL
         wsi_printf( PD_MAKESTR("**** getting time ****\n") );
#endif
         wrm->wrm_uid = curproc->p_cred->cr_ruid;

#ifdef DEBUG_WRCTL
         wsi_printf( PD_MAKESTR("Module #%d '%s' loaded\n"), wrm->wrm_id,
                     wrm->wrm_linked->linker_file_name );
#endif

         rval.r_val1 = wrm->wrm_id;
         return rval.r_vals;
      }

      case WRCTL_UNLOAD:
      {
         /* Unload wrapper module */
         int        type = 0;
         wss_mod_t *wrm;
         int        error;

         /* Find the module */
         if( ( wrm = wss_mod_find( id ) ) != NULL )
         {
            type = wrm->wrm_type;
            wss_mod_rele( wrm );
         }

         /* Delete the module */
         error = wss_mod_delete( id );

         if( error )
         {
            set_errno(error);
            return -error;
         }
         else
         {
            /*
             * Post module unloaded event
             */
            wss_wrvevent_post_modunloaded( id, type );

            rval.r_val1 = id;
            return rval.r_vals;
         }
      }

      case WRCTL_ENABLED:
      {
         /* Enable/disable wrapper module */
         wss_crit_t       *act;
         wss_mod_t        *wrm;
         static boolean_t  flag;

         flag = ((buffer==NULL)?B_FALSE:B_TRUE);
   
         /* Find the module */
         wrm = wss_mod_find( id );
         if( wrm == NULL )
         {
            set_errno( ENOENT );
            return -ENOENT;
         }

         /* Make sure it's a criteria module */
         if( ( wrm->wrm_type != WR_MOD_ACTIVATION ) &&
             ( wrm->wrm_type != WR_MOD_DEACTIVATION ) )
         {
            wss_mod_rele( wrm );
            wsi_printf( PD_MAKESTR("TBD: wrappers don't have an enabled bit to set (yet)\n") );
            set_errno( EINVAL );
            return -EINVAL;
         }

         /* Make sure the criteria exists and toggle the status accordinly */
         if( wrm->wrm_type == WR_MOD_ACTIVATION )
         {
            WSS_MUTEX_ENTER( wss_act_crit_list );
            for( act = wss_act_crit_list;
                 act && ( act->crit_module->wrm_id != id );
                 act = act->crit_next );
            if( act )
               act->crit_status = ( flag == B_TRUE )?WRACT_ENABLE:WRACT_DISABLE;
            WSS_MUTEX_EXIT( wss_act_crit_list );
         }
         else
         {
            WSS_MUTEX_ENTER( wss_deact_crit_list );
            for( act = wss_deact_crit_list;
                 act != NULL && ( act->crit_module->wrm_id != id );
                 act = act->crit_next );
            if( act )
               act->crit_status = ( flag == B_TRUE )?WRACT_ENABLE:WRACT_DISABLE;
            WSS_MUTEX_EXIT( wss_deact_crit_list );
         }
         wss_mod_rele( wrm );
      
         if( act == NULL )
         {
            set_errno( ENOENT );
            return -ENOENT;
         }

         return 0;
      }

      case WRCTL_STAT_LINKERS:
      {
         wss_linker_t *linker;
         int           num;

         WSS_MUTEX_ENTER( wss_linker_list );
         for( linker = wss_linker_list, num = 0;
              linker;
              linker = linker->linker_next, num++ )
         ;

         uprintf( "[ %d Linker%s Registered ]\n", num, (num==1)?"":"s" );
         for( linker = wss_linker_list;
              linker;
              linker = linker->linker_next )
         {
            uprintf( "%s @ 0x%x\n", linker->linker_description,
                     (u_int32_t)linker );
         }
         WSS_MUTEX_EXIT( wss_linker_list );

         return 0;
      }

      case WRCTL_STAT_REFS:
      {
         if( id & ( WRCTL_STAT_REFS_PROCESSES |
                    WRCTL_STAT_REFS_WRAPPER_INSTANCES ) )
         {
            wss_proc_t *wrp;

            WSS_MUTEX_ENTER( wss_proc_list );
            uprintf( "\n[ Process References ]\n" );
            for( wrp = wss_proc_list.lh_first;
                 wrp;
                 wrp = wrp->wrp_p_list.le_next )
            {
               uprintf( "%-50s = %d reference%s\n",
                        (wrp->wrp_progname==NULL)?"<null>":wrp->wrp_progname,
                        WSS_REF_COUNT(wrp->wrp_refcount),
                        (WSS_REF_COUNT(wrp->wrp_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
               if( id & WRCTL_STAT_REFS_PROCESSES_DUMP_CHAIN )
               {
                  wss_ref_t *ref;

                  for( ref = wrp->wrp_refcount.ref_next;
                       ref;
                       ref = ref->ref_next )
                     uprintf( "   %s:%d\n", ref->un.st.ref_file,
                              ref->un.st.ref_line );
               }
#endif /* DEBUG_REFS */

               if( id & WRCTL_STAT_REFS_WRAPPER_INSTANCES )
               {
                  wss_wrapper_instance_t *wi;

                  if( wrp->wrp_wi_list )
                  {
                     uprintf( "   === Wrapper Instances ===\n" );
                     for( wi = (wss_wrapper_instance_t *)wrp->wrp_wi_list;
                          wi;
                          wi = wi->wi_next )
                     {
                        uprintf( "   %-47s = %d reference%s\n",
                                 wi->wi_wrapper->w_module->wrm_linked->linker_file_name,
                                 WSS_REF_COUNT(wi->wi_refcount),
                                 (WSS_REF_COUNT(wi->wi_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
                        if( id & WRCTL_STAT_REFS_WRAPPER_INSTANCES_DUMP_CHAIN )
                        {
                           wss_ref_t *ref;

                           for( ref = wi->wi_refcount.ref_next;
                                ref;
                                ref = ref->ref_next )
                              uprintf( "   %s:%d\n", ref->un.st.ref_file,
                                       ref->un.st.ref_line );
                        }
#endif /* DEBUG_REFS */
                     }
                  }
               }
            }
            WSS_MUTEX_EXIT( wss_proc_list );
         }

         if( id & WRCTL_STAT_REFS_MODULES )
         {
            wss_mod_t *wrm;

            WSS_MUTEX_ENTER( wss_module_list );
            uprintf( "\n[ Module References ]\n" );
            for( wrm = wss_module_list; wrm; wrm = wrm->wrm_next )
            {
               uprintf( "%-50s = %d reference%s\n",
                        wrm->wrm_linked->linker_file_name,
                        WSS_REF_COUNT(wrm->wrm_refcount),
                        (WSS_REF_COUNT(wrm->wrm_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
               if( id & WRCTL_STAT_REFS_MODULES_DUMP_CHAIN )
               {
                  wss_ref_t *ref;

                  for( ref = wrm->wrm_refcount.ref_next;
                       ref;
                       ref = ref->ref_next )
                     uprintf( "   %s:%d\n", ref->un.st.ref_file,
                              ref->un.st.ref_line );
               }
#endif /* DEBUG_REFS */
            }
            WSS_MUTEX_EXIT( wss_module_list );
         }

         if( id & WRCTL_STAT_REFS_WRAPPERS )
         {
            wss_wrapper_t *w;

            WSS_MUTEX_ENTER( wss_wrapper_list );
            uprintf( "\n[ Wrapper References ]\n" );
            for( w = wss_wrapper_list; w; w = w->w_next )
            {
               uprintf( "%-50s = %d reference%s\n",
                        w->w_module->wrm_linked->linker_file_name,
                        WSS_REF_COUNT(w->w_refcount),
                        (WSS_REF_COUNT(w->w_refcount)==1)?"":"s" );
#ifdef DEBUG_REFS
               if( id & WRCTL_STAT_REFS_WRAPPERS_DUMP_CHAIN )
               {
                  wss_ref_t *ref;

                  for( ref = w->w_refcount.ref_next; ref; ref = ref->ref_next )
                     uprintf( "   %s:%d\n", ref->un.st.ref_file,
                              ref->un.st.ref_line );
               }
#endif /* DEBUG_REFS */
            }
            WSS_MUTEX_EXIT( wss_wrapper_list );
         }

         if( id & WRCTL_STAT_REFS_LINKER_FILES )
         {
            wss_linker_file_t *lf;

            WSS_MUTEX_ENTER( wss_linker_file_list );
            uprintf( "\n[ Linker File References ]\n" );
            for( lf = &wss_linker_file_list; lf; lf = lf->linker_file_next )
            {
               uprintf( "%-50s = %d reference%s\n",
                        lf->linker_file_name,
                        WSS_REF_COUNT(lf->linker_file_refcount),
                        (WSS_REF_COUNT(lf->linker_file_refcount)==1)?"":"s" );
            }
            WSS_MUTEX_EXIT( wss_linker_file_list );
         }
      }

      default:
         set_errno(EINVAL);
         return -EINVAL;
   }
}

