/*! \file wss_criteria.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_criteria.h,v 1.27 2001/06/22 04:35:18 stevek Exp $
 *
 * \brief   Prototypes and structures used by the activation and deactivation
 *          critera code
 * \author  Steve Kiernan
 * \version $Revision: 1.27 $
 * \date    $Date: 2001/06/22 04:35:18 $
 *
 * \sa wss_criteria.c
 */

#ifndef __WSS_CRITERIA_H
#define __WSS_CRITERIA_H   1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif

#ifndef __WSI_TYPES_H
#include <wsi_types.h>
#endif

#ifndef __WSS_REF_H
#include <wss_ref.h>
#endif

#ifndef __WSS_PROCESS_H
#include <wss_process.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/

typedef struct wss_crit wss_crit_t;

/*! \brief Represents a loaded activation or deactivation criterion */
struct wss_crit
{
   /* Fields protected by global wss_[act|deact]_crit_list lock */
   wss_crit_t        *crit_next;     /*!< Pointer to next criteria structure */

   WSS_MUTEX_PROTECT( crit );        /*!< Lock protecting crit_refcount and
                                          crit_status                         */
   /* Fields protected by crit lock */
   wss_ref_t          crit_refcount; /*!< Reference count for structure       */
   int                crit_status;   /*!< Criteria status (\b WRACT_ENABLE or
                                          \b WRACT_DISABLE)                   */
#define WRACT_ENABLE  1              /*!< Criteria is enabled flag            */
#define WRACT_DISABLE 0              /*!< Criteria is disabled flag           */

   /* Not protected by locks (exists same for life of criteria) */
   pd_char_t       **crit_implist;  /*!< Pointer to array of strings
                                          containing the names of the
                                          dependency modules                  */
   struct wss_mod    *crit_module;   /*!< Pointer to module structure for
                                          this criteria                       */
   void             (*crit_func)(wss_proc_t *,wss_crit_t *);
                                     /*!< Pointer to criteria function        */
};


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

extern wss_crit_t WSSLINKAGE *wss_act_crit_list;
WSS_MUTEX_EXPORT             ( wss_act_crit_list );
extern wss_crit_t WSSLINKAGE *wss_deact_crit_list;
WSS_MUTEX_EXPORT             ( wss_deact_crit_list );
extern pd_bool_t  WSSLINKAGE  wss_crit_admin_done;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int  WSSLINKAGE wss_crit_add               ( struct wss_mod *wrm, void *df,
                                             pd_char_t **imp_list );
int  WSSLINKAGE wss_crit_can_unload        ( wss_id_t id );
int  WSSLINKAGE wss_crit_check_activation  ( wss_proc_t *wrp );
int  WSSLINKAGE wss_crit_check_deactivation( wss_proc_t *wrp );
int  WSSLINKAGE wss_crit_delete            ( wss_id_t id );
void WSSLINKAGE wss_crit_free              ( wss_crit_t * );
void WSSLINKAGE wss_crit_init              ( void );
void WSSLINKAGE wss_crit_ref               ( wss_crit_t *crit );
void WSSLINKAGE wss_crit_rele              ( wss_crit_t *crit );
void WSSLINKAGE wss_crit_shutdown          ( void );


/***************************************************************************\
 * Functions implemented as macros                                         *
\***************************************************************************/

/*  wss_crit_ref */
/*! \brief Obtain reference to criteria structure
 *
 *  \param p Pointer to criteria structure
 */
#define wss_crit_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->crit_refcount) < 1) \
         panic("wss_crit_ref: reference count for 0x%x was %d < 1", \
               (unsigned int)(p), WSS_REF_COUNT((p)->crit_refcount)); \
      WSS_MUTEX_ENTER((p)->crit); \
      WSS_REF_INC((p)->crit_refcount); \
      WSS_MUTEX_EXIT((p)->crit); \
   } while(0)

/*  wss_crit_rele */
/*! \brief Release reference to criteria structure
 *
 *  \param p Pointer to criteria structure
 */
#define wss_crit_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->crit_refcount) < 1) \
         panic("wss_crit_rele: reference count for 0x%x was %d < 1", \
               (unsigned int)(p), WSS_REF_COUNT((p)->crit_refcount)); \
      WSS_MUTEX_ENTER((p)->crit); \
      if(WSS_REF_COUNT((p)->crit_refcount) > 1) { \
         WSS_REF_DEC((p)->crit_refcount); \
         WSS_MUTEX_EXIT((p)->crit); \
      } else { \
         wss_crit_free(p); \
      } \
   } while(0)


#ifdef __cplusplus
}
#endif

#endif /* __WSS_CRITERIA_H */

