/*
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_db.h,v 1.37 2001/02/23 19:02:47 stevek Exp $
 *
 * Author: Doug Kilpatrick and Lee Badger
 *
 * Purpose: contains the database representation structures.
 */
/*
 * Preliminary WQL data structures and accessing functions.
 *
 * In one contiguous region of memory, the DB is represented in the
 * most simple way: a list of tables; for each table, a list of rows;
 * for each row, a list of values.  All lists are NULL-terminated;
 * order within a list matters.  Each row of a table is an ordered
 * list of wql_val structures, which _correspond_ to the table's
 * ordered list of wql_col structures.  The entire DB can be searched
 * by starting at the first (well-known) table pointer; optimization
 * for speed will come later.  Of course, it's slightly more
 * complicated than that, but the simplest view is:
 */

#ifndef WR_SUPPRESS_WQL

#ifndef __WSS_DB_H
#define __WSS_DB_H 1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif /* WR_DEFS_INCLUDED */

#ifdef PD_IN_KERNEL
#ifndef __WSS_MUTEX_H
#include <wss_mutex.h>
#endif
#endif /* PD_IN_KERNEL */

#if defined(PD_IN_KERNEL) || defined(USER_MODE_TEST)
#ifndef __WSS_VM_H
#include <wss_vm.h>
#endif
#endif /* PD_IN_KERNEL || USER_MODE_TEST */

#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/

typedef unsigned long wss_db_handle_t;

#ifdef PD_IN_KERNEL
/*
 * Note: this structure is allocated in kernel memory, _NOT_ the virtual
 *       memory mapped for the database.  It should only contain transient
 *       information, _NOT_ data which should be persistent
 */
typedef struct wql_table_opaque
{
   /* Protects everything except "next".  Next is protected by
    * a mutex internal to wss_db.
    */
   WSS_MUTEX_PROTECT(table);
}
wqtabopaque, *wqtabopaque_t;
#endif /* PD_IN_KERNEL */

#if defined(PD_IN_KERNEL) || defined(USER_MODE_TEST)
typedef struct wss_db_head wss_db_head_t;

struct wss_db_head
{
   u_int32_t    db_version;
   wss_vmaddr_t db_tables;
};

#endif /* PD_IN_KERNEL || USER_MODE_TEST */

typedef struct wql_table
{
   /* In order to not polute user-space with kernel-space constructs, we
    * use an opaque pointer which contains the appropriate lock for the
    * table (see struct wql_table_opaque above)
    */
   void               *opaque;
   pd_char_t          *name;
   unsigned int        flags;
   int                 col_count;
   int                 max_rows;  /* TBD */
   int                 row_count; /* TBD */
   struct wql_col_def *col_defs;  /* defines this table's format */
   struct wql_row     *rows;
   struct wql_row     *tail;      /* TBD */
   struct wql_table   *next;
}
wqtab, *wqtab_t;

typedef enum wql_dtype { none_t, pd_char_t_t, int_t, float_t } wql_dtype;
typedef struct wql_col_def
{
   pd_char_t *name;
   wql_dtype   dtype;
   int         max_len;
}
wqcd, *wqcd_t;

typedef struct wql_row
{
   struct wql_val *vals;
   struct wql_row *next;
}
wqrow, *wqrow_t;

typedef struct wql_val
{
   union
   {
      pd_char_t  *cval;
      int         ival;
      float       fval;
   }
   v;
}
wqval, *wqval_t;

/* holds data returned from select */
/* This is a structure instead of a pointer to a structure so that it does
 * not need an extra creation step.
 */
struct wql_dbrow
{
   struct wql_table  results;
   struct wql_row   *cur;
};

/*
 * Definitions to support select/update/delete expressions.  An
 * expression is essentially a tree of structures.  Code in
 * select/update/delete traverses the tree to evaluate rows to act
 * upon.  It's a bit messy; will think about how to pretty it up
 * later.
 */

enum sxop { gt_op, gte_op, lt_op, lte_op, ne_op, eq_op, true_op };
enum sxarg { literal, field, variable };
typedef struct sxval
{
   union
   {
      union literal
      {
         float            fval;
         int              ival;
         const pd_char_t *cval;
      }
      l;
      struct field
      {
         int         column;  /* By index, starting at 0 */
      }
      f;
      struct variable
      {
         const void *ptr;
      }
      v;
   }
   v;
   wql_dtype  dtype;
   enum sxarg which;
   char owned;                /* boolean */
}
sxval, *sxval_t;

/*
 * Represents a relational expression such as ".count = 5" in a select
 * expression.
 */
typedef struct term_exp
{
   sxval_t left;
   sxval_t right;
   enum sxop sxop;
}
tx, *tx_t;

/* Hack.  not_op negates the left side, and ignores the right */
enum logop { and_op, or_op, not_op };
typedef struct nterm_exp
{
   struct sx_node *left;
   struct sx_node *right;
   enum logop      logop;
}
ntx, *ntx_t;


/*
 * A select (or delete or update) expression is a tree of these.  The tree
 * is evaluated at runtime to identify rows of a table that should be
 * returned/deleted/updated.
 */
enum nodekind { term, nonterm };
typedef struct sx_node
{
   union
   {
      tx_t  exp;
      ntx_t children;
   }
   un;                  /* Hideous hack to get around hideous brokenness. */
   enum nodekind which;
}
wqsx, *wqsx_t;

/*
 * Represents an update_set_list: a list of columns to set in an update
 * expression.
 */
typedef struct update_set_list
{
   struct sxval             *left;
   struct update_expression *right;
   struct update_set_list   *next;
}
wqusl, *wqusl_t;

/*
 * Represents a update_expression: a numeric value calculated from this
 * tree structure.
 */
enum numop {multiplication, division, mod, addition, subtraction, negation};
typedef struct update_expression
{
   union
   {
      sxval_t terminal;
      struct
      {
         struct update_expression *left;
         struct update_expression *right;
         enum numop                numop;
      }
      e;
   }
   w;
   enum nodekind which;
}
wquexp, *wquexp_t;


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

#if defined(PD_IN_KERNEL) || defined(USER_MODE_TEST)
extern wss_vmspace_t WSSLINKAGE *wss_db_space;
#endif /* PD_IN_KERNEL || USER_MODE_TEST */


/***************************************************************************\
 * Public macros                                                           *
\***************************************************************************/

#define WSS_DB_VERSION 1

#ifdef PD_IN_KERNEL
#define WQL_TABLE_OPAQUE(t) ((wqtabopaque_t)((t)->opaque))
#endif /* PD_IN_KERNEL */

#if defined(PD_IN_KERNEL) || defined(USER_MODE_TEST)
#define DBSPACEHEAD    (wss_db_space->vms_head)
#define DB_ABS2REL(ap) abs2rel(wss_db_space,(ap))
#define DB_REL2ABS(rp) rel2abs(wss_db_space,(rp))
#define DBHEAD         ((wss_db_head_t *)DB_REL2ABS(DBSPACEHEAD->vms_head_reserved))
#endif /* PD_IN_KERNEL || USER_MODE_TEST */

#define DB_FLAG_UNICODE       0x0001   /* Table contains Unicode characters */
#define DB_FLAG_LOCAL         0x0002   /* Table is a local table            */
#define DB_FLAG_IMMUTABLE     0x0004   /* Table cannot be altered           */
#define DB_FLAG_NONPERSISTENT 0x0008   /* Table does not survive reboot     */


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*
 * PD_SUPPRESS_FUNCTION_DECL is turned on by libraries.  This prevents
 * wrappers from attempting to reference symbols like wss_create, which
 * doesn't exist as far as the linker is concerned.
 */
#ifndef PD_SUPPRESS_FUNCTION_DECL

#if defined(PD_IN_KERNEL) || defined(USER_MODE_TEST) || defined(DBINTD_INTERFACE)
#if 0
int      WSSLINKAGE wr_dbrow_free                   __P(( struct wql_dbrow *rows ));
#endif

void     WSSLINKAGE wss_dbck                        __P(( void ));
int      WSSLINKAGE wss_db_init                     __P(( int ));

#if USER_MODE_TEST
wqtab_t             wss_tables                      __P(( void ));
#endif
wqtab_t  WSSLINKAGE wss_db_find_table_by_name       __P(( pd_char_t *name ));
wqtab_t  WSSLINKAGE wss_db_find_table_by_name_unsafe __P(( pd_char_t *name ));
unsigned WSSLINKAGE long wss_db_find_table_by_names __P(( pd_char_t *wrapper_name, int id,
                                                          pd_char_t *table_name ));
int      WSSLINKAGE wss_db_insert_nulls             __P(( u_long table_id ));
int      WSSLINKAGE wss_db_delete_by_handle         __P(( u_long table_id,
                                                          u_long row_id ));
wqtab_t  WSSLINKAGE wss_db_find_table_by_handle     __P(( u_long handle ));
int      WSSLINKAGE wss_create                      __P(( pd_char_t *wrapper_name,
                                                          int instance_id,
                                                          pd_char_t *table_name,
                                                          int free_col_defs,
                                                          struct wql_col_def *col_defs,
                                                          int col_count,
                                                          u_long *table_id ));
int      WSSLINKAGE wss_drop                        __P(( u_long table_id ));
int      WSSLINKAGE wss_insert                      __P(( u_long table_id, int free_vals,
                                                          wqrow_t vals ));
int      WSSLINKAGE wss_select                      __P(( u_long table_id, int free_scrit,
                                                          wqsx_t scrit, int take_first,
                                                          int order_by, int desc,
                                                          wqtab_t out_table ));
int      WSSLINKAGE wss_delete                      __P(( u_long table_id, int free_scrit,
                                                          wqsx_t scrit, int take_first ));
int      WSSLINKAGE wss_update                      __P(( u_long table_id,
                                                          int free_scripts, wqsx_t scrit,
                                                          int take_first,
                                                          wqusl_t set_script ));

/*
 * Support functions.
 * Used by the compilier
 */

wqcd_t   WSSLINKAGE wss_db_mcol_def                 __P(( int count, ... ));
wqcd_t   WSSLINKAGE wss_db_mcol_def_va              __P(( int count, va_list ap ));
wqrow_t  WSSLINKAGE wss_db_mrow                     __P(( wqval_t vals, wqrow_t next ));
wqsx_t   WSSLINKAGE wss_db_mwqsx_and                __P(( wqsx_t left, wqsx_t right ));
wqsx_t   WSSLINKAGE wss_db_mwqsx_or                 __P(( wqsx_t left, wqsx_t right ));
wqsx_t   WSSLINKAGE wss_db_mwqsx_terminal           __P(( tx_t terminal ));
wqsx_t   WSSLINKAGE wss_db_mwqsx_not                __P(( wqsx_t expression ));
tx_t     WSSLINKAGE wss_db_mtx                      __P(( sxval_t left, sxval_t right,
                                                          enum sxop sxop ));
sxval_t  WSSLINKAGE wss_db_msxval_field             __P(( int column ));
sxval_t  WSSLINKAGE wss_db_msxval_lit_int           __P(( int i ));
sxval_t  WSSLINKAGE wss_db_msxval_lit_float         __P(( double d ));
sxval_t  WSSLINKAGE wss_db_msxval_lit_string        __P(( const pd_char_t *str ));
#if 0
int      WSSLINKAGE wss_db_dbrow_start              __P(( struct wql_dbrow *dbrp ));
int      WSSLINKAGE wss_db_dbrow_next               __P(( struct wql_dbrow *dbrp ));
int      WSSLINKAGE wss_db_dbrow_isempty            __P(( struct wql_dbrow *dbrp ));
#endif
void     WSSLINKAGE wss_db_dbrow_finish             __P(( struct wql_dbrow *dbrp ));
wqusl_t  WSSLINKAGE wss_db_musl                     __P(( sxval_t left, wquexp_t right,
                                                          wqusl_t next ));
wquexp_t WSSLINKAGE wss_db_muexp                    __P(( wquexp_t left, wquexp_t right,
                                                          enum numop numop ));
wquexp_t WSSLINKAGE wss_db_muexp_term               __P(( sxval_t value ));
wqval_t  WSSLINKAGE wss_db_mvals                    __P(( int count, ... ));
wqval_t  WSSLINKAGE wss_db_mvals_va                 __P(( int count, va_list ap ));
void     WSSLINKAGE wss_db_free_criteria            __P(( wqsx_t expr ));
void     WSSLINKAGE wss_db_free_wqusl               __P(( wqusl_t list ));

/*
 * Table locking functions
 */
#ifdef PD_IN_KERNEL
void     WSSLINKAGE wss_db_lock_table               __P(( wqtab_t ));
void     WSSLINKAGE wss_db_unlock_table             __P(( wqtab_t ));
#endif /* PD_IN_KERNEL */

/*
 * Database locking functions
 */
void     WSSLINKAGE wss_db_lock                     __P(( void ));
void     WSSLINKAGE wss_db_unlock                   __P(( void ));

#endif /* PD_IN_KERNEL || USER_MODE_TEST || DBINTD_INTERFACE */


/***************************************************************************\
 * Functions implemented as macros                                         *
\***************************************************************************/
#ifdef PD_IN_KERNEL
#  ifdef DEBUG_DB_LOCKS
#define  wss_db_lock_table(t) WSS_MUTEX_ENTER_DEBUG(WQL_TABLE_OPAQUE(t)->table)
#define  wss_db_unlock_table(t) WSS_MUTEX_EXIT_DEBUG(WQL_TABLE_OPAQUE(t)->table)
#  else
#define  wss_db_lock_table(t) WSS_MUTEX_ENTER(WQL_TABLE_OPAQUE(t)->table)
#define  wss_db_unlock_table(t) WSS_MUTEX_EXIT(WQL_TABLE_OPAQUE(t)->table)
#  endif /* DEBUG_DB_LOCKS */
#endif /* PD_IN_KERNEL */

#endif /* PD_SUPPRESS_FUNCTION_DECL */

#ifdef __cplusplus
}
#endif

#endif /* __WSS_DB_H */
#endif /* WR_SUPPRESS_WQL */

