/*! \file wss_library.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_library.h,v 1.23 2001/02/23 19:02:47 stevek Exp $
 *
 * \brief   Prototypes and structures used by the module loading code in order
 *          to handle wrapper libraries
 * \author  Chris Vance, Steve Kiernan
 * \version $Revision: 1.23 $
 * \date    $Date: 2001/02/23 19:02:47 $
 *
 * \sa wss_library.c
 */

#ifndef __WSS_LIBRARY_H
#define __WSS_LIBRARY_H 1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif

#ifndef __WSI_TYPES_H
#include <wsi_types.h>
#endif

#ifndef __WSS_MUTEX_H
#include <wss_mutex.h>
#endif

#ifndef __WSS_REF_H
#include <wss_ref.h>
#endif

#ifndef __WSS_IFACE_H
#include <wss_iface.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************\
 * Forward references                                                      *
\***************************************************************************/

struct wss_mod;


/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/

typedef struct wss_lib wss_lib_t;

/*! Wrapper library structure */
struct wss_lib
{
   /* Fields protected by global wss_library_list lock */
   wss_lib_t        *lib_next;     /*!< Pointer to next library in list   */

   WSS_MUTEX_PROTECT(lib);         /*!< Lock protecting lib_refcount and
                                        lib_status                        */
   /* Fields protected by lib lock */
   wss_ref_t         lib_refcount; /*!< Reference count of structure      */
   int               lib_status;   /*!< Library status flags              */

   /* Not protected by locks (exists same for life of library) */
   wss_symbol_t     *lib_symtab;   /*!< Pointer to the exported symbols
                                        table for the library             */
   struct wss_mod   *lib_module;   /*!< Pointer to the wrapper module for
                                        the library                       */
   int               lib_major;    /*!< Major version of the library      */
   int               lib_minor;    /*!< Minor version of the library      */
   void            (*lib_fini)(void);/*!< Pointer to function to call when
                                        library is unloaded               */
};


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

wss_lib_t WSSLINKAGE *wss_library_list;
WSS_MUTEX_EXPORT    ( wss_library_list );


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int       WSSLINKAGE  wss_lib_can_unload    ( wss_id_t );
int       WSSLINKAGE  wss_lib_delete        ( wss_id_t );
wss_lib_t WSSLINKAGE *wss_lib_find          ( wss_id_t );
wss_lib_t WSSLINKAGE *wss_lib_find_by_name  ( const pd_char_t * );
void      WSSLINKAGE  wss_lib_free          ( wss_lib_t * );
void      WSSLINKAGE  wss_lib_init          ( void );
void      WSSLINKAGE  wss_lib_lock_symbols  ( wss_lib_t * );
void      WSSLINKAGE  wss_lib_ref           ( wss_lib_t * );
int       WSSLINKAGE  wss_lib_register      ( struct wss_mod *,
                                              wss_symbol_t *, void * );
void      WSSLINKAGE  wss_lib_rele          ( wss_lib_t * );
void      WSSLINKAGE  wss_lib_shutdown      ( void );
void      WSSLINKAGE  wss_lib_unlock_symbols( wss_lib_t * );


/***************************************************************************\
 * Functions implemented as macros                                         *
\***************************************************************************/

/*  wss_lib_ref */
/*! \brief Obtain reference to library structure
 *
 *  \param p Pointer to library structure
 */
#define wss_lib_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->lib_refcount) < 1) \
         panic("wss_lib_ref: reference count for 0x%x was %d < 1", \
               (unsigned int)(p), WSS_REF_COUNT((p)->lib_refcount)); \
      WSS_MUTEX_ENTER((p)->lib); \
      WSS_REF_INC((p)->lib_refcount); \
      WSS_MUTEX_EXIT((p)->lib); \
   } while(0)

/*  wss_lib_rele */
/*! \brief Release reference to library structure
 *
 *  \param p Pointer to library structure
 */
#define wss_lib_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->lib_refcount) < 1) \
         panic("wss_lib_rele: reference count for 0x%x was %d < 1", \
               (unsigned int)(p), WSS_REF_COUNT((p)->lib_refcount)); \
      WSS_MUTEX_ENTER((p)->lib); \
      if(WSS_REF_COUNT((p)->lib_refcount) > 1) { \
         WSS_REF_DEC((p)->lib_refcount); \
         WSS_MUTEX_EXIT((p)->lib); \
      } else { \
         wss_lib_free(p); \
      } \
   } while(0)


#ifdef __cplusplus
}
#endif

#endif /* __WSS_LIBRARY_H */

