/*! \file wss_module.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_module.h,v 1.41 2001/02/23 19:02:47 stevek Exp $
 *
 * \brief   Prototypes and structures used by the module loading code
 * \author  Chris Vance, Steve Kiernan
 * \version $Revision: 1.41 $
 * \date    $Date: 2001/02/23 19:02:47 $
 *
 * \sa wss_module.c
 */

#ifndef __WSS_MODULE_H
#define __WSS_MODULE_H  1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif

#ifndef __WSI_TYPES_H
#include <wsi_types.h>
#endif

#ifndef __WSS_REF_H
#include <wss_ref.h>
#endif
#ifndef __WSS_PROCESS_H
#include <wss_process.h>
#endif
#ifndef __WSS_WRAPPER_H
#include <wss_wrapper.h>
#endif
#ifndef __WSS_CRITERIA_H
#include <wss_criteria.h>
#endif
#ifndef __WSS_LIBRARY_H
#include <wss_library.h>
#endif
#ifndef __WSS_LINKER_H
#include <wss_linker.h>
#endif
#ifndef __WSS_FILE_H
#include <wss_file.h>
#endif


#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************\
 * Public macros                                                           *
\***************************************************************************/

#define WRLINK_MOD_TYPE             __wr_mytype
#define WRLINKNAME_MOD_TYPE         "__wr_mytype"


/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/

typedef enum wss_mod_type       wss_mod_type_t;
typedef struct wss_mod          wss_mod_t;
typedef struct wss_mod_file_map wss_mod_file_map_t;
typedef struct wss_mod_depend   wss_mod_depend_t;

/*! \brief Enumeration of available module types
 *
 *  The WR_MOD_* values are stored in the wr_module structure.  They
 *  are used to specify what type of module this is.  It should be set
 *  during the module's registration routine.
 */
enum wss_mod_type
{
   WR_MOD_WRAPPER      = 1,  /*!< Module is a wrapper type               */
   WR_MOD_ACTIVATION   = 2,  /*!< Module is an activation criteria type  */
   WR_MOD_LIBRARY      = 4,  /*!< Module is a wrapper library type       */
   WR_MOD_DEACTIVATION = 8   /*!< Module is a deactivation criteria type */
};

/*! \brief Structure used to keep track of files opened by the wrapper module */
struct wss_mod_file_map
{
   wss_file_t          map_file; /*!< Handle to file              */
   wss_mod_file_map_t *map_next; /*!< Pointer to next map in list */
};

/*! \brief The wrappers module structure
 *
 *  It stores information about modules that have been loaded into the WSS.
 *  Currently modules are either wrappers, activation criteria, or libraries.
 */
struct wss_mod
{
   /* Fields protected by global wss_module_list lock */
   wss_mod_t          *wrm_next;      /*<! Pointer to next module in list   */


   WSS_MUTEX_PROTECT  (wrm);          /*<! Mutex to protect wrm_refcount    */
   /* Fields protected by wrm lock */
   wss_ref_t           wrm_refcount;  /*<! Reference count for structure    */
   void              (*wrm_unload)(void);/*<! Pointer to module unload funtion*/


   WSS_MUTEX_PROTECT  (wrm_filemap);  /*<! Mutex to protect wrm_filemap     */
   /* Fields protected by wrm_filemap lock */
   wss_mod_file_map_t *wrm_filemap;   /*<! Pointer to a list of file maps   */


   WSS_MUTEX_PROTECT  (wrm_uses);     /*<! Mutex to protect wrm_uses        */
   /* Fields protected by wrm_uses lock */
   wss_mod_depend_t   *wrm_uses;      /*<! Pointer to a list of modules
                                           this module uses                 */

   /* Not protected by locks (exists same for life of module) */
   wss_id_t            wrm_id;        /*<! Unique identifier                */
   wss_mod_type_t      wrm_type;      /*<! Type of module                   */
   int                 wrm_uid;       /*<! Person who loaded this module    */
   long                wrm_loadtime;  /*<! Time module was loaded into WSS  */
   wss_linker_file_t  *wrm_linked;    /*<! Pointer to linker file structure */
   volatile void      *wrm_opaque;    /*<! System-dependent data pointer    */
};


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

extern wss_mod_t WSSLINKAGE *wss_module_list;
WSS_MUTEX_EXPORT           ( wss_module_list );


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

#ifndef WR_SUPPRESS_FUNCTION_DECL
int         WSSLINKAGE  wss_mod_add_dependency    ( wss_mod_t *, wss_mod_t * );
int         WSSLINKAGE  wss_mod_at_unload         ( void *, wss_id_t );
void        WSSLINKAGE  wss_mod_close_file        ( int, wss_file_t );
int         WSSLINKAGE  wss_mod_delete            ( wss_id_t );
#ifdef DEBUG_REFS
wss_mod_t   WSSLINKAGE *wss_mod_find_debug        ( wss_id_t,
                                                    const pd_char_t *, int );
#endif /* DEBUG_REFS */
wss_mod_t   WSSLINKAGE *wss_mod_find              ( wss_id_t );
#ifdef DEBUG_REFS
wss_mod_t   WSSLINKAGE *wss_mod_find_by_name_debug( pd_char_t *,
                                                    const pd_char_t *, int );
#endif /* DEBUG_REFS */
wss_mod_t   WSSLINKAGE *wss_mod_find_by_name      ( pd_char_t * );
void        WSSLINKAGE  wss_mod_free              ( wss_mod_t * );
void        WSSLINKAGE *wss_mod_get_file_map_addr ( int, wss_file_t );
size_t      WSSLINKAGE  wss_mod_get_file_size     ( int, wss_file_t );
void        WSSLINKAGE  wss_mod_init              ( void );
int         WSSLINKAGE  wss_mod_is_loaded         ( const pd_char_t * );
int         WSSLINKAGE  wss_mod_map_file          ( int, wss_file_t, size_t,
                                                    int, pd_offset_t );
wss_mod_t   WSSLINKAGE *wss_mod_new               ( pd_char_t *, int * );
wss_file_t  WSSLINKAGE  wss_mod_open_file         ( int, const pd_char_t *,
                                                    wss_file_mode_t );
int         WSSLINKAGE  wss_mod_read_file         ( int, wss_file_t,
                                                    pd_byte_t *, size_t );
pd_byte_t  WSSLINKAGE *wss_mod_read_file_alloc   ( int, wss_file_t, size_t * );
void        WSSLINKAGE  wss_mod_ref               ( wss_mod_t * );
void        WSSLINKAGE  wss_mod_rele              ( wss_mod_t * );
int         WSSLINKAGE  wss_mod_remove_at_unload  ( wss_id_t );
int         WSSLINKAGE  wss_mod_remove_dependency ( wss_mod_t *, wss_mod_t * );
int         WSSLINKAGE  wss_mod_seek_file         ( int, wss_file_t,
                                                    pd_offset_t, int );
void        WSSLINKAGE  wss_mod_shutdown          ( void );
void        WSSLINKAGE  wss_mod_unload_all        ( void );
void        WSSLINKAGE  wss_mod_unmap_file        ( int, wss_file_t );
int         WSSLINKAGE  wss_mod_write_file        ( int, wss_file_t,
                                                    pd_byte_t *, size_t );


/***************************************************************************\
 * Functions implemented as macros                                         *
\***************************************************************************/

#if defined(DEBUG_REFS) && !defined(DISABLE_MOD_DEBUG_REFS)
#  define wss_mod_find(p)         wss_mod_find_debug((p),__FILE__,__LINE__)
#  define wss_mod_find_by_name(p) wss_mod_find_by_name_debug((p),__FILE__,\
                                                             __LINE__)
#endif /* DEBUG_REFS */

#if defined(DEBUG_REFS) && !defined(DISABLE_MOD_DEBUG_REFS)

#  define wss_mod_ref_debug(p,f,l) \
   do { \
      if(WSS_REF_COUNT((p)->wrm_refcount) < 1) \
         panic("wss_mod_ref: reference count for 0x%x was %d < 1 [%s:%d]", \
               (unsigned int)(p), WSS_REF_COUNT((p)->wrm_refcount), \
               (f), (l)); \
      WSS_MUTEX_ENTER((p)->wrm); \
      wsi_printf( PD_MAKESTR( \
                  "wss_mod_ref: module #%d (0x%x) refcount = %d [%s:%d]\n"), \
                  (p)->wrm_id, (unsigned int)(p), \
                  WSS_REF_INC_DEBUG((p)->wrm_refcount,(f),(l)), (f), (l)); \
      WSS_MUTEX_EXIT((p)->wrm); \
   } while(0)

#  define wss_mod_ref(p) wss_mod_ref_debug(p,__FILE__,__LINE__)

#  define wss_mod_rele_debug(p,f,l) \
   do { \
      if(WSS_REF_COUNT((p)->wrm_refcount) < 1) \
         panic("wss_mod_rele: reference count for 0x%x was %d < 1 [%s:%d]", \
               (unsigned int)(p), WSS_REF_COUNT((p)->wrm_refcount), (f), (l)); \
      WSS_MUTEX_ENTER((p)->wrm); \
      if(WSS_REF_COUNT((p)->wrm_refcount) > 1) { \
         wsi_printf( PD_MAKESTR( \
                    "wss_mod_rele: module #%d (0x%x) refcount = %d [%s:%d]\n"),\
                     (p)->wrm_id, (unsigned int)(p), \
                     WSS_REF_DEC_DEBUG((p)->wrm_refcount,(f),(l)), (f), (l)); \
         WSS_MUTEX_EXIT((p)->wrm); \
      } else { \
         wss_mod_free(p); \
      } \
   } while(0)

#  define wss_mod_rele(p) wss_mod_rele_debug(p,__FILE__,__LINE__)

#else /* DEBUG_REFS */

/*  wss_mod_ref */
/*! \brief Obtain reference to module structure
 *
 *  \param p Pointer to module structure
 */
#  define wss_mod_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->wrm_refcount) < 1) \
         panic("wss_mod_ref: reference count for 0x%x was %d < 1 [%s:%d]", \
               (unsigned int)(p), WSS_REF_COUNT((p)->wrm_refcount), \
               __FILE__, __LINE__ ); \
      WSS_MUTEX_ENTER((p)->wrm); \
      WSS_REF_INC((p)->wrm_refcount); \
      WSS_MUTEX_EXIT((p)->wrm); \
   } while(0)

/*  wss_mod_rele */
/*! \brief Release reference to module structure
 *
 *  \param p Pointer to module structure
 */
#  define wss_mod_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->wrm_refcount) < 1) \
         panic("wss_mod_rele: reference count for 0x%x was %d < 1 [%s:%d]", \
               (unsigned int)(p), WSS_REF_COUNT((p)->wrm_refcount), \
               __FILE__, __LINE__ ); \
      WSS_MUTEX_ENTER((p)->wrm); \
      if(WSS_REF_COUNT((p)->wrm_refcount) > 1) { \
         WSS_REF_DEC((p)->wrm_refcount); \
         WSS_MUTEX_EXIT((p)->wrm); \
      } else { \
         wss_mod_free(p); \
      } \
   } while(0)

#endif /* DEBUG_REFS */


#ifndef __WSI_MODULE_H
#include <wsi_module.h>
#endif

#endif /* WR_SUPPRESS_FUNCTION_DECL */

#ifdef __cplusplus
}
#endif

#endif /* _WSS_MODULE_H */

