/*! \file wss_process.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_process.h,v 1.45 2001/06/23 00:07:42 stevek Exp $
 *
 * \brief   Prototypes and structures used by the process handling code
 * \author  Steve Kiernan
 * \version $Revision: 1.45 $
 * \date    $Date: 2001/06/23 00:07:42 $
 *
 * \sa wss_process.c
 */

#ifndef __WSS_PROCESS_H
#define __WSS_PROCESS_H 1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif

#ifndef __WSI_ERRNO_H
#include <wsi_errno.h>
#endif

#ifndef _SYS_QUEUE_H_
#include <pdpi/queue.h>
#endif

#ifndef __WSI_TYPES_H
#include <wsi_types.h>
#endif

#ifndef __WSS_REF_H
#include <wss_ref.h>
#endif

#ifndef __WSS_IFACE_H
#include <wss_iface.h>
#endif

#ifndef __WSI_PROCESS_H
#include <wsi_process.h>
#endif

#ifndef __WSS_MUTEX_H
#include <wss_mutex.h>
#endif

#ifndef __WSS_CV_H
#include <wss_cv.h>
#endif

#ifndef __WSI_RW_H
#include <wsi_rw.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/

typedef struct wss_arg  wss_arg_t;
typedef struct wss_proc wss_proc_t;

/*! \brief Argument replacement stack structure */
struct wss_arg
{
   int        index;  /*!< Index of argment                  */
   size_t     size;   /*!< Size of data to copy              */
   void      *old;    /*!< Pointer to old argument data      */
   wss_arg_t *next;   /*!< Pointer to next argument in stack */
};

/*! \brief This referred to as the "wrappers-specific process information"
 * 
 *  After the WSS has been installed, each process created is
 *  intercepted during the fork1() syscall.  At this time a wss_proc
 *  structure is created and added to the global list.  This structure
 *  stores process-related information.
 */
struct wss_proc
{
   /* Fields protected by global wss_proc_list lock */
   LIST_ENTRY(wss_proc)         wrp_p_list;    /*!< List of all processes    */


   WSS_MUTEX_PROTECT           (wrp);          /*!< Mutex protecting
                                                    wrp_refcount, wrp_flags,
                                                    wrp_nsyscall             */
   /* Fields protected by wrp lock */
   wss_ref_t                    wrp_refcount;  /*!< Reference count for
                                                    structure                */
   int                          wrp_flags;     /*!< Flags                    */
   int                          wrp_nsyscall;  /*!< Reference count for
                                                    calls                    */


   WSI_RW_PROTECT              (wrp_cwd);      /*!< Reader/writer lock
                                                    protecting wrp_cwd       */
   /* Fields protected by wrp_cwd lock */
   pd_char_t                  *wrp_cwd;       /*!< Pointer to current
                                                    working directory of
                                                    process. MALLOC'd memory,
                                                    NULL-terminated          */


   WSS_MUTEX_PROTECT           (wrp_wi_list);  /*!< Mutex protecting
                                                    wrp_wi_list              */
   /* Fields protected by wrp_wi_list lock */
   volatile
   struct wss_wrapper_instance *wrp_wi_list;   /*!< Pointer to list of
                                                    wrappers instances       */


   WSS_MUTEX_PROTECT           (wrp_arg_list); /*!< Mutex protecting
                                                    wrp_arg_list             */
   /* Fields protected by wrp_arglist lock */
   void                        *wrp_arg_list;  /*!< Pointer to stack of
                                                    replaced arguments       */


   WSS_MUTEX_PROTECT           (wrp_viewer);   /*!< Mutex protecting
                                                    wrp_viewer               */
   /* Fields protected by wrp_viewer lock */
   volatile struct wss_wrv     *wrp_viewer;    /*!< Pointer to the wrapper
                                                    viewer structure when
                                                    process is a wrapper
                                                    viewer                   */


   /* Not protected by locks (exists same for life of process)
    *
    * Note: wrp_progname can change during exec, but all threads should no
    *       longer have a pointer to the old area of memory which held the
    *       name of the program pre-exec                                     */
   pid_t                        wrp_pid;       /*!< Process identifier       */
   pd_char_t                  *wrp_path;      /*!< Pointer to full path of
                                                    executable. MALLOC'd
                                                    memory, NULL-terminated  */
   pd_char_t                  *wrp_oldpath;   /*!< Pointer to full path of
                                                    old program (parent).
                                                    MALLOC'd memory,
                                                    NULL-terminated          */
   pd_char_t                  *wrp_progname;  /*!< Pointer to program name
                                                    (no path info). MALLOC'd
                                                    memory, NULL-terminated  */
   proc_t                      *wrp_p;         /*!< Pointer to process
                                                    structure (may be unused)*/
   void                        *wrp_opaque;    /*!< System-dependent data    */
};


/***************************************************************************\
 * Public macros                                                           *
\***************************************************************************/

#define WRPF_GMASK   0x00000fff        /*!< Global flags mask                */
#define WRPF_WRAPPED 0x00000001        /*!< Process wrapped flag             */
#define WRPF_POSTED  0x00000002        /*!< Event posted flag                */

/* System-dependent parts can use flags within the WRPF_LMASK */
#define WRPF_LMASK   0xfffff000        /*!< Local flags mask                 */
#define WRPF_LOCAL   0x00001000        /*!< Starting value for local flags   */


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

PD_LIST_HEAD(wss_proclist, wss_proc);
extern struct wss_proclist WSSLINKAGE wss_proc_list;
WSS_MUTEX_EXPORT( wss_proc_list );

extern int WSSLINKAGE wss_proc_num_in_syscall;
WSS_MUTEX_EXPORT( wss_proc_num_in_syscall );
WSS_CV_EXPORT( wss_proc_num_in_syscall );


#ifndef WR_SUPPRESS_FUNCTION_DECL
/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

wss_proc_t WSSLINKAGE *wss_pfind                 ( register pid_t );
wss_proc_t WSSLINKAGE *wss_pfind_unsafe          ( register pid_t );
int        WSSLINKAGE  wss_proc_add_wrapper_blind( wss_proc_t *,
                                                   const pd_char_t * );
int        WSSLINKAGE  wss_proc_add_wrapper      ( wss_proc_t *,
                                                   const pd_char_t *,
                                                   struct wss_wrapper_instance ** );
void       WSSLINKAGE  wss_proc_del_all_wrappers ( wss_proc_t * );
int        WSSLINKAGE  wss_proc_del_wrapper      ( wss_proc_t *, pd_char_t * );
pd_char_t  WSSLINKAGE *wss_proc_get_cwd          ( pid_t );
pd_char_t  WSSLINKAGE *wss_proc_get_path         ( pid_t );
void       WSSLINKAGE  wss_proc_exit             ( proc_t * );
void       WSSLINKAGE  wss_proc_exit_stage1      ( proc_t * );
void       WSSLINKAGE  wss_proc_exit_stage2      ( pid_t );
void       WSSLINKAGE  wss_proc_fork             ( proc_t *, proc_t *, int );
void       WSSLINKAGE  wss_proc_free             ( wss_proc_t * );
void       WSSLINKAGE  wss_proc_init             ( void );
wss_proc_t WSSLINKAGE *wss_proc_new              ( proc_t * );
void       WSSLINKAGE  wss_proc_ref              ( wss_proc_t * );
void       WSSLINKAGE  wss_proc_rele             ( wss_proc_t * );
void       WSSLINKAGE  wss_proc_shutdown         ( void );


/***************************************************************************\
 * Functions implemented as macros                                         *
\***************************************************************************/

#if defined(DEBUG_REFS) && !defined(DISABLE_PROC_DEBUG_REFS)

#  define wss_proc_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->wrp_refcount) < 1) \
         panic("wss_proc_ref: reference count for 0x%x was %d < 1", \
                 (unsigned int)(p), WSS_REF_COUNT((p)->wrp_refcount)); \
      WSS_MUTEX_ENTER((p)->wrp); \
      wsi_printf( PD_MAKESTR( \
                  "wss_proc_ref: pid %d (0x%x) refcount = %d [%s:%d]\n"), \
                  wsi_proc_pid((p)->wrp_p), (unsigned int)(p), \
                  WSS_REF_INC((p)->wrp_refcount), __FILE__, __LINE__ ); \
      WSS_MUTEX_EXIT((p)->wrp); \
   } while(0)

#  define wss_proc_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->wrp_refcount) < 1) \
         panic("wss_proc_rele: reference count for 0x%x was %d < 1", \
                 (unsigned int)(p), WSS_REF_COUNT(p->wrp_refcount)); \
      WSS_MUTEX_ENTER((p)->wrp); \
      if(WSS_REF_COUNT((p)->wrp_refcount) > 1) { \
         wsi_printf( PD_MAKESTR( \
                     "wss_proc_rele: pid %d (0x%x) refcount = %d [%s:%d]\n"), \
                     wsi_proc_pid((p)->wrp_p), (unsigned int)(p), \
                     WSS_REF_DEC((p)->wrp_refcount), __FILE__, __LINE__ ); \
         WSS_MUTEX_EXIT((p)->wrp); \
      } else { \
         wsi_printf( PD_MAKESTR( \
                    "wss_proc_rele: freeing pid %d (0x%x) [%s:%d]\n"), \
                     wsi_proc_pid((p)->wrp_p), (unsigned int)(p), \
                     __FILE__, __LINE__ ); \
         wss_proc_free(p); \
      } \
   } while(0)

#else /* DEBUG_REFS */

/*  wss_proc_ref */
/*! \brief Obtain reference to wrapper process structure
 *
 *  \param p Pointer to wrapper process structure
 */
#  define wss_proc_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->wrp_refcount) < 1) \
         panic("wss_proc_ref: reference count for 0x%x was %d < 1", \
                 (unsigned int)(p), WSS_REF_COUNT((p)->wrp_refcount)); \
      WSS_MUTEX_ENTER((p)->wrp); \
      WSS_REF_INC((p)->wrp_refcount); \
      WSS_MUTEX_EXIT((p)->wrp); \
   } while(0)

/*  wss_proc_rele */
/*! \brief Release reference to wrapper process structure
 *
 *  \param p Pointer to wrapper process structure
 */

#  define wss_proc_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->wrp_refcount) < 1) \
         panic("wss_proc_rele: reference count for 0x%x was %d < 1", \
                 (unsigned int)(p), WSS_REF_COUNT(p->wrp_refcount)); \
      WSS_MUTEX_ENTER((p)->wrp); \
      if(WSS_REF_COUNT((p)->wrp_refcount) > 1) { \
         WSS_REF_DEC((p)->wrp_refcount); \
         WSS_MUTEX_EXIT((p)->wrp); \
      } else { \
         wss_proc_free(p); \
      } \
   } while(0)

#endif /* DEBUG_REFS */


#endif /* WR_SUPPRESS_FUNCTION_DECL */

#ifdef __cplusplus
}
#endif

#endif /* __WSS_PROCESS_H */

