/*! \file wss_vm.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_vm.h,v 1.30 2001/04/06 18:08:00 stevek Exp $
 * 
 * \author  Doug Kilpatrick
 * \brief   Declaration of WSS virtual memory services
 * \version $Revision: 1.30 $
 * \date    $Date: 2001/04/06 18:08:00 $
 *
 * Uses WSI services to map the file into memory.
 *
 * \sa wss_vm.c
 */

#ifndef __WSS_VM_H 
#define __WSS_VM_H 1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif

#ifndef __WSI_TYPES_H
#include <wsi_types.h>
#endif

#ifndef __WSS_FILE_H
#include <wss_file.h>
#endif

#ifndef __WSS_MUTEX_H
#include <wss_mutex.h>
#endif

#ifndef __WSI_VM_H
#include <wsi_vm.h>
#endif

#ifndef _SYS_QUEUE_H_
#include <pdpi/queue.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/

typedef struct wss_vmspace        wss_vmspace_t;
typedef struct wss_vmspace_header wss_vmspace_header_t;
typedef u_int32_t                 wss_vmaddr_t;

/*! \brief Memory-mapped virtual memory space */
struct wss_vmspace
{
   const pd_char_t         *vms_name;
   wss_file_t               vms_file;
   unsigned int             vms_flags;
   size_t                   vms_size;
   wss_vmspace_header_t    *vms_head;     /*!> Pointer to VM space header */
   size_t                   vms_headsize; /*!> Size of header area        */
   void                    *vms_allocp;   /*!> Last allocated block       */
   WSS_MUTEX_PROTECT      ( vms );
   LIST_ENTRY(wss_vmspace)  vms_list;
};

#define WSS_VMS_MARKSWEEP 0x01

struct wss_vmspace_header
{
   u_int32_t    vms_head_version;
   size_t       vms_head_free;
   wss_vmaddr_t vms_head_base;    /*!> First aligned block after header */
   wss_vmaddr_t vms_head_trace;
   wss_vmaddr_t vms_head_reserved;
};


#define WSS_VMS_VERSION 1

#define VMCK_OK      0
#define VMCK_REINIT  1


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int           WSSLINKAGE  wss_vmck                ( wss_vmspace_t * );
void          WSSLINKAGE  wss_vm_mark             ( wss_vmspace_t *, void * );
void          WSSLINKAGE  wss_vm_sweep            ( wss_vmspace_t * );
int           WSSLINKAGE  wss_vm_valid_str        ( wss_vmspace_t *,
                                                    const pd_char_t * );
int           WSSLINKAGE  wss_vm_valid_alloced    ( wss_vmspace_t *,
                                                    void * );

int           WSSLINKAGE  wss_vm_init             ( void );
void          WSSLINKAGE  wss_vm_shutdown         ( void );

wss_vmspace_t WSSLINKAGE *wss_vm_create           ( const pd_char_t *,
                                                    const pd_char_t *, size_t,
                                                    size_t );
void          WSSLINKAGE  wss_vm_destroy          ( wss_vmspace_t * );

void          WSSLINKAGE *wss_vm_malloc           ( wss_vmspace_t *, unsigned );
void          WSSLINKAGE  wss_vm_free             ( wss_vmspace_t *, void * );
void          WSSLINKAGE  wss_vm_free_unsafe      ( wss_vmspace_t *, void * );
#ifdef DEBUG_VM
void          WSSLINKAGE *wss_vm_malloc_debug     ( wss_vmspace_t *, unsigned,
                                                    const pd_char_t *, int );
void          WSSLINKAGE  wss_vm_free_debug       ( wss_vmspace_t *, void *,
                                                    const pd_char_t *, int );
void          WSSLINKAGE  wss_vm_free_unsafe_debug( wss_vmspace_t *, void *,
                                                    const pd_char_t *, int );
#endif /* DEBUG_VM */
void          WSSLINKAGE *wss_vm_first_free_block ( wss_vmspace_t * );
void          WSSLINKAGE *wss_vm_next_free_block  ( wss_vmspace_t *,
                                                    void *previous_block);
size_t        WSSLINKAGE wss_vm_header_size       ( void );
size_t        WSSLINKAGE wss_vm_free_block_size   ( void *header);

wss_vmaddr_t             abs2rel( wss_vmspace_t *, void * );
void                    *rel2abs( wss_vmspace_t *, wss_vmaddr_t );


/***************************************************************************\
 * Functions implemented as macros                                         *
\***************************************************************************/

/*! Convert absolute address to relative within the mapped memory space */
#define abs2rel(space,addr) ((wss_vmaddr_t)((char*)(addr) - \
                             (char*)space->vms_head))
/*! Convert relative address within the mapped memory space to absolute */
#define rel2abs(space,addr) ((pd_char_t *)((char*)(addr) + \
                             (int)space->vms_head))

#ifdef DEBUG_VM
#define wss_vm_malloc(s,n)      wss_vm_malloc_debug((s),(n),__FILE__,__LINE__)
#define wss_vm_free(s,p)        wss_vm_free_debug((s),(p),__FILE__,__LINE__)
#define wss_vm_free_unsafe(s,p) wss_vm_free_unsafe_debug((s),(p),__FILE__,__LINE__)
#endif /* DEBUG_VM */


#ifdef __cplusplus
}
#endif

#endif  /* __WSS_VM_H */

