/*! \file wss_wrapper.h
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_wrapper.h,v 1.36 2001/02/23 19:02:47 stevek Exp $
 *
 * \brief   Prototypes and structures used by the module loading code to
 *          handle wrappers
 * \author  Chris Vance
 * \version $Revision: 1.36 $
 * \date    $Date: 2001/02/23 19:02:47 $
 *
 * \sa wss_wrapper.c
 */

#ifndef __WSS_WRAPPER_H
#define __WSS_WRAPPER_H 1

#ifndef WR_DEFS_INCLUDED
#include <wr_conf.h>
#endif

#ifndef __WSI_TYPES_H
#include <wsi_types.h>
#endif

#ifndef __WSS_REF_H
#include <wss_ref.h>
#endif

#ifndef __WSI_WRAPPER_H
#include <wsi_wrapper.h>
#endif

#ifndef __WSS_MUTEX_H
#include <wss_mutex.h>
#endif

#ifndef __WSS_IFACE_H
#include <wss_iface.h>
#endif


#ifdef __cplusplus
extern "C" {
#endif

struct wss_proc;

/***************************************************************************\
 * Public structures and types                                             *
\***************************************************************************/


typedef struct wss_wrapper          wss_wrapper_t;
typedef struct wss_wrapper_data     wss_wrapper_data_t;
typedef struct wss_wrapper_instance wss_wrapper_instance_t;

typedef int(*wss_wrapper_data_func_t)( struct wss_proc * );
typedef int(*wss_wrapper_func_t)     ( struct wss_proc *,
                                       wss_wrapper_data_t * );
typedef int(*wss_wrapper_dispatch_t) ( int, pd_syscall_id_t, void **, void *,
                                       struct wss_proc *,
                                       struct wss_wrapper_data *,
                                       int );

/*! \brief Structure for holding addition data in wrappers */
struct wss_wrapper_data
{
   int  *errnum; /*!< Pointer to memory to contain error code */
   void *wiv;    /*!< Pointer to wrapper instance variables   */
   int  skip;    /*!< Skip should be initialized to 0 (by the WSS) or in the 
                     beginning of the 
                     dispatch_(pre or post) func.  It can be modified by any
                     of the preX or postX function.  If its value changed
                     to 1, the dispatch_pre or post function will skip the
                     remaining preX or postX function and return to the WSS 
                     the dispatcher immediately. It is kind of a hack added 
                     to WDL to allow us to generate code to recongize tokens 
                     in the IDwrapper project.   However, this also could be
                     a general feature for wrapper to allow us to skip
                     the rest of the actions on the intercepted event easily.
                  */
   /* add addition data here */
};

/*! \brief This is the structure used by wrappers when they are loaded into
 *         the WSS
 *  Each wrapper has a one-to-one correspondence with a
 *  wr_module structure, representing loaded module.  Since the unload
 *  function was specific to the unloading of the module, and not the
 *  unload of a wrapper, it's pointer is stored in the wr_module not
 *  the wrapper.
 */
struct wss_wrapper
{
   /* Fields protected by global wss_wrapper_list lock */
   wss_wrapper_t            *w_next;          /*!< Pointer to next wrapper
                                                   in the list               */

   WSS_MUTEX_PROTECT        (w);              /*!< Mutex protecting
                                                   w_refcount, w_procs       */
   /* Fields protected by w lock */
   wss_ref_t                 w_refcount;      /*!< Structure reference count */
   int                       w_procs;         /*!< Number of wrapped
                                                   processes                 */

   /* Not protected by locks (exists same for life of instance) */
   struct wss_mod           *w_module;        /*!< Pointer to wrapper module
                                                   structure                 */
   int                       w_max_space;     /*!< Max event space accepted  */
   wss_wrapper_dispatch_t    w_dispatch_pre;  /*!< Pointer to pre-call
                                                   dispatcher function       */
   wss_wrapper_dispatch_t    w_dispatch_post; /*!< Pointer to post-call
                                                   dispatcher function       */
   wss_wrapper_func_t        w_activate;      /*!< Pointer to activation
                                                   function                  */
   wss_wrapper_func_t        w_deactivate;    /*!< Pointer to deactivation
                                                   function                  */
   wss_wrapper_func_t        w_duplicate;     /*!< Pointer to duplication
                                                   function                  */
   wss_wrapper_data_func_t   w_new_data;      /*!< Pointer to data creation
                                                   function                  */
};

/*! \brief Structure for an activated wrapper
 *
 *  This struct has a one-to-one relationship with a wrapped process.
 */
struct wss_wrapper_instance
{
   /* Fields protected by wrp_wi_list in wss_proc structure */
   wss_wrapper_instance_t *wi_next;           /*!< Pointer to next wrapper
                                                   instance                  */
   wss_wrapper_instance_t *wi_prev;           /*!< Pointer to previous
                                                   wrapper instance          */

   WSS_MUTEX_PROTECT      (wi);
   /* Fields protected by wi lock */
   wss_ref_t               wi_refcount; /* Reference counter                */
   volatile int            wi_active:1; /* Bool: has activate() been called */
   void                   *wi_data;     /* Local data.                      */

   /* Not protected by locks (exists same for life of instance) */
   wss_wrapper_t          *wi_wrapper;  /* Pointer to the actual wrapper    */
};



/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

extern wss_wrapper_t WSSLINKAGE *wss_wrapper_list;
WSS_MUTEX_EXPORT               ( wss_wrapper_list );


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

int           WSSLINKAGE  wss_wrapper_add_new           __P(( pd_char_t * ));
int           WSSLINKAGE  wss_wrapper_call_activation   __P(( struct wss_proc * ));
int           WSSLINKAGE  wss_wrapper_call_deactivation __P(( struct wss_proc * ));
int           WSSLINKAGE  wss_wrapper_call_duplication  __P(( struct wss_proc * ));
int           WSSLINKAGE  wss_wrapper_can_unload        __P(( wss_id_t id ));
int           WSSLINKAGE  wss_wrapper_delete            __P(( wss_id_t id ));
#ifdef DEBUG_REFS
wss_wrapper_t WSSLINKAGE *wss_wrapper_find_debug        __P(( const pd_char_t *,
                                                              const pd_char_t *,
                                                              int ));
#endif /* DEBUG_REFS */
wss_wrapper_t WSSLINKAGE *wss_wrapper_find              __P(( const pd_char_t * ));
void          WSSLINKAGE  wss_wrapper_free              __P(( wss_wrapper_t * ));
void          WSSLINKAGE  wss_wrapper_init              __P(( void ));
void          WSSLINKAGE  wss_wrapper_instance_free     __P(( wss_wrapper_instance_t * ));
void          WSSLINKAGE  wss_wrapper_instance_ref      __P(( wss_wrapper_instance_t * ));
void          WSSLINKAGE  wss_wrapper_instance_rele     __P(( wss_wrapper_instance_t * ));
int           WSSLINKAGE  wss_wrapper_loaded            __P(( pd_char_t *name ));
void          WSSLINKAGE  wss_wrapper_ref               __P(( wss_wrapper_t * ));
int           WSSLINKAGE  wss_wrapper_register          __P(( struct wss_mod *,
                                                              void *, void *,
                                                              void *, void *,
                                                              void * ));
void          WSSLINKAGE  wss_wrapper_rele              __P(( wss_wrapper_t * ));
void          WSSLINKAGE  wss_wrapper_shutdown          __P(( void ));


/***************************************************************************\
 * Functions implemented as macros                                         *
\***************************************************************************/

#define WSS_GET_WIV_VAR(var) (((struct wr_wiv*)(data->wiv))->var)
#define WSS_SAVE_WIV(newwiv) (data->wiv = (newwiv))
#define WSS_GET_WIV()        ((struct wr_wiv*)(data->wiv))

#ifdef DEBUG_REFS
#  define wss_wrapper_find(p) wss_wrapper_find_debug((p),__FILE__,__LINE__)
#endif /* DEBUG_REFS */

#if defined(DEBUG_REFS) && !defined(DISABLE_WRAP_DEBUG_REFS)

#  define wss_wrapper_ref_debug(p,f,l) \
   do { \
      if(WSS_REF_COUNT((p)->w_refcount) < 1) \
         panic("wss_wrapper_ref: reference count for 0x%x was %d < 1 [%s:%d]", \
               (unsigned int)(p), WSS_REF_COUNT((p)->w_refcount), (f), (l)); \
      WSS_MUTEX_ENTER((p)->w); \
      wsi_printf(PD_MAKESTR("wss_wrapper_ref: wrapper #%d (0x%x) ") \
                 PD_MAKESTR("refcount = %d [%s:%d]\n"), \
                 (p)->w_module->wrm_id, (unsigned int)(p), \
                 WSS_REF_INC_DEBUG((p)->w_refcount, (f), (l)), (f), (l)); \
      WSS_MUTEX_EXIT((p)->w); \
   } while(0)

#  define wss_wrapper_ref(p) wss_wrapper_ref_debug((p), __FILE__, __LINE__)

#  define wss_wrapper_rele_debug(p,f,l) \
   do { \
      if(WSS_REF_COUNT((p)->w_refcount) < 1) \
         panic("wss_wrapper_rele: reference count for 0x%x was %d < 1 " \
               "[%s:%d]", (unsigned int)(p), \
               WSS_REF_COUNT((p)->w_refcount), (f), (l)); \
      WSS_MUTEX_ENTER((p)->w); \
      wsi_printf( PD_MAKESTR( \
         "wss_wrapper_rele: wrapper #%d (0x%x) refcount = %d [%s:%d]\n"),\
              (p)->w_module->wrm_id, (unsigned int)(p), \
              WSS_REF_COUNT((p)->w_refcount),(f),(l)); \
      if(WSS_REF_COUNT((p)->w_refcount) > 1) { \
         WSS_REF_DEC_DEBUG((p)->w_refcount, (f), (l)); \
         WSS_MUTEX_EXIT((p)->w); \
      } else { \
         wss_wrapper_free(p); \
      } \
   } while(0)

#  define wss_wrapper_rele(p) wss_wrapper_rele_debug((p), __FILE__, __LINE__)

#else /* DEBUG_REFS */

#  define wss_wrapper_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->w_refcount) < 1) \
         panic("wss_wrapper_ref: reference count for 0x%x was %d < 1", \
               (unsigned int)(p), WSS_REF_COUNT((p)->w_refcount)); \
      WSS_MUTEX_ENTER((p)->w); \
      WSS_REF_INC((p)->w_refcount); \
      WSS_MUTEX_EXIT((p)->w); \
   } while(0)

#  define wss_wrapper_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->w_refcount) < 1) \
         panic("wss_wrapper_rele: reference count for 0x%x was %d < 1", \
               (unsigned int)(p), WSS_REF_COUNT((p)->w_refcount)); \
      WSS_MUTEX_ENTER((p)->w); \
      if(WSS_REF_COUNT((p)->w_refcount) > 1) { \
         WSS_REF_DEC((p)->w_refcount); \
         WSS_MUTEX_EXIT((p)->w); \
      } else { \
         wss_wrapper_free(p); \
      } \
   } while(0)

#endif /* DEBUG_REFS */


#if defined(DEBUG_REFS) && !defined(DISABLE_WRAPINST_DEBUG_REFS)

#  define wss_wrapper_instance_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->wi_refcount) < 1) \
         panic("wss_wrapper_instance_ref: reference count for 0x%x was %d < 1",\
               (unsigned int)(p), WSS_REF_COUNT((p)->wi_refcount)); \
      WSS_MUTEX_ENTER((p)->wi); \
      wsi_printf( PD_MAKESTR("wss_wrapper_instance_ref: wrapper instance ") \
                  PD_MAKESTR("#%d (0x%x) refcount = %d [%s:%d]\n"), \
                  (p)->wi_wrapper->w_module->wrm_id, (unsigned int)(p), \
                  WSS_REF_INC((p)->wi_refcount), __FILE__, __LINE__ ); \
      WSS_MUTEX_EXIT((p)->wi); \
   } while(0)

#  define wss_wrapper_instance_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->wi_refcount) < 1) \
         panic("wss_wrapper_instance_rele: reference count for 0x%x was %d " \
               "< 1", (unsigned int)(p), WSS_REF_COUNT((p)->wi_refcount)); \
      WSS_MUTEX_ENTER((p)->wi); \
      if(WSS_REF_COUNT((p)->wi_refcount) > 1) { \
         wsi_printf( PD_MAKESTR("wss_wrapper_instance_rele: wrapper ") \
                     PD_MAKESTR("instance #%d (0x%x) refcount = %d ") \
                     PD_MAKESTR("[%s:%d]\n"), \
                     (p)->wi_wrapper->w_module->wrm_id, (unsigned int)(p), \
                     WSS_REF_DEC((p)->wi_refcount), __FILE__, __LINE__ ); \
         WSS_MUTEX_EXIT((p)->wi); \
      } else { \
         wss_wrapper_instance_free(p); \
      } \
   } while(0)

#else /* DEBUG_REFS */

#  define wss_wrapper_instance_ref(p) \
   do { \
      if(WSS_REF_COUNT((p)->wi_refcount) < 1) \
         panic("wss_wrapper_instance_ref: reference count for 0x%x was %d " \
               "< 1", (unsigned int)(p), WSS_REF_COUNT((p)->wi_refcount)); \
      WSS_MUTEX_ENTER((p)->wi); \
      WSS_REF_INC((p)->wi_refcount); \
      WSS_MUTEX_EXIT((p)->wi); \
   } while(0)

#  define wss_wrapper_instance_rele(p) \
   do { \
      if(WSS_REF_COUNT((p)->wi_refcount) < 1) \
         panic("wss_wrapper_instance_rele: reference count for 0x%x was %d " \
               "< 1", (unsigned int)(p), WSS_REF_COUNT((p)->wi_refcount)); \
      WSS_MUTEX_ENTER((p)->wi); \
      if(WSS_REF_COUNT((p)->wi_refcount) > 1) { \
         WSS_REF_DEC((p)->wi_refcount); \
         WSS_MUTEX_EXIT((p)->wi); \
      } else { \
         wss_wrapper_instance_free(p); \
      } \
   } while(0)

#endif /* DEBUG_REFS */

#ifdef __cplusplus
}
#endif

#endif /* __WSS_WRAPPER_H */

