/*! \file wss_arg.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_arg.c,v 1.28 2001/02/23 19:02:46 stevek Exp $
 *
 * \brief   Allow for replacement of call arguments from within a wrapper
 * \author  Chris Vance, Doug Kilpatrick
 * \version $Revision: 1.28 $
 * \date    $Date: 2001/02/23 19:02:46 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_arg.h>

#include <wss_utilities.h>


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*! \brief argument node type */
typedef struct arg_node arg_node_t;

/*! \brief argument node structure */
struct arg_node
{
    wss_util_rbnode_t   parent;   /*!< Parent red/black tree node.
                                       Key is handle/space/code. */
    int                 handle;   /*!< Thread id.  Needed to support
                                       multi-threaded platforms. */
    int                 space;    /*!< Event space. Needed to support
                                       active composition/synthetic events */
    pd_syscall_id_t     code;     /*!< Event (call) code. */
    wss_arg_t          *arg_list; /*!< Argument stack. */
};

static int
arg_compfunc(void *n1, void *n2)
{
    arg_node_t *node1 = n1;
    arg_node_t *node2 = n2;
    return (node1->handle > node2->handle ? 1  : 
            node1->handle < node2->handle ? -1 :
            node1->space  > node2->space  ? 1  :
            node1->code   < node2->code   ? -1 :
            node1->code   > node2->code   ? 1  :
            0);
}

static void *
arg_clonefunc(void *n)
{
    arg_node_t *retval = wss_malloc(sizeof(*retval));
    wss_util_rbnode_init(&(retval->parent));
    if (!n) {
        retval->handle = retval->space = retval->code = 0;
        retval->arg_list = 0;
    } else {
        arg_node_t *orig = n;
        retval->handle = orig->handle;
        retval->space  = orig->space ;
        retval->code   = orig->code  ;
        retval->arg_list = orig->arg_list;
    }
    return retval;
}

static wss_arg_t *
get_arglist(wss_proc_t *wrp, int handle, int space, pd_syscall_id_t code)
{
    arg_node_t temp, *results = NULL;
    
    if( wrp->wrp_arg_list )
    {
        wss_util_rbnode_init(&(temp.parent));
        temp.handle = handle;
        temp.space  = space;
        temp.code   = code;

        results = (arg_node_t *)wss_util_rbtree_find( wrp->wrp_arg_list,
                                                      &(temp.parent) );
    }
    
    return( results ? results->arg_list : NULL );
}

static void
set_arglist(wss_proc_t *wrp, int handle, int space, pd_syscall_id_t code,
            wss_arg_t *new)
{
    arg_node_t *new_node;
    
    /* Possible shortcut */
    if (!wrp->wrp_arg_list && !new)
        return;

    new_node = arg_clonefunc(NULL);
    new_node->handle    = handle;
    new_node->space     = space;
    new_node->code      = code;
    new_node->arg_list  = new;

    if (!wrp->wrp_arg_list) {
        /* Create a tree to hold the new list */
#ifdef MALLOC_DEBUG
        wrp->wrp_arg_list = wss_util_rbtree_new(
                arg_compfunc, arg_clonefunc, wss_free_debug);
#else
        wrp->wrp_arg_list = wss_util_rbtree_new(
                arg_compfunc, arg_clonefunc, wss_free);
#endif
    } else {
        /* Regardless of if new is NULL, we still have to remove
         * anything from that handle+space+code.
         * We don't deal with deep memory management... That's up to
         * the wss_arg functions that use this function
         * Just delete the node structure */
        wss_util_rbnode_t *old =
            wss_util_rbtree_delete(wrp->wrp_arg_list, &(new_node->parent));
        if (old)
            wss_free(old);

    }
    if (new)
        wss_util_rbtree_insert(wrp->wrp_arg_list, &(new_node->parent));
    else
        wss_free(new_node);
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_arg_init */
/*! \brief  Argument replacement module initialization routine
 *
 *  Place holder for an initialization function.  It may never be needed...
 */
void EXPORT
wss_arg_init()
{
   /* TBD: Nothing here yet! */
}


/*  wss_arg_shutdown */
/*! \brief  Argument replacement module shutdown routine
 *
 *  Place holder for a shutdown function.  It also may never be needed...
 */
void EXPORT
wss_arg_shutdown()
{
   /* TBD: Nothing here yet! */
}


/*  wss_arg_donesyscall */
/*!
 *  \brief Restores argument replacements at call completion
 *
 *  When the call is done, go through the list of saved arguments
 *  (in this case, the list is really a stack) undoing the replacements
 *  one by one.  If a single parameter was replaced multiple times, it
 *  will keep popping the replacements off the "stack" until things are
 *  back to the way they were in the beginning. 
 *
 *  \param wrp    Pointer to the wrapper process structure for this process
 *  \param handle Identifier of thread making call (0 on non-threaded platforms)
 *  \param space  Event space of call
 *  \param code   Code indentifier for call
 *  \param uap    Arguments of call
 */
void EXPORT
wss_arg_donesyscall( struct wss_proc *wrp, int handle, int space,
                     pd_syscall_id_t code, void *uap[] )
{
   struct wss_arg *a, *b;

   WSS_MUTEX_ENTER(wrp->wrp_arg_list);
   a = get_arglist(wrp, handle, space, code);
   while( a )
   {
      b = a;
      a = a->next;

      if( b->size != 0 )
         wsi_curproc_free( uap[b->index], b->size );
      uap[b->index] = b->old;
      wss_free( b );
   }
   set_arglist(wrp, handle, space, code, NULL);
   WSS_MUTEX_EXIT(wrp->wrp_arg_list);
}


/*  wss_arg_add */
/*! \brief Add a new argument to the argument stack.
 *
 *  \param wrp    Pointer to the wrapper process structure for this process
 *  \param handle Identifier of thread making call (0 on non-threaded platforms)
 *  \param space  Event space of call
 *  \param code   Code indentifier for call
 *  \param uap    Arguments of call
 *  \param index  Index of parameter to push on the stack
 *  \param size   If parameter is a pointer, this is the size of the memory
 *                pointed to by \a uap[index]
 *
 *  \return 0 on success,
 *  \return -1 on an error
 */
int EXPORT
wss_arg_add( struct wss_proc *wrp, int handle, int space, pd_syscall_id_t code,
            void *uap[], int index, size_t size )
{
   struct wss_arg *newarg;
   struct wss_arg *arglist;

   newarg = (struct wss_arg*)wss_malloc(sizeof(*newarg));
   if( newarg == NULL )
      return( -1 );
      
   newarg->index = index;
   newarg->old = uap[index];
   newarg->size = size;

   /* Blindly add to the head of the list. */
   WSS_MUTEX_ENTER(wrp->wrp_arg_list);
   arglist = get_arglist(wrp, handle, space, code);
   newarg->next = arglist;
   set_arglist(wrp, handle, space, code, newarg);
   WSS_MUTEX_EXIT(wrp->wrp_arg_list);
   
   return 0;
}

/*  wss_arg_replace */
/*! \brief Replace an argument in a call.
 *
 *  This function stores the old parameter on the replacement stack
 *  for restoration upon call completion.
 *
 *  \param wrp     Pointer to the wrapper process structure for this process
 *  \param handle  Identifier of thread making call (0 on non-threaded
 *                 platforms)
 *  \param space   Event space of call
 *  \param code    Code indentifier for call
 *  \param uap     Arguments of call
 *  \param new_arg Argument to be set (or copied) as the replacement
 *  \param index   Index of parameter to push on the stack
 *  \param size    If parameter is a pointer, this is the size of the memory
 *                 pointed to by \a uap[index]
 *
 *  \return 0 on success,
 *  \return -1 on error
 */
int EXPORT
wss_arg_replace( struct wss_proc *wrp, int handle, int space,
                 pd_syscall_id_t code, void *uap[], void *new_arg, int index,
                 size_t size )
{
   int error;

   error = wss_arg_add( wrp, handle, space, code, uap, index, size );
   if( error != 0 )
      return -1;
   
   if( size )
   {
      pd_byte_t *arg = (pd_byte_t *)wsi_curproc_malloc( size );

      copyout( new_arg, arg, size );

      uap[index] = arg;
   }
   else
      uap[index] = new_arg;

   return 0;
}

/*  wss_arg_copy */
/*! \brief Replace an argument in a call.
 *
 *  Just like wss_arg_replace, except it doesn't keep track of the old
 *  values, so you can't restore the old values.
 * 
 *  This exists because "out" parameters of function calls should NOT be
 *  restored.
 * 
 *  \param uap     Arguments of call
 *  \param new_arg Argument to be set (or copied) as the replacement
 *  \param index   Index of parameter to push on the stack
 *  \param size    If parameter is a pointer, this is the size of the memory
 *                 pointed to by \a uap[index]
 *
 *  \warning This function will "Do The Wrong Thing" if the new value takes
 *  \warning more space than the old value... I don't think this is avoidable.
 *
 *  \return 0 on success,
 *  \return -1 on error
 */
int EXPORT
wss_arg_copy( void *uap[], void *new_arg, int index, size_t size )
{
    if( size )
        copyout( new_arg, uap[index], size );
    else
        uap[index] = new_arg;
    return 0;
}


/*  wss_arg_dupin */
/*! \brief Wrapper version of "copyin".
 *
 *  Copies the block of memory in the current process pointed to by \a src
 *  into kernel space.
 *
 *  Needs to know the \a size of the data... the characterization data usually
 *  has this.
 *
 *  \param src  Pointer to data to by duplicated
 *  \param size Size of data to be duplicated
 *
 *  \return NULL on error,
 *  \return otherwise a pointer to the data in kernel space
 */
void EXPORT *
wss_arg_dupin( void *src, size_t size )
{
    int error;
    void *dst = wss_malloc( size );
    if( !dst )
        return NULL;
    error = copyin( src, dst, size );
    if( error )
    {
        wss_free( dst );
        return NULL;
    }
    return dst;
}

int EXPORT
wss_arg_replace_ntermlist(wss_proc_t *wrp, void *uap[], void *n_arg, int index)
{

    /* Um. .. Um ... Um ... Ow? */
#ifndef WR_WINNT
#warning TBD: Impliment replace-ntermlist.  This is such
#warning such a bad idea...  where to put new,
#warning event-space awareness, ...
#endif
    return 0;
}

void EXPORT wss_arg_free_ntermlist(void *kaddr)
{
    /* Created by dupin_ntermlist, so assume its well formed.
     * (I'm gonna eat these words, I know it... )
     */
    pd_char_t **k_argv = (pd_char_t **)kaddr;
    int i;

    for (i = 0 ; k_argv[i] != NULL && i < 100; ++i) {
        wss_free(k_argv[i]);
    }
    wss_free(k_argv);
}

void EXPORT *
wss_arg_dupin_ntermlist(void *src)
{
    /* This function returns an array of pointers to strings.  The
     * array is allocated, as are the strings that are pointed to.  The 
     * array is terminated by a null pointer to the string.
     */
    int count = 0;
    pd_char_t **u_argv = src;
    pd_char_t **k_argv = 0;
    pd_char_t **argv_i = (pd_char_t **)src;

    if (u_argv == NULL)
        return NULL;

    /* Step one, count argc */
    for (;count < 100; ++count) { /* Arbitrary limit */
        pd_char_t *p = NULL;
        int error;

        /* copy in one step at a time. */
        error = copyin((void*)argv_i, (void*)(&p), sizeof(p));
        if (error || !p)
            break;
        argv_i++;
    }

    /* Allocate the new array to hold the result */
    k_argv = (pd_char_t**)wss_malloc(sizeof(*k_argv) * (count+1));
    if (k_argv == NULL) return NULL;
    /* Terminate the list */
    k_argv[count] = NULL;
    
    /* Copy them in */
    for (--count;count >= 0; --count) {
        int error;

        k_argv[count] = wss_malloc(1024);
        error = copyinstr(u_argv[count], k_argv[count], 1024, NULL);
        if (error)
            goto error;
    }

    return k_argv;
error:
    wsi_printf(PD_MAKESTR("Error copying in string %d.\n"),count);
    /* Clean up and return NULL */
    for (; count < 100 && k_argv[count] != NULL; ++count)
        k_argv[count] = NULL;
    wss_free(k_argv);
    return NULL;
}

/*  wss_arg_dupout */
/*! \brief Wrapper version of "copyout".
 *
 *  Copies the block of kernel memory pointed to by \a src into the
 *  current process' address space.
 *
 *  The current process will have the memory required added via
 *  wsi_curproc_malloc(), the current implimentations of which allocate
 *  full pages per request.
 *
 *  \param src  Pointer to data to by duplicated
 *  \param size Size of data to be duplicated
 *
 *  \warning If the current process is not the process that will use
 *  \warning this data, you will need to hack things.
 * 
 *  \return NULL on error,
 *  \return otherwise a pointer to the data in the current processes space
 */
void EXPORT *
wss_arg_dupout( void *src, size_t size )
{
    int error;
    void *dst = wsi_curproc_malloc( size );
    if( !dst )
        return NULL;
    error = copyout( src, dst, size );
    if( error )
    {
        wsi_curproc_free( dst, size );
        return NULL;
    }
    return dst;
}

