/*! \file wss_criteria.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_criteria.c,v 1.30 2001/06/22 04:35:18 stevek Exp $
 *
 * \brief   Implementation code for the activation and deactivation criteria.
 * \author  Chris Vance, Doug Kilpatrick
 * \version $Revision: 1.30 $
 * \date    $Date: 2001/06/22 04:35:18 $
 */ 

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_criteria.h>
#include <wss_module.h>
#include <wss_malloc.h>

#include <eventfs.h>

#include <wrerrno.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern struct wss_mod *wss_module_list;


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/*! Activation criteria list */
wss_crit_t EXPORT *wss_act_crit_list = NULL;

/*! Mutex which protects the activation criteria list */
WSS_MUTEX_PROTECT( wss_act_crit_list );

/*! Deactivation criteria list */
wss_crit_t EXPORT *wss_deact_crit_list = NULL;

/*! Mutex which protects the deactivation criteria list */
WSS_MUTEX_PROTECT( wss_deact_crit_list );

/*! Flag to indicate if the WSS should process criteria */
pd_bool_t EXPORT   wss_crit_admin_done = FALSE;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_crit_init */
/*! \brief Criteria module initialization routine
 *
 * Initializes the mutexes protecting the activation and deactivation
 * criteria lists.
 */
void EXPORT
wss_crit_init()
{
   WSS_MUTEX_INIT( wss_act_crit_list );
   WSS_MUTEX_INIT( wss_deact_crit_list );
}


/*  wss_crit_shutdown */
/*! \brief Criteria module shutdown routine
 *
 * Destroys the mutexes protecting the activation and deactivation
 * criteria lists.
 */
void EXPORT
wss_crit_shutdown()
{
   WSS_MUTEX_DESTROY( wss_act_crit_list );
   WSS_MUTEX_DESTROY( wss_deact_crit_list );
}


/*  wss_crit_check_activation */
/*! \brief Evaulate activation criteria
 *
 *  Call activation criteria to see if process should be wrapped.
 *
 *  \param wrp Pointer to the wrapper process structure for this process
 *
 *  \return TRUE if process is now wrapped,
 *  \return FALSE otherwise
 */
int EXPORT
wss_crit_check_activation( wss_proc_t *wrp )
{
   void (*activate)( wss_proc_t *, wss_crit_t * );

   wss_crit_t *atmp;

   if( !wss_crit_admin_done )
      return FALSE;

   WSS_MUTEX_ENTER( wss_act_crit_list );
   atmp = wss_act_crit_list;
   while( atmp )
   {
      if( atmp->crit_status == WRACT_ENABLE )
      {
         activate = atmp->crit_func;
         (*activate)( wrp, atmp );
      }
      atmp = atmp->crit_next;
   }
   WSS_MUTEX_EXIT( wss_act_crit_list );

   return( ( wrp->wrp_wi_list != NULL ) ? TRUE : FALSE );
}


/*  wss_crit_check_deactivation */
/*! \brief Evaulate deactivation criteria
 *
 *  Call deactivation criteria to see if any wrappers should be deactivated
 *  for this process.
 *
 *  \param wrp Pointer to the wrapper process structure for this process
 *
 *  \return TRUE if process is no longer wrapped,
 *  \return FALSE otherwise
 */
int EXPORT
wss_crit_check_deactivation( wss_proc_t *wrp )
{
   void (*deactivate)( wss_proc_t *, wss_crit_t * );

   wss_crit_t *atmp;

   if( !wss_crit_admin_done )
      return FALSE;

   WSS_MUTEX_ENTER( wss_deact_crit_list );
   atmp = wss_deact_crit_list;
   while( atmp )
   {
      if( atmp->crit_status == WRACT_ENABLE )
      {
         deactivate = atmp->crit_func;
         (*deactivate)( wrp, atmp );
      }
      atmp = atmp->crit_next;
   }
   WSS_MUTEX_EXIT( wss_deact_crit_list );

   return( ( wrp->wrp_wi_list == NULL ) ? TRUE : FALSE );
}


/*  wss_crit_add */
/*! \brief Add a criteria to the activation or deactivation list
 *
 *  Adds a record to the kernel's list of criteria description records.
 *  The criteria's decision function is called by the WSS at exec-time.
 *  The prototype of the decision function is:
 *
 *  \code
 *    void activate( proc_t *p, wss_proc_t *wrp, wss_wrapper_data_t *data );
 *  \endcode
 *    
 *  \param df Pointer to the criterion's decision function.
 *
 *  \return 0 on success,
 *  \return ENOMEM on allocation failure
 */
int EXPORT
wss_crit_add( wss_mod_t *wrm, void *df, pd_char_t **imp_list )
{
   wss_crit_t *crit;
   pd_char_t **imp;

   if( ( crit = wss_malloc( sizeof( *crit ) ) ) == NULL )
      return ENOMEM;

   wss_ref_init( &(crit->crit_refcount) );
   wss_mod_ref( wrm );
   crit->crit_module = wrm;
   crit->crit_func = df;
   crit->crit_implist = imp_list;
   /* Enabled by default... for now */
   crit->crit_status = WRACT_ENABLE;
   WSS_MUTEX_INIT( crit->crit );

   /* Reference the dependency modules */
   for( imp = (pd_char_t **)crit->crit_implist; *imp; imp++ )
      wss_mod_find_by_name( *imp );

   /* blindly prepend to list */
   if( wrm->wrm_type == WR_MOD_ACTIVATION )
   {
      WSS_MUTEX_ENTER( wss_act_crit_list );
      crit->crit_next = wss_act_crit_list; /* assumed NULL initially */
      wss_act_crit_list = crit;
      WSS_MUTEX_EXIT( wss_act_crit_list );
   }
   else
   {
      WSS_MUTEX_ENTER( wss_deact_crit_list );
      crit->crit_next = wss_deact_crit_list;  /* assumed NULL initially */
      wss_deact_crit_list = crit;
      WSS_MUTEX_EXIT( wss_deact_crit_list );
   }

   return 0;
}

/*  wss_crit_can_unload */
/*! \brief Determines if criteria module is in proper state to be unloaded
 *
 *  \param id Indentifier of criteria module
 *
 *  \return TRUE if wrapper can be unloaded,
 *  \return FALSE otherwise
 */
int EXPORT
wss_crit_can_unload( wss_id_t id )
{
   wss_crit_t    *crit;
   pd_byte_t     deact = 0;

   /* See if the request is an activation criteria */
   WSS_MUTEX_ENTER( wss_act_crit_list );
   for( crit = wss_act_crit_list;
        crit && ( crit->crit_module->wrm_id != id );
        crit = crit->crit_next )
   ;

   if( crit == NULL )
   {
      WSS_MUTEX_EXIT( wss_act_crit_list );

      /* See if the request is a deactivation criteria */
      deact = 1;
      WSS_MUTEX_ENTER( wss_deact_crit_list );
      for( crit = wss_deact_crit_list;
           crit && ( crit->crit_module->wrm_id != id );
           crit = crit->crit_next )
      ;

      /* Return error if we didn't find it */
      if( crit == NULL )
      {
         WSS_MUTEX_EXIT( wss_deact_crit_list );
         return TRUE;
      }
   }

   /*
    * Can only remove a criteria if the refcount is one (or less).
    */
   if( WSS_REF_COUNT(crit->crit_refcount) > 1 )
   {
      if( deact )
         WSS_MUTEX_EXIT( wss_deact_crit_list );
      else
         WSS_MUTEX_EXIT( wss_act_crit_list );
      return FALSE;
   }

   if( deact )
      WSS_MUTEX_EXIT( wss_deact_crit_list );
   else
      WSS_MUTEX_EXIT( wss_act_crit_list );

   return TRUE;
}

/*  wss_crit_delete */
/*! \brief Removes a record that has been added via wss_crit_add.
 *
 *  \param id Indentifier of criteria module
 *
 *  \return 0 on success,
 *  \return \b WRE_NOTFOUND if \a id references a non-existant criteria module,
 *  \return \b WRE_REF if the criteria module is being referenced by something
 */
int EXPORT
wss_crit_delete( wss_id_t id )
{
   wss_crit_t  *crit;
   wss_crit_t **prev_crit_ptr = &wss_act_crit_list;
   pd_char_t  **imp;
   pd_byte_t    deact         = 0;

   /* See if the request is an activation criteria */
   WSS_MUTEX_ENTER( wss_act_crit_list );
   for( crit = wss_act_crit_list;
        crit && ( crit->crit_module->wrm_id != id );
        prev_crit_ptr = &(crit->crit_next), crit = crit->crit_next )
      continue;

   if( crit == NULL )
   {
      WSS_MUTEX_EXIT( wss_act_crit_list );

      /* See if the request is a deactivation criteria */
      deact = 1;
      WSS_MUTEX_ENTER( wss_deact_crit_list );
      prev_crit_ptr = &wss_deact_crit_list;
      for( crit = wss_deact_crit_list;
           crit && ( crit->crit_module->wrm_id != id );
           prev_crit_ptr = &(crit->crit_next), crit = crit->crit_next )
         continue;

      /* Return error if we didn't find it */
      if( crit == NULL )
      {
         WSS_MUTEX_EXIT( wss_deact_crit_list );
         return WRE_NOTFOUND;
      }
   }

   /* This check should always fail for now.  One day activation
      criteria may depend on each other and on that day this test
      will be significant! */
   if( WSS_REF_COUNT(crit->crit_refcount) > 1 )
   {
      if( deact )
         WSS_MUTEX_EXIT( wss_deact_crit_list );
      else
         WSS_MUTEX_EXIT( wss_act_crit_list );
      return WRE_REF;
   }

   /* Remove references the dependency modules */
   for( imp = (pd_char_t **)crit->crit_implist; *imp; imp++ )
   {
      wss_mod_t *wrm = wss_mod_find_by_name( *imp );
      if( wrm )
      {
         /* Release dependency reference */
         wss_mod_rele( wrm );

         /* Release reference from finding */
         wss_mod_rele( wrm );
      }
   }

   /* Release reference to module */
   wss_mod_rele( crit->crit_module );

   /* Remove it */
   *prev_crit_ptr = crit->crit_next;

   /* Unlock the appropriate list */
   if( deact )
      WSS_MUTEX_EXIT( wss_deact_crit_list );
   else
      WSS_MUTEX_EXIT( wss_act_crit_list );

   /* Release reference to criteria */
   wss_crit_rele( crit );

   return 0;
}

/*  wss_crit_free */
/*! \brief Free memory used by criteria structure
 *
 *  \param crit Pointer to the criteria structure
 */
void
wss_crit_free( wss_crit_t *crit )
{
   assert( crit != NULL );
   assert( WSS_REF_COUNT(crit->crit_refcount) == 1 );

   WSS_REF_DEC(crit->crit_refcount);

#ifdef DEBUG_CRITERIA
   wsi_printf( PD_MAKESTR("wss_crit_free: releasing memory for 0x%x\n"),
               (unsigned int)crit );
#endif /* DEBUG_CRITERIA */

   wss_ref_destroy( &(crit->crit_refcount) );

   WSS_MUTEX_DESTROY( crit->crit );
   wss_free( crit );
}

