/*! \file wss_db.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_db.c,v 1.96 2001/06/15 22:36:06 stevek Exp $ 
 *
 * \brief   Wrapper database subsystem
 * \author  Lee Badger, Doug Kilpatrick, Steve Kiernan
 * \version $Revision: 1.96 $
 * \date    $Date: 2001/06/15 22:36:06 $
 * 
 * A cut at the WQL interface After some thought, I realized that
 * designing the WQL interface is easier if the internal structure is
 * designed at the same time.  The reason is that the
 * select/delete/update operator row selection criteria clauses could
 * be complicated, and that doing them properly (and fast) would be
 * much easier if WDL generated a function that does the db-internal
 * selection criteria tests, and passed a pointer to that function
 * into the WQL interface.  This is also true for the update function.
 * The WDL parser is well-suited for creating such functions.  The
 * alternative would be to make the DB subsystem more complex and
 * slower since it would have to include an interpreter for selection
 * criteria and update expressions.
 * 
 * (Commentary from DSK:
 *   The way that things actually went is to create a structure that
 *   represents the criteria in its creation.
 *   A similar approach is used by update.  This is probally not too
 *   quick.)
 *
 * The DB will be locked during all accesses.  All DB routines return
 * \b copies of data, not references.  While this may be slower if
 * large objects get stored, it keeps dbrow variables tranquil, which
 * I think is necessary for sanity.  The VM region will contain a
 * simple memory allocator so that all structures, strings, etc., that
 * comprise the DB are included when the DB is paged to disk.  No
 * garbage collection, though possibly memory compaction at a later
 * time.
 */


#include <wr_conf.h>

#if !defined(USER_MODE_TEST) && !defined(DBINTD_INTERFACE)
#include <wsi_stdlib.h>
#include <wss_debug.h>
#include <wss_db.h>
#include <wss_malloc.h>
#include <wss_wrvevent.h>
#include <wss_vm.h>

#include <wrerrno.h>
#include <wss_utilities.h>

#ifdef WR_LINUX
#include <wss_dispatch.h>
#endif

#else
#include <stdarg.h>
#include <wsi_stdlib.h>
#include <wss_debug.h>
#include <stdio.h>
#include <stdlib.h>
#define wss_atoi(s) atoi(s)
#define wss_atof(s) atof(s)
#ifdef WR_UNIX
#include <unistd.h>
#endif
#include <stdarg.h>

#ifdef USER_MODE_TEST
#include <pdpi/queue.h>
#include <wss_malloc.h>
#include <wss_vm.h>
#endif /* USER_MODE_TEST */
#include <wss_db.h>
#include <wss_wrvevent.h>
#include <wrerrno.h>

#ifndef TRUE
#define TRUE 1
#endif
#endif

#ifdef DBINTD_INTERFACE
#include "dbinterface.h"
#endif /* DBINTD_INTERFACE */

#ifndef __GNUC__
#define __inline__
#endif

/*
 * Keeping in mind that we aren't trying to build a "real" DB, and
 * that we have big VM buffers to work with, the following is a cut at
 * DB data structures.
 */

/***************************************************************************\
 * External functions                                                      *
\***************************************************************************/

#if defined(WR_LINUX) && !defined(USER_MODE_TEST)
extern int      wsi_sysevents_getevspace        (void);
#endif

/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#ifdef USER_MODE_TEST
#undef WSS_MUTEX_ENTER
#undef WSS_MUTEX_EXIT
#define WSS_MUTEX_ENTER(m)
#define WSS_MUTEX_EXIT(m)
#define wss_wrvevent_post(a,b,c,e)
#define wss_db_lock_table(t)
#define wss_db_unlock_table(t)
#endif /* USER_MODE_TEST */

#if !defined(USER_MODE_TEST) && !defined(DBINTD_INTERFACE)
#define wss_db_lock()   WSS_MUTEX_ENTER( db )
#define wss_db_unlock() WSS_MUTEX_EXIT( db )
#endif /* !USER_MODE_TEST && !DBINTD_INTERFACE */

/*! Macro to determine if a relative pointer within the database is valid */
#define DB_VALID_PTR(rel) wss_vm_valid_alloced(wss_db_space,rel)

/*! Macro to determine if a relative string within the database is valid */
#define DB_VALID_STR(rel) wss_vm_valid_str(wss_db_space,rel)


/***************************************************************************\
 * Private types                                                           *
\***************************************************************************/

/*!
 * Enumeration of database check (dbck) operation steps
 */
typedef enum dbck_op
{
   op_check,  /*!< check blocks operation */
   op_mark,   /*!< mark blocks operation */
   op_end     /*!< dbck completed */
}
dbck_op;


/*!
 * The following structure supports wss_select, wss_update,
 * and wss_delete.
 */
typedef struct gval
{
   union
   {
      float             fval; /*!< float value */
      int               ival; /*!< integer value */
      const pd_char_t *cval; /*!< string value */
   }
   v;                         /*!< value data */
   enum wql_dtype dtype;      /*!< value data type */
}
gval, *gval_t;

typedef struct wql_dbrow dbrow;


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

static __inline__ pd_char_t  *strdup_abs         __P((const pd_char_t *string_abs));
static __inline__ pd_char_t  *strdup_rel         __P((const pd_char_t *string_abs));
static __inline__ wqcd_t col_defs_dup_abs_to_rel  __P((wqcd_t col_defs_abs,
                                                       int col_count));
static __inline__ wqcd_t col_defs_dup_rel_to_abs  __P((wqcd_t col_defs_rel,
                                                       int col_count));
pd_char_t EXPORT *wss_db_create_actual_table_name __P(( pd_char_t *wrapper_name, int id,
                                                         pd_char_t *table_name ));
void EXPORT              free_col_args            __P((wqcd_t cols, int count));
void EXPORT              free_row_args            __P((wqrow_t rows,
                                                       wqcd_t cols,
                                                       int col_count));
static int               row_sort                 __P((wqtab *table,
                                                       int column,
                                                       int decending));
static __inline__ enum wql_dtype get_val          __P((wqcd_t cols,
                                                       wqrow_t row,
                                                       sxval_t spec,
                                                       gval_t result));
static int               eval_term                __P((wqcd_t cols,
                                                       wqrow_t row,
                                                       tx_t term));
static int               eval_criteria            __P((wqcd_t cols,
                                                       wqrow_t row,
                                                       wqsx_t expr));
static void              free_sxval               __P((sxval_t term));
static void              free_wquexp              __P((wquexp_t exp));
static void              set_val                  __P((wqcd_t cols,
                                                       wqrow_t row,
                                                       sxval_t target,
                                                       sxval_t from));
static void              eval_expression          __P((wqcd_t cols,
                                                       wqrow_t row,
                                                       wquexp_t expr,
                                                       sxval_t retVal));
static void              eval_expression_list     __P((wqcd_t cols,
                                                       wqrow_t row,
                                                       wqusl_t expr_list));


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

/*! The database mutex which protects the list of DB tables */
WSS_MUTEX_PROTECT_EXPORTED(db);

/*! Mutex which protects the qsort variables used by select's order ability */
WSS_MUTEX_PROTECT(db_qsort);


/*
 * Globals: Sorry for the mess
 * The comparison function needs to know the type of the data.
 * TBD, duplication the comparison functions to ditch a global??
 * Lots of duplication in comparing stuff, see what can be factored
 * out.
 */
static enum wql_dtype sort_col_type;
static int sort_column;


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/*! Storage for the database VM space */
wss_vmspace_t WSSLINKAGE *wss_db_space    = NULL;

/*! Storage for the database filename (settable within configuration file) */
pd_char_t     WSSLINKAGE *wss_db_filename = NULL;

/*! Storage for the mapped database size (settable within configuration file) */
int           WSSLINKAGE  wss_db_size     = 0;


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

static __inline__ pd_char_t *
strdup_abs( const pd_char_t *string_abs )
{
   pd_char_t *cp;
   int len;

   if (string_abs == NULL) return NULL;
   len = ( wsi_strlen(string_abs) + 1 ) * sizeof(pd_char_t);
   cp = (pd_char_t *)wss_malloc(len);
   if( cp ) memcpy(cp, string_abs, len);
   return cp;
}

static __inline__ pd_char_t *
strdup_rel( const pd_char_t *string_abs )
{
   pd_char_t *cp;
   int len;

   if (string_abs == NULL) return NULL;
   len = ( wsi_strlen(string_abs) + 1 ) * sizeof(pd_char_t);
   cp = (pd_char_t *)wss_vm_malloc(wss_db_space, len);
   if( cp ) memcpy(cp, string_abs, len);
   return cp;
}

static __inline__ wqcd_t
col_defs_dup_abs_to_rel( wqcd_t col_defs_abs, int col_count )
{
   wqcd_t retval;
   int i;

   retval = (wqcd_t)wss_vm_malloc( wss_db_space, sizeof(*retval) * col_count );
   if( retval )
   {
      for( i = 0 ; i < col_count ; i++ )
      {
         wqcd_t new = retval + i;
         wqcd_t old = col_defs_abs + i;
        
         new->name = (pd_char_t *)DB_ABS2REL(strdup_rel(old->name));
         new->dtype = old->dtype;
         new->max_len = old->max_len;
      }
      retval = (wqcd_t)DB_ABS2REL(retval);
   }
   return retval;
}

static __inline__ wqcd_t
col_defs_dup_rel_to_abs( wqcd_t col_defs_rel, int col_count )
{
   wqcd_t retval;
   int i;
    
   col_defs_rel = (wqcd_t)DB_REL2ABS(col_defs_rel); /* Convert to abs */
   retval = (wqcd_t)wss_malloc(sizeof(*retval)*col_count);
   if( retval )
   {
      for( i = 0 ; i < col_count ; i++ )
      {
         wqcd_t new = retval + i;
         wqcd_t old = col_defs_rel + i;
        
         new->name = strdup_abs(DB_REL2ABS(old->name));
         new->dtype = old->dtype;
         new->max_len = old->max_len;
      }
   }
   return retval;
}

#ifdef DEBUG_DBCK
/*
 * Dump the pointers for each database table
 */
static void
dump_database_layout()
{
   wqtab_t tab;
   int     i;

   wss_db_lock();

   for( tab = (wqtab_t)(DBHEAD->db_tables?DB_REL2ABS(DBHEAD->db_tables):NULL),
        i = 0;
        tab;
        tab = (wqtab_t)((tab->next)?DB_REL2ABS(tab->next):NULL), i++ )
   {
      wqcd_t  cd;
      wqrow_t row;
      int     cnt;

      wss_db_lock_table( tab );
      wsi_printf( PD_MAKESTR("Table #%d: %p (%p)\n"), i, (void *) tab,
                  (void *) DB_ABS2REL(tab) );
      wsi_printf( PD_MAKESTR("Flags: ") );
      if( tab->flags & DB_FLAG_UNICODE )
         wsi_printf( PD_MAKESTR("Unicode ") );
      wsi_printf( PD_MAKESTR("\n") );
      wsi_printf( PD_MAKESTR("%d columns, %d rows\n"), tab->col_count,
                  tab->row_count );

      wsi_printf( PD_MAKESTR("Column Dump:\n") );
      for( cd = (wqcd_t)((tab->col_defs)?DB_REL2ABS(tab->col_defs):NULL), cnt = 0;
           ( cnt < tab->col_count );
           cd++, cnt++ )
      {
         if( cnt ) wsi_printf( PD_MAKESTR(", ") );
         wsi_printf( PD_MAKESTR("%p(%p)[%p:"), (void *) cd,
                     (void *) DB_ABS2REL(cd), (void *) cd->name );
         switch( cd->dtype )
         {
            case pd_char_t_t:
               wsi_printf( PD_MAKESTR("pd_char_t") );
               break;
            case int_t:
               wsi_printf( PD_MAKESTR("int") );
               break;
            case float_t:
               wsi_printf( PD_MAKESTR("float") );
               break;
            case none_t:
               wsi_printf( PD_MAKESTR("none") );
               break;
         }
         wsi_printf( PD_MAKESTR(":%d]"), cd->max_len );
      }
      wsi_printf( PD_MAKESTR("\n") );

      wsi_printf( PD_MAKESTR("Row Dump:\n") );
      for( row = (wqrow_t)((tab->rows)?DB_REL2ABS(tab->rows):NULL), cnt = 0;
           row;
           row = (wqrow_t)((row->next)?DB_REL2ABS(row->next):NULL), cnt++ )
      {
         if( cnt ) wsi_printf( PD_MAKESTR(", ") );
         wsi_printf( PD_MAKESTR("%p(%p)"), (void *) row,
                     (void *) DB_ABS2REL(row) );
      }
      wsi_printf( PD_MAKESTR("\n") );

      wss_db_unlock_table( tab );
   }

   wss_db_unlock();
}
#endif /* DEBUG_DBCK */

/*
 * Works like find_table, except it uses a table handle
 * for the match criteria
 * 
 * Assumes "db_lock" IS locked.
 * 
 * The table it returns IS locked.
 */
static wqtab_t
wss_db_find_table_by_handle_unsafe( u_long handle )
{
   wqtab_t tab;

   if( DBHEAD->db_tables == 0 )
      return NULL;

   for( tab = (wqtab_t)DB_REL2ABS(DBHEAD->db_tables);
        tab && ( (u_long)DB_ABS2REL(tab) != handle );
        tab = (wqtab_t)((tab->next)?DB_REL2ABS(tab->next):NULL) );

   if( tab )
       wss_db_lock_table( tab );

   return( tab );
}

/*
 * Row sort comparison functions for use with qsort
 */
static int
row_sort_compar( const void *one, const void *two )
{
   wqval_t a = ((wqrow_t)one)->vals + sort_column;
   wqval_t b = ((wqrow_t)two)->vals + sort_column;
   int retval = 0;

   switch( sort_col_type )
   {
      case pd_char_t_t:
         retval = wsi_strcmp(a->v.cval, b->v.cval);
         break;
      case int_t:
         retval = (a->v.ival == b->v.ival) ? 0 :
                  (a->v.ival > b->v.ival) ? 1 : -1;
         break;
      case float_t:
         retval = (a->v.ival == b->v.ival) ? 0 :
                  (a->v.ival > b->v.ival) ? 1 : -1;
         break;
      case none_t:
         retval = 0;
         break;
   }
   return retval;
}

/*
 * Takes a table, sorts it on the column named.
 * Don't touch table->next... so we won't care if it is accurate.
 * Don't touch table->table_lock, so we won't care if it is accurate.
 * Assumes tab is an absolute pointer, and its fields are absolute.
 * Assumes that the column name specified exists.
 * 
 */
static int
row_sort( wqtab *table, int column, int decending )
{
   int row_count = 0;
   int i;
   wqrow *row;
   wqrow **row_array;

#ifndef DBINTD_INTERFACE
   WSS_MUTEX_ENTER( db_qsort );
#endif /* DBINTD_INTERFACE */
   row_count = 0;
   for( row = table->rows; row; row = row->next )
      row_count++;
    
   /* A single row is already sorted */
   if( row_count < 2 )
   {
#ifndef DBINTD_INTERFACE
      WSS_MUTEX_EXIT( db_qsort );
#endif /* DBINTD_INTERFACE */
      return 1;
   }

   /* Create an array we can easily sort */
   row_array = (wqrow **)wss_malloc( sizeof(*row_array) * row_count );
   if( row_array == NULL )
   {
#ifndef DBINTD_INTERFACE
      WSS_MUTEX_EXIT( db_qsort );
#endif /* DBINTD_INTERFACE */
      return -WRE_NOMEM;
   }
   row = table->rows;

   for( i = 0; i < row_count; i++ )
   {
      row_array[i] = row;
      row = row->next;
   }
    
   /* Set the globals */
   sort_col_type = table->col_defs[column].dtype;
   sort_column = column;

   /* Sort the array */
   qsort( row_array, row_count, sizeof(row_array[0]), row_sort_compar );
    
   /* Redo the table in the order specified */
   if( !decending )
   {
      row = row_array[0];
      table->rows = row;
      for( i = 1; i < row_count ; i++ )
      {
         row->next = row_array[i];
         row = row_array[i];
      }
   }
   else
   {
      row = row_array[row_count - 1];
      table->rows = row;
      for( i = row_count - 2; i >= 0 ; i-- )
      {
         row->next = row_array[i];
         row = row_array[i];
      }
   }

   if( row )
      row->next = NULL;
    
   wss_free( row_array );
#ifndef DBINTD_INTERFACE
   WSS_MUTEX_EXIT( db_qsort );
#endif /* DBINTD_INTERFACE */
   return 1;
}


/*
 * The following routines support wss_select(), wss_update(), and wss_delete().
 */

/* Returns the value indicated by a select expression field reference or
 * literal value, and the type of the value.  Assumes all pointers are
 * absolute except for cols->next and row->*.  This should be much faster
 * now.
 * 
 * Assumes any locking has already been done.
 */
static __inline__ wql_dtype
get_val( wqcd_t cols, wqrow_t row, sxval_t spec, gval_t result )
{
   wqval_t val;

   if( spec->which == field )
   {
      int column;
      column = spec->v.f.column;
      val = (wqval_t)DB_REL2ABS(row->vals);
      result->dtype = cols[column].dtype;

      switch( result->dtype )
      {
         case pd_char_t_t:
            result->v.cval = DB_REL2ABS(val[column].v.cval);
            return pd_char_t_t;
            break;
         case int_t:
            result->v.ival = val[column].v.ival;
            return int_t;
            break;
         case float_t:
            result->v.fval = val[column].v.fval;
            return float_t;
            break;
         case none_t:
            return none_t;
            break;
      }
   }
   else
   {
      /* literal */

      /* copy it out of the spec */
      result->dtype = spec->dtype;
      switch( spec->dtype )
      {
         case pd_char_t_t:
            result->v.cval = spec->v.l.cval;
            return pd_char_t_t;
            break;
         case int_t:
            result->v.ival = spec->v.l.ival;
            return int_t;
            break;
         case float_t:
            result->v.fval = spec->v.l.fval;
            return float_t;
            break;
         case none_t:
            return none_t;
            break;
      }
   }

   return none_t;
}

/*
 * Evaluates a given binary expression (e.g., (.name <> "foo")) in
 * a select expression.  Returns !NULL iff the expression evaluates to
 * true.
 * 
 * Assumes any locking has already been done.
 */
static int
eval_term( wqcd_t cols, wqrow_t row, tx_t term )
{
   gval left, right;
   enum wql_dtype ltype, rtype;
   int comp;
   float lval, rval;

   /* The "true" comparison will be used to represent empty
    * select statements.
    * Place the test here so that the values will never be
    * dereferenced.
    */
   if( term->sxop == true_op )
      return 1;

   ltype = get_val( cols, row, term->left, &left );
   rtype = get_val( cols, row, term->right, &right );

   if( ltype == pd_char_t_t )
   {
      if ( !left.v.cval || !right.v.cval)
          comp = 1;
      else
          comp = wsi_strcmp( left.v.cval, right.v.cval );
      switch( term->sxop )
      {
         case gt_op:  return(comp > 0);
         case gte_op: return((comp > 0) || (comp == 0));
         case lt_op:  return(comp < 0);
         case lte_op: return((comp < 0) || (comp == 0));
         case ne_op:  return(comp != 0);
         case eq_op:  return(comp == 0);
         case true_op:   return 1;
      }
   }
   else
   {
      lval = (ltype == int_t) ? (float) left.v.ival : left.v.fval;
      rval = (rtype == int_t) ? (float) right.v.ival : right.v.fval;
      switch( term->sxop )
      {
         case gt_op:  return(lval > rval);
         case gte_op: return(lval >= rval);
         case lt_op:  return(lval < rval);
         case lte_op: return(lval <= rval);
         case ne_op:  return(lval != rval);
         case eq_op:  return(lval == rval);
         case true_op:   return 1;
       }
   }
   return 0;  /* to avoid warning: control should never reach here */
}

/*
 * Return !0 iff the row matches the criteria specified by expr.
 * Assumes that expr is properly constructed.
 * 
 * Assumes any locking has already been done.
 */
static int
eval_criteria( wqcd_t cols, wqrow_t row, wqsx_t expr )
{
   int res;

   if( expr == NULL )
      return 1;

   /* assume criteria are valid */
   if( expr->which == term )
   {
      return eval_term( cols, row, expr->un.exp );
   }
   else
   {
      res = eval_criteria(cols, row, expr->un.children->left);
      if( res && ( expr->un.children->logop == or_op ) )
      {
         return res;
      }
      else if( expr->un.children->logop == not_op )
      {
         return !res;
      }
      else if( !res && ( expr->un.children->logop == and_op ) )
      {
         return 0;
      }
      else
      {
         return eval_criteria( cols, row, expr->un.children->right );
      }
   }
}

static void
free_sxval( sxval_t term )
{
   if( term->which == literal && term->dtype == pd_char_t_t )
   {
        /* TBD: resolve this: wss_free(term->v.l.cval); */
   }
   wss_free( term );
}

static void
free_wquexp( wquexp_t exp )
{
   if( exp == NULL )
      return;

   if( exp->which == term )
   {
      free_sxval( exp->w.terminal );
   }
   else
   {
      free_wquexp( exp->w.e.left );
      if( exp->w.e.numop != negation )
         free_wquexp( exp->w.e.right );
   }
   wss_free( exp );
}

static void
set_val( wqcd_t cols, wqrow_t row, sxval_t target, sxval_t from )
{
   wqval_t val_to;
   enum wql_dtype val_to_type;

   /* Find the value to set. */
   switch( target->which )
   {
      case literal:
         wsi_printf( PD_MAKESTR("set_val: Error.  Can't set the value of a literal.\n") );
         /* Error. */
         return;

      case field:
         /* Find the value.   In this case, we also need to learn the type */
         val_to_type = cols[target->v.f.column].dtype;
         val_to = (wqval_t)DB_REL2ABS(row->vals) + target->v.f.column;

         /* Set the value. */
         if( from->which == literal )
         {
            switch( val_to_type )
            {
               case pd_char_t_t:
                  wss_vm_free(wss_db_space, DB_REL2ABS(val_to->v.cval));

                  /* Assume: from->v.l.cval is available to be claimed */
                  val_to->v.cval =
                     (pd_char_t *)DB_ABS2REL(strdup_rel( from->v.l.cval ));

                  /* Free the string. */
                  if( from->owned )
                  {
                     /* Cast away const */
                     wss_free( (void *)from->v.l.cval );
                     from->v.l.cval = 0;
                  }
                  from->v.l.cval = NULL;
                  break;

               case int_t:
                  if( from->dtype == int_t )
                     val_to->v.ival = from->v.l.ival;
                  else
                     val_to->v.ival = (int)from->v.l.fval;
                  break;

               case float_t:
                  if( from->dtype == int_t )
                     val_to->v.fval = (float)from->v.l.ival;
                  else
                     val_to->v.fval = from->v.l.fval;
                  break;

               case none_t:
                  wsi_printf( PD_MAKESTR("set_val: none_t typed column found.") );
                  break;
            }
         }
         else if (from->which == field)
         {
            wsi_printf( PD_MAKESTR("set_val: Warning.  Getting the lval of a field is not yet supported.\n") );
         }
         else
         {
            wsi_printf( PD_MAKESTR("set_val: Warning.  Getting the lval of a variable is not yet supported.\n") );
            /* Not implimented yet. */
            return;
         }
         break;

      case variable:
         wsi_printf( PD_MAKESTR("set_val: Warning.  Setting the value of a variable is not yet supported.\n") );
         /* Not implimented yet. */
         return;
   }
}

static void
eval_expression_list( wqcd_t cols, wqrow_t row, wqusl_t expr_list )
{
   wqusl_t usl = expr_list;
   sxval temp;
    
   /* Assume no type conflicts.  Ain't trust wonderfull. */
   while( usl != NULL )
   {
      eval_expression( cols, row, usl->right, &temp );
      set_val( cols, row, usl->left, &temp );
        
      usl = usl->next;
   }
}

static void
eval_expression( wqcd_t cols, wqrow_t row, wquexp_t expr, sxval_t retVal )
{
   if( expr->which == term )
   {
      sxval_t lval, rval;
      lval = rval = NULL;

      rval = expr->w.terminal;

      retVal->which = term;
      retVal->owned = 0;

      if( rval->which == literal )
      {
         /* Copy out value. */
         switch( rval->dtype )
         {
            case pd_char_t_t:
               retVal->dtype = pd_char_t_t;
               retVal->v.l.cval = rval->v.l.cval;
               retVal->owned = rval->owned;
               break;

            case int_t:
               retVal->dtype = int_t;
               retVal->v.l.ival = rval->v.l.ival;
               break;

            case float_t:
               retVal->dtype = float_t;
               retVal->v.l.fval = rval->v.l.fval;
               break;

            case none_t:
               retVal->dtype = none_t;
               break;
         }
      }
      else if( rval->which == field )
      {
         /* get the value from the field */
         enum wql_dtype val_from_type;
         wqval_t val_from;
            
         val_from_type = cols[rval->v.f.column].dtype;
         val_from = ((wqval_t)DB_REL2ABS(row->vals)) + rval->v.f.column;

         switch( val_from_type )
         {
            case pd_char_t_t:
               retVal->dtype = pd_char_t_t;
               retVal->v.l.cval = strdup_abs( DB_REL2ABS(val_from->v.cval) );
               retVal->owned = 1;
               break;

            case int_t:
               retVal->dtype = int_t;
               retVal->v.l.ival = val_from->v.ival;
               break;

            case float_t:
               retVal->dtype = float_t;
               retVal->v.l.fval = val_from->v.fval;
               break;

            case none_t:
               retVal->dtype = none_t;
               break;
         }
      }
      else
      {
         /* Is variable.  Not implimented yet. */
         wsi_printf( PD_MAKESTR("set_val: Warning.  Setting the value of a variable is not yet supported.\n") );
         retVal->dtype = int_t;
         retVal->v.l.ival = 0;
      }
   }
   else
   {
      double result, v1, v2 = 0;
      sxval lval, rval;

      /* Assume: lval and rval are returned terminals with values. */
      /* Also ignore string opps for time being. */

      eval_expression( cols, row, expr->w.e.left, &lval );
      if( lval.dtype == float_t )
         v1 = lval.v.l.fval;
      else if( lval.dtype == int_t )
         v1 = lval.v.l.ival;
      else
      {
         wsi_printf( PD_MAKESTR("Error: no string support in complex expressions.") );
         v1 = lval.v.l.ival;
         /* When you do implement this, don't forget string ownership */
      }

      if( expr->w.e.numop != negation )
      {
         eval_expression( cols, row, expr->w.e.right, &rval );
         if( rval.dtype == float_t )
            v2 = rval.v.l.fval;
         else
            v2 = rval.v.l.ival;
      }
        
      retVal->which = literal;

      switch( expr->w.e.numop )
      {
         case multiplication:
            result = v1 * v2;
            break;

         case division:
            result = v1 / v2;
            break;

         case mod:
            result = (int)v1 % (int)v2;
            break;

         case addition:
            result = v1 + v2;
            break;

         case subtraction:
            result = v1 - v2;
            break;

         case negation:
            result = -v1;
            break;

         default:
            result = 0;
      }
        
      /* Convert to int?  Only if both rval and lval ints. */
      if( ( rval.dtype == int_t ) &&
          ( expr->w.e.numop == negation || lval.dtype == int_t ) )
      {
         retVal->dtype = int_t;
         retVal->v.l.ival = (int)result;
      }
      else
      {
         retVal->dtype = float_t;
         retVal->v.l.fval = (float)result;
      }
   }
}

/***************************************************************************\
 * Privately exported functions                                            *
\***************************************************************************/

/*
 * Frees a list of column definitions.  Assumes that the cols are made with
 * normal mallocs and are absolute.
 */
void
free_col_args( wqcd_t cols, int count )
{
   int i;
   for( i = 0; i < count; i++ )
      wss_free( (void *)cols[i].name );

   wss_free( (void *)cols );
}

/*
 * Free a list of rows.  Assumes that the rows are made with normal
 * mallocs and are absolute.
 */
void
free_row_args( wqrow_t rows, wqcd_t cols, int col_count )
{
   int i;
   wqrow_t row;
   wqval_t val;
    
   while( rows )
   {
      row = rows;
      rows = rows->next;

      val = row->vals;
      for( i = 0; i < col_count; i++ )
      {
         if( cols[i].dtype == pd_char_t_t )
         {
            wss_free( (void*)val[i].v.cval );
         }
      }
      wss_free( (void*)val );
      wss_free( (void*)row );
   }
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_db_init */
/*! \brief Database module initialization routine
 *
 * Initializes the mutexes protecting the database table list and the
 * qsort globals.  Also, performs a check of the database to make sure
 * the data contained within is sane.
 *
 * \param run_dbck Flag indicating if dbck should be performed or not
 */
int EXPORT
wss_db_init( int run_dbck )
{
   /* Initialize mutexes */
   WSS_MUTEX_INIT( db );
   WSS_MUTEX_INIT( db_qsort );

   wss_db_space = wss_vm_create( PD_MAKESTR("WSS Database"), wss_db_filename,
                                 wss_db_size, sizeof(wss_db_head_t) );

   if( wss_db_space )
   {
      /* Check database for inconsistencies */
      if( run_dbck )
         wss_dbck();
      return (0);
   }
   else
      return (ENOENT);
}

/*  wss_db_create_actual_table_name */
/*! \brief Create a composite table name of the form
 *         <wrapper_name>.<id>.<table_name>
 *
 *  \param wrapper_name Pointer to string containing the wrapper name
 *  \param id           Identifier of process
 *  \param table_name   Pointer to string containing the table name
 *
 *  \warning The client must free the returned string
 * 
 *  \return Pointer to string containing the composite table name,
 *  \return NULL if memory is exhausted
 */
pd_char_t *
wss_db_create_actual_table_name( pd_char_t *wrapper_name, int id,
                                 pd_char_t *table_name )
{
   int len;
   pd_char_t *name;
#if !PD_BROKEN_SPRINTF_RETV
   pd_char_t ibuf[40];
#endif

   len = wsi_strlen(wrapper_name);
   if( id )
   {
#if PD_BROKEN_SPRINTF_RETV
      /* Maximum possible length of "id" in decimal representation +1 */
      len += 1 + 20;
#else
      /* This perfectly reasonable construct doesn't work
       * under Solaris because in Solaris's kernel the
       * return value of "sprintf" is a pointer to the buffer.
       */
      len += wsi_sprintf( ibuf, PD_MAKESTR(".%d"), id );
#endif
   }

   len += wsi_strlen(table_name) + 2;
   name = (pd_char_t*)wss_malloc( len * sizeof(pd_char_t) );
   if( name == NULL )
      return NULL;      /* memory exhausted */
   if( id )
   {
#if PD_BROKEN_SPRINTF_RETV
      wsi_sprintf( name, PD_MAKESTR("%s.%d.%s"), wrapper_name, id,
                   table_name );
#else
      wsi_sprintf( name, PD_MAKESTR("%s%s.%s"), wrapper_name, ibuf,
                   table_name );
#endif
   }
   else
   {
      wsi_sprintf( name, PD_MAKESTR("%s.%s"), wrapper_name, table_name );
   }
   return name;
}


/*  wss_db_find_table_by_name_unsafe */
/*! \brief Searches for the named table in the database
 *
 *  Searches the list of tables for the one named in the argument \a name.
 *
 *  \param name Full string name of table to find
 *
 *  \warning Assumes lock on db IS held by current thread.
 * 
 *  \return Pointer to absolute address of the named table,
 *  \return NULL if the table does not exist
 */
wqtab_t EXPORT
wss_db_find_table_by_name_unsafe( pd_char_t *name )
{
   wqtab_t tab = NULL;

   if( DBHEAD->db_tables )
   {
      wqtab_t curtab;

      for( curtab = (wqtab_t)DB_REL2ABS(DBHEAD->db_tables);
           curtab;
           curtab = curtab->next ? (wqtab_t)DB_REL2ABS(curtab->next) : NULL )
      {
#ifdef DEBUG_DB
         wsi_printf( PD_MAKESTR("wss_db_find_table_by_name: curtab = %p(%p)\n"), curtab, DB_ABS2REL(curtab) );
#endif /* DEBUG_DB */
         if( !wsi_strcmp( name, DB_REL2ABS(curtab->name) ) )
         {
            tab = curtab;
            wss_db_lock_table( tab );
            break;
         }
      }
   }

   return tab;
}


/*  wss_db_find_table_by_name */
/*! \brief Searches for the named table in the database
 *
 *  Searches the list of tables for the one named in the argument \a name.
 *
 *  \param name Full string name of table to find
 *
 *  \warning Assumes lock on db is NOT held by current thread.
 * 
 *  \return Pointer to absolute address of the named table,
 *  \return NULL if the table does not exist
 *
 *  \sa wss_db_find_table_by_name_unsafe
 */
wqtab_t EXPORT
wss_db_find_table_by_name( pd_char_t *name )
{
   wqtab_t tab;

   wss_db_lock();
   tab = wss_db_find_table_by_name_unsafe( name );
   wss_db_unlock();
   return tab;
}


/*  wss_db_find_table_by_names */
/*! \brief Searches for the named table in the database
 *
 *  Calls wss_db_find_table_by_name, constructing the table name from the
 *  three component parts first.
 *
 *  \param wrapper_name String name of wrapper module
 *  \param id           Identifier of instance
 *  \param table_name   String name of table
 *
 *  \warning Assumes lock on db is NOT held by current thread.
 *
 *  \return Relative pointer to table on success,
 *  \return 0 on failure
 */
u_long EXPORT
wss_db_find_table_by_names( pd_char_t *wrapper_name, int id, pd_char_t *table_name )
{
   wqtab_t result;
   pd_char_t *full_name;
    
   full_name = wss_db_create_actual_table_name( wrapper_name, id, table_name );
   if( full_name == NULL )
      return 0;
    
   result = wss_db_find_table_by_name( full_name );
   if( result != NULL )
      wss_db_unlock_table( result );
   wss_free( full_name );
   return result ? (u_long)DB_ABS2REL(result) : 0;
}

/*  wss_db_find_table_by_handle */
/*! \brief Searches for the identified table in the database
 *
 *  Works like wss_find_table_by_name, except it uses a table \a handle
 *  for the match criteria.
 *
 *  \param handle Handle identifier for table
 *
 *  \warning Assumes lock on db is NOT held by current thread.
 *
 *  \return Absolute pointer to table on success,
 *  \return NULL on failure
 */
wqtab_t EXPORT
wss_db_find_table_by_handle( u_long handle )
{
    wqtab_t retval;

    wss_db_lock();
    retval = wss_db_find_table_by_handle_unsafe( handle );
    wss_db_unlock();

    return retval;
}
    
/*  wss_dbck */
/*! \brief WSS database-equivalent to UNiX fsck for filesystems
 *
 *  Traverses the database list first checking for valid pointers and
 *  strings, second marking blocks for exemption from collection, and
 *  third collecting unused blocks.  This function tries to do for the
 *  WSS database what fsck does for UNiX filesystems.
 */
void EXPORT
wss_dbck()
{
#if ( !defined(USER_MODE_TEST) || defined(USER_MODE_DBCK) ) && !defined(DBINTD_INTERFACE)
   register wqtab_t *last;
   register wqtab_t tab;
   register wqtab_t atab;

   wss_db_lock();


   if( DBSPACEHEAD->vms_head_trace == 0 )
   {
      DBHEAD->db_version = WSS_DB_VERSION;
      DBHEAD->db_tables = 0;
   }

   if( DBHEAD->db_version != WSS_DB_VERSION )
   {
      wss_debug_log( PD_MAKESTR("wss_dbck:: Database version mismatch - ")
                     PD_MAKESTR("expected %d, detected %d\n"), WSS_DB_VERSION,
                     DBHEAD->db_version );
      wss_debug_log( PD_MAKESTR("           Dropping all tables\n") );

      DBHEAD->db_version = WSS_DB_VERSION;
      DBHEAD->db_tables = 0;
   }

   last = (wqtab_t *)&(DBHEAD->db_tables);
   tab = (wqtab_t)DBHEAD->db_tables;

   while( tab )
   {
      register wqcd_t cd    = NULL;
      register wqcd_t acd   = NULL;
      register wqrow_t row  = NULL;
      register wqrow_t arow = NULL;
      wqrow_t *lastRow      = NULL;
      dbck_op op;

      atab = (wqtab_t)DB_REL2ABS(tab);

      for( op = op_check; op < op_end; op++ )
      {
         register int i;

         if( op == op_check )
         {
            pd_char_t *name;

            /* Check table pointer */
            if( !DB_VALID_PTR(tab) )
            {
               wss_debug_log( PD_MAKESTR("wss_dbck:: table pointer out of range, truncating table list\n") );
               *last = NULL;
               goto sweep_tables;
            }

#ifndef USER_MODE_TEST
            if( sizeof(struct wql_table_opaque) > 0 )
            {
               atab->opaque = (void *)wss_malloc( sizeof(struct wql_table_opaque) );
               WSS_MUTEX_INIT(WQL_TABLE_OPAQUE(atab)->table);
#ifdef DEBUG_MUTEX
               wss_debug_log( PD_MAKESTR("wss_dbck:: initialized table lock 0x%x\n"),
                              WQL_TABLE_OPAQUE(atab)->table_lock );
#endif /* DEBUG_MUTEX */
            }
            else
               atab->opaque = NULL;
            wss_db_lock_table(atab);
#else
            atab->opaque = NULL;
#endif /* USER_MODE_TEST */
            if( !DB_VALID_STR(atab->name) )
            {
               wss_debug_log( PD_MAKESTR("wss_dbck:: table name out of range or not null-terminated, freeing table\n") );
#ifndef USER_MODE_TEST
               wss_db_unlock_table(atab);
#endif /* USER_MODE_TEST */
               *last = atab->next;
               goto next_tab;
            }

            for( name = DB_REL2ABS(atab->name); *name && (*name != '.'); name++ );
            if( *name )
            {
               for( name++;
                    *name && (*name != '.') && (*name >= '0' || *name <= '9');
                    name++ );
               if( *name == '.' )
               {
                  wss_debug_log( PD_MAKESTR("wss_dbck:: dropping local table '%s'\n"),
                                 DB_REL2ABS(atab->name) );
#ifndef USER_MODE_TEST
                  wss_db_unlock_table(atab);
#endif /* USER_MODE_TEST */
                  *last = atab->next;
                  goto next_tab;
               }
            }
            if( atab->flags & DB_FLAG_NONPERSISTENT )
            {
               wss_debug_log( PD_MAKESTR("wss_dbck:: dropping ")
                              PD_MAKESTR("non-persistent table '%s'\n"),
                              DB_REL2ABS(atab->name) );
#ifndef USER_MODE_TEST
               wss_db_unlock_table(atab);
#endif /* USER_MODE_TEST */
               *last = atab->next;
               goto next_tab;
            }
            wss_debug_log( PD_MAKESTR("wss_dbck:: Validating table '%s'\n"),
                           DB_REL2ABS(atab->name) );
         }
         else
         {
            wss_vm_mark( wss_db_space, tab );
            wss_vm_mark( wss_db_space, atab->name );
         }

         /* Check column defs for table */
         if( op == op_check )
         {
            if( !DB_VALID_PTR(atab->col_defs) )
            {
               wss_debug_log( PD_MAKESTR("wss_dbck:: col def pointer out of range, freeing table\n") );
#ifndef USER_MODE_TEST
               wss_db_unlock_table(atab);
#endif /* USER_MODE_TEST */
               *last = atab->next;
               goto next_tab;
            }
         }
         for( i = 0; i < atab->col_count; i++ )
         {
            cd = atab->col_defs + i;
            acd = (wqcd_t)DB_REL2ABS(cd);

            if( op == op_check )
            {
               if( !DB_VALID_STR(acd->name) )
               {
                  wss_debug_log( PD_MAKESTR("wss_dbck:: col def name out of range or not null-terminate, freeing table\n") );
#ifndef USER_MODE_TEST
                  wss_db_unlock_table(atab);
#endif /* USER_MODE_TEST */
                  *last = atab->next;
                  goto next_tab;
               }
            }
            else
               wss_vm_mark( wss_db_space, acd->name );
         }

         if( op == op_mark )
            wss_vm_mark( wss_db_space, atab->col_defs );

         /* Check rows for table */
         lastRow = &atab->rows;
         for( row = atab->rows; row;
              lastRow = &arow->next,
              row = (wqrow_t)(arow->next?arow->next:NULL) )
         {
            register wqval_t val;
            register wqval_t aval;

            arow = (wqrow_t)DB_REL2ABS(row);

            if( op == op_check )
            {
               if( !DB_VALID_PTR( row ) )
               {
                  wss_debug_log( PD_MAKESTR("wss_dbck:: row out of range, truncating rows in table\n") );
                  *lastRow = NULL;
                  continue;
               }
            }
            else
               wss_vm_mark( wss_db_space, row );

            if( op == op_check )
            {
               if( !DB_VALID_PTR( arow->vals ) )
               {
                  wss_debug_log( PD_MAKESTR("wss_dbck:: values out of range, dropping row\n") );
                  *lastRow = arow->next;
                  goto next_row;
               }
            }

            /* Check values for row */
            for( i = 0; i < atab->col_count; i++ )
            {
               wqcd_t acd = (wqcd_t)DB_REL2ABS(atab->col_defs) + i;
               val = arow->vals + i;
               aval = (wqval_t)DB_REL2ABS(val);

               if( op == op_check )
               {
                  if( ( acd->dtype == pd_char_t_t ) &&
                      !DB_VALID_STR( aval->v.cval ) )
                  {
                     wss_debug_log( PD_MAKESTR("wss_dbck:: string value data out of range or not null-terminated, dropping row\n") );
                     *lastRow = arow->next;
                     goto next_row;
                  }
               }
               else
               {
                  if( acd->dtype == pd_char_t_t )
                     wss_vm_mark( wss_db_space, aval->v.cval );
               }
            }

            if( op == op_mark )
               wss_vm_mark( wss_db_space, arow->vals );
next_row:
            ; /* semicolon here for compiler's sake */
         }
      }
#ifndef USER_MODE_TEST
      wss_db_unlock_table(atab);
#endif /* USER_MODE_TEST */
next_tab:
      last = &atab->next;
      tab = atab->next;
   }

sweep_tables:
   /* Sweep all unmarked blocks */
   wss_vm_sweep( wss_db_space );

#ifndef USER_MODE_TEST
   wss_db_unlock();
#endif /* USER_MODE_TEST */

#ifdef DEBUG_DBCK
   dump_database_layout();
#endif /* DEBUG_DBCK */
#endif /* !USER_MODE_TEST || USER_MODE_DBCK || DBINTD_INTERFACE */
}


/*  wss_create */
/*! \brief Create a table in the database.
 *
 *  The wrapper name is prepended to the table name, which prevents
 *  table-name collisions between wrappers.  If the instance_id
 *  parameter is NULL, this is a global table; otherwise, the
 *  instance_id is prepended to the table name to create a local table.
 *  It is likely an error for the wr_install routine to specify a local
 *  table since it will not have access to instance id's for wrapped
 *  processes.  Also, local and global tables must have unique names.
 *
 *  NOTE: For now, all parameters are copied into vm space using
 *  wss_vm_malloc.  As a speed optimization later, parameters could be
 *  originally allocated in vm space by the caller, TBD.  This
 *  function does not change its parameters.
 *
 *  \param wrapper_name  String name of wrapper module
 *  \param id            Identifier of instance
 *  \param table_name    String name of table
 *  \param free_col_defs Flag to indicate if passed-in column definition
 *                       should be freed when wss_create is finished with
 *                       them
 *  \param col_defs      Pointer to the column definition for the table
 *  \param col_count     Number of columns in table column definition
 *  \param table_id      Returned identifier for table
 *
 *  \return 0 on success,
 *  \return \b WRE_INVALID_PARAMETER when a max_len field is given for a data
 *          type other than char,
 *  \return \b WRE_DB_SPACE_EXHAUSTED when there is no more mapped memory for
 *          the database to use,
 *  \return \b WRE_EXIST if a table already exists with the provided name
 *  \return \b WRE_NOMEM if we cannot allocate enough memory
 */
int EXPORT
wss_create( pd_char_t *wrapper_name,
            int instance_id,          /* iff !NULL, local table */
            pd_char_t *table_name,
            int free_col_defs,        /* free them if !NULL */
            wqcd_t col_defs,
            int col_count,            /* null-terminated list */
            u_long *table_id )
{
   pd_char_t *name;
#if !PD_BROKEN_SPRINTF_RETV
   pd_char_t ibuf[40];
#endif
   int len;
   wqtab_t tab;

   /* No validation.  The wrapper compilier is perfect. <G> */

   /* Create the actual table name, allocating from VM space. */
   len = wsi_strlen( wrapper_name );
   if( instance_id )
   {
#if PD_BROKEN_SPRINTF_RETV
      /* Maximum possible length of "id" in decimal representation +1 */
      len += 1 + 20;
#else
      /* This perfectly reasonable construct doesn't work
       * under Solaris because in Solaris's kernel the
       * return value of "sprintf" is a pointer to the buffer.
       */
      len += wsi_sprintf( ibuf, PD_MAKESTR(".%d"), instance_id );
#endif
   }
   len += wsi_strlen( table_name ) + 2;
   name = (pd_char_t*)wss_vm_malloc( wss_db_space, len * sizeof(pd_char_t) );
   if( name == NULL )
      return -WRE_DB_SPACE_EXHAUSTED;

   if( instance_id )
   {
#if PD_BROKEN_SPRINTF_RETV
      wsi_sprintf(name, PD_MAKESTR("%s.%d.%s"), wrapper_name, instance_id, table_name);
#else
      wsi_sprintf(name, PD_MAKESTR("%s%s.%s"), wrapper_name, ibuf, table_name);
#endif
   }
   else
   {
      wsi_sprintf(name, PD_MAKESTR("%s.%s"), wrapper_name, table_name);
   }

   /* Lock now.  */
   wss_db_lock();

   /* Fail if a table by this name already exists. */
   if( ( tab = wss_db_find_table_by_name_unsafe( name ) ) != NULL )
   {
      wss_db_unlock_table( tab );
      wss_db_unlock();
      wss_vm_free( wss_db_space, name );
      return -WRE_EXIST;
   }
    
#ifdef DEBUG_DB
   wsi_printf( PD_MAKESTR("Creating new table.  name = %s\n"),name );
#endif

   /*
    * Make the table structure and fill it in using vm buffers and
    * relative pointers.
    */
   tab = (wqtab_t)wss_vm_malloc( wss_db_space, sizeof(*tab) );
   if( tab == NULL )
   {
      wss_db_unlock();
      wss_vm_free( wss_db_space, name );
      return -WRE_DB_SPACE_EXHAUSTED;
   }

#if !defined(USER_MODE_TEST) && !defined(DBINTD_INTERFACE)
   if( sizeof(struct wql_table_opaque) > 0 )
   {
      tab->opaque = (void *)wss_malloc( sizeof(struct wql_table_opaque) );
      if( tab->opaque == NULL )
      {
         wss_db_unlock();
         wss_vm_free( wss_db_space, name );
         wss_vm_free( wss_db_space, tab );
         return -WRE_NOMEM;
      }
      WSS_MUTEX_INIT(WQL_TABLE_OPAQUE(tab)->table);
   }
   else
      tab->opaque = NULL;
#else
   tab->opaque = NULL;
#endif /* !USER_MODE_TEST && !DBINTD_INTERFACE */
#ifdef UNICODE
   tab->flags = DB_FLAG_UNICODE;
#else
   tab->flags = 0;
#endif /* UNICODE */
   if( instance_id )
      tab->flags |= ( DB_FLAG_LOCAL | DB_FLAG_NONPERSISTENT );
   tab->name = (pd_char_t*)DB_ABS2REL(name);
   tab->rows = NULL;
   tab->tail = NULL;
   tab->col_count = col_count;
   tab->col_defs = col_defs_dup_abs_to_rel( col_defs, col_count );

   /* Add it to the DB. */
   tab->next = (wqtab_t)DBHEAD->db_tables;
#ifdef DEBUG_DB
   wsi_printf( PD_MAKESTR("Next table: %p\n"),tab->next );
#endif
   DBHEAD->db_tables = DB_ABS2REL(tab);

   if( free_col_defs )
      free_col_args( col_defs, col_count );
    
   if( table_id != NULL )
   {
      *table_id = (u_long)(tab ? DB_ABS2REL(tab) : 0);
#ifdef DEBUG_DB
      wsi_printf( PD_MAKESTR("New table_id = %lu\n"),*table_id );
#endif
   }

   /*
    * Post create table event
    */
   wss_db_unlock();
   wss_wrvevent_post_createtable( tab );
   return 0;
}

/*  wss_drop */
/*! \brief Remove a table from the database.
 *
 *  Removes all data associated with the table.
 *
 *  \param table_id Handle identifier for table
 *
 *  \warning It is the caller's responsibility to free buffers passed to
 *           this function.
 *
 *  \return 0 on success,
 *  \return \b WRE_NOTFOUND if the table did not exist,
 *  \return \b WRE_NOMEM if we cannot allocate enough memory
 */
int EXPORT
wss_drop( u_long table_id )
{
   wqtab_t tab;
   wqrow_t row, trow;
   wqcd_t cd;
   int i, max;

   wss_db_lock();
   tab = wss_db_find_table_by_handle_unsafe( table_id );
   if( tab == NULL )
   {
      wss_db_unlock();
      return -WRE_NOTFOUND;
   }

#ifdef DEBUG_DB
   wsi_printf( PD_MAKESTR("Dropping table.  name = %s\n"), (pd_char_t *)DB_REL2ABS(tab->name) );
#endif

   /* Unlink table */
   if( (wqtab_t)DB_REL2ABS(DBHEAD->db_tables) == tab )
   {
      DBHEAD->db_tables = 
         (unsigned int)((wqtab_t)DB_REL2ABS(DBHEAD->db_tables))->next;
   }
   else
   {
      wqtab_t curtab;

      curtab = (wqtab_t)DB_REL2ABS(DBHEAD->db_tables);
      while( (wqtab_t)DB_REL2ABS(curtab->next) != tab )
      {
         curtab = curtab->next ? (wqtab_t)DB_REL2ABS(curtab->next) : NULL;
      }
      curtab->next = tab->next;
   }
   wss_db_unlock();

   /* Remove rows. */
   row = tab->rows ? (wqrow_t)DB_REL2ABS(tab->rows) : NULL;
   cd = tab->col_defs ? (wqcd_t)DB_REL2ABS(tab->col_defs) : NULL;
   max = tab->col_count;
   while( row )
   {
      wqval_t val;

      trow = row;
      val = trow->vals ? (wqval_t)DB_REL2ABS(trow->vals) : NULL;
      for( i = 0 ; i < max ; i++ )
      {
         if( cd[i].dtype == pd_char_t_t )
         {
            wss_vm_free( wss_db_space, (void*)DB_REL2ABS(val[i].v.cval) );
         }
      }
      wss_vm_free( wss_db_space, (void*)val );
      trow = row;
      row = row->next ? (wqrow_t)DB_REL2ABS(row->next) :  NULL;
      wss_vm_free( wss_db_space, (void*)trow );
   }

   /* Remove column defs. */
   for( i = 0 ; i < max ; i++ )
   {
      wss_vm_free( wss_db_space, (void*)DB_REL2ABS(cd[i].name) );
   }
   wss_vm_free( wss_db_space, cd );

   wss_db_unlock_table( tab );
#ifndef USER_MODE_TEST
   WSS_MUTEX_DESTROY( WQL_TABLE_OPAQUE(tab)->table );
#endif /* USER_MODE_TEST */

   /*
    * Post drop table event
    */
   wss_wrvevent_post_droptable( tab );

   wss_vm_free( wss_db_space, DB_REL2ABS(tab->name) );
   if( tab->opaque )
      wss_free( tab->opaque );
   wss_vm_free( wss_db_space, tab );

   return 0;
}

/*  wss_insert */
/*! \brief Inserts a list of rows into the indicated table
 *
 *  Each row is represented by a NULL-terminated list of val_def structures.
 *
 *  \param table_id  Handle identifier for table
 *  \param free_vals Flag to indicate if passed-in values should be freed
 *                   when wss_insert is finished with them
 *  \param in_rows   List of rows to insert
 *
 *  \warning This function does not affect its parameters.  All data is
 *           copied into vm buffers.
 *
 *  \return 0 on success,
 *  \return \b WRE_PERM if the table is immutable,
 *  \return \b WRE_NOMEM if we could not allocate enough memory,
 *  \return \b WRE_NOTFOUND if the table did not exist,
 *  \return \b WRE_INVALID_PARAMETER if any of the values were invalid
 */
int EXPORT
wss_insert( u_long table_id, int free_vals, wqrow_t in_rows )
{
   wqtab_t tab;
   wqrow_t rows, nrow;
   wqcd_t  cd;
   int     i, col_count;
   int     rval = 0;

   tab = wss_db_find_table_by_handle( table_id );
   if( tab == NULL )
      return -WRE_NOTFOUND;

   /* Leave validation of parameters up to the wrapper compilier.
    * (and anyone else who calls this function)
    */

   /* Insert rows. */
   col_count = tab->col_count;

   if( tab->flags & DB_FLAG_IMMUTABLE )
   {
      rval = -WRE_PERM;
      goto badparam;
   }

   /* Insert rows. */
   rows = in_rows;

   while( rows )
   {
      nrow = (wqrow_t)wss_vm_malloc(wss_db_space, sizeof(wqrow));
      if( nrow == NULL )
      {
         rval = -WRE_NOMEM;
         break;
      }
      nrow->vals = (wqval_t)wss_vm_malloc(wss_db_space, sizeof(wqval)*col_count);
      if( nrow->vals == NULL )
      {
         wss_vm_free( wss_db_space, nrow );
         rval = -WRE_NOMEM;
         break;
      }

      nrow->next = NULL;

      for( cd = (wqcd_t)DB_REL2ABS(tab->col_defs), i = 0;
           i < col_count; i++, cd++ )
      {
         wqval_t nval, oval;

         /* Create new val struct and fill it in. */
         nval = nrow->vals + i;
         oval = rows->vals + i;
         switch( cd->dtype )
         {
            case int_t:
                nval->v.ival = oval->v.ival;
                break;

            case float_t:
                nval->v.fval = oval->v.fval;
                break;

            case pd_char_t_t:
                nval->v.cval = (pd_char_t *)DB_ABS2REL(strdup_rel(oval->v.cval));
                break;

            case none_t:
                wsi_printf( PD_MAKESTR("Invalid type in column.\n") );
                nval->v.ival = 0;
                rval = -WRE_INVALID_PARAMETER;
                goto badparam;
         }
      }
      nrow->vals = (wqval_t)DB_ABS2REL(nrow->vals);

      /* Add new row to the table. */
      nrow->next = tab->rows;
      tab->rows = (wqrow_t)DB_ABS2REL(nrow);
      tab->row_count++;

      /*
       * Post insert row event
       */
      wss_wrvevent_post_insertrow( tab, nrow );

#if defined(WR_LINUX) && !defined(USER_MODE_TEST)
      /*
       * Post system event
       */
      {
          static        int recursion = 0;
          if (!recursion) {
              wss_proc_t   *init_wrp = wss_pfind(0);
              if (init_wrp) {
                  recursion = 1;
                  wss_dispatch(wsi_sysevents_getevspace(), 3, init_wrp, 0,
                          1, DB_REL2ABS(tab->name));
                  wss_proc_rele(init_wrp);
                  recursion = 0;
              }
          }
      }
#endif
      rows = rows->next;
   }
badparam:

   if( free_vals )
      free_row_args( in_rows, (wqcd_t)DB_REL2ABS(tab->col_defs), col_count );

   wss_db_unlock_table( tab );
   return rval;
}

/*
 * The following routines and globals support sorting for wss_select().
 * 
 * This is setup so that I can just use qsort()
 */

/*  wss_db_free_criteria */
/*! \brief Free the memory used by a select criteria
 *
 *  \param expr Pointer to selection criteria expression
 */
void EXPORT
wss_db_free_criteria( wqsx_t expr )
{
   if( expr == NULL )
        return;

   if( expr->which == term )
   {
      tx_t term = expr->un.exp;
      if( term->sxop != true_op )
      {
         free_sxval( term->left );
         free_sxval( term->right );
      }
      wss_free( term );
   }
   else
   {
      ntx_t node = expr->un.children;
      wss_db_free_criteria( node->left );

      if( node->logop != not_op )
         wss_db_free_criteria( node->right );

      wss_free( node );
   }
   wss_free( expr );
}

/*  wss_db_free_wqusl */
/*! \brief Free the memory used by update selection list
 *
 *  \param list Pointer to an update selection list
 */
void EXPORT
wss_db_free_wqusl(wqusl_t list)
{
    wqusl_t next;
    while (list != NULL) {
        free_sxval(list->left);
        free_wquexp(list->right);
        next = list->next;
        wss_free(list);
        list = next;
    }
}

/*  wss_select */
/*! \brief Return data from a table
 *
 *  Data retrieved from a table is placed in a rows element of the
 *  wqtab structure pointed to by \a out_table.
 *  If any rows are returned, the schema is returned also in the
 *  in the \a out_table structure.
 * 
 *  For each row in a table, this function evaluates the \a scrit parameter,
 *  which is a boolean expression defined on field values and constants.
 * 
 *  \param table_id   Handle identifier for table
 *  \param free_scrit Flag which determines if the structures passed in
 *                    for the selection criteria will be freed with wss_free
 *  \param scrit      Pointer to selection criteria
 *  \param take_first Flag which determines if the search for rows that match
 *                    the selection criteria will stop after the first match
 *                    is found
 *  \param order_by   If non-NULL, it is the name of a field that should be
 *                    used to sort the returned rows
 *  \param desc       Flag which determines if the sort ordering should be
 *                    in descending order
 *  \param out_table  Resulting temporary table from select
 *
 *  \warning This function modifies only the rows parameter, which is
 *  allocated using a normal malloc() (not wss_vmalloc()).  It is the
 *  responsibility of the caller to free the rows parameter.
 *
 *  \return 0 on success,
 *  \return \b WRE_NOTFOUND if the table is not found,
 *  \return \b WRE_BAD_FIELD if the specified field does not exist in the table,
 *  \return \b WRE_TYPE_MISMATCH if incomparable types are compared
 *  \return \b WRE_NOMEM if we could not allocate enough memory,
 */
int EXPORT
wss_select( u_long table_id,
            int free_scrit,         /* Do I free the wqsx_t? */
            wqsx_t scrit,           /* select_criteria() ptr*/
            int take_first,         /* Do we stop on first match */
            int order_by,           /* -1 means don't order */
            int desc,
            wqtab_t out_table )     /* OUT, returned schema */
{
   wqtab_t tab;
   wqcd_t cd;
   wqrow_t row, nrows, nrow, last_row, cur_row;
   int count = 0; /* number of rows selected */
   int retVal, col_count, i;

   if( out_table )
   {
      out_table->col_defs = NULL;
      out_table->rows = NULL;
      out_table->col_count = 0;
      out_table->max_rows = 0;
      out_table->tail = NULL;
      out_table->next = NULL;
   }

   tab = wss_db_find_table_by_handle( table_id );
   if( tab == NULL )
   {
      retVal = -WRE_NOTFOUND;
      goto select_end;
   }

   /* Copy selected rows out to client-space. */
   row = tab->rows ? (wqrow_t)DB_REL2ABS(tab->rows) : NULL;
   cd = (wqcd_t)DB_REL2ABS(tab->col_defs);
   nrows = NULL;
   last_row = NULL;
   col_count = tab->col_count;
   while( row )
   {
      wqval_t vals;

      cur_row = row;
      row = row->next ? (wqrow_t)DB_REL2ABS(row->next) :  NULL;
      if( !eval_criteria( cd, cur_row, scrit ) )
         continue;

      count++;
      if( !out_table )
         continue;

      nrow = (wqrow_t)wss_malloc(sizeof(*nrow));
      if( nrow == NULL )
      {
         retVal = -WRE_NOMEM;
         goto select_cleanup_end;
      }

      nrow->vals = (wqval_t)wss_malloc(sizeof(wqval)*col_count);
      if( nrow->vals == NULL )
      {
         wss_free( nrow );
         retVal = -WRE_NOMEM;
         goto select_cleanup_end;
      }
      vals = (wqval_t)DB_REL2ABS(cur_row->vals);

      for( i = 0; i < col_count; i++ )
      {
         wqval_t nval, oval;

         nval = nrow->vals + i;
         oval = vals + i;
         switch( cd[i].dtype )
         {
            case pd_char_t_t :
               nval->v.cval = strdup_abs(DB_REL2ABS(oval->v.cval));
               break;

            case int_t :
               nval->v.ival = oval->v.ival;
               break;

            case float_t :
               nval->v.fval = oval->v.fval;
               break;

            case none_t:
               wsi_printf( PD_MAKESTR("none_t type column found in select\n") );
               break;
         }
      }

      /* Append; keep last_row absolute. */
      if( last_row )
      {
         last_row->next = (wqrow_t)(nrow);
         last_row = nrow;
      }
      else
      {
         nrows = (wqrow_t)(nrow);
         last_row = nrow;
      }
      if( take_first )
         break;
   }

   if( last_row )
      last_row->next = NULL;

   if( count && out_table )
   {
      out_table->rows = nrows;
      out_table->tail = last_row;
      out_table->row_count = count;
      out_table->col_defs = col_defs_dup_rel_to_abs(tab->col_defs,col_count);
      out_table->col_count = col_count;
   }

   retVal = count;
   wss_db_unlock_table( tab );

   /* If any rows were returned, and a sorting criteria was requested,
    * sort.
    */
   if( count && out_table && ( order_by >= 0 ) )
   {
      int rc = row_sort( out_table, order_by, desc );
      if( rc < 0 )
         retVal = rc;
   }

select_cleanup_end:
   if( ( retVal < 0 ) && nrows )
   {
      row = nrows;
      while( row )
      {
         for( i = 0; i < col_count; i++ )
         {
            wqval_t nval;

            nval = row->vals + i;
            if( cd[i].dtype == pd_char_t_t )
               wss_free( nval->v.cval );
         }

         wss_free( row->vals );

         nrow = row->next;
         wss_free( row );
         row = nrow;
      }
   }

select_end:
   if( free_scrit )
      wss_db_free_criteria( scrit );

   return retVal;
}

/*  wss_delete */
/*! \brief Delete specified rows from a table
 *
 *  This function is similar to wss_select(), except that no DB data is
 *  returned, and there is no order for the delete operation.  The \a scrit
 *  parameter is identical to that for wss_select(), but in this case
 *  the selected rows are deleted, not returned.  If the \a scrit parameter
 *  is NULL, all rows are deleted.
 *
 *  \param table_id   Handle identifier for table
 *  \param free_scrit Flag which determines if the structures passed in
 *                    for the selection criteria will be freed with wss_free
 *  \param scrit      Pointer to selection criteria
 *  \param take_first Flag which determines if the search for rows that match
 *                    the selection criteria will stop after the first match
 *                    is found
 *
 *  \return 0 on success,
 *  \return -1 if table does not exist
 */
int EXPORT
wss_delete( u_long table_id,
            int free_scrit,       /* Do I free the wqsx_t? */
            wqsx_t scrit,         /* select_criteria() ptr*/
            int take_first )      /* Stop after first match? */
{
   wqtab_t tab;
   wqrow_t row, next_row;
   wqrow_t *prev_ptr;
   wqcd_t  cd;
   int num_rows_deleted = 0;
   int retValue, i, col_count;

   tab = wss_db_find_table_by_handle( table_id );
   if( tab == NULL )
   {
      retValue = -WRE_NOTFOUND;
      goto delete_end;
   }

   if( tab->flags & DB_FLAG_IMMUTABLE )
   {
      retValue = -WRE_PERM;
      wss_db_unlock_table( tab );
      goto delete_end;
   }

   row = tab->rows ? (wqrow_t)DB_REL2ABS(tab->rows) : NULL;
   prev_ptr = &(tab->rows);

   cd = (wqcd_t)DB_REL2ABS(tab->col_defs);
   col_count = tab->col_count;
   while( row )
   {
      wqval_t val;

      val = (wqval_t)DB_REL2ABS(row->vals);

      if( !eval_criteria( cd, row, scrit ) )
      {
         *prev_ptr = (wqrow_t)DB_ABS2REL(row);
         prev_ptr = &(row->next);
         row = row->next ? (wqrow_t)DB_REL2ABS(row->next) : NULL;
         continue;
      }

      for( i = 0; i < col_count; i++ )
      {
         if( cd[i].dtype == pd_char_t_t )
            wss_vm_free( wss_db_space, DB_REL2ABS(val[i].v.cval) );
      }
      wss_vm_free( wss_db_space, val );
      next_row = row->next ? (wqrow_t)DB_REL2ABS(row->next) : NULL;
      tab->row_count--;
        
      /*
       * Post delete row event
       */
      wss_wrvevent_post_deleterow( tab, row );

      wss_vm_free( wss_db_space, row );
      row = next_row;
      num_rows_deleted++;
      if( take_first )
         break;
   }
   *prev_ptr = row ? (wqrow_t)DB_ABS2REL(row) : NULL;
   retValue = num_rows_deleted;
   wss_db_unlock_table( tab );

delete_end:
    if( free_scrit )
        wss_db_free_criteria( scrit );

    return retValue;
}


/*  wss_db_set_row_value_from_string */
/*! \brief Sets the value at the specified row and column
 *
 *  \param cols  Pointer to column definition for table
 *  \param row   Pointer to row structure
 *  \param col   Index of column of value to change
 *  \param value Pointer to string containing value to set
 *
 *  \warning Assumes locking has already been done
 */
void EXPORT
wss_db_set_row_value_from_string( wqcd_t cols, wqrow_t row, int col,
                                  pd_char_t *value )
{
   wqval_t val_to;
   enum wql_dtype val_to_type;

   /* Find the value.   In this case, we also need to learn the type */
   val_to_type = cols[col].dtype;
   val_to = (wqval_t)DB_REL2ABS( row->vals ) + col;

   switch( val_to_type )
   {
      case pd_char_t_t:
         wss_vm_free( wss_db_space, DB_REL2ABS( val_to->v.cval ) );
         if( value )
            val_to->v.cval = (pd_char_t *)DB_ABS2REL( strdup_rel( value ) );
         else
            val_to->v.cval = (pd_char_t *)DB_ABS2REL( strdup_rel( PD_MAKESTR("") ) );
         break;

      case int_t:
         val_to->v.ival = wss_atoi(value);
         break;

      case float_t:
         val_to->v.fval = (float) wss_atof(value); /* TDB: Should we use double
                                                           instead of floats? */
         break;

      case none_t:
      default:
         wsi_printf(PD_MAKESTR("wss_db_set_row_value_from_string: none_t typed column found.") );
         break;
   }
}


/*  wss_update */
/*! \brief Update values in a table
 *
 *  This function is similar to wss_select(), except that no DB data is
 *  returned, there is no order for the delete operation, and the DB is
 *  possibly modified.  The \a scrit parameter is identical to that for
 *  wss_select(), but in this case the selected rows are modified, not
 *  returned.  If the \a scrit parameter is NULL, all rows are selected.
 * 
 *  For each selected row, the \a set_script is "run" to update that row.
 * 
 *  If the \a take_first option is set, assume there will only be one
 *  row that matches the criteria.  Move that row to the front on the
 *  assumption that we may ask about it again shortly.
 *
 *  \param table_id   Handle identifier for table
 *  \param free_scrit Flag which determines if the structures passed in
 *                    for the selection criteria will be freed with wss_free
 *  \param scrit      Pointer to selection criteria
 *  \param take_first Flag which determines if the search for rows that match
 *                    the selection criteria will stop after the first match
 *                    is found.
 *  \param set_script Pointer to a list of assignment expressions
 * 
 *  \return The number of rows affected on success,
 *  \return -1 if the table did not exist
 */
int EXPORT
wss_update( u_long table_id,
            int free_scripts, /*Do I free the scrit _and_ the set_script?*/
            wqsx_t scrit,
            int take_first,   /* Do I stop after the first match? */
            wqusl_t set_script )
{
   wqtab_t tab;
   int count = 0; /* Number of rows selected by the where clause */
   wqrow_t row, next_row;
   wqcd_t col_defs_abs;
   int retVal;
   wqrow_t *prev_ptr;

   tab = wss_db_find_table_by_handle( table_id );
   if( tab == NULL )
   {
      retVal = -WRE_NOTFOUND;
      goto update_end;
   }

   if( tab->flags & DB_FLAG_IMMUTABLE )
   {
      retVal = -WRE_PERM;
      wss_db_unlock_table( tab );
      goto update_end;
   }

   /*
    * For every row: if if passes the select script, give it to the update
    * script.
    */
   row = tab->rows ? (wqrow_t)DB_REL2ABS(tab->rows) : NULL;

   col_defs_abs = (wqcd_t)DB_REL2ABS(tab->col_defs);
    
   prev_ptr = &(tab->rows);

   while( row )
   {
      next_row = (row->next) ? (wqrow_t)DB_REL2ABS(row->next) : NULL;
      if( !eval_criteria( col_defs_abs, row, scrit ) )
      {
         /* If we might be moving multiple rows, we'd
          * have to update *prev_ptr
          */
         prev_ptr = &(row->next);
         row = next_row;
         continue;
      }
      eval_expression_list( col_defs_abs, row, set_script );
      count++;
        
      /*
       * Post update row event
       */
      wss_wrvevent_post_updaterow( tab, row );

      if( take_first )
      {
         /* Removes it from the list */
         *prev_ptr = (wqrow_t)DB_ABS2REL(next_row);

         /* Add it to the start */
         row->next = tab->rows;
         tab->rows = (wqrow_t)DB_ABS2REL(row);
         break;
      }
      row = next_row;
   }
   retVal = count;
   wss_db_unlock_table( tab );

update_end:
   if( free_scripts )
   {
      wss_db_free_criteria( scrit );
      wss_db_free_wqusl( set_script );
   }

   return retVal;
}

/*
 * Utility functions.
 * 
 * These functions are used in the code generated for DB wrappers.
 * They have been moved here to help shrink the size of the generated
 * wrappers.  (8k is a little much!)
 *
 * Some utility functions have been left behind... the ones that
 * have been implemented as macros.
 *
 * The functions here follow the standard naming convention.  The in-kernel
 * linker magically makes them follow the old "wr_foo" names in the wrapper.
 * 
 * Locking issues will be dealt with by the compilier, essentially locking 
 * a per-wrapper mutex at the "{" of "WQL {" and unlocking it on the matching
 * "}"
 */

/*  wss_db_mcol_def */
/*! \brief Makes a column definition
 *
 *  The arguments passed in should match the following:
 *  \code
 *     ( pd_char_t *name, enum wql_dtype dt, int max_len)\{count\}
 *  \endcode
 *
 *  \param count Number of arguments
 *
 *  \return Pointer to column definition on success,
 *  \return NULL on failure
 *  \internal
 */
wqcd_t EXPORT
wss_db_mcol_def( int count, ... )
{
   va_list ap;
   wqcd_t  retval;

   va_start( ap, count );
   retval = wss_db_mcol_def_va( count, ap );
   va_end( ap );

   return retval;
}


/*  wss_db_mcol_def */
/*! \brief Makes a column definition
 *
 *  \param count Number of arguments
 *  \param ap    Variable argument list
 *
 *  \return Pointer to column definition on success,
 *  \return NULL on failure
 *  \internal
 */
wqcd_t EXPORT
wss_db_mcol_def_va( int count, va_list ap )
{
   wqcd_t retval = (wqcd_t)wss_malloc( sizeof(*retval) * count );
   int i;
    
#ifdef DEBUG_DB
   wsi_printf( PD_MAKESTR("Creating new column definition array.\n") );
#endif
   if( retval )
   {
      for( i = 0; i < count; i++ )
      {
         pd_char_t *name = va_arg( ap, pd_char_t * );
         int   type = va_arg( ap, int );
         int   len  = va_arg( ap, int );
        
         retval[i].name    = name;
         retval[i].dtype   = type;
         retval[i].max_len = len;
        
#ifdef DEBUG_DB
         wsi_printf( PD_MAKESTR("Col %d: name = %s, type = %d, max_len = %d\n"),
                 i, name, type, len );
#endif
      }
   }
   return retval;
}

/*  wss_db_mrow */
/*! \brief Creates a row
 *
 *  Just creates a new structure and sets the data elements to the
 *  values passed in.
 *
 *  \param vals Pointer to the list of values for this row
 *  \param next Pointer to the next row in the table after this one
 *
 *  \return Pointer to new row structure on success,
 *  \return NULL on failure
 *  \internal
 */
wqrow_t EXPORT
wss_db_mrow( wqval_t vals, wqrow_t next )
{
   wqrow_t row = (wqrow_t)wss_malloc( sizeof(wqrow) );
   if( row )
   {
      row->vals = vals;
      row->next = next;
   }
   return row;
}

/*
 * WQL support: wss_select.
 * These functions create the selection tree.  Most of them
 * just create a structure, fill in the elements from the function
 * arguments, and return the structure.  They exist to allow 
 * compact code to be generated for creating the data structure
 * wss_select needs
 * 
 * I could make things smaller by making the operations one of the arguments,
 * and I don't know why I havn't.
 */

/*  wss_db_mwqsx_and */
/*! \brief Create an AND operator node for a selection tree
 *
 *  \param left  Pointer to the left leaf of this node
 *  \param right Pointer to the right leaf of this node
 *
 *  \return Pointer to new selection tree node on success,
 *  \return NULL on failure
 *  \internal
 */
wqsx_t EXPORT
wss_db_mwqsx_and( wqsx_t left, wqsx_t right )
{
   ntx_t node;
   wqsx_t retVal = NULL;

   node = (ntx_t)wss_malloc( sizeof(*node) );
   if( node )
   {
      node->logop = and_op;
      node->left  = left;
      node->right = right;  

      retVal = (wqsx_t)wss_malloc( sizeof(*retVal) );
      if( retVal )
      {
         retVal->which = nonterm;
         retVal->un.children = node;
      }
      else
      {
         wss_free( node );
      }
   }
    
   return retVal;
}

/*  wss_db_mwqsx_or */
/*! \brief Create an OR operator node for a selection tree
 *
 *  \param left  Pointer to the left leaf of this node
 *  \param right Pointer to the right leaf of this node
 *
 *  \return Pointer to new selection tree node on success,
 *  \return NULL on failure
 *  \internal
 */
wqsx_t EXPORT
wss_db_mwqsx_or( wqsx_t left, wqsx_t right )
{
   ntx_t node;
   wqsx_t retVal = NULL;

   node = (ntx_t)wss_malloc( sizeof(*node) );
   if( node )
   {
      node->logop = or_op;
      node->left = left;
      node->right = right;  

      retVal = (wqsx_t)wss_malloc( sizeof(*retVal) );
      if( retVal )
      {
         retVal->which = nonterm;
         retVal->un.children = node;
      }
      else
      {
         wss_free( node );
      }
   }
    
   return retVal;
}

/*  wss_db_mwqsx_terminal */
/*! \brief Create a terminal node for a selection tree
 *
 *  \param terminal Pointer to the terminal expression
 *
 *  \return Pointer to new selection tree node on success,
 *  \return NULL on failure
 *  \internal
 */
wqsx_t EXPORT
wss_db_mwqsx_terminal( tx_t terminal )
{
   wqsx_t retVal = (wqsx_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->which = term;
      retVal->un.exp = terminal;
   }
   
   return retVal;
}

/*  wss_db_mwqsx_not */
/*! \brief Create a NOT operator node for a selection tree
 *
 *  \param expression Pointer to the expression node to negate
 *
 *  \return Pointer to new selection tree node on success,
 *  \return NULL on failure
 *  \internal
 */
wqsx_t EXPORT
wss_db_mwqsx_not( wqsx_t expression )
{
   ntx_t node;
   wqsx_t retVal = NULL;

   node = (ntx_t)wss_malloc( sizeof(*node) );
   if( node )
   {
      node->logop = not_op;
      node->left = expression;
      node->right = NULL;

      retVal = (wqsx_t)wss_malloc( sizeof(*retVal) );
      if( retVal )
      {
         retVal->which = nonterm;
         retVal->un.children = node;
      }
      else
      {
         wss_free( node );
      }
   }
    
   return retVal;
}

/*  wss_db_mtx */
/*! \brief Create a terminal expression
 *
 *  \param left  Pointer to the left leaf of this node
 *  \param right Pointer to the right leaf of this node
 *  \param sxop  Selection operation
 *
 *  \return Pointer to terminal expression on success,
 *  \return NULL on failure
 *  \internal
 */
tx_t EXPORT
wss_db_mtx( sxval_t left, sxval_t right, enum sxop sxop )
{
   tx_t retVal = (tx_t)wss_malloc( sizeof(*retVal) );
    
   if( retVal )
   {
      retVal->left = left;
      retVal->right = right;
      retVal->sxop = sxop;
   }

   return retVal;
}


/*  wss_db_msxval_field */
/*! \brief Create a selection expression field reference
 *
 *  \param column Column index of field
 *
 *  \return Pointer to selection expression value on success,
 *  \return NULL on failure
 *  \internal
 */
sxval_t EXPORT
wss_db_msxval_field( int column )
{
   sxval_t retVal = (sxval_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->dtype = pd_char_t_t;
      retVal->which = field;
      retVal->v.f.column = column;
      retVal->owned = 0;
   }

   return retVal;
}


/*  wss_db_msxval_lit_int */
/*! \brief Create an integer literal for a selection expression
 *
 *  \param i Integer value
 *
 *  \return Pointer to selection expression value on success,
 *  \return NULL on failure
 *  \internal
 */
sxval_t EXPORT
wss_db_msxval_lit_int( int i )
{
   sxval_t retVal = (sxval_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->dtype = int_t;
      retVal->which = literal;
      retVal->v.l.ival = i;
      retVal->owned = 0;
   }

   return retVal;
}

/*  wss_db_msxval_lit_float */
/*! \brief Create a float literal for a selection expression
 *
 *  \param i Float value
 *
 *  \return Pointer to selection expression value on success,
 *  \return NULL on failure
 *  \internal
 */
sxval_t EXPORT
wss_db_msxval_lit_float( double d )
{
   sxval_t retVal = (sxval_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->dtype = float_t;
      retVal->which = literal;
      retVal->v.l.fval = (float)d;
      retVal->owned = 0;
   }

   return retVal;
}

/*  wss_db_msxval_lit_string */
/*! \brief Create a string literal for a selection expression
 *
 *  \param str String value
 *
 *  \return Pointer to selection expression value on success,
 *  \return NULL on failure
 *  \internal
 */
sxval_t EXPORT
wss_db_msxval_lit_string( const pd_char_t *str )
{
   sxval_t retVal = (sxval_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->dtype = pd_char_t_t;
      retVal->which = literal;
      retVal->v.l.cval = str;
      retVal->owned = 0;
   }

   return retVal;
}

/*
 * WQL support: wss_select.
 * This structure and accompanying functions support the magic
 * DBROW variables.
 * 
 * The functions that are implimented as macros in codefrags.h
 * have a reference function implimentation here.
 */


#if 0
/*
 * Start should initialize the dbrow to the table used.
 * Next will point the current row pointer to the next row.
 * isempty will return TRUE if the cur pointer has iterated
 * the whole table.
 * Finish cleans up data.
 */
/* Non-macro version for documentation purposes */
int EXPORT
wss_db_dbrow_start( dbrow *dbrp )
{
   dbrp->cur = dbrp->results.rows;
   return( dbrp != NULL );
}

/* Non-macro version for documentation purposes */
int EXPORT
wss_db_dbrow_next( dbrow *dbrp )
{
   dbrp->cur = dbrp->cur->next;
   return( dbrp->cur != NULL );
}

/* Non-macro version for documentation purposes */
int EXPORT
wss_db_dbrow_isempty( dbrow *dbrp )
{
   return( dbrp->results.rows == NULL );
}
#endif


/*  wss_db_dbrow_finish */
/*! \brief Free the data used by selected rows
 *
 *  \param dbrp Pointer to a row structure
 *  \internal
 */
void EXPORT
wss_db_dbrow_finish( dbrow *dbrp )
{
   wqrow_t rows = dbrp->results.rows;
   wqcd_t cols = dbrp->results.col_defs;
   int count = dbrp->results.col_count;
   int i;

   if( !cols )
      return;

   dbrp->cur = NULL;
   /* Free the data. */

   while( rows )
   {
      wqrow_t row = rows;
      wqval_t vals = row->vals;

      rows = rows->next;
      for( i = 0; i < count; i++ )
      {
         if( cols[i].dtype == pd_char_t_t )
            wss_free( (void*)vals[i].v.cval );
      }
      wss_free( (void*)vals );
      wss_free( (void*)row );
   }

   for( i = 0; i < count; i++ )
      wss_free( (void*)cols[i].name );

   wss_free( (void*)cols );
    
   dbrp->results.rows = NULL;
   dbrp->results.col_defs = NULL;
   dbrp->results.col_count = 0;
}

/*
 * WQL support: wss_update.
 * These functions create the update operation tree.
 * Similar to the wss_select support functions, most of them
 * just create a structure and set its elements equal to the 
 * parameters passed in.
 */

/*  wss_db_musl */
/*! \brief Creates an update set list structure from the parameters.
 *
 *  \param left The lval in the expression, usually a column
 *  \param right The rval in the expression.
 *  \param next The next node in the list
 *
 *  \return a constructed and allocated list structure.
 *  \internal
 */
wqusl_t EXPORT
wss_db_musl( sxval_t left, wquexp_t right, wqusl_t next )
{
   wqusl_t retVal = (wqusl_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->left = left;
      retVal->right = right;
      retVal->next = next;
   }
    
   return retVal;
}

/*  wss_db_muexp */
/*! \brief Creates an update expression structure from the parameters.
 *
 *  \param left The left side of the expression
 *  \param right The right side of the expression, or NULL for a
 *  \param right unary operation.
 *  \param numop The type of operation.
 *
 *  \return a constructed and allocated expression structure.
 *  \internal
 */
wquexp_t EXPORT
wss_db_muexp( wquexp_t left, wquexp_t right, enum numop numop )
{
   wquexp_t retVal = (wquexp_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->which = nonterm;
      retVal->w.e.left = left;
      retVal->w.e.right = right;
      retVal->w.e.numop = numop;
   }
    
   return retVal;
}

/*  wss_db_muexp_term */
/*! \brief Creates an update expression terminal structure from the parameter.
 *
 *  \param left The value of the terminal node.
 *
 *  \return a constructed and allocated expression terminal structure.
 *  \internal
 */
wquexp_t EXPORT
wss_db_muexp_term( sxval_t value )
{
   wquexp_t retVal = (wquexp_t)wss_malloc( sizeof(*retVal) );

   if( retVal )
   {
      retVal->which = term;
      retVal->w.terminal = value;
   }
    
   return retVal;
}

/*
 * WQL support.  Probably not done; these casts are questionable.
 */

/*  wss_db_mvals */
/*! \brief Creates an array of wql values from the arguments passed in
 *
 *  \param count The number of (type,value) pairs passed in.
 *
 *  \return a constructed and allocated array of wql values.
 *  \internal
 */
wqval_t EXPORT
wss_db_mvals( int count, ... )
{
   va_list ap;
   wqval_t retval;

   va_start( ap, count );
   retval = wss_db_mvals_va( count, ap );
   va_end( ap );

   return retval;
}

/*  wss_db_mvals_va */
/*! \brief Creates an array of wql values from the arguments passed in
 *
 *  \param count The number of (type,value) pairs passed in.
 *  \param ap The va_list that contains the (type,value) pairs.
 *
 *  \return a constructed and allocated array of wql values.
 *  \internal
 */
wqval_t EXPORT
wss_db_mvals_va( int count, va_list ap )
{
   /*
    * (enum wql_dtype dt, {pd_char_t*|int|float} value){,count}
    */
   wqval_t retval = (wqval_t)wss_malloc( sizeof(*retval) * count );
   int i;

   if( retval )
   {
      for( i = 0; i < count; i++ )
      {
         int v; pd_char_t *c; float f;
         enum wql_dtype dt = va_arg( ap, int ); /* Assumes enum promotes to int */
         switch( dt )
         {
            case pd_char_t_t:
               c = va_arg( ap, pd_char_t * );
               retval[i].v.cval = c;
               break;

            case int_t:
               v = va_arg( ap, int );
               retval[i].v.ival = v;
               break;

            case float_t:
               /* float is promoted to double in ... */
               f = va_arg( ap, double );
               retval[i].v.fval = f;
               break;

            case none_t:
               break;
         }
      }
   }
   return retval;
}


/*  wss_db_insert_nulls */
/*! \brief Add a new row of null values to the specified table
 *
 *  \param table_id Handle identifier for table
 *
 *  \return 0 on success,
 *  \return \b WRE_NOMEM if we could not allocate enough memory,
 *  \return \b WRE_NOTFOUND if the table did not exist,
 *  \return \b WRE_INVALID_PARAMETER if any of the values were invalid
 */
int EXPORT
wss_db_insert_nulls( u_long table_id )
{
   int     i;
   wqtab_t tab;
   wqrow_t row;
   wqval_t vals;

   tab = wss_db_find_table_by_handle( table_id );
   if( tab == NULL )
      return WRE_NOTFOUND;

   row = (wqrow_t)wss_malloc( sizeof(*row) );
   if( row == NULL )
      return WRE_NOMEM;
   vals = (wqval_t)wss_malloc( sizeof(*vals) * tab->col_count );
   if( vals == NULL )
   {
      wss_free( row );
      return WRE_NOMEM;
   }
   row->vals = vals;
   row->next = NULL;

   for( i = 0; i < tab->col_count; i++ )
   {
      wqcd_t cd = (wqcd_t)DB_REL2ABS( tab->col_defs ) + i;

      switch( cd->dtype )
      {
         case pd_char_t_t:
#ifdef USER_MODE_TEST
            vals[i].v.cval = strdup(PD_MAKESTR(""));
#else
            vals[i].v.cval = wss_strdup(PD_MAKESTR(""));
#endif /* USER_MODE_TEST */
            break;

         case int_t:
            vals[i].v.ival = 0;
            break;

         case float_t:
            vals[i].v.fval = 0.0;
            break;

         case none_t:
            break;
      }
   }

   wss_db_unlock_table( tab );

   return wss_insert( table_id, TRUE, row );
}


/*  wss_db_delete_by_handle */
/*! \brief Delete specified row from a table
 *
 *  This function works like wss_delete(), except that it only drops a
 *  single row and the row is specified by its handle.
 *
 *  \param table_id Handle identifier for table
 *  \param row_id   Handle identifier for row
 *
 *  \return 0 on success,
 *  \return \b WRE_NOTFOUND if table is not found in the database
 */
int EXPORT
wss_db_delete_by_handle( u_long table_id, u_long row_id )
{
   wqtab_t  tab;
   wqrow_t  row;
   wqrow_t  arow;
   wqrow_t *last_row;

   tab = wss_db_find_table_by_handle( table_id );
   if( tab == NULL )
      return WRE_NOTFOUND;

   for( row = tab->rows, last_row = NULL; row;
        last_row = &arow->next,
        row = (wqrow_t)(arow->next?arow->next:NULL) )
   {
      arow = (wqrow_t)DB_REL2ABS( row );
      if( (u_long)row == row_id )
      {
         int     i;
         wqcd_t  cd;
         wqval_t val;

         if( last_row )
            *last_row = arow->next;
         else
            tab->rows = arow->next;

         cd = (wqcd_t)DB_REL2ABS( tab->col_defs );
         val = (wqval_t)DB_REL2ABS( arow->vals );

         for( i = 0; i < tab->col_count; i++ )
            if( cd[i].dtype == pd_char_t_t )
                wss_vm_free( wss_db_space, DB_REL2ABS( val[i].v.cval ) );

         wss_vm_free( wss_db_space, val );
         tab->row_count--;

         /*
          * Post delete row event
          */
         wss_wrvevent_post_deleterow( tab, arow );

         wss_vm_free( wss_db_space, arow );
      }
   }

   wss_db_unlock_table( tab );

   return 0;
}

