/*! \file wss_dispatch.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_dispatch.c,v 1.13 2001/02/23 19:02:46 stevek Exp $
 *
 * \brief   WSS implementation of the event dispatching functions
 * \author  Douglas Kilpatrick
 * \version $Revision: 1.13 $
 * \date    $Date: 2001/02/23 19:02:46 $
 *
 * \warning This does not include a "call real system call" step.  For that,
 *          see the WSI layer.
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>
#include <wss_malloc.h>
#include <wss_arg.h>

#include <wss_utilities.h>
#include <wss_dispatch.h>
#include <wss_wrapper.h>

#include <wsi_os.h>


/***************************************************************************\
 * Private Functions                                                       *
\***************************************************************************/

int EXPORT
_wss_dispatch_internal( int space, pd_syscall_id_t code, wss_proc_t *wrp,
                        int flags, int arg_count, void *uap[] )
{
    wss_wrapper_instance_t     *wi;
    wss_wrapper_instance_t     *wi_prev;
    wss_wrapper_data_t          wrd = {0,0,0};
    int                         rval = 0;
    int                         retval = 0;

    if (!wrp)
        return (1*4|WR_DENY);

    wi_prev = NULL;
    for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
         wi != NULL;
         wi_prev = wi, wi = wi->wi_next )
    {
        if (wi->wi_wrapper->w_dispatch_pre)
        {
            wrd.wiv = wi->wi_data;
#ifdef WR_DEBUG_DISPATCH
            wss_printf(PD_MAKESTR("%d/%d: Calling pre\n"), space, code);
#endif
            retval |= (*(wi->wi_wrapper->w_dispatch_pre))(space, code, uap, 
                    &rval, wrp, &wrd, flags);
#ifdef WR_DEBUG_DISPATCH
            wss_printf(PD_MAKESTR("%d/%d: rval = %d\n"), space, code, rval);
#endif
            if ( ( retval & WR_MASK1 ) == WR_DENY )
            {
#ifdef WR_DEBUG_DISPATCH
                wss_printf(PD_MAKESTR("%d/%d: *wrd.errnum = %d\n"),
                        space, code, *wrd.errnum);
#endif
                break;
            }
        }
    }
    /* No real system call to call */
    if (wi == NULL)
        wi = wi_prev;
    for ( ; wi ; wi = wi->wi_prev )
    {
        if (wi->wi_wrapper->w_dispatch_post)
        {
            wrd.wiv = wi->wi_data;
#ifdef WR_DEBUG_DISPATCH
            wss_printf(PD_MAKESTR("%d/%d: Calling post\n"), space, code);
#endif
            (*(wi->wi_wrapper->w_dispatch_post))(space, code, uap, &rval,
                                            wrp, &wrd, flags);
            /* Too late for something to try and signal failure now */
#ifdef WR_DEBUG_DISPATCH
            wss_printf(PD_MAKESTR("%d/%d: rval = %d\n"), space, code, rval);
#endif
        }

    }
    
    return retval;
}


/***************************************************************************\
 * Public Functions                                                        *
\***************************************************************************/

/*  wss_dispatch */
/*! \brief Thread-safe system call dispatch routine
 *
 *  \param space     Event space identifier
 *  \param code      Indentifier of calling thread
 *  \param wrp       Pointer to the wrapper process structure for this process
 *  \param flags     Flags to pass to routine
 *  \param arg_count Number of arguments in argument list
 *
 *  \return 0 if successful,
 *  \return error code otherwise
 */
int EXPORT
wss_dispatch( int space, pd_syscall_id_t code, wss_proc_t *wrp,
              int flags, int arg_count, ... )
{
    va_list  ap;
    void    *uap[10];
    int      i;
    int      retval;

    if (!wrp)
       return (1*4|WR_DENY);

    bzero( uap, sizeof(uap) );

    va_start(ap, arg_count);
    if (arg_count > sizeof(uap)/sizeof(uap[0]))
        arg_count = sizeof(uap)/sizeof(uap[0]);
    for (i = 0; i < arg_count ; i++)
    {
        uap[i] = va_arg(ap, void*);
    }
    va_end(ap);

    WSS_MUTEX_ENTER( wrp->wrp_wi_list );
    retval = _wss_dispatch_internal( space, code, wrp, flags, arg_count, uap );
    WSS_MUTEX_EXIT( wrp->wrp_wi_list );

    wss_arg_donesyscall( wrp, wsi_os_curthreadid(), space, code, uap );

    return retval;
}
                

/*  wss_dispatch_unsafe */
/*! \brief System call dispatch routine
 *
 *  This call is used by wrappers where the wrp_wi_list lock
 *  will already be held when the original calling system call
 *  passes through wss_dispatch().
 *
 *  \param space     Event space identifier
 *  \param code      Indentifier of calling thread
 *  \param wrp       Pointer to the wrapper process structure for this process
 *  \param flags     Flags to pass to routine
 *  \param arg_count Number of arguments in argument list
 *  \param ...       Variable argument list
 *
 *  \warning Assumes wrp_wi_list lock held on wrp
 *
 *  \return 0 if successful,
 *  \return error code otherwise
 *
 */
int EXPORT
wss_dispatch_unsafe( int space, pd_syscall_id_t code, wss_proc_t *wrp,
                    int flags, int arg_count, ... )
{
    va_list  ap;
    void    *uap[10];
    int      i;
    int      retval;

    if (!wrp)
       return (1*4|WR_DENY);

    bzero( uap, sizeof(uap) );

    va_start(ap,arg_count);

    if (arg_count > sizeof(uap)/sizeof(uap[0]))
        arg_count = sizeof(uap)/sizeof(uap[0]);

    for (i = 0; i < arg_count ; i++) {
        uap[i] = va_arg(ap, void*);
    }
    va_end(ap);

    retval = _wss_dispatch_internal( space, code, wrp, flags, arg_count, uap );

    wss_arg_donesyscall( wrp, wsi_os_curthreadid(), space, code, uap );

    return retval;
}

