/*! \file wss_evspace.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_evspace.c,v 1.16 2001/02/23 19:02:46 stevek Exp $
 *
 * \brief   Implementation of the event space functions
 * \author  Douglas Kilpatrick
 * \version $Revision: 1.16 $
 * \date    $Date: 2001/02/23 19:02:46 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_evspace.h>
#include <wss_malloc.h>
#include <wss_arg.h>
#include <wss_mutex.h>
#include <wss_utilities.h>

#include <eventfs.h>


/***************************************************************************\
 * Private structures and types                                            *
\***************************************************************************/

typedef struct evspace_node evspace_node_t;

struct evspace_node
{
    wss_util_rbnode_t   parent;
    pd_char_t         *name;
    int                 id;
};


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static wss_util_rbtree_t *wss_evspace_tree = NULL;
WSS_MUTEX_PROTECT       ( wss_evspace );


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/


static int
evspace_compfunc(void *n1, void *n2)
{
    evspace_node_t *node1 = n1;
    evspace_node_t *node2 = n2;
    return wsi_strcmp(node1->name, node2->name);
}

static void *
evspace_clonefunc(void *n)
{
    evspace_node_t *retval = wss_malloc(sizeof(*retval));
    wss_util_rbnode_init(&(retval->parent));
    if (!n) {
        retval->name = 0;
        retval->id = 0;
    } else {
        evspace_node_t *orig = n;
        retval->name = wss_strdup(orig->name);
        retval->id = orig->id;
    }
    return retval;
}

static void
evspace_free(void *n)
{
    evspace_node_t *node = n;
    if (node) {
        if (node->name)
            wss_free(node->name);
        wss_free(node);
    }
}

static int
find_space(pd_char_t *name)
{
    evspace_node_t temp, *results;

    WSS_MUTEX_ENTER(wss_evspace);
    if (!wss_evspace_tree)
        wss_evspace_tree = wss_util_rbtree_new( evspace_compfunc,
                                                evspace_clonefunc,
                                                evspace_free );

    wss_util_rbnode_init(&(temp.parent));
    temp.name = name;

    results = (evspace_node_t *)wss_util_rbtree_find(wss_evspace_tree,
                                                     &(temp.parent));
    WSS_MUTEX_EXIT(wss_evspace);

    return results ? results->id : -1;

}

static void
add_space(pd_char_t *name, int id)
{
    evspace_node_t *new_node;
#ifdef DEBUG_EVSPACE
    wss_util_rbnode_t *old;
#endif

    WSS_MUTEX_ENTER(wss_evspace);
    if (!wss_evspace_tree)
        wss_evspace_tree = wss_util_rbtree_new( evspace_compfunc,
                                                evspace_clonefunc,
                                                evspace_free );
    
    new_node = evspace_clonefunc(0);
    new_node->name = wss_strdup(name);
    new_node->id = id;
    
#ifdef DEBUG_EVSPACE
    old = wss_util_rbtree_delete(wss_evspace_tree, &(new_node->parent));
    if (old) {
        panic("duplicated space added.");
        evspace_free(old);
    }
#endif

    wss_util_rbtree_insert(wss_evspace_tree, &(new_node->parent));
    WSS_MUTEX_EXIT(wss_evspace);
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/


/*  wss_evspace_init */
/*! \brief Initialize the system event space
 *
 *  Constructs the mutex protecting the event space tree and adds initial
 *  event spaces for bsd, solaris, and win32
 */
void EXPORT
wss_evspace_init()
{
    WSS_MUTEX_INIT(wss_evspace);
#if defined(WR_FREEBSD)
    add_space(PD_MAKESTR("freebsd"), 0);
#elif defined(WR_LINUX)
    add_space(PD_MAKESTR("Linux"), 0);
#elif defined(WR_SOLARIS)
    add_space(PD_MAKESTR("solaris"), 0);
#elif defined(WR_WINNT)
    add_space(PD_MAKESTR("win32"), 0);
#endif
#ifdef DEBUG_EVSPACE
    wsi_printf(PD_MAKESTR("wss_evspace_init: finished.\n"));
#endif /* DEBUG_EVSPACE */
}


/*  wss_evspace_shutdown */
/*! \brief Shutdown the system event space
 *
 *  Destroys the system event space tree and the mutex protecting it.
 */
void EXPORT
wss_evspace_shutdown()
{
    WSS_MUTEX_ENTER(wss_evspace);
    wss_util_rbtree_free(wss_evspace_tree);
    wss_evspace_tree = NULL;
    WSS_MUTEX_EXIT(wss_evspace);

    WSS_MUTEX_DESTROY(wss_evspace);
}

/*  wss_evspace_resolve */
/*! \brief Resolve the identifier for an event space from the name
 *
 *  First we search for the event space.  If we can find an already
 *  existing event space, we return its identifier, otherwise we
 *  create a new event space and return its identifier.
 *
 *  \param name Name of event space
 *
 *  \return Identifier for the named event space
 */
int EXPORT
wss_evspace_resolve(pd_char_t *name)
{
    static int counter = 0;

    int retval = find_space(name);
    
    if (retval < 0) {
        WSS_MUTEX_ENTER(wss_evspace);
        retval = ++counter;
        WSS_MUTEX_EXIT(wss_evspace);
        add_space(name, retval);
    }
    return retval;
}
    
