/*! \file wss_init.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_init.c,v 1.54 2001/06/21 21:37:50 stevek Exp $
 * 
 * \brief   Handles initialization of all WSS functionality modules
 * \author  Doug Kilpatrick, Steve Kiernan
 * \version $Revision: 1.54 $
 * \date    $Date: 2001/06/21 21:37:50 $
 * 
 * Gets called at startup.  Put WSS initialization calls here.
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_arg.h>
#include <wss_criteria.h>
#include <wss_db.h>
#include <wss_log.h>
#include <wss_module.h>
#include <wss_process.h>
#include <wss_vm.h>
#include <wss_wrapper.h>
#include <wss_wrvevent.h>
#include <wss_audit.h>
#include <wss_init.h>
#include <wss_evspace.h>
#include <wss_linker.h>
#include <wss_utilities.h>
#include <wss_path.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern pd_char_t          *wss_db_filename;
extern int                  wss_db_size;
extern volatile pd_char_t *wss_log_path;


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/*! Handle to the global configuration database table */
u_long wss_config_db_handle;

/*! Home directory for WSS support files */
pd_char_t *wss_init_home = NULL;

/*! Directory where wrapper files are located */
pd_char_t *wss_wrap_dir = NULL;

/*! Initialization file */
pd_char_t *wss_conf_file = NULL;

/*!
 * Table of configuration entries available via wss.conf file
 *
 * \sa WSS_INIT_ENTRY
 */
wss_init_entry_t wss_init_globals[] =
{
   WSS_INIT_ENTRY( "dbpath",  WSS_INIT_TYPE_STRING, &wss_db_filename ),
   WSS_INIT_ENTRY( "dbsize",  WSS_INIT_TYPE_INT,    &wss_db_size ),
   WSS_INIT_ENTRY( "logpath", WSS_INIT_TYPE_STRING, &wss_log_path ),
   WSS_INIT_ENTRY( "home",    WSS_INIT_TYPE_STRING, &wss_init_home ),
   WSS_INIT_ENTRY( "wrapdir", WSS_INIT_TYPE_STRING, &wss_wrap_dir ),
   WSS_INIT_ENTRY( "wssconf", WSS_INIT_TYPE_STRING, &wss_conf_file ),

   { NULL, 0 , NULL }
};


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_init */
/*! \brief Initialize the WSS
 *
 *  Clean up if an error is encountered during initialization.
 *
 *  \param init_flag Flag to indicate wether or not the database should be
 *                   initialized and/or configuration data loaded
 *
 *  \return 0 on success,
 *  \return EPERM on failure
 */
int EXPORT
wss_init( int init_flags )
{
   int db_init_ret = 0;

   wss_malloc_init();

   /* Initialize default init entries */
   if( init_flags & WSS_INIT_DB )
   {
      wss_db_filename = wss_strdup( PD_MAKESTR(WSS_DB_FILENAME) );
#ifdef WR_LINUX
      wss_db_size = 8; /* DB size in megabytes.  Linux does not have
                        * pageable kernel memory.
                        */
#else
      wss_db_size = 16; /* DB size in megabytes */
#endif
   }
   wss_init_home = wss_strdup( PD_MAKESTR(WSS_HOME_PATH) );
   wss_log_path = wss_strdup( PD_MAKESTR(WSS_LOG_FILENAME) );

   /* Read init configuration */
   if( init_flags & WSS_INIT_CONF )
      wss_init_read_conf();

   /* Initialize the primary parts of the WSS */
   wss_proc_init();
   wss_log_init();
   wss_linker_init();
   wss_mod_init();
   wss_wrapper_init();
   wss_lib_init();
   wss_crit_init();
   wss_arg_init();
   wss_wrvevent_init();
   wss_audit_init();
   wss_evspace_init();

   /* Initialize the VM and map in the WSS database */
   wss_vm_init();
   if( init_flags & WSS_INIT_DB )
   {
      db_init_ret = wss_db_init( ( init_flags & WSS_INIT_DB ) ? TRUE : FALSE );

      if( db_init_ret == 0 )
      {
         wqcd_t cd;

         /* Initialize the configuration table */
         if( !wss_db_find_table_by_names( PD_MAKESTR(""), 0,
                                          PD_MAKESTR("CONFIG") ) )
         {
            wqtab *tab;
#if defined(WR_UNIX) && 0
            wqrow *first_row;
            wqrow *next_row;
#endif

            cd = wss_db_mcol_def( 2, PD_MAKESTR("Parameter"), pd_char_t_t, 64,
                                  PD_MAKESTR("Value"), pd_char_t_t, 1024 );
            wss_create( PD_MAKESTR(""), 0, PD_MAKESTR("CONFIG"), FALSE, cd, 2,
                        &wss_config_db_handle );

            /* Set the table non-persistent */
            tab = wss_db_find_table_by_handle( wss_config_db_handle );
            tab->flags |= DB_FLAG_NONPERSISTENT;
            wss_db_unlock_table( tab );

            wss_init_fill_db( wss_config_db_handle );

#ifdef WR_CONFIG_IMMUTABLE
            /* Now that it's filled, set it immutable */
            tab = wss_db_find_table_by_handle( wss_config_db_handle );
            tab->flags |= DB_FLAG_IMMUTABLE;
            wss_db_unlock_table( tab );
#endif /* WR_CONFIG_IMMUTABLE */
         
#ifdef WR_UNIX
#if 0
            first_row = next_row = wss_malloc(sizeof(*next_row));

            next_row->vals = wss_malloc(sizeof(wqval)*2);
            next_row->vals[0].v.cval =
                   wss_strdup(PD_MAKESTR("audit-kernel-record"));
            next_row->vals[1].v.cval = wss_strdup(PD_MAKESTR("true"));
            next_row->next = wss_malloc(sizeof(*next_row));
            next_row = next_row->next;

            next_row->vals = wss_malloc(sizeof(wqval)*2);
            next_row->vals[0].v.cval =wss_strdup(PD_MAKESTR("audit-kernel-path"));
            next_row->vals[1].v.cval =wss_strdup(PD_MAKESTR("/var/idw"));
            next_row->next = NULL;
         
            wss_insert( wss_config_db_handle, TRUE, first_row);
#endif /* 0 */
#endif /* WR_UNIX */
         }
      }
      else
      {
         wsi_printf( PD_MAKESTR("Shutting down...\n") );
         /* Error occurred, shut everything down */
         wsi_printf( PD_MAKESTR("* vm\n") );
         wss_vm_shutdown();
         wsi_printf( PD_MAKESTR("* evspace\n") );
         wss_evspace_shutdown();
         wsi_printf( PD_MAKESTR("* audit\n") );
         wss_audit_shutdown();
         wsi_printf( PD_MAKESTR("* wrvevent\n") );
         wss_wrvevent_shutdown();
         wsi_printf( PD_MAKESTR("* arg\n") );
         wss_arg_shutdown();
         wsi_printf( PD_MAKESTR("* crit\n") );
         wss_crit_shutdown();
         wsi_printf( PD_MAKESTR("* lib\n") );
         wss_lib_shutdown();
         wsi_printf( PD_MAKESTR("* wrapper\n") );
         wss_wrapper_shutdown();
         wsi_printf( PD_MAKESTR("* mod\n") );
         wss_mod_shutdown();
         wsi_printf( PD_MAKESTR("* linker\n") );
         wss_linker_shutdown();
         wsi_printf( PD_MAKESTR("* log\n") );
         wss_log_shutdown();
         wsi_printf( PD_MAKESTR("* proc\n") );
         wss_proc_shutdown();
         wsi_printf( PD_MAKESTR("* malloc\n") );
         wss_malloc_shutdown();
      }
   }

   return( ( init_flags & WSS_INIT_DB ) ? db_init_ret : 0 );
}

static wqrow *
_wsi_init_makerow( wss_init_entry_t *init_entry )
{
   pd_char_t *val;
   wqrow      *next_row;

   switch( init_entry->type )
   {
      case WSS_INIT_TYPE_STRING:
      {
         pd_char_t *str = *(pd_char_t **)init_entry->storage;

         if( str )
            val = wss_strdup( str );
         else
            val = wss_strdup( PD_MAKESTR("") );

         break;
      }

      case WSS_INIT_TYPE_INT:
      {
         int ival = *(int *)init_entry->storage;
         val = wss_malloc( 16 );
         if( val )
            wsi_sprintf( val, PD_MAKESTR("%d"), ival );

         break;
      }

      case WSS_INIT_TYPE_BOOL:
      {
         int ival = *(int *)init_entry->storage;

         if( ival )
            val = wss_strdup( PD_MAKESTR("true") );
         else
            val = wss_strdup( PD_MAKESTR("false") );

         break;
      }

      default:
         return NULL;
   }

   next_row = wss_malloc( sizeof(*next_row) );
   if( next_row )
   {
      next_row->vals = wss_malloc( sizeof(wqval) * 2 );
      if( next_row->vals )
      {
         next_row->vals[0].v.cval = wss_downcase( init_entry->key );
         next_row->vals[1].v.cval = val;
         next_row->next = NULL;
      }
      else
      {
         wss_free( next_row );
         wss_free( val );
         return NULL;
      }
   }
   else
      wss_free( val );

   return next_row;
}

/*  wss_init_fill_db */
/*! \brief Fill the CONFIG table in the database with initialization options
 */
int
wss_init_fill_db( u_long dbh )
{
   wss_init_entry_t *init_entry;
   wqrow            *first_row = NULL;
   wqrow            *last_row  = NULL;

   for( init_entry = wss_init_globals;
        ( init_entry->key != NULL );
        init_entry++ )
   {
      wqrow *next_row = _wsi_init_makerow( init_entry );

      if( last_row )
         last_row->next = next_row;
      else
         first_row = next_row;

      last_row = next_row;
   }

   for( init_entry = wsi_init_locals;
        ( init_entry->key != NULL );
        init_entry++ )
   {
      wqrow *next_row = _wsi_init_makerow( init_entry );

      if( last_row )
         last_row->next = next_row;
      else
         first_row = next_row;

      last_row = next_row;
   }

   if( first_row )
      return wss_insert( dbh, TRUE, first_row );

   return 0;
}

