/*! \file wss_library.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_library.c,v 1.24 2001/06/13 13:27:59 stevek Exp $
 *
 * \brief   Support for managing wrapper support libraries
 * \author  Steve Kiernan
 * \version $Revision: 1.24 $
 * \date    $Date: 2001/06/13 13:27:59 $
 */


#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_library.h>
#include <wss_module.h>
#include <wss_mutex.h>

#include <eventfs.h>

#include <wrerrno.h>


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

/*! List of loaded wrapper libraries */
wss_lib_t EXPORT  *wss_library_list = NULL;

/*! Mutex protecting list of libraries */
WSS_MUTEX_PROTECT( wss_library_list );


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_lib_init */
/*! \brief Initialize the library support code
 *
 *  Initializes the mutex protecting the wrapper library list.
 */
void EXPORT
wss_lib_init()
{
   WSS_MUTEX_INIT( wss_library_list );
}

/*  wss_lib_shutdown */
/*! \brief Shutdown the library support code
 *
 *  Cleans up the list of library structures and destroys the mutex
 *  protecting the list.
 */
void EXPORT
wss_lib_shutdown()
{
   wss_lib_t *lib;
   wss_lib_t *next;

   WSS_MUTEX_ENTER( wss_library_list );
   lib = wss_library_list;
   wss_library_list = NULL;
   WSS_MUTEX_EXIT( wss_library_list );

   while( lib )
   {
      next = lib->lib_next;
      wsi_printf( PD_MAKESTR("wss_lib_shutdown: Cleaning up '%s'\n"),
                  lib->lib_module->wrm_linked->linker_file_name );
      wss_free( lib );
      lib = next;
   }

   WSS_MUTEX_DESTROY( wss_library_list );
}


/*  wss_lib_find */
/*! \brief Find the library associated with the specified id
 *
 *  \param id Identifier of library module
 *
 *  \warning Library structure is returned as referenced and must
 *           be released when no longer needed
 *
 *  \return Pointer to library structure if found,
 *  \return NULL if not found
 */
wss_lib_t EXPORT *
wss_lib_find( int id )
{
   wss_lib_t *lib;

   WSS_MUTEX_ENTER( wss_library_list );
   for( lib = wss_library_list;
        lib && ( lib->lib_module->wrm_id != id );
        lib = lib->lib_next )
   ;
   if( lib )
      wss_lib_ref( lib );
   WSS_MUTEX_EXIT( wss_library_list );

   return lib;
}


/*  wss_lib_find_by_name */
/*! \brief Find the library associated with the specified name
 *
 *  \param name Name of library module
 *
 *  \return Pointer to library structure if found,
 *  \return NULL if not found
 */
wss_lib_t EXPORT *
wss_lib_find_by_name( const pd_char_t *name )
{
   wss_lib_t *lib;

   WSS_MUTEX_ENTER( wss_library_list );
   for( lib = wss_library_list;
        lib && ( wsi_strcmp( name,
                             lib->lib_module->wrm_linked->linker_file_name ) );
        lib = lib->lib_next )
   ;
   if( lib )
      wss_lib_ref( lib );
   WSS_MUTEX_EXIT( wss_library_list );

   return lib;
}


/*  wss_lib_register */
/*! \brief Register a new support library with the WSS
 *
 *  \param wrm    Pointer to the wrapper module for this library
 *  \param symtab Pointer to the library's symbol table
 *  \param fini   Pointer to the function to be called upon library unload
 *
 *  \return 0 on success,
 *  \return \b EINVAL if the module is not a library,
 *  \return \b ENOMEM if unable to allocate necessary memory
 *  \internal
 */
int EXPORT
wss_lib_register( wss_mod_t *wrm, wss_symbol_t *symtab, void *fini )
{
   wss_lib_t *lib;

   if( ( wrm == NULL ) || ( wrm->wrm_type != WR_MOD_LIBRARY ) )
      return EINVAL;

   if( ( lib = (wss_lib_t *)wss_malloc( sizeof( *lib ) ) ) == NULL )
   {
      wsi_printf( PD_MAKESTR("wss_lib_register: Can't malloc memory for library\n") );
      return ENOMEM;
   }

   bzero( lib, sizeof( *lib ) );
   wss_ref_init( &(lib->lib_refcount) );
   wss_mod_ref( wrm );
   lib->lib_module = wrm;
   lib->lib_symtab = symtab;
   lib->lib_fini = fini;

   /* Prepend to the library list */
   WSS_MUTEX_ENTER( wss_library_list );
   lib->lib_next = wss_library_list;
   wss_library_list = lib;
   WSS_MUTEX_EXIT( wss_library_list );

   return 0;
}


/*  wss_lib_can_unload */
/*! \brief Determines if library module is in proper state to be unloaded
 *
 *  \param id Identifier of library module
 *
 *  \return TRUE if wrapper can be unloaded (or not found),
 *  \return FALSE otherwise
 */
int EXPORT
wss_lib_can_unload( wss_id_t id )
{
   wss_lib_t *lib;

   WSS_MUTEX_ENTER( wss_library_list );
   for( lib = wss_library_list;
        lib && ( lib->lib_module->wrm_id != id );
        lib = lib->lib_next )
   ;

   if( lib == NULL )
   {
      WSS_MUTEX_EXIT( wss_library_list );
      return TRUE;
   }

   if( WSS_REF_COUNT(lib->lib_refcount) > 1 )
   {
#ifdef DEBUG_LIBRARY
      wsi_printf( PD_MAKESTR("wss_lib_free: refcount %d > 1\n"),
                  WSS_REF_COUNT(lib->lib_refcount) );
#endif /* DEBUG_LIBRARY */
      WSS_MUTEX_EXIT( wss_library_list );
      return FALSE;
   }

   WSS_MUTEX_EXIT( wss_library_list );
   return TRUE;
}


/*  wss_lib_delete */
/*! \brief Remove library from list for unload
 *
 *  \param id Identifier of library module
 *
 *  \return 0 on successful deletion,
 *  \return \b WRE_REF if the library is still being referenced,
 *  \return \b WRE_NOTFOUND if library can not be found
 */
int EXPORT
wss_lib_delete( wss_id_t id )
{
   wss_lib_t *lib;
   wss_lib_t *liblist, **last;

   WSS_MUTEX_ENTER( wss_library_list );
   for( lib = wss_library_list;
        lib && ( lib->lib_module->wrm_id != id );
        lib = lib->lib_next )
   ;
   if( lib == NULL )
   {
      WSS_MUTEX_EXIT( wss_library_list );
      return WRE_NOTFOUND;
   }

   if( WSS_REF_COUNT(lib->lib_refcount) > 1 )
   {
      WSS_MUTEX_EXIT( wss_library_list );
      return WRE_REF;
   }

   last = &wss_library_list;
   for( liblist = wss_library_list;
        liblist && ( liblist != lib );
        last = &liblist->lib_next, liblist = liblist->lib_next );
   *last = lib->lib_next;
   WSS_MUTEX_EXIT( wss_library_list );

   /* Release last reference to library */
   wss_lib_rele( lib );

   return 0;
}

/*  wss_lib_free */
/*! \brief Free memory used by library structure
 *
 *  \param lib Pointer to a library structure
 */
void EXPORT
wss_lib_free( wss_lib_t *lib )
{
   assert( lib != NULL );
   assert( WSS_REF_COUNT(lib->lib_refcount) == 1 );

   WSS_REF_DEC(lib->lib_refcount);

#ifdef DEBUG_LIBRARY
   wsi_printf( PD_MAKESTR("wss_lib_free: releasing memory for 0x%x\n"),
               (unsigned int)lib );
#endif /* DEBUG_LIBRARY */

   /* Call the unloaded function */
   if( lib->lib_fini )
      (*lib->lib_fini)();

   wss_ref_destroy( &(lib->lib_refcount) );

   WSS_MUTEX_DESTROY( lib->lib );

   wss_free( lib );
}

