/*! \file wss_linker.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_linker.c,v 1.15 2001/06/18 15:31:35 stevek Exp $
 *
 * \brief   Support for linking modules into the kernel
 * \author  Steve Kiernan
 * \version $Revision: 1.15 $
 * \date    $Date: 2001/06/18 15:31:35 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_linker.h>
#include <wss_export.h>
#include <wss_malloc.h>
#include <wss_mutex.h>

#include <eventfs.h>


/***************************************************************************\
 * External globals                                                        *
\***************************************************************************/

extern wss_symbol_t wss_export_table[];
extern int          wsi_unloading;


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

wss_linker_t EXPORT         *wss_linker_list = NULL;
WSS_MUTEX_PROTECT_EXPORTED( wss_linker_list );

static wss_linker_file_ops_t _wss_linker_global_file_ops =
{
   wss_export_lookup,
   wss_export_noop,
   wss_export_noop,
   wss_export_noop
};

wss_linker_file_t EXPORT wss_linker_file_list =
{
   { { 1 } },
   PD_MAKESTR("*WRAPPER SUPPORT SUBSYSTEM*"),
   PD_MAKESTR(""),
   { 0 },
   { 0 },
   0,
   &_wss_linker_global_file_ops,
   NULL,
   wss_export_table,
   PD_MAKESTR("\0"),
   NULL
};
WSS_MUTEX_PROTECT_EXPORTED( wss_linker_file_list );


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_linker_init */
/*! \brief Initialize module linker code
 *
 *  Constructs the mutex protecting the linker list.
 */
void EXPORT
wss_linker_init( void )
{
   WSS_MUTEX_INIT( wss_linker_list );
   WSS_MUTEX_INIT( wss_linker_file_list );

   WSS_MUTEX_INIT( wss_linker_file_list.linker_file );
}


/*  wss_linker_shutdown */
/*! \brief Shutdown module linker code
 *
 *  Destroys the mutex protecting the linker list.
 */
void EXPORT
wss_linker_shutdown( void )
{
   WSS_MUTEX_DESTROY( wss_linker_list );
   WSS_MUTEX_DESTROY( wss_linker_file_list );

   WSS_MUTEX_DESTROY( wss_linker_file_list.linker_file );
}


/*  wss_linker_register */
/*! \brief Register a module linker with the WSS
 *
 *  \param description Description of module linker
 *  \param ops         Pointer to linker ops structure
 *
 *  \return 0 if successfully loaded,
 *  \return EEXIST if a linker of the same name is already registered
 */                                                               
int EXPORT
wss_linker_register( const pd_char_t *description, wss_linker_ops_t *ops )
{
   wss_linker_t *linker;
   int           error  = 0;

   WSS_MUTEX_ENTER( wss_linker_list );
   for( linker = wss_linker_list; linker; linker = linker->linker_next )
      if( !wsi_strcmp( linker->linker_description, description ) )
      {
         error = EEXIST;
         break;
      }

   if( !error )
   {
      linker = (wss_linker_t *)wss_malloc( sizeof(*linker) );
      linker->linker_description = description;
      linker->linker_ops = ops;
      linker->linker_next = wss_linker_list;
      wss_linker_list = linker;
   }
   WSS_MUTEX_EXIT( wss_linker_list );

   return error;
}


/*  wss_linker_file_init */
/*! \brief Initialize a linker file structure
 *
 *  \param lf Pointer to linker file structure
 */                                                               
void EXPORT
wss_linker_file_init( wss_linker_file_t *lf )
{
   if( lf == NULL )
      return;

   bzero( lf, sizeof(*lf) );
   WSS_MUTEX_INIT( lf->linker_file );
   lf->linker_file_name = PD_MAKESTR("");
   lf->linker_file_path = PD_MAKESTR("");
   lf->linker_file_dependencies = PD_MAKESTR("\0");
   wss_ref_init( &(lf->linker_file_refcount) );
}


/*  wss_linker_file_destroy */
/*! \brief Destroy a linker file structure
 *
 *  \param lf Pointer to linker file structure
 *  \return 0 if successfully destroyed,
 *  \return EINVAL if an invalid argument is passed
 */
int EXPORT
wss_linker_file_destroy( wss_linker_file_t *lf )
{
   wss_linker_file_depend_t *ent, *next;

   if( lf == NULL )
      return EINVAL;

   assert( WSS_REF_COUNT(lf->linker_file_refcount) == 1 );

   WSS_REF_DEC(lf->linker_file_refcount);

   for( ent = lf->linker_file_dependency_list; ent; ent = next )
   {
      next = ent->linker_file_depend_next;
#ifdef DEBUG_LINKER
      wsi_printf(PD_MAKESTR("linker_file: releasing dependency '%s'\n"),
                 ent->linker_file_depend_entry->linker_file_name);
#endif
      wss_linker_file_rele( ent->linker_file_depend_entry );
      wss_free( (void *)ent );
   }

   if( lf->linker_file_symbols )
      wss_free( lf->linker_file_symbols );

   if( lf->linker_file_addr.addr )
      wss_free( lf->linker_file_addr.addr );

   if( lf->linker_file_image.addr )
      wss_free( lf->linker_file_image.addr );

   if( lf->linker_file_data )
      wss_free( lf->linker_file_data );

   if( lf->linker_file_path && *lf->linker_file_path )
      wss_free( (void *)lf->linker_file_path );

   wss_ref_destroy( &(lf->linker_file_refcount) );

   WSS_MUTEX_DESTROY( lf->linker_file );

   return 0;
}


/*  wss_linker_add_symbols */
/*! \brief Register a symbol set with the WSS
 *
 *  \param symsetname Description of the symbol set
 *  \param symtab     Pointer to symbol table
 *
 *  \warning WSS extensions should register symbol sets with an asterisk
 *           at the beginning of the name (Example: *MY SYMBOL SET*).
 *
 *  \return 0 if successfully loaded,
 *  \return EINVAL if an invalid argument is passed,
 *  \return EEXIST if a symbol set of the same name is already registered
 */
int EXPORT
wss_linker_add_symbols( const pd_char_t *symsetname, wss_symbol_t *symtab )
{
   wss_linker_file_t *lf;

   if( !symsetname || ( *symsetname != PD_MAKESTR('*') ) )
      return EINVAL;

   WSS_MUTEX_ENTER( wss_linker_file_list );
   for( lf = wss_linker_file_list.linker_file_next;
        lf;
        lf = lf->linker_file_next )
      if( !wsi_strcmp( lf->linker_file_name, symsetname ) )
      {
         WSS_MUTEX_EXIT( wss_linker_file_list );
         wsi_printf( PD_MAKESTR("Linker: Symbol set '%s' already exists.\n"),
                     symsetname );
         return EEXIST;
      }

   lf = (wss_linker_file_t *)wss_malloc( sizeof(*lf) );
   bzero( lf, sizeof(*lf) );

   wss_linker_file_init( lf );

   lf->linker_file_name = symsetname;
   lf->linker_file_symbols = symtab;
   lf->linker_file_ops = &_wss_linker_global_file_ops;

   lf->linker_file_next = wss_linker_file_list.linker_file_next;
   wss_linker_file_list.linker_file_next = lf;

   WSS_MUTEX_EXIT( wss_linker_file_list );

   return 0;
}


/*  wss_linker_load */
/*! \brief Load a wrapper module file
 *
 *  \param filename Pointer to string containing the path of the module
 *  \param result   Pointer to location to store resulting linker file
 *                  structure
 *
 *  \return 0 if successfully loaded,
 *  \return EINVAL if an invalid argument is passed,
 *  \return ENOENT if the path is not found,
 *  \return ENOMEM if there is not enough memory to load the module,
 *  \return ENOEXEC if there are no linkers which can handle the file format
 */
int EXPORT
wss_linker_load( const pd_char_t *filename, wss_linker_file_t *result )
{
   wss_linker_t *linker;
   int           error  = ENOEXEC;

   if( result == NULL )
      return EINVAL;

   result->linker_file_path = filename;

   WSS_MUTEX_ENTER( wss_linker_list );
   for( linker = wss_linker_list; linker; linker = linker->linker_next )
   {
      error = (linker->linker_ops->load_file)( filename, result );
      if( error == ENOENT )
         break;

      if( ( error == 0 ) && ( result != NULL ) )
      {
         wss_linker_addr_t addr;

         error =
            (result->linker_file_ops->lookup_symbol)(
               result, PD_MAKESTR(WRLINKNAME_MOD), &addr );

         if( error == 0 )
         {
            result->linker_file_name = *(const pd_char_t **)addr.addr;

            if( result->linker_file_name &&
                ( *(result->linker_file_name) == PD_MAKESTR('*') ) )
            {
               wsi_printf( PD_MAKESTR("Linker: Invalid module name\n") );
               error = EINVAL;
               break;
            }

            /* Add the linker file to the list */
            WSS_MUTEX_ENTER( wss_linker_file_list );
            result->linker_file_next = wss_linker_file_list.linker_file_next;
            wss_linker_file_list.linker_file_next = result;
            WSS_MUTEX_EXIT( wss_linker_file_list );
         }
         else
         {
            wsi_printf( PD_MAKESTR("Linker: Unable to determine module ")
                        PD_MAKESTR("name\n") );
         }

         break;
      }
   }
   WSS_MUTEX_EXIT( wss_linker_list );

   return error;
}


/*  wss_linker_find_file_unsafe */
/*! \brief Search for a specific linker file by name
 *
 *  \param name Pointer to string containing the name of the linker file to
 *              find
 *
 *  \return Pointer to linker file structure if found,
 *  \return NULL if not found
 */
wss_linker_file_t EXPORT *
wss_linker_find_file_unsafe( const pd_char_t *name )
{
   wss_linker_file_t *lf;

   for( lf = &wss_linker_file_list; lf; lf = lf->linker_file_next )
   {
      if( !wsi_strcmp( name, lf->linker_file_name ) )
      {
         wss_linker_file_ref( lf );
         return lf;
      }
   }
   return NULL;
}


/*  wss_linker_remove_file */
/*! \brief Remove (and unload) a linker file
 *
 *  \param lf Pointer to a linker file structure
 *
 *  \return 0 if successfully loaded,
 *  \return EINVAL if an invalid argument is passed,
 *  \return ENOENT if the linker file is not found
 */
int EXPORT
wss_linker_remove_file( wss_linker_file_t *lf )
{
   wss_linker_file_t  *linkfile;
   wss_linker_file_t **lastlink;

   if( lf == NULL )
      return EINVAL;

   WSS_MUTEX_ENTER( wss_linker_file_list );

   for( linkfile = wss_linker_file_list.linker_file_next,
           lastlink = &wss_linker_file_list.linker_file_next;
        linkfile;
        lastlink = &linkfile->linker_file_next,
           linkfile = linkfile->linker_file_next )
      if( ( linkfile == lf ) &&
          ( wsi_unloading || ( linkfile->linker_file_name == NULL ) ||
            ( *(linkfile->linker_file_name) != PD_MAKESTR('*') ) ) )
      {
         *lastlink = linkfile->linker_file_next;
         WSS_MUTEX_EXIT( wss_linker_file_list );

         wss_linker_file_rele( lf );
         return 0;
      }

   WSS_MUTEX_EXIT( wss_linker_file_list );

   return ENOENT;
}


/*  wss_linker_search */
/*! \brief Search the symbol table of a linker file and its dependencies
 *
 *  \param lf       Pointer to a linker file structure
 *  \param synmname Pointer to string containing name of symbol to find
 *  \param result   Pointer to location to store resulting address
 *
 *  \return 0 if successfully loaded,
 *  \return EINVAL if an invalid argument is passed,
 *  \return ENOENT if the symbol is not found
 */
int EXPORT
wss_linker_search( wss_linker_file_t *lf, const pd_char_t *symname,
                   wss_linker_addr_t *result )
{
   wss_linker_file_depend_t *depend;
   wss_linker_file_t        *deplf;

   if( ( lf == NULL ) || ( symname == NULL ) || ( result == NULL ) )
      return EINVAL;

   /* Search the WSS system component symbol tables first */
   WSS_MUTEX_ENTER( wss_linker_file_list );
   for( deplf = &wss_linker_file_list; deplf; deplf = deplf->linker_file_next )
      if( deplf->linker_file_name &&
          ( *deplf->linker_file_name == PD_MAKESTR('*') ) &&
          !(deplf->linker_file_ops->lookup_symbol)( deplf, symname, result ) )
      {
         WSS_MUTEX_EXIT( wss_linker_file_list );
         return 0;
      }
   WSS_MUTEX_EXIT( wss_linker_file_list );

   /* Now check the symbol tables for each dependency library */
   for( depend = lf->linker_file_dependency_list;
        depend;
        depend = depend->linker_file_depend_next )
   {
      wss_linker_file_t *deplf = depend->linker_file_depend_entry;
      if( !(deplf->linker_file_ops->lookup_symbol)( deplf, symname, result ) )
         return 0;
   }

   return ENOENT;
}


/*  wss_linker_construct_dependencies */
/*! \brief Construct the list of dependencies for symbol searching
 *
 *  \param lf       Pointer to a linker file structure
 *
 *  \return 0 if successfully loaded,
 *  \return EINVAL if an invalid argument is passed,
 *  \return ENOENT if a dependency file is not loaded
 */
int EXPORT
wss_linker_construct_dependencies( wss_linker_file_t *lf )
{
   const pd_char_t *depend;

   if( ( lf == NULL ) || ( lf->linker_file_dependency_list != NULL ) )
      return EINVAL;

   depend = lf->linker_file_dependencies;
   while( depend && *depend )
   {
      wss_linker_file_depend_t *ent;
      wss_linker_file_t        *deplf;

      WSS_MUTEX_ENTER( wss_linker_file_list );
      deplf = wss_linker_find_file_unsafe( depend );
      if( ( deplf == NULL ) || ( deplf->linker_file_symbols == NULL ) )
      {
         wss_linker_file_depend_t *next;

         if( deplf )
            wss_linker_file_rele( deplf );
         WSS_MUTEX_EXIT( wss_linker_file_list );
         wsi_printf( PD_MAKESTR("Linker: Dependency '%s' not found\n"),
                     depend );

         for( ent = lf->linker_file_dependency_list; ent; ent = next )
         {
            next = ent->linker_file_depend_next;
            wss_linker_file_rele( ent->linker_file_depend_entry );
            wss_free( (void *)ent );
         }

         return ENOENT;
      }

      ent = (wss_linker_file_depend_t *)wss_malloc( sizeof(*ent) );
      ent->linker_file_depend_entry = deplf;
      ent->linker_file_depend_next = lf->linker_file_dependency_list;
      lf->linker_file_dependency_list = ent;
      WSS_MUTEX_EXIT( wss_linker_file_list );

      while( *depend++ )
      ;
   }

   return 0;
}


/*  wss_linker_search_multiple */
/*! \brief Search the symbol table of a linker file and its dependencies
 *         for multiple entries
 *
 *  \param lf Pointer to a linker file structure
 *
 *  \return 0 if successfully loaded,
 *  \return EINVAL if an invalid argument is passed,
 *  \return ENOENT if the symbol is not found
 */
int EXPORT
wss_linker_search_multiple( wss_linker_file_t *lf, int warn_unresolved,
                            ... )
{
   va_list       ap;
   pd_char_t   *name;
   int           nunresolved = 0;
   int           error = 0;

   va_start( ap, warn_unresolved );
   while( ( name = va_arg( ap, pd_char_t * ) ) != NULL )
   {
      wss_linker_addr_t *addr = va_arg( ap, wss_linker_addr_t * );

      if( addr == NULL )
      {
         error = EINVAL;
         break;
      }

      if( wss_linker_search( lf, name, addr ) )
      {
         if( warn_unresolved )
            wsi_printf( PD_MAKESTR("Linker: Unresolved symbol '%s'\n"),
                        name );
         nunresolved++;
         error = ENOENT;
      }
   }
   va_end( ap );

   if( warn_unresolved && nunresolved )
      wsi_printf( PD_MAKESTR("Linker: [%d unresolved symbol(s)] Unable to ")
                  PD_MAKESTR("initialize module.\n"), nunresolved );

   return error;
}


/*  wss_linker_lookup_multiple */
/*! \brief Search the symbol table of a linker file for multiple entries
 *
 *  \param lf Pointer to a linker file structure
 *
 *  \return 0 if successfully loaded,
 *  \return EINVAL if an invalid argument is passed,
 *  \return ENOENT if the symbol is not found
 */
int EXPORT
wss_linker_lookup_multiple( wss_linker_file_t *lf, int warn_unresolved,
                            ... )
{
   va_list       ap;
   pd_char_t   *name;
   int           nunresolved = 0;
   int           error = 0;

   va_start( ap, warn_unresolved );
   while( ( name = va_arg( ap, pd_char_t * ) ) != NULL )
   {
      wss_linker_addr_t *addr = va_arg( ap, wss_linker_addr_t * );

      if( addr == NULL )
      {
         error = EINVAL;
         break;
      }

      if( (lf->linker_file_ops->lookup_symbol)( lf, name, addr ) )
      {
         if( warn_unresolved )
            wsi_printf( PD_MAKESTR("Linker: Unresolved symbol '%s'\n"),
                        name );
         nunresolved++;
         error = ENOENT;
      }
   }
   va_end( ap );

   if( warn_unresolved && nunresolved )
      wsi_printf( PD_MAKESTR("Linker: [%d unresolved symbol(s)] Unable to ")
                  PD_MAKESTR("initialize module.\n"), nunresolved );

   return error;
}

/*  wss_linker_file_has_dependency */
/*! \brief Determine if a linker file depends on another
 *
 *  \param lf    Pointer to a linker file structure
 *  \param deplf Pointer to dependency linker file structure
 *
 *  \return TRUE if \a lf depends on \a deplf,
 *  \return FALSE otherwise
 */
int EXPORT
wss_linker_file_has_dependency( wss_linker_file_t *lf,
                                wss_linker_file_t *deplf )
{
   wss_linker_file_depend_t  *depend;

   if( ( lf == NULL ) || ( deplf == NULL ) )
      return EINVAL;

   WSS_MUTEX_ENTER( lf->linker_file );
   for( depend = lf->linker_file_dependency_list;
        depend;
        depend = depend->linker_file_depend_next )
   {
      if( depend->linker_file_depend_entry == deplf )
      {
         WSS_MUTEX_EXIT( lf->linker_file );
         return TRUE;
      }
   }
   WSS_MUTEX_EXIT( lf->linker_file );

   return FALSE;
}

