/*! \file wss_log.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_log.c,v 1.14 2001/02/23 19:02:46 stevek Exp $
 *
 * \brief   Message logging functions
 * \author  Linda Thomas, Steve Kiernan
 * \version $Revision: 1.14 $
 * \date    $Date: 2001/02/23 19:02:46 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_file.h>
#include <wss_log.h>
#include <wss_malloc.h>
#include <wss_mutex.h>

#include <wsi_time.h>


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static volatile wss_file_t wrapperlog = WSS_FILE_INVALID;
static WSS_MUTEX_PROTECT ( wrapperlog );


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/*! Filename of log file */
volatile pd_char_t *wss_log_path = NULL;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_log_init */
/*! \brief Initialize logging
 *
 *  Construct mutex protecting the log file handle.
 *
 *  \return 0 (other values reserved for future expansion)
 */
int EXPORT
wss_log_init()
{
   WSS_MUTEX_INIT( wrapperlog );
   return 0;
}

/*  wss_log_shutdown */
/*! \brief Shutdown logging
 *
 *  Closes the log file and destroys the mutex protecting the
 *  log file handle.
 */
void EXPORT
wss_log_shutdown()
{
   wss_log_close();
   WSS_MUTEX_DESTROY( wrapperlog );
}


/*  wss_log_open */
/*! \brief Open the log file
 *
 *  \b TBD: Use an entry the WSS '.GLOBAL' database to store the location of
 *          the log file (currently we use an init file variable)
 *
 *  \return 0 if log file successfully opened,
 *  \return error code otherwise
 */
int EXPORT
wss_log_open()
{
   int err = 0;

   WSS_MUTEX_ENTER( wrapperlog );
   if( wrapperlog == WSS_FILE_INVALID )
   {
      wrapperlog = wss_file_open( (pd_char_t *)wss_log_path,
                                  WSS_FILE_CREAT | WSS_FILE_READ |
                                  WSS_FILE_WRITE | WSS_FILE_ROOTPRIV );
      if( wrapperlog != WSS_FILE_INVALID )
         err = wss_file_seek( wrapperlog, 0, WSS_FILE_SEEK_END );
   }
   WSS_MUTEX_EXIT( wrapperlog );

   return err;
}

/*  wss_log_close */
/*! \brief Close the log file
 *
 *  \return 0 (other values reserved for future expansion)
 */
int EXPORT
wss_log_close()
{
   WSS_MUTEX_ENTER( wrapperlog );
   if( wrapperlog != WSS_FILE_INVALID )
   {
      wss_file_close( wrapperlog );
      wrapperlog = WSS_FILE_INVALID;
   }
   WSS_MUTEX_EXIT( wrapperlog );

   return 0;
}

/*  wss_log_printf */
/*! \brief Log message to log file
 *
 *  This function uses a printf-style interface to add messages to the log.
 *
 *  \param format printf-style format string
 *
 *  \return 0 on success,
 *  \return error code on failure
 */
int EXPORT
wss_log_printf( const pd_char_t *format, ... )
{
   va_list ap;
   int     len;
   int     err = 0;

   if( !( err = wss_log_open() ) )
   {
      if( wrapperlog )
      {
         WSS_MUTEX_ENTER( wrapperlog );
         if( wrapperlog )
         {
            pd_char_t  buffer[2048];
            pd_char_t *timestr;

            timestr = wsi_time_format_alloc( wsi_time_get_current(NULL) );
            wsi_sprintf( buffer, PD_MAKESTR("%s:  "), timestr );
            wss_free( timestr );
            len = wsi_strlen( buffer ) * sizeof(pd_char_t);
            err = wsi_file_write( wrapperlog, (pd_byte_t *)buffer, len );

            va_start( ap, format );
            wsi_vsprintf( buffer, format, ap );
            va_end( ap );

            len = wsi_strlen( buffer ) * sizeof(pd_char_t);
            err = wss_file_write( wrapperlog, (pd_byte_t *)buffer, len );
            if( err < 0 )
               err = -err;
            else
               err = 0;
         }
         WSS_MUTEX_EXIT( wrapperlog );
      }
   }

   return err;
}

