/*! \file wss_malloc.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_malloc.c,v 1.23 2001/02/23 19:02:46 stevek Exp $
 * 
 * \brief   Memory management wrapper functions
 * \author  Doug Kilpatrick
 * \version $Revision: 1.23 $
 * \date    $Date: 2001/02/23 19:02:46 $
 */


#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_mutex.h>
#include <wss_utilities.h>


/***************************************************************************\
 * Private structures and globals                                          *
\***************************************************************************/

typedef struct malloc_record
{
    wss_util_rbnode_t  parent;
    pd_char_t         *file;
    int                line;
    void              *ptr;
} malloc_record_t;

wss_util_rbtree_t *wss_malloc_record_tree = NULL;
WSS_MUTEX_PROTECT( wss_malloc_record_tree );

#define init_mr(mr, f, l, b) \
    wss_util_rbnode_init(&((mr)->parent)); \
    (mr)->file = (f); \
    (mr)->line = (l); \
    (mr)->ptr  = (b)

/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

#ifdef MALLOC_DEBUG
/* RBTree stuff */
static int
malloc_compfunc(void *n1, void *n2)
{
    malloc_record_t *node1 = n1;
    malloc_record_t *node2 = n2;
    if (node1->ptr < node2->ptr )
        return -1;
    if (node1->ptr > node2->ptr )
        return 1;
    return 0;
}

static void *
malloc_clonefunc(void *n)
{
    malloc_record_t *retval = wsi_malloc(sizeof(*retval));
    if (!n) {
        init_mr(retval, 0, 0, 0);
    } else {
        malloc_record_t *orig = n;
        init_mr(retval, orig->file, orig->line, orig->ptr);
    }
    return retval;
}

static wss_util_rbtree_t *
new_tree(void)
{
    /* Do this manually since the one in wss_utilities uses wss_malloc
     * Using that would cause some recursion problems.
     */
    wss_util_rbtree_t *retval = wsi_malloc(sizeof(*retval));
    retval->Head = malloc_clonefunc(0);
    retval->Compare = malloc_compfunc;
    retval->Clone = malloc_clonefunc;
    retval->Free = wsi_free;
    
    return retval;
}


/*
 * Dump the allocation chain
 */
static int
malloc_dump_dofunc(void *n, int level)
{
    malloc_record_t *node = n;
    wsi_printf( PD_MAKESTR("Ptr: %p  allocated from %s:%d\n"), node->ptr,
            node->file, node->line );
    return 0;
}

void
wss_malloc_dump( void )
{
   WSS_MUTEX_ENTER( wss_malloc_record_tree );
   wss_util_rbtree_walk( wss_malloc_record_tree, malloc_dump_dofunc );
   WSS_MUTEX_EXIT( wss_malloc_record_tree );
}

#endif /* defined(MALLOC_DEBUG) */

/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_malloc_init */
/*! \brief Initialize the memory management code
 *
 *  If \b MALLOC_DEBUG is defined, intializes the mutex protecting the
 *  malloc record tree.
 */
void EXPORT
wss_malloc_init()
{
#ifdef MALLOC_DEBUG
   WSS_MUTEX_INIT( wss_malloc_record_tree );
#endif /* MALLOC_DEBUG */
}

/*  wss_malloc_shutdown */
/*! \brief Shutdown the memory management code
 *
 *  If \b MALLOC_DEBUG is defined, destroys the mutex protecting the
 *  malloc record tree.
 */
void EXPORT
wss_malloc_shutdown()
{
#ifdef MALLOC_DEBUG
   WSS_MUTEX_DESTROY( wss_malloc_record_tree );
#endif /* MALLOC_DEBUG */
}

#ifdef MALLOC_DEBUG


/*  wss_malloc_debug */
/*! \brief Allocate memory with debug tracking
 *
 *  \param size Size of memory block to allocate
 *  \param file Filename where calling function is contained
 *  \param line Line number of calling function
 *
 *  \return Allocated block of memory on success,
 *  \return NULL on failure
 */
void EXPORT *
wss_malloc_debug( size_t size, pd_char_t *file, int line )
{
   pd_byte_t *buf;
   malloc_record_t *mr;
   volatile int print = 0;

   /* Allocate block */
   buf = wsi_malloc( size );

   /* Create tracking structure */
   mr = wsi_malloc( sizeof( *mr ) );
   init_mr(mr, file, line, buf);

   WSS_MUTEX_ENTER( wss_malloc_record_tree );
   if( !wss_malloc_record_tree )
      wss_malloc_record_tree = new_tree();
   wss_util_rbtree_insert( wss_malloc_record_tree, &(mr->parent) );
   WSS_MUTEX_EXIT( wss_malloc_record_tree );

   /* Dump the allocation chain if debugging enabled */
   if( print )
      wss_malloc_dump();
   return( (void*)buf );
}

/*  wss_free_debug */
/*! \brief Free block of allocated memory
 *
 *  This function uses the malloc record tree to make sure the block of
 *  memory being freed was allocated by wss_malloc_debug().
 *
 *  \param ptr Pointer to block of memory to be freed
 */
void EXPORT
wss_free_debug( void *ptr )
{
   malloc_record_t *mr = wsi_malloc(sizeof(*mr));
   malloc_record_t *result;
   init_mr(mr, 0, 0, ptr);

   WSS_MUTEX_ENTER( wss_malloc_record_tree );
   result = (malloc_record_t *)wss_util_rbtree_delete(
           wss_malloc_record_tree, &(mr->parent));
   WSS_MUTEX_EXIT( wss_malloc_record_tree );
   wsi_free(mr);

   if( !result ) {
      wsi_printf(PD_MAKESTR("wss_free:: Freeing non-allocated memory.\n"));
      return;
   }
   wsi_free( ptr );
   wsi_free(result);
   return;
}

/*  wss_strdup_debug */
/*! \brief Debug tracking version of strdup
 *
 *  \param str  Pointer to string to be duplicated
 *  \param file Filename where calling function is contained
 *  \param line Line number of calling function
 *
 *  \return Pointer to copy of string contained in \a str,
 *  \return NULL if unable to allocate memory for string
 */
pd_char_t EXPORT *
wss_strdup_debug( const pd_char_t *str, pd_char_t *file, int line )
{
   int len = wsi_strlen( str );
   pd_char_t *buf = wss_malloc_string_allocator_debug( len, file, line );
   if( buf )
      wsi_strncpy( buf, str, len + 1 );

   return( buf );
}

/*  wss_malloc_string_allocator_debug */
/*! \brief Allocates a block of memory large enough to contain specified
 *         number of characters using debug tracking
 *
 *  This function handles the computation and allocation of a block of
 *  memory large enough to contain the specified number of characters.
 *  It will "Do The Right Thing" for Unicode character string on Unicode-
 *  enabled systems.
 *
 *  \param nchars Number of characters to allocate space for (excluding
 *                the terminating null character)
 *  \param file   Filename where calling function is contained
 *  \param line   Line number of calling function
 *
 *  \return Pointer to block of memory sufficiently sized to store the
 *          specified number of characters on success,
 *  \return NULL on allocation failure
 */
pd_char_t EXPORT *
wss_malloc_string_allocator_debug( size_t nchars, pd_char_t *file, int line)
{
   return wss_malloc_debug( ( nchars + 1 ) * sizeof( pd_char_t ), file, line);
}

#else /* MALLOC_DEBUG */

/*  wss_strdup */
/*! \brief WSS version of strdup exportable to wrappers
 *
 *  \param str  Pointer to string to be duplicated
 *
 *  \return Pointer to copy of string contained in \a str,
 *  \return NULL if unable to allocate memory for string
 */
pd_char_t EXPORT *
wss_strdup( const pd_char_t *str )
{
   int len = wsi_strlen( str );
   pd_char_t *buf = wss_malloc_string( len );
   if( buf )
      wsi_strncpy( buf, str, len + 1 );

   return buf;
}


/*  wss_malloc_string_allocator */
/*! \brief Allocates a block of memory large enough to contain specified
 *         number of characters
 *
 *  This function handles the computation and allocation of a block of
 *  memory large enough to contain the specified number of characters.
 *  It will "Do The Right Thing" for Unicode character string on Unicode-
 *  enabled systems.
 *
 *  \param nchars Number of characters to allocate space for (excluding
 *                the terminating null character)
 *
 *  \return Pointer to block of memory sufficiently sized to store the
 *          specified number of characters on success,
 *  \return NULL on allocation failure
 */
pd_char_t EXPORT *
wss_malloc_string_allocator( size_t nchars )
{
   return wss_malloc( ( nchars + 1 ) * sizeof( pd_char_t ) );
}

#endif /* MALLOC_DEBUG */

