/*! \file wss_module.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_module.c,v 1.73 2001/06/18 15:32:25 stevek Exp $
 *
 * \brief   Manage the WSS module list
 * \author  Chris Vance, Steve Kiernan
 * \version $Revision: 1.73 $
 * \date    $Date: 2001/06/18 15:32:25 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_file.h>
#include <wss_linker.h>
#include <wss_module.h>
#include <wss_malloc.h>
#include <wss_wrvevent.h>

#include <wsi_time.h>

/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static volatile int               wss_mod_next_id = 1;
wss_mod_t EXPORT                 *wss_module_list = NULL;
static volatile wss_mod_t        *wss_module_free_list = NULL;
WSS_MUTEX_PROTECT               ( wss_module_list );


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * _wss_mod_valid_file
 *
 * Description:
 *    Determine if file was opened by module
 *
 * Returns:
 *    TRUE  if file was opened by specified module
 *    FALSE otherwise
 */
static int
_wss_mod_valid_file( wss_mod_t *wrm, wss_file_t file )
{
   wss_mod_file_map_t *map;
   int                 found = FALSE;

   WSS_MUTEX_ENTER( wrm->wrm_filemap );
   for( map = wrm->wrm_filemap; map; map = map->map_next )
      if( map->map_file == file )
      {
         found = TRUE;
         break;
      }
   WSS_MUTEX_EXIT( wrm->wrm_filemap );

#ifdef DEBUG_WARN_BAD_FILE
   if( !found )
      wsi_printf( PD_MAKESTR("Warning: bad file handle 0x%x used by ")
                  PD_MAKESTR("wrapper module '%s'\n"), file, wrm->name );
#endif

   return found;
}

/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_mod_init */
/*! \brief Initialize module management code
 *
 *  Constructs the mutex protecting the module list.
 */
void EXPORT
wss_mod_init( void )
{
   WSS_MUTEX_INIT( wss_module_list );
}


/*  wss_mod_shutdown */
/*! \brief Shutdown module management code
 *
 *  Destroys the mutex protecting the module list.
 */
void EXPORT
wss_mod_shutdown( void )
{
   WSS_MUTEX_DESTROY( wss_module_list );
}


/*  wss_mod_find */
/*! \brief Find the module with the given id
 *
 *  \param id Identifier of module to find
 *
 *  \warning Module is reference before it is returned.  It is left
 *           to the caller to release the reference when the module
 *           structure is no longer needed.
 *
 *  \return Pointer to the module if it exists,
 *  \return NULL if it is not found
 */
wss_mod_t EXPORT *
#if defined(DEBUG_REFS) && !defined(DISABLE_MOD_DEBUG_REFS)
wss_mod_find_debug( wss_id_t id, const pd_char_t *file, int line )
#else
wss_mod_find( wss_id_t id )
#endif /* DEBUG_REFS */
{
   wss_mod_t *wrm;

   WSS_MUTEX_ENTER( wss_module_list );
   for( wrm = wss_module_list;
        wrm && ( wrm->wrm_id != id );
        wrm = wrm->wrm_next )
   ;
   if( wrm )
#if defined(DEBUG_REFS) && !defined(DISABLE_MOD_DEBUG_REFS)
      wss_mod_ref_debug( wrm, file, line );
#else
      wss_mod_ref( wrm );
#endif /* DEBUG_REFS */
   WSS_MUTEX_EXIT( wss_module_list );

   return wrm;
}


/*  wss_mod_find_by_name */
/*! \brief Find the module with the given name
 *
 *  \param name Name of module to find
 *
 *  \warning Module is reference before it is returned.  It is left
 *           to the caller to release the reference when the module
 *           structure is no longer needed.
 *
 *  \return Pointer to the module if it exists,
 *  \return NULL if it is not found
 */
wss_mod_t EXPORT *
#if defined(DEBUG_REFS) && !defined(DISABLE_MOD_DEBUG_REFS)
wss_mod_find_by_name_debug( pd_char_t *name, const pd_char_t *file, int line )
#else
wss_mod_find_by_name( pd_char_t *name )
#endif /* DEBUG_REFS */
{
   wss_mod_t *wrm;

   if( name == NULL )
      return NULL;

   WSS_MUTEX_ENTER( wss_module_list );
   for( wrm = wss_module_list;
        wrm && wrm->wrm_linked &&
           ( ( wrm->wrm_linked->linker_file_name == NULL ) ||
             wsi_strcmp( wrm->wrm_linked->linker_file_name, name ) );
        wrm = wrm->wrm_next )
   ;
   if( wrm )
#if defined(DEBUG_REFS) && !defined(DISABLE_MOD_DEBUG_REFS)
      wss_mod_ref_debug( wrm, file, line );
#else
      wss_mod_ref( wrm );
#endif /* DEBUG_REFS */
   WSS_MUTEX_EXIT( wss_module_list );

   return wrm;
}


/*  wss_mod_is_loaded */
/*! \brief Determine if named module has been loaded
 *
 *  \param name Name of module to check
 *
 *  \return TRUE if module is loaded,
 *  \return FALSE otherwise
 */
int EXPORT
wss_mod_is_loaded( const pd_char_t *name )
{
   wss_mod_t *wrm;
   int        isLoaded = FALSE;

   WSS_MUTEX_ENTER( wss_module_list );
   for( wrm = wss_module_list;
        wrm &&
           ( ( wrm->wrm_linked == NULL ) ||
             ( wrm->wrm_linked->linker_file_name == NULL ) ||
             wsi_strcmp( wrm->wrm_linked->linker_file_name, name ) );
        wrm = wrm->wrm_next )
   ;
   if( wrm )
      isLoaded = TRUE;
   WSS_MUTEX_EXIT( wss_module_list );

   return isLoaded;
}


/*  wss_mod_new */
/*! \brief Construct a new wrapper module
 *
 *  Load a module, relocates symbols, resolves any necessary data,
 *  and call the initialization function. 
 *
 *  \param filename Filename of module to load
 *  \param err      If non-NULL, the error code will be returned in the memory
 *                  pointed to by this parameter
 *
 *  \return Pointer to a newly-constructed module on success,
 *  \return NULL if an error occurs
 */
wss_mod_t EXPORT *
wss_mod_new( pd_char_t *filename, int *err )
{
   wss_mod_t *wrm = NULL;
   wss_id_t   id  = 0;

   /* Pull structure off the free list */
   if( wss_module_free_list )
   {
      WSS_MUTEX_ENTER( wss_module_list );
      if( wss_module_free_list )
      {
         wrm = (wss_mod_t *)wss_module_free_list;
         wss_module_free_list = wrm->wrm_next;

         id = wrm->wrm_id;
         bzero( wrm, sizeof( *wrm ) );

         /* Initialize the reference count */
         wss_ref_init( &(wrm->wrm_refcount) );
         WSS_MUTEX_INIT( wrm->wrm );

         /* Reset the identifier */
         wrm->wrm_id = id;

      }
      WSS_MUTEX_EXIT( wss_module_list );
   }

   /* Allocate memory for the module structure if nothing on the free list */
   if( wrm == NULL )
   {
      wrm = (wss_mod_t *)wss_malloc( sizeof( *wrm ) );

      if( wrm )
      {
         bzero( wrm, sizeof( *wrm ) );

         /* Initialize the reference count */
         wss_ref_init( &(wrm->wrm_refcount) );
         WSS_MUTEX_INIT( wrm->wrm );

         WSS_MUTEX_ENTER( wss_module_list );
         id = wrm->wrm_id = wss_mod_next_id++;
         WSS_MUTEX_EXIT( wss_module_list );
      }
   }

   if( wrm )
   {
      int error = 0;

      if( !error )
      {
         wss_linker_file_t *result =
               (wss_linker_file_t *)wss_malloc( sizeof(*result) );

         if( result )
         {
            if( filename )
               filename = wss_strdup( filename );

            wss_linker_file_init( result );

            WSS_MUTEX_INIT( wrm->wrm_filemap );
            WSS_MUTEX_INIT( wrm->wrm_uses );

            /*
             * Setup the wrapper module structure
             */
            wrm->wrm_loadtime = wsi_time_get_current(NULL);
            wrm->wrm_filemap = NULL;
            wrm->wrm_uses = NULL;

            WSS_MUTEX_ENTER( wss_module_list );
            if( !wss_module_list )
            {
               wss_module_list = wrm;
            }
            else
            {
               wrm->wrm_next = wss_module_list;
               wss_module_list = wrm;
            }
            wss_mod_ref( wrm );
            WSS_MUTEX_EXIT( wss_module_list );

            error = wss_linker_load( filename, result );
#ifdef DEBUG_LINKER
            if (error) {
                wsi_printf(PD_MAKESTR("Linker: wss_linker_load error.\n"));
            }
            if (!result) {
                wsi_printf(
                        PD_MAKESTR("Linker: wss_linker_load, null result.\n"));
            }
#endif
         }
         else
            error = ENOMEM;

         if( error == 0 )
         {
            wss_linker_addr_t      addr;
            const pd_char_t      *type;
            wss_linker_file_ops_t *ops;

            ops = result->linker_file_ops;

            if( wss_mod_is_loaded( result->linker_file_name ) )
            {
#ifdef DEBUG_LINKER
               wsi_printf( PD_MAKESTR("Linker: Module '%s' already loaded\n"),
                           (pd_char_t *)addr.addr );
#endif /* DEBUG_LINKER */
               error = EEXIST;
               goto modlinkdone;
            }

            /* Get the module's type */
            error = (ops->lookup_symbol)( result,
                                          PD_MAKESTR(WRLINKNAME_MOD_TYPE),
                                          &addr );
            if( error )
            {
#ifdef DEBUG_LINKER
               wsi_printf( PD_MAKESTR("Linker: Unable to determine symbol ")
                           PD_MAKESTR("'%s'\n"),
                           PD_MAKESTR(WRLINKNAME_MOD_TYPE) );
#endif /* DEBUG_LINKER */
               goto modlinkdone;
            }

            type = *(const pd_char_t **)addr.addr;

#ifdef DEBUG_LINKER
            wsi_printf( PD_MAKESTR("Linker: Module name '%s', type '%s' ")
                        PD_MAKESTR("initializing...\n"),
                        result->linker_file_name, type );
#endif /* DEBUG_LINKER */

            /* Set type */
            if( !wsi_strcmp( type, PD_MAKESTR("library") ) )
               wrm->wrm_type = WR_MOD_LIBRARY;
            else if( !wsi_strcmp( type, PD_MAKESTR("wrapper") ) )
               wrm->wrm_type = WR_MOD_WRAPPER;
            else if( !wsi_strcmp( type, PD_MAKESTR("activation") ) )
               wrm->wrm_type = WR_MOD_ACTIVATION;
            else if( !wsi_strcmp( type, PD_MAKESTR("deactivation") ) )
               wrm->wrm_type = WR_MOD_DEACTIVATION;
            else
            {
#ifdef DEBUG_LINKER
               wsi_printf( PD_MAKESTR("Linker: Invalid type encountered.\n") );
#endif /* DEBUG_LINKER */
               error = EINVAL;
               goto modlinkdone;
            }

            /* Initialize module-specific parts */
            switch( wrm->wrm_type )
            {
               case WR_MOD_WRAPPER:
               {
                  wss_linker_addr_t dispatch_pre, dispatch_post;
                  wss_linker_addr_t wr_activate, wr_deactivate;
                  wss_linker_addr_t wr_duplicate;      
                  wss_linker_addr_t wr_install, wr_uninstall;

                  bzero( &dispatch_pre, sizeof(dispatch_pre) );
                  bzero( &dispatch_post, sizeof(dispatch_post) );

                  error =
                     wss_linker_lookup_multiple( result, FALSE,
                                                 PD_MAKESTR("dispatch_pre"),
                                                 &dispatch_pre,
                                                 PD_MAKESTR("dispatch_post"),
                                                 &dispatch_post,
                                                 NULL );
                  /* Check for error, but ignore ENOENT.  We do not care
                     if there is a missing dispatch_pre or dispatch_post */
                  if( error && ( error != ENOENT ) )
                  {
#ifdef DEBUG_LINKER
                     wsi_printf( PD_MAKESTR("Linker: Unable to determine ")
                                 PD_MAKESTR("necessary wrapper symbols for ")
                                 PD_MAKESTR("'%s'\n"),
                                 wrm->wrm_linked->linker_file_name );
#endif /* DEBUG_LINKER */
                     goto modlinkdone;
                  }

                  error = 
                     wss_linker_lookup_multiple( result, TRUE,
                                                 PD_MAKESTR("wr_activate"),
                                                 &wr_activate,
                                                 PD_MAKESTR("wr_deactivate"),
                                                 &wr_deactivate,
                                                 PD_MAKESTR("wr_duplicate"),
                                                 &wr_duplicate,
                                                 PD_MAKESTR("wr_install"),
                                                 &wr_install,
                                                 PD_MAKESTR("wr_uninstall"),
                                                 &wr_uninstall,
                                                 NULL );
                  if( error )
                  {
#ifdef DEBUG_LINKER
                     wsi_printf( PD_MAKESTR("Linker: Unable to determine ")
                                 PD_MAKESTR("necessary wrapper symbols for ")
                                 PD_MAKESTR("'%s'\n"),
                                 wrm->wrm_linked->linker_file_name );
#endif /* DEBUG_LINKER */
                     goto modlinkdone;
                  }

                  /* Let system-specific actions be performed on module
                     before we finalize */
                  error = wsi_mod_new( wrm, result );
                  if( error )
                     goto modlinkdone;

                  /* Finalize usage of the file */
                  (ops->finalize)( result );
                  wss_linker_file_ref( result );
                  wrm->wrm_linked = result;

                  /* Register wrapper with the WSS */
                  wss_wrapper_register( wrm, dispatch_pre.addr, dispatch_post.addr,
                                        wr_activate.addr, wr_deactivate.addr,
                                        wr_duplicate.addr );
                  wss_mod_at_unload( wr_uninstall.addr, wrm->wrm_id );

                  /* Call wr_install */
#ifdef DEBUG_LINKER
                  wsi_printf( PD_MAKESTR("Linker: Calling wrapper ")
                              PD_MAKESTR("'wr_install'\n") );
#endif /* DEBUG_LINKER */
                  error = (*(int(*)(int))wr_install.addr)( wrm->wrm_id );

                  break;
               }

               case WR_MOD_LIBRARY:
               {
                  wss_linker_addr_t  init, fini;

                  (ops->export_symbols)( result );

                  bzero( &init, sizeof(init) );
                  bzero( &fini, sizeof(fini) );
                  (ops->lookup_symbol)( result, PD_MAKESTR("init"), &init );
                  (ops->lookup_symbol)( result, PD_MAKESTR("fini"), &fini );

                  /* Let system-specific actions be performed on module
                     before we finalize */
                  error = wsi_mod_new( wrm, result );
                  if( error )
                     goto modlinkdone;

                  /* Finalize usage of the file */
                  (ops->finalize)( result );
                  wss_linker_file_ref( result );
                  wrm->wrm_linked = result;

                  wss_lib_register( wrm, result->linker_file_symbols,
                                    fini.addr );

                  if( init.addr )
                  {
#ifdef DEBUG_LINKER  
                     wsi_printf( PD_MAKESTR("Linker: Calling library ")
                                 PD_MAKESTR("'init'\n") );
#endif /* DEBUG_LINKER */
                     (*(void(*)(void))init.addr)();
                  }

                  break;
               }

               case WR_MOD_ACTIVATION:
               case WR_MOD_DEACTIVATION:
               {
                  wss_linker_addr_t   activate, imp_list;
                  pd_char_t         **imp;
                  const pd_char_t    *actsym =
                     (wrm->wrm_type == WR_MOD_ACTIVATION)
                     ? PD_MAKESTR("activate")
                     : PD_MAKESTR("deactivate");

                  error = 
                     wss_linker_lookup_multiple( result, TRUE,
                                                 actsym, &activate,
                                                 PD_MAKESTR("imp_list"),
                                                 &imp_list, NULL );
                  if( error )
                  {
#ifdef DEBUG_LINKER
                     wsi_printf( PD_MAKESTR("Linker: Unable to determine ")
                                 PD_MAKESTR("necessary criteria symbols for ")
                                 PD_MAKESTR("'%s'\n"),
                                 wrm->wrm_linked->linker_file_name );
#endif /* DEBUG_LINKER */
                     goto modlinkdone;
                  }

                  for( imp = (pd_char_t **)imp_list.addr; *imp; imp++ )
                     if( !wss_mod_is_loaded( *imp ) )
                     {
#ifdef DEBUG_LINKER
                        wsi_printf( PD_MAKESTR("Linker: missing ")
                                    PD_MAKESTR("dependency module '%s'\n"),
                                    *imp );
#endif /* DEBUG_LINKER */
                        error = ENOENT;
                        goto modlinkdone;
                     }


                  /* Let system-specific actions be performed on module
                     before we finalize */
                  error = wsi_mod_new( wrm, result );
                  if( error )
                     goto modlinkdone;

                  /* Finalize usage of the file */
                  (ops->finalize)( result );
                  wss_linker_file_ref( result );
                  wrm->wrm_linked = result;

                  wss_crit_add( wrm, activate.addr,
                                (pd_char_t **)imp_list.addr );

                  break;
               }

            }
modlinkdone:
            if( error )
            {
               if( wrm->wrm_linked )
               {
                  wrm->wrm_linked = NULL;
                  wss_linker_file_rele( result );
               }
               wss_linker_file_rele( result );
               wss_free( result );
            }
         }
      }
      
      wss_mod_rele( wrm );

      if( error )
      {
         wrm = NULL;
         wss_mod_delete( id );

         if( err )
            *err = error;
      }
      else
         wss_wrvevent_post_modloaded( id, wrm->wrm_type );
   }

   return wrm;
}


/*  wss_mod_delete */
/*! \brief Removes the specified module from the WSS
 *
 *  The memory used by the module is deallocated while unloading.  All
 *  references to the module (excluding the module-specific structure
 *  references) must be released before calling this function.
 *
 *  \return 0 if successful,
 *  \return ENOENT if module identified by \a id not found,
 *  \return EBUSY if module is being used
 */
int EXPORT
wss_mod_delete( wss_id_t id )
{
   wss_mod_file_map_t *map, *nextmap;
   wss_mod_t          *wrm, *prev;
   wss_linker_file_t  *lf;

   /* Find the corresponding structure in the global list.  */
   WSS_MUTEX_ENTER( wss_module_list );
   for( prev = NULL, wrm = wss_module_list;
        wrm && wrm->wrm_id != id;
        prev = wrm, wrm = wrm->wrm_next )
   ;

   if( wrm == NULL )
   {
      WSS_MUTEX_EXIT( wss_module_list );
#ifdef DEBUG_MODULE
      wsi_printf( PD_MAKESTR("ERROR finding pre-loaded module id %d\n"), id );
#endif /* DEBUG_MODULE */
      return ENOENT;
   }

   /*
    * If reference count is > 2 (one for module self-reference, one
    * for the type structure reference), module is busy.
    */
   if( WSS_REF_COUNT(wrm->wrm_refcount) > 2 )
   {
      WSS_MUTEX_EXIT( wss_module_list );
#ifdef DEBUG_MODULE
      wsi_printf( PD_MAKESTR("ERROR: module #%d (0x%x) refcount > 2 on ")
                  PD_MAKESTR("delete\n"),
                  id, (unsigned int)wrm );
#endif /* DEBUG_MODULE */
      return EBUSY;
   }

   /*
    * If reference count for linker file > 2 (one for linker file
    * self-reference, one for module reference), module is busy.
    */
   if( wrm->wrm_linked &&
       ( WSS_REF_COUNT(wrm->wrm_linked->linker_file_refcount) > 2 ) )
   {
      WSS_MUTEX_EXIT( wss_module_list );
#ifdef DEBUG_MODULE
      wsi_printf( PD_MAKESTR("ERROR: module #%d (0x%x) linker file refcount")
                  PD_MAKESTR(" > 2 on delete\n"),
                  id, (unsigned int)wrm );
#endif /* DEBUG_MODULE */
      return EBUSY;
   }

   /* Make sure type-specific modules say it's okay to unload */
   switch( wrm->wrm_type )
   {
      case WR_MOD_WRAPPER:
         if( wss_wrapper_can_unload( wrm->wrm_id ) == FALSE )
         {
            WSS_MUTEX_EXIT( wss_module_list );
#ifdef DEBUG_MODULE
            wsi_printf( PD_MAKESTR("ERROR: module #%d (0x%x) wrapper busy\n"),
                        id, (unsigned int)wrm );
#endif /* DEBUG_MODULE */
            return EBUSY;
         }
         break;

      case WR_MOD_ACTIVATION:
      case WR_MOD_DEACTIVATION:
         if( wss_crit_can_unload( wrm->wrm_id ) == FALSE )
         {
            WSS_MUTEX_EXIT( wss_module_list );
#ifdef DEBUG_MODULE
            wsi_printf( PD_MAKESTR("ERROR: module #%d (0x%x) criteria busy\n"),
                        id, (unsigned int)wrm );
#endif /* DEBUG_MODULE */
            return EBUSY;
         }
         break;

      case WR_MOD_LIBRARY:
         if( wss_lib_can_unload( wrm->wrm_id ) == FALSE )
         {
            WSS_MUTEX_EXIT( wss_module_list );
#ifdef DEBUG_MODULE
            wsi_printf( PD_MAKESTR("ERROR: module #%d (0x%x) library busy\n"),
                        id, (unsigned int)wrm );
#endif /* DEBUG_MODULE */
            return EBUSY;
         }
         break;

      default:
         break;
   }

   /* Fix-up list pointers */
   if( prev == NULL )
      wss_module_list = wrm->wrm_next;
   else
      prev->wrm_next = wrm->wrm_next;
   WSS_MUTEX_EXIT( wss_module_list );

   /* Close and unmap any files associated with this module */
   WSS_MUTEX_ENTER( wrm->wrm_filemap );
   for( map = wrm->wrm_filemap; map; map = nextmap )
   {
      nextmap = map->map_next;

      /* wss_file_close will unmap any mapped file, we do not
         need to explicitly call wss_file_unmap */
      wss_file_close( map->map_file );

      wss_free( map );
   }
   wrm->wrm_filemap = NULL;
   WSS_MUTEX_EXIT( wrm->wrm_filemap );

   /* Call unload function if it exists */
   if( wrm->wrm_unload )
   {
      void (*unload)(void);

      WSS_MUTEX_ENTER( wrm->wrm );
      unload = wrm->wrm_unload;
      WSS_MUTEX_EXIT( wrm->wrm );

#ifdef DEBUG_MODULE
      wsi_printf( PD_MAKESTR("Executing module %d unload function\n"),
                  wrm->wrm_id );
#endif /* DEBUG_MODULE */

      /*
       * Call the unload function. 
       */
      (*unload)();

      wss_mod_remove_at_unload( wrm->wrm_id );
   }

   /* Handle type-specific unloading */
   switch( wrm->wrm_type )
   {
      case WR_MOD_WRAPPER:
         wss_wrapper_delete( wrm->wrm_id );
         break;

      case WR_MOD_ACTIVATION:
      case WR_MOD_DEACTIVATION:
         wss_crit_delete( wrm->wrm_id );
         break;

      case WR_MOD_LIBRARY:
         wss_lib_delete( wrm->wrm_id );
         break;

      default:
         break;
   }

   /* Release reference to linker file and remove from list */
   lf = wrm->wrm_linked;
   wrm->wrm_linked = NULL;
   if( lf )
   {
      wss_linker_remove_file( lf );
      wss_linker_file_rele( lf );
   }

   /* Delete system-dependent part of module structure */
   wsi_mod_delete( wrm );

   /* Release final reference */
   wss_mod_rele( wrm );

   return 0;
}

/*  wss_mod_free */
/*! \brief Add module structure to the free list
 *
 *  \param wrm Pointer to the module structure
 */
void
wss_mod_free( wss_mod_t *wrm )
{
   assert( wrm != NULL );
   assert( WSS_REF_COUNT(wrm->wrm_refcount) == 1 );

   WSS_REF_DEC( wrm->wrm_refcount );

   /* Destroy syncronization object protected the reference count */
   WSS_MUTEX_DESTROY( wrm->wrm );
   WSS_MUTEX_DESTROY( wrm->wrm_filemap );
   WSS_MUTEX_DESTROY( wrm->wrm_uses );

   /* Add module structure to the free list */
   WSS_MUTEX_ENTER( wss_module_list );
   wss_ref_destroy( &(wrm->wrm_refcount) );
   wrm->wrm_next = (wss_mod_t *)wss_module_free_list;
   wss_module_free_list = wrm;
   WSS_MUTEX_EXIT( wss_module_list );
}


/*  wss_mod_unload_all */
/*! \brief Unloads all loaded modules from the WSS
 *
 *  Only called when the WSS is being unloaded from the kernel
 */
void EXPORT
wss_mod_unload_all()
{
   wss_mod_t *wrm, *this;
   int        id;
        
   /* Lock the global module list */
   WSS_MUTEX_ENTER( wss_module_list );
   wrm = wss_module_list;
   wss_module_list = NULL;
   WSS_MUTEX_EXIT( wss_module_list );

   while( wrm )
   {
#ifdef DEBUG_MODULE
      wsi_printf( PD_MAKESTR("wss_mod_unload_all: unloading module....\n") );
#endif /* DEBUG_MODULE */
      this = wrm;
      wrm = wrm->wrm_next;
      id = this->wrm_id;

      /* Call unload function if it exists */
      if( this->wrm_unload )
      {
         void (*unload)(void);

         WSS_MUTEX_ENTER( this->wrm );
         unload = this->wrm_unload;
         WSS_MUTEX_EXIT( this->wrm );

         (*unload)();
      }
        
      /* Delete system-dependent part of module structure */
      wsi_mod_delete( this );

      /* Destroy synchronization objects */
      WSS_MUTEX_DESTROY( this->wrm );
      WSS_MUTEX_DESTROY( this->wrm_filemap );
      WSS_MUTEX_DESTROY( this->wrm_uses );

      /* Free memory used by structure */
      wss_free( this );

#ifdef DEBUG_MODULE
      wsi_printf( PD_MAKESTR("wss_mod_unload_all: unloaded module %d\n"), id );
#endif /* DEBUG_MODULE */
   }
}


/*  wss_mod_at_unload */
/*! \brief Install a function to be called by the kernel when the installing
 *         module is unloaded
 *
 *  The at unload function takes no arguments and has no return value.
 *
 *  \param uf Pointer to the unload function to call when module is unloaded
 *  \param id Identifier of module
 *
 *  \return 0 on success,
 *  \return ENOENT if the requested module could not be found
 *  \internal
 */
int EXPORT
wss_mod_at_unload( void *uf, wss_id_t id )
{
   wss_mod_t *wrm;

#ifdef DEBUG_MODULE
   wsi_printf( PD_MAKESTR("wss_mod_at_unload( 0x%x, %d )\n"),
               (u_int)uf, id );
#endif /* DEBUG_MODULE */

   /* Find and lock the module */
   if( ( wrm = wss_mod_find( id ) ) != NULL )
   {
      WSS_MUTEX_ENTER( wrm->wrm );
      /* Set the unload function blindly (this is usually only done once) */
      (void*)wrm->wrm_unload = uf;
      WSS_MUTEX_EXIT( wrm->wrm );

      /* Release reference to the module */
      wss_mod_rele( wrm );

      /* Return success */
      return 0;
   }

   /* Could not find the specified module */
   return ENOENT;
}


/*  wss_mod_remove_at_unload */
/*! \brief Deregister the at unload function.
 *
 *  TBD: This could be replaced by a macro which calls
 *       wss_mod_at_unload( NULL, id )
 *
 *  \param id Identifier of module
 *
 *  \return 0 on success,
 *  \return ENOENT if the requested module could not be found
 *  \internal
 */
int EXPORT
wss_mod_remove_at_unload( wss_id_t id )
{
   wss_mod_t *wrm;

#ifdef DEBUG_MODULE
   wsi_printf( PD_MAKESTR("wss_mod_remove_at_unload( %d )\n"), id );
#endif /* DEBUG_MODULE */

   if( ( wrm = wss_mod_find( id ) ) != NULL )
   {
      WSS_MUTEX_ENTER( wrm->wrm );
      (void*)wrm->wrm_unload = NULL;
      WSS_MUTEX_EXIT( wrm->wrm );

      wss_mod_rele( wrm );

      return 0;
   }       

   return ENOENT;
}      


/*  wss_mod_add_dependency */
/*! \brief Add a dependency between two modules
 *
 *  This function adds a module to the dependency the list of the
 *  module specified by \a wrm.  The \a uses module will have a
 *  reference held on it by the \a wrm module.
 *
 *  \param wrm  Module which using another
 *  \param uses Module which is used by former
 *
 *  \return 0 if successfully added,
 *  \return EINVAL if an invalid paramter is passed,
 *  \return EEXIST if the dependency already exists
 */
int EXPORT
wss_mod_add_dependency( wss_mod_t *wrm, wss_mod_t *uses )
{
   wss_linker_file_depend_t *depend;

   if( ( wrm == NULL ) || ( uses == NULL ) )
      return EINVAL;

   if( wss_linker_file_has_dependency( wrm->wrm_linked, uses->wrm_linked ) )
      return EEXIST;

   depend = (wss_linker_file_depend_t *)wss_malloc( sizeof(*depend) );
   wss_linker_file_ref( uses->wrm_linked );
   depend->linker_file_depend_entry = uses->wrm_linked;
   WSS_MUTEX_ENTER( wrm->wrm_linked->linker_file );
   depend->linker_file_depend_next =
      wrm->wrm_linked->linker_file_dependency_list;
   wrm->wrm_linked->linker_file_dependency_list = depend;
   WSS_MUTEX_EXIT( wrm->wrm_linked->linker_file );

   return 0;
}      


/*  wss_mod_remove_dependency */
/*! \brief Remove a dependency between two modules
 *
 *  This function removes a module from the dependency list of the \a wrm
 *  module and releases the reference on the \a used module held by \a wrm.
 *
 *  \param uses_id Module which using another
 *  \param used_id Module which is used by former
 *
 *  \return 0 if successfully added,
 *  \return EINVAL if an invalid paramter is passed,
 *  \return ENOENT if the dependency does not exist
 */
int EXPORT
wss_mod_remove_dependency( wss_mod_t *wrm, wss_mod_t *used )
{
   wss_linker_file_depend_t  *depend;
   wss_linker_file_depend_t **last;

   if( ( wrm == NULL ) || ( used == NULL ) )
      return EINVAL;

   WSS_MUTEX_ENTER( wrm->wrm_linked->linker_file );
   for( depend = wrm->wrm_linked->linker_file_dependency_list,
           last = &(wrm->wrm_linked->linker_file_dependency_list);
        depend;
        last = &(depend->linker_file_depend_next),
           depend = depend->linker_file_depend_next )
   {
      if( depend->linker_file_depend_entry == used->wrm_linked )
      {
         *last = depend->linker_file_depend_next;
         wss_linker_file_rele( depend->linker_file_depend_entry );
         wss_free( (void *)depend );
         WSS_MUTEX_EXIT( wrm->wrm_linked->linker_file );
         return 0;
      }
   }
   WSS_MUTEX_EXIT( wrm->wrm_linked->linker_file );

   return ENOENT;
}


/***************************************************************************\
 * Module-centric file operations                                          *
\***************************************************************************/

/*  wss_mod_close_file */
/*! \brief Close file associated to a module
 *
 *  \param id     Identifier of module
 *  \param handle Handle of open file
 */
void EXPORT
wss_mod_close_file( int id, wss_file_t handle )
{
   wss_mod_t *wrm = wss_mod_find( id );

   if( wrm )
   {
      wss_mod_file_map_t *map, **last;

      WSS_MUTEX_ENTER( wrm->wrm_filemap );
      for( map = wrm->wrm_filemap, last = &wrm->wrm_filemap;
           map;
           last = &map->map_next, map = map->map_next )
         if( map->map_file == handle )
         {
            /* Remove from list */
            *last = map->map_next;

            /* Release reference to file */
            wss_file_rele( handle );

            break;
         }
      WSS_MUTEX_EXIT( wrm->wrm_filemap );

      if( map )
      {
         /* Close file and free list node */
         wss_file_close( handle );
         wss_free( map );
      }

      wss_mod_rele( wrm );
   }
}


/*  wss_mod_get_file_map_addr */
/*! \brief Get memory-mapped address of file mapped by module
 *
 *  \param id     Identifier of module
 *  \param handle Handle of open file
 *
 *  \return Pointer to beginning of mapped view of file on success,
 *  \return NULL otherwise
 */
void EXPORT *
wss_mod_get_file_map_addr( int id, wss_file_t handle )
{
   wss_mod_t *wrm  = wss_mod_find( id );
   void      *addr = NULL;

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         addr = wss_file_get_map_addr( handle );

      wss_mod_rele( wrm );
   }

   return addr;
}


/*  wss_mod_get_file_size */
/*! \brief Get size of the file opened by the module
 *
 *  \param id     Identifier of module
 *  \param handle Handle of open file
 *
 *  \return Size of file on success,
 *  \return 0 on failure
 */
size_t EXPORT
wss_mod_get_file_size( int id, wss_file_t handle )
{
   wss_mod_t *wrm  = wss_mod_find( id );
   size_t     size = 0;

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         size = wss_file_get_size( handle );

      wss_mod_rele( wrm );
   }

   return size;
}


/*  wss_mod_map_file */
/*! \brief Map file opened by module into memory space
 *
 *  \param id      Identifier of module
 *  \param handle  Handle of open file
 *  \param size    Number of bytes to map in (some systems will round-up
 *                 to a page boundary)
 *  \param mapmode Mode to map file view
 *  \param pos     Offset into file to begin mapping
 *
 *  \return 0 on success,
 *  \return error code on failure
 */
int EXPORT
wss_mod_map_file( int id, wss_file_t handle, size_t size, int mapmode,
                  pd_offset_t pos )
{
   wss_mod_t *wrm    = wss_mod_find( id );
   int        retval = EINVAL;

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         retval = wss_file_map( handle, size, mapmode, pos );

      wss_mod_rele( wrm );
   }

   return retval;
}


/*  wss_mod_open_file */
/*! \brief Open named file within the module's context
 *
 *  Using wss_mod_open_file() instead of the wss_file_open() function
 *  allows for automatic cleanup of files associated with a module's context
 *  when a module is unloaded
 *
 *  \param id       Identifier of module
 *  \param filename Name of file to open
 *  \param mode     Mode with which to open the file
 *
 *  \return Handle for sucessfully opened file,
 *  \return \b WSS_FILE_INVALID on failure
 */
wss_file_t EXPORT
wss_mod_open_file( int id, const pd_char_t *filename, wss_file_mode_t mode )
{
   wss_mod_t  *wrm  = wss_mod_find( id );
   wss_file_t  file = WSS_FILE_INVALID;

   if( wrm )
   {
      if( ( file = wss_file_open( filename, mode ) ) != WSS_FILE_INVALID )
      {
         wss_mod_file_map_t *map;

         if( ( map = wss_malloc( sizeof(wss_mod_file_map_t) ) ) != NULL )
         {
            wss_file_ref( file );
            map->map_file = file;

            WSS_MUTEX_ENTER( wrm->wrm_filemap );
            map->map_next = wrm->wrm_filemap;
            wrm->wrm_filemap = map;
            WSS_MUTEX_EXIT( wrm->wrm_filemap );
         }
         else
         {
            wss_file_close( file );
            file = WSS_FILE_INVALID;
         }
      }

      wss_mod_rele( wrm );
   }

   return file;
}


/*  wss_mod_read_file */
/*! \brief Read bytes from file opened by module into the specified buffer
 *
 *  \param id     Identifier of module
 *  \param handle Handle of open file
 *  \param buffer Pointer to buffer which will contain the read data
 *  \param len    Number of bytes to read
 *
 *  \return Number of bytes read on success,
 *  \return 0 on failure
 */
int EXPORT
wss_mod_read_file( int id, wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   wss_mod_t *wrm    = wss_mod_find( id );
   int        retval = EINVAL;

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         retval = wss_file_read( handle, buffer, len );

      wss_mod_rele( wrm );
   }

   return retval;
}


/*  wss_mod_read_file_alloc */
/*! \brief Read bytes from file opened by module into an allocated buffer
 *
 *  \param id     Identifier of module
 *  \param handle Handle of open file
 *  \param len    Pointer to address which contains the number of bytes
 *                to read and will contain the number of bytes allocated
 *                on a sucessful read
 *
 *  \return Pointer to buffer allocated by wss_malloc() which contains the
 *          data read on success,
 *  \return NULL on failure
 */
pd_byte_t EXPORT *
wss_mod_read_file_alloc( int id, wss_file_t handle, size_t *len )
{
   wss_mod_t *wrm    = wss_mod_find( id );
   void      *buffer = NULL;

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         buffer = wss_file_read_alloc( handle, len );

      wss_mod_rele( wrm );
   }

   return buffer;
}


/*  wss_mod_seek_file */
/*! \brief Repositions the file offset pointer
 *
 *  This is equivalent to the C library function lseek(2)
 *
 *  \param id     Identifier of module
 *  \param handle Handle of open file
 *  \param offset Offset to seek within file
 *  \param whence Where to start seeking from (beginning, current position,
 *                or end of file)
 *
 *  \return 0 on success,
 *  \return EINVAL if an invalid argument is given
 */
int EXPORT
wss_mod_seek_file( int id, wss_file_t handle, pd_offset_t offset, int whence )
{
   wss_mod_t *wrm    = wss_mod_find( id );
   int        retval = EINVAL;

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         retval = wss_file_seek( handle, offset, whence );

      wss_mod_rele( wrm );
   }

   return retval;
}


/*  wss_mod_unmap_file */
/*! \brief Unmaps a file previously mapped into the WSS memory space
 *
 *  \param id     Identifier of module
 *  \param handle Handle to mapped file
 */
void EXPORT
wss_mod_unmap_file( int id, wss_file_t handle )
{
   wss_mod_t *wrm    = wss_mod_find( id );

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         wss_file_unmap( handle );

      wss_mod_rele( wrm );
   }
}


/*  wss_mod_write_file */
/*! \brief Writes bytes from the specified buffer to file opened by module 
 *
 *  \param id     Identifier of module
 *  \param handle Handle of open file
 *  \param buffer Pointer to buffer containing data to be written
 *  \param len    Number of bytes to write out
 *
 *  \return Number of bytes written to file on success,
 *  \return 0 on failure
 */
int EXPORT
wss_mod_write_file( int id, wss_file_t handle, pd_byte_t *buffer, size_t len )
{
   wss_mod_t *wrm    = wss_mod_find( id );
   int        retval = EINVAL;

   if( wrm )
   {
      if( _wss_mod_valid_file( wrm, handle ) )
         retval = wss_file_write( handle, buffer, len );

      wss_mod_rele( wrm );
   }

   return retval;
}

