/*! \file wss_process.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_process.c,v 1.61 2001/06/23 00:07:41 stevek Exp $
 *
 * \brief   Process management routines
 * \author  Steve Kiernan
 * \version $Revision: 1.61 $
 * \date    $Date: 2001/06/23 00:07:41 $
 */


#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_malloc.h>
#include <wss_process.h>
#include <wss_wrapper.h>
#include <wss_module.h>
#include <wss_iface.h>
#include <wss_path.h>
#include <wss_wrv.h>
#include <wss_mutex.h>
#include <wss_utilities.h>

#include <wrerrno.h>


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

/*! Wrapper process structure list */
struct wss_proclist EXPORT wss_proc_list;
/*! Mutex protecting the wrapper process structure list */
WSS_MUTEX_PROTECT         (wss_proc_list);


/*! Number of processes still within a system call */
int EXPORT        wss_proc_num_in_syscall = 0;
/*! Mutex used to synchronize WSS kernel module unloading */
WSS_MUTEX_PROTECT(wss_proc_num_in_syscall);
/*! Condition variable used to synchronize WSS kernel module unloading */
WSS_CV_PROTECT   (wss_proc_num_in_syscall);


/***************************************************************************\
 * External references                                                     *
\***************************************************************************/
extern int wsi_unloading;


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_pfind_unsafe */
/*! \brief Function which mimics the UNiX kernel's pfind function.
 *
 *  It searches the global list of wrappers-specific process information for
 *  the given PID.
 *
 *  \param pid PID of process to find
 *
 *  \warning Assumes lock on wss_proc_list already held by current thread
 *
 *  \return Pointer to the wrapper process structure if it was found,
 *  \return NULL if it was not located
 *
 */
wss_proc_t EXPORT *
wss_pfind_unsafe( register pid_t pid )
{
   register wss_proc_t *wrp = NULL;

   for( wrp = wss_proc_list.lh_first; wrp; wrp = wrp->wrp_p_list.le_next )
      if( wrp->wrp_pid == pid )
      {
         wss_proc_ref( wrp );
         break;
      }

   return wrp;
}


/*  wss_pfind */
/*! \brief Thread-safe function which mimics the UNiX kernel's pfind function.
 *
 *  It searches the global list of wrappers-specific process information for
 *  the given PID.
 *
 *  \param pid PID of process to find
 *
 *  \warning Assumes lock on wss_proc_list is \b NOT already held by current
 *           thread
 *
 *  \return Pointer to the wrapper process structure if it was found,
 *  \return NULL if it was not located
 *
 */
wss_proc_t EXPORT *
wss_pfind( register pid_t pid )
{
   register wss_proc_t *wrp;

   WSS_MUTEX_ENTER( wss_proc_list );
   wrp = wss_pfind_unsafe( pid );
   WSS_MUTEX_EXIT( wss_proc_list );

   return wrp;
}


/*  wss_proc_init */
/*! \brief The process handling module initialization routine
 *
 *  The wrappers-specific process list needs to be initialized before
 *  it can be used.  This simply calls a macro from queue.h and initializes
 *  the synchronization objects.
 */
void EXPORT
wss_proc_init()
{
   WSS_MUTEX_INIT( wss_proc_num_in_syscall );
   WSS_CV_INIT( wss_proc_num_in_syscall );
   WSS_MUTEX_INIT( wss_proc_list );
   LIST_INIT( &wss_proc_list );
}


/*  wss_proc_shutdown */
/*! \brief The process handling module shutdown routine
 *
 *  When the module is being unloaded, go through the list and free any
 *  malloc'd structures and try to put things back the way they were
 *  before we changed them.
 */
void EXPORT
wss_proc_shutdown()
{
   int i = 0;
   int j = 0;
   wss_proc_t *wrp,*wrl;
   wss_proc_t **last = &wss_proc_list.lh_first;

   WSS_MUTEX_ENTER( wss_proc_num_in_syscall );
   wss_proc_num_in_syscall = 0;
   WSS_MUTEX_EXIT( wss_proc_num_in_syscall );
   wrp = wss_proc_list.lh_first;
   while( wrp )
   {
      int insyscall;

      j++;
      if( ( insyscall = wsi_proc_insyscall( wrp ) ) == 0 )
      {
         i++;
         wsi_printf( PD_MAKESTR("cleaning up '%s'\n"), wrp->wrp_path );
         if( wrp->wrp_wi_list )
            wss_proc_del_all_wrappers( wrp );

         wrl = wrp;

         *last = wrp = wrp->wrp_p_list.le_next;
         if( wrl->wrp_path )
         {
            wss_free( (void *)wrl->wrp_path );
            wrl->wrp_path = NULL;
         }
         wsi_proc_free( (void *)wrl );
         wss_free( (void *)wrl );
      }
      else
      {
         WSS_MUTEX_ENTER( wss_proc_num_in_syscall );
         wss_proc_num_in_syscall++;
         WSS_MUTEX_EXIT( wss_proc_num_in_syscall );

         wsi_printf( PD_MAKESTR("pid %d has %d orphaned syscall(s)\n"),
                     (int)wrp->wrp_pid, insyscall );
         wsi_proc_force_syscall_exit( wrp );
         last = &( wrp->wrp_p_list.le_next );
         wrp = wrp->wrp_p_list.le_next;
      }
   }

   wsi_printf( PD_MAKESTR("Cleaned up %d/%d wss_proc entries\n"), i, j );
}


/*  wss_proc_exit_stage1 */
/*! \brief The stage 1 exit hook function
 *
 *  It calls deactivate on all wrappers which wrap the exiting process.
 *
 *  \param p Pointer to the process structure for the exiting process
 */
void EXPORT
wss_proc_exit_stage1( proc_t *p )
{
   wss_proc_t *wrp;
   int         err;

   wrp = wss_pfind( wsi_proc_pid(p) );
   if( wrp )
   {
      if( wrp->wrp_wi_list )
      {
         err = wss_wrapper_call_deactivation( wrp );
         if( err != 0 )
         {
            /*
             * TBD: Is there anything else we can
             * do here?
             */
            wsi_printf( PD_MAKESTR("Deactivation returned error!") );
         }
      }

      wss_proc_rele( wrp );
   }
}


/*  wss_proc_exit_stage2 */
/*! \brief The stage 2 exit hook function
 *
 *  It removes and reclaims memory used by the wss_proc structure associated
 *  with the exiting process.
 *
 *  \param pid PID of the exiting process
 */
void EXPORT
wss_proc_exit_stage2( pid_t pid )
{
   wss_proc_t *wrp;

   WSS_MUTEX_ENTER( wss_proc_list );
   wrp = wss_pfind_unsafe( pid );
   if( wrp )
   {
      int insyscall = wsi_proc_insyscall( wrp );

#ifdef DEBUG_WSS_PROC
      wsi_printf( PD_MAKESTR("wss_proc_exit: cleaning up pid %d\n"),
                  (int)pid );
#endif

      LIST_REMOVE( wrp, wrp_p_list );
      WSS_MUTEX_EXIT( wss_proc_list );

      /* We are unloading and the process is in at least one system call */
      if( wsi_unloading && insyscall )
      {
         WSS_MUTEX_ENTER( wss_proc_num_in_syscall );
         wss_proc_num_in_syscall--;
         WSS_CV_SIGNAL( wss_proc_num_in_syscall );
         WSS_MUTEX_EXIT( wss_proc_num_in_syscall );
      }

      /* Release reference from finding */
      wss_proc_rele( wrp );

      /* Release last reference */
      wss_proc_rele( wrp );

      return;
   }

   WSS_MUTEX_EXIT( wss_proc_list );
}


/*  wss_proc_free */
/*! \brief Reclaim memory used by the wss_proc structure associated with the
 *         exiting process.
 *
 *  \param wrp Pointer to the wrapper process structure for the exiting
 *             process
 */
void EXPORT
wss_proc_free( wss_proc_t *wrp )
{
   assert( wrp != NULL );
   assert( WSS_REF_COUNT(wrp->wrp_refcount) == 1 );

   WSS_REF_DEC(wrp->wrp_refcount);

#ifdef DEBUG_WSS_PROC
   wsi_printf( PD_MAKESTR("wss_proc_free: releasing memory for PID %d\n"),
               (int) wsi_proc_pid(wrp->wrp_p) );
#endif /* DEBUG_WSS_PROC */

   wss_ref_destroy( &(wrp->wrp_refcount) );

   if( wrp->wrp_path ) wss_free( (void*)wrp->wrp_path );
   if( wrp->wrp_oldpath ) wss_free( (void*)wrp->wrp_oldpath );
   if( wrp->wrp_progname ) wss_free( (void*)wrp->wrp_progname );
   if( wrp->wrp_cwd ) wss_free( (void*)wrp->wrp_cwd );

   if( wrp->wrp_arg_list )
      wss_util_rbtree_free( wrp->wrp_arg_list );

   if( wrp->wrp_wi_list )
      wss_proc_del_all_wrappers( wrp );

   wsi_proc_free( (void *)wrp );

   WSS_MUTEX_DESTROY( wrp->wrp );
   WSS_MUTEX_DESTROY( wrp->wrp_wi_list );
   WSS_MUTEX_DESTROY( wrp->wrp_arg_list );
   WSS_MUTEX_DESTROY( wrp->wrp_viewer );
   WSI_RW_DESTROY( wrp->wrp_cwd );

   wss_free( (void *)wrp );
}


/*  wss_proc_exit */
/*! \brief A convenience function used by operating systems which do not need
 *         to call the stage 1 and stage 2 exit functions separately.
 *
 *  \param p Pointer to the process structure for the exiting process
 */
void EXPORT
wss_proc_exit( proc_t *p )
{
   wss_proc_exit_stage1( p );
   wss_proc_exit_stage2( wsi_proc_pid(p) );
}


/*  wss_proc_new */
/*! \brief Creates a new wrapper process information structure based upon
 *         passed-in process structure
 *
 *  \param p Pointer to the process structure for the exiting process
 *
 *  \return Pointer to wrapper process structure on success,
 *  \return NULL on failure
 */
wss_proc_t EXPORT *
wss_proc_new( proc_t *p )
{
   wss_proc_t *wrp;

   wrp = (wss_proc_t *)wss_malloc( sizeof( *wrp ) );
   if( wrp )
   {
      bzero( wrp, sizeof( *wrp ) );

      WSS_MUTEX_INIT( wrp->wrp );
      WSS_MUTEX_INIT( wrp->wrp_wi_list );
      WSS_MUTEX_INIT( wrp->wrp_arg_list );
      WSS_MUTEX_INIT( wrp->wrp_viewer );
      WSI_RW_INIT( wrp->wrp_cwd );

      wss_ref_init( &(wrp->wrp_refcount) );
      wrp->wrp_pid = wsi_proc_pid(p);
      wrp->wrp_p = p;
      wsi_proc_new( wrp );
      wrp->wrp_cwd = wss_path_cwd( wrp );

      /* Reference the structure, since we're returning a pointer to it */
      wss_proc_ref( wrp );

      WSS_MUTEX_ENTER( wss_proc_list );
      LIST_INSERT_HEAD( &wss_proc_list, wrp, wrp_p_list );
      WSS_MUTEX_EXIT( wss_proc_list );
   }

   return( wrp );
}


/*  wss_proc_fork */
/*! \brief Callback function to handle process forking
 *
 *  This function is added to the kernel fork function list at WSS load
 *  time.  It will be called during a new process creation.  This is the
 *  function that performs the work for the fork/rfork/vfork/fork1/etc.
 *  system calls.
 *
 *  TBD: This description and name is Unix-centric.  Maybe we should rename it
 *       and have an appropriate description.  It is used on platforms other
 *       than Unix.
 *
 *  \param parent Pointer to the process structure for the parent process
 *  \param child  Pointer to the process structure for the child process
 *  \param flags  Forking flags
 */
void EXPORT
wss_proc_fork( proc_t *parent, proc_t *child, int flags )
{
   wss_proc_t             *wrp;
   wss_proc_t             *parent_wrp;
   wss_wrapper_instance_t *wi;

   if( !child )
   {
      wsi_printf( PD_MAKESTR("wss_proc_fork: null child proc structure!\n") );
      return;
   }

#ifdef DEBUG_WSS_PROC
   wsi_printf( PD_MAKESTR("wss_proc_fork: parent %d, child %d\n"),
               (int)wsi_proc_pid(parent), (int)wsi_proc_pid(child) );
#endif

   wrp = wss_proc_new( child );

   /* Find the parent of the forked process. */
   if( !( parent_wrp = wss_pfind( wsi_proc_pid(parent ) ) ) )
   {
      wss_proc_rele( wrp );
      return;
   }

   WSS_MUTEX_ENTER(wrp->wrp);
   wrp->wrp_nsyscall = wsi_proc_insyscall( parent_wrp );
   WSS_MUTEX_EXIT(wrp->wrp);

   /* Add a "progname" to the wrapper... execve will clear it. */
   if( parent_wrp->wrp_progname )
      wrp->wrp_progname = wss_strdup( parent_wrp->wrp_progname );
   if( parent_wrp->wrp_path )
      wrp->wrp_path = wss_strdup( parent_wrp->wrp_path );

   wsi_proc_fork( parent_wrp, wrp );

   /* Now duplicate all of the parent's wrappers for the child. */
        
   WSS_MUTEX_ENTER(parent_wrp->wrp_wi_list);
   for( wi = (wss_wrapper_instance_t *)(parent_wrp->wrp_wi_list);
        wi;
        wi = wi->wi_next )
   {
      wss_wrapper_instance_t *new_wi;

      wss_proc_add_wrapper( wrp, wi->wi_wrapper->w_module->wrm_linked->linker_file_name, &new_wi );

      /* Give the child access to the parent's WIV data.  The
       * wrapper's "duplicate" function will clone the data.
       *
       * There is a potential race condition here where we're 'borrowing'
       * the pointer to the parent's instance data and unlocking the
       * parent's instance list before actually using the pointer to
       * duplicate the data in wss_wrapper_call_duplication.  However,
       * we are not affected by this, as any threads in the parent should
       * be suspended during the part of fork that we get called.
       * (Or at least, this is the case on Solaris.)
       */
      new_wi->wi_data = wi->wi_data;

      wss_wrapper_instance_rele( new_wi );
   }
   WSS_MUTEX_EXIT(parent_wrp->wrp_wi_list);

   wss_wrapper_call_duplication( wrp );

   wss_proc_rele( parent_wrp );
   wss_proc_rele( wrp );
}


/*  wss_proc_add_wrapper_blind */
/*! \brief Adds the named wrapper to the active list for this process
 *
 *  This function is used for export to wrappers.
 *
 *  \param wrp  Pointer to the wrapper process structure for this process
 *  \param name Pointer to a string containing the name of the wrapper
 *              to add to active list
 *
 *  \return 0 on success,
 *  \return \b WRE_NOTFOUND if the named wrapper is not loaded,
 *  \return \b WRE_EXISTS if the wrapper is already on the active list
 *  \internal
 */
int EXPORT
wss_proc_add_wrapper_blind( wss_proc_t *wrp, const pd_char_t *name )
{
   return wss_proc_add_wrapper( wrp, name, NULL );
}

/*  wss_proc_add_wrapper */
/*! \brief Adds the named wrapper to the active list for this process
 *
 *  \param wrp    Pointer to the wrapper process structure for this process
 *  \param name   Pointer to a string containing the name of the wrapper
 *                to add to active list
 *  \param ret_wi If non-NULL, pointer to wrapper instance structure
 *                created by adding the named wrapper.  This instance will
 *                be referenced and it is left to the caller to release
 *                when no longer needed.
 *
 *  \return 0 on success,
 *  \return \b WRE_NOTFOUND if the named wrapper is not loaded,
 *  \return \b WRE_EXISTS if the wrapper is already on the active list
 */
int EXPORT
wss_proc_add_wrapper( wss_proc_t *wrp, const pd_char_t *name,
                      wss_wrapper_instance_t **ret_wi )
{
   wss_wrapper_t          *w;
   wss_wrapper_instance_t *wi, *prev_wi;
#ifdef WR_LINUX
   wss_proc_t             *init_wrp;
   wss_wrapper_instance_t *init_wi;
#endif

#ifdef USER_MODE_TEST
   wsi_printf( PD_MAKESTR("wss_proc_add_wrapper(%s)\n"), name );
#endif

   if( !( w = wss_wrapper_find( name ) ) )
      return WRE_NOTFOUND;
   
   /*
    * Check if this wrapper is already in the wrapper instance list
    * for this wss_proc structure.
    */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( prev_wi = NULL, wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
        wi && ( wi->wi_wrapper->w_module->wrm_id < w->w_module->wrm_id );
        prev_wi = wi, wi = wi->wi_next )
   ;
   if( wi )
      wss_wrapper_instance_ref( wi );
   if( prev_wi )
      wss_wrapper_instance_ref( prev_wi );
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   if( wi && ( wi->wi_wrapper->w_module->wrm_id == w->w_module->wrm_id ) )
   {
      wss_wrapper_instance_rele( wi );
      if( prev_wi )
         wss_wrapper_instance_rele( prev_wi );
      wss_wrapper_rele( w );
      return WRE_EXIST;
   }

   if( wi )
      wss_wrapper_instance_rele( wi );

   wi      = (wss_wrapper_instance_t *)wss_malloc( sizeof( *wi ) );
   if( wi == NULL )
   {
      wss_wrapper_rele( w );
      if( prev_wi )
         wss_wrapper_instance_rele( prev_wi );
      return WRE_NOMEM;
   }

   WSS_MUTEX_INIT( wi->wi );
   wss_ref_init( &(wi->wi_refcount) );
   wi->wi_data = NULL;
   wi->wi_active = 0;
   /* We have a reference to the wrapper, so we don't wss_wrapper_rele it */
   wi->wi_wrapper = w;

   WSS_MUTEX_ENTER( w->w );
   w->w_procs++;
   WSS_MUTEX_EXIT( w->w );

   /*
    * Add to the list so as to be in ascending order
    */
   if( prev_wi == NULL )
   {
      wi->wi_prev = NULL;
      wi->wi_next = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
      wrp->wrp_wi_list = wi;
      if( wi->wi_next != NULL )
         wi->wi_next->wi_prev = wi;
   }
   else
   {
      wi->wi_prev = prev_wi;
      wi->wi_next = prev_wi->wi_next;
      prev_wi->wi_next = wi;
      if( wi->wi_next != NULL )
         wi->wi_next->wi_prev = wi;
      wss_wrapper_instance_rele( prev_wi );
   }

   if( ret_wi != NULL )
   {
      /* If the caller wants a pointer to the newly-created
         wrapper instance, reference and return it */
      wss_wrapper_instance_ref( wi );
      *ret_wi = wi;
   }

#ifdef WR_LINUX
   /* Hack!
    * For Async events, we use the wrp structure that coresponds with
    * pid==1 (init).  We can't wrap init since we can not be loaded by
    * the time init is exec'd, so this should not remove any functionality.
    *
    * To implement this, All activated wrappers get an instance added to
    * init.  To avoid extra overhead, there is a shortcut in the wsi_dispatch
    * method for init (all pid < 10 actually)... So the only way those
    * intercepts can be called is by a wss_dispatch(... ,pfind(0)...)
    */
   init_wrp = wss_pfind(0);

   if (!init_wrp) {
       struct task_struct *inits_task = current;
       while (inits_task->p_pptr && inits_task->p_pptr != inits_task)
           inits_task = inits_task->p_pptr;
       init_wrp = wss_proc_new( inits_task );
       
       /* Fake a program name. */
       init_wrp->wrp_progname = wss_strdup("init");

       init_wrp->wrp_path = wss_path_cwd(init_wrp);
       if (!init_wrp->wrp_cwd)
           init_wrp->wrp_cwd = wss_strdup(PD_MAKESTR("/"));
   }
 
   WSS_MUTEX_ENTER( init_wrp->wrp_wi_list );
   for( prev_wi = NULL, wi = (wss_wrapper_instance_t *)(init_wrp->wrp_wi_list);
        wi && ( wi->wi_wrapper->w_module->wrm_id < w->w_module->wrm_id );
        prev_wi = wi, wi = wi->wi_next )
   ;

   if( wi ) {/* Just one per customer */
       WSS_MUTEX_EXIT( init_wrp->wrp_wi_list );
       goto linux_end;
   }

   if( prev_wi )
       wss_wrapper_instance_ref( prev_wi );
   WSS_MUTEX_EXIT( init_wrp->wrp_wi_list );

   init_wi = (wss_wrapper_instance_t *)wss_malloc( sizeof( *wi ) );
   if (!init_wi)
       goto linux_end; /* Valient try.  Hope they dont need it. */


   WSS_MUTEX_INIT( init_wi->wi );
   wss_ref_init( &(init_wi->wi_refcount) );
   init_wi->wi_data = NULL;
   init_wi->wi_active = 0;
   /* Remember to reference the wrapper */
   init_wi->wi_wrapper = w;
   wss_wrapper_ref( w );

   if( prev_wi == NULL ) {
      init_wi->wi_prev = NULL;
      init_wi->wi_next = (wss_wrapper_instance_t *)(init_wrp->wrp_wi_list);
      init_wrp->wrp_wi_list = init_wi;
      if( init_wi->wi_next != NULL )
         init_wi->wi_next->wi_prev = init_wi;
   } else {
      init_wi->wi_prev = prev_wi;
      init_wi->wi_next = prev_wi->wi_next;
      prev_wi->wi_next = init_wi;
      if( init_wi->wi_next != NULL )
         init_wi->wi_next->wi_prev = wi;
      wss_wrapper_instance_rele( prev_wi );
   }


linux_end:
   wss_proc_rele(init_wrp);
#endif

   return 0;
}


/*  wss_proc_del_wrapper */
/*! \brief Removes the named wrapper from the list of wrapper instances
 *         allociated with the given process
 *
 *  \param wrp    Pointer to the wrapper process structure for this process
 *  \param name   Pointer to a string containing the name of the wrapper
 *                to remove from the active list
 *
 *  \return 0 on success,
 *  \return WRE_NOTFOUND if the named wrapper is not loaded,
 *  \return WRE_ALREADY if the wrapper instance is not active,
 *  \return WRE_ERROR if an error occurs during deactivate
 */
int EXPORT
wss_proc_del_wrapper( wss_proc_t *wrp, pd_char_t *name )
{
   wss_wrapper_t          *w;
   wss_wrapper_instance_t *wi;

#ifdef USER_MODE_TEST
   wsi_printf( PD_MAKESTR("wss_proc_del_wrapper(%s)\n"), name );
#endif

   if( !( w = wss_wrapper_find( name ) ) )
      return WRE_NOTFOUND;

   /*
    * Check if this wrapper is already on the wss_proc wrapper list.
    */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( wi = NULL, wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
        wi && ( wi->wi_wrapper->w_module->wrm_id < w->w_module->wrm_id );
        wi = wi->wi_next )
   ;
   if( wi )
      wss_wrapper_instance_ref( wi );
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   if( !wi || ( wi->wi_wrapper->w_module->wrm_id != w->w_module->wrm_id ) )
   {
      if( wi )
         wss_wrapper_instance_rele( wi );
      wss_wrapper_rele( w );
      return WRE_NOTFOUND;
   }

   /* TBD: Hmm.. what does this have to do with deleting the wrapper? */
   if( !wi->wi_active )
   {
      wss_wrapper_instance_rele( wi );
      wss_wrapper_rele( w );
      return WRE_ALREADY;
   }

   if( wi->wi_wrapper != w )
   {
      wss_wrapper_instance_rele( wi );
      wss_wrapper_rele( w );
      return WRE_ERROR;
   }

   if( w->w_deactivate )
   {
       wss_wrapper_data_t wrd;
       int err;

       wrd.wiv = wi->wi_data;
       err = (*w->w_deactivate)(wrp, &wrd);
       wi->wi_data = wrd.wiv;

      if( err != WR_SUCCESS ) return WRE_ERROR;
   }

   WSS_MUTEX_ENTER( wi->wi_wrapper->w );
   wi->wi_wrapper->w_procs--;
   WSS_MUTEX_EXIT( wi->wi_wrapper->w );
   wi->wi_active = 0;

   /* Isolate wi */
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   wss_wrapper_instance_rele( wi );
   if( wi->wi_prev )
      wi->wi_prev->wi_next = wi->wi_next;
   else
      wrp->wrp_wi_list = wi->wi_next;

   if( wi->wi_next )
      wi->wi_next->wi_prev = wi->wi_prev;
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   /* Release reference to it (from being in the list) */
   wss_wrapper_instance_rele( wi );

   return 0;
}


/*  wss_proc_del_all_wrappers */
/*! \brief Deletes all wrapper instances from the specified process.
 *
 *  \param wrp Pointer to the wrapper process structure
 */
void EXPORT
wss_proc_del_all_wrappers( wss_proc_t *wrp )
{
   while( wrp->wrp_wi_list )
   {
      WSS_MUTEX_ENTER( wrp->wrp_wi_list );
      if( wrp->wrp_wi_list )
      {
         wss_wrapper_instance_t *wi =
            (wss_wrapper_instance_t *)(wrp->wrp_wi_list);

         WSS_MUTEX_ENTER( wi->wi_wrapper->w );
         wi->wi_wrapper->w_procs--;
         WSS_MUTEX_EXIT( wi->wi_wrapper->w );
         wi->wi_active = 0;

         /* Isolate wi */
         wrp->wrp_wi_list = wi->wi_next;
         if( wi->wi_next )
            wi->wi_next->wi_prev = wi->wi_prev;
         WSS_MUTEX_EXIT( wrp->wrp_wi_list );

         /* Release reference to it (from being in the list) */
         wss_wrapper_instance_rele( wi );
      }
      else
         WSS_MUTEX_EXIT( wrp->wrp_wi_list );
   }
}

pd_char_t EXPORT *
wss_proc_get_cwd( pid_t pid )
{
   wss_proc_t *wrp = wss_pfind( pid );
   if( wrp && wrp->wrp_cwd )
      return wss_strdup( wrp->wrp_cwd );

   return NULL;
}

pd_char_t EXPORT *
wss_proc_get_path( pid_t pid )
{
   wss_proc_t *wrp = wss_pfind( pid );
   if( wrp && wrp->wrp_path )
      return wss_strdup( wrp->wrp_path );

   return NULL;
}

