/*! \file wss_strings.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_strings.c,v 1.20 2001/02/23 19:02:46 stevek Exp $
 *
 * \brief   This file contains functions exported from the WSS to wrappers
 *          to be used for manipulating strings.
 * \author  Chris Vance, Ben Uecker, Mark Feldman
 * \version $Revision: 1.20 $
 * \date    $Date: 2001/02/23 19:02:46 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>
#include <wsi_types.h>

#include <wss_malloc.h>
#include <wss_strings.h>


/***************************************************************************\
 * Private structures                                                      *
\***************************************************************************/

/* Regular expression handling */
struct rex_node
{
   pd_char_t      *elem;
   struct rex_node *next;
};


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

/*
 * append_to_list
 */
static void
append_to_list( pd_char_t *word, struct rex_node **rex_list,
                struct rex_node **rex_tail )
{
   struct rex_node *tmp;
   tmp = (struct rex_node *)wss_malloc( sizeof( *tmp ) );

   tmp->elem = wss_strdup( word );
   tmp->next = NULL;

   if( *rex_tail )
      (*rex_tail)->next = tmp;

   *rex_tail = tmp;

   if( !*rex_list )
      *rex_list = tmp;
}

/*
 * reverse_list
 */
static void
reverse_list( struct rex_node **rex_list, struct rex_node **rex_tail )
{
   struct rex_node *t;
   t = *rex_list;

   /* 0 or 1 node */
   if( *rex_list == *rex_tail ) return;

   *rex_tail = NULL;
   t = *rex_list;
   while( *rex_list )
   {
      *rex_list = (*rex_list)->next;
      t->next = *rex_tail;
      *rex_tail = t;
      t = *rex_list;
   }

   *rex_list = *rex_tail;
   while( (*rex_tail)->next )
      *rex_tail = (*rex_tail)->next;
}

/*
 * init_list
 */
static void
init_list( struct rex_node **rex_list, struct rex_node **rex_tail,
           int *LEFT_ANCHOR, int *RIGHT_ANCHOR )
{
   *rex_list = NULL;
   *rex_tail = NULL;
   *LEFT_ANCHOR = *RIGHT_ANCHOR = 0;
}

/*
 * clear_list
 */
static void
clear_list( struct rex_node **rex_list, struct rex_node **rex_tail,
            int *LEFT_ANCHOR, int *RIGHT_ANCHOR )
{
   struct rex_node *t1, *t2;
   t1 = *rex_list;
   while( t1 )
   {
      t2 = t1;
      t1 = t1->next;
      wss_free( t2->elem );
      wss_free( t2 );
   }
   *rex_list = NULL;
   *rex_tail = NULL;

   *LEFT_ANCHOR = *RIGHT_ANCHOR = 0;
}

/*
 * parser
 */
static void
parser( pd_char_t *str, struct rex_node **rex_list, struct rex_node **rex_tail,
        int *LEFT_ANCHOR, int *RIGHT_ANCHOR )
{
   pd_char_t c;
   pd_char_t *buf, *p, *e;
   int len, cnt;

   len = wsi_strlen( str );

   if( ( c = *str ) == PD_MAKESTR('^') )
   {
      *LEFT_ANCHOR = 1;
   }
   if( *( e = ( str + len - 1 ) ) == PD_MAKESTR('$') )
   {
      *RIGHT_ANCHOR = 1;
      len--;
      *e = PD_MAKESTR('\0');
   }
   /* Don't want to advance str until after check for '$' */
   if( *LEFT_ANCHOR )
      c = *str++;

   for( ; ( c = *str++ ) != PD_MAKESTR('\0'); )
   {
      if( c == PD_MAKESTR('*') )
      {
         if( *str == PD_MAKESTR('*') )
         {
            append_to_list( PD_MAKESTR("**"), rex_list, rex_tail );
            str++;
         }
         else
            append_to_list( PD_MAKESTR("*"), rex_list, rex_tail );
      }
      else
      {
         p = str - 1;
         cnt = 0;
         while( ( c = *str++ ) != PD_MAKESTR('*') && c != PD_MAKESTR('\0') )
            cnt++;
         buf = (pd_char_t*)wss_malloc( (cnt + 2 )* sizeof(pd_char_t) );
         wss_strncpy( buf, p, cnt + 1 );
         *(buf + cnt + 1) = PD_MAKESTR('\0');
         append_to_list( buf, rex_list, rex_tail );
         wss_free( buf );

         if( c == PD_MAKESTR('\0') ) break;
         if( c == PD_MAKESTR('*') ) str--;
      }
   }
}


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_strchr */
/*! \brief strchr equivalent exported to wrappers
 *
 *  \param p  Pointer to string to search for character
 *  \param ch Character to search for
 *
 *  \return Pointer within string \a p where \a ch occurs,
 *  \return NULL if character is not found
 */
pd_char_t EXPORT *
wss_strchr( const pd_char_t *p, int ch )
{
   for( ;; ++p )
   {
      if( *p == ch )
         return( (pd_char_t *)p );
      if( !*p )
         return( (pd_char_t *)NULL );
   }
   /* NOTREACHED */
}

/*  wss_strrchr */
/*! \brief strrchr equivalent exported to wrappers
 * 
 *  \param p  Pointer to string to search for character
 *  \param ch Character to search for
 *
 *  \return Pointer within string \a p where last occurance of \a ch occurs,
 *  \return NULL if character is not found
 */
pd_char_t EXPORT *
wss_strrchr( const pd_char_t *p, int ch )
{
   pd_char_t *save;

   for( save = NULL;; ++p )
   {
      if( *p == ch )
         save = (pd_char_t *)p;
      if( !*p )
         return( save );
   }
   /* NOTREACHED */
}

/*  wss_strstr */
/*! \brief strstr equivalent exported to wrappers
 *
 *  \param s    Pointer to string to search within
 *  \param find Pointer to string to find
 *
 *  \return Pointer to first occurance of \a find in \a s,
 *  \return Pointer to \a s if \a find is NULL,
 *  \return NULL if \a find is not found within \a s
 */
pd_char_t EXPORT *
wss_strstr( const pd_char_t *s, const pd_char_t *find )
{
   register pd_char_t c, sc;
   register size_t len;

   if( ( c = *find++ ) != 0 )
   {
      len = wsi_strlen( find );
      do
      {
         do
         {
            if( ( sc = *s++ ) == 0 )
               return( NULL );
         }
         while( sc != c );
      }
      while( wsi_strncmp( s, find, len ) != 0 );
         s--;
   }
   return( (pd_char_t *)s );
}

/*  wss_strncpy */
/*! \brief Copy src to dst, truncating or null-padding to always copy n bytes
 *
 *  \param dst Pointer to buffer to copy into
 *  \param src Pointer to string to copy from
 *  \param n   Maximum number of characters to copy
 *
 *  \return Pointer to destination buffer
 */
pd_char_t EXPORT *
wss_strncpy( pd_char_t *dst, const pd_char_t *src, size_t n )
{
   if( n != 0 )
   {
      register pd_char_t *d = dst;
      register const pd_char_t *s = src;

      do
      {
         if( (*d++ = *s++) == 0 )
         {
            /* NUL pad the remaining n-1 bytes */
            while( --n != 0 )
               *d++ = 0;
            break;
         }
      }
      while( --n != 0 );
   }
   return dst;
}


/*  wss_match */
/*! \brief Regular-expression matching function
 *
 *  \param src Pointer to the source string
 *  \param pat Pointer to the pattern string
 *  \param beg
 */
int EXPORT
wss_match( pd_char_t *src, pd_char_t *pat, pd_char_t **beg, pd_char_t **end )
{
   struct rex_node *n;
   pd_char_t *lstart, *srcend;
   pd_char_t *Gstart = NULL;
   pd_char_t c;
   pd_char_t *source = src;
   int len = wsi_strlen( src );

   struct rex_node *rex_list;
   struct rex_node *rex_tail;
   int LEFT_ANCHOR = 0;
   int RIGHT_ANCHOR = 0;
        
   /* Ugly hack to allow "NULL" to be passed in to beg&end.  I'm
    * just doing this because its gonna be a PITA to add the variables
    * to activation criteria so they exist to be passed in.
    * Its not like any generated wrappers actually use the returned
    * values anyway.
    */
   pd_char_t *dummy_beg, *dummy_end;
   if( beg == NULL )
      beg = &dummy_beg;
   if( end == NULL )
      end = &dummy_end;

   srcend = src + len;
   init_list( &rex_list, &rex_tail, &LEFT_ANCHOR, &RIGHT_ANCHOR );
   parser( pat, &rex_list, &rex_tail, &LEFT_ANCHOR, &RIGHT_ANCHOR );

   for( n = rex_list; n != 0; n = n->next )
   {
      if( wsi_strcmp( n->elem, PD_MAKESTR("*") ) == 0 )
      {
         /*
          * if this * is the first element, it matches start
          * of src
          */
         if( !Gstart ) Gstart = src;

         /*
          * if this * is the last element, it matches the rest
          * of src
          */
         if( !n->next )
         {
            *end = srcend;
            /* kludge to avoid altering *end after loop */
            src = srcend; 
         }
         continue;
      }

      if( wsi_strcmp( n->elem, PD_MAKESTR("**") ) == 0 )
      {
         if( !Gstart ) Gstart = src;
         c = *src++;
         while( ( c != PD_MAKESTR('/') ) && ( c != PD_MAKESTR('\0') ) )
            c = *src++;
         if( c == PD_MAKESTR('/') )
         {  
            src--;
            continue;
         }
         if( c == PD_MAKESTR('\0') )
            *end = srcend;
      }
      else
      {
         /* matching a normal string */
         lstart = wss_strstr( src, n->elem );
         if( !lstart )
         {
            clear_list( &rex_list, &rex_tail, &LEFT_ANCHOR, &RIGHT_ANCHOR );
            return 0;
         }
         if( !Gstart )
            Gstart = lstart;
         src = lstart + wsi_strlen( n->elem );
      }
   }
   if( src <= srcend )
      *end = --src;

   *beg = Gstart;

   /* Check for left anchor - simple since we find the first match
    * from the left anyway.
    */
   if( LEFT_ANCHOR && ( *beg != source ) )
   {
      clear_list( &rex_list, &rex_tail, &LEFT_ANCHOR, &RIGHT_ANCHOR );
      return 0;
   }

   /* Check for right anchor. */
   if( RIGHT_ANCHOR )
   {
      pd_char_t *tptr;
      struct rex_node *sl;
      size_t tlen;

      /* Scan the src string again -- backwards */
      reverse_list( &rex_list, &rex_tail );
      tptr = srcend;
      for( sl = rex_list; sl != NULL; sl = sl->next )
      {
         tlen = wsi_strlen(sl->elem);
         if( wsi_strcmp( sl->elem, PD_MAKESTR("*") ) == 0 )
         { 
            *end = srcend;
            clear_list( &rex_list, &rex_tail, &LEFT_ANCHOR, &RIGHT_ANCHOR );
            return 1;
         }
         if( wsi_strcmp( sl->elem, PD_MAKESTR("**") ) == 0 )
         {
            tptr--;
            while( *tptr != PD_MAKESTR('/') && tptr >= source )
               tptr--;
            tptr++;
         }
         else
         {
            tptr = tptr - tlen;
            if( wsi_strncmp( sl->elem, tptr, tlen ) != 0 )
            {
               clear_list( &rex_list, &rex_tail, &LEFT_ANCHOR, &RIGHT_ANCHOR );
               return 0;
            }
         }
      }
   }

   /* Clean up */
   clear_list( &rex_list, &rex_tail, &LEFT_ANCHOR, &RIGHT_ANCHOR );
   return 1;
}

/*  wss_replace */
/*! \brief Substitute one string for a range of positions within a source
 *         string
 *
 *  \param src   Pointer to the source/changed string
 *  \param repl  Pointer to the replacement string
 *  \param start Starting character
 *  \param end   Ending character
 *
 *  \warning \a src points to the changed string. If needed, more space will
 *           be allocated for \a src.
 *
 *  \return TRUE on a sucessful replacement,
 *  \return FALSE on failure
 */
int EXPORT 
wss_replace( pd_char_t **src, pd_char_t *repl, pd_char_t *start, pd_char_t *end )
{
   u_int32_t rlen, slen, rest;
   pd_char_t *srcend;

   srcend = *src + wsi_strlen( *src );
   rest = ( ((u_int32_t)srcend - (u_int32_t)end)/ sizeof(pd_char_t)) - 1 ;

   rlen = wsi_strlen( repl );
   slen = ( ((u_int32_t)end - (u_int32_t)start)/ sizeof(pd_char_t)) + 1 ;
   
   /* Three cases: rlen = slen
    *    rlen > slen
    *    rlen < slen
    */

   if( rlen == slen )
   {
      wss_strncpy( start,repl,rlen );
      return TRUE;
   }

   if(rlen < slen)
   {
      wss_strncpy( start,repl,rlen );
      if ( rest>0 )
        wss_strncpy( ( start + rlen ), ( end + 1 ), rest );
      *( start + rlen + rest ) = PD_MAKESTR('\0');
      return TRUE;
   }

   if( rlen > slen )
   {
      pd_char_t *buf = (pd_char_t*)wss_malloc_string( wsi_strlen(*src) + rlen - slen + 1 );
      pd_char_t *bufend;
      int sz;

      /* copy up to `start'...*/
      sz = ((u_int32_t)start - (u_int32_t)*src)/sizeof(pd_char_t);
	  if( sz > 0 )
         wss_strncpy( buf, *src, sz );
      bufend = buf + sz;

      /* add substitution string...*/
      wss_strncpy( bufend, repl, rlen );
      bufend = bufend + rlen;

      /* tack on the rest */

      if ( rest>0 )
        wss_strncpy( bufend, end + 1, rest );

      *( bufend + rest ) = PD_MAKESTR('\0');

      wss_free( *src );
      *src = buf;
      return TRUE;
   }

   return FALSE;
}

/*  wss_substitute */
/*! \brief Substitutes one string for all occurrences of another string within
 *         the source string
 *
 *  \param src Pointer to the source/changed string
 *  \param txt Pointer to string occurrence to replace
 *  \param sub Pointer to string to substitute in place of \a txt
 * 
 *  \return Number of substitutions made
 */
int EXPORT
wss_substitute( pd_char_t **src, pd_char_t *txt, pd_char_t *sub )
{
   int ret = 0;
   int sublen, dist;
   pd_char_t *s, *beg, *end, *srcend;
   
   s = *src;
   sublen = wsi_strlen( sub );
   srcend = *src + wsi_strlen( *src );

   while( s < srcend )
   {
      if( wss_match( s, txt, &beg, &end ) )
      {
         /* distance from string start to beginning of match */
         dist = beg - *src;
         wss_replace( src, sub, beg, end );
         s = *src + dist + sublen;
         srcend = *src + wsi_strlen( *src );
         ret++;
      }
      else
         break;
   }

   return ret;
}

