/*! \file wss_utilities.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_utilities.c,v 1.14 2001/02/23 19:02:46 stevek Exp $
 * 
 * \brief   Collection of tools usefull in muliple parts of the wrappers
 *          system
 * \author  Douglas Kilpatrick
 * \version $Revision: 1.14 $
 * \date    $Date: 2001/02/23 19:02:46 $
 * 
 * Contains rbtree's.  These are used by the solaris thread_id -> arg
 * replacement structure and the malloc_debug stuff.
 * 
 * \warning These are \b NOT thread-safe, you must perform locking outside
 *          of the RB tree in order to play nice in a threaded environment.
 *          
 * Contains string-to-int and string-to-float functions.
 */

#ifdef USER_MODE_TEST
#include <assert.h>
#endif /* USER_MODE_TEST */

#include <wr_conf.h>
#include <wsi_stdlib.h>

#include <wss_utilities.h>

#ifdef USER_MODE_TEST
#define wss_malloc(s) malloc(s)
#define wss_free(p) free(p)
#ifdef MALLOC_DEBUG
pd_char_t *wss_strdup_debug __P((const pd_char_t *,pd_char_t *,int));
#define wss_strdup(s) wss_strdup_debug(s,__FILE__,__LINE__)
#else
pd_char_t *wss_strdup __P((const pd_char_t *s));
#endif /* MALLOC_DEBUG */
#else
#include <wss_malloc.h>
#endif

#ifndef WR_SUPPRESS_RBTREE

/*  wss_util_rbnode_init */
/*! \brief Initializes the passed in RBNode
 * 
 *  \param node         The RBNode to initialize.
 *  \return             The initialized RBNode
 */
wss_util_rbnode_t *wss_util_rbnode_init(wss_util_rbnode_t *node)
{
    node->link[WSS_UTIL_LEFT] = node->link[WSS_UTIL_RIGHT] = 0;
    node->red = 0;
    return node;
}

/*  wss_util_rbtree_new */
/*! \brief Creates an RBTree object from the specified arguments.
 * 
 *  \param cf           A function to compare two RBNodes.
 *  \param clone        A function to clone or create an RBNode.
 *  \param free         A function to free created RBNodes.
 *  \return             The created and initialized RBTree
 */
wss_util_rbtree_t *
wss_util_rbtree_new(
        wss_util_CompFunc cf,
        wss_util_CloneFunc clone,
        wss_util_FreeFunc free)
{
    wss_util_rbtree_t *retval = (wss_util_rbtree_t*)wss_malloc(sizeof(*retval));
    retval->Head = clone(0);
    retval->Compare = cf;
    retval->Clone = clone;
    retval->Free = free;
    
    return retval;
}

/*  wss_util_rbtree_find */
/*! \brief Finds a node equal to the given node in the given tree.
 * 
 *  \param tree         The RBTree to search.
 *  \param node         An node that is equivilent to the target node.
 *  \return             The target node on success.
 *  \return             NULL on failure
 */
wss_util_rbnode_t *
wss_util_rbtree_find(wss_util_rbtree_t *tree, wss_util_rbnode_t *node)
{
    wss_util_rbnode_t *retval;
    int dir;
    
    retval = tree->Head->link[WSS_UTIL_RIGHT];
    while (retval) {
        dir = tree->Compare(node, retval);
        if (dir == 0 && wss_util_rbnode_isempty(retval))
            break;
        dir = dir < 0;
        retval = retval->link[dir];
    }
    return retval;
}

/*
 * Rotation fun.
 */

#define NO_FLIP  0
#define FLIP_GCH 1
#define FLIP_CH  2

/* Rotate child(ch) and grandchild(gch) of root (r) along the path
 * specified by searching for n.
 */
static wss_util_rbnode_t * rotate(
        wss_util_rbtree_t *tree,
        wss_util_rbnode_t *n,
        wss_util_rbnode_t *r,
        int flip_mode)
{
    wss_util_rbnode_t *ch, *gch;
    int ch_dir, gch_dir;
    
    /* find child and grandchild */
    ch_dir = tree->Compare(n,r) < 0;
    if (flip_mode & FLIP_CH)
        ch_dir = !ch_dir;
    if (r == tree->Head)
        ch_dir = WSS_UTIL_RIGHT;
    ch = r->link[ch_dir];
    
    gch_dir = tree->Compare(n, ch) < 0;
    if (flip_mode) {
        if (flip_mode == FLIP_GCH)
            gch_dir = !gch_dir;
        else
            gch_dir = flip_mode & 1;
    }
    gch = ch->link[gch_dir];
    
    /* Rotate: move pointers */
    ch->link[gch_dir] = gch->link[!gch_dir];
    gch->link[!gch_dir] = ch;
    r->link[ch_dir] = gch;
    
    return gch;
}

/*
 * Take care of colors and balance.  It will color the current location
 * red, the currect location's children black, and then look to see if
 * two consecutive red nodes have been created.  If so, a single or double
 * rotation will be done to fix the tree.
 */

static
void split(
        wss_util_rbtree_t *tree,
        wss_util_rbnode_t *node,         /* node being inserted  */
        wss_util_rbnode_t **curloc,      /* current location     */
        wss_util_rbnode_t **parent,      /* its parent           */
        wss_util_rbnode_t *gparent,      /* its grandparent      */
        wss_util_rbnode_t *ggparent)     /* its gread grandparent*/
{
    assert(!tree->Head->red);
    (*curloc)->red = 1;
    if ((*curloc)->link[WSS_UTIL_LEFT])
        (*curloc)->link[WSS_UTIL_LEFT]->red = 0;
    if ((*curloc)->link[WSS_UTIL_RIGHT])
        (*curloc)->link[WSS_UTIL_RIGHT]->red = 0;
    
    /* Check to see if we have two red's ina row. */
    if ((*parent)->red) {
        gparent->red = 1;
        /* If the red links don't point in the same direction, 
         * we will need a double rotation.  The lower half
         * around the gp and then the upper half around the ggp.
         */
        if ((tree->Compare(node,gparent) < 0 ) != 
                    ( tree->Compare(node, *parent) < 0) )
            *parent = rotate(tree, node, gparent, NO_FLIP);
        *curloc = rotate(tree, node, ggparent, NO_FLIP);
        (*curloc)->red = 0;
    }
    tree->Head->link[WSS_UTIL_RIGHT]->red = 0;
}

/*  wss_util_rbtree_delete */
/*! \brief Deletes an equivilent node from the tree
 * 
 *  This function does not free the memory used by the node, it merely
 *  removes the node from the tree.
 * 
 *  \param tree         The RBTree.
 *  \param node         A node equivilent to the one to be deleted.
 *  
 *  \return             The node that was removed on success.
 *  \return             0 on failure.
 */
wss_util_rbnode_t *
wss_util_rbtree_delete(wss_util_rbtree_t *tree, wss_util_rbnode_t *node)
{
    wss_util_rbnode_t *s, *p, *g;
    int dir, next_dir;
    
    g = 0; p = tree->Head; s = p->link[WSS_UTIL_RIGHT];
    dir = WSS_UTIL_RIGHT;
    
    /* First check on the root. */
    if (!s)
        return 0;
    
    /* Check to make sure the root isn't an only child */
    if (wss_util_rbnode_isempty(s)) {
        if (tree->Compare(node, s) == 0) {
            /* deleting the node */
            p->link[WSS_UTIL_RIGHT] = NULL;
            return s;
        } else
            return 0;
    }
    
    /* Either the root or one of its children must be red */
    if (!s->link[WSS_UTIL_LEFT]->red && !s->link[WSS_UTIL_RIGHT]->red)
        s->red = 1; /* color the root red. */
    
    /* March down the tree, flipping colors on the way.  That way when
     * we find the leaf, its parent will be read, and its will be easier
     * to delete.
     */
    while (1) {
        if (wss_util_rbnode_isempty(s))
            break;
        
        next_dir = tree->Compare(node, s) < 0;

        /* If the current node or the next node is red, we can advance. */
        if ( s->red || s->link[next_dir]->red) {
            /* Empty */
        } else if ( s->link[!next_dir]->red) {
            /* If the current node is black and the next node is black but
             * the next node's sib is red:
             * Rotate from parent towards the red child.  This will lower
             * the current node, and give us a new grandparent (the old parent)
             * and a new parent (the sib that was red.  We paint the current
             * node red and the new parent is painted black.
             */
            g = p;
            p = rotate(tree, s->link[next_dir], p, FLIP_GCH);
            s->red = 1;
            p->red = 0;
        } else {
            /*
             * If the current node is black and the left child is black
             * and the right child is black
             * (a) the current node's parents must be red.
             * (b) its sub must be black (two in a row)
             * (c) we need to color the current node red.  To make
             * this possible we color the current node red, the parent black
             * and then check for tree imbalances.  Two cases:
             */
            wss_util_rbnode_t *sib;
            sib = p->link[!dir];

            assert(p->red);
            assert(!sib->red);
            assert(!wss_util_rbnode_isempty(sib));
            
            s->red = 1; p->red = 0;
            
            /* First case, black sib has 2 black kids.  Color the sib red.
             * (reverse a simple color flip) */
            if (!sib->link[WSS_UTIL_LEFT]->red && !sib->link[WSS_UTIL_RIGHT]->red)
                sib->red = 1;
            else {
                /* Second case: black sib has at least one red kid.
                 * We need to rotate to re-balance
                 */
                int redkid_dir;
                
                if (sib->link[WSS_UTIL_LEFT]->red)
                    redkid_dir = WSS_UTIL_LEFT;
                else
                    redkid_dir = WSS_UTIL_RIGHT;
                
                if (!dir == redkid_dir) {
                    sib->red = 1;
                    sib->link[redkid_dir]->red = 0;
                    g = rotate(tree, node, g, FLIP_GCH);
                } else {
                    rotate(tree, node, p, FLIP_CH + redkid_dir);
                    g = rotate(tree, node, g, FLIP_GCH);
                }
            }
        }
        
        /* advance pointer */
        dir = next_dir;
        g = p;
        p = s;
        s = s->link[dir];
    }
    
    /* Make the root black */
    tree->Head->link[WSS_UTIL_RIGHT]->red = 0;
    
    /* Delete it if it is a match.  Parent is red. */
    
    if (tree->Compare(s, node) == 0) {
        assert( p->red || p == tree->Head->link[WSS_UTIL_RIGHT]);
        
        g->link[(tree->Compare(s,g) < 0)] = p->link[(tree->Compare(s,p) >= 0)];
        tree->Free(p); /* Internal node */
        return s;
    }
    return 0;
}

/*  wss_util_rbtree_insert */
/*! \brief Adds the specified node to the tree.
 * 
 *  \param tree         The RBTree.
 *  \param node         A node equivilent to the one to be deleted.
 *  
 *  \return             0 on success.
 *  \return             -1 on failure.
 */

int wss_util_rbtree_insert(wss_util_rbtree_t *tree, wss_util_rbnode_t *node)
{
    int p_dir;
    wss_util_rbnode_t *p, *s;
    wss_util_rbnode_t *g = NULL;
    wss_util_rbnode_t *gg = NULL;
    
    /* Search for a leaf */
    p = tree->Head;
    p_dir = WSS_UTIL_RIGHT; /* always thus */
    s = p->link[WSS_UTIL_RIGHT];
    
    if (s) {
        wss_util_rbnode_t *temp;
        int dir;
        
        while (s->link[WSS_UTIL_RIGHT]) {
            if (s->link[WSS_UTIL_LEFT]->red && s->link[WSS_UTIL_RIGHT]->red)
                split(tree, node, &s, &p, g, gg);
            gg = g;
            g = p;
            p = s;
            p_dir = tree->Compare(node, s) < 0;
            s = s->link[p_dir];
        }
        
        dir = tree->Compare(node, s);
        if (dir == 0)
            return -1; /* Duplicate */
        
        dir = dir < 0;
        temp = tree->Clone(dir ? s : node);
        temp->link[dir] = node;
        temp->link[!dir] = s;
        node = temp;
    }
    
    /* Add the new node */
    p->link[p_dir] = node;
    
    split(tree, node, &node, &p, g, gg);
    return 0;
}

static
void rWalk(wss_util_rbnode_t *node, int level, wss_util_DoFunc df)
{
    if (node) {
        rWalk(node->link[WSS_UTIL_LEFT], level + 1, df);
        if (wss_util_rbnode_isempty(node))
            df(node, level);
        rWalk(node->link[WSS_UTIL_RIGHT], level + 1, df);
    }
}

int wss_util_rbtree_walk (wss_util_rbtree_t *tree, wss_util_DoFunc df)
{
    if (tree->Head->link[WSS_UTIL_RIGHT]) {
        rWalk(tree->Head->link[WSS_UTIL_RIGHT], 0, df);
        return 0;
    }
    return -1;
}

static void pWalkAndFree(wss_util_rbtree_t *tree, wss_util_rbnode_t *node)
{
    if (node) {
        pWalkAndFree(tree, node->link[WSS_UTIL_LEFT]);
        node->link[WSS_UTIL_LEFT] = 0;
        pWalkAndFree(tree, node->link[WSS_UTIL_RIGHT]);
        node->link[WSS_UTIL_RIGHT] = 0;
        tree->Free(node);
    }
}

/*  wss_util_rbtree_free */
/*! \brief Frees all memory assocated with the RBTree
 * 
 *  \param tree         The RBTree.
 */
void wss_util_rbtree_free(wss_util_rbtree_t *tree)
{
    pWalkAndFree(tree, tree->Head);
    wss_free(tree);
}

#undef wss_util_rbnode_isempty

/*  wss_util_rbnode_isempty */
/*! \brief Determines if the specified rbnode is empty.
 * 
 *  \param node         The RBNode.
 *  
 *  \return 1 if the node has no children.
 *  \return 0 if it has children.
 */
int wss_util_rbnode_isempty(wss_util_rbnode_t *node)
{
    return node->link[WSS_UTIL_LEFT] == 0;
}

#endif

int wss_atoi(const pd_char_t *string)
{
    int retval = 0;
    int sgn = 0;

    if (!string)
        return 0;
    
    if ( *string == '-' ) {
        string++;
        sgn++;
    }
    
    for( ; *string && ( *string >= '0' ) && ( *string <= '9' ); string++ ) {
        retval *= 10;
        retval += *string- '0';
    }
    
    if (sgn)
        retval = -retval;

    return retval;
}

double
wss_atof( const pd_char_t *string )
{
    double      retval = 0.0;
    int         sgn = 0;
    int         dp = 0;

    if (!string)
        return 0.0;

    if( *string == '-' )
        string++, sgn++;

    for( ; *string && ( ( ( *string >= '0' ) && ( *string <= '9' ) ) ||
                ( *string == '.' ) ); string ++ )
    {
        if( *string != '.' ) {
            if( dp ) {
                dp *= 10;
                retval += (*string - '0') / dp;
            } else {
                retval *= 10;
                retval += *string - '0';
            }
        }
        else if( dp++ )
            break;
    }
    
    if (sgn)
        retval = -retval;

    return retval;
}

pd_char_t EXPORT *
wss_downcase( const pd_char_t *str )
{
   pd_char_t *newstr = NULL;

   if( str )
   {
      pd_char_t *ptr;

      for( ptr = newstr = wss_strdup( str );
           ptr && *ptr;
           ptr++ )
         *ptr = wsi_tolower(*ptr);
   }

   return newstr;
}

pd_char_t EXPORT *
wss_upcase( const pd_char_t *str )
{
   pd_char_t *newstr = NULL;

   if( str )
   {
      pd_char_t *ptr;

      for( ptr = newstr = wss_strdup( str );
           ptr && *ptr;
           ptr++ )
         *ptr = wsi_toupper(*ptr);
   }

   return newstr;
}

