/*! \file wss_vm.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_vm.c,v 1.55 2001/06/21 17:46:07 stevek Exp $
 *
 * \brief   Functions for managing VM used by the WSS
 * \author  Lee Badger, Chris Vance, Steve Kiernan
 * \version $Revision: 1.55 $
 * \date    $Date: 2001/06/21 17:46:07 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <pdpi/queue.h>

#include <wss_debug.h>
#include <wss_db.h>
#include <wss_file.h>
#include <wss_malloc.h>
#include <wss_mutex.h>
#include <wss_vm.h>

#include <eventfs.h>

#include <wrerrno.h>


/***************************************************************************\
 * Private structures and types                                            *
\***************************************************************************/

/*
 * The following mem malloc/free structures adapted from "The C
 * Programming Language" by Kernighan and Ritchie, chapter 8.
 */
typedef double ALIGN;
typedef union header
{
   struct
   {
      union header *next;  /* next free block */
      unsigned      size;  /* size of this free block, in units */
   }
   s;
   ALIGN            x;     /* force alignment of blocks */
}
HEADER;

PD_LIST_HEAD(wss_vm_spaces_list, wss_vmspace);


/***************************************************************************\
 * Private globals                                                         *
\***************************************************************************/

static struct wss_vm_spaces_list wss_vm_spaces;
WSS_MUTEX_PROTECT(wss_vm_spaces);


/***************************************************************************\
 * Forward declarations                                                    *
\***************************************************************************/

void      *_wss_vm_map_file      __P(( const pd_char_t *filename,
                                       size_t size_in_bytes,
                                       wss_file_t *pfile, pd_bool_t *isnew ));
#ifdef DEBUG_VM
static int _wss_vm_aligned_ptr   __P(( wss_vmspace_t *space, wss_vmaddr_t rel,
                                       pd_char_t *file, int line ));
static int _wss_vm_valid_ptr     __P(( wss_vmspace_t *space, wss_vmaddr_t rel,
                                       pd_char_t *file, int line ));
static int _wss_vm_valid_abs_ptr __P(( wss_vmspace_t *space, void *abs,
                                       pd_char_t *file, int line ));
#endif /* DEBUG_VM */


/***************************************************************************\
 * Private macros                                                          *
\***************************************************************************/

#ifdef DEBUG_VM
#   define WR_VM_ALIGNED_PTR(space,rel)   \
              _wss_vm_aligned_ptr(space,rel,__FILE__,__LINE__)
#   define WR_VM_VALID_PTR(space,rel)     \
              _wss_vm_valid_ptr(space,rel,__FILE__,__LINE__)
#   define WR_VM_VALID_ABS_PTR(space,abs) \
              _wss_vm_valid_abs_ptr(space,abs,__FILE__,__LINE__)
#else
#   define WR_VM_ALIGNED_PTR(space,rel)   \
              ((((unsigned)rel)%sizeof(HEADER))==0)
#   define WR_VM_VALID_PTR(space,rel)     \
              ((rel>space->vms_head->vms_head_base)||((((vm_size_t)rel)<space->vms_size)&&WR_VM_ALIGNED_PTR(space,rel)))
#   define WR_VM_VALID_ABS_PTR(space,abs) \
              (((pd_byte_t*)abs>(pd_byte_t*)space->vms_head)&&WR_VM_VALID_PTR(space,abs2rel(space,abs)))
#endif /* !DEBUG_VM */

#ifdef USER_MODE_TEST
#define round_page(s) s
#endif

#define WR_VM_MARKED    (1<<(sizeof(unsigned)*8-1))
#define WR_VM_SIZE_MASK ~WR_VM_MARKED
#define WR_VM_SIZE(space,blk) (((space)->vms_flags & WSS_VMS_MARKSWEEP)? \
                               (blk)->s.size&WR_VM_SIZE_MASK:(blk)->s.size)


#ifdef __GNUC__
#define INLINE __inline__
#elif defined(WR_WINNT)
#define INLINE __inline
#else
#define INLINE
#endif


/***************************************************************************\
 * Private functions                                                       *
\***************************************************************************/

#ifdef DEBUG_VM
/**
 * _wss_vm_aligned_ptr
 *
 * Description:
 */
static int INLINE
_wss_vm_aligned_ptr( wss_vmspace_t *space, wss_vmaddr_t rel, pd_char_t *file,
                     int line )
{
   if( ( ((unsigned)rel) % sizeof(HEADER) ) != 0 )
   {
      wss_debug_log( PD_MAKESTR("_wss_vm_aligned_ptr: unaligned pointer ==> ")
                     PD_MAKESTR("%p (%s:%d)\n"), (void *)rel, file, line );
      return 0;
   }
   else
      return 1;
}

/**
 * _wss_vm_valid_ptr
 *
 * Description:
 */
static int INLINE
_wss_vm_valid_ptr( wss_vmspace_t *space, wss_vmaddr_t rel, pd_char_t *file,
                   int line )
{
   if( _wss_vm_aligned_ptr(space, rel, file, line) )
   {
      if( (rel > space->vms_head->vms_head_base) ||
          (((vm_size_t)rel) < space->vms_size) )
         return 1;
      else
      {
         wss_debug_log( PD_MAKESTR("_wss_vm_valid_ptr: relative pointer out ")
                        PD_MAKESTR("of range ==> %p (%s:%d)\n"),
                        (void *)rel, file, line );
         return 0;
      }
   }
   else
      return 0;
}

/**
 * _wss_vm_valid_abs_ptr
 *
 * Description:
 */
static int INLINE
_wss_vm_valid_abs_ptr( wss_vmspace_t *space, void *abs, pd_char_t *file,
                       int line )
{
   if( (pd_byte_t *)abs > (pd_byte_t *)space->vms_head)
      return( _wss_vm_valid_ptr( space, abs2rel(space,abs), file, line ) );
   else
   {
      wss_debug_log( PD_MAKESTR("_wss_vm_valid_abs_ptr: absolute pointer out ")
                     PD_MAKESTR("of range ==> %p <= %p (%s:%d)\n"),
                     abs, space->vms_head, file, line );
      return 0;
   }
}
#endif /* DEBUG_VM */

/*
 * This function is used by wss_vmck as the sort function parameter to qsort
 */
static int
block_sort_compr( const void *one, const void *two )
{
   HEADER *hone = *((HEADER **)one);
   HEADER *htwo = *((HEADER **)two);
   return( (hone==htwo)?0:((hone>htwo)?1:-1) );
}

#ifndef USER_MODE_TEST
/*
 * Maps the database file into the kernel
 */
void *
_wss_vm_map_file( const pd_char_t *filename, size_t size_in_bytes,
                  wss_file_t *pfile, pd_bool_t *isnew )
{
   wss_file_t  file;
   int         error;
   void       *mapaddr;
   pd_bool_t   newfile = FALSE;

   if( !( file = wss_file_open( filename, WSS_FILE_READ | WSS_FILE_WRITE |
                                WSS_FILE_ROOTPRIV ) ) )
   {
      if( !( file = wss_file_open( filename, WSS_FILE_READ | WSS_FILE_WRITE |
                                   WSS_FILE_CREAT | WSS_FILE_ROOTPRIV ) ) )
      {
         wsi_printf( PD_MAKESTR("Error: unable to create %s\n"), filename );
         return NULL;
      }

      newfile = TRUE;
   }
   else
   {
      size_t size = wss_file_get_size( file );

      if( ( size != 0 ) && ( size != size_in_bytes ) )
      {
         wsi_printf( PD_MAKESTR("Error: wss.db(%lu bytes) should be %lu bytes.\n"),
                     (unsigned long)size, (unsigned long)size_in_bytes );

         wss_file_rele( file );
         return NULL;
      }
   }

   if( ( error = wss_file_map( file, size_in_bytes,
                               WSS_FILE_MAP_READ | WSS_FILE_MAP_WRITE, 0 ) ) )
   {
      wsi_printf( PD_MAKESTR("Error %d: unable to map wss.db\n"), error );
      wss_file_close( file );
      return NULL;
   }

   mapaddr = wss_file_get_map_addr( file );
   *pfile = file;

   wss_debug_log( PD_MAKESTR("wss_vm: mapped wss.db @ 0x%x\n"),
                  (unsigned int)mapaddr );


   if( newfile )
   {
      wss_vmspace_header_t *header = (wss_vmspace_header_t *)mapaddr;
      bzero( header, sizeof(*header) );
      header->vms_head_version = WSS_VMS_VERSION;
   }

   if( isnew ) *isnew = newfile;

   return mapaddr;
}
#endif


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_vm_init */
/*! \brief VM management initialization function
 *
 *  \param size Size of the database in megabytes
 *
 *  \return TRUE if database file mapping was successful,
 *  \return FALSE on error
 */
int EXPORT
wss_vm_init()
{
   WSS_MUTEX_INIT( wss_vm_spaces );
   LIST_INIT( &wss_vm_spaces );

   return 0;
}

/*  wss_vm_shutdown */
/*! \brief VM management shutdown function
 */
void EXPORT
wss_vm_shutdown()
{
}

/*  wss_vm_create */
/*! \brief Virtual Memory space creation function
 *
 *  \param name     Name of the VM space
 *  \param filename Name of file used as backing store
 *  \param size     Size of the database in megabytes
 *  \param reserve  Size of memory to reserve from allocation
 *
 *  \return Pointer to VM space structure on success
 *  \return NULL on failure
 */
wss_vmspace_t EXPORT *
wss_vm_create( const pd_char_t *name, const pd_char_t *filename, size_t size,
               size_t reserve )
{
   void          *data;
   wss_vmspace_t *space          = NULL;
   wss_file_t     file           = WSS_FILE_INVALID;
   size_t         size_in_bytes  = size * 1024 * 1024;
   pd_bool_t      mark_and_sweep = TRUE;
   pd_bool_t      isnew          = FALSE;

   if( ( size_in_bytes / sizeof(HEADER) ) > WR_VM_MARKED )
   {
      wss_debug_log( PD_MAKESTR("wss_vm:: VM space '%s' too large to allow ")
                     PD_MAKESTR("mark-and-sweep operations\n"),
                     name );
      mark_and_sweep = FALSE;
   }

   /*
    * Request the specified size of mapped memory
    */
   data = _wss_vm_map_file( filename, size_in_bytes, &file, &isnew );
   if( data != NULL )
   {
      wss_vmspace_header_t *head     = (wss_vmspace_header_t *)data;
      register int          vmh_size = sizeof(struct wss_vmspace_header);
      register int          vmb_size = sizeof(HEADER);

      if( head->vms_head_version != WSS_VMS_VERSION )
      {
         wss_debug_log( PD_MAKESTR("wss_vm:: Unable to handle version %d VM ")
                        PD_MAKESTR("space '%s'\n"), head->vms_head_version,
                        name );
         wss_debug_log( PD_MAKESTR("         mapped from file '%s'\n"),
                        filename );
         wss_file_close( file );
      }
      else
      {
         space = (wss_vmspace_t *)wss_malloc( sizeof(*space) );
         space->vms_name = name;
         space->vms_file = file;
         space->vms_size = wsi_file_round_to_page( size_in_bytes );
         space->vms_head = head;
         space->vms_flags = ((mark_and_sweep)?WSS_VMS_MARKSWEEP:0);
         space->vms_allocp = NULL;

         WSS_MUTEX_INIT( space->vms );

         /* Determine size of header aligned to block size */
         space->vms_headsize = (vmh_size + vmb_size) & ~(vmb_size - 1);

         if( isnew )
         {
            if( reserve )
               space->vms_head->vms_head_reserved = space->vms_headsize;
            else
               space->vms_head->vms_head_reserved = 0;
         }

         if( reserve )
            space->vms_headsize += (reserve + vmb_size) & ~(vmb_size - 1);

         if( isnew )
         {
            HEADER *head_base;
            head_base = (HEADER *)((pd_byte_t *)space->vms_head +
                                   space->vms_headsize);
            space->vms_head->vms_head_base = space->vms_headsize;
            head_base->s.next = NULL;
         }

         /* Walk the allocation chain and validate */
         switch( wss_vmck( space ) )
         {
            case VMCK_REINIT:
            {
               HEADER *head_base;
               space->vms_headsize = (vmh_size + vmb_size) & ~(vmb_size - 1);
               space->vms_head->vms_head_version = WSS_VMS_VERSION;
               if( reserve )
               {
                  space->vms_head->vms_head_reserved = space->vms_headsize;
                  space->vms_headsize += (reserve + vmb_size) & ~(vmb_size - 1);
               }
               else
                  space->vms_head->vms_head_reserved = 0;
               head_base = (HEADER *)((pd_byte_t *)space->vms_head +
                                      space->vms_headsize);
               space->vms_head->vms_head_base = space->vms_headsize;
               head_base->s.next = NULL;
               space->vms_head->vms_head_trace = 0;
               space->vms_allocp = NULL;
               break;
            }

            default:
               break;
         }

         /* Add VM space to list */
         WSS_MUTEX_ENTER( wss_vm_spaces );
         LIST_INSERT_HEAD( &wss_vm_spaces, space, vms_list );
         WSS_MUTEX_EXIT( wss_vm_spaces );
      }
   }

   return space;
}

/*  wss_vm_destroy */
/*! \brief Virtual Memory space destruction function
 *
 *  \param space Pointer to the VM space
 */
void EXPORT
wss_vm_destroy( wss_vmspace_t *space )
{
   WSS_MUTEX_ENTER( wss_vm_spaces );
   LIST_REMOVE( space, vms_list );
   WSS_MUTEX_EXIT( wss_vm_spaces );

   wss_file_close( space->vms_file );
   space->vms_file = WSS_FILE_INVALID;

   wss_free( space );
}

/*  wss_vmck */
/*! \brief VM check routine
 *
 *  Makes sure allocated blocks and free blocks are all accounted for
 *  within the mapped memory space.
 *
 *  This routine is called after a VM space is mapped into memory.
 *
 *  \param space Pointer to the VM space
 *
 *  \warning This function does not catch all corruption in the database
 *           and can crash the WSS (and possibly your system) when checking
 *           a corrupt database.  It does, however increase the chance of
 *           recovering from a corrupted database.
 */
int EXPORT
wss_vmck( wss_vmspace_t *space )
{
   register HEADER *p;
   register HEADER **q;
   register int nunits;
   register HEADER *head_base;
   register HEADER *base;
   register int nblocks;
   register int i;

   q = (HEADER **)&(space->vms_head->vms_head_trace);
   p = base = (HEADER *)(space->vms_head->vms_head_trace);
   nblocks = 0;
   nunits = 0;

   /* Validate the header information */
   wss_debug_log( PD_MAKESTR("wss_vmck: validating VM space header\n") );
   if( ( space->vms_head->vms_head_base < space->vms_headsize ) ||
       !WR_VM_ALIGNED_PTR(space,space->vms_head->vms_head_base) )
   {
      wss_debug_log( PD_MAKESTR("wss_vmck: allocated block base invalid,")
                     PD_MAKESTR(" reinitializing\n") );

      return VMCK_REINIT;
   }

   if( space->vms_head->vms_head_reserved )
   {
      if( !WR_VM_ALIGNED_PTR(space,space->vms_head->vms_head_reserved) )
      {
         wss_debug_log( PD_MAKESTR("wss_vmck: misaligned reserved space,")
                        PD_MAKESTR(" reinitializing\n") );
         return VMCK_REINIT;
      }

      if( space->vms_head->vms_head_reserved >= space->vms_headsize )
      {
         wss_debug_log( PD_MAKESTR("wss_vmck: reserved space outside header")
                        PD_MAKESTR(" space 0x%x/0x%x, reinitializing\n"),
                        space->vms_head->vms_head_reserved,
                        space->vms_headsize);
         return VMCK_REINIT;
      }
   }

   if( !WR_VM_VALID_PTR(space,space->vms_head->vms_head_trace) )
   {
      wss_debug_log( PD_MAKESTR("wss_vmck: invalid start of block trace,")
                     PD_MAKESTR(" reinitializing\n") );
      return VMCK_REINIT;
   }

   /* Validate the memory map */
   wss_debug_log( PD_MAKESTR("wss_vmck: validating memory map\n") );
   while( p )
   {
      if( !WR_VM_VALID_PTR(space, (wss_vmaddr_t)p) )
      {
         wss_debug_log( PD_MAKESTR("wss_vmck: bad block %p encountered\n"), p );
         *q = NULL;

         if( nblocks )
         {
            wss_debug_log( PD_MAKESTR("          setting all remaining blocks free\n") );
            break;
         }
         else
         {
            wss_debug_log( PD_MAKESTR("          setting all blocks free\n") );

            return VMCK_REINIT;
         }
      }

      p = (HEADER *)rel2abs(space,p);
      p->s.size &= WR_VM_SIZE_MASK; /* make sure the mark is cleared */
#ifdef DEBUG_VM
      wss_debug_log( PD_MAKESTR("wss_vmck: allocated block of %d bytes @ %p\n"),
              p->s.size * sizeof(HEADER), (void *)abs2rel(space,p) );
#endif /* DEBUG_VM */
      q = &(p->s.next);
      p = (HEADER *)((p->s.next==NULL)?NULL:p->s.next);

      nblocks++;
   }

   if( nblocks )
   {
      q = (HEADER **)wss_malloc( sizeof(HEADER *) * nblocks );

      wss_debug_log( PD_MAKESTR("wss_vmck: setting up sort array\n") );
      for( i = 0, p = (HEADER *)rel2abs(space,base);
           i < nblocks;
           i++, p = (HEADER *)((p->s.next)?rel2abs(space,p->s.next):NULL) )
         q[i] = (HEADER *)abs2rel(space,p);

      head_base = (HEADER *)((pd_byte_t *)(space->vms_head) +
                             space->vms_headsize);
      space->vms_head->vms_head_base = space->vms_headsize;
      head_base->s.next = (HEADER *)space->vms_head->vms_head_base;

      wss_debug_log( PD_MAKESTR("wss_vmck: sorting %d block(s)\n"), nblocks );
      qsort( q, nblocks, sizeof(*q), block_sort_compr );

      space->vms_head->vms_head_trace = (u_int32_t)q[0];

      wss_debug_log( PD_MAKESTR("wss_vmck: adjusting pointers and free ")
                     PD_MAKESTR("blocks\n") );
readjust:
      for( i = 0; i < nblocks; i++ )
      {
         HEADER *fblock;
         fblock = (HEADER *)((pd_byte_t *)q[i] +
                             WR_VM_SIZE(space,(HEADER *)rel2abs(space,q[i])) *
                             sizeof(HEADER));

         if( (i+1) < nblocks )
            p = q[i+1];
         else
            p = (HEADER *)space->vms_size;

         if( fblock < p )
         {
            /* Compute free block */
            HEADER *afblock = (HEADER *)rel2abs(space,fblock);
            wss_debug_log( PD_MAKESTR("wss_vmck: computing free block %p ==> %p"),
                           fblock, p );
            afblock->s.size = ((unsigned)p - (unsigned)fblock) / sizeof(HEADER);
            afblock->s.next = head_base->s.next;
            head_base->s.next = fblock;
            nunits += afblock->s.size;
            wss_debug_log( PD_MAKESTR(" (%u blks * %u bytes/blk = %u bytes)\n"),
                           afblock->s.size,
                           sizeof(HEADER), afblock->s.size * sizeof(HEADER));
         }
         else if( fblock > p )
         {
            HEADER *afblock = (HEADER *)rel2abs(space,fblock);
            int j;

            /* We have overlapping blocks, remove this block and readjust pointers */
            wss_debug_log( PD_MAKESTR("wss_vmck: overlapping block %p + %u > %p, readjusting\n"),
                           q[i],
                           (fblock >= (HEADER *)space->vms_size)?
                              0:WR_VM_SIZE(space, afblock),
                           p );
            for( j = i + 1; j < nblocks; j++ ) q[i] = q[j];

            if( i == 0 ) space->vms_head->vms_head_trace = (u_int32_t)q[0];
            head_base->s.next = head_base;
            nblocks--;
            goto readjust;
         }

         if( (i+1) < nblocks )
            ((HEADER *)rel2abs(space,q[i]))->s.next = p;
         else
            ((HEADER *)rel2abs(space,q[i]))->s.next = NULL;
      }

      /* Compute last remaining free block */
      if( (pd_byte_t *)rel2abs(space,q[0]) > ((pd_byte_t *)(space->vms_head) +
                                        space->vms_headsize) )
      {
         base = (HEADER *)rel2abs(space,space->vms_head->vms_head_base);
         base->s.size = ( (unsigned)q[0] - ((unsigned)abs2rel(space,space->vms_head) +
                          space->vms_headsize )) / sizeof(HEADER);
         nunits += base->s.size;

         wss_debug_log( PD_MAKESTR("wss_vmck: computing free block %p ")
                        PD_MAKESTR("(%u blks * %u bytes/blk = %u bytes)\n"),
                        (void *)abs2rel(space,base), base->s.size, sizeof(HEADER),
                        base->s.size * sizeof(HEADER) );
      }

      /* Set free memory remaining */
      space->vms_head->vms_head_free = nunits * sizeof(HEADER);

      /* Free the sort array */
      wss_free( q );
   }
   else
   {
      /* No tracked blocks, set up same as new VM space */
      wss_debug_log( PD_MAKESTR("wss_vmck: warning: dropping all blocks, no")
                     PD_MAKESTR(" tracked memory found\n") );
      return VMCK_REINIT;
   }

   return VMCK_OK;
}

/*  wss_vm_valid_alloced */
/*! \brief Walk trace list to find the specified block
 *
 *  \param space Pointer to the VM space
 *  \param rel   Relative pointer to memory block
 *
 *  \warning String represented by \a rel must be a relative address
 *
 *  \return TRUE if we find the specified block,
 *  \return FALSE otherwise
 *
 */
int EXPORT
wss_vm_valid_alloced( wss_vmspace_t *space, void *rel )
{
   register HEADER *mem = (HEADER *)rel - 1;
   
   if( WR_VM_VALID_PTR(space, (wss_vmaddr_t)rel) &&
       ( space->vms_head->vms_head_trace != 0 ) )
   {
      register HEADER *trace = (HEADER *)space->vms_head->vms_head_trace;
      register HEADER *q;
      for( q = (HEADER *)rel2abs(space,trace); q;
           q = (HEADER*)((q->s.next)?rel2abs(space,q->s.next):NULL) )
      {
#ifdef DEBUG_VM
         if( !WR_VM_VALID_ABS_PTR(space, q) )
         {
            wss_debug_log( PD_MAKESTR("wr_wm_valid_alloced: ptr out of range in trace ==> %p\n"), q );
            panic( "wss_vm_valid_alloced" );
         }
#endif /* DEBUG_VM */
         if( q->s.next == mem )
            break;
      }

      if( q == NULL )
         return( ( mem == trace )?TRUE:FALSE );

      return TRUE;
   }
   else
      return FALSE;
}

/*  wss_vm_valid_str */
/*! \brief Check for valid null-terminated string within the VM space
 *
 *  String length must be less than size of block minus the header.
 *
 *  \param space Pointer to the VM space
 *  \param rel   Relative pointer to the address of the string
 *
 *  \warning String represented by \a rel must be a relative address
 *
 *  \return TRUE if the string is valid,
 *  \return FALSE otherwise
 */
int EXPORT
wss_vm_valid_str( wss_vmspace_t *space, const pd_char_t *rel )
{
   HEADER *blk = (HEADER *)rel2abs(space,rel) - 1;
   unsigned int i;
   pd_char_t *ptr;

   if( !wss_vm_valid_alloced( space, (void *)rel) ) return FALSE;

   for( i = 0, ptr = (pd_char_t *)(blk + 1);
        *ptr && ( i < (WR_VM_SIZE(space, blk) - 1) * sizeof(HEADER) );
        i += sizeof(pd_char_t), ptr++ );

   return( ( *ptr )?FALSE:TRUE );
}


/*  wss_vm_mark */
/*! \brief Mark block for exclusion from later sweeping operation
 *
 *  \param space Pointer to the VM space
 *  \param rel   Relative pointer to memory block
 *
 *  \warning Memory block represented by \a rel must be a relative address
 */
void EXPORT
wss_vm_mark( wss_vmspace_t *space, void *rel )
{
   register HEADER *mem = (HEADER *)rel - 1;

   if( ( space->vms_flags & WSS_VMS_MARKSWEEP ) &&
       wss_vm_valid_alloced( space, rel ) )
   {
#ifdef DEBUG_VM_MARK
      wss_debug_log( PD_MAKESTR("wss_vm_mark: marked %d byte block ==> %p (%p)\n"),
                     WR_VM_SIZE(space, mem) * sizeof(HEADER), mem,
                     (void *)abs2rel(space,mem) );
#endif /* DEBUG_VM_MARK */
      mem = (HEADER *)rel2abs(space,mem);
      mem->s.size |= WR_VM_MARKED;
   }
}


/*  wss_vm_sweep */
/*! \brief Sweep non-marked blocks onto the free list
 *
 *  \param space Pointer to the VM space
 */
void EXPORT
wss_vm_sweep( wss_vmspace_t *space )
{
   if( ( space->vms_flags & WSS_VMS_MARKSWEEP ) &&
       space->vms_head->vms_head_trace )
   {
      register HEADER *q;
      register HEADER *p;
      register int blocks = 0;

      WSS_MUTEX_ENTER( space->vms );
      for( p = (HEADER *)rel2abs(space,space->vms_head->vms_head_trace); p; p = q )
      {
         q = (HEADER *)(p->s.next?rel2abs(space,p->s.next):NULL);

         if( ( p->s.size & WR_VM_MARKED ) == 0 )
         {
#ifdef DEBUG_VM
            wss_debug_log( PD_MAKESTR("wss_vm_sweep: swept %d byte block ==> %p (%p)\n"),
                           WR_VM_SIZE(space, p) * sizeof(HEADER), p,
                           (void *)abs2rel(space,p) );
#endif /* DEBUG_VM */
            blocks += WR_VM_SIZE(space, p);
            wss_vm_free_unsafe( space, p + 1 );
         }
      }
      WSS_MUTEX_EXIT( space->vms );

#ifdef DEBUG_VM
      wss_debug_log( PD_MAKESTR("wss_vm_sweep: reclaimed %d bytes in free blocks\n"),
                     blocks * sizeof(HEADER) );
#endif /* DEBUG_VM */
   }
}


/*
 * The following mem malloc/free code adapted from "The C Programming
 * Language" by Kernighan and Ritchie, chapter 8.
 */


/*  wss_vm_malloc_debug */
/*  wss_vm_malloc */
/*! \brief malloc for the VM space
 *
 *  Like malloc(3), but allocates (absolute ptrs) from the VM region.
 *  This routine maintains all internal pointers as relative addresses
 *  within the VM region.  This means that memory allocated by DB
 *  structures in the VM region will be preserved across runs of the
 *  WSS.
 *
 *  \param nbytes Number of bytes of memory to allocation
 *
 *  \return Pointer to memory block,
 *  \return NULL on failure (i.e., if no memory was available)
 */
#ifdef DEBUG_VM
void EXPORT *
wss_vm_malloc_debug( wss_vmspace_t *space, unsigned nbytes,
                     const pd_char_t *file, int line )
#else
void EXPORT *
wss_vm_malloc( wss_vmspace_t *space, unsigned nbytes )
#endif /* DEBUG_VM */
{
   register HEADER *p, *q;
   register unsigned int nunits;
   register HEADER *base = NULL;

#ifdef DEBUG_VM
   wss_debug_log( PD_MAKESTR("wss_vm_malloc: Allocating %d bytes (%s:%d)\n"),
                  nbytes, file, line );
#endif /* DEBUG_VM */

   /* compute the number of needed units */
   nunits = 1 + (nbytes + sizeof(HEADER) - 1) / sizeof(HEADER);

   /* 
    * If this is the first malloc/free for this run of the wss,
    * initialize the header.
    */
   WSS_MUTEX_ENTER( space->vms );
   if( ( q = space->vms_allocp ) == NULL )
   {
      base = (HEADER *)rel2abs(space,space->vms_head->vms_head_base);
#ifdef DEBUG_VM
      wss_debug_log( PD_MAKESTR("wss_vm_malloc: Setting allocp to %p\n"), base );
#endif /* DEBUG_VM */
      if( base->s.next == NULL )
      {
         /* new wss.db file */
         base->s.next = (HEADER*)abs2rel(space,base);
         base->s.size = ( space->vms_size - space->vms_headsize ) /
                        sizeof(HEADER);
         space->vms_head->vms_head_trace = 0;
         space->vms_head->vms_head_free = base->s.size * sizeof(HEADER);
      }
      space->vms_allocp = q = base;
#ifdef DEBUG_VM
      if( space->vms_allocp == NULL )
         wss_debug_log( PD_MAKESTR("wss_vm_malloc: allocp set null\n") );
#endif /* DEBUG_VM */
   }

#ifdef DEBUG_VM
   if( !WR_VM_VALID_ABS_PTR(space, space->vms_allocp) )
   {
      wss_debug_log( PD_MAKESTR("wss_vm_malloc: allocp out of range ==> %p\n"),
                     space->vms_allocp );
      panic( "wss_vm_malloc" );
   }
   wss_debug_log( PD_MAKESTR("wss_vm_malloc: q = %p(%p), allocp = %p(%p), base = %p(%p)\n"),
                  q, (void *)abs2rel(space,q), (void *)space->vms_allocp,
                  (void *)abs2rel(space,space->vms_allocp), base,
                  ((base!=NULL)?(void *)abs2rel(space,base):(void *)base) );
#endif /* DEBUG_VM */

   /* Find a free block: use first one big enough. */
   for( p=(HEADER*)rel2abs(space,q->s.next); ; q=p, p=(HEADER*)rel2abs(space,p->s.next) )
   {
#ifdef DEBUG_VM
      wss_debug_log( PD_MAKESTR("wss_vm_malloc: Checking %p(%p)\n"),
                     p, (void *)abs2rel(space,p) );
      if( !WR_VM_VALID_ABS_PTR(space, p) )
      {
         wss_debug_log( PD_MAKESTR("wss_vm_malloc: free block out of range ==> %p\n"), p );
         panic( "wss_vm_malloc" );
      }
#endif /* DEBUG_VM */
      if( WR_VM_SIZE(space, p) >= nunits )
      {
         /* big enough */
         if( WR_VM_SIZE(space, p) == nunits )
         {
            /* exactly */
            q->s.next = p->s.next;  /* both addrs rel */
         }
         else
         {
            p->s.size = WR_VM_SIZE(space, p) - nunits;
            p += WR_VM_SIZE(space, p);
#ifdef DEBUG_VM
            if( !WR_VM_VALID_ABS_PTR(space, p) )
            {
               wss_debug_log( PD_MAKESTR("wss_vm_malloc: new block out of range ==> %p\n"), p );
               panic( "wss_vm_malloc" );
            }
#endif /* DEBUG_VM */
            p->s.size = nunits;
         }

         space->vms_head->vms_head_free -= nunits * sizeof(HEADER);

         /* Add block to trace list */
         p->s.next = (HEADER *)space->vms_head->vms_head_trace;
         space->vms_head->vms_head_trace = abs2rel(space,p);

         space->vms_allocp = q;
#ifdef DEBUG_VM
         if( space->vms_allocp == NULL )
            wss_debug_log( PD_MAKESTR("wss_vm_malloc: allocp set null\n") );
#endif /* DEBUG_VM */
         WSS_MUTEX_EXIT( space->vms );
         return( (void*) (p+1) );
      }
      if( p == space->vms_allocp )
      {
         WSS_MUTEX_EXIT( space->vms );
         /* wrapped around free list */
         return NULL;  /* out of memory */
      }
   }
}

/*  wss_vm_free_unsafe_debug */
/*  wss_vm_free_unsafe */
/*! \brief Free memory within the VM
 *
 *  Like free(3), but returns (absolute addressed) memory to the VM region
 *
 *  \param addr Pointer to block of memory to free
 */
#ifdef DEBUG_VM
void EXPORT
wss_vm_free_unsafe_debug( wss_vmspace_t *space, void *addr,
                          const pd_char_t *file, int line )
#else
void EXPORT
wss_vm_free_unsafe( wss_vmspace_t *space, void *addr )
#endif /* DEBUG_VM */
{
   register HEADER *mem, *p;

#ifdef DEBUG_VM
   wss_debug_log( PD_MAKESTR("wss_vm_free: Freeing block %p (%s:%d)\n"),
                  addr, file, line );
#endif /* DEBUG_VM */

   mem = (HEADER *)addr - 1;     /* point to header */

   /* 
    * If this is the first malloc/free for this run of the wss,
    * initialize the header.
    */
   if( space->vms_allocp == NULL )
   {
      space->vms_allocp = (HEADER *)rel2abs(space,space->vms_head->vms_head_base);
      if( ((HEADER *)space->vms_allocp)->s.next == NULL )
      {
         wss_debug_log( PD_MAKESTR("BAD Error: trying to free from VM before any allocations\n") );
         /* new wss.db file */
         ((HEADER *)space->vms_allocp)->s.next =
            (HEADER*)abs2rel(space,space->vms_allocp);
         ((HEADER *)space->vms_allocp)->s.size =
            ( space->vms_size - space->vms_headsize ) / sizeof(HEADER);
         space->vms_head->vms_head_trace = 0;
         space->vms_head->vms_head_free =
            ((HEADER *)space->vms_allocp)->s.size * sizeof(HEADER);
      }
#ifdef DEBUG_VM
      if( space->vms_allocp == NULL )
         wss_debug_log( PD_MAKESTR("wss_vm_free: allocp set null\n") );
#endif /* DEBUG_VM */
   }

#ifdef DEBUG_VM
   if( !WR_VM_VALID_ABS_PTR(space, space->vms_allocp) )
   {
      wss_debug_log( PD_MAKESTR("wss_wm_free: allocp out of range ==> %p\n"),
                     space->vms_allocp );
      panic( "wss_vm_free" );
   }
   if( !WR_VM_VALID_ABS_PTR(space, mem) )
   {
      wss_debug_log( PD_MAKESTR("wss_wm_free: free block out of range ==> %p\n"),
                     mem );
      panic( "wss_vm_free" );
   }
#endif /* DEBUG_VM */

   space->vms_head->vms_head_free += WR_VM_SIZE(space, mem) * sizeof(HEADER);

   if( space->vms_head->vms_head_trace != 0 )
   {
      register HEADER  *trace   = (HEADER *)space->vms_head->vms_head_trace;
      register HEADER **q       = (HEADER **)&(space->vms_head->vms_head_trace);
#ifdef DEBUG_VM
      HEADER           *q_debug = trace;
#ifdef DEBUG_VM_TRACK
      wss_debug_log( PD_MAKESTR("wss_vm_free: tracking %p:"), mem );
#endif /* DEBUG_VM_TRACK */
#endif /* DEBUG_VM */
      for( p = (HEADER *)rel2abs(space,trace); p;
           q = &p->s.next,
#ifdef DEBU_VM
           q_debug = (HEADER *)abs2rel(space,p),
#endif
           p = (HEADER*)((p->s.next)?rel2abs(space,p->s.next):NULL) )
      {
#ifdef DEBUG_VM
#ifdef DEBUG_VM_TRACK
         if( (unsigned)p <= (unsigned)mem )
            wss_debug_log( PD_MAKESTR(" %p==%p"), p, mem );
#endif /* DEBUG_VM_TRACK */
         if( !WR_VM_VALID_ABS_PTR(space, p) )
         {
#ifdef DEBUG_VM_TRACK
            wss_debug_log( PD_MAKESTR("\n") );
#endif
            wss_debug_log( PD_MAKESTR("wss_wm_free: ptr out of range in trace ==> %p\n"), p );
            panic( "wss_vm_free" );
         }
#endif /* DEBUG_VM */
         if( p == mem )
         {
#ifdef DEBUG_VM
#ifdef DEBUG_VM_TRACK
            wss_debug_log( PD_MAKESTR("\n") );
#endif
            wss_debug_log( PD_MAKESTR("wss_vm_free: Setting *q (&%p->s.next) = %p (next = %p)\n"),
                           q_debug, mem->s.next,
                           (mem->s.next)?
                              ((HEADER *)rel2abs(space,mem->s.next))->s.next:NULL );
#endif /* DEBUG_VM */
            *q = mem->s.next;
            break;
         }
      }
#ifdef DEBUG_VM
#ifdef DEBUG_VM_TRACK
      if( p == NULL )
         wss_debug_log( PD_MAKESTR("\n") );
#endif /* DEBUG_VM_TRACK */
#endif /* DEBUG_VM */

      if( p == NULL )
      {
         if( (HEADER *)abs2rel(space,mem) == trace )
            space->vms_head->vms_head_trace = (u_int32_t)mem->s.next;
#ifdef DEBUG_VM
         else
         {
            HEADER *head_base =
               (HEADER *)rel2abs(space,space->vms_head->vms_head_base);
            for( p = (HEADER *)rel2abs(space,head_base->s.next); p;
                 p = (HEADER*)rel2abs(space,p->s.next) )
            {
               if( p == mem )
                  panic( "wss_vm_free: already freed block ==> %p\n", mem );
               if( p == head_base ) break;
            }
            panic( "wss_vm_free: unable to find block in trace ==> %p\n", mem);
         }
#endif /* DEBUG_VM */
      }
   }

   /* find free blocks adjacent to the free'd memory*/
   for( p = space->vms_allocp;
        !( ( mem > p ) && ( mem < (HEADER*)rel2abs(space,p->s.next) ) );
        p = (HEADER*)rel2abs(space,p->s.next) )
   {
      if( ( p >= (HEADER*)rel2abs(space,p->s.next) ) &&
          ( ( mem > p ) || ( mem < (HEADER*)rel2abs(space,p->s.next) ) ) )
      {
         break;      /* at one end or the other */
      }
   }
   if( mem + WR_VM_SIZE(space, mem) == (HEADER*)rel2abs(space,p->s.next) )
   {
      /* join to upper free block */
      mem->s.size = WR_VM_SIZE(space, mem) +
                    WR_VM_SIZE(space, (HEADER*)rel2abs(space,p->s.next));
      mem->s.next = ((HEADER*)rel2abs(space,p->s.next))->s.next;
   }
   else
   {
      mem->s.next = p->s.next;
   }

   if( p + WR_VM_SIZE(space, p) == mem )
   {
      /* join to lower block */
      p->s.size = WR_VM_SIZE(space, p) + WR_VM_SIZE(space, mem);
      p->s.next = mem->s.next;
   }
   else
   {
      p->s.next = (HEADER*)abs2rel(space,mem);
   }
   space->vms_allocp = p;

#ifdef DEBUG_VM
   if( space->vms_allocp == NULL )
      wss_debug_log( PD_MAKESTR("wss_vm_free: allocp set null\n") );
#endif /* DEBUG_VM */
}

/*  wss_vm_free_debug */
/*  wss_vm_free */
/*! \brief Thread-safe version of free memory within the VM
 *
 *  Like free(3), but returns (absolute addressed) memory to the VM region
 *
 *  \param addr Pointer to block of memory to free
 */
#ifdef DEBUG_VM
void EXPORT
wss_vm_free_debug( wss_vmspace_t *space, void *addr, const pd_char_t *file,
                   int line )
#else
void EXPORT
wss_vm_free( wss_vmspace_t *space, void *addr )
#endif /* DEBUG_VM */
{
   WSS_MUTEX_ENTER( space->vms );
#ifdef DEBUG_VM
   wss_vm_free_unsafe_debug( space, addr, file, line );
#else
   wss_vm_free_unsafe( space, addr );
#endif /* DEBUG_VM */
   WSS_MUTEX_EXIT( space->vms );
}

/*  wss_vm_first_free_block() */
/*  wss_vm_first_free_block()
 *! \brief Returns the address of the first free block in VM space
 *
 *  in:    wss_vm_data (global) - points to the beginning of a "mapped"
 *                      file that contains the data structures for K&R
 *                      style memory management (as described in the
 *                      existing wss_vm.c comments).
 *  out:   nothing
 *  return: value       condition
 *          -----       -----------
 *          NULL        the "mapped" file indicated by `wss_vm_data'
 *                      did not contain the proper memory management
 *                      data structures, or the data structures were
 *                      not properly initialized.
 *          !NULL       the absolute address of the first free block in
 *                      the "mapped" file indicated by `wss_vm_data'.
 *
 *  This function returns the (absolute) address of the first free
 *  block in the "mapped" file indicated by `wss_vm_data'.  The "mapped"
 *  file must contain properly initialized K&R-style memory management
 *  data structures, or this function will return NULL.
 *
 *  This function does not aquire the lock associated with the "mapped"
 *  file's DB.  This allows a caller to lock the DB, and then make
 *  several calls to this function and wss_vm_next_free_block() in
 *  an atomic fashion.  This atomicity may be useful to an algorithm
 *  which flushes the entire DB.
 *
 *  \return The absolute address of the first free block in the VM space.
 *  \return NULL on failure
 *
 */

void EXPORT * wss_vm_first_free_block( wss_vmspace_t *space )
{
    /* The `base' field contains the address of the first HEADER *
     * in the free block list, relative to `wss_vm_data'.        */
    return rel2abs(space,space->vms_head->vms_head_base);

} /* wss_vm_first_free_block() */

/* wss_vm_next_free_block() */
/* wss_vm_next_free_block()
 *! \brief iterator of VM blocks
 *
 *  This function returns the (absolute) address of the free block
 *  which follows `p_previous_block'.  Note that the free block list
 *  is a ring, so if you want to avoid traversing the list forever,
 *  keep an eye out for the first free block, as identified by
 *  wss_vm_first_free_block().
 *
 *  This function does not aquire the lock associated with the "mapped"
 *  file's DB.  This allows a caller to lock the DB, and then make
 *  several calls to this function and wss_vm_next_free_block() in
 *  an atomic fashion.  This atomicity may be useful to an algorithm
 *  which flushes the entire DB.
 *
 *  \param p_previous_block absolute address of the previous free block
 *  \return the absolute address of the free block following `p_previous_block'
 *  \return NULL on error.
 */

void EXPORT *wss_vm_next_free_block( wss_vmspace_t *space,
                                     void *p_previous_block )
{
    HEADER *block = p_previous_block;

    return block ? block->s.next?rel2abs(space,block->s.next):NULL : NULL;
}

/*  wss_vm_header_size() */
/*  wss_vm_header_size()
 *
 *! \brief returns the size of the HEADER structure
 *
 *  This function always returns the same value - the size of the
 *  HEADER structure.  This structure is local to wss_vm.c.  This
 *  function reports the size without requiring the structure to be in
 *  public scope.
 *
 *  \return size of the HEADER structure.
 */

size_t EXPORT wss_vm_header_size( void )
{
    return( sizeof( HEADER ) );
}

/*  wss_vm_free_block_size() */
/*  wss_vm_free_block_size()
 *! \brief Reports the size of the free block with header 'p_header'
 *
 *  \param p_header pointer to free block header
 *  \return number of bytes in free block indicated by `p_header'.
 *
 */

size_t EXPORT wss_vm_free_block_size( void *p_header )
{
    HEADER *p_head = p_header;

    return p_head->s.size*sizeof(HEADER);
}

