/*! \file wss_wrapper.c
 * Copyright (C) 1999, 2000, 2001  Network Associates, Incorporated (NAI)
 * All rights reserved.
 *
 * This file is part of the NAI Generic Software Wrappers Toolkit
 * (GSWTK).  See ftp://ftp.tislabs.com/pub/wrappers/ for the latest
 * distribution.
 *
 * The GSWTK is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The GSWTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * $Id: wss_wrapper.c,v 1.50 2001/06/22 04:35:04 stevek Exp $
 *
 * \brief   Functions for manipulating wrapper modules
 * \author  Chris Vance, Steve Kiernan, Doug Kilpatrick
 * \version $Revision: 1.50 $
 * \date    $Date: 2001/06/22 04:35:04 $
 */

#include <wr_conf.h>

#include <wsi_stdlib.h>

#include <wss_wrapper.h>
#include <wss_debug.h>
#include <wss_iface.h>
#include <wss_malloc.h>
#include <wss_module.h>

#include <eventfs.h>

#include <wrerrno.h>


/***************************************************************************\
 * Public globals                                                          *
\***************************************************************************/

wss_wrapper_t EXPORT *wss_wrapper_list = NULL;
WSS_MUTEX_PROTECT   ( wss_wrapper_list );


/***************************************************************************\
 * Public functions                                                        *
\***************************************************************************/

/*  wss_wrapper_init */
/*! \brief Initialization routine for wrapper modules
 *
 *  Constructs the mutex protecting the wrapper list
 */
void EXPORT
wss_wrapper_init()
{
   WSS_MUTEX_INIT( wss_wrapper_list );
}


/*  wss_wrapper_shutdown */
/*! \brief A termination routine for wrapper modules
 *
 *  Iterates through the list of wrappers deallocating memory and
 *  destroys the mutex protecting the wrapper list.
 */
void EXPORT
wss_wrapper_shutdown()
{
   wss_wrapper_t *w, *this;

   WSS_MUTEX_ENTER( wss_wrapper_list );
   w = wss_wrapper_list;
   while( w )
   {
      this = w;
      w = w->w_next;
      wss_debug_log( PD_MAKESTR("wss_wrapper_shutdown: Cleaning up %s\n"),
                     this->w_module->wrm_linked->linker_file_name );
      wss_free( this );
   }
   wss_wrapper_list = NULL;
   WSS_MUTEX_EXIT( wss_wrapper_list );
   WSS_MUTEX_DESTROY( wss_wrapper_list );
}


/*  wss_wrapper_loaded */
/*! \brief Determines if the named wrapper is currently loaded
 *
 *  \param name Pointer to string containing the name of the wrapper
 *
 *  \return Identifier of named wrapper if it is currently loaded,
 *  \return 0 otherwise
 */
int EXPORT
wss_wrapper_loaded( pd_char_t *name )
{
   wss_wrapper_t *w;
   wss_id_t       id = 0;

   WSS_MUTEX_ENTER( wss_wrapper_list );
   for( w = wss_wrapper_list;
        w && ( wsi_strcmp( name,
                           w->w_module->wrm_linked->linker_file_name ) != 0 );
        w = w->w_next )
   ;
   if( w )
      id = w->w_module->wrm_id;
   WSS_MUTEX_EXIT( wss_wrapper_list );

   return id;
}


/*  wss_wrapper_register */
/*! \brief Register a wrapper with the given name
 *
 *  \param wrm        Pointer to the wrapper module structure
 *  \param pre        Pointer to the pre-call dispatch function
 *  \param post       Pointer to the post-call dispatch function
 *  \param activate   Pointer to the activation routine
 *  \param deactivate Pointer to the deactivation routine
 *  \param duplicate  Pointer to the duplication routine
 *
 *  \return 0 when wrapper is registered,
 *  \return \b WRE_NOMEM if wss_malloc fails to allocate memory for structure
 *
 *  \sa wss_wrapper_call_activation, wss_wrapper_call_deactivation,
 *      wss_wrapper_call_duplication
 *  \internal
 */
int EXPORT
wss_wrapper_register( wss_mod_t *wrm, void *pre, void *post,
                      void *activate, void *deactivate, void *duplicate )
{
   wss_wrapper_t *w;
   
   if( ( w = (wss_wrapper_t *)wss_malloc( sizeof( *w ) ) ) == NULL )
   {
      wsi_printf( PD_MAKESTR("wss_wrapper_register: Can't malloc memory for wrapper\n") );
      return WRE_NOMEM;
   }

   bzero( w, sizeof( *w ) );
   wss_ref_init( &(w->w_refcount) );
   wss_mod_ref( wrm );
   w->w_module = wrm;
   w->w_max_space = 0;
   w->w_dispatch_pre = pre;
   w->w_dispatch_post = post;
   w->w_procs = 0;
   w->w_activate = activate;
   w->w_deactivate = deactivate;
   w->w_duplicate = duplicate;
   WSS_MUTEX_INIT( w->w );

   /*
    * Blindly prepend to the wrapper list.
    */
   WSS_MUTEX_ENTER( wss_wrapper_list );
   w->w_next = wss_wrapper_list;
   wss_wrapper_list = w;
   WSS_MUTEX_EXIT( wss_wrapper_list );
   
   return 0;
}


/*  wss_wrapper_free */
/*! \brief Frees the memory associated with a wrapper
 *
 *  \param w Pointer to the wrapper structure
 */
void EXPORT
wss_wrapper_free( wss_wrapper_t *w )
{
   assert( w != NULL );
   assert( WSS_REF_COUNT(w->w_refcount) == 1 );

   WSS_REF_DEC(w->w_refcount);

#ifdef DEBUG_WRAPPER
   wsi_printf( PD_MAKESTR("wss_wrapper_free: releasing memory for 0x%x\n"),
               (unsigned int)w );
#endif /* DEBUG_WRAPPER */

   wss_ref_destroy( &(w->w_refcount) );

   wss_mod_rele( w->w_module );

   WSS_MUTEX_DESTROY( w->w );
   wss_free( (void *)w );
}


/*  wss_wrapper_can_unload */
/*! \brief Determines if wrapper module is in proper state to be unloaded
 *
 *  \param id Identifier of wrapper module
 *
 *  \return TRUE if wrapper can be unloaded,
 *  \return FALSE otherwise
 */
int EXPORT
wss_wrapper_can_unload( wss_id_t id )
{
   wss_wrapper_t *w, *prev_w;

   WSS_MUTEX_ENTER( wss_wrapper_list );
   for( prev_w = NULL, w = wss_wrapper_list;
        w && ( w->w_module->wrm_id != id );
        prev_w = w, w = w->w_next )
   ;

   /*
    * If not found, return TRUE.
    */
   if( w == NULL )
   {
      WSS_MUTEX_EXIT( wss_wrapper_list );
      return TRUE;
   }

#ifdef WR_LINUX
   /* Before we return true, remove our hook onto init's wrp struct.
    * Yes, side effects.  Yes, Evil.  I know.  The whole concept of
    * activation doesn't mix with async events anyway.
    */
   {
       wss_proc_t              *init_wrp;
       wss_wrapper_instance_t  *wi;
       wss_wrapper_instance_t  *prev_wi = NULL;
       int                      found_wi = 0;

       init_wrp = wss_pfind(0);
       /* If init_wrp == NULL, then there certainly ain't no 
        * wrapper_instance structs hanging off, cause the code
        * that adds the instances creates the wrp for init.
        */
       if (init_wrp != NULL) {
           WSS_MUTEX_ENTER( init_wrp->wrp_wi_list );
           wi = (wss_wrapper_instance_t *)(init_wrp->wrp_wi_list);
           for( ; wi && (
                    wi->wi_wrapper->w_module->wrm_id < w->w_module->wrm_id );
                   prev_wi = wi, wi = wi->wi_next )
               ;

           /* The wrapper instance we keep on "init" keeps a reference to
            * the wrapper.
            *
            * So, if we have an instance on init, allow an extra reference.
            */

           if (wi && wi->wi_wrapper->w_module->wrm_id == id) {
               found_wi = 1;
           }
           if ( WSS_REF_COUNT(w->w_refcount) > (1 + found_wi ) ) {
               WSS_MUTEX_EXIT( wss_wrapper_list );
               WSS_MUTEX_EXIT( init_wrp->wrp_wi_list );

               return FALSE;
           }


           if (found_wi) {
               if (prev_wi)
                   prev_wi->wi_next = wi->wi_next;
               else
                   init_wrp->wrp_wi_list = wi->wi_next;
               if (wi->wi_next)
                   wi->wi_next->wi_prev = wi->wi_prev;
               wi->wi_next = wi->wi_prev = NULL;
           }

           WSS_MUTEX_EXIT(  init_wrp->wrp_wi_list );

           if (found_wi)
               wss_wrapper_instance_rele(wi);
           wss_proc_rele(init_wrp);
       }
   }
#else
   /*
    * Can only remove a wrapper if the refcount is one or less.
    *
    * Note: all activation criteria and/or wrappers referencing
    *       this wrapper must go away
    */
   if( WSS_REF_COUNT(w->w_refcount) > 1 )
   {
      WSS_MUTEX_EXIT( wss_wrapper_list );
      return FALSE;
   }
#endif

   WSS_MUTEX_EXIT( wss_wrapper_list );

   return TRUE;
}


/*  wss_wrapper_delete */
/*! \brief Removes a record from the global list of wrappers
 *
 *  \param id Identifier of wrapper module
 *
 *  \return 0 on success,
 *  \return \b WRE_NOTFOUND if the wrapper identified by \a id is not found,
 *  \return \b WRE_REF if there are currently references to the wrapper
 */
int EXPORT
wss_wrapper_delete( wss_id_t id )
{
   wss_wrapper_t *w, *prev_w;

   WSS_MUTEX_ENTER( wss_wrapper_list );
   for( prev_w = NULL, w = wss_wrapper_list;
        w && ( w->w_module->wrm_id != id );
        prev_w = w, w = w->w_next )
   ;

   /*
    * If not found, return error.
    */
   if( w == NULL )
   {
      WSS_MUTEX_EXIT( wss_wrapper_list );
      return WRE_NOTFOUND;
   }

   /*
    * Can only remove a wrapper if the refcount is one or less.
    *
    * Note: all activation criteria and/or wrappers referencing
    *       this wrapper must go away
    */
   if( WSS_REF_COUNT(w->w_refcount) > 1 )
   {
      WSS_MUTEX_EXIT( wss_wrapper_list );
      return WRE_REF;
   }
   
   /*
    * Remove this wrapper from the list
    */
   if( prev_w )
      prev_w->w_next = w->w_next;
   else
      wss_wrapper_list = w->w_next;
   WSS_MUTEX_EXIT( wss_wrapper_list );

   /*
    * Release the reference from creation
    */
   wss_wrapper_rele( w );

   return 0;
}


/*  wss_wrapper_find */
/*! \brief Find the wrapper structure corresponding to the given name
 *
 *  \param name Pointer to string containing name of wrapper module
 *
 *  \return Pointer to wrapper structure if found,
 *  \return NULL if named wrapper is not found
 */
wss_wrapper_t EXPORT *
#ifdef DEBUG_REFS
wss_wrapper_find_debug( const pd_char_t *name, const pd_char_t *file,
                        int line  )
#else
wss_wrapper_find( const pd_char_t *name )
#endif /* DEBUG_REFS */
{
   wss_wrapper_t *w;

   WSS_MUTEX_ENTER( wss_wrapper_list );
   for( w = wss_wrapper_list;
        w && ( wsi_strcmp( w->w_module->wrm_linked->linker_file_name,
                           name ) != 0 );
        w = w->w_next )
   ;
   if( w )
#if defined(DEBUG_REFS) && !defined(DISABLE_WRAP_DEBUG_REFS)
      wss_wrapper_ref_debug( w, file, line );
#else
      wss_wrapper_ref( w );
#endif /* DEBUG_REFS */
   WSS_MUTEX_EXIT( wss_wrapper_list );

   return w;
}


/*  wss_wrapper_call_activation */
/*! \brief Run activation routines on all wrappers associated with this
 *         process.
 *
 *  Step through the list of wrappers associated with this process
 *  (generated from the activation criteria) and run the activate()
 *  function for each one.
 *
 *  Notes:
 *    Short-circuits -- won't run any more activation functions after one fails
 *    TBD: decide if this is what is desired
 *
 *  \param wrp Pointer to the wrapper process structure for this process
 *
 *  \return 0 if activate function calls were successful,
 *  \return error code on error
 *
 *  \sa wss_wrapper_call_deactivation, wss_wrapper_call_duplication,
 *      wss_wrapper_register
 */
int EXPORT
wss_wrapper_call_activation( wss_proc_t *wrp )
{
   int                     error;
   wss_wrapper_instance_t *wi;
   int (*activate)( wss_proc_t *wrproc, wss_wrapper_data_t *data );

recall_act:
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
        wi;
        wi = wi->wi_next )
   {
      /* Skip if instance is already active */
      if( wi->wi_active ) continue;

      /* Reference the wrapper instance */
      wss_wrapper_instance_ref( wi );

      activate = wi->wi_wrapper->w_activate;
      if( activate != NULL )
      {
         wss_wrapper_data_t wrd = { 0, 0, 0 };
         wrd.wiv = wi->wi_data;

         /* Unlock the list */
         WSS_MUTEX_EXIT( wrp->wrp_wi_list );

         /* Call activate function on wrapper */
         error = (*activate)( wrp, &wrd );
         wi->wi_data = wrd.wiv;

         if( error != WR_SUCCESS )
         {
            /* Release our reference to the wrapper instance */
            wss_wrapper_instance_rele( wi );

            WSS_MUTEX_EXIT( wrp->wrp_wi_list );
            return error;
         }
         else
         {
            /* Flag instance as active */
            wi->wi_active = 1;

            wsi_wrapper_activate( wrp, wi );

            /* Release our reference to the wrapper instance */
            wss_wrapper_instance_rele( wi );

            /* Start from the beginning again, in case something
               was added and/or removed during the time we did not
               have the lock acquired */
            goto recall_act;
         }
      }

      /* Flag instance as active */
      wi->wi_active = 1;

      /* Release our reference to the wrapper instance */
      wss_wrapper_instance_rele( wi );
   }

   /* Unlock the list */
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   return 0;
}


/*  wss_wrapper_call_deactivation */
/*! \brief Run deactivation routines on all wrappers associated with this
 *         process.
 *
 *  Like the activation routine, steps through all the wrappers
 *  associated with this process and calls the deactivate() function.
 *
 *  \param wrp Pointer to the wrapper process structure for this process
 *
 *  \return 0 if deactivate function calls were successful,
 *  \return error code on error
 *
 *  \sa wss_wrapper_call_activation, wss_wrapper_call_duplication,
 *      wss_wrapper_register
 */
int EXPORT
wss_wrapper_call_deactivation( wss_proc_t *wrp )
{
   int                     error;
   wss_wrapper_instance_t *wi;
   int (*deactivate)( wss_proc_t *wrproc, wss_wrapper_data_t *data );

recall_deact:
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
        wi;
        wi = wi->wi_next )
   {
      /* Skip if instance is not active */
      if( !wi->wi_active ) continue;

      /* Reference the wrapper instance */
      wss_wrapper_instance_ref( wi );

      deactivate = wi->wi_wrapper->w_deactivate;
      if( deactivate != NULL )
      {
         wss_wrapper_data_t wrd;
         wrd.wiv = wi->wi_data;

         /* Unlock the list */
         WSS_MUTEX_EXIT( wrp->wrp_wi_list );

         /* Call deactivate function on wrapper */
         error = (*deactivate)( wrp, &wrd );
         wi->wi_data = wrd.wiv;

         if( error != WR_SUCCESS )
         {
            /* Release our reference to the wrapper instance */
            wss_wrapper_instance_rele( wi );

            WSS_MUTEX_EXIT( wrp->wrp_wi_list );
            return error;
         }
         else
         {
            /* Flag instance as inactive */
            wi->wi_active = 0;

            wsi_wrapper_deactivate( wrp, wi );

            /* Release our reference to the wrapper instance */
            wss_wrapper_instance_rele( wi );

            /* Start from the beginning again, in case something
               was added and/or removed during the time we did not
               have the lock acquired */
            goto recall_deact;
         }
      }

      /* Flag instance as inactive */
      wi->wi_active = 0;

      /* Release our reference to the wrapper instance */
      wss_wrapper_instance_rele( wi );
   }

   /* Unlock the list */
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   return 0;
}


/*  wss_wrapper_call_duplication */
/*! \brief Run duplication routines on all wrappers associated with this
 *         process.
 *
 *  Like the activation routine, steps through all the wrappers
 *  associated with this process and calls the duplicate() function.
 *
 *  \param wrp Pointer to the wrapper process structure for this process
 *
 *  \return 0 if deactivate function calls were successful,
 *  \return error code on error
 *
 *  \sa wss_wrapper_call_activation, wss_wrapper_call_deactivation,
 *      wss_wrapper_register
 */
int EXPORT
wss_wrapper_call_duplication( wss_proc_t *wrp )
{
   int                     error;
   wss_wrapper_instance_t *wi;
   int (*duplicate)( wss_proc_t *wrproc, wss_wrapper_data_t *data );

recall_dup:
   WSS_MUTEX_ENTER( wrp->wrp_wi_list );
   for( wi = (wss_wrapper_instance_t *)(wrp->wrp_wi_list);
        wi;
        wi = wi->wi_next )
   {
      /* Skip if instance is already active */
      if( wi->wi_active ) continue;

      /* Reference the wrapper instance */
      wss_wrapper_instance_ref( wi );

      duplicate = wi->wi_wrapper->w_duplicate;
      if( duplicate != NULL )
      {
         wss_wrapper_data_t wrd;
         wrd.wiv = wi->wi_data;

         /* Unlock the list */
         WSS_MUTEX_EXIT( wrp->wrp_wi_list );

         /* Call duplicate function on wrapper */
         error = (*duplicate)( wrp, &wrd );
         wi->wi_data = wrd.wiv;

         if( error != WR_SUCCESS )
         {
            /* Release our reference to the wrapper instance */
            wss_wrapper_instance_rele( wi );

            WSS_MUTEX_EXIT( wrp->wrp_wi_list );
            return error;
         }
         else
         {
            /* Flag instance as active */
            wi->wi_active = 1;

            wsi_wrapper_duplicate( wrp, wi );

            /* Release our reference to the wrapper instance */
            wss_wrapper_instance_rele( wi );

            /* Start from the beginning again, in case something
               was added and/or removed during the time we did not
               have the lock acquired */
            goto recall_dup;
         }
      }

      /* Flag instance as active */
      wi->wi_active = 1;

      /* Release our reference to the wrapper instance */
      wss_wrapper_instance_rele( wi );
   }

   /* Unlock the list */
   WSS_MUTEX_EXIT( wrp->wrp_wi_list );

   return 0;
}


/*  wss_wrapper_add_new */
/*! \brief Adds the named wrapper to the active list for current process
 *
 *  Calls the activation function
 *
 *  This function can be called outside the context of a fork/exec
 *
 *  \param name Pointer to string containing the name of the wrapper module
 *
 *  \return 0 on success,
 *  \return error code on error
 */
int EXPORT
wss_wrapper_add_new( pd_char_t *name )
{
   int                     error = 1;
   wss_proc_t             *wrp;
   wss_wrapper_instance_t *wi;
   wss_wrapper_data_t      wrd = { 0, 0, 0 };

   wrp = wss_pfind( wsi_proc_get_current_pid() );

   assert( wrp != NULL );

   error = wss_proc_add_wrapper( wrp, name, &wi );
   if( error )
   {
      wss_proc_rele( wrp );
      return error;
   }

   /*
    * Assumption: Since "wss_proc_add_wrapper()" didn't fail, it added the
    * wrapper.  We had better find it...
    */
   assert( wi != NULL );

   /*
    * Call the activation function
    * Hope " curproc" is correct
    */
   wrd.wiv = wi->wi_data;
   error = (*(wi->wi_wrapper->w_activate))(wrp, &wrd);
   wi->wi_data = wrd.wiv;
   wss_proc_rele( wrp );
   
   if( error != WR_SUCCESS )
   {
      wss_wrapper_instance_rele( wi );
      return error;
   }

   wi->wi_active = 1;
   wss_wrapper_instance_rele( wi );

   return 0;
}


/*  wss_wrapper_instance_free */
/*! \brief Frees the memory associated with a wrapper instance
 *
 *  \param wi Pointer to the wrapper instance structure
 */
void EXPORT
wss_wrapper_instance_free( wss_wrapper_instance_t *wi )
{
   assert( wi != NULL );
   assert( WSS_REF_COUNT(wi->wi_refcount) == 1 );

   WSS_REF_DEC(wi->wi_refcount);

#ifdef DEBUG_WRAPPER
   wsi_printf( PD_MAKESTR("wss_wrapper_instance_free: releasing memory for 0x%x\n"),
               (unsigned int)wi );
#endif /* DEBUG_WRAPPER */

   wss_ref_destroy( &(wi->wi_refcount) );

   wss_wrapper_rele( wi->wi_wrapper );
   if( wi->wi_data )
      wss_free( (void *)wi->wi_data );
   wss_free( (void *)wi );
}

